(* Content-type: application/mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 6.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       145,          7]
NotebookDataLength[   1776417,      46938]
NotebookOptionsPosition[   1774196,      46866]
NotebookOutlinePosition[   1774852,      46892]
CellTagsIndexPosition[   1774768,      46887]
WindowFrame->Normal
ContainsDynamic->False*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["Arg Color Plot ", "Subsection",
 CellTags->"VQM`ArgColorPlot`"],

Cell["\<\
Plot the absolute value Abs[f[x]] of a complex-valued function f depending on \
a real variable x and fill the space between the plotted function and the \
x-axis with a color corresponding to the argument Arg[f[x]]. The saturation \
and brightness of the colors can be set using the options QSaturation and \
QBrightness. \
\>", "Text",
 CellTags->"VQM`ArgColorPlot`"],

Cell[BoxData[
 FormBox[
  StyleBox[
   FrameBox[GridBox[{
      {
       StyleBox[
        RowBox[{"QArgColorPlot", "[", 
         RowBox[{
          RowBox[{"f", "[", "x", "]"}], ",", 
          RowBox[{"{", 
           RowBox[{"x", ",", "x0", ",", "x1"}], "}"}], ",", "opts"}], "]"}],
        FontFamily->"Courier"], 
       StyleBox[
        RowBox[{" ", 
         RowBox[{
          RowBox[{
           RowBox[{
            RowBox[{
            "is", " ", "used", " ", "like", " ", "the", " ", "usual", " ", 
             "Plot", " ", 
             RowBox[{"command", ".", " ", "It"}], " ", "gives", " ", "a", " ",
              "two"}], "-", 
            RowBox[{
            "dimensional", " ", "plot", " ", "of", " ", "a", " ", "complex"}],
             "-", 
            RowBox[{
            "valued", " ", "function", " ", "f", " ", "of", " ", "a", " ", 
             "single", " ", "real", " ", "variable", " ", "x", " ", "in", " ",
              "the", " ", "range", " ", 
             RowBox[{
              RowBox[{"{", 
               RowBox[{"x0", ",", "x1"}], "}"}], ".", " ", "The"}], " ", 
             "plot", " ", "shows", " ", "the", " ", "curve", " ", 
             RowBox[{"Abs", "[", "f", "]"}], " ", "with", " ", "area", " ", 
             "between", " ", "the", " ", "curve", " ", "and", " ", "the", " ",
              "x"}], "-", 
            RowBox[{"axis", " ", "colored", " ", "by", " ", 
             RowBox[{
              RowBox[{"Hue", "[", 
               RowBox[{
                RowBox[{"Arg", "[", 
                 RowBox[{"f", "[", "x", "]"}], "]"}], "/", 
                RowBox[{"(", 
                 RowBox[{"2", " ", "Pi"}], ")"}]}], "]"}], ".", " ", "The"}], 
             " ", "default", " ", "options", " ", "of", " ", "Plot", " ", 
             "are", " ", "changed", " ", "to", " ", "Axes"}]}], "->", 
           RowBox[{"{", 
            RowBox[{"True", ",", "None"}], "}"}]}], ",", " ", 
          RowBox[{"Fame", "->", 
           RowBox[{"True", ".", " ", 
            RowBox[{"Package", ":", " ", "VQM`ArgColorPlot`"}]}]}]}]}],
        FontFamily->"Times"]},
      {
       StyleBox[
        RowBox[{"QListArgColorPlot", "[", 
         RowBox[{"f", ",", 
          RowBox[{"{", 
           RowBox[{"x", ",", "x0", ",", "x1"}], "}"}], ",", "opts"}], "]"}],
        FontFamily->"Courier"], 
       StyleBox[
        RowBox[{" ", 
         RowBox[{
          RowBox[{"plots", " ", "a", " ", 
           RowBox[{"Abs", "[", "f", "]"}]}], ",", " ", 
          RowBox[{
           RowBox[{
           "where", " ", "f", " ", "is", " ", "a", " ", "list", " ", "of", 
            " ", "complex", " ", 
            RowBox[{"numbers", ".", " ", "The"}], "  ", "points", " ", "of", 
            " ", "the", " ", "list", " ", 
            RowBox[{"Abs", "[", "f", "]"}], " ", "are", " ", "joined", " ", 
            "by", " ", "a", " ", 
            RowBox[{"line", ".", " ", "The"}], " ", "area", " ", "between", 
            " ", "the", " ", "curve", " ", "and", " ", "the", " ", "x"}], "-",
            
           RowBox[{
           "axis", " ", "is", " ", "colored", " ", "at", " ", "each", " ", 
            "point", " ", "by", " ", 
            RowBox[{
             RowBox[{"Hue", "[", 
              RowBox[{
               RowBox[{"Arg", "[", "f", "]"}], "/", 
               RowBox[{"(", 
                RowBox[{"2", " ", "Pi"}], ")"}]}], "]"}], ".", " ", 
             RowBox[{"Package", ":", " ", "VQM`ArgColorPlot`"}]}]}]}]}]}],
        FontFamily->"Times"]},
      {
       StyleBox[
        RowBox[{"QCombinedPlot", "[", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{
            RowBox[{"f", "[", "x", "]"}], ",", 
            RowBox[{"g", "[", "x", "]"}]}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{"x", ",", "x0", ",", "x1"}], "}"}], ",", "opts"}], "]"}],
        FontFamily->"Courier"], 
       StyleBox[
        RowBox[{" ", 
         RowBox[{
         "works", " ", "like", " ", "QArgColorPlot", " ", "with", " ", 
          "respect", " ", "to", " ", 
          RowBox[{"f", ".", " ", "The"}], " ", "curve", " ", "g", " ", "is", 
          " ", "drawn", " ", "in", " ", "front", " ", "of", " ", "the", " ", 
          "QArgColorPlot", " ", "of", " ", 
          RowBox[{"f", ".", " ", 
           RowBox[{"Package", ":", " ", "VQM`ArgColorPlot`"}]}]}]}],
        FontFamily->"Times"]},
      {
       StyleBox[
        RowBox[{"QListCombinedPlot", "[", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{"list", ",", 
            RowBox[{"f", "[", "x", "]"}]}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{"x", ",", "x0", ",", "x1"}], "}"}], ",", "opts"}], "]"}],
        FontFamily->"Courier"], 
       StyleBox[
        RowBox[{" ", 
         RowBox[{
         "works", " ", "like", " ", "QListArgColorPlot", " ", "with", " ", 
          "respect", " ", "to", " ", 
          RowBox[{"list", ".", " ", "It"}], " ", "is", " ", "assumed", " ", 
          "that", " ", "list", " ", "represents", " ", "the", " ", 
          "discretized", " ", "values", " ", "of", " ", "a", " ", "function", 
          " ", "defined", " ", "on", " ", "the", " ", 
          RowBox[{
           RowBox[{"interval", " ", "[", 
            RowBox[{"x0", ",", "x1"}], "]"}], ".", " ", "The"}], " ", "color",
           " ", "list", " ", "plot", " ", "is", " ", "then", " ", "combined", 
          " ", "with", " ", "an", " ", "ordinary", " ", "plot", " ", "of", 
          " ", "f", " ", "on", " ", "the", " ", "same", " ", "scale", " ", 
          "and", " ", "with", " ", "the", " ", "Ticks", " ", "automatically", 
          " ", 
          RowBox[{"adjusted", ".", " ", 
           RowBox[{"Package", ":", " ", "VQM`ArgColorPlot`"}]}]}]}],
        FontFamily->"Times"]},
      {
       StyleBox[
        RowBox[{"QSpinorPlot", "[", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{"func1", ",", "func2"}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{"x", ",", "x0", ",", "x1"}], "}"}], ",", "opts"}], "]"}],
        FontFamily->"Courier"], 
       StyleBox[
        RowBox[{" ", 
         RowBox[{
          RowBox[{
           RowBox[{
            RowBox[{
            "provides", " ", "a", " ", "method", " ", "to", " ", "visualize", 
             " ", 
             RowBox[{"C", "^", "2"}]}], "-", 
            RowBox[{"valued", " ", "functions", " ", 
             RowBox[{
              RowBox[{"(", 
               RowBox[{
                RowBox[{"for", " ", "example"}], ",", " ", 
                RowBox[{
                "spinor", " ", "wavefunctions", " ", "in", " ", "quantum", 
                 " ", "mechanics"}]}], ")"}], ".", " ", "The"}], " ", 
             "QSpinorPlot", " ", "combines", " ", "a", " ", "QArgColorPlot", 
             " ", "of", " ", "func1", " ", "with", " ", "a", " ", 
             "QArgColorPlot", " ", "of", " ", "func2", " ", 
             RowBox[{"(", 
              RowBox[{
               RowBox[{"upside", " ", "down"}], ",", " ", 
               RowBox[{"with", " ", "less", " ", "saturation"}]}], ")"}], " ",
              "Both", " ", "curves", " ", "are", " ", "plotted", " ", "with", 
             " ", "the", " ", "option", " ", "QSquared"}]}], "->", 
           RowBox[{"True", " ", 
            RowBox[{
             RowBox[{"(", 
              RowBox[{
               RowBox[{"that", " ", "is"}], ",", " ", 
               RowBox[{
               "a", " ", "plot", " ", "of", " ", "the", " ", "curve", " ", 
                RowBox[{
                 RowBox[{"Abs", "[", "func", "]"}], "^", "2"}], " ", "is", 
                " ", "filled", " ", "with", " ", "a", " ", "color", " ", 
                "describing", " ", "the", " ", "phase"}]}], ")"}], ".", " ", 
             "In"}], " ", "the", " ", "background"}]}], ",", " ", 
          RowBox[{
           RowBox[{"a", " ", "filled", " ", "plot", " ", "of", " ", 
            RowBox[{
             RowBox[{"Abs", "[", "func1", "]"}], "^", "2"}]}], " ", "+", " ", 
           
           RowBox[{
            RowBox[{
             RowBox[{"Abs", "[", "func2", "]"}], "^", "2"}], " ", "displays", 
            " ", "the", " ", "corresponding", " ", 
            RowBox[{"density", ".", " ", 
             RowBox[{"Package", ":", " ", "VQM`ArgColorPlot`"}]}]}]}]}]}],
        FontFamily->"Times"]},
      {
       StyleBox[
        RowBox[{"QListSpinorPlot", "[", 
         RowBox[{"list", ",", "opts"}], "]"}],
        FontFamily->"Courier"], 
       StyleBox[
        RowBox[{" ", 
         RowBox[{
          RowBox[{
           RowBox[{"visualizes", " ", "a", " ", "spinor"}], "-", 
           RowBox[{"valued", " ", "list", " ", "of", " ", "complex", " ", 
            RowBox[{"numbers", ".", " ", "Each"}], " ", "element", " ", "of", 
            " ", "list", " ", "is", " ", "a", " ", 
            RowBox[{"C", "^", 
             RowBox[{"-", "vector"}]}]}]}], ",", " ", 
          RowBox[{"that", " ", "is"}], ",", " ", 
          RowBox[{"list", " ", "=", " ", 
           RowBox[{
            RowBox[{"{", 
             RowBox[{
              RowBox[{"{", 
               RowBox[{"z11", ",", "z12"}], "}"}], ",", 
              RowBox[{"{", 
               RowBox[{"z21", ",", "z22"}], "}"}], ",", "..."}], "}"}], ".", 
            " ", "Alternatively"}]}], ",", " ", 
          RowBox[{"list", " ", "=", " ", 
           RowBox[{
            RowBox[{"{", 
             RowBox[{"list1", ",", "list2"}], "}"}], " ", "with", " ", "two", 
            " ", "lists", " ", "of", " ", "complex", " ", "numbers"}]}], ",", 
          " ", 
          RowBox[{
           RowBox[{
           "list1", " ", "giving", " ", "the", " ", "upper", " ", "component",
             " ", "of", " ", "the", " ", "spinor"}], "-", 
           RowBox[{"valued", " ", "wave", " ", "function"}]}], ",", " ", 
          RowBox[{
          "and", " ", "list2", " ", "giving", " ", "the", " ", "lower", " ", 
           RowBox[{"component", ".", " ", "The"}], " ", "lower", " ", 
           "component", " ", "is", " ", "plotted", " ", "upside", " ", "down",
            " ", "with", " ", "less", " ", 
           RowBox[{"saturation", ".", " ", "See"}], " ", "also", " ", "the", 
           " ", "description", " ", "of", " ", 
           RowBox[{"QSpinorPlot", ".", " ", 
            RowBox[{"Package", ":", " ", "VQM`ArgColorPlot`"}]}]}]}]}],
        FontFamily->"Times"]},
      {
       StyleBox[
        RowBox[{"QSpinorCombinedPlot", "[", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{"func1", ",", "func2"}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{"x", ",", "x0", ",", "x1"}], "}"}], ",", "opts"}], "]"}],
        FontFamily->"Courier"], 
       StyleBox[
        RowBox[{" ", 
         RowBox[{
          RowBox[{
          "combines", " ", "a", " ", "QSpinorPlot", " ", "of", " ", "func1", 
           " ", "with", " ", "an", " ", "ordinary", " ", "Plot", " ", "of", 
           " ", "a", " ", "real"}], "-", 
          RowBox[{"valued", " ", "function", " ", 
           RowBox[{"func2", ".", " ", "See"}], " ", "the", " ", "description",
            " ", "of", " ", "QCombinedPlot", " ", "and", " ", "of", " ", 
           RowBox[{"QSpinorPlot", ".", " ", 
            RowBox[{"Package", ":", " ", "VQM`ArgColorPlot`"}]}]}]}]}],
        FontFamily->"Times"]},
      {
       StyleBox[
        RowBox[{"QListSpinorCombinedPlot", "[", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{"list", ",", 
            RowBox[{"f", "[", "x", "]"}]}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{"x", ",", "x0", ",", "x1"}], "}"}], ",", "opts"}], "]"}],
        FontFamily->"Courier"], 
       StyleBox[
        RowBox[{" ", 
         RowBox[{
          RowBox[{
          "combines", " ", "a", " ", "QListSpinorPlot", " ", "of", " ", 
           "list1", " ", "with", " ", "an", " ", "ordinary", " ", "Plot", "  ",
            "of", " ", "a", " ", "real"}], "-", 
          RowBox[{"valued", " ", "function", " ", 
           RowBox[{"f", ".", " ", "See"}], " ", "the", " ", "description", 
           " ", "of", " ", "QListCombinedPlot", " ", "and", " ", "of", " ", 
           RowBox[{"QListSpinorPlot", ".", " ", 
            RowBox[{"Package", ":", " ", "VQM`ArgColorPlot`"}]}]}]}]}],
        FontFamily->"Times"]},
      {
       StyleBox[
        RowBox[{"QNiceTicks", "[", 
         RowBox[{"xmin", ",", "xmax", ",", "dx"}], "]"}],
        FontFamily->"Courier"], 
       StyleBox[
        RowBox[{" ", 
         RowBox[{
          RowBox[{
          "provides", " ", "a", " ", "list", " ", "of", " ", "nice", " ", 
           "positions", " ", "for", " ", "use", " ", "in", " ", "the", " ", 
           "Ticks", " ", "or", " ", "FrameTicks", " ", "option", " ", "in", 
           " ", "a", " ", "ListPlot"}], ",", " ", 
          RowBox[{
          "where", " ", "it", " ", "is", " ", "assumed", " ", "that", " ", 
           "the", " ", "list", " ", "of", " ", "values", " ", "ranges", " ", 
           "between", " ", "xmin", " ", "and", " ", "xmax", " ", "in", " ", 
           "steps", " ", 
           RowBox[{"dx", ".", " ", 
            RowBox[{"Package", ":", " ", "VQM`ArgColorPlot`"}]}]}]}]}],
        FontFamily->"Times"]}
     },
     GridBoxAlignment->{
      "Columns" -> {Right, {Left}}, "ColumnsIndexed" -> {}, 
       "Rows" -> {{Baseline}}, "RowsIndexed" -> {}},
     GridBoxDividers->{
      "Columns" -> {{False}}, "ColumnsIndexed" -> {}, "Rows" -> {{False}}, 
       "RowsIndexed" -> {}}]], "2ColumnBox"], TraditionalForm]], "Text",
 CellTags->"VQM`ArgColorPlot`"],

Cell["This loads the package", "MathCaption"],

Cell["Needs[\"VQM`ArgColorPlot`\"]", "Input",
 CellChangeTimes->{{3.39185763090625*^9, 3.391857634375*^9}},
 CellLabel->"In[30]:="],

Cell["This is an example", "MathCaption"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"msgonindet", "=", 
   RowBox[{
    RowBox[{"Head", "[", 
     RowBox[{"General", "::", "\"\<indet\>\""}], "]"}], "=!=", "$Off"}]}], 
  ";"}], "\n", 
 RowBox[{
  RowBox[{"Off", "[", 
   RowBox[{"Arg", "::", "\"\<indet\>\""}], "]"}], ";"}]}], "Input",
 CellLabel->"In[31]:="],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"QArgColorPlot", "[", 
  RowBox[{
   SuperscriptBox["\[ExponentialE]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "\[ImaginaryI]"}], " ", "6", " ", "x"}], "-", 
     FractionBox[
      SuperscriptBox["x", "2"], "2"]}]], ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", 
     RowBox[{"-", "4"}], ",", "4"}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{3.3918574275*^9},
 CellLabel->"In[34]:="],

Cell[BoxData[
 GraphicsBox[GraphicsComplexBox[CompressedData["
1:eJx13GdcjQ/cx/H2nnaEbBLSRPK7IiHaSkWp7JUoFYokSf1VIkJoGKVUFKV1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   "], {
    {GrayLevel[0.5], EdgeForm[None], 
     GraphicsGroupBox[PolygonBox[CompressedData["
1:eJwtmXncldPXh/e+z7lPgzRLkjSpVFKpNBoyNdKgUqk0kaRCEpKQSmmQIkmS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       "]]]}, {{}, {}, {
      {Hue[0.8197184782986047, 1, 1], LineBox[{1, 2}]}, 
      {Hue[0.817375481892275, 1, 1], LineBox[{2, 3}]}, 
      {Hue[0.8150324854859454, 1, 1], LineBox[{3, 4}]}, 
      {Hue[0.8103464926732862, 1, 1], LineBox[{4, 5}]}, 
      {Hue[0.8009745070479671, 1, 1], LineBox[{5, 6}]}, 
      {Hue[0.7822305357973296, 1, 1], LineBox[{6, 7}]}, 
      {Hue[0.779887539391, 1, 1], LineBox[{7, 8}]}, 
      {Hue[0.7775445429846704, 1, 1], LineBox[{8, 9}]}, 
      {Hue[0.7728585501720105, 1, 1], LineBox[{9, 10}]}, 
      {Hue[0.763486564546692, 1, 1], LineBox[{10, 11}]}, 
      {Hue[0.7447425932960545, 1, 1], LineBox[{11, 12}]}, 
      {Hue[0.7423995968897242, 1, 1], LineBox[{12, 13}]}, 
      {Hue[0.7400566004833946, 1, 1], LineBox[{13, 14}]}, 
      {Hue[0.7353706076707348, 1, 1], LineBox[{14, 15}]}, 
      {Hue[0.7259986220454163, 1, 1], LineBox[{15, 16}]}, 
      {Hue[0.7072546507947788, 1, 1], LineBox[{16, 17}]}, 
      {Hue[0.6697667082935036, 1, 1], LineBox[{17, 18}]}, 
      {Hue[0.6672266317029389, 1, 1], LineBox[{18, 19}]}, 
      {Hue[0.6646865551123748, 1, 1], LineBox[{19, 20}]}, 
      {Hue[0.6596064019312451, 1, 1], LineBox[{20, 21}]}, 
      {Hue[0.6494460955689871, 1, 1], LineBox[{21, 22}]}, 
      {Hue[0.6291254828444701, 1, 1], LineBox[{22, 23}]}, 
      {Hue[0.626585406253906, 1, 1], LineBox[{23, 24}]}, 
      {Hue[0.6240453296633414, 1, 1], LineBox[{24, 25}]}, 
      {Hue[0.6189651764822122, 1, 1], LineBox[{25, 26}]}, 
      {Hue[0.6088048701199542, 1, 1], LineBox[{26, 27}]}, 
      {Hue[0.5884842573954372, 1, 1], LineBox[{27, 28}]}, 
      {Hue[0.5859441808048731, 1, 1], LineBox[{28, 29}]}, 
      {Hue[0.5834041042143084, 1, 1], LineBox[{29, 30}]}, 
      {Hue[0.5783239510331792, 1, 1], LineBox[{30, 31}]}, 
      {Hue[0.5681636446709212, 1, 1], LineBox[{31, 32}]}, 
      {Hue[0.5478430319464042, 1, 1], LineBox[{32, 33}]}, 
      {Hue[0.5072018064973707, 1, 1], LineBox[{33, 34}]}, 
      {Hue[0.5048300564796699, 1, 1], LineBox[{34, 35}]}, 
      {Hue[0.5024583064619688, 1, 1], LineBox[{35, 36}]}, 
      {Hue[0.4977148064265675, 1, 1], LineBox[{36, 37}]}, 
      {Hue[0.48822780635576496`, 1, 1], LineBox[{37, 38}]}, 
      {Hue[0.4692538062141587, 1, 1], LineBox[{38, 39}]}, 
      {Hue[0.466882056196458, 1, 1], LineBox[{39, 40}]}, 
      {Hue[0.4645103061787574, 1, 1], LineBox[{40, 41}]}, 
      {Hue[0.45976680614335613`, 1, 1], LineBox[{41, 42}]}, 
      {Hue[0.45027980607255297`, 1, 1], LineBox[{42, 43}]}, 
      {Hue[0.43130580593094725`, 1, 1], LineBox[{43, 44}]}, 
      {Hue[0.4289340559132466, 1, 1], LineBox[{44, 45}]}, 
      {Hue[0.426562305895546, 1, 1], LineBox[{45, 46}]}, 
      {Hue[0.42181880586014414`, 1, 1], LineBox[{46, 47}]}, 
      {Hue[0.4123318057893416, 1, 1], LineBox[{47, 48}]}, 
      {Hue[0.39335780564773587`, 1, 1], LineBox[{48, 49}]}, 
      {Hue[0.3554098053645245, 1, 1], LineBox[{49, 50}]}, 
      {Hue[0.3530845796673215, 1, 1], LineBox[{50, 51}]}, 
      {Hue[0.35075935397011854`, 1, 1], LineBox[{51, 52}]}, 
      {Hue[0.3461089025757114, 1, 1], LineBox[{52, 53}]}, 
      {Hue[0.33680799978689946`, 1, 1], LineBox[{53, 54}]}, 
      {Hue[0.3182061942092745, 1, 1], LineBox[{54, 55}]}, 
      {Hue[0.3158809685120721, 1, 1], LineBox[{55, 56}]}, 
      {Hue[0.31355574281486853`, 1, 1], LineBox[{56, 57}]}, 
      {Hue[0.308905291420462, 1, 1], LineBox[{57, 58}]}, 
      {Hue[0.2996043886316495, 1, 1], LineBox[{58, 59}]}, 
      {Hue[0.2810025830540251, 1, 1], LineBox[{59, 60}]}, 
      {Hue[0.2786773573568221, 1, 1], LineBox[{60, 61}]}, 
      {Hue[0.2763521316596185, 1, 1], LineBox[{61, 62}]}, 
      {Hue[0.2717016802652126, 1, 1], LineBox[{62, 63}]}, 
      {Hue[0.26240077747640006`, 1, 1], LineBox[{63, 64}]}, 
      {Hue[0.24379897189877506`, 1, 1], LineBox[{64, 65}]}, 
      {Hue[0.2065953607435256, 1, 1], LineBox[{65, 66}]}, 
      {Hue[0.20407305486208763`, 1, 1], LineBox[{66, 67}]}, 
      {Hue[0.20155074898064965`, 1, 1], LineBox[{67, 68}]}, 
      {Hue[0.19650613721777366`, 1, 1], LineBox[{68, 69}]}, 
      {Hue[0.1864169136920217, 1, 1], LineBox[{69, 70}]}, 
      {Hue[0.16623846664051775`, 1, 1], LineBox[{70, 71}]}, 
      {Hue[0.16371616075907977`, 1, 1], LineBox[{71, 72}]}, 
      {Hue[0.1611938548776418, 1, 1], LineBox[{72, 73}]}, 
      {Hue[0.1561492431147658, 1, 1], LineBox[{73, 74}]}, 
      {Hue[0.14606001958901382`, 1, 1], LineBox[{74, 75}]}, 
      {Hue[0.12588157253750992`, 1, 1], LineBox[{75, 76}]}, 
      {Hue[0.12335926665607193`, 1, 1], LineBox[{76, 77}]}, 
      {Hue[0.12083696077463393`, 1, 1], LineBox[{77, 78}]}, 
      {Hue[0.11579234901175797`, 1, 1], LineBox[{78, 79}]}, 
      {Hue[0.105703125486006, 1, 1], LineBox[{79, 80}]}, 
      {Hue[0.0855246784345021, 1, 1], LineBox[{80, 81}]}, 
      {Hue[0.04516778433149423, 1, 1], LineBox[{81, 82}]}, 
      {Hue[0.04281380502292024, 1, 1], LineBox[{82, 83}]}, 
      {Hue[0.04045982571434625, 1, 1], LineBox[{83, 84}]}, 
      {Hue[0.035751867097198235`, 1, 1], LineBox[{84, 85}]}, 
      {Hue[0.026335949862901102`, 1, 1], LineBox[{85, 86}]}, 
      {Hue[0.007504115394307963, 1, 1], LineBox[{86, 87}]}, 
      {Hue[0.9698404464571223, 1, 1], LineBox[{87, 88}]}, 
      {Hue[0.9674864671485477, 1, 1], LineBox[{88, 89}]}, 
      {Hue[0.9651324878399743, 1, 1], LineBox[{89, 90}]}, 
      {Hue[0.9604245292228257, 1, 1], LineBox[{90, 91}]}, 
      {Hue[0.9510086119885292, 1, 1], LineBox[{91, 92}]}, 
      {Hue[0.9321767775199366, 1, 1], LineBox[{92, 93}]}, 
      {Hue[0.8945131085827502, 1, 1], LineBox[{93, 94}]}, 
      {Hue[0.8919620490899413, 1, 1], LineBox[{94, 95}]}, 
      {Hue[0.8894109895971323, 1, 1], LineBox[{95, 96}]}, 
      {Hue[0.8843088706115143, 1, 1], LineBox[{96, 97}]}, 
      {Hue[0.8741046326402782, 1, 1], LineBox[{97, 98}]}, 
      {Hue[0.8536961566978062, 1, 1], LineBox[{98, 99}]}, 
      {Hue[0.8128792048128621, 1, 1], LineBox[{99, 100}]}, 
      {Hue[0.810328145320053, 1, 1], LineBox[{100, 101}]}, 
      {Hue[0.8077770858272441, 1, 1], LineBox[{101, 102}]}, 
      {Hue[0.802674966841626, 1, 1], LineBox[{102, 103}]}, 
      {Hue[0.79247072887039, 1, 1], LineBox[{103, 104}]}, 
      {Hue[0.772062252927918, 1, 1], LineBox[{104, 105}]}, 
      {Hue[0.7312453010429738, 1, 1], LineBox[{105, 106}]}, 
      {Hue[0.7287407658706625, 1, 1], LineBox[{106, 107}]}, 
      {Hue[0.7262362306983512, 1, 1], LineBox[{107, 108}]}, 
      {Hue[0.7212271603537285, 1, 1], LineBox[{108, 109}]}, 
      {Hue[0.711209019664483, 1, 1], LineBox[{109, 110}]}, 
      {Hue[0.6911727382859922, 1, 1], LineBox[{110, 111}]}, 
      {Hue[0.6511001755290096, 1, 1], LineBox[{111, 112}]}, 
      {Hue[0.6485956403566981, 1, 1], LineBox[{112, 113}]}, 
      {Hue[0.646091105184387, 1, 1], LineBox[{113, 114}]}, 
      {Hue[0.6410820348397641, 1, 1], LineBox[{114, 115}]}, 
      {Hue[0.6310638941505182, 1, 1], LineBox[{115, 116}]}, 
      {Hue[0.6110276127720268, 1, 1], LineBox[{116, 117}]}, 
      {Hue[0.5709550500150447, 1, 1], LineBox[{117, 118}]}, 
      {Hue[0.5686188414155973, 1, 1], LineBox[{118, 119}]}, 
      {Hue[0.5662826328161494, 1, 1], LineBox[{119, 120}]}, 
      {Hue[0.5616102156172547, 1, 1], LineBox[{120, 121}]}, 
      {Hue[0.5522653812194647, 1, 1], LineBox[{121, 122}]}, 
      {Hue[0.5335757124238844, 1, 1], LineBox[{122, 123}]}, 
      {Hue[0.49619637483272355`, 1, 1], LineBox[{123, 124}]}, 
      {Hue[0.49386016623327594`, 1, 1], LineBox[{124, 125}]}, 
      {Hue[0.49152395763382856`, 1, 1], LineBox[{125, 126}]}, 
      {Hue[0.48685154043493384`, 1, 1], LineBox[{126, 127}]}, 
      {Hue[0.47750670603714357`, 1, 1], LineBox[{127, 128}]}, 
      {Hue[0.458817037241563, 1, 1], LineBox[{128, 129}]}, 
      {Hue[0.4214376996504027, 1, 1], LineBox[{129, 130}]}, 
      {Hue[0.4189044108667203, 1, 1], LineBox[{130, 131}]}, 
      {Hue[0.41637112208303795`, 1, 1], LineBox[{131, 132}]}, 
      {Hue[0.4113045445156732, 1, 1], LineBox[{132, 133}]}, 
      {Hue[0.4011713893809435, 1, 1], LineBox[{133, 134}]}, 
      {Hue[0.38090507911148397`, 1, 1], LineBox[{134, 135}]}, 
      {Hue[0.34037245857256526`, 1, 1], LineBox[{135, 136}]}, 
      {Hue[0.3378391697888829, 1, 1], LineBox[{136, 137}]}, 
      {Hue[0.33530588100520053`, 1, 1], LineBox[{137, 138}]}, 
      {Hue[0.3302393034378358, 1, 1], LineBox[{138, 139}]}, 
      {Hue[0.3201061483031061, 1, 1], LineBox[{139, 140}]}, 
      {Hue[0.2998398380336466, 1, 1], LineBox[{140, 141}]}, 
      {Hue[0.2593072174947282, 1, 1], LineBox[{141, 142}]}, 
      {Hue[0.2569422552839095, 1, 1], LineBox[{142, 143}]}, 
      {Hue[0.2545772930730911, 1, 1], LineBox[{143, 144}]}, 
      {Hue[0.24984736865145438`, 1, 1], LineBox[{144, 145}]}, 
      {Hue[0.24038751980818002`, 1, 1], LineBox[{145, 146}]}, 
      {Hue[0.22146782212163132`, 1, 1], LineBox[{146, 147}]}, 
      {Hue[0.18362842674853475`, 1, 1], LineBox[{147, 148}]}, 
      {Hue[0.18126346453771638`, 1, 1], LineBox[{148, 149}]}, 
      {Hue[0.17889850232689772`, 1, 1], LineBox[{149, 150}]}, 
      {Hue[0.17416857790526039`, 1, 1], LineBox[{150, 151}]}, 
      {Hue[0.16470872906198633`, 1, 1], LineBox[{151, 152}]}, 
      {Hue[0.14578903137543817`, 1, 1], LineBox[{152, 153}]}, 
      {Hue[0.10794963600234132`, 1, 1], LineBox[{153, 154}]}, 
      {Hue[0.10563119811202033`, 1, 1], LineBox[{154, 155}]}, 
      {Hue[0.10331276022169904`, 1, 1], LineBox[{155, 156}]}, 
      {Hue[0.09867588444105733, 1, 1], LineBox[{156, 157}]}, 
      {Hue[0.08940213287977387, 1, 1], LineBox[{157, 158}]}, 
      {Hue[0.07085462975720616, 1, 1], LineBox[{158, 159}]}, 
      {Hue[0.03375962351207098, 1, 1], LineBox[{159, 160}]}, 
      {Hue[0.03144118562175026, 1, 1], LineBox[{160, 161}]}, 
      {Hue[0.029122747731428977`, 1, 1], LineBox[{161, 162}]}, 
      {Hue[0.024485871950787247`, 1, 1], LineBox[{162, 163}]}, 
      {Hue[0.015212120389503814`, 1, 1], LineBox[{163, 164}]}, 
      {Hue[0.9966646172669361, 1, 1], LineBox[{164, 165}]}, 
      {Hue[0.9595696110218012, 1, 1], LineBox[{165, 166}]}, 
      {Hue[0.9570540929472452, 1, 1], LineBox[{166, 167}]}, 
      {Hue[0.9545385748726898, 1, 1], LineBox[{167, 168}]}, 
      {Hue[0.949507538723578, 1, 1], LineBox[{168, 169}]}, 
      {Hue[0.9394454664253545, 1, 1], LineBox[{169, 170}]}, 
      {Hue[0.9193213218289079, 1, 1], LineBox[{170, 171}]}, 
      {Hue[0.8790730326360152, 1, 1], LineBox[{171, 172}]}, 
      {Hue[0.7985764542502292, 1, 1], LineBox[{172, 173}]}, 
      {Hue[0.7962292627485371, 1, 1], LineBox[{173, 174}]}, 
      {Hue[0.7938820712468451, 1, 1], LineBox[{174, 175}]}, 
      {Hue[0.7891876882434611, 1, 1], LineBox[{175, 176}]}, 
      {Hue[0.7797989222366932, 1, 1], LineBox[{176, 177}]}, 
      {Hue[0.7610213902231577, 1, 1], LineBox[{177, 178}]}, 
      {Hue[0.7234663261960863, 1, 1], LineBox[{178, 179}]}, 
      {Hue[0.6483561981419439, 1, 1], LineBox[{179, 180}]}, 
      {Hue[0.6458119264560171, 1, 1], LineBox[{180, 181}]}, 
      {Hue[0.6432676547700904, 1, 1], LineBox[{181, 182}]}, 
      {Hue[0.6381791113982367, 1, 1], LineBox[{182, 183}]}, 
      {Hue[0.6280020246545293, 1, 1], LineBox[{183, 184}]}, 
      {Hue[0.6076478511671144, 1, 1], LineBox[{184, 185}]}, 
      {Hue[0.5669395041922853, 1, 1], LineBox[{185, 186}]}, 
      {Hue[0.4855228102426269, 1, 1], LineBox[{186, 187}]}, 
      {Hue[0.3256669788551567, 1, 1], LineBox[{187, 188}]}, 
      {Hue[0.17658404813097406`, 1, 1], LineBox[{188, 189}]}, 
      {Hue[0.014887985615758862`, 1, 1], LineBox[{189, 190}]}, 
      {Hue[0.8639648237638307, 1, 1], LineBox[{190, 191}]}, 
      {Hue[0.7004285301208703, 1, 1], LineBox[{191, 192}]}, 
      {Hue[0.6979197998531967, 1, 1], LineBox[{192, 193}]}, 
      {Hue[0.6954110695855231, 1, 1], LineBox[{193, 194}]}, 
      {Hue[0.6903936090501757, 1, 1], LineBox[{194, 195}]}, 
      {Hue[0.6803586879794812, 1, 1], LineBox[{195, 196}]}, 
      {Hue[0.660288845838092, 1, 1], LineBox[{196, 197}]}, 
      {Hue[0.6201491615553137, 1, 1], LineBox[{197, 198}]}, 
      {Hue[0.5398697929897572, 1, 1], LineBox[{198, 199}]}, 
      {Hue[0.5375293892949474, 1, 1], LineBox[{199, 200}]}, 
      {Hue[0.5351889856001377, 1, 1], LineBox[{200, 201}]}, 
      {Hue[0.5305081782105181, 1, 1], LineBox[{201, 202}]}, 
      {Hue[0.5211465634312791, 1, 1], LineBox[{202, 203}]}, 
      {Hue[0.5024233338728007, 1, 1], LineBox[{203, 204}]}, 
      {Hue[0.46497687475584426`, 1, 1], LineBox[{204, 205}]}, 
      {Hue[0.4626364710610345, 1, 1], LineBox[{205, 206}]}, 
      {Hue[0.46029606736622475`, 1, 1], LineBox[{206, 207}]}, 
      {Hue[0.45561525997660507`, 1, 1], LineBox[{207, 208}]}, 
      {Hue[0.446253645197366, 1, 1], LineBox[{208, 209}]}, 
      {Hue[0.4275304156388878, 1, 1], LineBox[{209, 210}]}, 
      {Hue[0.3900839565219313, 1, 1], LineBox[{210, 211}]}, 
      {Hue[0.38754647264288655`, 1, 1], LineBox[{211, 212}]}, 
      {Hue[0.38500898876384193`, 1, 1], LineBox[{212, 213}]}, 
      {Hue[0.37993402100575263`, 1, 1], LineBox[{213, 214}]}, 
      {Hue[0.3697840854895739, 1, 1], LineBox[{214, 215}]}, 
      {Hue[0.3494842144572167, 1, 1], LineBox[{215, 216}]}, 
      {Hue[0.34694673057817194`, 1, 1], LineBox[{216, 217}]}, 
      {Hue[0.34440924669912737`, 1, 1], LineBox[{217, 218}]}, 
      {Hue[0.339334278941038, 1, 1], LineBox[{218, 219}]}, 
      {Hue[0.32918434342485936`, 1, 1], LineBox[{219, 220}]}, 
      {Hue[0.3088844723925021, 1, 1], LineBox[{220, 221}]}, 
      {Hue[0.30634698851345743`, 1, 1], LineBox[{221, 222}]}, 
      {Hue[0.30380950463441275`, 1, 1], LineBox[{222, 223}]}, 
      {Hue[0.29873453687632345`, 1, 1], LineBox[{223, 224}]}, 
      {Hue[0.2885846013601448, 1, 1], LineBox[{224, 225}]}, 
      {Hue[0.2682847303277875, 1, 1], LineBox[{225, 226}]}, 
      {Hue[0.22768498826307296`, 1, 1], LineBox[{226, 227}]}, 
      {Hue[0.2253158309568922, 1, 1], LineBox[{227, 228}]}, 
      {Hue[0.22294667365071136`, 1, 1], LineBox[{228, 229}]}, 
      {Hue[0.21820835903834976`, 1, 1], LineBox[{229, 230}]}, 
      {Hue[0.20873172981362653`, 1, 1], LineBox[{230, 231}]}, 
      {Hue[0.18977847136418016`, 1, 1], LineBox[{231, 232}]}, 
      {Hue[0.18740931405799935`, 1, 1], LineBox[{232, 233}]}, 
      {Hue[0.18504015675181856`, 1, 1], LineBox[{233, 234}]}, 
      {Hue[0.18030184213945694`, 1, 1], LineBox[{234, 235}]}, 
      {Hue[0.17082521291473377`, 1, 1], LineBox[{235, 236}]}, 
      {Hue[0.1518719544652873, 1, 1], LineBox[{236, 237}]}, 
      {Hue[0.14950279715910653`, 1, 1], LineBox[{237, 238}]}, 
      {Hue[0.14713363985292574`, 1, 1], LineBox[{238, 239}]}, 
      {Hue[0.1423953252405641, 1, 1], LineBox[{239, 240}]}, 
      {Hue[0.1329186960158409, 1, 1], LineBox[{240, 241}]}, 
      {Hue[0.13054953870966016`, 1, 1], LineBox[{241, 242}]}, 
      {Hue[0.12818038140347934`, 1, 1], LineBox[{242, 243}]}, 
      {Hue[0.12344206679111772`, 1, 1], LineBox[{243, 244}]}, 
      {Hue[0.11396543756639452`, 1, 1], LineBox[{244, 245}]}, 
      {Hue[0.11159628026021372`, 1, 1], LineBox[{245, 246}]}, 
      {Hue[0.1092271229540329, 1, 1], LineBox[{246, 247}]}, 
      {Hue[0.10448880834167132`, 1, 1], LineBox[{247, 248}]}, 
      {Hue[0.09501217911694812, 1, 1], LineBox[{248, 249}]}, 
      {Hue[0.0926430218107673, 1, 1], LineBox[{249, 250}]}, 
      {Hue[0.09027386450458652, 1, 1], LineBox[{250, 251}]}, 
      {Hue[0.08553554989222491, 1, 1], LineBox[{251, 252}]}, 
      {Hue[0.07605892066750171, 1, 1], LineBox[{252, 253}]}, 
      {Hue[0.07373628768181852, 1, 1], LineBox[{253, 254}]}, 
      {Hue[0.07141365469613532, 1, 1], LineBox[{254, 255}]}, 
      {Hue[0.06676838872476895, 1, 1], LineBox[{255, 256}]}, 
      {Hue[0.06444575573908579, 1, 1], LineBox[{256, 257}]}, 
      {Hue[0.06212312275340258, 1, 1], LineBox[{257, 258}]}, 
      {Hue[0.05747785678203621, 1, 1], LineBox[{258, 259}]}, 
      {Hue[0.05515522379635303, 1, 1], LineBox[{259, 260}]}, 
      {Hue[0.052832590810669836`, 1, 1], LineBox[{260, 261}]}, 
      {Hue[0.04818732483930348, 1, 1], LineBox[{261, 262}]}, 
      {Hue[0.04586469185362028, 1, 1], LineBox[{262, 263}]}, 
      {Hue[0.043542058867937085`, 1, 1], LineBox[{263, 264}]}, 
      {Hue[0.03889679289657071, 1, 1], LineBox[{264, 265}]}, 
      {Hue[0.03657415991088753, 1, 1], LineBox[{265, 266}]}, 
      {Hue[0.03425152692520435, 1, 1], LineBox[{266, 267}]}, 
      {Hue[0.031928893939521155`, 1, 1], LineBox[{267, 268}]}, 
      {Hue[0.029606260953837966`, 1, 1], LineBox[{268, 269}]}, 
      {Hue[0.02728362796815478, 1, 1], LineBox[{269, 270}]}, 
      {Hue[0.024960994982471587`, 1, 1], LineBox[{270, 271}]}, 
      {Hue[0.0226383619967884, 1, 1], LineBox[{271, 272}]}, 
      {Hue[0.02031572901110522, 1, 1], LineBox[{272, 273}]}, 
      {Hue[0.017993096025422036`, 1, 1], LineBox[{273, 274}]}, 
      {Hue[0.015670463039738843`, 1, 1], LineBox[{274, 275}]}, 
      {Hue[0.013347830054055657`, 1, 1], LineBox[{275, 276}]}, 
      {Hue[0.011025197068372471`, 1, 1], LineBox[{276, 277}]}, 
      {Hue[0.008702564082689282, 1, 1], LineBox[{277, 278}]}, 
      {Hue[0.006379931097006096, 1, 1], LineBox[{278, 279}]}, 
      {Hue[0.004057298111322911, 1, 1], LineBox[{279, 280}]}, 
      {Hue[0.0017346651256397233`, 1, 1], LineBox[{280, 281}]}, 
      {Hue[0.9994120321399566, 1, 1], LineBox[{281, 282}]}, 
      {Hue[0.9970893991542733, 1, 1], LineBox[{282, 283}]}, 
      {Hue[0.9947667661685902, 1, 1], LineBox[{283, 284}]}, 
      {Hue[0.992444133182907, 1, 1], LineBox[{284, 285}]}, 
      {Hue[0.9901215001972238, 1, 1], LineBox[{285, 286}]}, 
      {Hue[0.9877988672115406, 1, 1], LineBox[{286, 287}]}, 
      {Hue[0.9854762342258574, 1, 1], LineBox[{287, 288}]}, 
      {Hue[0.9831536012401743, 1, 1], LineBox[{288, 289}]}, 
      {Hue[0.980830968254491, 1, 1], LineBox[{289, 290}]}, 
      {Hue[0.9785083352688079, 1, 1], LineBox[{290, 291}]}, 
      {Hue[0.9761857022831246, 1, 1], LineBox[{291, 292}]}, 
      {Hue[0.9738630692974415, 1, 1], LineBox[{292, 293}]}, 
      {Hue[0.9715404363117582, 1, 1], LineBox[{293, 294}]}, 
      {Hue[0.9692178033260751, 1, 1], LineBox[{294, 295}]}, 
      {Hue[0.9645725373547087, 1, 1], LineBox[{295, 296}]}, 
      {Hue[0.9622499043690256, 1, 1], LineBox[{296, 297}]}, 
      {Hue[0.9599272713833423, 1, 1], LineBox[{297, 298}]}, 
      {Hue[0.9552820054119759, 1, 1], LineBox[{298, 299}]}, 
      {Hue[0.9529593724262928, 1, 1], LineBox[{299, 300}]}, 
      {Hue[0.9506367394406096, 1, 1], LineBox[{300, 301}]}, 
      {Hue[0.9459914734692433, 1, 1], LineBox[{301, 302}]}, 
      {Hue[0.94366884048356, 1, 1], LineBox[{302, 303}]}, 
      {Hue[0.9413462074978769, 1, 1], LineBox[{303, 304}]}, 
      {Hue[0.9367009415265105, 1, 1], LineBox[{304, 305}]}, 
      {Hue[0.9274104095837777, 1, 1], LineBox[{305, 306}]}, 
      {Hue[0.9248906964138597, 1, 1], LineBox[{306, 307}]}, 
      {Hue[0.9223709832439416, 1, 1], LineBox[{307, 308}]}, 
      {Hue[0.9173315569041055, 1, 1], LineBox[{308, 309}]}, 
      {Hue[0.9072527042244332, 1, 1], LineBox[{309, 310}]}, 
      {Hue[0.9047329910545151, 1, 1], LineBox[{310, 311}]}, 
      {Hue[0.902213277884597, 1, 1], LineBox[{311, 312}]}, 
      {Hue[0.8971738515447609, 1, 1], LineBox[{312, 313}]}, 
      {Hue[0.8870949988650887, 1, 1], LineBox[{313, 314}]}, 
      {Hue[0.8845752856951706, 1, 1], LineBox[{314, 315}]}, 
      {Hue[0.8820555725252526, 1, 1], LineBox[{315, 316}]}, 
      {Hue[0.8770161461854165, 1, 1], LineBox[{316, 317}]}, 
      {Hue[0.8669372935057441, 1, 1], LineBox[{317, 318}]}, 
      {Hue[0.8467795881463996, 1, 1], LineBox[{318, 319}]}, 
      {Hue[0.8442598749764816, 1, 1], LineBox[{319, 320}]}, 
      {Hue[0.8417401618065634, 1, 1], LineBox[{320, 321}]}, 
      {Hue[0.8367007354667274, 1, 1], LineBox[{321, 322}]}, 
      {Hue[0.8266218827870551, 1, 1], LineBox[{322, 323}]}, 
      {Hue[0.8064641774277106, 1, 1], LineBox[{323, 324}]}, 
      {Hue[0.8039444642577924, 1, 1], LineBox[{324, 325}]}, 
      {Hue[0.8014247510878745, 1, 1], LineBox[{325, 326}]}, 
      {Hue[0.7963853247480384, 1, 1], LineBox[{326, 327}]}, 
      {Hue[0.786306472068366, 1, 1], LineBox[{327, 328}]}, 
      {Hue[0.7661487667090214, 1, 1], LineBox[{328, 329}]}, 
      {Hue[0.7637973801119673, 1, 1], LineBox[{329, 330}]}, 
      {Hue[0.7614459935149132, 1, 1], LineBox[{330, 331}]}, 
      {Hue[0.7567432203208047, 1, 1], LineBox[{331, 332}]}, 
      {Hue[0.7473376739325879, 1, 1], LineBox[{332, 333}]}, 
      {Hue[0.7285265811561541, 1, 1], LineBox[{333, 334}]}, 
      {Hue[0.690904395603287, 1, 1], LineBox[{334, 335}]}, 
      {Hue[0.6885530090062326, 1, 1], LineBox[{335, 336}]}, 
      {Hue[0.6862016224091785, 1, 1], LineBox[{336, 337}]}, 
      {Hue[0.68149884921507, 1, 1], LineBox[{337, 338}]}, 
      {Hue[0.6720933028268532, 1, 1], LineBox[{338, 339}]}, 
      {Hue[0.6532822100504196, 1, 1], LineBox[{339, 340}]}, 
      {Hue[0.6156600244975523, 1, 1], LineBox[{340, 341}]}, 
      {Hue[0.6131115577162632, 1, 1], LineBox[{341, 342}]}, 
      {Hue[0.6105630909349742, 1, 1], LineBox[{342, 343}]}, 
      {Hue[0.605466157372396, 1, 1], LineBox[{343, 344}]}, 
      {Hue[0.5952722902472397, 1, 1], LineBox[{344, 345}]}, 
      {Hue[0.5748845559969271, 1, 1], LineBox[{345, 346}]}, 
      {Hue[0.5341090874963016, 1, 1], LineBox[{346, 347}]}, 
      {Hue[0.45255815049505094`, 1, 1], LineBox[{347, 348}]}, 
      {Hue[0.4500562080342596, 1, 1], LineBox[{348, 349}]}, 
      {Hue[0.4475542655734681, 1, 1], LineBox[{349, 350}]}, 
      {Hue[0.4425503806518852, 1, 1], LineBox[{350, 351}]}, 
      {Hue[0.4325426108087192, 1, 1], LineBox[{351, 352}]}, 
      {Hue[0.41252707112238746`, 1, 1], LineBox[{352, 353}]}, 
      {Hue[0.372495991749724, 1, 1], LineBox[{353, 354}]}, 
      {Hue[0.2924338330043968, 1, 1], LineBox[{354, 355}]}, 
      {Hue[0.14308241617702985`, 1, 1], LineBox[{355, 356}]}, 
      {Hue[0.9811178675586304, 1, 1], LineBox[{356, 357}]}, 
      {Hue[0.829926219603518, 1, 1], LineBox[{357, 358}]}, 
      {Hue[0.681712128160253, 1, 1], LineBox[{358, 359}]}, 
      {Hue[0.520884904925956, 1, 1], LineBox[{359, 360}]}, 
      {Hue[0.5185403061357836, 1, 1], LineBox[{360, 361}]}, 
      {Hue[0.5161957073456116, 1, 1], LineBox[{361, 362}]}, 
      {Hue[0.5115065097652677, 1, 1], LineBox[{362, 363}]}, 
      {Hue[0.5021281146045796, 1, 1], LineBox[{363, 364}]}, 
      {Hue[0.48337132428320306`, 1, 1], LineBox[{364, 365}]}, 
      {Hue[0.4458577436404507, 1, 1], LineBox[{365, 366}]}, 
      {Hue[0.3708305823549457, 1, 1], LineBox[{366, 367}]}, 
      {Hue[0.3682889033805386, 1, 1], LineBox[{367, 368}]}, 
      {Hue[0.365747224406132, 1, 1], LineBox[{368, 369}]}, 
      {Hue[0.36066386645731824`, 1, 1], LineBox[{369, 370}]}, 
      {Hue[0.3504971505596905, 1, 1], LineBox[{370, 371}]}, 
      {Hue[0.33016371876443495`, 1, 1], LineBox[{371, 372}]}, 
      {Hue[0.2894968551739242, 1, 1], LineBox[{372, 373}]}, 
      {Hue[0.2081631279929033, 1, 1], LineBox[{373, 374}]}, 
      {Hue[0.2057897755913601, 1, 1], LineBox[{374, 375}]}, 
      {Hue[0.20341642318981695`, 1, 1], LineBox[{375, 376}]}, 
      {Hue[0.19866971838673092`, 1, 1], LineBox[{376, 377}]}, 
      {Hue[0.18917630878055883`, 1, 1], LineBox[{377, 378}]}, 
      {Hue[0.1701894895682144, 1, 1], LineBox[{378, 379}]}, 
      {Hue[0.1322158511435255, 1, 1], LineBox[{379, 380}]}, 
      {Hue[0.1298424987419823, 1, 1], LineBox[{380, 381}]}, 
      {Hue[0.12746914634043943`, 1, 1], LineBox[{381, 382}]}, 
      {Hue[0.12272244153735341`, 1, 1], LineBox[{382, 383}]}, 
      {Hue[0.11322903193118104`, 1, 1], LineBox[{383, 384}]}, 
      {Hue[0.09424221271883658, 1, 1], LineBox[{384, 385}]}, 
      {Hue[0.056268574294147965`, 1, 1], LineBox[{385, 386}]}, 
      {Hue[0.053941746213102444`, 1, 1], LineBox[{386, 387}]}, 
      {Hue[0.05161491813205696, 1, 1], LineBox[{387, 388}]}, 
      {Hue[0.046961261969965945`, 1, 1], LineBox[{388, 389}]}, 
      {Hue[0.03765394964578421, 1, 1], LineBox[{389, 390}]}, 
      {Hue[0.019039324997420736`, 1, 1], LineBox[{390, 391}]}, 
      {Hue[0.9818100757006936, 1, 1], LineBox[{391, 392}]}, 
      {Hue[0.979483247619648, 1, 1], LineBox[{392, 393}]}, 
      {Hue[0.9771564195386025, 1, 1], LineBox[{393, 394}]}, 
      {Hue[0.9725027633765115, 1, 1], LineBox[{394, 395}]}, 
      {Hue[0.9631954510523297, 1, 1], LineBox[{395, 396}]}, 
      {Hue[0.9445808264039662, 1, 1], LineBox[{396, 397}]}, 
      {Hue[0.9073515771072393, 1, 1], LineBox[{397, 398}]}, 
      {Hue[0.9048276688419594, 1, 1], LineBox[{398, 399}]}, 
      {Hue[0.9023037605766788, 1, 1], LineBox[{399, 400}]}, 
      {Hue[0.8972559440461179, 1, 1], LineBox[{400, 401}]}, 
      {Hue[0.8871603109849967, 1, 1], LineBox[{401, 402}]}, 
      {Hue[0.8669690448627543, 1, 1], LineBox[{402, 403}]}, 
      {Hue[0.8265865126182694, 1, 1], LineBox[{403, 404}]}, 
      {Hue[0.824062604352989, 1, 1], LineBox[{404, 405}]}, 
      {Hue[0.8215386960877087, 1, 1], LineBox[{405, 406}]}, 
      {Hue[0.8164908795571483, 1, 1], LineBox[{406, 407}]}, 
      {Hue[0.8063952464960271, 1, 1], LineBox[{407, 408}]}, 
      {Hue[0.7862039803737844, 1, 1], LineBox[{408, 409}]}, 
      {Hue[0.7458214481292991, 1, 1], LineBox[{409, 410}]}, 
      {Hue[0.7434658664368825, 1, 1], LineBox[{410, 411}]}, 
      {Hue[0.741110284744466, 1, 1], LineBox[{411, 412}]}, 
      {Hue[0.736399121359633, 1, 1], LineBox[{412, 413}]}, 
      {Hue[0.7269767945899674, 1, 1], LineBox[{413, 414}]}, 
      {Hue[0.7081321410506356, 1, 1], LineBox[{414, 415}]}, 
      {Hue[0.6704428339719721, 1, 1], LineBox[{415, 416}]}, 
      {Hue[0.6680872522795556, 1, 1], LineBox[{416, 417}]}, 
      {Hue[0.665731670587139, 1, 1], LineBox[{417, 418}]}, 
      {Hue[0.661020507202306, 1, 1], LineBox[{418, 419}]}, 
      {Hue[0.6515981804326405, 1, 1], LineBox[{419, 420}]}, 
      {Hue[0.6327535268933087, 1, 1], LineBox[{420, 421}]}, 
      {Hue[0.5950642198146456, 1, 1], LineBox[{421, 422}]}, 
      {Hue[0.5925115579379945, 1, 1], LineBox[{422, 423}]}, 
      {Hue[0.5899588960613429, 1, 1], LineBox[{423, 424}]}, 
      {Hue[0.5848535723080399, 1, 1], LineBox[{424, 425}]}, 
      {Hue[0.5746429248014346, 1, 1], LineBox[{425, 426}]}, 
      {Hue[0.5542216297882241, 1, 1], LineBox[{426, 427}]}, 
      {Hue[0.5133790397618024, 1, 1], LineBox[{427, 428}]}, 
      {Hue[0.5108263778851515, 1, 1], LineBox[{428, 429}]}, 
      {Hue[0.5082737160085, 1, 1], LineBox[{429, 430}]}, 
      {Hue[0.5031683922551968, 1, 1], LineBox[{430, 431}]}, 
      {Hue[0.4929577447485917, 1, 1], LineBox[{431, 432}]}, 
      {Hue[0.4725364497353812, 1, 1], LineBox[{432, 433}]}, 
      {Hue[0.4316938597089599, 1, 1], LineBox[{433, 434}]}, 
      {Hue[0.429187722152806, 1, 1], LineBox[{434, 435}]}, 
      {Hue[0.4266815845966521, 1, 1], LineBox[{435, 436}]}, 
      {Hue[0.42166930948434495`, 1, 1], LineBox[{436, 437}]}, 
      {Hue[0.41164475925973004`, 1, 1], LineBox[{437, 438}]}, 
      {Hue[0.39159565881050024`, 1, 1], LineBox[{438, 439}]}, 
      {Hue[0.35149745791204057`, 1, 1], LineBox[{439, 440}]}, 
      {Hue[0.3489913203558867, 1, 1], LineBox[{440, 441}]}, 
      {Hue[0.34648518279973284`, 1, 1], LineBox[{441, 442}]}, 
      {Hue[0.34147290768742566`, 1, 1], LineBox[{442, 443}]}, 
      {Hue[0.33144835746281076`, 1, 1], LineBox[{443, 444}]}, 
      {Hue[0.3113992570135809, 1, 1], LineBox[{444, 445}]}, 
      {Hue[0.27130105611512123`, 1, 1], LineBox[{445, 446}]}, 
      {Hue[0.26896324513183134`, 1, 1], LineBox[{446, 447}]}, 
      {Hue[0.26662543414854145`, 1, 1], LineBox[{447, 448}]}, 
      {Hue[0.26194981218196167`, 1, 1], LineBox[{448, 449}]}, 
      {Hue[0.25259856824880217`, 1, 1], LineBox[{449, 450}]}, 
      {Hue[0.23389608038248313`, 1, 1], LineBox[{450, 451}]}, 
      {Hue[0.19649110464984562`, 1, 1], LineBox[{451, 452}]}, 
      {Hue[0.19415329366655634`, 1, 1], LineBox[{452, 453}]}, 
      {Hue[0.19181548268326584`, 1, 1], LineBox[{453, 454}]}, 
      {Hue[0.18713986071668665`, 1, 1], LineBox[{454, 455}]}, 
      {Hue[0.17778861678352656`, 1, 1], LineBox[{455, 456}]}, 
      {Hue[0.1590861289172081, 1, 1], LineBox[{456, 457}]}, 
      {Hue[0.12168115318456944`, 1, 1], LineBox[{457, 458}]}, 
      {Hue[0.11914626201704458`, 1, 1], LineBox[{458, 459}]}, 
      {Hue[0.11661137084952025`, 1, 1], LineBox[{459, 460}]}, 
      {Hue[0.11154158851447046`, 1, 1], LineBox[{460, 461}]}, 
      {Hue[0.10140202384437148`, 1, 1], LineBox[{461, 462}]}, 
      {Hue[0.0811228945041741, 1, 1], LineBox[{462, 463}]}, 
      {Hue[0.0405646358237776, 1, 1], LineBox[{463, 464}]}, 
      {Hue[0.038029744656252716`, 1, 1], LineBox[{464, 465}]}, 
      {Hue[0.035494853488728405`, 1, 1], LineBox[{465, 466}]}, 
      {Hue[0.03042507115367863, 1, 1], LineBox[{466, 467}]}, 
      {Hue[0.020285506483580216`, 1, 1], LineBox[{467, 468}]}, 
      {Hue[6.377143381686833*^-6, 1, 1], LineBox[{468, 469}]}, 
      {Hue[0.9594481184629863, 1, 1], LineBox[{469, 470}]}, 
      {Hue[0.9570815538683255, 1, 1], LineBox[{470, 471}]}, 
      {Hue[0.9547149892736645, 1, 1], LineBox[{471, 472}]}, 
      {Hue[0.9499818600843427, 1, 1], LineBox[{472, 473}]}, 
      {Hue[0.9405156017056991, 1, 1], LineBox[{473, 474}]}, 
      {Hue[0.921583084948412, 1, 1], LineBox[{474, 475}]}, 
      {Hue[0.8837180514338376, 1, 1], LineBox[{475, 476}]}, 
      {Hue[0.8813514868391773, 1, 1], LineBox[{476, 477}]}, 
      {Hue[0.8789849222445164, 1, 1], LineBox[{477, 478}]}, 
      {Hue[0.8742517930551946, 1, 1], LineBox[{478, 479}]}, 
      {Hue[0.8647855346765511, 1, 1], LineBox[{479, 480}]}, 
      {Hue[0.8458530179192638, 1, 1], LineBox[{480, 481}]}, 
      {Hue[0.8079879844046894, 1, 1], LineBox[{481, 482}]}, 
      {Hue[0.8056679441305262, 1, 1], LineBox[{482, 483}]}, 
      {Hue[0.8033479038563625, 1, 1], LineBox[{483, 484}]}, 
      {Hue[0.7987078233080366, 1, 1], LineBox[{484, 485}]}, 
      {Hue[0.7894276622113829, 1, 1], LineBox[{485, 486}]}, 
      {Hue[0.7708673400180771, 1, 1], LineBox[{486, 487}]}, 
      {Hue[0.7685472997439133, 1, 1], LineBox[{487, 488}]}, 
      {Hue[0.76622725946975, 1, 1], LineBox[{488, 489}]}, 
      {Hue[0.7615871789214236, 1, 1], LineBox[{489, 490}]}, 
      {Hue[0.7523070178247706, 1, 1], LineBox[{490, 491}]}, 
      {Hue[0.7337466956314647, 1, 1], LineBox[{491, 492}]}, 
      {Hue[0.7314266553573009, 1, 1], LineBox[{492, 493}]}, 
      {Hue[0.7291066150831382, 1, 1], LineBox[{493, 494}]}, 
      {Hue[0.7244665345348111, 1, 1], LineBox[{494, 495}]}, 
      {Hue[0.7151863734381587, 1, 1], LineBox[{495, 496}]}, 
      {Hue[0.6966260512448523, 1, 1], LineBox[{496, 497}]}, 
      {Hue[0.65950540685824, 1, 1], LineBox[{497, 498}]}, 
      {Hue[0.6569882863998417, 1, 1], LineBox[{498, 499}]}, 
      {Hue[0.6544711659414435, 1, 1], LineBox[{499, 500}]}, 
      {Hue[0.6494369250246474, 1, 1], LineBox[{500, 501}]}, 
      {Hue[0.6393684431910545, 1, 1], LineBox[{501, 502}]}, 
      {Hue[0.6192314795238697, 1, 1], LineBox[{502, 503}]}, 
      {Hue[0.616714359065472, 1, 1], LineBox[{503, 504}]}, 
      {Hue[0.6141972386070738, 1, 1], LineBox[{504, 505}]}, 
      {Hue[0.6091629976902777, 1, 1], LineBox[{505, 506}]}, 
      {Hue[0.5990945158566849, 1, 1], LineBox[{506, 507}]}, 
      {Hue[0.5789575521894994, 1, 1], LineBox[{507, 508}]}, 
      {Hue[0.5764404317311012, 1, 1], LineBox[{508, 509}]}, 
      {Hue[0.5739233112727036, 1, 1], LineBox[{509, 510}]}, 
      {Hue[0.5688890703559071, 1, 1], LineBox[{510, 511}]}, 
      {Hue[0.5588205885223141, 1, 1], LineBox[{511, 512}]}, 
      {Hue[0.5386836248551292, 1, 1], LineBox[{512, 513}]}, 
      {Hue[0.4984096975207585, 1, 1], LineBox[{513, 514}]}, 
      {Hue[0.49606090363522365`, 1, 1], LineBox[{514, 515}]}, 
      {Hue[0.49371210974968993`, 1, 1], LineBox[{515, 516}]}, 
      {Hue[0.48901452197862144`, 1, 1], LineBox[{516, 517}]}, 
      {Hue[0.4796193464364844, 1, 1], LineBox[{517, 518}]}, 
      {Hue[0.4608289953522104, 1, 1], LineBox[{518, 519}]}, 
      {Hue[0.45848020146667556`, 1, 1], LineBox[{519, 520}]}, 
      {Hue[0.4561314075811413, 1, 1], LineBox[{520, 521}]}, 
      {Hue[0.4514338198100728, 1, 1], LineBox[{521, 522}]}, 
      {Hue[0.4420386442679352, 1, 1], LineBox[{522, 523}]}, 
      {Hue[0.42324829318366114`, 1, 1], LineBox[{523, 524}]}, 
      {Hue[0.4208994992981274, 1, 1], LineBox[{524, 525}]}, 
      {Hue[0.41855070541259265`, 1, 1], LineBox[{525, 526}]}, 
      {Hue[0.4138531176415241, 1, 1], LineBox[{526, 527}]}, 
      {Hue[0.40445794209938707`, 1, 1], LineBox[{527, 528}]}, 
      {Hue[0.3856675910151125, 1, 1], LineBox[{528, 529}]}, 
      {Hue[0.3480868888465638, 1, 1], LineBox[{529, 530}]}, 
      {Hue[0.3455410147767951, 1, 1], LineBox[{530, 531}]}, 
      {Hue[0.34299514070702586`, 1, 1], LineBox[{531, 532}]}, 
      {Hue[0.3379033925674879, 1, 1], LineBox[{532, 533}]}, 
      {Hue[0.32771989628841086`, 1, 1], LineBox[{533, 534}]}, 
      {Hue[0.30735290373025737`, 1, 1], LineBox[{534, 535}]}, 
      {Hue[0.30480702966048806`, 1, 1], LineBox[{535, 536}]}, 
      {Hue[0.30226115559071937`, 1, 1], LineBox[{536, 537}]}, 
      {Hue[0.2971694074511808, 1, 1], LineBox[{537, 538}]}, 
      {Hue[0.28698591117210376`, 1, 1], LineBox[{538, 539}]}, 
      {Hue[0.2666189186139508, 1, 1], LineBox[{539, 540}]}, 
      {Hue[0.26407304454418157`, 1, 1], LineBox[{540, 541}]}, 
      {Hue[0.2615271704744123, 1, 1], LineBox[{541, 542}]}, 
      {Hue[0.25643542233487376`, 1, 1], LineBox[{542, 543}]}, 
      {Hue[0.2462519260557973, 1, 1], LineBox[{543, 544}]}, 
      {Hue[0.22588493349764377`, 1, 1], LineBox[{544, 545}]}, 
      {Hue[0.18515094838133725`, 1, 1], LineBox[{545, 546}]}, 
      {Hue[0.185074863589463, 1, 1], LineBox[{546, 547}]}, 
      {Hue[0.18499877879758939`, 1, 1], LineBox[{547, 548}]}, 
      {Hue[0.18484660921384086`, 1, 1], LineBox[{548, 549}]}, 
      {Hue[0.1845422700463445, 1, 1], LineBox[{549, 550}]}, 
      {Hue[0.18393359171135223`, 1, 1], LineBox[{550, 551}]}, 
      {Hue[0.18271623504136603`, 1, 1], LineBox[{551, 552}]}, 
      {Hue[0.18264015024949232`, 1, 1], LineBox[{552, 553}]}, 
      {Hue[0.18256406545761808`, 1, 1], LineBox[{553, 554}]}, 
      {Hue[0.18241189587386958`, 1, 1], LineBox[{554, 555}]}, 
      {Hue[0.18210755670637377`, 1, 1], LineBox[{555, 556}]}, 
      {Hue[0.1814988783713804, 1, 1], LineBox[{556, 557}]}, 
      {Hue[0.1814227935795061, 1, 1], LineBox[{557, 558}]}, 
      {Hue[0.18134670878763243`, 1, 1], LineBox[{558, 559}]}, 
      {Hue[0.18119453920388395`, 1, 1], LineBox[{559, 560}]}, 
      {Hue[0.18089020003638753`, 1, 1], LineBox[{560, 561}]}, 
      {Hue[0.1808141152445139, 1, 1], LineBox[{561, 562}]}, 
      {Hue[0.18073803045263961`, 1, 1], LineBox[{562, 563}]}, 
      {Hue[0.18058586086889114`, 1, 1], LineBox[{563, 564}]}, 
      {Hue[0.18050977607701693`, 1, 1], LineBox[{564, 565}]}, 
      {Hue[0.18043369128514264`, 1, 1], LineBox[{565, 566}]}, 
      {Hue[0.18035760649326896`, 1, 1], LineBox[{566, 567}]}}}},
   VertexColors->{
     Hue[0.8197184782986047, 1, 1], 
     Hue[0.817375481892275, 1, 1], 
     Hue[0.8150324854859454, 1, 1], 
     Hue[0.8103464926732862, 1, 1], 
     Hue[0.8009745070479671, 1, 1], 
     Hue[0.7822305357973296, 1, 1], 
     Hue[0.779887539391, 1, 1], 
     Hue[0.7775445429846704, 1, 1], 
     Hue[0.7728585501720105, 1, 1], 
     Hue[0.763486564546692, 1, 1], 
     Hue[0.7447425932960545, 1, 1], 
     Hue[0.7423995968897242, 1, 1], 
     Hue[0.7400566004833946, 1, 1], 
     Hue[0.7353706076707348, 1, 1], 
     Hue[0.7259986220454163, 1, 1], 
     Hue[0.7072546507947788, 1, 1], 
     Hue[0.6697667082935036, 1, 1], 
     Hue[0.6672266317029389, 1, 1], 
     Hue[0.6646865551123748, 1, 1], 
     Hue[0.6596064019312451, 1, 1], 
     Hue[0.6494460955689871, 1, 1], 
     Hue[0.6291254828444701, 1, 1], 
     Hue[0.626585406253906, 1, 1], 
     Hue[0.6240453296633414, 1, 1], 
     Hue[0.6189651764822122, 1, 1], 
     Hue[0.6088048701199542, 1, 1], 
     Hue[0.5884842573954372, 1, 1], 
     Hue[0.5859441808048731, 1, 1], 
     Hue[0.5834041042143084, 1, 1], 
     Hue[0.5783239510331792, 1, 1], 
     Hue[0.5681636446709212, 1, 1], 
     Hue[0.5478430319464042, 1, 1], 
     Hue[0.5072018064973707, 1, 1], 
     Hue[0.5048300564796699, 1, 1], 
     Hue[0.5024583064619688, 1, 1], 
     Hue[0.4977148064265675, 1, 1], 
     Hue[0.48822780635576496`, 1, 1], 
     Hue[0.4692538062141587, 1, 1], 
     Hue[0.466882056196458, 1, 1], 
     Hue[0.4645103061787574, 1, 1], 
     Hue[0.45976680614335613`, 1, 1], 
     Hue[0.45027980607255297`, 1, 1], 
     Hue[0.43130580593094725`, 1, 1], 
     Hue[0.4289340559132466, 1, 1], 
     Hue[0.426562305895546, 1, 1], 
     Hue[0.42181880586014414`, 1, 1], 
     Hue[0.4123318057893416, 1, 1], 
     Hue[0.39335780564773587`, 1, 1], 
     Hue[0.3554098053645245, 1, 1], 
     Hue[0.3530845796673215, 1, 1], 
     Hue[0.35075935397011854`, 1, 1], 
     Hue[0.3461089025757114, 1, 1], 
     Hue[0.33680799978689946`, 1, 1], 
     Hue[0.3182061942092745, 1, 1], 
     Hue[0.3158809685120721, 1, 1], 
     Hue[0.31355574281486853`, 1, 1], 
     Hue[0.308905291420462, 1, 1], 
     Hue[0.2996043886316495, 1, 1], 
     Hue[0.2810025830540251, 1, 1], 
     Hue[0.2786773573568221, 1, 1], 
     Hue[0.2763521316596185, 1, 1], 
     Hue[0.2717016802652126, 1, 1], 
     Hue[0.26240077747640006`, 1, 1], 
     Hue[0.24379897189877506`, 1, 1], 
     Hue[0.2065953607435256, 1, 1], 
     Hue[0.20407305486208763`, 1, 1], 
     Hue[0.20155074898064965`, 1, 1], 
     Hue[0.19650613721777366`, 1, 1], 
     Hue[0.1864169136920217, 1, 1], 
     Hue[0.16623846664051775`, 1, 1], 
     Hue[0.16371616075907977`, 1, 1], 
     Hue[0.1611938548776418, 1, 1], 
     Hue[0.1561492431147658, 1, 1], 
     Hue[0.14606001958901382`, 1, 1], 
     Hue[0.12588157253750992`, 1, 1], 
     Hue[0.12335926665607193`, 1, 1], 
     Hue[0.12083696077463393`, 1, 1], 
     Hue[0.11579234901175797`, 1, 1], 
     Hue[0.105703125486006, 1, 1], 
     Hue[0.0855246784345021, 1, 1], 
     Hue[0.04516778433149423, 1, 1], 
     Hue[0.04281380502292024, 1, 1], 
     Hue[0.04045982571434625, 1, 1], 
     Hue[0.035751867097198235`, 1, 1], 
     Hue[0.026335949862901102`, 1, 1], 
     Hue[0.007504115394307963, 1, 1], 
     Hue[0.9698404464571223, 1, 1], 
     Hue[0.9674864671485477, 1, 1], 
     Hue[0.9651324878399743, 1, 1], 
     Hue[0.9604245292228257, 1, 1], 
     Hue[0.9510086119885292, 1, 1], 
     Hue[0.9321767775199366, 1, 1], 
     Hue[0.8945131085827502, 1, 1], 
     Hue[0.8919620490899413, 1, 1], 
     Hue[0.8894109895971323, 1, 1], 
     Hue[0.8843088706115143, 1, 1], 
     Hue[0.8741046326402782, 1, 1], 
     Hue[0.8536961566978062, 1, 1], 
     Hue[0.8128792048128621, 1, 1], 
     Hue[0.810328145320053, 1, 1], 
     Hue[0.8077770858272441, 1, 1], 
     Hue[0.802674966841626, 1, 1], 
     Hue[0.79247072887039, 1, 1], 
     Hue[0.772062252927918, 1, 1], 
     Hue[0.7312453010429738, 1, 1], 
     Hue[0.7287407658706625, 1, 1], 
     Hue[0.7262362306983512, 1, 1], 
     Hue[0.7212271603537285, 1, 1], 
     Hue[0.711209019664483, 1, 1], 
     Hue[0.6911727382859922, 1, 1], 
     Hue[0.6511001755290096, 1, 1], 
     Hue[0.6485956403566981, 1, 1], 
     Hue[0.646091105184387, 1, 1], 
     Hue[0.6410820348397641, 1, 1], 
     Hue[0.6310638941505182, 1, 1], 
     Hue[0.6110276127720268, 1, 1], 
     Hue[0.5709550500150447, 1, 1], 
     Hue[0.5686188414155973, 1, 1], 
     Hue[0.5662826328161494, 1, 1], 
     Hue[0.5616102156172547, 1, 1], 
     Hue[0.5522653812194647, 1, 1], 
     Hue[0.5335757124238844, 1, 1], 
     Hue[0.49619637483272355`, 1, 1], 
     Hue[0.49386016623327594`, 1, 1], 
     Hue[0.49152395763382856`, 1, 1], 
     Hue[0.48685154043493384`, 1, 1], 
     Hue[0.47750670603714357`, 1, 1], 
     Hue[0.458817037241563, 1, 1], 
     Hue[0.4214376996504027, 1, 1], 
     Hue[0.4189044108667203, 1, 1], 
     Hue[0.41637112208303795`, 1, 1], 
     Hue[0.4113045445156732, 1, 1], 
     Hue[0.4011713893809435, 1, 1], 
     Hue[0.38090507911148397`, 1, 1], 
     Hue[0.34037245857256526`, 1, 1], 
     Hue[0.3378391697888829, 1, 1], 
     Hue[0.33530588100520053`, 1, 1], 
     Hue[0.3302393034378358, 1, 1], 
     Hue[0.3201061483031061, 1, 1], 
     Hue[0.2998398380336466, 1, 1], 
     Hue[0.2593072174947282, 1, 1], 
     Hue[0.2569422552839095, 1, 1], 
     Hue[0.2545772930730911, 1, 1], 
     Hue[0.24984736865145438`, 1, 1], 
     Hue[0.24038751980818002`, 1, 1], 
     Hue[0.22146782212163132`, 1, 1], 
     Hue[0.18362842674853475`, 1, 1], 
     Hue[0.18126346453771638`, 1, 1], 
     Hue[0.17889850232689772`, 1, 1], 
     Hue[0.17416857790526039`, 1, 1], 
     Hue[0.16470872906198633`, 1, 1], 
     Hue[0.14578903137543817`, 1, 1], 
     Hue[0.10794963600234132`, 1, 1], 
     Hue[0.10563119811202033`, 1, 1], 
     Hue[0.10331276022169904`, 1, 1], 
     Hue[0.09867588444105733, 1, 1], 
     Hue[0.08940213287977387, 1, 1], 
     Hue[0.07085462975720616, 1, 1], 
     Hue[0.03375962351207098, 1, 1], 
     Hue[0.03144118562175026, 1, 1], 
     Hue[0.029122747731428977`, 1, 1], 
     Hue[0.024485871950787247`, 1, 1], 
     Hue[0.015212120389503814`, 1, 1], 
     Hue[0.9966646172669361, 1, 1], 
     Hue[0.9595696110218012, 1, 1], 
     Hue[0.9570540929472452, 1, 1], 
     Hue[0.9545385748726898, 1, 1], 
     Hue[0.949507538723578, 1, 1], 
     Hue[0.9394454664253545, 1, 1], 
     Hue[0.9193213218289079, 1, 1], 
     Hue[0.8790730326360152, 1, 1], 
     Hue[0.7985764542502292, 1, 1], 
     Hue[0.7962292627485371, 1, 1], 
     Hue[0.7938820712468451, 1, 1], 
     Hue[0.7891876882434611, 1, 1], 
     Hue[0.7797989222366932, 1, 1], 
     Hue[0.7610213902231577, 1, 1], 
     Hue[0.7234663261960863, 1, 1], 
     Hue[0.6483561981419439, 1, 1], 
     Hue[0.6458119264560171, 1, 1], 
     Hue[0.6432676547700904, 1, 1], 
     Hue[0.6381791113982367, 1, 1], 
     Hue[0.6280020246545293, 1, 1], 
     Hue[0.6076478511671144, 1, 1], 
     Hue[0.5669395041922853, 1, 1], 
     Hue[0.4855228102426269, 1, 1], 
     Hue[0.3256669788551567, 1, 1], 
     Hue[0.17658404813097406`, 1, 1], 
     Hue[0.014887985615758862`, 1, 1], 
     Hue[0.8639648237638307, 1, 1], 
     Hue[0.7004285301208703, 1, 1], 
     Hue[0.6979197998531967, 1, 1], 
     Hue[0.6954110695855231, 1, 1], 
     Hue[0.6903936090501757, 1, 1], 
     Hue[0.6803586879794812, 1, 1], 
     Hue[0.660288845838092, 1, 1], 
     Hue[0.6201491615553137, 1, 1], 
     Hue[0.5398697929897572, 1, 1], 
     Hue[0.5375293892949474, 1, 1], 
     Hue[0.5351889856001377, 1, 1], 
     Hue[0.5305081782105181, 1, 1], 
     Hue[0.5211465634312791, 1, 1], 
     Hue[0.5024233338728007, 1, 1], 
     Hue[0.46497687475584426`, 1, 1], 
     Hue[0.4626364710610345, 1, 1], 
     Hue[0.46029606736622475`, 1, 1], 
     Hue[0.45561525997660507`, 1, 1], 
     Hue[0.446253645197366, 1, 1], 
     Hue[0.4275304156388878, 1, 1], 
     Hue[0.3900839565219313, 1, 1], 
     Hue[0.38754647264288655`, 1, 1], 
     Hue[0.38500898876384193`, 1, 1], 
     Hue[0.37993402100575263`, 1, 1], 
     Hue[0.3697840854895739, 1, 1], 
     Hue[0.3494842144572167, 1, 1], 
     Hue[0.34694673057817194`, 1, 1], 
     Hue[0.34440924669912737`, 1, 1], 
     Hue[0.339334278941038, 1, 1], 
     Hue[0.32918434342485936`, 1, 1], 
     Hue[0.3088844723925021, 1, 1], 
     Hue[0.30634698851345743`, 1, 1], 
     Hue[0.30380950463441275`, 1, 1], 
     Hue[0.29873453687632345`, 1, 1], 
     Hue[0.2885846013601448, 1, 1], 
     Hue[0.2682847303277875, 1, 1], 
     Hue[0.22768498826307296`, 1, 1], 
     Hue[0.2253158309568922, 1, 1], 
     Hue[0.22294667365071136`, 1, 1], 
     Hue[0.21820835903834976`, 1, 1], 
     Hue[0.20873172981362653`, 1, 1], 
     Hue[0.18977847136418016`, 1, 1], 
     Hue[0.18740931405799935`, 1, 1], 
     Hue[0.18504015675181856`, 1, 1], 
     Hue[0.18030184213945694`, 1, 1], 
     Hue[0.17082521291473377`, 1, 1], 
     Hue[0.1518719544652873, 1, 1], 
     Hue[0.14950279715910653`, 1, 1], 
     Hue[0.14713363985292574`, 1, 1], 
     Hue[0.1423953252405641, 1, 1], 
     Hue[0.1329186960158409, 1, 1], 
     Hue[0.13054953870966016`, 1, 1], 
     Hue[0.12818038140347934`, 1, 1], 
     Hue[0.12344206679111772`, 1, 1], 
     Hue[0.11396543756639452`, 1, 1], 
     Hue[0.11159628026021372`, 1, 1], 
     Hue[0.1092271229540329, 1, 1], 
     Hue[0.10448880834167132`, 1, 1], 
     Hue[0.09501217911694812, 1, 1], 
     Hue[0.0926430218107673, 1, 1], 
     Hue[0.09027386450458652, 1, 1], 
     Hue[0.08553554989222491, 1, 1], 
     Hue[0.07605892066750171, 1, 1], 
     Hue[0.07373628768181852, 1, 1], 
     Hue[0.07141365469613532, 1, 1], 
     Hue[0.06676838872476895, 1, 1], 
     Hue[0.06444575573908579, 1, 1], 
     Hue[0.06212312275340258, 1, 1], 
     Hue[0.05747785678203621, 1, 1], 
     Hue[0.05515522379635303, 1, 1], 
     Hue[0.052832590810669836`, 1, 1], 
     Hue[0.04818732483930348, 1, 1], 
     Hue[0.04586469185362028, 1, 1], 
     Hue[0.043542058867937085`, 1, 1], 
     Hue[0.03889679289657071, 1, 1], 
     Hue[0.03657415991088753, 1, 1], 
     Hue[0.03425152692520435, 1, 1], 
     Hue[0.031928893939521155`, 1, 1], 
     Hue[0.029606260953837966`, 1, 1], 
     Hue[0.02728362796815478, 1, 1], 
     Hue[0.024960994982471587`, 1, 1], 
     Hue[0.0226383619967884, 1, 1], 
     Hue[0.02031572901110522, 1, 1], 
     Hue[0.017993096025422036`, 1, 1], 
     Hue[0.015670463039738843`, 1, 1], 
     Hue[0.013347830054055657`, 1, 1], 
     Hue[0.011025197068372471`, 1, 1], 
     Hue[0.008702564082689282, 1, 1], 
     Hue[0.006379931097006096, 1, 1], 
     Hue[0.004057298111322911, 1, 1], 
     Hue[0.0017346651256397233`, 1, 1], 
     Hue[0.9994120321399566, 1, 1], 
     Hue[0.9970893991542733, 1, 1], 
     Hue[0.9947667661685902, 1, 1], 
     Hue[0.992444133182907, 1, 1], 
     Hue[0.9901215001972238, 1, 1], 
     Hue[0.9877988672115406, 1, 1], 
     Hue[0.9854762342258574, 1, 1], 
     Hue[0.9831536012401743, 1, 1], 
     Hue[0.980830968254491, 1, 1], 
     Hue[0.9785083352688079, 1, 1], 
     Hue[0.9761857022831246, 1, 1], 
     Hue[0.9738630692974415, 1, 1], 
     Hue[0.9715404363117582, 1, 1], 
     Hue[0.9692178033260751, 1, 1], 
     Hue[0.9645725373547087, 1, 1], 
     Hue[0.9622499043690256, 1, 1], 
     Hue[0.9599272713833423, 1, 1], 
     Hue[0.9552820054119759, 1, 1], 
     Hue[0.9529593724262928, 1, 1], 
     Hue[0.9506367394406096, 1, 1], 
     Hue[0.9459914734692433, 1, 1], 
     Hue[0.94366884048356, 1, 1], 
     Hue[0.9413462074978769, 1, 1], 
     Hue[0.9367009415265105, 1, 1], 
     Hue[0.9274104095837777, 1, 1], 
     Hue[0.9248906964138597, 1, 1], 
     Hue[0.9223709832439416, 1, 1], 
     Hue[0.9173315569041055, 1, 1], 
     Hue[0.9072527042244332, 1, 1], 
     Hue[0.9047329910545151, 1, 1], 
     Hue[0.902213277884597, 1, 1], 
     Hue[0.8971738515447609, 1, 1], 
     Hue[0.8870949988650887, 1, 1], 
     Hue[0.8845752856951706, 1, 1], 
     Hue[0.8820555725252526, 1, 1], 
     Hue[0.8770161461854165, 1, 1], 
     Hue[0.8669372935057441, 1, 1], 
     Hue[0.8467795881463996, 1, 1], 
     Hue[0.8442598749764816, 1, 1], 
     Hue[0.8417401618065634, 1, 1], 
     Hue[0.8367007354667274, 1, 1], 
     Hue[0.8266218827870551, 1, 1], 
     Hue[0.8064641774277106, 1, 1], 
     Hue[0.8039444642577924, 1, 1], 
     Hue[0.8014247510878745, 1, 1], 
     Hue[0.7963853247480384, 1, 1], 
     Hue[0.786306472068366, 1, 1], 
     Hue[0.7661487667090214, 1, 1], 
     Hue[0.7637973801119673, 1, 1], 
     Hue[0.7614459935149132, 1, 1], 
     Hue[0.7567432203208047, 1, 1], 
     Hue[0.7473376739325879, 1, 1], 
     Hue[0.7285265811561541, 1, 1], 
     Hue[0.690904395603287, 1, 1], 
     Hue[0.6885530090062326, 1, 1], 
     Hue[0.6862016224091785, 1, 1], 
     Hue[0.68149884921507, 1, 1], 
     Hue[0.6720933028268532, 1, 1], 
     Hue[0.6532822100504196, 1, 1], 
     Hue[0.6156600244975523, 1, 1], 
     Hue[0.6131115577162632, 1, 1], 
     Hue[0.6105630909349742, 1, 1], 
     Hue[0.605466157372396, 1, 1], 
     Hue[0.5952722902472397, 1, 1], 
     Hue[0.5748845559969271, 1, 1], 
     Hue[0.5341090874963016, 1, 1], 
     Hue[0.45255815049505094`, 1, 1], 
     Hue[0.4500562080342596, 1, 1], 
     Hue[0.4475542655734681, 1, 1], 
     Hue[0.4425503806518852, 1, 1], 
     Hue[0.4325426108087192, 1, 1], 
     Hue[0.41252707112238746`, 1, 1], 
     Hue[0.372495991749724, 1, 1], 
     Hue[0.2924338330043968, 1, 1], 
     Hue[0.14308241617702985`, 1, 1], 
     Hue[0.9811178675586304, 1, 1], 
     Hue[0.829926219603518, 1, 1], 
     Hue[0.681712128160253, 1, 1], 
     Hue[0.520884904925956, 1, 1], 
     Hue[0.5185403061357836, 1, 1], 
     Hue[0.5161957073456116, 1, 1], 
     Hue[0.5115065097652677, 1, 1], 
     Hue[0.5021281146045796, 1, 1], 
     Hue[0.48337132428320306`, 1, 1], 
     Hue[0.4458577436404507, 1, 1], 
     Hue[0.3708305823549457, 1, 1], 
     Hue[0.3682889033805386, 1, 1], 
     Hue[0.365747224406132, 1, 1], 
     Hue[0.36066386645731824`, 1, 1], 
     Hue[0.3504971505596905, 1, 1], 
     Hue[0.33016371876443495`, 1, 1], 
     Hue[0.2894968551739242, 1, 1], 
     Hue[0.2081631279929033, 1, 1], 
     Hue[0.2057897755913601, 1, 1], 
     Hue[0.20341642318981695`, 1, 1], 
     Hue[0.19866971838673092`, 1, 1], 
     Hue[0.18917630878055883`, 1, 1], 
     Hue[0.1701894895682144, 1, 1], 
     Hue[0.1322158511435255, 1, 1], 
     Hue[0.1298424987419823, 1, 1], 
     Hue[0.12746914634043943`, 1, 1], 
     Hue[0.12272244153735341`, 1, 1], 
     Hue[0.11322903193118104`, 1, 1], 
     Hue[0.09424221271883658, 1, 1], 
     Hue[0.056268574294147965`, 1, 1], 
     Hue[0.053941746213102444`, 1, 1], 
     Hue[0.05161491813205696, 1, 1], 
     Hue[0.046961261969965945`, 1, 1], 
     Hue[0.03765394964578421, 1, 1], 
     Hue[0.019039324997420736`, 1, 1], 
     Hue[0.9818100757006936, 1, 1], 
     Hue[0.979483247619648, 1, 1], 
     Hue[0.9771564195386025, 1, 1], 
     Hue[0.9725027633765115, 1, 1], 
     Hue[0.9631954510523297, 1, 1], 
     Hue[0.9445808264039662, 1, 1], 
     Hue[0.9073515771072393, 1, 1], 
     Hue[0.9048276688419594, 1, 1], 
     Hue[0.9023037605766788, 1, 1], 
     Hue[0.8972559440461179, 1, 1], 
     Hue[0.8871603109849967, 1, 1], 
     Hue[0.8669690448627543, 1, 1], 
     Hue[0.8265865126182694, 1, 1], 
     Hue[0.824062604352989, 1, 1], 
     Hue[0.8215386960877087, 1, 1], 
     Hue[0.8164908795571483, 1, 1], 
     Hue[0.8063952464960271, 1, 1], 
     Hue[0.7862039803737844, 1, 1], 
     Hue[0.7458214481292991, 1, 1], 
     Hue[0.7434658664368825, 1, 1], 
     Hue[0.741110284744466, 1, 1], 
     Hue[0.736399121359633, 1, 1], 
     Hue[0.7269767945899674, 1, 1], 
     Hue[0.7081321410506356, 1, 1], 
     Hue[0.6704428339719721, 1, 1], 
     Hue[0.6680872522795556, 1, 1], 
     Hue[0.665731670587139, 1, 1], 
     Hue[0.661020507202306, 1, 1], 
     Hue[0.6515981804326405, 1, 1], 
     Hue[0.6327535268933087, 1, 1], 
     Hue[0.5950642198146456, 1, 1], 
     Hue[0.5925115579379945, 1, 1], 
     Hue[0.5899588960613429, 1, 1], 
     Hue[0.5848535723080399, 1, 1], 
     Hue[0.5746429248014346, 1, 1], 
     Hue[0.5542216297882241, 1, 1], 
     Hue[0.5133790397618024, 1, 1], 
     Hue[0.5108263778851515, 1, 1], 
     Hue[0.5082737160085, 1, 1], 
     Hue[0.5031683922551968, 1, 1], 
     Hue[0.4929577447485917, 1, 1], 
     Hue[0.4725364497353812, 1, 1], 
     Hue[0.4316938597089599, 1, 1], 
     Hue[0.429187722152806, 1, 1], 
     Hue[0.4266815845966521, 1, 1], 
     Hue[0.42166930948434495`, 1, 1], 
     Hue[0.41164475925973004`, 1, 1], 
     Hue[0.39159565881050024`, 1, 1], 
     Hue[0.35149745791204057`, 1, 1], 
     Hue[0.3489913203558867, 1, 1], 
     Hue[0.34648518279973284`, 1, 1], 
     Hue[0.34147290768742566`, 1, 1], 
     Hue[0.33144835746281076`, 1, 1], 
     Hue[0.3113992570135809, 1, 1], 
     Hue[0.27130105611512123`, 1, 1], 
     Hue[0.26896324513183134`, 1, 1], 
     Hue[0.26662543414854145`, 1, 1], 
     Hue[0.26194981218196167`, 1, 1], 
     Hue[0.25259856824880217`, 1, 1], 
     Hue[0.23389608038248313`, 1, 1], 
     Hue[0.19649110464984562`, 1, 1], 
     Hue[0.19415329366655634`, 1, 1], 
     Hue[0.19181548268326584`, 1, 1], 
     Hue[0.18713986071668665`, 1, 1], 
     Hue[0.17778861678352656`, 1, 1], 
     Hue[0.1590861289172081, 1, 1], 
     Hue[0.12168115318456944`, 1, 1], 
     Hue[0.11914626201704458`, 1, 1], 
     Hue[0.11661137084952025`, 1, 1], 
     Hue[0.11154158851447046`, 1, 1], 
     Hue[0.10140202384437148`, 1, 1], 
     Hue[0.0811228945041741, 1, 1], 
     Hue[0.0405646358237776, 1, 1], 
     Hue[0.038029744656252716`, 1, 1], 
     Hue[0.035494853488728405`, 1, 1], 
     Hue[0.03042507115367863, 1, 1], 
     Hue[0.020285506483580216`, 1, 1], 
     Hue[6.377143381686833*^-6, 1, 1], 
     Hue[0.9594481184629863, 1, 1], 
     Hue[0.9570815538683255, 1, 1], 
     Hue[0.9547149892736645, 1, 1], 
     Hue[0.9499818600843427, 1, 1], 
     Hue[0.9405156017056991, 1, 1], 
     Hue[0.921583084948412, 1, 1], 
     Hue[0.8837180514338376, 1, 1], 
     Hue[0.8813514868391773, 1, 1], 
     Hue[0.8789849222445164, 1, 1], 
     Hue[0.8742517930551946, 1, 1], 
     Hue[0.8647855346765511, 1, 1], 
     Hue[0.8458530179192638, 1, 1], 
     Hue[0.8079879844046894, 1, 1], 
     Hue[0.8056679441305262, 1, 1], 
     Hue[0.8033479038563625, 1, 1], 
     Hue[0.7987078233080366, 1, 1], 
     Hue[0.7894276622113829, 1, 1], 
     Hue[0.7708673400180771, 1, 1], 
     Hue[0.7685472997439133, 1, 1], 
     Hue[0.76622725946975, 1, 1], 
     Hue[0.7615871789214236, 1, 1], 
     Hue[0.7523070178247706, 1, 1], 
     Hue[0.7337466956314647, 1, 1], 
     Hue[0.7314266553573009, 1, 1], 
     Hue[0.7291066150831382, 1, 1], 
     Hue[0.7244665345348111, 1, 1], 
     Hue[0.7151863734381587, 1, 1], 
     Hue[0.6966260512448523, 1, 1], 
     Hue[0.65950540685824, 1, 1], 
     Hue[0.6569882863998417, 1, 1], 
     Hue[0.6544711659414435, 1, 1], 
     Hue[0.6494369250246474, 1, 1], 
     Hue[0.6393684431910545, 1, 1], 
     Hue[0.6192314795238697, 1, 1], 
     Hue[0.616714359065472, 1, 1], 
     Hue[0.6141972386070738, 1, 1], 
     Hue[0.6091629976902777, 1, 1], 
     Hue[0.5990945158566849, 1, 1], 
     Hue[0.5789575521894994, 1, 1], 
     Hue[0.5764404317311012, 1, 1], 
     Hue[0.5739233112727036, 1, 1], 
     Hue[0.5688890703559071, 1, 1], 
     Hue[0.5588205885223141, 1, 1], 
     Hue[0.5386836248551292, 1, 1], 
     Hue[0.4984096975207585, 1, 1], 
     Hue[0.49606090363522365`, 1, 1], 
     Hue[0.49371210974968993`, 1, 1], 
     Hue[0.48901452197862144`, 1, 1], 
     Hue[0.4796193464364844, 1, 1], 
     Hue[0.4608289953522104, 1, 1], 
     Hue[0.45848020146667556`, 1, 1], 
     Hue[0.4561314075811413, 1, 1], 
     Hue[0.4514338198100728, 1, 1], 
     Hue[0.4420386442679352, 1, 1], 
     Hue[0.42324829318366114`, 1, 1], 
     Hue[0.4208994992981274, 1, 1], 
     Hue[0.41855070541259265`, 1, 1], 
     Hue[0.4138531176415241, 1, 1], 
     Hue[0.40445794209938707`, 1, 1], 
     Hue[0.3856675910151125, 1, 1], 
     Hue[0.3480868888465638, 1, 1], 
     Hue[0.3455410147767951, 1, 1], 
     Hue[0.34299514070702586`, 1, 1], 
     Hue[0.3379033925674879, 1, 1], 
     Hue[0.32771989628841086`, 1, 1], 
     Hue[0.30735290373025737`, 1, 1], 
     Hue[0.30480702966048806`, 1, 1], 
     Hue[0.30226115559071937`, 1, 1], 
     Hue[0.2971694074511808, 1, 1], 
     Hue[0.28698591117210376`, 1, 1], 
     Hue[0.2666189186139508, 1, 1], 
     Hue[0.26407304454418157`, 1, 1], 
     Hue[0.2615271704744123, 1, 1], 
     Hue[0.25643542233487376`, 1, 1], 
     Hue[0.2462519260557973, 1, 1], 
     Hue[0.22588493349764377`, 1, 1], 
     Hue[0.18515094838133725`, 1, 1], 
     Hue[0.185074863589463, 1, 1], 
     Hue[0.18499877879758939`, 1, 1], 
     Hue[0.18484660921384086`, 1, 1], 
     Hue[0.1845422700463445, 1, 1], 
     Hue[0.18393359171135223`, 1, 1], 
     Hue[0.18271623504136603`, 1, 1], 
     Hue[0.18264015024949232`, 1, 1], 
     Hue[0.18256406545761808`, 1, 1], 
     Hue[0.18241189587386958`, 1, 1], 
     Hue[0.18210755670637377`, 1, 1], 
     Hue[0.1814988783713804, 1, 1], 
     Hue[0.1814227935795061, 1, 1], 
     Hue[0.18134670878763243`, 1, 1], 
     Hue[0.18119453920388395`, 1, 1], 
     Hue[0.18089020003638753`, 1, 1], 
     Hue[0.1808141152445139, 1, 1], 
     Hue[0.18073803045263961`, 1, 1], 
     Hue[0.18058586086889114`, 1, 1], 
     Hue[0.18050977607701693`, 1, 1], 
     Hue[0.18043369128514264`, 1, 1], 
     Hue[0.18035760649326896`, 1, 1], 
     Hue[0.18028152170139528`, 1, 1], 
     Hue[0.8197184782986047, 1, 1], 
     Hue[0.817375481892275, 1, 1], 
     Hue[0.817375481892275, 1, 1], 
     Hue[0.8150324854859454, 1, 1], 
     Hue[0.8150324854859454, 1, 1], 
     Hue[0.8103464926732862, 1, 1], 
     Hue[0.8103464926732862, 1, 1], 
     Hue[0.8009745070479671, 1, 1], 
     Hue[0.8009745070479671, 1, 1], 
     Hue[0.7822305357973296, 1, 1], 
     Hue[0.7822305357973296, 1, 1], 
     Hue[0.779887539391, 1, 1], 
     Hue[0.779887539391, 1, 1], 
     Hue[0.7775445429846704, 1, 1], 
     Hue[0.7775445429846704, 1, 1], 
     Hue[0.7728585501720105, 1, 1], 
     Hue[0.7728585501720105, 1, 1], 
     Hue[0.763486564546692, 1, 1], 
     Hue[0.763486564546692, 1, 1], 
     Hue[0.7447425932960545, 1, 1], 
     Hue[0.7447425932960545, 1, 1], 
     Hue[0.7423995968897242, 1, 1], 
     Hue[0.7423995968897242, 1, 1], 
     Hue[0.7400566004833946, 1, 1], 
     Hue[0.7400566004833946, 1, 1], 
     Hue[0.7353706076707348, 1, 1], 
     Hue[0.7353706076707348, 1, 1], 
     Hue[0.7259986220454163, 1, 1], 
     Hue[0.7259986220454163, 1, 1], 
     Hue[0.7072546507947788, 1, 1], 
     Hue[0.7072546507947788, 1, 1], 
     Hue[0.6697667082935036, 1, 1], 
     Hue[0.6697667082935036, 1, 1], 
     Hue[0.6672266317029389, 1, 1], 
     Hue[0.6672266317029389, 1, 1], 
     Hue[0.6646865551123748, 1, 1], 
     Hue[0.6646865551123748, 1, 1], 
     Hue[0.6596064019312451, 1, 1], 
     Hue[0.6596064019312451, 1, 1], 
     Hue[0.6494460955689871, 1, 1], 
     Hue[0.6494460955689871, 1, 1], 
     Hue[0.6291254828444701, 1, 1], 
     Hue[0.6291254828444701, 1, 1], 
     Hue[0.626585406253906, 1, 1], 
     Hue[0.626585406253906, 1, 1], 
     Hue[0.6240453296633414, 1, 1], 
     Hue[0.6240453296633414, 1, 1], 
     Hue[0.6189651764822122, 1, 1], 
     Hue[0.6189651764822122, 1, 1], 
     Hue[0.6088048701199542, 1, 1], 
     Hue[0.6088048701199542, 1, 1], 
     Hue[0.5884842573954372, 1, 1], 
     Hue[0.5884842573954372, 1, 1], 
     Hue[0.5859441808048731, 1, 1], 
     Hue[0.5859441808048731, 1, 1], 
     Hue[0.5834041042143084, 1, 1], 
     Hue[0.5834041042143084, 1, 1], 
     Hue[0.5783239510331792, 1, 1], 
     Hue[0.5783239510331792, 1, 1], 
     Hue[0.5681636446709212, 1, 1], 
     Hue[0.5681636446709212, 1, 1], 
     Hue[0.5478430319464042, 1, 1], 
     Hue[0.5478430319464042, 1, 1], 
     Hue[0.5072018064973707, 1, 1], 
     Hue[0.5072018064973707, 1, 1], 
     Hue[0.5048300564796699, 1, 1], 
     Hue[0.5048300564796699, 1, 1], 
     Hue[0.5024583064619688, 1, 1], 
     Hue[0.5024583064619688, 1, 1], 
     Hue[0.4977148064265675, 1, 1], 
     Hue[0.4977148064265675, 1, 1], 
     Hue[0.48822780635576496`, 1, 1], 
     Hue[0.48822780635576496`, 1, 1], 
     Hue[0.4692538062141587, 1, 1], 
     Hue[0.4692538062141587, 1, 1], 
     Hue[0.466882056196458, 1, 1], 
     Hue[0.466882056196458, 1, 1], 
     Hue[0.4645103061787574, 1, 1], 
     Hue[0.4645103061787574, 1, 1], 
     Hue[0.45976680614335613`, 1, 1], 
     Hue[0.45976680614335613`, 1, 1], 
     Hue[0.45027980607255297`, 1, 1], 
     Hue[0.45027980607255297`, 1, 1], 
     Hue[0.43130580593094725`, 1, 1], 
     Hue[0.43130580593094725`, 1, 1], 
     Hue[0.4289340559132466, 1, 1], 
     Hue[0.4289340559132466, 1, 1], 
     Hue[0.426562305895546, 1, 1], 
     Hue[0.426562305895546, 1, 1], 
     Hue[0.42181880586014414`, 1, 1], 
     Hue[0.42181880586014414`, 1, 1], 
     Hue[0.4123318057893416, 1, 1], 
     Hue[0.4123318057893416, 1, 1], 
     Hue[0.39335780564773587`, 1, 1], 
     Hue[0.39335780564773587`, 1, 1], 
     Hue[0.3554098053645245, 1, 1], 
     Hue[0.3554098053645245, 1, 1], 
     Hue[0.3530845796673215, 1, 1], 
     Hue[0.3530845796673215, 1, 1], 
     Hue[0.35075935397011854`, 1, 1], 
     Hue[0.35075935397011854`, 1, 1], 
     Hue[0.3461089025757114, 1, 1], 
     Hue[0.3461089025757114, 1, 1], 
     Hue[0.33680799978689946`, 1, 1], 
     Hue[0.33680799978689946`, 1, 1], 
     Hue[0.3182061942092745, 1, 1], 
     Hue[0.3182061942092745, 1, 1], 
     Hue[0.3158809685120721, 1, 1], 
     Hue[0.3158809685120721, 1, 1], 
     Hue[0.31355574281486853`, 1, 1], 
     Hue[0.31355574281486853`, 1, 1], 
     Hue[0.308905291420462, 1, 1], 
     Hue[0.308905291420462, 1, 1], 
     Hue[0.2996043886316495, 1, 1], 
     Hue[0.2996043886316495, 1, 1], 
     Hue[0.2810025830540251, 1, 1], 
     Hue[0.2810025830540251, 1, 1], 
     Hue[0.2786773573568221, 1, 1], 
     Hue[0.2786773573568221, 1, 1], 
     Hue[0.2763521316596185, 1, 1], 
     Hue[0.2763521316596185, 1, 1], 
     Hue[0.2717016802652126, 1, 1], 
     Hue[0.2717016802652126, 1, 1], 
     Hue[0.26240077747640006`, 1, 1], 
     Hue[0.26240077747640006`, 1, 1], 
     Hue[0.24379897189877506`, 1, 1], 
     Hue[0.24379897189877506`, 1, 1], 
     Hue[0.2065953607435256, 1, 1], 
     Hue[0.2065953607435256, 1, 1], 
     Hue[0.20407305486208763`, 1, 1], 
     Hue[0.20407305486208763`, 1, 1], 
     Hue[0.20155074898064965`, 1, 1], 
     Hue[0.20155074898064965`, 1, 1], 
     Hue[0.19650613721777366`, 1, 1], 
     Hue[0.19650613721777366`, 1, 1], 
     Hue[0.1864169136920217, 1, 1], 
     Hue[0.1864169136920217, 1, 1], 
     Hue[0.16623846664051775`, 1, 1], 
     Hue[0.16623846664051775`, 1, 1], 
     Hue[0.16371616075907977`, 1, 1], 
     Hue[0.16371616075907977`, 1, 1], 
     Hue[0.1611938548776418, 1, 1], 
     Hue[0.1611938548776418, 1, 1], 
     Hue[0.1561492431147658, 1, 1], 
     Hue[0.1561492431147658, 1, 1], 
     Hue[0.14606001958901382`, 1, 1], 
     Hue[0.14606001958901382`, 1, 1], 
     Hue[0.12588157253750992`, 1, 1], 
     Hue[0.12588157253750992`, 1, 1], 
     Hue[0.12335926665607193`, 1, 1], 
     Hue[0.12335926665607193`, 1, 1], 
     Hue[0.12083696077463393`, 1, 1], 
     Hue[0.12083696077463393`, 1, 1], 
     Hue[0.11579234901175797`, 1, 1], 
     Hue[0.11579234901175797`, 1, 1], 
     Hue[0.105703125486006, 1, 1], 
     Hue[0.105703125486006, 1, 1], 
     Hue[0.0855246784345021, 1, 1], 
     Hue[0.0855246784345021, 1, 1], 
     Hue[0.04516778433149423, 1, 1], 
     Hue[0.04516778433149423, 1, 1], 
     Hue[0.04281380502292024, 1, 1], 
     Hue[0.04281380502292024, 1, 1], 
     Hue[0.04045982571434625, 1, 1], 
     Hue[0.04045982571434625, 1, 1], 
     Hue[0.035751867097198235`, 1, 1], 
     Hue[0.035751867097198235`, 1, 1], 
     Hue[0.026335949862901102`, 1, 1], 
     Hue[0.026335949862901102`, 1, 1], 
     Hue[0.007504115394307963, 1, 1], 
     Hue[0.007504115394307963, 1, 1], 
     Hue[0.9698404464571223, 1, 1], 
     Hue[0.9698404464571223, 1, 1], 
     Hue[0.9674864671485477, 1, 1], 
     Hue[0.9674864671485477, 1, 1], 
     Hue[0.9651324878399743, 1, 1], 
     Hue[0.9651324878399743, 1, 1], 
     Hue[0.9604245292228257, 1, 1], 
     Hue[0.9604245292228257, 1, 1], 
     Hue[0.9510086119885292, 1, 1], 
     Hue[0.9510086119885292, 1, 1], 
     Hue[0.9321767775199366, 1, 1], 
     Hue[0.9321767775199366, 1, 1], 
     Hue[0.8945131085827502, 1, 1], 
     Hue[0.8945131085827502, 1, 1], 
     Hue[0.8919620490899413, 1, 1], 
     Hue[0.8919620490899413, 1, 1], 
     Hue[0.8894109895971323, 1, 1], 
     Hue[0.8894109895971323, 1, 1], 
     Hue[0.8843088706115143, 1, 1], 
     Hue[0.8843088706115143, 1, 1], 
     Hue[0.8741046326402782, 1, 1], 
     Hue[0.8741046326402782, 1, 1], 
     Hue[0.8536961566978062, 1, 1], 
     Hue[0.8536961566978062, 1, 1], 
     Hue[0.8128792048128621, 1, 1], 
     Hue[0.8128792048128621, 1, 1], 
     Hue[0.810328145320053, 1, 1], 
     Hue[0.810328145320053, 1, 1], 
     Hue[0.8077770858272441, 1, 1], 
     Hue[0.8077770858272441, 1, 1], 
     Hue[0.802674966841626, 1, 1], 
     Hue[0.802674966841626, 1, 1], 
     Hue[0.79247072887039, 1, 1], 
     Hue[0.79247072887039, 1, 1], 
     Hue[0.772062252927918, 1, 1], 
     Hue[0.772062252927918, 1, 1], 
     Hue[0.7312453010429738, 1, 1], 
     Hue[0.7312453010429738, 1, 1], 
     Hue[0.7287407658706625, 1, 1], 
     Hue[0.7287407658706625, 1, 1], 
     Hue[0.7262362306983512, 1, 1], 
     Hue[0.7262362306983512, 1, 1], 
     Hue[0.7212271603537285, 1, 1], 
     Hue[0.7212271603537285, 1, 1], 
     Hue[0.711209019664483, 1, 1], 
     Hue[0.711209019664483, 1, 1], 
     Hue[0.6911727382859922, 1, 1], 
     Hue[0.6911727382859922, 1, 1], 
     Hue[0.6511001755290096, 1, 1], 
     Hue[0.6511001755290096, 1, 1], 
     Hue[0.6485956403566981, 1, 1], 
     Hue[0.6485956403566981, 1, 1], 
     Hue[0.646091105184387, 1, 1], 
     Hue[0.646091105184387, 1, 1], 
     Hue[0.6410820348397641, 1, 1], 
     Hue[0.6410820348397641, 1, 1], 
     Hue[0.6310638941505182, 1, 1], 
     Hue[0.6310638941505182, 1, 1], 
     Hue[0.6110276127720268, 1, 1], 
     Hue[0.6110276127720268, 1, 1], 
     Hue[0.5709550500150447, 1, 1], 
     Hue[0.5709550500150447, 1, 1], 
     Hue[0.5686188414155973, 1, 1], 
     Hue[0.5686188414155973, 1, 1], 
     Hue[0.5662826328161494, 1, 1], 
     Hue[0.5662826328161494, 1, 1], 
     Hue[0.5616102156172547, 1, 1], 
     Hue[0.5616102156172547, 1, 1], 
     Hue[0.5522653812194647, 1, 1], 
     Hue[0.5522653812194647, 1, 1], 
     Hue[0.5335757124238844, 1, 1], 
     Hue[0.5335757124238844, 1, 1], 
     Hue[0.49619637483272355`, 1, 1], 
     Hue[0.49619637483272355`, 1, 1], 
     Hue[0.49386016623327594`, 1, 1], 
     Hue[0.49386016623327594`, 1, 1], 
     Hue[0.49152395763382856`, 1, 1], 
     Hue[0.49152395763382856`, 1, 1], 
     Hue[0.48685154043493384`, 1, 1], 
     Hue[0.48685154043493384`, 1, 1], 
     Hue[0.47750670603714357`, 1, 1], 
     Hue[0.47750670603714357`, 1, 1], 
     Hue[0.458817037241563, 1, 1], 
     Hue[0.458817037241563, 1, 1], 
     Hue[0.4214376996504027, 1, 1], 
     Hue[0.4214376996504027, 1, 1], 
     Hue[0.4189044108667203, 1, 1], 
     Hue[0.4189044108667203, 1, 1], 
     Hue[0.41637112208303795`, 1, 1], 
     Hue[0.41637112208303795`, 1, 1], 
     Hue[0.4113045445156732, 1, 1], 
     Hue[0.4113045445156732, 1, 1], 
     Hue[0.4011713893809435, 1, 1], 
     Hue[0.4011713893809435, 1, 1], 
     Hue[0.38090507911148397`, 1, 1], 
     Hue[0.38090507911148397`, 1, 1], 
     Hue[0.34037245857256526`, 1, 1], 
     Hue[0.34037245857256526`, 1, 1], 
     Hue[0.3378391697888829, 1, 1], 
     Hue[0.3378391697888829, 1, 1], 
     Hue[0.33530588100520053`, 1, 1], 
     Hue[0.33530588100520053`, 1, 1], 
     Hue[0.3302393034378358, 1, 1], 
     Hue[0.3302393034378358, 1, 1], 
     Hue[0.3201061483031061, 1, 1], 
     Hue[0.3201061483031061, 1, 1], 
     Hue[0.2998398380336466, 1, 1], 
     Hue[0.2998398380336466, 1, 1], 
     Hue[0.2593072174947282, 1, 1], 
     Hue[0.2593072174947282, 1, 1], 
     Hue[0.2569422552839095, 1, 1], 
     Hue[0.2569422552839095, 1, 1], 
     Hue[0.2545772930730911, 1, 1], 
     Hue[0.2545772930730911, 1, 1], 
     Hue[0.24984736865145438`, 1, 1], 
     Hue[0.24984736865145438`, 1, 1], 
     Hue[0.24038751980818002`, 1, 1], 
     Hue[0.24038751980818002`, 1, 1], 
     Hue[0.22146782212163132`, 1, 1], 
     Hue[0.22146782212163132`, 1, 1], 
     Hue[0.18362842674853475`, 1, 1], 
     Hue[0.18362842674853475`, 1, 1], 
     Hue[0.18126346453771638`, 1, 1], 
     Hue[0.18126346453771638`, 1, 1], 
     Hue[0.17889850232689772`, 1, 1], 
     Hue[0.17889850232689772`, 1, 1], 
     Hue[0.17416857790526039`, 1, 1], 
     Hue[0.17416857790526039`, 1, 1], 
     Hue[0.16470872906198633`, 1, 1], 
     Hue[0.16470872906198633`, 1, 1], 
     Hue[0.14578903137543817`, 1, 1], 
     Hue[0.14578903137543817`, 1, 1], 
     Hue[0.10794963600234132`, 1, 1], 
     Hue[0.10794963600234132`, 1, 1], 
     Hue[0.10563119811202033`, 1, 1], 
     Hue[0.10563119811202033`, 1, 1], 
     Hue[0.10331276022169904`, 1, 1], 
     Hue[0.10331276022169904`, 1, 1], 
     Hue[0.09867588444105733, 1, 1], 
     Hue[0.09867588444105733, 1, 1], 
     Hue[0.08940213287977387, 1, 1], 
     Hue[0.08940213287977387, 1, 1], 
     Hue[0.07085462975720616, 1, 1], 
     Hue[0.07085462975720616, 1, 1], 
     Hue[0.03375962351207098, 1, 1], 
     Hue[0.03375962351207098, 1, 1], 
     Hue[0.03144118562175026, 1, 1], 
     Hue[0.03144118562175026, 1, 1], 
     Hue[0.029122747731428977`, 1, 1], 
     Hue[0.029122747731428977`, 1, 1], 
     Hue[0.024485871950787247`, 1, 1], 
     Hue[0.024485871950787247`, 1, 1], 
     Hue[0.015212120389503814`, 1, 1], 
     Hue[0.015212120389503814`, 1, 1], 
     Hue[0.9966646172669361, 1, 1], 
     Hue[0.9966646172669361, 1, 1], 
     Hue[0.9595696110218012, 1, 1], 
     Hue[0.9595696110218012, 1, 1], 
     Hue[0.9570540929472452, 1, 1], 
     Hue[0.9570540929472452, 1, 1], 
     Hue[0.9545385748726898, 1, 1], 
     Hue[0.9545385748726898, 1, 1], 
     Hue[0.949507538723578, 1, 1], 
     Hue[0.949507538723578, 1, 1], 
     Hue[0.9394454664253545, 1, 1], 
     Hue[0.9394454664253545, 1, 1], 
     Hue[0.9193213218289079, 1, 1], 
     Hue[0.9193213218289079, 1, 1], 
     Hue[0.8790730326360152, 1, 1], 
     Hue[0.8790730326360152, 1, 1], 
     Hue[0.7985764542502292, 1, 1], 
     Hue[0.7985764542502292, 1, 1], 
     Hue[0.7962292627485371, 1, 1], 
     Hue[0.7962292627485371, 1, 1], 
     Hue[0.7938820712468451, 1, 1], 
     Hue[0.7938820712468451, 1, 1], 
     Hue[0.7891876882434611, 1, 1], 
     Hue[0.7891876882434611, 1, 1], 
     Hue[0.7797989222366932, 1, 1], 
     Hue[0.7797989222366932, 1, 1], 
     Hue[0.7610213902231577, 1, 1], 
     Hue[0.7610213902231577, 1, 1], 
     Hue[0.7234663261960863, 1, 1], 
     Hue[0.7234663261960863, 1, 1], 
     Hue[0.6483561981419439, 1, 1], 
     Hue[0.6483561981419439, 1, 1], 
     Hue[0.6458119264560171, 1, 1], 
     Hue[0.6458119264560171, 1, 1], 
     Hue[0.6432676547700904, 1, 1], 
     Hue[0.6432676547700904, 1, 1], 
     Hue[0.6381791113982367, 1, 1], 
     Hue[0.6381791113982367, 1, 1], 
     Hue[0.6280020246545293, 1, 1], 
     Hue[0.6280020246545293, 1, 1], 
     Hue[0.6076478511671144, 1, 1], 
     Hue[0.6076478511671144, 1, 1], 
     Hue[0.5669395041922853, 1, 1], 
     Hue[0.5669395041922853, 1, 1], 
     Hue[0.4855228102426269, 1, 1], 
     Hue[0.4855228102426269, 1, 1], 
     Hue[0.3256669788551567, 1, 1], 
     Hue[0.3256669788551567, 1, 1], 
     Hue[0.17658404813097406`, 1, 1], 
     Hue[0.17658404813097406`, 1, 1], 
     Hue[0.014887985615758862`, 1, 1], 
     Hue[0.014887985615758862`, 1, 1], 
     Hue[0.8639648237638307, 1, 1], 
     Hue[0.8639648237638307, 1, 1], 
     Hue[0.7004285301208703, 1, 1], 
     Hue[0.7004285301208703, 1, 1], 
     Hue[0.6979197998531967, 1, 1], 
     Hue[0.6979197998531967, 1, 1], 
     Hue[0.6954110695855231, 1, 1], 
     Hue[0.6954110695855231, 1, 1], 
     Hue[0.6903936090501757, 1, 1], 
     Hue[0.6903936090501757, 1, 1], 
     Hue[0.6803586879794812, 1, 1], 
     Hue[0.6803586879794812, 1, 1], 
     Hue[0.660288845838092, 1, 1], 
     Hue[0.660288845838092, 1, 1], 
     Hue[0.6201491615553137, 1, 1], 
     Hue[0.6201491615553137, 1, 1], 
     Hue[0.5398697929897572, 1, 1], 
     Hue[0.5398697929897572, 1, 1], 
     Hue[0.5375293892949474, 1, 1], 
     Hue[0.5375293892949474, 1, 1], 
     Hue[0.5351889856001377, 1, 1], 
     Hue[0.5351889856001377, 1, 1], 
     Hue[0.5305081782105181, 1, 1], 
     Hue[0.5305081782105181, 1, 1], 
     Hue[0.5211465634312791, 1, 1], 
     Hue[0.5211465634312791, 1, 1], 
     Hue[0.5024233338728007, 1, 1], 
     Hue[0.5024233338728007, 1, 1], 
     Hue[0.46497687475584426`, 1, 1], 
     Hue[0.46497687475584426`, 1, 1], 
     Hue[0.4626364710610345, 1, 1], 
     Hue[0.4626364710610345, 1, 1], 
     Hue[0.46029606736622475`, 1, 1], 
     Hue[0.46029606736622475`, 1, 1], 
     Hue[0.45561525997660507`, 1, 1], 
     Hue[0.45561525997660507`, 1, 1], 
     Hue[0.446253645197366, 1, 1], 
     Hue[0.446253645197366, 1, 1], 
     Hue[0.4275304156388878, 1, 1], 
     Hue[0.4275304156388878, 1, 1], 
     Hue[0.3900839565219313, 1, 1], 
     Hue[0.3900839565219313, 1, 1], 
     Hue[0.38754647264288655`, 1, 1], 
     Hue[0.38754647264288655`, 1, 1], 
     Hue[0.38500898876384193`, 1, 1], 
     Hue[0.38500898876384193`, 1, 1], 
     Hue[0.37993402100575263`, 1, 1], 
     Hue[0.37993402100575263`, 1, 1], 
     Hue[0.3697840854895739, 1, 1], 
     Hue[0.3697840854895739, 1, 1], 
     Hue[0.3494842144572167, 1, 1], 
     Hue[0.3494842144572167, 1, 1], 
     Hue[0.34694673057817194`, 1, 1], 
     Hue[0.34694673057817194`, 1, 1], 
     Hue[0.34440924669912737`, 1, 1], 
     Hue[0.34440924669912737`, 1, 1], 
     Hue[0.339334278941038, 1, 1], 
     Hue[0.339334278941038, 1, 1], 
     Hue[0.32918434342485936`, 1, 1], 
     Hue[0.32918434342485936`, 1, 1], 
     Hue[0.3088844723925021, 1, 1], 
     Hue[0.3088844723925021, 1, 1], 
     Hue[0.30634698851345743`, 1, 1], 
     Hue[0.30634698851345743`, 1, 1], 
     Hue[0.30380950463441275`, 1, 1], 
     Hue[0.30380950463441275`, 1, 1], 
     Hue[0.29873453687632345`, 1, 1], 
     Hue[0.29873453687632345`, 1, 1], 
     Hue[0.2885846013601448, 1, 1], 
     Hue[0.2885846013601448, 1, 1], 
     Hue[0.2682847303277875, 1, 1], 
     Hue[0.2682847303277875, 1, 1], 
     Hue[0.22768498826307296`, 1, 1], 
     Hue[0.22768498826307296`, 1, 1], 
     Hue[0.2253158309568922, 1, 1], 
     Hue[0.2253158309568922, 1, 1], 
     Hue[0.22294667365071136`, 1, 1], 
     Hue[0.22294667365071136`, 1, 1], 
     Hue[0.21820835903834976`, 1, 1], 
     Hue[0.21820835903834976`, 1, 1], 
     Hue[0.20873172981362653`, 1, 1], 
     Hue[0.20873172981362653`, 1, 1], 
     Hue[0.18977847136418016`, 1, 1], 
     Hue[0.18977847136418016`, 1, 1], 
     Hue[0.18740931405799935`, 1, 1], 
     Hue[0.18740931405799935`, 1, 1], 
     Hue[0.18504015675181856`, 1, 1], 
     Hue[0.18504015675181856`, 1, 1], 
     Hue[0.18030184213945694`, 1, 1], 
     Hue[0.18030184213945694`, 1, 1], 
     Hue[0.17082521291473377`, 1, 1], 
     Hue[0.17082521291473377`, 1, 1], 
     Hue[0.1518719544652873, 1, 1], 
     Hue[0.1518719544652873, 1, 1], 
     Hue[0.14950279715910653`, 1, 1], 
     Hue[0.14950279715910653`, 1, 1], 
     Hue[0.14713363985292574`, 1, 1], 
     Hue[0.14713363985292574`, 1, 1], 
     Hue[0.1423953252405641, 1, 1], 
     Hue[0.1423953252405641, 1, 1], 
     Hue[0.1329186960158409, 1, 1], 
     Hue[0.1329186960158409, 1, 1], 
     Hue[0.13054953870966016`, 1, 1], 
     Hue[0.13054953870966016`, 1, 1], 
     Hue[0.12818038140347934`, 1, 1], 
     Hue[0.12818038140347934`, 1, 1], 
     Hue[0.12344206679111772`, 1, 1], 
     Hue[0.12344206679111772`, 1, 1], 
     Hue[0.11396543756639452`, 1, 1], 
     Hue[0.11396543756639452`, 1, 1], 
     Hue[0.11159628026021372`, 1, 1], 
     Hue[0.11159628026021372`, 1, 1], 
     Hue[0.1092271229540329, 1, 1], 
     Hue[0.1092271229540329, 1, 1], 
     Hue[0.10448880834167132`, 1, 1], 
     Hue[0.10448880834167132`, 1, 1], 
     Hue[0.09501217911694812, 1, 1], 
     Hue[0.09501217911694812, 1, 1], 
     Hue[0.0926430218107673, 1, 1], 
     Hue[0.0926430218107673, 1, 1], 
     Hue[0.09027386450458652, 1, 1], 
     Hue[0.09027386450458652, 1, 1], 
     Hue[0.08553554989222491, 1, 1], 
     Hue[0.08553554989222491, 1, 1], 
     Hue[0.07605892066750171, 1, 1], 
     Hue[0.07605892066750171, 1, 1], 
     Hue[0.07373628768181852, 1, 1], 
     Hue[0.07373628768181852, 1, 1], 
     Hue[0.07141365469613532, 1, 1], 
     Hue[0.07141365469613532, 1, 1], 
     Hue[0.06676838872476895, 1, 1], 
     Hue[0.06676838872476895, 1, 1], 
     Hue[0.06444575573908579, 1, 1], 
     Hue[0.06444575573908579, 1, 1], 
     Hue[0.06212312275340258, 1, 1], 
     Hue[0.06212312275340258, 1, 1], 
     Hue[0.05747785678203621, 1, 1], 
     Hue[0.05747785678203621, 1, 1], 
     Hue[0.05515522379635303, 1, 1], 
     Hue[0.05515522379635303, 1, 1], 
     Hue[0.052832590810669836`, 1, 1], 
     Hue[0.052832590810669836`, 1, 1], 
     Hue[0.04818732483930348, 1, 1], 
     Hue[0.04818732483930348, 1, 1], 
     Hue[0.04586469185362028, 1, 1], 
     Hue[0.04586469185362028, 1, 1], 
     Hue[0.043542058867937085`, 1, 1], 
     Hue[0.043542058867937085`, 1, 1], 
     Hue[0.03889679289657071, 1, 1], 
     Hue[0.03889679289657071, 1, 1], 
     Hue[0.03657415991088753, 1, 1], 
     Hue[0.03657415991088753, 1, 1], 
     Hue[0.03425152692520435, 1, 1], 
     Hue[0.03425152692520435, 1, 1], 
     Hue[0.031928893939521155`, 1, 1], 
     Hue[0.031928893939521155`, 1, 1], 
     Hue[0.029606260953837966`, 1, 1], 
     Hue[0.029606260953837966`, 1, 1], 
     Hue[0.02728362796815478, 1, 1], 
     Hue[0.02728362796815478, 1, 1], 
     Hue[0.024960994982471587`, 1, 1], 
     Hue[0.024960994982471587`, 1, 1], 
     Hue[0.0226383619967884, 1, 1], 
     Hue[0.0226383619967884, 1, 1], 
     Hue[0.02031572901110522, 1, 1], 
     Hue[0.02031572901110522, 1, 1], 
     Hue[0.017993096025422036`, 1, 1], 
     Hue[0.017993096025422036`, 1, 1], 
     Hue[0.015670463039738843`, 1, 1], 
     Hue[0.015670463039738843`, 1, 1], 
     Hue[0.013347830054055657`, 1, 1], 
     Hue[0.013347830054055657`, 1, 1], 
     Hue[0.011025197068372471`, 1, 1], 
     Hue[0.011025197068372471`, 1, 1], 
     Hue[0.008702564082689282, 1, 1], 
     Hue[0.008702564082689282, 1, 1], 
     Hue[0.006379931097006096, 1, 1], 
     Hue[0.006379931097006096, 1, 1], 
     Hue[0.004057298111322911, 1, 1], 
     Hue[0.004057298111322911, 1, 1], 
     Hue[0.0017346651256397233`, 1, 1], 
     Hue[0.0017346651256397233`, 1, 1], 
     Hue[0.9994120321399566, 1, 1], 
     Hue[0.9994120321399566, 1, 1], 
     Hue[0.9970893991542733, 1, 1], 
     Hue[0.9970893991542733, 1, 1], 
     Hue[0.9947667661685902, 1, 1], 
     Hue[0.9947667661685902, 1, 1], 
     Hue[0.992444133182907, 1, 1], 
     Hue[0.992444133182907, 1, 1], 
     Hue[0.9901215001972238, 1, 1], 
     Hue[0.9901215001972238, 1, 1], 
     Hue[0.9877988672115406, 1, 1], 
     Hue[0.9877988672115406, 1, 1], 
     Hue[0.9854762342258574, 1, 1], 
     Hue[0.9854762342258574, 1, 1], 
     Hue[0.9831536012401743, 1, 1], 
     Hue[0.9831536012401743, 1, 1], 
     Hue[0.980830968254491, 1, 1], 
     Hue[0.980830968254491, 1, 1], 
     Hue[0.9785083352688079, 1, 1], 
     Hue[0.9785083352688079, 1, 1], 
     Hue[0.9761857022831246, 1, 1], 
     Hue[0.9761857022831246, 1, 1], 
     Hue[0.9738630692974415, 1, 1], 
     Hue[0.9738630692974415, 1, 1], 
     Hue[0.9715404363117582, 1, 1], 
     Hue[0.9715404363117582, 1, 1], 
     Hue[0.9692178033260751, 1, 1], 
     Hue[0.9692178033260751, 1, 1], 
     Hue[0.9645725373547087, 1, 1], 
     Hue[0.9645725373547087, 1, 1], 
     Hue[0.9622499043690256, 1, 1], 
     Hue[0.9622499043690256, 1, 1], 
     Hue[0.9599272713833423, 1, 1], 
     Hue[0.9599272713833423, 1, 1], 
     Hue[0.9552820054119759, 1, 1], 
     Hue[0.9552820054119759, 1, 1], 
     Hue[0.9529593724262928, 1, 1], 
     Hue[0.9529593724262928, 1, 1], 
     Hue[0.9506367394406096, 1, 1], 
     Hue[0.9506367394406096, 1, 1], 
     Hue[0.9459914734692433, 1, 1], 
     Hue[0.9459914734692433, 1, 1], 
     Hue[0.94366884048356, 1, 1], 
     Hue[0.94366884048356, 1, 1], 
     Hue[0.9413462074978769, 1, 1], 
     Hue[0.9413462074978769, 1, 1], 
     Hue[0.9367009415265105, 1, 1], 
     Hue[0.9367009415265105, 1, 1], 
     Hue[0.9274104095837777, 1, 1], 
     Hue[0.9274104095837777, 1, 1], 
     Hue[0.9248906964138597, 1, 1], 
     Hue[0.9248906964138597, 1, 1], 
     Hue[0.9223709832439416, 1, 1], 
     Hue[0.9223709832439416, 1, 1], 
     Hue[0.9173315569041055, 1, 1], 
     Hue[0.9173315569041055, 1, 1], 
     Hue[0.9072527042244332, 1, 1], 
     Hue[0.9072527042244332, 1, 1], 
     Hue[0.9047329910545151, 1, 1], 
     Hue[0.9047329910545151, 1, 1], 
     Hue[0.902213277884597, 1, 1], 
     Hue[0.902213277884597, 1, 1], 
     Hue[0.8971738515447609, 1, 1], 
     Hue[0.8971738515447609, 1, 1], 
     Hue[0.8870949988650887, 1, 1], 
     Hue[0.8870949988650887, 1, 1], 
     Hue[0.8845752856951706, 1, 1], 
     Hue[0.8845752856951706, 1, 1], 
     Hue[0.8820555725252526, 1, 1], 
     Hue[0.8820555725252526, 1, 1], 
     Hue[0.8770161461854165, 1, 1], 
     Hue[0.8770161461854165, 1, 1], 
     Hue[0.8669372935057441, 1, 1], 
     Hue[0.8669372935057441, 1, 1], 
     Hue[0.8467795881463996, 1, 1], 
     Hue[0.8467795881463996, 1, 1], 
     Hue[0.8442598749764816, 1, 1], 
     Hue[0.8442598749764816, 1, 1], 
     Hue[0.8417401618065634, 1, 1], 
     Hue[0.8417401618065634, 1, 1], 
     Hue[0.8367007354667274, 1, 1], 
     Hue[0.8367007354667274, 1, 1], 
     Hue[0.8266218827870551, 1, 1], 
     Hue[0.8266218827870551, 1, 1], 
     Hue[0.8064641774277106, 1, 1], 
     Hue[0.8064641774277106, 1, 1], 
     Hue[0.8039444642577924, 1, 1], 
     Hue[0.8039444642577924, 1, 1], 
     Hue[0.8014247510878745, 1, 1], 
     Hue[0.8014247510878745, 1, 1], 
     Hue[0.7963853247480384, 1, 1], 
     Hue[0.7963853247480384, 1, 1], 
     Hue[0.786306472068366, 1, 1], 
     Hue[0.786306472068366, 1, 1], 
     Hue[0.7661487667090214, 1, 1], 
     Hue[0.7661487667090214, 1, 1], 
     Hue[0.7637973801119673, 1, 1], 
     Hue[0.7637973801119673, 1, 1], 
     Hue[0.7614459935149132, 1, 1], 
     Hue[0.7614459935149132, 1, 1], 
     Hue[0.7567432203208047, 1, 1], 
     Hue[0.7567432203208047, 1, 1], 
     Hue[0.7473376739325879, 1, 1], 
     Hue[0.7473376739325879, 1, 1], 
     Hue[0.7285265811561541, 1, 1], 
     Hue[0.7285265811561541, 1, 1], 
     Hue[0.690904395603287, 1, 1], 
     Hue[0.690904395603287, 1, 1], 
     Hue[0.6885530090062326, 1, 1], 
     Hue[0.6885530090062326, 1, 1], 
     Hue[0.6862016224091785, 1, 1], 
     Hue[0.6862016224091785, 1, 1], 
     Hue[0.68149884921507, 1, 1], 
     Hue[0.68149884921507, 1, 1], 
     Hue[0.6720933028268532, 1, 1], 
     Hue[0.6720933028268532, 1, 1], 
     Hue[0.6532822100504196, 1, 1], 
     Hue[0.6532822100504196, 1, 1], 
     Hue[0.6156600244975523, 1, 1], 
     Hue[0.6156600244975523, 1, 1], 
     Hue[0.6131115577162632, 1, 1], 
     Hue[0.6131115577162632, 1, 1], 
     Hue[0.6105630909349742, 1, 1], 
     Hue[0.6105630909349742, 1, 1], 
     Hue[0.605466157372396, 1, 1], 
     Hue[0.605466157372396, 1, 1], 
     Hue[0.5952722902472397, 1, 1], 
     Hue[0.5952722902472397, 1, 1], 
     Hue[0.5748845559969271, 1, 1], 
     Hue[0.5748845559969271, 1, 1], 
     Hue[0.5341090874963016, 1, 1], 
     Hue[0.5341090874963016, 1, 1], 
     Hue[0.45255815049505094`, 1, 1], 
     Hue[0.45255815049505094`, 1, 1], 
     Hue[0.4500562080342596, 1, 1], 
     Hue[0.4500562080342596, 1, 1], 
     Hue[0.4475542655734681, 1, 1], 
     Hue[0.4475542655734681, 1, 1], 
     Hue[0.4425503806518852, 1, 1], 
     Hue[0.4425503806518852, 1, 1], 
     Hue[0.4325426108087192, 1, 1], 
     Hue[0.4325426108087192, 1, 1], 
     Hue[0.41252707112238746`, 1, 1], 
     Hue[0.41252707112238746`, 1, 1], 
     Hue[0.372495991749724, 1, 1], 
     Hue[0.372495991749724, 1, 1], 
     Hue[0.2924338330043968, 1, 1], 
     Hue[0.2924338330043968, 1, 1], 
     Hue[0.14308241617702985`, 1, 1], 
     Hue[0.14308241617702985`, 1, 1], 
     Hue[0.9811178675586304, 1, 1], 
     Hue[0.9811178675586304, 1, 1], 
     Hue[0.829926219603518, 1, 1], 
     Hue[0.829926219603518, 1, 1], 
     Hue[0.681712128160253, 1, 1], 
     Hue[0.681712128160253, 1, 1], 
     Hue[0.520884904925956, 1, 1], 
     Hue[0.520884904925956, 1, 1], 
     Hue[0.5185403061357836, 1, 1], 
     Hue[0.5185403061357836, 1, 1], 
     Hue[0.5161957073456116, 1, 1], 
     Hue[0.5161957073456116, 1, 1], 
     Hue[0.5115065097652677, 1, 1], 
     Hue[0.5115065097652677, 1, 1], 
     Hue[0.5021281146045796, 1, 1], 
     Hue[0.5021281146045796, 1, 1], 
     Hue[0.48337132428320306`, 1, 1], 
     Hue[0.48337132428320306`, 1, 1], 
     Hue[0.4458577436404507, 1, 1], 
     Hue[0.4458577436404507, 1, 1], 
     Hue[0.3708305823549457, 1, 1], 
     Hue[0.3708305823549457, 1, 1], 
     Hue[0.3682889033805386, 1, 1], 
     Hue[0.3682889033805386, 1, 1], 
     Hue[0.365747224406132, 1, 1], 
     Hue[0.365747224406132, 1, 1], 
     Hue[0.36066386645731824`, 1, 1], 
     Hue[0.36066386645731824`, 1, 1], 
     Hue[0.3504971505596905, 1, 1], 
     Hue[0.3504971505596905, 1, 1], 
     Hue[0.33016371876443495`, 1, 1], 
     Hue[0.33016371876443495`, 1, 1], 
     Hue[0.2894968551739242, 1, 1], 
     Hue[0.2894968551739242, 1, 1], 
     Hue[0.2081631279929033, 1, 1], 
     Hue[0.2081631279929033, 1, 1], 
     Hue[0.2057897755913601, 1, 1], 
     Hue[0.2057897755913601, 1, 1], 
     Hue[0.20341642318981695`, 1, 1], 
     Hue[0.20341642318981695`, 1, 1], 
     Hue[0.19866971838673092`, 1, 1], 
     Hue[0.19866971838673092`, 1, 1], 
     Hue[0.18917630878055883`, 1, 1], 
     Hue[0.18917630878055883`, 1, 1], 
     Hue[0.1701894895682144, 1, 1], 
     Hue[0.1701894895682144, 1, 1], 
     Hue[0.1322158511435255, 1, 1], 
     Hue[0.1322158511435255, 1, 1], 
     Hue[0.1298424987419823, 1, 1], 
     Hue[0.1298424987419823, 1, 1], 
     Hue[0.12746914634043943`, 1, 1], 
     Hue[0.12746914634043943`, 1, 1], 
     Hue[0.12272244153735341`, 1, 1], 
     Hue[0.12272244153735341`, 1, 1], 
     Hue[0.11322903193118104`, 1, 1], 
     Hue[0.11322903193118104`, 1, 1], 
     Hue[0.09424221271883658, 1, 1], 
     Hue[0.09424221271883658, 1, 1], 
     Hue[0.056268574294147965`, 1, 1], 
     Hue[0.056268574294147965`, 1, 1], 
     Hue[0.053941746213102444`, 1, 1], 
     Hue[0.053941746213102444`, 1, 1], 
     Hue[0.05161491813205696, 1, 1], 
     Hue[0.05161491813205696, 1, 1], 
     Hue[0.046961261969965945`, 1, 1], 
     Hue[0.046961261969965945`, 1, 1], 
     Hue[0.03765394964578421, 1, 1], 
     Hue[0.03765394964578421, 1, 1], 
     Hue[0.019039324997420736`, 1, 1], 
     Hue[0.019039324997420736`, 1, 1], 
     Hue[0.9818100757006936, 1, 1], 
     Hue[0.9818100757006936, 1, 1], 
     Hue[0.979483247619648, 1, 1], 
     Hue[0.979483247619648, 1, 1], 
     Hue[0.9771564195386025, 1, 1], 
     Hue[0.9771564195386025, 1, 1], 
     Hue[0.9725027633765115, 1, 1], 
     Hue[0.9725027633765115, 1, 1], 
     Hue[0.9631954510523297, 1, 1], 
     Hue[0.9631954510523297, 1, 1], 
     Hue[0.9445808264039662, 1, 1], 
     Hue[0.9445808264039662, 1, 1], 
     Hue[0.9073515771072393, 1, 1], 
     Hue[0.9073515771072393, 1, 1], 
     Hue[0.9048276688419594, 1, 1], 
     Hue[0.9048276688419594, 1, 1], 
     Hue[0.9023037605766788, 1, 1], 
     Hue[0.9023037605766788, 1, 1], 
     Hue[0.8972559440461179, 1, 1], 
     Hue[0.8972559440461179, 1, 1], 
     Hue[0.8871603109849967, 1, 1], 
     Hue[0.8871603109849967, 1, 1], 
     Hue[0.8669690448627543, 1, 1], 
     Hue[0.8669690448627543, 1, 1], 
     Hue[0.8265865126182694, 1, 1], 
     Hue[0.8265865126182694, 1, 1], 
     Hue[0.824062604352989, 1, 1], 
     Hue[0.824062604352989, 1, 1], 
     Hue[0.8215386960877087, 1, 1], 
     Hue[0.8215386960877087, 1, 1], 
     Hue[0.8164908795571483, 1, 1], 
     Hue[0.8164908795571483, 1, 1], 
     Hue[0.8063952464960271, 1, 1], 
     Hue[0.8063952464960271, 1, 1], 
     Hue[0.7862039803737844, 1, 1], 
     Hue[0.7862039803737844, 1, 1], 
     Hue[0.7458214481292991, 1, 1], 
     Hue[0.7458214481292991, 1, 1], 
     Hue[0.7434658664368825, 1, 1], 
     Hue[0.7434658664368825, 1, 1], 
     Hue[0.741110284744466, 1, 1], 
     Hue[0.741110284744466, 1, 1], 
     Hue[0.736399121359633, 1, 1], 
     Hue[0.736399121359633, 1, 1], 
     Hue[0.7269767945899674, 1, 1], 
     Hue[0.7269767945899674, 1, 1], 
     Hue[0.7081321410506356, 1, 1], 
     Hue[0.7081321410506356, 1, 1], 
     Hue[0.6704428339719721, 1, 1], 
     Hue[0.6704428339719721, 1, 1], 
     Hue[0.6680872522795556, 1, 1], 
     Hue[0.6680872522795556, 1, 1], 
     Hue[0.665731670587139, 1, 1], 
     Hue[0.665731670587139, 1, 1], 
     Hue[0.661020507202306, 1, 1], 
     Hue[0.661020507202306, 1, 1], 
     Hue[0.6515981804326405, 1, 1], 
     Hue[0.6515981804326405, 1, 1], 
     Hue[0.6327535268933087, 1, 1], 
     Hue[0.6327535268933087, 1, 1], 
     Hue[0.5950642198146456, 1, 1], 
     Hue[0.5950642198146456, 1, 1], 
     Hue[0.5925115579379945, 1, 1], 
     Hue[0.5925115579379945, 1, 1], 
     Hue[0.5899588960613429, 1, 1], 
     Hue[0.5899588960613429, 1, 1], 
     Hue[0.5848535723080399, 1, 1], 
     Hue[0.5848535723080399, 1, 1], 
     Hue[0.5746429248014346, 1, 1], 
     Hue[0.5746429248014346, 1, 1], 
     Hue[0.5542216297882241, 1, 1], 
     Hue[0.5542216297882241, 1, 1], 
     Hue[0.5133790397618024, 1, 1], 
     Hue[0.5133790397618024, 1, 1], 
     Hue[0.5108263778851515, 1, 1], 
     Hue[0.5108263778851515, 1, 1], 
     Hue[0.5082737160085, 1, 1], 
     Hue[0.5082737160085, 1, 1], 
     Hue[0.5031683922551968, 1, 1], 
     Hue[0.5031683922551968, 1, 1], 
     Hue[0.4929577447485917, 1, 1], 
     Hue[0.4929577447485917, 1, 1], 
     Hue[0.4725364497353812, 1, 1], 
     Hue[0.4725364497353812, 1, 1], 
     Hue[0.4316938597089599, 1, 1], 
     Hue[0.4316938597089599, 1, 1], 
     Hue[0.429187722152806, 1, 1], 
     Hue[0.429187722152806, 1, 1], 
     Hue[0.4266815845966521, 1, 1], 
     Hue[0.4266815845966521, 1, 1], 
     Hue[0.42166930948434495`, 1, 1], 
     Hue[0.42166930948434495`, 1, 1], 
     Hue[0.41164475925973004`, 1, 1], 
     Hue[0.41164475925973004`, 1, 1], 
     Hue[0.39159565881050024`, 1, 1], 
     Hue[0.39159565881050024`, 1, 1], 
     Hue[0.35149745791204057`, 1, 1], 
     Hue[0.35149745791204057`, 1, 1], 
     Hue[0.3489913203558867, 1, 1], 
     Hue[0.3489913203558867, 1, 1], 
     Hue[0.34648518279973284`, 1, 1], 
     Hue[0.34648518279973284`, 1, 1], 
     Hue[0.34147290768742566`, 1, 1], 
     Hue[0.34147290768742566`, 1, 1], 
     Hue[0.33144835746281076`, 1, 1], 
     Hue[0.33144835746281076`, 1, 1], 
     Hue[0.3113992570135809, 1, 1], 
     Hue[0.3113992570135809, 1, 1], 
     Hue[0.27130105611512123`, 1, 1], 
     Hue[0.27130105611512123`, 1, 1], 
     Hue[0.26896324513183134`, 1, 1], 
     Hue[0.26896324513183134`, 1, 1], 
     Hue[0.26662543414854145`, 1, 1], 
     Hue[0.26662543414854145`, 1, 1], 
     Hue[0.26194981218196167`, 1, 1], 
     Hue[0.26194981218196167`, 1, 1], 
     Hue[0.25259856824880217`, 1, 1], 
     Hue[0.25259856824880217`, 1, 1], 
     Hue[0.23389608038248313`, 1, 1], 
     Hue[0.23389608038248313`, 1, 1], 
     Hue[0.19649110464984562`, 1, 1], 
     Hue[0.19649110464984562`, 1, 1], 
     Hue[0.19415329366655634`, 1, 1], 
     Hue[0.19415329366655634`, 1, 1], 
     Hue[0.19181548268326584`, 1, 1], 
     Hue[0.19181548268326584`, 1, 1], 
     Hue[0.18713986071668665`, 1, 1], 
     Hue[0.18713986071668665`, 1, 1], 
     Hue[0.17778861678352656`, 1, 1], 
     Hue[0.17778861678352656`, 1, 1], 
     Hue[0.1590861289172081, 1, 1], 
     Hue[0.1590861289172081, 1, 1], 
     Hue[0.12168115318456944`, 1, 1], 
     Hue[0.12168115318456944`, 1, 1], 
     Hue[0.11914626201704458`, 1, 1], 
     Hue[0.11914626201704458`, 1, 1], 
     Hue[0.11661137084952025`, 1, 1], 
     Hue[0.11661137084952025`, 1, 1], 
     Hue[0.11154158851447046`, 1, 1], 
     Hue[0.11154158851447046`, 1, 1], 
     Hue[0.10140202384437148`, 1, 1], 
     Hue[0.10140202384437148`, 1, 1], 
     Hue[0.0811228945041741, 1, 1], 
     Hue[0.0811228945041741, 1, 1], 
     Hue[0.0405646358237776, 1, 1], 
     Hue[0.0405646358237776, 1, 1], 
     Hue[0.038029744656252716`, 1, 1], 
     Hue[0.038029744656252716`, 1, 1], 
     Hue[0.035494853488728405`, 1, 1], 
     Hue[0.035494853488728405`, 1, 1], 
     Hue[0.03042507115367863, 1, 1], 
     Hue[0.03042507115367863, 1, 1], 
     Hue[0.020285506483580216`, 1, 1], 
     Hue[0.020285506483580216`, 1, 1], 
     Hue[6.377143381686833*^-6, 1, 1], 
     Hue[6.377143381686833*^-6, 1, 1], 
     Hue[0.9594481184629863, 1, 1], 
     Hue[0.9594481184629863, 1, 1], 
     Hue[0.9570815538683255, 1, 1], 
     Hue[0.9570815538683255, 1, 1], 
     Hue[0.9547149892736645, 1, 1], 
     Hue[0.9547149892736645, 1, 1], 
     Hue[0.9499818600843427, 1, 1], 
     Hue[0.9499818600843427, 1, 1], 
     Hue[0.9405156017056991, 1, 1], 
     Hue[0.9405156017056991, 1, 1], 
     Hue[0.921583084948412, 1, 1], 
     Hue[0.921583084948412, 1, 1], 
     Hue[0.8837180514338376, 1, 1], 
     Hue[0.8837180514338376, 1, 1], 
     Hue[0.8813514868391773, 1, 1], 
     Hue[0.8813514868391773, 1, 1], 
     Hue[0.8789849222445164, 1, 1], 
     Hue[0.8789849222445164, 1, 1], 
     Hue[0.8742517930551946, 1, 1], 
     Hue[0.8742517930551946, 1, 1], 
     Hue[0.8647855346765511, 1, 1], 
     Hue[0.8647855346765511, 1, 1], 
     Hue[0.8458530179192638, 1, 1], 
     Hue[0.8458530179192638, 1, 1], 
     Hue[0.8079879844046894, 1, 1], 
     Hue[0.8079879844046894, 1, 1], 
     Hue[0.8056679441305262, 1, 1], 
     Hue[0.8056679441305262, 1, 1], 
     Hue[0.8033479038563625, 1, 1], 
     Hue[0.8033479038563625, 1, 1], 
     Hue[0.7987078233080366, 1, 1], 
     Hue[0.7987078233080366, 1, 1], 
     Hue[0.7894276622113829, 1, 1], 
     Hue[0.7894276622113829, 1, 1], 
     Hue[0.7708673400180771, 1, 1], 
     Hue[0.7708673400180771, 1, 1], 
     Hue[0.7685472997439133, 1, 1], 
     Hue[0.7685472997439133, 1, 1], 
     Hue[0.76622725946975, 1, 1], 
     Hue[0.76622725946975, 1, 1], 
     Hue[0.7615871789214236, 1, 1], 
     Hue[0.7615871789214236, 1, 1], 
     Hue[0.7523070178247706, 1, 1], 
     Hue[0.7523070178247706, 1, 1], 
     Hue[0.7337466956314647, 1, 1], 
     Hue[0.7337466956314647, 1, 1], 
     Hue[0.7314266553573009, 1, 1], 
     Hue[0.7314266553573009, 1, 1], 
     Hue[0.7291066150831382, 1, 1], 
     Hue[0.7291066150831382, 1, 1], 
     Hue[0.7244665345348111, 1, 1], 
     Hue[0.7244665345348111, 1, 1], 
     Hue[0.7151863734381587, 1, 1], 
     Hue[0.7151863734381587, 1, 1], 
     Hue[0.6966260512448523, 1, 1], 
     Hue[0.6966260512448523, 1, 1], 
     Hue[0.65950540685824, 1, 1], 
     Hue[0.65950540685824, 1, 1], 
     Hue[0.6569882863998417, 1, 1], 
     Hue[0.6569882863998417, 1, 1], 
     Hue[0.6544711659414435, 1, 1], 
     Hue[0.6544711659414435, 1, 1], 
     Hue[0.6494369250246474, 1, 1], 
     Hue[0.6494369250246474, 1, 1], 
     Hue[0.6393684431910545, 1, 1], 
     Hue[0.6393684431910545, 1, 1], 
     Hue[0.6192314795238697, 1, 1], 
     Hue[0.6192314795238697, 1, 1], 
     Hue[0.616714359065472, 1, 1], 
     Hue[0.616714359065472, 1, 1], 
     Hue[0.6141972386070738, 1, 1], 
     Hue[0.6141972386070738, 1, 1], 
     Hue[0.6091629976902777, 1, 1], 
     Hue[0.6091629976902777, 1, 1], 
     Hue[0.5990945158566849, 1, 1], 
     Hue[0.5990945158566849, 1, 1], 
     Hue[0.5789575521894994, 1, 1], 
     Hue[0.5789575521894994, 1, 1], 
     Hue[0.5764404317311012, 1, 1], 
     Hue[0.5764404317311012, 1, 1], 
     Hue[0.5739233112727036, 1, 1], 
     Hue[0.5739233112727036, 1, 1], 
     Hue[0.5688890703559071, 1, 1], 
     Hue[0.5688890703559071, 1, 1], 
     Hue[0.5588205885223141, 1, 1], 
     Hue[0.5588205885223141, 1, 1], 
     Hue[0.5386836248551292, 1, 1], 
     Hue[0.5386836248551292, 1, 1], 
     Hue[0.4984096975207585, 1, 1], 
     Hue[0.4984096975207585, 1, 1], 
     Hue[0.49606090363522365`, 1, 1], 
     Hue[0.49606090363522365`, 1, 1], 
     Hue[0.49371210974968993`, 1, 1], 
     Hue[0.49371210974968993`, 1, 1], 
     Hue[0.48901452197862144`, 1, 1], 
     Hue[0.48901452197862144`, 1, 1], 
     Hue[0.4796193464364844, 1, 1], 
     Hue[0.4796193464364844, 1, 1], 
     Hue[0.4608289953522104, 1, 1], 
     Hue[0.4608289953522104, 1, 1], 
     Hue[0.45848020146667556`, 1, 1], 
     Hue[0.45848020146667556`, 1, 1], 
     Hue[0.4561314075811413, 1, 1], 
     Hue[0.4561314075811413, 1, 1], 
     Hue[0.4514338198100728, 1, 1], 
     Hue[0.4514338198100728, 1, 1], 
     Hue[0.4420386442679352, 1, 1], 
     Hue[0.4420386442679352, 1, 1], 
     Hue[0.42324829318366114`, 1, 1], 
     Hue[0.42324829318366114`, 1, 1], 
     Hue[0.4208994992981274, 1, 1], 
     Hue[0.4208994992981274, 1, 1], 
     Hue[0.41855070541259265`, 1, 1], 
     Hue[0.41855070541259265`, 1, 1], 
     Hue[0.4138531176415241, 1, 1], 
     Hue[0.4138531176415241, 1, 1], 
     Hue[0.40445794209938707`, 1, 1], 
     Hue[0.40445794209938707`, 1, 1], 
     Hue[0.3856675910151125, 1, 1], 
     Hue[0.3856675910151125, 1, 1], 
     Hue[0.3480868888465638, 1, 1], 
     Hue[0.3480868888465638, 1, 1], 
     Hue[0.3455410147767951, 1, 1], 
     Hue[0.3455410147767951, 1, 1], 
     Hue[0.34299514070702586`, 1, 1], 
     Hue[0.34299514070702586`, 1, 1], 
     Hue[0.3379033925674879, 1, 1], 
     Hue[0.3379033925674879, 1, 1], 
     Hue[0.32771989628841086`, 1, 1], 
     Hue[0.32771989628841086`, 1, 1], 
     Hue[0.30735290373025737`, 1, 1], 
     Hue[0.30735290373025737`, 1, 1], 
     Hue[0.30480702966048806`, 1, 1], 
     Hue[0.30480702966048806`, 1, 1], 
     Hue[0.30226115559071937`, 1, 1], 
     Hue[0.30226115559071937`, 1, 1], 
     Hue[0.2971694074511808, 1, 1], 
     Hue[0.2971694074511808, 1, 1], 
     Hue[0.28698591117210376`, 1, 1], 
     Hue[0.28698591117210376`, 1, 1], 
     Hue[0.2666189186139508, 1, 1], 
     Hue[0.2666189186139508, 1, 1], 
     Hue[0.26407304454418157`, 1, 1], 
     Hue[0.26407304454418157`, 1, 1], 
     Hue[0.2615271704744123, 1, 1], 
     Hue[0.2615271704744123, 1, 1], 
     Hue[0.25643542233487376`, 1, 1], 
     Hue[0.25643542233487376`, 1, 1], 
     Hue[0.2462519260557973, 1, 1], 
     Hue[0.2462519260557973, 1, 1], 
     Hue[0.22588493349764377`, 1, 1], 
     Hue[0.22588493349764377`, 1, 1], 
     Hue[0.18515094838133725`, 1, 1], 
     Hue[0.18515094838133725`, 1, 1], 
     Hue[0.185074863589463, 1, 1], 
     Hue[0.185074863589463, 1, 1], 
     Hue[0.18499877879758939`, 1, 1], 
     Hue[0.18499877879758939`, 1, 1], 
     Hue[0.18484660921384086`, 1, 1], 
     Hue[0.18484660921384086`, 1, 1], 
     Hue[0.1845422700463445, 1, 1], 
     Hue[0.1845422700463445, 1, 1], 
     Hue[0.18393359171135223`, 1, 1], 
     Hue[0.18393359171135223`, 1, 1], 
     Hue[0.18271623504136603`, 1, 1], 
     Hue[0.18271623504136603`, 1, 1], 
     Hue[0.18264015024949232`, 1, 1], 
     Hue[0.18264015024949232`, 1, 1], 
     Hue[0.18256406545761808`, 1, 1], 
     Hue[0.18256406545761808`, 1, 1], 
     Hue[0.18241189587386958`, 1, 1], 
     Hue[0.18241189587386958`, 1, 1], 
     Hue[0.18210755670637377`, 1, 1], 
     Hue[0.18210755670637377`, 1, 1], 
     Hue[0.1814988783713804, 1, 1], 
     Hue[0.1814988783713804, 1, 1], 
     Hue[0.1814227935795061, 1, 1], 
     Hue[0.1814227935795061, 1, 1], 
     Hue[0.18134670878763243`, 1, 1], 
     Hue[0.18134670878763243`, 1, 1], 
     Hue[0.18119453920388395`, 1, 1], 
     Hue[0.18119453920388395`, 1, 1], 
     Hue[0.18089020003638753`, 1, 1], 
     Hue[0.18089020003638753`, 1, 1], 
     Hue[0.1808141152445139, 1, 1], 
     Hue[0.1808141152445139, 1, 1], 
     Hue[0.18073803045263961`, 1, 1], 
     Hue[0.18073803045263961`, 1, 1], 
     Hue[0.18058586086889114`, 1, 1], 
     Hue[0.18058586086889114`, 1, 1], 
     Hue[0.18050977607701693`, 1, 1], 
     Hue[0.18050977607701693`, 1, 1], 
     Hue[0.18043369128514264`, 1, 1], 
     Hue[0.18043369128514264`, 1, 1], 
     Hue[0.18035760649326896`, 1, 1], 
     Hue[0.18035760649326896`, 1, 1], 
     Hue[0.18028152170139528`, 1, 1]}],
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  Epilog->LineBox[CompressedData["
1:eJw12nc8VX/8B3DjWtd1L0pSKiMpyXaFeB8llbIje6UtFBllk4yvHUmSUSFC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    "]],
  PlotRange->{{-3.999999836734694, 3.999999836734694}, {0., 
   0.9999998104453803}},
  PlotRangeClipping->True,
  PlotRangePadding->{
    Scaled[0.02], 
    Scaled[0.02]}]], "Output",
 CellChangeTimes->{3.391857436265625*^9},
 CellLabel->"Out[34]=",
 ImageCache->GraphicsData["CompressedBitmap", "\<\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\
\>"]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"QArgColorPlot", "[", 
  RowBox[{
   SuperscriptBox["\[ExponentialE]", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "\[ImaginaryI]"}], " ", "6", " ", "x"}], "-", 
     FractionBox[
      SuperscriptBox["x", "2"], "2"]}]], ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", 
     RowBox[{"-", "4"}], ",", "4"}], "}"}], ",", 
   RowBox[{"PlotStyle", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"Thickness", "[", "0.025", "]"}], ",", 
      RowBox[{"GrayLevel", "[", "0.5", "]"}]}], "}"}]}], ",", 
   RowBox[{"QSaturation", "\[Rule]", "0.5"}], ",", 
   RowBox[{"QBrightness", "\[Rule]", "0.7"}], ",", 
   RowBox[{"PlotRange", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "0.2"}], ",", "1.2"}], "}"}]}], ",", 
   RowBox[{"Frame", "\[Rule]", "True"}], ",", 
   RowBox[{"Axes", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"True", ",", "False"}], "}"}]}]}], "]"}]], "Input",
 CellLabel->"In[35]:="],

Cell[BoxData[
 GraphicsBox[GraphicsComplexBox[CompressedData["
1:eJx13GdcjQ/cx/H2nnaEbBLSRPK7IiHaSkWp7JUoFYokSf1VIkJoGKVUFKV1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   "], {
    {GrayLevel[0.5], EdgeForm[None], 
     GraphicsGroupBox[PolygonBox[CompressedData["
1:eJwtmXncldPXh/e+z7lPgzRLkjSpVFKpNBoyNdKgUqk0kaRCEpKQSmmQIkmS
kJCQTAlJQpKKMhWhQYMGTSq91/f9/v64uD5r38/a6957rfM4nnK9BrUbmIQQ
6vGPLP+OUB9vAIrngcuRKyAvXg6G4XdD+eBnGvKPRpAJznERfjGkwbHGeJP/
5VeOK/GrIB+eH1rgLeEUvBBci3eAwsHPNMObJ35Wz7TCW0OBYK7Gr4FTg2mD
t4WCwbTD2yfOrZouwS+FXDBN8csSv6v27Ih3giLBXId3hqL4GdAPvxlK4afB
DXhPKIEXh+54j8RrivXCe8PpeEm4Eb8pcS7l6I/fAmcGn+k9+L1QITg2AL8V
SuNnwxD8TigbXFMXvCsUC+Z6vFviWvQzuuBBcFYwg/HboEwwt+N3JM6tnEPx
uxLftWrug/dNXLtqGo7fBxXxRvAq/ho0xs+BkfgDUAmvCuPw8VAtOPYg/hBU
xqvAaHwMnBscG4U/nHitDszCn4W6wTkexSdAdfw8mIw/BjWCmYI/DucHMxWf
BjWDeQJ/EmoFMx1/CmoHMwN/Gi4Irmks/kjid1FsJv5M4tpU02z8Oain3oZl
+Kfq8eAzWYC/rhnAL4S5+AtQP5gX8Zc0c8HMw1+GhsHMx19JfNbaYw7+fOJc
uoMR+P2Jz157LMTfgIs0u7AU/0gzHBx7E39LM6pZgMX4O5qJYN7F34OmwbyP
f6AZCWYJ/mHi3MqxCH87cS7t8TH+SeKz0Bksxz+DZro7zSsfCtfABDmsZW2d
ZlizpZlkLQ/cjDeHlax9AS3wlrAK/xpaBce+xL9KvKYc6/HvoA3eFjbiP0A7
zTZswX/XjAc/8z2+IfGzeuZH/CdoH8zP+C9wbTCb8M3QIZhf8d8S51ZNq/Fv
oHUwa/BvE7+r9vwD/xOuC2Yrvg064zfAIfww9NQswx78b8003hV24bsTrym2
F98H3fEe8A9+MHEu5TiCH4VewWeal7PNB/2DY/+ydgx64zdCwloGbgquaTtr
O6BLMH/hOxPXop85jp+APsH8h5+EvsHol0DMOLdyZvE047tWDfnxU+CWYArg
p8IA/EHNAN4IHsJvg5L4GXB78DMF8UJwazCF8SIwMJiieDEYFExx/DQYjN8L
1fHzYHhwrAR+esZ7aY9S+JlwB343VMHPhXuCY6Xxs2AIfieUxcvB0GDK4xXg
rmAq4ufAsGAq4ZUzzq0cZfCzM86lz7iJnOWkxJ912rMqa9Uyrl0ztYK1zxPP
is6oMWtNYBQ+AmrjF8D9wdTB68LIYOrhF8IDwdTHG2R89vdBTbxWxrl0RjXw
8zNe0x4X4RfDw/h4zQDeOuNZV+wS/FIYjY+By/ErYGwwV+JXwSPBNMObw7hg
WuAtM86tHE3xyzLOVZf3ns0v8edy/m8Yfaa0Ya0tTMQnaWbxDjA5mI54J3gM
n6aZw3vCE8Gx6/DOMAV/XDOId4OpwbEueNeM116Hqfg0WBicoxfeG57Ep2um
NEvwVDD99NkGM4Lpr96Hp4MZoF6GmcEMVO/CM8EMVm/CrOCauuM9Mn4XxW5X
r8KzwQxRL8HsYIaqF+E5/Hndqe4S5uJz1GPqrYzXFBuh3oEX8Bc1k5pFeCk4
NlK9k/GaYqPUCzAv+Az64H0zPgvFRuvu4GV8gc5UZ5nxWaqmYZqFjGuZrx7Q
3cMrwTyKT4BXg5mIT4LXgpmsu804t/YYq97KONf/35HuBt4IZrruBt4MZobu
At4KZqbOHhYFM0tnC28HM1tnCYuDmaOzg3eCmauzg3eDeVFnB+/hnygmh2XB
vI9/AJ/iH+gd9W6wJJgFOiv4MJiFehdYGsybqh0+CmaRaoWPg1ms2jLeWzXM
013A+/g6+Bn/BdYH17BEe8FyfBWsxr+Br/HPlFO5YEUwy1Q7fB7McvwzWBnM
Cq3BF8GslMOX4X/gX4ngGjbhm+G74BqW6t0y3ls1rMG/hdXBrMXXqcZg1utn
YU0w3+Mb9DPBbMR/gLXB/Ij/lPFZqIZVeteM312/Y/fzOXMg8e9a1fQra7/B
93jkC1R9vug0gATfQOwP1v6EjTpb2IP/rTMOjm3Ft6kG7Q1/4TtVQ3BsO74j
47VDUJ7cFeBwcI69rO2DTfhm+Ac/CL8Gcwg/rBqDOYIfhS3B/Isfg9+DOY6f
gD+C+Q8/qXcIrmkXvjvjd1FMX+wibA0mwTOwLZgsnsJ2vRvkx0+BnfgOyIvn
y3pNsQL4qbAL3w2F8SKwJzhWEC+U9ZpiRfFi8HfwGeyntgMZn4VixVk7Dfbi
B6EsXi7rs1RNOTxP1rXsg5L4GbA/mFL4mXAgmNL4WfBPMGXws7POrT1K4Kdn
nUt3VBE/B44EUwmvDEeDqYKfC/8GUxWvBseCqY6fB8eDqYGfDyeCqYnXgv+C
qY1fACeDqYPX/d+X+7xwJX6VzjyaZnhzyB/dsw3xRrrDaBrjTSAbzUX4xbrT
aC7BL4VcNE3xy3Sm0VyOX5H13qqhHn5h1rOi39HtuKv2Gf+uVg0tWGupHsGL
wfV4NyiOF4Cr8WvUI9G0wdtCwWja4e3VI9Fci3eAwtF0xDupp6K5Du8MRaNn
eItmIePZVQ2tWGud9d6qoTveQz0VzQ14TygRTS+8t3ogmj54XygZzY34Teqx
aPrhN0Op6Bq64F2zfnfF+uO3qAejGYDfCqWjGYgPUk/iFWA4fh9UxJvAQvwN
uAgvA7fjd6hn8bIwFL9LMxEdG4LfmfWacozA71cP45XgQfwh9TBeDR7FJ0D1
6GdG4g9k/ayeGYU/DFWiGY2PUc9HMxZ/BKpGMw4fn3Vu1TQMvxvKR3MPfm/W
76o9J+KTNCPRTMYfgxp4XZiNPwf18FowHX8KauM14Qn8yazXFJuBP60ZwuvA
LPzZrHMpxxz8efVw9Jm+ib+lmYiOzcVfgPp4I3gVfw0aR9c0BX9cMxzNVHxa
1rXoZ17EX4IG0czDX4aG0czHX8k6t3IuwF/P+q5VwyL8bbgkmsX4O5pRvBP8
gf8J1+mzAJbhn+ozIfqZd/H3oGk07+MfaKajWYJ/CJdHsxT/SDOuWdQM4Rs0
k9Gxj/FPst5LeyzHP4NmmmVYi6/TTEfHVuCfQ3N9FsCX+FfQMppV+NfQKprV
+Dea0WjW4N9mnVs5VuJfZJ1LdzoTfybru9We6/Hvsq79JfVrSr/CvOgz2sra
NuiMt4ef8V/g2mg24ZuhQzS/4r9Bx2i24L9nffbt4Ef8p6xz6Yw24j9kvaY9
tuM7oAt+AxzCD0PP6Nhf+E7oil8Pe/C/oVs0e/F90D2a/fgB6BHNP/jBrHMr
xy58d9a59M6jefcx8HL0nkdYOwq98N5wHD8BffC+oP+pGuFGvD/kx0+BW6Jj
CZ6Bm/B+kMPzwM3RsSyepl4bpZnCL9YdROcogJ8KA/BboTBeBAZGUxQvBoOi
KY6fBoOjKYGfDrdFUxI/A26PphR+JtwRXVNePF/qd1GsNH4WDImmDH423BlN
WbwcDMWHQSW8MtyN3wUV8XNSrylWBT8X7sHvher4eTA8OlYVr5Z6TbEa+Plw
X/QZFMQLpT4LxWritWAE/hA0xpukPkvVVB6vkLqW+6EOXhdGRlMPvxAeiKY+
3gAejKYh3ih1bu1RG78gdS71wH/0wsmse0F3dglrl8JofAxcjl8BY/EJ0AZv
CxOjY1fiV8Ej0TTDm8O4aFrgLWF8NK3w1vBo9B5N8ctS76XY1fg1qffSHu3w
9jAJn6aZwnvCE9Gxa/EOMBl/TDOPd4Yp0XTBu8Lj0VyPd4Op0XTHe6TOrRwd
8U6pc2lm/uVsjmU9O9qzF2u94Ul8umZEswFP4U/rjnW3MBOfoZnSLKVeU2yg
eh2ewWeph9W78Gx0bLB6PfWaYkPUqzA7es8+eN/Ueys2VL0Bz+Ev6s5116k/
C1VTP81q6lrmqEfVm/B8NMPVmzA3mhHqDXghmpHqrdS5tccwzUbqXPrOOV/f
7TP+rqvPnLHqBZiP59H/Z87RP5AXX6gz1NnBG/iXAv9K4K+oB3T38Cr+mu5E
dwELomMT1Qup15Rjus4e3sTf0hnrbGER/q7ORGcB70U/M0N3kfpZPTNLZw9v
RzNbZwmLo5mjs4J3opmrs0mdWzVNUW/B69FMVS+lflftOQ9/Gd6PZj7+CnyA
L1NMDp/iS1WjaoOP8A9hoc4q9Zpii1QrfIx/onfUu6XOpRxL8A9hefSZrsK/
hlXRsaXKDZ/hK2EF/jl8EV3TqzpbWBLNAr1b6lr0Mx/jn8CKaJapduWIZrly
p86tnCuVO/Vdq4bV+DeqKZo1+LewGv8D/sNPwp/4OvgZ/wXWRz+zFl+nHNGs
x7+DNdF8j29Qzmg24j/AWvwX2Ivvg03RsR/xn1LvpT024ZuVE/8JduG74efo
2K/4b/A9vkE1q1bYGM1WfJv2jGY7vgN+jOYvfGfq3MqxBf89dS7d6WL1Wuq7
1Z578L9T1/43FGeWToO90WekP4JG2Kra4BDPHlaN0RzBj8KWaP7Fj2nPaI7j
J1Kf/Wb4Bz+YOpfOaD9+IPWa9kjYKwPb8N1QGC8Ce6JjWTyF7fgOyIvng7+i
yY+fAjujKYCfCruiKYgXyjm3cuTwPDnn0mfKOGoZn/qzQnsWZa1Yzmej71CD
+ay+LevvSoehImvnwBG8iP5ug3eGovh+YqXwM+GA3h3K4GfDwehYafysnNeU
oxJeGY7qLKEqXg2O4SehDl5Xf5hO/EwV/Nycn9Uz1fHz4Hg0NfDz4UQ0NfFa
8F80tfELcs6tmsri5eBQNOXxCjm/q/ash1+onkhMfbwBJHg+/R0Pbw758VR/
98MvhRye1d8d8YtzXlOsKX5Zzp/l+gy/Er8q51zK0QJvqTtNfKZd8K66k8Sx
VnhrKIAX0t/h8A5QOHFNDfFG6qnENMab5FyLfuZq/Br1SGLa4G2hYGLa4e1z
zq2cHfFOOd+1arge7wbFE9Md76EZwqvBo/gEqI6Xgv74LeqJxM/cgPeEEonp
hfeG0xPTB+8LJRNzI34TnIFXhBH4/erBxLF++M0576U9BuC3Qmm8PNyD36s7
TRwbiA9SD+Jl4Hb8DvVoYobgd0LZxAzF71KPJGYYfnfOuZVjMH5bzrn0GVJC
7wL7ovccjt+Xc+2KldS75DwrOqOJ+CT1MF4ZRuEPQ5XEjMbHqOcTMxZ/BKom
Zhw+PuezrwQP4g/lnEtnNBJ/IOc17TEZfwxq4HVgFv5szn87UmwK/rhmCK8J
T+BPaoYSMx1/CmonZgb+tGYqMTPxZ3LOrRxT8Wk559LfpObgz2um8P8DLzRM
dQ==
       "]]]}, {{}, {}, 
     {Hue[0.8197184782986047, 0.5, 0.7], Thickness[0.025], LineBox[{1, 2}], 
      {Hue[0.817375481892275, 0.5, 0.7], LineBox[{2, 3}]}, 
      {Hue[0.8150324854859454, 0.5, 0.7], LineBox[{3, 4}]}, 
      {Hue[0.8103464926732862, 0.5, 0.7], LineBox[{4, 5}]}, 
      {Hue[0.8009745070479671, 0.5, 0.7], LineBox[{5, 6}]}, 
      {Hue[0.7822305357973296, 0.5, 0.7], LineBox[{6, 7}]}, 
      {Hue[0.779887539391, 0.5, 0.7], LineBox[{7, 8}]}, 
      {Hue[0.7775445429846704, 0.5, 0.7], LineBox[{8, 9}]}, 
      {Hue[0.7728585501720105, 0.5, 0.7], LineBox[{9, 10}]}, 
      {Hue[0.763486564546692, 0.5, 0.7], LineBox[{10, 11}]}, 
      {Hue[0.7447425932960545, 0.5, 0.7], LineBox[{11, 12}]}, 
      {Hue[0.7423995968897242, 0.5, 0.7], LineBox[{12, 13}]}, 
      {Hue[0.7400566004833946, 0.5, 0.7], LineBox[{13, 14}]}, 
      {Hue[0.7353706076707348, 0.5, 0.7], LineBox[{14, 15}]}, 
      {Hue[0.7259986220454163, 0.5, 0.7], LineBox[{15, 16}]}, 
      {Hue[0.7072546507947788, 0.5, 0.7], LineBox[{16, 17}]}, 
      {Hue[0.6697667082935036, 0.5, 0.7], LineBox[{17, 18}]}, 
      {Hue[0.6672266317029389, 0.5, 0.7], LineBox[{18, 19}]}, 
      {Hue[0.6646865551123748, 0.5, 0.7], LineBox[{19, 20}]}, 
      {Hue[0.6596064019312451, 0.5, 0.7], LineBox[{20, 21}]}, 
      {Hue[0.6494460955689871, 0.5, 0.7], LineBox[{21, 22}]}, 
      {Hue[0.6291254828444701, 0.5, 0.7], LineBox[{22, 23}]}, 
      {Hue[0.626585406253906, 0.5, 0.7], LineBox[{23, 24}]}, 
      {Hue[0.6240453296633414, 0.5, 0.7], LineBox[{24, 25}]}, 
      {Hue[0.6189651764822122, 0.5, 0.7], LineBox[{25, 26}]}, 
      {Hue[0.6088048701199542, 0.5, 0.7], LineBox[{26, 27}]}, 
      {Hue[0.5884842573954372, 0.5, 0.7], LineBox[{27, 28}]}, 
      {Hue[0.5859441808048731, 0.5, 0.7], LineBox[{28, 29}]}, 
      {Hue[0.5834041042143084, 0.5, 0.7], LineBox[{29, 30}]}, 
      {Hue[0.5783239510331792, 0.5, 0.7], LineBox[{30, 31}]}, 
      {Hue[0.5681636446709212, 0.5, 0.7], LineBox[{31, 32}]}, 
      {Hue[0.5478430319464042, 0.5, 0.7], LineBox[{32, 33}]}, 
      {Hue[0.5072018064973707, 0.5, 0.7], LineBox[{33, 34}]}, 
      {Hue[0.5048300564796699, 0.5, 0.7], LineBox[{34, 35}]}, 
      {Hue[0.5024583064619688, 0.5, 0.7], LineBox[{35, 36}]}, 
      {Hue[0.4977148064265675, 0.5, 0.7], LineBox[{36, 37}]}, 
      {Hue[0.48822780635576496`, 0.5, 0.7], LineBox[{37, 38}]}, 
      {Hue[0.4692538062141587, 0.5, 0.7], LineBox[{38, 39}]}, 
      {Hue[0.466882056196458, 0.5, 0.7], LineBox[{39, 40}]}, 
      {Hue[0.4645103061787574, 0.5, 0.7], LineBox[{40, 41}]}, 
      {Hue[0.45976680614335613`, 0.5, 0.7], LineBox[{41, 42}]}, 
      {Hue[0.45027980607255297`, 0.5, 0.7], LineBox[{42, 43}]}, 
      {Hue[0.43130580593094725`, 0.5, 0.7], LineBox[{43, 44}]}, 
      {Hue[0.4289340559132466, 0.5, 0.7], LineBox[{44, 45}]}, 
      {Hue[0.426562305895546, 0.5, 0.7], LineBox[{45, 46}]}, 
      {Hue[0.42181880586014414`, 0.5, 0.7], LineBox[{46, 47}]}, 
      {Hue[0.4123318057893416, 0.5, 0.7], LineBox[{47, 48}]}, 
      {Hue[0.39335780564773587`, 0.5, 0.7], LineBox[{48, 49}]}, 
      {Hue[0.3554098053645245, 0.5, 0.7], LineBox[{49, 50}]}, 
      {Hue[0.3530845796673215, 0.5, 0.7], LineBox[{50, 51}]}, 
      {Hue[0.35075935397011854`, 0.5, 0.7], LineBox[{51, 52}]}, 
      {Hue[0.3461089025757114, 0.5, 0.7], LineBox[{52, 53}]}, 
      {Hue[0.33680799978689946`, 0.5, 0.7], LineBox[{53, 54}]}, 
      {Hue[0.3182061942092745, 0.5, 0.7], LineBox[{54, 55}]}, 
      {Hue[0.3158809685120721, 0.5, 0.7], LineBox[{55, 56}]}, 
      {Hue[0.31355574281486853`, 0.5, 0.7], LineBox[{56, 57}]}, 
      {Hue[0.308905291420462, 0.5, 0.7], LineBox[{57, 58}]}, 
      {Hue[0.2996043886316495, 0.5, 0.7], LineBox[{58, 59}]}, 
      {Hue[0.2810025830540251, 0.5, 0.7], LineBox[{59, 60}]}, 
      {Hue[0.2786773573568221, 0.5, 0.7], LineBox[{60, 61}]}, 
      {Hue[0.2763521316596185, 0.5, 0.7], LineBox[{61, 62}]}, 
      {Hue[0.2717016802652126, 0.5, 0.7], LineBox[{62, 63}]}, 
      {Hue[0.26240077747640006`, 0.5, 0.7], LineBox[{63, 64}]}, 
      {Hue[0.24379897189877506`, 0.5, 0.7], LineBox[{64, 65}]}, 
      {Hue[0.2065953607435256, 0.5, 0.7], LineBox[{65, 66}]}, 
      {Hue[0.20407305486208763`, 0.5, 0.7], LineBox[{66, 67}]}, 
      {Hue[0.20155074898064965`, 0.5, 0.7], LineBox[{67, 68}]}, 
      {Hue[0.19650613721777366`, 0.5, 0.7], LineBox[{68, 69}]}, 
      {Hue[0.1864169136920217, 0.5, 0.7], LineBox[{69, 70}]}, 
      {Hue[0.16623846664051775`, 0.5, 0.7], LineBox[{70, 71}]}, 
      {Hue[0.16371616075907977`, 0.5, 0.7], LineBox[{71, 72}]}, 
      {Hue[0.1611938548776418, 0.5, 0.7], LineBox[{72, 73}]}, 
      {Hue[0.1561492431147658, 0.5, 0.7], LineBox[{73, 74}]}, 
      {Hue[0.14606001958901382`, 0.5, 0.7], LineBox[{74, 75}]}, 
      {Hue[0.12588157253750992`, 0.5, 0.7], LineBox[{75, 76}]}, 
      {Hue[0.12335926665607193`, 0.5, 0.7], LineBox[{76, 77}]}, 
      {Hue[0.12083696077463393`, 0.5, 0.7], LineBox[{77, 78}]}, 
      {Hue[0.11579234901175797`, 0.5, 0.7], LineBox[{78, 79}]}, 
      {Hue[0.105703125486006, 0.5, 0.7], LineBox[{79, 80}]}, 
      {Hue[0.0855246784345021, 0.5, 0.7], LineBox[{80, 81}]}, 
      {Hue[0.04516778433149423, 0.5, 0.7], LineBox[{81, 82}]}, 
      {Hue[0.04281380502292024, 0.5, 0.7], LineBox[{82, 83}]}, 
      {Hue[0.04045982571434625, 0.5, 0.7], LineBox[{83, 84}]}, 
      {Hue[0.035751867097198235`, 0.5, 0.7], LineBox[{84, 85}]}, 
      {Hue[0.026335949862901102`, 0.5, 0.7], LineBox[{85, 86}]}, 
      {Hue[0.007504115394307963, 0.5, 0.7], LineBox[{86, 87}]}, 
      {Hue[0.9698404464571223, 0.5, 0.7], LineBox[{87, 88}]}, 
      {Hue[0.9674864671485477, 0.5, 0.7], LineBox[{88, 89}]}, 
      {Hue[0.9651324878399743, 0.5, 0.7], LineBox[{89, 90}]}, 
      {Hue[0.9604245292228257, 0.5, 0.7], LineBox[{90, 91}]}, 
      {Hue[0.9510086119885292, 0.5, 0.7], LineBox[{91, 92}]}, 
      {Hue[0.9321767775199366, 0.5, 0.7], LineBox[{92, 93}]}, 
      {Hue[0.8945131085827502, 0.5, 0.7], LineBox[{93, 94}]}, 
      {Hue[0.8919620490899413, 0.5, 0.7], LineBox[{94, 95}]}, 
      {Hue[0.8894109895971323, 0.5, 0.7], LineBox[{95, 96}]}, 
      {Hue[0.8843088706115143, 0.5, 0.7], LineBox[{96, 97}]}, 
      {Hue[0.8741046326402782, 0.5, 0.7], LineBox[{97, 98}]}, 
      {Hue[0.8536961566978062, 0.5, 0.7], LineBox[{98, 99}]}, 
      {Hue[0.8128792048128621, 0.5, 0.7], LineBox[{99, 100}]}, 
      {Hue[0.810328145320053, 0.5, 0.7], LineBox[{100, 101}]}, 
      {Hue[0.8077770858272441, 0.5, 0.7], LineBox[{101, 102}]}, 
      {Hue[0.802674966841626, 0.5, 0.7], LineBox[{102, 103}]}, 
      {Hue[0.79247072887039, 0.5, 0.7], LineBox[{103, 104}]}, 
      {Hue[0.772062252927918, 0.5, 0.7], LineBox[{104, 105}]}, 
      {Hue[0.7312453010429738, 0.5, 0.7], LineBox[{105, 106}]}, 
      {Hue[0.7287407658706625, 0.5, 0.7], LineBox[{106, 107}]}, 
      {Hue[0.7262362306983512, 0.5, 0.7], LineBox[{107, 108}]}, 
      {Hue[0.7212271603537285, 0.5, 0.7], LineBox[{108, 109}]}, 
      {Hue[0.711209019664483, 0.5, 0.7], LineBox[{109, 110}]}, 
      {Hue[0.6911727382859922, 0.5, 0.7], LineBox[{110, 111}]}, 
      {Hue[0.6511001755290096, 0.5, 0.7], LineBox[{111, 112}]}, 
      {Hue[0.6485956403566981, 0.5, 0.7], LineBox[{112, 113}]}, 
      {Hue[0.646091105184387, 0.5, 0.7], LineBox[{113, 114}]}, 
      {Hue[0.6410820348397641, 0.5, 0.7], LineBox[{114, 115}]}, 
      {Hue[0.6310638941505182, 0.5, 0.7], LineBox[{115, 116}]}, 
      {Hue[0.6110276127720268, 0.5, 0.7], LineBox[{116, 117}]}, 
      {Hue[0.5709550500150447, 0.5, 0.7], LineBox[{117, 118}]}, 
      {Hue[0.5686188414155973, 0.5, 0.7], LineBox[{118, 119}]}, 
      {Hue[0.5662826328161494, 0.5, 0.7], LineBox[{119, 120}]}, 
      {Hue[0.5616102156172547, 0.5, 0.7], LineBox[{120, 121}]}, 
      {Hue[0.5522653812194647, 0.5, 0.7], LineBox[{121, 122}]}, 
      {Hue[0.5335757124238844, 0.5, 0.7], LineBox[{122, 123}]}, 
      {Hue[0.49619637483272355`, 0.5, 0.7], LineBox[{123, 124}]}, 
      {Hue[0.49386016623327594`, 0.5, 0.7], LineBox[{124, 125}]}, 
      {Hue[0.49152395763382856`, 0.5, 0.7], LineBox[{125, 126}]}, 
      {Hue[0.48685154043493384`, 0.5, 0.7], LineBox[{126, 127}]}, 
      {Hue[0.47750670603714357`, 0.5, 0.7], LineBox[{127, 128}]}, 
      {Hue[0.458817037241563, 0.5, 0.7], LineBox[{128, 129}]}, 
      {Hue[0.4214376996504027, 0.5, 0.7], LineBox[{129, 130}]}, 
      {Hue[0.4189044108667203, 0.5, 0.7], LineBox[{130, 131}]}, 
      {Hue[0.41637112208303795`, 0.5, 0.7], LineBox[{131, 132}]}, 
      {Hue[0.4113045445156732, 0.5, 0.7], LineBox[{132, 133}]}, 
      {Hue[0.4011713893809435, 0.5, 0.7], LineBox[{133, 134}]}, 
      {Hue[0.38090507911148397`, 0.5, 0.7], LineBox[{134, 135}]}, 
      {Hue[0.34037245857256526`, 0.5, 0.7], LineBox[{135, 136}]}, 
      {Hue[0.3378391697888829, 0.5, 0.7], LineBox[{136, 137}]}, 
      {Hue[0.33530588100520053`, 0.5, 0.7], LineBox[{137, 138}]}, 
      {Hue[0.3302393034378358, 0.5, 0.7], LineBox[{138, 139}]}, 
      {Hue[0.3201061483031061, 0.5, 0.7], LineBox[{139, 140}]}, 
      {Hue[0.2998398380336466, 0.5, 0.7], LineBox[{140, 141}]}, 
      {Hue[0.2593072174947282, 0.5, 0.7], LineBox[{141, 142}]}, 
      {Hue[0.2569422552839095, 0.5, 0.7], LineBox[{142, 143}]}, 
      {Hue[0.2545772930730911, 0.5, 0.7], LineBox[{143, 144}]}, 
      {Hue[0.24984736865145438`, 0.5, 0.7], LineBox[{144, 145}]}, 
      {Hue[0.24038751980818002`, 0.5, 0.7], LineBox[{145, 146}]}, 
      {Hue[0.22146782212163132`, 0.5, 0.7], LineBox[{146, 147}]}, 
      {Hue[0.18362842674853475`, 0.5, 0.7], LineBox[{147, 148}]}, 
      {Hue[0.18126346453771638`, 0.5, 0.7], LineBox[{148, 149}]}, 
      {Hue[0.17889850232689772`, 0.5, 0.7], LineBox[{149, 150}]}, 
      {Hue[0.17416857790526039`, 0.5, 0.7], LineBox[{150, 151}]}, 
      {Hue[0.16470872906198633`, 0.5, 0.7], LineBox[{151, 152}]}, 
      {Hue[0.14578903137543817`, 0.5, 0.7], LineBox[{152, 153}]}, 
      {Hue[0.10794963600234132`, 0.5, 0.7], LineBox[{153, 154}]}, 
      {Hue[0.10563119811202033`, 0.5, 0.7], LineBox[{154, 155}]}, 
      {Hue[0.10331276022169904`, 0.5, 0.7], LineBox[{155, 156}]}, 
      {Hue[0.09867588444105733, 0.5, 0.7], LineBox[{156, 157}]}, 
      {Hue[0.08940213287977387, 0.5, 0.7], LineBox[{157, 158}]}, 
      {Hue[0.07085462975720616, 0.5, 0.7], LineBox[{158, 159}]}, 
      {Hue[0.03375962351207098, 0.5, 0.7], LineBox[{159, 160}]}, 
      {Hue[0.03144118562175026, 0.5, 0.7], LineBox[{160, 161}]}, 
      {Hue[0.029122747731428977`, 0.5, 0.7], LineBox[{161, 162}]}, 
      {Hue[0.024485871950787247`, 0.5, 0.7], LineBox[{162, 163}]}, 
      {Hue[0.015212120389503814`, 0.5, 0.7], LineBox[{163, 164}]}, 
      {Hue[0.9966646172669361, 0.5, 0.7], LineBox[{164, 165}]}, 
      {Hue[0.9595696110218012, 0.5, 0.7], LineBox[{165, 166}]}, 
      {Hue[0.9570540929472452, 0.5, 0.7], LineBox[{166, 167}]}, 
      {Hue[0.9545385748726898, 0.5, 0.7], LineBox[{167, 168}]}, 
      {Hue[0.949507538723578, 0.5, 0.7], LineBox[{168, 169}]}, 
      {Hue[0.9394454664253545, 0.5, 0.7], LineBox[{169, 170}]}, 
      {Hue[0.9193213218289079, 0.5, 0.7], LineBox[{170, 171}]}, 
      {Hue[0.8790730326360152, 0.5, 0.7], LineBox[{171, 172}]}, 
      {Hue[0.7985764542502292, 0.5, 0.7], LineBox[{172, 173}]}, 
      {Hue[0.7962292627485371, 0.5, 0.7], LineBox[{173, 174}]}, 
      {Hue[0.7938820712468451, 0.5, 0.7], LineBox[{174, 175}]}, 
      {Hue[0.7891876882434611, 0.5, 0.7], LineBox[{175, 176}]}, 
      {Hue[0.7797989222366932, 0.5, 0.7], LineBox[{176, 177}]}, 
      {Hue[0.7610213902231577, 0.5, 0.7], LineBox[{177, 178}]}, 
      {Hue[0.7234663261960863, 0.5, 0.7], LineBox[{178, 179}]}, 
      {Hue[0.6483561981419439, 0.5, 0.7], LineBox[{179, 180}]}, 
      {Hue[0.6458119264560171, 0.5, 0.7], LineBox[{180, 181}]}, 
      {Hue[0.6432676547700904, 0.5, 0.7], LineBox[{181, 182}]}, 
      {Hue[0.6381791113982367, 0.5, 0.7], LineBox[{182, 183}]}, 
      {Hue[0.6280020246545293, 0.5, 0.7], LineBox[{183, 184}]}, 
      {Hue[0.6076478511671144, 0.5, 0.7], LineBox[{184, 185}]}, 
      {Hue[0.5669395041922853, 0.5, 0.7], LineBox[{185, 186}]}, 
      {Hue[0.4855228102426269, 0.5, 0.7], LineBox[{186, 187}]}, 
      {Hue[0.3256669788551567, 0.5, 0.7], LineBox[{187, 188}]}, 
      {Hue[0.17658404813097406`, 0.5, 0.7], LineBox[{188, 189}]}, 
      {Hue[0.014887985615758862`, 0.5, 0.7], LineBox[{189, 190}]}, 
      {Hue[0.8639648237638307, 0.5, 0.7], LineBox[{190, 191}]}, 
      {Hue[0.7004285301208703, 0.5, 0.7], LineBox[{191, 192}]}, 
      {Hue[0.6979197998531967, 0.5, 0.7], LineBox[{192, 193}]}, 
      {Hue[0.6954110695855231, 0.5, 0.7], LineBox[{193, 194}]}, 
      {Hue[0.6903936090501757, 0.5, 0.7], LineBox[{194, 195}]}, 
      {Hue[0.6803586879794812, 0.5, 0.7], LineBox[{195, 196}]}, 
      {Hue[0.660288845838092, 0.5, 0.7], LineBox[{196, 197}]}, 
      {Hue[0.6201491615553137, 0.5, 0.7], LineBox[{197, 198}]}, 
      {Hue[0.5398697929897572, 0.5, 0.7], LineBox[{198, 199}]}, 
      {Hue[0.5375293892949474, 0.5, 0.7], LineBox[{199, 200}]}, 
      {Hue[0.5351889856001377, 0.5, 0.7], LineBox[{200, 201}]}, 
      {Hue[0.5305081782105181, 0.5, 0.7], LineBox[{201, 202}]}, 
      {Hue[0.5211465634312791, 0.5, 0.7], LineBox[{202, 203}]}, 
      {Hue[0.5024233338728007, 0.5, 0.7], LineBox[{203, 204}]}, 
      {Hue[0.46497687475584426`, 0.5, 0.7], LineBox[{204, 205}]}, 
      {Hue[0.4626364710610345, 0.5, 0.7], LineBox[{205, 206}]}, 
      {Hue[0.46029606736622475`, 0.5, 0.7], LineBox[{206, 207}]}, 
      {Hue[0.45561525997660507`, 0.5, 0.7], LineBox[{207, 208}]}, 
      {Hue[0.446253645197366, 0.5, 0.7], LineBox[{208, 209}]}, 
      {Hue[0.4275304156388878, 0.5, 0.7], LineBox[{209, 210}]}, 
      {Hue[0.3900839565219313, 0.5, 0.7], LineBox[{210, 211}]}, 
      {Hue[0.38754647264288655`, 0.5, 0.7], LineBox[{211, 212}]}, 
      {Hue[0.38500898876384193`, 0.5, 0.7], LineBox[{212, 213}]}, 
      {Hue[0.37993402100575263`, 0.5, 0.7], LineBox[{213, 214}]}, 
      {Hue[0.3697840854895739, 0.5, 0.7], LineBox[{214, 215}]}, 
      {Hue[0.3494842144572167, 0.5, 0.7], LineBox[{215, 216}]}, 
      {Hue[0.34694673057817194`, 0.5, 0.7], LineBox[{216, 217}]}, 
      {Hue[0.34440924669912737`, 0.5, 0.7], LineBox[{217, 218}]}, 
      {Hue[0.339334278941038, 0.5, 0.7], LineBox[{218, 219}]}, 
      {Hue[0.32918434342485936`, 0.5, 0.7], LineBox[{219, 220}]}, 
      {Hue[0.3088844723925021, 0.5, 0.7], LineBox[{220, 221}]}, 
      {Hue[0.30634698851345743`, 0.5, 0.7], LineBox[{221, 222}]}, 
      {Hue[0.30380950463441275`, 0.5, 0.7], LineBox[{222, 223}]}, 
      {Hue[0.29873453687632345`, 0.5, 0.7], LineBox[{223, 224}]}, 
      {Hue[0.2885846013601448, 0.5, 0.7], LineBox[{224, 225}]}, 
      {Hue[0.2682847303277875, 0.5, 0.7], LineBox[{225, 226}]}, 
      {Hue[0.22768498826307296`, 0.5, 0.7], LineBox[{226, 227}]}, 
      {Hue[0.2253158309568922, 0.5, 0.7], LineBox[{227, 228}]}, 
      {Hue[0.22294667365071136`, 0.5, 0.7], LineBox[{228, 229}]}, 
      {Hue[0.21820835903834976`, 0.5, 0.7], LineBox[{229, 230}]}, 
      {Hue[0.20873172981362653`, 0.5, 0.7], LineBox[{230, 231}]}, 
      {Hue[0.18977847136418016`, 0.5, 0.7], LineBox[{231, 232}]}, 
      {Hue[0.18740931405799935`, 0.5, 0.7], LineBox[{232, 233}]}, 
      {Hue[0.18504015675181856`, 0.5, 0.7], LineBox[{233, 234}]}, 
      {Hue[0.18030184213945694`, 0.5, 0.7], LineBox[{234, 235}]}, 
      {Hue[0.17082521291473377`, 0.5, 0.7], LineBox[{235, 236}]}, 
      {Hue[0.1518719544652873, 0.5, 0.7], LineBox[{236, 237}]}, 
      {Hue[0.14950279715910653`, 0.5, 0.7], LineBox[{237, 238}]}, 
      {Hue[0.14713363985292574`, 0.5, 0.7], LineBox[{238, 239}]}, 
      {Hue[0.1423953252405641, 0.5, 0.7], LineBox[{239, 240}]}, 
      {Hue[0.1329186960158409, 0.5, 0.7], LineBox[{240, 241}]}, 
      {Hue[0.13054953870966016`, 0.5, 0.7], LineBox[{241, 242}]}, 
      {Hue[0.12818038140347934`, 0.5, 0.7], LineBox[{242, 243}]}, 
      {Hue[0.12344206679111772`, 0.5, 0.7], LineBox[{243, 244}]}, 
      {Hue[0.11396543756639452`, 0.5, 0.7], LineBox[{244, 245}]}, 
      {Hue[0.11159628026021372`, 0.5, 0.7], LineBox[{245, 246}]}, 
      {Hue[0.1092271229540329, 0.5, 0.7], LineBox[{246, 247}]}, 
      {Hue[0.10448880834167132`, 0.5, 0.7], LineBox[{247, 248}]}, 
      {Hue[0.09501217911694812, 0.5, 0.7], LineBox[{248, 249}]}, 
      {Hue[0.0926430218107673, 0.5, 0.7], LineBox[{249, 250}]}, 
      {Hue[0.09027386450458652, 0.5, 0.7], LineBox[{250, 251}]}, 
      {Hue[0.08553554989222491, 0.5, 0.7], LineBox[{251, 252}]}, 
      {Hue[0.07605892066750171, 0.5, 0.7], LineBox[{252, 253}]}, 
      {Hue[0.07373628768181852, 0.5, 0.7], LineBox[{253, 254}]}, 
      {Hue[0.07141365469613532, 0.5, 0.7], LineBox[{254, 255}]}, 
      {Hue[0.06676838872476895, 0.5, 0.7], LineBox[{255, 256}]}, 
      {Hue[0.06444575573908579, 0.5, 0.7], LineBox[{256, 257}]}, 
      {Hue[0.06212312275340258, 0.5, 0.7], LineBox[{257, 258}]}, 
      {Hue[0.05747785678203621, 0.5, 0.7], LineBox[{258, 259}]}, 
      {Hue[0.05515522379635303, 0.5, 0.7], LineBox[{259, 260}]}, 
      {Hue[0.052832590810669836`, 0.5, 0.7], LineBox[{260, 261}]}, 
      {Hue[0.04818732483930348, 0.5, 0.7], LineBox[{261, 262}]}, 
      {Hue[0.04586469185362028, 0.5, 0.7], LineBox[{262, 263}]}, 
      {Hue[0.043542058867937085`, 0.5, 0.7], LineBox[{263, 264}]}, 
      {Hue[0.03889679289657071, 0.5, 0.7], LineBox[{264, 265}]}, 
      {Hue[0.03657415991088753, 0.5, 0.7], LineBox[{265, 266}]}, 
      {Hue[0.03425152692520435, 0.5, 0.7], LineBox[{266, 267}]}, 
      {Hue[0.031928893939521155`, 0.5, 0.7], LineBox[{267, 268}]}, 
      {Hue[0.029606260953837966`, 0.5, 0.7], LineBox[{268, 269}]}, 
      {Hue[0.02728362796815478, 0.5, 0.7], LineBox[{269, 270}]}, 
      {Hue[0.024960994982471587`, 0.5, 0.7], LineBox[{270, 271}]}, 
      {Hue[0.0226383619967884, 0.5, 0.7], LineBox[{271, 272}]}, 
      {Hue[0.02031572901110522, 0.5, 0.7], LineBox[{272, 273}]}, 
      {Hue[0.017993096025422036`, 0.5, 0.7], LineBox[{273, 274}]}, 
      {Hue[0.015670463039738843`, 0.5, 0.7], LineBox[{274, 275}]}, 
      {Hue[0.013347830054055657`, 0.5, 0.7], LineBox[{275, 276}]}, 
      {Hue[0.011025197068372471`, 0.5, 0.7], LineBox[{276, 277}]}, 
      {Hue[0.008702564082689282, 0.5, 0.7], LineBox[{277, 278}]}, 
      {Hue[0.006379931097006096, 0.5, 0.7], LineBox[{278, 279}]}, 
      {Hue[0.004057298111322911, 0.5, 0.7], LineBox[{279, 280}]}, 
      {Hue[0.0017346651256397233`, 0.5, 0.7], LineBox[{280, 281}]}, 
      {Hue[0.9994120321399566, 0.5, 0.7], LineBox[{281, 282}]}, 
      {Hue[0.9970893991542733, 0.5, 0.7], LineBox[{282, 283}]}, 
      {Hue[0.9947667661685902, 0.5, 0.7], LineBox[{283, 284}]}, 
      {Hue[0.992444133182907, 0.5, 0.7], LineBox[{284, 285}]}, 
      {Hue[0.9901215001972238, 0.5, 0.7], LineBox[{285, 286}]}, 
      {Hue[0.9877988672115406, 0.5, 0.7], LineBox[{286, 287}]}, 
      {Hue[0.9854762342258574, 0.5, 0.7], LineBox[{287, 288}]}, 
      {Hue[0.9831536012401743, 0.5, 0.7], LineBox[{288, 289}]}, 
      {Hue[0.980830968254491, 0.5, 0.7], LineBox[{289, 290}]}, 
      {Hue[0.9785083352688079, 0.5, 0.7], LineBox[{290, 291}]}, 
      {Hue[0.9761857022831246, 0.5, 0.7], LineBox[{291, 292}]}, 
      {Hue[0.9738630692974415, 0.5, 0.7], LineBox[{292, 293}]}, 
      {Hue[0.9715404363117582, 0.5, 0.7], LineBox[{293, 294}]}, 
      {Hue[0.9692178033260751, 0.5, 0.7], LineBox[{294, 295}]}, 
      {Hue[0.9645725373547087, 0.5, 0.7], LineBox[{295, 296}]}, 
      {Hue[0.9622499043690256, 0.5, 0.7], LineBox[{296, 297}]}, 
      {Hue[0.9599272713833423, 0.5, 0.7], LineBox[{297, 298}]}, 
      {Hue[0.9552820054119759, 0.5, 0.7], LineBox[{298, 299}]}, 
      {Hue[0.9529593724262928, 0.5, 0.7], LineBox[{299, 300}]}, 
      {Hue[0.9506367394406096, 0.5, 0.7], LineBox[{300, 301}]}, 
      {Hue[0.9459914734692433, 0.5, 0.7], LineBox[{301, 302}]}, 
      {Hue[0.94366884048356, 0.5, 0.7], LineBox[{302, 303}]}, 
      {Hue[0.9413462074978769, 0.5, 0.7], LineBox[{303, 304}]}, 
      {Hue[0.9367009415265105, 0.5, 0.7], LineBox[{304, 305}]}, 
      {Hue[0.9274104095837777, 0.5, 0.7], LineBox[{305, 306}]}, 
      {Hue[0.9248906964138597, 0.5, 0.7], LineBox[{306, 307}]}, 
      {Hue[0.9223709832439416, 0.5, 0.7], LineBox[{307, 308}]}, 
      {Hue[0.9173315569041055, 0.5, 0.7], LineBox[{308, 309}]}, 
      {Hue[0.9072527042244332, 0.5, 0.7], LineBox[{309, 310}]}, 
      {Hue[0.9047329910545151, 0.5, 0.7], LineBox[{310, 311}]}, 
      {Hue[0.902213277884597, 0.5, 0.7], LineBox[{311, 312}]}, 
      {Hue[0.8971738515447609, 0.5, 0.7], LineBox[{312, 313}]}, 
      {Hue[0.8870949988650887, 0.5, 0.7], LineBox[{313, 314}]}, 
      {Hue[0.8845752856951706, 0.5, 0.7], LineBox[{314, 315}]}, 
      {Hue[0.8820555725252526, 0.5, 0.7], LineBox[{315, 316}]}, 
      {Hue[0.8770161461854165, 0.5, 0.7], LineBox[{316, 317}]}, 
      {Hue[0.8669372935057441, 0.5, 0.7], LineBox[{317, 318}]}, 
      {Hue[0.8467795881463996, 0.5, 0.7], LineBox[{318, 319}]}, 
      {Hue[0.8442598749764816, 0.5, 0.7], LineBox[{319, 320}]}, 
      {Hue[0.8417401618065634, 0.5, 0.7], LineBox[{320, 321}]}, 
      {Hue[0.8367007354667274, 0.5, 0.7], LineBox[{321, 322}]}, 
      {Hue[0.8266218827870551, 0.5, 0.7], LineBox[{322, 323}]}, 
      {Hue[0.8064641774277106, 0.5, 0.7], LineBox[{323, 324}]}, 
      {Hue[0.8039444642577924, 0.5, 0.7], LineBox[{324, 325}]}, 
      {Hue[0.8014247510878745, 0.5, 0.7], LineBox[{325, 326}]}, 
      {Hue[0.7963853247480384, 0.5, 0.7], LineBox[{326, 327}]}, 
      {Hue[0.786306472068366, 0.5, 0.7], LineBox[{327, 328}]}, 
      {Hue[0.7661487667090214, 0.5, 0.7], LineBox[{328, 329}]}, 
      {Hue[0.7637973801119673, 0.5, 0.7], LineBox[{329, 330}]}, 
      {Hue[0.7614459935149132, 0.5, 0.7], LineBox[{330, 331}]}, 
      {Hue[0.7567432203208047, 0.5, 0.7], LineBox[{331, 332}]}, 
      {Hue[0.7473376739325879, 0.5, 0.7], LineBox[{332, 333}]}, 
      {Hue[0.7285265811561541, 0.5, 0.7], LineBox[{333, 334}]}, 
      {Hue[0.690904395603287, 0.5, 0.7], LineBox[{334, 335}]}, 
      {Hue[0.6885530090062326, 0.5, 0.7], LineBox[{335, 336}]}, 
      {Hue[0.6862016224091785, 0.5, 0.7], LineBox[{336, 337}]}, 
      {Hue[0.68149884921507, 0.5, 0.7], LineBox[{337, 338}]}, 
      {Hue[0.6720933028268532, 0.5, 0.7], LineBox[{338, 339}]}, 
      {Hue[0.6532822100504196, 0.5, 0.7], LineBox[{339, 340}]}, 
      {Hue[0.6156600244975523, 0.5, 0.7], LineBox[{340, 341}]}, 
      {Hue[0.6131115577162632, 0.5, 0.7], LineBox[{341, 342}]}, 
      {Hue[0.6105630909349742, 0.5, 0.7], LineBox[{342, 343}]}, 
      {Hue[0.605466157372396, 0.5, 0.7], LineBox[{343, 344}]}, 
      {Hue[0.5952722902472397, 0.5, 0.7], LineBox[{344, 345}]}, 
      {Hue[0.5748845559969271, 0.5, 0.7], LineBox[{345, 346}]}, 
      {Hue[0.5341090874963016, 0.5, 0.7], LineBox[{346, 347}]}, 
      {Hue[0.45255815049505094`, 0.5, 0.7], LineBox[{347, 348}]}, 
      {Hue[0.4500562080342596, 0.5, 0.7], LineBox[{348, 349}]}, 
      {Hue[0.4475542655734681, 0.5, 0.7], LineBox[{349, 350}]}, 
      {Hue[0.4425503806518852, 0.5, 0.7], LineBox[{350, 351}]}, 
      {Hue[0.4325426108087192, 0.5, 0.7], LineBox[{351, 352}]}, 
      {Hue[0.41252707112238746`, 0.5, 0.7], LineBox[{352, 353}]}, 
      {Hue[0.372495991749724, 0.5, 0.7], LineBox[{353, 354}]}, 
      {Hue[0.2924338330043968, 0.5, 0.7], LineBox[{354, 355}]}, 
      {Hue[0.14308241617702985`, 0.5, 0.7], LineBox[{355, 356}]}, 
      {Hue[0.9811178675586304, 0.5, 0.7], LineBox[{356, 357}]}, 
      {Hue[0.829926219603518, 0.5, 0.7], LineBox[{357, 358}]}, 
      {Hue[0.681712128160253, 0.5, 0.7], LineBox[{358, 359}]}, 
      {Hue[0.520884904925956, 0.5, 0.7], LineBox[{359, 360}]}, 
      {Hue[0.5185403061357836, 0.5, 0.7], LineBox[{360, 361}]}, 
      {Hue[0.5161957073456116, 0.5, 0.7], LineBox[{361, 362}]}, 
      {Hue[0.5115065097652677, 0.5, 0.7], LineBox[{362, 363}]}, 
      {Hue[0.5021281146045796, 0.5, 0.7], LineBox[{363, 364}]}, 
      {Hue[0.48337132428320306`, 0.5, 0.7], LineBox[{364, 365}]}, 
      {Hue[0.4458577436404507, 0.5, 0.7], LineBox[{365, 366}]}, 
      {Hue[0.3708305823549457, 0.5, 0.7], LineBox[{366, 367}]}, 
      {Hue[0.3682889033805386, 0.5, 0.7], LineBox[{367, 368}]}, 
      {Hue[0.365747224406132, 0.5, 0.7], LineBox[{368, 369}]}, 
      {Hue[0.36066386645731824`, 0.5, 0.7], LineBox[{369, 370}]}, 
      {Hue[0.3504971505596905, 0.5, 0.7], LineBox[{370, 371}]}, 
      {Hue[0.33016371876443495`, 0.5, 0.7], LineBox[{371, 372}]}, 
      {Hue[0.2894968551739242, 0.5, 0.7], LineBox[{372, 373}]}, 
      {Hue[0.2081631279929033, 0.5, 0.7], LineBox[{373, 374}]}, 
      {Hue[0.2057897755913601, 0.5, 0.7], LineBox[{374, 375}]}, 
      {Hue[0.20341642318981695`, 0.5, 0.7], LineBox[{375, 376}]}, 
      {Hue[0.19866971838673092`, 0.5, 0.7], LineBox[{376, 377}]}, 
      {Hue[0.18917630878055883`, 0.5, 0.7], LineBox[{377, 378}]}, 
      {Hue[0.1701894895682144, 0.5, 0.7], LineBox[{378, 379}]}, 
      {Hue[0.1322158511435255, 0.5, 0.7], LineBox[{379, 380}]}, 
      {Hue[0.1298424987419823, 0.5, 0.7], LineBox[{380, 381}]}, 
      {Hue[0.12746914634043943`, 0.5, 0.7], LineBox[{381, 382}]}, 
      {Hue[0.12272244153735341`, 0.5, 0.7], LineBox[{382, 383}]}, 
      {Hue[0.11322903193118104`, 0.5, 0.7], LineBox[{383, 384}]}, 
      {Hue[0.09424221271883658, 0.5, 0.7], LineBox[{384, 385}]}, 
      {Hue[0.056268574294147965`, 0.5, 0.7], LineBox[{385, 386}]}, 
      {Hue[0.053941746213102444`, 0.5, 0.7], LineBox[{386, 387}]}, 
      {Hue[0.05161491813205696, 0.5, 0.7], LineBox[{387, 388}]}, 
      {Hue[0.046961261969965945`, 0.5, 0.7], LineBox[{388, 389}]}, 
      {Hue[0.03765394964578421, 0.5, 0.7], LineBox[{389, 390}]}, 
      {Hue[0.019039324997420736`, 0.5, 0.7], LineBox[{390, 391}]}, 
      {Hue[0.9818100757006936, 0.5, 0.7], LineBox[{391, 392}]}, 
      {Hue[0.979483247619648, 0.5, 0.7], LineBox[{392, 393}]}, 
      {Hue[0.9771564195386025, 0.5, 0.7], LineBox[{393, 394}]}, 
      {Hue[0.9725027633765115, 0.5, 0.7], LineBox[{394, 395}]}, 
      {Hue[0.9631954510523297, 0.5, 0.7], LineBox[{395, 396}]}, 
      {Hue[0.9445808264039662, 0.5, 0.7], LineBox[{396, 397}]}, 
      {Hue[0.9073515771072393, 0.5, 0.7], LineBox[{397, 398}]}, 
      {Hue[0.9048276688419594, 0.5, 0.7], LineBox[{398, 399}]}, 
      {Hue[0.9023037605766788, 0.5, 0.7], LineBox[{399, 400}]}, 
      {Hue[0.8972559440461179, 0.5, 0.7], LineBox[{400, 401}]}, 
      {Hue[0.8871603109849967, 0.5, 0.7], LineBox[{401, 402}]}, 
      {Hue[0.8669690448627543, 0.5, 0.7], LineBox[{402, 403}]}, 
      {Hue[0.8265865126182694, 0.5, 0.7], LineBox[{403, 404}]}, 
      {Hue[0.824062604352989, 0.5, 0.7], LineBox[{404, 405}]}, 
      {Hue[0.8215386960877087, 0.5, 0.7], LineBox[{405, 406}]}, 
      {Hue[0.8164908795571483, 0.5, 0.7], LineBox[{406, 407}]}, 
      {Hue[0.8063952464960271, 0.5, 0.7], LineBox[{407, 408}]}, 
      {Hue[0.7862039803737844, 0.5, 0.7], LineBox[{408, 409}]}, 
      {Hue[0.7458214481292991, 0.5, 0.7], LineBox[{409, 410}]}, 
      {Hue[0.7434658664368825, 0.5, 0.7], LineBox[{410, 411}]}, 
      {Hue[0.741110284744466, 0.5, 0.7], LineBox[{411, 412}]}, 
      {Hue[0.736399121359633, 0.5, 0.7], LineBox[{412, 413}]}, 
      {Hue[0.7269767945899674, 0.5, 0.7], LineBox[{413, 414}]}, 
      {Hue[0.7081321410506356, 0.5, 0.7], LineBox[{414, 415}]}, 
      {Hue[0.6704428339719721, 0.5, 0.7], LineBox[{415, 416}]}, 
      {Hue[0.6680872522795556, 0.5, 0.7], LineBox[{416, 417}]}, 
      {Hue[0.665731670587139, 0.5, 0.7], LineBox[{417, 418}]}, 
      {Hue[0.661020507202306, 0.5, 0.7], LineBox[{418, 419}]}, 
      {Hue[0.6515981804326405, 0.5, 0.7], LineBox[{419, 420}]}, 
      {Hue[0.6327535268933087, 0.5, 0.7], LineBox[{420, 421}]}, 
      {Hue[0.5950642198146456, 0.5, 0.7], LineBox[{421, 422}]}, 
      {Hue[0.5925115579379945, 0.5, 0.7], LineBox[{422, 423}]}, 
      {Hue[0.5899588960613429, 0.5, 0.7], LineBox[{423, 424}]}, 
      {Hue[0.5848535723080399, 0.5, 0.7], LineBox[{424, 425}]}, 
      {Hue[0.5746429248014346, 0.5, 0.7], LineBox[{425, 426}]}, 
      {Hue[0.5542216297882241, 0.5, 0.7], LineBox[{426, 427}]}, 
      {Hue[0.5133790397618024, 0.5, 0.7], LineBox[{427, 428}]}, 
      {Hue[0.5108263778851515, 0.5, 0.7], LineBox[{428, 429}]}, 
      {Hue[0.5082737160085, 0.5, 0.7], LineBox[{429, 430}]}, 
      {Hue[0.5031683922551968, 0.5, 0.7], LineBox[{430, 431}]}, 
      {Hue[0.4929577447485917, 0.5, 0.7], LineBox[{431, 432}]}, 
      {Hue[0.4725364497353812, 0.5, 0.7], LineBox[{432, 433}]}, 
      {Hue[0.4316938597089599, 0.5, 0.7], LineBox[{433, 434}]}, 
      {Hue[0.429187722152806, 0.5, 0.7], LineBox[{434, 435}]}, 
      {Hue[0.4266815845966521, 0.5, 0.7], LineBox[{435, 436}]}, 
      {Hue[0.42166930948434495`, 0.5, 0.7], LineBox[{436, 437}]}, 
      {Hue[0.41164475925973004`, 0.5, 0.7], LineBox[{437, 438}]}, 
      {Hue[0.39159565881050024`, 0.5, 0.7], LineBox[{438, 439}]}, 
      {Hue[0.35149745791204057`, 0.5, 0.7], LineBox[{439, 440}]}, 
      {Hue[0.3489913203558867, 0.5, 0.7], LineBox[{440, 441}]}, 
      {Hue[0.34648518279973284`, 0.5, 0.7], LineBox[{441, 442}]}, 
      {Hue[0.34147290768742566`, 0.5, 0.7], LineBox[{442, 443}]}, 
      {Hue[0.33144835746281076`, 0.5, 0.7], LineBox[{443, 444}]}, 
      {Hue[0.3113992570135809, 0.5, 0.7], LineBox[{444, 445}]}, 
      {Hue[0.27130105611512123`, 0.5, 0.7], LineBox[{445, 446}]}, 
      {Hue[0.26896324513183134`, 0.5, 0.7], LineBox[{446, 447}]}, 
      {Hue[0.26662543414854145`, 0.5, 0.7], LineBox[{447, 448}]}, 
      {Hue[0.26194981218196167`, 0.5, 0.7], LineBox[{448, 449}]}, 
      {Hue[0.25259856824880217`, 0.5, 0.7], LineBox[{449, 450}]}, 
      {Hue[0.23389608038248313`, 0.5, 0.7], LineBox[{450, 451}]}, 
      {Hue[0.19649110464984562`, 0.5, 0.7], LineBox[{451, 452}]}, 
      {Hue[0.19415329366655634`, 0.5, 0.7], LineBox[{452, 453}]}, 
      {Hue[0.19181548268326584`, 0.5, 0.7], LineBox[{453, 454}]}, 
      {Hue[0.18713986071668665`, 0.5, 0.7], LineBox[{454, 455}]}, 
      {Hue[0.17778861678352656`, 0.5, 0.7], LineBox[{455, 456}]}, 
      {Hue[0.1590861289172081, 0.5, 0.7], LineBox[{456, 457}]}, 
      {Hue[0.12168115318456944`, 0.5, 0.7], LineBox[{457, 458}]}, 
      {Hue[0.11914626201704458`, 0.5, 0.7], LineBox[{458, 459}]}, 
      {Hue[0.11661137084952025`, 0.5, 0.7], LineBox[{459, 460}]}, 
      {Hue[0.11154158851447046`, 0.5, 0.7], LineBox[{460, 461}]}, 
      {Hue[0.10140202384437148`, 0.5, 0.7], LineBox[{461, 462}]}, 
      {Hue[0.0811228945041741, 0.5, 0.7], LineBox[{462, 463}]}, 
      {Hue[0.0405646358237776, 0.5, 0.7], LineBox[{463, 464}]}, 
      {Hue[0.038029744656252716`, 0.5, 0.7], LineBox[{464, 465}]}, 
      {Hue[0.035494853488728405`, 0.5, 0.7], LineBox[{465, 466}]}, 
      {Hue[0.03042507115367863, 0.5, 0.7], LineBox[{466, 467}]}, 
      {Hue[0.020285506483580216`, 0.5, 0.7], LineBox[{467, 468}]}, 
      {Hue[6.377143381686833*^-6, 0.5, 0.7], LineBox[{468, 469}]}, 
      {Hue[0.9594481184629863, 0.5, 0.7], LineBox[{469, 470}]}, 
      {Hue[0.9570815538683255, 0.5, 0.7], LineBox[{470, 471}]}, 
      {Hue[0.9547149892736645, 0.5, 0.7], LineBox[{471, 472}]}, 
      {Hue[0.9499818600843427, 0.5, 0.7], LineBox[{472, 473}]}, 
      {Hue[0.9405156017056991, 0.5, 0.7], LineBox[{473, 474}]}, 
      {Hue[0.921583084948412, 0.5, 0.7], LineBox[{474, 475}]}, 
      {Hue[0.8837180514338376, 0.5, 0.7], LineBox[{475, 476}]}, 
      {Hue[0.8813514868391773, 0.5, 0.7], LineBox[{476, 477}]}, 
      {Hue[0.8789849222445164, 0.5, 0.7], LineBox[{477, 478}]}, 
      {Hue[0.8742517930551946, 0.5, 0.7], LineBox[{478, 479}]}, 
      {Hue[0.8647855346765511, 0.5, 0.7], LineBox[{479, 480}]}, 
      {Hue[0.8458530179192638, 0.5, 0.7], LineBox[{480, 481}]}, 
      {Hue[0.8079879844046894, 0.5, 0.7], LineBox[{481, 482}]}, 
      {Hue[0.8056679441305262, 0.5, 0.7], LineBox[{482, 483}]}, 
      {Hue[0.8033479038563625, 0.5, 0.7], LineBox[{483, 484}]}, 
      {Hue[0.7987078233080366, 0.5, 0.7], LineBox[{484, 485}]}, 
      {Hue[0.7894276622113829, 0.5, 0.7], LineBox[{485, 486}]}, 
      {Hue[0.7708673400180771, 0.5, 0.7], LineBox[{486, 487}]}, 
      {Hue[0.7685472997439133, 0.5, 0.7], LineBox[{487, 488}]}, 
      {Hue[0.76622725946975, 0.5, 0.7], LineBox[{488, 489}]}, 
      {Hue[0.7615871789214236, 0.5, 0.7], LineBox[{489, 490}]}, 
      {Hue[0.7523070178247706, 0.5, 0.7], LineBox[{490, 491}]}, 
      {Hue[0.7337466956314647, 0.5, 0.7], LineBox[{491, 492}]}, 
      {Hue[0.7314266553573009, 0.5, 0.7], LineBox[{492, 493}]}, 
      {Hue[0.7291066150831382, 0.5, 0.7], LineBox[{493, 494}]}, 
      {Hue[0.7244665345348111, 0.5, 0.7], LineBox[{494, 495}]}, 
      {Hue[0.7151863734381587, 0.5, 0.7], LineBox[{495, 496}]}, 
      {Hue[0.6966260512448523, 0.5, 0.7], LineBox[{496, 497}]}, 
      {Hue[0.65950540685824, 0.5, 0.7], LineBox[{497, 498}]}, 
      {Hue[0.6569882863998417, 0.5, 0.7], LineBox[{498, 499}]}, 
      {Hue[0.6544711659414435, 0.5, 0.7], LineBox[{499, 500}]}, 
      {Hue[0.6494369250246474, 0.5, 0.7], LineBox[{500, 501}]}, 
      {Hue[0.6393684431910545, 0.5, 0.7], LineBox[{501, 502}]}, 
      {Hue[0.6192314795238697, 0.5, 0.7], LineBox[{502, 503}]}, 
      {Hue[0.616714359065472, 0.5, 0.7], LineBox[{503, 504}]}, 
      {Hue[0.6141972386070738, 0.5, 0.7], LineBox[{504, 505}]}, 
      {Hue[0.6091629976902777, 0.5, 0.7], LineBox[{505, 506}]}, 
      {Hue[0.5990945158566849, 0.5, 0.7], LineBox[{506, 507}]}, 
      {Hue[0.5789575521894994, 0.5, 0.7], LineBox[{507, 508}]}, 
      {Hue[0.5764404317311012, 0.5, 0.7], LineBox[{508, 509}]}, 
      {Hue[0.5739233112727036, 0.5, 0.7], LineBox[{509, 510}]}, 
      {Hue[0.5688890703559071, 0.5, 0.7], LineBox[{510, 511}]}, 
      {Hue[0.5588205885223141, 0.5, 0.7], LineBox[{511, 512}]}, 
      {Hue[0.5386836248551292, 0.5, 0.7], LineBox[{512, 513}]}, 
      {Hue[0.4984096975207585, 0.5, 0.7], LineBox[{513, 514}]}, 
      {Hue[0.49606090363522365`, 0.5, 0.7], LineBox[{514, 515}]}, 
      {Hue[0.49371210974968993`, 0.5, 0.7], LineBox[{515, 516}]}, 
      {Hue[0.48901452197862144`, 0.5, 0.7], LineBox[{516, 517}]}, 
      {Hue[0.4796193464364844, 0.5, 0.7], LineBox[{517, 518}]}, 
      {Hue[0.4608289953522104, 0.5, 0.7], LineBox[{518, 519}]}, 
      {Hue[0.45848020146667556`, 0.5, 0.7], LineBox[{519, 520}]}, 
      {Hue[0.4561314075811413, 0.5, 0.7], LineBox[{520, 521}]}, 
      {Hue[0.4514338198100728, 0.5, 0.7], LineBox[{521, 522}]}, 
      {Hue[0.4420386442679352, 0.5, 0.7], LineBox[{522, 523}]}, 
      {Hue[0.42324829318366114`, 0.5, 0.7], LineBox[{523, 524}]}, 
      {Hue[0.4208994992981274, 0.5, 0.7], LineBox[{524, 525}]}, 
      {Hue[0.41855070541259265`, 0.5, 0.7], LineBox[{525, 526}]}, 
      {Hue[0.4138531176415241, 0.5, 0.7], LineBox[{526, 527}]}, 
      {Hue[0.40445794209938707`, 0.5, 0.7], LineBox[{527, 528}]}, 
      {Hue[0.3856675910151125, 0.5, 0.7], LineBox[{528, 529}]}, 
      {Hue[0.3480868888465638, 0.5, 0.7], LineBox[{529, 530}]}, 
      {Hue[0.3455410147767951, 0.5, 0.7], LineBox[{530, 531}]}, 
      {Hue[0.34299514070702586`, 0.5, 0.7], LineBox[{531, 532}]}, 
      {Hue[0.3379033925674879, 0.5, 0.7], LineBox[{532, 533}]}, 
      {Hue[0.32771989628841086`, 0.5, 0.7], LineBox[{533, 534}]}, 
      {Hue[0.30735290373025737`, 0.5, 0.7], LineBox[{534, 535}]}, 
      {Hue[0.30480702966048806`, 0.5, 0.7], LineBox[{535, 536}]}, 
      {Hue[0.30226115559071937`, 0.5, 0.7], LineBox[{536, 537}]}, 
      {Hue[0.2971694074511808, 0.5, 0.7], LineBox[{537, 538}]}, 
      {Hue[0.28698591117210376`, 0.5, 0.7], LineBox[{538, 539}]}, 
      {Hue[0.2666189186139508, 0.5, 0.7], LineBox[{539, 540}]}, 
      {Hue[0.26407304454418157`, 0.5, 0.7], LineBox[{540, 541}]}, 
      {Hue[0.2615271704744123, 0.5, 0.7], LineBox[{541, 542}]}, 
      {Hue[0.25643542233487376`, 0.5, 0.7], LineBox[{542, 543}]}, 
      {Hue[0.2462519260557973, 0.5, 0.7], LineBox[{543, 544}]}, 
      {Hue[0.22588493349764377`, 0.5, 0.7], LineBox[{544, 545}]}, 
      {Hue[0.18515094838133725`, 0.5, 0.7], LineBox[{545, 546}]}, 
      {Hue[0.185074863589463, 0.5, 0.7], LineBox[{546, 547}]}, 
      {Hue[0.18499877879758939`, 0.5, 0.7], LineBox[{547, 548}]}, 
      {Hue[0.18484660921384086`, 0.5, 0.7], LineBox[{548, 549}]}, 
      {Hue[0.1845422700463445, 0.5, 0.7], LineBox[{549, 550}]}, 
      {Hue[0.18393359171135223`, 0.5, 0.7], LineBox[{550, 551}]}, 
      {Hue[0.18271623504136603`, 0.5, 0.7], LineBox[{551, 552}]}, 
      {Hue[0.18264015024949232`, 0.5, 0.7], LineBox[{552, 553}]}, 
      {Hue[0.18256406545761808`, 0.5, 0.7], LineBox[{553, 554}]}, 
      {Hue[0.18241189587386958`, 0.5, 0.7], LineBox[{554, 555}]}, 
      {Hue[0.18210755670637377`, 0.5, 0.7], LineBox[{555, 556}]}, 
      {Hue[0.1814988783713804, 0.5, 0.7], LineBox[{556, 557}]}, 
      {Hue[0.1814227935795061, 0.5, 0.7], LineBox[{557, 558}]}, 
      {Hue[0.18134670878763243`, 0.5, 0.7], LineBox[{558, 559}]}, 
      {Hue[0.18119453920388395`, 0.5, 0.7], LineBox[{559, 560}]}, 
      {Hue[0.18089020003638753`, 0.5, 0.7], LineBox[{560, 561}]}, 
      {Hue[0.1808141152445139, 0.5, 0.7], LineBox[{561, 562}]}, 
      {Hue[0.18073803045263961`, 0.5, 0.7], LineBox[{562, 563}]}, 
      {Hue[0.18058586086889114`, 0.5, 0.7], LineBox[{563, 564}]}, 
      {Hue[0.18050977607701693`, 0.5, 0.7], LineBox[{564, 565}]}, 
      {Hue[0.18043369128514264`, 0.5, 0.7], LineBox[{565, 566}]}, 
      {Hue[0.18035760649326896`, 0.5, 0.7], LineBox[{566, 567}]}}}},
   VertexColors->{
     Hue[0.8197184782986047, 0.5, 0.7], 
     Hue[0.817375481892275, 0.5, 0.7], 
     Hue[0.8150324854859454, 0.5, 0.7], 
     Hue[0.8103464926732862, 0.5, 0.7], 
     Hue[0.8009745070479671, 0.5, 0.7], 
     Hue[0.7822305357973296, 0.5, 0.7], 
     Hue[0.779887539391, 0.5, 0.7], 
     Hue[0.7775445429846704, 0.5, 0.7], 
     Hue[0.7728585501720105, 0.5, 0.7], 
     Hue[0.763486564546692, 0.5, 0.7], 
     Hue[0.7447425932960545, 0.5, 0.7], 
     Hue[0.7423995968897242, 0.5, 0.7], 
     Hue[0.7400566004833946, 0.5, 0.7], 
     Hue[0.7353706076707348, 0.5, 0.7], 
     Hue[0.7259986220454163, 0.5, 0.7], 
     Hue[0.7072546507947788, 0.5, 0.7], 
     Hue[0.6697667082935036, 0.5, 0.7], 
     Hue[0.6672266317029389, 0.5, 0.7], 
     Hue[0.6646865551123748, 0.5, 0.7], 
     Hue[0.6596064019312451, 0.5, 0.7], 
     Hue[0.6494460955689871, 0.5, 0.7], 
     Hue[0.6291254828444701, 0.5, 0.7], 
     Hue[0.626585406253906, 0.5, 0.7], 
     Hue[0.6240453296633414, 0.5, 0.7], 
     Hue[0.6189651764822122, 0.5, 0.7], 
     Hue[0.6088048701199542, 0.5, 0.7], 
     Hue[0.5884842573954372, 0.5, 0.7], 
     Hue[0.5859441808048731, 0.5, 0.7], 
     Hue[0.5834041042143084, 0.5, 0.7], 
     Hue[0.5783239510331792, 0.5, 0.7], 
     Hue[0.5681636446709212, 0.5, 0.7], 
     Hue[0.5478430319464042, 0.5, 0.7], 
     Hue[0.5072018064973707, 0.5, 0.7], 
     Hue[0.5048300564796699, 0.5, 0.7], 
     Hue[0.5024583064619688, 0.5, 0.7], 
     Hue[0.4977148064265675, 0.5, 0.7], 
     Hue[0.48822780635576496`, 0.5, 0.7], 
     Hue[0.4692538062141587, 0.5, 0.7], 
     Hue[0.466882056196458, 0.5, 0.7], 
     Hue[0.4645103061787574, 0.5, 0.7], 
     Hue[0.45976680614335613`, 0.5, 0.7], 
     Hue[0.45027980607255297`, 0.5, 0.7], 
     Hue[0.43130580593094725`, 0.5, 0.7], 
     Hue[0.4289340559132466, 0.5, 0.7], 
     Hue[0.426562305895546, 0.5, 0.7], 
     Hue[0.42181880586014414`, 0.5, 0.7], 
     Hue[0.4123318057893416, 0.5, 0.7], 
     Hue[0.39335780564773587`, 0.5, 0.7], 
     Hue[0.3554098053645245, 0.5, 0.7], 
     Hue[0.3530845796673215, 0.5, 0.7], 
     Hue[0.35075935397011854`, 0.5, 0.7], 
     Hue[0.3461089025757114, 0.5, 0.7], 
     Hue[0.33680799978689946`, 0.5, 0.7], 
     Hue[0.3182061942092745, 0.5, 0.7], 
     Hue[0.3158809685120721, 0.5, 0.7], 
     Hue[0.31355574281486853`, 0.5, 0.7], 
     Hue[0.308905291420462, 0.5, 0.7], 
     Hue[0.2996043886316495, 0.5, 0.7], 
     Hue[0.2810025830540251, 0.5, 0.7], 
     Hue[0.2786773573568221, 0.5, 0.7], 
     Hue[0.2763521316596185, 0.5, 0.7], 
     Hue[0.2717016802652126, 0.5, 0.7], 
     Hue[0.26240077747640006`, 0.5, 0.7], 
     Hue[0.24379897189877506`, 0.5, 0.7], 
     Hue[0.2065953607435256, 0.5, 0.7], 
     Hue[0.20407305486208763`, 0.5, 0.7], 
     Hue[0.20155074898064965`, 0.5, 0.7], 
     Hue[0.19650613721777366`, 0.5, 0.7], 
     Hue[0.1864169136920217, 0.5, 0.7], 
     Hue[0.16623846664051775`, 0.5, 0.7], 
     Hue[0.16371616075907977`, 0.5, 0.7], 
     Hue[0.1611938548776418, 0.5, 0.7], 
     Hue[0.1561492431147658, 0.5, 0.7], 
     Hue[0.14606001958901382`, 0.5, 0.7], 
     Hue[0.12588157253750992`, 0.5, 0.7], 
     Hue[0.12335926665607193`, 0.5, 0.7], 
     Hue[0.12083696077463393`, 0.5, 0.7], 
     Hue[0.11579234901175797`, 0.5, 0.7], 
     Hue[0.105703125486006, 0.5, 0.7], 
     Hue[0.0855246784345021, 0.5, 0.7], 
     Hue[0.04516778433149423, 0.5, 0.7], 
     Hue[0.04281380502292024, 0.5, 0.7], 
     Hue[0.04045982571434625, 0.5, 0.7], 
     Hue[0.035751867097198235`, 0.5, 0.7], 
     Hue[0.026335949862901102`, 0.5, 0.7], 
     Hue[0.007504115394307963, 0.5, 0.7], 
     Hue[0.9698404464571223, 0.5, 0.7], 
     Hue[0.9674864671485477, 0.5, 0.7], 
     Hue[0.9651324878399743, 0.5, 0.7], 
     Hue[0.9604245292228257, 0.5, 0.7], 
     Hue[0.9510086119885292, 0.5, 0.7], 
     Hue[0.9321767775199366, 0.5, 0.7], 
     Hue[0.8945131085827502, 0.5, 0.7], 
     Hue[0.8919620490899413, 0.5, 0.7], 
     Hue[0.8894109895971323, 0.5, 0.7], 
     Hue[0.8843088706115143, 0.5, 0.7], 
     Hue[0.8741046326402782, 0.5, 0.7], 
     Hue[0.8536961566978062, 0.5, 0.7], 
     Hue[0.8128792048128621, 0.5, 0.7], 
     Hue[0.810328145320053, 0.5, 0.7], 
     Hue[0.8077770858272441, 0.5, 0.7], 
     Hue[0.802674966841626, 0.5, 0.7], 
     Hue[0.79247072887039, 0.5, 0.7], 
     Hue[0.772062252927918, 0.5, 0.7], 
     Hue[0.7312453010429738, 0.5, 0.7], 
     Hue[0.7287407658706625, 0.5, 0.7], 
     Hue[0.7262362306983512, 0.5, 0.7], 
     Hue[0.7212271603537285, 0.5, 0.7], 
     Hue[0.711209019664483, 0.5, 0.7], 
     Hue[0.6911727382859922, 0.5, 0.7], 
     Hue[0.6511001755290096, 0.5, 0.7], 
     Hue[0.6485956403566981, 0.5, 0.7], 
     Hue[0.646091105184387, 0.5, 0.7], 
     Hue[0.6410820348397641, 0.5, 0.7], 
     Hue[0.6310638941505182, 0.5, 0.7], 
     Hue[0.6110276127720268, 0.5, 0.7], 
     Hue[0.5709550500150447, 0.5, 0.7], 
     Hue[0.5686188414155973, 0.5, 0.7], 
     Hue[0.5662826328161494, 0.5, 0.7], 
     Hue[0.5616102156172547, 0.5, 0.7], 
     Hue[0.5522653812194647, 0.5, 0.7], 
     Hue[0.5335757124238844, 0.5, 0.7], 
     Hue[0.49619637483272355`, 0.5, 0.7], 
     Hue[0.49386016623327594`, 0.5, 0.7], 
     Hue[0.49152395763382856`, 0.5, 0.7], 
     Hue[0.48685154043493384`, 0.5, 0.7], 
     Hue[0.47750670603714357`, 0.5, 0.7], 
     Hue[0.458817037241563, 0.5, 0.7], 
     Hue[0.4214376996504027, 0.5, 0.7], 
     Hue[0.4189044108667203, 0.5, 0.7], 
     Hue[0.41637112208303795`, 0.5, 0.7], 
     Hue[0.4113045445156732, 0.5, 0.7], 
     Hue[0.4011713893809435, 0.5, 0.7], 
     Hue[0.38090507911148397`, 0.5, 0.7], 
     Hue[0.34037245857256526`, 0.5, 0.7], 
     Hue[0.3378391697888829, 0.5, 0.7], 
     Hue[0.33530588100520053`, 0.5, 0.7], 
     Hue[0.3302393034378358, 0.5, 0.7], 
     Hue[0.3201061483031061, 0.5, 0.7], 
     Hue[0.2998398380336466, 0.5, 0.7], 
     Hue[0.2593072174947282, 0.5, 0.7], 
     Hue[0.2569422552839095, 0.5, 0.7], 
     Hue[0.2545772930730911, 0.5, 0.7], 
     Hue[0.24984736865145438`, 0.5, 0.7], 
     Hue[0.24038751980818002`, 0.5, 0.7], 
     Hue[0.22146782212163132`, 0.5, 0.7], 
     Hue[0.18362842674853475`, 0.5, 0.7], 
     Hue[0.18126346453771638`, 0.5, 0.7], 
     Hue[0.17889850232689772`, 0.5, 0.7], 
     Hue[0.17416857790526039`, 0.5, 0.7], 
     Hue[0.16470872906198633`, 0.5, 0.7], 
     Hue[0.14578903137543817`, 0.5, 0.7], 
     Hue[0.10794963600234132`, 0.5, 0.7], 
     Hue[0.10563119811202033`, 0.5, 0.7], 
     Hue[0.10331276022169904`, 0.5, 0.7], 
     Hue[0.09867588444105733, 0.5, 0.7], 
     Hue[0.08940213287977387, 0.5, 0.7], 
     Hue[0.07085462975720616, 0.5, 0.7], 
     Hue[0.03375962351207098, 0.5, 0.7], 
     Hue[0.03144118562175026, 0.5, 0.7], 
     Hue[0.029122747731428977`, 0.5, 0.7], 
     Hue[0.024485871950787247`, 0.5, 0.7], 
     Hue[0.015212120389503814`, 0.5, 0.7], 
     Hue[0.9966646172669361, 0.5, 0.7], 
     Hue[0.9595696110218012, 0.5, 0.7], 
     Hue[0.9570540929472452, 0.5, 0.7], 
     Hue[0.9545385748726898, 0.5, 0.7], 
     Hue[0.949507538723578, 0.5, 0.7], 
     Hue[0.9394454664253545, 0.5, 0.7], 
     Hue[0.9193213218289079, 0.5, 0.7], 
     Hue[0.8790730326360152, 0.5, 0.7], 
     Hue[0.7985764542502292, 0.5, 0.7], 
     Hue[0.7962292627485371, 0.5, 0.7], 
     Hue[0.7938820712468451, 0.5, 0.7], 
     Hue[0.7891876882434611, 0.5, 0.7], 
     Hue[0.7797989222366932, 0.5, 0.7], 
     Hue[0.7610213902231577, 0.5, 0.7], 
     Hue[0.7234663261960863, 0.5, 0.7], 
     Hue[0.6483561981419439, 0.5, 0.7], 
     Hue[0.6458119264560171, 0.5, 0.7], 
     Hue[0.6432676547700904, 0.5, 0.7], 
     Hue[0.6381791113982367, 0.5, 0.7], 
     Hue[0.6280020246545293, 0.5, 0.7], 
     Hue[0.6076478511671144, 0.5, 0.7], 
     Hue[0.5669395041922853, 0.5, 0.7], 
     Hue[0.4855228102426269, 0.5, 0.7], 
     Hue[0.3256669788551567, 0.5, 0.7], 
     Hue[0.17658404813097406`, 0.5, 0.7], 
     Hue[0.014887985615758862`, 0.5, 0.7], 
     Hue[0.8639648237638307, 0.5, 0.7], 
     Hue[0.7004285301208703, 0.5, 0.7], 
     Hue[0.6979197998531967, 0.5, 0.7], 
     Hue[0.6954110695855231, 0.5, 0.7], 
     Hue[0.6903936090501757, 0.5, 0.7], 
     Hue[0.6803586879794812, 0.5, 0.7], 
     Hue[0.660288845838092, 0.5, 0.7], 
     Hue[0.6201491615553137, 0.5, 0.7], 
     Hue[0.5398697929897572, 0.5, 0.7], 
     Hue[0.5375293892949474, 0.5, 0.7], 
     Hue[0.5351889856001377, 0.5, 0.7], 
     Hue[0.5305081782105181, 0.5, 0.7], 
     Hue[0.5211465634312791, 0.5, 0.7], 
     Hue[0.5024233338728007, 0.5, 0.7], 
     Hue[0.46497687475584426`, 0.5, 0.7], 
     Hue[0.4626364710610345, 0.5, 0.7], 
     Hue[0.46029606736622475`, 0.5, 0.7], 
     Hue[0.45561525997660507`, 0.5, 0.7], 
     Hue[0.446253645197366, 0.5, 0.7], 
     Hue[0.4275304156388878, 0.5, 0.7], 
     Hue[0.3900839565219313, 0.5, 0.7], 
     Hue[0.38754647264288655`, 0.5, 0.7], 
     Hue[0.38500898876384193`, 0.5, 0.7], 
     Hue[0.37993402100575263`, 0.5, 0.7], 
     Hue[0.3697840854895739, 0.5, 0.7], 
     Hue[0.3494842144572167, 0.5, 0.7], 
     Hue[0.34694673057817194`, 0.5, 0.7], 
     Hue[0.34440924669912737`, 0.5, 0.7], 
     Hue[0.339334278941038, 0.5, 0.7], 
     Hue[0.32918434342485936`, 0.5, 0.7], 
     Hue[0.3088844723925021, 0.5, 0.7], 
     Hue[0.30634698851345743`, 0.5, 0.7], 
     Hue[0.30380950463441275`, 0.5, 0.7], 
     Hue[0.29873453687632345`, 0.5, 0.7], 
     Hue[0.2885846013601448, 0.5, 0.7], 
     Hue[0.2682847303277875, 0.5, 0.7], 
     Hue[0.22768498826307296`, 0.5, 0.7], 
     Hue[0.2253158309568922, 0.5, 0.7], 
     Hue[0.22294667365071136`, 0.5, 0.7], 
     Hue[0.21820835903834976`, 0.5, 0.7], 
     Hue[0.20873172981362653`, 0.5, 0.7], 
     Hue[0.18977847136418016`, 0.5, 0.7], 
     Hue[0.18740931405799935`, 0.5, 0.7], 
     Hue[0.18504015675181856`, 0.5, 0.7], 
     Hue[0.18030184213945694`, 0.5, 0.7], 
     Hue[0.17082521291473377`, 0.5, 0.7], 
     Hue[0.1518719544652873, 0.5, 0.7], 
     Hue[0.14950279715910653`, 0.5, 0.7], 
     Hue[0.14713363985292574`, 0.5, 0.7], 
     Hue[0.1423953252405641, 0.5, 0.7], 
     Hue[0.1329186960158409, 0.5, 0.7], 
     Hue[0.13054953870966016`, 0.5, 0.7], 
     Hue[0.12818038140347934`, 0.5, 0.7], 
     Hue[0.12344206679111772`, 0.5, 0.7], 
     Hue[0.11396543756639452`, 0.5, 0.7], 
     Hue[0.11159628026021372`, 0.5, 0.7], 
     Hue[0.1092271229540329, 0.5, 0.7], 
     Hue[0.10448880834167132`, 0.5, 0.7], 
     Hue[0.09501217911694812, 0.5, 0.7], 
     Hue[0.0926430218107673, 0.5, 0.7], 
     Hue[0.09027386450458652, 0.5, 0.7], 
     Hue[0.08553554989222491, 0.5, 0.7], 
     Hue[0.07605892066750171, 0.5, 0.7], 
     Hue[0.07373628768181852, 0.5, 0.7], 
     Hue[0.07141365469613532, 0.5, 0.7], 
     Hue[0.06676838872476895, 0.5, 0.7], 
     Hue[0.06444575573908579, 0.5, 0.7], 
     Hue[0.06212312275340258, 0.5, 0.7], 
     Hue[0.05747785678203621, 0.5, 0.7], 
     Hue[0.05515522379635303, 0.5, 0.7], 
     Hue[0.052832590810669836`, 0.5, 0.7], 
     Hue[0.04818732483930348, 0.5, 0.7], 
     Hue[0.04586469185362028, 0.5, 0.7], 
     Hue[0.043542058867937085`, 0.5, 0.7], 
     Hue[0.03889679289657071, 0.5, 0.7], 
     Hue[0.03657415991088753, 0.5, 0.7], 
     Hue[0.03425152692520435, 0.5, 0.7], 
     Hue[0.031928893939521155`, 0.5, 0.7], 
     Hue[0.029606260953837966`, 0.5, 0.7], 
     Hue[0.02728362796815478, 0.5, 0.7], 
     Hue[0.024960994982471587`, 0.5, 0.7], 
     Hue[0.0226383619967884, 0.5, 0.7], 
     Hue[0.02031572901110522, 0.5, 0.7], 
     Hue[0.017993096025422036`, 0.5, 0.7], 
     Hue[0.015670463039738843`, 0.5, 0.7], 
     Hue[0.013347830054055657`, 0.5, 0.7], 
     Hue[0.011025197068372471`, 0.5, 0.7], 
     Hue[0.008702564082689282, 0.5, 0.7], 
     Hue[0.006379931097006096, 0.5, 0.7], 
     Hue[0.004057298111322911, 0.5, 0.7], 
     Hue[0.0017346651256397233`, 0.5, 0.7], 
     Hue[0.9994120321399566, 0.5, 0.7], 
     Hue[0.9970893991542733, 0.5, 0.7], 
     Hue[0.9947667661685902, 0.5, 0.7], 
     Hue[0.992444133182907, 0.5, 0.7], 
     Hue[0.9901215001972238, 0.5, 0.7], 
     Hue[0.9877988672115406, 0.5, 0.7], 
     Hue[0.9854762342258574, 0.5, 0.7], 
     Hue[0.9831536012401743, 0.5, 0.7], 
     Hue[0.980830968254491, 0.5, 0.7], 
     Hue[0.9785083352688079, 0.5, 0.7], 
     Hue[0.9761857022831246, 0.5, 0.7], 
     Hue[0.9738630692974415, 0.5, 0.7], 
     Hue[0.9715404363117582, 0.5, 0.7], 
     Hue[0.9692178033260751, 0.5, 0.7], 
     Hue[0.9645725373547087, 0.5, 0.7], 
     Hue[0.9622499043690256, 0.5, 0.7], 
     Hue[0.9599272713833423, 0.5, 0.7], 
     Hue[0.9552820054119759, 0.5, 0.7], 
     Hue[0.9529593724262928, 0.5, 0.7], 
     Hue[0.9506367394406096, 0.5, 0.7], 
     Hue[0.9459914734692433, 0.5, 0.7], 
     Hue[0.94366884048356, 0.5, 0.7], 
     Hue[0.9413462074978769, 0.5, 0.7], 
     Hue[0.9367009415265105, 0.5, 0.7], 
     Hue[0.9274104095837777, 0.5, 0.7], 
     Hue[0.9248906964138597, 0.5, 0.7], 
     Hue[0.9223709832439416, 0.5, 0.7], 
     Hue[0.9173315569041055, 0.5, 0.7], 
     Hue[0.9072527042244332, 0.5, 0.7], 
     Hue[0.9047329910545151, 0.5, 0.7], 
     Hue[0.902213277884597, 0.5, 0.7], 
     Hue[0.8971738515447609, 0.5, 0.7], 
     Hue[0.8870949988650887, 0.5, 0.7], 
     Hue[0.8845752856951706, 0.5, 0.7], 
     Hue[0.8820555725252526, 0.5, 0.7], 
     Hue[0.8770161461854165, 0.5, 0.7], 
     Hue[0.8669372935057441, 0.5, 0.7], 
     Hue[0.8467795881463996, 0.5, 0.7], 
     Hue[0.8442598749764816, 0.5, 0.7], 
     Hue[0.8417401618065634, 0.5, 0.7], 
     Hue[0.8367007354667274, 0.5, 0.7], 
     Hue[0.8266218827870551, 0.5, 0.7], 
     Hue[0.8064641774277106, 0.5, 0.7], 
     Hue[0.8039444642577924, 0.5, 0.7], 
     Hue[0.8014247510878745, 0.5, 0.7], 
     Hue[0.7963853247480384, 0.5, 0.7], 
     Hue[0.786306472068366, 0.5, 0.7], 
     Hue[0.7661487667090214, 0.5, 0.7], 
     Hue[0.7637973801119673, 0.5, 0.7], 
     Hue[0.7614459935149132, 0.5, 0.7], 
     Hue[0.7567432203208047, 0.5, 0.7], 
     Hue[0.7473376739325879, 0.5, 0.7], 
     Hue[0.7285265811561541, 0.5, 0.7], 
     Hue[0.690904395603287, 0.5, 0.7], 
     Hue[0.6885530090062326, 0.5, 0.7], 
     Hue[0.6862016224091785, 0.5, 0.7], 
     Hue[0.68149884921507, 0.5, 0.7], 
     Hue[0.6720933028268532, 0.5, 0.7], 
     Hue[0.6532822100504196, 0.5, 0.7], 
     Hue[0.6156600244975523, 0.5, 0.7], 
     Hue[0.6131115577162632, 0.5, 0.7], 
     Hue[0.6105630909349742, 0.5, 0.7], 
     Hue[0.605466157372396, 0.5, 0.7], 
     Hue[0.5952722902472397, 0.5, 0.7], 
     Hue[0.5748845559969271, 0.5, 0.7], 
     Hue[0.5341090874963016, 0.5, 0.7], 
     Hue[0.45255815049505094`, 0.5, 0.7], 
     Hue[0.4500562080342596, 0.5, 0.7], 
     Hue[0.4475542655734681, 0.5, 0.7], 
     Hue[0.4425503806518852, 0.5, 0.7], 
     Hue[0.4325426108087192, 0.5, 0.7], 
     Hue[0.41252707112238746`, 0.5, 0.7], 
     Hue[0.372495991749724, 0.5, 0.7], 
     Hue[0.2924338330043968, 0.5, 0.7], 
     Hue[0.14308241617702985`, 0.5, 0.7], 
     Hue[0.9811178675586304, 0.5, 0.7], 
     Hue[0.829926219603518, 0.5, 0.7], 
     Hue[0.681712128160253, 0.5, 0.7], 
     Hue[0.520884904925956, 0.5, 0.7], 
     Hue[0.5185403061357836, 0.5, 0.7], 
     Hue[0.5161957073456116, 0.5, 0.7], 
     Hue[0.5115065097652677, 0.5, 0.7], 
     Hue[0.5021281146045796, 0.5, 0.7], 
     Hue[0.48337132428320306`, 0.5, 0.7], 
     Hue[0.4458577436404507, 0.5, 0.7], 
     Hue[0.3708305823549457, 0.5, 0.7], 
     Hue[0.3682889033805386, 0.5, 0.7], 
     Hue[0.365747224406132, 0.5, 0.7], 
     Hue[0.36066386645731824`, 0.5, 0.7], 
     Hue[0.3504971505596905, 0.5, 0.7], 
     Hue[0.33016371876443495`, 0.5, 0.7], 
     Hue[0.2894968551739242, 0.5, 0.7], 
     Hue[0.2081631279929033, 0.5, 0.7], 
     Hue[0.2057897755913601, 0.5, 0.7], 
     Hue[0.20341642318981695`, 0.5, 0.7], 
     Hue[0.19866971838673092`, 0.5, 0.7], 
     Hue[0.18917630878055883`, 0.5, 0.7], 
     Hue[0.1701894895682144, 0.5, 0.7], 
     Hue[0.1322158511435255, 0.5, 0.7], 
     Hue[0.1298424987419823, 0.5, 0.7], 
     Hue[0.12746914634043943`, 0.5, 0.7], 
     Hue[0.12272244153735341`, 0.5, 0.7], 
     Hue[0.11322903193118104`, 0.5, 0.7], 
     Hue[0.09424221271883658, 0.5, 0.7], 
     Hue[0.056268574294147965`, 0.5, 0.7], 
     Hue[0.053941746213102444`, 0.5, 0.7], 
     Hue[0.05161491813205696, 0.5, 0.7], 
     Hue[0.046961261969965945`, 0.5, 0.7], 
     Hue[0.03765394964578421, 0.5, 0.7], 
     Hue[0.019039324997420736`, 0.5, 0.7], 
     Hue[0.9818100757006936, 0.5, 0.7], 
     Hue[0.979483247619648, 0.5, 0.7], 
     Hue[0.9771564195386025, 0.5, 0.7], 
     Hue[0.9725027633765115, 0.5, 0.7], 
     Hue[0.9631954510523297, 0.5, 0.7], 
     Hue[0.9445808264039662, 0.5, 0.7], 
     Hue[0.9073515771072393, 0.5, 0.7], 
     Hue[0.9048276688419594, 0.5, 0.7], 
     Hue[0.9023037605766788, 0.5, 0.7], 
     Hue[0.8972559440461179, 0.5, 0.7], 
     Hue[0.8871603109849967, 0.5, 0.7], 
     Hue[0.8669690448627543, 0.5, 0.7], 
     Hue[0.8265865126182694, 0.5, 0.7], 
     Hue[0.824062604352989, 0.5, 0.7], 
     Hue[0.8215386960877087, 0.5, 0.7], 
     Hue[0.8164908795571483, 0.5, 0.7], 
     Hue[0.8063952464960271, 0.5, 0.7], 
     Hue[0.7862039803737844, 0.5, 0.7], 
     Hue[0.7458214481292991, 0.5, 0.7], 
     Hue[0.7434658664368825, 0.5, 0.7], 
     Hue[0.741110284744466, 0.5, 0.7], 
     Hue[0.736399121359633, 0.5, 0.7], 
     Hue[0.7269767945899674, 0.5, 0.7], 
     Hue[0.7081321410506356, 0.5, 0.7], 
     Hue[0.6704428339719721, 0.5, 0.7], 
     Hue[0.6680872522795556, 0.5, 0.7], 
     Hue[0.665731670587139, 0.5, 0.7], 
     Hue[0.661020507202306, 0.5, 0.7], 
     Hue[0.6515981804326405, 0.5, 0.7], 
     Hue[0.6327535268933087, 0.5, 0.7], 
     Hue[0.5950642198146456, 0.5, 0.7], 
     Hue[0.5925115579379945, 0.5, 0.7], 
     Hue[0.5899588960613429, 0.5, 0.7], 
     Hue[0.5848535723080399, 0.5, 0.7], 
     Hue[0.5746429248014346, 0.5, 0.7], 
     Hue[0.5542216297882241, 0.5, 0.7], 
     Hue[0.5133790397618024, 0.5, 0.7], 
     Hue[0.5108263778851515, 0.5, 0.7], 
     Hue[0.5082737160085, 0.5, 0.7], 
     Hue[0.5031683922551968, 0.5, 0.7], 
     Hue[0.4929577447485917, 0.5, 0.7], 
     Hue[0.4725364497353812, 0.5, 0.7], 
     Hue[0.4316938597089599, 0.5, 0.7], 
     Hue[0.429187722152806, 0.5, 0.7], 
     Hue[0.4266815845966521, 0.5, 0.7], 
     Hue[0.42166930948434495`, 0.5, 0.7], 
     Hue[0.41164475925973004`, 0.5, 0.7], 
     Hue[0.39159565881050024`, 0.5, 0.7], 
     Hue[0.35149745791204057`, 0.5, 0.7], 
     Hue[0.3489913203558867, 0.5, 0.7], 
     Hue[0.34648518279973284`, 0.5, 0.7], 
     Hue[0.34147290768742566`, 0.5, 0.7], 
     Hue[0.33144835746281076`, 0.5, 0.7], 
     Hue[0.3113992570135809, 0.5, 0.7], 
     Hue[0.27130105611512123`, 0.5, 0.7], 
     Hue[0.26896324513183134`, 0.5, 0.7], 
     Hue[0.26662543414854145`, 0.5, 0.7], 
     Hue[0.26194981218196167`, 0.5, 0.7], 
     Hue[0.25259856824880217`, 0.5, 0.7], 
     Hue[0.23389608038248313`, 0.5, 0.7], 
     Hue[0.19649110464984562`, 0.5, 0.7], 
     Hue[0.19415329366655634`, 0.5, 0.7], 
     Hue[0.19181548268326584`, 0.5, 0.7], 
     Hue[0.18713986071668665`, 0.5, 0.7], 
     Hue[0.17778861678352656`, 0.5, 0.7], 
     Hue[0.1590861289172081, 0.5, 0.7], 
     Hue[0.12168115318456944`, 0.5, 0.7], 
     Hue[0.11914626201704458`, 0.5, 0.7], 
     Hue[0.11661137084952025`, 0.5, 0.7], 
     Hue[0.11154158851447046`, 0.5, 0.7], 
     Hue[0.10140202384437148`, 0.5, 0.7], 
     Hue[0.0811228945041741, 0.5, 0.7], 
     Hue[0.0405646358237776, 0.5, 0.7], 
     Hue[0.038029744656252716`, 0.5, 0.7], 
     Hue[0.035494853488728405`, 0.5, 0.7], 
     Hue[0.03042507115367863, 0.5, 0.7], 
     Hue[0.020285506483580216`, 0.5, 0.7], 
     Hue[6.377143381686833*^-6, 0.5, 0.7], 
     Hue[0.9594481184629863, 0.5, 0.7], 
     Hue[0.9570815538683255, 0.5, 0.7], 
     Hue[0.9547149892736645, 0.5, 0.7], 
     Hue[0.9499818600843427, 0.5, 0.7], 
     Hue[0.9405156017056991, 0.5, 0.7], 
     Hue[0.921583084948412, 0.5, 0.7], 
     Hue[0.8837180514338376, 0.5, 0.7], 
     Hue[0.8813514868391773, 0.5, 0.7], 
     Hue[0.8789849222445164, 0.5, 0.7], 
     Hue[0.8742517930551946, 0.5, 0.7], 
     Hue[0.8647855346765511, 0.5, 0.7], 
     Hue[0.8458530179192638, 0.5, 0.7], 
     Hue[0.8079879844046894, 0.5, 0.7], 
     Hue[0.8056679441305262, 0.5, 0.7], 
     Hue[0.8033479038563625, 0.5, 0.7], 
     Hue[0.7987078233080366, 0.5, 0.7], 
     Hue[0.7894276622113829, 0.5, 0.7], 
     Hue[0.7708673400180771, 0.5, 0.7], 
     Hue[0.7685472997439133, 0.5, 0.7], 
     Hue[0.76622725946975, 0.5, 0.7], 
     Hue[0.7615871789214236, 0.5, 0.7], 
     Hue[0.7523070178247706, 0.5, 0.7], 
     Hue[0.7337466956314647, 0.5, 0.7], 
     Hue[0.7314266553573009, 0.5, 0.7], 
     Hue[0.7291066150831382, 0.5, 0.7], 
     Hue[0.7244665345348111, 0.5, 0.7], 
     Hue[0.7151863734381587, 0.5, 0.7], 
     Hue[0.6966260512448523, 0.5, 0.7], 
     Hue[0.65950540685824, 0.5, 0.7], 
     Hue[0.6569882863998417, 0.5, 0.7], 
     Hue[0.6544711659414435, 0.5, 0.7], 
     Hue[0.6494369250246474, 0.5, 0.7], 
     Hue[0.6393684431910545, 0.5, 0.7], 
     Hue[0.6192314795238697, 0.5, 0.7], 
     Hue[0.616714359065472, 0.5, 0.7], 
     Hue[0.6141972386070738, 0.5, 0.7], 
     Hue[0.6091629976902777, 0.5, 0.7], 
     Hue[0.5990945158566849, 0.5, 0.7], 
     Hue[0.5789575521894994, 0.5, 0.7], 
     Hue[0.5764404317311012, 0.5, 0.7], 
     Hue[0.5739233112727036, 0.5, 0.7], 
     Hue[0.5688890703559071, 0.5, 0.7], 
     Hue[0.5588205885223141, 0.5, 0.7], 
     Hue[0.5386836248551292, 0.5, 0.7], 
     Hue[0.4984096975207585, 0.5, 0.7], 
     Hue[0.49606090363522365`, 0.5, 0.7], 
     Hue[0.49371210974968993`, 0.5, 0.7], 
     Hue[0.48901452197862144`, 0.5, 0.7], 
     Hue[0.4796193464364844, 0.5, 0.7], 
     Hue[0.4608289953522104, 0.5, 0.7], 
     Hue[0.45848020146667556`, 0.5, 0.7], 
     Hue[0.4561314075811413, 0.5, 0.7], 
     Hue[0.4514338198100728, 0.5, 0.7], 
     Hue[0.4420386442679352, 0.5, 0.7], 
     Hue[0.42324829318366114`, 0.5, 0.7], 
     Hue[0.4208994992981274, 0.5, 0.7], 
     Hue[0.41855070541259265`, 0.5, 0.7], 
     Hue[0.4138531176415241, 0.5, 0.7], 
     Hue[0.40445794209938707`, 0.5, 0.7], 
     Hue[0.3856675910151125, 0.5, 0.7], 
     Hue[0.3480868888465638, 0.5, 0.7], 
     Hue[0.3455410147767951, 0.5, 0.7], 
     Hue[0.34299514070702586`, 0.5, 0.7], 
     Hue[0.3379033925674879, 0.5, 0.7], 
     Hue[0.32771989628841086`, 0.5, 0.7], 
     Hue[0.30735290373025737`, 0.5, 0.7], 
     Hue[0.30480702966048806`, 0.5, 0.7], 
     Hue[0.30226115559071937`, 0.5, 0.7], 
     Hue[0.2971694074511808, 0.5, 0.7], 
     Hue[0.28698591117210376`, 0.5, 0.7], 
     Hue[0.2666189186139508, 0.5, 0.7], 
     Hue[0.26407304454418157`, 0.5, 0.7], 
     Hue[0.2615271704744123, 0.5, 0.7], 
     Hue[0.25643542233487376`, 0.5, 0.7], 
     Hue[0.2462519260557973, 0.5, 0.7], 
     Hue[0.22588493349764377`, 0.5, 0.7], 
     Hue[0.18515094838133725`, 0.5, 0.7], 
     Hue[0.185074863589463, 0.5, 0.7], 
     Hue[0.18499877879758939`, 0.5, 0.7], 
     Hue[0.18484660921384086`, 0.5, 0.7], 
     Hue[0.1845422700463445, 0.5, 0.7], 
     Hue[0.18393359171135223`, 0.5, 0.7], 
     Hue[0.18271623504136603`, 0.5, 0.7], 
     Hue[0.18264015024949232`, 0.5, 0.7], 
     Hue[0.18256406545761808`, 0.5, 0.7], 
     Hue[0.18241189587386958`, 0.5, 0.7], 
     Hue[0.18210755670637377`, 0.5, 0.7], 
     Hue[0.1814988783713804, 0.5, 0.7], 
     Hue[0.1814227935795061, 0.5, 0.7], 
     Hue[0.18134670878763243`, 0.5, 0.7], 
     Hue[0.18119453920388395`, 0.5, 0.7], 
     Hue[0.18089020003638753`, 0.5, 0.7], 
     Hue[0.1808141152445139, 0.5, 0.7], 
     Hue[0.18073803045263961`, 0.5, 0.7], 
     Hue[0.18058586086889114`, 0.5, 0.7], 
     Hue[0.18050977607701693`, 0.5, 0.7], 
     Hue[0.18043369128514264`, 0.5, 0.7], 
     Hue[0.18035760649326896`, 0.5, 0.7], 
     Hue[0.18028152170139528`, 0.5, 0.7], 
     Hue[0.8197184782986047, 0.5, 0.7], 
     Hue[0.817375481892275, 0.5, 0.7], 
     Hue[0.817375481892275, 0.5, 0.7], 
     Hue[0.8150324854859454, 0.5, 0.7], 
     Hue[0.8150324854859454, 0.5, 0.7], 
     Hue[0.8103464926732862, 0.5, 0.7], 
     Hue[0.8103464926732862, 0.5, 0.7], 
     Hue[0.8009745070479671, 0.5, 0.7], 
     Hue[0.8009745070479671, 0.5, 0.7], 
     Hue[0.7822305357973296, 0.5, 0.7], 
     Hue[0.7822305357973296, 0.5, 0.7], 
     Hue[0.779887539391, 0.5, 0.7], 
     Hue[0.779887539391, 0.5, 0.7], 
     Hue[0.7775445429846704, 0.5, 0.7], 
     Hue[0.7775445429846704, 0.5, 0.7], 
     Hue[0.7728585501720105, 0.5, 0.7], 
     Hue[0.7728585501720105, 0.5, 0.7], 
     Hue[0.763486564546692, 0.5, 0.7], 
     Hue[0.763486564546692, 0.5, 0.7], 
     Hue[0.7447425932960545, 0.5, 0.7], 
     Hue[0.7447425932960545, 0.5, 0.7], 
     Hue[0.7423995968897242, 0.5, 0.7], 
     Hue[0.7423995968897242, 0.5, 0.7], 
     Hue[0.7400566004833946, 0.5, 0.7], 
     Hue[0.7400566004833946, 0.5, 0.7], 
     Hue[0.7353706076707348, 0.5, 0.7], 
     Hue[0.7353706076707348, 0.5, 0.7], 
     Hue[0.7259986220454163, 0.5, 0.7], 
     Hue[0.7259986220454163, 0.5, 0.7], 
     Hue[0.7072546507947788, 0.5, 0.7], 
     Hue[0.7072546507947788, 0.5, 0.7], 
     Hue[0.6697667082935036, 0.5, 0.7], 
     Hue[0.6697667082935036, 0.5, 0.7], 
     Hue[0.6672266317029389, 0.5, 0.7], 
     Hue[0.6672266317029389, 0.5, 0.7], 
     Hue[0.6646865551123748, 0.5, 0.7], 
     Hue[0.6646865551123748, 0.5, 0.7], 
     Hue[0.6596064019312451, 0.5, 0.7], 
     Hue[0.6596064019312451, 0.5, 0.7], 
     Hue[0.6494460955689871, 0.5, 0.7], 
     Hue[0.6494460955689871, 0.5, 0.7], 
     Hue[0.6291254828444701, 0.5, 0.7], 
     Hue[0.6291254828444701, 0.5, 0.7], 
     Hue[0.626585406253906, 0.5, 0.7], 
     Hue[0.626585406253906, 0.5, 0.7], 
     Hue[0.6240453296633414, 0.5, 0.7], 
     Hue[0.6240453296633414, 0.5, 0.7], 
     Hue[0.6189651764822122, 0.5, 0.7], 
     Hue[0.6189651764822122, 0.5, 0.7], 
     Hue[0.6088048701199542, 0.5, 0.7], 
     Hue[0.6088048701199542, 0.5, 0.7], 
     Hue[0.5884842573954372, 0.5, 0.7], 
     Hue[0.5884842573954372, 0.5, 0.7], 
     Hue[0.5859441808048731, 0.5, 0.7], 
     Hue[0.5859441808048731, 0.5, 0.7], 
     Hue[0.5834041042143084, 0.5, 0.7], 
     Hue[0.5834041042143084, 0.5, 0.7], 
     Hue[0.5783239510331792, 0.5, 0.7], 
     Hue[0.5783239510331792, 0.5, 0.7], 
     Hue[0.5681636446709212, 0.5, 0.7], 
     Hue[0.5681636446709212, 0.5, 0.7], 
     Hue[0.5478430319464042, 0.5, 0.7], 
     Hue[0.5478430319464042, 0.5, 0.7], 
     Hue[0.5072018064973707, 0.5, 0.7], 
     Hue[0.5072018064973707, 0.5, 0.7], 
     Hue[0.5048300564796699, 0.5, 0.7], 
     Hue[0.5048300564796699, 0.5, 0.7], 
     Hue[0.5024583064619688, 0.5, 0.7], 
     Hue[0.5024583064619688, 0.5, 0.7], 
     Hue[0.4977148064265675, 0.5, 0.7], 
     Hue[0.4977148064265675, 0.5, 0.7], 
     Hue[0.48822780635576496`, 0.5, 0.7], 
     Hue[0.48822780635576496`, 0.5, 0.7], 
     Hue[0.4692538062141587, 0.5, 0.7], 
     Hue[0.4692538062141587, 0.5, 0.7], 
     Hue[0.466882056196458, 0.5, 0.7], 
     Hue[0.466882056196458, 0.5, 0.7], 
     Hue[0.4645103061787574, 0.5, 0.7], 
     Hue[0.4645103061787574, 0.5, 0.7], 
     Hue[0.45976680614335613`, 0.5, 0.7], 
     Hue[0.45976680614335613`, 0.5, 0.7], 
     Hue[0.45027980607255297`, 0.5, 0.7], 
     Hue[0.45027980607255297`, 0.5, 0.7], 
     Hue[0.43130580593094725`, 0.5, 0.7], 
     Hue[0.43130580593094725`, 0.5, 0.7], 
     Hue[0.4289340559132466, 0.5, 0.7], 
     Hue[0.4289340559132466, 0.5, 0.7], 
     Hue[0.426562305895546, 0.5, 0.7], 
     Hue[0.426562305895546, 0.5, 0.7], 
     Hue[0.42181880586014414`, 0.5, 0.7], 
     Hue[0.42181880586014414`, 0.5, 0.7], 
     Hue[0.4123318057893416, 0.5, 0.7], 
     Hue[0.4123318057893416, 0.5, 0.7], 
     Hue[0.39335780564773587`, 0.5, 0.7], 
     Hue[0.39335780564773587`, 0.5, 0.7], 
     Hue[0.3554098053645245, 0.5, 0.7], 
     Hue[0.3554098053645245, 0.5, 0.7], 
     Hue[0.3530845796673215, 0.5, 0.7], 
     Hue[0.3530845796673215, 0.5, 0.7], 
     Hue[0.35075935397011854`, 0.5, 0.7], 
     Hue[0.35075935397011854`, 0.5, 0.7], 
     Hue[0.3461089025757114, 0.5, 0.7], 
     Hue[0.3461089025757114, 0.5, 0.7], 
     Hue[0.33680799978689946`, 0.5, 0.7], 
     Hue[0.33680799978689946`, 0.5, 0.7], 
     Hue[0.3182061942092745, 0.5, 0.7], 
     Hue[0.3182061942092745, 0.5, 0.7], 
     Hue[0.3158809685120721, 0.5, 0.7], 
     Hue[0.3158809685120721, 0.5, 0.7], 
     Hue[0.31355574281486853`, 0.5, 0.7], 
     Hue[0.31355574281486853`, 0.5, 0.7], 
     Hue[0.308905291420462, 0.5, 0.7], 
     Hue[0.308905291420462, 0.5, 0.7], 
     Hue[0.2996043886316495, 0.5, 0.7], 
     Hue[0.2996043886316495, 0.5, 0.7], 
     Hue[0.2810025830540251, 0.5, 0.7], 
     Hue[0.2810025830540251, 0.5, 0.7], 
     Hue[0.2786773573568221, 0.5, 0.7], 
     Hue[0.2786773573568221, 0.5, 0.7], 
     Hue[0.2763521316596185, 0.5, 0.7], 
     Hue[0.2763521316596185, 0.5, 0.7], 
     Hue[0.2717016802652126, 0.5, 0.7], 
     Hue[0.2717016802652126, 0.5, 0.7], 
     Hue[0.26240077747640006`, 0.5, 0.7], 
     Hue[0.26240077747640006`, 0.5, 0.7], 
     Hue[0.24379897189877506`, 0.5, 0.7], 
     Hue[0.24379897189877506`, 0.5, 0.7], 
     Hue[0.2065953607435256, 0.5, 0.7], 
     Hue[0.2065953607435256, 0.5, 0.7], 
     Hue[0.20407305486208763`, 0.5, 0.7], 
     Hue[0.20407305486208763`, 0.5, 0.7], 
     Hue[0.20155074898064965`, 0.5, 0.7], 
     Hue[0.20155074898064965`, 0.5, 0.7], 
     Hue[0.19650613721777366`, 0.5, 0.7], 
     Hue[0.19650613721777366`, 0.5, 0.7], 
     Hue[0.1864169136920217, 0.5, 0.7], 
     Hue[0.1864169136920217, 0.5, 0.7], 
     Hue[0.16623846664051775`, 0.5, 0.7], 
     Hue[0.16623846664051775`, 0.5, 0.7], 
     Hue[0.16371616075907977`, 0.5, 0.7], 
     Hue[0.16371616075907977`, 0.5, 0.7], 
     Hue[0.1611938548776418, 0.5, 0.7], 
     Hue[0.1611938548776418, 0.5, 0.7], 
     Hue[0.1561492431147658, 0.5, 0.7], 
     Hue[0.1561492431147658, 0.5, 0.7], 
     Hue[0.14606001958901382`, 0.5, 0.7], 
     Hue[0.14606001958901382`, 0.5, 0.7], 
     Hue[0.12588157253750992`, 0.5, 0.7], 
     Hue[0.12588157253750992`, 0.5, 0.7], 
     Hue[0.12335926665607193`, 0.5, 0.7], 
     Hue[0.12335926665607193`, 0.5, 0.7], 
     Hue[0.12083696077463393`, 0.5, 0.7], 
     Hue[0.12083696077463393`, 0.5, 0.7], 
     Hue[0.11579234901175797`, 0.5, 0.7], 
     Hue[0.11579234901175797`, 0.5, 0.7], 
     Hue[0.105703125486006, 0.5, 0.7], 
     Hue[0.105703125486006, 0.5, 0.7], 
     Hue[0.0855246784345021, 0.5, 0.7], 
     Hue[0.0855246784345021, 0.5, 0.7], 
     Hue[0.04516778433149423, 0.5, 0.7], 
     Hue[0.04516778433149423, 0.5, 0.7], 
     Hue[0.04281380502292024, 0.5, 0.7], 
     Hue[0.04281380502292024, 0.5, 0.7], 
     Hue[0.04045982571434625, 0.5, 0.7], 
     Hue[0.04045982571434625, 0.5, 0.7], 
     Hue[0.035751867097198235`, 0.5, 0.7], 
     Hue[0.035751867097198235`, 0.5, 0.7], 
     Hue[0.026335949862901102`, 0.5, 0.7], 
     Hue[0.026335949862901102`, 0.5, 0.7], 
     Hue[0.007504115394307963, 0.5, 0.7], 
     Hue[0.007504115394307963, 0.5, 0.7], 
     Hue[0.9698404464571223, 0.5, 0.7], 
     Hue[0.9698404464571223, 0.5, 0.7], 
     Hue[0.9674864671485477, 0.5, 0.7], 
     Hue[0.9674864671485477, 0.5, 0.7], 
     Hue[0.9651324878399743, 0.5, 0.7], 
     Hue[0.9651324878399743, 0.5, 0.7], 
     Hue[0.9604245292228257, 0.5, 0.7], 
     Hue[0.9604245292228257, 0.5, 0.7], 
     Hue[0.9510086119885292, 0.5, 0.7], 
     Hue[0.9510086119885292, 0.5, 0.7], 
     Hue[0.9321767775199366, 0.5, 0.7], 
     Hue[0.9321767775199366, 0.5, 0.7], 
     Hue[0.8945131085827502, 0.5, 0.7], 
     Hue[0.8945131085827502, 0.5, 0.7], 
     Hue[0.8919620490899413, 0.5, 0.7], 
     Hue[0.8919620490899413, 0.5, 0.7], 
     Hue[0.8894109895971323, 0.5, 0.7], 
     Hue[0.8894109895971323, 0.5, 0.7], 
     Hue[0.8843088706115143, 0.5, 0.7], 
     Hue[0.8843088706115143, 0.5, 0.7], 
     Hue[0.8741046326402782, 0.5, 0.7], 
     Hue[0.8741046326402782, 0.5, 0.7], 
     Hue[0.8536961566978062, 0.5, 0.7], 
     Hue[0.8536961566978062, 0.5, 0.7], 
     Hue[0.8128792048128621, 0.5, 0.7], 
     Hue[0.8128792048128621, 0.5, 0.7], 
     Hue[0.810328145320053, 0.5, 0.7], 
     Hue[0.810328145320053, 0.5, 0.7], 
     Hue[0.8077770858272441, 0.5, 0.7], 
     Hue[0.8077770858272441, 0.5, 0.7], 
     Hue[0.802674966841626, 0.5, 0.7], 
     Hue[0.802674966841626, 0.5, 0.7], 
     Hue[0.79247072887039, 0.5, 0.7], 
     Hue[0.79247072887039, 0.5, 0.7], 
     Hue[0.772062252927918, 0.5, 0.7], 
     Hue[0.772062252927918, 0.5, 0.7], 
     Hue[0.7312453010429738, 0.5, 0.7], 
     Hue[0.7312453010429738, 0.5, 0.7], 
     Hue[0.7287407658706625, 0.5, 0.7], 
     Hue[0.7287407658706625, 0.5, 0.7], 
     Hue[0.7262362306983512, 0.5, 0.7], 
     Hue[0.7262362306983512, 0.5, 0.7], 
     Hue[0.7212271603537285, 0.5, 0.7], 
     Hue[0.7212271603537285, 0.5, 0.7], 
     Hue[0.711209019664483, 0.5, 0.7], 
     Hue[0.711209019664483, 0.5, 0.7], 
     Hue[0.6911727382859922, 0.5, 0.7], 
     Hue[0.6911727382859922, 0.5, 0.7], 
     Hue[0.6511001755290096, 0.5, 0.7], 
     Hue[0.6511001755290096, 0.5, 0.7], 
     Hue[0.6485956403566981, 0.5, 0.7], 
     Hue[0.6485956403566981, 0.5, 0.7], 
     Hue[0.646091105184387, 0.5, 0.7], 
     Hue[0.646091105184387, 0.5, 0.7], 
     Hue[0.6410820348397641, 0.5, 0.7], 
     Hue[0.6410820348397641, 0.5, 0.7], 
     Hue[0.6310638941505182, 0.5, 0.7], 
     Hue[0.6310638941505182, 0.5, 0.7], 
     Hue[0.6110276127720268, 0.5, 0.7], 
     Hue[0.6110276127720268, 0.5, 0.7], 
     Hue[0.5709550500150447, 0.5, 0.7], 
     Hue[0.5709550500150447, 0.5, 0.7], 
     Hue[0.5686188414155973, 0.5, 0.7], 
     Hue[0.5686188414155973, 0.5, 0.7], 
     Hue[0.5662826328161494, 0.5, 0.7], 
     Hue[0.5662826328161494, 0.5, 0.7], 
     Hue[0.5616102156172547, 0.5, 0.7], 
     Hue[0.5616102156172547, 0.5, 0.7], 
     Hue[0.5522653812194647, 0.5, 0.7], 
     Hue[0.5522653812194647, 0.5, 0.7], 
     Hue[0.5335757124238844, 0.5, 0.7], 
     Hue[0.5335757124238844, 0.5, 0.7], 
     Hue[0.49619637483272355`, 0.5, 0.7], 
     Hue[0.49619637483272355`, 0.5, 0.7], 
     Hue[0.49386016623327594`, 0.5, 0.7], 
     Hue[0.49386016623327594`, 0.5, 0.7], 
     Hue[0.49152395763382856`, 0.5, 0.7], 
     Hue[0.49152395763382856`, 0.5, 0.7], 
     Hue[0.48685154043493384`, 0.5, 0.7], 
     Hue[0.48685154043493384`, 0.5, 0.7], 
     Hue[0.47750670603714357`, 0.5, 0.7], 
     Hue[0.47750670603714357`, 0.5, 0.7], 
     Hue[0.458817037241563, 0.5, 0.7], 
     Hue[0.458817037241563, 0.5, 0.7], 
     Hue[0.4214376996504027, 0.5, 0.7], 
     Hue[0.4214376996504027, 0.5, 0.7], 
     Hue[0.4189044108667203, 0.5, 0.7], 
     Hue[0.4189044108667203, 0.5, 0.7], 
     Hue[0.41637112208303795`, 0.5, 0.7], 
     Hue[0.41637112208303795`, 0.5, 0.7], 
     Hue[0.4113045445156732, 0.5, 0.7], 
     Hue[0.4113045445156732, 0.5, 0.7], 
     Hue[0.4011713893809435, 0.5, 0.7], 
     Hue[0.4011713893809435, 0.5, 0.7], 
     Hue[0.38090507911148397`, 0.5, 0.7], 
     Hue[0.38090507911148397`, 0.5, 0.7], 
     Hue[0.34037245857256526`, 0.5, 0.7], 
     Hue[0.34037245857256526`, 0.5, 0.7], 
     Hue[0.3378391697888829, 0.5, 0.7], 
     Hue[0.3378391697888829, 0.5, 0.7], 
     Hue[0.33530588100520053`, 0.5, 0.7], 
     Hue[0.33530588100520053`, 0.5, 0.7], 
     Hue[0.3302393034378358, 0.5, 0.7], 
     Hue[0.3302393034378358, 0.5, 0.7], 
     Hue[0.3201061483031061, 0.5, 0.7], 
     Hue[0.3201061483031061, 0.5, 0.7], 
     Hue[0.2998398380336466, 0.5, 0.7], 
     Hue[0.2998398380336466, 0.5, 0.7], 
     Hue[0.2593072174947282, 0.5, 0.7], 
     Hue[0.2593072174947282, 0.5, 0.7], 
     Hue[0.2569422552839095, 0.5, 0.7], 
     Hue[0.2569422552839095, 0.5, 0.7], 
     Hue[0.2545772930730911, 0.5, 0.7], 
     Hue[0.2545772930730911, 0.5, 0.7], 
     Hue[0.24984736865145438`, 0.5, 0.7], 
     Hue[0.24984736865145438`, 0.5, 0.7], 
     Hue[0.24038751980818002`, 0.5, 0.7], 
     Hue[0.24038751980818002`, 0.5, 0.7], 
     Hue[0.22146782212163132`, 0.5, 0.7], 
     Hue[0.22146782212163132`, 0.5, 0.7], 
     Hue[0.18362842674853475`, 0.5, 0.7], 
     Hue[0.18362842674853475`, 0.5, 0.7], 
     Hue[0.18126346453771638`, 0.5, 0.7], 
     Hue[0.18126346453771638`, 0.5, 0.7], 
     Hue[0.17889850232689772`, 0.5, 0.7], 
     Hue[0.17889850232689772`, 0.5, 0.7], 
     Hue[0.17416857790526039`, 0.5, 0.7], 
     Hue[0.17416857790526039`, 0.5, 0.7], 
     Hue[0.16470872906198633`, 0.5, 0.7], 
     Hue[0.16470872906198633`, 0.5, 0.7], 
     Hue[0.14578903137543817`, 0.5, 0.7], 
     Hue[0.14578903137543817`, 0.5, 0.7], 
     Hue[0.10794963600234132`, 0.5, 0.7], 
     Hue[0.10794963600234132`, 0.5, 0.7], 
     Hue[0.10563119811202033`, 0.5, 0.7], 
     Hue[0.10563119811202033`, 0.5, 0.7], 
     Hue[0.10331276022169904`, 0.5, 0.7], 
     Hue[0.10331276022169904`, 0.5, 0.7], 
     Hue[0.09867588444105733, 0.5, 0.7], 
     Hue[0.09867588444105733, 0.5, 0.7], 
     Hue[0.08940213287977387, 0.5, 0.7], 
     Hue[0.08940213287977387, 0.5, 0.7], 
     Hue[0.07085462975720616, 0.5, 0.7], 
     Hue[0.07085462975720616, 0.5, 0.7], 
     Hue[0.03375962351207098, 0.5, 0.7], 
     Hue[0.03375962351207098, 0.5, 0.7], 
     Hue[0.03144118562175026, 0.5, 0.7], 
     Hue[0.03144118562175026, 0.5, 0.7], 
     Hue[0.029122747731428977`, 0.5, 0.7], 
     Hue[0.029122747731428977`, 0.5, 0.7], 
     Hue[0.024485871950787247`, 0.5, 0.7], 
     Hue[0.024485871950787247`, 0.5, 0.7], 
     Hue[0.015212120389503814`, 0.5, 0.7], 
     Hue[0.015212120389503814`, 0.5, 0.7], 
     Hue[0.9966646172669361, 0.5, 0.7], 
     Hue[0.9966646172669361, 0.5, 0.7], 
     Hue[0.9595696110218012, 0.5, 0.7], 
     Hue[0.9595696110218012, 0.5, 0.7], 
     Hue[0.9570540929472452, 0.5, 0.7], 
     Hue[0.9570540929472452, 0.5, 0.7], 
     Hue[0.9545385748726898, 0.5, 0.7], 
     Hue[0.9545385748726898, 0.5, 0.7], 
     Hue[0.949507538723578, 0.5, 0.7], 
     Hue[0.949507538723578, 0.5, 0.7], 
     Hue[0.9394454664253545, 0.5, 0.7], 
     Hue[0.9394454664253545, 0.5, 0.7], 
     Hue[0.9193213218289079, 0.5, 0.7], 
     Hue[0.9193213218289079, 0.5, 0.7], 
     Hue[0.8790730326360152, 0.5, 0.7], 
     Hue[0.8790730326360152, 0.5, 0.7], 
     Hue[0.7985764542502292, 0.5, 0.7], 
     Hue[0.7985764542502292, 0.5, 0.7], 
     Hue[0.7962292627485371, 0.5, 0.7], 
     Hue[0.7962292627485371, 0.5, 0.7], 
     Hue[0.7938820712468451, 0.5, 0.7], 
     Hue[0.7938820712468451, 0.5, 0.7], 
     Hue[0.7891876882434611, 0.5, 0.7], 
     Hue[0.7891876882434611, 0.5, 0.7], 
     Hue[0.7797989222366932, 0.5, 0.7], 
     Hue[0.7797989222366932, 0.5, 0.7], 
     Hue[0.7610213902231577, 0.5, 0.7], 
     Hue[0.7610213902231577, 0.5, 0.7], 
     Hue[0.7234663261960863, 0.5, 0.7], 
     Hue[0.7234663261960863, 0.5, 0.7], 
     Hue[0.6483561981419439, 0.5, 0.7], 
     Hue[0.6483561981419439, 0.5, 0.7], 
     Hue[0.6458119264560171, 0.5, 0.7], 
     Hue[0.6458119264560171, 0.5, 0.7], 
     Hue[0.6432676547700904, 0.5, 0.7], 
     Hue[0.6432676547700904, 0.5, 0.7], 
     Hue[0.6381791113982367, 0.5, 0.7], 
     Hue[0.6381791113982367, 0.5, 0.7], 
     Hue[0.6280020246545293, 0.5, 0.7], 
     Hue[0.6280020246545293, 0.5, 0.7], 
     Hue[0.6076478511671144, 0.5, 0.7], 
     Hue[0.6076478511671144, 0.5, 0.7], 
     Hue[0.5669395041922853, 0.5, 0.7], 
     Hue[0.5669395041922853, 0.5, 0.7], 
     Hue[0.4855228102426269, 0.5, 0.7], 
     Hue[0.4855228102426269, 0.5, 0.7], 
     Hue[0.3256669788551567, 0.5, 0.7], 
     Hue[0.3256669788551567, 0.5, 0.7], 
     Hue[0.17658404813097406`, 0.5, 0.7], 
     Hue[0.17658404813097406`, 0.5, 0.7], 
     Hue[0.014887985615758862`, 0.5, 0.7], 
     Hue[0.014887985615758862`, 0.5, 0.7], 
     Hue[0.8639648237638307, 0.5, 0.7], 
     Hue[0.8639648237638307, 0.5, 0.7], 
     Hue[0.7004285301208703, 0.5, 0.7], 
     Hue[0.7004285301208703, 0.5, 0.7], 
     Hue[0.6979197998531967, 0.5, 0.7], 
     Hue[0.6979197998531967, 0.5, 0.7], 
     Hue[0.6954110695855231, 0.5, 0.7], 
     Hue[0.6954110695855231, 0.5, 0.7], 
     Hue[0.6903936090501757, 0.5, 0.7], 
     Hue[0.6903936090501757, 0.5, 0.7], 
     Hue[0.6803586879794812, 0.5, 0.7], 
     Hue[0.6803586879794812, 0.5, 0.7], 
     Hue[0.660288845838092, 0.5, 0.7], 
     Hue[0.660288845838092, 0.5, 0.7], 
     Hue[0.6201491615553137, 0.5, 0.7], 
     Hue[0.6201491615553137, 0.5, 0.7], 
     Hue[0.5398697929897572, 0.5, 0.7], 
     Hue[0.5398697929897572, 0.5, 0.7], 
     Hue[0.5375293892949474, 0.5, 0.7], 
     Hue[0.5375293892949474, 0.5, 0.7], 
     Hue[0.5351889856001377, 0.5, 0.7], 
     Hue[0.5351889856001377, 0.5, 0.7], 
     Hue[0.5305081782105181, 0.5, 0.7], 
     Hue[0.5305081782105181, 0.5, 0.7], 
     Hue[0.5211465634312791, 0.5, 0.7], 
     Hue[0.5211465634312791, 0.5, 0.7], 
     Hue[0.5024233338728007, 0.5, 0.7], 
     Hue[0.5024233338728007, 0.5, 0.7], 
     Hue[0.46497687475584426`, 0.5, 0.7], 
     Hue[0.46497687475584426`, 0.5, 0.7], 
     Hue[0.4626364710610345, 0.5, 0.7], 
     Hue[0.4626364710610345, 0.5, 0.7], 
     Hue[0.46029606736622475`, 0.5, 0.7], 
     Hue[0.46029606736622475`, 0.5, 0.7], 
     Hue[0.45561525997660507`, 0.5, 0.7], 
     Hue[0.45561525997660507`, 0.5, 0.7], 
     Hue[0.446253645197366, 0.5, 0.7], 
     Hue[0.446253645197366, 0.5, 0.7], 
     Hue[0.4275304156388878, 0.5, 0.7], 
     Hue[0.4275304156388878, 0.5, 0.7], 
     Hue[0.3900839565219313, 0.5, 0.7], 
     Hue[0.3900839565219313, 0.5, 0.7], 
     Hue[0.38754647264288655`, 0.5, 0.7], 
     Hue[0.38754647264288655`, 0.5, 0.7], 
     Hue[0.38500898876384193`, 0.5, 0.7], 
     Hue[0.38500898876384193`, 0.5, 0.7], 
     Hue[0.37993402100575263`, 0.5, 0.7], 
     Hue[0.37993402100575263`, 0.5, 0.7], 
     Hue[0.3697840854895739, 0.5, 0.7], 
     Hue[0.3697840854895739, 0.5, 0.7], 
     Hue[0.3494842144572167, 0.5, 0.7], 
     Hue[0.3494842144572167, 0.5, 0.7], 
     Hue[0.34694673057817194`, 0.5, 0.7], 
     Hue[0.34694673057817194`, 0.5, 0.7], 
     Hue[0.34440924669912737`, 0.5, 0.7], 
     Hue[0.34440924669912737`, 0.5, 0.7], 
     Hue[0.339334278941038, 0.5, 0.7], 
     Hue[0.339334278941038, 0.5, 0.7], 
     Hue[0.32918434342485936`, 0.5, 0.7], 
     Hue[0.32918434342485936`, 0.5, 0.7], 
     Hue[0.3088844723925021, 0.5, 0.7], 
     Hue[0.3088844723925021, 0.5, 0.7], 
     Hue[0.30634698851345743`, 0.5, 0.7], 
     Hue[0.30634698851345743`, 0.5, 0.7], 
     Hue[0.30380950463441275`, 0.5, 0.7], 
     Hue[0.30380950463441275`, 0.5, 0.7], 
     Hue[0.29873453687632345`, 0.5, 0.7], 
     Hue[0.29873453687632345`, 0.5, 0.7], 
     Hue[0.2885846013601448, 0.5, 0.7], 
     Hue[0.2885846013601448, 0.5, 0.7], 
     Hue[0.2682847303277875, 0.5, 0.7], 
     Hue[0.2682847303277875, 0.5, 0.7], 
     Hue[0.22768498826307296`, 0.5, 0.7], 
     Hue[0.22768498826307296`, 0.5, 0.7], 
     Hue[0.2253158309568922, 0.5, 0.7], 
     Hue[0.2253158309568922, 0.5, 0.7], 
     Hue[0.22294667365071136`, 0.5, 0.7], 
     Hue[0.22294667365071136`, 0.5, 0.7], 
     Hue[0.21820835903834976`, 0.5, 0.7], 
     Hue[0.21820835903834976`, 0.5, 0.7], 
     Hue[0.20873172981362653`, 0.5, 0.7], 
     Hue[0.20873172981362653`, 0.5, 0.7], 
     Hue[0.18977847136418016`, 0.5, 0.7], 
     Hue[0.18977847136418016`, 0.5, 0.7], 
     Hue[0.18740931405799935`, 0.5, 0.7], 
     Hue[0.18740931405799935`, 0.5, 0.7], 
     Hue[0.18504015675181856`, 0.5, 0.7], 
     Hue[0.18504015675181856`, 0.5, 0.7], 
     Hue[0.18030184213945694`, 0.5, 0.7], 
     Hue[0.18030184213945694`, 0.5, 0.7], 
     Hue[0.17082521291473377`, 0.5, 0.7], 
     Hue[0.17082521291473377`, 0.5, 0.7], 
     Hue[0.1518719544652873, 0.5, 0.7], 
     Hue[0.1518719544652873, 0.5, 0.7], 
     Hue[0.14950279715910653`, 0.5, 0.7], 
     Hue[0.14950279715910653`, 0.5, 0.7], 
     Hue[0.14713363985292574`, 0.5, 0.7], 
     Hue[0.14713363985292574`, 0.5, 0.7], 
     Hue[0.1423953252405641, 0.5, 0.7], 
     Hue[0.1423953252405641, 0.5, 0.7], 
     Hue[0.1329186960158409, 0.5, 0.7], 
     Hue[0.1329186960158409, 0.5, 0.7], 
     Hue[0.13054953870966016`, 0.5, 0.7], 
     Hue[0.13054953870966016`, 0.5, 0.7], 
     Hue[0.12818038140347934`, 0.5, 0.7], 
     Hue[0.12818038140347934`, 0.5, 0.7], 
     Hue[0.12344206679111772`, 0.5, 0.7], 
     Hue[0.12344206679111772`, 0.5, 0.7], 
     Hue[0.11396543756639452`, 0.5, 0.7], 
     Hue[0.11396543756639452`, 0.5, 0.7], 
     Hue[0.11159628026021372`, 0.5, 0.7], 
     Hue[0.11159628026021372`, 0.5, 0.7], 
     Hue[0.1092271229540329, 0.5, 0.7], 
     Hue[0.1092271229540329, 0.5, 0.7], 
     Hue[0.10448880834167132`, 0.5, 0.7], 
     Hue[0.10448880834167132`, 0.5, 0.7], 
     Hue[0.09501217911694812, 0.5, 0.7], 
     Hue[0.09501217911694812, 0.5, 0.7], 
     Hue[0.0926430218107673, 0.5, 0.7], 
     Hue[0.0926430218107673, 0.5, 0.7], 
     Hue[0.09027386450458652, 0.5, 0.7], 
     Hue[0.09027386450458652, 0.5, 0.7], 
     Hue[0.08553554989222491, 0.5, 0.7], 
     Hue[0.08553554989222491, 0.5, 0.7], 
     Hue[0.07605892066750171, 0.5, 0.7], 
     Hue[0.07605892066750171, 0.5, 0.7], 
     Hue[0.07373628768181852, 0.5, 0.7], 
     Hue[0.07373628768181852, 0.5, 0.7], 
     Hue[0.07141365469613532, 0.5, 0.7], 
     Hue[0.07141365469613532, 0.5, 0.7], 
     Hue[0.06676838872476895, 0.5, 0.7], 
     Hue[0.06676838872476895, 0.5, 0.7], 
     Hue[0.06444575573908579, 0.5, 0.7], 
     Hue[0.06444575573908579, 0.5, 0.7], 
     Hue[0.06212312275340258, 0.5, 0.7], 
     Hue[0.06212312275340258, 0.5, 0.7], 
     Hue[0.05747785678203621, 0.5, 0.7], 
     Hue[0.05747785678203621, 0.5, 0.7], 
     Hue[0.05515522379635303, 0.5, 0.7], 
     Hue[0.05515522379635303, 0.5, 0.7], 
     Hue[0.052832590810669836`, 0.5, 0.7], 
     Hue[0.052832590810669836`, 0.5, 0.7], 
     Hue[0.04818732483930348, 0.5, 0.7], 
     Hue[0.04818732483930348, 0.5, 0.7], 
     Hue[0.04586469185362028, 0.5, 0.7], 
     Hue[0.04586469185362028, 0.5, 0.7], 
     Hue[0.043542058867937085`, 0.5, 0.7], 
     Hue[0.043542058867937085`, 0.5, 0.7], 
     Hue[0.03889679289657071, 0.5, 0.7], 
     Hue[0.03889679289657071, 0.5, 0.7], 
     Hue[0.03657415991088753, 0.5, 0.7], 
     Hue[0.03657415991088753, 0.5, 0.7], 
     Hue[0.03425152692520435, 0.5, 0.7], 
     Hue[0.03425152692520435, 0.5, 0.7], 
     Hue[0.031928893939521155`, 0.5, 0.7], 
     Hue[0.031928893939521155`, 0.5, 0.7], 
     Hue[0.029606260953837966`, 0.5, 0.7], 
     Hue[0.029606260953837966`, 0.5, 0.7], 
     Hue[0.02728362796815478, 0.5, 0.7], 
     Hue[0.02728362796815478, 0.5, 0.7], 
     Hue[0.024960994982471587`, 0.5, 0.7], 
     Hue[0.024960994982471587`, 0.5, 0.7], 
     Hue[0.0226383619967884, 0.5, 0.7], 
     Hue[0.0226383619967884, 0.5, 0.7], 
     Hue[0.02031572901110522, 0.5, 0.7], 
     Hue[0.02031572901110522, 0.5, 0.7], 
     Hue[0.017993096025422036`, 0.5, 0.7], 
     Hue[0.017993096025422036`, 0.5, 0.7], 
     Hue[0.015670463039738843`, 0.5, 0.7], 
     Hue[0.015670463039738843`, 0.5, 0.7], 
     Hue[0.013347830054055657`, 0.5, 0.7], 
     Hue[0.013347830054055657`, 0.5, 0.7], 
     Hue[0.011025197068372471`, 0.5, 0.7], 
     Hue[0.011025197068372471`, 0.5, 0.7], 
     Hue[0.008702564082689282, 0.5, 0.7], 
     Hue[0.008702564082689282, 0.5, 0.7], 
     Hue[0.006379931097006096, 0.5, 0.7], 
     Hue[0.006379931097006096, 0.5, 0.7], 
     Hue[0.004057298111322911, 0.5, 0.7], 
     Hue[0.004057298111322911, 0.5, 0.7], 
     Hue[0.0017346651256397233`, 0.5, 0.7], 
     Hue[0.0017346651256397233`, 0.5, 0.7], 
     Hue[0.9994120321399566, 0.5, 0.7], 
     Hue[0.9994120321399566, 0.5, 0.7], 
     Hue[0.9970893991542733, 0.5, 0.7], 
     Hue[0.9970893991542733, 0.5, 0.7], 
     Hue[0.9947667661685902, 0.5, 0.7], 
     Hue[0.9947667661685902, 0.5, 0.7], 
     Hue[0.992444133182907, 0.5, 0.7], 
     Hue[0.992444133182907, 0.5, 0.7], 
     Hue[0.9901215001972238, 0.5, 0.7], 
     Hue[0.9901215001972238, 0.5, 0.7], 
     Hue[0.9877988672115406, 0.5, 0.7], 
     Hue[0.9877988672115406, 0.5, 0.7], 
     Hue[0.9854762342258574, 0.5, 0.7], 
     Hue[0.9854762342258574, 0.5, 0.7], 
     Hue[0.9831536012401743, 0.5, 0.7], 
     Hue[0.9831536012401743, 0.5, 0.7], 
     Hue[0.980830968254491, 0.5, 0.7], 
     Hue[0.980830968254491, 0.5, 0.7], 
     Hue[0.9785083352688079, 0.5, 0.7], 
     Hue[0.9785083352688079, 0.5, 0.7], 
     Hue[0.9761857022831246, 0.5, 0.7], 
     Hue[0.9761857022831246, 0.5, 0.7], 
     Hue[0.9738630692974415, 0.5, 0.7], 
     Hue[0.9738630692974415, 0.5, 0.7], 
     Hue[0.9715404363117582, 0.5, 0.7], 
     Hue[0.9715404363117582, 0.5, 0.7], 
     Hue[0.9692178033260751, 0.5, 0.7], 
     Hue[0.9692178033260751, 0.5, 0.7], 
     Hue[0.9645725373547087, 0.5, 0.7], 
     Hue[0.9645725373547087, 0.5, 0.7], 
     Hue[0.9622499043690256, 0.5, 0.7], 
     Hue[0.9622499043690256, 0.5, 0.7], 
     Hue[0.9599272713833423, 0.5, 0.7], 
     Hue[0.9599272713833423, 0.5, 0.7], 
     Hue[0.9552820054119759, 0.5, 0.7], 
     Hue[0.9552820054119759, 0.5, 0.7], 
     Hue[0.9529593724262928, 0.5, 0.7], 
     Hue[0.9529593724262928, 0.5, 0.7], 
     Hue[0.9506367394406096, 0.5, 0.7], 
     Hue[0.9506367394406096, 0.5, 0.7], 
     Hue[0.9459914734692433, 0.5, 0.7], 
     Hue[0.9459914734692433, 0.5, 0.7], 
     Hue[0.94366884048356, 0.5, 0.7], 
     Hue[0.94366884048356, 0.5, 0.7], 
     Hue[0.9413462074978769, 0.5, 0.7], 
     Hue[0.9413462074978769, 0.5, 0.7], 
     Hue[0.9367009415265105, 0.5, 0.7], 
     Hue[0.9367009415265105, 0.5, 0.7], 
     Hue[0.9274104095837777, 0.5, 0.7], 
     Hue[0.9274104095837777, 0.5, 0.7], 
     Hue[0.9248906964138597, 0.5, 0.7], 
     Hue[0.9248906964138597, 0.5, 0.7], 
     Hue[0.9223709832439416, 0.5, 0.7], 
     Hue[0.9223709832439416, 0.5, 0.7], 
     Hue[0.9173315569041055, 0.5, 0.7], 
     Hue[0.9173315569041055, 0.5, 0.7], 
     Hue[0.9072527042244332, 0.5, 0.7], 
     Hue[0.9072527042244332, 0.5, 0.7], 
     Hue[0.9047329910545151, 0.5, 0.7], 
     Hue[0.9047329910545151, 0.5, 0.7], 
     Hue[0.902213277884597, 0.5, 0.7], 
     Hue[0.902213277884597, 0.5, 0.7], 
     Hue[0.8971738515447609, 0.5, 0.7], 
     Hue[0.8971738515447609, 0.5, 0.7], 
     Hue[0.8870949988650887, 0.5, 0.7], 
     Hue[0.8870949988650887, 0.5, 0.7], 
     Hue[0.8845752856951706, 0.5, 0.7], 
     Hue[0.8845752856951706, 0.5, 0.7], 
     Hue[0.8820555725252526, 0.5, 0.7], 
     Hue[0.8820555725252526, 0.5, 0.7], 
     Hue[0.8770161461854165, 0.5, 0.7], 
     Hue[0.8770161461854165, 0.5, 0.7], 
     Hue[0.8669372935057441, 0.5, 0.7], 
     Hue[0.8669372935057441, 0.5, 0.7], 
     Hue[0.8467795881463996, 0.5, 0.7], 
     Hue[0.8467795881463996, 0.5, 0.7], 
     Hue[0.8442598749764816, 0.5, 0.7], 
     Hue[0.8442598749764816, 0.5, 0.7], 
     Hue[0.8417401618065634, 0.5, 0.7], 
     Hue[0.8417401618065634, 0.5, 0.7], 
     Hue[0.8367007354667274, 0.5, 0.7], 
     Hue[0.8367007354667274, 0.5, 0.7], 
     Hue[0.8266218827870551, 0.5, 0.7], 
     Hue[0.8266218827870551, 0.5, 0.7], 
     Hue[0.8064641774277106, 0.5, 0.7], 
     Hue[0.8064641774277106, 0.5, 0.7], 
     Hue[0.8039444642577924, 0.5, 0.7], 
     Hue[0.8039444642577924, 0.5, 0.7], 
     Hue[0.8014247510878745, 0.5, 0.7], 
     Hue[0.8014247510878745, 0.5, 0.7], 
     Hue[0.7963853247480384, 0.5, 0.7], 
     Hue[0.7963853247480384, 0.5, 0.7], 
     Hue[0.786306472068366, 0.5, 0.7], 
     Hue[0.786306472068366, 0.5, 0.7], 
     Hue[0.7661487667090214, 0.5, 0.7], 
     Hue[0.7661487667090214, 0.5, 0.7], 
     Hue[0.7637973801119673, 0.5, 0.7], 
     Hue[0.7637973801119673, 0.5, 0.7], 
     Hue[0.7614459935149132, 0.5, 0.7], 
     Hue[0.7614459935149132, 0.5, 0.7], 
     Hue[0.7567432203208047, 0.5, 0.7], 
     Hue[0.7567432203208047, 0.5, 0.7], 
     Hue[0.7473376739325879, 0.5, 0.7], 
     Hue[0.7473376739325879, 0.5, 0.7], 
     Hue[0.7285265811561541, 0.5, 0.7], 
     Hue[0.7285265811561541, 0.5, 0.7], 
     Hue[0.690904395603287, 0.5, 0.7], 
     Hue[0.690904395603287, 0.5, 0.7], 
     Hue[0.6885530090062326, 0.5, 0.7], 
     Hue[0.6885530090062326, 0.5, 0.7], 
     Hue[0.6862016224091785, 0.5, 0.7], 
     Hue[0.6862016224091785, 0.5, 0.7], 
     Hue[0.68149884921507, 0.5, 0.7], 
     Hue[0.68149884921507, 0.5, 0.7], 
     Hue[0.6720933028268532, 0.5, 0.7], 
     Hue[0.6720933028268532, 0.5, 0.7], 
     Hue[0.6532822100504196, 0.5, 0.7], 
     Hue[0.6532822100504196, 0.5, 0.7], 
     Hue[0.6156600244975523, 0.5, 0.7], 
     Hue[0.6156600244975523, 0.5, 0.7], 
     Hue[0.6131115577162632, 0.5, 0.7], 
     Hue[0.6131115577162632, 0.5, 0.7], 
     Hue[0.6105630909349742, 0.5, 0.7], 
     Hue[0.6105630909349742, 0.5, 0.7], 
     Hue[0.605466157372396, 0.5, 0.7], 
     Hue[0.605466157372396, 0.5, 0.7], 
     Hue[0.5952722902472397, 0.5, 0.7], 
     Hue[0.5952722902472397, 0.5, 0.7], 
     Hue[0.5748845559969271, 0.5, 0.7], 
     Hue[0.5748845559969271, 0.5, 0.7], 
     Hue[0.5341090874963016, 0.5, 0.7], 
     Hue[0.5341090874963016, 0.5, 0.7], 
     Hue[0.45255815049505094`, 0.5, 0.7], 
     Hue[0.45255815049505094`, 0.5, 0.7], 
     Hue[0.4500562080342596, 0.5, 0.7], 
     Hue[0.4500562080342596, 0.5, 0.7], 
     Hue[0.4475542655734681, 0.5, 0.7], 
     Hue[0.4475542655734681, 0.5, 0.7], 
     Hue[0.4425503806518852, 0.5, 0.7], 
     Hue[0.4425503806518852, 0.5, 0.7], 
     Hue[0.4325426108087192, 0.5, 0.7], 
     Hue[0.4325426108087192, 0.5, 0.7], 
     Hue[0.41252707112238746`, 0.5, 0.7], 
     Hue[0.41252707112238746`, 0.5, 0.7], 
     Hue[0.372495991749724, 0.5, 0.7], 
     Hue[0.372495991749724, 0.5, 0.7], 
     Hue[0.2924338330043968, 0.5, 0.7], 
     Hue[0.2924338330043968, 0.5, 0.7], 
     Hue[0.14308241617702985`, 0.5, 0.7], 
     Hue[0.14308241617702985`, 0.5, 0.7], 
     Hue[0.9811178675586304, 0.5, 0.7], 
     Hue[0.9811178675586304, 0.5, 0.7], 
     Hue[0.829926219603518, 0.5, 0.7], 
     Hue[0.829926219603518, 0.5, 0.7], 
     Hue[0.681712128160253, 0.5, 0.7], 
     Hue[0.681712128160253, 0.5, 0.7], 
     Hue[0.520884904925956, 0.5, 0.7], 
     Hue[0.520884904925956, 0.5, 0.7], 
     Hue[0.5185403061357836, 0.5, 0.7], 
     Hue[0.5185403061357836, 0.5, 0.7], 
     Hue[0.5161957073456116, 0.5, 0.7], 
     Hue[0.5161957073456116, 0.5, 0.7], 
     Hue[0.5115065097652677, 0.5, 0.7], 
     Hue[0.5115065097652677, 0.5, 0.7], 
     Hue[0.5021281146045796, 0.5, 0.7], 
     Hue[0.5021281146045796, 0.5, 0.7], 
     Hue[0.48337132428320306`, 0.5, 0.7], 
     Hue[0.48337132428320306`, 0.5, 0.7], 
     Hue[0.4458577436404507, 0.5, 0.7], 
     Hue[0.4458577436404507, 0.5, 0.7], 
     Hue[0.3708305823549457, 0.5, 0.7], 
     Hue[0.3708305823549457, 0.5, 0.7], 
     Hue[0.3682889033805386, 0.5, 0.7], 
     Hue[0.3682889033805386, 0.5, 0.7], 
     Hue[0.365747224406132, 0.5, 0.7], 
     Hue[0.365747224406132, 0.5, 0.7], 
     Hue[0.36066386645731824`, 0.5, 0.7], 
     Hue[0.36066386645731824`, 0.5, 0.7], 
     Hue[0.3504971505596905, 0.5, 0.7], 
     Hue[0.3504971505596905, 0.5, 0.7], 
     Hue[0.33016371876443495`, 0.5, 0.7], 
     Hue[0.33016371876443495`, 0.5, 0.7], 
     Hue[0.2894968551739242, 0.5, 0.7], 
     Hue[0.2894968551739242, 0.5, 0.7], 
     Hue[0.2081631279929033, 0.5, 0.7], 
     Hue[0.2081631279929033, 0.5, 0.7], 
     Hue[0.2057897755913601, 0.5, 0.7], 
     Hue[0.2057897755913601, 0.5, 0.7], 
     Hue[0.20341642318981695`, 0.5, 0.7], 
     Hue[0.20341642318981695`, 0.5, 0.7], 
     Hue[0.19866971838673092`, 0.5, 0.7], 
     Hue[0.19866971838673092`, 0.5, 0.7], 
     Hue[0.18917630878055883`, 0.5, 0.7], 
     Hue[0.18917630878055883`, 0.5, 0.7], 
     Hue[0.1701894895682144, 0.5, 0.7], 
     Hue[0.1701894895682144, 0.5, 0.7], 
     Hue[0.1322158511435255, 0.5, 0.7], 
     Hue[0.1322158511435255, 0.5, 0.7], 
     Hue[0.1298424987419823, 0.5, 0.7], 
     Hue[0.1298424987419823, 0.5, 0.7], 
     Hue[0.12746914634043943`, 0.5, 0.7], 
     Hue[0.12746914634043943`, 0.5, 0.7], 
     Hue[0.12272244153735341`, 0.5, 0.7], 
     Hue[0.12272244153735341`, 0.5, 0.7], 
     Hue[0.11322903193118104`, 0.5, 0.7], 
     Hue[0.11322903193118104`, 0.5, 0.7], 
     Hue[0.09424221271883658, 0.5, 0.7], 
     Hue[0.09424221271883658, 0.5, 0.7], 
     Hue[0.056268574294147965`, 0.5, 0.7], 
     Hue[0.056268574294147965`, 0.5, 0.7], 
     Hue[0.053941746213102444`, 0.5, 0.7], 
     Hue[0.053941746213102444`, 0.5, 0.7], 
     Hue[0.05161491813205696, 0.5, 0.7], 
     Hue[0.05161491813205696, 0.5, 0.7], 
     Hue[0.046961261969965945`, 0.5, 0.7], 
     Hue[0.046961261969965945`, 0.5, 0.7], 
     Hue[0.03765394964578421, 0.5, 0.7], 
     Hue[0.03765394964578421, 0.5, 0.7], 
     Hue[0.019039324997420736`, 0.5, 0.7], 
     Hue[0.019039324997420736`, 0.5, 0.7], 
     Hue[0.9818100757006936, 0.5, 0.7], 
     Hue[0.9818100757006936, 0.5, 0.7], 
     Hue[0.979483247619648, 0.5, 0.7], 
     Hue[0.979483247619648, 0.5, 0.7], 
     Hue[0.9771564195386025, 0.5, 0.7], 
     Hue[0.9771564195386025, 0.5, 0.7], 
     Hue[0.9725027633765115, 0.5, 0.7], 
     Hue[0.9725027633765115, 0.5, 0.7], 
     Hue[0.9631954510523297, 0.5, 0.7], 
     Hue[0.9631954510523297, 0.5, 0.7], 
     Hue[0.9445808264039662, 0.5, 0.7], 
     Hue[0.9445808264039662, 0.5, 0.7], 
     Hue[0.9073515771072393, 0.5, 0.7], 
     Hue[0.9073515771072393, 0.5, 0.7], 
     Hue[0.9048276688419594, 0.5, 0.7], 
     Hue[0.9048276688419594, 0.5, 0.7], 
     Hue[0.9023037605766788, 0.5, 0.7], 
     Hue[0.9023037605766788, 0.5, 0.7], 
     Hue[0.8972559440461179, 0.5, 0.7], 
     Hue[0.8972559440461179, 0.5, 0.7], 
     Hue[0.8871603109849967, 0.5, 0.7], 
     Hue[0.8871603109849967, 0.5, 0.7], 
     Hue[0.8669690448627543, 0.5, 0.7], 
     Hue[0.8669690448627543, 0.5, 0.7], 
     Hue[0.8265865126182694, 0.5, 0.7], 
     Hue[0.8265865126182694, 0.5, 0.7], 
     Hue[0.824062604352989, 0.5, 0.7], 
     Hue[0.824062604352989, 0.5, 0.7], 
     Hue[0.8215386960877087, 0.5, 0.7], 
     Hue[0.8215386960877087, 0.5, 0.7], 
     Hue[0.8164908795571483, 0.5, 0.7], 
     Hue[0.8164908795571483, 0.5, 0.7], 
     Hue[0.8063952464960271, 0.5, 0.7], 
     Hue[0.8063952464960271, 0.5, 0.7], 
     Hue[0.7862039803737844, 0.5, 0.7], 
     Hue[0.7862039803737844, 0.5, 0.7], 
     Hue[0.7458214481292991, 0.5, 0.7], 
     Hue[0.7458214481292991, 0.5, 0.7], 
     Hue[0.7434658664368825, 0.5, 0.7], 
     Hue[0.7434658664368825, 0.5, 0.7], 
     Hue[0.741110284744466, 0.5, 0.7], 
     Hue[0.741110284744466, 0.5, 0.7], 
     Hue[0.736399121359633, 0.5, 0.7], 
     Hue[0.736399121359633, 0.5, 0.7], 
     Hue[0.7269767945899674, 0.5, 0.7], 
     Hue[0.7269767945899674, 0.5, 0.7], 
     Hue[0.7081321410506356, 0.5, 0.7], 
     Hue[0.7081321410506356, 0.5, 0.7], 
     Hue[0.6704428339719721, 0.5, 0.7], 
     Hue[0.6704428339719721, 0.5, 0.7], 
     Hue[0.6680872522795556, 0.5, 0.7], 
     Hue[0.6680872522795556, 0.5, 0.7], 
     Hue[0.665731670587139, 0.5, 0.7], 
     Hue[0.665731670587139, 0.5, 0.7], 
     Hue[0.661020507202306, 0.5, 0.7], 
     Hue[0.661020507202306, 0.5, 0.7], 
     Hue[0.6515981804326405, 0.5, 0.7], 
     Hue[0.6515981804326405, 0.5, 0.7], 
     Hue[0.6327535268933087, 0.5, 0.7], 
     Hue[0.6327535268933087, 0.5, 0.7], 
     Hue[0.5950642198146456, 0.5, 0.7], 
     Hue[0.5950642198146456, 0.5, 0.7], 
     Hue[0.5925115579379945, 0.5, 0.7], 
     Hue[0.5925115579379945, 0.5, 0.7], 
     Hue[0.5899588960613429, 0.5, 0.7], 
     Hue[0.5899588960613429, 0.5, 0.7], 
     Hue[0.5848535723080399, 0.5, 0.7], 
     Hue[0.5848535723080399, 0.5, 0.7], 
     Hue[0.5746429248014346, 0.5, 0.7], 
     Hue[0.5746429248014346, 0.5, 0.7], 
     Hue[0.5542216297882241, 0.5, 0.7], 
     Hue[0.5542216297882241, 0.5, 0.7], 
     Hue[0.5133790397618024, 0.5, 0.7], 
     Hue[0.5133790397618024, 0.5, 0.7], 
     Hue[0.5108263778851515, 0.5, 0.7], 
     Hue[0.5108263778851515, 0.5, 0.7], 
     Hue[0.5082737160085, 0.5, 0.7], 
     Hue[0.5082737160085, 0.5, 0.7], 
     Hue[0.5031683922551968, 0.5, 0.7], 
     Hue[0.5031683922551968, 0.5, 0.7], 
     Hue[0.4929577447485917, 0.5, 0.7], 
     Hue[0.4929577447485917, 0.5, 0.7], 
     Hue[0.4725364497353812, 0.5, 0.7], 
     Hue[0.4725364497353812, 0.5, 0.7], 
     Hue[0.4316938597089599, 0.5, 0.7], 
     Hue[0.4316938597089599, 0.5, 0.7], 
     Hue[0.429187722152806, 0.5, 0.7], 
     Hue[0.429187722152806, 0.5, 0.7], 
     Hue[0.4266815845966521, 0.5, 0.7], 
     Hue[0.4266815845966521, 0.5, 0.7], 
     Hue[0.42166930948434495`, 0.5, 0.7], 
     Hue[0.42166930948434495`, 0.5, 0.7], 
     Hue[0.41164475925973004`, 0.5, 0.7], 
     Hue[0.41164475925973004`, 0.5, 0.7], 
     Hue[0.39159565881050024`, 0.5, 0.7], 
     Hue[0.39159565881050024`, 0.5, 0.7], 
     Hue[0.35149745791204057`, 0.5, 0.7], 
     Hue[0.35149745791204057`, 0.5, 0.7], 
     Hue[0.3489913203558867, 0.5, 0.7], 
     Hue[0.3489913203558867, 0.5, 0.7], 
     Hue[0.34648518279973284`, 0.5, 0.7], 
     Hue[0.34648518279973284`, 0.5, 0.7], 
     Hue[0.34147290768742566`, 0.5, 0.7], 
     Hue[0.34147290768742566`, 0.5, 0.7], 
     Hue[0.33144835746281076`, 0.5, 0.7], 
     Hue[0.33144835746281076`, 0.5, 0.7], 
     Hue[0.3113992570135809, 0.5, 0.7], 
     Hue[0.3113992570135809, 0.5, 0.7], 
     Hue[0.27130105611512123`, 0.5, 0.7], 
     Hue[0.27130105611512123`, 0.5, 0.7], 
     Hue[0.26896324513183134`, 0.5, 0.7], 
     Hue[0.26896324513183134`, 0.5, 0.7], 
     Hue[0.26662543414854145`, 0.5, 0.7], 
     Hue[0.26662543414854145`, 0.5, 0.7], 
     Hue[0.26194981218196167`, 0.5, 0.7], 
     Hue[0.26194981218196167`, 0.5, 0.7], 
     Hue[0.25259856824880217`, 0.5, 0.7], 
     Hue[0.25259856824880217`, 0.5, 0.7], 
     Hue[0.23389608038248313`, 0.5, 0.7], 
     Hue[0.23389608038248313`, 0.5, 0.7], 
     Hue[0.19649110464984562`, 0.5, 0.7], 
     Hue[0.19649110464984562`, 0.5, 0.7], 
     Hue[0.19415329366655634`, 0.5, 0.7], 
     Hue[0.19415329366655634`, 0.5, 0.7], 
     Hue[0.19181548268326584`, 0.5, 0.7], 
     Hue[0.19181548268326584`, 0.5, 0.7], 
     Hue[0.18713986071668665`, 0.5, 0.7], 
     Hue[0.18713986071668665`, 0.5, 0.7], 
     Hue[0.17778861678352656`, 0.5, 0.7], 
     Hue[0.17778861678352656`, 0.5, 0.7], 
     Hue[0.1590861289172081, 0.5, 0.7], 
     Hue[0.1590861289172081, 0.5, 0.7], 
     Hue[0.12168115318456944`, 0.5, 0.7], 
     Hue[0.12168115318456944`, 0.5, 0.7], 
     Hue[0.11914626201704458`, 0.5, 0.7], 
     Hue[0.11914626201704458`, 0.5, 0.7], 
     Hue[0.11661137084952025`, 0.5, 0.7], 
     Hue[0.11661137084952025`, 0.5, 0.7], 
     Hue[0.11154158851447046`, 0.5, 0.7], 
     Hue[0.11154158851447046`, 0.5, 0.7], 
     Hue[0.10140202384437148`, 0.5, 0.7], 
     Hue[0.10140202384437148`, 0.5, 0.7], 
     Hue[0.0811228945041741, 0.5, 0.7], 
     Hue[0.0811228945041741, 0.5, 0.7], 
     Hue[0.0405646358237776, 0.5, 0.7], 
     Hue[0.0405646358237776, 0.5, 0.7], 
     Hue[0.038029744656252716`, 0.5, 0.7], 
     Hue[0.038029744656252716`, 0.5, 0.7], 
     Hue[0.035494853488728405`, 0.5, 0.7], 
     Hue[0.035494853488728405`, 0.5, 0.7], 
     Hue[0.03042507115367863, 0.5, 0.7], 
     Hue[0.03042507115367863, 0.5, 0.7], 
     Hue[0.020285506483580216`, 0.5, 0.7], 
     Hue[0.020285506483580216`, 0.5, 0.7], 
     Hue[6.377143381686833*^-6, 0.5, 0.7], 
     Hue[6.377143381686833*^-6, 0.5, 0.7], 
     Hue[0.9594481184629863, 0.5, 0.7], 
     Hue[0.9594481184629863, 0.5, 0.7], 
     Hue[0.9570815538683255, 0.5, 0.7], 
     Hue[0.9570815538683255, 0.5, 0.7], 
     Hue[0.9547149892736645, 0.5, 0.7], 
     Hue[0.9547149892736645, 0.5, 0.7], 
     Hue[0.9499818600843427, 0.5, 0.7], 
     Hue[0.9499818600843427, 0.5, 0.7], 
     Hue[0.9405156017056991, 0.5, 0.7], 
     Hue[0.9405156017056991, 0.5, 0.7], 
     Hue[0.921583084948412, 0.5, 0.7], 
     Hue[0.921583084948412, 0.5, 0.7], 
     Hue[0.8837180514338376, 0.5, 0.7], 
     Hue[0.8837180514338376, 0.5, 0.7], 
     Hue[0.8813514868391773, 0.5, 0.7], 
     Hue[0.8813514868391773, 0.5, 0.7], 
     Hue[0.8789849222445164, 0.5, 0.7], 
     Hue[0.8789849222445164, 0.5, 0.7], 
     Hue[0.8742517930551946, 0.5, 0.7], 
     Hue[0.8742517930551946, 0.5, 0.7], 
     Hue[0.8647855346765511, 0.5, 0.7], 
     Hue[0.8647855346765511, 0.5, 0.7], 
     Hue[0.8458530179192638, 0.5, 0.7], 
     Hue[0.8458530179192638, 0.5, 0.7], 
     Hue[0.8079879844046894, 0.5, 0.7], 
     Hue[0.8079879844046894, 0.5, 0.7], 
     Hue[0.8056679441305262, 0.5, 0.7], 
     Hue[0.8056679441305262, 0.5, 0.7], 
     Hue[0.8033479038563625, 0.5, 0.7], 
     Hue[0.8033479038563625, 0.5, 0.7], 
     Hue[0.7987078233080366, 0.5, 0.7], 
     Hue[0.7987078233080366, 0.5, 0.7], 
     Hue[0.7894276622113829, 0.5, 0.7], 
     Hue[0.7894276622113829, 0.5, 0.7], 
     Hue[0.7708673400180771, 0.5, 0.7], 
     Hue[0.7708673400180771, 0.5, 0.7], 
     Hue[0.7685472997439133, 0.5, 0.7], 
     Hue[0.7685472997439133, 0.5, 0.7], 
     Hue[0.76622725946975, 0.5, 0.7], 
     Hue[0.76622725946975, 0.5, 0.7], 
     Hue[0.7615871789214236, 0.5, 0.7], 
     Hue[0.7615871789214236, 0.5, 0.7], 
     Hue[0.7523070178247706, 0.5, 0.7], 
     Hue[0.7523070178247706, 0.5, 0.7], 
     Hue[0.7337466956314647, 0.5, 0.7], 
     Hue[0.7337466956314647, 0.5, 0.7], 
     Hue[0.7314266553573009, 0.5, 0.7], 
     Hue[0.7314266553573009, 0.5, 0.7], 
     Hue[0.7291066150831382, 0.5, 0.7], 
     Hue[0.7291066150831382, 0.5, 0.7], 
     Hue[0.7244665345348111, 0.5, 0.7], 
     Hue[0.7244665345348111, 0.5, 0.7], 
     Hue[0.7151863734381587, 0.5, 0.7], 
     Hue[0.7151863734381587, 0.5, 0.7], 
     Hue[0.6966260512448523, 0.5, 0.7], 
     Hue[0.6966260512448523, 0.5, 0.7], 
     Hue[0.65950540685824, 0.5, 0.7], 
     Hue[0.65950540685824, 0.5, 0.7], 
     Hue[0.6569882863998417, 0.5, 0.7], 
     Hue[0.6569882863998417, 0.5, 0.7], 
     Hue[0.6544711659414435, 0.5, 0.7], 
     Hue[0.6544711659414435, 0.5, 0.7], 
     Hue[0.6494369250246474, 0.5, 0.7], 
     Hue[0.6494369250246474, 0.5, 0.7], 
     Hue[0.6393684431910545, 0.5, 0.7], 
     Hue[0.6393684431910545, 0.5, 0.7], 
     Hue[0.6192314795238697, 0.5, 0.7], 
     Hue[0.6192314795238697, 0.5, 0.7], 
     Hue[0.616714359065472, 0.5, 0.7], 
     Hue[0.616714359065472, 0.5, 0.7], 
     Hue[0.6141972386070738, 0.5, 0.7], 
     Hue[0.6141972386070738, 0.5, 0.7], 
     Hue[0.6091629976902777, 0.5, 0.7], 
     Hue[0.6091629976902777, 0.5, 0.7], 
     Hue[0.5990945158566849, 0.5, 0.7], 
     Hue[0.5990945158566849, 0.5, 0.7], 
     Hue[0.5789575521894994, 0.5, 0.7], 
     Hue[0.5789575521894994, 0.5, 0.7], 
     Hue[0.5764404317311012, 0.5, 0.7], 
     Hue[0.5764404317311012, 0.5, 0.7], 
     Hue[0.5739233112727036, 0.5, 0.7], 
     Hue[0.5739233112727036, 0.5, 0.7], 
     Hue[0.5688890703559071, 0.5, 0.7], 
     Hue[0.5688890703559071, 0.5, 0.7], 
     Hue[0.5588205885223141, 0.5, 0.7], 
     Hue[0.5588205885223141, 0.5, 0.7], 
     Hue[0.5386836248551292, 0.5, 0.7], 
     Hue[0.5386836248551292, 0.5, 0.7], 
     Hue[0.4984096975207585, 0.5, 0.7], 
     Hue[0.4984096975207585, 0.5, 0.7], 
     Hue[0.49606090363522365`, 0.5, 0.7], 
     Hue[0.49606090363522365`, 0.5, 0.7], 
     Hue[0.49371210974968993`, 0.5, 0.7], 
     Hue[0.49371210974968993`, 0.5, 0.7], 
     Hue[0.48901452197862144`, 0.5, 0.7], 
     Hue[0.48901452197862144`, 0.5, 0.7], 
     Hue[0.4796193464364844, 0.5, 0.7], 
     Hue[0.4796193464364844, 0.5, 0.7], 
     Hue[0.4608289953522104, 0.5, 0.7], 
     Hue[0.4608289953522104, 0.5, 0.7], 
     Hue[0.45848020146667556`, 0.5, 0.7], 
     Hue[0.45848020146667556`, 0.5, 0.7], 
     Hue[0.4561314075811413, 0.5, 0.7], 
     Hue[0.4561314075811413, 0.5, 0.7], 
     Hue[0.4514338198100728, 0.5, 0.7], 
     Hue[0.4514338198100728, 0.5, 0.7], 
     Hue[0.4420386442679352, 0.5, 0.7], 
     Hue[0.4420386442679352, 0.5, 0.7], 
     Hue[0.42324829318366114`, 0.5, 0.7], 
     Hue[0.42324829318366114`, 0.5, 0.7], 
     Hue[0.4208994992981274, 0.5, 0.7], 
     Hue[0.4208994992981274, 0.5, 0.7], 
     Hue[0.41855070541259265`, 0.5, 0.7], 
     Hue[0.41855070541259265`, 0.5, 0.7], 
     Hue[0.4138531176415241, 0.5, 0.7], 
     Hue[0.4138531176415241, 0.5, 0.7], 
     Hue[0.40445794209938707`, 0.5, 0.7], 
     Hue[0.40445794209938707`, 0.5, 0.7], 
     Hue[0.3856675910151125, 0.5, 0.7], 
     Hue[0.3856675910151125, 0.5, 0.7], 
     Hue[0.3480868888465638, 0.5, 0.7], 
     Hue[0.3480868888465638, 0.5, 0.7], 
     Hue[0.3455410147767951, 0.5, 0.7], 
     Hue[0.3455410147767951, 0.5, 0.7], 
     Hue[0.34299514070702586`, 0.5, 0.7], 
     Hue[0.34299514070702586`, 0.5, 0.7], 
     Hue[0.3379033925674879, 0.5, 0.7], 
     Hue[0.3379033925674879, 0.5, 0.7], 
     Hue[0.32771989628841086`, 0.5, 0.7], 
     Hue[0.32771989628841086`, 0.5, 0.7], 
     Hue[0.30735290373025737`, 0.5, 0.7], 
     Hue[0.30735290373025737`, 0.5, 0.7], 
     Hue[0.30480702966048806`, 0.5, 0.7], 
     Hue[0.30480702966048806`, 0.5, 0.7], 
     Hue[0.30226115559071937`, 0.5, 0.7], 
     Hue[0.30226115559071937`, 0.5, 0.7], 
     Hue[0.2971694074511808, 0.5, 0.7], 
     Hue[0.2971694074511808, 0.5, 0.7], 
     Hue[0.28698591117210376`, 0.5, 0.7], 
     Hue[0.28698591117210376`, 0.5, 0.7], 
     Hue[0.2666189186139508, 0.5, 0.7], 
     Hue[0.2666189186139508, 0.5, 0.7], 
     Hue[0.26407304454418157`, 0.5, 0.7], 
     Hue[0.26407304454418157`, 0.5, 0.7], 
     Hue[0.2615271704744123, 0.5, 0.7], 
     Hue[0.2615271704744123, 0.5, 0.7], 
     Hue[0.25643542233487376`, 0.5, 0.7], 
     Hue[0.25643542233487376`, 0.5, 0.7], 
     Hue[0.2462519260557973, 0.5, 0.7], 
     Hue[0.2462519260557973, 0.5, 0.7], 
     Hue[0.22588493349764377`, 0.5, 0.7], 
     Hue[0.22588493349764377`, 0.5, 0.7], 
     Hue[0.18515094838133725`, 0.5, 0.7], 
     Hue[0.18515094838133725`, 0.5, 0.7], 
     Hue[0.185074863589463, 0.5, 0.7], 
     Hue[0.185074863589463, 0.5, 0.7], 
     Hue[0.18499877879758939`, 0.5, 0.7], 
     Hue[0.18499877879758939`, 0.5, 0.7], 
     Hue[0.18484660921384086`, 0.5, 0.7], 
     Hue[0.18484660921384086`, 0.5, 0.7], 
     Hue[0.1845422700463445, 0.5, 0.7], 
     Hue[0.1845422700463445, 0.5, 0.7], 
     Hue[0.18393359171135223`, 0.5, 0.7], 
     Hue[0.18393359171135223`, 0.5, 0.7], 
     Hue[0.18271623504136603`, 0.5, 0.7], 
     Hue[0.18271623504136603`, 0.5, 0.7], 
     Hue[0.18264015024949232`, 0.5, 0.7], 
     Hue[0.18264015024949232`, 0.5, 0.7], 
     Hue[0.18256406545761808`, 0.5, 0.7], 
     Hue[0.18256406545761808`, 0.5, 0.7], 
     Hue[0.18241189587386958`, 0.5, 0.7], 
     Hue[0.18241189587386958`, 0.5, 0.7], 
     Hue[0.18210755670637377`, 0.5, 0.7], 
     Hue[0.18210755670637377`, 0.5, 0.7], 
     Hue[0.1814988783713804, 0.5, 0.7], 
     Hue[0.1814988783713804, 0.5, 0.7], 
     Hue[0.1814227935795061, 0.5, 0.7], 
     Hue[0.1814227935795061, 0.5, 0.7], 
     Hue[0.18134670878763243`, 0.5, 0.7], 
     Hue[0.18134670878763243`, 0.5, 0.7], 
     Hue[0.18119453920388395`, 0.5, 0.7], 
     Hue[0.18119453920388395`, 0.5, 0.7], 
     Hue[0.18089020003638753`, 0.5, 0.7], 
     Hue[0.18089020003638753`, 0.5, 0.7], 
     Hue[0.1808141152445139, 0.5, 0.7], 
     Hue[0.1808141152445139, 0.5, 0.7], 
     Hue[0.18073803045263961`, 0.5, 0.7], 
     Hue[0.18073803045263961`, 0.5, 0.7], 
     Hue[0.18058586086889114`, 0.5, 0.7], 
     Hue[0.18058586086889114`, 0.5, 0.7], 
     Hue[0.18050977607701693`, 0.5, 0.7], 
     Hue[0.18050977607701693`, 0.5, 0.7], 
     Hue[0.18043369128514264`, 0.5, 0.7], 
     Hue[0.18043369128514264`, 0.5, 0.7], 
     Hue[0.18035760649326896`, 0.5, 0.7], 
     Hue[0.18035760649326896`, 0.5, 0.7], 
     Hue[0.18028152170139528`, 0.5, 0.7]}],
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->{True, False},
  AxesOrigin->{0, 0},
  Epilog->{
    Thickness[0.025], 
    GrayLevel[0.5], 
    LineBox[CompressedData["
1:eJw12nc8VX/8B3DjWtd1L0pSKiMpyXaFeB8llbIje6UtFBllk4yvHUmSUSFC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     "]]},
  Frame->True,
  PlotRange->{-0.2, 1.2},
  PlotRangeClipping->True,
  PlotRangePadding->{
    Scaled[0.02], Automatic}]], "Output",
 CellChangeTimes->{3.391857468015625*^9},
 CellLabel->"Out[35]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"mytab", "=", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Sin", "[", 
       RowBox[{"\[Pi]", " ", "x"}], "]"}], " ", 
      SuperscriptBox["\[ExponentialE]", 
       RowBox[{"2", " ", "\[ImaginaryI]", " ", "\[Pi]", " ", "x"}]]}], ",", 
     RowBox[{"{", 
      RowBox[{"x", ",", 
       RowBox[{"-", "2"}], ",", "2", ",", "0.1"}], "}"}]}], "]"}]}], 
  ";"}], "\n", 
 RowBox[{"QListArgColorPlot", "[", 
  RowBox[{"mytab", ",", 
   RowBox[{"Axes", "\[Rule]", "True"}], ",", 
   RowBox[{"QHorizontalRange", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "2"}], ",", "2"}], "}"}]}]}], "]"}], "\n", 
 RowBox[{"QListArgColorPlot", "[", 
  RowBox[{"mytab", ",", 
   RowBox[{"Axes", "\[Rule]", "True"}], ",", 
   RowBox[{"QHorizontalRange", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "2"}], ",", "2"}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "1"}], ",", "0"}], "}"}]}], "}"}]}]}], "]"}], "\n", 
 RowBox[{"QListArgColorPlot", "[", 
  RowBox[{"mytab", ",", 
   RowBox[{"Axes", "\[Rule]", "True"}], ",", 
   RowBox[{"QHorizontalRange", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "2"}], ",", "2"}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "4"}], ",", "4"}], "}"}]}], "}"}]}]}], "]"}]}], "Input",
 CellChangeTimes->{{3.391857449765625*^9, 3.391857452953125*^9}, 
   3.391857620453125*^9},
 CellLabel->"In[58]:="],

Cell[BoxData[
 GraphicsBox[GraphicsComplexBox[CompressedData["
1:eJyFlLFKA1EQRRextvELxA+wsLAcC7GwNFGIIoqgbbAQP0XUVkHxC1KNlRBQ
LFKlCwaEBdEEg0QiUZOZ+2QiNw4su4e3b/buvLkzs1te3ZvIsuzg5xrcPW7w
sD+MvhbyuYXX25rc3w3iU3c2DidnHx5lfhg9veocnx01czk9GcSHZmtL01Pl
F7EsXbV7Syzfe1q3fB298P2W7023PL/la+uyf9/zKPTZ/mct+bq9n+u277fv
PellyN8c0ddI+mx/Pa3b+7Wkz9aruhn0VXQl6PsNX5f1oK8qsX41ifWry1cx
6MN/Q79An/+fRH25QJ/XR6K+lkR9bSn6utc/6fPzkWvP7+cnI+eb8vr5S6xf
T859v/ePlML59KUQ67c42n+M0Y+M0Z+MU78Sjv37l1M/E0Z/M0a/M0b/M0b/
M4YfGMMfjJNfCFs0KCc/EYa/GMNv47nyH8t4rlKGPxnDr4y9HpSTnwnD34zh
d8bwP2P4nzHmAWPMB8ZpXhC26FJO84Qw5gtjzBvGmD+MvwEbnD01
   "], {
    {GrayLevel[0.5], EdgeForm[None], 
     GraphicsGroupBox[{
       PolygonBox[{{1, 42, 2}, {21, 78, 22}, {10, 59, 11}, {11, 60, 12}, {31, 
        96, 32}, {30, 95, 31}, {20, 77, 21}, {40, 113, 41}}], 
       PolygonBox[{{9, 57, 58, 10}, {27, 89, 90, 28}, {4, 47, 48, 5}, {3, 45, 
        46, 4}, {2, 43, 44, 3}, {28, 91, 92, 29}, {18, 73, 74, 19}, {17, 71, 
        72, 18}, {12, 61, 62, 13}, {39, 111, 112, 40}, {14, 65, 66, 15}, {15, 
        67, 68, 16}, {16, 69, 70, 17}, {5, 49, 50, 6}, {23, 81, 82, 24}, {8, 
        55, 56, 9}, {7, 53, 54, 8}, {38, 109, 110, 39}, {37, 107, 108, 38}, {
        32, 97, 98, 33}, {22, 79, 80, 23}, {24, 83, 84, 25}, {26, 87, 88, 
        27}, {36, 105, 106, 37}, {34, 101, 102, 35}, {19, 75, 76, 20}, {13, 
        63, 64, 14}, {29, 93, 94, 30}, {35, 103, 104, 36}, {33, 99, 100, 
        34}, {6, 51, 52, 7}, {25, 85, 86, 26}}]}]}, {{}, {}, {
      {Hue[0., 1, 1], LineBox[{1, 2}]}, 
      {Hue[0.10000000000000009`, 1, 1], LineBox[{2, 3}]}, 
      {Hue[0.20000000000000007`, 1, 1], LineBox[{3, 4}]}, 
      {Hue[0.3000000000000001, 1, 1], LineBox[{4, 5}]}, 
      {Hue[0.4000000000000001, 1, 1], LineBox[{5, 6}]}, 
      {Hue[0.5, 1, 1], LineBox[{6, 7}]}, 
      {Hue[0.6000000000000001, 1, 1], LineBox[{7, 8}]}, 
      {Hue[0.7, 1, 1], LineBox[{8, 9}]}, 
      {Hue[0.8, 1, 1], LineBox[{9, 10}]}, 
      {Hue[0.9000000000000001, 1, 1], LineBox[{10, 11}]}, 
      {Hue[0., 1, 1], LineBox[{11, 12}]}, 
      {Hue[0.6000000000000001, 1, 1], LineBox[{12, 13}]}, 
      {Hue[0.7, 1, 1], LineBox[{13, 14}]}, 
      {Hue[0.8, 1, 1], LineBox[{14, 15}]}, 
      {Hue[0.9000000000000001, 1, 1], LineBox[{15, 16}]}, 
      {Hue[0., 1, 1], LineBox[{16, 17}]}, 
      {Hue[0.10000000000000009`, 1, 1], LineBox[{17, 18}]}, 
      {Hue[0.20000000000000007`, 1, 1], LineBox[{18, 19}]}, 
      {Hue[0.30000000000000016`, 1, 1], LineBox[{19, 20}]}, 
      {Hue[0.40000000000000013`, 1, 1], LineBox[{20, 21}]}, 
      {Hue[0., 1, 1], LineBox[{21, 22}]}, 
      {Hue[0.10000000000000013`, 1, 1], LineBox[{22, 23}]}, 
      {Hue[0.20000000000000012`, 1, 1], LineBox[{23, 24}]}, 
      {Hue[0.30000000000000016`, 1, 1], LineBox[{24, 25}]}, 
      {Hue[0.40000000000000013`, 1, 1], LineBox[{25, 26}]}, 
      {Hue[0.5, 1, 1], LineBox[{26, 27}]}, 
      {Hue[0.6000000000000001, 1, 1], LineBox[{27, 28}]}, 
      {Hue[0.7000000000000002, 1, 1], LineBox[{28, 29}]}, 
      {Hue[0.8, 1, 1], LineBox[{29, 30}]}, 
      {Hue[0.9000000000000001, 1, 1], LineBox[{30, 31}]}, 
      {Hue[0., 1, 1], LineBox[{31, 32}]}, 
      {Hue[0.6000000000000001, 1, 1], LineBox[{32, 33}]}, 
      {Hue[0.7000000000000002, 1, 1], LineBox[{33, 34}]}, 
      {Hue[0.8000000000000003, 1, 1], LineBox[{34, 35}]}, 
      {Hue[0.8999999999999999, 1, 1], LineBox[{35, 36}]}, 
      {Hue[0., 1, 1], LineBox[{36, 37}]}, 
      {Hue[0.09999999999999994, 1, 1], LineBox[{37, 38}]}, 
      {Hue[0.2000000000000002, 1, 1], LineBox[{38, 39}]}, 
      {Hue[0.3000000000000002, 1, 1], LineBox[{39, 40}]}, 
      {Hue[0.3999999999999999, 1, 1], LineBox[{40, 41}]}}}},
   VertexColors->{
     Hue[0., 1, 1], 
     Hue[0.10000000000000009`, 1, 1], 
     Hue[0.20000000000000007`, 1, 1], 
     Hue[0.3000000000000001, 1, 1], 
     Hue[0.4000000000000001, 1, 1], 
     Hue[0.5, 1, 1], 
     Hue[0.6000000000000001, 1, 1], 
     Hue[0.7, 1, 1], 
     Hue[0.8, 1, 1], 
     Hue[0.9000000000000001, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.6000000000000001, 1, 1], 
     Hue[0.7, 1, 1], 
     Hue[0.8, 1, 1], 
     Hue[0.9000000000000001, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.10000000000000009`, 1, 1], 
     Hue[0.20000000000000007`, 1, 1], 
     Hue[0.30000000000000016`, 1, 1], 
     Hue[0.40000000000000013`, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.10000000000000013`, 1, 1], 
     Hue[0.20000000000000012`, 1, 1], 
     Hue[0.30000000000000016`, 1, 1], 
     Hue[0.40000000000000013`, 1, 1], 
     Hue[0.5, 1, 1], 
     Hue[0.6000000000000001, 1, 1], 
     Hue[0.7000000000000002, 1, 1], 
     Hue[0.8, 1, 1], 
     Hue[0.9000000000000001, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.6000000000000001, 1, 1], 
     Hue[0.7000000000000002, 1, 1], 
     Hue[0.8000000000000003, 1, 1], 
     Hue[0.8999999999999999, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.09999999999999994, 1, 1], 
     Hue[0.2000000000000002, 1, 1], 
     Hue[0.3000000000000002, 1, 1], 
     Hue[0.3999999999999999, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.10000000000000009`, 1, 1], 
     Hue[0.10000000000000009`, 1, 1], 
     Hue[0.20000000000000007`, 1, 1], 
     Hue[0.20000000000000007`, 1, 1], 
     Hue[0.3000000000000001, 1, 1], 
     Hue[0.3000000000000001, 1, 1], 
     Hue[0.4000000000000001, 1, 1], 
     Hue[0.4000000000000001, 1, 1], 
     Hue[0.5, 1, 1], 
     Hue[0.5, 1, 1], 
     Hue[0.6000000000000001, 1, 1], 
     Hue[0.6000000000000001, 1, 1], 
     Hue[0.7, 1, 1], 
     Hue[0.7, 1, 1], 
     Hue[0.8, 1, 1], 
     Hue[0.8, 1, 1], 
     Hue[0.9000000000000001, 1, 1], 
     Hue[0.9000000000000001, 1, 1], 
     Hue[0.6000000000000001, 1, 1], 
     Hue[0.6000000000000001, 1, 1], 
     Hue[0.7, 1, 1], 
     Hue[0.7, 1, 1], 
     Hue[0.8, 1, 1], 
     Hue[0.8, 1, 1], 
     Hue[0.9000000000000001, 1, 1], 
     Hue[0.9000000000000001, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.10000000000000009`, 1, 1], 
     Hue[0.10000000000000009`, 1, 1], 
     Hue[0.20000000000000007`, 1, 1], 
     Hue[0.20000000000000007`, 1, 1], 
     Hue[0.30000000000000016`, 1, 1], 
     Hue[0.30000000000000016`, 1, 1], 
     Hue[0.40000000000000013`, 1, 1], 
     Hue[0.40000000000000013`, 1, 1], 
     Hue[0.10000000000000013`, 1, 1], 
     Hue[0.10000000000000013`, 1, 1], 
     Hue[0.20000000000000012`, 1, 1], 
     Hue[0.20000000000000012`, 1, 1], 
     Hue[0.30000000000000016`, 1, 1], 
     Hue[0.30000000000000016`, 1, 1], 
     Hue[0.40000000000000013`, 1, 1], 
     Hue[0.40000000000000013`, 1, 1], 
     Hue[0.5, 1, 1], 
     Hue[0.5, 1, 1], 
     Hue[0.6000000000000001, 1, 1], 
     Hue[0.6000000000000001, 1, 1], 
     Hue[0.7000000000000002, 1, 1], 
     Hue[0.7000000000000002, 1, 1], 
     Hue[0.8, 1, 1], 
     Hue[0.8, 1, 1], 
     Hue[0.9000000000000001, 1, 1], 
     Hue[0.9000000000000001, 1, 1], 
     Hue[0.6000000000000001, 1, 1], 
     Hue[0.6000000000000001, 1, 1], 
     Hue[0.7000000000000002, 1, 1], 
     Hue[0.7000000000000002, 1, 1], 
     Hue[0.8000000000000003, 1, 1], 
     Hue[0.8000000000000003, 1, 1], 
     Hue[0.8999999999999999, 1, 1], 
     Hue[0.8999999999999999, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.09999999999999994, 1, 1], 
     Hue[0.09999999999999994, 1, 1], 
     Hue[0.2000000000000002, 1, 1], 
     Hue[0.2000000000000002, 1, 1], 
     Hue[0.3000000000000002, 1, 1], 
     Hue[0.3000000000000002, 1, 1], 
     Hue[0.3999999999999999, 1, 1], 
     Hue[0.3999999999999999, 1, 1]}],
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  Epilog->LineBox[
    NCache[{{-2, 0.}, {
       Rational[-19, 10], 0.3090169943749476}, {
       Rational[-9, 5], 0.5877852522924734}, {
       Rational[-17, 10], 0.8090169943749476}, {
       Rational[-8, 5], 0.9510565162951536}, {
       Rational[-3, 2], 1.}, {
       Rational[-7, 5], 0.9510565162951536}, {
       Rational[-13, 10], 0.8090169943749471}, {
       Rational[-6, 5], 0.587785252292473}, {
       Rational[-11, 10], 0.3090169943749469}, {-1, 0.}, {
       Rational[-9, 10], 0.30901699437494795`}, {
       Rational[-4, 5], 0.5877852522924732}, {
       Rational[-7, 10], 0.8090169943749475}, {
       Rational[-3, 5], 0.9510565162951536}, {
       Rational[-1, 2], 1.}, {
       Rational[-2, 5], 0.9510565162951536}, {
       Rational[-3, 10], 0.8090169943749471}, {
       Rational[-1, 5], 0.5877852522924728}, {
       Rational[-1, 10], 0.3090169943749471}, {0, 0.}, {
       Rational[1, 10], 0.3090169943749478}, {
       Rational[1, 5], 0.5877852522924734}, {
       Rational[3, 10], 0.8090169943749476}, {
       Rational[2, 5], 0.9510565162951535}, {
       Rational[1, 2], 1.}, {
       Rational[3, 5], 0.9510565162951536}, {
       Rational[7, 10], 0.8090169943749471}, {
       Rational[4, 5], 0.5877852522924728}, {
       Rational[9, 10], 0.3090169943749471}, {1, 0.}, {
       Rational[11, 10], 0.3090169943749477}, {
       Rational[6, 5], 0.5877852522924734}, {
       Rational[13, 10], 0.8090169943749478}, {
       Rational[7, 5], 0.9510565162951536}, {
       Rational[3, 2], 1.}, {
       Rational[8, 5], 0.9510565162951536}, {
       Rational[17, 10], 0.809016994374947}, {
       Rational[9, 5], 0.5877852522924726}, {
       Rational[19, 10], 0.3090169943749476}, {2, 0.}}, {{-2, 0.}, {-1.9, 
      0.3090169943749476}, {-1.8, 0.5877852522924734}, {-1.7, 
      0.8090169943749476}, {-1.6, 0.9510565162951536}, {-1.5, 1.}, {-1.4, 
      0.9510565162951536}, {-1.3, 0.8090169943749471}, {-1.2, 
      0.587785252292473}, {-1.1, 0.3090169943749469}, {-1, 0.}, {-0.9, 
      0.30901699437494795`}, {-0.8, 0.5877852522924732}, {-0.7, 
      0.8090169943749475}, {-0.6, 0.9510565162951536}, {-0.5, 1.}, {-0.4, 
      0.9510565162951536}, {-0.3, 0.8090169943749471}, {-0.2, 
      0.5877852522924728}, {-0.1, 0.3090169943749471}, {0, 0.}, {0.1, 
      0.3090169943749478}, {0.2, 0.5877852522924734}, {0.3, 
      0.8090169943749476}, {0.4, 0.9510565162951535}, {0.5, 1.}, {0.6, 
      0.9510565162951536}, {0.7, 0.8090169943749471}, {0.8, 
      0.5877852522924728}, {0.9, 0.3090169943749471}, {1, 0.}, {1.1, 
      0.3090169943749477}, {1.2, 0.5877852522924734}, {1.3, 
      0.8090169943749478}, {1.4, 0.9510565162951536}, {1.5, 1.}, {1.6, 
      0.9510565162951536}, {1.7, 0.809016994374947}, {1.8, 
      0.5877852522924726}, {1.9, 0.3090169943749476}, {2, 0.}}]],
  PlotRange->{{-2., 2.}, {0., 1.}},
  PlotRangeClipping->True,
  PlotRangePadding->{
    Scaled[0.02], 
    Scaled[0.02]}]], "Output",
 CellChangeTimes->{3.39185747940625*^9, 3.391857621828125*^9},
 CellLabel->"Out[59]="],

Cell[BoxData[
 GraphicsBox[
  GraphicsComplexBox[{{-1., 0.}, {-0.9, 0.30901699437494795`}, {-0.8, 
   0.5877852522924732}, {-0.7, 0.8090169943749475}, {-0.6, 
   0.9510565162951536}, {-0.5, 1.}, {-0.4, 0.9510565162951536}, {-0.3, 
   0.8090169943749471}, {-0.2, 0.5877852522924728}, {-0.1, 
   0.3090169943749471}, {0., 0.}, {-0.9, 0.}, {-0.9, 0.}, {-0.8, 0.}, {-0.8, 
   0.}, {-0.7, 0.}, {-0.7, 0.}, {-0.6, 0.}, {-0.6, 0.}, {-0.5, 0.}, {-0.5, 
   0.}, {-0.4, 0.}, {-0.4, 0.}, {-0.3, 0.}, {-0.3, 0.}, {-0.2, 0.}, {-0.2, 
   0.}, {-0.1, 0.}, {-0.1, 0.}}, {
    {GrayLevel[0.5], EdgeForm[None], 
     GraphicsGroupBox[{PolygonBox[{{1, 12, 2}, {10, 29, 11}}], 
       PolygonBox[{{7, 23, 24, 8}, {2, 13, 14, 3}, {9, 27, 28, 10}, {4, 17, 
        18, 5}, {3, 15, 16, 4}, {6, 21, 22, 7}, {8, 25, 26, 9}, {5, 19, 20, 
        6}}]}]}, {{}, {}, {
      {Hue[0., 1, 1], LineBox[{1, 2}]}, 
      {Hue[0.6000000000000001, 1, 1], LineBox[{2, 3}]}, 
      {Hue[0.7, 1, 1], LineBox[{3, 4}]}, 
      {Hue[0.8, 1, 1], LineBox[{4, 5}]}, 
      {Hue[0.9000000000000001, 1, 1], LineBox[{5, 6}]}, 
      {Hue[0., 1, 1], LineBox[{6, 7}]}, 
      {Hue[0.10000000000000009`, 1, 1], LineBox[{7, 8}]}, 
      {Hue[0.20000000000000007`, 1, 1], LineBox[{8, 9}]}, 
      {Hue[0.30000000000000016`, 1, 1], LineBox[{9, 10}]}, 
      {Hue[0.40000000000000013`, 1, 1], LineBox[{10, 11}]}}}},
   VertexColors->{
     Hue[0., 1, 1], 
     Hue[0.6000000000000001, 1, 1], 
     Hue[0.7, 1, 1], 
     Hue[0.8, 1, 1], 
     Hue[0.9000000000000001, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.10000000000000009`, 1, 1], 
     Hue[0.20000000000000007`, 1, 1], 
     Hue[0.30000000000000016`, 1, 1], 
     Hue[0.40000000000000013`, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.6000000000000001, 1, 1], 
     Hue[0.6000000000000001, 1, 1], 
     Hue[0.7, 1, 1], 
     Hue[0.7, 1, 1], 
     Hue[0.8, 1, 1], 
     Hue[0.8, 1, 1], 
     Hue[0.9000000000000001, 1, 1], 
     Hue[0.9000000000000001, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.10000000000000009`, 1, 1], 
     Hue[0.10000000000000009`, 1, 1], 
     Hue[0.20000000000000007`, 1, 1], 
     Hue[0.20000000000000007`, 1, 1], 
     Hue[0.30000000000000016`, 1, 1], 
     Hue[0.30000000000000016`, 1, 1], 
     Hue[0.40000000000000013`, 1, 1], 
     Hue[0.40000000000000013`, 1, 1]}],
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  Epilog->LineBox[
    NCache[{{-1, 0.}, {
       Rational[-9, 10], 0.30901699437494795`}, {
       Rational[-4, 5], 0.5877852522924732}, {
       Rational[-7, 10], 0.8090169943749475}, {
       Rational[-3, 5], 0.9510565162951536}, {
       Rational[-1, 2], 1.}, {
       Rational[-2, 5], 0.9510565162951536}, {
       Rational[-3, 10], 0.8090169943749471}, {
       Rational[-1, 5], 0.5877852522924728}, {
       Rational[-1, 10], 0.3090169943749471}, {0, 0.}}, {{-1, 0.}, {-0.9, 
      0.30901699437494795`}, {-0.8, 0.5877852522924732}, {-0.7, 
      0.8090169943749475}, {-0.6, 0.9510565162951536}, {-0.5, 1.}, {-0.4, 
      0.9510565162951536}, {-0.3, 0.8090169943749471}, {-0.2, 
      0.5877852522924728}, {-0.1, 0.3090169943749471}, {0, 0.}}]],
  PlotRange->{{-1., 0.}, {0., 1.}},
  PlotRangeClipping->True,
  PlotRangePadding->{
    Scaled[0.02], 
    Scaled[0.02]}]], "Output",
 CellChangeTimes->{3.39185747940625*^9, 3.39185762196875*^9},
 CellLabel->"Out[60]="],

Cell[BoxData[
 GraphicsBox[GraphicsComplexBox[CompressedData["
1:eJyFlLFKA1EQRRexloBfIH6AhYXlWIiFpYlCFFEEbS1E/BRRWwXFL0j1rISA
YpEqXTAgLIgmGCQSiZrM3JEJ3DiwLIe3b/a+eXNnZvdgdW8iy7LD32fw1ijc
ZjGc94fRT8V8buHtriYP94P4SjsbR5Ozj08yP4xeuu6cnB83czk7HcRnytaW
pqcOXkWzdJO+W6L5Pnxd83XSpe3XfO9py/JrvnZatv9bngR9uv8llW1dv8/T
tu3X/z2nq5C/OaKv4fp0f93X9fua69P1atoM+ippJej7C1uX9aCvKrF+NYn1
q8t3KejDuaFfoM/OJ1FfLtBn9ZGoryVRX1tKtm71d312P3Jj+e3+ZOR+Pa/d
v8T69eTC9lv/SDncT1+KsX6LsYoFZ/QjY/QnY/QrY+9fwnZeyt7fhNHvjNH/
jOEHxvADY/iDMfzC2P1DWKNB2f1FGH5jDP+N58p/LOO5Shl+ZQz/MrZ6UHZ/
E4bfGcP/jDEPGGMeMMZ8YIx5wdjnB2GNLmWfL4Qxbxhj/jDGPGL8AxFjPlc=

   "], {
    {GrayLevel[0.5], EdgeForm[None], 
     GraphicsGroupBox[{
       PolygonBox[{{2, 44, 3}, {22, 80, 23}, {11, 61, 12}, {12, 62, 13}, {32, 
        98, 33}, {31, 97, 32}, {21, 79, 22}, {41, 115, 42}}], 
       PolygonBox[{{10, 59, 60, 11}, {28, 91, 92, 29}, {5, 49, 50, 6}, {4, 47,
         48, 5}, {3, 45, 46, 4}, {29, 93, 94, 30}, {19, 75, 76, 20}, {18, 73, 
        74, 19}, {13, 63, 64, 14}, {40, 113, 114, 41}, {15, 67, 68, 16}, {16, 
        69, 70, 17}, {17, 71, 72, 18}, {6, 51, 52, 7}, {24, 83, 84, 25}, {9, 
        57, 58, 10}, {8, 55, 56, 9}, {39, 111, 112, 40}, {38, 109, 110, 39}, {
        33, 99, 100, 34}, {23, 81, 82, 24}, {25, 85, 86, 26}, {27, 89, 90, 
        28}, {37, 107, 108, 38}, {35, 103, 104, 36}, {20, 77, 78, 21}, {14, 
        65, 66, 15}, {30, 95, 96, 31}, {36, 105, 106, 37}, {34, 101, 102, 
        35}, {7, 53, 54, 8}, {26, 87, 88, 27}}]}]}, {{}, {}, {
      {Hue[0., 1, 1], LineBox[{1, 2}]}, 
      {Hue[0., 1, 1], LineBox[{2, 3}]}, 
      {Hue[0.10000000000000009`, 1, 1], LineBox[{3, 4}]}, 
      {Hue[0.20000000000000007`, 1, 1], LineBox[{4, 5}]}, 
      {Hue[0.3000000000000001, 1, 1], LineBox[{5, 6}]}, 
      {Hue[0.4000000000000001, 1, 1], LineBox[{6, 7}]}, 
      {Hue[0.5, 1, 1], LineBox[{7, 8}]}, 
      {Hue[0.6000000000000001, 1, 1], LineBox[{8, 9}]}, 
      {Hue[0.7, 1, 1], LineBox[{9, 10}]}, 
      {Hue[0.8, 1, 1], LineBox[{10, 11}]}, 
      {Hue[0.9000000000000001, 1, 1], LineBox[{11, 12}]}, 
      {Hue[0., 1, 1], LineBox[{12, 13}]}, 
      {Hue[0.6000000000000001, 1, 1], LineBox[{13, 14}]}, 
      {Hue[0.7, 1, 1], LineBox[{14, 15}]}, 
      {Hue[0.8, 1, 1], LineBox[{15, 16}]}, 
      {Hue[0.9000000000000001, 1, 1], LineBox[{16, 17}]}, 
      {Hue[0., 1, 1], LineBox[{17, 18}]}, 
      {Hue[0.10000000000000009`, 1, 1], LineBox[{18, 19}]}, 
      {Hue[0.20000000000000007`, 1, 1], LineBox[{19, 20}]}, 
      {Hue[0.30000000000000016`, 1, 1], LineBox[{20, 21}]}, 
      {Hue[0.40000000000000013`, 1, 1], LineBox[{21, 22}]}, 
      {Hue[0., 1, 1], LineBox[{22, 23}]}, 
      {Hue[0.10000000000000013`, 1, 1], LineBox[{23, 24}]}, 
      {Hue[0.20000000000000012`, 1, 1], LineBox[{24, 25}]}, 
      {Hue[0.30000000000000016`, 1, 1], LineBox[{25, 26}]}, 
      {Hue[0.40000000000000013`, 1, 1], LineBox[{26, 27}]}, 
      {Hue[0.5, 1, 1], LineBox[{27, 28}]}, 
      {Hue[0.6000000000000001, 1, 1], LineBox[{28, 29}]}, 
      {Hue[0.7000000000000002, 1, 1], LineBox[{29, 30}]}, 
      {Hue[0.8, 1, 1], LineBox[{30, 31}]}, 
      {Hue[0.9000000000000001, 1, 1], LineBox[{31, 32}]}, 
      {Hue[0., 1, 1], LineBox[{32, 33}]}, 
      {Hue[0.6000000000000001, 1, 1], LineBox[{33, 34}]}, 
      {Hue[0.7000000000000002, 1, 1], LineBox[{34, 35}]}, 
      {Hue[0.8000000000000003, 1, 1], LineBox[{35, 36}]}, 
      {Hue[0.8999999999999999, 1, 1], LineBox[{36, 37}]}, 
      {Hue[0., 1, 1], LineBox[{37, 38}]}, 
      {Hue[0.09999999999999994, 1, 1], LineBox[{38, 39}]}, 
      {Hue[0.2000000000000002, 1, 1], LineBox[{39, 40}]}, 
      {Hue[0.3000000000000002, 1, 1], LineBox[{40, 41}]}, 
      {Hue[0.3999999999999999, 1, 1], LineBox[{41, 42}]}, 
      {Hue[0., 1, 1], LineBox[{42, 43}]}}}},
   VertexColors->{
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.10000000000000009`, 1, 1], 
     Hue[0.20000000000000007`, 1, 1], 
     Hue[0.3000000000000001, 1, 1], 
     Hue[0.4000000000000001, 1, 1], 
     Hue[0.5, 1, 1], 
     Hue[0.6000000000000001, 1, 1], 
     Hue[0.7, 1, 1], 
     Hue[0.8, 1, 1], 
     Hue[0.9000000000000001, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.6000000000000001, 1, 1], 
     Hue[0.7, 1, 1], 
     Hue[0.8, 1, 1], 
     Hue[0.9000000000000001, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.10000000000000009`, 1, 1], 
     Hue[0.20000000000000007`, 1, 1], 
     Hue[0.30000000000000016`, 1, 1], 
     Hue[0.40000000000000013`, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.10000000000000013`, 1, 1], 
     Hue[0.20000000000000012`, 1, 1], 
     Hue[0.30000000000000016`, 1, 1], 
     Hue[0.40000000000000013`, 1, 1], 
     Hue[0.5, 1, 1], 
     Hue[0.6000000000000001, 1, 1], 
     Hue[0.7000000000000002, 1, 1], 
     Hue[0.8, 1, 1], 
     Hue[0.9000000000000001, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.6000000000000001, 1, 1], 
     Hue[0.7000000000000002, 1, 1], 
     Hue[0.8000000000000003, 1, 1], 
     Hue[0.8999999999999999, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.09999999999999994, 1, 1], 
     Hue[0.2000000000000002, 1, 1], 
     Hue[0.3000000000000002, 1, 1], 
     Hue[0.3999999999999999, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.10000000000000009`, 1, 1], 
     Hue[0.10000000000000009`, 1, 1], 
     Hue[0.20000000000000007`, 1, 1], 
     Hue[0.20000000000000007`, 1, 1], 
     Hue[0.3000000000000001, 1, 1], 
     Hue[0.3000000000000001, 1, 1], 
     Hue[0.4000000000000001, 1, 1], 
     Hue[0.4000000000000001, 1, 1], 
     Hue[0.5, 1, 1], 
     Hue[0.5, 1, 1], 
     Hue[0.6000000000000001, 1, 1], 
     Hue[0.6000000000000001, 1, 1], 
     Hue[0.7, 1, 1], 
     Hue[0.7, 1, 1], 
     Hue[0.8, 1, 1], 
     Hue[0.8, 1, 1], 
     Hue[0.9000000000000001, 1, 1], 
     Hue[0.9000000000000001, 1, 1], 
     Hue[0.6000000000000001, 1, 1], 
     Hue[0.6000000000000001, 1, 1], 
     Hue[0.7, 1, 1], 
     Hue[0.7, 1, 1], 
     Hue[0.8, 1, 1], 
     Hue[0.8, 1, 1], 
     Hue[0.9000000000000001, 1, 1], 
     Hue[0.9000000000000001, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.10000000000000009`, 1, 1], 
     Hue[0.10000000000000009`, 1, 1], 
     Hue[0.20000000000000007`, 1, 1], 
     Hue[0.20000000000000007`, 1, 1], 
     Hue[0.30000000000000016`, 1, 1], 
     Hue[0.30000000000000016`, 1, 1], 
     Hue[0.40000000000000013`, 1, 1], 
     Hue[0.40000000000000013`, 1, 1], 
     Hue[0.10000000000000013`, 1, 1], 
     Hue[0.10000000000000013`, 1, 1], 
     Hue[0.20000000000000012`, 1, 1], 
     Hue[0.20000000000000012`, 1, 1], 
     Hue[0.30000000000000016`, 1, 1], 
     Hue[0.30000000000000016`, 1, 1], 
     Hue[0.40000000000000013`, 1, 1], 
     Hue[0.40000000000000013`, 1, 1], 
     Hue[0.5, 1, 1], 
     Hue[0.5, 1, 1], 
     Hue[0.6000000000000001, 1, 1], 
     Hue[0.6000000000000001, 1, 1], 
     Hue[0.7000000000000002, 1, 1], 
     Hue[0.7000000000000002, 1, 1], 
     Hue[0.8, 1, 1], 
     Hue[0.8, 1, 1], 
     Hue[0.9000000000000001, 1, 1], 
     Hue[0.9000000000000001, 1, 1], 
     Hue[0.6000000000000001, 1, 1], 
     Hue[0.6000000000000001, 1, 1], 
     Hue[0.7000000000000002, 1, 1], 
     Hue[0.7000000000000002, 1, 1], 
     Hue[0.8000000000000003, 1, 1], 
     Hue[0.8000000000000003, 1, 1], 
     Hue[0.8999999999999999, 1, 1], 
     Hue[0.8999999999999999, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.09999999999999994, 1, 1], 
     Hue[0.09999999999999994, 1, 1], 
     Hue[0.2000000000000002, 1, 1], 
     Hue[0.2000000000000002, 1, 1], 
     Hue[0.3000000000000002, 1, 1], 
     Hue[0.3000000000000002, 1, 1], 
     Hue[0.3999999999999999, 1, 1], 
     Hue[0.3999999999999999, 1, 1]}],
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  Epilog->LineBox[
    NCache[{{-4, 0.}, {-2, 0.}, {
       Rational[-19, 10], 0.3090169943749476}, {
       Rational[-9, 5], 0.5877852522924734}, {
       Rational[-17, 10], 0.8090169943749476}, {
       Rational[-8, 5], 0.9510565162951536}, {
       Rational[-3, 2], 1.}, {
       Rational[-7, 5], 0.9510565162951536}, {
       Rational[-13, 10], 0.8090169943749471}, {
       Rational[-6, 5], 0.587785252292473}, {
       Rational[-11, 10], 0.3090169943749469}, {-1, 0.}, {
       Rational[-9, 10], 0.30901699437494795`}, {
       Rational[-4, 5], 0.5877852522924732}, {
       Rational[-7, 10], 0.8090169943749475}, {
       Rational[-3, 5], 0.9510565162951536}, {
       Rational[-1, 2], 1.}, {
       Rational[-2, 5], 0.9510565162951536}, {
       Rational[-3, 10], 0.8090169943749471}, {
       Rational[-1, 5], 0.5877852522924728}, {
       Rational[-1, 10], 0.3090169943749471}, {0, 0.}, {
       Rational[1, 10], 0.3090169943749478}, {
       Rational[1, 5], 0.5877852522924734}, {
       Rational[3, 10], 0.8090169943749476}, {
       Rational[2, 5], 0.9510565162951535}, {
       Rational[1, 2], 1.}, {
       Rational[3, 5], 0.9510565162951536}, {
       Rational[7, 10], 0.8090169943749471}, {
       Rational[4, 5], 0.5877852522924728}, {
       Rational[9, 10], 0.3090169943749471}, {1, 0.}, {
       Rational[11, 10], 0.3090169943749477}, {
       Rational[6, 5], 0.5877852522924734}, {
       Rational[13, 10], 0.8090169943749478}, {
       Rational[7, 5], 0.9510565162951536}, {
       Rational[3, 2], 1.}, {
       Rational[8, 5], 0.9510565162951536}, {
       Rational[17, 10], 0.809016994374947}, {
       Rational[9, 5], 0.5877852522924726}, {
       Rational[19, 10], 0.3090169943749476}, {2, 0.}, {
      4, 0.}}, {{-4, 0.}, {-2, 0.}, {-1.9, 0.3090169943749476}, {-1.8, 
      0.5877852522924734}, {-1.7, 0.8090169943749476}, {-1.6, 
      0.9510565162951536}, {-1.5, 1.}, {-1.4, 0.9510565162951536}, {-1.3, 
      0.8090169943749471}, {-1.2, 0.587785252292473}, {-1.1, 
      0.3090169943749469}, {-1, 0.}, {-0.9, 0.30901699437494795`}, {-0.8, 
      0.5877852522924732}, {-0.7, 0.8090169943749475}, {-0.6, 
      0.9510565162951536}, {-0.5, 1.}, {-0.4, 0.9510565162951536}, {-0.3, 
      0.8090169943749471}, {-0.2, 0.5877852522924728}, {-0.1, 
      0.3090169943749471}, {0, 0.}, {0.1, 0.3090169943749478}, {0.2, 
      0.5877852522924734}, {0.3, 0.8090169943749476}, {0.4, 
      0.9510565162951535}, {0.5, 1.}, {0.6, 0.9510565162951536}, {0.7, 
      0.8090169943749471}, {0.8, 0.5877852522924728}, {0.9, 
      0.3090169943749471}, {1, 0.}, {1.1, 0.3090169943749477}, {1.2, 
      0.5877852522924734}, {1.3, 0.8090169943749478}, {1.4, 
      0.9510565162951536}, {1.5, 1.}, {1.6, 0.9510565162951536}, {1.7, 
      0.809016994374947}, {1.8, 0.5877852522924726}, {1.9, 
      0.3090169943749476}, {2, 0.}, {4, 0.}}]],
  PlotRange->{{-4., 4.}, {0., 1.}},
  PlotRangeClipping->True,
  PlotRangePadding->{
    Scaled[0.02], 
    Scaled[0.02]}]], "Output",
 CellChangeTimes->{3.39185747940625*^9, 3.39185762221875*^9},
 CellLabel->"Out[61]="]
}, Open  ]],

Cell[BoxData[
 RowBox[{"QCombinedPlot", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"Sin", "[", 
      RowBox[{"\[Pi]", " ", "x"}], "]"}], ",", 
     RowBox[{"Sin", "[", 
      RowBox[{"\[Pi]", " ", "x"}], "]"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", 
     RowBox[{"-", "4"}], ",", "3"}], "}"}]}], "]"}]], "Input",
 CellChangeTimes->{3.3918574554375*^9}],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"mytab", "=", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"Sin", "[", 
      RowBox[{"\[Pi]", " ", "x"}], "]"}], ",", 
     RowBox[{"{", 
      RowBox[{"x", ",", 
       RowBox[{"-", "4"}], ",", "3", ",", "0.1"}], "}"}]}], "]"}]}], 
  ";"}], "\n", 
 RowBox[{"QListCombinedPlot", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"mytab", ",", 
     RowBox[{"Sin", "[", 
      RowBox[{"\[Pi]", " ", "x"}], "]"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", 
     RowBox[{"-", "4"}], ",", "3"}], "}"}]}], "]"}], "\n", 
 RowBox[{"QListCombinedPlot", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"mytab", ",", 
     RowBox[{"Sin", "[", 
      RowBox[{"\[Pi]", " ", "x"}], "]"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", "0", ",", "3"}], "}"}]}], "]"}], "\n", 
 RowBox[{"QListCombinedPlot", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"mytab", ",", 
     RowBox[{"Sin", "[", 
      RowBox[{"\[Pi]", " ", "x"}], "]"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", "0", ",", "3"}], "}"}], ",", 
   RowBox[{"QHorizontalRange", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "4"}], ",", "3"}], "}"}]}]}], "]"}], "\n", 
 RowBox[{"QListCombinedPlot", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"mytab", ",", 
     RowBox[{"Sin", "[", 
      RowBox[{"\[Pi]", " ", "x"}], "]"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", "0", ",", "3"}], "}"}], ",", 
   RowBox[{"QHorizontalRange", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"-", "4"}], ",", "3"}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"1", ",", "2"}], "}"}]}], "}"}]}]}], "]"}]}], "Input",
 CellChangeTimes->{
  3.3918574565625*^9, {3.39185748825*^9, 3.391857492390625*^9}, 
   3.39185761490625*^9},
 CellLabel->"In[53]:="],

Cell[BoxData[
 GraphicsBox[{GraphicsComplexBox[CompressedData["
1:eJyFlUFrU1EQhV/S2FatVfAXiD/AhYsu5yJFEHfWlkZdxIpduOlK/Cmi7kQF
q3+gq7kbhYDiIqvuioVCoGiLQSKxUdOZc2EenDoQko/37rxz7+Sdc2lt49bD
ZlVVH/59Jt9WFzJ+XT2u83m1f2Xh+8eerB/XfL5/53Hr8pev8vzZpM7lzcHT
F092+/L506TmcmN58eL8xjexLnPe70Cs39k8vm3Xrd+Z/MbXW7/T+a73t36z
+YY/3/rM1vTN5JWgbzp3gr5TNX2tXAV9rZq+qfwn6Gvml0FfIy8HfVW+GfRV
RZ+tH+uSX7f7f+u6r7fnjfRt6P9Lo76hQp/1+6nQZ/0GCn3W74euhP0f6vWg
70Chz9bva9uv2/19XfP19rw9fRf672qc707RZ+u39cj12f09fR3219V20Lel
i/H8Svl1uRf0deVBOL+evA/9t6WmD/uGfjkK892TV2F9X1bDfPflWjw/ifoO
iz4/f+mE9YOiz+dX0zcsfX3+Es9vVPT5/0faYT5jqZ1fivOt0qNwfo0U59tM
cb5TKeprJejz9yeNXJ+/X2kz7G86dcJ8ZtJCfH+LPvgLY/gNY/gPY/gRYyvO
8CvG8C/G8DPG8DfG8DfG8DvG8D/GxQ8JW3GGXzKGfzKGnzKGvzKGvzKG3zKG
/zIufkzYakgZfs0Y/s0Yfs4Y/s4Y/s4Yfs8Y/s8YecDYaocy8oIx8oMx8uRk
3vofy8ncpYz8YYw8YuznQRl5xRj5xRh5xhj5xhj5xhh5xxj5xxh5yNhqSBl5
yRj5yRh5yhj5yhj5yhh5yxj5yxh5zNiqRRl5zRj5zRh5zhj5zvgvRnqtTg==

    "], {
     {GrayLevel[0.5], EdgeForm[None], 
      GraphicsGroupBox[{
        PolygonBox[{{1, 72, 2}, {11, 90, 12}, {61, 180, 62}, {60, 179, 61}, {
         51, 162, 52}, {10, 89, 11}, {31, 126, 32}, {30, 125, 31}, {21, 108, 
         22}, {20, 107, 21}, {41, 144, 42}, {70, 197, 71}, {40, 143, 41}, {50,
          161, 51}}], 
        PolygonBox[{{9, 87, 88, 10}, {8, 85, 86, 9}, {69, 195, 196, 70}, {17, 
         101, 102, 18}, {6, 81, 82, 7}, {5, 79, 80, 6}, {19, 105, 106, 20}, {
         7, 83, 84, 8}, {2, 73, 74, 3}, {3, 75, 76, 4}, {16, 99, 100, 17}, {4,
          77, 78, 5}, {29, 123, 124, 30}, {48, 157, 158, 49}, {22, 109, 110, 
         23}, {23, 111, 112, 24}, {24, 113, 114, 25}, {47, 155, 156, 48}, {46,
          153, 154, 47}, {25, 115, 116, 26}, {27, 119, 120, 28}, {32, 127, 
         128, 33}, {49, 159, 160, 50}, {33, 129, 130, 34}, {18, 103, 104, 
         19}, {28, 121, 122, 29}, {39, 141, 142, 40}, {15, 97, 98, 16}, {13, 
         93, 94, 14}, {66, 189, 190, 67}, {67, 191, 192, 68}, {68, 193, 194, 
         69}, {65, 187, 188, 66}, {63, 183, 184, 64}, {64, 185, 186, 65}, {53,
          165, 166, 54}, {42, 145, 146, 43}, {38, 139, 140, 39}, {52, 163, 
         164, 53}, {54, 167, 168, 55}, {37, 137, 138, 38}, {62, 181, 182, 
         63}, {43, 147, 148, 44}, {34, 131, 132, 35}, {45, 151, 152, 46}, {12,
          91, 92, 13}, {26, 117, 118, 27}, {36, 135, 136, 37}, {44, 149, 150, 
         45}, {59, 177, 178, 60}, {55, 169, 170, 56}, {14, 95, 96, 15}, {56, 
         171, 172, 57}, {58, 175, 176, 59}, {57, 173, 174, 58}, {35, 133, 134,
          36}}]}]}, {{}, {}, {
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{1, 2}]}, 
       {Hue[0, 1, 1], LineBox[{2, 3}]}, 
       {Hue[0, 1, 1], LineBox[{3, 4}]}, 
       {Hue[0, 1, 1], LineBox[{4, 5}]}, 
       {Hue[0, 1, 1], LineBox[{5, 6}]}, 
       {Hue[0, 1, 1], LineBox[{6, 7}]}, 
       {Hue[0, 1, 1], LineBox[{7, 8}]}, 
       {Hue[0, 1, 1], LineBox[{8, 9}]}, 
       {Hue[0, 1, 1], LineBox[{9, 10}]}, 
       {Hue[0, 1, 1], LineBox[{10, 11}]}, 
       {Hue[0, 1, 1], LineBox[{11, 12}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{12, 13}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{13, 14}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{14, 15}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{15, 16}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{16, 17}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{17, 18}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{18, 19}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{19, 20}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{20, 21}]}, 
       {Hue[0, 1, 1], LineBox[{21, 22}]}, 
       {Hue[0, 1, 1], LineBox[{22, 23}]}, 
       {Hue[0, 1, 1], LineBox[{23, 24}]}, 
       {Hue[0, 1, 1], LineBox[{24, 25}]}, 
       {Hue[0, 1, 1], LineBox[{25, 26}]}, 
       {Hue[0, 1, 1], LineBox[{26, 27}]}, 
       {Hue[0, 1, 1], LineBox[{27, 28}]}, 
       {Hue[0, 1, 1], LineBox[{28, 29}]}, 
       {Hue[0, 1, 1], LineBox[{29, 30}]}, 
       {Hue[0, 1, 1], LineBox[{30, 31}]}, 
       {Hue[0, 1, 1], LineBox[{31, 32}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{32, 33}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{33, 34}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{34, 35}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{35, 36}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{36, 37}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{37, 38}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{38, 39}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{39, 40}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{40, 41}]}, 
       {Hue[0, 1, 1], LineBox[{41, 42}]}, 
       {Hue[0, 1, 1], LineBox[{42, 43}]}, 
       {Hue[0, 1, 1], LineBox[{43, 44}]}, 
       {Hue[0, 1, 1], LineBox[{44, 45}]}, 
       {Hue[0, 1, 1], LineBox[{45, 46}]}, 
       {Hue[0, 1, 1], LineBox[{46, 47}]}, 
       {Hue[0, 1, 1], LineBox[{47, 48}]}, 
       {Hue[0, 1, 1], LineBox[{48, 49}]}, 
       {Hue[0, 1, 1], LineBox[{49, 50}]}, 
       {Hue[0, 1, 1], LineBox[{50, 51}]}, 
       {Hue[0, 1, 1], LineBox[{51, 52}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{52, 53}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{53, 54}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{54, 55}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{55, 56}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{56, 57}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{57, 58}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{58, 59}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{59, 60}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{60, 61}]}, 
       {Hue[0, 1, 1], LineBox[{61, 62}]}, 
       {Hue[0, 1, 1], LineBox[{62, 63}]}, 
       {Hue[0, 1, 1], LineBox[{63, 64}]}, 
       {Hue[0, 1, 1], LineBox[{64, 65}]}, 
       {Hue[0, 1, 1], LineBox[{65, 66}]}, 
       {Hue[0, 1, 1], LineBox[{66, 67}]}, 
       {Hue[0, 1, 1], LineBox[{67, 68}]}, 
       {Hue[0, 1, 1], LineBox[{68, 69}]}, 
       {Hue[0, 1, 1], LineBox[{69, 70}]}, 
       {Hue[0, 1, 1], LineBox[{70, 71}]}}}},
    VertexColors->{
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1]}], {{}, {}, 
    {Hue[0.67, 0.6, 0.6], LineBox[CompressedData["
1:eJwlnHk8Vd/3/3EvMg+RIUqEZEgqobKOMmQWkiSVIkOZk7wLKVNpUCqpJBqE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      "]]}}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  Epilog->LineBox[
    NCache[{{-4, 0.}, {
       Rational[-39, 10], 0.3090169943749479}, {
       Rational[-19, 5], 0.5877852522924735}, {
       Rational[-37, 10], 0.8090169943749477}, {
       Rational[-18, 5], 0.9510565162951538}, {
       Rational[-7, 2], 1.}, {
       Rational[-17, 5], 0.9510565162951534}, {
       Rational[-33, 10], 0.8090169943749472}, {
       Rational[-16, 5], 0.5877852522924728}, {
       Rational[-31, 10], 0.30901699437494706`}, {-3, 0.}, {
       Rational[-29, 10], 0.3090169943749478}, {
       Rational[-14, 5], 0.5877852522924734}, {
       Rational[-27, 10], 0.8090169943749477}, {
       Rational[-13, 5], 0.9510565162951536}, {
       Rational[-5, 2], 1.}, {
       Rational[-12, 5], 0.9510565162951535}, {
       Rational[-23, 10], 0.8090169943749468}, {
       Rational[-11, 5], 0.5877852522924721}, {
       Rational[-21, 10], 0.3090169943749472}, {-2, 0.}, {
       Rational[-19, 10], 0.3090169943749476}, {
       Rational[-9, 5], 0.587785252292474}, {
       Rational[-17, 10], 0.8090169943749476}, {
       Rational[-8, 5], 0.9510565162951536}, {
       Rational[-3, 2], 1.}, {
       Rational[-7, 5], 0.9510565162951535}, {
       Rational[-13, 10], 0.8090169943749468}, {
       Rational[-6, 5], 0.5877852522924722}, {
       Rational[-11, 10], 0.3090169943749469}, {-1, 0.}, {
       Rational[-9, 10], 0.30901699437494795`}, {
       Rational[-4, 5], 0.5877852522924736}, {
       Rational[-7, 10], 0.8090169943749478}, {
       Rational[-3, 5], 0.9510565162951538}, {
       Rational[-1, 2], 1.}, {
       Rational[-2, 5], 0.9510565162951533}, {
       Rational[-3, 10], 0.8090169943749471}, {
       Rational[-1, 5], 0.5877852522924726}, {
       Rational[-1, 10], 0.3090169943749468}, {0, 0.}, {
       Rational[1, 10], 0.3090169943749481}, {
       Rational[1, 5], 0.5877852522924737}, {
       Rational[3, 10], 0.8090169943749479}, {
       Rational[2, 5], 0.9510565162951538}, {
       Rational[1, 2], 1.}, {
       Rational[3, 5], 0.9510565162951533}, {
       Rational[7, 10], 0.809016994374947}, {
       Rational[4, 5], 0.5877852522924725}, {
       Rational[9, 10], 0.3090169943749467}, {1, 0.}, {
       Rational[11, 10], 0.3090169943749481}, {
       Rational[6, 5], 0.5877852522924734}, {
       Rational[13, 10], 0.8090169943749479}, {
       Rational[7, 5], 0.9510565162951538}, {
       Rational[3, 2], 1.}, {
       Rational[8, 5], 0.9510565162951533}, {
       Rational[17, 10], 0.809016994374947}, {
       Rational[9, 5], 0.5877852522924726}, {
       Rational[19, 10], 0.3090169943749468}, {2, 0.}, {
       Rational[21, 10], 0.3090169943749489}, {
       Rational[11, 5], 0.5877852522924736}, {
       Rational[23, 10], 0.8090169943749478}, {
       Rational[12, 5], 0.9510565162951538}, {
       Rational[5, 2], 1.}, {
       Rational[13, 5], 0.9510565162951531}, {
       Rational[27, 10], 0.8090169943749477}, {
       Rational[14, 5], 0.5877852522924734}, {
       Rational[29, 10], 0.30901699437494606`}, {3, 0.}}, {{-4, 0.}, {-3.9, 
      0.3090169943749479}, {-3.8, 0.5877852522924735}, {-3.7, 
      0.8090169943749477}, {-3.6, 0.9510565162951538}, {-3.5, 1.}, {-3.4, 
      0.9510565162951534}, {-3.3, 0.8090169943749472}, {-3.2, 
      0.5877852522924728}, {-3.1, 0.30901699437494706`}, {-3, 0.}, {-2.9, 
      0.3090169943749478}, {-2.8, 0.5877852522924734}, {-2.7, 
      0.8090169943749477}, {-2.6, 0.9510565162951536}, {-2.5, 1.}, {-2.4, 
      0.9510565162951535}, {-2.3, 0.8090169943749468}, {-2.2, 
      0.5877852522924721}, {-2.1, 0.3090169943749472}, {-2, 0.}, {-1.9, 
      0.3090169943749476}, {-1.8, 0.587785252292474}, {-1.7, 
      0.8090169943749476}, {-1.6, 0.9510565162951536}, {-1.5, 1.}, {-1.4, 
      0.9510565162951535}, {-1.3, 0.8090169943749468}, {-1.2, 
      0.5877852522924722}, {-1.1, 0.3090169943749469}, {-1, 0.}, {-0.9, 
      0.30901699437494795`}, {-0.8, 0.5877852522924736}, {-0.7, 
      0.8090169943749478}, {-0.6, 0.9510565162951538}, {-0.5, 1.}, {-0.4, 
      0.9510565162951533}, {-0.3, 0.8090169943749471}, {-0.2, 
      0.5877852522924726}, {-0.1, 0.3090169943749468}, {0, 0.}, {0.1, 
      0.3090169943749481}, {0.2, 0.5877852522924737}, {0.3, 
      0.8090169943749479}, {0.4, 0.9510565162951538}, {0.5, 1.}, {0.6, 
      0.9510565162951533}, {0.7, 0.809016994374947}, {0.8, 
      0.5877852522924725}, {0.9, 0.3090169943749467}, {1, 0.}, {1.1, 
      0.3090169943749481}, {1.2, 0.5877852522924734}, {1.3, 
      0.8090169943749479}, {1.4, 0.9510565162951538}, {1.5, 1.}, {1.6, 
      0.9510565162951533}, {1.7, 0.809016994374947}, {1.8, 
      0.5877852522924726}, {1.9, 0.3090169943749468}, {2, 0.}, {2.1, 
      0.3090169943749489}, {2.2, 0.5877852522924736}, {2.3, 
      0.8090169943749478}, {2.4, 0.9510565162951538}, {2.5, 1.}, {2.6, 
      0.9510565162951531}, {2.7, 0.8090169943749477}, {2.8, 
      0.5877852522924734}, {2.9, 0.30901699437494606`}, {3, 0.}}]],
  PlotRange->Automatic,
  PlotRangeClipping->True,
  PlotRangePadding->{
    Scaled[0.02], 
    Scaled[0.02]}]], "Output",
 CellChangeTimes->{3.391857494046875*^9, 3.391857615484375*^9},
 CellLabel->"Out[54]="],

Cell[BoxData[
 GraphicsBox[{GraphicsComplexBox[CompressedData["
1:eJx9lU9IlFEUxb8Zx7GVLcJ2LiJbFbRoIdjiPmUIhjZSaVqBZdQiCFfSpp2t
bOdGqk1FhmlgG2snbWxTUaBBIEQ5RVmhTmqN/8bm++49D17D6eOJ/Hi++857
93nOvt6+E5fSURRNV37i3/9+e3uKDT3FEelaONy89GJGlCflwpn+zP4389Je
118Zz2VsdfjOtcKCzU9LqiO3p75vUeaeXq2Ml6LVlu3v30r5lM4fSb4ZeRis
fydnrf764ERpcOK95G1/rTcn0Dc1f7AyPkinzWv9j3Le1rc1zrY2zn7y+nS/
gkSm70B+qCk/9Nnr0/2/yI7pS8q1f5X7tl71fJMOq//9bn085Hig74fX9/pV
/P2Ukzavehflsq1/UooLLskjq6/6l72+RF5T0evT8xS9vqOjAy2jA7+8Pj3f
inRa/UPJhityzPbX8656fUm51jXpDvq7Jr22fndywN8ybvX1Pv74/mqVkten
91OSbdMXV2voWZcRW6/3tSHdVr8QX8fUhuSC+9v0+hJ5jVtyzub1Prfkoq2f
Pn29MrblsdXX+y17fc+S51L2+vS+d7w+rR+5B8H9R67L6ut+kWuz/bUfKQd9
un/KQZ/2J+Xw/lRP2kGf9ivtQn1pB33avxoHfaq3xuvTftY43F9SriXjckF/
M17fveR5ZtwVm9d+1zr09/at+Kt148H/X63Xlxw3n/X69D1k3abpu5m0O+vG
gvdR589/I5Y3WueabX+ts8vrg7/8nycpw38Yw48Yh/5UzfArxvAvxvAzxvA3
xvA3xvA7xvA/xt4PCYf+WM3wS8bwT8bwU8bwV8bwV8bwW8bwX8bejwmH/lzN
8GvG8G/G8HPG8HfG8HfG8HvG8H/GyAPG+pUoIy8YIz8YI08YI18YI18YI28Y
I38YI48Yh/lUzcgrxsgvxsgzxsg3xsg3xsg7xsg/xshDxmE+VjPykjHykzHy
lDHylTHylTHyljHylzHymHGYz9WMvGaM/GaMPGeMfGf8F4RUwtw=
    "], {
     {GrayLevel[0.5], EdgeForm[None], 
      GraphicsGroupBox[{
        PolygonBox[{{1, 72, 2}, {11, 90, 12}, {61, 180, 62}, {60, 179, 61}, {
         51, 162, 52}, {10, 89, 11}, {31, 126, 32}, {30, 125, 31}, {21, 108, 
         22}, {20, 107, 21}, {41, 144, 42}, {70, 197, 71}, {40, 143, 41}, {50,
          161, 51}}], 
        PolygonBox[{{9, 87, 88, 10}, {8, 85, 86, 9}, {69, 195, 196, 70}, {17, 
         101, 102, 18}, {6, 81, 82, 7}, {5, 79, 80, 6}, {19, 105, 106, 20}, {
         7, 83, 84, 8}, {2, 73, 74, 3}, {3, 75, 76, 4}, {16, 99, 100, 17}, {4,
          77, 78, 5}, {29, 123, 124, 30}, {48, 157, 158, 49}, {22, 109, 110, 
         23}, {23, 111, 112, 24}, {24, 113, 114, 25}, {47, 155, 156, 48}, {46,
          153, 154, 47}, {25, 115, 116, 26}, {27, 119, 120, 28}, {32, 127, 
         128, 33}, {49, 159, 160, 50}, {33, 129, 130, 34}, {18, 103, 104, 
         19}, {28, 121, 122, 29}, {39, 141, 142, 40}, {15, 97, 98, 16}, {13, 
         93, 94, 14}, {66, 189, 190, 67}, {67, 191, 192, 68}, {68, 193, 194, 
         69}, {65, 187, 188, 66}, {63, 183, 184, 64}, {64, 185, 186, 65}, {53,
          165, 166, 54}, {42, 145, 146, 43}, {38, 139, 140, 39}, {52, 163, 
         164, 53}, {54, 167, 168, 55}, {37, 137, 138, 38}, {62, 181, 182, 
         63}, {43, 147, 148, 44}, {34, 131, 132, 35}, {45, 151, 152, 46}, {12,
          91, 92, 13}, {26, 117, 118, 27}, {36, 135, 136, 37}, {44, 149, 150, 
         45}, {59, 177, 178, 60}, {55, 169, 170, 56}, {14, 95, 96, 15}, {56, 
         171, 172, 57}, {58, 175, 176, 59}, {57, 173, 174, 58}, {35, 133, 134,
          36}}]}]}, {{}, {}, {
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{1, 2}]}, 
       {Hue[0, 1, 1], LineBox[{2, 3}]}, 
       {Hue[0, 1, 1], LineBox[{3, 4}]}, 
       {Hue[0, 1, 1], LineBox[{4, 5}]}, 
       {Hue[0, 1, 1], LineBox[{5, 6}]}, 
       {Hue[0, 1, 1], LineBox[{6, 7}]}, 
       {Hue[0, 1, 1], LineBox[{7, 8}]}, 
       {Hue[0, 1, 1], LineBox[{8, 9}]}, 
       {Hue[0, 1, 1], LineBox[{9, 10}]}, 
       {Hue[0, 1, 1], LineBox[{10, 11}]}, 
       {Hue[0, 1, 1], LineBox[{11, 12}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{12, 13}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{13, 14}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{14, 15}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{15, 16}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{16, 17}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{17, 18}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{18, 19}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{19, 20}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{20, 21}]}, 
       {Hue[0, 1, 1], LineBox[{21, 22}]}, 
       {Hue[0, 1, 1], LineBox[{22, 23}]}, 
       {Hue[0, 1, 1], LineBox[{23, 24}]}, 
       {Hue[0, 1, 1], LineBox[{24, 25}]}, 
       {Hue[0, 1, 1], LineBox[{25, 26}]}, 
       {Hue[0, 1, 1], LineBox[{26, 27}]}, 
       {Hue[0, 1, 1], LineBox[{27, 28}]}, 
       {Hue[0, 1, 1], LineBox[{28, 29}]}, 
       {Hue[0, 1, 1], LineBox[{29, 30}]}, 
       {Hue[0, 1, 1], LineBox[{30, 31}]}, 
       {Hue[0, 1, 1], LineBox[{31, 32}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{32, 33}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{33, 34}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{34, 35}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{35, 36}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{36, 37}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{37, 38}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{38, 39}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{39, 40}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{40, 41}]}, 
       {Hue[0, 1, 1], LineBox[{41, 42}]}, 
       {Hue[0, 1, 1], LineBox[{42, 43}]}, 
       {Hue[0, 1, 1], LineBox[{43, 44}]}, 
       {Hue[0, 1, 1], LineBox[{44, 45}]}, 
       {Hue[0, 1, 1], LineBox[{45, 46}]}, 
       {Hue[0, 1, 1], LineBox[{46, 47}]}, 
       {Hue[0, 1, 1], LineBox[{47, 48}]}, 
       {Hue[0, 1, 1], LineBox[{48, 49}]}, 
       {Hue[0, 1, 1], LineBox[{49, 50}]}, 
       {Hue[0, 1, 1], LineBox[{50, 51}]}, 
       {Hue[0, 1, 1], LineBox[{51, 52}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{52, 53}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{53, 54}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{54, 55}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{55, 56}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{56, 57}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{57, 58}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{58, 59}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{59, 60}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{60, 61}]}, 
       {Hue[0, 1, 1], LineBox[{61, 62}]}, 
       {Hue[0, 1, 1], LineBox[{62, 63}]}, 
       {Hue[0, 1, 1], LineBox[{63, 64}]}, 
       {Hue[0, 1, 1], LineBox[{64, 65}]}, 
       {Hue[0, 1, 1], LineBox[{65, 66}]}, 
       {Hue[0, 1, 1], LineBox[{66, 67}]}, 
       {Hue[0, 1, 1], LineBox[{67, 68}]}, 
       {Hue[0, 1, 1], LineBox[{68, 69}]}, 
       {Hue[0, 1, 1], LineBox[{69, 70}]}, 
       {Hue[0, 1, 1], LineBox[{70, 71}]}}}},
    VertexColors->{
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1]}], {{}, {}, 
    {Hue[0.67, 0.6, 0.6], LineBox[CompressedData["
1:eJwVmnk8VO8XxzFji5KQCEUSSiotJM5FkZSlsiRJisjuaw/JniVLtrIrokWy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      "]]}}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  Epilog->LineBox[
    NCache[{{0, 0.}, {
       Rational[3, 70], 0.3090169943749479}, {
       Rational[3, 35], 0.5877852522924735}, {
       Rational[9, 70], 0.8090169943749477}, {
       Rational[6, 35], 0.9510565162951538}, {
       Rational[3, 14], 1.}, {
       Rational[9, 35], 0.9510565162951534}, {
       Rational[3, 10], 0.8090169943749472}, {
       Rational[12, 35], 0.5877852522924728}, {
       Rational[27, 70], 0.30901699437494706`}, {
       Rational[3, 7], 0.}, {
       Rational[33, 70], 0.3090169943749478}, {
       Rational[18, 35], 0.5877852522924734}, {
       Rational[39, 70], 0.8090169943749477}, {
       Rational[3, 5], 0.9510565162951536}, {
       Rational[9, 14], 1.}, {
       Rational[24, 35], 0.9510565162951535}, {
       Rational[51, 70], 0.8090169943749468}, {
       Rational[27, 35], 0.5877852522924721}, {
       Rational[57, 70], 0.3090169943749472}, {
       Rational[6, 7], 0.}, {
       Rational[9, 10], 0.3090169943749476}, {
       Rational[33, 35], 0.587785252292474}, {
       Rational[69, 70], 0.8090169943749476}, {
       Rational[36, 35], 0.9510565162951536}, {
       Rational[15, 14], 1.}, {
       Rational[39, 35], 0.9510565162951535}, {
       Rational[81, 70], 0.8090169943749468}, {
       Rational[6, 5], 0.5877852522924722}, {
       Rational[87, 70], 0.3090169943749469}, {
       Rational[9, 7], 0.}, {
       Rational[93, 70], 0.30901699437494795`}, {
       Rational[48, 35], 0.5877852522924736}, {
       Rational[99, 70], 0.8090169943749478}, {
       Rational[51, 35], 0.9510565162951538}, {
       Rational[3, 2], 1.}, {
       Rational[54, 35], 0.9510565162951533}, {
       Rational[111, 70], 0.8090169943749471}, {
       Rational[57, 35], 0.5877852522924726}, {
       Rational[117, 70], 0.3090169943749468}, {
       Rational[12, 7], 0.}, {
       Rational[123, 70], 0.3090169943749481}, {
       Rational[9, 5], 0.5877852522924737}, {
       Rational[129, 70], 0.8090169943749479}, {
       Rational[66, 35], 0.9510565162951538}, {
       Rational[27, 14], 1.}, {
       Rational[69, 35], 0.9510565162951533}, {
       Rational[141, 70], 0.809016994374947}, {
       Rational[72, 35], 0.5877852522924725}, {
       Rational[21, 10], 0.3090169943749467}, {
       Rational[15, 7], 0.}, {
       Rational[153, 70], 0.3090169943749481}, {
       Rational[78, 35], 0.5877852522924734}, {
       Rational[159, 70], 0.8090169943749479}, {
       Rational[81, 35], 0.9510565162951538}, {
       Rational[33, 14], 1.}, {
       Rational[12, 5], 0.9510565162951533}, {
       Rational[171, 70], 0.809016994374947}, {
       Rational[87, 35], 0.5877852522924726}, {
       Rational[177, 70], 0.3090169943749468}, {
       Rational[18, 7], 0.}, {
       Rational[183, 70], 0.3090169943749489}, {
       Rational[93, 35], 0.5877852522924736}, {
       Rational[27, 10], 0.8090169943749478}, {
       Rational[96, 35], 0.9510565162951538}, {
       Rational[39, 14], 1.}, {
       Rational[99, 35], 0.9510565162951531}, {
       Rational[201, 70], 0.8090169943749477}, {
       Rational[102, 35], 0.5877852522924734}, {
       Rational[207, 70], 0.30901699437494606`}, {3, 0.}}, {{0, 0.}, {
      0.04285714285714286, 0.3090169943749479}, {0.08571428571428572, 
      0.5877852522924735}, {0.12857142857142856`, 0.8090169943749477}, {
      0.17142857142857143`, 0.9510565162951538}, {0.21428571428571427`, 1.}, {
      0.2571428571428571, 0.9510565162951534}, {0.3, 0.8090169943749472}, {
      0.34285714285714286`, 0.5877852522924728}, {0.38571428571428573`, 
      0.30901699437494706`}, {0.42857142857142855`, 0.}, {0.4714285714285714, 
      0.3090169943749478}, {0.5142857142857142, 0.5877852522924734}, {
      0.5571428571428572, 0.8090169943749477}, {0.6, 0.9510565162951536}, {
      0.6428571428571429, 1.}, {0.6857142857142857, 0.9510565162951535}, {
      0.7285714285714285, 0.8090169943749468}, {0.7714285714285715, 
      0.5877852522924721}, {0.8142857142857143, 0.3090169943749472}, {
      0.8571428571428571, 0.}, {0.9, 0.3090169943749476}, {0.9428571428571428,
       0.587785252292474}, {0.9857142857142858, 0.8090169943749476}, {
      1.0285714285714285`, 0.9510565162951536}, {1.0714285714285714`, 1.}, {
      1.1142857142857143`, 0.9510565162951535}, {1.1571428571428573`, 
      0.8090169943749468}, {1.2, 0.5877852522924722}, {1.2428571428571429`, 
      0.3090169943749469}, {1.2857142857142858`, 0.}, {1.3285714285714285`, 
      0.30901699437494795`}, {1.3714285714285714`, 0.5877852522924736}, {
      1.4142857142857144`, 0.8090169943749478}, {1.457142857142857, 
      0.9510565162951538}, {1.5, 1.}, {1.542857142857143, 
      0.9510565162951533}, {1.5857142857142856`, 0.8090169943749471}, {
      1.6285714285714286`, 0.5877852522924726}, {1.6714285714285715`, 
      0.3090169943749468}, {1.7142857142857142`, 0.}, {1.7571428571428571`, 
      0.3090169943749481}, {1.8, 0.5877852522924737}, {1.8428571428571427`, 
      0.8090169943749479}, {1.8857142857142857`, 0.9510565162951538}, {
      1.9285714285714286`, 1.}, {1.9714285714285715`, 0.9510565162951533}, {
      2.0142857142857142`, 0.809016994374947}, {2.057142857142857, 
      0.5877852522924725}, {2.1, 0.3090169943749467}, {2.142857142857143, 
      0.}, {2.1857142857142855`, 0.3090169943749481}, {2.2285714285714286`, 
      0.5877852522924734}, {2.2714285714285714`, 0.8090169943749479}, {
      2.3142857142857145`, 0.9510565162951538}, {2.357142857142857, 1.}, {2.4,
       0.9510565162951533}, {2.442857142857143, 0.809016994374947}, {
      2.4857142857142858`, 0.5877852522924726}, {2.5285714285714285`, 
      0.3090169943749468}, {2.5714285714285716`, 0.}, {2.6142857142857143`, 
      0.3090169943749489}, {2.657142857142857, 0.5877852522924736}, {2.7, 
      0.8090169943749478}, {2.742857142857143, 0.9510565162951538}, {
      2.7857142857142856`, 1.}, {2.8285714285714287`, 0.9510565162951531}, {
      2.8714285714285714`, 0.8090169943749477}, {2.914285714285714, 
      0.5877852522924734}, {2.9571428571428573`, 0.30901699437494606`}, {
      3, 0.}}]],
  PlotRange->Automatic,
  PlotRangeClipping->True,
  PlotRangePadding->{
    Scaled[0.02], 
    Scaled[0.02]}]], "Output",
 CellChangeTimes->{3.391857494046875*^9, 3.39185761578125*^9},
 CellLabel->"Out[55]="],

Cell[BoxData[
 GraphicsBox[{GraphicsComplexBox[CompressedData["
1:eJx9k7FKQ0EQRTfJU1PZ+AXiB6SwSDmD5AvEgIpFiGhhk0r8FFFbK/UH/AEh
oFikshMDwgMxikEiUdGX2bswkevAYzns7t27b+8stjurO+UQQvP3K8bpOj4q
6lK28lp9cNUT465sb+5nS7cPsjypnlwMD08O+nmcv5NSs7Ew33kWU7mP40tc
35evNZvfndSjnLr9uaxH/Zvrop5kJZ4PHe/vNfkz/Tdpuf3D5M/Oe5/yN0q6
pveR/JneOPkzvU/ZiPqm9y0N5y8o/Nn6oHvu/5W07faX9Tzq23kV9f4yhT/T
y3Qc/ZnejJ65+81qy73PnNadv2ryh/f9n7uU8f6MkQfGPh9/GXlhjPwwRp4Y
I1+MkS/GyBtj5I8x8sjYakQZeWWM/DJGnhkj34yRb8bIO2PknzH6gbFVRhn9
whj9wxj9xBj9xfgHuNiDCg==
    "], {
     {GrayLevel[0.5], EdgeForm[None], 
      GraphicsGroupBox[{
        PolygonBox[{{1, 32, 2}, {10, 49, 11}, {11, 50, 12}, {21, 68, 22}, {30,
          85, 31}, {20, 67, 21}}], 
        PolygonBox[{{22, 69, 70, 23}, {3, 35, 36, 4}, {2, 33, 34, 3}, {29, 83,
          84, 30}, {7, 43, 44, 8}, {9, 47, 48, 10}, {13, 53, 54, 14}, {4, 37, 
         38, 5}, {12, 51, 52, 13}, {8, 45, 46, 9}, {18, 63, 64, 19}, {6, 41, 
         42, 7}, {14, 55, 56, 15}, {28, 81, 82, 29}, {24, 73, 74, 25}, {23, 
         71, 72, 24}, {17, 61, 62, 18}, {27, 79, 80, 28}, {25, 75, 76, 26}, {
         19, 65, 66, 20}, {16, 59, 60, 17}, {26, 77, 78, 27}, {5, 39, 40, 
         6}, {15, 57, 58, 16}}]}]}, {{}, {}, {
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{1, 2}]}, 
       {Hue[0, 1, 1], LineBox[{2, 3}]}, 
       {Hue[0, 1, 1], LineBox[{3, 4}]}, 
       {Hue[0, 1, 1], LineBox[{4, 5}]}, 
       {Hue[0, 1, 1], LineBox[{5, 6}]}, 
       {Hue[0, 1, 1], LineBox[{6, 7}]}, 
       {Hue[0, 1, 1], LineBox[{7, 8}]}, 
       {Hue[0, 1, 1], LineBox[{8, 9}]}, 
       {Hue[0, 1, 1], LineBox[{9, 10}]}, 
       {Hue[0, 1, 1], LineBox[{10, 11}]}, 
       {Hue[0, 1, 1], LineBox[{11, 12}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{12, 13}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{13, 14}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{14, 15}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{15, 16}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{16, 17}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{17, 18}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{18, 19}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{19, 20}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{20, 21}]}, 
       {Hue[0, 1, 1], LineBox[{21, 22}]}, 
       {Hue[0, 1, 1], LineBox[{22, 23}]}, 
       {Hue[0, 1, 1], LineBox[{23, 24}]}, 
       {Hue[0, 1, 1], LineBox[{24, 25}]}, 
       {Hue[0, 1, 1], LineBox[{25, 26}]}, 
       {Hue[0, 1, 1], LineBox[{26, 27}]}, 
       {Hue[0, 1, 1], LineBox[{27, 28}]}, 
       {Hue[0, 1, 1], LineBox[{28, 29}]}, 
       {Hue[0, 1, 1], LineBox[{29, 30}]}, 
       {Hue[0, 1, 1], LineBox[{30, 31}]}}}},
    VertexColors->{
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1], 
      Hue[0, 1, 1]}], {{}, {}, 
    {Hue[0.67, 0.6, 0.6], LineBox[CompressedData["
1:eJwVmnk8VO8XxzFji5KQCEUSSiotJM5FkZSlsiRJisjuaw/JniVLtrIrokWy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      "]]}}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  Epilog->LineBox[
    NCache[{{0, 0.}, {
       Rational[1, 10], 0.3090169943749481}, {
       Rational[1, 5], 0.5877852522924737}, {
       Rational[3, 10], 0.8090169943749479}, {
       Rational[2, 5], 0.9510565162951538}, {
       Rational[1, 2], 1.}, {
       Rational[3, 5], 0.9510565162951533}, {
       Rational[7, 10], 0.809016994374947}, {
       Rational[4, 5], 0.5877852522924725}, {
       Rational[9, 10], 0.3090169943749467}, {1, 0.}, {
       Rational[11, 10], 0.3090169943749481}, {
       Rational[6, 5], 0.5877852522924734}, {
       Rational[13, 10], 0.8090169943749479}, {
       Rational[7, 5], 0.9510565162951538}, {
       Rational[3, 2], 1.}, {
       Rational[8, 5], 0.9510565162951533}, {
       Rational[17, 10], 0.809016994374947}, {
       Rational[9, 5], 0.5877852522924726}, {
       Rational[19, 10], 0.3090169943749468}, {2, 0.}, {
       Rational[21, 10], 0.3090169943749489}, {
       Rational[11, 5], 0.5877852522924736}, {
       Rational[23, 10], 0.8090169943749478}, {
       Rational[12, 5], 0.9510565162951538}, {
       Rational[5, 2], 1.}, {
       Rational[13, 5], 0.9510565162951531}, {
       Rational[27, 10], 0.8090169943749477}, {
       Rational[14, 5], 0.5877852522924734}, {
       Rational[29, 10], 0.30901699437494606`}, {3, 0.}}, {{0, 0.}, {0.1, 
      0.3090169943749481}, {0.2, 0.5877852522924737}, {0.3, 
      0.8090169943749479}, {0.4, 0.9510565162951538}, {0.5, 1.}, {0.6, 
      0.9510565162951533}, {0.7, 0.809016994374947}, {0.8, 
      0.5877852522924725}, {0.9, 0.3090169943749467}, {1, 0.}, {1.1, 
      0.3090169943749481}, {1.2, 0.5877852522924734}, {1.3, 
      0.8090169943749479}, {1.4, 0.9510565162951538}, {1.5, 1.}, {1.6, 
      0.9510565162951533}, {1.7, 0.809016994374947}, {1.8, 
      0.5877852522924726}, {1.9, 0.3090169943749468}, {2, 0.}, {2.1, 
      0.3090169943749489}, {2.2, 0.5877852522924736}, {2.3, 
      0.8090169943749478}, {2.4, 0.9510565162951538}, {2.5, 1.}, {2.6, 
      0.9510565162951531}, {2.7, 0.8090169943749477}, {2.8, 
      0.5877852522924734}, {2.9, 0.30901699437494606`}, {3, 0.}}]],
  PlotRange->Automatic,
  PlotRangeClipping->True,
  PlotRangePadding->{
    Scaled[0.02], 
    Scaled[0.02]}]], "Output",
 CellChangeTimes->{3.391857494046875*^9, 3.391857615953125*^9},
 CellLabel->"Out[56]="],

Cell[BoxData[
 GraphicsBox[{
   GraphicsComplexBox[{{1., 0.}, {1.1, 0.3090169943749481}, {1.2, 
    0.5877852522924734}, {1.3, 0.8090169943749479}, {1.4, 
    0.9510565162951538}, {1.5, 1.}, {1.6, 0.9510565162951533}, {1.7, 
    0.809016994374947}, {1.8, 0.5877852522924726}, {1.9, 
    0.3090169943749468}, {2., 0.}, {1.1, 0.}, {1.1, 0.}, {1.2, 0.}, {1.2, 
    0.}, {1.3, 0.}, {1.3, 0.}, {1.4, 0.}, {1.4, 0.}, {1.5, 0.}, {1.5, 0.}, {
    1.6, 0.}, {1.6, 0.}, {1.7, 0.}, {1.7, 0.}, {1.8, 0.}, {1.8, 0.}, {1.9, 
    0.}, {1.9, 0.}}, {
     {GrayLevel[0.5], EdgeForm[None], 
      GraphicsGroupBox[{PolygonBox[{{1, 12, 2}, {10, 29, 11}}], 
        PolygonBox[{{7, 23, 24, 8}, {2, 13, 14, 3}, {9, 27, 28, 10}, {4, 17, 
         18, 5}, {3, 15, 16, 4}, {6, 21, 22, 7}, {8, 25, 26, 9}, {5, 19, 20, 
         6}}]}]}, {{}, {}, {
       {Hue[0, 1, 1], LineBox[{1, 2}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{2, 3}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{3, 4}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{4, 5}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{5, 6}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{6, 7}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{7, 8}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{8, 9}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{9, 10}]}, 
       {Hue[
         NCache[
          Rational[1, 2], 0.5], 1, 1], LineBox[{10, 11}]}}}},
    VertexColors->{
      Hue[0, 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[0, 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1], 
      Hue[
       Rational[1, 2], 1, 1]}], {{}, {}, 
    {Hue[0.67, 0.6, 0.6], LineBox[CompressedData["
1:eJwVmnk8VO8XxzFji5KQCEUSSiotJM5FkZSlsiRJisjuaw/JniVLtrIrokWy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      "]]}}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{1., 0},
  Epilog->LineBox[
    NCache[{{1, 0.}, {
       Rational[11, 10], 0.3090169943749481}, {
       Rational[6, 5], 0.5877852522924734}, {
       Rational[13, 10], 0.8090169943749479}, {
       Rational[7, 5], 0.9510565162951538}, {
       Rational[3, 2], 1.}, {
       Rational[8, 5], 0.9510565162951533}, {
       Rational[17, 10], 0.809016994374947}, {
       Rational[9, 5], 0.5877852522924726}, {
       Rational[19, 10], 0.3090169943749468}, {2, 0.}}, {{1, 0.}, {1.1, 
      0.3090169943749481}, {1.2, 0.5877852522924734}, {1.3, 
      0.8090169943749479}, {1.4, 0.9510565162951538}, {1.5, 1.}, {1.6, 
      0.9510565162951533}, {1.7, 0.809016994374947}, {1.8, 
      0.5877852522924726}, {1.9, 0.3090169943749468}, {2, 0.}}]],
  PlotRange->Automatic,
  PlotRangeClipping->True,
  PlotRangePadding->{
    Scaled[0.02], 
    Scaled[0.02]}]], "Output",
 CellChangeTimes->{3.391857494046875*^9, 3.391857616171875*^9},
 CellLabel->"Out[57]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Show", "[", 
  RowBox[{
   RowBox[{"QArgColorPlot", "[", 
    RowBox[{
     RowBox[{"Sin", "[", 
      RowBox[{"x", "+", "\[ImaginaryI]"}], "]"}], ",", 
     RowBox[{"{", 
      RowBox[{"x", ",", 
       RowBox[{"-", "3"}], ",", "3"}], "}"}], ",", 
     RowBox[{"QBottomLine", "\[Rule]", "1"}], ",", 
     RowBox[{"Epilog", "\[Rule]", 
      RowBox[{"Line", "[", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"{", 
          RowBox[{
           RowBox[{"-", "3"}], ",", "1"}], "}"}], ",", 
         RowBox[{"{", 
          RowBox[{"3", ",", "1"}], "}"}]}], "}"}], "]"}]}]}], "]"}], ",", 
   RowBox[{"QArgColorPlot", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Cos", "[", 
       RowBox[{"x", "+", "\[ImaginaryI]"}], "]"}], "-", 
      SuperscriptBox["\[ExponentialE]", 
       RowBox[{"\[ImaginaryI]", " ", 
        RowBox[{"Arg", "[", 
         RowBox[{"Cos", "[", 
          RowBox[{"x", "+", "\[ImaginaryI]"}], "]"}], "]"}]}]]}], ",", 
     RowBox[{"{", 
      RowBox[{"x", ",", 
       RowBox[{"-", "3"}], ",", "3"}], "}"}]}], "]"}], ",", 
   RowBox[{"PlotRange", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"-", "0.2"}], ",", "1.7"}], "}"}]}], ",", 
   RowBox[{"Frame", "\[Rule]", "True"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.3918575193125*^9, 3.39185752365625*^9}, 
   3.39185755484375*^9},
 CellLabel->"In[46]:="],

Cell[BoxData[
 GraphicsBox[{GraphicsComplexBox[CompressedData["
1:eJx13HlYjP37PvCKmmwhRCFLlmSJZK1cQ9mzJyJkFyFEKUqJVMq+tCChRSlk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    "], {
     {GrayLevel[0.5], EdgeForm[None], 
      GraphicsGroupBox[PolygonBox[CompressedData["
1:eJw12nWgFlUeh/Fz5n3nvXR3d3d3dwqIiIiIgIiIiIiIqIiIgIhISamASElJ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        "]]]}, {{}, {}, {
       {Hue[0.7205518817558729, 1, 1], LineBox[{1, 2}]}, 
       {Hue[0.7201728632810981, 1, 1], LineBox[{2, 3}]}, 
       {Hue[0.7197939855871971, 1, 1], LineBox[{3, 4}]}, 
       {Hue[0.7190366591317676, 1, 1], LineBox[{4, 5}]}, 
       {Hue[0.7175237677526916, 1, 1], LineBox[{5, 6}]}, 
       {Hue[0.7145053924996501, 1, 1], LineBox[{6, 7}]}, 
       {Hue[0.7141288210543033, 1, 1], LineBox[{7, 8}]}, 
       {Hue[0.7137524160490083, 1, 1], LineBox[{8, 9}]}, 
       {Hue[0.7130001115521821, 1, 1], LineBox[{9, 10}]}, 
       {Hue[0.7114975675954697, 1, 1], LineBox[{10, 11}]}, 
       {Hue[0.7085010779576801, 1, 1], LineBox[{11, 12}]}, 
       {Hue[0.7081273517673412, 1, 1], LineBox[{12, 13}]}, 
       {Hue[0.707753815997507, 1, 1], LineBox[{13, 14}]}, 
       {Hue[0.7070073214666093, 1, 1], LineBox[{14, 15}]}, 
       {Hue[0.7055166802582684, 1, 1], LineBox[{15, 16}]}, 
       {Hue[0.7025451015155051, 1, 1], LineBox[{16, 17}]}, 
       {Hue[0.6966431533753163, 1, 1], LineBox[{17, 18}]}, 
       {Hue[0.6962453407646028, 1, 1], LineBox[{18, 19}]}, 
       {Hue[0.6958478017085719, 1, 1], LineBox[{19, 20}]}, 
       {Hue[0.6950535502982287, 1, 1], LineBox[{20, 21}]}, 
       {Hue[0.6934683959066561, 1, 1], LineBox[{21, 22}]}, 
       {Hue[0.69031180412586, 1, 1], LineBox[{22, 23}]}, 
       {Hue[0.6840559180388815, 1, 1], LineBox[{23, 24}]}, 
       {Hue[0.6717902321513325, 1, 1], LineBox[{24, 25}]}, 
       {Hue[0.6714375103634247, 1, 1], LineBox[{25, 26}]}, 
       {Hue[0.671085090804056, 1, 1], LineBox[{26, 27}]}, 
       {Hue[0.6703811608448678, 1, 1], LineBox[{27, 28}]}, 
       {Hue[0.668976954394541, 1, 1], LineBox[{28, 29}]}, 
       {Hue[0.666183282738352, 1, 1], LineBox[{29, 30}]}, 
       {Hue[0.6606558087846726, 1, 1], LineBox[{30, 31}]}, 
       {Hue[0.6498458025017817, 1, 1], LineBox[{31, 32}]}, 
       {Hue[0.6296151700839989, 1, 1], LineBox[{32, 33}]}, 
       {Hue[0.6090856243791776, 1, 1], LineBox[{33, 34}]}, 
       {Hue[0.5911506349521957, 1, 1], LineBox[{34, 35}]}, 
       {Hue[0.5728738758396617, 1, 1], LineBox[{35, 36}]}, 
       {Hue[0.5726018283453903, 1, 1], LineBox[{36, 37}]}, 
       {Hue[0.5723300141524164, 1, 1], LineBox[{37, 38}]}, 
       {Hue[0.5717870824013804, 1, 1], LineBox[{38, 39}]}, 
       {Hue[0.570703982533093, 1, 1], LineBox[{39, 40}]}, 
       {Hue[0.5685486537726839, 1, 1], LineBox[{40, 41}]}, 
       {Hue[0.5642800076517256, 1, 1], LineBox[{41, 42}]}, 
       {Hue[0.5558989435773015, 1, 1], LineBox[{42, 43}]}, 
       {Hue[0.5556575555942079, 1, 1], LineBox[{43, 44}]}, 
       {Hue[0.5554163247938143, 1, 1], LineBox[{44, 45}]}, 
       {Hue[0.5549343320656529, 1, 1], LineBox[{45, 46}]}, 
       {Hue[0.5539722033773958, 1, 1], LineBox[{46, 47}]}, 
       {Hue[0.5520552233913043, 1, 1], LineBox[{47, 48}]}, 
       {Hue[0.5482491773582174, 1, 1], LineBox[{48, 49}]}, 
       {Hue[0.5480124880738455, 1, 1], LineBox[{49, 50}]}, 
       {Hue[0.5477759346340769, 1, 1], LineBox[{50, 51}]}, 
       {Hue[0.5473032326321359, 1, 1], LineBox[{51, 52}]}, 
       {Hue[0.5463594295560026, 1, 1], LineBox[{52, 53}]}, 
       {Hue[0.5444780786214425, 1, 1], LineBox[{53, 54}]}, 
       {Hue[0.5407392124960726, 1, 1], LineBox[{54, 55}]}, 
       {Hue[0.5404869192549508, 1, 1], LineBox[{55, 56}]}, 
       {Hue[0.5402347608129789, 1, 1], LineBox[{56, 57}]}, 
       {Hue[0.5397308449751576, 1, 1], LineBox[{57, 58}]}, 
       {Hue[0.5387245940428427, 1, 1], LineBox[{58, 59}]}, 
       {Hue[0.5367182278963121, 1, 1], LineBox[{59, 60}]}, 
       {Hue[0.5327285448595175, 1, 1], LineBox[{60, 61}]}, 
       {Hue[0.5324801512102756, 1, 1], LineBox[{61, 62}]}, 
       {Hue[0.5322318657097549, 1, 1], LineBox[{62, 63}]}, 
       {Hue[0.5317356158456428, 1, 1], LineBox[{63, 64}]}, 
       {Hue[0.5307443775187145, 1, 1], LineBox[{64, 65}]}, 
       {Hue[0.5287667615698055, 1, 1], LineBox[{65, 66}]}, 
       {Hue[0.5248294975419259, 1, 1], LineBox[{66, 67}]}, 
       {Hue[0.5246004552449229, 1, 1], LineBox[{67, 68}]}, 
       {Hue[0.5243714843111781, 1, 1], LineBox[{68, 69}]}, 
       {Hue[0.5239137538730607, 1, 1], LineBox[{69, 70}]}, 
       {Hue[0.5229991201035351, 1, 1], LineBox[{70, 71}]}, 
       {Hue[0.5211730122817956, 1, 1], LineBox[{71, 72}]}, 
       {Hue[0.5209450320969299, 1, 1], LineBox[{72, 73}]}, 
       {Hue[0.5207171126580099, 1, 1], LineBox[{73, 74}]}, 
       {Hue[0.5202614533707344, 1, 1], LineBox[{74, 75}]}, 
       {Hue[0.5193508346382325, 1, 1], LineBox[{75, 76}]}, 
       {Hue[0.5175322485357188, 1, 1], LineBox[{76, 77}]}, 
       {Hue[0.5173051609668913, 1, 1], LineBox[{77, 78}]}, 
       {Hue[0.5170781235763294, 1, 1], LineBox[{78, 79}]}, 
       {Hue[0.516624196694218, 1, 1], LineBox[{79, 80}]}, 
       {Hue[0.5157169160846813, 1, 1], LineBox[{80, 81}]}, 
       {Hue[0.5154902114994349, 1, 1], LineBox[{81, 82}]}, 
       {Hue[0.5152635518250663, 1, 1], LineBox[{82, 83}]}, 
       {Hue[0.5148103645782285, 1, 1], LineBox[{83, 84}]}, 
       {Hue[0.5139045000852724, 1, 1], LineBox[{84, 85}]}, 
       {Hue[0.513678136379099, 1, 1], LineBox[{85, 86}]}, 
       {Hue[0.5134518123261133, 1, 1], LineBox[{86, 87}]}, 
       {Hue[0.5129992805535211, 1, 1], LineBox[{87, 88}]}, 
       {Hue[0.5120946639659323, 1, 1], LineBox[{88, 89}]}, 
       {Hue[0.5102870717187569, 1, 1], LineBox[{89, 90}]}, 
       {Hue[0.5100657056310821, 1, 1], LineBox[{90, 91}]}, 
       {Hue[0.5098443675824635, 1, 1], LineBox[{91, 92}]}, 
       {Hue[0.5094017731352407, 1, 1], LineBox[{92, 93}]}, 
       {Hue[0.50851689357583, 1, 1], LineBox[{93, 94}]}, 
       {Hue[0.5082957345349084, 1, 1], LineBox[{94, 95}]}, 
       {Hue[0.508074598601023, 1, 1], LineBox[{95, 96}]}, 
       {Hue[0.5076323935899204, 1, 1], LineBox[{96, 97}]}, 
       {Hue[0.5067482337474685, 1, 1], LineBox[{97, 98}]}, 
       {Hue[0.5065272423153528, 1, 1], LineBox[{98, 99}]}, 
       {Hue[0.5063062690632634, 1, 1], LineBox[{99, 100}]}, 
       {Hue[0.5058643746369205, 1, 1], LineBox[{100, 101}]}, 
       {Hue[0.5056434522316935, 1, 1], LineBox[{101, 102}]}, 
       {Hue[0.5054225455445416, 1, 1], LineBox[{102, 103}]}, 
       {Hue[0.504980776863122, 1, 1], LineBox[{103, 104}]}, 
       {Hue[0.504759913638311, 1, 1], LineBox[{104, 105}]}, 
       {Hue[0.5045390636704838, 1, 1], LineBox[{105, 106}]}, 
       {Hue[0.5040974010452097, 1, 1], LineBox[{106, 107}]}, 
       {Hue[0.5038765871575825, 1, 1], LineBox[{107, 108}]}, 
       {Hue[0.5036557840665759, 1, 1], LineBox[{108, 109}]}, 
       {Hue[0.5032142078144854, 1, 1], LineBox[{109, 110}]}, 
       {Hue[0.5029934334235187, 1, 1], LineBox[{110, 111}]}, 
       {Hue[0.5027726673694013, 1, 1], LineBox[{111, 112}]}, 
       {Hue[0.5025519090372679, 1, 1], LineBox[{112, 113}]}, 
       {Hue[0.5023311578122791, 1, 1], LineBox[{113, 114}]}, 
       {Hue[0.5021104130796202, 1, 1], LineBox[{114, 115}]}, 
       {Hue[0.5018896742244989, 1, 1], LineBox[{115, 116}]}, 
       {Hue[0.5016689406321433, 1, 1], LineBox[{116, 117}]}, 
       {Hue[0.501448211687799, 1, 1], LineBox[{117, 118}]}, 
       {Hue[0.5012274867767283, 1, 1], LineBox[{118, 119}]}, 
       {Hue[0.5010067652842074, 1, 1], LineBox[{119, 120}]}, 
       {Hue[0.5007860465955236, 1, 1], LineBox[{120, 121}]}, 
       {Hue[0.5005653300959744, 1, 1], LineBox[{121, 122}]}, 
       {Hue[0.5003446151708648, 1, 1], LineBox[{122, 123}]}, 
       {Hue[0.5001239012055051, 1, 1], LineBox[{123, 124}]}, 
       {Hue[0.499903187585209, 1, 1], LineBox[{124, 125}]}, 
       {Hue[0.4996824736952913, 1, 1], LineBox[{125, 126}]}, 
       {Hue[0.4994617589210659, 1, 1], LineBox[{126, 127}]}, 
       {Hue[0.4992410426478438, 1, 1], LineBox[{127, 128}]}, 
       {Hue[0.4990203242609306, 1, 1], LineBox[{128, 129}]}, 
       {Hue[0.4987996031456248, 1, 1], LineBox[{129, 130}]}, 
       {Hue[0.49857887868721534`, 1, 1], LineBox[{130, 131}]}, 
       {Hue[0.4983581502709798, 1, 1], LineBox[{131, 132}]}, 
       {Hue[0.4981374172821819, 1, 1], LineBox[{132, 133}]}, 
       {Hue[0.4979166791060699, 1, 1], LineBox[{133, 134}]}, 
       {Hue[0.49769593512787397`, 1, 1], LineBox[{134, 135}]}, 
       {Hue[0.4974751847328044, 1, 1], LineBox[{135, 136}]}, 
       {Hue[0.49725442730604935`, 1, 1], LineBox[{136, 137}]}, 
       {Hue[0.4970336622327728, 1, 1], LineBox[{137, 138}]}, 
       {Hue[0.49681288889811237`, 1, 1], LineBox[{138, 139}]}, 
       {Hue[0.49659210668717735`, 1, 1], LineBox[{139, 140}]}, 
       {Hue[0.4961505131767657, 1, 1], LineBox[{140, 141}]}, 
       {Hue[0.49591092983354623`, 1, 1], LineBox[{141, 142}]}, 
       {Hue[0.4956713330834243, 1, 1], LineBox[{142, 143}]}, 
       {Hue[0.4951920962198805, 1, 1], LineBox[{143, 144}]}, 
       {Hue[0.49495245453501585`, 1, 1], LineBox[{144, 145}]}, 
       {Hue[0.494712796300357, 1, 1], LineBox[{145, 146}]}, 
       {Hue[0.49423342703805917`, 1, 1], LineBox[{146, 147}]}, 
       {Hue[0.4939937144384435, 1, 1], LineBox[{147, 148}]}, 
       {Hue[0.49375398214507416`, 1, 1], LineBox[{148, 149}]}, 
       {Hue[0.4932744553322717, 1, 1], LineBox[{149, 150}]}, 
       {Hue[0.4923151307840086, 1, 1], LineBox[{150, 151}]}, 
       {Hue[0.4920752386164939, 1, 1], LineBox[{151, 152}]}, 
       {Hue[0.4918353204633067, 1, 1], LineBox[{152, 153}]}, 
       {Hue[0.491355403052108, 1, 1], LineBox[{153, 154}]}, 
       {Hue[0.4903952217696123, 1, 1], LineBox[{154, 155}]}, 
       {Hue[0.49015509967662196`, 1, 1], LineBox[{155, 156}]}, 
       {Hue[0.4899149452993977, 1, 1], LineBox[{156, 157}]}, 
       {Hue[0.4894345365406593, 1, 1], LineBox[{157, 158}]}, 
       {Hue[0.488473296937411, 1, 1], LineBox[{158, 159}]}, 
       {Hue[0.4882328945028128, 1, 1], LineBox[{159, 160}]}, 
       {Hue[0.4879924534772371, 1, 1], LineBox[{160, 161}]}, 
       {Hue[0.4875114524970251, 1, 1], LineBox[{161, 162}]}, 
       {Hue[0.48654895271867454`, 1, 1], LineBox[{162, 163}]}, 
       {Hue[0.4863082194556099, 1, 1], LineBox[{163, 164}]}, 
       {Hue[0.48606744128514545`, 1, 1], LineBox[{164, 165}]}, 
       {Hue[0.4855857470606176, 1, 1], LineBox[{165, 166}]}, 
       {Hue[0.48462178493732505`, 1, 1], LineBox[{166, 167}]}, 
       {Hue[0.4843806702764197, 1, 1], LineBox[{167, 168}]}, 
       {Hue[0.4841395043805711, 1, 1], LineBox[{168, 169}]}, 
       {Hue[0.4836570157166681, 1, 1], LineBox[{169, 170}]}, 
       {Hue[0.4826913887171722, 1, 1], LineBox[{170, 171}]}, 
       {Hue[0.480757358393055, 1, 1], LineBox[{171, 172}]}, 
       {Hue[0.4805315263020154, 1, 1], LineBox[{172, 173}]}, 
       {Hue[0.4803056385978369, 1, 1], LineBox[{173, 174}]}, 
       {Hue[0.47985369376570597`, 1, 1], LineBox[{174, 175}]}, 
       {Hue[0.47894910830569265`, 1, 1], LineBox[{175, 176}]}, 
       {Hue[0.47713700924086944`, 1, 1], LineBox[{176, 177}]}, 
       {Hue[0.4769102097107648, 1, 1], LineBox[{177, 178}]}, 
       {Hue[0.47668334420836855`, 1, 1], LineBox[{178, 179}]}, 
       {Hue[0.47622941269007835`, 1, 1], LineBox[{179, 180}]}, 
       {Hue[0.47532072941166953`, 1, 1], LineBox[{180, 181}]}, 
       {Hue[0.47349993620919084`, 1, 1], LineBox[{181, 182}]}, 
       {Hue[0.47327200311344897`, 1, 1], LineBox[{182, 183}]}, 
       {Hue[0.47304399363458505`, 1, 1], LineBox[{183, 184}]}, 
       {Hue[0.4725877429172142, 1, 1], LineBox[{184, 185}]}, 
       {Hue[0.47167429615942985`, 1, 1], LineBox[{185, 186}]}, 
       {Hue[0.4698434748853157, 1, 1], LineBox[{186, 187}]}, 
       {Hue[0.46616494645249384`, 1, 1], LineBox[{187, 188}]}, 
       {Hue[0.4659148691575225, 1, 1], LineBox[{188, 189}]}, 
       {Hue[0.46566467738003314`, 1, 1], LineBox[{189, 190}]}, 
       {Hue[0.4651639470144475, 1, 1], LineBox[{190, 191}]}, 
       {Hue[0.4641610754806451, 1, 1], LineBox[{191, 192}]}, 
       {Hue[0.46214950043805014`, 1, 1], LineBox[{192, 193}]}, 
       {Hue[0.4581015078231806, 1, 1], LineBox[{193, 194}]}, 
       {Hue[0.4578473484445837, 1, 1], LineBox[{194, 195}]}, 
       {Hue[0.4575930475155824, 1, 1], LineBox[{195, 196}]}, 
       {Hue[0.45708401760083006`, 1, 1], LineBox[{196, 197}]}, 
       {Hue[0.4560642216878684, 1, 1], LineBox[{197, 198}]}, 
       {Hue[0.4540174944156169, 1, 1], LineBox[{198, 199}]}, 
       {Hue[0.4498939651897446, 1, 1], LineBox[{199, 200}]}, 
       {Hue[0.4496395920714616, 1, 1], LineBox[{200, 201}]}, 
       {Hue[0.449385056150846, 1, 1], LineBox[{201, 202}]}, 
       {Hue[0.4488754926464263, 1, 1], LineBox[{202, 203}]}, 
       {Hue[0.4478543761796728, 1, 1], LineBox[{203, 204}]}, 
       {Hue[0.4458040028063383, 1, 1], LineBox[{204, 205}]}, 
       {Hue[0.44166923112517276`, 1, 1], LineBox[{205, 206}]}, 
       {Hue[0.4332518596871854, 1, 1], LineBox[{206, 207}]}, 
       {Hue[0.4169447860407514, 1, 1], LineBox[{207, 208}]}, 
       {Hue[0.39819770478061217`, 1, 1], LineBox[{208, 209}]}, 
       {Hue[0.3795344901876919, 1, 1], LineBox[{209, 210}]}, 
       {Hue[0.3578952936955762, 1, 1], LineBox[{210, 211}]}, 
       {Hue[0.35755137633603035`, 1, 1], LineBox[{211, 212}]}, 
       {Hue[0.35720709405282786`, 1, 1], LineBox[{212, 213}]}, 
       {Hue[0.3565174345978122, 1, 1], LineBox[{213, 214}]}, 
       {Hue[0.3551337357577113, 1, 1], LineBox[{214, 215}]}, 
       {Hue[0.3523488221943377, 1, 1], LineBox[{215, 216}]}, 
       {Hue[0.3467090755102115, 1, 1], LineBox[{216, 217}]}, 
       {Hue[0.33515298547996475`, 1, 1], LineBox[{217, 218}]}, 
       {Hue[0.334810643178493, 1, 1], LineBox[{218, 219}]}, 
       {Hue[0.3344679970554599, 1, 1], LineBox[{219, 220}]}, 
       {Hue[0.33378179518226486`, 1, 1], LineBox[{220, 221}]}, 
       {Hue[0.3324057662380327, 1, 1], LineBox[{221, 222}]}, 
       {Hue[0.32963932081463193`, 1, 1], LineBox[{222, 223}]}, 
       {Hue[0.3240498957473761, 1, 1], LineBox[{223, 224}]}, 
       {Hue[0.32369809827084883`, 1, 1], LineBox[{224, 225}]}, 
       {Hue[0.32334601565736026`, 1, 1], LineBox[{225, 226}]}, 
       {Hue[0.3226409979132937, 1, 1], LineBox[{226, 227}]}, 
       {Hue[0.32122757308827055`, 1, 1], LineBox[{227, 228}]}, 
       {Hue[0.3183873390867039, 1, 1], LineBox[{228, 229}]}, 
       {Hue[0.3126548323856401, 1, 1], LineBox[{229, 230}]}, 
       {Hue[0.31226394002872365`, 1, 1], LineBox[{230, 231}]}, 
       {Hue[0.3118727446949302, 1, 1], LineBox[{231, 232}]}, 
       {Hue[0.31108945018381584`, 1, 1], LineBox[{232, 233}]}, 
       {Hue[0.30951928205543044`, 1, 1], LineBox[{233, 234}]}, 
       {Hue[0.3063649293548802, 1, 1], LineBox[{234, 235}]}, 
       {Hue[0.30596934819271, 1, 1], LineBox[{235, 236}]}, 
       {Hue[0.3055734857590331, 1, 1], LineBox[{236, 237}]}, 
       {Hue[0.30478092290960934`, 1, 1], LineBox[{237, 238}]}, 
       {Hue[0.30319248675089605`, 1, 1], LineBox[{238, 239}]}, 
       {Hue[0.3000027138831849, 1, 1], LineBox[{239, 240}]}, 
       {Hue[0.29960281338186984`, 1, 1], LineBox[{240, 241}]}, 
       {Hue[0.29920265625388204`, 1, 1], LineBox[{241, 242}]}, 
       {Hue[0.29840157866966915`, 1, 1], LineBox[{242, 243}]}, 
       {Hue[0.2967964166611863, 1, 1], LineBox[{243, 244}]}, 
       {Hue[0.2935744461038008, 1, 1], LineBox[{244, 245}]}, 
       {Hue[0.2870871058179891, 1, 1], LineBox[{245, 246}]}, 
       {Hue[0.2867068783556848, 1, 1], LineBox[{246, 247}]}, 
       {Hue[0.28632647720527604`, 1, 1], LineBox[{247, 248}]}, 
       {Hue[0.28556516022402634`, 1, 1], LineBox[{248, 249}]}, 
       {Hue[0.284040512612144, 1, 1], LineBox[{249, 250}]}, 
       {Hue[0.28098352900329876`, 1, 1], LineBox[{250, 251}]}, 
       {Hue[0.2806007176236879, 1, 1], LineBox[{251, 252}]}, 
       {Hue[0.2802177588652316, 1, 1], LineBox[{252, 253}]}, 
       {Hue[0.27945140601201585`, 1, 1], LineBox[{253, 254}]}, 
       {Hue[0.2779170068960574, 1, 1], LineBox[{254, 255}]}, 
       {Hue[0.2748418233531421, 1, 1], LineBox[{255, 256}]}, 
       {Hue[0.27445686099902716`, 1, 1], LineBox[{256, 257}]}, 
       {Hue[0.2740717791307105, 1, 1], LineBox[{257, 258}]}, 
       {Hue[0.2733012640056608, 1, 1], LineBox[{258, 259}]}, 
       {Hue[0.2717588785561085, 1, 1], LineBox[{259, 260}]}, 
       {Hue[0.27137301040675016`, 1, 1], LineBox[{260, 261}]}, 
       {Hue[0.2709870371770742, 1, 1], LineBox[{261, 262}]}, 
       {Hue[0.2702147827817822, 1, 1], LineBox[{262, 263}]}, 
       {Hue[0.26866909362519453`, 1, 1], LineBox[{263, 264}]}, 
       {Hue[0.26828243617916075`, 1, 1], LineBox[{264, 265}]}, 
       {Hue[0.2678956883732717, 1, 1], LineBox[{265, 266}]}, 
       {Hue[0.2671219291190381, 1, 1], LineBox[{266, 267}]}, 
       {Hue[0.26557340830942106`, 1, 1], LineBox[{267, 268}]}, 
       {Hue[0.2651860802254811, 1, 1], LineBox[{268, 269}]}, 
       {Hue[0.264798676750572, 1, 1], LineBox[{269, 270}]}, 
       {Hue[0.2640236511774876, 1, 1], LineBox[{270, 271}]}, 
       {Hue[0.26247277855953854`, 1, 1], LineBox[{271, 272}]}, 
       {Hue[0.26209251788900434`, 1, 1], LineBox[{272, 273}]}, 
       {Hue[0.26171219931070255`, 1, 1], LineBox[{273, 274}]}, 
       {Hue[0.2609513956309302, 1, 1], LineBox[{274, 275}]}, 
       {Hue[0.2605709141353225, 1, 1], LineBox[{275, 276}]}, 
       {Hue[0.26019038194352423`, 1, 1], LineBox[{276, 277}]}, 
       {Hue[0.2594291727069773, 1, 1], LineBox[{277, 278}]}, 
       {Hue[0.2590484992850898, 1, 1], LineBox[{278, 279}]}, 
       {Hue[0.25866778241258415`, 1, 1], LineBox[{279, 280}]}, 
       {Hue[0.2579062255820407, 1, 1], LineBox[{280, 281}]}, 
       {Hue[0.2575253892614616, 1, 1], LineBox[{281, 282}]}, 
       {Hue[0.2571445167650305, 1, 1], LineBox[{282, 283}]}, 
       {Hue[0.2563826705368009, 1, 1], LineBox[{283, 284}]}, 
       {Hue[0.256001700454635, 1, 1], LineBox[{284, 285}]}, 
       {Hue[0.25562070149572913`, 1, 1], LineBox[{285, 286}]}, 
       {Hue[0.2552396754880936, 1, 1], LineBox[{286, 287}]}, 
       {Hue[0.25485862426087846`, 1, 1], LineBox[{287, 288}]}, 
       {Hue[0.2544775496442966, 1, 1], LineBox[{288, 289}]}, 
       {Hue[0.254096453469547, 1, 1], LineBox[{289, 290}]}, 
       {Hue[0.2537153375687379, 1, 1], LineBox[{290, 291}]}, 
       {Hue[0.2533342037748092, 1, 1], LineBox[{291, 292}]}, 
       {Hue[0.2529530539214558, 1, 1], LineBox[{292, 293}]}, 
       {Hue[0.2525718898430503, 1, 1], LineBox[{293, 294}]}, 
       {Hue[0.2521907133745654, 1, 1], LineBox[{294, 295}]}, 
       {Hue[0.2518095263514968, 1, 1], LineBox[{295, 296}]}, 
       {Hue[0.2514283306097857, 1, 1], LineBox[{296, 297}]}, 
       {Hue[0.2510471279857414, 1, 1], LineBox[{297, 298}]}, 
       {Hue[0.2506659203159637, 1, 1], LineBox[{298, 299}]}, 
       {Hue[0.25028470943726544`, 1, 1], LineBox[{299, 300}]}, 
       {Hue[0.24990349718659496`, 1, 1], LineBox[{300, 301}]}, 
       {Hue[0.24952228540095855`, 1, 1], LineBox[{301, 302}]}, 
       {Hue[0.24914107591734283`, 1, 1], LineBox[{302, 303}]}, 
       {Hue[0.24875987057263726`, 1, 1], LineBox[{303, 304}]}, 
       {Hue[0.24837867120355647`, 1, 1], LineBox[{304, 305}]}, 
       {Hue[0.2479974796465629, 1, 1], LineBox[{305, 306}]}, 
       {Hue[0.2476162977377892, 1, 1], LineBox[{306, 307}]}, 
       {Hue[0.2472351273129606, 1, 1], LineBox[{307, 308}]}, 
       {Hue[0.24685397020731778`, 1, 1], LineBox[{308, 309}]}, 
       {Hue[0.24647282825553923`, 1, 1], LineBox[{309, 310}]}, 
       {Hue[0.24609170329166405`, 1, 1], LineBox[{310, 311}]}, 
       {Hue[0.24571059714901458`, 1, 1], LineBox[{311, 312}]}, 
       {Hue[0.24532951166011932`, 1, 1], LineBox[{312, 313}]}, 
       {Hue[0.24494844865663576`, 1, 1], LineBox[{313, 314}]}, 
       {Hue[0.24418639742771617`, 1, 1], LineBox[{314, 315}]}, 
       {Hue[0.24380541286054708`, 1, 1], LineBox[{315, 316}]}, 
       {Hue[0.24342445809517, 1, 1], LineBox[{316, 317}]}, 
       {Hue[0.24266264527305542`, 1, 1], LineBox[{317, 318}]}, 
       {Hue[0.24228179086454416`, 1, 1], LineBox[{318, 319}]}, 
       {Hue[0.24190097355412485`, 1, 1], LineBox[{319, 320}]}, 
       {Hue[0.2411394575073841, 1, 1], LineBox[{320, 321}]}, 
       {Hue[0.2407587624068073, 1, 1], LineBox[{321, 322}]}, 
       {Hue[0.24037811167566045`, 1, 1], LineBox[{322, 323}]}, 
       {Hue[0.23961695057318147`, 1, 1], LineBox[{323, 324}]}, 
       {Hue[0.2380952404554197, 1, 1], LineBox[{324, 325}]}, 
       {Hue[0.23768268530800246`, 1, 1], LineBox[{325, 326}]}, 
       {Hue[0.23727019956300532`, 1, 1], LineBox[{326, 327}]}, 
       {Hue[0.23644544544623258`, 1, 1], LineBox[{327, 328}]}, 
       {Hue[0.23479687066374988`, 1, 1], LineBox[{328, 329}]}, 
       {Hue[0.23438493471701696`, 1, 1], LineBox[{329, 330}]}, 
       {Hue[0.23397308640684428`, 1, 1], LineBox[{330, 331}]}, 
       {Hue[0.23314966173624715`, 1, 1], LineBox[{331, 332}]}, 
       {Hue[0.23150396322258707`, 1, 1], LineBox[{332, 333}]}, 
       {Hue[0.23109279142615208`, 1, 1], LineBox[{333, 334}]}, 
       {Hue[0.23068172522628413`, 1, 1], LineBox[{334, 335}]}, 
       {Hue[0.22985991850312365`, 1, 1], LineBox[{335, 336}]}, 
       {Hue[0.2282176696727462, 1, 1], LineBox[{336, 337}]}, 
       {Hue[0.22493912100615956`, 1, 1], LineBox[{337, 338}]}, 
       {Hue[0.22452990240719808`, 1, 1], LineBox[{338, 339}]}, 
       {Hue[0.22412082428986946`, 1, 1], LineBox[{339, 340}]}, 
       {Hue[0.22330309800330872`, 1, 1], LineBox[{340, 341}]}, 
       {Hue[0.2216694243601005, 1, 1], LineBox[{341, 342}]}, 
       {Hue[0.21840965988103805`, 1, 1], LineBox[{342, 343}]}, 
       {Hue[0.21800294032724274`, 1, 1], LineBox[{343, 344}]}, 
       {Hue[0.21759639443153517`, 1, 1], LineBox[{344, 345}]}, 
       {Hue[0.21678383164015788`, 1, 1], LineBox[{345, 346}]}, 
       {Hue[0.21516087777741738`, 1, 1], LineBox[{346, 347}]}, 
       {Hue[0.2119240955782303, 1, 1], LineBox[{347, 348}]}, 
       {Hue[0.21154735354250084`, 1, 1], LineBox[{348, 349}]}, 
       {Hue[0.21117078972624026`, 1, 1], LineBox[{349, 350}]}, 
       {Hue[0.21041820282444174`, 1, 1], LineBox[{350, 351}]}, 
       {Hue[0.20891523405310197`, 1, 1], LineBox[{351, 352}]}, 
       {Hue[0.2059184474441574, 1, 1], LineBox[{352, 353}]}, 
       {Hue[0.19996402390747361`, 1, 1], LineBox[{353, 354}]}, 
       {Hue[0.19959370230262743`, 1, 1], LineBox[{354, 355}]}, 
       {Hue[0.1992236038568319, 1, 1], LineBox[{355, 356}]}, 
       {Hue[0.19848408153459976`, 1, 1], LineBox[{356, 357}]}, 
       {Hue[0.19700777035336184`, 1, 1], LineBox[{357, 358}]}, 
       {Hue[0.19406635637185005`, 1, 1], LineBox[{358, 359}]}, 
       {Hue[0.18823044513520215`, 1, 1], LineBox[{359, 360}]}, 
       {Hue[0.18783747724681005`, 1, 1], LineBox[{360, 361}]}, 
       {Hue[0.1874448146329372, 1, 1], LineBox[{361, 362}]}, 
       {Hue[0.18666041032284467`, 1, 1], LineBox[{362, 363}]}, 
       {Hue[0.1850953203601452, 1, 1], LineBox[{363, 364}]}, 
       {Hue[0.18198028486572737`, 1, 1], LineBox[{364, 365}]}, 
       {Hue[0.1758127883673826, 1, 1], LineBox[{365, 366}]}, 
       {Hue[0.1637414997039794, 1, 1], LineBox[{366, 367}]}, 
       {Hue[0.16337692046153715`, 1, 1], LineBox[{367, 368}]}, 
       {Hue[0.16301269104984262`, 1, 1], LineBox[{368, 369}]}, 
       {Hue[0.16228528369910886`, 1, 1], LineBox[{369, 370}]}, 
       {Hue[0.1608346881817319, 1, 1], LineBox[{370, 371}]}, 
       {Hue[0.1579504717808601, 1, 1], LineBox[{371, 372}]}, 
       {Hue[0.15225058847550843`, 1, 1], LineBox[{372, 373}]}, 
       {Hue[0.14112821611198173`, 1, 1], LineBox[{373, 374}]}, 
       {Hue[0.12136682583576486`, 1, 1], LineBox[{374, 375}]}, 
       {Hue[0.10132435565811605`, 1, 1], LineBox[{375, 376}]}, 
       {Hue[0.08378739001554548, 1, 1], LineBox[{376, 377}]}, 
       {Hue[0.06753590541558296, 1, 1], LineBox[{377, 378}]}, 
       {Hue[0.0672674238272465, 1, 1], LineBox[{378, 379}]}, 
       {Hue[0.06699916076804159, 1, 1], LineBox[{379, 380}]}, 
       {Hue[0.06646328691034005, 1, 1], LineBox[{380, 381}]}, 
       {Hue[0.06539412494024377, 1, 1], LineBox[{381, 382}]}, 
       {Hue[0.06326595746667119, 1, 1], LineBox[{382, 383}]}, 
       {Hue[0.05904875488299526, 1, 1], LineBox[{383, 384}]}, 
       {Hue[0.05878685308360073, 1, 1], LineBox[{384, 385}]}, 
       {Hue[0.05852514314748131, 1, 1], LineBox[{385, 386}]}, 
       {Hue[0.05800229553060262, 1, 1], LineBox[{386, 387}]}, 
       {Hue[0.05695886598100607, 1, 1], LineBox[{387, 388}]}, 
       {Hue[0.05488088299658431, 1, 1], LineBox[{388, 389}]}, 
       {Hue[0.05075893020904179, 1, 1], LineBox[{389, 390}]}, 
       {Hue[0.05051990962755181, 1, 1], LineBox[{390, 391}]}, 
       {Hue[0.0502810329455785, 1, 1], LineBox[{391, 392}]}, 
       {Hue[0.049803708587567615`, 1, 1], LineBox[{392, 393}]}, 
       {Hue[0.048850757056675125`, 1, 1], LineBox[{393, 394}]}, 
       {Hue[0.04695149217825662, 1, 1], LineBox[{394, 395}]}, 
       {Hue[0.043178313304186866`, 1, 1], LineBox[{395, 396}]}, 
       {Hue[0.04294356452866228, 1, 1], LineBox[{396, 397}]}, 
       {Hue[0.04270893820573536, 1, 1], LineBox[{397, 398}]}, 
       {Hue[0.04224005025099981, 1, 1], LineBox[{398, 399}]}, 
       {Hue[0.041303714452091696`, 1, 1], LineBox[{399, 400}]}, 
       {Hue[0.039436654259988296`, 1, 1], LineBox[{400, 401}]}, 
       {Hue[0.03572379038379056, 1, 1], LineBox[{401, 402}]}, 
       {Hue[0.03547320455100585, 1, 1], LineBox[{402, 403}]}, 
       {Hue[0.03522273762146896, 1, 1], LineBox[{403, 404}]}, 
       {Hue[0.03472215711393221, 1, 1], LineBox[{404, 405}]}, 
       {Hue[0.03372238601633399, 1, 1], LineBox[{405, 406}]}, 
       {Hue[0.03172821586249482, 1, 1], LineBox[{406, 407}]}, 
       {Hue[0.03147943191275808, 1, 1], LineBox[{407, 408}]}, 
       {Hue[0.03123075347217772, 1, 1], LineBox[{408, 409}]}, 
       {Hue[0.030733709781557574`, 1, 1], LineBox[{409, 410}]}, 
       {Hue[0.029740851822258815`, 1, 1], LineBox[{410, 411}]}, 
       {Hue[0.027759867141148174`, 1, 1], LineBox[{411, 412}]}, 
       {Hue[0.027512671401734252`, 1, 1], LineBox[{412, 413}]}, 
       {Hue[0.027265567861157906`, 1, 1], LineBox[{413, 414}]}, 
       {Hue[0.02677163406000022, 1, 1], LineBox[{414, 415}]}, 
       {Hue[0.02578483637921635, 1, 1], LineBox[{415, 416}]}, 
       {Hue[0.0238153353695229, 1, 1], LineBox[{416, 417}]}, 
       {Hue[0.023569515476446326`, 1, 1], LineBox[{417, 418}]}, 
       {Hue[0.023323774550983816`, 1, 1], LineBox[{418, 419}]}, 
       {Hue[0.022832526305280633`, 1, 1], LineBox[{419, 420}]}, 
       {Hue[0.021850941166985965`, 1, 1], LineBox[{420, 421}]}, 
       {Hue[0.019891231988468473`, 1, 1], LineBox[{421, 422}]}, 
       {Hue[0.019662777413417473`, 1, 1], LineBox[{422, 423}]}, 
       {Hue[0.0194343802662687, 1, 1], LineBox[{423, 424}]}, 
       {Hue[0.01897775558442008, 1, 1], LineBox[{424, 425}]}, 
       {Hue[0.018065165982028755`, 1, 1], LineBox[{425, 426}]}, 
       {Hue[0.0162424764644601, 1, 1], LineBox[{426, 427}]}, 
       {Hue[0.01601486069341581, 1, 1], LineBox[{427, 428}]}, 
       {Hue[0.0157872916858071, 1, 1], LineBox[{428, 429}]}, 
       {Hue[0.015332291300638774`, 1, 1], LineBox[{429, 430}]}, 
       {Hue[0.014422822437338929`, 1, 1], LineBox[{430, 431}]}, 
       {Hue[0.01260586355789195, 1, 1], LineBox[{431, 432}]}, 
       {Hue[0.012378916273309075`, 1, 1], LineBox[{432, 433}]}, 
       {Hue[0.012152005128142402`, 1, 1], LineBox[{433, 434}]}, 
       {Hue[0.011698288604802218`, 1, 1], LineBox[{434, 435}]}, 
       {Hue[0.010791260074104256`, 1, 1], LineBox[{435, 436}]}, 
       {Hue[0.01056458335196563, 1, 1], LineBox[{436, 437}]}, 
       {Hue[0.010337937469823545`, 1, 1], LineBox[{437, 438}]}, 
       {Hue[0.009884735577967426, 1, 1], LineBox[{438, 439}]}, 
       {Hue[0.008978672756857162, 1, 1], LineBox[{439, 440}]}, 
       {Hue[0.008752224226806537, 1, 1], LineBox[{440, 441}]}, 
       {Hue[0.008525801244265314, 1, 1], LineBox[{441, 442}]}, 
       {Hue[0.008073029277279837, 1, 1], LineBox[{442, 443}]}, 
       {Hue[0.007846678970838494, 1, 1], LineBox[{443, 444}]}, 
       {Hue[0.007620351567907857, 1, 1], LineBox[{444, 445}]}, 
       {Hue[0.0071677628294967485`, 1, 1], LineBox[{445, 446}]}, 
       {Hue[0.006941500172671094, 1, 1], LineBox[{446, 447}]}, 
       {Hue[0.006715257776661196, 1, 1], LineBox[{447, 448}]}, 
       {Hue[0.006262831125209884, 1, 1], LineBox[{448, 449}]}, 
       {Hue[0.005358191895222516, 1, 1], LineBox[{449, 450}]}, 
       {Hue[0.005136505252619112, 1, 1], LineBox[{450, 451}]}, 
       {Hue[0.004914833019706234, 1, 1], LineBox[{451, 452}]}, 
       {Hue[0.004471529295255007, 1, 1], LineBox[{452, 453}]}, 
       {Hue[0.004249896559984951, 1, 1], LineBox[{453, 454}]}, 
       {Hue[0.004028275746937781, 1, 1], LineBox[{454, 455}]}, 
       {Hue[0.003585067400516932, 1, 1], LineBox[{455, 456}]}, 
       {Hue[0.0033634786237412393`, 1, 1], LineBox[{456, 457}]}, 
       {Hue[0.0031418992823801332`, 1, 1], LineBox[{457, 458}]}, 
       {Hue[0.002920328754818563, 1, 1], LineBox[{458, 459}]}, 
       {Hue[0.0026987664194720344`, 1, 1], LineBox[{459, 460}]}, 
       {Hue[0.002477211654784387, 1, 1], LineBox[{460, 461}]}, 
       {Hue[0.0022556638392256965`, 1, 1], LineBox[{461, 462}]}, 
       {Hue[0.0020341223512901885`, 1, 1], LineBox[{462, 463}]}, 
       {Hue[0.0018125865694940044`, 1, 1], LineBox[{463, 464}]}, 
       {Hue[0.0015910558723730781`, 1, 1], LineBox[{464, 465}]}, 
       {Hue[0.0013695296384809846`, 1, 1], LineBox[{465, 466}]}, 
       {Hue[0.0011480072463867883`, 1, 1], LineBox[{466, 467}]}, 
       {Hue[0.0009264880746729118, 1, 1], LineBox[{467, 468}]}, 
       {Hue[0.0007049715019329034, 1, 1], LineBox[{468, 469}]}, 
       {Hue[0.0004834569067693328, 1, 1], LineBox[{469, 470}]}, 
       {Hue[0.00026194366779169054`, 1, 1], LineBox[{470, 471}]}, 
       {Hue[0.000040431163614147966`, 1, 1], LineBox[{471, 472}]}, 
       {Hue[0.9998189187728534, 1, 1], LineBox[{472, 473}]}, 
       {Hue[0.9995974058741266, 1, 1], LineBox[{473, 474}]}, 
       {Hue[0.9993758918460491, 1, 1], LineBox[{474, 475}]}, 
       {Hue[0.9991543760672325, 1, 1], LineBox[{475, 476}]}, 
       {Hue[0.9989328579162817, 1, 1], LineBox[{476, 477}]}, 
       {Hue[0.9987113367717942, 1, 1], LineBox[{477, 478}]}, 
       {Hue[0.9984898120123568, 1, 1], LineBox[{478, 479}]}, 
       {Hue[0.9982682830165432, 1, 1], LineBox[{479, 480}]}, 
       {Hue[0.9980467491629134, 1, 1], LineBox[{480, 481}]}, 
       {Hue[0.9978252098300097, 1, 1], LineBox[{481, 482}]}, 
       {Hue[0.9976036643963558, 1, 1], LineBox[{482, 483}]}, 
       {Hue[0.9973821122404543, 1, 1], LineBox[{483, 484}]}, 
       {Hue[0.9971605527407842, 1, 1], LineBox[{484, 485}]}, 
       {Hue[0.9969389852757992, 1, 1], LineBox[{485, 486}]}, 
       {Hue[0.9964958239635591, 1, 1], LineBox[{486, 487}]}, 
       {Hue[0.9962742288730648, 1, 1], LineBox[{487, 488}]}, 
       {Hue[0.996052623330773, 1, 1], LineBox[{488, 489}]}, 
       {Hue[0.9956093784039354, 1, 1], LineBox[{489, 490}]}, 
       {Hue[0.995387737775861, 1, 1], LineBox[{490, 491}]}, 
       {Hue[0.9951660842089277, 1, 1], LineBox[{491, 492}]}, 
       {Hue[0.9947227357709507, 1, 1], LineBox[{492, 493}]}, 
       {Hue[0.9945010396560208, 1, 1], LineBox[{493, 494}]}, 
       {Hue[0.9942793281144557, 1, 1], LineBox[{494, 495}]}, 
       {Hue[0.9938358562630776, 1, 1], LineBox[{495, 496}]}, 
       {Hue[0.9936140947089531, 1, 1], LineBox[{496, 497}]}, 
       {Hue[0.9933923152395664, 1, 1], LineBox[{497, 498}]}, 
       {Hue[0.9929487000657207, 1, 1], LineBox[{498, 499}]}, 
       {Hue[0.9927268631164576, 1, 1], LineBox[{499, 500}]}, 
       {Hue[0.99250500576232, 1, 1], LineBox[{500, 501}]}, 
       {Hue[0.9920612273490585, 1, 1], LineBox[{501, 502}]}, 
       {Hue[0.9911733982659021, 1, 1], LineBox[{502, 503}]}, 
       {Hue[0.9909325747967033, 1, 1], LineBox[{503, 504}]}, 
       {Hue[0.9906917213950743, 1, 1], LineBox[{504, 505}]}, 
       {Hue[0.9902099216129358, 1, 1], LineBox[{505, 506}]}, 
       {Hue[0.989245927854953, 1, 1], LineBox[{506, 507}]}, 
       {Hue[0.9890048426489481, 1, 1], LineBox[{507, 508}]}, 
       {Hue[0.9887637211439682, 1, 1], LineBox[{508, 509}]}, 
       {Hue[0.9882813660511997, 1, 1], LineBox[{509, 510}]}, 
       {Hue[0.9873161852248517, 1, 1], LineBox[{510, 511}]}, 
       {Hue[0.9870747873184581, 1, 1], LineBox[{511, 512}]}, 
       {Hue[0.9868333467373395, 1, 1], LineBox[{512, 513}]}, 
       {Hue[0.9863503343599933, 1, 1], LineBox[{513, 514}]}, 
       {Hue[0.9853837623986506, 1, 1], LineBox[{514, 515}]}, 
       {Hue[0.9834482507269322, 1, 1], LineBox[{515, 516}]}, 
       {Hue[0.9832060742017218, 1, 1], LineBox[{516, 517}]}, 
       {Hue[0.9829638422167695, 1, 1], LineBox[{517, 518}]}, 
       {Hue[0.9824792086644553, 1, 1], LineBox[{518, 519}]}, 
       {Hue[0.9815092407958094, 1, 1], LineBox[{519, 520}]}, 
       {Hue[0.9795663223392861, 1, 1], LineBox[{520, 521}]}, 
       {Hue[0.9793231622450247, 1, 1], LineBox[{521, 522}]}, 
       {Hue[0.9790799338521377, 1, 1], LineBox[{522, 523}]}, 
       {Hue[0.9785932689524265, 1, 1], LineBox[{523, 524}]}, 
       {Hue[0.9776190841566166, 1, 1], LineBox[{524, 525}]}, 
       {Hue[0.975667114035326, 1, 1], LineBox[{525, 526}]}, 
       {Hue[0.9754390633641128, 1, 1], LineBox[{526, 527}]}, 
       {Hue[0.9752109420085299, 1, 1], LineBox[{527, 528}]}, 
       {Hue[0.9747544846142472, 1, 1], LineBox[{528, 529}]}, 
       {Hue[0.9738406926710507, 1, 1], LineBox[{529, 530}]}, 
       {Hue[0.9720094526039389, 1, 1], LineBox[{530, 531}]}, 
       {Hue[0.9683311642017814, 1, 1], LineBox[{531, 532}]}, 
       {Hue[0.9681005244293676, 1, 1], LineBox[{532, 533}]}, 
       {Hue[0.9678697928221847, 1, 1], LineBox[{533, 534}]}, 
       {Hue[0.9674080514439187, 1, 1], LineBox[{534, 535}]}, 
       {Hue[0.966483437400964, 1, 1], LineBox[{535, 536}]}, 
       {Hue[0.9646295349167414, 1, 1], LineBox[{536, 537}]}, 
       {Hue[0.9609018350177954, 1, 1], LineBox[{537, 538}]}, 
       {Hue[0.9606483503513359, 1, 1], LineBox[{538, 539}]}, 
       {Hue[0.9603947326474463, 1, 1], LineBox[{539, 540}]}, 
       {Hue[0.9598870947025527, 1, 1], LineBox[{540, 541}]}, 
       {Hue[0.9588701846718123, 1, 1], LineBox[{541, 542}]}, 
       {Hue[0.9568296358286239, 1, 1], LineBox[{542, 543}]}, 
       {Hue[0.9527200809416663, 1, 1], LineBox[{543, 544}]}, 
       {Hue[0.9443747422042437, 1, 1], LineBox[{544, 545}]}, 
       {Hue[0.9441158128316757, 1, 1], LineBox[{545, 546}]}, 
       {Hue[0.943856701855986, 1, 1], LineBox[{546, 547}]}, 
       {Hue[0.9433379317912788, 1, 1], LineBox[{547, 548}]}, 
       {Hue[0.942298176074474, 1, 1], LineBox[{548, 549}]}, 
       {Hue[0.9402096171393628, 1, 1], LineBox[{549, 550}]}, 
       {Hue[0.9359948273114629, 1, 1], LineBox[{550, 551}]}, 
       {Hue[0.9274027124575679, 1, 1], LineBox[{551, 552}]}, 
       {Hue[0.9107129293179048, 1, 1], LineBox[{552, 553}]}, 
       {Hue[0.8914807352640294, 1, 1], LineBox[{553, 554}]}, 
       {Hue[0.8722988598223519, 1, 1], LineBox[{554, 555}]}, 
       {Hue[0.8522396379588035, 1, 1], LineBox[{555, 556}]}, 
       {Hue[0.8518885980552959, 1, 1], LineBox[{556, 557}]}, 
       {Hue[0.8515371912952256, 1, 1], LineBox[{557, 558}]}, 
       {Hue[0.8508332776347656, 1, 1], LineBox[{558, 559}]}, 
       {Hue[0.8494210532542403, 1, 1], LineBox[{559, 560}]}, 
       {Hue[0.8465790518299141, 1, 1], LineBox[{560, 561}]}, 
       {Hue[0.8408252475632318, 1, 1], LineBox[{561, 562}]}, 
       {Hue[0.8290438106889065, 1, 1], LineBox[{562, 563}]}, 
       {Hue[0.8286949133355983, 1, 1], LineBox[{563, 564}]}, 
       {Hue[0.8283457193462858, 1, 1], LineBox[{564, 565}]}, 
       {Hue[0.8276464439003608, 1, 1], LineBox[{565, 566}]}, 
       {Hue[0.8262443606965498, 1, 1], LineBox[{566, 567}]}, 
       {Hue[0.8234262126203606, 1, 1], LineBox[{567, 568}]}, 
       {Hue[0.8177352853514953, 1, 1], LineBox[{568, 569}]}, 
       {Hue[0.8173772394635447, 1, 1], LineBox[{569, 570}]}, 
       {Hue[0.8170189206330951, 1, 1], LineBox[{570, 571}]}, 
       {Hue[0.8163014676846723, 1, 1], LineBox[{571, 572}]}, 
       {Hue[0.8148633259038316, 1, 1], LineBox[{572, 573}]}, 
       {Hue[0.8119743082263935, 1, 1], LineBox[{573, 574}]}, 
       {Hue[0.8061471281010334, 1, 1], LineBox[{574, 575}]}, 
       {Hue[0.8057500829027806, 1, 1], LineBox[{575, 576}]}, 
       {Hue[0.8053527553799907, 1, 1], LineBox[{576, 577}]}, 
       {Hue[0.8045572592770968, 1, 1], LineBox[{577, 578}]}, 
       {Hue[0.8029629449050588, 1, 1], LineBox[{578, 579}]}, 
       {Hue[0.7997613736541621, 1, 1], LineBox[{579, 580}]}, 
       {Hue[0.799359994854902, 1, 1], LineBox[{580, 581}]}, 
       {Hue[0.7989583587267372, 1, 1], LineBox[{581, 582}]}, 
       {Hue[0.7981543211276124, 1, 1], LineBox[{582, 583}]}, 
       {Hue[0.7965432316826655, 1, 1], LineBox[{583, 584}]}, 
       {Hue[0.7933093819342499, 1, 1], LineBox[{584, 585}]}, 
       {Hue[0.7929040909701478, 1, 1], LineBox[{585, 586}]}, 
       {Hue[0.7924985705100261, 1, 1], LineBox[{586, 587}]}, 
       {Hue[0.7916868484298172, 1, 1], LineBox[{587, 588}]}, 
       {Hue[0.7900607314967867, 1, 1], LineBox[{588, 589}]}, 
       {Hue[0.7867982298702154, 1, 1], LineBox[{589, 590}]}, 
       {Hue[0.7802356708666068, 1, 1], LineBox[{590, 591}]}, 
       {Hue[0.7802233700537657, 1, 1], LineBox[{591, 592}]}, 
       {Hue[0.7802110690906503, 1, 1], LineBox[{592, 593}]}, 
       {Hue[0.7801864667138221, 1, 1], LineBox[{593, 594}]}, 
       {Hue[0.7801372601593551, 1, 1], LineBox[{594, 595}]}, 
       {Hue[0.7800388398598123, 1, 1], LineBox[{595, 596}]}, 
       {Hue[0.7798419705994404, 1, 1], LineBox[{596, 597}]}, 
       {Hue[0.7798296650058097, 1, 1], LineBox[{597, 598}]}, 
       {Hue[0.7798173592637114, 1, 1], LineBox[{598, 599}]}, 
       {Hue[0.7797927473343383, 1, 1], LineBox[{599, 600}]}, 
       {Hue[0.7797435216964705, 1, 1], LineBox[{600, 601}]}, 
       {Hue[0.7796450633169278, 1, 1], LineBox[{601, 602}]}, 
       {Hue[0.7796327553546072, 1, 1], LineBox[{602, 603}]}, 
       {Hue[0.7796204472447248, 1, 1], LineBox[{603, 604}]}, 
       {Hue[0.7795958305825013, 1, 1], LineBox[{604, 605}]}, 
       {Hue[0.7795465954898065, 1, 1], LineBox[{605, 606}]}, 
       {Hue[0.7795342863485786, 1, 1], LineBox[{606, 607}]}, 
       {Hue[0.7795219770602422, 1, 1], LineBox[{607, 608}]}, 
       {Hue[0.7794973580424717, 1, 1], LineBox[{608, 609}]}, 
       {Hue[0.7794850483131508, 1, 1], LineBox[{609, 610}]}, 
       {Hue[0.7794727384369486, 1, 1], LineBox[{610, 611}]}, 
       {Hue[0.7794604284139217, 1, 1], LineBox[{611, 612}]}}}},
    VertexColors->{
      Hue[0.7205518817558729, 1, 1], 
      Hue[0.7201728632810981, 1, 1], 
      Hue[0.7197939855871971, 1, 1], 
      Hue[0.7190366591317676, 1, 1], 
      Hue[0.7175237677526916, 1, 1], 
      Hue[0.7145053924996501, 1, 1], 
      Hue[0.7141288210543033, 1, 1], 
      Hue[0.7137524160490083, 1, 1], 
      Hue[0.7130001115521821, 1, 1], 
      Hue[0.7114975675954697, 1, 1], 
      Hue[0.7085010779576801, 1, 1], 
      Hue[0.7081273517673412, 1, 1], 
      Hue[0.707753815997507, 1, 1], 
      Hue[0.7070073214666093, 1, 1], 
      Hue[0.7055166802582684, 1, 1], 
      Hue[0.7025451015155051, 1, 1], 
      Hue[0.6966431533753163, 1, 1], 
      Hue[0.6962453407646028, 1, 1], 
      Hue[0.6958478017085719, 1, 1], 
      Hue[0.6950535502982287, 1, 1], 
      Hue[0.6934683959066561, 1, 1], 
      Hue[0.69031180412586, 1, 1], 
      Hue[0.6840559180388815, 1, 1], 
      Hue[0.6717902321513325, 1, 1], 
      Hue[0.6714375103634247, 1, 1], 
      Hue[0.671085090804056, 1, 1], 
      Hue[0.6703811608448678, 1, 1], 
      Hue[0.668976954394541, 1, 1], 
      Hue[0.666183282738352, 1, 1], 
      Hue[0.6606558087846726, 1, 1], 
      Hue[0.6498458025017817, 1, 1], 
      Hue[0.6296151700839989, 1, 1], 
      Hue[0.6090856243791776, 1, 1], 
      Hue[0.5911506349521957, 1, 1], 
      Hue[0.5728738758396617, 1, 1], 
      Hue[0.5726018283453903, 1, 1], 
      Hue[0.5723300141524164, 1, 1], 
      Hue[0.5717870824013804, 1, 1], 
      Hue[0.570703982533093, 1, 1], 
      Hue[0.5685486537726839, 1, 1], 
      Hue[0.5642800076517256, 1, 1], 
      Hue[0.5558989435773015, 1, 1], 
      Hue[0.5556575555942079, 1, 1], 
      Hue[0.5554163247938143, 1, 1], 
      Hue[0.5549343320656529, 1, 1], 
      Hue[0.5539722033773958, 1, 1], 
      Hue[0.5520552233913043, 1, 1], 
      Hue[0.5482491773582174, 1, 1], 
      Hue[0.5480124880738455, 1, 1], 
      Hue[0.5477759346340769, 1, 1], 
      Hue[0.5473032326321359, 1, 1], 
      Hue[0.5463594295560026, 1, 1], 
      Hue[0.5444780786214425, 1, 1], 
      Hue[0.5407392124960726, 1, 1], 
      Hue[0.5404869192549508, 1, 1], 
      Hue[0.5402347608129789, 1, 1], 
      Hue[0.5397308449751576, 1, 1], 
      Hue[0.5387245940428427, 1, 1], 
      Hue[0.5367182278963121, 1, 1], 
      Hue[0.5327285448595175, 1, 1], 
      Hue[0.5324801512102756, 1, 1], 
      Hue[0.5322318657097549, 1, 1], 
      Hue[0.5317356158456428, 1, 1], 
      Hue[0.5307443775187145, 1, 1], 
      Hue[0.5287667615698055, 1, 1], 
      Hue[0.5248294975419259, 1, 1], 
      Hue[0.5246004552449229, 1, 1], 
      Hue[0.5243714843111781, 1, 1], 
      Hue[0.5239137538730607, 1, 1], 
      Hue[0.5229991201035351, 1, 1], 
      Hue[0.5211730122817956, 1, 1], 
      Hue[0.5209450320969299, 1, 1], 
      Hue[0.5207171126580099, 1, 1], 
      Hue[0.5202614533707344, 1, 1], 
      Hue[0.5193508346382325, 1, 1], 
      Hue[0.5175322485357188, 1, 1], 
      Hue[0.5173051609668913, 1, 1], 
      Hue[0.5170781235763294, 1, 1], 
      Hue[0.516624196694218, 1, 1], 
      Hue[0.5157169160846813, 1, 1], 
      Hue[0.5154902114994349, 1, 1], 
      Hue[0.5152635518250663, 1, 1], 
      Hue[0.5148103645782285, 1, 1], 
      Hue[0.5139045000852724, 1, 1], 
      Hue[0.513678136379099, 1, 1], 
      Hue[0.5134518123261133, 1, 1], 
      Hue[0.5129992805535211, 1, 1], 
      Hue[0.5120946639659323, 1, 1], 
      Hue[0.5102870717187569, 1, 1], 
      Hue[0.5100657056310821, 1, 1], 
      Hue[0.5098443675824635, 1, 1], 
      Hue[0.5094017731352407, 1, 1], 
      Hue[0.50851689357583, 1, 1], 
      Hue[0.5082957345349084, 1, 1], 
      Hue[0.508074598601023, 1, 1], 
      Hue[0.5076323935899204, 1, 1], 
      Hue[0.5067482337474685, 1, 1], 
      Hue[0.5065272423153528, 1, 1], 
      Hue[0.5063062690632634, 1, 1], 
      Hue[0.5058643746369205, 1, 1], 
      Hue[0.5056434522316935, 1, 1], 
      Hue[0.5054225455445416, 1, 1], 
      Hue[0.504980776863122, 1, 1], 
      Hue[0.504759913638311, 1, 1], 
      Hue[0.5045390636704838, 1, 1], 
      Hue[0.5040974010452097, 1, 1], 
      Hue[0.5038765871575825, 1, 1], 
      Hue[0.5036557840665759, 1, 1], 
      Hue[0.5032142078144854, 1, 1], 
      Hue[0.5029934334235187, 1, 1], 
      Hue[0.5027726673694013, 1, 1], 
      Hue[0.5025519090372679, 1, 1], 
      Hue[0.5023311578122791, 1, 1], 
      Hue[0.5021104130796202, 1, 1], 
      Hue[0.5018896742244989, 1, 1], 
      Hue[0.5016689406321433, 1, 1], 
      Hue[0.501448211687799, 1, 1], 
      Hue[0.5012274867767283, 1, 1], 
      Hue[0.5010067652842074, 1, 1], 
      Hue[0.5007860465955236, 1, 1], 
      Hue[0.5005653300959744, 1, 1], 
      Hue[0.5003446151708648, 1, 1], 
      Hue[0.5001239012055051, 1, 1], 
      Hue[0.499903187585209, 1, 1], 
      Hue[0.4996824736952913, 1, 1], 
      Hue[0.4994617589210659, 1, 1], 
      Hue[0.4992410426478438, 1, 1], 
      Hue[0.4990203242609306, 1, 1], 
      Hue[0.4987996031456248, 1, 1], 
      Hue[0.49857887868721534`, 1, 1], 
      Hue[0.4983581502709798, 1, 1], 
      Hue[0.4981374172821819, 1, 1], 
      Hue[0.4979166791060699, 1, 1], 
      Hue[0.49769593512787397`, 1, 1], 
      Hue[0.4974751847328044, 1, 1], 
      Hue[0.49725442730604935`, 1, 1], 
      Hue[0.4970336622327728, 1, 1], 
      Hue[0.49681288889811237`, 1, 1], 
      Hue[0.49659210668717735`, 1, 1], 
      Hue[0.4961505131767657, 1, 1], 
      Hue[0.49591092983354623`, 1, 1], 
      Hue[0.4956713330834243, 1, 1], 
      Hue[0.4951920962198805, 1, 1], 
      Hue[0.49495245453501585`, 1, 1], 
      Hue[0.494712796300357, 1, 1], 
      Hue[0.49423342703805917`, 1, 1], 
      Hue[0.4939937144384435, 1, 1], 
      Hue[0.49375398214507416`, 1, 1], 
      Hue[0.4932744553322717, 1, 1], 
      Hue[0.4923151307840086, 1, 1], 
      Hue[0.4920752386164939, 1, 1], 
      Hue[0.4918353204633067, 1, 1], 
      Hue[0.491355403052108, 1, 1], 
      Hue[0.4903952217696123, 1, 1], 
      Hue[0.49015509967662196`, 1, 1], 
      Hue[0.4899149452993977, 1, 1], 
      Hue[0.4894345365406593, 1, 1], 
      Hue[0.488473296937411, 1, 1], 
      Hue[0.4882328945028128, 1, 1], 
      Hue[0.4879924534772371, 1, 1], 
      Hue[0.4875114524970251, 1, 1], 
      Hue[0.48654895271867454`, 1, 1], 
      Hue[0.4863082194556099, 1, 1], 
      Hue[0.48606744128514545`, 1, 1], 
      Hue[0.4855857470606176, 1, 1], 
      Hue[0.48462178493732505`, 1, 1], 
      Hue[0.4843806702764197, 1, 1], 
      Hue[0.4841395043805711, 1, 1], 
      Hue[0.4836570157166681, 1, 1], 
      Hue[0.4826913887171722, 1, 1], 
      Hue[0.480757358393055, 1, 1], 
      Hue[0.4805315263020154, 1, 1], 
      Hue[0.4803056385978369, 1, 1], 
      Hue[0.47985369376570597`, 1, 1], 
      Hue[0.47894910830569265`, 1, 1], 
      Hue[0.47713700924086944`, 1, 1], 
      Hue[0.4769102097107648, 1, 1], 
      Hue[0.47668334420836855`, 1, 1], 
      Hue[0.47622941269007835`, 1, 1], 
      Hue[0.47532072941166953`, 1, 1], 
      Hue[0.47349993620919084`, 1, 1], 
      Hue[0.47327200311344897`, 1, 1], 
      Hue[0.47304399363458505`, 1, 1], 
      Hue[0.4725877429172142, 1, 1], 
      Hue[0.47167429615942985`, 1, 1], 
      Hue[0.4698434748853157, 1, 1], 
      Hue[0.46616494645249384`, 1, 1], 
      Hue[0.4659148691575225, 1, 1], 
      Hue[0.46566467738003314`, 1, 1], 
      Hue[0.4651639470144475, 1, 1], 
      Hue[0.4641610754806451, 1, 1], 
      Hue[0.46214950043805014`, 1, 1], 
      Hue[0.4581015078231806, 1, 1], 
      Hue[0.4578473484445837, 1, 1], 
      Hue[0.4575930475155824, 1, 1], 
      Hue[0.45708401760083006`, 1, 1], 
      Hue[0.4560642216878684, 1, 1], 
      Hue[0.4540174944156169, 1, 1], 
      Hue[0.4498939651897446, 1, 1], 
      Hue[0.4496395920714616, 1, 1], 
      Hue[0.449385056150846, 1, 1], 
      Hue[0.4488754926464263, 1, 1], 
      Hue[0.4478543761796728, 1, 1], 
      Hue[0.4458040028063383, 1, 1], 
      Hue[0.44166923112517276`, 1, 1], 
      Hue[0.4332518596871854, 1, 1], 
      Hue[0.4169447860407514, 1, 1], 
      Hue[0.39819770478061217`, 1, 1], 
      Hue[0.3795344901876919, 1, 1], 
      Hue[0.3578952936955762, 1, 1], 
      Hue[0.35755137633603035`, 1, 1], 
      Hue[0.35720709405282786`, 1, 1], 
      Hue[0.3565174345978122, 1, 1], 
      Hue[0.3551337357577113, 1, 1], 
      Hue[0.3523488221943377, 1, 1], 
      Hue[0.3467090755102115, 1, 1], 
      Hue[0.33515298547996475`, 1, 1], 
      Hue[0.334810643178493, 1, 1], 
      Hue[0.3344679970554599, 1, 1], 
      Hue[0.33378179518226486`, 1, 1], 
      Hue[0.3324057662380327, 1, 1], 
      Hue[0.32963932081463193`, 1, 1], 
      Hue[0.3240498957473761, 1, 1], 
      Hue[0.32369809827084883`, 1, 1], 
      Hue[0.32334601565736026`, 1, 1], 
      Hue[0.3226409979132937, 1, 1], 
      Hue[0.32122757308827055`, 1, 1], 
      Hue[0.3183873390867039, 1, 1], 
      Hue[0.3126548323856401, 1, 1], 
      Hue[0.31226394002872365`, 1, 1], 
      Hue[0.3118727446949302, 1, 1], 
      Hue[0.31108945018381584`, 1, 1], 
      Hue[0.30951928205543044`, 1, 1], 
      Hue[0.3063649293548802, 1, 1], 
      Hue[0.30596934819271, 1, 1], 
      Hue[0.3055734857590331, 1, 1], 
      Hue[0.30478092290960934`, 1, 1], 
      Hue[0.30319248675089605`, 1, 1], 
      Hue[0.3000027138831849, 1, 1], 
      Hue[0.29960281338186984`, 1, 1], 
      Hue[0.29920265625388204`, 1, 1], 
      Hue[0.29840157866966915`, 1, 1], 
      Hue[0.2967964166611863, 1, 1], 
      Hue[0.2935744461038008, 1, 1], 
      Hue[0.2870871058179891, 1, 1], 
      Hue[0.2867068783556848, 1, 1], 
      Hue[0.28632647720527604`, 1, 1], 
      Hue[0.28556516022402634`, 1, 1], 
      Hue[0.284040512612144, 1, 1], 
      Hue[0.28098352900329876`, 1, 1], 
      Hue[0.2806007176236879, 1, 1], 
      Hue[0.2802177588652316, 1, 1], 
      Hue[0.27945140601201585`, 1, 1], 
      Hue[0.2779170068960574, 1, 1], 
      Hue[0.2748418233531421, 1, 1], 
      Hue[0.27445686099902716`, 1, 1], 
      Hue[0.2740717791307105, 1, 1], 
      Hue[0.2733012640056608, 1, 1], 
      Hue[0.2717588785561085, 1, 1], 
      Hue[0.27137301040675016`, 1, 1], 
      Hue[0.2709870371770742, 1, 1], 
      Hue[0.2702147827817822, 1, 1], 
      Hue[0.26866909362519453`, 1, 1], 
      Hue[0.26828243617916075`, 1, 1], 
      Hue[0.2678956883732717, 1, 1], 
      Hue[0.2671219291190381, 1, 1], 
      Hue[0.26557340830942106`, 1, 1], 
      Hue[0.2651860802254811, 1, 1], 
      Hue[0.264798676750572, 1, 1], 
      Hue[0.2640236511774876, 1, 1], 
      Hue[0.26247277855953854`, 1, 1], 
      Hue[0.26209251788900434`, 1, 1], 
      Hue[0.26171219931070255`, 1, 1], 
      Hue[0.2609513956309302, 1, 1], 
      Hue[0.2605709141353225, 1, 1], 
      Hue[0.26019038194352423`, 1, 1], 
      Hue[0.2594291727069773, 1, 1], 
      Hue[0.2590484992850898, 1, 1], 
      Hue[0.25866778241258415`, 1, 1], 
      Hue[0.2579062255820407, 1, 1], 
      Hue[0.2575253892614616, 1, 1], 
      Hue[0.2571445167650305, 1, 1], 
      Hue[0.2563826705368009, 1, 1], 
      Hue[0.256001700454635, 1, 1], 
      Hue[0.25562070149572913`, 1, 1], 
      Hue[0.2552396754880936, 1, 1], 
      Hue[0.25485862426087846`, 1, 1], 
      Hue[0.2544775496442966, 1, 1], 
      Hue[0.254096453469547, 1, 1], 
      Hue[0.2537153375687379, 1, 1], 
      Hue[0.2533342037748092, 1, 1], 
      Hue[0.2529530539214558, 1, 1], 
      Hue[0.2525718898430503, 1, 1], 
      Hue[0.2521907133745654, 1, 1], 
      Hue[0.2518095263514968, 1, 1], 
      Hue[0.2514283306097857, 1, 1], 
      Hue[0.2510471279857414, 1, 1], 
      Hue[0.2506659203159637, 1, 1], 
      Hue[0.25028470943726544`, 1, 1], 
      Hue[0.24990349718659496`, 1, 1], 
      Hue[0.24952228540095855`, 1, 1], 
      Hue[0.24914107591734283`, 1, 1], 
      Hue[0.24875987057263726`, 1, 1], 
      Hue[0.24837867120355647`, 1, 1], 
      Hue[0.2479974796465629, 1, 1], 
      Hue[0.2476162977377892, 1, 1], 
      Hue[0.2472351273129606, 1, 1], 
      Hue[0.24685397020731778`, 1, 1], 
      Hue[0.24647282825553923`, 1, 1], 
      Hue[0.24609170329166405`, 1, 1], 
      Hue[0.24571059714901458`, 1, 1], 
      Hue[0.24532951166011932`, 1, 1], 
      Hue[0.24494844865663576`, 1, 1], 
      Hue[0.24418639742771617`, 1, 1], 
      Hue[0.24380541286054708`, 1, 1], 
      Hue[0.24342445809517, 1, 1], 
      Hue[0.24266264527305542`, 1, 1], 
      Hue[0.24228179086454416`, 1, 1], 
      Hue[0.24190097355412485`, 1, 1], 
      Hue[0.2411394575073841, 1, 1], 
      Hue[0.2407587624068073, 1, 1], 
      Hue[0.24037811167566045`, 1, 1], 
      Hue[0.23961695057318147`, 1, 1], 
      Hue[0.2380952404554197, 1, 1], 
      Hue[0.23768268530800246`, 1, 1], 
      Hue[0.23727019956300532`, 1, 1], 
      Hue[0.23644544544623258`, 1, 1], 
      Hue[0.23479687066374988`, 1, 1], 
      Hue[0.23438493471701696`, 1, 1], 
      Hue[0.23397308640684428`, 1, 1], 
      Hue[0.23314966173624715`, 1, 1], 
      Hue[0.23150396322258707`, 1, 1], 
      Hue[0.23109279142615208`, 1, 1], 
      Hue[0.23068172522628413`, 1, 1], 
      Hue[0.22985991850312365`, 1, 1], 
      Hue[0.2282176696727462, 1, 1], 
      Hue[0.22493912100615956`, 1, 1], 
      Hue[0.22452990240719808`, 1, 1], 
      Hue[0.22412082428986946`, 1, 1], 
      Hue[0.22330309800330872`, 1, 1], 
      Hue[0.2216694243601005, 1, 1], 
      Hue[0.21840965988103805`, 1, 1], 
      Hue[0.21800294032724274`, 1, 1], 
      Hue[0.21759639443153517`, 1, 1], 
      Hue[0.21678383164015788`, 1, 1], 
      Hue[0.21516087777741738`, 1, 1], 
      Hue[0.2119240955782303, 1, 1], 
      Hue[0.21154735354250084`, 1, 1], 
      Hue[0.21117078972624026`, 1, 1], 
      Hue[0.21041820282444174`, 1, 1], 
      Hue[0.20891523405310197`, 1, 1], 
      Hue[0.2059184474441574, 1, 1], 
      Hue[0.19996402390747361`, 1, 1], 
      Hue[0.19959370230262743`, 1, 1], 
      Hue[0.1992236038568319, 1, 1], 
      Hue[0.19848408153459976`, 1, 1], 
      Hue[0.19700777035336184`, 1, 1], 
      Hue[0.19406635637185005`, 1, 1], 
      Hue[0.18823044513520215`, 1, 1], 
      Hue[0.18783747724681005`, 1, 1], 
      Hue[0.1874448146329372, 1, 1], 
      Hue[0.18666041032284467`, 1, 1], 
      Hue[0.1850953203601452, 1, 1], 
      Hue[0.18198028486572737`, 1, 1], 
      Hue[0.1758127883673826, 1, 1], 
      Hue[0.1637414997039794, 1, 1], 
      Hue[0.16337692046153715`, 1, 1], 
      Hue[0.16301269104984262`, 1, 1], 
      Hue[0.16228528369910886`, 1, 1], 
      Hue[0.1608346881817319, 1, 1], 
      Hue[0.1579504717808601, 1, 1], 
      Hue[0.15225058847550843`, 1, 1], 
      Hue[0.14112821611198173`, 1, 1], 
      Hue[0.12136682583576486`, 1, 1], 
      Hue[0.10132435565811605`, 1, 1], 
      Hue[0.08378739001554548, 1, 1], 
      Hue[0.06753590541558296, 1, 1], 
      Hue[0.0672674238272465, 1, 1], 
      Hue[0.06699916076804159, 1, 1], 
      Hue[0.06646328691034005, 1, 1], 
      Hue[0.06539412494024377, 1, 1], 
      Hue[0.06326595746667119, 1, 1], 
      Hue[0.05904875488299526, 1, 1], 
      Hue[0.05878685308360073, 1, 1], 
      Hue[0.05852514314748131, 1, 1], 
      Hue[0.05800229553060262, 1, 1], 
      Hue[0.05695886598100607, 1, 1], 
      Hue[0.05488088299658431, 1, 1], 
      Hue[0.05075893020904179, 1, 1], 
      Hue[0.05051990962755181, 1, 1], 
      Hue[0.0502810329455785, 1, 1], 
      Hue[0.049803708587567615`, 1, 1], 
      Hue[0.048850757056675125`, 1, 1], 
      Hue[0.04695149217825662, 1, 1], 
      Hue[0.043178313304186866`, 1, 1], 
      Hue[0.04294356452866228, 1, 1], 
      Hue[0.04270893820573536, 1, 1], 
      Hue[0.04224005025099981, 1, 1], 
      Hue[0.041303714452091696`, 1, 1], 
      Hue[0.039436654259988296`, 1, 1], 
      Hue[0.03572379038379056, 1, 1], 
      Hue[0.03547320455100585, 1, 1], 
      Hue[0.03522273762146896, 1, 1], 
      Hue[0.03472215711393221, 1, 1], 
      Hue[0.03372238601633399, 1, 1], 
      Hue[0.03172821586249482, 1, 1], 
      Hue[0.03147943191275808, 1, 1], 
      Hue[0.03123075347217772, 1, 1], 
      Hue[0.030733709781557574`, 1, 1], 
      Hue[0.029740851822258815`, 1, 1], 
      Hue[0.027759867141148174`, 1, 1], 
      Hue[0.027512671401734252`, 1, 1], 
      Hue[0.027265567861157906`, 1, 1], 
      Hue[0.02677163406000022, 1, 1], 
      Hue[0.02578483637921635, 1, 1], 
      Hue[0.0238153353695229, 1, 1], 
      Hue[0.023569515476446326`, 1, 1], 
      Hue[0.023323774550983816`, 1, 1], 
      Hue[0.022832526305280633`, 1, 1], 
      Hue[0.021850941166985965`, 1, 1], 
      Hue[0.019891231988468473`, 1, 1], 
      Hue[0.019662777413417473`, 1, 1], 
      Hue[0.0194343802662687, 1, 1], 
      Hue[0.01897775558442008, 1, 1], 
      Hue[0.018065165982028755`, 1, 1], 
      Hue[0.0162424764644601, 1, 1], 
      Hue[0.01601486069341581, 1, 1], 
      Hue[0.0157872916858071, 1, 1], 
      Hue[0.015332291300638774`, 1, 1], 
      Hue[0.014422822437338929`, 1, 1], 
      Hue[0.01260586355789195, 1, 1], 
      Hue[0.012378916273309075`, 1, 1], 
      Hue[0.012152005128142402`, 1, 1], 
      Hue[0.011698288604802218`, 1, 1], 
      Hue[0.010791260074104256`, 1, 1], 
      Hue[0.01056458335196563, 1, 1], 
      Hue[0.010337937469823545`, 1, 1], 
      Hue[0.009884735577967426, 1, 1], 
      Hue[0.008978672756857162, 1, 1], 
      Hue[0.008752224226806537, 1, 1], 
      Hue[0.008525801244265314, 1, 1], 
      Hue[0.008073029277279837, 1, 1], 
      Hue[0.007846678970838494, 1, 1], 
      Hue[0.007620351567907857, 1, 1], 
      Hue[0.0071677628294967485`, 1, 1], 
      Hue[0.006941500172671094, 1, 1], 
      Hue[0.006715257776661196, 1, 1], 
      Hue[0.006262831125209884, 1, 1], 
      Hue[0.005358191895222516, 1, 1], 
      Hue[0.005136505252619112, 1, 1], 
      Hue[0.004914833019706234, 1, 1], 
      Hue[0.004471529295255007, 1, 1], 
      Hue[0.004249896559984951, 1, 1], 
      Hue[0.004028275746937781, 1, 1], 
      Hue[0.003585067400516932, 1, 1], 
      Hue[0.0033634786237412393`, 1, 1], 
      Hue[0.0031418992823801332`, 1, 1], 
      Hue[0.002920328754818563, 1, 1], 
      Hue[0.0026987664194720344`, 1, 1], 
      Hue[0.002477211654784387, 1, 1], 
      Hue[0.0022556638392256965`, 1, 1], 
      Hue[0.0020341223512901885`, 1, 1], 
      Hue[0.0018125865694940044`, 1, 1], 
      Hue[0.0015910558723730781`, 1, 1], 
      Hue[0.0013695296384809846`, 1, 1], 
      Hue[0.0011480072463867883`, 1, 1], 
      Hue[0.0009264880746729118, 1, 1], 
      Hue[0.0007049715019329034, 1, 1], 
      Hue[0.0004834569067693328, 1, 1], 
      Hue[0.00026194366779169054`, 1, 1], 
      Hue[0.000040431163614147966`, 1, 1], 
      Hue[0.9998189187728534, 1, 1], 
      Hue[0.9995974058741266, 1, 1], 
      Hue[0.9993758918460491, 1, 1], 
      Hue[0.9991543760672325, 1, 1], 
      Hue[0.9989328579162817, 1, 1], 
      Hue[0.9987113367717942, 1, 1], 
      Hue[0.9984898120123568, 1, 1], 
      Hue[0.9982682830165432, 1, 1], 
      Hue[0.9980467491629134, 1, 1], 
      Hue[0.9978252098300097, 1, 1], 
      Hue[0.9976036643963558, 1, 1], 
      Hue[0.9973821122404543, 1, 1], 
      Hue[0.9971605527407842, 1, 1], 
      Hue[0.9969389852757992, 1, 1], 
      Hue[0.9964958239635591, 1, 1], 
      Hue[0.9962742288730648, 1, 1], 
      Hue[0.996052623330773, 1, 1], 
      Hue[0.9956093784039354, 1, 1], 
      Hue[0.995387737775861, 1, 1], 
      Hue[0.9951660842089277, 1, 1], 
      Hue[0.9947227357709507, 1, 1], 
      Hue[0.9945010396560208, 1, 1], 
      Hue[0.9942793281144557, 1, 1], 
      Hue[0.9938358562630776, 1, 1], 
      Hue[0.9936140947089531, 1, 1], 
      Hue[0.9933923152395664, 1, 1], 
      Hue[0.9929487000657207, 1, 1], 
      Hue[0.9927268631164576, 1, 1], 
      Hue[0.99250500576232, 1, 1], 
      Hue[0.9920612273490585, 1, 1], 
      Hue[0.9911733982659021, 1, 1], 
      Hue[0.9909325747967033, 1, 1], 
      Hue[0.9906917213950743, 1, 1], 
      Hue[0.9902099216129358, 1, 1], 
      Hue[0.989245927854953, 1, 1], 
      Hue[0.9890048426489481, 1, 1], 
      Hue[0.9887637211439682, 1, 1], 
      Hue[0.9882813660511997, 1, 1], 
      Hue[0.9873161852248517, 1, 1], 
      Hue[0.9870747873184581, 1, 1], 
      Hue[0.9868333467373395, 1, 1], 
      Hue[0.9863503343599933, 1, 1], 
      Hue[0.9853837623986506, 1, 1], 
      Hue[0.9834482507269322, 1, 1], 
      Hue[0.9832060742017218, 1, 1], 
      Hue[0.9829638422167695, 1, 1], 
      Hue[0.9824792086644553, 1, 1], 
      Hue[0.9815092407958094, 1, 1], 
      Hue[0.9795663223392861, 1, 1], 
      Hue[0.9793231622450247, 1, 1], 
      Hue[0.9790799338521377, 1, 1], 
      Hue[0.9785932689524265, 1, 1], 
      Hue[0.9776190841566166, 1, 1], 
      Hue[0.975667114035326, 1, 1], 
      Hue[0.9754390633641128, 1, 1], 
      Hue[0.9752109420085299, 1, 1], 
      Hue[0.9747544846142472, 1, 1], 
      Hue[0.9738406926710507, 1, 1], 
      Hue[0.9720094526039389, 1, 1], 
      Hue[0.9683311642017814, 1, 1], 
      Hue[0.9681005244293676, 1, 1], 
      Hue[0.9678697928221847, 1, 1], 
      Hue[0.9674080514439187, 1, 1], 
      Hue[0.966483437400964, 1, 1], 
      Hue[0.9646295349167414, 1, 1], 
      Hue[0.9609018350177954, 1, 1], 
      Hue[0.9606483503513359, 1, 1], 
      Hue[0.9603947326474463, 1, 1], 
      Hue[0.9598870947025527, 1, 1], 
      Hue[0.9588701846718123, 1, 1], 
      Hue[0.9568296358286239, 1, 1], 
      Hue[0.9527200809416663, 1, 1], 
      Hue[0.9443747422042437, 1, 1], 
      Hue[0.9441158128316757, 1, 1], 
      Hue[0.943856701855986, 1, 1], 
      Hue[0.9433379317912788, 1, 1], 
      Hue[0.942298176074474, 1, 1], 
      Hue[0.9402096171393628, 1, 1], 
      Hue[0.9359948273114629, 1, 1], 
      Hue[0.9274027124575679, 1, 1], 
      Hue[0.9107129293179048, 1, 1], 
      Hue[0.8914807352640294, 1, 1], 
      Hue[0.8722988598223519, 1, 1], 
      Hue[0.8522396379588035, 1, 1], 
      Hue[0.8518885980552959, 1, 1], 
      Hue[0.8515371912952256, 1, 1], 
      Hue[0.8508332776347656, 1, 1], 
      Hue[0.8494210532542403, 1, 1], 
      Hue[0.8465790518299141, 1, 1], 
      Hue[0.8408252475632318, 1, 1], 
      Hue[0.8290438106889065, 1, 1], 
      Hue[0.8286949133355983, 1, 1], 
      Hue[0.8283457193462858, 1, 1], 
      Hue[0.8276464439003608, 1, 1], 
      Hue[0.8262443606965498, 1, 1], 
      Hue[0.8234262126203606, 1, 1], 
      Hue[0.8177352853514953, 1, 1], 
      Hue[0.8173772394635447, 1, 1], 
      Hue[0.8170189206330951, 1, 1], 
      Hue[0.8163014676846723, 1, 1], 
      Hue[0.8148633259038316, 1, 1], 
      Hue[0.8119743082263935, 1, 1], 
      Hue[0.8061471281010334, 1, 1], 
      Hue[0.8057500829027806, 1, 1], 
      Hue[0.8053527553799907, 1, 1], 
      Hue[0.8045572592770968, 1, 1], 
      Hue[0.8029629449050588, 1, 1], 
      Hue[0.7997613736541621, 1, 1], 
      Hue[0.799359994854902, 1, 1], 
      Hue[0.7989583587267372, 1, 1], 
      Hue[0.7981543211276124, 1, 1], 
      Hue[0.7965432316826655, 1, 1], 
      Hue[0.7933093819342499, 1, 1], 
      Hue[0.7929040909701478, 1, 1], 
      Hue[0.7924985705100261, 1, 1], 
      Hue[0.7916868484298172, 1, 1], 
      Hue[0.7900607314967867, 1, 1], 
      Hue[0.7867982298702154, 1, 1], 
      Hue[0.7802356708666068, 1, 1], 
      Hue[0.7802233700537657, 1, 1], 
      Hue[0.7802110690906503, 1, 1], 
      Hue[0.7801864667138221, 1, 1], 
      Hue[0.7801372601593551, 1, 1], 
      Hue[0.7800388398598123, 1, 1], 
      Hue[0.7798419705994404, 1, 1], 
      Hue[0.7798296650058097, 1, 1], 
      Hue[0.7798173592637114, 1, 1], 
      Hue[0.7797927473343383, 1, 1], 
      Hue[0.7797435216964705, 1, 1], 
      Hue[0.7796450633169278, 1, 1], 
      Hue[0.7796327553546072, 1, 1], 
      Hue[0.7796204472447248, 1, 1], 
      Hue[0.7795958305825013, 1, 1], 
      Hue[0.7795465954898065, 1, 1], 
      Hue[0.7795342863485786, 1, 1], 
      Hue[0.7795219770602422, 1, 1], 
      Hue[0.7794973580424717, 1, 1], 
      Hue[0.7794850483131508, 1, 1], 
      Hue[0.7794727384369486, 1, 1], 
      Hue[0.7794604284139217, 1, 1], 
      Hue[0.779448118244127, 1, 1], 
      Hue[0.7205518817558729, 1, 1], 
      Hue[0.7201728632810981, 1, 1], 
      Hue[0.7201728632810981, 1, 1], 
      Hue[0.7197939855871971, 1, 1], 
      Hue[0.7197939855871971, 1, 1], 
      Hue[0.7190366591317676, 1, 1], 
      Hue[0.7190366591317676, 1, 1], 
      Hue[0.7175237677526916, 1, 1], 
      Hue[0.7175237677526916, 1, 1], 
      Hue[0.7145053924996501, 1, 1], 
      Hue[0.7145053924996501, 1, 1], 
      Hue[0.7141288210543033, 1, 1], 
      Hue[0.7141288210543033, 1, 1], 
      Hue[0.7137524160490083, 1, 1], 
      Hue[0.7137524160490083, 1, 1], 
      Hue[0.7130001115521821, 1, 1], 
      Hue[0.7130001115521821, 1, 1], 
      Hue[0.7114975675954697, 1, 1], 
      Hue[0.7114975675954697, 1, 1], 
      Hue[0.7085010779576801, 1, 1], 
      Hue[0.7085010779576801, 1, 1], 
      Hue[0.7081273517673412, 1, 1], 
      Hue[0.7081273517673412, 1, 1], 
      Hue[0.707753815997507, 1, 1], 
      Hue[0.707753815997507, 1, 1], 
      Hue[0.7070073214666093, 1, 1], 
      Hue[0.7070073214666093, 1, 1], 
      Hue[0.7055166802582684, 1, 1], 
      Hue[0.7055166802582684, 1, 1], 
      Hue[0.7025451015155051, 1, 1], 
      Hue[0.7025451015155051, 1, 1], 
      Hue[0.6966431533753163, 1, 1], 
      Hue[0.6966431533753163, 1, 1], 
      Hue[0.6962453407646028, 1, 1], 
      Hue[0.6962453407646028, 1, 1], 
      Hue[0.6958478017085719, 1, 1], 
      Hue[0.6958478017085719, 1, 1], 
      Hue[0.6950535502982287, 1, 1], 
      Hue[0.6950535502982287, 1, 1], 
      Hue[0.6934683959066561, 1, 1], 
      Hue[0.6934683959066561, 1, 1], 
      Hue[0.69031180412586, 1, 1], 
      Hue[0.69031180412586, 1, 1], 
      Hue[0.6840559180388815, 1, 1], 
      Hue[0.6840559180388815, 1, 1], 
      Hue[0.6717902321513325, 1, 1], 
      Hue[0.6717902321513325, 1, 1], 
      Hue[0.6714375103634247, 1, 1], 
      Hue[0.6714375103634247, 1, 1], 
      Hue[0.671085090804056, 1, 1], 
      Hue[0.671085090804056, 1, 1], 
      Hue[0.6703811608448678, 1, 1], 
      Hue[0.6703811608448678, 1, 1], 
      Hue[0.668976954394541, 1, 1], 
      Hue[0.668976954394541, 1, 1], 
      Hue[0.666183282738352, 1, 1], 
      Hue[0.666183282738352, 1, 1], 
      Hue[0.6606558087846726, 1, 1], 
      Hue[0.6606558087846726, 1, 1], 
      Hue[0.6498458025017817, 1, 1], 
      Hue[0.6498458025017817, 1, 1], 
      Hue[0.6296151700839989, 1, 1], 
      Hue[0.6296151700839989, 1, 1], 
      Hue[0.6090856243791776, 1, 1], 
      Hue[0.6090856243791776, 1, 1], 
      Hue[0.5911506349521957, 1, 1], 
      Hue[0.5911506349521957, 1, 1], 
      Hue[0.5728738758396617, 1, 1], 
      Hue[0.5728738758396617, 1, 1], 
      Hue[0.5726018283453903, 1, 1], 
      Hue[0.5726018283453903, 1, 1], 
      Hue[0.5723300141524164, 1, 1], 
      Hue[0.5723300141524164, 1, 1], 
      Hue[0.5717870824013804, 1, 1], 
      Hue[0.5717870824013804, 1, 1], 
      Hue[0.570703982533093, 1, 1], 
      Hue[0.570703982533093, 1, 1], 
      Hue[0.5685486537726839, 1, 1], 
      Hue[0.5685486537726839, 1, 1], 
      Hue[0.5642800076517256, 1, 1], 
      Hue[0.5642800076517256, 1, 1], 
      Hue[0.5558989435773015, 1, 1], 
      Hue[0.5558989435773015, 1, 1], 
      Hue[0.5556575555942079, 1, 1], 
      Hue[0.5556575555942079, 1, 1], 
      Hue[0.5554163247938143, 1, 1], 
      Hue[0.5554163247938143, 1, 1], 
      Hue[0.5549343320656529, 1, 1], 
      Hue[0.5549343320656529, 1, 1], 
      Hue[0.5539722033773958, 1, 1], 
      Hue[0.5539722033773958, 1, 1], 
      Hue[0.5520552233913043, 1, 1], 
      Hue[0.5520552233913043, 1, 1], 
      Hue[0.5482491773582174, 1, 1], 
      Hue[0.5482491773582174, 1, 1], 
      Hue[0.5480124880738455, 1, 1], 
      Hue[0.5480124880738455, 1, 1], 
      Hue[0.5477759346340769, 1, 1], 
      Hue[0.5477759346340769, 1, 1], 
      Hue[0.5473032326321359, 1, 1], 
      Hue[0.5473032326321359, 1, 1], 
      Hue[0.5463594295560026, 1, 1], 
      Hue[0.5463594295560026, 1, 1], 
      Hue[0.5444780786214425, 1, 1], 
      Hue[0.5444780786214425, 1, 1], 
      Hue[0.5407392124960726, 1, 1], 
      Hue[0.5407392124960726, 1, 1], 
      Hue[0.5404869192549508, 1, 1], 
      Hue[0.5404869192549508, 1, 1], 
      Hue[0.5402347608129789, 1, 1], 
      Hue[0.5402347608129789, 1, 1], 
      Hue[0.5397308449751576, 1, 1], 
      Hue[0.5397308449751576, 1, 1], 
      Hue[0.5387245940428427, 1, 1], 
      Hue[0.5387245940428427, 1, 1], 
      Hue[0.5367182278963121, 1, 1], 
      Hue[0.5367182278963121, 1, 1], 
      Hue[0.5327285448595175, 1, 1], 
      Hue[0.5327285448595175, 1, 1], 
      Hue[0.5324801512102756, 1, 1], 
      Hue[0.5324801512102756, 1, 1], 
      Hue[0.5322318657097549, 1, 1], 
      Hue[0.5322318657097549, 1, 1], 
      Hue[0.5317356158456428, 1, 1], 
      Hue[0.5317356158456428, 1, 1], 
      Hue[0.5307443775187145, 1, 1], 
      Hue[0.5307443775187145, 1, 1], 
      Hue[0.5287667615698055, 1, 1], 
      Hue[0.5287667615698055, 1, 1], 
      Hue[0.5248294975419259, 1, 1], 
      Hue[0.5248294975419259, 1, 1], 
      Hue[0.5246004552449229, 1, 1], 
      Hue[0.5246004552449229, 1, 1], 
      Hue[0.5243714843111781, 1, 1], 
      Hue[0.5243714843111781, 1, 1], 
      Hue[0.5239137538730607, 1, 1], 
      Hue[0.5239137538730607, 1, 1], 
      Hue[0.5229991201035351, 1, 1], 
      Hue[0.5229991201035351, 1, 1], 
      Hue[0.5211730122817956, 1, 1], 
      Hue[0.5211730122817956, 1, 1], 
      Hue[0.5209450320969299, 1, 1], 
      Hue[0.5209450320969299, 1, 1], 
      Hue[0.5207171126580099, 1, 1], 
      Hue[0.5207171126580099, 1, 1], 
      Hue[0.5202614533707344, 1, 1], 
      Hue[0.5202614533707344, 1, 1], 
      Hue[0.5193508346382325, 1, 1], 
      Hue[0.5193508346382325, 1, 1], 
      Hue[0.5175322485357188, 1, 1], 
      Hue[0.5175322485357188, 1, 1], 
      Hue[0.5173051609668913, 1, 1], 
      Hue[0.5173051609668913, 1, 1], 
      Hue[0.5170781235763294, 1, 1], 
      Hue[0.5170781235763294, 1, 1], 
      Hue[0.516624196694218, 1, 1], 
      Hue[0.516624196694218, 1, 1], 
      Hue[0.5157169160846813, 1, 1], 
      Hue[0.5157169160846813, 1, 1], 
      Hue[0.5154902114994349, 1, 1], 
      Hue[0.5154902114994349, 1, 1], 
      Hue[0.5152635518250663, 1, 1], 
      Hue[0.5152635518250663, 1, 1], 
      Hue[0.5148103645782285, 1, 1], 
      Hue[0.5148103645782285, 1, 1], 
      Hue[0.5139045000852724, 1, 1], 
      Hue[0.5139045000852724, 1, 1], 
      Hue[0.513678136379099, 1, 1], 
      Hue[0.513678136379099, 1, 1], 
      Hue[0.5134518123261133, 1, 1], 
      Hue[0.5134518123261133, 1, 1], 
      Hue[0.5129992805535211, 1, 1], 
      Hue[0.5129992805535211, 1, 1], 
      Hue[0.5120946639659323, 1, 1], 
      Hue[0.5120946639659323, 1, 1], 
      Hue[0.5102870717187569, 1, 1], 
      Hue[0.5102870717187569, 1, 1], 
      Hue[0.5100657056310821, 1, 1], 
      Hue[0.5100657056310821, 1, 1], 
      Hue[0.5098443675824635, 1, 1], 
      Hue[0.5098443675824635, 1, 1], 
      Hue[0.5094017731352407, 1, 1], 
      Hue[0.5094017731352407, 1, 1], 
      Hue[0.50851689357583, 1, 1], 
      Hue[0.50851689357583, 1, 1], 
      Hue[0.5082957345349084, 1, 1], 
      Hue[0.5082957345349084, 1, 1], 
      Hue[0.508074598601023, 1, 1], 
      Hue[0.508074598601023, 1, 1], 
      Hue[0.5076323935899204, 1, 1], 
      Hue[0.5076323935899204, 1, 1], 
      Hue[0.5067482337474685, 1, 1], 
      Hue[0.5067482337474685, 1, 1], 
      Hue[0.5065272423153528, 1, 1], 
      Hue[0.5065272423153528, 1, 1], 
      Hue[0.5063062690632634, 1, 1], 
      Hue[0.5063062690632634, 1, 1], 
      Hue[0.5058643746369205, 1, 1], 
      Hue[0.5058643746369205, 1, 1], 
      Hue[0.5056434522316935, 1, 1], 
      Hue[0.5056434522316935, 1, 1], 
      Hue[0.5054225455445416, 1, 1], 
      Hue[0.5054225455445416, 1, 1], 
      Hue[0.504980776863122, 1, 1], 
      Hue[0.504980776863122, 1, 1], 
      Hue[0.504759913638311, 1, 1], 
      Hue[0.504759913638311, 1, 1], 
      Hue[0.5045390636704838, 1, 1], 
      Hue[0.5045390636704838, 1, 1], 
      Hue[0.5040974010452097, 1, 1], 
      Hue[0.5040974010452097, 1, 1], 
      Hue[0.5038765871575825, 1, 1], 
      Hue[0.5038765871575825, 1, 1], 
      Hue[0.5036557840665759, 1, 1], 
      Hue[0.5036557840665759, 1, 1], 
      Hue[0.5032142078144854, 1, 1], 
      Hue[0.5032142078144854, 1, 1], 
      Hue[0.5029934334235187, 1, 1], 
      Hue[0.5029934334235187, 1, 1], 
      Hue[0.5027726673694013, 1, 1], 
      Hue[0.5027726673694013, 1, 1], 
      Hue[0.5025519090372679, 1, 1], 
      Hue[0.5025519090372679, 1, 1], 
      Hue[0.5023311578122791, 1, 1], 
      Hue[0.5023311578122791, 1, 1], 
      Hue[0.5021104130796202, 1, 1], 
      Hue[0.5021104130796202, 1, 1], 
      Hue[0.5018896742244989, 1, 1], 
      Hue[0.5018896742244989, 1, 1], 
      Hue[0.5016689406321433, 1, 1], 
      Hue[0.5016689406321433, 1, 1], 
      Hue[0.501448211687799, 1, 1], 
      Hue[0.501448211687799, 1, 1], 
      Hue[0.5012274867767283, 1, 1], 
      Hue[0.5012274867767283, 1, 1], 
      Hue[0.5010067652842074, 1, 1], 
      Hue[0.5010067652842074, 1, 1], 
      Hue[0.5007860465955236, 1, 1], 
      Hue[0.5007860465955236, 1, 1], 
      Hue[0.5005653300959744, 1, 1], 
      Hue[0.5005653300959744, 1, 1], 
      Hue[0.5003446151708648, 1, 1], 
      Hue[0.5003446151708648, 1, 1], 
      Hue[0.5001239012055051, 1, 1], 
      Hue[0.5001239012055051, 1, 1], 
      Hue[0.499903187585209, 1, 1], 
      Hue[0.499903187585209, 1, 1], 
      Hue[0.4996824736952913, 1, 1], 
      Hue[0.4996824736952913, 1, 1], 
      Hue[0.4994617589210659, 1, 1], 
      Hue[0.4994617589210659, 1, 1], 
      Hue[0.4992410426478438, 1, 1], 
      Hue[0.4992410426478438, 1, 1], 
      Hue[0.4990203242609306, 1, 1], 
      Hue[0.4990203242609306, 1, 1], 
      Hue[0.4987996031456248, 1, 1], 
      Hue[0.4987996031456248, 1, 1], 
      Hue[0.49857887868721534`, 1, 1], 
      Hue[0.49857887868721534`, 1, 1], 
      Hue[0.4983581502709798, 1, 1], 
      Hue[0.4983581502709798, 1, 1], 
      Hue[0.4981374172821819, 1, 1], 
      Hue[0.4981374172821819, 1, 1], 
      Hue[0.4979166791060699, 1, 1], 
      Hue[0.4979166791060699, 1, 1], 
      Hue[0.49769593512787397`, 1, 1], 
      Hue[0.49769593512787397`, 1, 1], 
      Hue[0.4974751847328044, 1, 1], 
      Hue[0.4974751847328044, 1, 1], 
      Hue[0.49725442730604935`, 1, 1], 
      Hue[0.49725442730604935`, 1, 1], 
      Hue[0.4970336622327728, 1, 1], 
      Hue[0.4970336622327728, 1, 1], 
      Hue[0.49681288889811237`, 1, 1], 
      Hue[0.49681288889811237`, 1, 1], 
      Hue[0.49659210668717735`, 1, 1], 
      Hue[0.49659210668717735`, 1, 1], 
      Hue[0.4961505131767657, 1, 1], 
      Hue[0.4961505131767657, 1, 1], 
      Hue[0.49591092983354623`, 1, 1], 
      Hue[0.49591092983354623`, 1, 1], 
      Hue[0.4956713330834243, 1, 1], 
      Hue[0.4956713330834243, 1, 1], 
      Hue[0.4951920962198805, 1, 1], 
      Hue[0.4951920962198805, 1, 1], 
      Hue[0.49495245453501585`, 1, 1], 
      Hue[0.49495245453501585`, 1, 1], 
      Hue[0.494712796300357, 1, 1], 
      Hue[0.494712796300357, 1, 1], 
      Hue[0.49423342703805917`, 1, 1], 
      Hue[0.49423342703805917`, 1, 1], 
      Hue[0.4939937144384435, 1, 1], 
      Hue[0.4939937144384435, 1, 1], 
      Hue[0.49375398214507416`, 1, 1], 
      Hue[0.49375398214507416`, 1, 1], 
      Hue[0.4932744553322717, 1, 1], 
      Hue[0.4932744553322717, 1, 1], 
      Hue[0.4923151307840086, 1, 1], 
      Hue[0.4923151307840086, 1, 1], 
      Hue[0.4920752386164939, 1, 1], 
      Hue[0.4920752386164939, 1, 1], 
      Hue[0.4918353204633067, 1, 1], 
      Hue[0.4918353204633067, 1, 1], 
      Hue[0.491355403052108, 1, 1], 
      Hue[0.491355403052108, 1, 1], 
      Hue[0.4903952217696123, 1, 1], 
      Hue[0.4903952217696123, 1, 1], 
      Hue[0.49015509967662196`, 1, 1], 
      Hue[0.49015509967662196`, 1, 1], 
      Hue[0.4899149452993977, 1, 1], 
      Hue[0.4899149452993977, 1, 1], 
      Hue[0.4894345365406593, 1, 1], 
      Hue[0.4894345365406593, 1, 1], 
      Hue[0.488473296937411, 1, 1], 
      Hue[0.488473296937411, 1, 1], 
      Hue[0.4882328945028128, 1, 1], 
      Hue[0.4882328945028128, 1, 1], 
      Hue[0.4879924534772371, 1, 1], 
      Hue[0.4879924534772371, 1, 1], 
      Hue[0.4875114524970251, 1, 1], 
      Hue[0.4875114524970251, 1, 1], 
      Hue[0.48654895271867454`, 1, 1], 
      Hue[0.48654895271867454`, 1, 1], 
      Hue[0.4863082194556099, 1, 1], 
      Hue[0.4863082194556099, 1, 1], 
      Hue[0.48606744128514545`, 1, 1], 
      Hue[0.48606744128514545`, 1, 1], 
      Hue[0.4855857470606176, 1, 1], 
      Hue[0.4855857470606176, 1, 1], 
      Hue[0.48462178493732505`, 1, 1], 
      Hue[0.48462178493732505`, 1, 1], 
      Hue[0.4843806702764197, 1, 1], 
      Hue[0.4843806702764197, 1, 1], 
      Hue[0.4841395043805711, 1, 1], 
      Hue[0.4841395043805711, 1, 1], 
      Hue[0.4836570157166681, 1, 1], 
      Hue[0.4836570157166681, 1, 1], 
      Hue[0.4826913887171722, 1, 1], 
      Hue[0.4826913887171722, 1, 1], 
      Hue[0.480757358393055, 1, 1], 
      Hue[0.480757358393055, 1, 1], 
      Hue[0.4805315263020154, 1, 1], 
      Hue[0.4805315263020154, 1, 1], 
      Hue[0.4803056385978369, 1, 1], 
      Hue[0.4803056385978369, 1, 1], 
      Hue[0.47985369376570597`, 1, 1], 
      Hue[0.47985369376570597`, 1, 1], 
      Hue[0.47894910830569265`, 1, 1], 
      Hue[0.47894910830569265`, 1, 1], 
      Hue[0.47713700924086944`, 1, 1], 
      Hue[0.47713700924086944`, 1, 1], 
      Hue[0.4769102097107648, 1, 1], 
      Hue[0.4769102097107648, 1, 1], 
      Hue[0.47668334420836855`, 1, 1], 
      Hue[0.47668334420836855`, 1, 1], 
      Hue[0.47622941269007835`, 1, 1], 
      Hue[0.47622941269007835`, 1, 1], 
      Hue[0.47532072941166953`, 1, 1], 
      Hue[0.47532072941166953`, 1, 1], 
      Hue[0.47349993620919084`, 1, 1], 
      Hue[0.47349993620919084`, 1, 1], 
      Hue[0.47327200311344897`, 1, 1], 
      Hue[0.47327200311344897`, 1, 1], 
      Hue[0.47304399363458505`, 1, 1], 
      Hue[0.47304399363458505`, 1, 1], 
      Hue[0.4725877429172142, 1, 1], 
      Hue[0.4725877429172142, 1, 1], 
      Hue[0.47167429615942985`, 1, 1], 
      Hue[0.47167429615942985`, 1, 1], 
      Hue[0.4698434748853157, 1, 1], 
      Hue[0.4698434748853157, 1, 1], 
      Hue[0.46616494645249384`, 1, 1], 
      Hue[0.46616494645249384`, 1, 1], 
      Hue[0.4659148691575225, 1, 1], 
      Hue[0.4659148691575225, 1, 1], 
      Hue[0.46566467738003314`, 1, 1], 
      Hue[0.46566467738003314`, 1, 1], 
      Hue[0.4651639470144475, 1, 1], 
      Hue[0.4651639470144475, 1, 1], 
      Hue[0.4641610754806451, 1, 1], 
      Hue[0.4641610754806451, 1, 1], 
      Hue[0.46214950043805014`, 1, 1], 
      Hue[0.46214950043805014`, 1, 1], 
      Hue[0.4581015078231806, 1, 1], 
      Hue[0.4581015078231806, 1, 1], 
      Hue[0.4578473484445837, 1, 1], 
      Hue[0.4578473484445837, 1, 1], 
      Hue[0.4575930475155824, 1, 1], 
      Hue[0.4575930475155824, 1, 1], 
      Hue[0.45708401760083006`, 1, 1], 
      Hue[0.45708401760083006`, 1, 1], 
      Hue[0.4560642216878684, 1, 1], 
      Hue[0.4560642216878684, 1, 1], 
      Hue[0.4540174944156169, 1, 1], 
      Hue[0.4540174944156169, 1, 1], 
      Hue[0.4498939651897446, 1, 1], 
      Hue[0.4498939651897446, 1, 1], 
      Hue[0.4496395920714616, 1, 1], 
      Hue[0.4496395920714616, 1, 1], 
      Hue[0.449385056150846, 1, 1], 
      Hue[0.449385056150846, 1, 1], 
      Hue[0.4488754926464263, 1, 1], 
      Hue[0.4488754926464263, 1, 1], 
      Hue[0.4478543761796728, 1, 1], 
      Hue[0.4478543761796728, 1, 1], 
      Hue[0.4458040028063383, 1, 1], 
      Hue[0.4458040028063383, 1, 1], 
      Hue[0.44166923112517276`, 1, 1], 
      Hue[0.44166923112517276`, 1, 1], 
      Hue[0.4332518596871854, 1, 1], 
      Hue[0.4332518596871854, 1, 1], 
      Hue[0.4169447860407514, 1, 1], 
      Hue[0.4169447860407514, 1, 1], 
      Hue[0.39819770478061217`, 1, 1], 
      Hue[0.39819770478061217`, 1, 1], 
      Hue[0.3795344901876919, 1, 1], 
      Hue[0.3795344901876919, 1, 1], 
      Hue[0.3578952936955762, 1, 1], 
      Hue[0.3578952936955762, 1, 1], 
      Hue[0.35755137633603035`, 1, 1], 
      Hue[0.35755137633603035`, 1, 1], 
      Hue[0.35720709405282786`, 1, 1], 
      Hue[0.35720709405282786`, 1, 1], 
      Hue[0.3565174345978122, 1, 1], 
      Hue[0.3565174345978122, 1, 1], 
      Hue[0.3551337357577113, 1, 1], 
      Hue[0.3551337357577113, 1, 1], 
      Hue[0.3523488221943377, 1, 1], 
      Hue[0.3523488221943377, 1, 1], 
      Hue[0.3467090755102115, 1, 1], 
      Hue[0.3467090755102115, 1, 1], 
      Hue[0.33515298547996475`, 1, 1], 
      Hue[0.33515298547996475`, 1, 1], 
      Hue[0.334810643178493, 1, 1], 
      Hue[0.334810643178493, 1, 1], 
      Hue[0.3344679970554599, 1, 1], 
      Hue[0.3344679970554599, 1, 1], 
      Hue[0.33378179518226486`, 1, 1], 
      Hue[0.33378179518226486`, 1, 1], 
      Hue[0.3324057662380327, 1, 1], 
      Hue[0.3324057662380327, 1, 1], 
      Hue[0.32963932081463193`, 1, 1], 
      Hue[0.32963932081463193`, 1, 1], 
      Hue[0.3240498957473761, 1, 1], 
      Hue[0.3240498957473761, 1, 1], 
      Hue[0.32369809827084883`, 1, 1], 
      Hue[0.32369809827084883`, 1, 1], 
      Hue[0.32334601565736026`, 1, 1], 
      Hue[0.32334601565736026`, 1, 1], 
      Hue[0.3226409979132937, 1, 1], 
      Hue[0.3226409979132937, 1, 1], 
      Hue[0.32122757308827055`, 1, 1], 
      Hue[0.32122757308827055`, 1, 1], 
      Hue[0.3183873390867039, 1, 1], 
      Hue[0.3183873390867039, 1, 1], 
      Hue[0.3126548323856401, 1, 1], 
      Hue[0.3126548323856401, 1, 1], 
      Hue[0.31226394002872365`, 1, 1], 
      Hue[0.31226394002872365`, 1, 1], 
      Hue[0.3118727446949302, 1, 1], 
      Hue[0.3118727446949302, 1, 1], 
      Hue[0.31108945018381584`, 1, 1], 
      Hue[0.31108945018381584`, 1, 1], 
      Hue[0.30951928205543044`, 1, 1], 
      Hue[0.30951928205543044`, 1, 1], 
      Hue[0.3063649293548802, 1, 1], 
      Hue[0.3063649293548802, 1, 1], 
      Hue[0.30596934819271, 1, 1], 
      Hue[0.30596934819271, 1, 1], 
      Hue[0.3055734857590331, 1, 1], 
      Hue[0.3055734857590331, 1, 1], 
      Hue[0.30478092290960934`, 1, 1], 
      Hue[0.30478092290960934`, 1, 1], 
      Hue[0.30319248675089605`, 1, 1], 
      Hue[0.30319248675089605`, 1, 1], 
      Hue[0.3000027138831849, 1, 1], 
      Hue[0.3000027138831849, 1, 1], 
      Hue[0.29960281338186984`, 1, 1], 
      Hue[0.29960281338186984`, 1, 1], 
      Hue[0.29920265625388204`, 1, 1], 
      Hue[0.29920265625388204`, 1, 1], 
      Hue[0.29840157866966915`, 1, 1], 
      Hue[0.29840157866966915`, 1, 1], 
      Hue[0.2967964166611863, 1, 1], 
      Hue[0.2967964166611863, 1, 1], 
      Hue[0.2935744461038008, 1, 1], 
      Hue[0.2935744461038008, 1, 1], 
      Hue[0.2870871058179891, 1, 1], 
      Hue[0.2870871058179891, 1, 1], 
      Hue[0.2867068783556848, 1, 1], 
      Hue[0.2867068783556848, 1, 1], 
      Hue[0.28632647720527604`, 1, 1], 
      Hue[0.28632647720527604`, 1, 1], 
      Hue[0.28556516022402634`, 1, 1], 
      Hue[0.28556516022402634`, 1, 1], 
      Hue[0.284040512612144, 1, 1], 
      Hue[0.284040512612144, 1, 1], 
      Hue[0.28098352900329876`, 1, 1], 
      Hue[0.28098352900329876`, 1, 1], 
      Hue[0.2806007176236879, 1, 1], 
      Hue[0.2806007176236879, 1, 1], 
      Hue[0.2802177588652316, 1, 1], 
      Hue[0.2802177588652316, 1, 1], 
      Hue[0.27945140601201585`, 1, 1], 
      Hue[0.27945140601201585`, 1, 1], 
      Hue[0.2779170068960574, 1, 1], 
      Hue[0.2779170068960574, 1, 1], 
      Hue[0.2748418233531421, 1, 1], 
      Hue[0.2748418233531421, 1, 1], 
      Hue[0.27445686099902716`, 1, 1], 
      Hue[0.27445686099902716`, 1, 1], 
      Hue[0.2740717791307105, 1, 1], 
      Hue[0.2740717791307105, 1, 1], 
      Hue[0.2733012640056608, 1, 1], 
      Hue[0.2733012640056608, 1, 1], 
      Hue[0.2717588785561085, 1, 1], 
      Hue[0.2717588785561085, 1, 1], 
      Hue[0.27137301040675016`, 1, 1], 
      Hue[0.27137301040675016`, 1, 1], 
      Hue[0.2709870371770742, 1, 1], 
      Hue[0.2709870371770742, 1, 1], 
      Hue[0.2702147827817822, 1, 1], 
      Hue[0.2702147827817822, 1, 1], 
      Hue[0.26866909362519453`, 1, 1], 
      Hue[0.26866909362519453`, 1, 1], 
      Hue[0.26828243617916075`, 1, 1], 
      Hue[0.26828243617916075`, 1, 1], 
      Hue[0.2678956883732717, 1, 1], 
      Hue[0.2678956883732717, 1, 1], 
      Hue[0.2671219291190381, 1, 1], 
      Hue[0.2671219291190381, 1, 1], 
      Hue[0.26557340830942106`, 1, 1], 
      Hue[0.26557340830942106`, 1, 1], 
      Hue[0.2651860802254811, 1, 1], 
      Hue[0.2651860802254811, 1, 1], 
      Hue[0.264798676750572, 1, 1], 
      Hue[0.264798676750572, 1, 1], 
      Hue[0.2640236511774876, 1, 1], 
      Hue[0.2640236511774876, 1, 1], 
      Hue[0.26247277855953854`, 1, 1], 
      Hue[0.26247277855953854`, 1, 1], 
      Hue[0.26209251788900434`, 1, 1], 
      Hue[0.26209251788900434`, 1, 1], 
      Hue[0.26171219931070255`, 1, 1], 
      Hue[0.26171219931070255`, 1, 1], 
      Hue[0.2609513956309302, 1, 1], 
      Hue[0.2609513956309302, 1, 1], 
      Hue[0.2605709141353225, 1, 1], 
      Hue[0.2605709141353225, 1, 1], 
      Hue[0.26019038194352423`, 1, 1], 
      Hue[0.26019038194352423`, 1, 1], 
      Hue[0.2594291727069773, 1, 1], 
      Hue[0.2594291727069773, 1, 1], 
      Hue[0.2590484992850898, 1, 1], 
      Hue[0.2590484992850898, 1, 1], 
      Hue[0.25866778241258415`, 1, 1], 
      Hue[0.25866778241258415`, 1, 1], 
      Hue[0.2579062255820407, 1, 1], 
      Hue[0.2579062255820407, 1, 1], 
      Hue[0.2575253892614616, 1, 1], 
      Hue[0.2575253892614616, 1, 1], 
      Hue[0.2571445167650305, 1, 1], 
      Hue[0.2571445167650305, 1, 1], 
      Hue[0.2563826705368009, 1, 1], 
      Hue[0.2563826705368009, 1, 1], 
      Hue[0.256001700454635, 1, 1], 
      Hue[0.256001700454635, 1, 1], 
      Hue[0.25562070149572913`, 1, 1], 
      Hue[0.25562070149572913`, 1, 1], 
      Hue[0.2552396754880936, 1, 1], 
      Hue[0.2552396754880936, 1, 1], 
      Hue[0.25485862426087846`, 1, 1], 
      Hue[0.25485862426087846`, 1, 1], 
      Hue[0.2544775496442966, 1, 1], 
      Hue[0.2544775496442966, 1, 1], 
      Hue[0.254096453469547, 1, 1], 
      Hue[0.254096453469547, 1, 1], 
      Hue[0.2537153375687379, 1, 1], 
      Hue[0.2537153375687379, 1, 1], 
      Hue[0.2533342037748092, 1, 1], 
      Hue[0.2533342037748092, 1, 1], 
      Hue[0.2529530539214558, 1, 1], 
      Hue[0.2529530539214558, 1, 1], 
      Hue[0.2525718898430503, 1, 1], 
      Hue[0.2525718898430503, 1, 1], 
      Hue[0.2521907133745654, 1, 1], 
      Hue[0.2521907133745654, 1, 1], 
      Hue[0.2518095263514968, 1, 1], 
      Hue[0.2518095263514968, 1, 1], 
      Hue[0.2514283306097857, 1, 1], 
      Hue[0.2514283306097857, 1, 1], 
      Hue[0.2510471279857414, 1, 1], 
      Hue[0.2510471279857414, 1, 1], 
      Hue[0.2506659203159637, 1, 1], 
      Hue[0.2506659203159637, 1, 1], 
      Hue[0.25028470943726544`, 1, 1], 
      Hue[0.25028470943726544`, 1, 1], 
      Hue[0.24990349718659496`, 1, 1], 
      Hue[0.24990349718659496`, 1, 1], 
      Hue[0.24952228540095855`, 1, 1], 
      Hue[0.24952228540095855`, 1, 1], 
      Hue[0.24914107591734283`, 1, 1], 
      Hue[0.24914107591734283`, 1, 1], 
      Hue[0.24875987057263726`, 1, 1], 
      Hue[0.24875987057263726`, 1, 1], 
      Hue[0.24837867120355647`, 1, 1], 
      Hue[0.24837867120355647`, 1, 1], 
      Hue[0.2479974796465629, 1, 1], 
      Hue[0.2479974796465629, 1, 1], 
      Hue[0.2476162977377892, 1, 1], 
      Hue[0.2476162977377892, 1, 1], 
      Hue[0.2472351273129606, 1, 1], 
      Hue[0.2472351273129606, 1, 1], 
      Hue[0.24685397020731778`, 1, 1], 
      Hue[0.24685397020731778`, 1, 1], 
      Hue[0.24647282825553923`, 1, 1], 
      Hue[0.24647282825553923`, 1, 1], 
      Hue[0.24609170329166405`, 1, 1], 
      Hue[0.24609170329166405`, 1, 1], 
      Hue[0.24571059714901458`, 1, 1], 
      Hue[0.24571059714901458`, 1, 1], 
      Hue[0.24532951166011932`, 1, 1], 
      Hue[0.24532951166011932`, 1, 1], 
      Hue[0.24494844865663576`, 1, 1], 
      Hue[0.24494844865663576`, 1, 1], 
      Hue[0.24418639742771617`, 1, 1], 
      Hue[0.24418639742771617`, 1, 1], 
      Hue[0.24380541286054708`, 1, 1], 
      Hue[0.24380541286054708`, 1, 1], 
      Hue[0.24342445809517, 1, 1], 
      Hue[0.24342445809517, 1, 1], 
      Hue[0.24266264527305542`, 1, 1], 
      Hue[0.24266264527305542`, 1, 1], 
      Hue[0.24228179086454416`, 1, 1], 
      Hue[0.24228179086454416`, 1, 1], 
      Hue[0.24190097355412485`, 1, 1], 
      Hue[0.24190097355412485`, 1, 1], 
      Hue[0.2411394575073841, 1, 1], 
      Hue[0.2411394575073841, 1, 1], 
      Hue[0.2407587624068073, 1, 1], 
      Hue[0.2407587624068073, 1, 1], 
      Hue[0.24037811167566045`, 1, 1], 
      Hue[0.24037811167566045`, 1, 1], 
      Hue[0.23961695057318147`, 1, 1], 
      Hue[0.23961695057318147`, 1, 1], 
      Hue[0.2380952404554197, 1, 1], 
      Hue[0.2380952404554197, 1, 1], 
      Hue[0.23768268530800246`, 1, 1], 
      Hue[0.23768268530800246`, 1, 1], 
      Hue[0.23727019956300532`, 1, 1], 
      Hue[0.23727019956300532`, 1, 1], 
      Hue[0.23644544544623258`, 1, 1], 
      Hue[0.23644544544623258`, 1, 1], 
      Hue[0.23479687066374988`, 1, 1], 
      Hue[0.23479687066374988`, 1, 1], 
      Hue[0.23438493471701696`, 1, 1], 
      Hue[0.23438493471701696`, 1, 1], 
      Hue[0.23397308640684428`, 1, 1], 
      Hue[0.23397308640684428`, 1, 1], 
      Hue[0.23314966173624715`, 1, 1], 
      Hue[0.23314966173624715`, 1, 1], 
      Hue[0.23150396322258707`, 1, 1], 
      Hue[0.23150396322258707`, 1, 1], 
      Hue[0.23109279142615208`, 1, 1], 
      Hue[0.23109279142615208`, 1, 1], 
      Hue[0.23068172522628413`, 1, 1], 
      Hue[0.23068172522628413`, 1, 1], 
      Hue[0.22985991850312365`, 1, 1], 
      Hue[0.22985991850312365`, 1, 1], 
      Hue[0.2282176696727462, 1, 1], 
      Hue[0.2282176696727462, 1, 1], 
      Hue[0.22493912100615956`, 1, 1], 
      Hue[0.22493912100615956`, 1, 1], 
      Hue[0.22452990240719808`, 1, 1], 
      Hue[0.22452990240719808`, 1, 1], 
      Hue[0.22412082428986946`, 1, 1], 
      Hue[0.22412082428986946`, 1, 1], 
      Hue[0.22330309800330872`, 1, 1], 
      Hue[0.22330309800330872`, 1, 1], 
      Hue[0.2216694243601005, 1, 1], 
      Hue[0.2216694243601005, 1, 1], 
      Hue[0.21840965988103805`, 1, 1], 
      Hue[0.21840965988103805`, 1, 1], 
      Hue[0.21800294032724274`, 1, 1], 
      Hue[0.21800294032724274`, 1, 1], 
      Hue[0.21759639443153517`, 1, 1], 
      Hue[0.21759639443153517`, 1, 1], 
      Hue[0.21678383164015788`, 1, 1], 
      Hue[0.21678383164015788`, 1, 1], 
      Hue[0.21516087777741738`, 1, 1], 
      Hue[0.21516087777741738`, 1, 1], 
      Hue[0.2119240955782303, 1, 1], 
      Hue[0.2119240955782303, 1, 1], 
      Hue[0.21154735354250084`, 1, 1], 
      Hue[0.21154735354250084`, 1, 1], 
      Hue[0.21117078972624026`, 1, 1], 
      Hue[0.21117078972624026`, 1, 1], 
      Hue[0.21041820282444174`, 1, 1], 
      Hue[0.21041820282444174`, 1, 1], 
      Hue[0.20891523405310197`, 1, 1], 
      Hue[0.20891523405310197`, 1, 1], 
      Hue[0.2059184474441574, 1, 1], 
      Hue[0.2059184474441574, 1, 1], 
      Hue[0.19996402390747361`, 1, 1], 
      Hue[0.19996402390747361`, 1, 1], 
      Hue[0.19959370230262743`, 1, 1], 
      Hue[0.19959370230262743`, 1, 1], 
      Hue[0.1992236038568319, 1, 1], 
      Hue[0.1992236038568319, 1, 1], 
      Hue[0.19848408153459976`, 1, 1], 
      Hue[0.19848408153459976`, 1, 1], 
      Hue[0.19700777035336184`, 1, 1], 
      Hue[0.19700777035336184`, 1, 1], 
      Hue[0.19406635637185005`, 1, 1], 
      Hue[0.19406635637185005`, 1, 1], 
      Hue[0.18823044513520215`, 1, 1], 
      Hue[0.18823044513520215`, 1, 1], 
      Hue[0.18783747724681005`, 1, 1], 
      Hue[0.18783747724681005`, 1, 1], 
      Hue[0.1874448146329372, 1, 1], 
      Hue[0.1874448146329372, 1, 1], 
      Hue[0.18666041032284467`, 1, 1], 
      Hue[0.18666041032284467`, 1, 1], 
      Hue[0.1850953203601452, 1, 1], 
      Hue[0.1850953203601452, 1, 1], 
      Hue[0.18198028486572737`, 1, 1], 
      Hue[0.18198028486572737`, 1, 1], 
      Hue[0.1758127883673826, 1, 1], 
      Hue[0.1758127883673826, 1, 1], 
      Hue[0.1637414997039794, 1, 1], 
      Hue[0.1637414997039794, 1, 1], 
      Hue[0.16337692046153715`, 1, 1], 
      Hue[0.16337692046153715`, 1, 1], 
      Hue[0.16301269104984262`, 1, 1], 
      Hue[0.16301269104984262`, 1, 1], 
      Hue[0.16228528369910886`, 1, 1], 
      Hue[0.16228528369910886`, 1, 1], 
      Hue[0.1608346881817319, 1, 1], 
      Hue[0.1608346881817319, 1, 1], 
      Hue[0.1579504717808601, 1, 1], 
      Hue[0.1579504717808601, 1, 1], 
      Hue[0.15225058847550843`, 1, 1], 
      Hue[0.15225058847550843`, 1, 1], 
      Hue[0.14112821611198173`, 1, 1], 
      Hue[0.14112821611198173`, 1, 1], 
      Hue[0.12136682583576486`, 1, 1], 
      Hue[0.12136682583576486`, 1, 1], 
      Hue[0.10132435565811605`, 1, 1], 
      Hue[0.10132435565811605`, 1, 1], 
      Hue[0.08378739001554548, 1, 1], 
      Hue[0.08378739001554548, 1, 1], 
      Hue[0.06753590541558296, 1, 1], 
      Hue[0.06753590541558296, 1, 1], 
      Hue[0.0672674238272465, 1, 1], 
      Hue[0.0672674238272465, 1, 1], 
      Hue[0.06699916076804159, 1, 1], 
      Hue[0.06699916076804159, 1, 1], 
      Hue[0.06646328691034005, 1, 1], 
      Hue[0.06646328691034005, 1, 1], 
      Hue[0.06539412494024377, 1, 1], 
      Hue[0.06539412494024377, 1, 1], 
      Hue[0.06326595746667119, 1, 1], 
      Hue[0.06326595746667119, 1, 1], 
      Hue[0.05904875488299526, 1, 1], 
      Hue[0.05904875488299526, 1, 1], 
      Hue[0.05878685308360073, 1, 1], 
      Hue[0.05878685308360073, 1, 1], 
      Hue[0.05852514314748131, 1, 1], 
      Hue[0.05852514314748131, 1, 1], 
      Hue[0.05800229553060262, 1, 1], 
      Hue[0.05800229553060262, 1, 1], 
      Hue[0.05695886598100607, 1, 1], 
      Hue[0.05695886598100607, 1, 1], 
      Hue[0.05488088299658431, 1, 1], 
      Hue[0.05488088299658431, 1, 1], 
      Hue[0.05075893020904179, 1, 1], 
      Hue[0.05075893020904179, 1, 1], 
      Hue[0.05051990962755181, 1, 1], 
      Hue[0.05051990962755181, 1, 1], 
      Hue[0.0502810329455785, 1, 1], 
      Hue[0.0502810329455785, 1, 1], 
      Hue[0.049803708587567615`, 1, 1], 
      Hue[0.049803708587567615`, 1, 1], 
      Hue[0.048850757056675125`, 1, 1], 
      Hue[0.048850757056675125`, 1, 1], 
      Hue[0.04695149217825662, 1, 1], 
      Hue[0.04695149217825662, 1, 1], 
      Hue[0.043178313304186866`, 1, 1], 
      Hue[0.043178313304186866`, 1, 1], 
      Hue[0.04294356452866228, 1, 1], 
      Hue[0.04294356452866228, 1, 1], 
      Hue[0.04270893820573536, 1, 1], 
      Hue[0.04270893820573536, 1, 1], 
      Hue[0.04224005025099981, 1, 1], 
      Hue[0.04224005025099981, 1, 1], 
      Hue[0.041303714452091696`, 1, 1], 
      Hue[0.041303714452091696`, 1, 1], 
      Hue[0.039436654259988296`, 1, 1], 
      Hue[0.039436654259988296`, 1, 1], 
      Hue[0.03572379038379056, 1, 1], 
      Hue[0.03572379038379056, 1, 1], 
      Hue[0.03547320455100585, 1, 1], 
      Hue[0.03547320455100585, 1, 1], 
      Hue[0.03522273762146896, 1, 1], 
      Hue[0.03522273762146896, 1, 1], 
      Hue[0.03472215711393221, 1, 1], 
      Hue[0.03472215711393221, 1, 1], 
      Hue[0.03372238601633399, 1, 1], 
      Hue[0.03372238601633399, 1, 1], 
      Hue[0.03172821586249482, 1, 1], 
      Hue[0.03172821586249482, 1, 1], 
      Hue[0.03147943191275808, 1, 1], 
      Hue[0.03147943191275808, 1, 1], 
      Hue[0.03123075347217772, 1, 1], 
      Hue[0.03123075347217772, 1, 1], 
      Hue[0.030733709781557574`, 1, 1], 
      Hue[0.030733709781557574`, 1, 1], 
      Hue[0.029740851822258815`, 1, 1], 
      Hue[0.029740851822258815`, 1, 1], 
      Hue[0.027759867141148174`, 1, 1], 
      Hue[0.027759867141148174`, 1, 1], 
      Hue[0.027512671401734252`, 1, 1], 
      Hue[0.027512671401734252`, 1, 1], 
      Hue[0.027265567861157906`, 1, 1], 
      Hue[0.027265567861157906`, 1, 1], 
      Hue[0.02677163406000022, 1, 1], 
      Hue[0.02677163406000022, 1, 1], 
      Hue[0.02578483637921635, 1, 1], 
      Hue[0.02578483637921635, 1, 1], 
      Hue[0.0238153353695229, 1, 1], 
      Hue[0.0238153353695229, 1, 1], 
      Hue[0.023569515476446326`, 1, 1], 
      Hue[0.023569515476446326`, 1, 1], 
      Hue[0.023323774550983816`, 1, 1], 
      Hue[0.023323774550983816`, 1, 1], 
      Hue[0.022832526305280633`, 1, 1], 
      Hue[0.022832526305280633`, 1, 1], 
      Hue[0.021850941166985965`, 1, 1], 
      Hue[0.021850941166985965`, 1, 1], 
      Hue[0.019891231988468473`, 1, 1], 
      Hue[0.019891231988468473`, 1, 1], 
      Hue[0.019662777413417473`, 1, 1], 
      Hue[0.019662777413417473`, 1, 1], 
      Hue[0.0194343802662687, 1, 1], 
      Hue[0.0194343802662687, 1, 1], 
      Hue[0.01897775558442008, 1, 1], 
      Hue[0.01897775558442008, 1, 1], 
      Hue[0.018065165982028755`, 1, 1], 
      Hue[0.018065165982028755`, 1, 1], 
      Hue[0.0162424764644601, 1, 1], 
      Hue[0.0162424764644601, 1, 1], 
      Hue[0.01601486069341581, 1, 1], 
      Hue[0.01601486069341581, 1, 1], 
      Hue[0.0157872916858071, 1, 1], 
      Hue[0.0157872916858071, 1, 1], 
      Hue[0.015332291300638774`, 1, 1], 
      Hue[0.015332291300638774`, 1, 1], 
      Hue[0.014422822437338929`, 1, 1], 
      Hue[0.014422822437338929`, 1, 1], 
      Hue[0.01260586355789195, 1, 1], 
      Hue[0.01260586355789195, 1, 1], 
      Hue[0.012378916273309075`, 1, 1], 
      Hue[0.012378916273309075`, 1, 1], 
      Hue[0.012152005128142402`, 1, 1], 
      Hue[0.012152005128142402`, 1, 1], 
      Hue[0.011698288604802218`, 1, 1], 
      Hue[0.011698288604802218`, 1, 1], 
      Hue[0.010791260074104256`, 1, 1], 
      Hue[0.010791260074104256`, 1, 1], 
      Hue[0.01056458335196563, 1, 1], 
      Hue[0.01056458335196563, 1, 1], 
      Hue[0.010337937469823545`, 1, 1], 
      Hue[0.010337937469823545`, 1, 1], 
      Hue[0.009884735577967426, 1, 1], 
      Hue[0.009884735577967426, 1, 1], 
      Hue[0.008978672756857162, 1, 1], 
      Hue[0.008978672756857162, 1, 1], 
      Hue[0.008752224226806537, 1, 1], 
      Hue[0.008752224226806537, 1, 1], 
      Hue[0.008525801244265314, 1, 1], 
      Hue[0.008525801244265314, 1, 1], 
      Hue[0.008073029277279837, 1, 1], 
      Hue[0.008073029277279837, 1, 1], 
      Hue[0.007846678970838494, 1, 1], 
      Hue[0.007846678970838494, 1, 1], 
      Hue[0.007620351567907857, 1, 1], 
      Hue[0.007620351567907857, 1, 1], 
      Hue[0.0071677628294967485`, 1, 1], 
      Hue[0.0071677628294967485`, 1, 1], 
      Hue[0.006941500172671094, 1, 1], 
      Hue[0.006941500172671094, 1, 1], 
      Hue[0.006715257776661196, 1, 1], 
      Hue[0.006715257776661196, 1, 1], 
      Hue[0.006262831125209884, 1, 1], 
      Hue[0.006262831125209884, 1, 1], 
      Hue[0.005358191895222516, 1, 1], 
      Hue[0.005358191895222516, 1, 1], 
      Hue[0.005136505252619112, 1, 1], 
      Hue[0.005136505252619112, 1, 1], 
      Hue[0.004914833019706234, 1, 1], 
      Hue[0.004914833019706234, 1, 1], 
      Hue[0.004471529295255007, 1, 1], 
      Hue[0.004471529295255007, 1, 1], 
      Hue[0.004249896559984951, 1, 1], 
      Hue[0.004249896559984951, 1, 1], 
      Hue[0.004028275746937781, 1, 1], 
      Hue[0.004028275746937781, 1, 1], 
      Hue[0.003585067400516932, 1, 1], 
      Hue[0.003585067400516932, 1, 1], 
      Hue[0.0033634786237412393`, 1, 1], 
      Hue[0.0033634786237412393`, 1, 1], 
      Hue[0.0031418992823801332`, 1, 1], 
      Hue[0.0031418992823801332`, 1, 1], 
      Hue[0.002920328754818563, 1, 1], 
      Hue[0.002920328754818563, 1, 1], 
      Hue[0.0026987664194720344`, 1, 1], 
      Hue[0.0026987664194720344`, 1, 1], 
      Hue[0.002477211654784387, 1, 1], 
      Hue[0.002477211654784387, 1, 1], 
      Hue[0.0022556638392256965`, 1, 1], 
      Hue[0.0022556638392256965`, 1, 1], 
      Hue[0.0020341223512901885`, 1, 1], 
      Hue[0.0020341223512901885`, 1, 1], 
      Hue[0.0018125865694940044`, 1, 1], 
      Hue[0.0018125865694940044`, 1, 1], 
      Hue[0.0015910558723730781`, 1, 1], 
      Hue[0.0015910558723730781`, 1, 1], 
      Hue[0.0013695296384809846`, 1, 1], 
      Hue[0.0013695296384809846`, 1, 1], 
      Hue[0.0011480072463867883`, 1, 1], 
      Hue[0.0011480072463867883`, 1, 1], 
      Hue[0.0009264880746729118, 1, 1], 
      Hue[0.0009264880746729118, 1, 1], 
      Hue[0.0007049715019329034, 1, 1], 
      Hue[0.0007049715019329034, 1, 1], 
      Hue[0.0004834569067693328, 1, 1], 
      Hue[0.0004834569067693328, 1, 1], 
      Hue[0.00026194366779169054`, 1, 1], 
      Hue[0.00026194366779169054`, 1, 1], 
      Hue[0.000040431163614147966`, 1, 1], 
      Hue[0.000040431163614147966`, 1, 1], 
      Hue[0.9998189187728534, 1, 1], 
      Hue[0.9998189187728534, 1, 1], 
      Hue[0.9995974058741266, 1, 1], 
      Hue[0.9995974058741266, 1, 1], 
      Hue[0.9993758918460491, 1, 1], 
      Hue[0.9993758918460491, 1, 1], 
      Hue[0.9991543760672325, 1, 1], 
      Hue[0.9991543760672325, 1, 1], 
      Hue[0.9989328579162817, 1, 1], 
      Hue[0.9989328579162817, 1, 1], 
      Hue[0.9987113367717942, 1, 1], 
      Hue[0.9987113367717942, 1, 1], 
      Hue[0.9984898120123568, 1, 1], 
      Hue[0.9984898120123568, 1, 1], 
      Hue[0.9982682830165432, 1, 1], 
      Hue[0.9982682830165432, 1, 1], 
      Hue[0.9980467491629134, 1, 1], 
      Hue[0.9980467491629134, 1, 1], 
      Hue[0.9978252098300097, 1, 1], 
      Hue[0.9978252098300097, 1, 1], 
      Hue[0.9976036643963558, 1, 1], 
      Hue[0.9976036643963558, 1, 1], 
      Hue[0.9973821122404543, 1, 1], 
      Hue[0.9973821122404543, 1, 1], 
      Hue[0.9971605527407842, 1, 1], 
      Hue[0.9971605527407842, 1, 1], 
      Hue[0.9969389852757992, 1, 1], 
      Hue[0.9969389852757992, 1, 1], 
      Hue[0.9964958239635591, 1, 1], 
      Hue[0.9964958239635591, 1, 1], 
      Hue[0.9962742288730648, 1, 1], 
      Hue[0.9962742288730648, 1, 1], 
      Hue[0.996052623330773, 1, 1], 
      Hue[0.996052623330773, 1, 1], 
      Hue[0.9956093784039354, 1, 1], 
      Hue[0.9956093784039354, 1, 1], 
      Hue[0.995387737775861, 1, 1], 
      Hue[0.995387737775861, 1, 1], 
      Hue[0.9951660842089277, 1, 1], 
      Hue[0.9951660842089277, 1, 1], 
      Hue[0.9947227357709507, 1, 1], 
      Hue[0.9947227357709507, 1, 1], 
      Hue[0.9945010396560208, 1, 1], 
      Hue[0.9945010396560208, 1, 1], 
      Hue[0.9942793281144557, 1, 1], 
      Hue[0.9942793281144557, 1, 1], 
      Hue[0.9938358562630776, 1, 1], 
      Hue[0.9938358562630776, 1, 1], 
      Hue[0.9936140947089531, 1, 1], 
      Hue[0.9936140947089531, 1, 1], 
      Hue[0.9933923152395664, 1, 1], 
      Hue[0.9933923152395664, 1, 1], 
      Hue[0.9929487000657207, 1, 1], 
      Hue[0.9929487000657207, 1, 1], 
      Hue[0.9927268631164576, 1, 1], 
      Hue[0.9927268631164576, 1, 1], 
      Hue[0.99250500576232, 1, 1], 
      Hue[0.99250500576232, 1, 1], 
      Hue[0.9920612273490585, 1, 1], 
      Hue[0.9920612273490585, 1, 1], 
      Hue[0.9911733982659021, 1, 1], 
      Hue[0.9911733982659021, 1, 1], 
      Hue[0.9909325747967033, 1, 1], 
      Hue[0.9909325747967033, 1, 1], 
      Hue[0.9906917213950743, 1, 1], 
      Hue[0.9906917213950743, 1, 1], 
      Hue[0.9902099216129358, 1, 1], 
      Hue[0.9902099216129358, 1, 1], 
      Hue[0.989245927854953, 1, 1], 
      Hue[0.989245927854953, 1, 1], 
      Hue[0.9890048426489481, 1, 1], 
      Hue[0.9890048426489481, 1, 1], 
      Hue[0.9887637211439682, 1, 1], 
      Hue[0.9887637211439682, 1, 1], 
      Hue[0.9882813660511997, 1, 1], 
      Hue[0.9882813660511997, 1, 1], 
      Hue[0.9873161852248517, 1, 1], 
      Hue[0.9873161852248517, 1, 1], 
      Hue[0.9870747873184581, 1, 1], 
      Hue[0.9870747873184581, 1, 1], 
      Hue[0.9868333467373395, 1, 1], 
      Hue[0.9868333467373395, 1, 1], 
      Hue[0.9863503343599933, 1, 1], 
      Hue[0.9863503343599933, 1, 1], 
      Hue[0.9853837623986506, 1, 1], 
      Hue[0.9853837623986506, 1, 1], 
      Hue[0.9834482507269322, 1, 1], 
      Hue[0.9834482507269322, 1, 1], 
      Hue[0.9832060742017218, 1, 1], 
      Hue[0.9832060742017218, 1, 1], 
      Hue[0.9829638422167695, 1, 1], 
      Hue[0.9829638422167695, 1, 1], 
      Hue[0.9824792086644553, 1, 1], 
      Hue[0.9824792086644553, 1, 1], 
      Hue[0.9815092407958094, 1, 1], 
      Hue[0.9815092407958094, 1, 1], 
      Hue[0.9795663223392861, 1, 1], 
      Hue[0.9795663223392861, 1, 1], 
      Hue[0.9793231622450247, 1, 1], 
      Hue[0.9793231622450247, 1, 1], 
      Hue[0.9790799338521377, 1, 1], 
      Hue[0.9790799338521377, 1, 1], 
      Hue[0.9785932689524265, 1, 1], 
      Hue[0.9785932689524265, 1, 1], 
      Hue[0.9776190841566166, 1, 1], 
      Hue[0.9776190841566166, 1, 1], 
      Hue[0.975667114035326, 1, 1], 
      Hue[0.975667114035326, 1, 1], 
      Hue[0.9754390633641128, 1, 1], 
      Hue[0.9754390633641128, 1, 1], 
      Hue[0.9752109420085299, 1, 1], 
      Hue[0.9752109420085299, 1, 1], 
      Hue[0.9747544846142472, 1, 1], 
      Hue[0.9747544846142472, 1, 1], 
      Hue[0.9738406926710507, 1, 1], 
      Hue[0.9738406926710507, 1, 1], 
      Hue[0.9720094526039389, 1, 1], 
      Hue[0.9720094526039389, 1, 1], 
      Hue[0.9683311642017814, 1, 1], 
      Hue[0.9683311642017814, 1, 1], 
      Hue[0.9681005244293676, 1, 1], 
      Hue[0.9681005244293676, 1, 1], 
      Hue[0.9678697928221847, 1, 1], 
      Hue[0.9678697928221847, 1, 1], 
      Hue[0.9674080514439187, 1, 1], 
      Hue[0.9674080514439187, 1, 1], 
      Hue[0.966483437400964, 1, 1], 
      Hue[0.966483437400964, 1, 1], 
      Hue[0.9646295349167414, 1, 1], 
      Hue[0.9646295349167414, 1, 1], 
      Hue[0.9609018350177954, 1, 1], 
      Hue[0.9609018350177954, 1, 1], 
      Hue[0.9606483503513359, 1, 1], 
      Hue[0.9606483503513359, 1, 1], 
      Hue[0.9603947326474463, 1, 1], 
      Hue[0.9603947326474463, 1, 1], 
      Hue[0.9598870947025527, 1, 1], 
      Hue[0.9598870947025527, 1, 1], 
      Hue[0.9588701846718123, 1, 1], 
      Hue[0.9588701846718123, 1, 1], 
      Hue[0.9568296358286239, 1, 1], 
      Hue[0.9568296358286239, 1, 1], 
      Hue[0.9527200809416663, 1, 1], 
      Hue[0.9527200809416663, 1, 1], 
      Hue[0.9443747422042437, 1, 1], 
      Hue[0.9443747422042437, 1, 1], 
      Hue[0.9441158128316757, 1, 1], 
      Hue[0.9441158128316757, 1, 1], 
      Hue[0.943856701855986, 1, 1], 
      Hue[0.943856701855986, 1, 1], 
      Hue[0.9433379317912788, 1, 1], 
      Hue[0.9433379317912788, 1, 1], 
      Hue[0.942298176074474, 1, 1], 
      Hue[0.942298176074474, 1, 1], 
      Hue[0.9402096171393628, 1, 1], 
      Hue[0.9402096171393628, 1, 1], 
      Hue[0.9359948273114629, 1, 1], 
      Hue[0.9359948273114629, 1, 1], 
      Hue[0.9274027124575679, 1, 1], 
      Hue[0.9274027124575679, 1, 1], 
      Hue[0.9107129293179048, 1, 1], 
      Hue[0.9107129293179048, 1, 1], 
      Hue[0.8914807352640294, 1, 1], 
      Hue[0.8914807352640294, 1, 1], 
      Hue[0.8722988598223519, 1, 1], 
      Hue[0.8722988598223519, 1, 1], 
      Hue[0.8522396379588035, 1, 1], 
      Hue[0.8522396379588035, 1, 1], 
      Hue[0.8518885980552959, 1, 1], 
      Hue[0.8518885980552959, 1, 1], 
      Hue[0.8515371912952256, 1, 1], 
      Hue[0.8515371912952256, 1, 1], 
      Hue[0.8508332776347656, 1, 1], 
      Hue[0.8508332776347656, 1, 1], 
      Hue[0.8494210532542403, 1, 1], 
      Hue[0.8494210532542403, 1, 1], 
      Hue[0.8465790518299141, 1, 1], 
      Hue[0.8465790518299141, 1, 1], 
      Hue[0.8408252475632318, 1, 1], 
      Hue[0.8408252475632318, 1, 1], 
      Hue[0.8290438106889065, 1, 1], 
      Hue[0.8290438106889065, 1, 1], 
      Hue[0.8286949133355983, 1, 1], 
      Hue[0.8286949133355983, 1, 1], 
      Hue[0.8283457193462858, 1, 1], 
      Hue[0.8283457193462858, 1, 1], 
      Hue[0.8276464439003608, 1, 1], 
      Hue[0.8276464439003608, 1, 1], 
      Hue[0.8262443606965498, 1, 1], 
      Hue[0.8262443606965498, 1, 1], 
      Hue[0.8234262126203606, 1, 1], 
      Hue[0.8234262126203606, 1, 1], 
      Hue[0.8177352853514953, 1, 1], 
      Hue[0.8177352853514953, 1, 1], 
      Hue[0.8173772394635447, 1, 1], 
      Hue[0.8173772394635447, 1, 1], 
      Hue[0.8170189206330951, 1, 1], 
      Hue[0.8170189206330951, 1, 1], 
      Hue[0.8163014676846723, 1, 1], 
      Hue[0.8163014676846723, 1, 1], 
      Hue[0.8148633259038316, 1, 1], 
      Hue[0.8148633259038316, 1, 1], 
      Hue[0.8119743082263935, 1, 1], 
      Hue[0.8119743082263935, 1, 1], 
      Hue[0.8061471281010334, 1, 1], 
      Hue[0.8061471281010334, 1, 1], 
      Hue[0.8057500829027806, 1, 1], 
      Hue[0.8057500829027806, 1, 1], 
      Hue[0.8053527553799907, 1, 1], 
      Hue[0.8053527553799907, 1, 1], 
      Hue[0.8045572592770968, 1, 1], 
      Hue[0.8045572592770968, 1, 1], 
      Hue[0.8029629449050588, 1, 1], 
      Hue[0.8029629449050588, 1, 1], 
      Hue[0.7997613736541621, 1, 1], 
      Hue[0.7997613736541621, 1, 1], 
      Hue[0.799359994854902, 1, 1], 
      Hue[0.799359994854902, 1, 1], 
      Hue[0.7989583587267372, 1, 1], 
      Hue[0.7989583587267372, 1, 1], 
      Hue[0.7981543211276124, 1, 1], 
      Hue[0.7981543211276124, 1, 1], 
      Hue[0.7965432316826655, 1, 1], 
      Hue[0.7965432316826655, 1, 1], 
      Hue[0.7933093819342499, 1, 1], 
      Hue[0.7933093819342499, 1, 1], 
      Hue[0.7929040909701478, 1, 1], 
      Hue[0.7929040909701478, 1, 1], 
      Hue[0.7924985705100261, 1, 1], 
      Hue[0.7924985705100261, 1, 1], 
      Hue[0.7916868484298172, 1, 1], 
      Hue[0.7916868484298172, 1, 1], 
      Hue[0.7900607314967867, 1, 1], 
      Hue[0.7900607314967867, 1, 1], 
      Hue[0.7867982298702154, 1, 1], 
      Hue[0.7867982298702154, 1, 1], 
      Hue[0.7802356708666068, 1, 1], 
      Hue[0.7802356708666068, 1, 1], 
      Hue[0.7802233700537657, 1, 1], 
      Hue[0.7802233700537657, 1, 1], 
      Hue[0.7802110690906503, 1, 1], 
      Hue[0.7802110690906503, 1, 1], 
      Hue[0.7801864667138221, 1, 1], 
      Hue[0.7801864667138221, 1, 1], 
      Hue[0.7801372601593551, 1, 1], 
      Hue[0.7801372601593551, 1, 1], 
      Hue[0.7800388398598123, 1, 1], 
      Hue[0.7800388398598123, 1, 1], 
      Hue[0.7798419705994404, 1, 1], 
      Hue[0.7798419705994404, 1, 1], 
      Hue[0.7798296650058097, 1, 1], 
      Hue[0.7798296650058097, 1, 1], 
      Hue[0.7798173592637114, 1, 1], 
      Hue[0.7798173592637114, 1, 1], 
      Hue[0.7797927473343383, 1, 1], 
      Hue[0.7797927473343383, 1, 1], 
      Hue[0.7797435216964705, 1, 1], 
      Hue[0.7797435216964705, 1, 1], 
      Hue[0.7796450633169278, 1, 1], 
      Hue[0.7796450633169278, 1, 1], 
      Hue[0.7796327553546072, 1, 1], 
      Hue[0.7796327553546072, 1, 1], 
      Hue[0.7796204472447248, 1, 1], 
      Hue[0.7796204472447248, 1, 1], 
      Hue[0.7795958305825013, 1, 1], 
      Hue[0.7795958305825013, 1, 1], 
      Hue[0.7795465954898065, 1, 1], 
      Hue[0.7795465954898065, 1, 1], 
      Hue[0.7795342863485786, 1, 1], 
      Hue[0.7795342863485786, 1, 1], 
      Hue[0.7795219770602422, 1, 1], 
      Hue[0.7795219770602422, 1, 1], 
      Hue[0.7794973580424717, 1, 1], 
      Hue[0.7794973580424717, 1, 1], 
      Hue[0.7794850483131508, 1, 1], 
      Hue[0.7794850483131508, 1, 1], 
      Hue[0.7794727384369486, 1, 1], 
      Hue[0.7794727384369486, 1, 1], 
      Hue[0.7794604284139217, 1, 1], 
      Hue[0.7794604284139217, 1, 1], 
      Hue[0.779448118244127, 1, 1]}], GraphicsComplexBox[CompressedData["
1:eJx13HlYTH3YB/CKmlRIiQoV2QrxEEXlHmTfdxGyRkhISip7RMmSytJCEpKl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    "], {
     {GrayLevel[0.5], EdgeForm[None], 
      GraphicsGroupBox[PolygonBox[CompressedData["
1:eJw12nfAT+X/x/HrfT7D3jshMyub7B2RmSQkIUmSJKkkLUmykiSJkiRJkiQh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        "]]]}, {{}, {}, {
       {Hue[0.4827891120223487, 1, 1], LineBox[{1, 2}]}, 
       {Hue[0.482564154597886, 1, 1], LineBox[{2, 3}]}, 
       {Hue[0.4823391475506281, 1, 1], LineBox[{3, 4}]}, 
       {Hue[0.4818889820233611, 1, 1], LineBox[{4, 5}]}, 
       {Hue[0.48098802727801265`, 1, 1], LineBox[{5, 6}]}, 
       {Hue[0.4791834792023881, 1, 1], LineBox[{6, 7}]}, 
       {Hue[0.47895765079255526`, 1, 1], LineBox[{7, 8}]}, 
       {Hue[0.47873176248257115`, 1, 1], LineBox[{8, 9}]}, 
       {Hue[0.4782798035865052, 1, 1], LineBox[{9, 10}]}, 
       {Hue[0.4773751386497867, 1, 1], LineBox[{10, 11}]}, 
       {Hue[0.4755626757123231, 1, 1], LineBox[{11, 12}]}, 
       {Hue[0.47533581152777804`, 1, 1], LineBox[{12, 13}]}, 
       {Hue[0.47510887711774313`, 1, 1], LineBox[{13, 14}]}, 
       {Hue[0.47465479503274677`, 1, 1], LineBox[{14, 15}]}, 
       {Hue[0.4737457596719586, 1, 1], LineBox[{15, 16}]}, 
       {Hue[0.4719240589571304, 1, 1], LineBox[{16, 17}]}, 
       {Hue[0.46826497272567535`, 1, 1], LineBox[{17, 18}]}, 
       {Hue[0.46801623645615187`, 1, 1], LineBox[{18, 19}]}, 
       {Hue[0.4677673931206509, 1, 1], LineBox[{19, 20}]}, 
       {Hue[0.4672693819164649, 1, 1], LineBox[{20, 21}]}, 
       {Hue[0.46627203801009476`, 1, 1], LineBox[{21, 22}]}, 
       {Hue[0.4642718770043928, 1, 1], LineBox[{22, 23}]}, 
       {Hue[0.4602481601269848, 1, 1], LineBox[{23, 24}]}, 
       {Hue[0.4599955825663628, 1, 1], LineBox[{24, 25}]}, 
       {Hue[0.45974287109529516`, 1, 1], LineBox[{25, 26}]}, 
       {Hue[0.45923704304401985`, 1, 1], LineBox[{26, 27}]}, 
       {Hue[0.45822374284207923`, 1, 1], LineBox[{27, 28}]}, 
       {Hue[0.4561903764679997, 1, 1], LineBox[{28, 29}]}, 
       {Hue[0.45209505906550335`, 1, 1], LineBox[{29, 30}]}, 
       {Hue[0.4518548306823689, 1, 1], LineBox[{30, 31}]}, 
       {Hue[0.4516144619060803, 1, 1], LineBox[{31, 32}]}, 
       {Hue[0.45113330039310223`, 1, 1], LineBox[{32, 33}]}, 
       {Hue[0.45016926204801466`, 1, 1], LineBox[{33, 34}]}, 
       {Hue[0.44823416810320266`, 1, 1], LineBox[{34, 35}]}, 
       {Hue[0.44433466107445957`, 1, 1], LineBox[{35, 36}]}, 
       {Hue[0.43640834183084554`, 1, 1], LineBox[{36, 37}]}, 
       {Hue[0.42029717691584667`, 1, 1], LineBox[{37, 38}]}, 
       {Hue[0.40180141253448876`, 1, 1], LineBox[{38, 39}]}, 
       {Hue[0.3834099295601948, 1, 1], LineBox[{39, 40}]}, 
       {Hue[0.36209016028713203`, 1, 1], LineBox[{40, 41}]}, 
       {Hue[0.3396709561709671, 1, 1], LineBox[{41, 42}]}, 
       {Hue[0.3393332954116793, 1, 1], LineBox[{42, 43}]}, 
       {Hue[0.3389953266080553, 1, 1], LineBox[{43, 44}]}, 
       {Hue[0.3383184662821582, 1, 1], LineBox[{44, 45}]}, 
       {Hue[0.3369610650839929, 1, 1], LineBox[{45, 46}]}, 
       {Hue[0.33423162960793346`, 1, 1], LineBox[{46, 47}]}, 
       {Hue[0.32871504472731383`, 1, 1], LineBox[{47, 48}]}, 
       {Hue[0.3283677395765662, 1, 1], LineBox[{48, 49}]}, 
       {Hue[0.3280201415235637, 1, 1], LineBox[{49, 50}]}, 
       {Hue[0.3273240691434086, 1, 1], LineBox[{50, 51}]}, 
       {Hue[0.3259284367777426, 1, 1], LineBox[{51, 52}]}, 
       {Hue[0.3231233685349689, 1, 1], LineBox[{52, 53}]}, 
       {Hue[0.3174593058882333, 1, 1], LineBox[{53, 54}]}, 
       {Hue[0.3170728973764113, 1, 1], LineBox[{54, 55}]}, 
       {Hue[0.31668617223125795`, 1, 1], LineBox[{55, 56}]}, 
       {Hue[0.3159117765219359, 1, 1], LineBox[{56, 57}]}, 
       {Hue[0.31435923545210387`, 1, 1], LineBox[{57, 58}]}, 
       {Hue[0.31123942077030387`, 1, 1], LineBox[{58, 59}]}, 
       {Hue[0.3049431469648374, 1, 1], LineBox[{59, 60}]}, 
       {Hue[0.3045472184743325, 1, 1], LineBox[{60, 61}]}, 
       {Hue[0.304151014900581, 1, 1], LineBox[{61, 62}]}, 
       {Hue[0.3033577884702304, 1, 1], LineBox[{62, 63}]}, 
       {Hue[0.3017681008777704, 1, 1], LineBox[{63, 64}]}, 
       {Hue[0.2985761353478784, 1, 1], LineBox[{64, 65}]}, 
       {Hue[0.2921447786295377, 1, 1], LineBox[{65, 66}]}, 
       {Hue[0.29176768769392497`, 1, 1], LineBox[{66, 67}]}, 
       {Hue[0.29139040316372905`, 1, 1], LineBox[{67, 68}]}, 
       {Hue[0.2906352592835539, 1, 1], LineBox[{68, 69}]}, 
       {Hue[0.28912271440992654`, 1, 1], LineBox[{69, 70}]}, 
       {Hue[0.28608893968771626`, 1, 1], LineBox[{70, 71}]}, 
       {Hue[0.285708934056643, 1, 1], LineBox[{71, 72}]}, 
       {Hue[0.285328759532689, 1, 1], LineBox[{72, 73}]}, 
       {Hue[0.28456791022786804`, 1, 1], LineBox[{73, 74}]}, 
       {Hue[0.28304425591740834`, 1, 1], LineBox[{74, 75}]}, 
       {Hue[0.279989492581923, 1, 1], LineBox[{75, 76}]}, 
       {Hue[0.27960698079228, 1, 1], LineBox[{76, 77}]}, 
       {Hue[0.27922432654828644`, 1, 1], LineBox[{77, 78}]}, 
       {Hue[0.27845859752293123`, 1, 1], LineBox[{78, 79}]}, 
       {Hue[0.27692550542522587`, 1, 1], LineBox[{79, 80}]}, 
       {Hue[0.27654190206159385`, 1, 1], LineBox[{80, 81}]}, 
       {Hue[0.27615817004289983`, 1, 1], LineBox[{81, 82}]}, 
       {Hue[0.27539032704479266`, 1, 1], LineBox[{82, 83}]}, 
       {Hue[0.27385317453066826`, 1, 1], LineBox[{83, 84}]}, 
       {Hue[0.2734685912133791, 1, 1], LineBox[{84, 85}]}, 
       {Hue[0.27308389338429445`, 1, 1], LineBox[{85, 86}]}, 
       {Hue[0.27231416135705505`, 1, 1], LineBox[{86, 87}]}, 
       {Hue[0.2707734022533922, 1, 1], LineBox[{87, 88}]}, 
       {Hue[0.2676871110167407, 1, 1], LineBox[{88, 89}]}, 
       {Hue[0.26730851114187276`, 1, 1], LineBox[{89, 90}]}, 
       {Hue[0.2669298292222692, 1, 1], LineBox[{90, 91}]}, 
       {Hue[0.2661722262534576, 1, 1], LineBox[{91, 92}]}, 
       {Hue[0.264656113011379, 1, 1], LineBox[{92, 93}]}, 
       {Hue[0.26427690596839526`, 1, 1], LineBox[{93, 94}]}, 
       {Hue[0.2638976309721529, 1, 1], LineBox[{94, 95}]}, 
       {Hue[0.26313888422214005`, 1, 1], LineBox[{95, 96}]}, 
       {Hue[0.26162065355766034`, 1, 1], LineBox[{96, 97}]}, 
       {Hue[0.2612409527340028, 1, 1], LineBox[{97, 98}]}, 
       {Hue[0.26086119822901244`, 1, 1], LineBox[{98, 99}]}, 
       {Hue[0.2601015353565505, 1, 1], LineBox[{99, 100}]}, 
       {Hue[0.2597216305851791, 1, 1], LineBox[{100, 101}]}, 
       {Hue[0.25934167932452723`, 1, 1], LineBox[{101, 102}]}, 
       {Hue[0.2585816445494831, 1, 1], LineBox[{102, 103}]}, 
       {Hue[0.25820156464674043`, 1, 1], LineBox[{103, 104}]}, 
       {Hue[0.25782144547786634`, 1, 1], LineBox[{104, 105}]}, 
       {Hue[0.25706109658363874`, 1, 1], LineBox[{105, 106}]}, 
       {Hue[0.25668087048309157`, 1, 1], LineBox[{106, 107}]}, 
       {Hue[0.25630061236587415`, 1, 1], LineBox[{107, 108}]}, 
       {Hue[0.2555400073463437, 1, 1], LineBox[{108, 109}]}, 
       {Hue[0.2551596640795802, 1, 1], LineBox[{109, 110}]}, 
       {Hue[0.25477929606709404`, 1, 1], LineBox[{110, 111}]}, 
       {Hue[0.2543989051294105, 1, 1], LineBox[{111, 112}]}, 
       {Hue[0.2540184930880174, 1, 1], LineBox[{112, 113}]}, 
       {Hue[0.25363806176528997`, 1, 1], LineBox[{113, 114}]}, 
       {Hue[0.2532576129844135, 1, 1], LineBox[{114, 115}]}, 
       {Hue[0.2528771485693071, 1, 1], LineBox[{115, 116}]}, 
       {Hue[0.2524966703445472, 1, 1], LineBox[{116, 117}]}, 
       {Hue[0.25211618013529086`, 1, 1], LineBox[{117, 118}]}, 
       {Hue[0.25173567976719907`, 1, 1], LineBox[{118, 119}]}, 
       {Hue[0.2513551710663605, 1, 1], LineBox[{119, 120}]}, 
       {Hue[0.25097465585921375`, 1, 1], LineBox[{120, 121}]}, 
       {Hue[0.2505941359724715, 1, 1], LineBox[{121, 122}]}, 
       {Hue[0.25021361323304303`, 1, 1], LineBox[{122, 123}]}, 
       {Hue[0.24983308946795788`, 1, 1], LineBox[{123, 124}]}, 
       {Hue[0.2494525665042888, 1, 1], LineBox[{124, 125}]}, 
       {Hue[0.2490720461690742, 1, 1], LineBox[{125, 126}]}, 
       {Hue[0.2486915302892427, 1, 1], LineBox[{126, 127}]}, 
       {Hue[0.2483110206915351, 1, 1], LineBox[{127, 128}]}, 
       {Hue[0.2479305192024281, 1, 1], LineBox[{128, 129}]}, 
       {Hue[0.24755002764805686`, 1, 1], LineBox[{129, 130}]}, 
       {Hue[0.24716954785413933`, 1, 1], LineBox[{130, 131}]}, 
       {Hue[0.24678908164589852`, 1, 1], LineBox[{131, 132}]}, 
       {Hue[0.24640863084798625`, 1, 1], LineBox[{132, 133}]}, 
       {Hue[0.24602819728440656`, 1, 1], LineBox[{133, 134}]}, 
       {Hue[0.2456477827784387, 1, 1], LineBox[{134, 135}]}, 
       {Hue[0.24526738915256116`, 1, 1], LineBox[{135, 136}]}, 
       {Hue[0.24488701822837544`, 1, 1], LineBox[{136, 137}]}, 
       {Hue[0.24450667182652894`, 1, 1], LineBox[{137, 138}]}, 
       {Hue[0.24412635176663902`, 1, 1], LineBox[{138, 139}]}, 
       {Hue[0.243365797945593, 1, 1], LineBox[{139, 140}]}, 
       {Hue[0.24295324765529228`, 1, 1], LineBox[{140, 141}]}, 
       {Hue[0.2425407371117956, 1, 1], LineBox[{141, 142}]}, 
       {Hue[0.24171584452690567`, 1, 1], LineBox[{142, 143}]}, 
       {Hue[0.24130346711105533`, 1, 1], LineBox[{143, 144}]}, 
       {Hue[0.2408911386928689, 1, 1], LineBox[{144, 145}]}, 
       {Hue[0.24006663807480177`, 1, 1], LineBox[{145, 146}]}, 
       {Hue[0.23965447048114782`, 1, 1], LineBox[{146, 147}]}, 
       {Hue[0.2392423610973872, 1, 1], LineBox[{147, 148}]}, 
       {Hue[0.23841832614129013`, 1, 1], LineBox[{148, 149}]}, 
       {Hue[0.23677105557424935`, 1, 1], LineBox[{149, 150}]}, 
       {Hue[0.2363594177784766, 1, 1], LineBox[{150, 151}]}, 
       {Hue[0.23594785647257394`, 1, 1], LineBox[{151, 152}]}, 
       {Hue[0.23512497240373834`, 1, 1], LineBox[{152, 153}]}, 
       {Hue[0.23348022172980518`, 1, 1], LineBox[{153, 154}]}, 
       {Hue[0.23306925917031257`, 1, 1], LineBox[{154, 155}]}, 
       {Hue[0.2326583911412925, 1, 1], LineBox[{155, 156}]}, 
       {Hue[0.23183694761200105`, 1, 1], LineBox[{156, 157}]}, 
       {Hue[0.23019529296078733`, 1, 1], LineBox[{157, 158}]}, 
       {Hue[0.22978514896890515`, 1, 1], LineBox[{158, 159}]}, 
       {Hue[0.2293751172542093, 1, 1], LineBox[{159, 160}]}, 
       {Hue[0.22855539943102374`, 1, 1], LineBox[{160, 161}]}, 
       {Hue[0.22691740731771895`, 1, 1], LineBox[{161, 162}]}, 
       {Hue[0.2236476811129669, 1, 1], LineBox[{162, 163}]}, 
       {Hue[0.22323959369566776`, 1, 1], LineBox[{163, 164}]}, 
       {Hue[0.2228316529602289, 1, 1], LineBox[{164, 165}]}, 
       {Hue[0.2220162199090814, 1, 1], LineBox[{165, 166}]}, 
       {Hue[0.22038720570674894`, 1, 1], LineBox[{166, 167}]}, 
       {Hue[0.21713704446893195`, 1, 1], LineBox[{167, 168}]}, 
       {Hue[0.21675867941709884`, 1, 1], LineBox[{168, 169}]}, 
       {Hue[0.2163804703713593, 1, 1], LineBox[{169, 170}]}, 
       {Hue[0.21562452670544383`, 1, 1], LineBox[{170, 171}]}, 
       {Hue[0.2141145815706027, 1, 1], LineBox[{171, 172}]}, 
       {Hue[0.2111028105643298, 1, 1], LineBox[{172, 173}]}, 
       {Hue[0.20511445765265002`, 1, 1], LineBox[{173, 174}]}, 
       {Hue[0.2047418469969295, 1, 1], LineBox[{174, 175}]}, 
       {Hue[0.20436944028448414`, 1, 1], LineBox[{175, 176}]}, 
       {Hue[0.2036252441940914, 1, 1], LineBox[{176, 177}]}, 
       {Hue[0.2021393594502079, 1, 1], LineBox[{177, 178}]}, 
       {Hue[0.19917791790822034`, 1, 1], LineBox[{178, 179}]}, 
       {Hue[0.19329862431015446`, 1, 1], LineBox[{179, 180}]}, 
       {Hue[0.19290251120305676`, 1, 1], LineBox[{180, 181}]}, 
       {Hue[0.19250668498575932`, 1, 1], LineBox[{181, 182}]}, 
       {Hue[0.19171589889733506`, 1, 1], LineBox[{182, 183}]}, 
       {Hue[0.19013783118538485`, 1, 1], LineBox[{183, 184}]}, 
       {Hue[0.18699601635630125`, 1, 1], LineBox[{184, 185}]}, 
       {Hue[0.18077193135702532`, 1, 1], LineBox[{185, 186}]}, 
       {Hue[0.1685774743268824, 1, 1], LineBox[{186, 187}]}, 
       {Hue[0.1682089558570103, 1, 1], LineBox[{187, 188}]}, 
       {Hue[0.1678407785456452, 1, 1], LineBox[{188, 189}]}, 
       {Hue[0.16710544991494183`, 1, 1], LineBox[{189, 190}]}, 
       {Hue[0.16563891364948222`, 1, 1], LineBox[{190, 191}]}, 
       {Hue[0.16272245258452597`, 1, 1], LineBox[{191, 192}]}, 
       {Hue[0.15695685884954436`, 1, 1], LineBox[{192, 193}]}, 
       {Hue[0.14569993963152494`, 1, 1], LineBox[{193, 194}]}, 
       {Hue[0.1256893641497182, 1, 1], LineBox[{194, 195}]}, 
       {Hue[0.10539044861568864`, 1, 1], LineBox[{195, 196}]}, 
       {Hue[0.0876450931568508, 1, 1], LineBox[{196, 197}]}, 
       {Hue[0.06954252802227788, 1, 1], LineBox[{197, 198}]}, 
       {Hue[0.06927284345491812, 1, 1], LineBox[{198, 199}]}, 
       {Hue[0.06900338285499233, 1, 1], LineBox[{199, 200}]}, 
       {Hue[0.06846513025460943, 1, 1], LineBox[{200, 201}]}, 
       {Hue[0.06739127631685438, 1, 1], LineBox[{201, 202}]}, 
       {Hue[0.06525398821771008, 1, 1], LineBox[{202, 203}]}, 
       {Hue[0.061019606605566934`, 1, 1], LineBox[{203, 204}]}, 
       {Hue[0.060756678464624916`, 1, 1], LineBox[{204, 205}]}, 
       {Hue[0.06049394778226548, 1, 1], LineBox[{205, 206}]}, 
       {Hue[0.05996907547438736, 1, 1], LineBox[{206, 207}]}, 
       {Hue[0.05892166385474366, 1, 1], LineBox[{207, 208}]}, 
       {Hue[0.05683598678500967, 1, 1], LineBox[{208, 209}]}, 
       {Hue[0.05269973189924155, 1, 1], LineBox[{209, 210}]}, 
       {Hue[0.052459950972470006`, 1, 1], LineBox[{210, 211}]}, 
       {Hue[0.05222031886942626, 1, 1], LineBox[{211, 212}]}, 
       {Hue[0.05174149845122005, 1, 1], LineBox[{212, 213}]}, 
       {Hue[0.05078561399160116, 1, 1], LineBox[{213, 214}]}, 
       {Hue[0.048880720514997804`, 1, 1], LineBox[{214, 215}]}, 
       {Hue[0.045097237160123156`, 1, 1], LineBox[{215, 216}]}, 
       {Hue[0.04486188666864251, 1, 1], LineBox[{216, 217}]}, 
       {Hue[0.0446266636188734, 1, 1], LineBox[{217, 218}]}, 
       {Hue[0.04415659718485805, 1, 1], LineBox[{218, 219}]}, 
       {Hue[0.04321796437381417, 1, 1], LineBox[{219, 220}]}, 
       {Hue[0.041346550324896265`, 1, 1], LineBox[{220, 221}]}, 
       {Hue[0.03762594234630728, 1, 1], LineBox[{221, 222}]}, 
       {Hue[0.0373748435349179, 1, 1], LineBox[{222, 223}]}, 
       {Hue[0.0371238695861517, 1, 1], LineBox[{223, 224}]}, 
       {Hue[0.03662229292476846, 1, 1], LineBox[{224, 225}]}, 
       {Hue[0.03562060110319645, 1, 1], LineBox[{225, 226}]}, 
       {Hue[0.033622876197495356`, 1, 1], LineBox[{226, 227}]}, 
       {Hue[0.033373674812982916`, 1, 1], LineBox[{227, 228}]}, 
       {Hue[0.03312458492284089, 1, 1], LineBox[{228, 229}]}, 
       {Hue[0.032626736295236373`, 1, 1], LineBox[{229, 230}]}, 
       {Hue[0.03163234035619638, 1, 1], LineBox[{230, 231}]}, 
       {Hue[0.02964856766170453, 1, 1], LineBox[{231, 232}]}, 
       {Hue[0.029401050448112517`, 1, 1], LineBox[{232, 233}]}, 
       {Hue[0.029153631445108612`, 1, 1], LineBox[{233, 234}]}, 
       {Hue[0.02865908476120701, 1, 1], LineBox[{234, 235}]}, 
       {Hue[0.027671133531875818`, 1, 1], LineBox[{235, 236}]}, 
       {Hue[0.025699614688751985`, 1, 1], LineBox[{236, 237}]}, 
       {Hue[0.021772636012389598`, 1, 1], LineBox[{237, 238}]}, 
       {Hue[0.021544088339672494`, 1, 1], LineBox[{238, 239}]}, 
       {Hue[0.021315603374490812`, 1, 1], LineBox[{239, 240}]}, 
       {Hue[0.02085881890328848, 1, 1], LineBox[{240, 241}]}, 
       {Hue[0.019945973164421232`, 1, 1], LineBox[{241, 242}]}, 
       {Hue[0.018123025596737827`, 1, 1], LineBox[{242, 243}]}, 
       {Hue[0.01789540144331209, 1, 1], LineBox[{243, 244}]}, 
       {Hue[0.01766782936571196, 1, 1], LineBox[{244, 245}]}, 
       {Hue[0.017212838786401, 1, 1], LineBox[{245, 246}]}, 
       {Hue[0.0163034533796867, 1, 1], LineBox[{246, 247}]}, 
       {Hue[0.01448691729759583, 1, 1], LineBox[{247, 248}]}, 
       {Hue[0.014260046896887183`, 1, 1], LineBox[{248, 249}]}, 
       {Hue[0.014033217984415945`, 1, 1], LineBox[{249, 250}]}, 
       {Hue[0.013579681982588002`, 1, 1], LineBox[{250, 251}]}, 
       {Hue[0.012673078787923713`, 1, 1], LineBox[{251, 252}]}, 
       {Hue[0.012446521808640131`, 1, 1], LineBox[{252, 253}]}, 
       {Hue[0.012220001037920017`, 1, 1], LineBox[{253, 254}]}, 
       {Hue[0.011767065484800906`, 1, 1], LineBox[{254, 255}]}, 
       {Hue[0.010861599877047007`, 1, 1], LineBox[{255, 256}]}, 
       {Hue[0.010635314110578184`, 1, 1], LineBox[{256, 257}]}, 
       {Hue[0.010409059281020351`, 1, 1], LineBox[{257, 258}]}, 
       {Hue[0.009956639798957091, 1, 1], LineBox[{258, 259}]}, 
       {Hue[0.009052143113095098, 1, 1], LineBox[{259, 260}]}, 
       {Hue[0.008826086411195665, 1, 1], LineBox[{260, 261}]}, 
       {Hue[0.008600055381481772, 1, 1], LineBox[{261, 262}]}, 
       {Hue[0.008148067708061165, 1, 1], LineBox[{262, 263}]}, 
       {Hue[0.007244371496228025, 1, 1], LineBox[{263, 264}]}, 
       {Hue[0.007022937079816843, 1, 1], LineBox[{264, 265}]}, 
       {Hue[0.006801522295207302, 1, 1], LineBox[{265, 266}]}, 
       {Hue[0.0063587491451943855`, 1, 1], LineBox[{266, 267}]}, 
       {Hue[0.006137389541852155, 1, 1], LineBox[{267, 268}]}, 
       {Hue[0.0059160470944297034`, 1, 1], LineBox[{268, 269}]}, 
       {Hue[0.005473411192127852, 1, 1], LineBox[{269, 270}]}, 
       {Hue[0.0052521164997800344`, 1, 1], LineBox[{270, 271}]}, 
       {Hue[0.005030836488410986, 1, 1], LineBox[{271, 272}]}, 
       {Hue[0.00458831803424194, 1, 1], LineBox[{272, 273}]}, 
       {Hue[0.004367078354377108, 1, 1], LineBox[{273, 274}]}, 
       {Hue[0.004145850881357151, 1, 1], LineBox[{274, 275}]}, 
       {Hue[0.0037034300821993217`, 1, 1], LineBox[{275, 276}]}, 
       {Hue[0.0034822355193328733`, 1, 1], LineBox[{276, 277}]}, 
       {Hue[0.0032610506898499213`, 1, 1], LineBox[{277, 278}]}, 
       {Hue[0.003039874975474352, 1, 1], LineBox[{278, 279}]}, 
       {Hue[0.002818707757961507, 1, 1], LineBox[{279, 280}]}, 
       {Hue[0.0025975484190960568`, 1, 1], LineBox[{280, 281}]}, 
       {Hue[0.0023763963406898754`, 1, 1], LineBox[{281, 282}]}, 
       {Hue[0.0021552509045799197`, 1, 1], LineBox[{282, 283}]}, 
       {Hue[0.0019341114926260908`, 1, 1], LineBox[{283, 284}]}, 
       {Hue[0.0017129774867091097`, 1, 1], LineBox[{284, 285}]}, 
       {Hue[0.0014918482687283844`, 1, 1], LineBox[{285, 286}]}, 
       {Hue[0.0012707232205998733`, 1, 1], LineBox[{286, 287}]}, 
       {Hue[0.001049601724253957, 1, 1], LineBox[{287, 288}]}, 
       {Hue[0.0008284831616332965, 1, 1], LineBox[{288, 289}]}, 
       {Hue[0.0006073669146907009, 1, 1], LineBox[{289, 290}]}, 
       {Hue[0.00038625236538699054`, 1, 1], LineBox[{290, 291}]}, 
       {Hue[0.00016513889568885723`, 1, 1], LineBox[{291, 292}]}, 
       {Hue[0.9999440258875667, 1, 1], LineBox[{292, 293}]}, 
       {Hue[0.9997229127229926, 1, 1], LineBox[{293, 294}]}, 
       {Hue[0.9995017987839381, 1, 1], LineBox[{294, 295}]}, 
       {Hue[0.9992806834523718, 1, 1], LineBox[{295, 296}]}, 
       {Hue[0.9990595661102578, 1, 1], LineBox[{296, 297}]}, 
       {Hue[0.9988384461395532, 1, 1], LineBox[{297, 298}]}, 
       {Hue[0.9986173229222058, 1, 1], LineBox[{298, 299}]}, 
       {Hue[0.9983961958401524, 1, 1], LineBox[{299, 300}]}, 
       {Hue[0.9981750642753163, 1, 1], LineBox[{300, 301}]}, 
       {Hue[0.9979539276096054, 1, 1], LineBox[{301, 302}]}, 
       {Hue[0.9977327852249099, 1, 1], LineBox[{302, 303}]}, 
       {Hue[0.9975116365031003, 1, 1], LineBox[{303, 304}]}, 
       {Hue[0.997290480826025, 1, 1], LineBox[{304, 305}]}, 
       {Hue[0.9970693175755087, 1, 1], LineBox[{305, 306}]}, 
       {Hue[0.9966269658813178, 1, 1], LineBox[{306, 307}]}, 
       {Hue[0.9964057762011531, 1, 1], LineBox[{307, 308}]}, 
       {Hue[0.9961845764745626, 1, 1], LineBox[{308, 309}]}, 
       {Hue[0.9957421444087574, 1, 1], LineBox[{309, 310}]}, 
       {Hue[0.9955209108327753, 1, 1], LineBox[{310, 311}]}, 
       {Hue[0.9952996647368284, 1, 1], LineBox[{311, 312}]}, 
       {Hue[0.9948571325110459, 1, 1], LineBox[{312, 313}]}, 
       {Hue[0.9946358451440981, 1, 1], LineBox[{313, 314}]}, 
       {Hue[0.9944145427829568, 1, 1], LineBox[{314, 315}]}, 
       {Hue[0.9939718906033176, 1, 1], LineBox[{315, 316}]}, 
       {Hue[0.9930863790880589, 1, 1], LineBox[{316, 317}]}, 
       {Hue[0.9928461651955212, 1, 1], LineBox[{317, 318}]}, 
       {Hue[0.9926059277675018, 1, 1], LineBox[{318, 319}]}, 
       {Hue[0.9921253791427431, 1, 1], LineBox[{319, 320}]}, 
       {Hue[0.9911639646767813, 1, 1], LineBox[{320, 321}]}, 
       {Hue[0.9909235403594647, 1, 1], LineBox[{321, 322}]}, 
       {Hue[0.9906830861794046, 1, 1], LineBox[{322, 323}]}, 
       {Hue[0.9902020850652838, 1, 1], LineBox[{323, 324}]}, 
       {Hue[0.9892396896536435, 1, 1], LineBox[{324, 325}]}, 
       {Hue[0.9889990042677944, 1, 1], LineBox[{325, 326}]}, 
       {Hue[0.9887582826843183, 1, 1], LineBox[{326, 327}]}, 
       {Hue[0.9882767277544471, 1, 1], LineBox[{327, 328}]}, 
       {Hue[0.9873131486444305, 1, 1], LineBox[{328, 329}]}, 
       {Hue[0.9853839357017069, 1, 1], LineBox[{329, 330}]}, 
       {Hue[0.9851425759700883, 1, 1], LineBox[{330, 331}]}, 
       {Hue[0.9849011673421513, 1, 1], LineBox[{331, 332}]}, 
       {Hue[0.984418200216528, 1, 1], LineBox[{332, 333}]}, 
       {Hue[0.9834516442098782, 1, 1], LineBox[{333, 334}]}, 
       {Hue[0.9815158667939559, 1, 1], LineBox[{334, 335}]}, 
       {Hue[0.9812736291913834, 1, 1], LineBox[{335, 336}]}, 
       {Hue[0.9810313299456908, 1, 1], LineBox[{336, 337}]}, 
       {Hue[0.9805465433306404, 1, 1], LineBox[{337, 338}]}, 
       {Hue[0.9795761952326396, 1, 1], LineBox[{338, 339}]}, 
       {Hue[0.9776322203763465, 1, 1], LineBox[{339, 340}]}, 
       {Hue[0.9774051575452036, 1, 1], LineBox[{340, 341}]}, 
       {Hue[0.9771780299268937, 1, 1], LineBox[{341, 342}]}, 
       {Hue[0.9767235777199724, 1, 1], LineBox[{342, 343}]}, 
       {Hue[0.9758138671519275, 1, 1], LineBox[{343, 344}]}, 
       {Hue[0.9739910750405399, 1, 1], LineBox[{344, 345}]}, 
       {Hue[0.9703308331526667, 1, 1], LineBox[{345, 346}]}, 
       {Hue[0.9701013726173703, 1, 1], LineBox[{346, 347}]}, 
       {Hue[0.9698718263196648, 1, 1], LineBox[{347, 348}]}, 
       {Hue[0.9694124737999488, 1, 1], LineBox[{348, 349}]}, 
       {Hue[0.9684927105904569, 1, 1], LineBox[{349, 350}]}, 
       {Hue[0.9666488035074778, 1, 1], LineBox[{350, 351}]}, 
       {Hue[0.9629422795583775, 1, 1], LineBox[{351, 352}]}, 
       {Hue[0.96269025198017, 1, 1], LineBox[{352, 353}]}, 
       {Hue[0.9624380986753572, 1, 1], LineBox[{353, 354}]}, 
       {Hue[0.9619334114888105, 1, 1], LineBox[{354, 355}]}, 
       {Hue[0.9609224910093952, 1, 1], LineBox[{355, 356}]}, 
       {Hue[0.9588942749542262, 1, 1], LineBox[{356, 357}]}, 
       {Hue[0.9548108126517543, 1, 1], LineBox[{357, 358}]}, 
       {Hue[0.946523461974924, 1, 1], LineBox[{358, 359}]}, 
       {Hue[0.9462664435525957, 1, 1], LineBox[{359, 360}]}, 
       {Hue[0.946009250984037, 1, 1], LineBox[{360, 361}]}, 
       {Hue[0.9454943401255437, 1, 1], LineBox[{361, 362}]}, 
       {Hue[0.944462392535523, 1, 1], LineBox[{362, 363}]}, 
       {Hue[0.9423898098455793, 1, 1], LineBox[{363, 364}]}, 
       {Hue[0.9382084208650068, 1, 1], LineBox[{364, 365}]}, 
       {Hue[0.9296889548412226, 1, 1], LineBox[{365, 366}]}, 
       {Hue[0.9131577192100314, 1, 1], LineBox[{366, 367}]}, 
       {Hue[0.8941248259641966, 1, 1], LineBox[{367, 368}]}, 
       {Hue[0.8751555374447492, 1, 1], LineBox[{368, 369}]}, 
       {Hue[0.8553207001162826, 1, 1], LineBox[{369, 370}]}, 
       {Hue[0.8549734798782153, 1, 1], LineBox[{370, 371}]}, 
       {Hue[0.8546258934543335, 1, 1], LineBox[{371, 372}]}, 
       {Hue[0.8539296221742921, 1, 1], LineBox[{372, 373}]}, 
       {Hue[0.8525326872315117, 1, 1], LineBox[{373, 374}]}, 
       {Hue[0.8497212657634408, 1, 1], LineBox[{374, 375}]}, 
       {Hue[0.8440284786662016, 1, 1], LineBox[{375, 376}]}, 
       {Hue[0.8323672368187282, 1, 1], LineBox[{376, 377}]}, 
       {Hue[0.832021825341367, 1, 1], LineBox[{377, 378}]}, 
       {Hue[0.8316761129089216, 1, 1], LineBox[{378, 379}]}, 
       {Hue[0.830983787289033, 1, 1], LineBox[{379, 380}]}, 
       {Hue[0.8295955482610037, 1, 1], LineBox[{380, 381}]}, 
       {Hue[0.8268048478525195, 1, 1], LineBox[{381, 382}]}, 
       {Hue[0.8211677004737814, 1, 1], LineBox[{382, 383}]}, 
       {Hue[0.8208129603290649, 1, 1], LineBox[{383, 384}]}, 
       {Hue[0.8204579401889733, 1, 1], LineBox[{384, 385}]}, 
       {Hue[0.8197470631114766, 1, 1], LineBox[{385, 386}]}, 
       {Hue[0.8183219845703645, 1, 1], LineBox[{386, 387}]}, 
       {Hue[0.8154587196009777, 1, 1], LineBox[{387, 388}]}, 
       {Hue[0.8096813911841393, 1, 1], LineBox[{388, 389}]}, 
       {Hue[0.8092875819890374, 1, 1], LineBox[{389, 390}]}, 
       {Hue[0.808893478770385, 1, 1], LineBox[{390, 391}]}, 
       {Hue[0.8081043957324173, 1, 1], LineBox[{391, 392}]}, 
       {Hue[0.8065227621999305, 1, 1], LineBox[{392, 393}]}, 
       {Hue[0.8033459466645283, 1, 1], LineBox[{393, 394}]}, 
       {Hue[0.802947603375167, 1, 1], LineBox[{394, 395}]}, 
       {Hue[0.8025489892916752, 1, 1], LineBox[{395, 396}]}, 
       {Hue[0.8017509549521107, 1, 1], LineBox[{396, 397}]}, 
       {Hue[0.8001517056838474, 1, 1], LineBox[{397, 398}]}, 
       {Hue[0.7969408469898973, 1, 1], LineBox[{398, 399}]}, 
       {Hue[0.7965383632273197, 1, 1], LineBox[{399, 400}]}, 
       {Hue[0.7961356348032806, 1, 1], LineBox[{400, 401}]}, 
       {Hue[0.7953294508865575, 1, 1], LineBox[{401, 402}]}, 
       {Hue[0.7937142237832816, 1, 1], LineBox[{402, 403}]}, 
       {Hue[0.7904727333124814, 1, 1], LineBox[{403, 404}]}, 
       {Hue[0.7900665479973398, 1, 1], LineBox[{404, 405}]}, 
       {Hue[0.7896601468913598, 1, 1], LineBox[{405, 406}]}, 
       {Hue[0.78884670487822, 1, 1], LineBox[{406, 407}]}, 
       {Hue[0.7872173151887687, 1, 1], LineBox[{407, 408}]}, 
       {Hue[0.7839489494382963, 1, 1], LineBox[{408, 409}]}, 
       {Hue[0.7835666629003442, 1, 1], LineBox[{409, 410}]}, 
       {Hue[0.7831842154488271, 1, 1], LineBox[{410, 411}]}, 
       {Hue[0.7824188444458442, 1, 1], LineBox[{411, 412}]}, 
       {Hue[0.7808862448877505, 1, 1], LineBox[{412, 413}]}, 
       {Hue[0.7778139956224613, 1, 1], LineBox[{413, 414}]}, 
       {Hue[0.7774293370436219, 1, 1], LineBox[{414, 415}]}, 
       {Hue[0.7770445448384399, 1, 1], LineBox[{415, 416}]}, 
       {Hue[0.7762745665778351, 1, 1], LineBox[{416, 417}]}, 
       {Hue[0.774733084163743, 1, 1], LineBox[{417, 418}]}, 
       {Hue[0.7716444162444386, 1, 1], LineBox[{418, 419}]}, 
       {Hue[0.7712578333211608, 1, 1], LineBox[{419, 420}]}, 
       {Hue[0.770871145491256, 1, 1], LineBox[{420, 421}]}, 
       {Hue[0.7700974624607736, 1, 1], LineBox[{421, 422}]}, 
       {Hue[0.7685489186047042, 1, 1], LineBox[{422, 423}]}, 
       {Hue[0.7681615480840327, 1, 1], LineBox[{423, 424}]}, 
       {Hue[0.7677740874657797, 1, 1], LineBox[{424, 425}]}, 
       {Hue[0.7669989034180303, 1, 1], LineBox[{425, 426}]}, 
       {Hue[0.765447536657087, 1, 1], LineBox[{426, 427}]}, 
       {Hue[0.7650594979075211, 1, 1], LineBox[{427, 428}]}, 
       {Hue[0.7646713841180712, 1, 1], LineBox[{428, 429}]}, 
       {Hue[0.7638949390135434, 1, 1], LineBox[{429, 430}]}, 
       {Hue[0.7635066115033468, 1, 1], LineBox[{430, 431}]}, 
       {Hue[0.7631182165628669, 1, 1], LineBox[{431, 432}]}, 
       {Hue[0.7623412320336995, 1, 1], LineBox[{432, 433}]}, 
       {Hue[0.7619526462733915, 1, 1], LineBox[{433, 434}]}, 
       {Hue[0.7615640007393966, 1, 1], LineBox[{434, 435}]}, 
       {Hue[0.7607865380363968, 1, 1], LineBox[{435, 436}]}, 
       {Hue[0.7592309800296734, 1, 1], LineBox[{436, 437}]}, 
       {Hue[0.758849596113294, 1, 1], LineBox[{437, 438}]}, 
       {Hue[0.7584681695412607, 1, 1], LineBox[{438, 439}]}, 
       {Hue[0.7577051957396829, 1, 1], LineBox[{439, 440}]}, 
       {Hue[0.7573236521690917, 1, 1], LineBox[{440, 441}]}, 
       {Hue[0.7569420732606006, 1, 1], LineBox[{441, 442}]}, 
       {Hue[0.7561788167647769, 1, 1], LineBox[{442, 443}]}, 
       {Hue[0.7557971428483347, 1, 1], LineBox[{443, 444}]}, 
       {Hue[0.755415440935619, 1, 1], LineBox[{444, 445}]}, 
       {Hue[0.7550337128651892, 1, 1], LineBox[{445, 446}]}, 
       {Hue[0.7546519604767108, 1, 1], LineBox[{446, 447}]}, 
       {Hue[0.7542701856108777, 1, 1], LineBox[{447, 448}]}, 
       {Hue[0.7538883901093358, 1, 1], LineBox[{448, 449}]}, 
       {Hue[0.7535065758146038, 1, 1], LineBox[{449, 450}]}, 
       {Hue[0.7531247445699961, 1, 1], LineBox[{450, 451}]}, 
       {Hue[0.7527428982195451, 1, 1], LineBox[{451, 452}]}, 
       {Hue[0.7523610386079226, 1, 1], LineBox[{452, 453}]}, 
       {Hue[0.7519791675803619, 1, 1], LineBox[{453, 454}]}, 
       {Hue[0.7515972869825804, 1, 1], LineBox[{454, 455}]}, 
       {Hue[0.7512153986607009, 1, 1], LineBox[{455, 456}]}, 
       {Hue[0.7508335044611727, 1, 1], LineBox[{456, 457}]}, 
       {Hue[0.7504516062306952, 1, 1], LineBox[{457, 458}]}, 
       {Hue[0.7500697058161379, 1, 1], LineBox[{458, 459}]}, 
       {Hue[0.749687805064463, 1, 1], LineBox[{459, 460}]}, 
       {Hue[0.7493059058226473, 1, 1], LineBox[{460, 461}]}, 
       {Hue[0.7489240099376031, 1, 1], LineBox[{461, 462}]}, 
       {Hue[0.748542119256101, 1, 1], LineBox[{462, 463}]}, 
       {Hue[0.7481602356246906, 1, 1], LineBox[{463, 464}]}, 
       {Hue[0.7477783608896231, 1, 1], LineBox[{464, 465}]}, 
       {Hue[0.7473964968967727, 1, 1], LineBox[{465, 466}]}, 
       {Hue[0.7470146454915587, 1, 1], LineBox[{466, 467}]}, 
       {Hue[0.7466328085188669, 1, 1], LineBox[{467, 468}]}, 
       {Hue[0.7462509878229721, 1, 1], LineBox[{468, 469}]}, 
       {Hue[0.7458691852474597, 1, 1], LineBox[{469, 470}]}, 
       {Hue[0.7454874026351481, 1, 1], LineBox[{470, 471}]}, 
       {Hue[0.7451056418280108, 1, 1], LineBox[{471, 472}]}, 
       {Hue[0.7447239046670979, 1, 1], LineBox[{472, 473}]}, 
       {Hue[0.7439605086430686, 1, 1], LineBox[{473, 474}]}, 
       {Hue[0.7435788534567404, 1, 1], LineBox[{474, 475}]}, 
       {Hue[0.743197229270059, 1, 1], LineBox[{475, 476}]}, 
       {Hue[0.7424340812353439, 1, 1], LineBox[{476, 477}]}, 
       {Hue[0.7420525610536186, 1, 1], LineBox[{477, 478}]}, 
       {Hue[0.7416710792040031, 1, 1], LineBox[{478, 479}]}, 
       {Hue[0.7409082378164619, 1, 1], LineBox[{479, 480}]}, 
       {Hue[0.740526881931904, 1, 1], LineBox[{480, 481}]}, 
       {Hue[0.7401455716860399, 1, 1], LineBox[{481, 482}]}, 
       {Hue[0.7393830953965116, 1, 1], LineBox[{482, 483}]}, 
       {Hue[0.7390019329908355, 1, 1], LineBox[{483, 484}]}, 
       {Hue[0.7386208234996777, 1, 1], LineBox[{484, 485}]}, 
       {Hue[0.7378587705129496, 1, 1], LineBox[{485, 486}]}, 
       {Hue[0.7374778306375714, 1, 1], LineBox[{486, 487}]}, 
       {Hue[0.7370969509169467, 1, 1], LineBox[{487, 488}]}, 
       {Hue[0.7363353791531093, 1, 1], LineBox[{488, 489}]}, 
       {Hue[0.7348130366775789, 1, 1], LineBox[{489, 490}]}, 
       {Hue[0.7344004115686685, 1, 1], LineBox[{490, 491}]}, 
       {Hue[0.7339878743030456, 1, 1], LineBox[{491, 492}]}, 
       {Hue[0.7331630723875937, 1, 1], LineBox[{492, 493}]}, 
       {Hue[0.7315146223679854, 1, 1], LineBox[{493, 494}]}, 
       {Hue[0.7311027633829322, 1, 1], LineBox[{494, 495}]}, 
       {Hue[0.7306910102921902, 1, 1], LineBox[{495, 496}]}, 
       {Hue[0.729867830725456, 1, 1], LineBox[{496, 497}]}, 
       {Hue[0.7282228402733528, 1, 1], LineBox[{497, 498}]}, 
       {Hue[0.7278118907005582, 1, 1], LineBox[{498, 499}]}, 
       {Hue[0.7274010647432104, 1, 1], LineBox[{499, 500}]}, 
       {Hue[0.726579792426144, 1, 1], LineBox[{500, 501}]}, 
       {Hue[0.7249388270964126, 1, 1], LineBox[{501, 502}]}, 
       {Hue[0.7216636955311848, 1, 1], LineBox[{502, 503}]}, 
       {Hue[0.7212549829546681, 1, 1], LineBox[{503, 504}]}, 
       {Hue[0.7208464282417213, 1, 1], LineBox[{504, 505}]}, 
       {Hue[0.7200298007228656, 1, 1], LineBox[{505, 506}]}, 
       {Hue[0.7183985311005171, 1, 1], LineBox[{506, 507}]}, 
       {Hue[0.7151443891844453, 1, 1], LineBox[{507, 508}]}, 
       {Hue[0.7147384477657693, 1, 1], LineBox[{508, 509}]}, 
       {Hue[0.7143326965547425, 1, 1], LineBox[{509, 510}]}, 
       {Hue[0.713521772548821, 1, 1], LineBox[{510, 511}]}, 
       {Hue[0.7119022922564333, 1, 1], LineBox[{511, 512}]}, 
       {Hue[0.7086732273419383, 1, 1], LineBox[{512, 513}]}, 
       {Hue[0.7082974065478993, 1, 1], LineBox[{513, 514}]}, 
       {Hue[0.7079217775613169, 1, 1], LineBox[{514, 515}]}, 
       {Hue[0.7071711008642876, 1, 1], LineBox[{515, 516}]}, 
       {Hue[0.7056721131307586, 1, 1], LineBox[{516, 517}]}, 
       {Hue[0.7026839183442184, 1, 1], LineBox[{517, 518}]}, 
       {Hue[0.6967490903085825, 1, 1], LineBox[{518, 519}]}, 
       {Hue[0.6963800957883017, 1, 1], LineBox[{519, 520}]}, 
       {Hue[0.6960113360845944, 1, 1], LineBox[{520, 521}]}, 
       {Hue[0.6952745259565233, 1, 1], LineBox[{521, 522}]}, 
       {Hue[0.6938037761512619, 1, 1], LineBox[{522, 523}]}, 
       {Hue[0.6908740178550813, 1, 1], LineBox[{523, 524}]}, 
       {Hue[0.6850634268348434, 1, 1], LineBox[{524, 525}]}, 
       {Hue[0.6846723194047373, 1, 1], LineBox[{525, 526}]}, 
       {Hue[0.6842815282627351, 1, 1], LineBox[{526, 527}]}, 
       {Hue[0.6835008995719081, 1, 1], LineBox[{527, 528}]}, 
       {Hue[0.6819434889866574, 1, 1], LineBox[{528, 529}]}, 
       {Hue[0.6788443042180083, 1, 1], LineBox[{529, 530}]}, 
       {Hue[0.6727103116791255, 1, 1], LineBox[{530, 531}]}, 
       {Hue[0.6607119238065423, 1, 1], LineBox[{531, 532}]}, 
       {Hue[0.6603496703678688, 1, 1], LineBox[{532, 533}]}, 
       {Hue[0.6599877717876945, 1, 1], LineBox[{533, 534}]}, 
       {Hue[0.6592650408555486, 1, 1], LineBox[{534, 535}]}, 
       {Hue[0.6578238549160509, 1, 1], LineBox[{535, 536}]}, 
       {Hue[0.6549586665777667, 1, 1], LineBox[{536, 537}]}, 
       {Hue[0.6492975339672553, 1, 1], LineBox[{537, 538}]}, 
       {Hue[0.6382543738240635, 1, 1], LineBox[{538, 539}]}, 
       {Hue[0.6186381533904819, 1, 1], LineBox[{539, 540}]}, 
       {Hue[0.5987450483813834, 1, 1], LineBox[{540, 541}]}, 
       {Hue[0.581327200613792, 1, 1], LineBox[{541, 542}]}, 
       {Hue[0.5651708351621385, 1, 1], LineBox[{542, 543}]}, 
       {Hue[0.5649038110512727, 1, 1], LineBox[{543, 544}]}, 
       {Hue[0.5646369988185738, 1, 1], LineBox[{544, 545}]}, 
       {Hue[0.5641040066386009, 1, 1], LineBox[{545, 546}]}, 
       {Hue[0.5630405279707895, 1, 1], LineBox[{546, 547}]}, 
       {Hue[0.5609234057403701, 1, 1], LineBox[{547, 548}]}, 
       {Hue[0.5567270000080751, 1, 1], LineBox[{548, 549}]}, 
       {Hue[0.5564663404531357, 1, 1], LineBox[{549, 550}]}, 
       {Hue[0.5562058659678826, 1, 1], LineBox[{550, 551}]}, 
       {Hue[0.5556854688584882, 1, 1], LineBox[{551, 552}]}, 
       {Hue[0.5546468586555297, 1, 1], LineBox[{552, 553}]}, 
       {Hue[0.5525781870036823, 1, 1], LineBox[{553, 554}]}, 
       {Hue[0.548473542566891, 1, 1], LineBox[{554, 555}]}, 
       {Hue[0.5482354485858063, 1, 1], LineBox[{555, 556}]}, 
       {Hue[0.5479974925493978, 1, 1], LineBox[{556, 557}]}, 
       {Hue[0.5475219916105107, 1, 1], LineBox[{557, 558}]}, 
       {Hue[0.5465726153787527, 1, 1], LineBox[{558, 559}]}, 
       {Hue[0.5446802145516939, 1, 1], LineBox[{559, 560}]}, 
       {Hue[0.540919614724254, 1, 1], LineBox[{560, 561}]}, 
       {Hue[0.5406856011644792, 1, 1], LineBox[{561, 562}]}, 
       {Hue[0.5404517040519298, 1, 1], LineBox[{562, 563}]}, 
       {Hue[0.5399842564964968, 1, 1], LineBox[{563, 564}]}, 
       {Hue[0.539050729373585, 1, 1], LineBox[{564, 565}]}, 
       {Hue[0.5371889977559043, 1, 1], LineBox[{565, 566}]}, 
       {Hue[0.5334856337081378, 1, 1], LineBox[{566, 567}]}, 
       {Hue[0.5332356702272592, 1, 1], LineBox[{567, 568}]}, 
       {Hue[0.5329858185148917, 1, 1], LineBox[{568, 569}]}, 
       {Hue[0.5324864470328554, 1, 1], LineBox[{569, 570}]}, 
       {Hue[0.5314890083184363, 1, 1], LineBox[{570, 571}]}, 
       {Hue[0.5294991599976908, 1, 1], LineBox[{571, 572}]}, 
       {Hue[0.5292508841040008, 1, 1], LineBox[{572, 573}]}, 
       {Hue[0.529002706565882, 1, 1], LineBox[{573, 574}]}, 
       {Hue[0.5285066432145402, 1, 1], LineBox[{574, 575}]}, 
       {Hue[0.5275156600365587, 1, 1], LineBox[{575, 576}]}, 
       {Hue[0.5255380810443776, 1, 1], LineBox[{576, 577}]}, 
       {Hue[0.525291278752398, 1, 1], LineBox[{577, 578}]}, 
       {Hue[0.5250445614850563, 1, 1], LineBox[{578, 579}]}, 
       {Hue[0.5245513787022311, 1, 1], LineBox[{579, 580}]}, 
       {Hue[0.5235659969060595, 1, 1], LineBox[{580, 581}]}, 
       {Hue[0.5215989827293759, 1, 1], LineBox[{581, 582}]}, 
       {Hue[0.5213534413229897, 1, 1], LineBox[{582, 583}]}, 
       {Hue[0.5211079716861946, 1, 1], LineBox[{583, 584}]}, 
       {Hue[0.5206172444172416, 1, 1], LineBox[{584, 585}]}, 
       {Hue[0.5196366147834943, 1, 1], LineBox[{585, 586}]}, 
       {Hue[0.5176784704295874, 1, 1], LineBox[{586, 587}]}, 
       {Hue[0.517671162797467, 1, 1], LineBox[{587, 588}]}, 
       {Hue[0.5176638552183818, 1, 1], LineBox[{588, 589}]}, 
       {Hue[0.5176492402192279, 1, 1], LineBox[{589, 590}]}, 
       {Hue[0.5176200108563732, 1, 1], LineBox[{590, 591}]}, 
       {Hue[0.5175615546675576, 1, 1], LineBox[{591, 592}]}, 
       {Hue[0.5174446523981632, 1, 1], LineBox[{592, 593}]}, 
       {Hue[0.5174373464522712, 1, 1], LineBox[{593, 594}]}, 
       {Hue[0.5174300405587118, 1, 1], LineBox[{594, 595}]}, 
       {Hue[0.5174154289285023, 1, 1], LineBox[{595, 596}]}, 
       {Hue[0.5173862062951068, 1, 1], LineBox[{596, 597}]}, 
       {Hue[0.5173277635314939, 1, 1], LineBox[{597, 598}]}, 
       {Hue[0.5173204584202873, 1, 1], LineBox[{598, 599}]}, 
       {Hue[0.5173131533610619, 1, 1], LineBox[{599, 600}]}, 
       {Hue[0.5172985433984671, 1, 1], LineBox[{600, 601}]}, 
       {Hue[0.5172693240960875, 1, 1], LineBox[{601, 602}]}, 
       {Hue[0.5172620194001166, 1, 1], LineBox[{602, 603}]}, 
       {Hue[0.5172547147559514, 1, 1], LineBox[{603, 604}]}, 
       {Hue[0.5172401056229503, 1, 1], LineBox[{604, 605}]}, 
       {Hue[0.5172328011340706, 1, 1], LineBox[{605, 606}]}, 
       {Hue[0.5172254966969089, 1, 1], LineBox[{606, 607}]}, 
       {Hue[0.5172181923114431, 1, 1], LineBox[{607, 608}]}}}},
    VertexColors->{
      Hue[0.4827891120223487, 1, 1], 
      Hue[0.482564154597886, 1, 1], 
      Hue[0.4823391475506281, 1, 1], 
      Hue[0.4818889820233611, 1, 1], 
      Hue[0.48098802727801265`, 1, 1], 
      Hue[0.4791834792023881, 1, 1], 
      Hue[0.47895765079255526`, 1, 1], 
      Hue[0.47873176248257115`, 1, 1], 
      Hue[0.4782798035865052, 1, 1], 
      Hue[0.4773751386497867, 1, 1], 
      Hue[0.4755626757123231, 1, 1], 
      Hue[0.47533581152777804`, 1, 1], 
      Hue[0.47510887711774313`, 1, 1], 
      Hue[0.47465479503274677`, 1, 1], 
      Hue[0.4737457596719586, 1, 1], 
      Hue[0.4719240589571304, 1, 1], 
      Hue[0.46826497272567535`, 1, 1], 
      Hue[0.46801623645615187`, 1, 1], 
      Hue[0.4677673931206509, 1, 1], 
      Hue[0.4672693819164649, 1, 1], 
      Hue[0.46627203801009476`, 1, 1], 
      Hue[0.4642718770043928, 1, 1], 
      Hue[0.4602481601269848, 1, 1], 
      Hue[0.4599955825663628, 1, 1], 
      Hue[0.45974287109529516`, 1, 1], 
      Hue[0.45923704304401985`, 1, 1], 
      Hue[0.45822374284207923`, 1, 1], 
      Hue[0.4561903764679997, 1, 1], 
      Hue[0.45209505906550335`, 1, 1], 
      Hue[0.4518548306823689, 1, 1], 
      Hue[0.4516144619060803, 1, 1], 
      Hue[0.45113330039310223`, 1, 1], 
      Hue[0.45016926204801466`, 1, 1], 
      Hue[0.44823416810320266`, 1, 1], 
      Hue[0.44433466107445957`, 1, 1], 
      Hue[0.43640834183084554`, 1, 1], 
      Hue[0.42029717691584667`, 1, 1], 
      Hue[0.40180141253448876`, 1, 1], 
      Hue[0.3834099295601948, 1, 1], 
      Hue[0.36209016028713203`, 1, 1], 
      Hue[0.3396709561709671, 1, 1], 
      Hue[0.3393332954116793, 1, 1], 
      Hue[0.3389953266080553, 1, 1], 
      Hue[0.3383184662821582, 1, 1], 
      Hue[0.3369610650839929, 1, 1], 
      Hue[0.33423162960793346`, 1, 1], 
      Hue[0.32871504472731383`, 1, 1], 
      Hue[0.3283677395765662, 1, 1], 
      Hue[0.3280201415235637, 1, 1], 
      Hue[0.3273240691434086, 1, 1], 
      Hue[0.3259284367777426, 1, 1], 
      Hue[0.3231233685349689, 1, 1], 
      Hue[0.3174593058882333, 1, 1], 
      Hue[0.3170728973764113, 1, 1], 
      Hue[0.31668617223125795`, 1, 1], 
      Hue[0.3159117765219359, 1, 1], 
      Hue[0.31435923545210387`, 1, 1], 
      Hue[0.31123942077030387`, 1, 1], 
      Hue[0.3049431469648374, 1, 1], 
      Hue[0.3045472184743325, 1, 1], 
      Hue[0.304151014900581, 1, 1], 
      Hue[0.3033577884702304, 1, 1], 
      Hue[0.3017681008777704, 1, 1], 
      Hue[0.2985761353478784, 1, 1], 
      Hue[0.2921447786295377, 1, 1], 
      Hue[0.29176768769392497`, 1, 1], 
      Hue[0.29139040316372905`, 1, 1], 
      Hue[0.2906352592835539, 1, 1], 
      Hue[0.28912271440992654`, 1, 1], 
      Hue[0.28608893968771626`, 1, 1], 
      Hue[0.285708934056643, 1, 1], 
      Hue[0.285328759532689, 1, 1], 
      Hue[0.28456791022786804`, 1, 1], 
      Hue[0.28304425591740834`, 1, 1], 
      Hue[0.279989492581923, 1, 1], 
      Hue[0.27960698079228, 1, 1], 
      Hue[0.27922432654828644`, 1, 1], 
      Hue[0.27845859752293123`, 1, 1], 
      Hue[0.27692550542522587`, 1, 1], 
      Hue[0.27654190206159385`, 1, 1], 
      Hue[0.27615817004289983`, 1, 1], 
      Hue[0.27539032704479266`, 1, 1], 
      Hue[0.27385317453066826`, 1, 1], 
      Hue[0.2734685912133791, 1, 1], 
      Hue[0.27308389338429445`, 1, 1], 
      Hue[0.27231416135705505`, 1, 1], 
      Hue[0.2707734022533922, 1, 1], 
      Hue[0.2676871110167407, 1, 1], 
      Hue[0.26730851114187276`, 1, 1], 
      Hue[0.2669298292222692, 1, 1], 
      Hue[0.2661722262534576, 1, 1], 
      Hue[0.264656113011379, 1, 1], 
      Hue[0.26427690596839526`, 1, 1], 
      Hue[0.2638976309721529, 1, 1], 
      Hue[0.26313888422214005`, 1, 1], 
      Hue[0.26162065355766034`, 1, 1], 
      Hue[0.2612409527340028, 1, 1], 
      Hue[0.26086119822901244`, 1, 1], 
      Hue[0.2601015353565505, 1, 1], 
      Hue[0.2597216305851791, 1, 1], 
      Hue[0.25934167932452723`, 1, 1], 
      Hue[0.2585816445494831, 1, 1], 
      Hue[0.25820156464674043`, 1, 1], 
      Hue[0.25782144547786634`, 1, 1], 
      Hue[0.25706109658363874`, 1, 1], 
      Hue[0.25668087048309157`, 1, 1], 
      Hue[0.25630061236587415`, 1, 1], 
      Hue[0.2555400073463437, 1, 1], 
      Hue[0.2551596640795802, 1, 1], 
      Hue[0.25477929606709404`, 1, 1], 
      Hue[0.2543989051294105, 1, 1], 
      Hue[0.2540184930880174, 1, 1], 
      Hue[0.25363806176528997`, 1, 1], 
      Hue[0.2532576129844135, 1, 1], 
      Hue[0.2528771485693071, 1, 1], 
      Hue[0.2524966703445472, 1, 1], 
      Hue[0.25211618013529086`, 1, 1], 
      Hue[0.25173567976719907`, 1, 1], 
      Hue[0.2513551710663605, 1, 1], 
      Hue[0.25097465585921375`, 1, 1], 
      Hue[0.2505941359724715, 1, 1], 
      Hue[0.25021361323304303`, 1, 1], 
      Hue[0.24983308946795788`, 1, 1], 
      Hue[0.2494525665042888, 1, 1], 
      Hue[0.2490720461690742, 1, 1], 
      Hue[0.2486915302892427, 1, 1], 
      Hue[0.2483110206915351, 1, 1], 
      Hue[0.2479305192024281, 1, 1], 
      Hue[0.24755002764805686`, 1, 1], 
      Hue[0.24716954785413933`, 1, 1], 
      Hue[0.24678908164589852`, 1, 1], 
      Hue[0.24640863084798625`, 1, 1], 
      Hue[0.24602819728440656`, 1, 1], 
      Hue[0.2456477827784387, 1, 1], 
      Hue[0.24526738915256116`, 1, 1], 
      Hue[0.24488701822837544`, 1, 1], 
      Hue[0.24450667182652894`, 1, 1], 
      Hue[0.24412635176663902`, 1, 1], 
      Hue[0.243365797945593, 1, 1], 
      Hue[0.24295324765529228`, 1, 1], 
      Hue[0.2425407371117956, 1, 1], 
      Hue[0.24171584452690567`, 1, 1], 
      Hue[0.24130346711105533`, 1, 1], 
      Hue[0.2408911386928689, 1, 1], 
      Hue[0.24006663807480177`, 1, 1], 
      Hue[0.23965447048114782`, 1, 1], 
      Hue[0.2392423610973872, 1, 1], 
      Hue[0.23841832614129013`, 1, 1], 
      Hue[0.23677105557424935`, 1, 1], 
      Hue[0.2363594177784766, 1, 1], 
      Hue[0.23594785647257394`, 1, 1], 
      Hue[0.23512497240373834`, 1, 1], 
      Hue[0.23348022172980518`, 1, 1], 
      Hue[0.23306925917031257`, 1, 1], 
      Hue[0.2326583911412925, 1, 1], 
      Hue[0.23183694761200105`, 1, 1], 
      Hue[0.23019529296078733`, 1, 1], 
      Hue[0.22978514896890515`, 1, 1], 
      Hue[0.2293751172542093, 1, 1], 
      Hue[0.22855539943102374`, 1, 1], 
      Hue[0.22691740731771895`, 1, 1], 
      Hue[0.2236476811129669, 1, 1], 
      Hue[0.22323959369566776`, 1, 1], 
      Hue[0.2228316529602289, 1, 1], 
      Hue[0.2220162199090814, 1, 1], 
      Hue[0.22038720570674894`, 1, 1], 
      Hue[0.21713704446893195`, 1, 1], 
      Hue[0.21675867941709884`, 1, 1], 
      Hue[0.2163804703713593, 1, 1], 
      Hue[0.21562452670544383`, 1, 1], 
      Hue[0.2141145815706027, 1, 1], 
      Hue[0.2111028105643298, 1, 1], 
      Hue[0.20511445765265002`, 1, 1], 
      Hue[0.2047418469969295, 1, 1], 
      Hue[0.20436944028448414`, 1, 1], 
      Hue[0.2036252441940914, 1, 1], 
      Hue[0.2021393594502079, 1, 1], 
      Hue[0.19917791790822034`, 1, 1], 
      Hue[0.19329862431015446`, 1, 1], 
      Hue[0.19290251120305676`, 1, 1], 
      Hue[0.19250668498575932`, 1, 1], 
      Hue[0.19171589889733506`, 1, 1], 
      Hue[0.19013783118538485`, 1, 1], 
      Hue[0.18699601635630125`, 1, 1], 
      Hue[0.18077193135702532`, 1, 1], 
      Hue[0.1685774743268824, 1, 1], 
      Hue[0.1682089558570103, 1, 1], 
      Hue[0.1678407785456452, 1, 1], 
      Hue[0.16710544991494183`, 1, 1], 
      Hue[0.16563891364948222`, 1, 1], 
      Hue[0.16272245258452597`, 1, 1], 
      Hue[0.15695685884954436`, 1, 1], 
      Hue[0.14569993963152494`, 1, 1], 
      Hue[0.1256893641497182, 1, 1], 
      Hue[0.10539044861568864`, 1, 1], 
      Hue[0.0876450931568508, 1, 1], 
      Hue[0.06954252802227788, 1, 1], 
      Hue[0.06927284345491812, 1, 1], 
      Hue[0.06900338285499233, 1, 1], 
      Hue[0.06846513025460943, 1, 1], 
      Hue[0.06739127631685438, 1, 1], 
      Hue[0.06525398821771008, 1, 1], 
      Hue[0.061019606605566934`, 1, 1], 
      Hue[0.060756678464624916`, 1, 1], 
      Hue[0.06049394778226548, 1, 1], 
      Hue[0.05996907547438736, 1, 1], 
      Hue[0.05892166385474366, 1, 1], 
      Hue[0.05683598678500967, 1, 1], 
      Hue[0.05269973189924155, 1, 1], 
      Hue[0.052459950972470006`, 1, 1], 
      Hue[0.05222031886942626, 1, 1], 
      Hue[0.05174149845122005, 1, 1], 
      Hue[0.05078561399160116, 1, 1], 
      Hue[0.048880720514997804`, 1, 1], 
      Hue[0.045097237160123156`, 1, 1], 
      Hue[0.04486188666864251, 1, 1], 
      Hue[0.0446266636188734, 1, 1], 
      Hue[0.04415659718485805, 1, 1], 
      Hue[0.04321796437381417, 1, 1], 
      Hue[0.041346550324896265`, 1, 1], 
      Hue[0.03762594234630728, 1, 1], 
      Hue[0.0373748435349179, 1, 1], 
      Hue[0.0371238695861517, 1, 1], 
      Hue[0.03662229292476846, 1, 1], 
      Hue[0.03562060110319645, 1, 1], 
      Hue[0.033622876197495356`, 1, 1], 
      Hue[0.033373674812982916`, 1, 1], 
      Hue[0.03312458492284089, 1, 1], 
      Hue[0.032626736295236373`, 1, 1], 
      Hue[0.03163234035619638, 1, 1], 
      Hue[0.02964856766170453, 1, 1], 
      Hue[0.029401050448112517`, 1, 1], 
      Hue[0.029153631445108612`, 1, 1], 
      Hue[0.02865908476120701, 1, 1], 
      Hue[0.027671133531875818`, 1, 1], 
      Hue[0.025699614688751985`, 1, 1], 
      Hue[0.021772636012389598`, 1, 1], 
      Hue[0.021544088339672494`, 1, 1], 
      Hue[0.021315603374490812`, 1, 1], 
      Hue[0.02085881890328848, 1, 1], 
      Hue[0.019945973164421232`, 1, 1], 
      Hue[0.018123025596737827`, 1, 1], 
      Hue[0.01789540144331209, 1, 1], 
      Hue[0.01766782936571196, 1, 1], 
      Hue[0.017212838786401, 1, 1], 
      Hue[0.0163034533796867, 1, 1], 
      Hue[0.01448691729759583, 1, 1], 
      Hue[0.014260046896887183`, 1, 1], 
      Hue[0.014033217984415945`, 1, 1], 
      Hue[0.013579681982588002`, 1, 1], 
      Hue[0.012673078787923713`, 1, 1], 
      Hue[0.012446521808640131`, 1, 1], 
      Hue[0.012220001037920017`, 1, 1], 
      Hue[0.011767065484800906`, 1, 1], 
      Hue[0.010861599877047007`, 1, 1], 
      Hue[0.010635314110578184`, 1, 1], 
      Hue[0.010409059281020351`, 1, 1], 
      Hue[0.009956639798957091, 1, 1], 
      Hue[0.009052143113095098, 1, 1], 
      Hue[0.008826086411195665, 1, 1], 
      Hue[0.008600055381481772, 1, 1], 
      Hue[0.008148067708061165, 1, 1], 
      Hue[0.007244371496228025, 1, 1], 
      Hue[0.007022937079816843, 1, 1], 
      Hue[0.006801522295207302, 1, 1], 
      Hue[0.0063587491451943855`, 1, 1], 
      Hue[0.006137389541852155, 1, 1], 
      Hue[0.0059160470944297034`, 1, 1], 
      Hue[0.005473411192127852, 1, 1], 
      Hue[0.0052521164997800344`, 1, 1], 
      Hue[0.005030836488410986, 1, 1], 
      Hue[0.00458831803424194, 1, 1], 
      Hue[0.004367078354377108, 1, 1], 
      Hue[0.004145850881357151, 1, 1], 
      Hue[0.0037034300821993217`, 1, 1], 
      Hue[0.0034822355193328733`, 1, 1], 
      Hue[0.0032610506898499213`, 1, 1], 
      Hue[0.003039874975474352, 1, 1], 
      Hue[0.002818707757961507, 1, 1], 
      Hue[0.0025975484190960568`, 1, 1], 
      Hue[0.0023763963406898754`, 1, 1], 
      Hue[0.0021552509045799197`, 1, 1], 
      Hue[0.0019341114926260908`, 1, 1], 
      Hue[0.0017129774867091097`, 1, 1], 
      Hue[0.0014918482687283844`, 1, 1], 
      Hue[0.0012707232205998733`, 1, 1], 
      Hue[0.001049601724253957, 1, 1], 
      Hue[0.0008284831616332965, 1, 1], 
      Hue[0.0006073669146907009, 1, 1], 
      Hue[0.00038625236538699054`, 1, 1], 
      Hue[0.00016513889568885723`, 1, 1], 
      Hue[0.9999440258875667, 1, 1], 
      Hue[0.9997229127229926, 1, 1], 
      Hue[0.9995017987839381, 1, 1], 
      Hue[0.9992806834523718, 1, 1], 
      Hue[0.9990595661102578, 1, 1], 
      Hue[0.9988384461395532, 1, 1], 
      Hue[0.9986173229222058, 1, 1], 
      Hue[0.9983961958401524, 1, 1], 
      Hue[0.9981750642753163, 1, 1], 
      Hue[0.9979539276096054, 1, 1], 
      Hue[0.9977327852249099, 1, 1], 
      Hue[0.9975116365031003, 1, 1], 
      Hue[0.997290480826025, 1, 1], 
      Hue[0.9970693175755087, 1, 1], 
      Hue[0.9966269658813178, 1, 1], 
      Hue[0.9964057762011531, 1, 1], 
      Hue[0.9961845764745626, 1, 1], 
      Hue[0.9957421444087574, 1, 1], 
      Hue[0.9955209108327753, 1, 1], 
      Hue[0.9952996647368284, 1, 1], 
      Hue[0.9948571325110459, 1, 1], 
      Hue[0.9946358451440981, 1, 1], 
      Hue[0.9944145427829568, 1, 1], 
      Hue[0.9939718906033176, 1, 1], 
      Hue[0.9930863790880589, 1, 1], 
      Hue[0.9928461651955212, 1, 1], 
      Hue[0.9926059277675018, 1, 1], 
      Hue[0.9921253791427431, 1, 1], 
      Hue[0.9911639646767813, 1, 1], 
      Hue[0.9909235403594647, 1, 1], 
      Hue[0.9906830861794046, 1, 1], 
      Hue[0.9902020850652838, 1, 1], 
      Hue[0.9892396896536435, 1, 1], 
      Hue[0.9889990042677944, 1, 1], 
      Hue[0.9887582826843183, 1, 1], 
      Hue[0.9882767277544471, 1, 1], 
      Hue[0.9873131486444305, 1, 1], 
      Hue[0.9853839357017069, 1, 1], 
      Hue[0.9851425759700883, 1, 1], 
      Hue[0.9849011673421513, 1, 1], 
      Hue[0.984418200216528, 1, 1], 
      Hue[0.9834516442098782, 1, 1], 
      Hue[0.9815158667939559, 1, 1], 
      Hue[0.9812736291913834, 1, 1], 
      Hue[0.9810313299456908, 1, 1], 
      Hue[0.9805465433306404, 1, 1], 
      Hue[0.9795761952326396, 1, 1], 
      Hue[0.9776322203763465, 1, 1], 
      Hue[0.9774051575452036, 1, 1], 
      Hue[0.9771780299268937, 1, 1], 
      Hue[0.9767235777199724, 1, 1], 
      Hue[0.9758138671519275, 1, 1], 
      Hue[0.9739910750405399, 1, 1], 
      Hue[0.9703308331526667, 1, 1], 
      Hue[0.9701013726173703, 1, 1], 
      Hue[0.9698718263196648, 1, 1], 
      Hue[0.9694124737999488, 1, 1], 
      Hue[0.9684927105904569, 1, 1], 
      Hue[0.9666488035074778, 1, 1], 
      Hue[0.9629422795583775, 1, 1], 
      Hue[0.96269025198017, 1, 1], 
      Hue[0.9624380986753572, 1, 1], 
      Hue[0.9619334114888105, 1, 1], 
      Hue[0.9609224910093952, 1, 1], 
      Hue[0.9588942749542262, 1, 1], 
      Hue[0.9548108126517543, 1, 1], 
      Hue[0.946523461974924, 1, 1], 
      Hue[0.9462664435525957, 1, 1], 
      Hue[0.946009250984037, 1, 1], 
      Hue[0.9454943401255437, 1, 1], 
      Hue[0.944462392535523, 1, 1], 
      Hue[0.9423898098455793, 1, 1], 
      Hue[0.9382084208650068, 1, 1], 
      Hue[0.9296889548412226, 1, 1], 
      Hue[0.9131577192100314, 1, 1], 
      Hue[0.8941248259641966, 1, 1], 
      Hue[0.8751555374447492, 1, 1], 
      Hue[0.8553207001162826, 1, 1], 
      Hue[0.8549734798782153, 1, 1], 
      Hue[0.8546258934543335, 1, 1], 
      Hue[0.8539296221742921, 1, 1], 
      Hue[0.8525326872315117, 1, 1], 
      Hue[0.8497212657634408, 1, 1], 
      Hue[0.8440284786662016, 1, 1], 
      Hue[0.8323672368187282, 1, 1], 
      Hue[0.832021825341367, 1, 1], 
      Hue[0.8316761129089216, 1, 1], 
      Hue[0.830983787289033, 1, 1], 
      Hue[0.8295955482610037, 1, 1], 
      Hue[0.8268048478525195, 1, 1], 
      Hue[0.8211677004737814, 1, 1], 
      Hue[0.8208129603290649, 1, 1], 
      Hue[0.8204579401889733, 1, 1], 
      Hue[0.8197470631114766, 1, 1], 
      Hue[0.8183219845703645, 1, 1], 
      Hue[0.8154587196009777, 1, 1], 
      Hue[0.8096813911841393, 1, 1], 
      Hue[0.8092875819890374, 1, 1], 
      Hue[0.808893478770385, 1, 1], 
      Hue[0.8081043957324173, 1, 1], 
      Hue[0.8065227621999305, 1, 1], 
      Hue[0.8033459466645283, 1, 1], 
      Hue[0.802947603375167, 1, 1], 
      Hue[0.8025489892916752, 1, 1], 
      Hue[0.8017509549521107, 1, 1], 
      Hue[0.8001517056838474, 1, 1], 
      Hue[0.7969408469898973, 1, 1], 
      Hue[0.7965383632273197, 1, 1], 
      Hue[0.7961356348032806, 1, 1], 
      Hue[0.7953294508865575, 1, 1], 
      Hue[0.7937142237832816, 1, 1], 
      Hue[0.7904727333124814, 1, 1], 
      Hue[0.7900665479973398, 1, 1], 
      Hue[0.7896601468913598, 1, 1], 
      Hue[0.78884670487822, 1, 1], 
      Hue[0.7872173151887687, 1, 1], 
      Hue[0.7839489494382963, 1, 1], 
      Hue[0.7835666629003442, 1, 1], 
      Hue[0.7831842154488271, 1, 1], 
      Hue[0.7824188444458442, 1, 1], 
      Hue[0.7808862448877505, 1, 1], 
      Hue[0.7778139956224613, 1, 1], 
      Hue[0.7774293370436219, 1, 1], 
      Hue[0.7770445448384399, 1, 1], 
      Hue[0.7762745665778351, 1, 1], 
      Hue[0.774733084163743, 1, 1], 
      Hue[0.7716444162444386, 1, 1], 
      Hue[0.7712578333211608, 1, 1], 
      Hue[0.770871145491256, 1, 1], 
      Hue[0.7700974624607736, 1, 1], 
      Hue[0.7685489186047042, 1, 1], 
      Hue[0.7681615480840327, 1, 1], 
      Hue[0.7677740874657797, 1, 1], 
      Hue[0.7669989034180303, 1, 1], 
      Hue[0.765447536657087, 1, 1], 
      Hue[0.7650594979075211, 1, 1], 
      Hue[0.7646713841180712, 1, 1], 
      Hue[0.7638949390135434, 1, 1], 
      Hue[0.7635066115033468, 1, 1], 
      Hue[0.7631182165628669, 1, 1], 
      Hue[0.7623412320336995, 1, 1], 
      Hue[0.7619526462733915, 1, 1], 
      Hue[0.7615640007393966, 1, 1], 
      Hue[0.7607865380363968, 1, 1], 
      Hue[0.7592309800296734, 1, 1], 
      Hue[0.758849596113294, 1, 1], 
      Hue[0.7584681695412607, 1, 1], 
      Hue[0.7577051957396829, 1, 1], 
      Hue[0.7573236521690917, 1, 1], 
      Hue[0.7569420732606006, 1, 1], 
      Hue[0.7561788167647769, 1, 1], 
      Hue[0.7557971428483347, 1, 1], 
      Hue[0.755415440935619, 1, 1], 
      Hue[0.7550337128651892, 1, 1], 
      Hue[0.7546519604767108, 1, 1], 
      Hue[0.7542701856108777, 1, 1], 
      Hue[0.7538883901093358, 1, 1], 
      Hue[0.7535065758146038, 1, 1], 
      Hue[0.7531247445699961, 1, 1], 
      Hue[0.7527428982195451, 1, 1], 
      Hue[0.7523610386079226, 1, 1], 
      Hue[0.7519791675803619, 1, 1], 
      Hue[0.7515972869825804, 1, 1], 
      Hue[0.7512153986607009, 1, 1], 
      Hue[0.7508335044611727, 1, 1], 
      Hue[0.7504516062306952, 1, 1], 
      Hue[0.7500697058161379, 1, 1], 
      Hue[0.749687805064463, 1, 1], 
      Hue[0.7493059058226473, 1, 1], 
      Hue[0.7489240099376031, 1, 1], 
      Hue[0.748542119256101, 1, 1], 
      Hue[0.7481602356246906, 1, 1], 
      Hue[0.7477783608896231, 1, 1], 
      Hue[0.7473964968967727, 1, 1], 
      Hue[0.7470146454915587, 1, 1], 
      Hue[0.7466328085188669, 1, 1], 
      Hue[0.7462509878229721, 1, 1], 
      Hue[0.7458691852474597, 1, 1], 
      Hue[0.7454874026351481, 1, 1], 
      Hue[0.7451056418280108, 1, 1], 
      Hue[0.7447239046670979, 1, 1], 
      Hue[0.7439605086430686, 1, 1], 
      Hue[0.7435788534567404, 1, 1], 
      Hue[0.743197229270059, 1, 1], 
      Hue[0.7424340812353439, 1, 1], 
      Hue[0.7420525610536186, 1, 1], 
      Hue[0.7416710792040031, 1, 1], 
      Hue[0.7409082378164619, 1, 1], 
      Hue[0.740526881931904, 1, 1], 
      Hue[0.7401455716860399, 1, 1], 
      Hue[0.7393830953965116, 1, 1], 
      Hue[0.7390019329908355, 1, 1], 
      Hue[0.7386208234996777, 1, 1], 
      Hue[0.7378587705129496, 1, 1], 
      Hue[0.7374778306375714, 1, 1], 
      Hue[0.7370969509169467, 1, 1], 
      Hue[0.7363353791531093, 1, 1], 
      Hue[0.7348130366775789, 1, 1], 
      Hue[0.7344004115686685, 1, 1], 
      Hue[0.7339878743030456, 1, 1], 
      Hue[0.7331630723875937, 1, 1], 
      Hue[0.7315146223679854, 1, 1], 
      Hue[0.7311027633829322, 1, 1], 
      Hue[0.7306910102921902, 1, 1], 
      Hue[0.729867830725456, 1, 1], 
      Hue[0.7282228402733528, 1, 1], 
      Hue[0.7278118907005582, 1, 1], 
      Hue[0.7274010647432104, 1, 1], 
      Hue[0.726579792426144, 1, 1], 
      Hue[0.7249388270964126, 1, 1], 
      Hue[0.7216636955311848, 1, 1], 
      Hue[0.7212549829546681, 1, 1], 
      Hue[0.7208464282417213, 1, 1], 
      Hue[0.7200298007228656, 1, 1], 
      Hue[0.7183985311005171, 1, 1], 
      Hue[0.7151443891844453, 1, 1], 
      Hue[0.7147384477657693, 1, 1], 
      Hue[0.7143326965547425, 1, 1], 
      Hue[0.713521772548821, 1, 1], 
      Hue[0.7119022922564333, 1, 1], 
      Hue[0.7086732273419383, 1, 1], 
      Hue[0.7082974065478993, 1, 1], 
      Hue[0.7079217775613169, 1, 1], 
      Hue[0.7071711008642876, 1, 1], 
      Hue[0.7056721131307586, 1, 1], 
      Hue[0.7026839183442184, 1, 1], 
      Hue[0.6967490903085825, 1, 1], 
      Hue[0.6963800957883017, 1, 1], 
      Hue[0.6960113360845944, 1, 1], 
      Hue[0.6952745259565233, 1, 1], 
      Hue[0.6938037761512619, 1, 1], 
      Hue[0.6908740178550813, 1, 1], 
      Hue[0.6850634268348434, 1, 1], 
      Hue[0.6846723194047373, 1, 1], 
      Hue[0.6842815282627351, 1, 1], 
      Hue[0.6835008995719081, 1, 1], 
      Hue[0.6819434889866574, 1, 1], 
      Hue[0.6788443042180083, 1, 1], 
      Hue[0.6727103116791255, 1, 1], 
      Hue[0.6607119238065423, 1, 1], 
      Hue[0.6603496703678688, 1, 1], 
      Hue[0.6599877717876945, 1, 1], 
      Hue[0.6592650408555486, 1, 1], 
      Hue[0.6578238549160509, 1, 1], 
      Hue[0.6549586665777667, 1, 1], 
      Hue[0.6492975339672553, 1, 1], 
      Hue[0.6382543738240635, 1, 1], 
      Hue[0.6186381533904819, 1, 1], 
      Hue[0.5987450483813834, 1, 1], 
      Hue[0.581327200613792, 1, 1], 
      Hue[0.5651708351621385, 1, 1], 
      Hue[0.5649038110512727, 1, 1], 
      Hue[0.5646369988185738, 1, 1], 
      Hue[0.5641040066386009, 1, 1], 
      Hue[0.5630405279707895, 1, 1], 
      Hue[0.5609234057403701, 1, 1], 
      Hue[0.5567270000080751, 1, 1], 
      Hue[0.5564663404531357, 1, 1], 
      Hue[0.5562058659678826, 1, 1], 
      Hue[0.5556854688584882, 1, 1], 
      Hue[0.5546468586555297, 1, 1], 
      Hue[0.5525781870036823, 1, 1], 
      Hue[0.548473542566891, 1, 1], 
      Hue[0.5482354485858063, 1, 1], 
      Hue[0.5479974925493978, 1, 1], 
      Hue[0.5475219916105107, 1, 1], 
      Hue[0.5465726153787527, 1, 1], 
      Hue[0.5446802145516939, 1, 1], 
      Hue[0.540919614724254, 1, 1], 
      Hue[0.5406856011644792, 1, 1], 
      Hue[0.5404517040519298, 1, 1], 
      Hue[0.5399842564964968, 1, 1], 
      Hue[0.539050729373585, 1, 1], 
      Hue[0.5371889977559043, 1, 1], 
      Hue[0.5334856337081378, 1, 1], 
      Hue[0.5332356702272592, 1, 1], 
      Hue[0.5329858185148917, 1, 1], 
      Hue[0.5324864470328554, 1, 1], 
      Hue[0.5314890083184363, 1, 1], 
      Hue[0.5294991599976908, 1, 1], 
      Hue[0.5292508841040008, 1, 1], 
      Hue[0.529002706565882, 1, 1], 
      Hue[0.5285066432145402, 1, 1], 
      Hue[0.5275156600365587, 1, 1], 
      Hue[0.5255380810443776, 1, 1], 
      Hue[0.525291278752398, 1, 1], 
      Hue[0.5250445614850563, 1, 1], 
      Hue[0.5245513787022311, 1, 1], 
      Hue[0.5235659969060595, 1, 1], 
      Hue[0.5215989827293759, 1, 1], 
      Hue[0.5213534413229897, 1, 1], 
      Hue[0.5211079716861946, 1, 1], 
      Hue[0.5206172444172416, 1, 1], 
      Hue[0.5196366147834943, 1, 1], 
      Hue[0.5176784704295874, 1, 1], 
      Hue[0.517671162797467, 1, 1], 
      Hue[0.5176638552183818, 1, 1], 
      Hue[0.5176492402192279, 1, 1], 
      Hue[0.5176200108563732, 1, 1], 
      Hue[0.5175615546675576, 1, 1], 
      Hue[0.5174446523981632, 1, 1], 
      Hue[0.5174373464522712, 1, 1], 
      Hue[0.5174300405587118, 1, 1], 
      Hue[0.5174154289285023, 1, 1], 
      Hue[0.5173862062951068, 1, 1], 
      Hue[0.5173277635314939, 1, 1], 
      Hue[0.5173204584202873, 1, 1], 
      Hue[0.5173131533610619, 1, 1], 
      Hue[0.5172985433984671, 1, 1], 
      Hue[0.5172693240960875, 1, 1], 
      Hue[0.5172620194001166, 1, 1], 
      Hue[0.5172547147559514, 1, 1], 
      Hue[0.5172401056229503, 1, 1], 
      Hue[0.5172328011340706, 1, 1], 
      Hue[0.5172254966969089, 1, 1], 
      Hue[0.5172181923114431, 1, 1], 
      Hue[0.5172108879776514, 1, 1], 
      Hue[0.4827891120223487, 1, 1], 
      Hue[0.482564154597886, 1, 1], 
      Hue[0.482564154597886, 1, 1], 
      Hue[0.4823391475506281, 1, 1], 
      Hue[0.4823391475506281, 1, 1], 
      Hue[0.4818889820233611, 1, 1], 
      Hue[0.4818889820233611, 1, 1], 
      Hue[0.48098802727801265`, 1, 1], 
      Hue[0.48098802727801265`, 1, 1], 
      Hue[0.4791834792023881, 1, 1], 
      Hue[0.4791834792023881, 1, 1], 
      Hue[0.47895765079255526`, 1, 1], 
      Hue[0.47895765079255526`, 1, 1], 
      Hue[0.47873176248257115`, 1, 1], 
      Hue[0.47873176248257115`, 1, 1], 
      Hue[0.4782798035865052, 1, 1], 
      Hue[0.4782798035865052, 1, 1], 
      Hue[0.4773751386497867, 1, 1], 
      Hue[0.4773751386497867, 1, 1], 
      Hue[0.4755626757123231, 1, 1], 
      Hue[0.4755626757123231, 1, 1], 
      Hue[0.47533581152777804`, 1, 1], 
      Hue[0.47533581152777804`, 1, 1], 
      Hue[0.47510887711774313`, 1, 1], 
      Hue[0.47510887711774313`, 1, 1], 
      Hue[0.47465479503274677`, 1, 1], 
      Hue[0.47465479503274677`, 1, 1], 
      Hue[0.4737457596719586, 1, 1], 
      Hue[0.4737457596719586, 1, 1], 
      Hue[0.4719240589571304, 1, 1], 
      Hue[0.4719240589571304, 1, 1], 
      Hue[0.46826497272567535`, 1, 1], 
      Hue[0.46826497272567535`, 1, 1], 
      Hue[0.46801623645615187`, 1, 1], 
      Hue[0.46801623645615187`, 1, 1], 
      Hue[0.4677673931206509, 1, 1], 
      Hue[0.4677673931206509, 1, 1], 
      Hue[0.4672693819164649, 1, 1], 
      Hue[0.4672693819164649, 1, 1], 
      Hue[0.46627203801009476`, 1, 1], 
      Hue[0.46627203801009476`, 1, 1], 
      Hue[0.4642718770043928, 1, 1], 
      Hue[0.4642718770043928, 1, 1], 
      Hue[0.4602481601269848, 1, 1], 
      Hue[0.4602481601269848, 1, 1], 
      Hue[0.4599955825663628, 1, 1], 
      Hue[0.4599955825663628, 1, 1], 
      Hue[0.45974287109529516`, 1, 1], 
      Hue[0.45974287109529516`, 1, 1], 
      Hue[0.45923704304401985`, 1, 1], 
      Hue[0.45923704304401985`, 1, 1], 
      Hue[0.45822374284207923`, 1, 1], 
      Hue[0.45822374284207923`, 1, 1], 
      Hue[0.4561903764679997, 1, 1], 
      Hue[0.4561903764679997, 1, 1], 
      Hue[0.45209505906550335`, 1, 1], 
      Hue[0.45209505906550335`, 1, 1], 
      Hue[0.4518548306823689, 1, 1], 
      Hue[0.4518548306823689, 1, 1], 
      Hue[0.4516144619060803, 1, 1], 
      Hue[0.4516144619060803, 1, 1], 
      Hue[0.45113330039310223`, 1, 1], 
      Hue[0.45113330039310223`, 1, 1], 
      Hue[0.45016926204801466`, 1, 1], 
      Hue[0.45016926204801466`, 1, 1], 
      Hue[0.44823416810320266`, 1, 1], 
      Hue[0.44823416810320266`, 1, 1], 
      Hue[0.44433466107445957`, 1, 1], 
      Hue[0.44433466107445957`, 1, 1], 
      Hue[0.43640834183084554`, 1, 1], 
      Hue[0.43640834183084554`, 1, 1], 
      Hue[0.42029717691584667`, 1, 1], 
      Hue[0.42029717691584667`, 1, 1], 
      Hue[0.40180141253448876`, 1, 1], 
      Hue[0.40180141253448876`, 1, 1], 
      Hue[0.3834099295601948, 1, 1], 
      Hue[0.3834099295601948, 1, 1], 
      Hue[0.36209016028713203`, 1, 1], 
      Hue[0.36209016028713203`, 1, 1], 
      Hue[0.3396709561709671, 1, 1], 
      Hue[0.3396709561709671, 1, 1], 
      Hue[0.3393332954116793, 1, 1], 
      Hue[0.3393332954116793, 1, 1], 
      Hue[0.3389953266080553, 1, 1], 
      Hue[0.3389953266080553, 1, 1], 
      Hue[0.3383184662821582, 1, 1], 
      Hue[0.3383184662821582, 1, 1], 
      Hue[0.3369610650839929, 1, 1], 
      Hue[0.3369610650839929, 1, 1], 
      Hue[0.33423162960793346`, 1, 1], 
      Hue[0.33423162960793346`, 1, 1], 
      Hue[0.32871504472731383`, 1, 1], 
      Hue[0.32871504472731383`, 1, 1], 
      Hue[0.3283677395765662, 1, 1], 
      Hue[0.3283677395765662, 1, 1], 
      Hue[0.3280201415235637, 1, 1], 
      Hue[0.3280201415235637, 1, 1], 
      Hue[0.3273240691434086, 1, 1], 
      Hue[0.3273240691434086, 1, 1], 
      Hue[0.3259284367777426, 1, 1], 
      Hue[0.3259284367777426, 1, 1], 
      Hue[0.3231233685349689, 1, 1], 
      Hue[0.3231233685349689, 1, 1], 
      Hue[0.3174593058882333, 1, 1], 
      Hue[0.3174593058882333, 1, 1], 
      Hue[0.3170728973764113, 1, 1], 
      Hue[0.3170728973764113, 1, 1], 
      Hue[0.31668617223125795`, 1, 1], 
      Hue[0.31668617223125795`, 1, 1], 
      Hue[0.3159117765219359, 1, 1], 
      Hue[0.3159117765219359, 1, 1], 
      Hue[0.31435923545210387`, 1, 1], 
      Hue[0.31435923545210387`, 1, 1], 
      Hue[0.31123942077030387`, 1, 1], 
      Hue[0.31123942077030387`, 1, 1], 
      Hue[0.3049431469648374, 1, 1], 
      Hue[0.3049431469648374, 1, 1], 
      Hue[0.3045472184743325, 1, 1], 
      Hue[0.3045472184743325, 1, 1], 
      Hue[0.304151014900581, 1, 1], 
      Hue[0.304151014900581, 1, 1], 
      Hue[0.3033577884702304, 1, 1], 
      Hue[0.3033577884702304, 1, 1], 
      Hue[0.3017681008777704, 1, 1], 
      Hue[0.3017681008777704, 1, 1], 
      Hue[0.2985761353478784, 1, 1], 
      Hue[0.2985761353478784, 1, 1], 
      Hue[0.2921447786295377, 1, 1], 
      Hue[0.2921447786295377, 1, 1], 
      Hue[0.29176768769392497`, 1, 1], 
      Hue[0.29176768769392497`, 1, 1], 
      Hue[0.29139040316372905`, 1, 1], 
      Hue[0.29139040316372905`, 1, 1], 
      Hue[0.2906352592835539, 1, 1], 
      Hue[0.2906352592835539, 1, 1], 
      Hue[0.28912271440992654`, 1, 1], 
      Hue[0.28912271440992654`, 1, 1], 
      Hue[0.28608893968771626`, 1, 1], 
      Hue[0.28608893968771626`, 1, 1], 
      Hue[0.285708934056643, 1, 1], 
      Hue[0.285708934056643, 1, 1], 
      Hue[0.285328759532689, 1, 1], 
      Hue[0.285328759532689, 1, 1], 
      Hue[0.28456791022786804`, 1, 1], 
      Hue[0.28456791022786804`, 1, 1], 
      Hue[0.28304425591740834`, 1, 1], 
      Hue[0.28304425591740834`, 1, 1], 
      Hue[0.279989492581923, 1, 1], 
      Hue[0.279989492581923, 1, 1], 
      Hue[0.27960698079228, 1, 1], 
      Hue[0.27960698079228, 1, 1], 
      Hue[0.27922432654828644`, 1, 1], 
      Hue[0.27922432654828644`, 1, 1], 
      Hue[0.27845859752293123`, 1, 1], 
      Hue[0.27845859752293123`, 1, 1], 
      Hue[0.27692550542522587`, 1, 1], 
      Hue[0.27692550542522587`, 1, 1], 
      Hue[0.27654190206159385`, 1, 1], 
      Hue[0.27654190206159385`, 1, 1], 
      Hue[0.27615817004289983`, 1, 1], 
      Hue[0.27615817004289983`, 1, 1], 
      Hue[0.27539032704479266`, 1, 1], 
      Hue[0.27539032704479266`, 1, 1], 
      Hue[0.27385317453066826`, 1, 1], 
      Hue[0.27385317453066826`, 1, 1], 
      Hue[0.2734685912133791, 1, 1], 
      Hue[0.2734685912133791, 1, 1], 
      Hue[0.27308389338429445`, 1, 1], 
      Hue[0.27308389338429445`, 1, 1], 
      Hue[0.27231416135705505`, 1, 1], 
      Hue[0.27231416135705505`, 1, 1], 
      Hue[0.2707734022533922, 1, 1], 
      Hue[0.2707734022533922, 1, 1], 
      Hue[0.2676871110167407, 1, 1], 
      Hue[0.2676871110167407, 1, 1], 
      Hue[0.26730851114187276`, 1, 1], 
      Hue[0.26730851114187276`, 1, 1], 
      Hue[0.2669298292222692, 1, 1], 
      Hue[0.2669298292222692, 1, 1], 
      Hue[0.2661722262534576, 1, 1], 
      Hue[0.2661722262534576, 1, 1], 
      Hue[0.264656113011379, 1, 1], 
      Hue[0.264656113011379, 1, 1], 
      Hue[0.26427690596839526`, 1, 1], 
      Hue[0.26427690596839526`, 1, 1], 
      Hue[0.2638976309721529, 1, 1], 
      Hue[0.2638976309721529, 1, 1], 
      Hue[0.26313888422214005`, 1, 1], 
      Hue[0.26313888422214005`, 1, 1], 
      Hue[0.26162065355766034`, 1, 1], 
      Hue[0.26162065355766034`, 1, 1], 
      Hue[0.2612409527340028, 1, 1], 
      Hue[0.2612409527340028, 1, 1], 
      Hue[0.26086119822901244`, 1, 1], 
      Hue[0.26086119822901244`, 1, 1], 
      Hue[0.2601015353565505, 1, 1], 
      Hue[0.2601015353565505, 1, 1], 
      Hue[0.2597216305851791, 1, 1], 
      Hue[0.2597216305851791, 1, 1], 
      Hue[0.25934167932452723`, 1, 1], 
      Hue[0.25934167932452723`, 1, 1], 
      Hue[0.2585816445494831, 1, 1], 
      Hue[0.2585816445494831, 1, 1], 
      Hue[0.25820156464674043`, 1, 1], 
      Hue[0.25820156464674043`, 1, 1], 
      Hue[0.25782144547786634`, 1, 1], 
      Hue[0.25782144547786634`, 1, 1], 
      Hue[0.25706109658363874`, 1, 1], 
      Hue[0.25706109658363874`, 1, 1], 
      Hue[0.25668087048309157`, 1, 1], 
      Hue[0.25668087048309157`, 1, 1], 
      Hue[0.25630061236587415`, 1, 1], 
      Hue[0.25630061236587415`, 1, 1], 
      Hue[0.2555400073463437, 1, 1], 
      Hue[0.2555400073463437, 1, 1], 
      Hue[0.2551596640795802, 1, 1], 
      Hue[0.2551596640795802, 1, 1], 
      Hue[0.25477929606709404`, 1, 1], 
      Hue[0.25477929606709404`, 1, 1], 
      Hue[0.2543989051294105, 1, 1], 
      Hue[0.2543989051294105, 1, 1], 
      Hue[0.2540184930880174, 1, 1], 
      Hue[0.2540184930880174, 1, 1], 
      Hue[0.25363806176528997`, 1, 1], 
      Hue[0.25363806176528997`, 1, 1], 
      Hue[0.2532576129844135, 1, 1], 
      Hue[0.2532576129844135, 1, 1], 
      Hue[0.2528771485693071, 1, 1], 
      Hue[0.2528771485693071, 1, 1], 
      Hue[0.2524966703445472, 1, 1], 
      Hue[0.2524966703445472, 1, 1], 
      Hue[0.25211618013529086`, 1, 1], 
      Hue[0.25211618013529086`, 1, 1], 
      Hue[0.25173567976719907`, 1, 1], 
      Hue[0.25173567976719907`, 1, 1], 
      Hue[0.2513551710663605, 1, 1], 
      Hue[0.2513551710663605, 1, 1], 
      Hue[0.25097465585921375`, 1, 1], 
      Hue[0.25097465585921375`, 1, 1], 
      Hue[0.2505941359724715, 1, 1], 
      Hue[0.2505941359724715, 1, 1], 
      Hue[0.25021361323304303`, 1, 1], 
      Hue[0.25021361323304303`, 1, 1], 
      Hue[0.24983308946795788`, 1, 1], 
      Hue[0.24983308946795788`, 1, 1], 
      Hue[0.2494525665042888, 1, 1], 
      Hue[0.2494525665042888, 1, 1], 
      Hue[0.2490720461690742, 1, 1], 
      Hue[0.2490720461690742, 1, 1], 
      Hue[0.2486915302892427, 1, 1], 
      Hue[0.2486915302892427, 1, 1], 
      Hue[0.2483110206915351, 1, 1], 
      Hue[0.2483110206915351, 1, 1], 
      Hue[0.2479305192024281, 1, 1], 
      Hue[0.2479305192024281, 1, 1], 
      Hue[0.24755002764805686`, 1, 1], 
      Hue[0.24755002764805686`, 1, 1], 
      Hue[0.24716954785413933`, 1, 1], 
      Hue[0.24716954785413933`, 1, 1], 
      Hue[0.24678908164589852`, 1, 1], 
      Hue[0.24678908164589852`, 1, 1], 
      Hue[0.24640863084798625`, 1, 1], 
      Hue[0.24640863084798625`, 1, 1], 
      Hue[0.24602819728440656`, 1, 1], 
      Hue[0.24602819728440656`, 1, 1], 
      Hue[0.2456477827784387, 1, 1], 
      Hue[0.2456477827784387, 1, 1], 
      Hue[0.24526738915256116`, 1, 1], 
      Hue[0.24526738915256116`, 1, 1], 
      Hue[0.24488701822837544`, 1, 1], 
      Hue[0.24488701822837544`, 1, 1], 
      Hue[0.24450667182652894`, 1, 1], 
      Hue[0.24450667182652894`, 1, 1], 
      Hue[0.24412635176663902`, 1, 1], 
      Hue[0.24412635176663902`, 1, 1], 
      Hue[0.243365797945593, 1, 1], 
      Hue[0.243365797945593, 1, 1], 
      Hue[0.24295324765529228`, 1, 1], 
      Hue[0.24295324765529228`, 1, 1], 
      Hue[0.2425407371117956, 1, 1], 
      Hue[0.2425407371117956, 1, 1], 
      Hue[0.24171584452690567`, 1, 1], 
      Hue[0.24171584452690567`, 1, 1], 
      Hue[0.24130346711105533`, 1, 1], 
      Hue[0.24130346711105533`, 1, 1], 
      Hue[0.2408911386928689, 1, 1], 
      Hue[0.2408911386928689, 1, 1], 
      Hue[0.24006663807480177`, 1, 1], 
      Hue[0.24006663807480177`, 1, 1], 
      Hue[0.23965447048114782`, 1, 1], 
      Hue[0.23965447048114782`, 1, 1], 
      Hue[0.2392423610973872, 1, 1], 
      Hue[0.2392423610973872, 1, 1], 
      Hue[0.23841832614129013`, 1, 1], 
      Hue[0.23841832614129013`, 1, 1], 
      Hue[0.23677105557424935`, 1, 1], 
      Hue[0.23677105557424935`, 1, 1], 
      Hue[0.2363594177784766, 1, 1], 
      Hue[0.2363594177784766, 1, 1], 
      Hue[0.23594785647257394`, 1, 1], 
      Hue[0.23594785647257394`, 1, 1], 
      Hue[0.23512497240373834`, 1, 1], 
      Hue[0.23512497240373834`, 1, 1], 
      Hue[0.23348022172980518`, 1, 1], 
      Hue[0.23348022172980518`, 1, 1], 
      Hue[0.23306925917031257`, 1, 1], 
      Hue[0.23306925917031257`, 1, 1], 
      Hue[0.2326583911412925, 1, 1], 
      Hue[0.2326583911412925, 1, 1], 
      Hue[0.23183694761200105`, 1, 1], 
      Hue[0.23183694761200105`, 1, 1], 
      Hue[0.23019529296078733`, 1, 1], 
      Hue[0.23019529296078733`, 1, 1], 
      Hue[0.22978514896890515`, 1, 1], 
      Hue[0.22978514896890515`, 1, 1], 
      Hue[0.2293751172542093, 1, 1], 
      Hue[0.2293751172542093, 1, 1], 
      Hue[0.22855539943102374`, 1, 1], 
      Hue[0.22855539943102374`, 1, 1], 
      Hue[0.22691740731771895`, 1, 1], 
      Hue[0.22691740731771895`, 1, 1], 
      Hue[0.2236476811129669, 1, 1], 
      Hue[0.2236476811129669, 1, 1], 
      Hue[0.22323959369566776`, 1, 1], 
      Hue[0.22323959369566776`, 1, 1], 
      Hue[0.2228316529602289, 1, 1], 
      Hue[0.2228316529602289, 1, 1], 
      Hue[0.2220162199090814, 1, 1], 
      Hue[0.2220162199090814, 1, 1], 
      Hue[0.22038720570674894`, 1, 1], 
      Hue[0.22038720570674894`, 1, 1], 
      Hue[0.21713704446893195`, 1, 1], 
      Hue[0.21713704446893195`, 1, 1], 
      Hue[0.21675867941709884`, 1, 1], 
      Hue[0.21675867941709884`, 1, 1], 
      Hue[0.2163804703713593, 1, 1], 
      Hue[0.2163804703713593, 1, 1], 
      Hue[0.21562452670544383`, 1, 1], 
      Hue[0.21562452670544383`, 1, 1], 
      Hue[0.2141145815706027, 1, 1], 
      Hue[0.2141145815706027, 1, 1], 
      Hue[0.2111028105643298, 1, 1], 
      Hue[0.2111028105643298, 1, 1], 
      Hue[0.20511445765265002`, 1, 1], 
      Hue[0.20511445765265002`, 1, 1], 
      Hue[0.2047418469969295, 1, 1], 
      Hue[0.2047418469969295, 1, 1], 
      Hue[0.20436944028448414`, 1, 1], 
      Hue[0.20436944028448414`, 1, 1], 
      Hue[0.2036252441940914, 1, 1], 
      Hue[0.2036252441940914, 1, 1], 
      Hue[0.2021393594502079, 1, 1], 
      Hue[0.2021393594502079, 1, 1], 
      Hue[0.19917791790822034`, 1, 1], 
      Hue[0.19917791790822034`, 1, 1], 
      Hue[0.19329862431015446`, 1, 1], 
      Hue[0.19329862431015446`, 1, 1], 
      Hue[0.19290251120305676`, 1, 1], 
      Hue[0.19290251120305676`, 1, 1], 
      Hue[0.19250668498575932`, 1, 1], 
      Hue[0.19250668498575932`, 1, 1], 
      Hue[0.19171589889733506`, 1, 1], 
      Hue[0.19171589889733506`, 1, 1], 
      Hue[0.19013783118538485`, 1, 1], 
      Hue[0.19013783118538485`, 1, 1], 
      Hue[0.18699601635630125`, 1, 1], 
      Hue[0.18699601635630125`, 1, 1], 
      Hue[0.18077193135702532`, 1, 1], 
      Hue[0.18077193135702532`, 1, 1], 
      Hue[0.1685774743268824, 1, 1], 
      Hue[0.1685774743268824, 1, 1], 
      Hue[0.1682089558570103, 1, 1], 
      Hue[0.1682089558570103, 1, 1], 
      Hue[0.1678407785456452, 1, 1], 
      Hue[0.1678407785456452, 1, 1], 
      Hue[0.16710544991494183`, 1, 1], 
      Hue[0.16710544991494183`, 1, 1], 
      Hue[0.16563891364948222`, 1, 1], 
      Hue[0.16563891364948222`, 1, 1], 
      Hue[0.16272245258452597`, 1, 1], 
      Hue[0.16272245258452597`, 1, 1], 
      Hue[0.15695685884954436`, 1, 1], 
      Hue[0.15695685884954436`, 1, 1], 
      Hue[0.14569993963152494`, 1, 1], 
      Hue[0.14569993963152494`, 1, 1], 
      Hue[0.1256893641497182, 1, 1], 
      Hue[0.1256893641497182, 1, 1], 
      Hue[0.10539044861568864`, 1, 1], 
      Hue[0.10539044861568864`, 1, 1], 
      Hue[0.0876450931568508, 1, 1], 
      Hue[0.0876450931568508, 1, 1], 
      Hue[0.06954252802227788, 1, 1], 
      Hue[0.06954252802227788, 1, 1], 
      Hue[0.06927284345491812, 1, 1], 
      Hue[0.06927284345491812, 1, 1], 
      Hue[0.06900338285499233, 1, 1], 
      Hue[0.06900338285499233, 1, 1], 
      Hue[0.06846513025460943, 1, 1], 
      Hue[0.06846513025460943, 1, 1], 
      Hue[0.06739127631685438, 1, 1], 
      Hue[0.06739127631685438, 1, 1], 
      Hue[0.06525398821771008, 1, 1], 
      Hue[0.06525398821771008, 1, 1], 
      Hue[0.061019606605566934`, 1, 1], 
      Hue[0.061019606605566934`, 1, 1], 
      Hue[0.060756678464624916`, 1, 1], 
      Hue[0.060756678464624916`, 1, 1], 
      Hue[0.06049394778226548, 1, 1], 
      Hue[0.06049394778226548, 1, 1], 
      Hue[0.05996907547438736, 1, 1], 
      Hue[0.05996907547438736, 1, 1], 
      Hue[0.05892166385474366, 1, 1], 
      Hue[0.05892166385474366, 1, 1], 
      Hue[0.05683598678500967, 1, 1], 
      Hue[0.05683598678500967, 1, 1], 
      Hue[0.05269973189924155, 1, 1], 
      Hue[0.05269973189924155, 1, 1], 
      Hue[0.052459950972470006`, 1, 1], 
      Hue[0.052459950972470006`, 1, 1], 
      Hue[0.05222031886942626, 1, 1], 
      Hue[0.05222031886942626, 1, 1], 
      Hue[0.05174149845122005, 1, 1], 
      Hue[0.05174149845122005, 1, 1], 
      Hue[0.05078561399160116, 1, 1], 
      Hue[0.05078561399160116, 1, 1], 
      Hue[0.048880720514997804`, 1, 1], 
      Hue[0.048880720514997804`, 1, 1], 
      Hue[0.045097237160123156`, 1, 1], 
      Hue[0.045097237160123156`, 1, 1], 
      Hue[0.04486188666864251, 1, 1], 
      Hue[0.04486188666864251, 1, 1], 
      Hue[0.0446266636188734, 1, 1], 
      Hue[0.0446266636188734, 1, 1], 
      Hue[0.04415659718485805, 1, 1], 
      Hue[0.04415659718485805, 1, 1], 
      Hue[0.04321796437381417, 1, 1], 
      Hue[0.04321796437381417, 1, 1], 
      Hue[0.041346550324896265`, 1, 1], 
      Hue[0.041346550324896265`, 1, 1], 
      Hue[0.03762594234630728, 1, 1], 
      Hue[0.03762594234630728, 1, 1], 
      Hue[0.0373748435349179, 1, 1], 
      Hue[0.0373748435349179, 1, 1], 
      Hue[0.0371238695861517, 1, 1], 
      Hue[0.0371238695861517, 1, 1], 
      Hue[0.03662229292476846, 1, 1], 
      Hue[0.03662229292476846, 1, 1], 
      Hue[0.03562060110319645, 1, 1], 
      Hue[0.03562060110319645, 1, 1], 
      Hue[0.033622876197495356`, 1, 1], 
      Hue[0.033622876197495356`, 1, 1], 
      Hue[0.033373674812982916`, 1, 1], 
      Hue[0.033373674812982916`, 1, 1], 
      Hue[0.03312458492284089, 1, 1], 
      Hue[0.03312458492284089, 1, 1], 
      Hue[0.032626736295236373`, 1, 1], 
      Hue[0.032626736295236373`, 1, 1], 
      Hue[0.03163234035619638, 1, 1], 
      Hue[0.03163234035619638, 1, 1], 
      Hue[0.02964856766170453, 1, 1], 
      Hue[0.02964856766170453, 1, 1], 
      Hue[0.029401050448112517`, 1, 1], 
      Hue[0.029401050448112517`, 1, 1], 
      Hue[0.029153631445108612`, 1, 1], 
      Hue[0.029153631445108612`, 1, 1], 
      Hue[0.02865908476120701, 1, 1], 
      Hue[0.02865908476120701, 1, 1], 
      Hue[0.027671133531875818`, 1, 1], 
      Hue[0.027671133531875818`, 1, 1], 
      Hue[0.025699614688751985`, 1, 1], 
      Hue[0.025699614688751985`, 1, 1], 
      Hue[0.021772636012389598`, 1, 1], 
      Hue[0.021772636012389598`, 1, 1], 
      Hue[0.021544088339672494`, 1, 1], 
      Hue[0.021544088339672494`, 1, 1], 
      Hue[0.021315603374490812`, 1, 1], 
      Hue[0.021315603374490812`, 1, 1], 
      Hue[0.02085881890328848, 1, 1], 
      Hue[0.02085881890328848, 1, 1], 
      Hue[0.019945973164421232`, 1, 1], 
      Hue[0.019945973164421232`, 1, 1], 
      Hue[0.018123025596737827`, 1, 1], 
      Hue[0.018123025596737827`, 1, 1], 
      Hue[0.01789540144331209, 1, 1], 
      Hue[0.01789540144331209, 1, 1], 
      Hue[0.01766782936571196, 1, 1], 
      Hue[0.01766782936571196, 1, 1], 
      Hue[0.017212838786401, 1, 1], 
      Hue[0.017212838786401, 1, 1], 
      Hue[0.0163034533796867, 1, 1], 
      Hue[0.0163034533796867, 1, 1], 
      Hue[0.01448691729759583, 1, 1], 
      Hue[0.01448691729759583, 1, 1], 
      Hue[0.014260046896887183`, 1, 1], 
      Hue[0.014260046896887183`, 1, 1], 
      Hue[0.014033217984415945`, 1, 1], 
      Hue[0.014033217984415945`, 1, 1], 
      Hue[0.013579681982588002`, 1, 1], 
      Hue[0.013579681982588002`, 1, 1], 
      Hue[0.012673078787923713`, 1, 1], 
      Hue[0.012673078787923713`, 1, 1], 
      Hue[0.012446521808640131`, 1, 1], 
      Hue[0.012446521808640131`, 1, 1], 
      Hue[0.012220001037920017`, 1, 1], 
      Hue[0.012220001037920017`, 1, 1], 
      Hue[0.011767065484800906`, 1, 1], 
      Hue[0.011767065484800906`, 1, 1], 
      Hue[0.010861599877047007`, 1, 1], 
      Hue[0.010861599877047007`, 1, 1], 
      Hue[0.010635314110578184`, 1, 1], 
      Hue[0.010635314110578184`, 1, 1], 
      Hue[0.010409059281020351`, 1, 1], 
      Hue[0.010409059281020351`, 1, 1], 
      Hue[0.009956639798957091, 1, 1], 
      Hue[0.009956639798957091, 1, 1], 
      Hue[0.009052143113095098, 1, 1], 
      Hue[0.009052143113095098, 1, 1], 
      Hue[0.008826086411195665, 1, 1], 
      Hue[0.008826086411195665, 1, 1], 
      Hue[0.008600055381481772, 1, 1], 
      Hue[0.008600055381481772, 1, 1], 
      Hue[0.008148067708061165, 1, 1], 
      Hue[0.008148067708061165, 1, 1], 
      Hue[0.007244371496228025, 1, 1], 
      Hue[0.007244371496228025, 1, 1], 
      Hue[0.007022937079816843, 1, 1], 
      Hue[0.007022937079816843, 1, 1], 
      Hue[0.006801522295207302, 1, 1], 
      Hue[0.006801522295207302, 1, 1], 
      Hue[0.0063587491451943855`, 1, 1], 
      Hue[0.0063587491451943855`, 1, 1], 
      Hue[0.006137389541852155, 1, 1], 
      Hue[0.006137389541852155, 1, 1], 
      Hue[0.0059160470944297034`, 1, 1], 
      Hue[0.0059160470944297034`, 1, 1], 
      Hue[0.005473411192127852, 1, 1], 
      Hue[0.005473411192127852, 1, 1], 
      Hue[0.0052521164997800344`, 1, 1], 
      Hue[0.0052521164997800344`, 1, 1], 
      Hue[0.005030836488410986, 1, 1], 
      Hue[0.005030836488410986, 1, 1], 
      Hue[0.00458831803424194, 1, 1], 
      Hue[0.00458831803424194, 1, 1], 
      Hue[0.004367078354377108, 1, 1], 
      Hue[0.004367078354377108, 1, 1], 
      Hue[0.004145850881357151, 1, 1], 
      Hue[0.004145850881357151, 1, 1], 
      Hue[0.0037034300821993217`, 1, 1], 
      Hue[0.0037034300821993217`, 1, 1], 
      Hue[0.0034822355193328733`, 1, 1], 
      Hue[0.0034822355193328733`, 1, 1], 
      Hue[0.0032610506898499213`, 1, 1], 
      Hue[0.0032610506898499213`, 1, 1], 
      Hue[0.003039874975474352, 1, 1], 
      Hue[0.003039874975474352, 1, 1], 
      Hue[0.002818707757961507, 1, 1], 
      Hue[0.002818707757961507, 1, 1], 
      Hue[0.0025975484190960568`, 1, 1], 
      Hue[0.0025975484190960568`, 1, 1], 
      Hue[0.0023763963406898754`, 1, 1], 
      Hue[0.0023763963406898754`, 1, 1], 
      Hue[0.0021552509045799197`, 1, 1], 
      Hue[0.0021552509045799197`, 1, 1], 
      Hue[0.0019341114926260908`, 1, 1], 
      Hue[0.0019341114926260908`, 1, 1], 
      Hue[0.0017129774867091097`, 1, 1], 
      Hue[0.0017129774867091097`, 1, 1], 
      Hue[0.0014918482687283844`, 1, 1], 
      Hue[0.0014918482687283844`, 1, 1], 
      Hue[0.0012707232205998733`, 1, 1], 
      Hue[0.0012707232205998733`, 1, 1], 
      Hue[0.001049601724253957, 1, 1], 
      Hue[0.001049601724253957, 1, 1], 
      Hue[0.0008284831616332965, 1, 1], 
      Hue[0.0008284831616332965, 1, 1], 
      Hue[0.0006073669146907009, 1, 1], 
      Hue[0.0006073669146907009, 1, 1], 
      Hue[0.00038625236538699054`, 1, 1], 
      Hue[0.00038625236538699054`, 1, 1], 
      Hue[0.00016513889568885723`, 1, 1], 
      Hue[0.00016513889568885723`, 1, 1], 
      Hue[0.9999440258875667, 1, 1], 
      Hue[0.9999440258875667, 1, 1], 
      Hue[0.9997229127229926, 1, 1], 
      Hue[0.9997229127229926, 1, 1], 
      Hue[0.9995017987839381, 1, 1], 
      Hue[0.9995017987839381, 1, 1], 
      Hue[0.9992806834523718, 1, 1], 
      Hue[0.9992806834523718, 1, 1], 
      Hue[0.9990595661102578, 1, 1], 
      Hue[0.9990595661102578, 1, 1], 
      Hue[0.9988384461395532, 1, 1], 
      Hue[0.9988384461395532, 1, 1], 
      Hue[0.9986173229222058, 1, 1], 
      Hue[0.9986173229222058, 1, 1], 
      Hue[0.9983961958401524, 1, 1], 
      Hue[0.9983961958401524, 1, 1], 
      Hue[0.9981750642753163, 1, 1], 
      Hue[0.9981750642753163, 1, 1], 
      Hue[0.9979539276096054, 1, 1], 
      Hue[0.9979539276096054, 1, 1], 
      Hue[0.9977327852249099, 1, 1], 
      Hue[0.9977327852249099, 1, 1], 
      Hue[0.9975116365031003, 1, 1], 
      Hue[0.9975116365031003, 1, 1], 
      Hue[0.997290480826025, 1, 1], 
      Hue[0.997290480826025, 1, 1], 
      Hue[0.9970693175755087, 1, 1], 
      Hue[0.9970693175755087, 1, 1], 
      Hue[0.9966269658813178, 1, 1], 
      Hue[0.9966269658813178, 1, 1], 
      Hue[0.9964057762011531, 1, 1], 
      Hue[0.9964057762011531, 1, 1], 
      Hue[0.9961845764745626, 1, 1], 
      Hue[0.9961845764745626, 1, 1], 
      Hue[0.9957421444087574, 1, 1], 
      Hue[0.9957421444087574, 1, 1], 
      Hue[0.9955209108327753, 1, 1], 
      Hue[0.9955209108327753, 1, 1], 
      Hue[0.9952996647368284, 1, 1], 
      Hue[0.9952996647368284, 1, 1], 
      Hue[0.9948571325110459, 1, 1], 
      Hue[0.9948571325110459, 1, 1], 
      Hue[0.9946358451440981, 1, 1], 
      Hue[0.9946358451440981, 1, 1], 
      Hue[0.9944145427829568, 1, 1], 
      Hue[0.9944145427829568, 1, 1], 
      Hue[0.9939718906033176, 1, 1], 
      Hue[0.9939718906033176, 1, 1], 
      Hue[0.9930863790880589, 1, 1], 
      Hue[0.9930863790880589, 1, 1], 
      Hue[0.9928461651955212, 1, 1], 
      Hue[0.9928461651955212, 1, 1], 
      Hue[0.9926059277675018, 1, 1], 
      Hue[0.9926059277675018, 1, 1], 
      Hue[0.9921253791427431, 1, 1], 
      Hue[0.9921253791427431, 1, 1], 
      Hue[0.9911639646767813, 1, 1], 
      Hue[0.9911639646767813, 1, 1], 
      Hue[0.9909235403594647, 1, 1], 
      Hue[0.9909235403594647, 1, 1], 
      Hue[0.9906830861794046, 1, 1], 
      Hue[0.9906830861794046, 1, 1], 
      Hue[0.9902020850652838, 1, 1], 
      Hue[0.9902020850652838, 1, 1], 
      Hue[0.9892396896536435, 1, 1], 
      Hue[0.9892396896536435, 1, 1], 
      Hue[0.9889990042677944, 1, 1], 
      Hue[0.9889990042677944, 1, 1], 
      Hue[0.9887582826843183, 1, 1], 
      Hue[0.9887582826843183, 1, 1], 
      Hue[0.9882767277544471, 1, 1], 
      Hue[0.9882767277544471, 1, 1], 
      Hue[0.9873131486444305, 1, 1], 
      Hue[0.9873131486444305, 1, 1], 
      Hue[0.9853839357017069, 1, 1], 
      Hue[0.9853839357017069, 1, 1], 
      Hue[0.9851425759700883, 1, 1], 
      Hue[0.9851425759700883, 1, 1], 
      Hue[0.9849011673421513, 1, 1], 
      Hue[0.9849011673421513, 1, 1], 
      Hue[0.984418200216528, 1, 1], 
      Hue[0.984418200216528, 1, 1], 
      Hue[0.9834516442098782, 1, 1], 
      Hue[0.9834516442098782, 1, 1], 
      Hue[0.9815158667939559, 1, 1], 
      Hue[0.9815158667939559, 1, 1], 
      Hue[0.9812736291913834, 1, 1], 
      Hue[0.9812736291913834, 1, 1], 
      Hue[0.9810313299456908, 1, 1], 
      Hue[0.9810313299456908, 1, 1], 
      Hue[0.9805465433306404, 1, 1], 
      Hue[0.9805465433306404, 1, 1], 
      Hue[0.9795761952326396, 1, 1], 
      Hue[0.9795761952326396, 1, 1], 
      Hue[0.9776322203763465, 1, 1], 
      Hue[0.9776322203763465, 1, 1], 
      Hue[0.9774051575452036, 1, 1], 
      Hue[0.9774051575452036, 1, 1], 
      Hue[0.9771780299268937, 1, 1], 
      Hue[0.9771780299268937, 1, 1], 
      Hue[0.9767235777199724, 1, 1], 
      Hue[0.9767235777199724, 1, 1], 
      Hue[0.9758138671519275, 1, 1], 
      Hue[0.9758138671519275, 1, 1], 
      Hue[0.9739910750405399, 1, 1], 
      Hue[0.9739910750405399, 1, 1], 
      Hue[0.9703308331526667, 1, 1], 
      Hue[0.9703308331526667, 1, 1], 
      Hue[0.9701013726173703, 1, 1], 
      Hue[0.9701013726173703, 1, 1], 
      Hue[0.9698718263196648, 1, 1], 
      Hue[0.9698718263196648, 1, 1], 
      Hue[0.9694124737999488, 1, 1], 
      Hue[0.9694124737999488, 1, 1], 
      Hue[0.9684927105904569, 1, 1], 
      Hue[0.9684927105904569, 1, 1], 
      Hue[0.9666488035074778, 1, 1], 
      Hue[0.9666488035074778, 1, 1], 
      Hue[0.9629422795583775, 1, 1], 
      Hue[0.9629422795583775, 1, 1], 
      Hue[0.96269025198017, 1, 1], 
      Hue[0.96269025198017, 1, 1], 
      Hue[0.9624380986753572, 1, 1], 
      Hue[0.9624380986753572, 1, 1], 
      Hue[0.9619334114888105, 1, 1], 
      Hue[0.9619334114888105, 1, 1], 
      Hue[0.9609224910093952, 1, 1], 
      Hue[0.9609224910093952, 1, 1], 
      Hue[0.9588942749542262, 1, 1], 
      Hue[0.9588942749542262, 1, 1], 
      Hue[0.9548108126517543, 1, 1], 
      Hue[0.9548108126517543, 1, 1], 
      Hue[0.946523461974924, 1, 1], 
      Hue[0.946523461974924, 1, 1], 
      Hue[0.9462664435525957, 1, 1], 
      Hue[0.9462664435525957, 1, 1], 
      Hue[0.946009250984037, 1, 1], 
      Hue[0.946009250984037, 1, 1], 
      Hue[0.9454943401255437, 1, 1], 
      Hue[0.9454943401255437, 1, 1], 
      Hue[0.944462392535523, 1, 1], 
      Hue[0.944462392535523, 1, 1], 
      Hue[0.9423898098455793, 1, 1], 
      Hue[0.9423898098455793, 1, 1], 
      Hue[0.9382084208650068, 1, 1], 
      Hue[0.9382084208650068, 1, 1], 
      Hue[0.9296889548412226, 1, 1], 
      Hue[0.9296889548412226, 1, 1], 
      Hue[0.9131577192100314, 1, 1], 
      Hue[0.9131577192100314, 1, 1], 
      Hue[0.8941248259641966, 1, 1], 
      Hue[0.8941248259641966, 1, 1], 
      Hue[0.8751555374447492, 1, 1], 
      Hue[0.8751555374447492, 1, 1], 
      Hue[0.8553207001162826, 1, 1], 
      Hue[0.8553207001162826, 1, 1], 
      Hue[0.8549734798782153, 1, 1], 
      Hue[0.8549734798782153, 1, 1], 
      Hue[0.8546258934543335, 1, 1], 
      Hue[0.8546258934543335, 1, 1], 
      Hue[0.8539296221742921, 1, 1], 
      Hue[0.8539296221742921, 1, 1], 
      Hue[0.8525326872315117, 1, 1], 
      Hue[0.8525326872315117, 1, 1], 
      Hue[0.8497212657634408, 1, 1], 
      Hue[0.8497212657634408, 1, 1], 
      Hue[0.8440284786662016, 1, 1], 
      Hue[0.8440284786662016, 1, 1], 
      Hue[0.8323672368187282, 1, 1], 
      Hue[0.8323672368187282, 1, 1], 
      Hue[0.832021825341367, 1, 1], 
      Hue[0.832021825341367, 1, 1], 
      Hue[0.8316761129089216, 1, 1], 
      Hue[0.8316761129089216, 1, 1], 
      Hue[0.830983787289033, 1, 1], 
      Hue[0.830983787289033, 1, 1], 
      Hue[0.8295955482610037, 1, 1], 
      Hue[0.8295955482610037, 1, 1], 
      Hue[0.8268048478525195, 1, 1], 
      Hue[0.8268048478525195, 1, 1], 
      Hue[0.8211677004737814, 1, 1], 
      Hue[0.8211677004737814, 1, 1], 
      Hue[0.8208129603290649, 1, 1], 
      Hue[0.8208129603290649, 1, 1], 
      Hue[0.8204579401889733, 1, 1], 
      Hue[0.8204579401889733, 1, 1], 
      Hue[0.8197470631114766, 1, 1], 
      Hue[0.8197470631114766, 1, 1], 
      Hue[0.8183219845703645, 1, 1], 
      Hue[0.8183219845703645, 1, 1], 
      Hue[0.8154587196009777, 1, 1], 
      Hue[0.8154587196009777, 1, 1], 
      Hue[0.8096813911841393, 1, 1], 
      Hue[0.8096813911841393, 1, 1], 
      Hue[0.8092875819890374, 1, 1], 
      Hue[0.8092875819890374, 1, 1], 
      Hue[0.808893478770385, 1, 1], 
      Hue[0.808893478770385, 1, 1], 
      Hue[0.8081043957324173, 1, 1], 
      Hue[0.8081043957324173, 1, 1], 
      Hue[0.8065227621999305, 1, 1], 
      Hue[0.8065227621999305, 1, 1], 
      Hue[0.8033459466645283, 1, 1], 
      Hue[0.8033459466645283, 1, 1], 
      Hue[0.802947603375167, 1, 1], 
      Hue[0.802947603375167, 1, 1], 
      Hue[0.8025489892916752, 1, 1], 
      Hue[0.8025489892916752, 1, 1], 
      Hue[0.8017509549521107, 1, 1], 
      Hue[0.8017509549521107, 1, 1], 
      Hue[0.8001517056838474, 1, 1], 
      Hue[0.8001517056838474, 1, 1], 
      Hue[0.7969408469898973, 1, 1], 
      Hue[0.7969408469898973, 1, 1], 
      Hue[0.7965383632273197, 1, 1], 
      Hue[0.7965383632273197, 1, 1], 
      Hue[0.7961356348032806, 1, 1], 
      Hue[0.7961356348032806, 1, 1], 
      Hue[0.7953294508865575, 1, 1], 
      Hue[0.7953294508865575, 1, 1], 
      Hue[0.7937142237832816, 1, 1], 
      Hue[0.7937142237832816, 1, 1], 
      Hue[0.7904727333124814, 1, 1], 
      Hue[0.7904727333124814, 1, 1], 
      Hue[0.7900665479973398, 1, 1], 
      Hue[0.7900665479973398, 1, 1], 
      Hue[0.7896601468913598, 1, 1], 
      Hue[0.7896601468913598, 1, 1], 
      Hue[0.78884670487822, 1, 1], 
      Hue[0.78884670487822, 1, 1], 
      Hue[0.7872173151887687, 1, 1], 
      Hue[0.7872173151887687, 1, 1], 
      Hue[0.7839489494382963, 1, 1], 
      Hue[0.7839489494382963, 1, 1], 
      Hue[0.7835666629003442, 1, 1], 
      Hue[0.7835666629003442, 1, 1], 
      Hue[0.7831842154488271, 1, 1], 
      Hue[0.7831842154488271, 1, 1], 
      Hue[0.7824188444458442, 1, 1], 
      Hue[0.7824188444458442, 1, 1], 
      Hue[0.7808862448877505, 1, 1], 
      Hue[0.7808862448877505, 1, 1], 
      Hue[0.7778139956224613, 1, 1], 
      Hue[0.7778139956224613, 1, 1], 
      Hue[0.7774293370436219, 1, 1], 
      Hue[0.7774293370436219, 1, 1], 
      Hue[0.7770445448384399, 1, 1], 
      Hue[0.7770445448384399, 1, 1], 
      Hue[0.7762745665778351, 1, 1], 
      Hue[0.7762745665778351, 1, 1], 
      Hue[0.774733084163743, 1, 1], 
      Hue[0.774733084163743, 1, 1], 
      Hue[0.7716444162444386, 1, 1], 
      Hue[0.7716444162444386, 1, 1], 
      Hue[0.7712578333211608, 1, 1], 
      Hue[0.7712578333211608, 1, 1], 
      Hue[0.770871145491256, 1, 1], 
      Hue[0.770871145491256, 1, 1], 
      Hue[0.7700974624607736, 1, 1], 
      Hue[0.7700974624607736, 1, 1], 
      Hue[0.7685489186047042, 1, 1], 
      Hue[0.7685489186047042, 1, 1], 
      Hue[0.7681615480840327, 1, 1], 
      Hue[0.7681615480840327, 1, 1], 
      Hue[0.7677740874657797, 1, 1], 
      Hue[0.7677740874657797, 1, 1], 
      Hue[0.7669989034180303, 1, 1], 
      Hue[0.7669989034180303, 1, 1], 
      Hue[0.765447536657087, 1, 1], 
      Hue[0.765447536657087, 1, 1], 
      Hue[0.7650594979075211, 1, 1], 
      Hue[0.7650594979075211, 1, 1], 
      Hue[0.7646713841180712, 1, 1], 
      Hue[0.7646713841180712, 1, 1], 
      Hue[0.7638949390135434, 1, 1], 
      Hue[0.7638949390135434, 1, 1], 
      Hue[0.7635066115033468, 1, 1], 
      Hue[0.7635066115033468, 1, 1], 
      Hue[0.7631182165628669, 1, 1], 
      Hue[0.7631182165628669, 1, 1], 
      Hue[0.7623412320336995, 1, 1], 
      Hue[0.7623412320336995, 1, 1], 
      Hue[0.7619526462733915, 1, 1], 
      Hue[0.7619526462733915, 1, 1], 
      Hue[0.7615640007393966, 1, 1], 
      Hue[0.7615640007393966, 1, 1], 
      Hue[0.7607865380363968, 1, 1], 
      Hue[0.7607865380363968, 1, 1], 
      Hue[0.7592309800296734, 1, 1], 
      Hue[0.7592309800296734, 1, 1], 
      Hue[0.758849596113294, 1, 1], 
      Hue[0.758849596113294, 1, 1], 
      Hue[0.7584681695412607, 1, 1], 
      Hue[0.7584681695412607, 1, 1], 
      Hue[0.7577051957396829, 1, 1], 
      Hue[0.7577051957396829, 1, 1], 
      Hue[0.7573236521690917, 1, 1], 
      Hue[0.7573236521690917, 1, 1], 
      Hue[0.7569420732606006, 1, 1], 
      Hue[0.7569420732606006, 1, 1], 
      Hue[0.7561788167647769, 1, 1], 
      Hue[0.7561788167647769, 1, 1], 
      Hue[0.7557971428483347, 1, 1], 
      Hue[0.7557971428483347, 1, 1], 
      Hue[0.755415440935619, 1, 1], 
      Hue[0.755415440935619, 1, 1], 
      Hue[0.7550337128651892, 1, 1], 
      Hue[0.7550337128651892, 1, 1], 
      Hue[0.7546519604767108, 1, 1], 
      Hue[0.7546519604767108, 1, 1], 
      Hue[0.7542701856108777, 1, 1], 
      Hue[0.7542701856108777, 1, 1], 
      Hue[0.7538883901093358, 1, 1], 
      Hue[0.7538883901093358, 1, 1], 
      Hue[0.7535065758146038, 1, 1], 
      Hue[0.7535065758146038, 1, 1], 
      Hue[0.7531247445699961, 1, 1], 
      Hue[0.7531247445699961, 1, 1], 
      Hue[0.7527428982195451, 1, 1], 
      Hue[0.7527428982195451, 1, 1], 
      Hue[0.7523610386079226, 1, 1], 
      Hue[0.7523610386079226, 1, 1], 
      Hue[0.7519791675803619, 1, 1], 
      Hue[0.7519791675803619, 1, 1], 
      Hue[0.7515972869825804, 1, 1], 
      Hue[0.7515972869825804, 1, 1], 
      Hue[0.7512153986607009, 1, 1], 
      Hue[0.7512153986607009, 1, 1], 
      Hue[0.7508335044611727, 1, 1], 
      Hue[0.7508335044611727, 1, 1], 
      Hue[0.7504516062306952, 1, 1], 
      Hue[0.7504516062306952, 1, 1], 
      Hue[0.7500697058161379, 1, 1], 
      Hue[0.7500697058161379, 1, 1], 
      Hue[0.749687805064463, 1, 1], 
      Hue[0.749687805064463, 1, 1], 
      Hue[0.7493059058226473, 1, 1], 
      Hue[0.7493059058226473, 1, 1], 
      Hue[0.7489240099376031, 1, 1], 
      Hue[0.7489240099376031, 1, 1], 
      Hue[0.748542119256101, 1, 1], 
      Hue[0.748542119256101, 1, 1], 
      Hue[0.7481602356246906, 1, 1], 
      Hue[0.7481602356246906, 1, 1], 
      Hue[0.7477783608896231, 1, 1], 
      Hue[0.7477783608896231, 1, 1], 
      Hue[0.7473964968967727, 1, 1], 
      Hue[0.7473964968967727, 1, 1], 
      Hue[0.7470146454915587, 1, 1], 
      Hue[0.7470146454915587, 1, 1], 
      Hue[0.7466328085188669, 1, 1], 
      Hue[0.7466328085188669, 1, 1], 
      Hue[0.7462509878229721, 1, 1], 
      Hue[0.7462509878229721, 1, 1], 
      Hue[0.7458691852474597, 1, 1], 
      Hue[0.7458691852474597, 1, 1], 
      Hue[0.7454874026351481, 1, 1], 
      Hue[0.7454874026351481, 1, 1], 
      Hue[0.7451056418280108, 1, 1], 
      Hue[0.7451056418280108, 1, 1], 
      Hue[0.7447239046670979, 1, 1], 
      Hue[0.7447239046670979, 1, 1], 
      Hue[0.7439605086430686, 1, 1], 
      Hue[0.7439605086430686, 1, 1], 
      Hue[0.7435788534567404, 1, 1], 
      Hue[0.7435788534567404, 1, 1], 
      Hue[0.743197229270059, 1, 1], 
      Hue[0.743197229270059, 1, 1], 
      Hue[0.7424340812353439, 1, 1], 
      Hue[0.7424340812353439, 1, 1], 
      Hue[0.7420525610536186, 1, 1], 
      Hue[0.7420525610536186, 1, 1], 
      Hue[0.7416710792040031, 1, 1], 
      Hue[0.7416710792040031, 1, 1], 
      Hue[0.7409082378164619, 1, 1], 
      Hue[0.7409082378164619, 1, 1], 
      Hue[0.740526881931904, 1, 1], 
      Hue[0.740526881931904, 1, 1], 
      Hue[0.7401455716860399, 1, 1], 
      Hue[0.7401455716860399, 1, 1], 
      Hue[0.7393830953965116, 1, 1], 
      Hue[0.7393830953965116, 1, 1], 
      Hue[0.7390019329908355, 1, 1], 
      Hue[0.7390019329908355, 1, 1], 
      Hue[0.7386208234996777, 1, 1], 
      Hue[0.7386208234996777, 1, 1], 
      Hue[0.7378587705129496, 1, 1], 
      Hue[0.7378587705129496, 1, 1], 
      Hue[0.7374778306375714, 1, 1], 
      Hue[0.7374778306375714, 1, 1], 
      Hue[0.7370969509169467, 1, 1], 
      Hue[0.7370969509169467, 1, 1], 
      Hue[0.7363353791531093, 1, 1], 
      Hue[0.7363353791531093, 1, 1], 
      Hue[0.7348130366775789, 1, 1], 
      Hue[0.7348130366775789, 1, 1], 
      Hue[0.7344004115686685, 1, 1], 
      Hue[0.7344004115686685, 1, 1], 
      Hue[0.7339878743030456, 1, 1], 
      Hue[0.7339878743030456, 1, 1], 
      Hue[0.7331630723875937, 1, 1], 
      Hue[0.7331630723875937, 1, 1], 
      Hue[0.7315146223679854, 1, 1], 
      Hue[0.7315146223679854, 1, 1], 
      Hue[0.7311027633829322, 1, 1], 
      Hue[0.7311027633829322, 1, 1], 
      Hue[0.7306910102921902, 1, 1], 
      Hue[0.7306910102921902, 1, 1], 
      Hue[0.729867830725456, 1, 1], 
      Hue[0.729867830725456, 1, 1], 
      Hue[0.7282228402733528, 1, 1], 
      Hue[0.7282228402733528, 1, 1], 
      Hue[0.7278118907005582, 1, 1], 
      Hue[0.7278118907005582, 1, 1], 
      Hue[0.7274010647432104, 1, 1], 
      Hue[0.7274010647432104, 1, 1], 
      Hue[0.726579792426144, 1, 1], 
      Hue[0.726579792426144, 1, 1], 
      Hue[0.7249388270964126, 1, 1], 
      Hue[0.7249388270964126, 1, 1], 
      Hue[0.7216636955311848, 1, 1], 
      Hue[0.7216636955311848, 1, 1], 
      Hue[0.7212549829546681, 1, 1], 
      Hue[0.7212549829546681, 1, 1], 
      Hue[0.7208464282417213, 1, 1], 
      Hue[0.7208464282417213, 1, 1], 
      Hue[0.7200298007228656, 1, 1], 
      Hue[0.7200298007228656, 1, 1], 
      Hue[0.7183985311005171, 1, 1], 
      Hue[0.7183985311005171, 1, 1], 
      Hue[0.7151443891844453, 1, 1], 
      Hue[0.7151443891844453, 1, 1], 
      Hue[0.7147384477657693, 1, 1], 
      Hue[0.7147384477657693, 1, 1], 
      Hue[0.7143326965547425, 1, 1], 
      Hue[0.7143326965547425, 1, 1], 
      Hue[0.713521772548821, 1, 1], 
      Hue[0.713521772548821, 1, 1], 
      Hue[0.7119022922564333, 1, 1], 
      Hue[0.7119022922564333, 1, 1], 
      Hue[0.7086732273419383, 1, 1], 
      Hue[0.7086732273419383, 1, 1], 
      Hue[0.7082974065478993, 1, 1], 
      Hue[0.7082974065478993, 1, 1], 
      Hue[0.7079217775613169, 1, 1], 
      Hue[0.7079217775613169, 1, 1], 
      Hue[0.7071711008642876, 1, 1], 
      Hue[0.7071711008642876, 1, 1], 
      Hue[0.7056721131307586, 1, 1], 
      Hue[0.7056721131307586, 1, 1], 
      Hue[0.7026839183442184, 1, 1], 
      Hue[0.7026839183442184, 1, 1], 
      Hue[0.6967490903085825, 1, 1], 
      Hue[0.6967490903085825, 1, 1], 
      Hue[0.6963800957883017, 1, 1], 
      Hue[0.6963800957883017, 1, 1], 
      Hue[0.6960113360845944, 1, 1], 
      Hue[0.6960113360845944, 1, 1], 
      Hue[0.6952745259565233, 1, 1], 
      Hue[0.6952745259565233, 1, 1], 
      Hue[0.6938037761512619, 1, 1], 
      Hue[0.6938037761512619, 1, 1], 
      Hue[0.6908740178550813, 1, 1], 
      Hue[0.6908740178550813, 1, 1], 
      Hue[0.6850634268348434, 1, 1], 
      Hue[0.6850634268348434, 1, 1], 
      Hue[0.6846723194047373, 1, 1], 
      Hue[0.6846723194047373, 1, 1], 
      Hue[0.6842815282627351, 1, 1], 
      Hue[0.6842815282627351, 1, 1], 
      Hue[0.6835008995719081, 1, 1], 
      Hue[0.6835008995719081, 1, 1], 
      Hue[0.6819434889866574, 1, 1], 
      Hue[0.6819434889866574, 1, 1], 
      Hue[0.6788443042180083, 1, 1], 
      Hue[0.6788443042180083, 1, 1], 
      Hue[0.6727103116791255, 1, 1], 
      Hue[0.6727103116791255, 1, 1], 
      Hue[0.6607119238065423, 1, 1], 
      Hue[0.6607119238065423, 1, 1], 
      Hue[0.6603496703678688, 1, 1], 
      Hue[0.6603496703678688, 1, 1], 
      Hue[0.6599877717876945, 1, 1], 
      Hue[0.6599877717876945, 1, 1], 
      Hue[0.6592650408555486, 1, 1], 
      Hue[0.6592650408555486, 1, 1], 
      Hue[0.6578238549160509, 1, 1], 
      Hue[0.6578238549160509, 1, 1], 
      Hue[0.6549586665777667, 1, 1], 
      Hue[0.6549586665777667, 1, 1], 
      Hue[0.6492975339672553, 1, 1], 
      Hue[0.6492975339672553, 1, 1], 
      Hue[0.6382543738240635, 1, 1], 
      Hue[0.6382543738240635, 1, 1], 
      Hue[0.6186381533904819, 1, 1], 
      Hue[0.6186381533904819, 1, 1], 
      Hue[0.5987450483813834, 1, 1], 
      Hue[0.5987450483813834, 1, 1], 
      Hue[0.581327200613792, 1, 1], 
      Hue[0.581327200613792, 1, 1], 
      Hue[0.5651708351621385, 1, 1], 
      Hue[0.5651708351621385, 1, 1], 
      Hue[0.5649038110512727, 1, 1], 
      Hue[0.5649038110512727, 1, 1], 
      Hue[0.5646369988185738, 1, 1], 
      Hue[0.5646369988185738, 1, 1], 
      Hue[0.5641040066386009, 1, 1], 
      Hue[0.5641040066386009, 1, 1], 
      Hue[0.5630405279707895, 1, 1], 
      Hue[0.5630405279707895, 1, 1], 
      Hue[0.5609234057403701, 1, 1], 
      Hue[0.5609234057403701, 1, 1], 
      Hue[0.5567270000080751, 1, 1], 
      Hue[0.5567270000080751, 1, 1], 
      Hue[0.5564663404531357, 1, 1], 
      Hue[0.5564663404531357, 1, 1], 
      Hue[0.5562058659678826, 1, 1], 
      Hue[0.5562058659678826, 1, 1], 
      Hue[0.5556854688584882, 1, 1], 
      Hue[0.5556854688584882, 1, 1], 
      Hue[0.5546468586555297, 1, 1], 
      Hue[0.5546468586555297, 1, 1], 
      Hue[0.5525781870036823, 1, 1], 
      Hue[0.5525781870036823, 1, 1], 
      Hue[0.548473542566891, 1, 1], 
      Hue[0.548473542566891, 1, 1], 
      Hue[0.5482354485858063, 1, 1], 
      Hue[0.5482354485858063, 1, 1], 
      Hue[0.5479974925493978, 1, 1], 
      Hue[0.5479974925493978, 1, 1], 
      Hue[0.5475219916105107, 1, 1], 
      Hue[0.5475219916105107, 1, 1], 
      Hue[0.5465726153787527, 1, 1], 
      Hue[0.5465726153787527, 1, 1], 
      Hue[0.5446802145516939, 1, 1], 
      Hue[0.5446802145516939, 1, 1], 
      Hue[0.540919614724254, 1, 1], 
      Hue[0.540919614724254, 1, 1], 
      Hue[0.5406856011644792, 1, 1], 
      Hue[0.5406856011644792, 1, 1], 
      Hue[0.5404517040519298, 1, 1], 
      Hue[0.5404517040519298, 1, 1], 
      Hue[0.5399842564964968, 1, 1], 
      Hue[0.5399842564964968, 1, 1], 
      Hue[0.539050729373585, 1, 1], 
      Hue[0.539050729373585, 1, 1], 
      Hue[0.5371889977559043, 1, 1], 
      Hue[0.5371889977559043, 1, 1], 
      Hue[0.5334856337081378, 1, 1], 
      Hue[0.5334856337081378, 1, 1], 
      Hue[0.5332356702272592, 1, 1], 
      Hue[0.5332356702272592, 1, 1], 
      Hue[0.5329858185148917, 1, 1], 
      Hue[0.5329858185148917, 1, 1], 
      Hue[0.5324864470328554, 1, 1], 
      Hue[0.5324864470328554, 1, 1], 
      Hue[0.5314890083184363, 1, 1], 
      Hue[0.5314890083184363, 1, 1], 
      Hue[0.5294991599976908, 1, 1], 
      Hue[0.5294991599976908, 1, 1], 
      Hue[0.5292508841040008, 1, 1], 
      Hue[0.5292508841040008, 1, 1], 
      Hue[0.529002706565882, 1, 1], 
      Hue[0.529002706565882, 1, 1], 
      Hue[0.5285066432145402, 1, 1], 
      Hue[0.5285066432145402, 1, 1], 
      Hue[0.5275156600365587, 1, 1], 
      Hue[0.5275156600365587, 1, 1], 
      Hue[0.5255380810443776, 1, 1], 
      Hue[0.5255380810443776, 1, 1], 
      Hue[0.525291278752398, 1, 1], 
      Hue[0.525291278752398, 1, 1], 
      Hue[0.5250445614850563, 1, 1], 
      Hue[0.5250445614850563, 1, 1], 
      Hue[0.5245513787022311, 1, 1], 
      Hue[0.5245513787022311, 1, 1], 
      Hue[0.5235659969060595, 1, 1], 
      Hue[0.5235659969060595, 1, 1], 
      Hue[0.5215989827293759, 1, 1], 
      Hue[0.5215989827293759, 1, 1], 
      Hue[0.5213534413229897, 1, 1], 
      Hue[0.5213534413229897, 1, 1], 
      Hue[0.5211079716861946, 1, 1], 
      Hue[0.5211079716861946, 1, 1], 
      Hue[0.5206172444172416, 1, 1], 
      Hue[0.5206172444172416, 1, 1], 
      Hue[0.5196366147834943, 1, 1], 
      Hue[0.5196366147834943, 1, 1], 
      Hue[0.5176784704295874, 1, 1], 
      Hue[0.5176784704295874, 1, 1], 
      Hue[0.517671162797467, 1, 1], 
      Hue[0.517671162797467, 1, 1], 
      Hue[0.5176638552183818, 1, 1], 
      Hue[0.5176638552183818, 1, 1], 
      Hue[0.5176492402192279, 1, 1], 
      Hue[0.5176492402192279, 1, 1], 
      Hue[0.5176200108563732, 1, 1], 
      Hue[0.5176200108563732, 1, 1], 
      Hue[0.5175615546675576, 1, 1], 
      Hue[0.5175615546675576, 1, 1], 
      Hue[0.5174446523981632, 1, 1], 
      Hue[0.5174446523981632, 1, 1], 
      Hue[0.5174373464522712, 1, 1], 
      Hue[0.5174373464522712, 1, 1], 
      Hue[0.5174300405587118, 1, 1], 
      Hue[0.5174300405587118, 1, 1], 
      Hue[0.5174154289285023, 1, 1], 
      Hue[0.5174154289285023, 1, 1], 
      Hue[0.5173862062951068, 1, 1], 
      Hue[0.5173862062951068, 1, 1], 
      Hue[0.5173277635314939, 1, 1], 
      Hue[0.5173277635314939, 1, 1], 
      Hue[0.5173204584202873, 1, 1], 
      Hue[0.5173204584202873, 1, 1], 
      Hue[0.5173131533610619, 1, 1], 
      Hue[0.5173131533610619, 1, 1], 
      Hue[0.5172985433984671, 1, 1], 
      Hue[0.5172985433984671, 1, 1], 
      Hue[0.5172693240960875, 1, 1], 
      Hue[0.5172693240960875, 1, 1], 
      Hue[0.5172620194001166, 1, 1], 
      Hue[0.5172620194001166, 1, 1], 
      Hue[0.5172547147559514, 1, 1], 
      Hue[0.5172547147559514, 1, 1], 
      Hue[0.5172401056229503, 1, 1], 
      Hue[0.5172401056229503, 1, 1], 
      Hue[0.5172328011340706, 1, 1], 
      Hue[0.5172328011340706, 1, 1], 
      Hue[0.5172254966969089, 1, 1], 
      Hue[0.5172254966969089, 1, 1], 
      Hue[0.5172181923114431, 1, 1], 
      Hue[0.5172181923114431, 1, 1], 
      Hue[0.5172108879776514, 1, 1]}]},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 1.2},
  Epilog->LineBox[{{-3, 1}, {3, 1}}],
  Frame->True,
  PlotRange->{-0.2, 1.7},
  PlotRangeClipping->True,
  PlotRangePadding->{
    Scaled[0.02], 
    Scaled[0.02]}]], "Output",
 CellChangeTimes->{{3.391857542859375*^9, 3.3918575586875*^9}},
 CellLabel->"Out[46]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"spinorfunction", "[", "x_", "]"}], ":=", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{
     RowBox[{"Cos", "[", 
      RowBox[{"3", " ", "x"}], "]"}], " ", 
     SuperscriptBox["\[ExponentialE]", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         FractionBox["1", "3"]}], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"x", "-", 
           FractionBox["1", "4"]}], ")"}], "2"]}], "+", 
       RowBox[{"\[ImaginaryI]", " ", "x"}]}]]}], ",", 
    RowBox[{
     RowBox[{"Sin", "[", 
      RowBox[{"4", " ", "x"}], "]"}], " ", 
     SuperscriptBox["\[ExponentialE]", 
      RowBox[{
       RowBox[{
        RowBox[{"-", 
         FractionBox["1", "2"]}], " ", 
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{"x", "+", 
           FractionBox["1", "4"]}], ")"}], "2"]}], "+", 
       RowBox[{"2", " ", "\[ImaginaryI]", " ", "x"}]}]]}]}], "}"}]}], "\n", 
 RowBox[{"QSpinorPlot", "[", 
  RowBox[{
   RowBox[{"Evaluate", "[", 
    RowBox[{"spinorfunction", "[", "x", "]"}], "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", 
     RowBox[{"-", "3."}], ",", "3."}], "}"}], ",", 
   RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
   RowBox[{"Frame", "\[Rule]", "True"}], ",", 
   RowBox[{"Axes", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"True", ",", "False"}], "}"}]}]}], "]"}]}], "Input",
 CellChangeTimes->{3.391857503734375*^9},
 CellLabel->"In[47]:="],

Cell[BoxData[
 GraphicsBox[{GraphicsComplexBox[CompressedData["
1:eJwUV3c81e8XFxfZCpEiIyMtRZH0PSollKRBmZWoKLJSFFGZyYpIZolLiMzk
+dh7z+xxXePi3mtvv8/vL6/3y+ee57zP+33Ocx7xezZ6DxgZGBhieRgY/v/3
fnogdWuLFbNam3V6diUa+NeEpU+vs2JHWRu1JJ9EQ9nF70ZOy6zY5NeDzIv+
0SDdn189RWPFAi9+Z9usi4ZJzpGEf0Os2ITWsbIwnRiwfXzCIKuUFbtKjm+e
PBgLr6W6MQsvVuzd7crjCWpxEPlFMriemxVTmXL+0fIhAYjyEXG57KyY3yPx
dJ34BCgo5/4Vx8KKGbV4NTbkJEA3daXRcZMFQxe7L/0bSABh9SYu0RkWDN7O
Ht5+/Bt8mXL1eVrPgj23upF+teMbRJ3pcOfxZ8EWhLiYFIQTIbVFO3DFiwXz
HWWU5DqSCIWWxTEjnizYmJHa+c6zidAblFqU68KCPRC/EqH8KBFEyB7rplYs
2PvLs9PCuYkQHXDUOUOLBWMZTQzL1vkBMQPeNtfYWDDOZaM9PZZJYMaxvdGd
mQW70zv2SuZ5EkgoeR3J2MaCbbKxeOm+T4LEgHfT3KvMmCkGHCe/JUHaGQ+r
uklmbGuEaH+mPwkKv7hYatQxY9yKEVeytZLhdeVqhVMVM5bReiPAQz8Z1OZe
SCeWMWOsnmEdEubJUKblPEr4y4zFX7x5rts1GWpXHO6X/GTGLnasaAukJkO3
/lPTMx+ZMa5rvfs1txEhynO6yNqPGbPkeiuwnYMIJunWolFezJifp9LuYH4i
DLNY9a26MWPHn3MVMMsQYSLb0jDPlhlzoL4tMdMkAnGIXDBmjcc7phX+5BoR
rLkshAUfMWMJx3ijzt0mAs3c/J/jXWbsmMxObvWHRFjmu6uvoMeModNvnL96
EqEABnPu6eDxK9Enkg8RXK1MdwVrMWP5qUKcsx+JwFBq3E47x4yt1R93Nv1C
hJKZXkUxYMZCy9dPoFgivBU2Cr16mhkzlNHsI30nAqvdnetpCsyYb+/v/Dfp
RKj6+i+z7ygzJmevOLacRQTfaoOdXIeYMQvdm9QTeUS4vND5TFWWGWtl7upX
LiQCt7h+s9V+ZqyvyrF6G0aEpssd8l/EmDEvJb6awFIiBDvfDKzZy4xFHp5n
JFcQYVfT9asHBJgxtmFi3UgdEXY3yHRa7WDGFHOcVtwaibC3bs0kjYsZIwdb
Px5tJoJoTSOZxsaMHX5z8wF/GxHEqxKeKrAwY4Uu60pcHUTYX/F80ZGRGdM3
N1du6iSCdJn267xNAta9uVlm/I8IB0r2saytErDXcyT1nG4iHMLmPpxZImDU
Lm7Z7h4iHC2qFHCfI2C5Fz+OVfcS4Xjhl68lVAImSKtnft9HBMUCGynmKQL2
rqKUm6efCEp5539qjBOwKQnlzw9wrJIjeMKXRMDY5jfP+uJY9TelsG6QgD2m
KuW8wjFkInWePgJmXBKQrIbjcxkhddf+EbDU/1x7WvH4F9Isb4S2EzCCm/cu
ZRxfSj3d29FMwC5tfJB/iuejReQx391AwKJDN9if4/leSRqhGNYQsCZLyZDr
OB/dxFz76AoCFpJgXULA+ep981sbLCFgizvTov3wetyMN/WURATM2cVKZaKd
CAaxChwWfwiY8u992aJ4PQ2jWUOScgnY+uN3pw+34H6O6hGmZBGwUa9B1l1N
RDD/7Clnm0rAdg7QYl1qiWAZpp+ZmUTALEnPZdeqiPA49KDKwjcCJjF82PoO
rq9NYKvmy68E7BcpXa0I94NdwI/mwggCZit9kVDxlwgO/i63tz4RsJjKk5NZ
BUR46S356G0AXu+GsOOav4nw3t3eO8CNgGmF/VgISCSCz2sN3mYXAraisffr
ejwR/F33fOZzJmDgQ7tiFkOEIOfSH59tCZhP7voOvs94f9ryVcbfJWD8L6qP
W+H9EfN0TGfUmIAtyHb/rn9HhHjrPx0ydwhYzi5npOJBhB8P75NT9QiYEe9+
KTUXvD8tlJ5SdfB4I/4B/c+JkGrOsXhMm4BpP3T5GuJAhEyzLObc8wTsgrRD
is5TImSbeH1YAQI2/vyMiaEVEfKMDAVUVQnYo+59ov54/xbcOfr1tTIBy3wh
9IL0gAh/DZikihUJ2N/xpR0P7xMBu9WZynSMgBHvX48Vvov3640UxYuHcb/+
e0LbMiFCuZ5bofcBAhZ+RWhkjzHen7rX1WulCNhZcZ8ndoZEkPu883uhOAHj
shTooeDzw3+gmTlNhIAVXe9Q/W1AhBnpIIuY3QSM2T+R2qSP++WpbmWgAAHL
polYaeA4M5tH1mMHAUueZTwrgmP+jQZvey4CpktO2muAYyf1gAlzNgK2ka1S
soHjLr8rWreYcX6mdC5+/DyVVs4UjW24v3NOzMXewesvXMdxaoMJK15KdSEa
EWHzrp+13AoTZlVztOKkKRHMkrXq9ywwYcOmKcyX7+F8aWxHuOhM2MbrKTca
Xp/9ytUBm1NM2I87N/zlHuN+cPOmUseZMBGF58EMNkQYr9DQHSIxYXx937a/
wPXQ4mb91TLIhNEeGdpFvsT1ulmxo6yXCVu9/M3E6g0+n76+s8vuYsKIrkYW
s95EsCWptya2MWGXFB7KnQwiQstBguLnJiZMX7D12plIvN/tS0N96piwuII+
L+5v+Pzddu6WdRkTlqv5u/0A7tc7mttyjTEmrH849OIb3O+FgZjg1UImTPbh
pP3fVnxedbk5q+UxYb7xHzdJQ0Rw3wf/jv1mwoZCv73epOH3g8XmKckMJuwP
m9ob3m0poJ72N5I/lQl7qLTnvfjOFNh+RtVoKYEJ0074qH5HKQUev10rHI9h
wip9kr8HaKdAXW2BSPcXJmy/xa3rPWYpEHjn1EBhCBNmubl3sD0gBWbjliHt
I/576TfRH3+kwI2J3NgYPybMZzH2/oviFBByPnnPw5MJE27jN5xbToGXaKHE
3o0Jq0vbaxKwKxV6WbIlH7gwYYsmVsw2J1Ih9pPCqIY9E9bOnCUl8DwVZLPk
H3LdZ8Le3WmQcGL/Cb6r1KpNEyZs7rREN3b8J0ydTT9Au8OEiWYkaTkY/YSM
psOUFj0m7NxkwAnJ3z9BeUbu6edzTJjXXbv8RYc0iDwx2eDzHxNGoAwd7E9K
g3XX5KMuKrgfItfT1PvTAOOQpRsfZ8IGb3KIhF1Jh0sHpBwkJZgwydHqJuV9
GWDEOrP/jihev53ZK3xnMsB2NKc9UBj3w1/JuQjDDAiPu6S0uZMJ05EeKk6M
zADybuvVLiYm7JXNpcUSsV+wuqSYwsPAhDGLrmvPn/8F3B0bhhfXGTH1p7RP
jZa/4ETwx79Z84yYl565vcevX/CWPcs9YJQR0+p7YVx7ORM+j7scKx9ixJIE
vSr+c8yE1Ar14bU+RmzRjpimH50JrR4d5x91MGIF+lmjTrOZILG2wqJeyYgZ
fpQ8eT0+C4opah9WkhgxyYZwV+K+bGirZvtP/jsj9vb+z3usetkw/qNlxiKO
EWu+wF258jYbeM3NddsiGDGx+VsnIqeywazXiy/dlxEbTeaZTSzJgc26hghz
a0YshXjbpNE7D1TSjBIbjzJivY9moM++EH6288pnH2TEQtkpDrN/CkFsoyw/
UoYRI1h2srwl/AWWy4frH+xjxHwYfDv5w/9Cy8TG3Bo3nn/NndW84iJ4LBVz
VnZmG8byMiM88xMGkVFDve6p2zDO4yFCluKlsOprwXf8wDas245xdtarCtLe
2pTLSW3DCi1cBymZVXDvtfNzSfFtmJ0w0+hgXxXUPPPt4d+9DXtrsVhQoFAN
kQZpCUvbt2GDk7ekuQeq4ZT0osLfcQasT/BcuqJCLTgXv7txKYkBe7iGJV6p
rodDfwJYzn5jwPK+3CX50eth8Hd43qlYBsw+ytIzdncDXEpK3nvwMwNm0Hpo
RuNRAwh+rCdxeTNgf76wvrNmaYQcIwHHVksGDC6+N3ixpwkWlr59MpFhwIg9
F2IYFJvhjndueJQkA1a6FH9T8XwzYEI1Ed37GDDsSMsBdb1m8D9F/XpLiAEb
nE7gXLRpBgmXU4k6bAyYTY3MYEhyM8Suf3nyqmMLSe94W68m0ALWIsXPVhq2
UHXw3/RF8RZQ+o/s6Fi5hQomGY67H2mBBjf510/yttAY84tlDo0WWGcsDzCO
3EL7wlvQkmML6LPNpJ8x2UJ35WfViDUtICHH9zv/1hZa5H00cKetBWa0lPNO
XN1CGnEb2R19LfDW3wM7rLaF1JtDlg/TWiCTR7BZRGILiV95ZXJ5Zyu8lldt
jxDeQgZOShqXhFtB89rdfwJ8W0intEiTQ6IVBoNThrgIW4jvjngISb4VuHep
zW6MbiJq40XODu1W6D75YPFF/ya62S0Vtn6tFb7r+64udGwiP6vQY2T9VlCN
aGOcqdxEF693xTLcbwXWghWWx9gmCvuR+lz+YSu0dItykPM2UWIIv4X4k1Z4
tPcRX3/yJiqa7A9Vd2oFxTMBgnfiNxFrjESv48tWYDDJ2tMRuYkOJpRcsnjd
CrWvu/bphWyi9cHXpgIerRAWsyHZ4LeJPk7Ff/d41wr3MAlZrbeb6IgIW9xP
71Y4PKRxqMJ1E/X9EOn65NeKz/cn8uccN5Hjn6FmlYBWKJMIVix6son4o2cE
ogJbQafjw2T4g03kxZqWPhLcCl0+PrHPjDeRZBK/V3coHv/Mu1vaNzcRc/Uu
ZvewVpiiuXNKXcHzY/wQ0hreCk7fXEs21TeRA1X5a9tnPH8DZ+cu1U0kOlb9
wSOiFXw5HI5kKm6irF9Dzb045kM2JL9DmwjO7+4cxnGUnVXkg/2bKMWRmTMU
x9LSlrqwF69H2elBKh4v4989lt38m6iJRb9sAT9P5YNJ4SzHJqqSXz/wHc+n
TO2OXR3TJnrNOeC5geerM39TNnFtA6muVlzY/D+fH9f63eY2EKSFkBJxvvcM
r4TepmwgAerZjsUPOB9uTS2FkQ207QYtaNoX51OizsDVs4HOMpE++nrhfJzU
csgtG4hg+0Gz3hPnc0DVGqvZQPOCH05lu7UCf5+SRGTJBjqlltdz1qUVogMV
uuwLNlDRL5Gfdri+v5bk1GWIG8jfl7uryArPP0V6lSF+A+1J2nb83wM8fxOJ
jO6IDfR7irXwkymeb/nuvQE+G8irzNlmDvfbvRcCLZZvNpAbT1tCiBae76Ed
3mdf4PjmYdG2c3h+odvn5x9uIBn3UTbF43h+GgRig9kGMjG7EaR3AM9vbcs0
yWADlb6b9OURawXZe0u1hpc2UNDJb7z3OVshU2DuzQm1DcRyyL9oeRvu5+oZ
JR7lDfT++WGJg0stcFWenFAis4GOc46bmA+2QNfI0O2ofRtIYZ7L4jHej/fC
+3icBDfQpyAeN46qFnDabHM5wLqBdMK0b4iktUB0Xen1QPI6Wm0TG3PA+1vW
HbE97l9HyUav6y9Z4P2q8Aed71hH2h2n6n7faoHyyMyDS+XrKDvVM+f6yRaY
ehjHZPJ9HYWKSO39RmsGp71f85W+rqOLs0LhR/ubgaHps82OT/jvNZJidWub
gV8psKfs7Trasi4/o/etGU4zu2UdNF9HR59dF718rRlkyR/YAo3wfDT9+mdV
m2FXZZTp/I11NGCiyntathlo3gUcfy+so0bXJ9zpG02QyLl4X0dmHaVeNM8X
+9YEO/mf8D+bXMMfp9YH+IsagWHB5XH78BrC6N97Hvs0wnS7L3aqZw3x3v2P
n+FGI1SFJ1kz1q2hhH0WDqPkBni1h1QW8nMNLVizKHxmaoBxCUPHHNs19J9I
soyfUB10MD2uFX60hk49CDha1FkLpSRncbe7a4jsylByJKwWohPD6jX01lDY
1EpZFW8tXJdrkfqnsIbS6yUrh9eroeiYZsfa4irauR4t3phRCaFqSkpnX60i
8rd07ps+JcBUQe2rdVxFAaq58abHS8BOO+ntraerqD7V8GhedzFcu7W7xcp0
FXGGWKvdPFAM3NZr1mFnV5Edr8N4lCsC73AUT2FeRcLvOAnvP+bD0l5nTafN
FeQSbM40WpsHFvHyNIalFRQruKt6jiUPLvyMO7NrfAVdrXn5TfpFDjCWenap
1aygU8l2h7l0ssB1RoM7LGAFIZc9Ld+dkoFiz5At5rWClPP6FUYqfsCdlTzD
FLcVVOjoM5LDnwinCHLJmO0KImH9U/LJ8ZDkM3xV+9EK4ho+eXmAGgtCPF8W
2++uoNzVQLaHD6NhcTenOkVvBW0dtrxfYhUOD2LKJh21V9Cg+PDFt+QQaNv/
KohBfQVJS15q+W0UCOrEE8p+qito6O9oZH6UD2QdnekXOLGCLv34d5r9uQdI
ZCe+iz28groZ/ff0BDpBsIrpoYPSK4hnNe1S53kz2IYJtmaLrqBZwwet11/e
BtsLTS/UBFfQDgnZs29oDjBQ4y1Wy7OCHtbvSvtX/QZ0dM9W3ty+gkLNxJwY
9/vAYcNMfquVZRT7XSi9ezkEJm1D0z/1LCOBLx1/9rrGwf0og9HJmGX0+WuU
2f4nGRD4e+xMa/gykomKGg1y/QVFdU5hfz4uozr15S2eD5kgvBGq4e+2jPx/
S6ge/fUbmoybkw+bLqPJpi+/Apnz4b99Wk9tRZdRIied+mcag93xp5cWopZQ
z0q/mxlPDTT+EGXbFrmIBCsnGmbcOsBsGtvmELyI3j8I5NtX2QGzx++vkn0X
UdixH680uTuBv+gHpc5lEZ1TltfwjuoEgzb5hs/G+Penv0kF5HXBMMP5EHnx
RST8jeVq4XQ3LN62FDFLWkAih0SV/dUHwCuGbVdr7AL6aqE2UuU9AEKjKdwX
IxaQdX1mRnfdAJy2pW0d9F1Ad3a9dVWVGwT39y+Gl6wWEJfaZgjKGAT2LL8f
H48uIJQT9bUtbwhEOTOOoZx5lONSNSH8awSekmR7P6bNoy/mRaG05hEoKox7
b5Y4j1rE8hbDZkfA+ElIz7aweRTuoLSgp0CCqAand+qO82ihmP7RPpMEu4PO
/KtWmEfLqy6lOYmj8PBRjmfkwXm08nNn0NmyUcg7e/SIleQ8enJ0NSNoaBQM
ZsU8Ofnmkd9LAxm3PWQIu044fHV2DgWHX+xx8ScD/65a97aMOZRXZVVrcW8M
zGfOH/yeNIdyTYd5F13G4HdFYbtj7BxKdT3Pp/ZpDK4/T5MTDJxDLs3bG7ZV
jkFwV1DbbZs5ZP71zD85mXEYzmB3k7OcQ2pniv+8/m8cjvl4HlgzmUNHxwxf
eN8ch6ZTjq+jdOaQw5D04T8e48DzxUB24PAcipTOu/Xh3ziY2De3pEvNofdn
3MbdZsYhTVvrlbvIHBqJlKiSZJoAnfXTLeJcc6idNHrh1cEJ8Dfe52o+NYvW
c4bvP3SeAHYxcuMEcRZ9dVDP/sk4Cb1R22sH42cRlh537sjOSfi5+2BFZ+Qs
4s+4e91QfBJ0+WwLy31n0aHIEv38/yYhjGU1Ke7xLJoRM/3zwnESHnru/fb5
3iyaf3xF5rrHJJxigJiPd2bRHz+dxfYAPP6K56dXWrOo5LqSbkfiJEhMc3nc
lptFnSmdd/NaJmH+kfwrXYlZ9FcuWKO1dxLKyXrOGsKzSNPyzHgoGY8/GP70
BPss6n5z5RDHyiSoGBc8OsQ4i44UtZ1B2yjA0d1rLrlKR5MaUWQhdgqktYob
7piko5vO9vQ/uyngrqt+a/swHak0p7QQxClwrd7i2tY/OkJ97pqzMhSQ0PK5
vNhMR0/7CXveHqHAfEWKxnQ1HcnYvucuUKRAxfmGc6RiOnonFTYdokKBzxjt
TE8+HSnIcr7doUaB0wUnFKuT6ejAL76dS5oU4FQyOIrF0dHVmJlUEx0K9Ge9
lMuNoCMnuYYLj/QokCH/VSotiI5Cd1z8vesWBTx+IrHvPnTEdeRGi8VtCtyQ
G94T9YaOOvUtXG8bUUDqB0Ew5AUdDd4wC5oxocCipMxO32d0FNEWNyt1lwJV
sZpcbx7R0exzzrdb9ygQKWK93fkuHTmS2Q65mFPAKjKAyeY2HR2mB7V+fkAB
VcFfmw+u0VFJqbyFgQUFuENbV4w06ahgP1dLPo4HeRfnr5+lI6XyGda/OP71
QYimdYqOlOPU1+7i2IP9NOXsMTr6Z3swPgmPd8PLmKx8gI7MDjosBOLnSRPc
h46K0xHLVyJN+D4Fltzje6V30xED797gi//Pd6OsU2QHHdHZ+rv5TPF8X461
8LPRkZFoe4MHztd6ia2BYxsdyXb+cQjB63HG4VA14woN7ft9oOMCXi9uuk7Z
Co2GLpyL3IrA6zn45BmijdOQ1GUHhkC83pmTIQVjgzTEs38HRUaLAp6WOdn9
XTRkyFbWdvcCni+pK6O9iYYak4faz+D6Sd9dS6mroqGrWCVbIa7vcp/Ij1KM
hpabSQG9uP7Vd9TiC/JoiJUDex+D+8P6xrvPSUk0NLvubM6F+ylus0BBLo6G
sh7HMKkIU6AjidaQEkFDMiYbzQJ8FFDbMGRO96Uhpvk70T1MFHD8ERQr70lD
2OinufK1SSBeqzyd6UJDQp/8pM3mJoE/8bhdtjWez2kR2rehSdDUfcil9ICG
7oTYzV7tmoTXq1+T8oxpqHmRkZjSMAnjOmwDf3RoKCr3yWGngkn4uzSgXSJP
Q9PeexpM8P6bjRMYO3+AhrpF32eE4f0pc1nbo1ychuwPcL944zQJQbE5eVU7
aaigr4P9mfEk1AYLa0dPUNGuoTvrxTKTsPek42buMyr69evjR8+0CXjvO6j4
5x4VeXZ+NhCMngBav7ZV0XUqEnl0lmj+YQLKvCS6yk5Qkbv+fcIRqwl48q/x
V/PKDNIcCD/gJzkBRa5yDyhvZpDSj/e3lr3GQbb505cZuxn056x/q5X9OARL
bWuh359Bne9s2BJMxsGiofO/5QszqLV4tNBUEZ+HYu+EmNln0BkhC54DfWNg
VjpQKxo8jboiolOK94+BxpnlB31vptHNrw5bjexjcCSXlyHq2TRSN/KtTqKR
YZ14VnH3tWkkOP/9HvpDhs/BCVF8vNNo6nDWVPJVMjTdtXyyPWAKHVB//6fH
ZhTOMlC5Z30oiKV2q2s9egRkX7ImZ7ygoOMnByyr3owA7/y+8zaPKMgzyCch
2nwE+sd0n09doqBbZ+L1nORG4GVj5gCZlYKuSX5g3Z87DJnRThk97ybRMKlH
VLFzCCKEPmpFOk4iA6UuYaHCIXAP/kEyeDCJ0k485FOJGwKd911CHeqTKEH1
I7poPQSTT1TeNDFNoonLTxWmmYdA4szmtfI3E8hhqWAvSWsQmnacORRhN4G6
ixKaR04OwmuyC8uT+xOov36pVVdyELo/rhTwX5hAz6uD0xjWBsDbXPnT+IkJ
NLLdTy2kYwBOnnpuUyg9gbDcSu89WQMQMjwvab59AuVffsUlazMA53IVNpVW
xtHcIldovs4A0Pzsujgmx9FjyVxTw6MDcPkEzT+rFsdVM58n5vphle2opVfh
ONIvPZE93NkPSf1Pzhr+HEdRp7ieMv/tB4I3ZYHx4zha0WVqGvDth0wjuaYO
t3HEp3ntcIJ9P5gde0Qk2o4jQbKmzw/jfuBmSXr7+u44wt7V35q91A+F3WQT
Pb1xdIzU/M/7RD88Tpc6JX1+HPn8095vK9kPQm/N+VYVxhEpWorv585+qDBI
mK7fP45CGy5oqzP1g+Ph4co4gXG0fi/lpMpCH+xnFI93ZBlHMevx20LG+6Cl
w9RVc2kM/cjc/Vmvrw/cU6JviYyPocmYHQmurX1wxL1Pnt41hljOq77nr+2D
3ht7Ocqrx1B1TJGxSFkf/h41HP1cMIa6eiXWI4r6QHkzAlmnjKHQ7pfj/gV9
QG7pilCLGkMSizlGy7l9EPpD0IH/wxi626Wg3p/TB+dcb+mMvxpDjm1vVNTw
/9N0P8kWPh1Du+Ytfffl90G0VBtjoOkYmjfff/hVYR9cXt3Zd193DL30plTf
Lu6D1YZruUpnx1CxXWJ2dmUfJCUEBnEcH0Nqh+mzgY19oO/caDUgMYbYIgyO
U7r6gOUK98UsvjGUcKOzpXSkD36LXxHzIowhp5DA4d20Pri36Ld6Z4GMTt2v
3De10Qe8tTVtR8hkpGDDu6rM1Q9FMWzpjJ1kdJzyJ2RTpB+sHS75dFSSkd6s
Q6mKfD9UiVSceZ1MRszmUU2ct/vBaZYgpBdJRrfcbGajbPphf+X5WSk/MvJk
iVT67NUP7rbFifXWZBRyzL27+E8/HLnA8CbOmIzKZAOcxnE/9e4GQ0cdMrpT
I5D3eKEflEsLeUTk8e83rc/bKg4AOXxtgiZGRs0f0vdO3hyAUGuVsrIdZES7
cwlLdcb9K5DnbD03ig6TLsYvFQ9A8sOsoT85o4iZYKH57tggtN2L2aedOIqy
SS/cpi8NAoOxn3H3p1GUeXddeN1sEG5du/9v2WEU8d94yb4RPAiMKnzNJxRG
0XM5BY99m4Nwh8MeS08noblNyi/L6SF4x2y6+V8MCRVX2V8w5BqGjC0t1YYA
EnppzzaUfXgYWOYl8qaekBBTnU5roA0+L3pb0g8cwvHgYwWHpWFgTzsek5A0
gizOPiBW85PgRJJo3/HPIyghpYLsdIoEZvHse0q8RpCP0w2kZ0KCnLDhsEGL
EeTVe7XXJ4kE99yCA0SkRhBHgMLfuP9GoUB39lV43DDS6+ux3WZDhsdzv4z8
IofQz+0nSlQaxyHiJP20pe8QUtA1fEGfG4eqF/J7zr8cQlnPXx3xE5oAKYb0
f6u3h1B7kDbz3bsTMMiZeuvx7iFUPxx36jF9Am5KJ+pqfh5EFUKGEclsFHj7
aPSolPcg8oozEuw5iN/Xqft5tjkPooFRSlrXFQrwKiTU5+kPIuFPp00UgihQ
C7FasoKD6FsCz+uLAlPwJlPRh/vQAGLneSG5sGMa+pWzNCo4BhCspH8xl58G
VXSM5TWlH/0p2//to840LNYe8Zwm9iM+jmmCgO80WI/Kvqo70I9+zDO+qdiY
BgMhkWe+Mn3oAK9bgVjnDOREfzl6jrUPsf24rT9DmwE+KeGZFXIv2u28fsGa
nQoNxwStHiX2ojJiMt1JlQrntXc8uCTViy68SXHIj6LCkdfMt1kke1CE6dWW
KmkaCBc0hY8TehBRd2zdQp4GzItfOmrI3ah/SiJh4BQNep8cv/GR2I38MpwW
TS/TwNfY9KrQ8W6UbcUqFmlDA6cIuYBVvm4kxRXryeZMg7vtC3W9C/9QVfz5
bD13Gihf8deKy/+HBoj5WnaBNBhTzb8gd/YfkmOt5R1No0GL89u3nJL/ENmw
OUU8hwZFv6+WzhD+oYs9GQ1Kf2kQdoisllXVhWIljv6g1dBAfS+fqqpuF2r1
kO4OJdFA3qD/pejxLmRY3jzwdZIGe0KT8xn4u1DZ2Up/dxoNZjnUlMo7OxFF
U5y9dY0GsWvWx3VMO9G5njbZNF46+CspP5M/24lGfPbOvxOgw3N7poydkp0o
1FR9UkmYDjqUiMOd5A70NNo6R06SDioyD6zzqzpQDPswspKhg/R9+ZQvxA6k
oiBm6HWQDhvdFbJmTztQ12BRkY4CHcZ3BVue0+1ADwXaK1ZP0qFNzzhx//EO
FNwnMvhGhQ6pNXOSEwvtSOmqTYf4WTp8ZkH3ajvbkW3VzS0VdTp4nvON+5nf
jh7WyDYf1aDDnQKxfXav2tGDH+OCPy/T4eIixfiGaTvyLcJIJ67S4djx3KiT
Z9vR7c3b7F+u0WHvU48eIcl2JEG1yOu9ToftxCvCa4R2dGSvoNjaTTrMjwrd
7iO3oc/cuxwW9ekwIE4KR1Vt6ISa73jDbTrUGKd3xBHbEAcLf9ZbQzrkRLwU
eOvfhuaf1bELGdMhvv3CDYunbejgQXVZHxM6fNixI+SSbhsSOeNxocuUDs5X
epvljrchrV7GaOa7dDD3+cHLxd+GssVynu64Rwfdcrur1IVWZJO9ubCGY9Vt
/wU0d7YikmuGffl9OsieYavPym9F9w7wHHhqTge+F20cYV9akVt1ncEijpe3
VnZQX7Yiu+8FgQ8e0KH//T6hS3da0bmKuJPpOC7luiAad6oVZW2T+96K46TQ
x/tXhVqRk/aU8v9xwJ5AuevLLeiCqK1uGo7t47PlUztbkE2d7OX/xzM40HOS
ObcFbc+1C1jGz/svg+GMSVgLan8c6PIQx/uVpM/nOragjbgfL7LxfNmKtDV5
b7ag7ANFtAGcz4z6s6uPFFtQIzFHaxzn31YbdrOED49/48pcgxkd8vUKDffM
NaMsS36DULxe0f+G7jq0NKOQBNtxZbyeb81YH9b/akatHz8J/DWiw6OxQ0+l
g5oRW+IbPXFcD52neg5uts2oRX+VzwLXS3Hh+cuuq82Ih+7V5IfruclY+t6X
uxmdDssVccX1J/mM+49MNyHPBzu+aeD+qOblDlGtb0IChmTeRR06hIrejpnx
a0I5cktfNjTp8PL76+8aVk1odEnzrz7uP9ND31JitZrQxsnTtWG4P+VUqLl6
bE1ITbWqrPw/OrxOYY97wtWINkZiDi4eo4Nm/s2f++oa0Kmq5IT3R+jAXxmb
3+zbgH7+vWbKgPdL8vDJlhPbG5Cvnd34L7y/2nbfZ9pkrEdBoeKvRPjx+N5/
HgQs16C9B8M/RtFpsPCJxU4ttwZlDebJbp+mAZZw7fWsYw36HP8z1WKcBjfR
WNituWpkFriTcaufBu6LfFWiM1Wor/8GOa2WBp3m1nLpIxVIJ8Xnw9cEfN70
6SrlJFYgBuol+d5oGiTeOqH+91EFqmA0dhWKpIHDpU3jWmo5ctb0VPLH5xfv
ocCgsbUyRIx9p27zigaac1nL+/hL8fm6pvBFH59H1p+ZZTpL0IKt9Pt5PRoI
jbruPBJZgpa4w9eu69CA3HHhkKpYCbpVedpQ5gINPAo6TQ0OFSPP+8ysRxRo
4Mi09EHcpwjlMBzRlN2Bz7sF9OMoTxFK6x+dEuWiwc4x7+Izn/4i1QhlY0k2
GqTVCC/cjitEHdKXjPS34fGD/jMKzi9Atr5MSzMkKvz2ZHWKhQKk/HzmSVMf
FTwcmz6mlecjYEjeau2gwr7b90prmvMQ//MDD9WrqWAg9l6OaTIHRdmvEvrT
qSC986r6Dtsc9NR7uwwpmQrzTEIm+xazUdQ86tjxjQpBY8lBpxmz0TeN02fI
n6lg9s8uRdP7N5o7pLgrIAS/b2pPl+tz/0a3/AWKTQOoUJNWv2wvnIW62CvC
PN5S4XNs2E6P2ExEOsaT3+ZGBYtg00OB0pnoWhBPlJ4LFZic6Kapx38hWnHK
zVY7KjRbFrwoyMtAlv6Zix1PqRBz2zOk6r8MlPhOZprVigrPX8oFVH9KQ6SE
uSNz96mgM0PrSE/8iR5av+IrMsP53MvdF5abirxXQ7Oyjamw2f7qoWtVCvqF
PdbqvkOFdk31X/f+EdGtXH+uQwZUSP3LvnppMhlxOfUfSrlJhbfHms8dXUtC
+nLBLfrXqWD0PdxPgDMJJT2VkDx+jQoKu03a1vb+QJNMV86oXKUC+4f9IsOH
E1GoDbeK7RUqDDNQHlT99x0xukcfatOmQoHDr7S0q9/Qw7wnEhZaeD3Hny+F
miWgD98lJaU1qfDQ6D81l2fxaHhPtDz/JSqoNRF87nrEIRdT+kV5DSoIqdc2
a4TEIp6Tj4xeXqRCv43kCaNLMSh82Pfu7AUqfNDPqJB9/BWJswud/YxjVThj
sOD3BZHddw5a4ZgiXTNR/DMCPd7ERJ7gOJJb3yWgMRyJ5v5u/YJjzcURTkP6
J5T3r7F9CcfLfbbRMnyhiGXm8qw7ft6P8o2j84rBqGhnJf0kns+tn77F2K1A
lCV4Pnw3ni+73NFEYbo/knx2Klb2/3xWLlVnKfogOkNc2X2cb0XVvanLzu/Q
vszVnBq8HpKfXXnIf96guGHj/cZ4vRgbDz3Y3HRBD8Lf/RbG6/n3iVmK/HsH
dMqpn40Jr3dTqujADb1b6Kb/D3s9XB/CXW0x7/P/Qa0NPTNPnwrJR+Kfcxbc
A5FRNa5LuL59/x6lx+k4g0HzhWODplSQOH3ykU2lGzwxmdoYu4frp/MmSDTW
E/KpsZRdFrif0q9oTwn5gf9h2tqYNRWEBxVkr7cHgEzt4pq3LRVMtn9g7LsU
BMac3F06DlRgOGU/cJ75E7zk+Zt20ZUK6pEQseN1FHh8UPnd9oEKnzIzWS8Y
RcMSL3+wbzAV7Ed3nlYViQXHLwWLpuFUKNt4l2yzFgcOw2p3LeLwfjhi/b4j
9ztw2fXerMuhAvGjsto3+Z/QfJL9mQYZ51sWU8iqmAajktsjxKapYLrMcspK
KR2P16jOM08F34xJod0xGfDNoecsPz4f/tudv2PweyYY1cTXh/Lg827yJtPT
rBwoulHUnXKYBq1vfYr3nciFPIs3ja7HaMCw769bc04u7M8MatE5QYM71/ev
KxbkgUbIF885VRrwFM7Or2EFEP/Ln/MIvh86f/hI9mkoAgM1wZmeRzT4LlP6
7bQuAo7uyDG7J/i+V7x4b7oZwfeTWVsMz2iway3NJukxBt6aK6IM+P5oe3TX
FMfTYqi/Xn35xTsavF8/WktKK4b/zhg+yvOmwddqTeJfajHcVsgenPCjQc39
Vw9tnpXg90beR/FgGkh9JpFaHEpBtN5qvfUrDVQfbJamZJcChSd34VMsDfSO
CyW8XSyFSM2HlRfw+e9ep333pHMZuPEq1Fgn4ftmxAM1nvwyOKxSOjFMpMFP
C7d94ytlICKRXaT+kwbdDFl9ES7lMGXK4Vj9iwa0+rpCu8JyKL6ipEvJogHr
F/IX7Y1yeDlj1raYTQPFE8J3Nl5XwHrIn56qfBpoMSqe6kAVsJTYbB7wB9+X
G68IpTNUQnFyQfxpfN8NePSmw8yjEqyVNO9oYzTwrt2WTSqqhFSBLMGwYvy+
OOwR8nCtEn6+9ttVU0ID14+MdlNKVZAn9W2FXEqD53QPXVuHKohXF3wwVUYD
u+tMR+czcPxVa/NfOQ2ssz25nKerYO+9Wav0ChpYCBKm1g5Uw/Qm3y3rShqY
vXhb42ZRDdyN2y14q3C9ewjJTAnVIJDtfvELjm+ceeflNVANj6LeJnBW00An
htmCY28NFAw7y5nj+NK29+ofDWrgMXGfbxyOz91nkeT7VAPfdn2xK8Oxavn7
beHNNZD4zOZTHY5PyrAOCnPXwsaZd9/zcXzMx6soWqsWNBQO6vvh+BCF9auE
Vy3skxEyO4tj6SveLomltdA4bP28B89HLH37HTmGOjAlZ9+8jWPhHT7Kaap1
kL57+ncBzoffnk3w+Is6YLN20d3C+XK3+yxkZ9fB9D2VUWkcb1dibzs1Wwch
nx2PHcPrwxjhm/n3SD1kyE3Ni+D1W19lDzprVQ8Cwi20Kbzei0Z+NuU/6iH3
S/PYV1wPWhGHjiapHoINOaPlcb0oYv6H6sUaQCZz779vRfh97MHJcc24AWrr
dpmtFtJgkOQ/0RbRAB+rfdbkcf3bkj4k9u5shHMTnhFquTRoYOd+Z3a1EVj1
IhR34/6psg64T/JrhD7eiwItmTQw+azB68TYBNfKVsZHcD8aVntt5Yo1gfWt
D+6qKTQwWK2cWfmvCa4KtCY64H6+bnSp/pVLE0hYyK14xNNA94N3IYpoAqY/
g+0GMbh+RVUp2/KaQFcsNJ4rCtdPTNP33XwTMH5efMX3CdeLpKnx4UkzvJvR
zfDH+/GUgO/JRr9mmPWjO956g+t3sUZqB7EZbtSYKG264vtPkhbhE7kZ2suz
dMfsabDfWrs4yrQFOBi1k0XMaCD+1S+j/1ULcL/6yN1+hwaiDbUxYlEtcJJ7
re/RTXxfkr/8+ltXC6i/p3w8qEUDrvnLp1OvtYJkPN2S+zi+f3Ufv//crhW8
JFWZpQ/i+mJCfudCWkGPfKyOfz+uvz+pu6utFYpmqwefC9BAdr/rS2aDNqjt
2ipNnqLCQfZ78U3ObbD7NXHhci++n9A0ar5EtEHNpMfOmFoqnCjkEz7e0wZ/
cgpSdfF959yNlHwTk3aIG6BkzuD7wwWVoKEDbu2Q2sxsw61Dxev1nG0hph2K
LohU9qtQ4erUudt+Q+2wVCrsHMyH7wdvu5dzzTvgsnLGzqvFM+D0m015h1Un
KFDU8hy5Z+BFJNW0168T2IW92moXp8HVvd3rR2onHBzuvvqvfxo8L8d1npnp
hJUUes7etGkIJCk/f/ysC9jnOSNvaU5DCv/DnNLn/+Dg+4vXjtlPwaBDhaLT
2x6QNlYh3I+chN4qZuGkbz3Ac7X7Yo7LJHTtvbD1r6wHzp892tFvNAlNZaXV
Z5h7QWIjKvy36CRg/MUmzO97oZVRUvBT7ATE/P7jHerVBw6dV2KuRY+DyXxG
b6bvABiK2C9e9SfDnUu0YhJxAPyuPfXntSLDzaijP3bVDoDl8ctjXZpkkJbo
VP7LPAj9Pq/kfVjJIFpzk0n84iDMSX2rTfAcBW7h65/HKwah1MP36B93EnCW
zfpu6xwE9dILndvukoD9afAr4bFBGKy6reJ4lgQsJc33LrMOwc4LSVFTjCTY
fKR7OENjCMa5pTweeo/ATP6VkudVQ+AtwXuUKXYYpNgqHVw6h2Afee95pffD
YGSgJuNGHgKryMIT+dbDULN43P89YRhWc746ZqgMw3cFoVuf1IahO5WF8Kt7
CHo9grZHXB2Gb19MIj6X4ue2sP+JMhkGasqntKHUIXC33dj3zXUYxpc8ZOPd
h8Do5/BEZt4wjFCsXcQOD0HIumFUTuUwxBjPKxjsHoIa7Xadgo5hSFmg/Saw
DIHyZGVW8fwwuBK2Xy0fHASbU2ctyplGYHKQd66vYRASvQuEqneOQHLfb3uH
v4PAJ/PzVZP8CCjuYP8tGTUIWk7S8m0wAoNnxVPU/AfBvTxmuFNnBKilDZQR
10GYuR+sMWA9AsWGr/eXmw6CVBbH6rDLCBg8tBlh0RsEI8Z3qWTfEfD4vM25
Q30QamKf75hJGoGuZ35zUocGYRuNVkrPHYGXf6f5ksUGQRkeOy1U4N8vc7lg
Ani+ASOyK+0jMHveKfgJB55vn1HPOmkEtBpKfv7aNgh9hzo+MMyPwFu9Qr69
9AHgd72qRmAigVPPh8+sgwPwRvjcdw5xEoS3h5x1Kx6AvEd/9HnkcXzhCzPP
7wGYyVNk5wMSuN6JbO36MQBS29MKd+mQwCIxYKUvagCM9GVshI1JwG/W+HFf
yACEJMaKi1qT4PqLy87RuE9rFna3ibuQACSmjQ09B2DbhZD3Ur4kiMrh/n3r
1QAoh3KeOhBBAhvS/pxA5wGwGXlHOZREAt0UF0s2xwFIPL71VT6XBC/i9oRg
dgPQ98ZZV7EC911eU232M5xPM51RuZ0Efn7XrcdxrC1mlX2aRALlrdBwffsB
8LAhWcIcCU6NHDzF6ITzKzIWPs84CtO0EeaxFwNA5eqsu7hjFP5YP3EluA3g
favrpiU2CqV2h8gG73B+qdXHdI6OwvlGaYNhf5zf2jnStf9GoU2I53f0J5yf
VmHYzSujwBNllewXg/OLPKF522gU0LvOnES8L5Un0taMrEZhZ6TRwekcnJ+y
bJrZy1GwSR2RMSrD+XnFmZn7jILlQn4srWUAzH49SU+NGIXI3fxa20kDINxz
anM+eRQu57pse7Y4AAFHWr68qxmFg6o134N4BuGSQfRkffcoHEjr+vJLeBAY
PR6f2kUZhTf2wo620oPg1M7YmchBhjqGUXkMBkF+q0GKuocMfsq5QdpXBmFS
9ouD0iEyFArKtT82HAQTV4WdVZfJ4LXt3IXbLwbhoqT5lYkPZMhfpSbF/hkE
hivyUceiycD1UuW5U/0gFDitT75IIwPtob1z5cAgHKkJ9WZvJIN5kZ5oA/MQ
7HpWXnqQdwxOZu723qk/BOQiaZWnwWMQOtXV9IVjGGLHZ71z4sfAoecZa7Tk
MNzZiTo3M8cg4b1e9tjpYWh4oO/4sXUMRLam9l14Mgw5nN4ZGfzjeF0rw1pb
huH9nQnpubBxiKv038pNHAGpxRQ+5y8TsDGSP2JnOgpC5dckUogT0GXzLVz9
xSiwhy7J9+dPgCMHLZo7ZBRox87rqHdNgIezavC7ilEofNLtzSswCUceGs23
43W5Mbp9MylgEhhtdz6qXyCDxu80zt6vk1C5b9LDhXsMVDxv7OH5OQlx1574
ccqMgZh4rLJj7SRc5ygeatEfA4qRkv3Z7RQY49+QbMkdA4+2B+P/3lBgl4nr
VK7tODgmcCxxBlGgsufooqDXODy0+8WsFkuBjgHv/dpfx0GHd0MisYgCaPFa
nEz1OOy+HGpst0YBOXbGsXjRCcgoLWlhd5wCGTNBlU00AQnBD4fOvJ0C21a9
wROtExB2l5tmGzIFnTvdz58gT4Dr1m2uzl9TYK9/ZNybYxI0TtM1EmamgK14
ZdX++iT0Z+77e/rhNFREyeiSOyehXtlOTuL5NHi0jrZxjk1CYVFZ+Pb300B7
qpy8sDAJX2oe2bUnTENeSVIpJx8FDIazZGwGpuEC+cTwRS0KaDxk+XRzZhru
QMecoz4FTs4YMKpuTIP0Wt0Xc3MK8K9u9G7fMwPpEhQ3t1cUaN5xKTj+1gw+
l8nNCakUwMIjt7wfzMAH7c9FlbkUyBCZtrZxmIH//LftSy6hQMCBYA3VYByb
BF5P6KTAq3TSb4m4GdiS1Mn/M0QB6xNKEmwZM/C6Yzw7kEIBrbO96+31M9Bx
t+SnzSYFTlUeeVzYOwOXtEW1X7BOgeyVN53xlBmoiXtf+h/vFLDels60YaOC
140fMetiU1Bma2vJhr/DQ4V3TQBMQdZiSRv1OhV8TpQq7r84BfGuAuc68Hds
+KWis6WXp+CNd8HeBHcqnIm4cPLYbVwXbi5fn49U+DaYNL1lMgVmoaZLNtFU
2LyyxexjPgX/xRJaVAup4MSewVnzdAoOS+uDJL4XqV1e6nqD7xd7U5NT2bqp
kC3NVDL3fAo4j6/vpo1TQXOE+88+1ylYy9Xx6liiwvJrpnxm9ylYiNZ9acKE
78Vf7TO/e07B9NBKRjM7DdZ8w2O23k8BaX/CmPpOGgyd03zJ5zsFvZaXRfN2
0+CNL8N/Q/5T0EZcuHFQnAaSvI1jDz9OQe10tF+0LA3Ohxs5EYOmoET+UskO
eRq4HBmZ+I77rMCevvxWiQbPbA6qG36agsycyKPL/9HgZpW3d03YFCSvnLew
uoi/sxra/8yFT0Gs6nRU/xUa/Ii0JXV8noLPbmGt1/A99M8hJXaHiCkILAH2
ciP8XfvP82Qtjr2YJ9SUzfE9Gx1w6MXx60vBz1OsaJB//lnjDxw7+Z1OE8X3
3j/h/LcO4/hpA4kU9BI/T11V8Ake32JHwB5mD/ydgVwkLfHzTW4o6Tn74Hst
Y4P/Hjy/m+GD3pRA/N0kH2zhGzoFARpZXNsiaPBkz6Knd/AUVC29C94VR4O6
E0UzioFTsC3JQPBQMs5XXrT844cpUDE4GHUWf5dKmu+XicDrab99U0wff3dK
036fvOY1Bal5Td+t8XeL1PBbwTxcD9HdTunhzTSYZkonBbtMgX71JcWf//B3
ZpeROCuub+CLPfklQzRYGhtrlLObAqZurGSajr8TZa6ffGM5Baq+IRpMqzT4
16J7+fddnL+KRZ0QIx0KnhsHhhtOwXgkR8f5nXQI2nfgmY3OFIhr99+5LUwH
ysjpF+YaU3BnLWPgqQQd+MTX37GrTUGd4a2JiON08GDI+PRYfgqYOQ88TVeh
w4LVV14pWdyfhWuzZefoIMst6R60bwp+7Y1bp+rR4aNIypk3XFMwWWf/htmQ
Djt4MB5mwhTsf3WRZc99Oth98uk/vUqBsD4Kz0V7Ov4+MfMtHaXAy+iTklEh
dJB/f5TvfD4FsnTYkn59oUOwE8qz+EmBqc2eQ5UJdNjV7p4OcRQwMX1zcjaL
Dunztm07vSlwXqxO81IbHXJD4kU19Cjg2hTdYNxLh/v1XJai6hTIcX923Z5E
hzkB2eSfJyggO7TLOHqeDoo5De8WBCnAGX/Xdp5/FmLd5ZsV/03CBT3FBTaR
WfjFusbdUTUJrxlZX+6TmgVjlb8qx/MmgXYv1VPrxCx4HuWh7A2bhHbJpbDY
m7NwJq++qUJnEqK/+/+9HDYLBa2Fi1VZE+Agezm+8+sstKy3X0nB93nNFA6v
e99nIZSh8LL+hwmYz/DVdf49C3uzyrO5LCZAu9B7JKFlFiJ4rrzZ5J+A5Za3
bGvcc5BzMPY/9UfjcI3h1c1krznYk5fcvncG3+89VVUUP84Bi52c8WwbGdaZ
10VR2BwcM56RiP5DhiQOl/G273OwV5BF8KU3GTZ3vXBhKJuDs5Hjz66KkYF4
2DFOn2EegsV6rjReGgWC4ZNp5hfzcHKgh/2A+wjs0b4lpvBmHqZcbwWKmI3A
8dNw3cxnHo6ThS/T8b3XbM+O/IKIeTDKtU5Q3hqGvz2/39oUzEMt/03XW/he
7mS0vuff2jwMW53cc/nZEEwY+2mlvl4AfmL5o+2h/dBglvxD7+0iFKixKRwk
toOHpJvKgt8iRCR2buezbIcT5Bv14SGLkCwaZL4h2Q5RVgxzvfGLcKGzsqP3
axtYOen/Z1m8CPunM0N7g1phuz9zu8vmInxiW172c2mGc7l3Gb87L+H3dueH
/OwaWHyh9EnDfQmOsA58EVSogWRVLtlJryXwC2JIS86oBt6SvCtHw5egSX5L
0Sa1CgbqeCLys5dA4IZWxrfECnAd/nu0YXYJZP/clfqXWALZXMLGS9bL8Lrw
VUnf32wgH7wYY++wDNQuLu8Tjb9BSMtuiOqyDOfovSynh7LA9X3Ng3HfZcg2
W92pz5oJ6lsvn/5LWgZ6jMJ/fHmp4CTyI+NmxjIwRsnxcJCJkHS6dbY5dxka
r/9ecBVIBg7ng89rKpbBzH+J2Of8HVpoPW5/SMuwwH00cSYhCgg820uUppYh
TirAQFMoAk4eViT8nluGkCEBhuaPnyDikb93KuMK6AtqvyTs+AA1Xnk1suwr
ECUSeduO7z2sfydxft+xAkmdloIDyq/hSBnvVfHdK+Dh1iC208AWzIZVg76K
rYCFDcvJwv3qEMzwqHW37AqI/DT8OONhAWWinwTCjq5Ai0Ga5LZVZ1hULdbf
qbQCpyduhzvpeYKM4XREwH8rUMar/GiHuC/MtH6NdzNaActBp/4N30gQm60m
rd1fgUdzD33rB7+CHu+itLPVCph9ZdrzPiIWsi/rpNi+XIGpe44MZuQEcC3f
zLr7eQUuzpwtUXhMBPZc0/LzbSswN3bDc7t/FkhHio+zXl6FgE+bOjtfI4iQ
eu4sfWMVqmz+vmeQwIDzV932C0ar4PvTjM3oLQZzlc8PeFivAkje1DyjUQzY
Qv3jdf9V6DjEyxlZXgKG115O0+pXIX7zOUX3Wzk09Da+4ulYhW/+Qas6G+Vw
7qEU95H+VWDftxNLvlkBsm5NR6xmVkHHh75oT6iExVRp21HuNfhiVWR97U4V
BG5vnf13dQ26eCrOTZFrgClU1nNZfw06ofLZxxO14LTvNZ+g2RoIJ5FyfT1r
wfjkAcUbtmtQYPgEeYvUgZy5m2ND4BpIzS9WfNWsh3J0cLmkeQ3ueUk7cno0
wqrT242U6+tQ/egXd0l+M7C3V0yD4TrIvmQTDG1rBmEFtv7We+vwWLdYey+1
GU5RPxStPVsHgfgY86uSLeBsGe6mHbgOqfT/9XVnYTW2cfvHV+NKEw2iIplS
CiENxHU/laQyROhJEfFUxpBEmZJKKiEpDVQkSWqZCkXJlGSoViFTkmikhMZ3
bby/9//fOdn7HBy3+7padRztnN+RGf+EvmSdyzLEGkt72CfbHKG7VTmTvt70
fU9FD7t8unD2aedypqY6qVa5RuSs0V+Oepezac+vFE1v7GEHIpYOT4wvZz42
+QfCBvQyzXM/+PNbytkP0+dS46172T2PL1cnB1cw8ZPKnfnzetlmy9+JTScr
mPJPx3qHJb1M/PfoPW4XKtjknFcPdq7pZfwTvEvbnlQwb71PIY8De5nf48QL
dvKVbF/IWD/XsF7mL52p/1KzkkV99vD8frSX2Y34/VBZv5JdTm6yUU/uZa4+
ZU1FNpWsRb1zgNedXnbhjoLTp32VrG+HaVfPg172TH7uUBZZyRSFu75FlfWy
c3WeM5fGV7IJx/pLct/2sljnp//duFrJNsgOCJfp6WXa9ZJ6tnWVLMDTLiBB
oo+N3m/Yf7C1koU/iNhgKNfHfs4WXNjTVckyA5XnOWn0saUOD2eeHyhkjV2a
iummfeyB2/unr42FzPxBiP07ro+5pm0uu8cJWeTR9jDVuX1M5oFtyUpbITPU
fcLf/28fe9r+8U6Oi5AFthtb31jVx6YtsvnHc62QlRekBDV79bHB5rZPX24U
Mt8lu8Scd/WxScJEzfu7heyh6Gd+VGAfK0soXjfvoJANbXLY8yCsj/GNJ7wL
CxeyW4F63VPi+5hJDrdI55SQyc8/YeaV2sfmDb5UF3RGyFaoi/mdvtjHHvq8
LohLEzJednWH3O0+FlPm5/MlR8gW+VtNtSjuY5k853rtG0KWap29xa+0j018
Wl2qdFvIOpSGZWdV9LGGM9E+N+8KmfXbkJa6GtH5TfPGa94XspPp7Qaan/tE
v6f3mBs9FrKGbSvXOzSLzt+Q+EfqqZAdljVpyO/tY41bj1+qKBeymsoUnQ6p
fvZ5eL/lS6GQTUxWXDtesZ8tsW4vOvRKyPZu2JXqptbPgpbO9O9+I2TPTeo/
xmj1s3kaugXj3gnZKIlF2k91+tnGhp1vFT8I2bay/BUSk/qZ/dOt+jkfhaw4
Ti/RzKSfnb1R0qX4ScjU1p54s5n1s8Npl5N064TMw1BMI21OP3N6dW9dn8i5
3Rucahb0swliw65EfRYy2YfVMcpO/UzHZ2bpe5GXH7OqtHHrZ1arh8k1iZzp
mq2y17OfjZHXuH9N5D7dYYuuefezhd3iC2eKvKAjJKrRr5+tMVcQPyB6/pk7
7WUj9/ezX6eEk0NE7/MjbKWC06F+NvX9amPbWiGzXPrELvJoP/tXQcOjVHSe
6JEmYcVx/awj7YGm4nshq29KedSV3M9eNh/PHPhWyExyFfmTM/rZN135rS9e
C1nogV2zPQT9jNeodn9xtZC9nl9/IPFmP9Nc/qsvplLI9DUWFZUX9bMPe2x3
xr8UsoDP+TzZJ/2sIFgqctUzISvL1mNceT9r1fQ5+fWJkHnPEbud+amf+T/n
2rliIStS3thV29jP8nyrugeJvv4q76pN1TtE77/qyNqzt0T3M736tYU4j+PV
TC9VEwiZpLZil6s0j5OsqtrfmSm6Lykr9Z2yPE7i7OcB8edF9/X88rLLyjxu
y69iKaME0f14hFRojuZx5bzEb5cDhUzG3/PJ6HE87r3Zd/uP/qL3iZxbpK/P
444uz9tR4yNk7Jpc9oypPE6q7Uzk5P+E7J141OHlljxO/r7EOTdr0edTbUug
+xweN3fXOYNvM4UsS2/RzvV2PC7S8IkDN03Idi9U9fBfzOPiFRJrF48WsmFJ
sRbx7jxuptowvVG9lSwoZ6dZqgeP029bm6bRXsmaip0NL67nccYDRifXNlSy
W43DtG5t43G6G47try2vZM5myX9eH+Bx/a80hmSer2SxFReyNc6K3o+7ZJhr
Xcl4DWHnR6XzOK/PU/UrzCqZZ/f6pPGZPG6OvPTO+waVzGTUxPDpV3ncf7yw
YD3lSibcLPBwLuZxCslGV8TeVDBVuVtap+p43JEGvTyrtRUs6p/ScPWxYtzR
CZoH01eUM+uC1Kk1umKc9yNLx8Xzy1nPdP/XSQZinLmOn9GnmeXM02i87hgj
Ma5+bebSK8PKGTcu9N5ESzFu8Ng3a4a/esna5K26LFeJceGTG3Y6z3vJFlTf
8tiUKMbtez1OIXDsC6a4+aLFvcHinL7sj5UVNk9ZgVVu/nUNCU5pb51vVWEh
c/94asZBLQnu8oDJmY4nC5nMnt15i0dJcO6yko4fNhSyRTcsrrXpSXBO5Wkp
Y4YWsnq9p5njzSS45zZKeRs232WKA2vjk5ZJcJP/G9ySYVHAVryS2xUcLcFV
JrseThyfxyR8W345xkpwG1VWGhW9zmXpyi+2j06Q4LpzE7SGH85lP+xittxN
leC+FKtU+zXdYMEF2l5dAglO53Rm09uc6ywrdZrTphcSXMLN1F3KNldZ78aV
xksVJbmUWRf7Niy8zFabNh9WUpbkXg+0z3ctzmIPxP0/lg6W5HL/zStqNc1i
kSdjwi2HS3LOkidPi425xLQKn9YaGkhyOlMT5umKX2SzVM2PyNlKctPOSDs5
fExjKe8e1z2YJ8lt7zz/44FbGpO+sGx6oIMk90KDGzzk4zlWNnPb599Okpxi
T+EF17qzbIVHxox6D0ku4NyijumdKWzvzaENdw9Kck3S5+cVJiaxuqC0mQGH
JLnJUWvHn7uTyGwWGB03iZDk7s/58SyiLoENqlswKytakuspH5WpZBrPziiE
RMenSnIJ7dXtDnKxTLJa9dvS85Lcx3DbX9dsTzLPlBSmfFH0fpdLdD9HxjBD
k4JvoQJJTt7tRGuvzgkWLWbPWV2X5DyMQ36c949mv5+8OtGfJ8mNn6BemCQ8
zu66/eR8CyW5i53fBy5KPsbG6B+ImXxfktNWvLpKRfkYC/05qKnpkSTX5VJU
oXf4KGu6k/RPeqkkt+/i+kl6CkfZwjCDk+7PJbni4E61C7FR7KrjzSatCkmu
L+CWjcXEKDZ0hI3F6yrR+3vbjFJLPMICvlaePPFGkvNeX6Gk0xHJPlxxb174
XpJ7O9N2y+3Vkcxqz3cL+U+S3NZog/w7XyNYus3e2If1klxiZo/Lt0MRTF5F
oSXwmyRXOm/Jj7uWEcz77SnLWS2S3PK9lmvDhkSwivO6cX++S3Kxm8zCbKUi
mOnW6y1Xf4q+XjpzpO7KR7AEcysr7z+if39C7draCRFMjP8yTr9Xknsj+zx4
x38RTDrUR81UTIq7bto/v6YggskOGHLcSkqK66yuN1o+OZINDMsb6DBAilvS
6KI593ak6PvZJdxVQYpTsI2e+db1CFMP75NZpyTFxZ0uXa0tG8W0FJIP+g6W
4vQsHo4r2xTFdAfW7zkyXIqTEf7pl3c4yiZEhXbHj5TiNCLuvWktPcqmKOn7
pY+V4pynh3/pnH+MzVDx3lI4QYo7Llm09PLq42z+kC739plS3L/fI9LTck+w
xbEJH/r/keJ4g6dcU1kSw5zUmau8tRQXHVnaENgew1ZrBi0dO1+Ku8B9XGQw
LZZtH6E4d9kKKa65dOiIjOJTLGHcqIm39khxp3Pfuic0n2bJ6cUZDwOluDqv
YolLNmdYmp7HuIpg0X1kRCr/c+oMu6yfqd0cKcU1VcW0eJons6JJxiojTov+
/rrykm7/FPbVxPZ34B3R+23NPlb+6SwzttlaZMuT5gTlD997Wlxgh6bItSpJ
SnOKKwY7Re+5wGqGndV8xZfm8odt0jK4eYEFfq/08RgozV3frlKYNCmDlZ0y
GxekJc2NP/qo7KDKRebRzAvPN5fm2hXemYy+k8luVsXlBnHS3KHfw+Nbv2cy
xaIpn+2spLl/j69+N0b0/X01Zs2s13bSXJWiuUpJ8CXG4x61dTpLc89MRmXy
Zmex2GNHlhjulOYWfLALNz13mTUG6Ab+ChA9b4h2fWLZZSb65TarYJ80l3F0
/dxNvy6zuhnt/Hmh0lxzzPG49NHZzPDz0pteJ6W5q2Ye2Sre2ezAs9Z6w3hp
buObWSt9A7OZMC9U5XeSNPdRfVz+kehsFhB5c0NwmjQ31HnT1C83stljE60R
qdekufPiStITu7PZsFE37NblSXOTekeOtJPLYd7yC/0m54ueN8V1tbZmDlP7
uP/FnWJpbmn3Sp0Osxy2OuzzgZpyaU7G7sVRa58cds1nT3ZqlTR33H7f1fGB
OUxm5ZC3696IzjujXvDoSA7LmjrX+E+tNKcufWPC1Iwc1lWT2TD4hzS3q+WK
4NGbHNay29l8Yqc090qqQMK/IYfVaslEWXeJ7nNt6+S3HTnssZu7yQ5xPtfO
+71JIC9gt8WUDh+R5nMZZuljdIYKWHZKwbvzsnxu1eRxUnajBSy2Tj24WpnP
LZYvSLxgKmDhBx++alPjc8PDd0nUWQjYPp3tEwZo8rnlB2a3PrcXME/P5xVm
o/mc0KLz9r2VAmbSFDw6fiqfi9ErHjwkUMDGRxjtuGLC595d2ni49ZCAaU2s
LXkyg89Nunh/1e6jAibtPXNrjyWf+zPhasbN0wLWNajxvqoNn6vZ2/5mXZqA
teTEqk+w53OlO/UbX2QKmLC9/a6rI58Tc1gz7kGugJVEJ6v6OoneV3BniGOB
gBVMW+AZ6cLnjgkLN5y6J2ACYc+tNDc+p3N026G4RwJ2bkfGwDtr+NyK7NUv
HJ6KzjvUyb3Kk8+NvJ6Ufu+F6Lx5UjdaN/C5ZQnXtrVWCphPt9sK7e18bqCp
0i7ft6LzJigKTHfyuY0Pnu57+UHAXGbelnLYLXrenfnK3z4J2MJ3Xv967edz
gWHrT+TXC5jV3iGX9h/kcz0RHlvmfRUwU+37vFOH+KKfn60KpxoFzKBwq6Mg
gs81Gw0uT20WMO3V2uklR/ncvqUL9Na3CpiKRFl37Qk+J/5n2862NtF9nfVf
0B3H53qH3jAw+iG6Lyu9VJUkPhcZfT7TuF10X5+FnfopfM5nk4H7b5Frg4Ns
rdL4XLx64M2dHaL7GzclySWDz/33xGT4LZFLHr3/7pPF50YIDTpvi1zgFTE7
QsDn+gZEZu8VWSA7I+7cdT6XObz8hoTIaRcbmvJv8rnJw2/vmCt6/in7GE5Y
wOc63893cRC9T2SzZXRLEZ97HVTxQvO7gO2P/P5F+qHo+T+FVhdE5/GZdHrG
iCd8TnOj1qgu0Xk9n9sfMXkmeh+5y0LFJtF9bumqXVDO526vffDsi+i+Fiqn
G3tW8blHo1/tOvhFdJ9XloTte8Pnik9HzW6oE92no8S72Pd87nD4o8RBtaL7
/Jk9OecTn1uZHP69753oPmNWHHz8hc+12KpdvvJGwFRN5F99bORzL1aVWE2q
FjCZ6jyDrlbR57veeOTWCgFrVR9cMf636L5DJnstKBX9fzIHdG1F5g0MiJso
8qnOtgAvka30NPMVRLYuLx2bLvLp4c7vnjwRsKTwoB1j/4ieP7stdI7I9v3t
GtpdfK4pdtxe8xIBu1D3ctXgHj73+KHyz2Giz2PfS+66kchRwz2Pdj0UMMfC
LFlHkdc5jB9ZLTIvKezKMZG1XX9rHBfZaZmF1KBePuds+3yHjMjSJTnpsn18
ruuSelXbfdH95Y7o1RM5fVBsxlORBWkRDnNFvrTzyJYMkV0PeHWFiPzHs+mV
u8hXzUfaS/WLzivloyosFjBZ/SNnxoicVPvsgEBkN/XeDkuRZ7/JqI8UWb6z
OjFQZGPdj8FzRF5dZ/0jWeQ7YrOKR4uc+/KqdaHIz3nDf/aLvv8UC0fFfxBZ
uWGPeo3I/7unyRnsWjM59//5Lu9//9C+JjLtbSLT/iYy7XEi0z4nMu11ItN+
JzLteSLTvicy7X0i0/4nMu2BItM+KDLthSLTfigy7Yki074oMu2NItP+KDLt
kSLTPiky7ZUi034pMu2ZItO+KTLtnSLT/iky7aEi0z4qMu2lItN+KjLtqSLT
vioy7a0i0/4qMu2xItM+KzLttSLTfisy7bki074rMu29ItP+KzLtwSLTPiwy
7cUi034sMu3JItO+LDLtzSLT/iwy7dEi0z4tMu3VItN+LTLt2SLTvi0y7d0i
0/4tMu3hItM+LjLt5SLTfi4y7eki074uMu3tItP+LjLt8SLTPi8y7fUi034v
Mu35ItO+LzLt/SLT/i8y7QEj0z4wMu0FI9N+MDLtCSPTvjAy7Q0j0/4wMu0R
I9M+MTLtFSPTfjEy7Rkj074xMu0dI9P+MTLtISPTPjIy7SUj034yMu0pI9O+
MjLtLSPT/jIy7TEj0z4zMu01I9N+MzLtOSPTvjMy7T0j0/4zMu1BI9M+NDLt
RSPTfjQy7Ukj0740Mu1NI9P+NDLtUSPTPjUy7VUj0341Mu1ZI9O+NTLtXSPT
/jUy7WEj0z42Mu1lI9N+NjLtaSPTvjYy7W0j0/42Mu1xI9M+NzLtdSPTfjcy
7Xkj0743Mu19I9P+NzLtgSPTPjgy7YUj0344Mu2JI9O+ODLtjSPT/jgy7ZEj
0z45Mu2VI9N+OTLtmSPTvjky7Z0j0/45Mu2hI9M+OjLtpSPTfjoy7akj0746
Mu2tI9P+OjLtsSPTPjsy7bUj0347Mu25I9O+OzLtvSPT/jsy7cEj0z48Mu3F
I9N+PDLtySPTvjwy7c0j0/48Mu3RI9M+PTLt1SPTfj0y7dkj0749Mu3dI9P+
PTLt4SPTPj4y7eUj034+Mu3pI9O+PjLt7SPT/j4y7fEj0z4/Mu31I9N+PzLt
+SPTvj8y7f0j0/4/MvUAkKkPgEy9AGTqByBTTwCZ+gLI1BtApv4AMvUIkKlP
gEy9AmTqFyBTzwCZ+gbI1DtApv4BMvUQkKmPgEy9BGTqJyBTTwGZ+grI1FtA
pv4CMvUYkKnPgEy9BmTqNyBTzwGZ+g7I1HtApv4DMvUgkKkPgUy9CGTqRyBT
TwKZ+hLI1JtApv4EMvUokKlPgUy9CmTqVyBTzwKZ+hbI1LtApv4FMvUwkKmP
gUy9DGTqZyBTTwOZ+hrI1NtApv4GMvU4kKnPgUy9DmTqdyBTzwOZ+h7I1PtA
pv4HMvVAkKkPgky9EGTqhyBTTwSZ+iLI1BtBpv4IMvVIkKlPgky9EmTqlyBT
zwSZ+ibI1DtBpv4JMvVQkKmPgky9FGTqpyBTTwWZ+irI1FtBpv4KMvVYkKnP
gky9FmTqtyBTzwWZ+i7I1HtBpv4LMvVgkKkPg0y9GGTqxyBTTwaZ+jLI1JtB
pv4MMvVokKlPg0y9GmTq1yBTzwaZ+jbI1LtBpv4NMvVwkKmPg0y9HGTq5yBT
TweZ+jrI1NtBpv4OMvV4kKnPg0y9HmTq9yBTzweZ+j7I1PtBpv4PMvWAkKkP
hEy9IGTqByFTTwiZ+kLI1BtCpv4QMvWIkKlPhEy9ImTqFyFTzwiZ+kbI1DtC
pv4RMvWQkKmPhEy9JGTqJyFTTwmZ+krI1FtCpv4SMvWYkKnPhEy9JmTqNyFT
zwmZ+k7I1HtCpv4TMvWgkKkPhUy9KGTqRyFTTwqZ+lLI1JtCpv4UMvWokKlP
hUy9KmTqVyFTzwqZ+lbI1LtCpv4VMvWwkKmPhUy9LGTqZyFTTwuZ+lrI1NtC
pv4WMvW4kKnPhUy9LmTqdyFTzwv5//pewNT7Qqb+FzL1wJCpD4ZMvTBk6och
U08MmfpiyNQbQ6b+GDL1yJCpT4ZMvTJk6pchU88MmfpmyNQ7Q6b+GTL10JCp
j4ZMvTRk6qchU08NmfpqyNRbQ6b+GjL12JCpz4ZMvTZk6rchU88NmfpuyNR7
Q6b+GzL14JCpD4dMvThk6schU08OmfpyyNSbQ6b+HDL16JCpT4dMvTpk6tch
U88Omfp2yNS7Q6b+HTL18JCpj4dMvTxk6uchU08Pmfp6yNTbQ6b+HjL1+JCp
z4dMvT5k6vchU88Pmfp+yNT7Q6b+HzL1AJGpD4hMvUBk6gciU08QmfqCyNQb
RKb+IDL1CJGpT4hMvUJk6hciU88QmfqGyNQ7RKb+ITL1EJGpj4hMvURk6ici
U08RmfqKyNRbRKb+IjL1GJGpz4hMvUZk6jciU88RmfqOyNR7RKb+IzL1IJGp
D4lMvUhk6kciU08SmfqSyNSbRKb+JDL1KJGpT4lMvUpk6lciU88SmfqWyNS7
RKb+JTL1MJGpj4lMvUxk6mciU08TmfqayNTbRKb+JjL1OJGpz4lMvU5k6nci
U88TmfqeyNT7RKb+JzL1QJGpD4pMvVBk6ociU08UmfqiyNQbRab+KDL1SJGp
T4pMvVJk6pciU88UmfqmyNQ7Rab+KTL1UJGpj4pMvVRk6qciU08VmfqqyNRb
Rab+KjL1WJGpz4pMvVZk6rciU88VmfquyNR7Rab+KzL1YJGpD4tMvVhk6sci
U08WmfqyyNSbRab+LDL1aJGpT4tMvVpk6tciU88Wmfq2yNS7Rab+LTL1cJGp
j4tMvVxk6uciU08Xmfq6yNTbRab+LjL1eJGpz4tMvV5k6vciU88Xmfq+yNT7
Rab+LzL1gJGpD4xMvWBk6gcjU08YmfrCyNQbRqb+MDL1iJGpT4xMvWJk6hcj
U88YmfrGyNQ7Rqb+MTL1kJGpj4xMvWRk6icjU08ZmfrKyNRbRqb+MjL1mJGp
z4xMvWZk6jcjU88Z+f/6zsDUe0am/jMy9aCRqQ+NTL1oZOpHI1NPGpn60sjU
m0am/jQy9aiRqU+NTL1qZOpXI1PPGpn61sjUu0am/jUy9bCRqY+NTL1sZOpn
I1NPG5n62sjU20am/jYy9biRqc+NTL1uZOp3I1PPG5n63sjU+0am/jcy9cCR
qQ+OTL1wZOqHI1NPHJn64sjUG0em/jgy9ciRqU+OTL1yZOqXI1PPHJn65sjU
O0em/jky9dCRqY+OTL10ZOqnI1NPHZn66sjUW0em/joy9diRqc+OTL12ZOq3
I1PPHZn67sjUe0em/jsy9eCRqQ+PTL14ZOrHI1NPHpn68sjUm0em/jwy9eiR
qU+PTL16ZOrXI1PPHpn69sjUu0fmjcjf++I6tvPiMT1GN3OhB97+0dF99ya0
X8SR+kNlBdDnxt07O2PhHeiXhZ2rm19gq3VnbU5fdxfae5Jak9ymQujgnklP
6rKwEx/PzchvxS5x3+25eUsR9NjYurqXPvegzdf23bt4DXvRlKGpQZ3Y+0rt
Vhn7FUPHxK3lBuZhX/pv74iGP9iveVfexvnfh257Wnp7621sfnx9vF0vttE0
DefePQ+gbcWNzIR3sFc9mzf0Mu8hdKTXfqFbIHboE7FrdQXYgRMCj3t2Ywcc
Ed/aZPIIesf3wIXePthbF0tM6sjG3nDtgIJfM/Z/QySbuvUeQ7vtDCrZ+x+2
8xvJCxKp2I4zD4aEvMeef1rqP7lhJdA2YsFWR5ywLdylR6ucwDa/Hyx28gW2
8Tj+Bw3FJ9CTD4UUJNliGzTyE0eFYOvMC/VPu4etfVnGeTyvFFpD6ZBpljm2
6rYBQ6bsxFasPPTz2jVsGRPZCrMf2OJxYYL8iU+he7pkj/6zHrvT5fDm++ex
2wrk5s+tw27UDjd4ql0GXR8oL+fgiv2hLvxrRRx2RXpEWo3yM+gyWcWDbguw
H22IdK87jL0ids4gX/Hn0Msfh/Tf0MZ26nrY8mcW9mIXm6e7/bEXRoTevhOH
Pb/g0UWxXGwb7blhBzuwzevmzonY+ALabHCY8bPD2MbWJWOVMrAN020lT9Rj
j9lgV5iw8iX0yMTD2e92Y2uVPTmtnYA91NB+z9lqbIUO+xmZDuXQg15Pcd+x
FVv17tDDFsexNcLrXldXYOuOCdgl5VQBrS+7OuW5H/bEtjkl8XHY026raEx5
g23heDFvxYpK6NnTj37U24tto71jwM/T2AuaLP49/BHbJej17xtrhNC+VweY
Kq2vgt55qnVlzWHsgH2VIeczsQ/YJ1fNbMGOqjPdsW5LNfRFVc/r93a8gv7g
88DIN+gNdM0jKY30s9jVw2b3vyrGfl587/FMqRrou6qFK6SCsU9fvRUaHfIW
ekVHdo0g7D20s01bYV0G9pKESefVnmDrjKoyzZf6AK1VskRipDW2osbi2IYH
2PLFP8LEqrBlNx3brfEFW7roxWp7/kfoPq+FE7LnYLfkzSva8Qh77ICHPv5V
2C5O3Li99dglnVPCgyVroc9NHbr0BIddE3hUJm4BtvJL2VsJK7D3efeOOBuA
7XKp9qsgF/t4z/KE6w+xS+wq598UYpt+e3ilsAN7s9k//92X+ASdFnpz6GNl
bJVxl3Y/N8S29dUxrGDY++6frq2aj93ifmzO+w3YY6/IddX6Y7uIH8ysD8Mu
ObNDqSUdW6yt7d73G9imbJ3vzwfYmyM/6f6pxE576/Kmpw77rYEwgteBrRqw
gJOUqIPer2FxTm4kdq7XrWUDDbFbco1kVRj2WJms22rzsV2Wjdus4Yp9PO3M
SK0N2CU/1StG+mOLzT4ePDYM2zRa3kwvDnvzp4ONBunYaVP6Ew1vYL/d77fQ
6AG26ovv4qaV2Hba66/NqMMO3FznwdqxcwtcNSzFP0O3KlSVWith67gu3Gur
je2S+Xjy/EnYx7st6hxmYZfY3o5ZMg9b7NS0uf+6YJt+zep2WY+92VQ3y20X
dlpIstuaQ9huORsvZ8Zha7wx6+u4gB058WX8wRJsG6ekb09fY4sHrjNTa8T2
rRSvSpOrhzbsLxvbqon9TTfex8QAe0XAVOVH9tjWo9fM+xqBzZtnmDA5Cfum
b8+3nVnYE0uiQ2WfYattuX9Pf9AX6PoCnembjmGfafgRej0F21n5TlWfALts
7bLtR8qxr8uHZmerNkAHO3/VaY/BHtt5UcUv/iv00PsOoy5mYMtG/zJ8l4fd
NtlyvlU19u2Nr0MHDf4G7fhZpi89EnvO1Sz5mkTs6QccNQdewtYeecZ0+xPs
RheTbf/INEIHVqxteLUfe3uq3C/5o9ieW3OkuDPY8wf1jkorwFa3j3bd2o2d
fa/opez2JujUY54fZwZhx6xSbPM+jh3Q/69CVQ72nBnf56S2YL8TjMif4dkM
/dR06/hRO7BvFxSflAnGji/x2lqZiu1Ue2Xc5vfYczylTyxpwTZucRI378VW
7eqtkdFsgX6hZHMsZSn23ZOn+kPXYmcPb96w2Qc7Uu/YHPNj2Lsv110dlYy9
YZrJqAHZ2Lb/1PRUPsU2ezhx3e0abN15+6tSGrH5/+oINg9ohS729vYYMBv7
SmdRReti7JSAwRbC1dj7Q28OS92H7a2oEHboCLZb9Mpfm5OwZ52RfGl+G3uC
zjI2+gn2sMwLmQNeY8tP6VFva8DuvjE/RPgL+2fSwl0rJNqgmz/+yX4hi103
JvWLlTJ2jYe9Vq46dkXGT0f9kdhPmpMOJ+liFxnaFCkZYt/c9v13kAm24Pqp
Sb9nYV/4Y/nfemvsM+bNCe/mYcfujSl3WIIdVcRk77tgh0h95UzXYO+xObbj
4nps38MzsrS2YW8qq6s7ugv7P6VITalA7BWOJov8DmEvOfkhtDEKO3LOFQWx
OOxHvw4eU0vGFkt3GmJwAXu6k37CPznY22T6tJflYWfmPj+3oRBbS9338skX
2Mse2xhdeoUdtVMzr+gjtsTru0XN37HNw47PkejC9p3+X+lQ8e/QDafkhJbK
2CPt3jn/q4Ht3J39ftMo7NLlS7/GTcGWktfbdHk69qzb3T+KLbBzhiX3tC7C
/la6bb/Ucuwxu62lNd2xY942DrTehr0ryXh0wnHsK/MHpOfEYzf1vTF4mIq9
YuV+4x9XsC21S+faVGAHPE8qc63Bvr5vy+Jtddi6H9Vckzqw5VNWeXeo/oCe
vcjo54Dh2HvE+btGjMVuW515wHYaduXoXzFnlmAnnQvPt4/B9tG1T6lKxJ57
US5k9TnsjuywhX5Xse1uh35KfYn9+2XQgG7FdmgH3u4lF0KwdQ6YTzc6gt0j
1aN1JwY7Xc6/oeIcdp/aTn9eMXbGhO3Jy3gd0JLLNzZL7cTWtFuqPXU/9pQZ
bLHbIWw3TaW8m3HY+W+uBm2+ie3r0qP5qhv7q+th28w9P6HL3C6cXxTUCR04
eu/0n4exp9U7Pj15HDthPa+9JgV7ve+yWR6F2DLhUpX+fdgWN1aJn/P7Bd25
0+TEnH3YF8wVdL+FYA8qyp036ST2+9KBcXnXsANq8yeV/cC+pqDh+mvDb+h6
fevT23ywh9pu/djqjx0QXLK2IQzbqn/Xplfp2L7Dz2cvycZOn1H+48UNbDk/
/R0lD7Bftr3Ze6sOW3KgTJFJE7bxBCPJq+3YcV7hoZnif6BLQnJLdGWxe87V
yZ9Twp5YPGjBSHVst1rzo4na2Md4XuXqutjFWicGx0zC7jQvXKZsgj1ueXNc
5CzslvLElL0u2No/Htd1u2MvGtSp47ce+5r9/Iveu7AD7vddWRWLLXtj5X3L
CmydUyMb+PZd0HFjd/jpOGLL55TKzHbBbn+4Qy9wA/bdn0/X9YRjL3fY1dz2
FLus5tnugUJsC8+xihPfYevufT5xfQt2Z6aO92fFbugomfIfrxZgS0TrHvi9
DNt3xB6VIW7YrsZ6Ro7e2OPX7N1eFoV9/47+76IX2F2+Qb0XF/dAy1Y+aGbL
sTWmDnhXvhrbrDWioHsLtp/Hyb12UdidyzLEGkuxpa83fd9Tga2mOqlWuQZ7
2vMrRdMbsX1s8g+EDeiF/mH6XGq8Nbb4SeXO/HnYyj8d6x2WYE/OefVg5xps
b71PIY8DsfeFjPVzDcOO+uzh+f0o9uXkJhv1ZOwW9c4BXnew+3aYdvU8wFYU
7voWVYY94Vh/Se5b7A2yA8JlerADPO0CEiT6oMMfRGwwlMPODFSe56SB3dil
qZhuim3+IMT+HYcdebQ9THUutqHuE/7+f7ED242tb6zCLi9ICWr2wvZdskvM
eRf2Q+16FhWIPbTJYc+DMOxbgXrdU+Kx5eefMPNKxV6hLuZ3+iI2L7u6Q+42
9iJ/q6kWxdip1tlb/EqxO5SGZWdVYFu/DWmpq8E+md5uoPkZu2HbyvUOzdiH
ZU0a8nuxaypTdDqk+qEnJiuuHa+IvXfDrlQ3NeznJvUfY7SwR0ks0n6qg72t
LH+FxCTs4ji9RDMTbLW1J95sZtgehmIaaXOwc7s3ONUswJZ9WB2j7IS9/JhV
pY0bdqZrtspeT+w+3WGLrnljL+gIiWr0wz5zp71s5H7sH2ErFZwOYVsufWIX
eRQ7eqRJWHEcdn1TyqOuZGyTXEX+5Azs0AO7ZnsIsF/Prz+QeBNbX2NRUXkR
dsDnfJ7sE+yybD3GlWN7zxG7nfkJu0h5Y1dtI7bKu2pT9Q7s5dOrX1uI8zhk
SW3FLldp7EwpK/Wdsth9zy8vu6yMfcYjpEJzNLaMv+eT0eOwvSPnFunrY7Nr
ctkzpmK/E486vNwS21ptS6D7HOwsvUU719th716o6uG/GHtYUqxFvDt2UM5O
s1QP7KZiZ8OL67FvNQ7TurUN29ks+c/rA9ixFReyNc5i8xrCzo9Kx/bsXp80
PhPbZNTE8OlXsYWbBR7Oxdiqcre0TtVhR/1TGq4+VgzauiB1ao0uds90/9dJ
BtieRuN1xxhhc+NC7020xG6Tt+qyXIW9oPqWx6ZEbMXNFy3uDRaHLrDKzb+u
IQHt/vHUjINa2DJ7ductHoW96IbFtTY97Hq9p5njzbAVB9bGJy3DXvFKbldw
NLaEb8svx1jsdOUX20cnYP+wi9lyNxU7uEDbq0uAnZU6zWnTC+zejSuNlypK
Qq82bT6spIz9QNz/Y+lg7MiTMeGWw7G1Cp/WGhpgz1I1PyJni53y7nHdg3nY
0heWTQ90wC6bue3zbyfsFR4ZM+o9sPfeHNpw9yB2XVDazIBD2DYLjI6bRGAP
qlswKysa+4xCSHR8KrZkteq3peexPVNSmPJFbEOTgm+hAuxoMXvO6jr27yev
TvTnYd91+8n5FmKP0T8QM/k+dujPQU1Nj7Cb7iT9k16KvTDM4KT7c+yrjjeb
tCqwh46wsXhdhR3wtfLkiTfYH664Ny98j22157uF/CfsdJu9sQ/rseVVFFoC
v2F7vz1lOasFu+K8btyf79imW6+3XP2JnWBuZeX9B1uM/zJOvxdbOtRHzVRM
Clp2wJDjVlLYA8PyBjoMwFaVcwl3VcBWD++TWaeEraWQfNB3MLbuwPo9R4Zj
T4gK7Y4fiT1FSd8vfSz2DBXvLYUTsOcP6XJvn4m9ODbhQ/8/2E7qzFXeGnu1
ZtDSsfOxt49QnLtsBXbCuFETb+3BTk4vzngYiJ2m5zGuIhj7sn6mdnMkdtEk
Y5URp7G/mtj+DryDbWyztciWJw19aIpcq5Ikds2ws5qv+NiB3yt9PAZil50y
Gxekhe3RzAvPN8e+WRWXG8RhKxZN+WxnhX01Zs2s13bYPO5RW6czduyxI0sM
d2I3BugG/grAnuVRmFWwD7tuRjt/Xii24eelN71OYh941lpvGI8tzAtV+Z2E
HRB5c0NwGvZjE60Rqdewh426YbcuD9tbfqHf5HxstY/7X9wpxl4d9vlATTn2
NZ892alV2DIrh7xd9wY7a+pc4z+12F01mQ2Df2C37HY2n9iJXaslE2Xdhf3Y
zd1khzgf+raY0uEj0tjZKQXvzstix9apB1crY4cffPiqTQ17n872CQM0sT09
n1eYjcY2aQoeHT8Ve3yE0Y4rJthaE2tLnszAlvaeubXHErtrUON9VRvslpxY
9Qn22ML29ruujtgl0cmqvk7YBdMWeEa6YAuEPbfS3LDP7cgYeGcNduxQJ/cq
T+zwPKkbrRuwfbrdVmhvx/ZMUBSY7sR2mXlbymE39sJ3Xv967ce22jvk0v6D
2Kba93mnDmEbFG51FERga6/WTi85iq0iUdZdewJb+qz/gu447C4rvVSVJOyW
z8JO/RTs2uAgW6s0bOG4KUkuGdglj95/98nCLvCKmB0hwBbIzog7dx077WJD
U/5N7FP2MZywADuy2TK6pQh7f+T3L9IPsX0mnZ4x4gm253P7IybPsF22dNUu
KMdeqJxu7FmFbXVlSdi+N9imjhLvYt9jG/zMnpzzCVs7ZsXBx1+wVU3kX31s
xJapzjPoasVuVR9cMf43tpXMAV3bv/hUZ1uA119sXV46Nv0vTgoP2jH2D7Z9
f7uGdhf2hbqXqwb3YPe95K4b/cWOhVmyjn8xLynsyrG/2GmZhdSgXmzpkpx0
2T5sl9wRvXp/sSAtwmHuX+x6wKsr5C++aj7SXqofW1b/yJkxf7Gbem+H5V8s
31mdGPgXr66z/pH8F+e+vGpd+BcrFo6K//AXu1+Oau3///w/tKJTHQ==
    "], {
     {GrayLevel[0.5], EdgeForm[None], 
      GraphicsGroupBox[PolygonBox[CompressedData["
1:eJwtnAfcl2P/hn/3+O3xSJIkSSVJS4VKJSJaUtHQHpIkIQlZSZIkCUlWdlbI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        
        "]]]}, {{}, {}, {
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1, 2}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{2, 3}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{3, 4}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{4, 5}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{5, 6}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{6, 7}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{7, 8}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{8, 9}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{9, 10}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{10, 11}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{11, 12}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{12, 13}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{13, 14}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{14, 15}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{15, 16}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{16, 17}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{17, 18}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{18, 19}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{19, 20}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{20, 21}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{21, 22}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{22, 23}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{23, 24}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{24, 25}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{25, 26}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{26, 27}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{27, 28}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{28, 29}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{29, 30}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{30, 31}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{31, 32}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{32, 33}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{33, 34}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{34, 35}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{35, 36}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{36, 37}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{37, 38}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{38, 39}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{39, 40}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{40, 41}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{41, 42}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{42, 43}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{43, 44}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{44, 45}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{45, 46}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{46, 47}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{47, 48}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{48, 49}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{49, 50}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{50, 51}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{51, 52}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{52, 53}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{53, 54}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{54, 55}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{55, 56}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{56, 57}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{57, 58}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{58, 59}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{59, 60}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{60, 61}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{61, 62}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{62, 63}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{63, 64}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{64, 65}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{65, 66}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{66, 67}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{67, 68}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{68, 69}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{69, 70}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{70, 71}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{71, 72}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{72, 73}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{73, 74}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{74, 75}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{75, 76}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{76, 77}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{77, 78}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{78, 79}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{79, 80}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{80, 81}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{81, 82}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{82, 83}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{83, 84}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{84, 85}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{85, 86}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{86, 87}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{87, 88}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{88, 89}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{89, 90}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{90, 91}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{91, 92}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{92, 93}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{93, 94}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{94, 95}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{95, 96}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{96, 97}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{97, 98}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{98, 99}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{99, 100}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{100, 101}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{101, 102}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{102, 103}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{103, 104}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{104, 105}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{105, 106}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{106, 107}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{107, 108}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{108, 109}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{109, 110}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{110, 111}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{111, 112}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{112, 113}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{113, 114}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{114, 115}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{115, 116}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{116, 117}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{117, 118}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{118, 119}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{119, 120}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{120, 121}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{121, 122}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{122, 123}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{123, 124}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{124, 125}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{125, 126}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{126, 127}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{127, 128}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{128, 129}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{129, 130}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{130, 131}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{131, 132}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{132, 133}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{133, 134}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{134, 135}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{135, 136}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{136, 137}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{137, 138}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{138, 139}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{139, 140}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{140, 141}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{141, 142}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{142, 143}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{143, 144}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{144, 145}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{145, 146}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{146, 147}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{147, 148}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{148, 149}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{149, 150}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{150, 151}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{151, 152}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{152, 153}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{153, 154}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{154, 155}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{155, 156}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{156, 157}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{157, 158}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{158, 159}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{159, 160}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{160, 161}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{161, 162}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{162, 163}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{163, 164}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{164, 165}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{165, 166}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{166, 167}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{167, 168}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{168, 169}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{169, 170}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{170, 171}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{171, 172}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{172, 173}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{173, 174}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{174, 175}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{175, 176}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{176, 177}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{177, 178}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{178, 179}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{179, 180}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{180, 181}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{181, 182}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{182, 183}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{183, 184}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{184, 185}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{185, 186}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{186, 187}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{187, 188}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{188, 189}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{189, 190}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{190, 191}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{191, 192}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{192, 193}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{193, 194}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{194, 195}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{195, 196}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{196, 197}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{197, 198}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{198, 199}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{199, 200}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{200, 201}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{201, 202}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{202, 203}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{203, 204}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{204, 205}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{205, 206}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{206, 207}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{207, 208}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{208, 209}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{209, 210}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{210, 211}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{211, 212}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{212, 213}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{213, 214}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{214, 215}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{215, 216}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{216, 217}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{217, 218}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{218, 219}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{219, 220}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{220, 221}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{221, 222}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{222, 223}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{223, 224}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{224, 225}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{225, 226}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{226, 227}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{227, 228}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{228, 229}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{229, 230}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{230, 231}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{231, 232}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{232, 233}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{233, 234}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{234, 235}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{235, 236}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{236, 237}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{237, 238}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{238, 239}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{239, 240}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{240, 241}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{241, 242}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{242, 243}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{243, 244}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{244, 245}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{245, 246}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{246, 247}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{247, 248}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{248, 249}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{249, 250}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{250, 251}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{251, 252}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{252, 253}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{253, 254}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{254, 255}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{255, 256}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{256, 257}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{257, 258}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{258, 259}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{259, 260}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{260, 261}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{261, 262}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{262, 263}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{263, 264}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{264, 265}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{265, 266}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{266, 267}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{267, 268}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{268, 269}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{269, 270}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{270, 271}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{271, 272}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{272, 273}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{273, 274}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{274, 275}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{275, 276}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{276, 277}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{277, 278}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{278, 279}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{279, 280}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{280, 281}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{281, 282}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{282, 283}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{283, 284}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{284, 285}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{285, 286}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{286, 287}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{287, 288}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{288, 289}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{289, 290}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{290, 291}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{291, 292}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{292, 293}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{293, 294}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{294, 295}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{295, 296}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{296, 297}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{297, 298}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{298, 299}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{299, 300}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{300, 301}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{301, 302}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{302, 303}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{303, 304}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{304, 305}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{305, 306}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{306, 307}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{307, 308}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{308, 309}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{309, 310}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{310, 311}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{311, 312}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{312, 313}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{313, 314}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{314, 315}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{315, 316}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{316, 317}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{317, 318}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{318, 319}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{319, 320}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{320, 321}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{321, 322}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{322, 323}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{323, 324}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{324, 325}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{325, 326}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{326, 327}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{327, 328}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{328, 329}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{329, 330}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{330, 331}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{331, 332}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{332, 333}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{333, 334}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{334, 335}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{335, 336}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{336, 337}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{337, 338}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{338, 339}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{339, 340}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{340, 341}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{341, 342}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{342, 343}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{343, 344}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{344, 345}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{345, 346}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{346, 347}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{347, 348}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{348, 349}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{349, 350}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{350, 351}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{351, 352}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{352, 353}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{353, 354}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{354, 355}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{355, 356}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{356, 357}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{357, 358}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{358, 359}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{359, 360}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{360, 361}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{361, 362}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{362, 363}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{363, 364}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{364, 365}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{365, 366}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{366, 367}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{367, 368}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{368, 369}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{369, 370}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{370, 371}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{371, 372}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{372, 373}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{373, 374}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{374, 375}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{375, 376}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{376, 377}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{377, 378}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{378, 379}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{379, 380}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{380, 381}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{381, 382}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{382, 383}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{383, 384}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{384, 385}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{385, 386}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{386, 387}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{387, 388}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{388, 389}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{389, 390}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{390, 391}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{391, 392}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{392, 393}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{393, 394}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{394, 395}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{395, 396}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{396, 397}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{397, 398}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{398, 399}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{399, 400}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{400, 401}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{401, 402}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{402, 403}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{403, 404}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{404, 405}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{405, 406}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{406, 407}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{407, 408}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{408, 409}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{409, 410}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{410, 411}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{411, 412}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{412, 413}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{413, 414}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{414, 415}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{415, 416}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{416, 417}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{417, 418}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{418, 419}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{419, 420}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{420, 421}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{421, 422}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{422, 423}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{423, 424}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{424, 425}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{425, 426}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{426, 427}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{427, 428}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{428, 429}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{429, 430}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{430, 431}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{431, 432}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{432, 433}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{433, 434}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{434, 435}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{435, 436}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{436, 437}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{437, 438}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{438, 439}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{439, 440}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{440, 441}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{441, 442}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{442, 443}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{443, 444}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{444, 445}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{445, 446}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{446, 447}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{447, 448}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{448, 449}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{449, 450}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{450, 451}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{451, 452}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{452, 453}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{453, 454}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{454, 455}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{455, 456}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{456, 457}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{457, 458}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{458, 459}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{459, 460}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{460, 461}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{461, 462}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{462, 463}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{463, 464}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{464, 465}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{465, 466}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{466, 467}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{467, 468}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{468, 469}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{469, 470}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{470, 471}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{471, 472}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{472, 473}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{473, 474}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{474, 475}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{475, 476}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{476, 477}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{477, 478}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{478, 479}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{479, 480}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{480, 481}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{481, 482}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{482, 483}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{483, 484}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{484, 485}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{485, 486}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{486, 487}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{487, 488}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{488, 489}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{489, 490}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{490, 491}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{491, 492}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{492, 493}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{493, 494}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{494, 495}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{495, 496}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{496, 497}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{497, 498}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{498, 499}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{499, 500}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{500, 501}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{501, 502}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{502, 503}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{503, 504}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{504, 505}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{505, 506}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{506, 507}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{507, 508}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{508, 509}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{509, 510}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{510, 511}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{511, 512}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{512, 513}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{513, 514}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{514, 515}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{515, 516}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{516, 517}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{517, 518}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{518, 519}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{519, 520}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{520, 521}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{521, 522}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{522, 523}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{523, 524}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{524, 525}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{525, 526}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{526, 527}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{527, 528}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{528, 529}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{529, 530}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{530, 531}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{531, 532}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{532, 533}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{533, 534}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{534, 535}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{535, 536}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{536, 537}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{537, 538}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{538, 539}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{539, 540}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{540, 541}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{541, 542}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{542, 543}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{543, 544}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{544, 545}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{545, 546}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{546, 547}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{547, 548}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{548, 549}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{549, 550}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{550, 551}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{551, 552}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{552, 553}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{553, 554}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{554, 555}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{555, 556}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{556, 557}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{557, 558}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{558, 559}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{559, 560}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{560, 561}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{561, 562}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{562, 563}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{563, 564}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{564, 565}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{565, 566}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{566, 567}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{567, 568}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{568, 569}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{569, 570}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{570, 571}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{571, 572}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{572, 573}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{573, 574}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{574, 575}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{575, 576}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{576, 577}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{577, 578}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{578, 579}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{579, 580}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{580, 581}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{581, 582}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{582, 583}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{583, 584}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{584, 585}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{585, 586}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{586, 587}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{587, 588}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{588, 589}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{589, 590}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{590, 591}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{591, 592}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{592, 593}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{593, 594}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{594, 595}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{595, 596}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{596, 597}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{597, 598}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{598, 599}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{599, 600}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{600, 601}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{601, 602}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{602, 603}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{603, 604}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{604, 605}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{605, 606}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{606, 607}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{607, 608}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{608, 609}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{609, 610}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{610, 611}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{611, 612}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{612, 613}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{613, 614}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{614, 615}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{615, 616}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{616, 617}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{617, 618}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{618, 619}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{619, 620}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{620, 621}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{621, 622}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{622, 623}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{623, 624}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{624, 625}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{625, 626}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{626, 627}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{627, 628}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{628, 629}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{629, 630}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{630, 631}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{631, 632}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{632, 633}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{633, 634}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{634, 635}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{635, 636}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{636, 637}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{637, 638}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{638, 639}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{639, 640}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{640, 641}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{641, 642}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{642, 643}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{643, 644}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{644, 645}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{645, 646}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{646, 647}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{647, 648}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{648, 649}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{649, 650}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{650, 651}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{651, 652}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{652, 653}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{653, 654}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{654, 655}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{655, 656}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{656, 657}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{657, 658}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{658, 659}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{659, 660}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{660, 661}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{661, 662}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{662, 663}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{663, 664}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{664, 665}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{665, 666}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{666, 667}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{667, 668}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{668, 669}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{669, 670}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{670, 671}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{671, 672}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{672, 673}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{673, 674}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{674, 675}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{675, 676}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{676, 677}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{677, 678}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{678, 679}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{679, 680}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{680, 681}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{681, 682}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{682, 683}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{683, 684}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{684, 685}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{685, 686}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{686, 687}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{687, 688}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{688, 689}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{689, 690}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{690, 691}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{691, 692}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{692, 693}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{693, 694}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{694, 695}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{695, 696}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{696, 697}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{697, 698}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{698, 699}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{699, 700}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{700, 701}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{701, 702}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{702, 703}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{703, 704}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{704, 705}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{705, 706}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{706, 707}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{707, 708}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{708, 709}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{709, 710}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{710, 711}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{711, 712}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{712, 713}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{713, 714}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{714, 715}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{715, 716}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{716, 717}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{717, 718}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{718, 719}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{719, 720}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{720, 721}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{721, 722}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{722, 723}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{723, 724}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{724, 725}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{725, 726}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{726, 727}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{727, 728}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{728, 729}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{729, 730}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{730, 731}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{731, 732}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{732, 733}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{733, 734}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{734, 735}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{735, 736}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{736, 737}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{737, 738}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{738, 739}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{739, 740}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{740, 741}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{741, 742}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{742, 743}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{743, 744}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{744, 745}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{745, 746}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{746, 747}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{747, 748}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{748, 749}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{749, 750}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{750, 751}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{751, 752}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{752, 753}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{753, 754}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{754, 755}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{755, 756}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{756, 757}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{757, 758}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{758, 759}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{759, 760}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{760, 761}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{761, 762}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{762, 763}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{763, 764}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{764, 765}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{765, 766}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{766, 767}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{767, 768}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{768, 769}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{769, 770}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{770, 771}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{771, 772}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{772, 773}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{773, 774}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{774, 775}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{775, 776}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{776, 777}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{777, 778}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{778, 779}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{779, 780}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{780, 781}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{781, 782}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{782, 783}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{783, 784}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{784, 785}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{785, 786}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{786, 787}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{787, 788}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{788, 789}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{789, 790}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{790, 791}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{791, 792}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{792, 793}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{793, 794}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{794, 795}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{795, 796}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{796, 797}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{797, 798}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{798, 799}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{799, 800}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{800, 801}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{801, 802}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{802, 803}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{803, 804}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{804, 805}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{805, 806}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{806, 807}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{807, 808}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{808, 809}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{809, 810}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{810, 811}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{811, 812}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{812, 813}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{813, 814}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{814, 815}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{815, 816}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{816, 817}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{817, 818}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{818, 819}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{819, 820}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{820, 821}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{821, 822}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{822, 823}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{823, 824}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{824, 825}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{825, 826}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{826, 827}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{827, 828}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{828, 829}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{829, 830}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{830, 831}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{831, 832}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{832, 833}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{833, 834}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{834, 835}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{835, 836}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{836, 837}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{837, 838}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{838, 839}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{839, 840}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{840, 841}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{841, 842}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{842, 843}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{843, 844}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{844, 845}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{845, 846}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{846, 847}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{847, 848}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{848, 849}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{849, 850}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{850, 851}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{851, 852}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{852, 853}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{853, 854}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{854, 855}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{855, 856}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{856, 857}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{857, 858}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{858, 859}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{859, 860}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{860, 861}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{861, 862}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{862, 863}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{863, 864}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{864, 865}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{865, 866}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{866, 867}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{867, 868}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{868, 869}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{869, 870}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{870, 871}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{871, 872}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{872, 873}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{873, 874}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{874, 875}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{875, 876}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{876, 877}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{877, 878}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{878, 879}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{879, 880}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{880, 881}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{881, 882}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{882, 883}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{883, 884}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{884, 885}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{885, 886}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{886, 887}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{887, 888}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{888, 889}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{889, 890}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{890, 891}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{891, 892}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{892, 893}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{893, 894}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{894, 895}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{895, 896}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{896, 897}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{897, 898}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{898, 899}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{899, 900}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{900, 901}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{901, 902}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{902, 903}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{903, 904}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{904, 905}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{905, 906}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{906, 907}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{907, 908}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{908, 909}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{909, 910}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{910, 911}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{911, 912}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{912, 913}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{913, 914}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{914, 915}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{915, 916}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{916, 917}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{917, 918}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{918, 919}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{919, 920}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{920, 921}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{921, 922}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{922, 923}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{923, 924}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{924, 925}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{925, 926}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{926, 927}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{927, 928}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{928, 929}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{929, 930}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{930, 931}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{931, 932}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{932, 933}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{933, 934}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{934, 935}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{935, 936}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{936, 937}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{937, 938}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{938, 939}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{939, 940}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{940, 941}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{941, 942}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{942, 943}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{943, 944}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{944, 945}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{945, 946}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{946, 947}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{947, 948}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{948, 949}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{949, 950}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{950, 951}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{951, 952}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{952, 953}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{953, 954}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{954, 955}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{955, 956}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{956, 957}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{957, 958}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{958, 959}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{959, 960}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{960, 961}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{961, 962}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{962, 963}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{963, 964}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{964, 965}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{965, 966}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{966, 967}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{967, 968}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{968, 969}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{969, 970}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{970, 971}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{971, 972}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{972, 973}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{973, 974}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{974, 975}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{975, 976}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{976, 977}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{977, 978}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{978, 979}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{979, 980}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{980, 981}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{981, 982}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{982, 983}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{983, 984}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{984, 985}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{985, 986}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{986, 987}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{987, 988}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{988, 989}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{989, 990}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{990, 991}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{991, 992}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{992, 993}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{993, 994}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{994, 995}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{995, 996}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{996, 997}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{997, 998}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{998, 999}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{999, 1000}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1000, 1001}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1001, 1002}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1002, 1003}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1003, 1004}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1004, 1005}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1005, 1006}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1006, 1007}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1007, 1008}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1008, 1009}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1009, 1010}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1010, 1011}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1011, 1012}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1012, 1013}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1013, 1014}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1014, 1015}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1015, 1016}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1016, 1017}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1017, 1018}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1018, 1019}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1019, 1020}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1020, 1021}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1021, 1022}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1022, 1023}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1023, 1024}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1024, 1025}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1025, 1026}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1026, 1027}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1027, 1028}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1028, 1029}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1029, 1030}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1030, 1031}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1031, 1032}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1032, 1033}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1033, 1034}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1034, 1035}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1035, 1036}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1036, 1037}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1037, 1038}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1038, 1039}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1039, 1040}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1040, 1041}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1041, 1042}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1042, 1043}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1043, 1044}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1044, 1045}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1045, 1046}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1046, 1047}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1047, 1048}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1048, 1049}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1049, 1050}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1050, 1051}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1051, 1052}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1052, 1053}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1053, 1054}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1054, 1055}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1055, 1056}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1056, 1057}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1057, 1058}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1058, 1059}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1059, 1060}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1060, 1061}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1061, 1062}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1062, 1063}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1063, 1064}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1064, 1065}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1065, 1066}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1066, 1067}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1067, 1068}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1068, 1069}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1069, 1070}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1070, 1071}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1071, 1072}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1072, 1073}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1073, 1074}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1074, 1075}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1075, 1076}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1076, 1077}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1077, 1078}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1078, 1079}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1079, 1080}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1080, 1081}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1081, 1082}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1082, 1083}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1083, 1084}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1084, 1085}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1085, 1086}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1086, 1087}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1087, 1088}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1088, 1089}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1089, 1090}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1090, 1091}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1091, 1092}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1092, 1093}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1093, 1094}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1094, 1095}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1095, 1096}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1096, 1097}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1097, 1098}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1098, 1099}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1099, 1100}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1100, 1101}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1101, 1102}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1102, 1103}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1103, 1104}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1104, 1105}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1105, 1106}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1106, 1107}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1107, 1108}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1108, 1109}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1109, 1110}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1110, 1111}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1111, 1112}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1112, 1113}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1113, 1114}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1114, 1115}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1115, 1116}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1116, 1117}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1117, 1118}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1118, 1119}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1119, 1120}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1120, 1121}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1121, 1122}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1122, 1123}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1123, 1124}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1124, 1125}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1125, 1126}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1126, 1127}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1127, 1128}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1128, 1129}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1129, 1130}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1130, 1131}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1131, 1132}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1132, 1133}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1133, 1134}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1134, 1135}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1135, 1136}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1136, 1137}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1137, 1138}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1138, 1139}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1139, 1140}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1140, 1141}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1141, 1142}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1142, 1143}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1143, 1144}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1144, 1145}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1145, 1146}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1146, 1147}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1147, 1148}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1148, 1149}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1149, 1150}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1150, 1151}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1151, 1152}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1152, 1153}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1153, 1154}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1154, 1155}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1155, 1156}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1156, 1157}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1157, 1158}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1158, 1159}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1159, 1160}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1160, 1161}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1161, 1162}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1162, 1163}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1163, 1164}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1164, 1165}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1165, 1166}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1166, 1167}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1167, 1168}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1168, 1169}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1169, 1170}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1170, 1171}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1171, 1172}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1172, 1173}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1173, 1174}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1174, 1175}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1175, 1176}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1176, 1177}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1177, 1178}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1178, 1179}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1179, 1180}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1180, 1181}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1181, 1182}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1182, 1183}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1183, 1184}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1184, 1185}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1185, 1186}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1186, 1187}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1187, 1188}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1188, 1189}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1189, 1190}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1190, 1191}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1191, 1192}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1192, 1193}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1193, 1194}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1194, 1195}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1195, 1196}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1196, 1197}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1197, 1198}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1198, 1199}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1199, 1200}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1200, 1201}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1201, 1202}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1202, 1203}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1203, 1204}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1204, 1205}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1205, 1206}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1206, 1207}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1207, 1208}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1208, 1209}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1209, 1210}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1210, 1211}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1211, 1212}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1212, 1213}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1213, 1214}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1214, 1215}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1215, 1216}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1216, 1217}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1217, 1218}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1218, 1219}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1219, 1220}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1220, 1221}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1221, 1222}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1222, 1223}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1223, 1224}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1224, 1225}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1225, 1226}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1226, 1227}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1227, 1228}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1228, 1229}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1229, 1230}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1230, 1231}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1231, 1232}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1232, 1233}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1233, 1234}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1234, 1235}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1235, 1236}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1236, 1237}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1237, 1238}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1238, 1239}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1239, 1240}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1240, 1241}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1241, 1242}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1242, 1243}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1243, 1244}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1244, 1245}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1245, 1246}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1246, 1247}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1247, 1248}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1248, 1249}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1249, 1250}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1250, 1251}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1251, 1252}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1252, 1253}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1253, 1254}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1254, 1255}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1255, 1256}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1256, 1257}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1257, 1258}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1258, 1259}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1259, 1260}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1260, 1261}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1261, 1262}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1262, 1263}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1263, 1264}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1264, 1265}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1265, 1266}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1266, 1267}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1267, 1268}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1268, 1269}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1269, 1270}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1270, 1271}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1271, 1272}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1272, 1273}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1273, 1274}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1274, 1275}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1275, 1276}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1276, 1277}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1277, 1278}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1278, 1279}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1279, 1280}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1280, 1281}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1281, 1282}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1282, 1283}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1283, 1284}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1284, 1285}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1285, 1286}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1286, 1287}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1287, 1288}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1288, 1289}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1289, 1290}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1290, 1291}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1291, 1292}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1292, 1293}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1293, 1294}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1294, 1295}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1295, 1296}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1296, 1297}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1297, 1298}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1298, 1299}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1299, 1300}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1300, 1301}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1301, 1302}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1302, 1303}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1303, 1304}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1304, 1305}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1305, 1306}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1306, 1307}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1307, 1308}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1308, 1309}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1309, 1310}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1310, 1311}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1311, 1312}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1312, 1313}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1313, 1314}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1314, 1315}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1315, 1316}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1316, 1317}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1317, 1318}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1318, 1319}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1319, 1320}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1320, 1321}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1321, 1322}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1322, 1323}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1323, 1324}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1324, 1325}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1325, 1326}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1326, 1327}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1327, 1328}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1328, 1329}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1329, 1330}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1330, 1331}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1331, 1332}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1332, 1333}]}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], LineBox[{1333, 1334}]}}}},
    VertexColors->{
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]], 
      Hue[0, 0, 
       Rational[1, 4]]}], GraphicsComplexBox[CompressedData["
1:eJwUV3c81P8fP9w5KxUSyqyEUoq+RfS6UCqJIoRQkrKpkERUViWULTMyCiFJ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    "], {
     {GrayLevel[0.5], EdgeForm[None], 
      GraphicsGroupBox[PolygonBox[CompressedData["
1:eJwt3Af4TnXjx/Ez7n3fPysyslc2WZE9S0uSdqmkpV1SSUsLlUqSkERDKkmF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        "]]]}, {{}, {}, {
       {Hue[0.022535190212674343`, 1, 1], LineBox[{1, 2}]}, 
       {Hue[0.022828064763465546`, 1, 1], LineBox[{2, 3}]}, 
       {Hue[0.02312093931425682, 1, 1], LineBox[{3, 4}]}, 
       {Hue[0.023706688415839228`, 1, 1], LineBox[{4, 5}]}, 
       {Hue[0.02399956296663043, 1, 1], LineBox[{5, 6}]}, 
       {Hue[0.024292437517421633`, 1, 1], LineBox[{6, 7}]}, 
       {Hue[0.024878186619004036`, 1, 1], LineBox[{7, 8}]}, 
       {Hue[0.02517106116979524, 1, 1], LineBox[{8, 9}]}, 
       {Hue[0.025463935720586448`, 1, 1], LineBox[{9, 10}]}, 
       {Hue[0.026049684822168923`, 1, 1], LineBox[{10, 11}]}, 
       {Hue[0.026342559372960126`, 1, 1], LineBox[{11, 12}]}, 
       {Hue[0.02663543392375139, 1, 1], LineBox[{12, 13}]}, 
       {Hue[0.027221183025333804`, 1, 1], LineBox[{13, 14}]}, 
       {Hue[0.027514057576125007`, 1, 1], LineBox[{14, 15}]}, 
       {Hue[0.027806932126916206`, 1, 1], LineBox[{15, 16}]}, 
       {Hue[0.028099806677707416`, 1, 1], LineBox[{16, 17}]}, 
       {Hue[0.028392681228498615`, 1, 1], LineBox[{17, 18}]}, 
       {Hue[0.02868555577928982, 1, 1], LineBox[{18, 19}]}, 
       {Hue[0.028978430330081024`, 1, 1], LineBox[{19, 20}]}, 
       {Hue[0.0292713048808723, 1, 1], LineBox[{20, 21}]}, 
       {Hue[0.029564179431663503`, 1, 1], LineBox[{21, 22}]}, 
       {Hue[0.029857053982454702`, 1, 1], LineBox[{22, 23}]}, 
       {Hue[0.030149928533245978`, 1, 1], LineBox[{23, 24}]}, 
       {Hue[0.030442803084037184`, 1, 1], LineBox[{24, 25}]}, 
       {Hue[0.030735677634828387`, 1, 1], LineBox[{25, 26}]}, 
       {Hue[0.031028552185619583`, 1, 1], LineBox[{26, 27}]}, 
       {Hue[0.031321426736410786`, 1, 1], LineBox[{27, 28}]}, 
       {Hue[0.03161430128720199, 1, 1], LineBox[{28, 29}]}, 
       {Hue[0.031907175837993206`, 1, 1], LineBox[{29, 30}]}, 
       {Hue[0.0322000503887844, 1, 1], LineBox[{30, 31}]}, 
       {Hue[0.0324929249395756, 1, 1], LineBox[{31, 32}]}, 
       {Hue[0.03278579949036681, 1, 1], LineBox[{32, 33}]}, 
       {Hue[0.033078674041158, 1, 1], LineBox[{33, 34}]}, 
       {Hue[0.03337154859194921, 1, 1], LineBox[{34, 35}]}, 
       {Hue[0.033664423142740416`, 1, 1], LineBox[{35, 36}]}, 
       {Hue[0.033957297693531695`, 1, 1], LineBox[{36, 37}]}, 
       {Hue[0.0342501722443229, 1, 1], LineBox[{37, 38}]}, 
       {Hue[0.0345430467951141, 1, 1], LineBox[{38, 39}]}, 
       {Hue[0.034835921345905366`, 1, 1], LineBox[{39, 40}]}, 
       {Hue[0.035128795896696575`, 1, 1], LineBox[{40, 41}]}, 
       {Hue[0.03542167044748777, 1, 1], LineBox[{41, 42}]}, 
       {Hue[0.03571454499827898, 1, 1], LineBox[{42, 43}]}, 
       {Hue[0.03600741954907018, 1, 1], LineBox[{43, 44}]}, 
       {Hue[0.03630029409986138, 1, 1], LineBox[{44, 45}]}, 
       {Hue[0.03659316865065259, 1, 1], LineBox[{45, 46}]}, 
       {Hue[0.0368860432014438, 1, 1], LineBox[{46, 47}]}, 
       {Hue[0.037178917752234995`, 1, 1], LineBox[{47, 48}]}, 
       {Hue[0.0374717923030262, 1, 1], LineBox[{48, 49}]}, 
       {Hue[0.0377646668538174, 1, 1], LineBox[{49, 50}]}, 
       {Hue[0.03805754140460861, 1, 1], LineBox[{50, 51}]}, 
       {Hue[0.03835041595539981, 1, 1], LineBox[{51, 52}]}, 
       {Hue[0.03893616505698228, 1, 1], LineBox[{52, 53}]}, 
       {Hue[0.03922903960777349, 1, 1], LineBox[{53, 54}]}, 
       {Hue[0.039521914158564764`, 1, 1], LineBox[{54, 55}]}, 
       {Hue[0.040107663260147156`, 1, 1], LineBox[{55, 56}]}, 
       {Hue[0.041279161463311974`, 1, 1], LineBox[{56, 57}]}, 
       {Hue[0.041596671037132556`, 1, 1], LineBox[{57, 58}]}, 
       {Hue[0.04191418061095314, 1, 1], LineBox[{58, 59}]}, 
       {Hue[0.0425491997585943, 1, 1], LineBox[{59, 60}]}, 
       {Hue[0.04286670933241487, 1, 1], LineBox[{60, 61}]}, 
       {Hue[0.04318421890623545, 1, 1], LineBox[{61, 62}]}, 
       {Hue[0.04381923805387661, 1, 1], LineBox[{62, 63}]}, 
       {Hue[0.04413674762769719, 1, 1], LineBox[{63, 64}]}, 
       {Hue[0.044454257201517765`, 1, 1], LineBox[{64, 65}]}, 
       {Hue[0.04508927634915892, 1, 1], LineBox[{65, 66}]}, 
       {Hue[0.04635931464444118, 1, 1], LineBox[{66, 67}]}, 
       {Hue[0.04667682421826168, 1, 1], LineBox[{67, 68}]}, 
       {Hue[0.04699433379208225, 1, 1], LineBox[{68, 69}]}, 
       {Hue[0.04762935293972342, 1, 1], LineBox[{69, 70}]}, 
       {Hue[0.04889939123500573, 1, 1], LineBox[{70, 71}]}, 
       {Hue[0.0492169008088263, 1, 1], LineBox[{71, 72}]}, 
       {Hue[0.04953441038264688, 1, 1], LineBox[{72, 73}]}, 
       {Hue[0.05016942953028804, 1, 1], LineBox[{73, 74}]}, 
       {Hue[0.05143946782557036, 1, 1], LineBox[{74, 75}]}, 
       {Hue[0.05175697739939094, 1, 1], LineBox[{75, 76}]}, 
       {Hue[0.05207448697321152, 1, 1], LineBox[{76, 77}]}, 
       {Hue[0.05270950612085268, 1, 1], LineBox[{77, 78}]}, 
       {Hue[0.05397954441613492, 1, 1], LineBox[{78, 79}]}, 
       {Hue[0.054297053989955435`, 1, 1], LineBox[{79, 80}]}, 
       {Hue[0.05461456356377601, 1, 1], LineBox[{80, 81}]}, 
       {Hue[0.05524958271141718, 1, 1], LineBox[{81, 82}]}, 
       {Hue[0.05651962100669947, 1, 1], LineBox[{82, 83}]}, 
       {Hue[0.05683713058052007, 1, 1], LineBox[{83, 84}]}, 
       {Hue[0.05715464015434064, 1, 1], LineBox[{84, 85}]}, 
       {Hue[0.057789659301981794`, 1, 1], LineBox[{85, 86}]}, 
       {Hue[0.059059697597264114`, 1, 1], LineBox[{86, 87}]}, 
       {Hue[0.061599774187828665`, 1, 1], LineBox[{87, 88}]}, 
       {Hue[0.061896242940041334`, 1, 1], LineBox[{88, 89}]}, 
       {Hue[0.062192711692253905`, 1, 1], LineBox[{89, 90}]}, 
       {Hue[0.06278564919667907, 1, 1], LineBox[{90, 91}]}, 
       {Hue[0.0639715242055294, 1, 1], LineBox[{91, 92}]}, 
       {Hue[0.06634327422323016, 1, 1], LineBox[{92, 93}]}, 
       {Hue[0.06663974297544276, 1, 1], LineBox[{93, 94}]}, 
       {Hue[0.06693621172765533, 1, 1], LineBox[{94, 95}]}, 
       {Hue[0.06752914923208049, 1, 1], LineBox[{95, 96}]}, 
       {Hue[0.06871502424093089, 1, 1], LineBox[{96, 97}]}, 
       {Hue[0.0710867742586316, 1, 1], LineBox[{97, 98}]}, 
       {Hue[0.07138324301084417, 1, 1], LineBox[{98, 99}]}, 
       {Hue[0.07167971176305678, 1, 1], LineBox[{99, 100}]}, 
       {Hue[0.0722726492674819, 1, 1], LineBox[{100, 101}]}, 
       {Hue[0.0734585242763323, 1, 1], LineBox[{101, 102}]}, 
       {Hue[0.07583027429403302, 1, 1], LineBox[{102, 103}]}, 
       {Hue[0.0761267430462456, 1, 1], LineBox[{103, 104}]}, 
       {Hue[0.07642321179845818, 1, 1], LineBox[{104, 105}]}, 
       {Hue[0.07701614930288335, 1, 1], LineBox[{105, 106}]}, 
       {Hue[0.07820202431173374, 1, 1], LineBox[{106, 107}]}, 
       {Hue[0.08057377432943442, 1, 1], LineBox[{107, 108}]}, 
       {Hue[0.0808644275415848, 1, 1], LineBox[{108, 109}]}, 
       {Hue[0.08115508075373519, 1, 1], LineBox[{109, 110}]}, 
       {Hue[0.08173638717803601, 1, 1], LineBox[{110, 111}]}, 
       {Hue[0.08202704039018638, 1, 1], LineBox[{111, 112}]}, 
       {Hue[0.08231769360233682, 1, 1], LineBox[{112, 113}]}, 
       {Hue[0.08260834681448721, 1, 1], LineBox[{113, 114}]}, 
       {Hue[0.08289900002663757, 1, 1], LineBox[{114, 115}]}, 
       {Hue[0.08318965323878796, 1, 1], LineBox[{115, 116}]}, 
       {Hue[0.5834803064509383, 1, 1], LineBox[{116, 117}]}, 
       {Hue[0.5837709596630887, 1, 1], LineBox[{117, 118}]}, 
       {Hue[0.5840616128752392, 1, 1], LineBox[{118, 119}]}, 
       {Hue[0.5843522660873894, 1, 1], LineBox[{119, 120}]}, 
       {Hue[0.5846429192995399, 1, 1], LineBox[{120, 121}]}, 
       {Hue[0.5852242257238407, 1, 1], LineBox[{121, 122}]}, 
       {Hue[0.585514878935991, 1, 1], LineBox[{122, 123}]}, 
       {Hue[0.5858055321481415, 1, 1], LineBox[{123, 124}]}, 
       {Hue[0.5863868385724422, 1, 1], LineBox[{124, 125}]}, 
       {Hue[0.5875494514210438, 1, 1], LineBox[{125, 126}]}, 
       {Hue[0.5898746771182468, 1, 1], LineBox[{126, 127}]}, 
       {Hue[0.5901653303303972, 1, 1], LineBox[{127, 128}]}, 
       {Hue[0.5904559835425476, 1, 1], LineBox[{128, 129}]}, 
       {Hue[0.5910372899668483, 1, 1], LineBox[{129, 130}]}, 
       {Hue[0.5921999028154499, 1, 1], LineBox[{130, 131}]}, 
       {Hue[0.594525128512653, 1, 1], LineBox[{131, 132}]}, 
       {Hue[0.5991755799070593, 1, 1], LineBox[{132, 133}]}, 
       {Hue[0.5994908681422391, 1, 1], LineBox[{133, 134}]}, 
       {Hue[0.5998061563774189, 1, 1], LineBox[{134, 135}]}, 
       {Hue[0.6004367328477782, 1, 1], LineBox[{135, 136}]}, 
       {Hue[0.6016978857884973, 1, 1], LineBox[{136, 137}]}, 
       {Hue[0.6042201916699352, 1, 1], LineBox[{137, 138}]}, 
       {Hue[0.6092648034328112, 1, 1], LineBox[{138, 139}]}, 
       {Hue[0.6193540269585631, 1, 1], LineBox[{139, 140}]}, 
       {Hue[0.6381858614271563, 1, 1], LineBox[{140, 141}]}, 
       {Hue[0.6385047438637573, 1, 1], LineBox[{141, 142}]}, 
       {Hue[0.6388236263003584, 1, 1], LineBox[{142, 143}]}, 
       {Hue[0.6394613911735607, 1, 1], LineBox[{143, 144}]}, 
       {Hue[0.6407369209199651, 1, 1], LineBox[{144, 145}]}, 
       {Hue[0.6432879804127742, 1, 1], LineBox[{145, 146}]}, 
       {Hue[0.6483900993983922, 1, 1], LineBox[{146, 147}]}, 
       {Hue[0.6487089818349934, 1, 1], LineBox[{147, 148}]}, 
       {Hue[0.6490278642715945, 1, 1], LineBox[{148, 149}]}, 
       {Hue[0.6496656291447966, 1, 1], LineBox[{149, 150}]}, 
       {Hue[0.6509411588912013, 1, 1], LineBox[{150, 151}]}, 
       {Hue[0.6534922183840102, 1, 1], LineBox[{151, 152}]}, 
       {Hue[0.6585943373696282, 1, 1], LineBox[{152, 153}]}, 
       {Hue[0.6589074042661671, 1, 1], LineBox[{153, 154}]}, 
       {Hue[0.6592204711627061, 1, 1], LineBox[{154, 155}]}, 
       {Hue[0.6598466049557838, 1, 1], LineBox[{155, 156}]}, 
       {Hue[0.6610988725419396, 1, 1], LineBox[{156, 157}]}, 
       {Hue[0.663603407714251, 1, 1], LineBox[{157, 158}]}, 
       {Hue[0.66391647461079, 1, 1], LineBox[{158, 159}]}, 
       {Hue[0.6642295415073289, 1, 1], LineBox[{159, 160}]}, 
       {Hue[0.6648556753004067, 1, 1], LineBox[{160, 161}]}, 
       {Hue[0.6661079428865624, 1, 1], LineBox[{161, 162}]}, 
       {Hue[0.6686124780588738, 1, 1], LineBox[{162, 163}]}, 
       {Hue[0.6689255449554127, 1, 1], LineBox[{163, 164}]}, 
       {Hue[0.6692386118519515, 1, 1], LineBox[{164, 165}]}, 
       {Hue[0.6698647456450295, 1, 1], LineBox[{165, 166}]}, 
       {Hue[0.6711170132311852, 1, 1], LineBox[{166, 167}]}, 
       {Hue[0.6714300801277242, 1, 1], LineBox[{167, 168}]}, 
       {Hue[0.6717431470242632, 1, 1], LineBox[{168, 169}]}, 
       {Hue[0.6723692808173409, 1, 1], LineBox[{169, 170}]}, 
       {Hue[0.6736215484034966, 1, 1], LineBox[{170, 171}]}, 
       {Hue[0.6739346153000356, 1, 1], LineBox[{171, 172}]}, 
       {Hue[0.6742476821965744, 1, 1], LineBox[{172, 173}]}, 
       {Hue[0.6748738159896523, 1, 1], LineBox[{173, 174}]}, 
       {Hue[0.6761260835758081, 1, 1], LineBox[{174, 175}]}, 
       {Hue[0.6786306187481195, 1, 1], LineBox[{175, 176}]}, 
       {Hue[0.6789226448230503, 1, 1], LineBox[{176, 177}]}, 
       {Hue[0.6792146708979814, 1, 1], LineBox[{177, 178}]}, 
       {Hue[0.6797987230478432, 1, 1], LineBox[{178, 179}]}, 
       {Hue[0.680966827347567, 1, 1], LineBox[{179, 180}]}, 
       {Hue[0.6812588534224979, 1, 1], LineBox[{180, 181}]}, 
       {Hue[0.6815508794974288, 1, 1], LineBox[{181, 182}]}, 
       {Hue[0.6821349316472907, 1, 1], LineBox[{182, 183}]}, 
       {Hue[0.6833030359470145, 1, 1], LineBox[{183, 184}]}, 
       {Hue[0.6835950620219453, 1, 1], LineBox[{184, 185}]}, 
       {Hue[0.6838870880968764, 1, 1], LineBox[{185, 186}]}, 
       {Hue[0.6844711402467383, 1, 1], LineBox[{186, 187}]}, 
       {Hue[0.6847631663216691, 1, 1], LineBox[{187, 188}]}, 
       {Hue[0.6850551923966001, 1, 1], LineBox[{188, 189}]}, 
       {Hue[0.685639244546462, 1, 1], LineBox[{189, 190}]}, 
       {Hue[0.6859312706213929, 1, 1], LineBox[{190, 191}]}, 
       {Hue[0.6862232966963239, 1, 1], LineBox[{191, 192}]}, 
       {Hue[0.6865153227712548, 1, 1], LineBox[{192, 193}]}, 
       {Hue[0.6868073488461857, 1, 1], LineBox[{193, 194}]}, 
       {Hue[0.6870993749211167, 1, 1], LineBox[{194, 195}]}, 
       {Hue[0.6873914009960476, 1, 1], LineBox[{195, 196}]}, 
       {Hue[0.6876834270709786, 1, 1], LineBox[{196, 197}]}, 
       {Hue[0.6879754531459095, 1, 1], LineBox[{197, 198}]}, 
       {Hue[0.6882674792208405, 1, 1], LineBox[{198, 199}]}, 
       {Hue[0.6885595052957714, 1, 1], LineBox[{199, 200}]}, 
       {Hue[0.6888515313707023, 1, 1], LineBox[{200, 201}]}, 
       {Hue[0.6891435574456333, 1, 1], LineBox[{201, 202}]}, 
       {Hue[0.6894355835205642, 1, 1], LineBox[{202, 203}]}, 
       {Hue[0.6897276095954952, 1, 1], LineBox[{203, 204}]}, 
       {Hue[0.6900196356704261, 1, 1], LineBox[{204, 205}]}, 
       {Hue[0.6903116617453571, 1, 1], LineBox[{205, 206}]}, 
       {Hue[0.6906036878202879, 1, 1], LineBox[{206, 207}]}, 
       {Hue[0.690895713895219, 1, 1], LineBox[{207, 208}]}, 
       {Hue[0.6911877399701498, 1, 1], LineBox[{208, 209}]}, 
       {Hue[0.6914797660450808, 1, 1], LineBox[{209, 210}]}, 
       {Hue[0.6917717921200117, 1, 1], LineBox[{210, 211}]}, 
       {Hue[0.6920638181949427, 1, 1], LineBox[{211, 212}]}, 
       {Hue[0.6923558442698736, 1, 1], LineBox[{212, 213}]}, 
       {Hue[0.6926478703448045, 1, 1], LineBox[{213, 214}]}, 
       {Hue[0.6929398964197355, 1, 1], LineBox[{214, 215}]}, 
       {Hue[0.6932319224946664, 1, 1], LineBox[{215, 216}]}, 
       {Hue[0.6935239485695974, 1, 1], LineBox[{216, 217}]}, 
       {Hue[0.6938159746445283, 1, 1], LineBox[{217, 218}]}, 
       {Hue[0.6941080007194593, 1, 1], LineBox[{218, 219}]}, 
       {Hue[0.6944000267943902, 1, 1], LineBox[{219, 220}]}, 
       {Hue[0.6949840789442521, 1, 1], LineBox[{220, 221}]}, 
       {Hue[0.695276105019183, 1, 1], LineBox[{221, 222}]}, 
       {Hue[0.695568131094114, 1, 1], LineBox[{222, 223}]}, 
       {Hue[0.6961521832439759, 1, 1], LineBox[{223, 224}]}, 
       {Hue[0.6964442093189067, 1, 1], LineBox[{224, 225}]}, 
       {Hue[0.6967362353938378, 1, 1], LineBox[{225, 226}]}, 
       {Hue[0.6973202875436997, 1, 1], LineBox[{226, 227}]}, 
       {Hue[0.6976369486416599, 1, 1], LineBox[{227, 228}]}, 
       {Hue[0.6979536097396202, 1, 1], LineBox[{228, 229}]}, 
       {Hue[0.6985869319355408, 1, 1], LineBox[{229, 230}]}, 
       {Hue[0.698903593033501, 1, 1], LineBox[{230, 231}]}, 
       {Hue[0.6992202541314614, 1, 1], LineBox[{231, 232}]}, 
       {Hue[0.6998535763273821, 1, 1], LineBox[{232, 233}]}, 
       {Hue[0.7001702374253423, 1, 1], LineBox[{233, 234}]}, 
       {Hue[0.7004868985233026, 1, 1], LineBox[{234, 235}]}, 
       {Hue[0.7011202207192233, 1, 1], LineBox[{235, 236}]}, 
       {Hue[0.7023868651110644, 1, 1], LineBox[{236, 237}]}, 
       {Hue[0.7027035262090247, 1, 1], LineBox[{237, 238}]}, 
       {Hue[0.703020187306985, 1, 1], LineBox[{238, 239}]}, 
       {Hue[0.7036535095029056, 1, 1], LineBox[{239, 240}]}, 
       {Hue[0.7049201538947468, 1, 1], LineBox[{240, 241}]}, 
       {Hue[0.7052368149927071, 1, 1], LineBox[{241, 242}]}, 
       {Hue[0.7055534760906674, 1, 1], LineBox[{242, 243}]}, 
       {Hue[0.706186798286588, 1, 1], LineBox[{243, 244}]}, 
       {Hue[0.7074534426784292, 1, 1], LineBox[{244, 245}]}, 
       {Hue[0.7077701037763895, 1, 1], LineBox[{245, 246}]}, 
       {Hue[0.7080867648743498, 1, 1], LineBox[{246, 247}]}, 
       {Hue[0.7087200870702703, 1, 1], LineBox[{247, 248}]}, 
       {Hue[0.7099867314621116, 1, 1], LineBox[{248, 249}]}, 
       {Hue[0.7103033925600719, 1, 1], LineBox[{249, 250}]}, 
       {Hue[0.7106200536580323, 1, 1], LineBox[{250, 251}]}, 
       {Hue[0.7112533758539528, 1, 1], LineBox[{251, 252}]}, 
       {Hue[0.712520020245794, 1, 1], LineBox[{252, 253}]}, 
       {Hue[0.7128366813437543, 1, 1], LineBox[{253, 254}]}, 
       {Hue[0.7131533424417147, 1, 1], LineBox[{254, 255}]}, 
       {Hue[0.7137866646376352, 1, 1], LineBox[{255, 256}]}, 
       {Hue[0.7150533090294764, 1, 1], LineBox[{256, 257}]}, 
       {Hue[0.7175865978131588, 1, 1], LineBox[{257, 258}]}, 
       {Hue[0.7178822180895112, 1, 1], LineBox[{258, 259}]}, 
       {Hue[0.7181778383658635, 1, 1], LineBox[{259, 260}]}, 
       {Hue[0.7187690789185681, 1, 1], LineBox[{260, 261}]}, 
       {Hue[0.7199515600239774, 1, 1], LineBox[{261, 262}]}, 
       {Hue[0.7223165222347959, 1, 1], LineBox[{262, 263}]}, 
       {Hue[0.7226121425111482, 1, 1], LineBox[{263, 264}]}, 
       {Hue[0.7229077627875006, 1, 1], LineBox[{264, 265}]}, 
       {Hue[0.7234990033402051, 1, 1], LineBox[{265, 266}]}, 
       {Hue[0.7246814844456144, 1, 1], LineBox[{266, 267}]}, 
       {Hue[0.727046446656433, 1, 1], LineBox[{267, 268}]}, 
       {Hue[0.7273420669327854, 1, 1], LineBox[{268, 269}]}, 
       {Hue[0.7276376872091377, 1, 1], LineBox[{269, 270}]}, 
       {Hue[0.7282289277618423, 1, 1], LineBox[{270, 271}]}, 
       {Hue[0.7294114088672516, 1, 1], LineBox[{271, 272}]}, 
       {Hue[0.7317763710780701, 1, 1], LineBox[{272, 273}]}, 
       {Hue[0.7365062954997073, 1, 1], LineBox[{273, 274}]}, 
       {Hue[0.7367961002359974, 1, 1], LineBox[{274, 275}]}, 
       {Hue[0.7370859049722875, 1, 1], LineBox[{275, 276}]}, 
       {Hue[0.7376655144448677, 1, 1], LineBox[{276, 277}]}, 
       {Hue[0.7388247333900282, 1, 1], LineBox[{277, 278}]}, 
       {Hue[0.7411431712803491, 1, 1], LineBox[{278, 279}]}, 
       {Hue[0.7414329760166392, 1, 1], LineBox[{279, 280}]}, 
       {Hue[0.7417227807529294, 1, 1], LineBox[{280, 281}]}, 
       {Hue[0.7423023902255096, 1, 1], LineBox[{281, 282}]}, 
       {Hue[0.7434616091706701, 1, 1], LineBox[{282, 283}]}, 
       {Hue[0.745780047060991, 1, 1], LineBox[{283, 284}]}, 
       {Hue[0.746069851797281, 1, 1], LineBox[{284, 285}]}, 
       {Hue[0.7463596565335713, 1, 1], LineBox[{285, 286}]}, 
       {Hue[0.7469392660061515, 1, 1], LineBox[{286, 287}]}, 
       {Hue[0.7472290707424416, 1, 1], LineBox[{287, 288}]}, 
       {Hue[0.7475188754787316, 1, 1], LineBox[{288, 289}]}, 
       {Hue[0.7480984849513119, 1, 1], LineBox[{289, 290}]}, 
       {Hue[0.748388289687602, 1, 1], LineBox[{290, 291}]}, 
       {Hue[0.7486780944238922, 1, 1], LineBox[{291, 292}]}, 
       {Hue[0.7489678991601822, 1, 1], LineBox[{292, 293}]}, 
       {Hue[0.7492577038964724, 1, 1], LineBox[{293, 294}]}, 
       {Hue[0.7495475086327625, 1, 1], LineBox[{294, 295}]}, 
       {Hue[0.7498373133690526, 1, 1], LineBox[{295, 296}]}, 
       {Hue[0.25012711810534277`, 1, 1], LineBox[{296, 297}]}, 
       {Hue[0.25041692284163286`, 1, 1], LineBox[{297, 298}]}, 
       {Hue[0.250706727577923, 1, 1], LineBox[{298, 299}]}, 
       {Hue[0.25099653231421315`, 1, 1], LineBox[{299, 300}]}, 
       {Hue[0.25157614178679333`, 1, 1], LineBox[{300, 301}]}, 
       {Hue[0.2527353607319538, 1, 1], LineBox[{301, 302}]}, 
       {Hue[0.2550537986222748, 1, 1], LineBox[{302, 303}]}, 
       {Hue[0.2553682383815943, 1, 1], LineBox[{303, 304}]}, 
       {Hue[0.2556826781409137, 1, 1], LineBox[{304, 305}]}, 
       {Hue[0.25631155765955266`, 1, 1], LineBox[{305, 306}]}, 
       {Hue[0.2575693166968306, 1, 1], LineBox[{306, 307}]}, 
       {Hue[0.2600848347713864, 1, 1], LineBox[{307, 308}]}, 
       {Hue[0.26511587092049804`, 1, 1], LineBox[{308, 309}]}, 
       {Hue[0.27517794321872124`, 1, 1], LineBox[{309, 310}]}, 
       {Hue[0.2939554752322569, 1, 1], LineBox[{310, 311}]}, 
       {Hue[0.2942735091929977, 1, 1], LineBox[{311, 312}]}, 
       {Hue[0.2945915431537386, 1, 1], LineBox[{312, 313}]}, 
       {Hue[0.2952276110752203, 1, 1], LineBox[{313, 314}]}, 
       {Hue[0.29649974691818365`, 1, 1], LineBox[{314, 315}]}, 
       {Hue[0.29904401860411056`, 1, 1], LineBox[{315, 316}]}, 
       {Hue[0.3041325619759642, 1, 1], LineBox[{316, 317}]}, 
       {Hue[0.30445059593670504`, 1, 1], LineBox[{317, 318}]}, 
       {Hue[0.30476862989744585`, 1, 1], LineBox[{318, 319}]}, 
       {Hue[0.3054046978189276, 1, 1], LineBox[{319, 320}]}, 
       {Hue[0.306676833661891, 1, 1], LineBox[{320, 321}]}, 
       {Hue[0.3092211053478179, 1, 1], LineBox[{321, 322}]}, 
       {Hue[0.31430964871967154`, 1, 1], LineBox[{322, 323}]}, 
       {Hue[0.3146218671403502, 1, 1], LineBox[{323, 324}]}, 
       {Hue[0.3149340855610288, 1, 1], LineBox[{324, 325}]}, 
       {Hue[0.31555852240238613`, 1, 1], LineBox[{325, 326}]}, 
       {Hue[0.3168073960851008, 1, 1], LineBox[{326, 327}]}, 
       {Hue[0.31930514345052996`, 1, 1], LineBox[{327, 328}]}, 
       {Hue[0.3196173618712086, 1, 1], LineBox[{328, 329}]}, 
       {Hue[0.3199295802918873, 1, 1], LineBox[{329, 330}]}, 
       {Hue[0.32055401713324455`, 1, 1], LineBox[{330, 331}]}, 
       {Hue[0.3218028908159592, 1, 1], LineBox[{331, 332}]}, 
       {Hue[0.3243006381813884, 1, 1], LineBox[{332, 333}]}, 
       {Hue[0.324612856602067, 1, 1], LineBox[{333, 334}]}, 
       {Hue[0.32492507502274565`, 1, 1], LineBox[{334, 335}]}, 
       {Hue[0.32554951186410297`, 1, 1], LineBox[{335, 336}]}, 
       {Hue[0.32679838554681756`, 1, 1], LineBox[{336, 337}]}, 
       {Hue[0.3271106039674962, 1, 1], LineBox[{337, 338}]}, 
       {Hue[0.3274228223881749, 1, 1], LineBox[{338, 339}]}, 
       {Hue[0.32804725922953215`, 1, 1], LineBox[{339, 340}]}, 
       {Hue[0.3292961329122468, 1, 1], LineBox[{340, 341}]}, 
       {Hue[0.32960835133292543`, 1, 1], LineBox[{341, 342}]}, 
       {Hue[0.3299205697536041, 1, 1], LineBox[{342, 343}]}, 
       {Hue[0.3305450065949614, 1, 1], LineBox[{343, 344}]}, 
       {Hue[0.331793880277676, 1, 1], LineBox[{344, 345}]}, 
       {Hue[0.3342916276431052, 1, 1], LineBox[{345, 346}]}, 
       {Hue[0.33458280524217593`, 1, 1], LineBox[{346, 347}]}, 
       {Hue[0.3348739828412466, 1, 1], LineBox[{347, 348}]}, 
       {Hue[0.3354563380393879, 1, 1], LineBox[{348, 349}]}, 
       {Hue[0.3366210484356706, 1, 1], LineBox[{349, 350}]}, 
       {Hue[0.3369122260347413, 1, 1], LineBox[{350, 351}]}, 
       {Hue[0.33720340363381196`, 1, 1], LineBox[{351, 352}]}, 
       {Hue[0.3377857588319533, 1, 1], LineBox[{352, 353}]}, 
       {Hue[0.33895046922823596`, 1, 1], LineBox[{353, 354}]}, 
       {Hue[0.3392416468273066, 1, 1], LineBox[{354, 355}]}, 
       {Hue[0.33953282442637733`, 1, 1], LineBox[{355, 356}]}, 
       {Hue[0.34011517962451865`, 1, 1], LineBox[{356, 357}]}, 
       {Hue[0.3404063572235893, 1, 1], LineBox[{357, 358}]}, 
       {Hue[0.34069753482266, 1, 1], LineBox[{358, 359}]}, 
       {Hue[0.34127989002080134`, 1, 1], LineBox[{359, 360}]}, 
       {Hue[0.341571067619872, 1, 1], LineBox[{360, 361}]}, 
       {Hue[0.34186224521894265`, 1, 1], LineBox[{361, 362}]}, 
       {Hue[0.34215342281801336`, 1, 1], LineBox[{362, 363}]}, 
       {Hue[0.34244460041708397`, 1, 1], LineBox[{363, 364}]}, 
       {Hue[0.3427357780161547, 1, 1], LineBox[{364, 365}]}, 
       {Hue[0.34302695561522534`, 1, 1], LineBox[{365, 366}]}, 
       {Hue[0.34360931081336665`, 1, 1], LineBox[{366, 367}]}, 
       {Hue[0.34390048841243737`, 1, 1], LineBox[{367, 368}]}, 
       {Hue[0.344191666011508, 1, 1], LineBox[{368, 369}]}, 
       {Hue[0.34448284361057874`, 1, 1], LineBox[{369, 370}]}, 
       {Hue[0.34477402120964934`, 1, 1], LineBox[{370, 371}]}, 
       {Hue[0.34506519880872005`, 1, 1], LineBox[{371, 372}]}, 
       {Hue[0.3453563764077907, 1, 1], LineBox[{372, 373}]}, 
       {Hue[0.3456475540068614, 1, 1], LineBox[{373, 374}]}, 
       {Hue[0.345938731605932, 1, 1], LineBox[{374, 375}]}, 
       {Hue[0.3462299092050027, 1, 1], LineBox[{375, 376}]}, 
       {Hue[0.3465210868040734, 1, 1], LineBox[{376, 377}]}, 
       {Hue[0.346812264403144, 1, 1], LineBox[{377, 378}]}, 
       {Hue[0.3471034420022147, 1, 1], LineBox[{378, 379}]}, 
       {Hue[0.34739461960128537`, 1, 1], LineBox[{379, 380}]}, 
       {Hue[0.3476857972003561, 1, 1], LineBox[{380, 381}]}, 
       {Hue[0.34797697479942674`, 1, 1], LineBox[{381, 382}]}, 
       {Hue[0.3482681523984974, 1, 1], LineBox[{382, 383}]}, 
       {Hue[0.34855932999756806`, 1, 1], LineBox[{383, 384}]}, 
       {Hue[0.34885050759663877`, 1, 1], LineBox[{384, 385}]}, 
       {Hue[0.34914168519570943`, 1, 1], LineBox[{385, 386}]}, 
       {Hue[0.3494328627947801, 1, 1], LineBox[{386, 387}]}, 
       {Hue[0.3497240403938508, 1, 1], LineBox[{387, 388}]}, 
       {Hue[0.35001521799292146`, 1, 1], LineBox[{388, 389}]}, 
       {Hue[0.3503063955919921, 1, 1], LineBox[{389, 390}]}, 
       {Hue[0.3505975731910628, 1, 1], LineBox[{390, 391}]}, 
       {Hue[0.35088875079013343`, 1, 1], LineBox[{391, 392}]}, 
       {Hue[0.3511799283892041, 1, 1], LineBox[{392, 393}]}, 
       {Hue[0.3514711059882748, 1, 1], LineBox[{393, 394}]}, 
       {Hue[0.3517622835873454, 1, 1], LineBox[{394, 395}]}, 
       {Hue[0.3520534611864161, 1, 1], LineBox[{395, 396}]}, 
       {Hue[0.3523446387854868, 1, 1], LineBox[{396, 397}]}, 
       {Hue[0.3529269939836281, 1, 1], LineBox[{397, 398}]}, 
       {Hue[0.3532428066057282, 1, 1], LineBox[{398, 399}]}, 
       {Hue[0.3535586192278282, 1, 1], LineBox[{399, 400}]}, 
       {Hue[0.35419024447202824`, 1, 1], LineBox[{400, 401}]}, 
       {Hue[0.3545060570941283, 1, 1], LineBox[{401, 402}]}, 
       {Hue[0.3548218697162283, 1, 1], LineBox[{402, 403}]}, 
       {Hue[0.35545349496042833`, 1, 1], LineBox[{403, 404}]}, 
       {Hue[0.35576930758252845`, 1, 1], LineBox[{404, 405}]}, 
       {Hue[0.35608512020462846`, 1, 1], LineBox[{405, 406}]}, 
       {Hue[0.3567167454488285, 1, 1], LineBox[{406, 407}]}, 
       {Hue[0.35797999593722857`, 1, 1], LineBox[{407, 408}]}, 
       {Hue[0.3582958085593287, 1, 1], LineBox[{408, 409}]}, 
       {Hue[0.3586116211814287, 1, 1], LineBox[{409, 410}]}, 
       {Hue[0.3592432464256287, 1, 1], LineBox[{410, 411}]}, 
       {Hue[0.3605064969140288, 1, 1], LineBox[{411, 412}]}, 
       {Hue[0.3608223095361289, 1, 1], LineBox[{412, 413}]}, 
       {Hue[0.36113812215822894`, 1, 1], LineBox[{413, 414}]}, 
       {Hue[0.36176974740242895`, 1, 1], LineBox[{414, 415}]}, 
       {Hue[0.36303299789082905`, 1, 1], LineBox[{415, 416}]}, 
       {Hue[0.3633488105129291, 1, 1], LineBox[{416, 417}]}, 
       {Hue[0.3636646231350291, 1, 1], LineBox[{417, 418}]}, 
       {Hue[0.3642962483792292, 1, 1], LineBox[{418, 419}]}, 
       {Hue[0.36555949886762934`, 1, 1], LineBox[{419, 420}]}, 
       {Hue[0.36587531148972935`, 1, 1], LineBox[{420, 421}]}, 
       {Hue[0.36619112411182936`, 1, 1], LineBox[{421, 422}]}, 
       {Hue[0.36682274935602943`, 1, 1], LineBox[{422, 423}]}, 
       {Hue[0.3680859998444296, 1, 1], LineBox[{423, 424}]}, 
       {Hue[0.3684018124665296, 1, 1], LineBox[{424, 425}]}, 
       {Hue[0.3687176250886296, 1, 1], LineBox[{425, 426}]}, 
       {Hue[0.36934925033282967`, 1, 1], LineBox[{426, 427}]}, 
       {Hue[0.3706125008212298, 1, 1], LineBox[{427, 428}]}, 
       {Hue[0.37313900179803006`, 1, 1], LineBox[{428, 429}]}, 
       {Hue[0.37343377359852203`, 1, 1], LineBox[{429, 430}]}, 
       {Hue[0.3737285453990141, 1, 1], LineBox[{430, 431}]}, 
       {Hue[0.37431808899999824`, 1, 1], LineBox[{431, 432}]}, 
       {Hue[0.37549717620196643`, 1, 1], LineBox[{432, 433}]}, 
       {Hue[0.37785535060590275`, 1, 1], LineBox[{433, 434}]}, 
       {Hue[0.37815012240639484`, 1, 1], LineBox[{434, 435}]}, 
       {Hue[0.37844489420688693`, 1, 1], LineBox[{435, 436}]}, 
       {Hue[0.37903443780787094`, 1, 1], LineBox[{436, 437}]}, 
       {Hue[0.3802135250098391, 1, 1], LineBox[{437, 438}]}, 
       {Hue[0.38257169941377556`, 1, 1], LineBox[{438, 439}]}, 
       {Hue[0.38286647121426753`, 1, 1], LineBox[{439, 440}]}, 
       {Hue[0.3831612430147596, 1, 1], LineBox[{440, 441}]}, 
       {Hue[0.38375078661574374`, 1, 1], LineBox[{441, 442}]}, 
       {Hue[0.38492987381771193`, 1, 1], LineBox[{442, 443}]}, 
       {Hue[0.38728804822164825`, 1, 1], LineBox[{443, 444}]}, 
       {Hue[0.39200439702952106`, 1, 1], LineBox[{444, 445}]}, 
       {Hue[0.39232380385304244`, 1, 1], LineBox[{445, 446}]}, 
       {Hue[0.3926432106765638, 1, 1], LineBox[{446, 447}]}, 
       {Hue[0.3932820243236067, 1, 1], LineBox[{447, 448}]}, 
       {Hue[0.3945596516176923, 1, 1], LineBox[{448, 449}]}, 
       {Hue[0.39711490620586354`, 1, 1], LineBox[{449, 450}]}, 
       {Hue[0.4022254153822061, 1, 1], LineBox[{450, 451}]}, 
       {Hue[0.40254482220572746`, 1, 1], LineBox[{451, 452}]}, 
       {Hue[0.40286422902924884`, 1, 1], LineBox[{452, 453}]}, 
       {Hue[0.4035030426762917, 1, 1], LineBox[{453, 454}]}, 
       {Hue[0.4047806699703773, 1, 1], LineBox[{454, 455}]}, 
       {Hue[0.4073359245585486, 1, 1], LineBox[{455, 456}]}, 
       {Hue[0.40765533138207, 1, 1], LineBox[{456, 457}]}, 
       {Hue[0.4079747382055914, 1, 1], LineBox[{457, 458}]}, 
       {Hue[0.4086135518526342, 1, 1], LineBox[{458, 459}]}, 
       {Hue[0.40989117914671985`, 1, 1], LineBox[{459, 460}]}, 
       {Hue[0.4124464337348911, 1, 1], LineBox[{460, 461}]}, 
       {Hue[0.4127600250183503, 1, 1], LineBox[{461, 462}]}, 
       {Hue[0.4130736163018095, 1, 1], LineBox[{462, 463}]}, 
       {Hue[0.4137007988687279, 1, 1], LineBox[{463, 464}]}, 
       {Hue[0.4140143901521871, 1, 1], LineBox[{464, 465}]}, 
       {Hue[0.41432798143564636`, 1, 1], LineBox[{465, 466}]}, 
       {Hue[0.4149551640025647, 1, 1], LineBox[{466, 467}]}, 
       {Hue[0.41526875528602397`, 1, 1], LineBox[{467, 468}]}, 
       {Hue[0.4155823465694832, 1, 1], LineBox[{468, 469}]}, 
       {Hue[0.4158959378529423, 1, 1], LineBox[{469, 470}]}, 
       {Hue[0.4162095291364016, 1, 1], LineBox[{470, 471}]}, 
       {Hue[0.4165231204198608, 1, 1], LineBox[{471, 472}]}, 
       {Hue[0.9168367117033199, 1, 1], LineBox[{472, 473}]}, 
       {Hue[0.9171503029867791, 1, 1], LineBox[{473, 474}]}, 
       {Hue[0.9174638942702383, 1, 1], LineBox[{474, 475}]}, 
       {Hue[0.9177774855536975, 1, 1], LineBox[{475, 476}]}, 
       {Hue[0.9180910768371567, 1, 1], LineBox[{476, 477}]}, 
       {Hue[0.9187182594040751, 1, 1], LineBox[{477, 478}]}, 
       {Hue[0.919972624537912, 1, 1], LineBox[{478, 479}]}, 
       {Hue[0.9224813548055857, 1, 1], LineBox[{479, 480}]}, 
       {Hue[0.9227949460890449, 1, 1], LineBox[{480, 481}]}, 
       {Hue[0.9231085373725041, 1, 1], LineBox[{481, 482}]}, 
       {Hue[0.9237357199394225, 1, 1], LineBox[{482, 483}]}, 
       {Hue[0.9249900850732593, 1, 1], LineBox[{483, 484}]}, 
       {Hue[0.9274988153409329, 1, 1], LineBox[{484, 485}]}, 
       {Hue[0.9325162758762802, 1, 1], LineBox[{485, 486}]}, 
       {Hue[0.9328088263381314, 1, 1], LineBox[{486, 487}]}, 
       {Hue[0.9331013767999826, 1, 1], LineBox[{487, 488}]}, 
       {Hue[0.9336864777236851, 1, 1], LineBox[{488, 489}]}, 
       {Hue[0.93485667957109, 1, 1], LineBox[{489, 490}]}, 
       {Hue[0.9371970832658998, 1, 1], LineBox[{490, 491}]}, 
       {Hue[0.9418778906555193, 1, 1], LineBox[{491, 492}]}, 
       {Hue[0.9421704411173706, 1, 1], LineBox[{492, 493}]}, 
       {Hue[0.9424629915792218, 1, 1], LineBox[{493, 494}]}, 
       {Hue[0.9430480925029242, 1, 1], LineBox[{494, 495}]}, 
       {Hue[0.9442182943503291, 1, 1], LineBox[{495, 496}]}, 
       {Hue[0.9465586980451389, 1, 1], LineBox[{496, 497}]}, 
       {Hue[0.9512395054347584, 1, 1], LineBox[{497, 498}]}, 
       {Hue[0.951556690919639, 1, 1], LineBox[{498, 499}]}, 
       {Hue[0.9518738764045196, 1, 1], LineBox[{499, 500}]}, 
       {Hue[0.9525082473742809, 1, 1], LineBox[{500, 501}]}, 
       {Hue[0.9537769893138032, 1, 1], LineBox[{501, 502}]}, 
       {Hue[0.9563144731928478, 1, 1], LineBox[{502, 503}]}, 
       {Hue[0.9613894409509371, 1, 1], LineBox[{503, 504}]}, 
       {Hue[0.9617066264358177, 1, 1], LineBox[{504, 505}]}, 
       {Hue[0.9620238119206983, 1, 1], LineBox[{505, 506}]}, 
       {Hue[0.9626581828904595, 1, 1], LineBox[{506, 507}]}, 
       {Hue[0.9639269248299818, 1, 1], LineBox[{507, 508}]}, 
       {Hue[0.9664644087090264, 1, 1], LineBox[{508, 509}]}, 
       {Hue[0.9715393764671157, 1, 1], LineBox[{509, 510}]}, 
       {Hue[0.9718355211303884, 1, 1], LineBox[{510, 511}]}, 
       {Hue[0.972131665793661, 1, 1], LineBox[{511, 512}]}, 
       {Hue[0.9727239551202062, 1, 1], LineBox[{512, 513}]}, 
       {Hue[0.9739085337732966, 1, 1], LineBox[{513, 514}]}, 
       {Hue[0.9762776910794774, 1, 1], LineBox[{514, 515}]}, 
       {Hue[0.9765738357427499, 1, 1], LineBox[{515, 516}]}, 
       {Hue[0.9768699804060226, 1, 1], LineBox[{516, 517}]}, 
       {Hue[0.9774622697325678, 1, 1], LineBox[{517, 518}]}, 
       {Hue[0.9786468483856582, 1, 1], LineBox[{518, 519}]}, 
       {Hue[0.981016005691839, 1, 1], LineBox[{519, 520}]}, 
       {Hue[0.9813121503551115, 1, 1], LineBox[{520, 521}]}, 
       {Hue[0.9816082950183842, 1, 1], LineBox[{521, 522}]}, 
       {Hue[0.9822005843449294, 1, 1], LineBox[{522, 523}]}, 
       {Hue[0.9833851629980198, 1, 1], LineBox[{523, 524}]}, 
       {Hue[0.9836813076612924, 1, 1], LineBox[{524, 525}]}, 
       {Hue[0.983977452324565, 1, 1], LineBox[{525, 526}]}, 
       {Hue[0.9845697416511101, 1, 1], LineBox[{526, 527}]}, 
       {Hue[0.9857543203042005, 1, 1], LineBox[{527, 528}]}, 
       {Hue[0.9860504649674732, 1, 1], LineBox[{528, 529}]}, 
       {Hue[0.9863466096307457, 1, 1], LineBox[{529, 530}]}, 
       {Hue[0.9869388989572909, 1, 1], LineBox[{530, 531}]}, 
       {Hue[0.9881234776103813, 1, 1], LineBox[{531, 532}]}, 
       {Hue[0.988419622273654, 1, 1], LineBox[{532, 533}]}, 
       {Hue[0.9887157669369265, 1, 1], LineBox[{533, 534}]}, 
       {Hue[0.9893080562634717, 1, 1], LineBox[{534, 535}]}, 
       {Hue[0.9904926349165621, 1, 1], LineBox[{535, 536}]}, 
       {Hue[0.9907829640397726, 1, 1], LineBox[{536, 537}]}, 
       {Hue[0.991073293162983, 1, 1], LineBox[{537, 538}]}, 
       {Hue[0.9916539514094037, 1, 1], LineBox[{538, 539}]}, 
       {Hue[0.9919442805326142, 1, 1], LineBox[{539, 540}]}, 
       {Hue[0.9922346096558246, 1, 1], LineBox[{540, 541}]}, 
       {Hue[0.9928152679022454, 1, 1], LineBox[{541, 542}]}, 
       {Hue[0.9931055970254558, 1, 1], LineBox[{542, 543}]}, 
       {Hue[0.9933959261486661, 1, 1], LineBox[{543, 544}]}, 
       {Hue[0.993976584395087, 1, 1], LineBox[{544, 545}]}, 
       {Hue[0.9951379008879285, 1, 1], LineBox[{545, 546}]}, 
       {Hue[0.9954282300111389, 1, 1], LineBox[{546, 547}]}, 
       {Hue[0.9957185591343494, 1, 1], LineBox[{547, 548}]}, 
       {Hue[0.9962992173807701, 1, 1], LineBox[{548, 549}]}, 
       {Hue[0.9965895465039806, 1, 1], LineBox[{549, 550}]}, 
       {Hue[0.9968798756271909, 1, 1], LineBox[{550, 551}]}, 
       {Hue[0.9974605338736118, 1, 1], LineBox[{551, 552}]}, 
       {Hue[0.9977508629968221, 1, 1], LineBox[{552, 553}]}, 
       {Hue[0.9980411921200325, 1, 1], LineBox[{553, 554}]}, 
       {Hue[0.9983315212432429, 1, 1], LineBox[{554, 555}]}, 
       {Hue[0.9986218503664533, 1, 1], LineBox[{555, 556}]}, 
       {Hue[0.9989121794896637, 1, 1], LineBox[{556, 557}]}, 
       {Hue[0.9992025086128741, 1, 1], LineBox[{557, 558}]}, 
       {Hue[0.9994928377360845, 1, 1], LineBox[{558, 559}]}, 
       {Hue[0.9997831668592949, 1, 1], LineBox[{559, 560}]}, 
       {Hue[0.00007349598250531426, 1, 1], LineBox[{560, 561}]}, 
       {Hue[0.00036382510571571257`, 1, 1], LineBox[{561, 562}]}, 
       {Hue[0.000654154228926111, 1, 1], LineBox[{562, 563}]}, 
       {Hue[0.0009444833521365094, 1, 1], LineBox[{563, 564}]}, 
       {Hue[0.0012348124753469077`, 1, 1], LineBox[{564, 565}]}, 
       {Hue[0.0015251415985573066`, 1, 1], LineBox[{565, 566}]}, 
       {Hue[0.0018154707217677046`, 1, 1], LineBox[{566, 567}]}, 
       {Hue[0.002105799844978103, 1, 1], LineBox[{567, 568}]}, 
       {Hue[0.0023961289681885016`, 1, 1], LineBox[{568, 569}]}, 
       {Hue[0.0026864580913989, 1, 1], LineBox[{569, 570}]}, 
       {Hue[0.0029767872146092976`, 1, 1], LineBox[{570, 571}]}, 
       {Hue[0.0032671163378196967`, 1, 1], LineBox[{571, 572}]}, 
       {Hue[0.003557445461030095, 1, 1], LineBox[{572, 573}]}, 
       {Hue[0.003847774584240493, 1, 1], LineBox[{573, 574}]}, 
       {Hue[0.004138103707450892, 1, 1], LineBox[{574, 575}]}, 
       {Hue[0.004428432830661289, 1, 1], LineBox[{575, 576}]}, 
       {Hue[0.0047187619538716875`, 1, 1], LineBox[{576, 577}]}, 
       {Hue[0.005009091077082086, 1, 1], LineBox[{577, 578}]}, 
       {Hue[0.005299420200292486, 1, 1], LineBox[{578, 579}]}, 
       {Hue[0.005589749323502882, 1, 1], LineBox[{579, 580}]}, 
       {Hue[0.005880078446713281, 1, 1], LineBox[{580, 581}]}, 
       {Hue[0.0061704075699236804`, 1, 1], LineBox[{581, 582}]}, 
       {Hue[0.006751065816344477, 1, 1], LineBox[{582, 583}]}, 
       {Hue[0.007041394939554876, 1, 1], LineBox[{583, 584}]}, 
       {Hue[0.007331724062765272, 1, 1], LineBox[{584, 585}]}, 
       {Hue[0.00791238230918607, 1, 1], LineBox[{585, 586}]}, 
       {Hue[0.008202711432396469, 1, 1], LineBox[{586, 587}]}, 
       {Hue[0.008493040555606868, 1, 1], LineBox[{587, 588}]}, 
       {Hue[0.009073698802027665, 1, 1], LineBox[{588, 589}]}, 
       {Hue[0.009388662948267424, 1, 1], LineBox[{589, 590}]}, 
       {Hue[0.00970362709450718, 1, 1], LineBox[{590, 591}]}, 
       {Hue[0.0103335553869867, 1, 1], LineBox[{591, 592}]}, 
       {Hue[0.010648519533226455`, 1, 1], LineBox[{592, 593}]}, 
       {Hue[0.010963483679466215`, 1, 1], LineBox[{593, 594}]}, 
       {Hue[0.011593411971945728`, 1, 1], LineBox[{594, 595}]}, 
       {Hue[0.011908376118185491`, 1, 1], LineBox[{595, 596}]}, 
       {Hue[0.012223340264425247`, 1, 1], LineBox[{596, 597}]}, 
       {Hue[0.012853268556904763`, 1, 1], LineBox[{597, 598}]}, 
       {Hue[0.014113125141863794`, 1, 1], LineBox[{598, 599}]}, 
       {Hue[0.014428089288103553`, 1, 1], LineBox[{599, 600}]}, 
       {Hue[0.014743053434343314`, 1, 1], LineBox[{600, 601}]}, 
       {Hue[0.01537298172682283, 1, 1], LineBox[{601, 602}]}, 
       {Hue[0.01663283831178186, 1, 1], LineBox[{602, 603}]}, 
       {Hue[0.016947802458021622`, 1, 1], LineBox[{603, 604}]}, 
       {Hue[0.017262766604261383`, 1, 1], LineBox[{604, 605}]}, 
       {Hue[0.017892694896740897`, 1, 1], LineBox[{605, 606}]}, 
       {Hue[0.019152551481699927`, 1, 1], LineBox[{606, 607}]}, 
       {Hue[0.019467515627939688`, 1, 1], LineBox[{607, 608}]}, 
       {Hue[0.019782479774179445`, 1, 1], LineBox[{608, 609}]}, 
       {Hue[0.020412408066658963`, 1, 1], LineBox[{609, 610}]}, 
       {Hue[0.02167226465161799, 1, 1], LineBox[{610, 611}]}, 
       {Hue[0.02419197782153607, 1, 1], LineBox[{611, 612}]}, 
       {Hue[0.02450694196777583, 1, 1], LineBox[{612, 613}]}, 
       {Hue[0.02482190611401558, 1, 1], LineBox[{613, 614}]}, 
       {Hue[0.0254518344064951, 1, 1], LineBox[{614, 615}]}, 
       {Hue[0.026711690991454134`, 1, 1], LineBox[{615, 616}]}, 
       {Hue[0.0292314041613722, 1, 1], LineBox[{616, 617}]}, 
       {Hue[0.029525327486003974`, 1, 1], LineBox[{617, 618}]}, 
       {Hue[0.02981925081063575, 1, 1], LineBox[{618, 619}]}, 
       {Hue[0.030407097459899297`, 1, 1], LineBox[{619, 620}]}, 
       {Hue[0.03158279075842641, 1, 1], LineBox[{620, 621}]}, 
       {Hue[0.033934177355480606`, 1, 1], LineBox[{621, 622}]}, 
       {Hue[0.03863695054958902, 1, 1], LineBox[{622, 623}]}, 
       {Hue[0.0389308738742208, 1, 1], LineBox[{623, 624}]}, 
       {Hue[0.03922479719885257, 1, 1], LineBox[{624, 625}]}, 
       {Hue[0.03981264384811612, 1, 1], LineBox[{625, 626}]}, 
       {Hue[0.04098833714664322, 1, 1], LineBox[{626, 627}]}, 
       {Hue[0.043339723743697434`, 1, 1], LineBox[{627, 628}]}, 
       {Hue[0.048042496937805844`, 1, 1], LineBox[{628, 629}]}, 
       {Hue[0.048361055285466976`, 1, 1], LineBox[{629, 630}]}, 
       {Hue[0.048679613633128116`, 1, 1], LineBox[{630, 631}]}, 
       {Hue[0.049316730328450395`, 1, 1], LineBox[{631, 632}]}, 
       {Hue[0.05059096371909494, 1, 1], LineBox[{632, 633}]}, 
       {Hue[0.05313943050038401, 1, 1], LineBox[{633, 634}]}, 
       {Hue[0.058236364062962176`, 1, 1], LineBox[{634, 635}]}, 
       {Hue[0.05855492241062331, 1, 1], LineBox[{635, 636}]}, 
       {Hue[0.05887348075828444, 1, 1], LineBox[{636, 637}]}, 
       {Hue[0.05951059745360673, 1, 1], LineBox[{637, 638}]}, 
       {Hue[0.06078483084425125, 1, 1], LineBox[{638, 639}]}, 
       {Hue[0.06333329762554035, 1, 1], LineBox[{639, 640}]}, 
       {Hue[0.06843023118811851, 1, 1], LineBox[{640, 641}]}, 
       {Hue[0.06874297399571745, 1, 1], LineBox[{641, 642}]}, 
       {Hue[0.06905571680331637, 1, 1], LineBox[{642, 643}]}, 
       {Hue[0.06968120241851426, 1, 1], LineBox[{643, 644}]}, 
       {Hue[0.07093217364891, 1, 1], LineBox[{644, 645}]}, 
       {Hue[0.07343411610970148, 1, 1], LineBox[{645, 646}]}, 
       {Hue[0.0737468589173004, 1, 1], LineBox[{646, 647}]}, 
       {Hue[0.07405960172489932, 1, 1], LineBox[{647, 648}]}, 
       {Hue[0.07468508734009718, 1, 1], LineBox[{648, 649}]}, 
       {Hue[0.07593605857049295, 1, 1], LineBox[{649, 650}]}, 
       {Hue[0.0784380010312844, 1, 1], LineBox[{650, 651}]}, 
       {Hue[0.07875074383888335, 1, 1], LineBox[{651, 652}]}, 
       {Hue[0.07906348664648226, 1, 1], LineBox[{652, 653}]}, 
       {Hue[0.07968897226168013, 1, 1], LineBox[{653, 654}]}, 
       {Hue[0.08093994349207585, 1, 1], LineBox[{654, 655}]}, 
       {Hue[0.08125268629967479, 1, 1], LineBox[{655, 656}]}, 
       {Hue[0.08156542910727374, 1, 1], LineBox[{656, 657}]}, 
       {Hue[0.08187817191487268, 1, 1], LineBox[{657, 658}]}, 
       {Hue[0.08219091472247161, 1, 1], LineBox[{658, 659}]}, 
       {Hue[0.08250365753007054, 1, 1], LineBox[{659, 660}]}, 
       {Hue[0.0828164003376695, 1, 1], LineBox[{660, 661}]}, 
       {Hue[0.08312914314526841, 1, 1], LineBox[{661, 662}]}, 
       {Hue[0.5834418859528673, 1, 1], LineBox[{662, 663}]}, 
       {Hue[0.5837546287604662, 1, 1], LineBox[{663, 664}]}, 
       {Hue[0.5840673715680651, 1, 1], LineBox[{664, 665}]}, 
       {Hue[0.584692857183263, 1, 1], LineBox[{665, 666}]}, 
       {Hue[0.5859438284136588, 1, 1], LineBox[{666, 667}]}, 
       {Hue[0.5884457708744503, 1, 1], LineBox[{667, 668}]}, 
       {Hue[0.5887374728604412, 1, 1], LineBox[{668, 669}]}, 
       {Hue[0.5890291748464322, 1, 1], LineBox[{669, 670}]}, 
       {Hue[0.5896125788184141, 1, 1], LineBox[{670, 671}]}, 
       {Hue[0.5907793867623778, 1, 1], LineBox[{671, 672}]}, 
       {Hue[0.5931130026503055, 1, 1], LineBox[{672, 673}]}, 
       {Hue[0.5977802344261607, 1, 1], LineBox[{673, 674}]}, 
       {Hue[0.5980719364121516, 1, 1], LineBox[{674, 675}]}, 
       {Hue[0.5983636383981427, 1, 1], LineBox[{675, 676}]}, 
       {Hue[0.5989470423701245, 1, 1], LineBox[{676, 677}]}, 
       {Hue[0.6001138503140884, 1, 1], LineBox[{677, 678}]}, 
       {Hue[0.602447466202016, 1, 1], LineBox[{678, 679}]}, 
       {Hue[0.6071146979778712, 1, 1], LineBox[{679, 680}]}, 
       {Hue[0.6074310349868914, 1, 1], LineBox[{680, 681}]}, 
       {Hue[0.6077473719959118, 1, 1], LineBox[{681, 682}]}, 
       {Hue[0.6083800460139523, 1, 1], LineBox[{682, 683}]}, 
       {Hue[0.6096453940500337, 1, 1], LineBox[{683, 684}]}, 
       {Hue[0.6121760901221962, 1, 1], LineBox[{684, 685}]}, 
       {Hue[0.6172374822665212, 1, 1], LineBox[{685, 686}]}, 
       {Hue[0.6175538192755414, 1, 1], LineBox[{686, 687}]}, 
       {Hue[0.6178701562845618, 1, 1], LineBox[{687, 688}]}, 
       {Hue[0.6185028303026023, 1, 1], LineBox[{688, 689}]}, 
       {Hue[0.6197681783386836, 1, 1], LineBox[{689, 690}]}, 
       {Hue[0.6222988744108461, 1, 1], LineBox[{690, 691}]}, 
       {Hue[0.6273602665551711, 1, 1], LineBox[{691, 692}]}, 
       {Hue[0.6276555627425835, 1, 1], LineBox[{692, 693}]}, 
       {Hue[0.6279508589299957, 1, 1], LineBox[{693, 694}]}, 
       {Hue[0.6285414513048204, 1, 1], LineBox[{694, 695}]}, 
       {Hue[0.6297226360544697, 1, 1], LineBox[{695, 696}]}, 
       {Hue[0.6320850055537683, 1, 1], LineBox[{696, 697}]}, 
       {Hue[0.6323803017411807, 1, 1], LineBox[{697, 698}]}, 
       {Hue[0.632675597928593, 1, 1], LineBox[{698, 699}]}, 
       {Hue[0.6332661903034177, 1, 1], LineBox[{699, 700}]}, 
       {Hue[0.634447375053067, 1, 1], LineBox[{700, 701}]}, 
       {Hue[0.6368097445523656, 1, 1], LineBox[{701, 702}]}, 
       {Hue[0.637105040739778, 1, 1], LineBox[{702, 703}]}, 
       {Hue[0.6374003369271903, 1, 1], LineBox[{703, 704}]}, 
       {Hue[0.637990929302015, 1, 1], LineBox[{704, 705}]}, 
       {Hue[0.6391721140516642, 1, 1], LineBox[{705, 706}]}, 
       {Hue[0.6394674102390766, 1, 1], LineBox[{706, 707}]}, 
       {Hue[0.6397627064264888, 1, 1], LineBox[{707, 708}]}, 
       {Hue[0.6403532988013136, 1, 1], LineBox[{708, 709}]}, 
       {Hue[0.6415344835509629, 1, 1], LineBox[{709, 710}]}, 
       {Hue[0.6418297797383752, 1, 1], LineBox[{710, 711}]}, 
       {Hue[0.6421250759257875, 1, 1], LineBox[{711, 712}]}, 
       {Hue[0.6427156683006122, 1, 1], LineBox[{712, 713}]}, 
       {Hue[0.6438968530502616, 1, 1], LineBox[{713, 714}]}, 
       {Hue[0.6441921492376739, 1, 1], LineBox[{714, 715}]}, 
       {Hue[0.6444874454250862, 1, 1], LineBox[{715, 716}]}, 
       {Hue[0.6450780377999108, 1, 1], LineBox[{716, 717}]}, 
       {Hue[0.64625922254956, 1, 1], LineBox[{717, 718}]}, 
       {Hue[0.6465487031969103, 1, 1], LineBox[{718, 719}]}, 
       {Hue[0.6468381838442604, 1, 1], LineBox[{719, 720}]}, 
       {Hue[0.6474171451389605, 1, 1], LineBox[{720, 721}]}, 
       {Hue[0.6477066257863107, 1, 1], LineBox[{721, 722}]}, 
       {Hue[0.6479961064336609, 1, 1], LineBox[{722, 723}]}, 
       {Hue[0.6485750677283612, 1, 1], LineBox[{723, 724}]}, 
       {Hue[0.6488645483757112, 1, 1], LineBox[{724, 725}]}, 
       {Hue[0.6491540290230614, 1, 1], LineBox[{725, 726}]}, 
       {Hue[0.6497329903177617, 1, 1], LineBox[{726, 727}]}, 
       {Hue[0.6508909129071622, 1, 1], LineBox[{727, 728}]}, 
       {Hue[0.6511803935545123, 1, 1], LineBox[{728, 729}]}, 
       {Hue[0.6514698742018624, 1, 1], LineBox[{729, 730}]}, 
       {Hue[0.6520488354965627, 1, 1], LineBox[{730, 731}]}, 
       {Hue[0.6523383161439128, 1, 1], LineBox[{731, 732}]}, 
       {Hue[0.6526277967912629, 1, 1], LineBox[{732, 733}]}, 
       {Hue[0.6532067580859631, 1, 1], LineBox[{733, 734}]}, 
       {Hue[0.6534962387333133, 1, 1], LineBox[{734, 735}]}, 
       {Hue[0.6537857193806634, 1, 1], LineBox[{735, 736}]}, 
       {Hue[0.6540752000280136, 1, 1], LineBox[{736, 737}]}, 
       {Hue[0.6543646806753636, 1, 1], LineBox[{737, 738}]}, 
       {Hue[0.6546541613227137, 1, 1], LineBox[{738, 739}]}, 
       {Hue[0.6549436419700639, 1, 1], LineBox[{739, 740}]}, 
       {Hue[0.6552331226174141, 1, 1], LineBox[{740, 741}]}, 
       {Hue[0.6555226032647642, 1, 1], LineBox[{741, 742}]}, 
       {Hue[0.6558120839121143, 1, 1], LineBox[{742, 743}]}, 
       {Hue[0.6561015645594643, 1, 1], LineBox[{743, 744}]}, 
       {Hue[0.6563910452068146, 1, 1], LineBox[{744, 745}]}, 
       {Hue[0.6566805258541647, 1, 1], LineBox[{745, 746}]}, 
       {Hue[0.6569700065015148, 1, 1], LineBox[{746, 747}]}, 
       {Hue[0.6572594871488651, 1, 1], LineBox[{747, 748}]}, 
       {Hue[0.6575489677962151, 1, 1], LineBox[{748, 749}]}, 
       {Hue[0.6578384484435652, 1, 1], LineBox[{749, 750}]}, 
       {Hue[0.6581279290909153, 1, 1], LineBox[{750, 751}]}, 
       {Hue[0.6584174097382655, 1, 1], LineBox[{751, 752}]}, 
       {Hue[0.6587068903856157, 1, 1], LineBox[{752, 753}]}, 
       {Hue[0.6589963710329658, 1, 1], LineBox[{753, 754}]}, 
       {Hue[0.6592858516803158, 1, 1], LineBox[{754, 755}]}, 
       {Hue[0.659575332327666, 1, 1], LineBox[{755, 756}]}, 
       {Hue[0.6598648129750161, 1, 1], LineBox[{756, 757}]}, 
       {Hue[0.6601542936223663, 1, 1], LineBox[{757, 758}]}, 
       {Hue[0.6604437742697165, 1, 1], LineBox[{758, 759}]}, 
       {Hue[0.6607332549170665, 1, 1], LineBox[{759, 760}]}, 
       {Hue[0.6610227355644166, 1, 1], LineBox[{760, 761}]}, 
       {Hue[0.6613122162117667, 1, 1], LineBox[{761, 762}]}, 
       {Hue[0.661601696859117, 1, 1], LineBox[{762, 763}]}, 
       {Hue[0.661891177506467, 1, 1], LineBox[{763, 764}]}, 
       {Hue[0.6624701388011672, 1, 1], LineBox[{764, 765}]}, 
       {Hue[0.6627596194485174, 1, 1], LineBox[{765, 766}]}, 
       {Hue[0.6630491000958674, 1, 1], LineBox[{766, 767}]}, 
       {Hue[0.6636280613905677, 1, 1], LineBox[{767, 768}]}, 
       {Hue[0.6639175420379179, 1, 1], LineBox[{768, 769}]}, 
       {Hue[0.664207022685268, 1, 1], LineBox[{769, 770}]}, 
       {Hue[0.6647859839799682, 1, 1], LineBox[{770, 771}]}, 
       {Hue[0.6651000996503478, 1, 1], LineBox[{771, 772}]}, 
       {Hue[0.6654142153207272, 1, 1], LineBox[{772, 773}]}, 
       {Hue[0.6660424466614863, 1, 1], LineBox[{773, 774}]}, 
       {Hue[0.6663565623318657, 1, 1], LineBox[{774, 775}]}, 
       {Hue[0.6666706780022451, 1, 1], LineBox[{775, 776}]}, 
       {Hue[0.6672989093430042, 1, 1], LineBox[{776, 777}]}, 
       {Hue[0.6676130250133836, 1, 1], LineBox[{777, 778}]}, 
       {Hue[0.6679271406837631, 1, 1], LineBox[{778, 779}]}, 
       {Hue[0.6685553720245221, 1, 1], LineBox[{779, 780}]}, 
       {Hue[0.66981183470604, 1, 1], LineBox[{780, 781}]}, 
       {Hue[0.6701259503764195, 1, 1], LineBox[{781, 782}]}, 
       {Hue[0.6704400660467991, 1, 1], LineBox[{782, 783}]}, 
       {Hue[0.671068297387558, 1, 1], LineBox[{783, 784}]}, 
       {Hue[0.6723247600690759, 1, 1], LineBox[{784, 785}]}, 
       {Hue[0.6748376854321119, 1, 1], LineBox[{785, 786}]}, 
       {Hue[0.6751518011024913, 1, 1], LineBox[{786, 787}]}, 
       {Hue[0.6754659167728708, 1, 1], LineBox[{787, 788}]}, 
       {Hue[0.6760941481136298, 1, 1], LineBox[{788, 789}]}, 
       {Hue[0.6773506107951477, 1, 1], LineBox[{789, 790}]}, 
       {Hue[0.6798635361581836, 1, 1], LineBox[{790, 791}]}, 
       {Hue[0.680177651828563, 1, 1], LineBox[{791, 792}]}, 
       {Hue[0.6804917674989426, 1, 1], LineBox[{792, 793}]}, 
       {Hue[0.6811199988397015, 1, 1], LineBox[{793, 794}]}, 
       {Hue[0.6823764615212196, 1, 1], LineBox[{794, 795}]}, 
       {Hue[0.6848893868842554, 1, 1], LineBox[{795, 796}]}, 
       {Hue[0.6851824617330269, 1, 1], LineBox[{796, 797}]}, 
       {Hue[0.6854755365817984, 1, 1], LineBox[{797, 798}]}, 
       {Hue[0.6860616862793414, 1, 1], LineBox[{798, 799}]}, 
       {Hue[0.6872339856744274, 1, 1], LineBox[{799, 800}]}, 
       {Hue[0.6895785844645994, 1, 1], LineBox[{800, 801}]}, 
       {Hue[0.6942677820449434, 1, 1], LineBox[{801, 802}]}, 
       {Hue[0.6945608568937149, 1, 1], LineBox[{802, 803}]}, 
       {Hue[0.6948539317424864, 1, 1], LineBox[{803, 804}]}, 
       {Hue[0.6954400814400294, 1, 1], LineBox[{804, 805}]}, 
       {Hue[0.6966123808351155, 1, 1], LineBox[{805, 806}]}, 
       {Hue[0.6989569796252876, 1, 1], LineBox[{806, 807}]}, 
       {Hue[0.7036461772056316, 1, 1], LineBox[{807, 808}]}, 
       {Hue[0.7039638870774325, 1, 1], LineBox[{808, 809}]}, 
       {Hue[0.7042815969492333, 1, 1], LineBox[{809, 810}]}, 
       {Hue[0.704917016692835, 1, 1], LineBox[{810, 811}]}, 
       {Hue[0.7061878561800385, 1, 1], LineBox[{811, 812}]}, 
       {Hue[0.7087295351544455, 1, 1], LineBox[{812, 813}]}, 
       {Hue[0.7138128931032592, 1, 1], LineBox[{813, 814}]}, 
       {Hue[0.723979609000887, 1, 1], LineBox[{814, 815}]}, 
       {Hue[0.7242762780510799, 1, 1], LineBox[{815, 816}]}, 
       {Hue[0.7245729471012727, 1, 1], LineBox[{816, 817}]}, 
       {Hue[0.7251662852016585, 1, 1], LineBox[{817, 818}]}, 
       {Hue[0.72635296140243, 1, 1], LineBox[{818, 819}]}, 
       {Hue[0.728726313803973, 1, 1], LineBox[{819, 820}]}, 
       {Hue[0.7334730186070592, 1, 1], LineBox[{820, 821}]}, 
       {Hue[0.7337696876572521, 1, 1], LineBox[{821, 822}]}, 
       {Hue[0.7340663567074449, 1, 1], LineBox[{822, 823}]}, 
       {Hue[0.7346596948078307, 1, 1], LineBox[{823, 824}]}, 
       {Hue[0.7358463710086022, 1, 1], LineBox[{824, 825}]}, 
       {Hue[0.7382197234101453, 1, 1], LineBox[{825, 826}]}, 
       {Hue[0.7429664282132313, 1, 1], LineBox[{826, 827}]}, 
       {Hue[0.743257281723362, 1, 1], LineBox[{827, 828}]}, 
       {Hue[0.7435481352334927, 1, 1], LineBox[{828, 829}]}, 
       {Hue[0.7441298422537541, 1, 1], LineBox[{829, 830}]}, 
       {Hue[0.7452932562942768, 1, 1], LineBox[{830, 831}]}, 
       {Hue[0.7455841098044075, 1, 1], LineBox[{831, 832}]}, 
       {Hue[0.7458749633145382, 1, 1], LineBox[{832, 833}]}, 
       {Hue[0.7464566703347995, 1, 1], LineBox[{833, 834}]}, 
       {Hue[0.7476200843753222, 1, 1], LineBox[{834, 835}]}, 
       {Hue[0.7479109378854529, 1, 1], LineBox[{835, 836}]}, 
       {Hue[0.7482017913955836, 1, 1], LineBox[{836, 837}]}, 
       {Hue[0.748783498415845, 1, 1], LineBox[{837, 838}]}, 
       {Hue[0.7490743519259756, 1, 1], LineBox[{838, 839}]}, 
       {Hue[0.7493652054361063, 1, 1], LineBox[{839, 840}]}, 
       {Hue[0.749656058946237, 1, 1], LineBox[{840, 841}]}, 
       {Hue[0.7499469124563677, 1, 1], LineBox[{841, 842}]}, 
       {Hue[0.2502377659664984, 1, 1], LineBox[{842, 843}]}, 
       {Hue[0.2505286194766291, 1, 1], LineBox[{843, 844}]}, 
       {Hue[0.2508194729867598, 1, 1], LineBox[{844, 845}]}, 
       {Hue[0.2511103264968904, 1, 1], LineBox[{845, 846}]}, 
       {Hue[0.25227374053741314`, 1, 1], LineBox[{846, 847}]}, 
       {Hue[0.25256459404754383`, 1, 1], LineBox[{847, 848}]}, 
       {Hue[0.25285544755767453`, 1, 1], LineBox[{848, 849}]}, 
       {Hue[0.25343715457793586`, 1, 1], LineBox[{849, 850}]}, 
       {Hue[0.2546005686184586, 1, 1], LineBox[{850, 851}]}, 
       {Hue[0.25692739669950404`, 1, 1], LineBox[{851, 852}]}, 
       {Hue[0.2615810528615949, 1, 1], LineBox[{852, 853}]}, 
       {Hue[0.26189654139475493`, 1, 1], LineBox[{853, 854}]}, 
       {Hue[0.2622120299279149, 1, 1], LineBox[{854, 855}]}, 
       {Hue[0.262843006994235, 1, 1], LineBox[{855, 856}]}, 
       {Hue[0.2641049611268752, 1, 1], LineBox[{856, 857}]}, 
       {Hue[0.2666288693921555, 1, 1], LineBox[{857, 858}]}, 
       {Hue[0.26694435792531557`, 1, 1], LineBox[{858, 859}]}, 
       {Hue[0.2672598464584756, 1, 1], LineBox[{859, 860}]}, 
       {Hue[0.26789082352479565`, 1, 1], LineBox[{860, 861}]}, 
       {Hue[0.26915277765743584`, 1, 1], LineBox[{861, 862}]}, 
       {Hue[0.27167668592271615`, 1, 1], LineBox[{862, 863}]}, 
       {Hue[0.2719921744558762, 1, 1], LineBox[{863, 864}]}, 
       {Hue[0.2723076629890362, 1, 1], LineBox[{864, 865}]}, 
       {Hue[0.2729386400553563, 1, 1], LineBox[{865, 866}]}, 
       {Hue[0.27420059418799647`, 1, 1], LineBox[{866, 867}]}, 
       {Hue[0.2767245024532768, 1, 1], LineBox[{867, 868}]}, 
       {Hue[0.2817723189838374, 1, 1], LineBox[{868, 869}]}, 
       {Hue[0.2820667666953895, 1, 1], LineBox[{869, 870}]}, 
       {Hue[0.28236121440694156`, 1, 1], LineBox[{870, 871}]}, 
       {Hue[0.28295010983004565`, 1, 1], LineBox[{871, 872}]}, 
       {Hue[0.2841279006762539, 1, 1], LineBox[{872, 873}]}, 
       {Hue[0.28648348236867033`, 1, 1], LineBox[{873, 874}]}, 
       {Hue[0.2867779300802224, 1, 1], LineBox[{874, 875}]}, 
       {Hue[0.28707237779177447`, 1, 1], LineBox[{875, 876}]}, 
       {Hue[0.28766127321487855`, 1, 1], LineBox[{876, 877}]}, 
       {Hue[0.2888390640610868, 1, 1], LineBox[{877, 878}]}, 
       {Hue[0.29119464575350323`, 1, 1], LineBox[{878, 879}]}, 
       {Hue[0.2914890934650553, 1, 1], LineBox[{879, 880}]}, 
       {Hue[0.2917835411766074, 1, 1], LineBox[{880, 881}]}, 
       {Hue[0.29237243659971146`, 1, 1], LineBox[{881, 882}]}, 
       {Hue[0.2935502274459197, 1, 1], LineBox[{882, 883}]}, 
       {Hue[0.29384467515747176`, 1, 1], LineBox[{883, 884}]}, 
       {Hue[0.29413912286902383`, 1, 1], LineBox[{884, 885}]}, 
       {Hue[0.2947280182921279, 1, 1], LineBox[{885, 886}]}, 
       {Hue[0.29590580913833614`, 1, 1], LineBox[{886, 887}]}, 
       {Hue[0.2962002568498882, 1, 1], LineBox[{887, 888}]}, 
       {Hue[0.2964947045614402, 1, 1], LineBox[{888, 889}]}, 
       {Hue[0.29708359998454437`, 1, 1], LineBox[{889, 890}]}, 
       {Hue[0.2982613908307526, 1, 1], LineBox[{890, 891}]}, 
       {Hue[0.29855583854230466`, 1, 1], LineBox[{891, 892}]}, 
       {Hue[0.2988502862538567, 1, 1], LineBox[{892, 893}]}, 
       {Hue[0.2994391816769608, 1, 1], LineBox[{893, 894}]}, 
       {Hue[0.30061697252316905`, 1, 1], LineBox[{894, 895}]}, 
       {Hue[0.30093605525775047`, 1, 1], LineBox[{895, 896}]}, 
       {Hue[0.3012551379923319, 1, 1], LineBox[{896, 897}]}, 
       {Hue[0.30189330346149473`, 1, 1], LineBox[{897, 898}]}, 
       {Hue[0.30221238619607615`, 1, 1], LineBox[{898, 899}]}, 
       {Hue[0.3025314689306576, 1, 1], LineBox[{899, 900}]}, 
       {Hue[0.3031696343998204, 1, 1], LineBox[{900, 901}]}, 
       {Hue[0.30348871713440184`, 1, 1], LineBox[{901, 902}]}, 
       {Hue[0.3038077998689832, 1, 1], LineBox[{902, 903}]}, 
       {Hue[0.30444596533814605`, 1, 1], LineBox[{903, 904}]}, 
       {Hue[0.30572229627647174`, 1, 1], LineBox[{904, 905}]}, 
       {Hue[0.30604137901105316`, 1, 1], LineBox[{905, 906}]}, 
       {Hue[0.3063604617456346, 1, 1], LineBox[{906, 907}]}, 
       {Hue[0.3069986272147974, 1, 1], LineBox[{907, 908}]}, 
       {Hue[0.30731770994937885`, 1, 1], LineBox[{908, 909}]}, 
       {Hue[0.3076367926839602, 1, 1], LineBox[{909, 910}]}, 
       {Hue[0.30827495815312306`, 1, 1], LineBox[{910, 911}]}, 
       {Hue[0.3085940408877045, 1, 1], LineBox[{911, 912}]}, 
       {Hue[0.3089131236222859, 1, 1], LineBox[{912, 913}]}, 
       {Hue[0.3092322063568673, 1, 1], LineBox[{913, 914}]}, 
       {Hue[0.30955128909144874`, 1, 1], LineBox[{914, 915}]}, 
       {Hue[0.30987037182603017`, 1, 1], LineBox[{915, 916}]}, 
       {Hue[0.3101894545606116, 1, 1], LineBox[{916, 917}]}, 
       {Hue[0.310508537295193, 1, 1], LineBox[{917, 918}]}, 
       {Hue[0.31082762002977443`, 1, 1], LineBox[{918, 919}]}, 
       {Hue[0.31114670276435585`, 1, 1], LineBox[{919, 920}]}, 
       {Hue[0.3114657854989372, 1, 1], LineBox[{920, 921}]}, 
       {Hue[0.3117848682335187, 1, 1], LineBox[{921, 922}]}, 
       {Hue[0.31210395096810006`, 1, 1], LineBox[{922, 923}]}, 
       {Hue[0.3124230337026815, 1, 1], LineBox[{923, 924}]}, 
       {Hue[0.3127421164372629, 1, 1], LineBox[{924, 925}]}, 
       {Hue[0.31306119917184433`, 1, 1], LineBox[{925, 926}]}, 
       {Hue[0.31338028190642575`, 1, 1], LineBox[{926, 927}]}, 
       {Hue[0.3136993646410072, 1, 1], LineBox[{927, 928}]}, 
       {Hue[0.3140184473755886, 1, 1], LineBox[{928, 929}]}, 
       {Hue[0.31433753011017, 1, 1], LineBox[{929, 930}]}, 
       {Hue[0.31465661284475144`, 1, 1], LineBox[{930, 931}]}, 
       {Hue[0.31497569557933286`, 1, 1], LineBox[{931, 932}]}, 
       {Hue[0.3152947783139142, 1, 1], LineBox[{932, 933}]}, 
       {Hue[0.3156138610484957, 1, 1], LineBox[{933, 934}]}, 
       {Hue[0.31593294378307707`, 1, 1], LineBox[{934, 935}]}, 
       {Hue[0.3162520265176585, 1, 1], LineBox[{935, 936}]}, 
       {Hue[0.3165711092522399, 1, 1], LineBox[{936, 937}]}, 
       {Hue[0.31689019198682133`, 1, 1], LineBox[{937, 938}]}, 
       {Hue[0.31720927472140276`, 1, 1], LineBox[{938, 939}]}, 
       {Hue[0.3175283574559842, 1, 1], LineBox[{939, 940}]}, 
       {Hue[0.3178474401905656, 1, 1], LineBox[{940, 941}]}, 
       {Hue[0.31848560565972844`, 1, 1], LineBox[{941, 942}]}, 
       {Hue[0.31880468839430987`, 1, 1], LineBox[{942, 943}]}, 
       {Hue[0.31912377112889123`, 1, 1], LineBox[{943, 944}]}, 
       {Hue[0.31976193659805413`, 1, 1], LineBox[{944, 945}]}, 
       {Hue[0.3200810193326355, 1, 1], LineBox[{945, 946}]}, 
       {Hue[0.320400102067217, 1, 1], LineBox[{946, 947}]}, 
       {Hue[0.32103826753637976`, 1, 1], LineBox[{947, 948}]}, 
       {Hue[0.321351534730899, 1, 1], LineBox[{948, 949}]}, 
       {Hue[0.32166480192541824`, 1, 1], LineBox[{949, 950}]}, 
       {Hue[0.3222913363144566, 1, 1], LineBox[{950, 951}]}, 
       {Hue[0.32354440509253346`, 1, 1], LineBox[{951, 952}]}, 
       {Hue[0.3238576722870527, 1, 1], LineBox[{952, 953}]}, 
       {Hue[0.32417093948157194`, 1, 1], LineBox[{953, 954}]}, 
       {Hue[0.32479747387061036`, 1, 1], LineBox[{954, 955}]}, 
       {Hue[0.3260505426486872, 1, 1], LineBox[{955, 956}]}, 
       {Hue[0.32636380984320645`, 1, 1], LineBox[{956, 957}]}, 
       {Hue[0.3266770770377257, 1, 1], LineBox[{957, 958}]}, 
       {Hue[0.32730361142676406`, 1, 1], LineBox[{958, 959}]}, 
       {Hue[0.32855668020484097`, 1, 1], LineBox[{959, 960}]}, 
       {Hue[0.33106281776099467`, 1, 1], LineBox[{960, 961}]}, 
       {Hue[0.3313760849555139, 1, 1], LineBox[{961, 962}]}, 
       {Hue[0.33168935215003315`, 1, 1], LineBox[{962, 963}]}, 
       {Hue[0.3323158865390715, 1, 1], LineBox[{963, 964}]}, 
       {Hue[0.3335689553171484, 1, 1], LineBox[{964, 965}]}, 
       {Hue[0.3360750928733021, 1, 1], LineBox[{965, 966}]}, 
       {Hue[0.33638836006782136`, 1, 1], LineBox[{966, 967}]}, 
       {Hue[0.3367016272623406, 1, 1], LineBox[{967, 968}]}, 
       {Hue[0.337328161651379, 1, 1], LineBox[{968, 969}]}, 
       {Hue[0.33858123042945587`, 1, 1], LineBox[{969, 970}]}, 
       {Hue[0.34108736798560957`, 1, 1], LineBox[{970, 971}]}, 
       {Hue[0.34137959435852083`, 1, 1], LineBox[{971, 972}]}, 
       {Hue[0.34167182073143204`, 1, 1], LineBox[{972, 973}]}, 
       {Hue[0.3422562734772545, 1, 1], LineBox[{973, 974}]}, 
       {Hue[0.34342517896889946`, 1, 1], LineBox[{974, 975}]}, 
       {Hue[0.34576298995218935`, 1, 1], LineBox[{975, 976}]}, 
       {Hue[0.350438611918769, 1, 1], LineBox[{976, 977}]}, 
       {Hue[0.35978985585192846`, 1, 1], LineBox[{977, 978}]}, 
       {Hue[0.38006898519212645`, 1, 1], LineBox[{978, 979}]}, 
       {Hue[0.38036480576645904`, 1, 1], LineBox[{979, 980}]}, 
       {Hue[0.3806606263407917, 1, 1], LineBox[{980, 981}]}, 
       {Hue[0.3812522674894569, 1, 1], LineBox[{981, 982}]}, 
       {Hue[0.38243554978678734`, 1, 1], LineBox[{982, 983}]}, 
       {Hue[0.38480211438144823`, 1, 1], LineBox[{983, 984}]}, 
       {Hue[0.38953524357077, 1, 1], LineBox[{984, 985}]}, 
       {Hue[0.38983106414510266`, 1, 1], LineBox[{985, 986}]}, 
       {Hue[0.39012688471943524`, 1, 1], LineBox[{986, 987}]}, 
       {Hue[0.39071852586810046`, 1, 1], LineBox[{987, 988}]}, 
       {Hue[0.3919018081654309, 1, 1], LineBox[{988, 989}]}, 
       {Hue[0.39426837276009186`, 1, 1], LineBox[{989, 990}]}, 
       {Hue[0.39900150194941353`, 1, 1], LineBox[{990, 991}]}, 
       {Hue[0.39929150698368404`, 1, 1], LineBox[{991, 992}]}, 
       {Hue[0.39958151201795444`, 1, 1], LineBox[{992, 993}]}, 
       {Hue[0.40016152208649525`, 1, 1], LineBox[{993, 994}]}, 
       {Hue[0.40132154222357685`, 1, 1], LineBox[{994, 995}]}, 
       {Hue[0.4036415824977401, 1, 1], LineBox[{995, 996}]}, 
       {Hue[0.4039315875320105, 1, 1], LineBox[{996, 997}]}, 
       {Hue[0.40422159256628093`, 1, 1], LineBox[{997, 998}]}, 
       {Hue[0.40480160263482173`, 1, 1], LineBox[{998, 999}]}, 
       {Hue[0.40596162277190334`, 1, 1], LineBox[{999, 1000}]}, 
       {Hue[0.40828166304606667`, 1, 1], LineBox[{1000, 1001}]}, 
       {Hue[0.4085716680803371, 1, 1], LineBox[{1001, 1002}]}, 
       {Hue[0.4088616731146075, 1, 1], LineBox[{1002, 1003}]}, 
       {Hue[0.40944168318314833`, 1, 1], LineBox[{1003, 1004}]}, 
       {Hue[0.41060170332022994`, 1, 1], LineBox[{1004, 1005}]}, 
       {Hue[0.4108917083545004, 1, 1], LineBox[{1005, 1006}]}, 
       {Hue[0.4111817133887708, 1, 1], LineBox[{1006, 1007}]}, 
       {Hue[0.4117617234573116, 1, 1], LineBox[{1007, 1008}]}, 
       {Hue[0.4129217435943932, 1, 1], LineBox[{1008, 1009}]}, 
       {Hue[0.4132117486286636, 1, 1], LineBox[{1009, 1010}]}, 
       {Hue[0.413501753662934, 1, 1], LineBox[{1010, 1011}]}, 
       {Hue[0.4140817637314748, 1, 1], LineBox[{1011, 1012}]}, 
       {Hue[0.4143717687657452, 1, 1], LineBox[{1012, 1013}]}, 
       {Hue[0.4146617738000156, 1, 1], LineBox[{1013, 1014}]}, 
       {Hue[0.4152417838685564, 1, 1], LineBox[{1014, 1015}]}, 
       {Hue[0.4155317889028268, 1, 1], LineBox[{1015, 1016}]}, 
       {Hue[0.41582179393709723`, 1, 1], LineBox[{1016, 1017}]}, 
       {Hue[0.41611179897136763`, 1, 1], LineBox[{1017, 1018}]}, 
       {Hue[0.4164018040056381, 1, 1], LineBox[{1018, 1019}]}, 
       {Hue[0.9166918090399084, 1, 1], LineBox[{1019, 1020}]}, 
       {Hue[0.9169818140741789, 1, 1], LineBox[{1020, 1021}]}, 
       {Hue[0.9172718191084492, 1, 1], LineBox[{1021, 1022}]}, 
       {Hue[0.9175618241427197, 1, 1], LineBox[{1022, 1023}]}, 
       {Hue[0.9178764642000194, 1, 1], LineBox[{1023, 1024}]}, 
       {Hue[0.9181911042573192, 1, 1], LineBox[{1024, 1025}]}, 
       {Hue[0.9188203843719188, 1, 1], LineBox[{1025, 1026}]}, 
       {Hue[0.9200789446011179, 1, 1], LineBox[{1026, 1027}]}, 
       {Hue[0.922596065059516, 1, 1], LineBox[{1027, 1028}]}, 
       {Hue[0.9229107051168157, 1, 1], LineBox[{1028, 1029}]}, 
       {Hue[0.9232253451741156, 1, 1], LineBox[{1029, 1030}]}, 
       {Hue[0.9238546252887151, 1, 1], LineBox[{1030, 1031}]}, 
       {Hue[0.9251131855179142, 1, 1], LineBox[{1031, 1032}]}, 
       {Hue[0.9276303059763124, 1, 1], LineBox[{1032, 1033}]}, 
       {Hue[0.9279449460336121, 1, 1], LineBox[{1033, 1034}]}, 
       {Hue[0.928259586090912, 1, 1], LineBox[{1034, 1035}]}, 
       {Hue[0.9288888662055114, 1, 1], LineBox[{1035, 1036}]}, 
       {Hue[0.9301474264347105, 1, 1], LineBox[{1036, 1037}]}, 
       {Hue[0.9326645468931086, 1, 1], LineBox[{1037, 1038}]}, 
       {Hue[0.9376987878099049, 1, 1], LineBox[{1038, 1039}]}, 
       {Hue[0.9379923870455967, 1, 1], LineBox[{1039, 1040}]}, 
       {Hue[0.9382859862812885, 1, 1], LineBox[{1040, 1041}]}, 
       {Hue[0.938873184752672, 1, 1], LineBox[{1041, 1042}]}, 
       {Hue[0.9400475816954391, 1, 1], LineBox[{1042, 1043}]}, 
       {Hue[0.9423963755809734, 1, 1], LineBox[{1043, 1044}]}, 
       {Hue[0.9426899748166653, 1, 1], LineBox[{1044, 1045}]}, 
       {Hue[0.9429835740523571, 1, 1], LineBox[{1045, 1046}]}, 
       {Hue[0.9435707725237407, 1, 1], LineBox[{1046, 1047}]}, 
       {Hue[0.9447451694665078, 1, 1], LineBox[{1047, 1048}]}, 
       {Hue[0.9470939633520421, 1, 1], LineBox[{1048, 1049}]}, 
       {Hue[0.9473875625877338, 1, 1], LineBox[{1049, 1050}]}, 
       {Hue[0.9476811618234257, 1, 1], LineBox[{1050, 1051}]}, 
       {Hue[0.9482683602948092, 1, 1], LineBox[{1051, 1052}]}, 
       {Hue[0.9494427572375763, 1, 1], LineBox[{1052, 1053}]}, 
       {Hue[0.9497363564732682, 1, 1], LineBox[{1053, 1054}]}, 
       {Hue[0.9500299557089599, 1, 1], LineBox[{1054, 1055}]}, 
       {Hue[0.9506171541803434, 1, 1], LineBox[{1055, 1056}]}, 
       {Hue[0.9517915511231106, 1, 1], LineBox[{1056, 1057}]}, 
       {Hue[0.9520851503588025, 1, 1], LineBox[{1057, 1058}]}, 
       {Hue[0.9523787495944943, 1, 1], LineBox[{1058, 1059}]}, 
       {Hue[0.9529659480658779, 1, 1], LineBox[{1059, 1060}]}, 
       {Hue[0.954140345008645, 1, 1], LineBox[{1060, 1061}]}, 
       {Hue[0.9544339442443368, 1, 1], LineBox[{1061, 1062}]}, 
       {Hue[0.9547275434800285, 1, 1], LineBox[{1062, 1063}]}, 
       {Hue[0.9553147419514121, 1, 1], LineBox[{1063, 1064}]}, 
       {Hue[0.9564891388941792, 1, 1], LineBox[{1064, 1065}]}, 
       {Hue[0.9568073731529004, 1, 1], LineBox[{1065, 1066}]}, 
       {Hue[0.9571256074116216, 1, 1], LineBox[{1066, 1067}]}, 
       {Hue[0.9577620759290638, 1, 1], LineBox[{1067, 1068}]}, 
       {Hue[0.9580803101877848, 1, 1], LineBox[{1068, 1069}]}, 
       {Hue[0.9583985444465061, 1, 1], LineBox[{1069, 1070}]}, 
       {Hue[0.9590350129639483, 1, 1], LineBox[{1070, 1071}]}, 
       {Hue[0.9593532472226696, 1, 1], LineBox[{1071, 1072}]}, 
       {Hue[0.9596714814813907, 1, 1], LineBox[{1072, 1073}]}, 
       {Hue[0.960307949998833, 1, 1], LineBox[{1073, 1074}]}, 
       {Hue[0.9615808870337176, 1, 1], LineBox[{1074, 1075}]}, 
       {Hue[0.9618991212924386, 1, 1], LineBox[{1075, 1076}]}, 
       {Hue[0.9622173555511598, 1, 1], LineBox[{1076, 1077}]}, 
       {Hue[0.9628538240686021, 1, 1], LineBox[{1077, 1078}]}, 
       {Hue[0.9631720583273232, 1, 1], LineBox[{1078, 1079}]}, 
       {Hue[0.9634902925860445, 1, 1], LineBox[{1079, 1080}]}, 
       {Hue[0.9641267611034867, 1, 1], LineBox[{1080, 1081}]}, 
       {Hue[0.9644449953622078, 1, 1], LineBox[{1081, 1082}]}, 
       {Hue[0.9647632296209291, 1, 1], LineBox[{1082, 1083}]}, 
       {Hue[0.9650814638796502, 1, 1], LineBox[{1083, 1084}]}, 
       {Hue[0.9653996981383713, 1, 1], LineBox[{1084, 1085}]}, 
       {Hue[0.9657179323970924, 1, 1], LineBox[{1085, 1086}]}, 
       {Hue[0.9660361666558135, 1, 1], LineBox[{1086, 1087}]}, 
       {Hue[0.9666726351732559, 1, 1], LineBox[{1087, 1088}]}, 
       {Hue[0.966990869431977, 1, 1], LineBox[{1088, 1089}]}, 
       {Hue[0.9673091036906982, 1, 1], LineBox[{1089, 1090}]}, 
       {Hue[0.9676273379494194, 1, 1], LineBox[{1090, 1091}]}, 
       {Hue[0.9679455722081405, 1, 1], LineBox[{1091, 1092}]}, 
       {Hue[0.9682638064668616, 1, 1], LineBox[{1092, 1093}]}, 
       {Hue[0.9685820407255828, 1, 1], LineBox[{1093, 1094}]}, 
       {Hue[0.968900274984304, 1, 1], LineBox[{1094, 1095}]}, 
       {Hue[0.9692185092430251, 1, 1], LineBox[{1095, 1096}]}, 
       {Hue[0.9695367435017461, 1, 1], LineBox[{1096, 1097}]}, 
       {Hue[0.9698549777604673, 1, 1], LineBox[{1097, 1098}]}, 
       {Hue[0.9701732120191885, 1, 1], LineBox[{1098, 1099}]}, 
       {Hue[0.9704914462779096, 1, 1], LineBox[{1099, 1100}]}, 
       {Hue[0.9708096805366307, 1, 1], LineBox[{1100, 1101}]}, 
       {Hue[0.9711279147953519, 1, 1], LineBox[{1101, 1102}]}, 
       {Hue[0.9714461490540731, 1, 1], LineBox[{1102, 1103}]}, 
       {Hue[0.9717643833127942, 1, 1], LineBox[{1103, 1104}]}, 
       {Hue[0.9720826175715154, 1, 1], LineBox[{1104, 1105}]}, 
       {Hue[0.9724008518302366, 1, 1], LineBox[{1105, 1106}]}, 
       {Hue[0.9727190860889577, 1, 1], LineBox[{1106, 1107}]}, 
       {Hue[0.9730373203476788, 1, 1], LineBox[{1107, 1108}]}, 
       {Hue[0.9733555546063999, 1, 1], LineBox[{1108, 1109}]}, 
       {Hue[0.973673788865121, 1, 1], LineBox[{1109, 1110}]}, 
       {Hue[0.9739920231238423, 1, 1], LineBox[{1110, 1111}]}, 
       {Hue[0.9743102573825634, 1, 1], LineBox[{1111, 1112}]}, 
       {Hue[0.9746284916412845, 1, 1], LineBox[{1112, 1113}]}, 
       {Hue[0.9749467259000056, 1, 1], LineBox[{1113, 1114}]}, 
       {Hue[0.9752649601587269, 1, 1], LineBox[{1114, 1115}]}, 
       {Hue[0.975583194417448, 1, 1], LineBox[{1115, 1116}]}, 
       {Hue[0.9759014286761691, 1, 1], LineBox[{1116, 1117}]}, 
       {Hue[0.9762196629348903, 1, 1], LineBox[{1117, 1118}]}, 
       {Hue[0.9768561314523325, 1, 1], LineBox[{1118, 1119}]}, 
       {Hue[0.9768656420513168, 1, 1], LineBox[{1119, 1120}]}, 
       {Hue[0.976875152650301, 1, 1], LineBox[{1120, 1121}]}, 
       {Hue[0.9768941738482696, 1, 1], LineBox[{1121, 1122}]}, 
       {Hue[0.9769322162442068, 1, 1], LineBox[{1122, 1123}]}, 
       {Hue[0.9770083010360808, 1, 1], LineBox[{1123, 1124}]}, 
       {Hue[0.9771604706198291, 1, 1], LineBox[{1124, 1125}]}, 
       {Hue[0.9771699812188133, 1, 1], LineBox[{1125, 1126}]}, 
       {Hue[0.9771794918177976, 1, 1], LineBox[{1126, 1127}]}, 
       {Hue[0.9771985130157662, 1, 1], LineBox[{1127, 1128}]}, 
       {Hue[0.9772365554117033, 1, 1], LineBox[{1128, 1129}]}, 
       {Hue[0.9773126402035773, 1, 1], LineBox[{1129, 1130}]}, 
       {Hue[0.9773221508025617, 1, 1], LineBox[{1130, 1131}]}, 
       {Hue[0.9773316614015459, 1, 1], LineBox[{1131, 1132}]}, 
       {Hue[0.9773506825995144, 1, 1], LineBox[{1132, 1133}]}, 
       {Hue[0.9773887249954515, 1, 1], LineBox[{1133, 1134}]}, 
       {Hue[0.9773982355944357, 1, 1], LineBox[{1134, 1135}]}, 
       {Hue[0.9774077461934201, 1, 1], LineBox[{1135, 1136}]}, 
       {Hue[0.9774267673913886, 1, 1], LineBox[{1136, 1137}]}, 
       {Hue[0.9774362779903729, 1, 1], LineBox[{1137, 1138}]}, 
       {Hue[0.9774457885893572, 1, 1], LineBox[{1138, 1139}]}, 
       {Hue[0.9774552991883414, 1, 1], LineBox[{1139, 1140}]}}}},
    VertexColors->{
      Hue[0.022535190212674343`, 1, 1], 
      Hue[0.022828064763465546`, 1, 1], 
      Hue[0.02312093931425682, 1, 1], 
      Hue[0.023706688415839228`, 1, 1], 
      Hue[0.02399956296663043, 1, 1], 
      Hue[0.024292437517421633`, 1, 1], 
      Hue[0.024878186619004036`, 1, 1], 
      Hue[0.02517106116979524, 1, 1], 
      Hue[0.025463935720586448`, 1, 1], 
      Hue[0.026049684822168923`, 1, 1], 
      Hue[0.026342559372960126`, 1, 1], 
      Hue[0.02663543392375139, 1, 1], 
      Hue[0.027221183025333804`, 1, 1], 
      Hue[0.027514057576125007`, 1, 1], 
      Hue[0.027806932126916206`, 1, 1], 
      Hue[0.028099806677707416`, 1, 1], 
      Hue[0.028392681228498615`, 1, 1], 
      Hue[0.02868555577928982, 1, 1], 
      Hue[0.028978430330081024`, 1, 1], 
      Hue[0.0292713048808723, 1, 1], 
      Hue[0.029564179431663503`, 1, 1], 
      Hue[0.029857053982454702`, 1, 1], 
      Hue[0.030149928533245978`, 1, 1], 
      Hue[0.030442803084037184`, 1, 1], 
      Hue[0.030735677634828387`, 1, 1], 
      Hue[0.031028552185619583`, 1, 1], 
      Hue[0.031321426736410786`, 1, 1], 
      Hue[0.03161430128720199, 1, 1], 
      Hue[0.031907175837993206`, 1, 1], 
      Hue[0.0322000503887844, 1, 1], 
      Hue[0.0324929249395756, 1, 1], 
      Hue[0.03278579949036681, 1, 1], 
      Hue[0.033078674041158, 1, 1], 
      Hue[0.03337154859194921, 1, 1], 
      Hue[0.033664423142740416`, 1, 1], 
      Hue[0.033957297693531695`, 1, 1], 
      Hue[0.0342501722443229, 1, 1], 
      Hue[0.0345430467951141, 1, 1], 
      Hue[0.034835921345905366`, 1, 1], 
      Hue[0.035128795896696575`, 1, 1], 
      Hue[0.03542167044748777, 1, 1], 
      Hue[0.03571454499827898, 1, 1], 
      Hue[0.03600741954907018, 1, 1], 
      Hue[0.03630029409986138, 1, 1], 
      Hue[0.03659316865065259, 1, 1], 
      Hue[0.0368860432014438, 1, 1], 
      Hue[0.037178917752234995`, 1, 1], 
      Hue[0.0374717923030262, 1, 1], 
      Hue[0.0377646668538174, 1, 1], 
      Hue[0.03805754140460861, 1, 1], 
      Hue[0.03835041595539981, 1, 1], 
      Hue[0.03893616505698228, 1, 1], 
      Hue[0.03922903960777349, 1, 1], 
      Hue[0.039521914158564764`, 1, 1], 
      Hue[0.040107663260147156`, 1, 1], 
      Hue[0.041279161463311974`, 1, 1], 
      Hue[0.041596671037132556`, 1, 1], 
      Hue[0.04191418061095314, 1, 1], 
      Hue[0.0425491997585943, 1, 1], 
      Hue[0.04286670933241487, 1, 1], 
      Hue[0.04318421890623545, 1, 1], 
      Hue[0.04381923805387661, 1, 1], 
      Hue[0.04413674762769719, 1, 1], 
      Hue[0.044454257201517765`, 1, 1], 
      Hue[0.04508927634915892, 1, 1], 
      Hue[0.04635931464444118, 1, 1], 
      Hue[0.04667682421826168, 1, 1], 
      Hue[0.04699433379208225, 1, 1], 
      Hue[0.04762935293972342, 1, 1], 
      Hue[0.04889939123500573, 1, 1], 
      Hue[0.0492169008088263, 1, 1], 
      Hue[0.04953441038264688, 1, 1], 
      Hue[0.05016942953028804, 1, 1], 
      Hue[0.05143946782557036, 1, 1], 
      Hue[0.05175697739939094, 1, 1], 
      Hue[0.05207448697321152, 1, 1], 
      Hue[0.05270950612085268, 1, 1], 
      Hue[0.05397954441613492, 1, 1], 
      Hue[0.054297053989955435`, 1, 1], 
      Hue[0.05461456356377601, 1, 1], 
      Hue[0.05524958271141718, 1, 1], 
      Hue[0.05651962100669947, 1, 1], 
      Hue[0.05683713058052007, 1, 1], 
      Hue[0.05715464015434064, 1, 1], 
      Hue[0.057789659301981794`, 1, 1], 
      Hue[0.059059697597264114`, 1, 1], 
      Hue[0.061599774187828665`, 1, 1], 
      Hue[0.061896242940041334`, 1, 1], 
      Hue[0.062192711692253905`, 1, 1], 
      Hue[0.06278564919667907, 1, 1], 
      Hue[0.0639715242055294, 1, 1], 
      Hue[0.06634327422323016, 1, 1], 
      Hue[0.06663974297544276, 1, 1], 
      Hue[0.06693621172765533, 1, 1], 
      Hue[0.06752914923208049, 1, 1], 
      Hue[0.06871502424093089, 1, 1], 
      Hue[0.0710867742586316, 1, 1], 
      Hue[0.07138324301084417, 1, 1], 
      Hue[0.07167971176305678, 1, 1], 
      Hue[0.0722726492674819, 1, 1], 
      Hue[0.0734585242763323, 1, 1], 
      Hue[0.07583027429403302, 1, 1], 
      Hue[0.0761267430462456, 1, 1], 
      Hue[0.07642321179845818, 1, 1], 
      Hue[0.07701614930288335, 1, 1], 
      Hue[0.07820202431173374, 1, 1], 
      Hue[0.08057377432943442, 1, 1], 
      Hue[0.0808644275415848, 1, 1], 
      Hue[0.08115508075373519, 1, 1], 
      Hue[0.08173638717803601, 1, 1], 
      Hue[0.08202704039018638, 1, 1], 
      Hue[0.08231769360233682, 1, 1], 
      Hue[0.08260834681448721, 1, 1], 
      Hue[0.08289900002663757, 1, 1], 
      Hue[0.08318965323878796, 1, 1], 
      Hue[0.5834803064509383, 1, 1], 
      Hue[0.5837709596630887, 1, 1], 
      Hue[0.5840616128752392, 1, 1], 
      Hue[0.5843522660873894, 1, 1], 
      Hue[0.5846429192995399, 1, 1], 
      Hue[0.5852242257238407, 1, 1], 
      Hue[0.585514878935991, 1, 1], 
      Hue[0.5858055321481415, 1, 1], 
      Hue[0.5863868385724422, 1, 1], 
      Hue[0.5875494514210438, 1, 1], 
      Hue[0.5898746771182468, 1, 1], 
      Hue[0.5901653303303972, 1, 1], 
      Hue[0.5904559835425476, 1, 1], 
      Hue[0.5910372899668483, 1, 1], 
      Hue[0.5921999028154499, 1, 1], 
      Hue[0.594525128512653, 1, 1], 
      Hue[0.5991755799070593, 1, 1], 
      Hue[0.5994908681422391, 1, 1], 
      Hue[0.5998061563774189, 1, 1], 
      Hue[0.6004367328477782, 1, 1], 
      Hue[0.6016978857884973, 1, 1], 
      Hue[0.6042201916699352, 1, 1], 
      Hue[0.6092648034328112, 1, 1], 
      Hue[0.6193540269585631, 1, 1], 
      Hue[0.6381858614271563, 1, 1], 
      Hue[0.6385047438637573, 1, 1], 
      Hue[0.6388236263003584, 1, 1], 
      Hue[0.6394613911735607, 1, 1], 
      Hue[0.6407369209199651, 1, 1], 
      Hue[0.6432879804127742, 1, 1], 
      Hue[0.6483900993983922, 1, 1], 
      Hue[0.6487089818349934, 1, 1], 
      Hue[0.6490278642715945, 1, 1], 
      Hue[0.6496656291447966, 1, 1], 
      Hue[0.6509411588912013, 1, 1], 
      Hue[0.6534922183840102, 1, 1], 
      Hue[0.6585943373696282, 1, 1], 
      Hue[0.6589074042661671, 1, 1], 
      Hue[0.6592204711627061, 1, 1], 
      Hue[0.6598466049557838, 1, 1], 
      Hue[0.6610988725419396, 1, 1], 
      Hue[0.663603407714251, 1, 1], 
      Hue[0.66391647461079, 1, 1], 
      Hue[0.6642295415073289, 1, 1], 
      Hue[0.6648556753004067, 1, 1], 
      Hue[0.6661079428865624, 1, 1], 
      Hue[0.6686124780588738, 1, 1], 
      Hue[0.6689255449554127, 1, 1], 
      Hue[0.6692386118519515, 1, 1], 
      Hue[0.6698647456450295, 1, 1], 
      Hue[0.6711170132311852, 1, 1], 
      Hue[0.6714300801277242, 1, 1], 
      Hue[0.6717431470242632, 1, 1], 
      Hue[0.6723692808173409, 1, 1], 
      Hue[0.6736215484034966, 1, 1], 
      Hue[0.6739346153000356, 1, 1], 
      Hue[0.6742476821965744, 1, 1], 
      Hue[0.6748738159896523, 1, 1], 
      Hue[0.6761260835758081, 1, 1], 
      Hue[0.6786306187481195, 1, 1], 
      Hue[0.6789226448230503, 1, 1], 
      Hue[0.6792146708979814, 1, 1], 
      Hue[0.6797987230478432, 1, 1], 
      Hue[0.680966827347567, 1, 1], 
      Hue[0.6812588534224979, 1, 1], 
      Hue[0.6815508794974288, 1, 1], 
      Hue[0.6821349316472907, 1, 1], 
      Hue[0.6833030359470145, 1, 1], 
      Hue[0.6835950620219453, 1, 1], 
      Hue[0.6838870880968764, 1, 1], 
      Hue[0.6844711402467383, 1, 1], 
      Hue[0.6847631663216691, 1, 1], 
      Hue[0.6850551923966001, 1, 1], 
      Hue[0.685639244546462, 1, 1], 
      Hue[0.6859312706213929, 1, 1], 
      Hue[0.6862232966963239, 1, 1], 
      Hue[0.6865153227712548, 1, 1], 
      Hue[0.6868073488461857, 1, 1], 
      Hue[0.6870993749211167, 1, 1], 
      Hue[0.6873914009960476, 1, 1], 
      Hue[0.6876834270709786, 1, 1], 
      Hue[0.6879754531459095, 1, 1], 
      Hue[0.6882674792208405, 1, 1], 
      Hue[0.6885595052957714, 1, 1], 
      Hue[0.6888515313707023, 1, 1], 
      Hue[0.6891435574456333, 1, 1], 
      Hue[0.6894355835205642, 1, 1], 
      Hue[0.6897276095954952, 1, 1], 
      Hue[0.6900196356704261, 1, 1], 
      Hue[0.6903116617453571, 1, 1], 
      Hue[0.6906036878202879, 1, 1], 
      Hue[0.690895713895219, 1, 1], 
      Hue[0.6911877399701498, 1, 1], 
      Hue[0.6914797660450808, 1, 1], 
      Hue[0.6917717921200117, 1, 1], 
      Hue[0.6920638181949427, 1, 1], 
      Hue[0.6923558442698736, 1, 1], 
      Hue[0.6926478703448045, 1, 1], 
      Hue[0.6929398964197355, 1, 1], 
      Hue[0.6932319224946664, 1, 1], 
      Hue[0.6935239485695974, 1, 1], 
      Hue[0.6938159746445283, 1, 1], 
      Hue[0.6941080007194593, 1, 1], 
      Hue[0.6944000267943902, 1, 1], 
      Hue[0.6949840789442521, 1, 1], 
      Hue[0.695276105019183, 1, 1], 
      Hue[0.695568131094114, 1, 1], 
      Hue[0.6961521832439759, 1, 1], 
      Hue[0.6964442093189067, 1, 1], 
      Hue[0.6967362353938378, 1, 1], 
      Hue[0.6973202875436997, 1, 1], 
      Hue[0.6976369486416599, 1, 1], 
      Hue[0.6979536097396202, 1, 1], 
      Hue[0.6985869319355408, 1, 1], 
      Hue[0.698903593033501, 1, 1], 
      Hue[0.6992202541314614, 1, 1], 
      Hue[0.6998535763273821, 1, 1], 
      Hue[0.7001702374253423, 1, 1], 
      Hue[0.7004868985233026, 1, 1], 
      Hue[0.7011202207192233, 1, 1], 
      Hue[0.7023868651110644, 1, 1], 
      Hue[0.7027035262090247, 1, 1], 
      Hue[0.703020187306985, 1, 1], 
      Hue[0.7036535095029056, 1, 1], 
      Hue[0.7049201538947468, 1, 1], 
      Hue[0.7052368149927071, 1, 1], 
      Hue[0.7055534760906674, 1, 1], 
      Hue[0.706186798286588, 1, 1], 
      Hue[0.7074534426784292, 1, 1], 
      Hue[0.7077701037763895, 1, 1], 
      Hue[0.7080867648743498, 1, 1], 
      Hue[0.7087200870702703, 1, 1], 
      Hue[0.7099867314621116, 1, 1], 
      Hue[0.7103033925600719, 1, 1], 
      Hue[0.7106200536580323, 1, 1], 
      Hue[0.7112533758539528, 1, 1], 
      Hue[0.712520020245794, 1, 1], 
      Hue[0.7128366813437543, 1, 1], 
      Hue[0.7131533424417147, 1, 1], 
      Hue[0.7137866646376352, 1, 1], 
      Hue[0.7150533090294764, 1, 1], 
      Hue[0.7175865978131588, 1, 1], 
      Hue[0.7178822180895112, 1, 1], 
      Hue[0.7181778383658635, 1, 1], 
      Hue[0.7187690789185681, 1, 1], 
      Hue[0.7199515600239774, 1, 1], 
      Hue[0.7223165222347959, 1, 1], 
      Hue[0.7226121425111482, 1, 1], 
      Hue[0.7229077627875006, 1, 1], 
      Hue[0.7234990033402051, 1, 1], 
      Hue[0.7246814844456144, 1, 1], 
      Hue[0.727046446656433, 1, 1], 
      Hue[0.7273420669327854, 1, 1], 
      Hue[0.7276376872091377, 1, 1], 
      Hue[0.7282289277618423, 1, 1], 
      Hue[0.7294114088672516, 1, 1], 
      Hue[0.7317763710780701, 1, 1], 
      Hue[0.7365062954997073, 1, 1], 
      Hue[0.7367961002359974, 1, 1], 
      Hue[0.7370859049722875, 1, 1], 
      Hue[0.7376655144448677, 1, 1], 
      Hue[0.7388247333900282, 1, 1], 
      Hue[0.7411431712803491, 1, 1], 
      Hue[0.7414329760166392, 1, 1], 
      Hue[0.7417227807529294, 1, 1], 
      Hue[0.7423023902255096, 1, 1], 
      Hue[0.7434616091706701, 1, 1], 
      Hue[0.745780047060991, 1, 1], 
      Hue[0.746069851797281, 1, 1], 
      Hue[0.7463596565335713, 1, 1], 
      Hue[0.7469392660061515, 1, 1], 
      Hue[0.7472290707424416, 1, 1], 
      Hue[0.7475188754787316, 1, 1], 
      Hue[0.7480984849513119, 1, 1], 
      Hue[0.748388289687602, 1, 1], 
      Hue[0.7486780944238922, 1, 1], 
      Hue[0.7489678991601822, 1, 1], 
      Hue[0.7492577038964724, 1, 1], 
      Hue[0.7495475086327625, 1, 1], 
      Hue[0.7498373133690526, 1, 1], 
      Hue[0.25012711810534277`, 1, 1], 
      Hue[0.25041692284163286`, 1, 1], 
      Hue[0.250706727577923, 1, 1], 
      Hue[0.25099653231421315`, 1, 1], 
      Hue[0.25157614178679333`, 1, 1], 
      Hue[0.2527353607319538, 1, 1], 
      Hue[0.2550537986222748, 1, 1], 
      Hue[0.2553682383815943, 1, 1], 
      Hue[0.2556826781409137, 1, 1], 
      Hue[0.25631155765955266`, 1, 1], 
      Hue[0.2575693166968306, 1, 1], 
      Hue[0.2600848347713864, 1, 1], 
      Hue[0.26511587092049804`, 1, 1], 
      Hue[0.27517794321872124`, 1, 1], 
      Hue[0.2939554752322569, 1, 1], 
      Hue[0.2942735091929977, 1, 1], 
      Hue[0.2945915431537386, 1, 1], 
      Hue[0.2952276110752203, 1, 1], 
      Hue[0.29649974691818365`, 1, 1], 
      Hue[0.29904401860411056`, 1, 1], 
      Hue[0.3041325619759642, 1, 1], 
      Hue[0.30445059593670504`, 1, 1], 
      Hue[0.30476862989744585`, 1, 1], 
      Hue[0.3054046978189276, 1, 1], 
      Hue[0.306676833661891, 1, 1], 
      Hue[0.3092211053478179, 1, 1], 
      Hue[0.31430964871967154`, 1, 1], 
      Hue[0.3146218671403502, 1, 1], 
      Hue[0.3149340855610288, 1, 1], 
      Hue[0.31555852240238613`, 1, 1], 
      Hue[0.3168073960851008, 1, 1], 
      Hue[0.31930514345052996`, 1, 1], 
      Hue[0.3196173618712086, 1, 1], 
      Hue[0.3199295802918873, 1, 1], 
      Hue[0.32055401713324455`, 1, 1], 
      Hue[0.3218028908159592, 1, 1], 
      Hue[0.3243006381813884, 1, 1], 
      Hue[0.324612856602067, 1, 1], 
      Hue[0.32492507502274565`, 1, 1], 
      Hue[0.32554951186410297`, 1, 1], 
      Hue[0.32679838554681756`, 1, 1], 
      Hue[0.3271106039674962, 1, 1], 
      Hue[0.3274228223881749, 1, 1], 
      Hue[0.32804725922953215`, 1, 1], 
      Hue[0.3292961329122468, 1, 1], 
      Hue[0.32960835133292543`, 1, 1], 
      Hue[0.3299205697536041, 1, 1], 
      Hue[0.3305450065949614, 1, 1], 
      Hue[0.331793880277676, 1, 1], 
      Hue[0.3342916276431052, 1, 1], 
      Hue[0.33458280524217593`, 1, 1], 
      Hue[0.3348739828412466, 1, 1], 
      Hue[0.3354563380393879, 1, 1], 
      Hue[0.3366210484356706, 1, 1], 
      Hue[0.3369122260347413, 1, 1], 
      Hue[0.33720340363381196`, 1, 1], 
      Hue[0.3377857588319533, 1, 1], 
      Hue[0.33895046922823596`, 1, 1], 
      Hue[0.3392416468273066, 1, 1], 
      Hue[0.33953282442637733`, 1, 1], 
      Hue[0.34011517962451865`, 1, 1], 
      Hue[0.3404063572235893, 1, 1], 
      Hue[0.34069753482266, 1, 1], 
      Hue[0.34127989002080134`, 1, 1], 
      Hue[0.341571067619872, 1, 1], 
      Hue[0.34186224521894265`, 1, 1], 
      Hue[0.34215342281801336`, 1, 1], 
      Hue[0.34244460041708397`, 1, 1], 
      Hue[0.3427357780161547, 1, 1], 
      Hue[0.34302695561522534`, 1, 1], 
      Hue[0.34360931081336665`, 1, 1], 
      Hue[0.34390048841243737`, 1, 1], 
      Hue[0.344191666011508, 1, 1], 
      Hue[0.34448284361057874`, 1, 1], 
      Hue[0.34477402120964934`, 1, 1], 
      Hue[0.34506519880872005`, 1, 1], 
      Hue[0.3453563764077907, 1, 1], 
      Hue[0.3456475540068614, 1, 1], 
      Hue[0.345938731605932, 1, 1], 
      Hue[0.3462299092050027, 1, 1], 
      Hue[0.3465210868040734, 1, 1], 
      Hue[0.346812264403144, 1, 1], 
      Hue[0.3471034420022147, 1, 1], 
      Hue[0.34739461960128537`, 1, 1], 
      Hue[0.3476857972003561, 1, 1], 
      Hue[0.34797697479942674`, 1, 1], 
      Hue[0.3482681523984974, 1, 1], 
      Hue[0.34855932999756806`, 1, 1], 
      Hue[0.34885050759663877`, 1, 1], 
      Hue[0.34914168519570943`, 1, 1], 
      Hue[0.3494328627947801, 1, 1], 
      Hue[0.3497240403938508, 1, 1], 
      Hue[0.35001521799292146`, 1, 1], 
      Hue[0.3503063955919921, 1, 1], 
      Hue[0.3505975731910628, 1, 1], 
      Hue[0.35088875079013343`, 1, 1], 
      Hue[0.3511799283892041, 1, 1], 
      Hue[0.3514711059882748, 1, 1], 
      Hue[0.3517622835873454, 1, 1], 
      Hue[0.3520534611864161, 1, 1], 
      Hue[0.3523446387854868, 1, 1], 
      Hue[0.3529269939836281, 1, 1], 
      Hue[0.3532428066057282, 1, 1], 
      Hue[0.3535586192278282, 1, 1], 
      Hue[0.35419024447202824`, 1, 1], 
      Hue[0.3545060570941283, 1, 1], 
      Hue[0.3548218697162283, 1, 1], 
      Hue[0.35545349496042833`, 1, 1], 
      Hue[0.35576930758252845`, 1, 1], 
      Hue[0.35608512020462846`, 1, 1], 
      Hue[0.3567167454488285, 1, 1], 
      Hue[0.35797999593722857`, 1, 1], 
      Hue[0.3582958085593287, 1, 1], 
      Hue[0.3586116211814287, 1, 1], 
      Hue[0.3592432464256287, 1, 1], 
      Hue[0.3605064969140288, 1, 1], 
      Hue[0.3608223095361289, 1, 1], 
      Hue[0.36113812215822894`, 1, 1], 
      Hue[0.36176974740242895`, 1, 1], 
      Hue[0.36303299789082905`, 1, 1], 
      Hue[0.3633488105129291, 1, 1], 
      Hue[0.3636646231350291, 1, 1], 
      Hue[0.3642962483792292, 1, 1], 
      Hue[0.36555949886762934`, 1, 1], 
      Hue[0.36587531148972935`, 1, 1], 
      Hue[0.36619112411182936`, 1, 1], 
      Hue[0.36682274935602943`, 1, 1], 
      Hue[0.3680859998444296, 1, 1], 
      Hue[0.3684018124665296, 1, 1], 
      Hue[0.3687176250886296, 1, 1], 
      Hue[0.36934925033282967`, 1, 1], 
      Hue[0.3706125008212298, 1, 1], 
      Hue[0.37313900179803006`, 1, 1], 
      Hue[0.37343377359852203`, 1, 1], 
      Hue[0.3737285453990141, 1, 1], 
      Hue[0.37431808899999824`, 1, 1], 
      Hue[0.37549717620196643`, 1, 1], 
      Hue[0.37785535060590275`, 1, 1], 
      Hue[0.37815012240639484`, 1, 1], 
      Hue[0.37844489420688693`, 1, 1], 
      Hue[0.37903443780787094`, 1, 1], 
      Hue[0.3802135250098391, 1, 1], 
      Hue[0.38257169941377556`, 1, 1], 
      Hue[0.38286647121426753`, 1, 1], 
      Hue[0.3831612430147596, 1, 1], 
      Hue[0.38375078661574374`, 1, 1], 
      Hue[0.38492987381771193`, 1, 1], 
      Hue[0.38728804822164825`, 1, 1], 
      Hue[0.39200439702952106`, 1, 1], 
      Hue[0.39232380385304244`, 1, 1], 
      Hue[0.3926432106765638, 1, 1], 
      Hue[0.3932820243236067, 1, 1], 
      Hue[0.3945596516176923, 1, 1], 
      Hue[0.39711490620586354`, 1, 1], 
      Hue[0.4022254153822061, 1, 1], 
      Hue[0.40254482220572746`, 1, 1], 
      Hue[0.40286422902924884`, 1, 1], 
      Hue[0.4035030426762917, 1, 1], 
      Hue[0.4047806699703773, 1, 1], 
      Hue[0.4073359245585486, 1, 1], 
      Hue[0.40765533138207, 1, 1], 
      Hue[0.4079747382055914, 1, 1], 
      Hue[0.4086135518526342, 1, 1], 
      Hue[0.40989117914671985`, 1, 1], 
      Hue[0.4124464337348911, 1, 1], 
      Hue[0.4127600250183503, 1, 1], 
      Hue[0.4130736163018095, 1, 1], 
      Hue[0.4137007988687279, 1, 1], 
      Hue[0.4140143901521871, 1, 1], 
      Hue[0.41432798143564636`, 1, 1], 
      Hue[0.4149551640025647, 1, 1], 
      Hue[0.41526875528602397`, 1, 1], 
      Hue[0.4155823465694832, 1, 1], 
      Hue[0.4158959378529423, 1, 1], 
      Hue[0.4162095291364016, 1, 1], 
      Hue[0.4165231204198608, 1, 1], 
      Hue[0.9168367117033199, 1, 1], 
      Hue[0.9171503029867791, 1, 1], 
      Hue[0.9174638942702383, 1, 1], 
      Hue[0.9177774855536975, 1, 1], 
      Hue[0.9180910768371567, 1, 1], 
      Hue[0.9187182594040751, 1, 1], 
      Hue[0.919972624537912, 1, 1], 
      Hue[0.9224813548055857, 1, 1], 
      Hue[0.9227949460890449, 1, 1], 
      Hue[0.9231085373725041, 1, 1], 
      Hue[0.9237357199394225, 1, 1], 
      Hue[0.9249900850732593, 1, 1], 
      Hue[0.9274988153409329, 1, 1], 
      Hue[0.9325162758762802, 1, 1], 
      Hue[0.9328088263381314, 1, 1], 
      Hue[0.9331013767999826, 1, 1], 
      Hue[0.9336864777236851, 1, 1], 
      Hue[0.93485667957109, 1, 1], 
      Hue[0.9371970832658998, 1, 1], 
      Hue[0.9418778906555193, 1, 1], 
      Hue[0.9421704411173706, 1, 1], 
      Hue[0.9424629915792218, 1, 1], 
      Hue[0.9430480925029242, 1, 1], 
      Hue[0.9442182943503291, 1, 1], 
      Hue[0.9465586980451389, 1, 1], 
      Hue[0.9512395054347584, 1, 1], 
      Hue[0.951556690919639, 1, 1], 
      Hue[0.9518738764045196, 1, 1], 
      Hue[0.9525082473742809, 1, 1], 
      Hue[0.9537769893138032, 1, 1], 
      Hue[0.9563144731928478, 1, 1], 
      Hue[0.9613894409509371, 1, 1], 
      Hue[0.9617066264358177, 1, 1], 
      Hue[0.9620238119206983, 1, 1], 
      Hue[0.9626581828904595, 1, 1], 
      Hue[0.9639269248299818, 1, 1], 
      Hue[0.9664644087090264, 1, 1], 
      Hue[0.9715393764671157, 1, 1], 
      Hue[0.9718355211303884, 1, 1], 
      Hue[0.972131665793661, 1, 1], 
      Hue[0.9727239551202062, 1, 1], 
      Hue[0.9739085337732966, 1, 1], 
      Hue[0.9762776910794774, 1, 1], 
      Hue[0.9765738357427499, 1, 1], 
      Hue[0.9768699804060226, 1, 1], 
      Hue[0.9774622697325678, 1, 1], 
      Hue[0.9786468483856582, 1, 1], 
      Hue[0.981016005691839, 1, 1], 
      Hue[0.9813121503551115, 1, 1], 
      Hue[0.9816082950183842, 1, 1], 
      Hue[0.9822005843449294, 1, 1], 
      Hue[0.9833851629980198, 1, 1], 
      Hue[0.9836813076612924, 1, 1], 
      Hue[0.983977452324565, 1, 1], 
      Hue[0.9845697416511101, 1, 1], 
      Hue[0.9857543203042005, 1, 1], 
      Hue[0.9860504649674732, 1, 1], 
      Hue[0.9863466096307457, 1, 1], 
      Hue[0.9869388989572909, 1, 1], 
      Hue[0.9881234776103813, 1, 1], 
      Hue[0.988419622273654, 1, 1], 
      Hue[0.9887157669369265, 1, 1], 
      Hue[0.9893080562634717, 1, 1], 
      Hue[0.9904926349165621, 1, 1], 
      Hue[0.9907829640397726, 1, 1], 
      Hue[0.991073293162983, 1, 1], 
      Hue[0.9916539514094037, 1, 1], 
      Hue[0.9919442805326142, 1, 1], 
      Hue[0.9922346096558246, 1, 1], 
      Hue[0.9928152679022454, 1, 1], 
      Hue[0.9931055970254558, 1, 1], 
      Hue[0.9933959261486661, 1, 1], 
      Hue[0.993976584395087, 1, 1], 
      Hue[0.9951379008879285, 1, 1], 
      Hue[0.9954282300111389, 1, 1], 
      Hue[0.9957185591343494, 1, 1], 
      Hue[0.9962992173807701, 1, 1], 
      Hue[0.9965895465039806, 1, 1], 
      Hue[0.9968798756271909, 1, 1], 
      Hue[0.9974605338736118, 1, 1], 
      Hue[0.9977508629968221, 1, 1], 
      Hue[0.9980411921200325, 1, 1], 
      Hue[0.9983315212432429, 1, 1], 
      Hue[0.9986218503664533, 1, 1], 
      Hue[0.9989121794896637, 1, 1], 
      Hue[0.9992025086128741, 1, 1], 
      Hue[0.9994928377360845, 1, 1], 
      Hue[0.9997831668592949, 1, 1], 
      Hue[0.00007349598250531426, 1, 1], 
      Hue[0.00036382510571571257`, 1, 1], 
      Hue[0.000654154228926111, 1, 1], 
      Hue[0.0009444833521365094, 1, 1], 
      Hue[0.0012348124753469077`, 1, 1], 
      Hue[0.0015251415985573066`, 1, 1], 
      Hue[0.0018154707217677046`, 1, 1], 
      Hue[0.002105799844978103, 1, 1], 
      Hue[0.0023961289681885016`, 1, 1], 
      Hue[0.0026864580913989, 1, 1], 
      Hue[0.0029767872146092976`, 1, 1], 
      Hue[0.0032671163378196967`, 1, 1], 
      Hue[0.003557445461030095, 1, 1], 
      Hue[0.003847774584240493, 1, 1], 
      Hue[0.004138103707450892, 1, 1], 
      Hue[0.004428432830661289, 1, 1], 
      Hue[0.0047187619538716875`, 1, 1], 
      Hue[0.005009091077082086, 1, 1], 
      Hue[0.005299420200292486, 1, 1], 
      Hue[0.005589749323502882, 1, 1], 
      Hue[0.005880078446713281, 1, 1], 
      Hue[0.0061704075699236804`, 1, 1], 
      Hue[0.006751065816344477, 1, 1], 
      Hue[0.007041394939554876, 1, 1], 
      Hue[0.007331724062765272, 1, 1], 
      Hue[0.00791238230918607, 1, 1], 
      Hue[0.008202711432396469, 1, 1], 
      Hue[0.008493040555606868, 1, 1], 
      Hue[0.009073698802027665, 1, 1], 
      Hue[0.009388662948267424, 1, 1], 
      Hue[0.00970362709450718, 1, 1], 
      Hue[0.0103335553869867, 1, 1], 
      Hue[0.010648519533226455`, 1, 1], 
      Hue[0.010963483679466215`, 1, 1], 
      Hue[0.011593411971945728`, 1, 1], 
      Hue[0.011908376118185491`, 1, 1], 
      Hue[0.012223340264425247`, 1, 1], 
      Hue[0.012853268556904763`, 1, 1], 
      Hue[0.014113125141863794`, 1, 1], 
      Hue[0.014428089288103553`, 1, 1], 
      Hue[0.014743053434343314`, 1, 1], 
      Hue[0.01537298172682283, 1, 1], 
      Hue[0.01663283831178186, 1, 1], 
      Hue[0.016947802458021622`, 1, 1], 
      Hue[0.017262766604261383`, 1, 1], 
      Hue[0.017892694896740897`, 1, 1], 
      Hue[0.019152551481699927`, 1, 1], 
      Hue[0.019467515627939688`, 1, 1], 
      Hue[0.019782479774179445`, 1, 1], 
      Hue[0.020412408066658963`, 1, 1], 
      Hue[0.02167226465161799, 1, 1], 
      Hue[0.02419197782153607, 1, 1], 
      Hue[0.02450694196777583, 1, 1], 
      Hue[0.02482190611401558, 1, 1], 
      Hue[0.0254518344064951, 1, 1], 
      Hue[0.026711690991454134`, 1, 1], 
      Hue[0.0292314041613722, 1, 1], 
      Hue[0.029525327486003974`, 1, 1], 
      Hue[0.02981925081063575, 1, 1], 
      Hue[0.030407097459899297`, 1, 1], 
      Hue[0.03158279075842641, 1, 1], 
      Hue[0.033934177355480606`, 1, 1], 
      Hue[0.03863695054958902, 1, 1], 
      Hue[0.0389308738742208, 1, 1], 
      Hue[0.03922479719885257, 1, 1], 
      Hue[0.03981264384811612, 1, 1], 
      Hue[0.04098833714664322, 1, 1], 
      Hue[0.043339723743697434`, 1, 1], 
      Hue[0.048042496937805844`, 1, 1], 
      Hue[0.048361055285466976`, 1, 1], 
      Hue[0.048679613633128116`, 1, 1], 
      Hue[0.049316730328450395`, 1, 1], 
      Hue[0.05059096371909494, 1, 1], 
      Hue[0.05313943050038401, 1, 1], 
      Hue[0.058236364062962176`, 1, 1], 
      Hue[0.05855492241062331, 1, 1], 
      Hue[0.05887348075828444, 1, 1], 
      Hue[0.05951059745360673, 1, 1], 
      Hue[0.06078483084425125, 1, 1], 
      Hue[0.06333329762554035, 1, 1], 
      Hue[0.06843023118811851, 1, 1], 
      Hue[0.06874297399571745, 1, 1], 
      Hue[0.06905571680331637, 1, 1], 
      Hue[0.06968120241851426, 1, 1], 
      Hue[0.07093217364891, 1, 1], 
      Hue[0.07343411610970148, 1, 1], 
      Hue[0.0737468589173004, 1, 1], 
      Hue[0.07405960172489932, 1, 1], 
      Hue[0.07468508734009718, 1, 1], 
      Hue[0.07593605857049295, 1, 1], 
      Hue[0.0784380010312844, 1, 1], 
      Hue[0.07875074383888335, 1, 1], 
      Hue[0.07906348664648226, 1, 1], 
      Hue[0.07968897226168013, 1, 1], 
      Hue[0.08093994349207585, 1, 1], 
      Hue[0.08125268629967479, 1, 1], 
      Hue[0.08156542910727374, 1, 1], 
      Hue[0.08187817191487268, 1, 1], 
      Hue[0.08219091472247161, 1, 1], 
      Hue[0.08250365753007054, 1, 1], 
      Hue[0.0828164003376695, 1, 1], 
      Hue[0.08312914314526841, 1, 1], 
      Hue[0.5834418859528673, 1, 1], 
      Hue[0.5837546287604662, 1, 1], 
      Hue[0.5840673715680651, 1, 1], 
      Hue[0.584692857183263, 1, 1], 
      Hue[0.5859438284136588, 1, 1], 
      Hue[0.5884457708744503, 1, 1], 
      Hue[0.5887374728604412, 1, 1], 
      Hue[0.5890291748464322, 1, 1], 
      Hue[0.5896125788184141, 1, 1], 
      Hue[0.5907793867623778, 1, 1], 
      Hue[0.5931130026503055, 1, 1], 
      Hue[0.5977802344261607, 1, 1], 
      Hue[0.5980719364121516, 1, 1], 
      Hue[0.5983636383981427, 1, 1], 
      Hue[0.5989470423701245, 1, 1], 
      Hue[0.6001138503140884, 1, 1], 
      Hue[0.602447466202016, 1, 1], 
      Hue[0.6071146979778712, 1, 1], 
      Hue[0.6074310349868914, 1, 1], 
      Hue[0.6077473719959118, 1, 1], 
      Hue[0.6083800460139523, 1, 1], 
      Hue[0.6096453940500337, 1, 1], 
      Hue[0.6121760901221962, 1, 1], 
      Hue[0.6172374822665212, 1, 1], 
      Hue[0.6175538192755414, 1, 1], 
      Hue[0.6178701562845618, 1, 1], 
      Hue[0.6185028303026023, 1, 1], 
      Hue[0.6197681783386836, 1, 1], 
      Hue[0.6222988744108461, 1, 1], 
      Hue[0.6273602665551711, 1, 1], 
      Hue[0.6276555627425835, 1, 1], 
      Hue[0.6279508589299957, 1, 1], 
      Hue[0.6285414513048204, 1, 1], 
      Hue[0.6297226360544697, 1, 1], 
      Hue[0.6320850055537683, 1, 1], 
      Hue[0.6323803017411807, 1, 1], 
      Hue[0.632675597928593, 1, 1], 
      Hue[0.6332661903034177, 1, 1], 
      Hue[0.634447375053067, 1, 1], 
      Hue[0.6368097445523656, 1, 1], 
      Hue[0.637105040739778, 1, 1], 
      Hue[0.6374003369271903, 1, 1], 
      Hue[0.637990929302015, 1, 1], 
      Hue[0.6391721140516642, 1, 1], 
      Hue[0.6394674102390766, 1, 1], 
      Hue[0.6397627064264888, 1, 1], 
      Hue[0.6403532988013136, 1, 1], 
      Hue[0.6415344835509629, 1, 1], 
      Hue[0.6418297797383752, 1, 1], 
      Hue[0.6421250759257875, 1, 1], 
      Hue[0.6427156683006122, 1, 1], 
      Hue[0.6438968530502616, 1, 1], 
      Hue[0.6441921492376739, 1, 1], 
      Hue[0.6444874454250862, 1, 1], 
      Hue[0.6450780377999108, 1, 1], 
      Hue[0.64625922254956, 1, 1], 
      Hue[0.6465487031969103, 1, 1], 
      Hue[0.6468381838442604, 1, 1], 
      Hue[0.6474171451389605, 1, 1], 
      Hue[0.6477066257863107, 1, 1], 
      Hue[0.6479961064336609, 1, 1], 
      Hue[0.6485750677283612, 1, 1], 
      Hue[0.6488645483757112, 1, 1], 
      Hue[0.6491540290230614, 1, 1], 
      Hue[0.6497329903177617, 1, 1], 
      Hue[0.6508909129071622, 1, 1], 
      Hue[0.6511803935545123, 1, 1], 
      Hue[0.6514698742018624, 1, 1], 
      Hue[0.6520488354965627, 1, 1], 
      Hue[0.6523383161439128, 1, 1], 
      Hue[0.6526277967912629, 1, 1], 
      Hue[0.6532067580859631, 1, 1], 
      Hue[0.6534962387333133, 1, 1], 
      Hue[0.6537857193806634, 1, 1], 
      Hue[0.6540752000280136, 1, 1], 
      Hue[0.6543646806753636, 1, 1], 
      Hue[0.6546541613227137, 1, 1], 
      Hue[0.6549436419700639, 1, 1], 
      Hue[0.6552331226174141, 1, 1], 
      Hue[0.6555226032647642, 1, 1], 
      Hue[0.6558120839121143, 1, 1], 
      Hue[0.6561015645594643, 1, 1], 
      Hue[0.6563910452068146, 1, 1], 
      Hue[0.6566805258541647, 1, 1], 
      Hue[0.6569700065015148, 1, 1], 
      Hue[0.6572594871488651, 1, 1], 
      Hue[0.6575489677962151, 1, 1], 
      Hue[0.6578384484435652, 1, 1], 
      Hue[0.6581279290909153, 1, 1], 
      Hue[0.6584174097382655, 1, 1], 
      Hue[0.6587068903856157, 1, 1], 
      Hue[0.6589963710329658, 1, 1], 
      Hue[0.6592858516803158, 1, 1], 
      Hue[0.659575332327666, 1, 1], 
      Hue[0.6598648129750161, 1, 1], 
      Hue[0.6601542936223663, 1, 1], 
      Hue[0.6604437742697165, 1, 1], 
      Hue[0.6607332549170665, 1, 1], 
      Hue[0.6610227355644166, 1, 1], 
      Hue[0.6613122162117667, 1, 1], 
      Hue[0.661601696859117, 1, 1], 
      Hue[0.661891177506467, 1, 1], 
      Hue[0.6624701388011672, 1, 1], 
      Hue[0.6627596194485174, 1, 1], 
      Hue[0.6630491000958674, 1, 1], 
      Hue[0.6636280613905677, 1, 1], 
      Hue[0.6639175420379179, 1, 1], 
      Hue[0.664207022685268, 1, 1], 
      Hue[0.6647859839799682, 1, 1], 
      Hue[0.6651000996503478, 1, 1], 
      Hue[0.6654142153207272, 1, 1], 
      Hue[0.6660424466614863, 1, 1], 
      Hue[0.6663565623318657, 1, 1], 
      Hue[0.6666706780022451, 1, 1], 
      Hue[0.6672989093430042, 1, 1], 
      Hue[0.6676130250133836, 1, 1], 
      Hue[0.6679271406837631, 1, 1], 
      Hue[0.6685553720245221, 1, 1], 
      Hue[0.66981183470604, 1, 1], 
      Hue[0.6701259503764195, 1, 1], 
      Hue[0.6704400660467991, 1, 1], 
      Hue[0.671068297387558, 1, 1], 
      Hue[0.6723247600690759, 1, 1], 
      Hue[0.6748376854321119, 1, 1], 
      Hue[0.6751518011024913, 1, 1], 
      Hue[0.6754659167728708, 1, 1], 
      Hue[0.6760941481136298, 1, 1], 
      Hue[0.6773506107951477, 1, 1], 
      Hue[0.6798635361581836, 1, 1], 
      Hue[0.680177651828563, 1, 1], 
      Hue[0.6804917674989426, 1, 1], 
      Hue[0.6811199988397015, 1, 1], 
      Hue[0.6823764615212196, 1, 1], 
      Hue[0.6848893868842554, 1, 1], 
      Hue[0.6851824617330269, 1, 1], 
      Hue[0.6854755365817984, 1, 1], 
      Hue[0.6860616862793414, 1, 1], 
      Hue[0.6872339856744274, 1, 1], 
      Hue[0.6895785844645994, 1, 1], 
      Hue[0.6942677820449434, 1, 1], 
      Hue[0.6945608568937149, 1, 1], 
      Hue[0.6948539317424864, 1, 1], 
      Hue[0.6954400814400294, 1, 1], 
      Hue[0.6966123808351155, 1, 1], 
      Hue[0.6989569796252876, 1, 1], 
      Hue[0.7036461772056316, 1, 1], 
      Hue[0.7039638870774325, 1, 1], 
      Hue[0.7042815969492333, 1, 1], 
      Hue[0.704917016692835, 1, 1], 
      Hue[0.7061878561800385, 1, 1], 
      Hue[0.7087295351544455, 1, 1], 
      Hue[0.7138128931032592, 1, 1], 
      Hue[0.723979609000887, 1, 1], 
      Hue[0.7242762780510799, 1, 1], 
      Hue[0.7245729471012727, 1, 1], 
      Hue[0.7251662852016585, 1, 1], 
      Hue[0.72635296140243, 1, 1], 
      Hue[0.728726313803973, 1, 1], 
      Hue[0.7334730186070592, 1, 1], 
      Hue[0.7337696876572521, 1, 1], 
      Hue[0.7340663567074449, 1, 1], 
      Hue[0.7346596948078307, 1, 1], 
      Hue[0.7358463710086022, 1, 1], 
      Hue[0.7382197234101453, 1, 1], 
      Hue[0.7429664282132313, 1, 1], 
      Hue[0.743257281723362, 1, 1], 
      Hue[0.7435481352334927, 1, 1], 
      Hue[0.7441298422537541, 1, 1], 
      Hue[0.7452932562942768, 1, 1], 
      Hue[0.7455841098044075, 1, 1], 
      Hue[0.7458749633145382, 1, 1], 
      Hue[0.7464566703347995, 1, 1], 
      Hue[0.7476200843753222, 1, 1], 
      Hue[0.7479109378854529, 1, 1], 
      Hue[0.7482017913955836, 1, 1], 
      Hue[0.748783498415845, 1, 1], 
      Hue[0.7490743519259756, 1, 1], 
      Hue[0.7493652054361063, 1, 1], 
      Hue[0.749656058946237, 1, 1], 
      Hue[0.7499469124563677, 1, 1], 
      Hue[0.2502377659664984, 1, 1], 
      Hue[0.2505286194766291, 1, 1], 
      Hue[0.2508194729867598, 1, 1], 
      Hue[0.2511103264968904, 1, 1], 
      Hue[0.25227374053741314`, 1, 1], 
      Hue[0.25256459404754383`, 1, 1], 
      Hue[0.25285544755767453`, 1, 1], 
      Hue[0.25343715457793586`, 1, 1], 
      Hue[0.2546005686184586, 1, 1], 
      Hue[0.25692739669950404`, 1, 1], 
      Hue[0.2615810528615949, 1, 1], 
      Hue[0.26189654139475493`, 1, 1], 
      Hue[0.2622120299279149, 1, 1], 
      Hue[0.262843006994235, 1, 1], 
      Hue[0.2641049611268752, 1, 1], 
      Hue[0.2666288693921555, 1, 1], 
      Hue[0.26694435792531557`, 1, 1], 
      Hue[0.2672598464584756, 1, 1], 
      Hue[0.26789082352479565`, 1, 1], 
      Hue[0.26915277765743584`, 1, 1], 
      Hue[0.27167668592271615`, 1, 1], 
      Hue[0.2719921744558762, 1, 1], 
      Hue[0.2723076629890362, 1, 1], 
      Hue[0.2729386400553563, 1, 1], 
      Hue[0.27420059418799647`, 1, 1], 
      Hue[0.2767245024532768, 1, 1], 
      Hue[0.2817723189838374, 1, 1], 
      Hue[0.2820667666953895, 1, 1], 
      Hue[0.28236121440694156`, 1, 1], 
      Hue[0.28295010983004565`, 1, 1], 
      Hue[0.2841279006762539, 1, 1], 
      Hue[0.28648348236867033`, 1, 1], 
      Hue[0.2867779300802224, 1, 1], 
      Hue[0.28707237779177447`, 1, 1], 
      Hue[0.28766127321487855`, 1, 1], 
      Hue[0.2888390640610868, 1, 1], 
      Hue[0.29119464575350323`, 1, 1], 
      Hue[0.2914890934650553, 1, 1], 
      Hue[0.2917835411766074, 1, 1], 
      Hue[0.29237243659971146`, 1, 1], 
      Hue[0.2935502274459197, 1, 1], 
      Hue[0.29384467515747176`, 1, 1], 
      Hue[0.29413912286902383`, 1, 1], 
      Hue[0.2947280182921279, 1, 1], 
      Hue[0.29590580913833614`, 1, 1], 
      Hue[0.2962002568498882, 1, 1], 
      Hue[0.2964947045614402, 1, 1], 
      Hue[0.29708359998454437`, 1, 1], 
      Hue[0.2982613908307526, 1, 1], 
      Hue[0.29855583854230466`, 1, 1], 
      Hue[0.2988502862538567, 1, 1], 
      Hue[0.2994391816769608, 1, 1], 
      Hue[0.30061697252316905`, 1, 1], 
      Hue[0.30093605525775047`, 1, 1], 
      Hue[0.3012551379923319, 1, 1], 
      Hue[0.30189330346149473`, 1, 1], 
      Hue[0.30221238619607615`, 1, 1], 
      Hue[0.3025314689306576, 1, 1], 
      Hue[0.3031696343998204, 1, 1], 
      Hue[0.30348871713440184`, 1, 1], 
      Hue[0.3038077998689832, 1, 1], 
      Hue[0.30444596533814605`, 1, 1], 
      Hue[0.30572229627647174`, 1, 1], 
      Hue[0.30604137901105316`, 1, 1], 
      Hue[0.3063604617456346, 1, 1], 
      Hue[0.3069986272147974, 1, 1], 
      Hue[0.30731770994937885`, 1, 1], 
      Hue[0.3076367926839602, 1, 1], 
      Hue[0.30827495815312306`, 1, 1], 
      Hue[0.3085940408877045, 1, 1], 
      Hue[0.3089131236222859, 1, 1], 
      Hue[0.3092322063568673, 1, 1], 
      Hue[0.30955128909144874`, 1, 1], 
      Hue[0.30987037182603017`, 1, 1], 
      Hue[0.3101894545606116, 1, 1], 
      Hue[0.310508537295193, 1, 1], 
      Hue[0.31082762002977443`, 1, 1], 
      Hue[0.31114670276435585`, 1, 1], 
      Hue[0.3114657854989372, 1, 1], 
      Hue[0.3117848682335187, 1, 1], 
      Hue[0.31210395096810006`, 1, 1], 
      Hue[0.3124230337026815, 1, 1], 
      Hue[0.3127421164372629, 1, 1], 
      Hue[0.31306119917184433`, 1, 1], 
      Hue[0.31338028190642575`, 1, 1], 
      Hue[0.3136993646410072, 1, 1], 
      Hue[0.3140184473755886, 1, 1], 
      Hue[0.31433753011017, 1, 1], 
      Hue[0.31465661284475144`, 1, 1], 
      Hue[0.31497569557933286`, 1, 1], 
      Hue[0.3152947783139142, 1, 1], 
      Hue[0.3156138610484957, 1, 1], 
      Hue[0.31593294378307707`, 1, 1], 
      Hue[0.3162520265176585, 1, 1], 
      Hue[0.3165711092522399, 1, 1], 
      Hue[0.31689019198682133`, 1, 1], 
      Hue[0.31720927472140276`, 1, 1], 
      Hue[0.3175283574559842, 1, 1], 
      Hue[0.3178474401905656, 1, 1], 
      Hue[0.31848560565972844`, 1, 1], 
      Hue[0.31880468839430987`, 1, 1], 
      Hue[0.31912377112889123`, 1, 1], 
      Hue[0.31976193659805413`, 1, 1], 
      Hue[0.3200810193326355, 1, 1], 
      Hue[0.320400102067217, 1, 1], 
      Hue[0.32103826753637976`, 1, 1], 
      Hue[0.321351534730899, 1, 1], 
      Hue[0.32166480192541824`, 1, 1], 
      Hue[0.3222913363144566, 1, 1], 
      Hue[0.32354440509253346`, 1, 1], 
      Hue[0.3238576722870527, 1, 1], 
      Hue[0.32417093948157194`, 1, 1], 
      Hue[0.32479747387061036`, 1, 1], 
      Hue[0.3260505426486872, 1, 1], 
      Hue[0.32636380984320645`, 1, 1], 
      Hue[0.3266770770377257, 1, 1], 
      Hue[0.32730361142676406`, 1, 1], 
      Hue[0.32855668020484097`, 1, 1], 
      Hue[0.33106281776099467`, 1, 1], 
      Hue[0.3313760849555139, 1, 1], 
      Hue[0.33168935215003315`, 1, 1], 
      Hue[0.3323158865390715, 1, 1], 
      Hue[0.3335689553171484, 1, 1], 
      Hue[0.3360750928733021, 1, 1], 
      Hue[0.33638836006782136`, 1, 1], 
      Hue[0.3367016272623406, 1, 1], 
      Hue[0.337328161651379, 1, 1], 
      Hue[0.33858123042945587`, 1, 1], 
      Hue[0.34108736798560957`, 1, 1], 
      Hue[0.34137959435852083`, 1, 1], 
      Hue[0.34167182073143204`, 1, 1], 
      Hue[0.3422562734772545, 1, 1], 
      Hue[0.34342517896889946`, 1, 1], 
      Hue[0.34576298995218935`, 1, 1], 
      Hue[0.350438611918769, 1, 1], 
      Hue[0.35978985585192846`, 1, 1], 
      Hue[0.38006898519212645`, 1, 1], 
      Hue[0.38036480576645904`, 1, 1], 
      Hue[0.3806606263407917, 1, 1], 
      Hue[0.3812522674894569, 1, 1], 
      Hue[0.38243554978678734`, 1, 1], 
      Hue[0.38480211438144823`, 1, 1], 
      Hue[0.38953524357077, 1, 1], 
      Hue[0.38983106414510266`, 1, 1], 
      Hue[0.39012688471943524`, 1, 1], 
      Hue[0.39071852586810046`, 1, 1], 
      Hue[0.3919018081654309, 1, 1], 
      Hue[0.39426837276009186`, 1, 1], 
      Hue[0.39900150194941353`, 1, 1], 
      Hue[0.39929150698368404`, 1, 1], 
      Hue[0.39958151201795444`, 1, 1], 
      Hue[0.40016152208649525`, 1, 1], 
      Hue[0.40132154222357685`, 1, 1], 
      Hue[0.4036415824977401, 1, 1], 
      Hue[0.4039315875320105, 1, 1], 
      Hue[0.40422159256628093`, 1, 1], 
      Hue[0.40480160263482173`, 1, 1], 
      Hue[0.40596162277190334`, 1, 1], 
      Hue[0.40828166304606667`, 1, 1], 
      Hue[0.4085716680803371, 1, 1], 
      Hue[0.4088616731146075, 1, 1], 
      Hue[0.40944168318314833`, 1, 1], 
      Hue[0.41060170332022994`, 1, 1], 
      Hue[0.4108917083545004, 1, 1], 
      Hue[0.4111817133887708, 1, 1], 
      Hue[0.4117617234573116, 1, 1], 
      Hue[0.4129217435943932, 1, 1], 
      Hue[0.4132117486286636, 1, 1], 
      Hue[0.413501753662934, 1, 1], 
      Hue[0.4140817637314748, 1, 1], 
      Hue[0.4143717687657452, 1, 1], 
      Hue[0.4146617738000156, 1, 1], 
      Hue[0.4152417838685564, 1, 1], 
      Hue[0.4155317889028268, 1, 1], 
      Hue[0.41582179393709723`, 1, 1], 
      Hue[0.41611179897136763`, 1, 1], 
      Hue[0.4164018040056381, 1, 1], 
      Hue[0.9166918090399084, 1, 1], 
      Hue[0.9169818140741789, 1, 1], 
      Hue[0.9172718191084492, 1, 1], 
      Hue[0.9175618241427197, 1, 1], 
      Hue[0.9178764642000194, 1, 1], 
      Hue[0.9181911042573192, 1, 1], 
      Hue[0.9188203843719188, 1, 1], 
      Hue[0.9200789446011179, 1, 1], 
      Hue[0.922596065059516, 1, 1], 
      Hue[0.9229107051168157, 1, 1], 
      Hue[0.9232253451741156, 1, 1], 
      Hue[0.9238546252887151, 1, 1], 
      Hue[0.9251131855179142, 1, 1], 
      Hue[0.9276303059763124, 1, 1], 
      Hue[0.9279449460336121, 1, 1], 
      Hue[0.928259586090912, 1, 1], 
      Hue[0.9288888662055114, 1, 1], 
      Hue[0.9301474264347105, 1, 1], 
      Hue[0.9326645468931086, 1, 1], 
      Hue[0.9376987878099049, 1, 1], 
      Hue[0.9379923870455967, 1, 1], 
      Hue[0.9382859862812885, 1, 1], 
      Hue[0.938873184752672, 1, 1], 
      Hue[0.9400475816954391, 1, 1], 
      Hue[0.9423963755809734, 1, 1], 
      Hue[0.9426899748166653, 1, 1], 
      Hue[0.9429835740523571, 1, 1], 
      Hue[0.9435707725237407, 1, 1], 
      Hue[0.9447451694665078, 1, 1], 
      Hue[0.9470939633520421, 1, 1], 
      Hue[0.9473875625877338, 1, 1], 
      Hue[0.9476811618234257, 1, 1], 
      Hue[0.9482683602948092, 1, 1], 
      Hue[0.9494427572375763, 1, 1], 
      Hue[0.9497363564732682, 1, 1], 
      Hue[0.9500299557089599, 1, 1], 
      Hue[0.9506171541803434, 1, 1], 
      Hue[0.9517915511231106, 1, 1], 
      Hue[0.9520851503588025, 1, 1], 
      Hue[0.9523787495944943, 1, 1], 
      Hue[0.9529659480658779, 1, 1], 
      Hue[0.954140345008645, 1, 1], 
      Hue[0.9544339442443368, 1, 1], 
      Hue[0.9547275434800285, 1, 1], 
      Hue[0.9553147419514121, 1, 1], 
      Hue[0.9564891388941792, 1, 1], 
      Hue[0.9568073731529004, 1, 1], 
      Hue[0.9571256074116216, 1, 1], 
      Hue[0.9577620759290638, 1, 1], 
      Hue[0.9580803101877848, 1, 1], 
      Hue[0.9583985444465061, 1, 1], 
      Hue[0.9590350129639483, 1, 1], 
      Hue[0.9593532472226696, 1, 1], 
      Hue[0.9596714814813907, 1, 1], 
      Hue[0.960307949998833, 1, 1], 
      Hue[0.9615808870337176, 1, 1], 
      Hue[0.9618991212924386, 1, 1], 
      Hue[0.9622173555511598, 1, 1], 
      Hue[0.9628538240686021, 1, 1], 
      Hue[0.9631720583273232, 1, 1], 
      Hue[0.9634902925860445, 1, 1], 
      Hue[0.9641267611034867, 1, 1], 
      Hue[0.9644449953622078, 1, 1], 
      Hue[0.9647632296209291, 1, 1], 
      Hue[0.9650814638796502, 1, 1], 
      Hue[0.9653996981383713, 1, 1], 
      Hue[0.9657179323970924, 1, 1], 
      Hue[0.9660361666558135, 1, 1], 
      Hue[0.9666726351732559, 1, 1], 
      Hue[0.966990869431977, 1, 1], 
      Hue[0.9673091036906982, 1, 1], 
      Hue[0.9676273379494194, 1, 1], 
      Hue[0.9679455722081405, 1, 1], 
      Hue[0.9682638064668616, 1, 1], 
      Hue[0.9685820407255828, 1, 1], 
      Hue[0.968900274984304, 1, 1], 
      Hue[0.9692185092430251, 1, 1], 
      Hue[0.9695367435017461, 1, 1], 
      Hue[0.9698549777604673, 1, 1], 
      Hue[0.9701732120191885, 1, 1], 
      Hue[0.9704914462779096, 1, 1], 
      Hue[0.9708096805366307, 1, 1], 
      Hue[0.9711279147953519, 1, 1], 
      Hue[0.9714461490540731, 1, 1], 
      Hue[0.9717643833127942, 1, 1], 
      Hue[0.9720826175715154, 1, 1], 
      Hue[0.9724008518302366, 1, 1], 
      Hue[0.9727190860889577, 1, 1], 
      Hue[0.9730373203476788, 1, 1], 
      Hue[0.9733555546063999, 1, 1], 
      Hue[0.973673788865121, 1, 1], 
      Hue[0.9739920231238423, 1, 1], 
      Hue[0.9743102573825634, 1, 1], 
      Hue[0.9746284916412845, 1, 1], 
      Hue[0.9749467259000056, 1, 1], 
      Hue[0.9752649601587269, 1, 1], 
      Hue[0.975583194417448, 1, 1], 
      Hue[0.9759014286761691, 1, 1], 
      Hue[0.9762196629348903, 1, 1], 
      Hue[0.9768561314523325, 1, 1], 
      Hue[0.9768656420513168, 1, 1], 
      Hue[0.976875152650301, 1, 1], 
      Hue[0.9768941738482696, 1, 1], 
      Hue[0.9769322162442068, 1, 1], 
      Hue[0.9770083010360808, 1, 1], 
      Hue[0.9771604706198291, 1, 1], 
      Hue[0.9771699812188133, 1, 1], 
      Hue[0.9771794918177976, 1, 1], 
      Hue[0.9771985130157662, 1, 1], 
      Hue[0.9772365554117033, 1, 1], 
      Hue[0.9773126402035773, 1, 1], 
      Hue[0.9773221508025617, 1, 1], 
      Hue[0.9773316614015459, 1, 1], 
      Hue[0.9773506825995144, 1, 1], 
      Hue[0.9773887249954515, 1, 1], 
      Hue[0.9773982355944357, 1, 1], 
      Hue[0.9774077461934201, 1, 1], 
      Hue[0.9774267673913886, 1, 1], 
      Hue[0.9774362779903729, 1, 1], 
      Hue[0.9774457885893572, 1, 1], 
      Hue[0.9774552991883414, 1, 1], 
      Hue[0.9774648097873256, 1, 1], 
      Hue[0.022535190212674343`, 1, 1], 
      Hue[0.022828064763465546`, 1, 1], 
      Hue[0.022828064763465546`, 1, 1], 
      Hue[0.02312093931425682, 1, 1], 
      Hue[0.02312093931425682, 1, 1], 
      Hue[0.023706688415839228`, 1, 1], 
      Hue[0.023706688415839228`, 1, 1], 
      Hue[0.02399956296663043, 1, 1], 
      Hue[0.02399956296663043, 1, 1], 
      Hue[0.024292437517421633`, 1, 1], 
      Hue[0.024292437517421633`, 1, 1], 
      Hue[0.024878186619004036`, 1, 1], 
      Hue[0.024878186619004036`, 1, 1], 
      Hue[0.02517106116979524, 1, 1], 
      Hue[0.02517106116979524, 1, 1], 
      Hue[0.025463935720586448`, 1, 1], 
      Hue[0.025463935720586448`, 1, 1], 
      Hue[0.026049684822168923`, 1, 1], 
      Hue[0.026049684822168923`, 1, 1], 
      Hue[0.026342559372960126`, 1, 1], 
      Hue[0.026342559372960126`, 1, 1], 
      Hue[0.02663543392375139, 1, 1], 
      Hue[0.02663543392375139, 1, 1], 
      Hue[0.027221183025333804`, 1, 1], 
      Hue[0.027221183025333804`, 1, 1], 
      Hue[0.027514057576125007`, 1, 1], 
      Hue[0.027514057576125007`, 1, 1], 
      Hue[0.027806932126916206`, 1, 1], 
      Hue[0.027806932126916206`, 1, 1], 
      Hue[0.028099806677707416`, 1, 1], 
      Hue[0.028099806677707416`, 1, 1], 
      Hue[0.028392681228498615`, 1, 1], 
      Hue[0.028392681228498615`, 1, 1], 
      Hue[0.02868555577928982, 1, 1], 
      Hue[0.02868555577928982, 1, 1], 
      Hue[0.028978430330081024`, 1, 1], 
      Hue[0.028978430330081024`, 1, 1], 
      Hue[0.0292713048808723, 1, 1], 
      Hue[0.0292713048808723, 1, 1], 
      Hue[0.029564179431663503`, 1, 1], 
      Hue[0.029564179431663503`, 1, 1], 
      Hue[0.029857053982454702`, 1, 1], 
      Hue[0.029857053982454702`, 1, 1], 
      Hue[0.030149928533245978`, 1, 1], 
      Hue[0.030149928533245978`, 1, 1], 
      Hue[0.030442803084037184`, 1, 1], 
      Hue[0.030442803084037184`, 1, 1], 
      Hue[0.030735677634828387`, 1, 1], 
      Hue[0.030735677634828387`, 1, 1], 
      Hue[0.031028552185619583`, 1, 1], 
      Hue[0.031028552185619583`, 1, 1], 
      Hue[0.031321426736410786`, 1, 1], 
      Hue[0.031321426736410786`, 1, 1], 
      Hue[0.03161430128720199, 1, 1], 
      Hue[0.03161430128720199, 1, 1], 
      Hue[0.031907175837993206`, 1, 1], 
      Hue[0.031907175837993206`, 1, 1], 
      Hue[0.0322000503887844, 1, 1], 
      Hue[0.0322000503887844, 1, 1], 
      Hue[0.0324929249395756, 1, 1], 
      Hue[0.0324929249395756, 1, 1], 
      Hue[0.03278579949036681, 1, 1], 
      Hue[0.03278579949036681, 1, 1], 
      Hue[0.033078674041158, 1, 1], 
      Hue[0.033078674041158, 1, 1], 
      Hue[0.03337154859194921, 1, 1], 
      Hue[0.03337154859194921, 1, 1], 
      Hue[0.033664423142740416`, 1, 1], 
      Hue[0.033664423142740416`, 1, 1], 
      Hue[0.033957297693531695`, 1, 1], 
      Hue[0.033957297693531695`, 1, 1], 
      Hue[0.0342501722443229, 1, 1], 
      Hue[0.0342501722443229, 1, 1], 
      Hue[0.0345430467951141, 1, 1], 
      Hue[0.0345430467951141, 1, 1], 
      Hue[0.034835921345905366`, 1, 1], 
      Hue[0.034835921345905366`, 1, 1], 
      Hue[0.035128795896696575`, 1, 1], 
      Hue[0.035128795896696575`, 1, 1], 
      Hue[0.03542167044748777, 1, 1], 
      Hue[0.03542167044748777, 1, 1], 
      Hue[0.03571454499827898, 1, 1], 
      Hue[0.03571454499827898, 1, 1], 
      Hue[0.03600741954907018, 1, 1], 
      Hue[0.03600741954907018, 1, 1], 
      Hue[0.03630029409986138, 1, 1], 
      Hue[0.03630029409986138, 1, 1], 
      Hue[0.03659316865065259, 1, 1], 
      Hue[0.03659316865065259, 1, 1], 
      Hue[0.0368860432014438, 1, 1], 
      Hue[0.0368860432014438, 1, 1], 
      Hue[0.037178917752234995`, 1, 1], 
      Hue[0.037178917752234995`, 1, 1], 
      Hue[0.0374717923030262, 1, 1], 
      Hue[0.0374717923030262, 1, 1], 
      Hue[0.0377646668538174, 1, 1], 
      Hue[0.0377646668538174, 1, 1], 
      Hue[0.03805754140460861, 1, 1], 
      Hue[0.03805754140460861, 1, 1], 
      Hue[0.03835041595539981, 1, 1], 
      Hue[0.03835041595539981, 1, 1], 
      Hue[0.03893616505698228, 1, 1], 
      Hue[0.03893616505698228, 1, 1], 
      Hue[0.03922903960777349, 1, 1], 
      Hue[0.03922903960777349, 1, 1], 
      Hue[0.039521914158564764`, 1, 1], 
      Hue[0.039521914158564764`, 1, 1], 
      Hue[0.040107663260147156`, 1, 1], 
      Hue[0.040107663260147156`, 1, 1], 
      Hue[0.041279161463311974`, 1, 1], 
      Hue[0.041279161463311974`, 1, 1], 
      Hue[0.041596671037132556`, 1, 1], 
      Hue[0.041596671037132556`, 1, 1], 
      Hue[0.04191418061095314, 1, 1], 
      Hue[0.04191418061095314, 1, 1], 
      Hue[0.0425491997585943, 1, 1], 
      Hue[0.0425491997585943, 1, 1], 
      Hue[0.04286670933241487, 1, 1], 
      Hue[0.04286670933241487, 1, 1], 
      Hue[0.04318421890623545, 1, 1], 
      Hue[0.04318421890623545, 1, 1], 
      Hue[0.04381923805387661, 1, 1], 
      Hue[0.04381923805387661, 1, 1], 
      Hue[0.04413674762769719, 1, 1], 
      Hue[0.04413674762769719, 1, 1], 
      Hue[0.044454257201517765`, 1, 1], 
      Hue[0.044454257201517765`, 1, 1], 
      Hue[0.04508927634915892, 1, 1], 
      Hue[0.04508927634915892, 1, 1], 
      Hue[0.04635931464444118, 1, 1], 
      Hue[0.04635931464444118, 1, 1], 
      Hue[0.04667682421826168, 1, 1], 
      Hue[0.04667682421826168, 1, 1], 
      Hue[0.04699433379208225, 1, 1], 
      Hue[0.04699433379208225, 1, 1], 
      Hue[0.04762935293972342, 1, 1], 
      Hue[0.04762935293972342, 1, 1], 
      Hue[0.04889939123500573, 1, 1], 
      Hue[0.04889939123500573, 1, 1], 
      Hue[0.0492169008088263, 1, 1], 
      Hue[0.0492169008088263, 1, 1], 
      Hue[0.04953441038264688, 1, 1], 
      Hue[0.04953441038264688, 1, 1], 
      Hue[0.05016942953028804, 1, 1], 
      Hue[0.05016942953028804, 1, 1], 
      Hue[0.05143946782557036, 1, 1], 
      Hue[0.05143946782557036, 1, 1], 
      Hue[0.05175697739939094, 1, 1], 
      Hue[0.05175697739939094, 1, 1], 
      Hue[0.05207448697321152, 1, 1], 
      Hue[0.05207448697321152, 1, 1], 
      Hue[0.05270950612085268, 1, 1], 
      Hue[0.05270950612085268, 1, 1], 
      Hue[0.05397954441613492, 1, 1], 
      Hue[0.05397954441613492, 1, 1], 
      Hue[0.054297053989955435`, 1, 1], 
      Hue[0.054297053989955435`, 1, 1], 
      Hue[0.05461456356377601, 1, 1], 
      Hue[0.05461456356377601, 1, 1], 
      Hue[0.05524958271141718, 1, 1], 
      Hue[0.05524958271141718, 1, 1], 
      Hue[0.05651962100669947, 1, 1], 
      Hue[0.05651962100669947, 1, 1], 
      Hue[0.05683713058052007, 1, 1], 
      Hue[0.05683713058052007, 1, 1], 
      Hue[0.05715464015434064, 1, 1], 
      Hue[0.05715464015434064, 1, 1], 
      Hue[0.057789659301981794`, 1, 1], 
      Hue[0.057789659301981794`, 1, 1], 
      Hue[0.059059697597264114`, 1, 1], 
      Hue[0.059059697597264114`, 1, 1], 
      Hue[0.061599774187828665`, 1, 1], 
      Hue[0.061599774187828665`, 1, 1], 
      Hue[0.061896242940041334`, 1, 1], 
      Hue[0.061896242940041334`, 1, 1], 
      Hue[0.062192711692253905`, 1, 1], 
      Hue[0.062192711692253905`, 1, 1], 
      Hue[0.06278564919667907, 1, 1], 
      Hue[0.06278564919667907, 1, 1], 
      Hue[0.0639715242055294, 1, 1], 
      Hue[0.0639715242055294, 1, 1], 
      Hue[0.06634327422323016, 1, 1], 
      Hue[0.06634327422323016, 1, 1], 
      Hue[0.06663974297544276, 1, 1], 
      Hue[0.06663974297544276, 1, 1], 
      Hue[0.06693621172765533, 1, 1], 
      Hue[0.06693621172765533, 1, 1], 
      Hue[0.06752914923208049, 1, 1], 
      Hue[0.06752914923208049, 1, 1], 
      Hue[0.06871502424093089, 1, 1], 
      Hue[0.06871502424093089, 1, 1], 
      Hue[0.0710867742586316, 1, 1], 
      Hue[0.0710867742586316, 1, 1], 
      Hue[0.07138324301084417, 1, 1], 
      Hue[0.07138324301084417, 1, 1], 
      Hue[0.07167971176305678, 1, 1], 
      Hue[0.07167971176305678, 1, 1], 
      Hue[0.0722726492674819, 1, 1], 
      Hue[0.0722726492674819, 1, 1], 
      Hue[0.0734585242763323, 1, 1], 
      Hue[0.0734585242763323, 1, 1], 
      Hue[0.07583027429403302, 1, 1], 
      Hue[0.07583027429403302, 1, 1], 
      Hue[0.0761267430462456, 1, 1], 
      Hue[0.0761267430462456, 1, 1], 
      Hue[0.07642321179845818, 1, 1], 
      Hue[0.07642321179845818, 1, 1], 
      Hue[0.07701614930288335, 1, 1], 
      Hue[0.07701614930288335, 1, 1], 
      Hue[0.07820202431173374, 1, 1], 
      Hue[0.07820202431173374, 1, 1], 
      Hue[0.08057377432943442, 1, 1], 
      Hue[0.08057377432943442, 1, 1], 
      Hue[0.0808644275415848, 1, 1], 
      Hue[0.0808644275415848, 1, 1], 
      Hue[0.08115508075373519, 1, 1], 
      Hue[0.08115508075373519, 1, 1], 
      Hue[0.08173638717803601, 1, 1], 
      Hue[0.08173638717803601, 1, 1], 
      Hue[0.08202704039018638, 1, 1], 
      Hue[0.08202704039018638, 1, 1], 
      Hue[0.08231769360233682, 1, 1], 
      Hue[0.08231769360233682, 1, 1], 
      Hue[0.08260834681448721, 1, 1], 
      Hue[0.08260834681448721, 1, 1], 
      Hue[0.08289900002663757, 1, 1], 
      Hue[0.08289900002663757, 1, 1], 
      Hue[0.08318965323878796, 1, 1], 
      Hue[0.08318965323878796, 1, 1], 
      Hue[0.5834803064509383, 1, 1], 
      Hue[0.5834803064509383, 1, 1], 
      Hue[0.5837709596630887, 1, 1], 
      Hue[0.5837709596630887, 1, 1], 
      Hue[0.5840616128752392, 1, 1], 
      Hue[0.5840616128752392, 1, 1], 
      Hue[0.5843522660873894, 1, 1], 
      Hue[0.5843522660873894, 1, 1], 
      Hue[0.5846429192995399, 1, 1], 
      Hue[0.5846429192995399, 1, 1], 
      Hue[0.5852242257238407, 1, 1], 
      Hue[0.5852242257238407, 1, 1], 
      Hue[0.585514878935991, 1, 1], 
      Hue[0.585514878935991, 1, 1], 
      Hue[0.5858055321481415, 1, 1], 
      Hue[0.5858055321481415, 1, 1], 
      Hue[0.5863868385724422, 1, 1], 
      Hue[0.5863868385724422, 1, 1], 
      Hue[0.5875494514210438, 1, 1], 
      Hue[0.5875494514210438, 1, 1], 
      Hue[0.5898746771182468, 1, 1], 
      Hue[0.5898746771182468, 1, 1], 
      Hue[0.5901653303303972, 1, 1], 
      Hue[0.5901653303303972, 1, 1], 
      Hue[0.5904559835425476, 1, 1], 
      Hue[0.5904559835425476, 1, 1], 
      Hue[0.5910372899668483, 1, 1], 
      Hue[0.5910372899668483, 1, 1], 
      Hue[0.5921999028154499, 1, 1], 
      Hue[0.5921999028154499, 1, 1], 
      Hue[0.594525128512653, 1, 1], 
      Hue[0.594525128512653, 1, 1], 
      Hue[0.5991755799070593, 1, 1], 
      Hue[0.5991755799070593, 1, 1], 
      Hue[0.5994908681422391, 1, 1], 
      Hue[0.5994908681422391, 1, 1], 
      Hue[0.5998061563774189, 1, 1], 
      Hue[0.5998061563774189, 1, 1], 
      Hue[0.6004367328477782, 1, 1], 
      Hue[0.6004367328477782, 1, 1], 
      Hue[0.6016978857884973, 1, 1], 
      Hue[0.6016978857884973, 1, 1], 
      Hue[0.6042201916699352, 1, 1], 
      Hue[0.6042201916699352, 1, 1], 
      Hue[0.6092648034328112, 1, 1], 
      Hue[0.6092648034328112, 1, 1], 
      Hue[0.6193540269585631, 1, 1], 
      Hue[0.6193540269585631, 1, 1], 
      Hue[0.6381858614271563, 1, 1], 
      Hue[0.6381858614271563, 1, 1], 
      Hue[0.6385047438637573, 1, 1], 
      Hue[0.6385047438637573, 1, 1], 
      Hue[0.6388236263003584, 1, 1], 
      Hue[0.6388236263003584, 1, 1], 
      Hue[0.6394613911735607, 1, 1], 
      Hue[0.6394613911735607, 1, 1], 
      Hue[0.6407369209199651, 1, 1], 
      Hue[0.6407369209199651, 1, 1], 
      Hue[0.6432879804127742, 1, 1], 
      Hue[0.6432879804127742, 1, 1], 
      Hue[0.6483900993983922, 1, 1], 
      Hue[0.6483900993983922, 1, 1], 
      Hue[0.6487089818349934, 1, 1], 
      Hue[0.6487089818349934, 1, 1], 
      Hue[0.6490278642715945, 1, 1], 
      Hue[0.6490278642715945, 1, 1], 
      Hue[0.6496656291447966, 1, 1], 
      Hue[0.6496656291447966, 1, 1], 
      Hue[0.6509411588912013, 1, 1], 
      Hue[0.6509411588912013, 1, 1], 
      Hue[0.6534922183840102, 1, 1], 
      Hue[0.6534922183840102, 1, 1], 
      Hue[0.6585943373696282, 1, 1], 
      Hue[0.6585943373696282, 1, 1], 
      Hue[0.6589074042661671, 1, 1], 
      Hue[0.6589074042661671, 1, 1], 
      Hue[0.6592204711627061, 1, 1], 
      Hue[0.6592204711627061, 1, 1], 
      Hue[0.6598466049557838, 1, 1], 
      Hue[0.6598466049557838, 1, 1], 
      Hue[0.6610988725419396, 1, 1], 
      Hue[0.6610988725419396, 1, 1], 
      Hue[0.663603407714251, 1, 1], 
      Hue[0.663603407714251, 1, 1], 
      Hue[0.66391647461079, 1, 1], 
      Hue[0.66391647461079, 1, 1], 
      Hue[0.6642295415073289, 1, 1], 
      Hue[0.6642295415073289, 1, 1], 
      Hue[0.6648556753004067, 1, 1], 
      Hue[0.6648556753004067, 1, 1], 
      Hue[0.6661079428865624, 1, 1], 
      Hue[0.6661079428865624, 1, 1], 
      Hue[0.6686124780588738, 1, 1], 
      Hue[0.6686124780588738, 1, 1], 
      Hue[0.6689255449554127, 1, 1], 
      Hue[0.6689255449554127, 1, 1], 
      Hue[0.6692386118519515, 1, 1], 
      Hue[0.6692386118519515, 1, 1], 
      Hue[0.6698647456450295, 1, 1], 
      Hue[0.6698647456450295, 1, 1], 
      Hue[0.6711170132311852, 1, 1], 
      Hue[0.6711170132311852, 1, 1], 
      Hue[0.6714300801277242, 1, 1], 
      Hue[0.6714300801277242, 1, 1], 
      Hue[0.6717431470242632, 1, 1], 
      Hue[0.6717431470242632, 1, 1], 
      Hue[0.6723692808173409, 1, 1], 
      Hue[0.6723692808173409, 1, 1], 
      Hue[0.6736215484034966, 1, 1], 
      Hue[0.6736215484034966, 1, 1], 
      Hue[0.6739346153000356, 1, 1], 
      Hue[0.6739346153000356, 1, 1], 
      Hue[0.6742476821965744, 1, 1], 
      Hue[0.6742476821965744, 1, 1], 
      Hue[0.6748738159896523, 1, 1], 
      Hue[0.6748738159896523, 1, 1], 
      Hue[0.6761260835758081, 1, 1], 
      Hue[0.6761260835758081, 1, 1], 
      Hue[0.6786306187481195, 1, 1], 
      Hue[0.6786306187481195, 1, 1], 
      Hue[0.6789226448230503, 1, 1], 
      Hue[0.6789226448230503, 1, 1], 
      Hue[0.6792146708979814, 1, 1], 
      Hue[0.6792146708979814, 1, 1], 
      Hue[0.6797987230478432, 1, 1], 
      Hue[0.6797987230478432, 1, 1], 
      Hue[0.680966827347567, 1, 1], 
      Hue[0.680966827347567, 1, 1], 
      Hue[0.6812588534224979, 1, 1], 
      Hue[0.6812588534224979, 1, 1], 
      Hue[0.6815508794974288, 1, 1], 
      Hue[0.6815508794974288, 1, 1], 
      Hue[0.6821349316472907, 1, 1], 
      Hue[0.6821349316472907, 1, 1], 
      Hue[0.6833030359470145, 1, 1], 
      Hue[0.6833030359470145, 1, 1], 
      Hue[0.6835950620219453, 1, 1], 
      Hue[0.6835950620219453, 1, 1], 
      Hue[0.6838870880968764, 1, 1], 
      Hue[0.6838870880968764, 1, 1], 
      Hue[0.6844711402467383, 1, 1], 
      Hue[0.6844711402467383, 1, 1], 
      Hue[0.6847631663216691, 1, 1], 
      Hue[0.6847631663216691, 1, 1], 
      Hue[0.6850551923966001, 1, 1], 
      Hue[0.6850551923966001, 1, 1], 
      Hue[0.685639244546462, 1, 1], 
      Hue[0.685639244546462, 1, 1], 
      Hue[0.6859312706213929, 1, 1], 
      Hue[0.6859312706213929, 1, 1], 
      Hue[0.6862232966963239, 1, 1], 
      Hue[0.6862232966963239, 1, 1], 
      Hue[0.6865153227712548, 1, 1], 
      Hue[0.6865153227712548, 1, 1], 
      Hue[0.6868073488461857, 1, 1], 
      Hue[0.6868073488461857, 1, 1], 
      Hue[0.6870993749211167, 1, 1], 
      Hue[0.6870993749211167, 1, 1], 
      Hue[0.6873914009960476, 1, 1], 
      Hue[0.6873914009960476, 1, 1], 
      Hue[0.6876834270709786, 1, 1], 
      Hue[0.6876834270709786, 1, 1], 
      Hue[0.6879754531459095, 1, 1], 
      Hue[0.6879754531459095, 1, 1], 
      Hue[0.6882674792208405, 1, 1], 
      Hue[0.6882674792208405, 1, 1], 
      Hue[0.6885595052957714, 1, 1], 
      Hue[0.6885595052957714, 1, 1], 
      Hue[0.6888515313707023, 1, 1], 
      Hue[0.6888515313707023, 1, 1], 
      Hue[0.6891435574456333, 1, 1], 
      Hue[0.6891435574456333, 1, 1], 
      Hue[0.6894355835205642, 1, 1], 
      Hue[0.6894355835205642, 1, 1], 
      Hue[0.6897276095954952, 1, 1], 
      Hue[0.6897276095954952, 1, 1], 
      Hue[0.6900196356704261, 1, 1], 
      Hue[0.6900196356704261, 1, 1], 
      Hue[0.6903116617453571, 1, 1], 
      Hue[0.6903116617453571, 1, 1], 
      Hue[0.6906036878202879, 1, 1], 
      Hue[0.6906036878202879, 1, 1], 
      Hue[0.690895713895219, 1, 1], 
      Hue[0.690895713895219, 1, 1], 
      Hue[0.6911877399701498, 1, 1], 
      Hue[0.6911877399701498, 1, 1], 
      Hue[0.6914797660450808, 1, 1], 
      Hue[0.6914797660450808, 1, 1], 
      Hue[0.6917717921200117, 1, 1], 
      Hue[0.6917717921200117, 1, 1], 
      Hue[0.6920638181949427, 1, 1], 
      Hue[0.6920638181949427, 1, 1], 
      Hue[0.6923558442698736, 1, 1], 
      Hue[0.6923558442698736, 1, 1], 
      Hue[0.6926478703448045, 1, 1], 
      Hue[0.6926478703448045, 1, 1], 
      Hue[0.6929398964197355, 1, 1], 
      Hue[0.6929398964197355, 1, 1], 
      Hue[0.6932319224946664, 1, 1], 
      Hue[0.6932319224946664, 1, 1], 
      Hue[0.6935239485695974, 1, 1], 
      Hue[0.6935239485695974, 1, 1], 
      Hue[0.6938159746445283, 1, 1], 
      Hue[0.6938159746445283, 1, 1], 
      Hue[0.6941080007194593, 1, 1], 
      Hue[0.6941080007194593, 1, 1], 
      Hue[0.6944000267943902, 1, 1], 
      Hue[0.6944000267943902, 1, 1], 
      Hue[0.6949840789442521, 1, 1], 
      Hue[0.6949840789442521, 1, 1], 
      Hue[0.695276105019183, 1, 1], 
      Hue[0.695276105019183, 1, 1], 
      Hue[0.695568131094114, 1, 1], 
      Hue[0.695568131094114, 1, 1], 
      Hue[0.6961521832439759, 1, 1], 
      Hue[0.6961521832439759, 1, 1], 
      Hue[0.6964442093189067, 1, 1], 
      Hue[0.6964442093189067, 1, 1], 
      Hue[0.6967362353938378, 1, 1], 
      Hue[0.6967362353938378, 1, 1], 
      Hue[0.6973202875436997, 1, 1], 
      Hue[0.6973202875436997, 1, 1], 
      Hue[0.6976369486416599, 1, 1], 
      Hue[0.6976369486416599, 1, 1], 
      Hue[0.6979536097396202, 1, 1], 
      Hue[0.6979536097396202, 1, 1], 
      Hue[0.6985869319355408, 1, 1], 
      Hue[0.6985869319355408, 1, 1], 
      Hue[0.698903593033501, 1, 1], 
      Hue[0.698903593033501, 1, 1], 
      Hue[0.6992202541314614, 1, 1], 
      Hue[0.6992202541314614, 1, 1], 
      Hue[0.6998535763273821, 1, 1], 
      Hue[0.6998535763273821, 1, 1], 
      Hue[0.7001702374253423, 1, 1], 
      Hue[0.7001702374253423, 1, 1], 
      Hue[0.7004868985233026, 1, 1], 
      Hue[0.7004868985233026, 1, 1], 
      Hue[0.7011202207192233, 1, 1], 
      Hue[0.7011202207192233, 1, 1], 
      Hue[0.7023868651110644, 1, 1], 
      Hue[0.7023868651110644, 1, 1], 
      Hue[0.7027035262090247, 1, 1], 
      Hue[0.7027035262090247, 1, 1], 
      Hue[0.703020187306985, 1, 1], 
      Hue[0.703020187306985, 1, 1], 
      Hue[0.7036535095029056, 1, 1], 
      Hue[0.7036535095029056, 1, 1], 
      Hue[0.7049201538947468, 1, 1], 
      Hue[0.7049201538947468, 1, 1], 
      Hue[0.7052368149927071, 1, 1], 
      Hue[0.7052368149927071, 1, 1], 
      Hue[0.7055534760906674, 1, 1], 
      Hue[0.7055534760906674, 1, 1], 
      Hue[0.706186798286588, 1, 1], 
      Hue[0.706186798286588, 1, 1], 
      Hue[0.7074534426784292, 1, 1], 
      Hue[0.7074534426784292, 1, 1], 
      Hue[0.7077701037763895, 1, 1], 
      Hue[0.7077701037763895, 1, 1], 
      Hue[0.7080867648743498, 1, 1], 
      Hue[0.7080867648743498, 1, 1], 
      Hue[0.7087200870702703, 1, 1], 
      Hue[0.7087200870702703, 1, 1], 
      Hue[0.7099867314621116, 1, 1], 
      Hue[0.7099867314621116, 1, 1], 
      Hue[0.7103033925600719, 1, 1], 
      Hue[0.7103033925600719, 1, 1], 
      Hue[0.7106200536580323, 1, 1], 
      Hue[0.7106200536580323, 1, 1], 
      Hue[0.7112533758539528, 1, 1], 
      Hue[0.7112533758539528, 1, 1], 
      Hue[0.712520020245794, 1, 1], 
      Hue[0.712520020245794, 1, 1], 
      Hue[0.7128366813437543, 1, 1], 
      Hue[0.7128366813437543, 1, 1], 
      Hue[0.7131533424417147, 1, 1], 
      Hue[0.7131533424417147, 1, 1], 
      Hue[0.7137866646376352, 1, 1], 
      Hue[0.7137866646376352, 1, 1], 
      Hue[0.7150533090294764, 1, 1], 
      Hue[0.7150533090294764, 1, 1], 
      Hue[0.7175865978131588, 1, 1], 
      Hue[0.7175865978131588, 1, 1], 
      Hue[0.7178822180895112, 1, 1], 
      Hue[0.7178822180895112, 1, 1], 
      Hue[0.7181778383658635, 1, 1], 
      Hue[0.7181778383658635, 1, 1], 
      Hue[0.7187690789185681, 1, 1], 
      Hue[0.7187690789185681, 1, 1], 
      Hue[0.7199515600239774, 1, 1], 
      Hue[0.7199515600239774, 1, 1], 
      Hue[0.7223165222347959, 1, 1], 
      Hue[0.7223165222347959, 1, 1], 
      Hue[0.7226121425111482, 1, 1], 
      Hue[0.7226121425111482, 1, 1], 
      Hue[0.7229077627875006, 1, 1], 
      Hue[0.7229077627875006, 1, 1], 
      Hue[0.7234990033402051, 1, 1], 
      Hue[0.7234990033402051, 1, 1], 
      Hue[0.7246814844456144, 1, 1], 
      Hue[0.7246814844456144, 1, 1], 
      Hue[0.727046446656433, 1, 1], 
      Hue[0.727046446656433, 1, 1], 
      Hue[0.7273420669327854, 1, 1], 
      Hue[0.7273420669327854, 1, 1], 
      Hue[0.7276376872091377, 1, 1], 
      Hue[0.7276376872091377, 1, 1], 
      Hue[0.7282289277618423, 1, 1], 
      Hue[0.7282289277618423, 1, 1], 
      Hue[0.7294114088672516, 1, 1], 
      Hue[0.7294114088672516, 1, 1], 
      Hue[0.7317763710780701, 1, 1], 
      Hue[0.7317763710780701, 1, 1], 
      Hue[0.7365062954997073, 1, 1], 
      Hue[0.7365062954997073, 1, 1], 
      Hue[0.7367961002359974, 1, 1], 
      Hue[0.7367961002359974, 1, 1], 
      Hue[0.7370859049722875, 1, 1], 
      Hue[0.7370859049722875, 1, 1], 
      Hue[0.7376655144448677, 1, 1], 
      Hue[0.7376655144448677, 1, 1], 
      Hue[0.7388247333900282, 1, 1], 
      Hue[0.7388247333900282, 1, 1], 
      Hue[0.7411431712803491, 1, 1], 
      Hue[0.7411431712803491, 1, 1], 
      Hue[0.7414329760166392, 1, 1], 
      Hue[0.7414329760166392, 1, 1], 
      Hue[0.7417227807529294, 1, 1], 
      Hue[0.7417227807529294, 1, 1], 
      Hue[0.7423023902255096, 1, 1], 
      Hue[0.7423023902255096, 1, 1], 
      Hue[0.7434616091706701, 1, 1], 
      Hue[0.7434616091706701, 1, 1], 
      Hue[0.745780047060991, 1, 1], 
      Hue[0.745780047060991, 1, 1], 
      Hue[0.746069851797281, 1, 1], 
      Hue[0.746069851797281, 1, 1], 
      Hue[0.7463596565335713, 1, 1], 
      Hue[0.7463596565335713, 1, 1], 
      Hue[0.7469392660061515, 1, 1], 
      Hue[0.7469392660061515, 1, 1], 
      Hue[0.7472290707424416, 1, 1], 
      Hue[0.7472290707424416, 1, 1], 
      Hue[0.7475188754787316, 1, 1], 
      Hue[0.7475188754787316, 1, 1], 
      Hue[0.7480984849513119, 1, 1], 
      Hue[0.7480984849513119, 1, 1], 
      Hue[0.748388289687602, 1, 1], 
      Hue[0.748388289687602, 1, 1], 
      Hue[0.7486780944238922, 1, 1], 
      Hue[0.7486780944238922, 1, 1], 
      Hue[0.7489678991601822, 1, 1], 
      Hue[0.7489678991601822, 1, 1], 
      Hue[0.7492577038964724, 1, 1], 
      Hue[0.7492577038964724, 1, 1], 
      Hue[0.7495475086327625, 1, 1], 
      Hue[0.7495475086327625, 1, 1], 
      Hue[0.7498373133690526, 1, 1], 
      Hue[0.7498373133690526, 1, 1], 
      Hue[0.25012711810534277`, 1, 1], 
      Hue[0.25012711810534277`, 1, 1], 
      Hue[0.25041692284163286`, 1, 1], 
      Hue[0.25041692284163286`, 1, 1], 
      Hue[0.250706727577923, 1, 1], 
      Hue[0.250706727577923, 1, 1], 
      Hue[0.25099653231421315`, 1, 1], 
      Hue[0.25099653231421315`, 1, 1], 
      Hue[0.25157614178679333`, 1, 1], 
      Hue[0.25157614178679333`, 1, 1], 
      Hue[0.2527353607319538, 1, 1], 
      Hue[0.2527353607319538, 1, 1], 
      Hue[0.2550537986222748, 1, 1], 
      Hue[0.2550537986222748, 1, 1], 
      Hue[0.2553682383815943, 1, 1], 
      Hue[0.2553682383815943, 1, 1], 
      Hue[0.2556826781409137, 1, 1], 
      Hue[0.2556826781409137, 1, 1], 
      Hue[0.25631155765955266`, 1, 1], 
      Hue[0.25631155765955266`, 1, 1], 
      Hue[0.2575693166968306, 1, 1], 
      Hue[0.2575693166968306, 1, 1], 
      Hue[0.2600848347713864, 1, 1], 
      Hue[0.2600848347713864, 1, 1], 
      Hue[0.26511587092049804`, 1, 1], 
      Hue[0.26511587092049804`, 1, 1], 
      Hue[0.27517794321872124`, 1, 1], 
      Hue[0.27517794321872124`, 1, 1], 
      Hue[0.2939554752322569, 1, 1], 
      Hue[0.2939554752322569, 1, 1], 
      Hue[0.2942735091929977, 1, 1], 
      Hue[0.2942735091929977, 1, 1], 
      Hue[0.2945915431537386, 1, 1], 
      Hue[0.2945915431537386, 1, 1], 
      Hue[0.2952276110752203, 1, 1], 
      Hue[0.2952276110752203, 1, 1], 
      Hue[0.29649974691818365`, 1, 1], 
      Hue[0.29649974691818365`, 1, 1], 
      Hue[0.29904401860411056`, 1, 1], 
      Hue[0.29904401860411056`, 1, 1], 
      Hue[0.3041325619759642, 1, 1], 
      Hue[0.3041325619759642, 1, 1], 
      Hue[0.30445059593670504`, 1, 1], 
      Hue[0.30445059593670504`, 1, 1], 
      Hue[0.30476862989744585`, 1, 1], 
      Hue[0.30476862989744585`, 1, 1], 
      Hue[0.3054046978189276, 1, 1], 
      Hue[0.3054046978189276, 1, 1], 
      Hue[0.306676833661891, 1, 1], 
      Hue[0.306676833661891, 1, 1], 
      Hue[0.3092211053478179, 1, 1], 
      Hue[0.3092211053478179, 1, 1], 
      Hue[0.31430964871967154`, 1, 1], 
      Hue[0.31430964871967154`, 1, 1], 
      Hue[0.3146218671403502, 1, 1], 
      Hue[0.3146218671403502, 1, 1], 
      Hue[0.3149340855610288, 1, 1], 
      Hue[0.3149340855610288, 1, 1], 
      Hue[0.31555852240238613`, 1, 1], 
      Hue[0.31555852240238613`, 1, 1], 
      Hue[0.3168073960851008, 1, 1], 
      Hue[0.3168073960851008, 1, 1], 
      Hue[0.31930514345052996`, 1, 1], 
      Hue[0.31930514345052996`, 1, 1], 
      Hue[0.3196173618712086, 1, 1], 
      Hue[0.3196173618712086, 1, 1], 
      Hue[0.3199295802918873, 1, 1], 
      Hue[0.3199295802918873, 1, 1], 
      Hue[0.32055401713324455`, 1, 1], 
      Hue[0.32055401713324455`, 1, 1], 
      Hue[0.3218028908159592, 1, 1], 
      Hue[0.3218028908159592, 1, 1], 
      Hue[0.3243006381813884, 1, 1], 
      Hue[0.3243006381813884, 1, 1], 
      Hue[0.324612856602067, 1, 1], 
      Hue[0.324612856602067, 1, 1], 
      Hue[0.32492507502274565`, 1, 1], 
      Hue[0.32492507502274565`, 1, 1], 
      Hue[0.32554951186410297`, 1, 1], 
      Hue[0.32554951186410297`, 1, 1], 
      Hue[0.32679838554681756`, 1, 1], 
      Hue[0.32679838554681756`, 1, 1], 
      Hue[0.3271106039674962, 1, 1], 
      Hue[0.3271106039674962, 1, 1], 
      Hue[0.3274228223881749, 1, 1], 
      Hue[0.3274228223881749, 1, 1], 
      Hue[0.32804725922953215`, 1, 1], 
      Hue[0.32804725922953215`, 1, 1], 
      Hue[0.3292961329122468, 1, 1], 
      Hue[0.3292961329122468, 1, 1], 
      Hue[0.32960835133292543`, 1, 1], 
      Hue[0.32960835133292543`, 1, 1], 
      Hue[0.3299205697536041, 1, 1], 
      Hue[0.3299205697536041, 1, 1], 
      Hue[0.3305450065949614, 1, 1], 
      Hue[0.3305450065949614, 1, 1], 
      Hue[0.331793880277676, 1, 1], 
      Hue[0.331793880277676, 1, 1], 
      Hue[0.3342916276431052, 1, 1], 
      Hue[0.3342916276431052, 1, 1], 
      Hue[0.33458280524217593`, 1, 1], 
      Hue[0.33458280524217593`, 1, 1], 
      Hue[0.3348739828412466, 1, 1], 
      Hue[0.3348739828412466, 1, 1], 
      Hue[0.3354563380393879, 1, 1], 
      Hue[0.3354563380393879, 1, 1], 
      Hue[0.3366210484356706, 1, 1], 
      Hue[0.3366210484356706, 1, 1], 
      Hue[0.3369122260347413, 1, 1], 
      Hue[0.3369122260347413, 1, 1], 
      Hue[0.33720340363381196`, 1, 1], 
      Hue[0.33720340363381196`, 1, 1], 
      Hue[0.3377857588319533, 1, 1], 
      Hue[0.3377857588319533, 1, 1], 
      Hue[0.33895046922823596`, 1, 1], 
      Hue[0.33895046922823596`, 1, 1], 
      Hue[0.3392416468273066, 1, 1], 
      Hue[0.3392416468273066, 1, 1], 
      Hue[0.33953282442637733`, 1, 1], 
      Hue[0.33953282442637733`, 1, 1], 
      Hue[0.34011517962451865`, 1, 1], 
      Hue[0.34011517962451865`, 1, 1], 
      Hue[0.3404063572235893, 1, 1], 
      Hue[0.3404063572235893, 1, 1], 
      Hue[0.34069753482266, 1, 1], 
      Hue[0.34069753482266, 1, 1], 
      Hue[0.34127989002080134`, 1, 1], 
      Hue[0.34127989002080134`, 1, 1], 
      Hue[0.341571067619872, 1, 1], 
      Hue[0.341571067619872, 1, 1], 
      Hue[0.34186224521894265`, 1, 1], 
      Hue[0.34186224521894265`, 1, 1], 
      Hue[0.34215342281801336`, 1, 1], 
      Hue[0.34215342281801336`, 1, 1], 
      Hue[0.34244460041708397`, 1, 1], 
      Hue[0.34244460041708397`, 1, 1], 
      Hue[0.3427357780161547, 1, 1], 
      Hue[0.3427357780161547, 1, 1], 
      Hue[0.34302695561522534`, 1, 1], 
      Hue[0.34302695561522534`, 1, 1], 
      Hue[0.34360931081336665`, 1, 1], 
      Hue[0.34360931081336665`, 1, 1], 
      Hue[0.34390048841243737`, 1, 1], 
      Hue[0.34390048841243737`, 1, 1], 
      Hue[0.344191666011508, 1, 1], 
      Hue[0.344191666011508, 1, 1], 
      Hue[0.34448284361057874`, 1, 1], 
      Hue[0.34448284361057874`, 1, 1], 
      Hue[0.34477402120964934`, 1, 1], 
      Hue[0.34477402120964934`, 1, 1], 
      Hue[0.34506519880872005`, 1, 1], 
      Hue[0.34506519880872005`, 1, 1], 
      Hue[0.3453563764077907, 1, 1], 
      Hue[0.3453563764077907, 1, 1], 
      Hue[0.3456475540068614, 1, 1], 
      Hue[0.3456475540068614, 1, 1], 
      Hue[0.345938731605932, 1, 1], 
      Hue[0.345938731605932, 1, 1], 
      Hue[0.3462299092050027, 1, 1], 
      Hue[0.3462299092050027, 1, 1], 
      Hue[0.3465210868040734, 1, 1], 
      Hue[0.3465210868040734, 1, 1], 
      Hue[0.346812264403144, 1, 1], 
      Hue[0.346812264403144, 1, 1], 
      Hue[0.3471034420022147, 1, 1], 
      Hue[0.3471034420022147, 1, 1], 
      Hue[0.34739461960128537`, 1, 1], 
      Hue[0.34739461960128537`, 1, 1], 
      Hue[0.3476857972003561, 1, 1], 
      Hue[0.3476857972003561, 1, 1], 
      Hue[0.34797697479942674`, 1, 1], 
      Hue[0.34797697479942674`, 1, 1], 
      Hue[0.3482681523984974, 1, 1], 
      Hue[0.3482681523984974, 1, 1], 
      Hue[0.34855932999756806`, 1, 1], 
      Hue[0.34855932999756806`, 1, 1], 
      Hue[0.34885050759663877`, 1, 1], 
      Hue[0.34885050759663877`, 1, 1], 
      Hue[0.34914168519570943`, 1, 1], 
      Hue[0.34914168519570943`, 1, 1], 
      Hue[0.3494328627947801, 1, 1], 
      Hue[0.3494328627947801, 1, 1], 
      Hue[0.3497240403938508, 1, 1], 
      Hue[0.3497240403938508, 1, 1], 
      Hue[0.35001521799292146`, 1, 1], 
      Hue[0.35001521799292146`, 1, 1], 
      Hue[0.3503063955919921, 1, 1], 
      Hue[0.3503063955919921, 1, 1], 
      Hue[0.3505975731910628, 1, 1], 
      Hue[0.3505975731910628, 1, 1], 
      Hue[0.35088875079013343`, 1, 1], 
      Hue[0.35088875079013343`, 1, 1], 
      Hue[0.3511799283892041, 1, 1], 
      Hue[0.3511799283892041, 1, 1], 
      Hue[0.3514711059882748, 1, 1], 
      Hue[0.3514711059882748, 1, 1], 
      Hue[0.3517622835873454, 1, 1], 
      Hue[0.3517622835873454, 1, 1], 
      Hue[0.3520534611864161, 1, 1], 
      Hue[0.3520534611864161, 1, 1], 
      Hue[0.3523446387854868, 1, 1], 
      Hue[0.3523446387854868, 1, 1], 
      Hue[0.3529269939836281, 1, 1], 
      Hue[0.3529269939836281, 1, 1], 
      Hue[0.3532428066057282, 1, 1], 
      Hue[0.3532428066057282, 1, 1], 
      Hue[0.3535586192278282, 1, 1], 
      Hue[0.3535586192278282, 1, 1], 
      Hue[0.35419024447202824`, 1, 1], 
      Hue[0.35419024447202824`, 1, 1], 
      Hue[0.3545060570941283, 1, 1], 
      Hue[0.3545060570941283, 1, 1], 
      Hue[0.3548218697162283, 1, 1], 
      Hue[0.3548218697162283, 1, 1], 
      Hue[0.35545349496042833`, 1, 1], 
      Hue[0.35545349496042833`, 1, 1], 
      Hue[0.35576930758252845`, 1, 1], 
      Hue[0.35576930758252845`, 1, 1], 
      Hue[0.35608512020462846`, 1, 1], 
      Hue[0.35608512020462846`, 1, 1], 
      Hue[0.3567167454488285, 1, 1], 
      Hue[0.3567167454488285, 1, 1], 
      Hue[0.35797999593722857`, 1, 1], 
      Hue[0.35797999593722857`, 1, 1], 
      Hue[0.3582958085593287, 1, 1], 
      Hue[0.3582958085593287, 1, 1], 
      Hue[0.3586116211814287, 1, 1], 
      Hue[0.3586116211814287, 1, 1], 
      Hue[0.3592432464256287, 1, 1], 
      Hue[0.3592432464256287, 1, 1], 
      Hue[0.3605064969140288, 1, 1], 
      Hue[0.3605064969140288, 1, 1], 
      Hue[0.3608223095361289, 1, 1], 
      Hue[0.3608223095361289, 1, 1], 
      Hue[0.36113812215822894`, 1, 1], 
      Hue[0.36113812215822894`, 1, 1], 
      Hue[0.36176974740242895`, 1, 1], 
      Hue[0.36176974740242895`, 1, 1], 
      Hue[0.36303299789082905`, 1, 1], 
      Hue[0.36303299789082905`, 1, 1], 
      Hue[0.3633488105129291, 1, 1], 
      Hue[0.3633488105129291, 1, 1], 
      Hue[0.3636646231350291, 1, 1], 
      Hue[0.3636646231350291, 1, 1], 
      Hue[0.3642962483792292, 1, 1], 
      Hue[0.3642962483792292, 1, 1], 
      Hue[0.36555949886762934`, 1, 1], 
      Hue[0.36555949886762934`, 1, 1], 
      Hue[0.36587531148972935`, 1, 1], 
      Hue[0.36587531148972935`, 1, 1], 
      Hue[0.36619112411182936`, 1, 1], 
      Hue[0.36619112411182936`, 1, 1], 
      Hue[0.36682274935602943`, 1, 1], 
      Hue[0.36682274935602943`, 1, 1], 
      Hue[0.3680859998444296, 1, 1], 
      Hue[0.3680859998444296, 1, 1], 
      Hue[0.3684018124665296, 1, 1], 
      Hue[0.3684018124665296, 1, 1], 
      Hue[0.3687176250886296, 1, 1], 
      Hue[0.3687176250886296, 1, 1], 
      Hue[0.36934925033282967`, 1, 1], 
      Hue[0.36934925033282967`, 1, 1], 
      Hue[0.3706125008212298, 1, 1], 
      Hue[0.3706125008212298, 1, 1], 
      Hue[0.37313900179803006`, 1, 1], 
      Hue[0.37313900179803006`, 1, 1], 
      Hue[0.37343377359852203`, 1, 1], 
      Hue[0.37343377359852203`, 1, 1], 
      Hue[0.3737285453990141, 1, 1], 
      Hue[0.3737285453990141, 1, 1], 
      Hue[0.37431808899999824`, 1, 1], 
      Hue[0.37431808899999824`, 1, 1], 
      Hue[0.37549717620196643`, 1, 1], 
      Hue[0.37549717620196643`, 1, 1], 
      Hue[0.37785535060590275`, 1, 1], 
      Hue[0.37785535060590275`, 1, 1], 
      Hue[0.37815012240639484`, 1, 1], 
      Hue[0.37815012240639484`, 1, 1], 
      Hue[0.37844489420688693`, 1, 1], 
      Hue[0.37844489420688693`, 1, 1], 
      Hue[0.37903443780787094`, 1, 1], 
      Hue[0.37903443780787094`, 1, 1], 
      Hue[0.3802135250098391, 1, 1], 
      Hue[0.3802135250098391, 1, 1], 
      Hue[0.38257169941377556`, 1, 1], 
      Hue[0.38257169941377556`, 1, 1], 
      Hue[0.38286647121426753`, 1, 1], 
      Hue[0.38286647121426753`, 1, 1], 
      Hue[0.3831612430147596, 1, 1], 
      Hue[0.3831612430147596, 1, 1], 
      Hue[0.38375078661574374`, 1, 1], 
      Hue[0.38375078661574374`, 1, 1], 
      Hue[0.38492987381771193`, 1, 1], 
      Hue[0.38492987381771193`, 1, 1], 
      Hue[0.38728804822164825`, 1, 1], 
      Hue[0.38728804822164825`, 1, 1], 
      Hue[0.39200439702952106`, 1, 1], 
      Hue[0.39200439702952106`, 1, 1], 
      Hue[0.39232380385304244`, 1, 1], 
      Hue[0.39232380385304244`, 1, 1], 
      Hue[0.3926432106765638, 1, 1], 
      Hue[0.3926432106765638, 1, 1], 
      Hue[0.3932820243236067, 1, 1], 
      Hue[0.3932820243236067, 1, 1], 
      Hue[0.3945596516176923, 1, 1], 
      Hue[0.3945596516176923, 1, 1], 
      Hue[0.39711490620586354`, 1, 1], 
      Hue[0.39711490620586354`, 1, 1], 
      Hue[0.4022254153822061, 1, 1], 
      Hue[0.4022254153822061, 1, 1], 
      Hue[0.40254482220572746`, 1, 1], 
      Hue[0.40254482220572746`, 1, 1], 
      Hue[0.40286422902924884`, 1, 1], 
      Hue[0.40286422902924884`, 1, 1], 
      Hue[0.4035030426762917, 1, 1], 
      Hue[0.4035030426762917, 1, 1], 
      Hue[0.4047806699703773, 1, 1], 
      Hue[0.4047806699703773, 1, 1], 
      Hue[0.4073359245585486, 1, 1], 
      Hue[0.4073359245585486, 1, 1], 
      Hue[0.40765533138207, 1, 1], 
      Hue[0.40765533138207, 1, 1], 
      Hue[0.4079747382055914, 1, 1], 
      Hue[0.4079747382055914, 1, 1], 
      Hue[0.4086135518526342, 1, 1], 
      Hue[0.4086135518526342, 1, 1], 
      Hue[0.40989117914671985`, 1, 1], 
      Hue[0.40989117914671985`, 1, 1], 
      Hue[0.4124464337348911, 1, 1], 
      Hue[0.4124464337348911, 1, 1], 
      Hue[0.4127600250183503, 1, 1], 
      Hue[0.4127600250183503, 1, 1], 
      Hue[0.4130736163018095, 1, 1], 
      Hue[0.4130736163018095, 1, 1], 
      Hue[0.4137007988687279, 1, 1], 
      Hue[0.4137007988687279, 1, 1], 
      Hue[0.4140143901521871, 1, 1], 
      Hue[0.4140143901521871, 1, 1], 
      Hue[0.41432798143564636`, 1, 1], 
      Hue[0.41432798143564636`, 1, 1], 
      Hue[0.4149551640025647, 1, 1], 
      Hue[0.4149551640025647, 1, 1], 
      Hue[0.41526875528602397`, 1, 1], 
      Hue[0.41526875528602397`, 1, 1], 
      Hue[0.4155823465694832, 1, 1], 
      Hue[0.4155823465694832, 1, 1], 
      Hue[0.4158959378529423, 1, 1], 
      Hue[0.4158959378529423, 1, 1], 
      Hue[0.4162095291364016, 1, 1], 
      Hue[0.4162095291364016, 1, 1], 
      Hue[0.4165231204198608, 1, 1], 
      Hue[0.4165231204198608, 1, 1], 
      Hue[0.9168367117033199, 1, 1], 
      Hue[0.9168367117033199, 1, 1], 
      Hue[0.9171503029867791, 1, 1], 
      Hue[0.9171503029867791, 1, 1], 
      Hue[0.9174638942702383, 1, 1], 
      Hue[0.9174638942702383, 1, 1], 
      Hue[0.9177774855536975, 1, 1], 
      Hue[0.9177774855536975, 1, 1], 
      Hue[0.9180910768371567, 1, 1], 
      Hue[0.9180910768371567, 1, 1], 
      Hue[0.9187182594040751, 1, 1], 
      Hue[0.9187182594040751, 1, 1], 
      Hue[0.919972624537912, 1, 1], 
      Hue[0.919972624537912, 1, 1], 
      Hue[0.9224813548055857, 1, 1], 
      Hue[0.9224813548055857, 1, 1], 
      Hue[0.9227949460890449, 1, 1], 
      Hue[0.9227949460890449, 1, 1], 
      Hue[0.9231085373725041, 1, 1], 
      Hue[0.9231085373725041, 1, 1], 
      Hue[0.9237357199394225, 1, 1], 
      Hue[0.9237357199394225, 1, 1], 
      Hue[0.9249900850732593, 1, 1], 
      Hue[0.9249900850732593, 1, 1], 
      Hue[0.9274988153409329, 1, 1], 
      Hue[0.9274988153409329, 1, 1], 
      Hue[0.9325162758762802, 1, 1], 
      Hue[0.9325162758762802, 1, 1], 
      Hue[0.9328088263381314, 1, 1], 
      Hue[0.9328088263381314, 1, 1], 
      Hue[0.9331013767999826, 1, 1], 
      Hue[0.9331013767999826, 1, 1], 
      Hue[0.9336864777236851, 1, 1], 
      Hue[0.9336864777236851, 1, 1], 
      Hue[0.93485667957109, 1, 1], 
      Hue[0.93485667957109, 1, 1], 
      Hue[0.9371970832658998, 1, 1], 
      Hue[0.9371970832658998, 1, 1], 
      Hue[0.9418778906555193, 1, 1], 
      Hue[0.9418778906555193, 1, 1], 
      Hue[0.9421704411173706, 1, 1], 
      Hue[0.9421704411173706, 1, 1], 
      Hue[0.9424629915792218, 1, 1], 
      Hue[0.9424629915792218, 1, 1], 
      Hue[0.9430480925029242, 1, 1], 
      Hue[0.9430480925029242, 1, 1], 
      Hue[0.9442182943503291, 1, 1], 
      Hue[0.9442182943503291, 1, 1], 
      Hue[0.9465586980451389, 1, 1], 
      Hue[0.9465586980451389, 1, 1], 
      Hue[0.9512395054347584, 1, 1], 
      Hue[0.9512395054347584, 1, 1], 
      Hue[0.951556690919639, 1, 1], 
      Hue[0.951556690919639, 1, 1], 
      Hue[0.9518738764045196, 1, 1], 
      Hue[0.9518738764045196, 1, 1], 
      Hue[0.9525082473742809, 1, 1], 
      Hue[0.9525082473742809, 1, 1], 
      Hue[0.9537769893138032, 1, 1], 
      Hue[0.9537769893138032, 1, 1], 
      Hue[0.9563144731928478, 1, 1], 
      Hue[0.9563144731928478, 1, 1], 
      Hue[0.9613894409509371, 1, 1], 
      Hue[0.9613894409509371, 1, 1], 
      Hue[0.9617066264358177, 1, 1], 
      Hue[0.9617066264358177, 1, 1], 
      Hue[0.9620238119206983, 1, 1], 
      Hue[0.9620238119206983, 1, 1], 
      Hue[0.9626581828904595, 1, 1], 
      Hue[0.9626581828904595, 1, 1], 
      Hue[0.9639269248299818, 1, 1], 
      Hue[0.9639269248299818, 1, 1], 
      Hue[0.9664644087090264, 1, 1], 
      Hue[0.9664644087090264, 1, 1], 
      Hue[0.9715393764671157, 1, 1], 
      Hue[0.9715393764671157, 1, 1], 
      Hue[0.9718355211303884, 1, 1], 
      Hue[0.9718355211303884, 1, 1], 
      Hue[0.972131665793661, 1, 1], 
      Hue[0.972131665793661, 1, 1], 
      Hue[0.9727239551202062, 1, 1], 
      Hue[0.9727239551202062, 1, 1], 
      Hue[0.9739085337732966, 1, 1], 
      Hue[0.9739085337732966, 1, 1], 
      Hue[0.9762776910794774, 1, 1], 
      Hue[0.9762776910794774, 1, 1], 
      Hue[0.9765738357427499, 1, 1], 
      Hue[0.9765738357427499, 1, 1], 
      Hue[0.9768699804060226, 1, 1], 
      Hue[0.9768699804060226, 1, 1], 
      Hue[0.9774622697325678, 1, 1], 
      Hue[0.9774622697325678, 1, 1], 
      Hue[0.9786468483856582, 1, 1], 
      Hue[0.9786468483856582, 1, 1], 
      Hue[0.981016005691839, 1, 1], 
      Hue[0.981016005691839, 1, 1], 
      Hue[0.9813121503551115, 1, 1], 
      Hue[0.9813121503551115, 1, 1], 
      Hue[0.9816082950183842, 1, 1], 
      Hue[0.9816082950183842, 1, 1], 
      Hue[0.9822005843449294, 1, 1], 
      Hue[0.9822005843449294, 1, 1], 
      Hue[0.9833851629980198, 1, 1], 
      Hue[0.9833851629980198, 1, 1], 
      Hue[0.9836813076612924, 1, 1], 
      Hue[0.9836813076612924, 1, 1], 
      Hue[0.983977452324565, 1, 1], 
      Hue[0.983977452324565, 1, 1], 
      Hue[0.9845697416511101, 1, 1], 
      Hue[0.9845697416511101, 1, 1], 
      Hue[0.9857543203042005, 1, 1], 
      Hue[0.9857543203042005, 1, 1], 
      Hue[0.9860504649674732, 1, 1], 
      Hue[0.9860504649674732, 1, 1], 
      Hue[0.9863466096307457, 1, 1], 
      Hue[0.9863466096307457, 1, 1], 
      Hue[0.9869388989572909, 1, 1], 
      Hue[0.9869388989572909, 1, 1], 
      Hue[0.9881234776103813, 1, 1], 
      Hue[0.9881234776103813, 1, 1], 
      Hue[0.988419622273654, 1, 1], 
      Hue[0.988419622273654, 1, 1], 
      Hue[0.9887157669369265, 1, 1], 
      Hue[0.9887157669369265, 1, 1], 
      Hue[0.9893080562634717, 1, 1], 
      Hue[0.9893080562634717, 1, 1], 
      Hue[0.9904926349165621, 1, 1], 
      Hue[0.9904926349165621, 1, 1], 
      Hue[0.9907829640397726, 1, 1], 
      Hue[0.9907829640397726, 1, 1], 
      Hue[0.991073293162983, 1, 1], 
      Hue[0.991073293162983, 1, 1], 
      Hue[0.9916539514094037, 1, 1], 
      Hue[0.9916539514094037, 1, 1], 
      Hue[0.9919442805326142, 1, 1], 
      Hue[0.9919442805326142, 1, 1], 
      Hue[0.9922346096558246, 1, 1], 
      Hue[0.9922346096558246, 1, 1], 
      Hue[0.9928152679022454, 1, 1], 
      Hue[0.9928152679022454, 1, 1], 
      Hue[0.9931055970254558, 1, 1], 
      Hue[0.9931055970254558, 1, 1], 
      Hue[0.9933959261486661, 1, 1], 
      Hue[0.9933959261486661, 1, 1], 
      Hue[0.993976584395087, 1, 1], 
      Hue[0.993976584395087, 1, 1], 
      Hue[0.9951379008879285, 1, 1], 
      Hue[0.9951379008879285, 1, 1], 
      Hue[0.9954282300111389, 1, 1], 
      Hue[0.9954282300111389, 1, 1], 
      Hue[0.9957185591343494, 1, 1], 
      Hue[0.9957185591343494, 1, 1], 
      Hue[0.9962992173807701, 1, 1], 
      Hue[0.9962992173807701, 1, 1], 
      Hue[0.9965895465039806, 1, 1], 
      Hue[0.9965895465039806, 1, 1], 
      Hue[0.9968798756271909, 1, 1], 
      Hue[0.9968798756271909, 1, 1], 
      Hue[0.9974605338736118, 1, 1], 
      Hue[0.9974605338736118, 1, 1], 
      Hue[0.9977508629968221, 1, 1], 
      Hue[0.9977508629968221, 1, 1], 
      Hue[0.9980411921200325, 1, 1], 
      Hue[0.9980411921200325, 1, 1], 
      Hue[0.9983315212432429, 1, 1], 
      Hue[0.9983315212432429, 1, 1], 
      Hue[0.9986218503664533, 1, 1], 
      Hue[0.9986218503664533, 1, 1], 
      Hue[0.9989121794896637, 1, 1], 
      Hue[0.9989121794896637, 1, 1], 
      Hue[0.9992025086128741, 1, 1], 
      Hue[0.9992025086128741, 1, 1], 
      Hue[0.9994928377360845, 1, 1], 
      Hue[0.9994928377360845, 1, 1], 
      Hue[0.9997831668592949, 1, 1], 
      Hue[0.9997831668592949, 1, 1], 
      Hue[0.00007349598250531426, 1, 1], 
      Hue[0.00007349598250531426, 1, 1], 
      Hue[0.00036382510571571257`, 1, 1], 
      Hue[0.00036382510571571257`, 1, 1], 
      Hue[0.000654154228926111, 1, 1], 
      Hue[0.000654154228926111, 1, 1], 
      Hue[0.0009444833521365094, 1, 1], 
      Hue[0.0009444833521365094, 1, 1], 
      Hue[0.0012348124753469077`, 1, 1], 
      Hue[0.0012348124753469077`, 1, 1], 
      Hue[0.0015251415985573066`, 1, 1], 
      Hue[0.0015251415985573066`, 1, 1], 
      Hue[0.0018154707217677046`, 1, 1], 
      Hue[0.0018154707217677046`, 1, 1], 
      Hue[0.002105799844978103, 1, 1], 
      Hue[0.002105799844978103, 1, 1], 
      Hue[0.0023961289681885016`, 1, 1], 
      Hue[0.0023961289681885016`, 1, 1], 
      Hue[0.0026864580913989, 1, 1], 
      Hue[0.0026864580913989, 1, 1], 
      Hue[0.0029767872146092976`, 1, 1], 
      Hue[0.0029767872146092976`, 1, 1], 
      Hue[0.0032671163378196967`, 1, 1], 
      Hue[0.0032671163378196967`, 1, 1], 
      Hue[0.003557445461030095, 1, 1], 
      Hue[0.003557445461030095, 1, 1], 
      Hue[0.003847774584240493, 1, 1], 
      Hue[0.003847774584240493, 1, 1], 
      Hue[0.004138103707450892, 1, 1], 
      Hue[0.004138103707450892, 1, 1], 
      Hue[0.004428432830661289, 1, 1], 
      Hue[0.004428432830661289, 1, 1], 
      Hue[0.0047187619538716875`, 1, 1], 
      Hue[0.0047187619538716875`, 1, 1], 
      Hue[0.005009091077082086, 1, 1], 
      Hue[0.005009091077082086, 1, 1], 
      Hue[0.005299420200292486, 1, 1], 
      Hue[0.005299420200292486, 1, 1], 
      Hue[0.005589749323502882, 1, 1], 
      Hue[0.005589749323502882, 1, 1], 
      Hue[0.005880078446713281, 1, 1], 
      Hue[0.005880078446713281, 1, 1], 
      Hue[0.0061704075699236804`, 1, 1], 
      Hue[0.0061704075699236804`, 1, 1], 
      Hue[0.006751065816344477, 1, 1], 
      Hue[0.006751065816344477, 1, 1], 
      Hue[0.007041394939554876, 1, 1], 
      Hue[0.007041394939554876, 1, 1], 
      Hue[0.007331724062765272, 1, 1], 
      Hue[0.007331724062765272, 1, 1], 
      Hue[0.00791238230918607, 1, 1], 
      Hue[0.00791238230918607, 1, 1], 
      Hue[0.008202711432396469, 1, 1], 
      Hue[0.008202711432396469, 1, 1], 
      Hue[0.008493040555606868, 1, 1], 
      Hue[0.008493040555606868, 1, 1], 
      Hue[0.009073698802027665, 1, 1], 
      Hue[0.009073698802027665, 1, 1], 
      Hue[0.009388662948267424, 1, 1], 
      Hue[0.009388662948267424, 1, 1], 
      Hue[0.00970362709450718, 1, 1], 
      Hue[0.00970362709450718, 1, 1], 
      Hue[0.0103335553869867, 1, 1], 
      Hue[0.0103335553869867, 1, 1], 
      Hue[0.010648519533226455`, 1, 1], 
      Hue[0.010648519533226455`, 1, 1], 
      Hue[0.010963483679466215`, 1, 1], 
      Hue[0.010963483679466215`, 1, 1], 
      Hue[0.011593411971945728`, 1, 1], 
      Hue[0.011593411971945728`, 1, 1], 
      Hue[0.011908376118185491`, 1, 1], 
      Hue[0.011908376118185491`, 1, 1], 
      Hue[0.012223340264425247`, 1, 1], 
      Hue[0.012223340264425247`, 1, 1], 
      Hue[0.012853268556904763`, 1, 1], 
      Hue[0.012853268556904763`, 1, 1], 
      Hue[0.014113125141863794`, 1, 1], 
      Hue[0.014113125141863794`, 1, 1], 
      Hue[0.014428089288103553`, 1, 1], 
      Hue[0.014428089288103553`, 1, 1], 
      Hue[0.014743053434343314`, 1, 1], 
      Hue[0.014743053434343314`, 1, 1], 
      Hue[0.01537298172682283, 1, 1], 
      Hue[0.01537298172682283, 1, 1], 
      Hue[0.01663283831178186, 1, 1], 
      Hue[0.01663283831178186, 1, 1], 
      Hue[0.016947802458021622`, 1, 1], 
      Hue[0.016947802458021622`, 1, 1], 
      Hue[0.017262766604261383`, 1, 1], 
      Hue[0.017262766604261383`, 1, 1], 
      Hue[0.017892694896740897`, 1, 1], 
      Hue[0.017892694896740897`, 1, 1], 
      Hue[0.019152551481699927`, 1, 1], 
      Hue[0.019152551481699927`, 1, 1], 
      Hue[0.019467515627939688`, 1, 1], 
      Hue[0.019467515627939688`, 1, 1], 
      Hue[0.019782479774179445`, 1, 1], 
      Hue[0.019782479774179445`, 1, 1], 
      Hue[0.020412408066658963`, 1, 1], 
      Hue[0.020412408066658963`, 1, 1], 
      Hue[0.02167226465161799, 1, 1], 
      Hue[0.02167226465161799, 1, 1], 
      Hue[0.02419197782153607, 1, 1], 
      Hue[0.02419197782153607, 1, 1], 
      Hue[0.02450694196777583, 1, 1], 
      Hue[0.02450694196777583, 1, 1], 
      Hue[0.02482190611401558, 1, 1], 
      Hue[0.02482190611401558, 1, 1], 
      Hue[0.0254518344064951, 1, 1], 
      Hue[0.0254518344064951, 1, 1], 
      Hue[0.026711690991454134`, 1, 1], 
      Hue[0.026711690991454134`, 1, 1], 
      Hue[0.0292314041613722, 1, 1], 
      Hue[0.0292314041613722, 1, 1], 
      Hue[0.029525327486003974`, 1, 1], 
      Hue[0.029525327486003974`, 1, 1], 
      Hue[0.02981925081063575, 1, 1], 
      Hue[0.02981925081063575, 1, 1], 
      Hue[0.030407097459899297`, 1, 1], 
      Hue[0.030407097459899297`, 1, 1], 
      Hue[0.03158279075842641, 1, 1], 
      Hue[0.03158279075842641, 1, 1], 
      Hue[0.033934177355480606`, 1, 1], 
      Hue[0.033934177355480606`, 1, 1], 
      Hue[0.03863695054958902, 1, 1], 
      Hue[0.03863695054958902, 1, 1], 
      Hue[0.0389308738742208, 1, 1], 
      Hue[0.0389308738742208, 1, 1], 
      Hue[0.03922479719885257, 1, 1], 
      Hue[0.03922479719885257, 1, 1], 
      Hue[0.03981264384811612, 1, 1], 
      Hue[0.03981264384811612, 1, 1], 
      Hue[0.04098833714664322, 1, 1], 
      Hue[0.04098833714664322, 1, 1], 
      Hue[0.043339723743697434`, 1, 1], 
      Hue[0.043339723743697434`, 1, 1], 
      Hue[0.048042496937805844`, 1, 1], 
      Hue[0.048042496937805844`, 1, 1], 
      Hue[0.048361055285466976`, 1, 1], 
      Hue[0.048361055285466976`, 1, 1], 
      Hue[0.048679613633128116`, 1, 1], 
      Hue[0.048679613633128116`, 1, 1], 
      Hue[0.049316730328450395`, 1, 1], 
      Hue[0.049316730328450395`, 1, 1], 
      Hue[0.05059096371909494, 1, 1], 
      Hue[0.05059096371909494, 1, 1], 
      Hue[0.05313943050038401, 1, 1], 
      Hue[0.05313943050038401, 1, 1], 
      Hue[0.058236364062962176`, 1, 1], 
      Hue[0.058236364062962176`, 1, 1], 
      Hue[0.05855492241062331, 1, 1], 
      Hue[0.05855492241062331, 1, 1], 
      Hue[0.05887348075828444, 1, 1], 
      Hue[0.05887348075828444, 1, 1], 
      Hue[0.05951059745360673, 1, 1], 
      Hue[0.05951059745360673, 1, 1], 
      Hue[0.06078483084425125, 1, 1], 
      Hue[0.06078483084425125, 1, 1], 
      Hue[0.06333329762554035, 1, 1], 
      Hue[0.06333329762554035, 1, 1], 
      Hue[0.06843023118811851, 1, 1], 
      Hue[0.06843023118811851, 1, 1], 
      Hue[0.06874297399571745, 1, 1], 
      Hue[0.06874297399571745, 1, 1], 
      Hue[0.06905571680331637, 1, 1], 
      Hue[0.06905571680331637, 1, 1], 
      Hue[0.06968120241851426, 1, 1], 
      Hue[0.06968120241851426, 1, 1], 
      Hue[0.07093217364891, 1, 1], 
      Hue[0.07093217364891, 1, 1], 
      Hue[0.07343411610970148, 1, 1], 
      Hue[0.07343411610970148, 1, 1], 
      Hue[0.0737468589173004, 1, 1], 
      Hue[0.0737468589173004, 1, 1], 
      Hue[0.07405960172489932, 1, 1], 
      Hue[0.07405960172489932, 1, 1], 
      Hue[0.07468508734009718, 1, 1], 
      Hue[0.07468508734009718, 1, 1], 
      Hue[0.07593605857049295, 1, 1], 
      Hue[0.07593605857049295, 1, 1], 
      Hue[0.0784380010312844, 1, 1], 
      Hue[0.0784380010312844, 1, 1], 
      Hue[0.07875074383888335, 1, 1], 
      Hue[0.07875074383888335, 1, 1], 
      Hue[0.07906348664648226, 1, 1], 
      Hue[0.07906348664648226, 1, 1], 
      Hue[0.07968897226168013, 1, 1], 
      Hue[0.07968897226168013, 1, 1], 
      Hue[0.08093994349207585, 1, 1], 
      Hue[0.08093994349207585, 1, 1], 
      Hue[0.08125268629967479, 1, 1], 
      Hue[0.08125268629967479, 1, 1], 
      Hue[0.08156542910727374, 1, 1], 
      Hue[0.08156542910727374, 1, 1], 
      Hue[0.08187817191487268, 1, 1], 
      Hue[0.08187817191487268, 1, 1], 
      Hue[0.08219091472247161, 1, 1], 
      Hue[0.08219091472247161, 1, 1], 
      Hue[0.08250365753007054, 1, 1], 
      Hue[0.08250365753007054, 1, 1], 
      Hue[0.0828164003376695, 1, 1], 
      Hue[0.0828164003376695, 1, 1], 
      Hue[0.08312914314526841, 1, 1], 
      Hue[0.08312914314526841, 1, 1], 
      Hue[0.5834418859528673, 1, 1], 
      Hue[0.5834418859528673, 1, 1], 
      Hue[0.5837546287604662, 1, 1], 
      Hue[0.5837546287604662, 1, 1], 
      Hue[0.5840673715680651, 1, 1], 
      Hue[0.5840673715680651, 1, 1], 
      Hue[0.584692857183263, 1, 1], 
      Hue[0.584692857183263, 1, 1], 
      Hue[0.5859438284136588, 1, 1], 
      Hue[0.5859438284136588, 1, 1], 
      Hue[0.5884457708744503, 1, 1], 
      Hue[0.5884457708744503, 1, 1], 
      Hue[0.5887374728604412, 1, 1], 
      Hue[0.5887374728604412, 1, 1], 
      Hue[0.5890291748464322, 1, 1], 
      Hue[0.5890291748464322, 1, 1], 
      Hue[0.5896125788184141, 1, 1], 
      Hue[0.5896125788184141, 1, 1], 
      Hue[0.5907793867623778, 1, 1], 
      Hue[0.5907793867623778, 1, 1], 
      Hue[0.5931130026503055, 1, 1], 
      Hue[0.5931130026503055, 1, 1], 
      Hue[0.5977802344261607, 1, 1], 
      Hue[0.5977802344261607, 1, 1], 
      Hue[0.5980719364121516, 1, 1], 
      Hue[0.5980719364121516, 1, 1], 
      Hue[0.5983636383981427, 1, 1], 
      Hue[0.5983636383981427, 1, 1], 
      Hue[0.5989470423701245, 1, 1], 
      Hue[0.5989470423701245, 1, 1], 
      Hue[0.6001138503140884, 1, 1], 
      Hue[0.6001138503140884, 1, 1], 
      Hue[0.602447466202016, 1, 1], 
      Hue[0.602447466202016, 1, 1], 
      Hue[0.6071146979778712, 1, 1], 
      Hue[0.6071146979778712, 1, 1], 
      Hue[0.6074310349868914, 1, 1], 
      Hue[0.6074310349868914, 1, 1], 
      Hue[0.6077473719959118, 1, 1], 
      Hue[0.6077473719959118, 1, 1], 
      Hue[0.6083800460139523, 1, 1], 
      Hue[0.6083800460139523, 1, 1], 
      Hue[0.6096453940500337, 1, 1], 
      Hue[0.6096453940500337, 1, 1], 
      Hue[0.6121760901221962, 1, 1], 
      Hue[0.6121760901221962, 1, 1], 
      Hue[0.6172374822665212, 1, 1], 
      Hue[0.6172374822665212, 1, 1], 
      Hue[0.6175538192755414, 1, 1], 
      Hue[0.6175538192755414, 1, 1], 
      Hue[0.6178701562845618, 1, 1], 
      Hue[0.6178701562845618, 1, 1], 
      Hue[0.6185028303026023, 1, 1], 
      Hue[0.6185028303026023, 1, 1], 
      Hue[0.6197681783386836, 1, 1], 
      Hue[0.6197681783386836, 1, 1], 
      Hue[0.6222988744108461, 1, 1], 
      Hue[0.6222988744108461, 1, 1], 
      Hue[0.6273602665551711, 1, 1], 
      Hue[0.6273602665551711, 1, 1], 
      Hue[0.6276555627425835, 1, 1], 
      Hue[0.6276555627425835, 1, 1], 
      Hue[0.6279508589299957, 1, 1], 
      Hue[0.6279508589299957, 1, 1], 
      Hue[0.6285414513048204, 1, 1], 
      Hue[0.6285414513048204, 1, 1], 
      Hue[0.6297226360544697, 1, 1], 
      Hue[0.6297226360544697, 1, 1], 
      Hue[0.6320850055537683, 1, 1], 
      Hue[0.6320850055537683, 1, 1], 
      Hue[0.6323803017411807, 1, 1], 
      Hue[0.6323803017411807, 1, 1], 
      Hue[0.632675597928593, 1, 1], 
      Hue[0.632675597928593, 1, 1], 
      Hue[0.6332661903034177, 1, 1], 
      Hue[0.6332661903034177, 1, 1], 
      Hue[0.634447375053067, 1, 1], 
      Hue[0.634447375053067, 1, 1], 
      Hue[0.6368097445523656, 1, 1], 
      Hue[0.6368097445523656, 1, 1], 
      Hue[0.637105040739778, 1, 1], 
      Hue[0.637105040739778, 1, 1], 
      Hue[0.6374003369271903, 1, 1], 
      Hue[0.6374003369271903, 1, 1], 
      Hue[0.637990929302015, 1, 1], 
      Hue[0.637990929302015, 1, 1], 
      Hue[0.6391721140516642, 1, 1], 
      Hue[0.6391721140516642, 1, 1], 
      Hue[0.6394674102390766, 1, 1], 
      Hue[0.6394674102390766, 1, 1], 
      Hue[0.6397627064264888, 1, 1], 
      Hue[0.6397627064264888, 1, 1], 
      Hue[0.6403532988013136, 1, 1], 
      Hue[0.6403532988013136, 1, 1], 
      Hue[0.6415344835509629, 1, 1], 
      Hue[0.6415344835509629, 1, 1], 
      Hue[0.6418297797383752, 1, 1], 
      Hue[0.6418297797383752, 1, 1], 
      Hue[0.6421250759257875, 1, 1], 
      Hue[0.6421250759257875, 1, 1], 
      Hue[0.6427156683006122, 1, 1], 
      Hue[0.6427156683006122, 1, 1], 
      Hue[0.6438968530502616, 1, 1], 
      Hue[0.6438968530502616, 1, 1], 
      Hue[0.6441921492376739, 1, 1], 
      Hue[0.6441921492376739, 1, 1], 
      Hue[0.6444874454250862, 1, 1], 
      Hue[0.6444874454250862, 1, 1], 
      Hue[0.6450780377999108, 1, 1], 
      Hue[0.6450780377999108, 1, 1], 
      Hue[0.64625922254956, 1, 1], 
      Hue[0.64625922254956, 1, 1], 
      Hue[0.6465487031969103, 1, 1], 
      Hue[0.6465487031969103, 1, 1], 
      Hue[0.6468381838442604, 1, 1], 
      Hue[0.6468381838442604, 1, 1], 
      Hue[0.6474171451389605, 1, 1], 
      Hue[0.6474171451389605, 1, 1], 
      Hue[0.6477066257863107, 1, 1], 
      Hue[0.6477066257863107, 1, 1], 
      Hue[0.6479961064336609, 1, 1], 
      Hue[0.6479961064336609, 1, 1], 
      Hue[0.6485750677283612, 1, 1], 
      Hue[0.6485750677283612, 1, 1], 
      Hue[0.6488645483757112, 1, 1], 
      Hue[0.6488645483757112, 1, 1], 
      Hue[0.6491540290230614, 1, 1], 
      Hue[0.6491540290230614, 1, 1], 
      Hue[0.6497329903177617, 1, 1], 
      Hue[0.6497329903177617, 1, 1], 
      Hue[0.6508909129071622, 1, 1], 
      Hue[0.6508909129071622, 1, 1], 
      Hue[0.6511803935545123, 1, 1], 
      Hue[0.6511803935545123, 1, 1], 
      Hue[0.6514698742018624, 1, 1], 
      Hue[0.6514698742018624, 1, 1], 
      Hue[0.6520488354965627, 1, 1], 
      Hue[0.6520488354965627, 1, 1], 
      Hue[0.6523383161439128, 1, 1], 
      Hue[0.6523383161439128, 1, 1], 
      Hue[0.6526277967912629, 1, 1], 
      Hue[0.6526277967912629, 1, 1], 
      Hue[0.6532067580859631, 1, 1], 
      Hue[0.6532067580859631, 1, 1], 
      Hue[0.6534962387333133, 1, 1], 
      Hue[0.6534962387333133, 1, 1], 
      Hue[0.6537857193806634, 1, 1], 
      Hue[0.6537857193806634, 1, 1], 
      Hue[0.6540752000280136, 1, 1], 
      Hue[0.6540752000280136, 1, 1], 
      Hue[0.6543646806753636, 1, 1], 
      Hue[0.6543646806753636, 1, 1], 
      Hue[0.6546541613227137, 1, 1], 
      Hue[0.6546541613227137, 1, 1], 
      Hue[0.6549436419700639, 1, 1], 
      Hue[0.6549436419700639, 1, 1], 
      Hue[0.6552331226174141, 1, 1], 
      Hue[0.6552331226174141, 1, 1], 
      Hue[0.6555226032647642, 1, 1], 
      Hue[0.6555226032647642, 1, 1], 
      Hue[0.6558120839121143, 1, 1], 
      Hue[0.6558120839121143, 1, 1], 
      Hue[0.6561015645594643, 1, 1], 
      Hue[0.6561015645594643, 1, 1], 
      Hue[0.6563910452068146, 1, 1], 
      Hue[0.6563910452068146, 1, 1], 
      Hue[0.6566805258541647, 1, 1], 
      Hue[0.6566805258541647, 1, 1], 
      Hue[0.6569700065015148, 1, 1], 
      Hue[0.6569700065015148, 1, 1], 
      Hue[0.6572594871488651, 1, 1], 
      Hue[0.6572594871488651, 1, 1], 
      Hue[0.6575489677962151, 1, 1], 
      Hue[0.6575489677962151, 1, 1], 
      Hue[0.6578384484435652, 1, 1], 
      Hue[0.6578384484435652, 1, 1], 
      Hue[0.6581279290909153, 1, 1], 
      Hue[0.6581279290909153, 1, 1], 
      Hue[0.6584174097382655, 1, 1], 
      Hue[0.6584174097382655, 1, 1], 
      Hue[0.6587068903856157, 1, 1], 
      Hue[0.6587068903856157, 1, 1], 
      Hue[0.6589963710329658, 1, 1], 
      Hue[0.6589963710329658, 1, 1], 
      Hue[0.6592858516803158, 1, 1], 
      Hue[0.6592858516803158, 1, 1], 
      Hue[0.659575332327666, 1, 1], 
      Hue[0.659575332327666, 1, 1], 
      Hue[0.6598648129750161, 1, 1], 
      Hue[0.6598648129750161, 1, 1], 
      Hue[0.6601542936223663, 1, 1], 
      Hue[0.6601542936223663, 1, 1], 
      Hue[0.6604437742697165, 1, 1], 
      Hue[0.6604437742697165, 1, 1], 
      Hue[0.6607332549170665, 1, 1], 
      Hue[0.6607332549170665, 1, 1], 
      Hue[0.6610227355644166, 1, 1], 
      Hue[0.6610227355644166, 1, 1], 
      Hue[0.6613122162117667, 1, 1], 
      Hue[0.6613122162117667, 1, 1], 
      Hue[0.661601696859117, 1, 1], 
      Hue[0.661601696859117, 1, 1], 
      Hue[0.661891177506467, 1, 1], 
      Hue[0.661891177506467, 1, 1], 
      Hue[0.6624701388011672, 1, 1], 
      Hue[0.6624701388011672, 1, 1], 
      Hue[0.6627596194485174, 1, 1], 
      Hue[0.6627596194485174, 1, 1], 
      Hue[0.6630491000958674, 1, 1], 
      Hue[0.6630491000958674, 1, 1], 
      Hue[0.6636280613905677, 1, 1], 
      Hue[0.6636280613905677, 1, 1], 
      Hue[0.6639175420379179, 1, 1], 
      Hue[0.6639175420379179, 1, 1], 
      Hue[0.664207022685268, 1, 1], 
      Hue[0.664207022685268, 1, 1], 
      Hue[0.6647859839799682, 1, 1], 
      Hue[0.6647859839799682, 1, 1], 
      Hue[0.6651000996503478, 1, 1], 
      Hue[0.6651000996503478, 1, 1], 
      Hue[0.6654142153207272, 1, 1], 
      Hue[0.6654142153207272, 1, 1], 
      Hue[0.6660424466614863, 1, 1], 
      Hue[0.6660424466614863, 1, 1], 
      Hue[0.6663565623318657, 1, 1], 
      Hue[0.6663565623318657, 1, 1], 
      Hue[0.6666706780022451, 1, 1], 
      Hue[0.6666706780022451, 1, 1], 
      Hue[0.6672989093430042, 1, 1], 
      Hue[0.6672989093430042, 1, 1], 
      Hue[0.6676130250133836, 1, 1], 
      Hue[0.6676130250133836, 1, 1], 
      Hue[0.6679271406837631, 1, 1], 
      Hue[0.6679271406837631, 1, 1], 
      Hue[0.6685553720245221, 1, 1], 
      Hue[0.6685553720245221, 1, 1], 
      Hue[0.66981183470604, 1, 1], 
      Hue[0.66981183470604, 1, 1], 
      Hue[0.6701259503764195, 1, 1], 
      Hue[0.6701259503764195, 1, 1], 
      Hue[0.6704400660467991, 1, 1], 
      Hue[0.6704400660467991, 1, 1], 
      Hue[0.671068297387558, 1, 1], 
      Hue[0.671068297387558, 1, 1], 
      Hue[0.6723247600690759, 1, 1], 
      Hue[0.6723247600690759, 1, 1], 
      Hue[0.6748376854321119, 1, 1], 
      Hue[0.6748376854321119, 1, 1], 
      Hue[0.6751518011024913, 1, 1], 
      Hue[0.6751518011024913, 1, 1], 
      Hue[0.6754659167728708, 1, 1], 
      Hue[0.6754659167728708, 1, 1], 
      Hue[0.6760941481136298, 1, 1], 
      Hue[0.6760941481136298, 1, 1], 
      Hue[0.6773506107951477, 1, 1], 
      Hue[0.6773506107951477, 1, 1], 
      Hue[0.6798635361581836, 1, 1], 
      Hue[0.6798635361581836, 1, 1], 
      Hue[0.680177651828563, 1, 1], 
      Hue[0.680177651828563, 1, 1], 
      Hue[0.6804917674989426, 1, 1], 
      Hue[0.6804917674989426, 1, 1], 
      Hue[0.6811199988397015, 1, 1], 
      Hue[0.6811199988397015, 1, 1], 
      Hue[0.6823764615212196, 1, 1], 
      Hue[0.6823764615212196, 1, 1], 
      Hue[0.6848893868842554, 1, 1], 
      Hue[0.6848893868842554, 1, 1], 
      Hue[0.6851824617330269, 1, 1], 
      Hue[0.6851824617330269, 1, 1], 
      Hue[0.6854755365817984, 1, 1], 
      Hue[0.6854755365817984, 1, 1], 
      Hue[0.6860616862793414, 1, 1], 
      Hue[0.6860616862793414, 1, 1], 
      Hue[0.6872339856744274, 1, 1], 
      Hue[0.6872339856744274, 1, 1], 
      Hue[0.6895785844645994, 1, 1], 
      Hue[0.6895785844645994, 1, 1], 
      Hue[0.6942677820449434, 1, 1], 
      Hue[0.6942677820449434, 1, 1], 
      Hue[0.6945608568937149, 1, 1], 
      Hue[0.6945608568937149, 1, 1], 
      Hue[0.6948539317424864, 1, 1], 
      Hue[0.6948539317424864, 1, 1], 
      Hue[0.6954400814400294, 1, 1], 
      Hue[0.6954400814400294, 1, 1], 
      Hue[0.6966123808351155, 1, 1], 
      Hue[0.6966123808351155, 1, 1], 
      Hue[0.6989569796252876, 1, 1], 
      Hue[0.6989569796252876, 1, 1], 
      Hue[0.7036461772056316, 1, 1], 
      Hue[0.7036461772056316, 1, 1], 
      Hue[0.7039638870774325, 1, 1], 
      Hue[0.7039638870774325, 1, 1], 
      Hue[0.7042815969492333, 1, 1], 
      Hue[0.7042815969492333, 1, 1], 
      Hue[0.704917016692835, 1, 1], 
      Hue[0.704917016692835, 1, 1], 
      Hue[0.7061878561800385, 1, 1], 
      Hue[0.7061878561800385, 1, 1], 
      Hue[0.7087295351544455, 1, 1], 
      Hue[0.7087295351544455, 1, 1], 
      Hue[0.7138128931032592, 1, 1], 
      Hue[0.7138128931032592, 1, 1], 
      Hue[0.723979609000887, 1, 1], 
      Hue[0.723979609000887, 1, 1], 
      Hue[0.7242762780510799, 1, 1], 
      Hue[0.7242762780510799, 1, 1], 
      Hue[0.7245729471012727, 1, 1], 
      Hue[0.7245729471012727, 1, 1], 
      Hue[0.7251662852016585, 1, 1], 
      Hue[0.7251662852016585, 1, 1], 
      Hue[0.72635296140243, 1, 1], 
      Hue[0.72635296140243, 1, 1], 
      Hue[0.728726313803973, 1, 1], 
      Hue[0.728726313803973, 1, 1], 
      Hue[0.7334730186070592, 1, 1], 
      Hue[0.7334730186070592, 1, 1], 
      Hue[0.7337696876572521, 1, 1], 
      Hue[0.7337696876572521, 1, 1], 
      Hue[0.7340663567074449, 1, 1], 
      Hue[0.7340663567074449, 1, 1], 
      Hue[0.7346596948078307, 1, 1], 
      Hue[0.7346596948078307, 1, 1], 
      Hue[0.7358463710086022, 1, 1], 
      Hue[0.7358463710086022, 1, 1], 
      Hue[0.7382197234101453, 1, 1], 
      Hue[0.7382197234101453, 1, 1], 
      Hue[0.7429664282132313, 1, 1], 
      Hue[0.7429664282132313, 1, 1], 
      Hue[0.743257281723362, 1, 1], 
      Hue[0.743257281723362, 1, 1], 
      Hue[0.7435481352334927, 1, 1], 
      Hue[0.7435481352334927, 1, 1], 
      Hue[0.7441298422537541, 1, 1], 
      Hue[0.7441298422537541, 1, 1], 
      Hue[0.7452932562942768, 1, 1], 
      Hue[0.7452932562942768, 1, 1], 
      Hue[0.7455841098044075, 1, 1], 
      Hue[0.7455841098044075, 1, 1], 
      Hue[0.7458749633145382, 1, 1], 
      Hue[0.7458749633145382, 1, 1], 
      Hue[0.7464566703347995, 1, 1], 
      Hue[0.7464566703347995, 1, 1], 
      Hue[0.7476200843753222, 1, 1], 
      Hue[0.7476200843753222, 1, 1], 
      Hue[0.7479109378854529, 1, 1], 
      Hue[0.7479109378854529, 1, 1], 
      Hue[0.7482017913955836, 1, 1], 
      Hue[0.7482017913955836, 1, 1], 
      Hue[0.748783498415845, 1, 1], 
      Hue[0.748783498415845, 1, 1], 
      Hue[0.7490743519259756, 1, 1], 
      Hue[0.7490743519259756, 1, 1], 
      Hue[0.7493652054361063, 1, 1], 
      Hue[0.7493652054361063, 1, 1], 
      Hue[0.749656058946237, 1, 1], 
      Hue[0.749656058946237, 1, 1], 
      Hue[0.7499469124563677, 1, 1], 
      Hue[0.7499469124563677, 1, 1], 
      Hue[0.2502377659664984, 1, 1], 
      Hue[0.2502377659664984, 1, 1], 
      Hue[0.2505286194766291, 1, 1], 
      Hue[0.2505286194766291, 1, 1], 
      Hue[0.2508194729867598, 1, 1], 
      Hue[0.2508194729867598, 1, 1], 
      Hue[0.2511103264968904, 1, 1], 
      Hue[0.2511103264968904, 1, 1], 
      Hue[0.25227374053741314`, 1, 1], 
      Hue[0.25227374053741314`, 1, 1], 
      Hue[0.25256459404754383`, 1, 1], 
      Hue[0.25256459404754383`, 1, 1], 
      Hue[0.25285544755767453`, 1, 1], 
      Hue[0.25285544755767453`, 1, 1], 
      Hue[0.25343715457793586`, 1, 1], 
      Hue[0.25343715457793586`, 1, 1], 
      Hue[0.2546005686184586, 1, 1], 
      Hue[0.2546005686184586, 1, 1], 
      Hue[0.25692739669950404`, 1, 1], 
      Hue[0.25692739669950404`, 1, 1], 
      Hue[0.2615810528615949, 1, 1], 
      Hue[0.2615810528615949, 1, 1], 
      Hue[0.26189654139475493`, 1, 1], 
      Hue[0.26189654139475493`, 1, 1], 
      Hue[0.2622120299279149, 1, 1], 
      Hue[0.2622120299279149, 1, 1], 
      Hue[0.262843006994235, 1, 1], 
      Hue[0.262843006994235, 1, 1], 
      Hue[0.2641049611268752, 1, 1], 
      Hue[0.2641049611268752, 1, 1], 
      Hue[0.2666288693921555, 1, 1], 
      Hue[0.2666288693921555, 1, 1], 
      Hue[0.26694435792531557`, 1, 1], 
      Hue[0.26694435792531557`, 1, 1], 
      Hue[0.2672598464584756, 1, 1], 
      Hue[0.2672598464584756, 1, 1], 
      Hue[0.26789082352479565`, 1, 1], 
      Hue[0.26789082352479565`, 1, 1], 
      Hue[0.26915277765743584`, 1, 1], 
      Hue[0.26915277765743584`, 1, 1], 
      Hue[0.27167668592271615`, 1, 1], 
      Hue[0.27167668592271615`, 1, 1], 
      Hue[0.2719921744558762, 1, 1], 
      Hue[0.2719921744558762, 1, 1], 
      Hue[0.2723076629890362, 1, 1], 
      Hue[0.2723076629890362, 1, 1], 
      Hue[0.2729386400553563, 1, 1], 
      Hue[0.2729386400553563, 1, 1], 
      Hue[0.27420059418799647`, 1, 1], 
      Hue[0.27420059418799647`, 1, 1], 
      Hue[0.2767245024532768, 1, 1], 
      Hue[0.2767245024532768, 1, 1], 
      Hue[0.2817723189838374, 1, 1], 
      Hue[0.2817723189838374, 1, 1], 
      Hue[0.2820667666953895, 1, 1], 
      Hue[0.2820667666953895, 1, 1], 
      Hue[0.28236121440694156`, 1, 1], 
      Hue[0.28236121440694156`, 1, 1], 
      Hue[0.28295010983004565`, 1, 1], 
      Hue[0.28295010983004565`, 1, 1], 
      Hue[0.2841279006762539, 1, 1], 
      Hue[0.2841279006762539, 1, 1], 
      Hue[0.28648348236867033`, 1, 1], 
      Hue[0.28648348236867033`, 1, 1], 
      Hue[0.2867779300802224, 1, 1], 
      Hue[0.2867779300802224, 1, 1], 
      Hue[0.28707237779177447`, 1, 1], 
      Hue[0.28707237779177447`, 1, 1], 
      Hue[0.28766127321487855`, 1, 1], 
      Hue[0.28766127321487855`, 1, 1], 
      Hue[0.2888390640610868, 1, 1], 
      Hue[0.2888390640610868, 1, 1], 
      Hue[0.29119464575350323`, 1, 1], 
      Hue[0.29119464575350323`, 1, 1], 
      Hue[0.2914890934650553, 1, 1], 
      Hue[0.2914890934650553, 1, 1], 
      Hue[0.2917835411766074, 1, 1], 
      Hue[0.2917835411766074, 1, 1], 
      Hue[0.29237243659971146`, 1, 1], 
      Hue[0.29237243659971146`, 1, 1], 
      Hue[0.2935502274459197, 1, 1], 
      Hue[0.2935502274459197, 1, 1], 
      Hue[0.29384467515747176`, 1, 1], 
      Hue[0.29384467515747176`, 1, 1], 
      Hue[0.29413912286902383`, 1, 1], 
      Hue[0.29413912286902383`, 1, 1], 
      Hue[0.2947280182921279, 1, 1], 
      Hue[0.2947280182921279, 1, 1], 
      Hue[0.29590580913833614`, 1, 1], 
      Hue[0.29590580913833614`, 1, 1], 
      Hue[0.2962002568498882, 1, 1], 
      Hue[0.2962002568498882, 1, 1], 
      Hue[0.2964947045614402, 1, 1], 
      Hue[0.2964947045614402, 1, 1], 
      Hue[0.29708359998454437`, 1, 1], 
      Hue[0.29708359998454437`, 1, 1], 
      Hue[0.2982613908307526, 1, 1], 
      Hue[0.2982613908307526, 1, 1], 
      Hue[0.29855583854230466`, 1, 1], 
      Hue[0.29855583854230466`, 1, 1], 
      Hue[0.2988502862538567, 1, 1], 
      Hue[0.2988502862538567, 1, 1], 
      Hue[0.2994391816769608, 1, 1], 
      Hue[0.2994391816769608, 1, 1], 
      Hue[0.30061697252316905`, 1, 1], 
      Hue[0.30061697252316905`, 1, 1], 
      Hue[0.30093605525775047`, 1, 1], 
      Hue[0.30093605525775047`, 1, 1], 
      Hue[0.3012551379923319, 1, 1], 
      Hue[0.3012551379923319, 1, 1], 
      Hue[0.30189330346149473`, 1, 1], 
      Hue[0.30189330346149473`, 1, 1], 
      Hue[0.30221238619607615`, 1, 1], 
      Hue[0.30221238619607615`, 1, 1], 
      Hue[0.3025314689306576, 1, 1], 
      Hue[0.3025314689306576, 1, 1], 
      Hue[0.3031696343998204, 1, 1], 
      Hue[0.3031696343998204, 1, 1], 
      Hue[0.30348871713440184`, 1, 1], 
      Hue[0.30348871713440184`, 1, 1], 
      Hue[0.3038077998689832, 1, 1], 
      Hue[0.3038077998689832, 1, 1], 
      Hue[0.30444596533814605`, 1, 1], 
      Hue[0.30444596533814605`, 1, 1], 
      Hue[0.30572229627647174`, 1, 1], 
      Hue[0.30572229627647174`, 1, 1], 
      Hue[0.30604137901105316`, 1, 1], 
      Hue[0.30604137901105316`, 1, 1], 
      Hue[0.3063604617456346, 1, 1], 
      Hue[0.3063604617456346, 1, 1], 
      Hue[0.3069986272147974, 1, 1], 
      Hue[0.3069986272147974, 1, 1], 
      Hue[0.30731770994937885`, 1, 1], 
      Hue[0.30731770994937885`, 1, 1], 
      Hue[0.3076367926839602, 1, 1], 
      Hue[0.3076367926839602, 1, 1], 
      Hue[0.30827495815312306`, 1, 1], 
      Hue[0.30827495815312306`, 1, 1], 
      Hue[0.3085940408877045, 1, 1], 
      Hue[0.3085940408877045, 1, 1], 
      Hue[0.3089131236222859, 1, 1], 
      Hue[0.3089131236222859, 1, 1], 
      Hue[0.3092322063568673, 1, 1], 
      Hue[0.3092322063568673, 1, 1], 
      Hue[0.30955128909144874`, 1, 1], 
      Hue[0.30955128909144874`, 1, 1], 
      Hue[0.30987037182603017`, 1, 1], 
      Hue[0.30987037182603017`, 1, 1], 
      Hue[0.3101894545606116, 1, 1], 
      Hue[0.3101894545606116, 1, 1], 
      Hue[0.310508537295193, 1, 1], 
      Hue[0.310508537295193, 1, 1], 
      Hue[0.31082762002977443`, 1, 1], 
      Hue[0.31082762002977443`, 1, 1], 
      Hue[0.31114670276435585`, 1, 1], 
      Hue[0.31114670276435585`, 1, 1], 
      Hue[0.3114657854989372, 1, 1], 
      Hue[0.3114657854989372, 1, 1], 
      Hue[0.3117848682335187, 1, 1], 
      Hue[0.3117848682335187, 1, 1], 
      Hue[0.31210395096810006`, 1, 1], 
      Hue[0.31210395096810006`, 1, 1], 
      Hue[0.3124230337026815, 1, 1], 
      Hue[0.3124230337026815, 1, 1], 
      Hue[0.3127421164372629, 1, 1], 
      Hue[0.3127421164372629, 1, 1], 
      Hue[0.31306119917184433`, 1, 1], 
      Hue[0.31306119917184433`, 1, 1], 
      Hue[0.31338028190642575`, 1, 1], 
      Hue[0.31338028190642575`, 1, 1], 
      Hue[0.3136993646410072, 1, 1], 
      Hue[0.3136993646410072, 1, 1], 
      Hue[0.3140184473755886, 1, 1], 
      Hue[0.3140184473755886, 1, 1], 
      Hue[0.31433753011017, 1, 1], 
      Hue[0.31433753011017, 1, 1], 
      Hue[0.31465661284475144`, 1, 1], 
      Hue[0.31465661284475144`, 1, 1], 
      Hue[0.31497569557933286`, 1, 1], 
      Hue[0.31497569557933286`, 1, 1], 
      Hue[0.3152947783139142, 1, 1], 
      Hue[0.3152947783139142, 1, 1], 
      Hue[0.3156138610484957, 1, 1], 
      Hue[0.3156138610484957, 1, 1], 
      Hue[0.31593294378307707`, 1, 1], 
      Hue[0.31593294378307707`, 1, 1], 
      Hue[0.3162520265176585, 1, 1], 
      Hue[0.3162520265176585, 1, 1], 
      Hue[0.3165711092522399, 1, 1], 
      Hue[0.3165711092522399, 1, 1], 
      Hue[0.31689019198682133`, 1, 1], 
      Hue[0.31689019198682133`, 1, 1], 
      Hue[0.31720927472140276`, 1, 1], 
      Hue[0.31720927472140276`, 1, 1], 
      Hue[0.3175283574559842, 1, 1], 
      Hue[0.3175283574559842, 1, 1], 
      Hue[0.3178474401905656, 1, 1], 
      Hue[0.3178474401905656, 1, 1], 
      Hue[0.31848560565972844`, 1, 1], 
      Hue[0.31848560565972844`, 1, 1], 
      Hue[0.31880468839430987`, 1, 1], 
      Hue[0.31880468839430987`, 1, 1], 
      Hue[0.31912377112889123`, 1, 1], 
      Hue[0.31912377112889123`, 1, 1], 
      Hue[0.31976193659805413`, 1, 1], 
      Hue[0.31976193659805413`, 1, 1], 
      Hue[0.3200810193326355, 1, 1], 
      Hue[0.3200810193326355, 1, 1], 
      Hue[0.320400102067217, 1, 1], 
      Hue[0.320400102067217, 1, 1], 
      Hue[0.32103826753637976`, 1, 1], 
      Hue[0.32103826753637976`, 1, 1], 
      Hue[0.321351534730899, 1, 1], 
      Hue[0.321351534730899, 1, 1], 
      Hue[0.32166480192541824`, 1, 1], 
      Hue[0.32166480192541824`, 1, 1], 
      Hue[0.3222913363144566, 1, 1], 
      Hue[0.3222913363144566, 1, 1], 
      Hue[0.32354440509253346`, 1, 1], 
      Hue[0.32354440509253346`, 1, 1], 
      Hue[0.3238576722870527, 1, 1], 
      Hue[0.3238576722870527, 1, 1], 
      Hue[0.32417093948157194`, 1, 1], 
      Hue[0.32417093948157194`, 1, 1], 
      Hue[0.32479747387061036`, 1, 1], 
      Hue[0.32479747387061036`, 1, 1], 
      Hue[0.3260505426486872, 1, 1], 
      Hue[0.3260505426486872, 1, 1], 
      Hue[0.32636380984320645`, 1, 1], 
      Hue[0.32636380984320645`, 1, 1], 
      Hue[0.3266770770377257, 1, 1], 
      Hue[0.3266770770377257, 1, 1], 
      Hue[0.32730361142676406`, 1, 1], 
      Hue[0.32730361142676406`, 1, 1], 
      Hue[0.32855668020484097`, 1, 1], 
      Hue[0.32855668020484097`, 1, 1], 
      Hue[0.33106281776099467`, 1, 1], 
      Hue[0.33106281776099467`, 1, 1], 
      Hue[0.3313760849555139, 1, 1], 
      Hue[0.3313760849555139, 1, 1], 
      Hue[0.33168935215003315`, 1, 1], 
      Hue[0.33168935215003315`, 1, 1], 
      Hue[0.3323158865390715, 1, 1], 
      Hue[0.3323158865390715, 1, 1], 
      Hue[0.3335689553171484, 1, 1], 
      Hue[0.3335689553171484, 1, 1], 
      Hue[0.3360750928733021, 1, 1], 
      Hue[0.3360750928733021, 1, 1], 
      Hue[0.33638836006782136`, 1, 1], 
      Hue[0.33638836006782136`, 1, 1], 
      Hue[0.3367016272623406, 1, 1], 
      Hue[0.3367016272623406, 1, 1], 
      Hue[0.337328161651379, 1, 1], 
      Hue[0.337328161651379, 1, 1], 
      Hue[0.33858123042945587`, 1, 1], 
      Hue[0.33858123042945587`, 1, 1], 
      Hue[0.34108736798560957`, 1, 1], 
      Hue[0.34108736798560957`, 1, 1], 
      Hue[0.34137959435852083`, 1, 1], 
      Hue[0.34137959435852083`, 1, 1], 
      Hue[0.34167182073143204`, 1, 1], 
      Hue[0.34167182073143204`, 1, 1], 
      Hue[0.3422562734772545, 1, 1], 
      Hue[0.3422562734772545, 1, 1], 
      Hue[0.34342517896889946`, 1, 1], 
      Hue[0.34342517896889946`, 1, 1], 
      Hue[0.34576298995218935`, 1, 1], 
      Hue[0.34576298995218935`, 1, 1], 
      Hue[0.350438611918769, 1, 1], 
      Hue[0.350438611918769, 1, 1], 
      Hue[0.35978985585192846`, 1, 1], 
      Hue[0.35978985585192846`, 1, 1], 
      Hue[0.38006898519212645`, 1, 1], 
      Hue[0.38006898519212645`, 1, 1], 
      Hue[0.38036480576645904`, 1, 1], 
      Hue[0.38036480576645904`, 1, 1], 
      Hue[0.3806606263407917, 1, 1], 
      Hue[0.3806606263407917, 1, 1], 
      Hue[0.3812522674894569, 1, 1], 
      Hue[0.3812522674894569, 1, 1], 
      Hue[0.38243554978678734`, 1, 1], 
      Hue[0.38243554978678734`, 1, 1], 
      Hue[0.38480211438144823`, 1, 1], 
      Hue[0.38480211438144823`, 1, 1], 
      Hue[0.38953524357077, 1, 1], 
      Hue[0.38953524357077, 1, 1], 
      Hue[0.38983106414510266`, 1, 1], 
      Hue[0.38983106414510266`, 1, 1], 
      Hue[0.39012688471943524`, 1, 1], 
      Hue[0.39012688471943524`, 1, 1], 
      Hue[0.39071852586810046`, 1, 1], 
      Hue[0.39071852586810046`, 1, 1], 
      Hue[0.3919018081654309, 1, 1], 
      Hue[0.3919018081654309, 1, 1], 
      Hue[0.39426837276009186`, 1, 1], 
      Hue[0.39426837276009186`, 1, 1], 
      Hue[0.39900150194941353`, 1, 1], 
      Hue[0.39900150194941353`, 1, 1], 
      Hue[0.39929150698368404`, 1, 1], 
      Hue[0.39929150698368404`, 1, 1], 
      Hue[0.39958151201795444`, 1, 1], 
      Hue[0.39958151201795444`, 1, 1], 
      Hue[0.40016152208649525`, 1, 1], 
      Hue[0.40016152208649525`, 1, 1], 
      Hue[0.40132154222357685`, 1, 1], 
      Hue[0.40132154222357685`, 1, 1], 
      Hue[0.4036415824977401, 1, 1], 
      Hue[0.4036415824977401, 1, 1], 
      Hue[0.4039315875320105, 1, 1], 
      Hue[0.4039315875320105, 1, 1], 
      Hue[0.40422159256628093`, 1, 1], 
      Hue[0.40422159256628093`, 1, 1], 
      Hue[0.40480160263482173`, 1, 1], 
      Hue[0.40480160263482173`, 1, 1], 
      Hue[0.40596162277190334`, 1, 1], 
      Hue[0.40596162277190334`, 1, 1], 
      Hue[0.40828166304606667`, 1, 1], 
      Hue[0.40828166304606667`, 1, 1], 
      Hue[0.4085716680803371, 1, 1], 
      Hue[0.4085716680803371, 1, 1], 
      Hue[0.4088616731146075, 1, 1], 
      Hue[0.4088616731146075, 1, 1], 
      Hue[0.40944168318314833`, 1, 1], 
      Hue[0.40944168318314833`, 1, 1], 
      Hue[0.41060170332022994`, 1, 1], 
      Hue[0.41060170332022994`, 1, 1], 
      Hue[0.4108917083545004, 1, 1], 
      Hue[0.4108917083545004, 1, 1], 
      Hue[0.4111817133887708, 1, 1], 
      Hue[0.4111817133887708, 1, 1], 
      Hue[0.4117617234573116, 1, 1], 
      Hue[0.4117617234573116, 1, 1], 
      Hue[0.4129217435943932, 1, 1], 
      Hue[0.4129217435943932, 1, 1], 
      Hue[0.4132117486286636, 1, 1], 
      Hue[0.4132117486286636, 1, 1], 
      Hue[0.413501753662934, 1, 1], 
      Hue[0.413501753662934, 1, 1], 
      Hue[0.4140817637314748, 1, 1], 
      Hue[0.4140817637314748, 1, 1], 
      Hue[0.4143717687657452, 1, 1], 
      Hue[0.4143717687657452, 1, 1], 
      Hue[0.4146617738000156, 1, 1], 
      Hue[0.4146617738000156, 1, 1], 
      Hue[0.4152417838685564, 1, 1], 
      Hue[0.4152417838685564, 1, 1], 
      Hue[0.4155317889028268, 1, 1], 
      Hue[0.4155317889028268, 1, 1], 
      Hue[0.41582179393709723`, 1, 1], 
      Hue[0.41582179393709723`, 1, 1], 
      Hue[0.41611179897136763`, 1, 1], 
      Hue[0.41611179897136763`, 1, 1], 
      Hue[0.4164018040056381, 1, 1], 
      Hue[0.4164018040056381, 1, 1], 
      Hue[0.9166918090399084, 1, 1], 
      Hue[0.9166918090399084, 1, 1], 
      Hue[0.9169818140741789, 1, 1], 
      Hue[0.9169818140741789, 1, 1], 
      Hue[0.9172718191084492, 1, 1], 
      Hue[0.9172718191084492, 1, 1], 
      Hue[0.9175618241427197, 1, 1], 
      Hue[0.9175618241427197, 1, 1], 
      Hue[0.9178764642000194, 1, 1], 
      Hue[0.9178764642000194, 1, 1], 
      Hue[0.9181911042573192, 1, 1], 
      Hue[0.9181911042573192, 1, 1], 
      Hue[0.9188203843719188, 1, 1], 
      Hue[0.9188203843719188, 1, 1], 
      Hue[0.9200789446011179, 1, 1], 
      Hue[0.9200789446011179, 1, 1], 
      Hue[0.922596065059516, 1, 1], 
      Hue[0.922596065059516, 1, 1], 
      Hue[0.9229107051168157, 1, 1], 
      Hue[0.9229107051168157, 1, 1], 
      Hue[0.9232253451741156, 1, 1], 
      Hue[0.9232253451741156, 1, 1], 
      Hue[0.9238546252887151, 1, 1], 
      Hue[0.9238546252887151, 1, 1], 
      Hue[0.9251131855179142, 1, 1], 
      Hue[0.9251131855179142, 1, 1], 
      Hue[0.9276303059763124, 1, 1], 
      Hue[0.9276303059763124, 1, 1], 
      Hue[0.9279449460336121, 1, 1], 
      Hue[0.9279449460336121, 1, 1], 
      Hue[0.928259586090912, 1, 1], 
      Hue[0.928259586090912, 1, 1], 
      Hue[0.9288888662055114, 1, 1], 
      Hue[0.9288888662055114, 1, 1], 
      Hue[0.9301474264347105, 1, 1], 
      Hue[0.9301474264347105, 1, 1], 
      Hue[0.9326645468931086, 1, 1], 
      Hue[0.9326645468931086, 1, 1], 
      Hue[0.9376987878099049, 1, 1], 
      Hue[0.9376987878099049, 1, 1], 
      Hue[0.9379923870455967, 1, 1], 
      Hue[0.9379923870455967, 1, 1], 
      Hue[0.9382859862812885, 1, 1], 
      Hue[0.9382859862812885, 1, 1], 
      Hue[0.938873184752672, 1, 1], 
      Hue[0.938873184752672, 1, 1], 
      Hue[0.9400475816954391, 1, 1], 
      Hue[0.9400475816954391, 1, 1], 
      Hue[0.9423963755809734, 1, 1], 
      Hue[0.9423963755809734, 1, 1], 
      Hue[0.9426899748166653, 1, 1], 
      Hue[0.9426899748166653, 1, 1], 
      Hue[0.9429835740523571, 1, 1], 
      Hue[0.9429835740523571, 1, 1], 
      Hue[0.9435707725237407, 1, 1], 
      Hue[0.9435707725237407, 1, 1], 
      Hue[0.9447451694665078, 1, 1], 
      Hue[0.9447451694665078, 1, 1], 
      Hue[0.9470939633520421, 1, 1], 
      Hue[0.9470939633520421, 1, 1], 
      Hue[0.9473875625877338, 1, 1], 
      Hue[0.9473875625877338, 1, 1], 
      Hue[0.9476811618234257, 1, 1], 
      Hue[0.9476811618234257, 1, 1], 
      Hue[0.9482683602948092, 1, 1], 
      Hue[0.9482683602948092, 1, 1], 
      Hue[0.9494427572375763, 1, 1], 
      Hue[0.9494427572375763, 1, 1], 
      Hue[0.9497363564732682, 1, 1], 
      Hue[0.9497363564732682, 1, 1], 
      Hue[0.9500299557089599, 1, 1], 
      Hue[0.9500299557089599, 1, 1], 
      Hue[0.9506171541803434, 1, 1], 
      Hue[0.9506171541803434, 1, 1], 
      Hue[0.9517915511231106, 1, 1], 
      Hue[0.9517915511231106, 1, 1], 
      Hue[0.9520851503588025, 1, 1], 
      Hue[0.9520851503588025, 1, 1], 
      Hue[0.9523787495944943, 1, 1], 
      Hue[0.9523787495944943, 1, 1], 
      Hue[0.9529659480658779, 1, 1], 
      Hue[0.9529659480658779, 1, 1], 
      Hue[0.954140345008645, 1, 1], 
      Hue[0.954140345008645, 1, 1], 
      Hue[0.9544339442443368, 1, 1], 
      Hue[0.9544339442443368, 1, 1], 
      Hue[0.9547275434800285, 1, 1], 
      Hue[0.9547275434800285, 1, 1], 
      Hue[0.9553147419514121, 1, 1], 
      Hue[0.9553147419514121, 1, 1], 
      Hue[0.9564891388941792, 1, 1], 
      Hue[0.9564891388941792, 1, 1], 
      Hue[0.9568073731529004, 1, 1], 
      Hue[0.9568073731529004, 1, 1], 
      Hue[0.9571256074116216, 1, 1], 
      Hue[0.9571256074116216, 1, 1], 
      Hue[0.9577620759290638, 1, 1], 
      Hue[0.9577620759290638, 1, 1], 
      Hue[0.9580803101877848, 1, 1], 
      Hue[0.9580803101877848, 1, 1], 
      Hue[0.9583985444465061, 1, 1], 
      Hue[0.9583985444465061, 1, 1], 
      Hue[0.9590350129639483, 1, 1], 
      Hue[0.9590350129639483, 1, 1], 
      Hue[0.9593532472226696, 1, 1], 
      Hue[0.9593532472226696, 1, 1], 
      Hue[0.9596714814813907, 1, 1], 
      Hue[0.9596714814813907, 1, 1], 
      Hue[0.960307949998833, 1, 1], 
      Hue[0.960307949998833, 1, 1], 
      Hue[0.9615808870337176, 1, 1], 
      Hue[0.9615808870337176, 1, 1], 
      Hue[0.9618991212924386, 1, 1], 
      Hue[0.9618991212924386, 1, 1], 
      Hue[0.9622173555511598, 1, 1], 
      Hue[0.9622173555511598, 1, 1], 
      Hue[0.9628538240686021, 1, 1], 
      Hue[0.9628538240686021, 1, 1], 
      Hue[0.9631720583273232, 1, 1], 
      Hue[0.9631720583273232, 1, 1], 
      Hue[0.9634902925860445, 1, 1], 
      Hue[0.9634902925860445, 1, 1], 
      Hue[0.9641267611034867, 1, 1], 
      Hue[0.9641267611034867, 1, 1], 
      Hue[0.9644449953622078, 1, 1], 
      Hue[0.9644449953622078, 1, 1], 
      Hue[0.9647632296209291, 1, 1], 
      Hue[0.9647632296209291, 1, 1], 
      Hue[0.9650814638796502, 1, 1], 
      Hue[0.9650814638796502, 1, 1], 
      Hue[0.9653996981383713, 1, 1], 
      Hue[0.9653996981383713, 1, 1], 
      Hue[0.9657179323970924, 1, 1], 
      Hue[0.9657179323970924, 1, 1], 
      Hue[0.9660361666558135, 1, 1], 
      Hue[0.9660361666558135, 1, 1], 
      Hue[0.9666726351732559, 1, 1], 
      Hue[0.9666726351732559, 1, 1], 
      Hue[0.966990869431977, 1, 1], 
      Hue[0.966990869431977, 1, 1], 
      Hue[0.9673091036906982, 1, 1], 
      Hue[0.9673091036906982, 1, 1], 
      Hue[0.9676273379494194, 1, 1], 
      Hue[0.9676273379494194, 1, 1], 
      Hue[0.9679455722081405, 1, 1], 
      Hue[0.9679455722081405, 1, 1], 
      Hue[0.9682638064668616, 1, 1], 
      Hue[0.9682638064668616, 1, 1], 
      Hue[0.9685820407255828, 1, 1], 
      Hue[0.9685820407255828, 1, 1], 
      Hue[0.968900274984304, 1, 1], 
      Hue[0.968900274984304, 1, 1], 
      Hue[0.9692185092430251, 1, 1], 
      Hue[0.9692185092430251, 1, 1], 
      Hue[0.9695367435017461, 1, 1], 
      Hue[0.9695367435017461, 1, 1], 
      Hue[0.9698549777604673, 1, 1], 
      Hue[0.9698549777604673, 1, 1], 
      Hue[0.9701732120191885, 1, 1], 
      Hue[0.9701732120191885, 1, 1], 
      Hue[0.9704914462779096, 1, 1], 
      Hue[0.9704914462779096, 1, 1], 
      Hue[0.9708096805366307, 1, 1], 
      Hue[0.9708096805366307, 1, 1], 
      Hue[0.9711279147953519, 1, 1], 
      Hue[0.9711279147953519, 1, 1], 
      Hue[0.9714461490540731, 1, 1], 
      Hue[0.9714461490540731, 1, 1], 
      Hue[0.9717643833127942, 1, 1], 
      Hue[0.9717643833127942, 1, 1], 
      Hue[0.9720826175715154, 1, 1], 
      Hue[0.9720826175715154, 1, 1], 
      Hue[0.9724008518302366, 1, 1], 
      Hue[0.9724008518302366, 1, 1], 
      Hue[0.9727190860889577, 1, 1], 
      Hue[0.9727190860889577, 1, 1], 
      Hue[0.9730373203476788, 1, 1], 
      Hue[0.9730373203476788, 1, 1], 
      Hue[0.9733555546063999, 1, 1], 
      Hue[0.9733555546063999, 1, 1], 
      Hue[0.973673788865121, 1, 1], 
      Hue[0.973673788865121, 1, 1], 
      Hue[0.9739920231238423, 1, 1], 
      Hue[0.9739920231238423, 1, 1], 
      Hue[0.9743102573825634, 1, 1], 
      Hue[0.9743102573825634, 1, 1], 
      Hue[0.9746284916412845, 1, 1], 
      Hue[0.9746284916412845, 1, 1], 
      Hue[0.9749467259000056, 1, 1], 
      Hue[0.9749467259000056, 1, 1], 
      Hue[0.9752649601587269, 1, 1], 
      Hue[0.9752649601587269, 1, 1], 
      Hue[0.975583194417448, 1, 1], 
      Hue[0.975583194417448, 1, 1], 
      Hue[0.9759014286761691, 1, 1], 
      Hue[0.9759014286761691, 1, 1], 
      Hue[0.9762196629348903, 1, 1], 
      Hue[0.9762196629348903, 1, 1], 
      Hue[0.9768561314523325, 1, 1], 
      Hue[0.9768561314523325, 1, 1], 
      Hue[0.9768656420513168, 1, 1], 
      Hue[0.9768656420513168, 1, 1], 
      Hue[0.976875152650301, 1, 1], 
      Hue[0.976875152650301, 1, 1], 
      Hue[0.9768941738482696, 1, 1], 
      Hue[0.9768941738482696, 1, 1], 
      Hue[0.9769322162442068, 1, 1], 
      Hue[0.9769322162442068, 1, 1], 
      Hue[0.9770083010360808, 1, 1], 
      Hue[0.9770083010360808, 1, 1], 
      Hue[0.9771604706198291, 1, 1], 
      Hue[0.9771604706198291, 1, 1], 
      Hue[0.9771699812188133, 1, 1], 
      Hue[0.9771699812188133, 1, 1], 
      Hue[0.9771794918177976, 1, 1], 
      Hue[0.9771794918177976, 1, 1], 
      Hue[0.9771985130157662, 1, 1], 
      Hue[0.9771985130157662, 1, 1], 
      Hue[0.9772365554117033, 1, 1], 
      Hue[0.9772365554117033, 1, 1], 
      Hue[0.9773126402035773, 1, 1], 
      Hue[0.9773126402035773, 1, 1], 
      Hue[0.9773221508025617, 1, 1], 
      Hue[0.9773221508025617, 1, 1], 
      Hue[0.9773316614015459, 1, 1], 
      Hue[0.9773316614015459, 1, 1], 
      Hue[0.9773506825995144, 1, 1], 
      Hue[0.9773506825995144, 1, 1], 
      Hue[0.9773887249954515, 1, 1], 
      Hue[0.9773887249954515, 1, 1], 
      Hue[0.9773982355944357, 1, 1], 
      Hue[0.9773982355944357, 1, 1], 
      Hue[0.9774077461934201, 1, 1], 
      Hue[0.9774077461934201, 1, 1], 
      Hue[0.9774267673913886, 1, 1], 
      Hue[0.9774267673913886, 1, 1], 
      Hue[0.9774362779903729, 1, 1], 
      Hue[0.9774362779903729, 1, 1], 
      Hue[0.9774457885893572, 1, 1], 
      Hue[0.9774457885893572, 1, 1], 
      Hue[0.9774552991883414, 1, 1], 
      Hue[0.9774552991883414, 1, 1], 
      Hue[0.9774648097873256, 1, 1]}], GraphicsComplexBox[CompressedData["
1:eJwUV3c8lt8bNl67pCUkhUgaRDKiY1VWWkRWEoV8S2Rnyw5J9s58ySwruZ/s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    "], {
     {GrayLevel[0.5], EdgeForm[None], 
      GraphicsGroupBox[PolygonBox[CompressedData["
1:eJwtnAm4jdXbh8+wh3dPEkIyJ1JpniTNElIiJDNllhSZQ0JRZCbzPIbMMs9T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        "]]]}, {{}, {}, {
       {Hue[0.04507038042534869, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1, 2}]}, 
       {Hue[0.045656129526931086`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{2, 3}]}, 
       {Hue[0.04624187862851364, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{3, 4}]}, 
       {Hue[0.047413376831678455`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{4, 5}]}, 
       {Hue[0.04975637323800807, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{5, 6}]}, 
       {Hue[0.054442366050667615`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{6, 7}]}, 
       {Hue[0.06381435167598638, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{7, 8}]}, 
       {Hue[0.08255832292662395, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{8, 9}]}, 
       {Hue[0.08319334207426511, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{9, 10}]}, 
       {Hue[0.08382836122190629, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{10, 11}]}, 
       {Hue[0.08509839951718859, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{11, 12}]}, 
       {Hue[0.08763847610775323, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{12, 13}]}, 
       {Hue[0.09271862928888232, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{13, 14}]}, 
       {Hue[0.10287893565114072`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{14, 15}]}, 
       {Hue[0.10351395479878188`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{15, 16}]}, 
       {Hue[0.10414897394642304`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{16, 17}]}, 
       {Hue[0.10541901224170536`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{17, 18}]}, 
       {Hue[0.10795908883226986`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{18, 19}]}, 
       {Hue[0.11303924201339897`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{19, 20}]}, 
       {Hue[0.12319954837565734`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{20, 21}]}, 
       {Hue[0.12379248588008265`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{21, 22}]}, 
       {Hue[0.12438542338450781`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{22, 23}]}, 
       {Hue[0.12557129839335812`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{23, 24}]}, 
       {Hue[0.12794304841105877`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{24, 25}]}, 
       {Hue[0.13268654844646033`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{25, 26}]}, 
       {Hue[0.1332794859508855, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{26, 27}]}, 
       {Hue[0.13387242345531064`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{27, 28}]}, 
       {Hue[0.13505829846416098`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{28, 29}]}, 
       {Hue[0.13743004848186177`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{29, 30}]}, 
       {Hue[0.14217354851726322`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{30, 31}]}, 
       {Hue[0.14276648602168834`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{31, 32}]}, 
       {Hue[0.1433594235261135, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{32, 33}]}, 
       {Hue[0.14454529853496384`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{33, 34}]}, 
       {Hue[0.1469170485526646, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{34, 35}]}, 
       {Hue[0.14750998605708993`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{35, 36}]}, 
       {Hue[0.14810292356151508`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{36, 37}]}, 
       {Hue[0.1492887985703654, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{37, 38}]}, 
       {Hue[0.15166054858806602`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{38, 39}]}, 
       {Hue[0.1522534860924912, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{39, 40}]}, 
       {Hue[0.1528464235969164, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{40, 41}]}, 
       {Hue[0.15403229860576667`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{41, 42}]}, 
       {Hue[0.15462523611019183`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{42, 43}]}, 
       {Hue[0.155218173614617, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{43, 44}]}, 
       {Hue[0.15640404862346746`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{44, 45}]}, 
       {Hue[0.15699698612789276`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{45, 46}]}, 
       {Hue[0.15758992363231794`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{46, 47}]}, 
       {Hue[0.15818286113674307`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{47, 48}]}, 
       {Hue[0.15877579864116825`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{48, 49}]}, 
       {Hue[0.1593687361455934, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{49, 50}]}, 
       {Hue[0.15996167365001857`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{50, 51}]}, 
       {Hue[0.16114754865886888`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{51, 52}]}, 
       {Hue[0.16172885508316964`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{52, 53}]}, 
       {Hue[0.16231016150747038`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{53, 54}]}, 
       {Hue[0.16289146793177126`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{54, 55}]}, 
       {Hue[0.16347277435607202`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{55, 56}]}, 
       {Hue[0.16405408078037276`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{56, 57}]}, 
       {Hue[0.16463538720467363`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{57, 58}]}, 
       {Hue[0.1652166936289744, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{58, 59}]}, 
       {Hue[0.16579800005327514`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{59, 60}]}, 
       {Hue[0.16637930647757587`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{60, 61}]}, 
       {Hue[0.16696061290187664`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{61, 62}]}, 
       {Hue[0.1675419193261775, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{62, 63}]}, 
       {Hue[0.16812322575047825`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{63, 64}]}, 
       {Hue[0.16870453217477901`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{64, 65}]}, 
       {Hue[0.1692858385990799, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{65, 66}]}, 
       {Hue[0.16986714502338063`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{66, 67}]}, 
       {Hue[0.1704484514476814, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{67, 68}]}, 
       {Hue[0.17102975787198213`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{68, 69}]}, 
       {Hue[0.17161106429628287`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{69, 70}]}, 
       {Hue[0.17219237072058377`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{70, 71}]}, 
       {Hue[0.1727736771448845, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{71, 72}]}, 
       {Hue[0.17335498356918527`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{72, 73}]}, 
       {Hue[0.17393628999348615`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{73, 74}]}, 
       {Hue[0.17451759641778689`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{74, 75}]}, 
       {Hue[0.17509890284208765`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{75, 76}]}, 
       {Hue[0.1756802092663884, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{76, 77}]}, 
       {Hue[0.17626151569068912`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{77, 78}]}, 
       {Hue[0.1768428221149899, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{78, 79}]}, 
       {Hue[0.17742412853929063`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{79, 80}]}, 
       {Hue[0.17800543496359136`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{80, 81}]}, 
       {Hue[0.1785867413878921, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{81, 82}]}, 
       {Hue[0.17974935423649374`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{82, 83}]}, 
       {Hue[0.1803306606607945, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{83, 84}]}, 
       {Hue[0.18091196708509538`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{84, 85}]}, 
       {Hue[0.18207457993369688`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{85, 86}]}, 
       {Hue[0.18265588635799762`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{86, 87}]}, 
       {Hue[0.18323719278229836`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{87, 88}]}, 
       {Hue[0.18439980563089986`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{88, 89}]}, 
       {Hue[0.1849811120552006, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{89, 90}]}, 
       {Hue[0.18556241847950136`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{90, 91}]}, 
       {Hue[0.18672503132810297`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{91, 92}]}, 
       {Hue[0.18905025702530612`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{92, 93}]}, 
       {Hue[0.18963156344960685`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{93, 94}]}, 
       {Hue[0.1902128698739076, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{94, 95}]}, 
       {Hue[0.19137548272250923`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{95, 96}]}, 
       {Hue[0.19370070841971235`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{96, 97}]}, 
       {Hue[0.1942820148440131, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{97, 98}]}, 
       {Hue[0.19486332126831385`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{98, 99}]}, 
       {Hue[0.1960259341169155, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{99, 100}]}, 
       {Hue[0.1983511598141186, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{100, 101}]}, 
       {Hue[0.1989817362844781, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{101, 102}]}, 
       {Hue[0.1996123127548376, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{102, 103}]}, 
       {Hue[0.2008734656955566, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{103, 104}]}, 
       {Hue[0.2033957715769946, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{104, 105}]}, 
       {Hue[0.2040263480473541, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{105, 106}]}, 
       {Hue[0.20465692451771358`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{106, 107}]}, 
       {Hue[0.20591807745843257`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{107, 108}]}, 
       {Hue[0.20844038333987058`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{108, 109}]}, 
       {Hue[0.20907095981023008`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{109, 110}]}, 
       {Hue[0.20970153628058957`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{110, 111}]}, 
       {Hue[0.21096268922130856`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{111, 112}]}, 
       {Hue[0.21348499510274654`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{112, 113}]}, 
       {Hue[0.21852960686562253`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{113, 114}]}, 
       {Hue[0.21916018333598203`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{114, 115}]}, 
       {Hue[0.21979075980634152`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{115, 116}]}, 
       {Hue[0.2210519127470605, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{116, 117}]}, 
       {Hue[0.22357421862849852`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{117, 118}]}, 
       {Hue[0.22861883039137448`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{118, 119}]}, 
       {Hue[0.22924940686173398`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{119, 120}]}, 
       {Hue[0.22987998333209347`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{120, 121}]}, 
       {Hue[0.2311411362728125, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{121, 122}]}, 
       {Hue[0.23366344215425047`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{122, 123}]}, 
       {Hue[0.23870805391712646`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{123, 124}]}, 
       {Hue[0.23929654874426995`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{124, 125}]}, 
       {Hue[0.23988504357141346`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{125, 126}]}, 
       {Hue[0.2410620332257006, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{126, 127}]}, 
       {Hue[0.24341601253427475`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{127, 128}]}, 
       {Hue[0.24400450736141824`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{128, 129}]}, 
       {Hue[0.24459300218856173`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{129, 130}]}, 
       {Hue[0.24576999184284887`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{130, 131}]}, 
       {Hue[0.24635848666999252`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{131, 132}]}, 
       {Hue[0.246946981497136, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{132, 133}]}, 
       {Hue[0.248123971151423, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{133, 134}]}, 
       {Hue[0.24871246597856653`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{134, 135}]}, 
       {Hue[0.24930096080571001`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{135, 136}]}, 
       {Hue[0.24988945563285353`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{136, 137}]}, 
       {Hue[0.7504779504599972, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{137, 138}]}, 
       {Hue[0.7510664452871407, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{138, 139}]}, 
       {Hue[0.7516549401142842, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{139, 140}]}, 
       {Hue[0.7528319297685713, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{140, 141}]}, 
       {Hue[0.7534204245957148, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{141, 142}]}, 
       {Hue[0.7540089194228583, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{142, 143}]}, 
       {Hue[0.7551859090771453, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{143, 144}]}, 
       {Hue[0.7575398883857194, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{144, 145}]}, 
       {Hue[0.7581283832128631, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{145, 146}]}, 
       {Hue[0.7587168780400065, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{146, 147}]}, 
       {Hue[0.7598938676942936, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{147, 148}]}, 
       {Hue[0.7622478470028675, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{148, 149}]}, 
       {Hue[0.7669558056200159, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{149, 150}]}, 
       {Hue[0.7763717228543123, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{150, 151}]}, 
       {Hue[0.7770094877275147, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{151, 152}]}, 
       {Hue[0.777647252600717, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{152, 153}]}, 
       {Hue[0.7789227823471214, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{153, 154}]}, 
       {Hue[0.7814738418399304, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{154, 155}]}, 
       {Hue[0.7865759608255485, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{155, 156}]}, 
       {Hue[0.7967801987967844, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{156, 157}]}, 
       {Hue[0.8171886747392565, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{157, 158}]}, 
       {Hue[0.8178148085323343, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{158, 159}]}, 
       {Hue[0.8184409423254122, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{159, 160}]}, 
       {Hue[0.8196932099115679, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{160, 161}]}, 
       {Hue[0.8221977450838792, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{161, 162}]}, 
       {Hue[0.827206815428502, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{162, 163}]}, 
       {Hue[0.8372249561177476, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{163, 164}]}, 
       {Hue[0.8378510899108255, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{164, 165}]}, 
       {Hue[0.8384772237039033, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{165, 166}]}, 
       {Hue[0.8397294912900589, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{166, 167}]}, 
       {Hue[0.8422340264623704, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{167, 168}]}, 
       {Hue[0.8472430968069933, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{168, 169}]}, 
       {Hue[0.8572612374962388, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{169, 170}]}, 
       {Hue[0.8578452896461009, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{170, 171}]}, 
       {Hue[0.8584293417959626, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{171, 172}]}, 
       {Hue[0.8595974460956863, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{172, 173}]}, 
       {Hue[0.8619336546951339, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{173, 174}]}, 
       {Hue[0.8666060718940289, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{174, 175}]}, 
       {Hue[0.8671901240438908, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{175, 176}]}, 
       {Hue[0.8677741761937527, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{176, 177}]}, 
       {Hue[0.8689422804934764, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{177, 178}]}, 
       {Hue[0.8712784890929239, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{178, 179}]}, 
       {Hue[0.875950906291819, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{179, 180}]}, 
       {Hue[0.876534958441681, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{180, 181}]}, 
       {Hue[0.8771190105915428, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{181, 182}]}, 
       {Hue[0.8782871148912667, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{182, 183}]}, 
       {Hue[0.8806233234907141, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{183, 184}]}, 
       {Hue[0.8812073756405759, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{184, 185}]}, 
       {Hue[0.8817914277904378, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{185, 186}]}, 
       {Hue[0.8829595320901616, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{186, 187}]}, 
       {Hue[0.8852957406896091, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{187, 188}]}, 
       {Hue[0.8858797928394709, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{188, 189}]}, 
       {Hue[0.8864638449893328, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{189, 190}]}, 
       {Hue[0.8876319492890566, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{190, 191}]}, 
       {Hue[0.8882160014389185, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{191, 192}]}, 
       {Hue[0.8888000535887804, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{192, 193}]}, 
       {Hue[0.8899681578885041, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{193, 194}]}, 
       {Hue[0.890552210038366, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{194, 195}]}, 
       {Hue[0.891136262188228, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{195, 196}]}, 
       {Hue[0.8923043664879518, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{196, 197}]}, 
       {Hue[0.8928884186378137, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{197, 198}]}, 
       {Hue[0.8934724707876756, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{198, 199}]}, 
       {Hue[0.8946405750873992, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{199, 200}]}, 
       {Hue[0.8952738972833199, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{200, 201}]}, 
       {Hue[0.8959072194792405, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{201, 202}]}, 
       {Hue[0.8965405416751611, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{202, 203}]}, 
       {Hue[0.8971738638710816, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{203, 204}]}, 
       {Hue[0.8978071860670023, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{204, 205}]}, 
       {Hue[0.8984405082629228, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{205, 206}]}, 
       {Hue[0.8990738304588434, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{206, 207}]}, 
       {Hue[0.8997071526547641, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{207, 208}]}, 
       {Hue[0.9003404748506847, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{208, 209}]}, 
       {Hue[0.9009737970466053, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{209, 210}]}, 
       {Hue[0.9016071192425259, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{210, 211}]}, 
       {Hue[0.9022404414384465, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{211, 212}]}, 
       {Hue[0.9028737636343671, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{212, 213}]}, 
       {Hue[0.9035070858302877, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{213, 214}]}, 
       {Hue[0.9041404080262083, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{214, 215}]}, 
       {Hue[0.9047737302221288, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{215, 216}]}, 
       {Hue[0.9054070524180494, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{216, 217}]}, 
       {Hue[0.90604037461397, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{217, 218}]}, 
       {Hue[0.9066736968098906, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{218, 219}]}, 
       {Hue[0.9073070190058112, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{219, 220}]}, 
       {Hue[0.9079403412017318, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{220, 221}]}, 
       {Hue[0.9085736633976524, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{221, 222}]}, 
       {Hue[0.909206985593573, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{222, 223}]}, 
       {Hue[0.9098403077894937, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{223, 224}]}, 
       {Hue[0.9104736299854144, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{224, 225}]}, 
       {Hue[0.9111069521813349, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{225, 226}]}, 
       {Hue[0.9117402743772555, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{226, 227}]}, 
       {Hue[0.9123735965731761, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{227, 228}]}, 
       {Hue[0.9130069187690967, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{228, 229}]}, 
       {Hue[0.9136402409650173, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{229, 230}]}, 
       {Hue[0.9149068853568585, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{230, 231}]}, 
       {Hue[0.9155402075527791, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{231, 232}]}, 
       {Hue[0.9161735297486997, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{232, 233}]}, 
       {Hue[0.9174401741405409, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{233, 234}]}, 
       {Hue[0.9180734963364614, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{234, 235}]}, 
       {Hue[0.918706818532382, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{235, 236}]}, 
       {Hue[0.9199734629242233, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{236, 237}]}, 
       {Hue[0.9206067851201439, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{237, 238}]}, 
       {Hue[0.9212401073160645, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{238, 239}]}, 
       {Hue[0.9225067517079057, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{239, 240}]}, 
       {Hue[0.9250400404915881, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{240, 241}]}, 
       {Hue[0.9256733626875087, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{241, 242}]}, 
       {Hue[0.9263066848834293, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{242, 243}]}, 
       {Hue[0.9275733292752706, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{243, 244}]}, 
       {Hue[0.9301066180589529, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{244, 245}]}, 
       {Hue[0.9307399402548735, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{245, 246}]}, 
       {Hue[0.9313732624507941, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{246, 247}]}, 
       {Hue[0.9326399068426353, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{247, 248}]}, 
       {Hue[0.9351731956263177, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{248, 249}]}, 
       {Hue[0.9357644361790224, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{249, 250}]}, 
       {Hue[0.936355676731727, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{250, 251}]}, 
       {Hue[0.9375381578371363, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{251, 252}]}, 
       {Hue[0.9399031200479548, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{252, 253}]}, 
       {Hue[0.9404943606006595, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{253, 254}]}, 
       {Hue[0.9410856011533641, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{254, 255}]}, 
       {Hue[0.9422680822587733, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{255, 256}]}, 
       {Hue[0.9446330444695918, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{256, 257}]}, 
       {Hue[0.9452242850222965, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{257, 258}]}, 
       {Hue[0.9458155255750011, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{258, 259}]}, 
       {Hue[0.9469980066804105, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{259, 260}]}, 
       {Hue[0.949362968891229, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{260, 261}]}, 
       {Hue[0.9540928933128661, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{261, 262}]}, 
       {Hue[0.9546841338655707, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{262, 263}]}, 
       {Hue[0.9552753744182754, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{263, 264}]}, 
       {Hue[0.9564578555236847, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{264, 265}]}, 
       {Hue[0.9588228177345032, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{265, 266}]}, 
       {Hue[0.9635527421561403, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{266, 267}]}, 
       {Hue[0.9730125909994145, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{267, 268}]}, 
       {Hue[0.9735922004719948, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{268, 269}]}, 
       {Hue[0.974171809944575, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{269, 270}]}, 
       {Hue[0.9753310288897354, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{270, 271}]}, 
       {Hue[0.9776494667800564, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{271, 272}]}, 
       {Hue[0.9822863425606982, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{272, 273}]}, 
       {Hue[0.9828659520332785, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{273, 274}]}, 
       {Hue[0.9834455615058587, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{274, 275}]}, 
       {Hue[0.9846047804510192, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{275, 276}]}, 
       {Hue[0.9869232183413401, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{276, 277}]}, 
       {Hue[0.991560094121982, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{277, 278}]}, 
       {Hue[0.9921397035945623, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{278, 279}]}, 
       {Hue[0.9927193130671425, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{279, 280}]}, 
       {Hue[0.9938785320123029, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{280, 281}]}, 
       {Hue[0.9944581414848832, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{281, 282}]}, 
       {Hue[0.9950377509574634, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{282, 283}]}, 
       {Hue[0.9961969699026239, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{283, 284}]}, 
       {Hue[0.996776579375204, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{284, 285}]}, 
       {Hue[0.9973561888477843, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{285, 286}]}, 
       {Hue[0.9979357983203645, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{286, 287}]}, 
       {Hue[0.9985154077929448, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{287, 288}]}, 
       {Hue[0.999095017265525, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{288, 289}]}, 
       {Hue[0.9996746267381051, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{289, 290}]}, 
       {Hue[0.5002542362106854, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{290, 291}]}, 
       {Hue[0.5008338456832657, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{291, 292}]}, 
       {Hue[0.5014134551558459, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{292, 293}]}, 
       {Hue[0.5019930646284262, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{293, 294}]}, 
       {Hue[0.5031522835735867, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{294, 295}]}, 
       {Hue[0.5054707214639076, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{295, 296}]}, 
       {Hue[0.5101075972445495, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{296, 297}]}, 
       {Hue[0.5107364767631885, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{297, 298}]}, 
       {Hue[0.5113653562818274, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{298, 299}]}, 
       {Hue[0.5126231153191052, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{299, 300}]}, 
       {Hue[0.5151386333936612, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{300, 301}]}, 
       {Hue[0.5201696695427727, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{301, 302}]}, 
       {Hue[0.530231741840996, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{302, 303}]}, 
       {Hue[0.5503558864374425, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{303, 304}]}, 
       {Hue[0.5509426843128655, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{304, 305}]}, 
       {Hue[0.5515294821882883, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{305, 306}]}, 
       {Hue[0.5527030779391344, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{306, 307}]}, 
       {Hue[0.5550502694408264, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{307, 308}]}, 
       {Hue[0.5597446524442103, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{308, 309}]}, 
       {Hue[0.5691334184509782, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{309, 310}]}, 
       {Hue[0.5697202163264011, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{310, 311}]}, 
       {Hue[0.5703070142018241, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{311, 312}]}, 
       {Hue[0.5714806099526701, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{312, 313}]}, 
       {Hue[0.5738278014543621, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{313, 314}]}, 
       {Hue[0.578522184457746, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{314, 315}]}, 
       {Hue[0.5879109504645137, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{315, 316}]}, 
       {Hue[0.5885470183859953, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{316, 317}]}, 
       {Hue[0.5891830863074771, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{317, 318}]}, 
       {Hue[0.5904552221504404, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{318, 319}]}, 
       {Hue[0.5929994938363673, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{319, 320}]}, 
       {Hue[0.598088037208221, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{320, 321}]}, 
       {Hue[0.5987241051297028, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{321, 322}]}, 
       {Hue[0.5993601730511844, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{322, 323}]}, 
       {Hue[0.6006323088941479, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{323, 324}]}, 
       {Hue[0.6031765805800747, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{324, 325}]}, 
       {Hue[0.6082651239519283, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{325, 326}]}, 
       {Hue[0.60890119187341, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{326, 327}]}, 
       {Hue[0.6095372597948917, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{327, 328}]}, 
       {Hue[0.6108093956378551, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{328, 329}]}, 
       {Hue[0.6133536673237819, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{329, 330}]}, 
       {Hue[0.6139897352452637, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{330, 331}]}, 
       {Hue[0.6146258031667454, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{331, 332}]}, 
       {Hue[0.6158979390097088, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{332, 333}]}, 
       {Hue[0.6184422106956357, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{333, 334}]}, 
       {Hue[0.6190782786171174, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{334, 335}]}, 
       {Hue[0.619714346538599, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{335, 336}]}, 
       {Hue[0.6209864823815625, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{336, 337}]}, 
       {Hue[0.6216225503030441, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{337, 338}]}, 
       {Hue[0.6222586182245259, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{338, 339}]}, 
       {Hue[0.6235307540674894, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{339, 340}]}, 
       {Hue[0.6241668219889711, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{340, 341}]}, 
       {Hue[0.6248028899104527, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{341, 342}]}, 
       {Hue[0.6260750257534162, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{342, 343}]}, 
       {Hue[0.6267110936748979, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{343, 344}]}, 
       {Hue[0.6273471615963796, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{344, 345}]}, 
       {Hue[0.628619297439343, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{345, 346}]}, 
       {Hue[0.6292437342807002, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{346, 347}]}, 
       {Hue[0.6298681711220575, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{347, 348}]}, 
       {Hue[0.6311170448047723, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{348, 349}]}, 
       {Hue[0.6317414816461295, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{349, 350}]}, 
       {Hue[0.6323659184874868, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{350, 351}]}, 
       {Hue[0.6336147921702014, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{351, 352}]}, 
       {Hue[0.6342392290115587, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{352, 353}]}, 
       {Hue[0.6348636658529161, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{353, 354}]}, 
       {Hue[0.6354881026942734, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{354, 355}]}, 
       {Hue[0.6361125395356306, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{355, 356}]}, 
       {Hue[0.6367369763769879, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{356, 357}]}, 
       {Hue[0.6373614132183453, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{357, 358}]}, 
       {Hue[0.6379858500597027, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{358, 359}]}, 
       {Hue[0.6386102869010599, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{359, 360}]}, 
       {Hue[0.6392347237424172, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{360, 361}]}, 
       {Hue[0.6398591605837745, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{361, 362}]}, 
       {Hue[0.6404835974251318, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{362, 363}]}, 
       {Hue[0.6411080342664891, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{363, 364}]}, 
       {Hue[0.6417324711078464, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{364, 365}]}, 
       {Hue[0.6423569079492037, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{365, 366}]}, 
       {Hue[0.642981344790561, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{366, 367}]}, 
       {Hue[0.6436057816319183, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{367, 368}]}, 
       {Hue[0.6442302184732756, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{368, 369}]}, 
       {Hue[0.6448546553146329, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{369, 370}]}, 
       {Hue[0.6454790921559903, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{370, 371}]}, 
       {Hue[0.6461035289973476, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{371, 372}]}, 
       {Hue[0.6467279658387048, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{372, 373}]}, 
       {Hue[0.6473524026800621, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{373, 374}]}, 
       {Hue[0.6479768395214194, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{374, 375}]}, 
       {Hue[0.6486012763627766, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{375, 376}]}, 
       {Hue[0.649225713204134, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{376, 377}]}, 
       {Hue[0.6498501500454913, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{377, 378}]}, 
       {Hue[0.6504745868868487, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{378, 379}]}, 
       {Hue[0.6510990237282059, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{379, 380}]}, 
       {Hue[0.6517234605695632, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{380, 381}]}, 
       {Hue[0.6523478974109205, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{381, 382}]}, 
       {Hue[0.653596771093635, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{382, 383}]}, 
       {Hue[0.6542212079349923, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{383, 384}]}, 
       {Hue[0.6548456447763498, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{384, 385}]}, 
       {Hue[0.6560945184590643, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{385, 386}]}, 
       {Hue[0.6567189553004216, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{386, 387}]}, 
       {Hue[0.657343392141779, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{387, 388}]}, 
       {Hue[0.6585922658244936, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{388, 389}]}, 
       {Hue[0.6592167026658509, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{389, 390}]}, 
       {Hue[0.6598411395072081, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{390, 391}]}, 
       {Hue[0.6610900131899228, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{391, 392}]}, 
       {Hue[0.663587760555352, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{392, 393}]}, 
       {Hue[0.6642121973967092, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{393, 394}]}, 
       {Hue[0.6648366342380666, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{394, 395}]}, 
       {Hue[0.6660855079207813, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{395, 396}]}, 
       {Hue[0.6685832552862104, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{396, 397}]}, 
       {Hue[0.6691656104843517, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{397, 398}]}, 
       {Hue[0.6697479656824932, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{398, 399}]}, 
       {Hue[0.6709126760787758, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{399, 400}]}, 
       {Hue[0.6732420968713412, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{400, 401}]}, 
       {Hue[0.6738244520694825, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{401, 402}]}, 
       {Hue[0.6744068072676239, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{402, 403}]}, 
       {Hue[0.6755715176639066, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{403, 404}]}, 
       {Hue[0.6779009384564718, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{404, 405}]}, 
       {Hue[0.6784832936546132, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{405, 406}]}, 
       {Hue[0.6790656488527546, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{406, 407}]}, 
       {Hue[0.6802303592490373, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{407, 408}]}, 
       {Hue[0.6825597800416026, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{408, 409}]}, 
       {Hue[0.6872186216267333, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{409, 410}]}, 
       {Hue[0.6878009768248747, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{410, 411}]}, 
       {Hue[0.6883833320230159, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{411, 412}]}, 
       {Hue[0.6895480424192987, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{412, 413}]}, 
       {Hue[0.691877463211864, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{413, 414}]}, 
       {Hue[0.6965363047969948, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{414, 415}]}, 
       {Hue[0.7058539879672562, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{415, 416}]}, 
       {Hue[0.7064856132114563, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{416, 417}]}, 
       {Hue[0.7071172384556563, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{417, 418}]}, 
       {Hue[0.7083804889440564, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{418, 419}]}, 
       {Hue[0.7109069899208567, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{419, 420}]}, 
       {Hue[0.7159599918744571, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{420, 421}]}, 
       {Hue[0.7260659957816581, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{421, 422}]}, 
       {Hue[0.7266976210258582, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{422, 423}]}, 
       {Hue[0.7273292462700582, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{423, 424}]}, 
       {Hue[0.7285924967584583, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{424, 425}]}, 
       {Hue[0.7311189977352586, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{425, 426}]}, 
       {Hue[0.736171999688859, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{426, 427}]}, 
       {Hue[0.7368036249330591, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{427, 428}]}, 
       {Hue[0.7374352501772592, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{428, 429}]}, 
       {Hue[0.7386985006656592, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{429, 430}]}, 
       {Hue[0.7412250016424595, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{430, 431}]}, 
       {Hue[0.7418566268866595, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{431, 432}]}, 
       {Hue[0.7424882521308596, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{432, 433}]}, 
       {Hue[0.7437515026192598, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{433, 434}]}, 
       {Hue[0.74627800359606, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{434, 435}]}, 
       {Hue[0.7468675471970441, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{435, 436}]}, 
       {Hue[0.7474570907980282, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{436, 437}]}, 
       {Hue[0.7480466343990122, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{437, 438}]}, 
       {Hue[0.7486361779999964, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{438, 439}]}, 
       {Hue[0.7492257216009806, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{439, 440}]}, 
       {Hue[0.7498152652019645, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{440, 441}]}, 
       {Hue[0.25040480880294874`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{441, 442}]}, 
       {Hue[0.2509943524039328, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{442, 443}]}, 
       {Hue[0.2515838960049169, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{443, 444}]}, 
       {Hue[0.25217343960590094`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{444, 445}]}, 
       {Hue[0.2533525268078691, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{445, 446}]}, 
       {Hue[0.25571070121180556`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{446, 447}]}, 
       {Hue[0.2563002448127896, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{447, 448}]}, 
       {Hue[0.25688978841377375`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{448, 449}]}, 
       {Hue[0.25806887561574193`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{449, 450}]}, 
       {Hue[0.2604270500196783, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{450, 451}]}, 
       {Hue[0.26514339882755106`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{451, 452}]}, 
       {Hue[0.2657329424285351, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{452, 453}]}, 
       {Hue[0.26632248602951925`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{453, 454}]}, 
       {Hue[0.26750157323148743`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{454, 455}]}, 
       {Hue[0.2698597476354238, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{455, 456}]}, 
       {Hue[0.27457609644329656`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{456, 457}]}, 
       {Hue[0.28400879405904206`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{457, 458}]}, 
       {Hue[0.2846476077060849, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{458, 459}]}, 
       {Hue[0.2852864213531277, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{459, 460}]}, 
       {Hue[0.2865640486472133, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{460, 461}]}, 
       {Hue[0.28911930323538454`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{461, 462}]}, 
       {Hue[0.2942298124117271, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{462, 463}]}, 
       {Hue[0.3044508307644121, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{463, 464}]}, 
       {Hue[0.3050896444114549, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{464, 465}]}, 
       {Hue[0.30572845805849774`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{465, 466}]}, 
       {Hue[0.3070060853525834, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{466, 467}]}, 
       {Hue[0.3095613399407546, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{467, 468}]}, 
       {Hue[0.3146718491170971, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{468, 469}]}, 
       {Hue[0.32489286746978213`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{469, 470}]}, 
       {Hue[0.3255200500367006, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{470, 471}]}, 
       {Hue[0.32614723260361894`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{471, 472}]}, 
       {Hue[0.3274015977374558, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{472, 473}]}, 
       {Hue[0.3299103280051294, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{473, 474}]}, 
       {Hue[0.3349277885404768, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{474, 475}]}, 
       {Hue[0.33555497110739513`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{475, 476}]}, 
       {Hue[0.3361821536743136, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{476, 477}]}, 
       {Hue[0.33743651880815034`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{477, 478}]}, 
       {Hue[0.339945249075824, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{478, 479}]}, 
       {Hue[0.3449627096111713, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{479, 480}]}, 
       {Hue[0.3455898921780897, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{480, 481}]}, 
       {Hue[0.3462170747450081, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{481, 482}]}, 
       {Hue[0.347471439878845, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{482, 483}]}, 
       {Hue[0.34998017014651855`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{483, 484}]}, 
       {Hue[0.350607352713437, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{484, 485}]}, 
       {Hue[0.35123453528035536`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{485, 486}]}, 
       {Hue[0.35248890041419223`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{486, 487}]}, 
       {Hue[0.35499763068186585`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{487, 488}]}, 
       {Hue[0.3556248132487843, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{488, 489}]}, 
       {Hue[0.35625199581570266`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{489, 490}]}, 
       {Hue[0.3575063609495395, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{490, 491}]}, 
       {Hue[0.358133543516458, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{491, 492}]}, 
       {Hue[0.35876072608337634`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{492, 493}]}, 
       {Hue[0.36001509121721315`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{493, 494}]}, 
       {Hue[0.36064227378413155`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{494, 495}]}, 
       {Hue[0.36126945635105, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{495, 496}]}, 
       {Hue[0.3625238214848868, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{496, 497}]}, 
       {Hue[0.3631510040518052, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{497, 498}]}, 
       {Hue[0.3637781866187237, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{498, 499}]}, 
       {Hue[0.3650325517525605, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{499, 500}]}, 
       {Hue[0.3656176526762629, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{500, 501}]}, 
       {Hue[0.36620275359996535`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{501, 502}]}, 
       {Hue[0.3673729554473702, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{502, 503}]}, 
       {Hue[0.36795805637107265`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{503, 504}]}, 
       {Hue[0.3685431572947751, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{504, 505}]}, 
       {Hue[0.36971335914218006`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{505, 506}]}, 
       {Hue[0.37029846006588246`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{506, 507}]}, 
       {Hue[0.3708835609895849, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{507, 508}]}, 
       {Hue[0.37146866191328737`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{508, 509}]}, 
       {Hue[0.3720537628369898, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{509, 510}]}, 
       {Hue[0.3726388637606922, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{510, 511}]}, 
       {Hue[0.37322396468439467`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{511, 512}]}, 
       {Hue[0.3738090656080972, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{512, 513}]}, 
       {Hue[0.3743941665317996, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{513, 514}]}, 
       {Hue[0.3749792674555021, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{514, 515}]}, 
       {Hue[0.3755643683792045, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{515, 516}]}, 
       {Hue[0.37614946930290694`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{516, 517}]}, 
       {Hue[0.3767345702266094, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{517, 518}]}, 
       {Hue[0.3773196711503118, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{518, 519}]}, 
       {Hue[0.37790477207401424`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{519, 520}]}, 
       {Hue[0.3784898729977167, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{520, 521}]}, 
       {Hue[0.3790749739214192, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{521, 522}]}, 
       {Hue[0.37966007484512154`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{522, 523}]}, 
       {Hue[0.38024517576882405`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{523, 524}]}, 
       {Hue[0.3808302766925265, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{524, 525}]}, 
       {Hue[0.38141537761622896`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{525, 526}]}, 
       {Hue[0.3820004785399314, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{526, 527}]}, 
       {Hue[0.3825855794636338, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{527, 528}]}, 
       {Hue[0.38317068038733626`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{528, 529}]}, 
       {Hue[0.3837557813110387, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{529, 530}]}, 
       {Hue[0.38434088223474117`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{530, 531}]}, 
       {Hue[0.3849259831584437, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{531, 532}]}, 
       {Hue[0.38551108408214607`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{532, 533}]}, 
       {Hue[0.3860961850058485, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{533, 534}]}, 
       {Hue[0.386681285929551, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{534, 535}]}, 
       {Hue[0.38726638685325343`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{535, 536}]}, 
       {Hue[0.3884365887006583, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{536, 537}]}, 
       {Hue[0.38902168962436073`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{537, 538}]}, 
       {Hue[0.38960679054806313`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{538, 539}]}, 
       {Hue[0.39077699239546804`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{539, 540}]}, 
       {Hue[0.3913620933191705, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{540, 541}]}, 
       {Hue[0.391947194242873, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{541, 542}]}, 
       {Hue[0.39311739609027785`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{542, 543}]}, 
       {Hue[0.3937024970139803, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{543, 544}]}, 
       {Hue[0.39428759793768275`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{544, 545}]}, 
       {Hue[0.3954577997850876, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{545, 546}]}, 
       {Hue[0.3977982034798974, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{546, 547}]}, 
       {Hue[0.39838330440359987`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{547, 548}]}, 
       {Hue[0.3989684053273023, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{548, 549}]}, 
       {Hue[0.4001386071747072, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{549, 550}]}, 
       {Hue[0.402479010869517, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{550, 551}]}, 
       {Hue[0.40311338183927814`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{551, 552}]}, 
       {Hue[0.4037477528090393, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{552, 553}]}, 
       {Hue[0.4050164947485616, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{553, 554}]}, 
       {Hue[0.40755397862760634`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{554, 555}]}, 
       {Hue[0.40818834959736744`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{555, 556}]}, 
       {Hue[0.4088227205671286, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{556, 557}]}, 
       {Hue[0.4100914625066509, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{557, 558}]}, 
       {Hue[0.41262894638569564`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{558, 559}]}, 
       {Hue[0.4132633173554568, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{559, 560}]}, 
       {Hue[0.413897688325218, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{560, 561}]}, 
       {Hue[0.4151664302647403, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{561, 562}]}, 
       {Hue[0.41770391414378494`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{562, 563}]}, 
       {Hue[0.4227788819018743, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{563, 564}]}, 
       {Hue[0.4234132528716354, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{564, 565}]}, 
       {Hue[0.4240476238413966, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{565, 566}]}, 
       {Hue[0.4253163657809189, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{566, 567}]}, 
       {Hue[0.4278538496599636, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{567, 568}]}, 
       {Hue[0.4329288174180529, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{568, 569}]}, 
       {Hue[0.44307875293423155`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{569, 570}]}, 
       {Hue[0.4436710422607767, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{570, 571}]}, 
       {Hue[0.44426333158732195`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{571, 572}]}, 
       {Hue[0.44544791024041236`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{572, 573}]}, 
       {Hue[0.4478170675465931, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{573, 574}]}, 
       {Hue[0.45255538215895474`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{574, 575}]}, 
       {Hue[0.46203201138367794`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{575, 576}]}, 
       {Hue[0.46262430071022315`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{576, 577}]}, 
       {Hue[0.46321659003676835`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{577, 578}]}, 
       {Hue[0.46440116868985876`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{578, 579}]}, 
       {Hue[0.46677032599603957`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{579, 580}]}, 
       {Hue[0.4715086406084012, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{580, 581}]}, 
       {Hue[0.4809852698331244, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{581, 582}]}, 
       {Hue[0.4815659280795452, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{582, 583}]}, 
       {Hue[0.48214658632596596`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{583, 584}]}, 
       {Hue[0.4833079028188076, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{584, 585}]}, 
       {Hue[0.4856305358044908, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{585, 586}]}, 
       {Hue[0.4902758017758571, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{586, 587}]}, 
       {Hue[0.4908564600222779, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{587, 588}]}, 
       {Hue[0.4914371182686987, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{588, 589}]}, 
       {Hue[0.4925984347615403, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{589, 590}]}, 
       {Hue[0.4949210677472235, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{590, 591}]}, 
       {Hue[0.4955017259936443, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{591, 592}]}, 
       {Hue[0.49608238424006507`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{592, 593}]}, 
       {Hue[0.4972437007329067, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{593, 594}]}, 
       {Hue[0.4978243589793275, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{594, 595}]}, 
       {Hue[0.4984050172257483, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{595, 596}]}, 
       {Hue[0.4989856754721691, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{596, 597}]}, 
       {Hue[0.4995663337185899, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{597, 598}]}, 
       {Hue[0.00014699196501062852`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{598, 599}]}, 
       {Hue[0.0007276502114314253, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{599, 600}]}, 
       {Hue[0.001308308457852222, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{600, 601}]}, 
       {Hue[0.0018889667042730188`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{601, 602}]}, 
       {Hue[0.002469624950693816, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{602, 603}]}, 
       {Hue[0.0030502831971146124`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{603, 604}]}, 
       {Hue[0.004211599689956207, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{604, 605}]}, 
       {Hue[0.004792257936377003, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{605, 606}]}, 
       {Hue[0.005372916182797801, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{606, 607}]}, 
       {Hue[0.006534232675639393, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{607, 608}]}, 
       {Hue[0.00885686566132258, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{608, 609}]}, 
       {Hue[0.01814739760405533, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{609, 610}]}, 
       {Hue[0.018777325896534847`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{610, 611}]}, 
       {Hue[0.019407254189014365`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{611, 612}]}, 
       {Hue[0.020667110773973398`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{612, 613}]}, 
       {Hue[0.02318682394389146, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{613, 614}]}, 
       {Hue[0.028226250283727595`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{614, 615}]}, 
       {Hue[0.03830510296339986, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{615, 616}]}, 
       {Hue[0.03893503125587938, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{616, 617}]}, 
       {Hue[0.03956495954835889, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{617, 618}]}, 
       {Hue[0.04082481613331793, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{618, 619}]}, 
       {Hue[0.04334452930323598, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{619, 620}]}, 
       {Hue[0.048383955643072116`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{620, 621}]}, 
       {Hue[0.058462808322744406`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{621, 622}]}, 
       {Hue[0.05905065497200795, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{622, 623}]}, 
       {Hue[0.05963850162127149, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{623, 624}]}, 
       {Hue[0.0608141949197986, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{624, 625}]}, 
       {Hue[0.06316558151685281, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{625, 626}]}, 
       {Hue[0.06786835471096123, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{626, 627}]}, 
       {Hue[0.06845620136022479, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{627, 628}]}, 
       {Hue[0.06904404800948832, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{628, 629}]}, 
       {Hue[0.07021974130801542, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{629, 630}]}, 
       {Hue[0.07257112790506963, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{630, 631}]}, 
       {Hue[0.07727390109917805, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{631, 632}]}, 
       {Hue[0.0778617477484416, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{632, 633}]}, 
       {Hue[0.07844959439770514, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{633, 634}]}, 
       {Hue[0.07962528769623224, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{634, 635}]}, 
       {Hue[0.08197667429328648, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{635, 636}]}, 
       {Hue[0.08256452094255003, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{636, 637}]}, 
       {Hue[0.08315236759181359, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{637, 638}]}, 
       {Hue[0.08432806089034067, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{638, 639}]}, 
       {Hue[0.08667944748739487, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{639, 640}]}, 
       {Hue[0.08726729413665844, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{640, 641}]}, 
       {Hue[0.08785514078592198, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{641, 642}]}, 
       {Hue[0.08903083408444905, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{642, 643}]}, 
       {Hue[0.09138222068150327, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{643, 644}]}, 
       {Hue[0.09197006733076682, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{644, 645}]}, 
       {Hue[0.09255791398003038, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{645, 646}]}, 
       {Hue[0.0937336072785575, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{646, 647}]}, 
       {Hue[0.09608499387561169, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{647, 648}]}, 
       {Hue[0.09672211057093397, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{648, 649}]}, 
       {Hue[0.09735922726625623, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{649, 650}]}, 
       {Hue[0.09863346065690076, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{650, 651}]}, 
       {Hue[0.09927057735222307, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{651, 652}]}, 
       {Hue[0.09990769404754532, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{652, 653}]}, 
       {Hue[0.10118192743818985`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{653, 654}]}, 
       {Hue[0.10181904413351212`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{654, 655}]}, 
       {Hue[0.1024561608288344, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{655, 656}]}, 
       {Hue[0.10309327752415666`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{656, 657}]}, 
       {Hue[0.10373039421947897`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{657, 658}]}, 
       {Hue[0.10436751091480122`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{658, 659}]}, 
       {Hue[0.10500462761012351`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{659, 660}]}, 
       {Hue[0.10564174430544576`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{660, 661}]}, 
       {Hue[0.10627886100076801`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{661, 662}]}, 
       {Hue[0.10691597769609032`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{662, 663}]}, 
       {Hue[0.10755309439141254`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{663, 664}]}, 
       {Hue[0.10819021108673485`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{664, 665}]}, 
       {Hue[0.10882732778205713`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{665, 666}]}, 
       {Hue[0.10946444447737938`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{666, 667}]}, 
       {Hue[0.11010156117270166`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{667, 668}]}, 
       {Hue[0.11073867786802392`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{668, 669}]}, 
       {Hue[0.11137579456334618`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{669, 670}]}, 
       {Hue[0.11201291125866845`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{670, 671}]}, 
       {Hue[0.11265002795399072`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{671, 672}]}, 
       {Hue[0.11328714464931298`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{672, 673}]}, 
       {Hue[0.11392426134463528`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{673, 674}]}, 
       {Hue[0.11456137803995754`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{674, 675}]}, 
       {Hue[0.11519849473527982`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{675, 676}]}, 
       {Hue[0.11583561143060207`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{676, 677}]}, 
       {Hue[0.11647272812592435`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{677, 678}]}, 
       {Hue[0.11710984482124662`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{678, 679}]}, 
       {Hue[0.11774696151656887`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{679, 680}]}, 
       {Hue[0.11838407821189115`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{680, 681}]}, 
       {Hue[0.11902119490721341`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{681, 682}]}, 
       {Hue[0.1196583116025357, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{682, 683}]}, 
       {Hue[0.12029542829785798`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{683, 684}]}, 
       {Hue[0.1215696616885025, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{684, 685}]}, 
       {Hue[0.12220677838382481`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{685, 686}]}, 
       {Hue[0.12284389507914709`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{686, 687}]}, 
       {Hue[0.12411812846979162`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{687, 688}]}, 
       {Hue[0.12475524516511388`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{688, 689}]}, 
       {Hue[0.12539236186043612`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{689, 690}]}, 
       {Hue[0.1266665952510807, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{690, 691}]}, 
       {Hue[0.12730371194640297`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{691, 692}]}, 
       {Hue[0.12794082864172526`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{692, 693}]}, 
       {Hue[0.1292150620323698, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{693, 694}]}, 
       {Hue[0.12985217872769206`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{694, 695}]}, 
       {Hue[0.13048929542301432`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{695, 696}]}, 
       {Hue[0.13176352881365888`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{696, 697}]}, 
       {Hue[0.13240064550898115`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{697, 698}]}, 
       {Hue[0.13303776220430344`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{698, 699}]}, 
       {Hue[0.13431199559494797`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{699, 700}]}, 
       {Hue[0.13686046237623703`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{700, 701}]}, 
       {Hue[0.1374859479914349, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{701, 702}]}, 
       {Hue[0.13811143360663275`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{702, 703}]}, 
       {Hue[0.13936240483702853`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{703, 704}]}, 
       {Hue[0.14186434729781996`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{704, 705}]}, 
       {Hue[0.14248983291301787`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{705, 706}]}, 
       {Hue[0.14311531852821574`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{706, 707}]}, 
       {Hue[0.14436628975861146`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{707, 708}]}, 
       {Hue[0.14686823221940293`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{708, 709}]}, 
       {Hue[0.14749371783460077`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{709, 710}]}, 
       {Hue[0.14811920344979865`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{710, 711}]}, 
       {Hue[0.1493701746801944, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{711, 712}]}, 
       {Hue[0.15187211714098586`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{712, 713}]}, 
       {Hue[0.15687600206256883`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{713, 714}]}, 
       {Hue[0.1575014876777667, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{714, 715}]}, 
       {Hue[0.15812697329296455`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{715, 716}]}, 
       {Hue[0.15937794452336027`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{716, 717}]}, 
       {Hue[0.16187988698415173`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{717, 718}]}, 
       {Hue[0.1668837719057347, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{718, 719}]}, 
       {Hue[0.1768915417489006, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{719, 720}]}, 
       {Hue[0.17747494572088252`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{720, 721}]}, 
       {Hue[0.1780583496928644, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{721, 722}]}, 
       {Hue[0.17922515763682823`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{722, 723}]}, 
       {Hue[0.18155877352475583`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{723, 724}]}, 
       {Hue[0.18622600530061104`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{724, 725}]}, 
       {Hue[0.19556046885232148`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{725, 726}]}, 
       {Hue[0.1961438728243034, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{726, 727}]}, 
       {Hue[0.1967272767962853, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{727, 728}]}, 
       {Hue[0.1978940847402491, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{728, 729}]}, 
       {Hue[0.2002277006281767, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{729, 730}]}, 
       {Hue[0.20489493240403192`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{730, 731}]}, 
       {Hue[0.2142293959557424, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{731, 732}]}, 
       {Hue[0.214862069973783, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{732, 733}]}, 
       {Hue[0.2154947439918236, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{733, 734}]}, 
       {Hue[0.21676009202790486`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{734, 735}]}, 
       {Hue[0.21929078810006736`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{735, 736}]}, 
       {Hue[0.22435218024439235`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{736, 737}]}, 
       {Hue[0.22498485426243298`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{737, 738}]}, 
       {Hue[0.2256175282804736, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{738, 739}]}, 
       {Hue[0.22688287631655485`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{739, 740}]}, 
       {Hue[0.22941357238871735`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{740, 741}]}, 
       {Hue[0.2344749645330423, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{741, 742}]}, 
       {Hue[0.23510763855108294`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{742, 743}]}, 
       {Hue[0.23574031256912356`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{743, 744}]}, 
       {Hue[0.2370056606052048, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{744, 745}]}, 
       {Hue[0.2395363566773673, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{745, 746}]}, 
       {Hue[0.24016903069540793`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{746, 747}]}, 
       {Hue[0.24080170471344856`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{747, 748}]}, 
       {Hue[0.2420670527495298, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{748, 749}]}, 
       {Hue[0.2445977488216923, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{749, 750}]}, 
       {Hue[0.24523042283973293`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{750, 751}]}, 
       {Hue[0.24586309685777355`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{751, 752}]}, 
       {Hue[0.2471284448938548, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{752, 753}]}, 
       {Hue[0.24776111891189542`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{753, 754}]}, 
       {Hue[0.24839379292993605`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{754, 755}]}, 
       {Hue[0.24902646694797667`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{755, 756}]}, 
       {Hue[0.2496591409660173, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{756, 757}]}, 
       {Hue[0.7502918149840578, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{757, 758}]}, 
       {Hue[0.7509244890020985, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{758, 759}]}, 
       {Hue[0.7515571630201391, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{759, 760}]}, 
       {Hue[0.7521898370381798, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{760, 761}]}, 
       {Hue[0.7528225110562203, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{761, 762}]}, 
       {Hue[0.753455185074261, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{762, 763}]}, 
       {Hue[0.7547205331103422, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{763, 764}]}, 
       {Hue[0.7553111254851669, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{764, 765}]}, 
       {Hue[0.7559017178599915, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{765, 766}]}, 
       {Hue[0.7570829026096408, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{766, 767}]}, 
       {Hue[0.7594452721089394, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{767, 768}]}, 
       {Hue[0.7641700111075367, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{768, 769}]}, 
       {Hue[0.7647606034823614, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{769, 770}]}, 
       {Hue[0.765351195857186, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{770, 771}]}, 
       {Hue[0.7665323806068354, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{771, 772}]}, 
       {Hue[0.7688947501061341, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{772, 773}]}, 
       {Hue[0.7736194891047312, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{773, 774}]}, 
       {Hue[0.774210081479556, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{774, 775}]}, 
       {Hue[0.7748006738543806, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{775, 776}]}, 
       {Hue[0.7759818586040299, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{776, 777}]}, 
       {Hue[0.7783442281033285, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{777, 778}]}, 
       {Hue[0.7830689671019258, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{778, 779}]}, 
       {Hue[0.7925184450991203, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{779, 780}]}, 
       {Hue[0.7930974063938205, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{780, 781}]}, 
       {Hue[0.7936763676885208, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{781, 782}]}, 
       {Hue[0.7948342902779213, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{782, 783}]}, 
       {Hue[0.7971501354567223, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{783, 784}]}, 
       {Hue[0.8017818258143243, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{784, 785}]}, 
       {Hue[0.8023607871090246, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{785, 786}]}, 
       {Hue[0.8029397484037248, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{786, 787}]}, 
       {Hue[0.8040976709931253, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{787, 788}]}, 
       {Hue[0.8064135161719264, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{788, 789}]}, 
       {Hue[0.8110452065295284, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{789, 790}]}, 
       {Hue[0.8116241678242286, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{790, 791}]}, 
       {Hue[0.8122031291189289, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{791, 792}]}, 
       {Hue[0.8133610517083294, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{792, 793}]}, 
       {Hue[0.8156768968871304, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{793, 794}]}, 
       {Hue[0.8162558581818307, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{794, 795}]}, 
       {Hue[0.8168348194765309, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{795, 796}]}, 
       {Hue[0.8179927420659314, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{796, 797}]}, 
       {Hue[0.8203085872447325, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{797, 798}]}, 
       {Hue[0.8208875485394328, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{798, 799}]}, 
       {Hue[0.8214665098341329, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{799, 800}]}, 
       {Hue[0.8226244324235334, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{800, 801}]}, 
       {Hue[0.8249402776023345, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{801, 802}]}, 
       {Hue[0.8255192388970347, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{802, 803}]}, 
       {Hue[0.826098200191735, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{803, 804}]}, 
       {Hue[0.8272561227811355, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{804, 805}]}, 
       {Hue[0.8295719679599365, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{805, 806}]}, 
       {Hue[0.8302001993006954, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{806, 807}]}, 
       {Hue[0.8308284306414545, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{807, 808}]}, 
       {Hue[0.8320848933229724, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{808, 809}]}, 
       {Hue[0.8327131246637314, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{809, 810}]}, 
       {Hue[0.8333413560044904, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{810, 811}]}, 
       {Hue[0.8345978186860082, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{811, 812}]}, 
       {Hue[0.8352260500267673, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{812, 813}]}, 
       {Hue[0.8358542813675263, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{813, 814}]}, 
       {Hue[0.8371107440490442, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{814, 815}]}, 
       {Hue[0.8377389753898031, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{815, 816}]}, 
       {Hue[0.8383672067305622, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{816, 817}]}, 
       {Hue[0.8396236694120801, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{817, 818}]}, 
       {Hue[0.8402519007528391, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{818, 819}]}, 
       {Hue[0.8408801320935981, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{819, 820}]}, 
       {Hue[0.841508363434357, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{820, 821}]}, 
       {Hue[0.842136594775116, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{821, 822}]}, 
       {Hue[0.8427648261158749, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{822, 823}]}, 
       {Hue[0.843393057456634, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{823, 824}]}, 
       {Hue[0.8440212887973929, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{824, 825}]}, 
       {Hue[0.8446495201381519, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{825, 826}]}, 
       {Hue[0.8452777514789109, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{826, 827}]}, 
       {Hue[0.8459059828196699, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{827, 828}]}, 
       {Hue[0.8465342141604288, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{828, 829}]}, 
       {Hue[0.8471624455011878, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{829, 830}]}, 
       {Hue[0.8477906768419468, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{830, 831}]}, 
       {Hue[0.8484189081827058, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{831, 832}]}, 
       {Hue[0.8490471395234647, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{832, 833}]}, 
       {Hue[0.8496753708642237, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{833, 834}]}, 
       {Hue[0.8503036022049827, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{834, 835}]}, 
       {Hue[0.8509318335457416, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{835, 836}]}, 
       {Hue[0.8515600648865006, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{836, 837}]}, 
       {Hue[0.8521882962272596, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{837, 838}]}, 
       {Hue[0.8528165275680185, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{838, 839}]}, 
       {Hue[0.8534447589087775, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{839, 840}]}, 
       {Hue[0.8540729902495365, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{840, 841}]}, 
       {Hue[0.8547012215902955, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{841, 842}]}, 
       {Hue[0.8553294529310544, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{842, 843}]}, 
       {Hue[0.8559576842718135, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{843, 844}]}, 
       {Hue[0.8565859156125724, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{844, 845}]}, 
       {Hue[0.8572141469533314, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{845, 846}]}, 
       {Hue[0.8578423782940904, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{846, 847}]}, 
       {Hue[0.8584706096348493, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{847, 848}]}, 
       {Hue[0.8597270723163672, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{848, 849}]}, 
       {Hue[0.8603553036571262, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{849, 850}]}, 
       {Hue[0.8609835349978852, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{850, 851}]}, 
       {Hue[0.8622399976794031, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{851, 852}]}, 
       {Hue[0.862868229020162, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{852, 853}]}, 
       {Hue[0.863496460360921, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{853, 854}]}, 
       {Hue[0.864752923042439, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{854, 855}]}, 
       {Hue[0.865381154383198, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{855, 856}]}, 
       {Hue[0.866009385723957, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{856, 857}]}, 
       {Hue[0.8672658484054749, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{857, 858}]}, 
       {Hue[0.8697787737685108, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{858, 859}]}, 
       {Hue[0.8703649234660538, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{859, 860}]}, 
       {Hue[0.8709510731635968, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{860, 861}]}, 
       {Hue[0.8721233725586828, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{861, 862}]}, 
       {Hue[0.8744679713488549, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{862, 863}]}, 
       {Hue[0.8750541210463978, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{863, 864}]}, 
       {Hue[0.8756402707439409, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{864, 865}]}, 
       {Hue[0.8768125701390269, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{865, 866}]}, 
       {Hue[0.8791571689291989, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{866, 867}]}, 
       {Hue[0.879743318626742, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{867, 868}]}, 
       {Hue[0.8803294683242849, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{868, 869}]}, 
       {Hue[0.8815017677193709, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{869, 870}]}, 
       {Hue[0.883846366509543, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{870, 871}]}, 
       {Hue[0.8885355640898871, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{871, 872}]}, 
       {Hue[0.88912171378743, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{872, 873}]}, 
       {Hue[0.8897078634849731, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{873, 874}]}, 
       {Hue[0.8908801628800591, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{874, 875}]}, 
       {Hue[0.8932247616702311, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{875, 876}]}, 
       {Hue[0.8979139592505752, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{876, 877}]}, 
       {Hue[0.9072923544112633, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{877, 878}]}, 
       {Hue[0.907927774154865, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{878, 879}]}, 
       {Hue[0.9085631938984667, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{879, 880}]}, 
       {Hue[0.9098340333856701, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{880, 881}]}, 
       {Hue[0.9123757123600771, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{881, 882}]}, 
       {Hue[0.917459070308891, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{882, 883}]}, 
       {Hue[0.9276257862065186, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{883, 884}]}, 
       {Hue[0.947959218001774, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{884, 885}]}, 
       {Hue[0.9485525561021597, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{885, 886}]}, 
       {Hue[0.9491458942025454, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{886, 887}]}, 
       {Hue[0.9503325704033169, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{887, 888}]}, 
       {Hue[0.95270592280486, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{888, 889}]}, 
       {Hue[0.9574526276079461, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{889, 890}]}, 
       {Hue[0.9669460372141183, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{890, 891}]}, 
       {Hue[0.9675393753145042, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{891, 892}]}, 
       {Hue[0.9681327134148898, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{892, 893}]}, 
       {Hue[0.9693193896156613, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{893, 894}]}, 
       {Hue[0.9716927420172045, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{894, 895}]}, 
       {Hue[0.9764394468202906, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{895, 896}]}, 
       {Hue[0.9859328564264628, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{896, 897}]}, 
       {Hue[0.9865145634467241, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{897, 898}]}, 
       {Hue[0.9870962704669854, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{898, 899}]}, 
       {Hue[0.9882596845075081, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{899, 900}]}, 
       {Hue[0.9905865125885537, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{900, 901}]}, 
       {Hue[0.9911682196088151, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{901, 902}]}, 
       {Hue[0.9917499266290765, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{902, 903}]}, 
       {Hue[0.9929133406695991, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{903, 904}]}, 
       {Hue[0.9952401687506446, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{904, 905}]}, 
       {Hue[0.9958218757709059, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{905, 906}]}, 
       {Hue[0.9964035827911673, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{906, 907}]}, 
       {Hue[0.9975669968316899, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{907, 908}]}, 
       {Hue[0.9981487038519513, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{908, 909}]}, 
       {Hue[0.9987304108722127, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{909, 910}]}, 
       {Hue[0.999312117892474, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{910, 911}]}, 
       {Hue[0.9998938249127354, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{911, 912}]}, 
       {Hue[0.5004755319329968, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{912, 913}]}, 
       {Hue[0.5010572389532582, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{913, 914}]}, 
       {Hue[0.5016389459735195, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{914, 915}]}, 
       {Hue[0.5022206529937807, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{915, 916}]}, 
       {Hue[0.5045474810748263, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{916, 917}]}, 
       {Hue[0.5051291880950876, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{917, 918}]}, 
       {Hue[0.505710895115349, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{918, 919}]}, 
       {Hue[0.5068743091558716, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{919, 920}]}, 
       {Hue[0.5092011372369171, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{920, 921}]}, 
       {Hue[0.513854793399008, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{921, 922}]}, 
       {Hue[0.5231621057231897, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{922, 923}]}, 
       {Hue[0.5237930827895098, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{923, 924}]}, 
       {Hue[0.5244240598558298, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{924, 925}]}, 
       {Hue[0.52568601398847, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{925, 926}]}, 
       {Hue[0.5282099222537503, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{926, 927}]}, 
       {Hue[0.5332577387843109, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{927, 928}]}, 
       {Hue[0.533888715850631, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{928, 929}]}, 
       {Hue[0.5345196929169511, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{929, 930}]}, 
       {Hue[0.5357816470495913, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{930, 931}]}, 
       {Hue[0.5383055553148717, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{931, 932}]}, 
       {Hue[0.5433533718454322, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{932, 933}]}, 
       {Hue[0.5439843489117524, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{933, 934}]}, 
       {Hue[0.5446153259780724, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{934, 935}]}, 
       {Hue[0.5458772801107126, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{935, 936}]}, 
       {Hue[0.5484011883759928, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{936, 937}]}, 
       {Hue[0.5490321654423129, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{937, 938}]}, 
       {Hue[0.549663142508633, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{938, 939}]}, 
       {Hue[0.5509250966412733, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{939, 940}]}, 
       {Hue[0.5534490049065535, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{940, 941}]}, 
       {Hue[0.5540799819728736, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{941, 942}]}, 
       {Hue[0.5547109590391937, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{942, 943}]}, 
       {Hue[0.5559729131718338, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{943, 944}]}, 
       {Hue[0.5584968214371142, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{944, 945}]}, 
       {Hue[0.5591277985034342, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{945, 946}]}, 
       {Hue[0.5597587755697544, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{946, 947}]}, 
       {Hue[0.5610207297023944, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{947, 948}]}, 
       {Hue[0.5635446379676747, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{948, 949}]}, 
       {Hue[0.5641335333907789, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{949, 950}]}, 
       {Hue[0.564722428813883, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{950, 951}]}, 
       {Hue[0.5659002196600912, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{951, 952}]}, 
       {Hue[0.5664891150831952, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{952, 953}]}, 
       {Hue[0.5670780105062995, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{953, 954}]}, 
       {Hue[0.5682558013525076, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{954, 955}]}, 
       {Hue[0.5688446967756118, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{955, 956}]}, 
       {Hue[0.5694335921987159, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{956, 957}]}, 
       {Hue[0.5706113830449241, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{957, 958}]}, 
       {Hue[0.5712002784680281, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{958, 959}]}, 
       {Hue[0.5717891738911324, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{959, 960}]}, 
       {Hue[0.5729669647373405, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{960, 961}]}, 
       {Hue[0.5735558601604447, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{961, 962}]}, 
       {Hue[0.5741447555835488, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{962, 963}]}, 
       {Hue[0.5753225464297571, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{963, 964}]}, 
       {Hue[0.575911441852861, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{964, 965}]}, 
       {Hue[0.5765003372759653, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{965, 966}]}, 
       {Hue[0.5776781281221735, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{966, 967}]}, 
       {Hue[0.5782670235452776, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{967, 968}]}, 
       {Hue[0.5788559189683817, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{968, 969}]}, 
       {Hue[0.5794448143914859, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{969, 970}]}, 
       {Hue[0.5800337098145899, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{970, 971}]}, 
       {Hue[0.580622605237694, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{971, 972}]}, 
       {Hue[0.5812115006607981, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{972, 973}]}, 
       {Hue[0.5818003960839022, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{973, 974}]}, 
       {Hue[0.5823892915070064, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{974, 975}]}, 
       {Hue[0.5829781869301105, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{975, 976}]}, 
       {Hue[0.5835670823532146, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{976, 977}]}, 
       {Hue[0.5841559777763188, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{977, 978}]}, 
       {Hue[0.5847448731994228, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{978, 979}]}, 
       {Hue[0.585333768622527, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{979, 980}]}, 
       {Hue[0.585922664045631, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{980, 981}]}, 
       {Hue[0.5865115594687351, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{981, 982}]}, 
       {Hue[0.5871004548918393, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{982, 983}]}, 
       {Hue[0.5876893503149434, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{983, 984}]}, 
       {Hue[0.5882782457380475, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{984, 985}]}, 
       {Hue[0.5888671411611517, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{985, 986}]}, 
       {Hue[0.5894560365842557, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{986, 987}]}, 
       {Hue[0.5900449320073597, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{987, 988}]}, 
       {Hue[0.5906338274304639, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{988, 989}]}, 
       {Hue[0.591222722853568, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{989, 990}]}, 
       {Hue[0.5918116182766722, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{990, 991}]}, 
       {Hue[0.5924005136997763, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{991, 992}]}, 
       {Hue[0.5929894091228805, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{992, 993}]}, 
       {Hue[0.5935783045459846, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{993, 994}]}, 
       {Hue[0.5941671999690887, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{994, 995}]}, 
       {Hue[0.5947560953921927, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{995, 996}]}, 
       {Hue[0.5953449908152968, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{996, 997}]}, 
       {Hue[0.5965227816615051, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{997, 998}]}, 
       {Hue[0.5971116770846092, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{998, 999}]}, 
       {Hue[0.5977005725077134, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{999, 1000}]}, 
       {Hue[0.5988783633539216, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1000, 1001}]}, 
       {Hue[0.5994672587770258, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1001, 1002}]}, 
       {Hue[0.6000561542001298, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1002, 1003}]}, 
       {Hue[0.601233945046338, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1003, 1004}]}, 
       {Hue[0.6018721105155009, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1004, 1005}]}, 
       {Hue[0.6025102759846637, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1005, 1006}]}, 
       {Hue[0.6037866069229894, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1006, 1007}]}, 
       {Hue[0.6063392687996407, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1007, 1008}]}, 
       {Hue[0.6069774342688035, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1008, 1009}]}, 
       {Hue[0.6076155997379664, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1009, 1010}]}, 
       {Hue[0.6088919306762921, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1010, 1011}]}, 
       {Hue[0.6114445925529435, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1011, 1012}]}, 
       {Hue[0.6120827580221063, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1012, 1013}]}, 
       {Hue[0.612720923491269, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1013, 1014}]}, 
       {Hue[0.6139972544295947, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1014, 1015}]}, 
       {Hue[0.6165499163062461, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1015, 1016}]}, 
       {Hue[0.6216552400595488, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1016, 1017}]}, 
       {Hue[0.6222934055287116, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1017, 1018}]}, 
       {Hue[0.6229315709978744, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1018, 1019}]}, 
       {Hue[0.6242079019362001, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1019, 1020}]}, 
       {Hue[0.6267605638128514, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1020, 1021}]}, 
       {Hue[0.6318658875661541, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1021, 1022}]}, 
       {Hue[0.6420765350727594, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1022, 1023}]}, 
       {Hue[0.6427030694617979, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1023, 1024}]}, 
       {Hue[0.6433296038508364, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1024, 1025}]}, 
       {Hue[0.6445826726289132, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1025, 1026}]}, 
       {Hue[0.6470888101850669, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1026, 1027}]}, 
       {Hue[0.6521010852973743, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1027, 1028}]}, 
       {Hue[0.6621256355219893, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1028, 1029}]}, 
       {Hue[0.6821747359712191, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1029, 1030}]}, 
       {Hue[0.6827591887170417, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1030, 1031}]}, 
       {Hue[0.6833436414628641, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1031, 1032}]}, 
       {Hue[0.684512546954509, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1032, 1033}]}, 
       {Hue[0.6868503579377989, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1033, 1034}]}, 
       {Hue[0.6915259799043787, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1034, 1035}]}, 
       {Hue[0.700877223837538, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1035, 1036}]}, 
       {Hue[0.7014616765833603, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1036, 1037}]}, 
       {Hue[0.7020461293291829, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1037, 1038}]}, 
       {Hue[0.7032150348208277, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1038, 1039}]}, 
       {Hue[0.7055528458041176, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1039, 1040}]}, 
       {Hue[0.7102284677706974, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1040, 1041}]}, 
       {Hue[0.7195797117038569, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1041, 1042}]}, 
       {Hue[0.7202134344957382, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1042, 1043}]}, 
       {Hue[0.7208471572876194, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1043, 1044}]}, 
       {Hue[0.7221146028713816, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1044, 1045}]}, 
       {Hue[0.7246494940389063, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1045, 1046}]}, 
       {Hue[0.729719276373956, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1046, 1047}]}, 
       {Hue[0.7303529991658371, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1047, 1048}]}, 
       {Hue[0.7309867219577184, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1048, 1049}]}, 
       {Hue[0.7322541675414806, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1049, 1050}]}, 
       {Hue[0.7347890587090053, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1050, 1051}]}, 
       {Hue[0.7398588410440549, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1051, 1052}]}, 
       {Hue[0.7404925638359361, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1052, 1053}]}, 
       {Hue[0.7411262866278173, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1053, 1054}]}, 
       {Hue[0.7423937322115797, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1054, 1055}]}, 
       {Hue[0.7449286233791044, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1055, 1056}]}, 
       {Hue[0.7455623461709856, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1056, 1057}]}, 
       {Hue[0.7461960689628668, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1057, 1058}]}, 
       {Hue[0.7474635145466291, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1058, 1059}]}, 
       {Hue[0.7480972373385104, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1059, 1060}]}, 
       {Hue[0.7487309601303914, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1060, 1061}]}, 
       {Hue[0.7493646829222727, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1061, 1062}]}, 
       {Hue[0.7499984057141539, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1062, 1063}]}, 
       {Hue[0.2506321285060351, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1063, 1064}]}, 
       {Hue[0.2512658512979163, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1064, 1065}]}, 
       {Hue[0.25253329688167864`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1065, 1066}]}, 
       {Hue[0.25506818804920334`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1066, 1067}]}, 
       {Hue[0.26013797038425285`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1067, 1068}]}, 
       {Hue[0.26072961153291807`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1068, 1069}]}, 
       {Hue[0.2613212526815833, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1069, 1070}]}, 
       {Hue[0.26250453497891374`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1070, 1071}]}, 
       {Hue[0.26487109957357463`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1071, 1072}]}, 
       {Hue[0.2696042287628964, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1072, 1073}]}, 
       {Hue[0.2701958699115617, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1073, 1074}]}, 
       {Hue[0.2707875110602269, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1074, 1075}]}, 
       {Hue[0.27197079335755736`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1075, 1076}]}, 
       {Hue[0.27433735795221825`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1076, 1077}]}, 
       {Hue[0.27907048714154004`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1077, 1078}]}, 
       {Hue[0.2796621282902053, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1078, 1079}]}, 
       {Hue[0.2802537694388705, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1079, 1080}]}, 
       {Hue[0.281437051736201, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1080, 1081}]}, 
       {Hue[0.2838036163308618, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1081, 1082}]}, 
       {Hue[0.28439525747952704`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1082, 1083}]}, 
       {Hue[0.28498689862819226`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1083, 1084}]}, 
       {Hue[0.2861701809255227, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1084, 1085}]}, 
       {Hue[0.2885367455201836, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1085, 1086}]}, 
       {Hue[0.2891283866688488, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1086, 1087}]}, 
       {Hue[0.2897200278175141, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1087, 1088}]}, 
       {Hue[0.2909033101148445, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1088, 1089}]}, 
       {Hue[0.2932698747095054, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1089, 1090}]}, 
       {Hue[0.29386151585817066`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1090, 1091}]}, 
       {Hue[0.29445315700683583`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1091, 1092}]}, 
       {Hue[0.29563643930416633`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1092, 1093}]}, 
       {Hue[0.2980030038988271, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1093, 1094}]}, 
       {Hue[0.298583013967368, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1094, 1095}]}, 
       {Hue[0.29916302403590883`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1095, 1096}]}, 
       {Hue[0.30032304417299044`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1096, 1097}]}, 
       {Hue[0.30090305424153124`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1097, 1098}]}, 
       {Hue[0.30148306431007205`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1098, 1099}]}, 
       {Hue[0.3026430844471537, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1099, 1100}]}, 
       {Hue[0.3032230945156945, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1100, 1101}]}, 
       {Hue[0.3038031045842353, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1101, 1102}]}, 
       {Hue[0.3049631247213169, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1102, 1103}]}, 
       {Hue[0.30728316499548014`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1103, 1104}]}, 
       {Hue[0.30786317506402094`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1104, 1105}]}, 
       {Hue[0.3084431851325618, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1105, 1106}]}, 
       {Hue[0.3096032052696434, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1106, 1107}]}, 
       {Hue[0.31018321533818427`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1107, 1108}]}, 
       {Hue[0.310763225406725, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1108, 1109}]}, 
       {Hue[0.3119232455438066, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1109, 1110}]}, 
       {Hue[0.31250325561234743`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1110, 1111}]}, 
       {Hue[0.31308326568088823`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1111, 1112}]}, 
       {Hue[0.3142432858179699, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1112, 1113}]}, 
       {Hue[0.3148232958865107, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1113, 1114}]}, 
       {Hue[0.3154033059550515, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1114, 1115}]}, 
       {Hue[0.3165633260921333, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1115, 1116}]}, 
       {Hue[0.31714333616067425`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1116, 1117}]}, 
       {Hue[0.31772334622921505`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1117, 1118}]}, 
       {Hue[0.31830335629775586`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1118, 1119}]}, 
       {Hue[0.31888336636629666`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1119, 1120}]}, 
       {Hue[0.31946337643483746`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1120, 1121}]}, 
       {Hue[0.32004338650337827`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1121, 1122}]}, 
       {Hue[0.32062339657191913`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1122, 1123}]}, 
       {Hue[0.32120340664045993`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1123, 1124}]}, 
       {Hue[0.32178341670900074`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1124, 1125}]}, 
       {Hue[0.32236342677754154`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1125, 1126}]}, 
       {Hue[0.3229434368460823, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1126, 1127}]}, 
       {Hue[0.32352344691462315`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1127, 1128}]}, 
       {Hue[0.32410345698316395`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1128, 1129}]}, 
       {Hue[0.32468346705170475`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1129, 1130}]}, 
       {Hue[0.32526347712024556`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1130, 1131}]}, 
       {Hue[0.3258434871887864, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1131, 1132}]}, 
       {Hue[0.3264234972573272, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1132, 1133}]}, 
       {Hue[0.327003507325868, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1133, 1134}]}, 
       {Hue[0.32758351739440883`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1134, 1135}]}, 
       {Hue[0.32816352746294963`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1135, 1136}]}, 
       {Hue[0.32874353753149044`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1136, 1137}]}, 
       {Hue[0.32932354760003124`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1137, 1138}]}, 
       {Hue[0.32990355766857205`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1138, 1139}]}, 
       {Hue[0.33048356773711285`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1139, 1140}]}, 
       {Hue[0.3310635778056537, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1140, 1141}]}, 
       {Hue[0.3316435878741945, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1141, 1142}]}, 
       {Hue[0.3322235979427353, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1142, 1143}]}, 
       {Hue[0.3328036080112761, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1143, 1144}]}, 
       {Hue[0.3333836180798169, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1144, 1145}]}, 
       {Hue[0.3339636281483577, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1145, 1146}]}, 
       {Hue[0.3351236482854395, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1146, 1147}]}, 
       {Hue[0.33575292840003906`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1147, 1148}]}, 
       {Hue[0.33638220851463846`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1148, 1149}]}, 
       {Hue[0.3376407687438376, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1149, 1150}]}, 
       {Hue[0.33827004885843714`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1150, 1151}]}, 
       {Hue[0.3388993289730367, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1151, 1152}]}, 
       {Hue[0.34015788920223583`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1152, 1153}]}, 
       {Hue[0.3407871693168354, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1153, 1154}]}, 
       {Hue[0.34141644943143495`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1154, 1155}]}, 
       {Hue[0.34267500966063397`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1155, 1156}]}, 
       {Hue[0.34519213011903205`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1156, 1157}]}, 
       {Hue[0.3458214102336316, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1157, 1158}]}, 
       {Hue[0.34645069034823117`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1158, 1159}]}, 
       {Hue[0.3477092505774303, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1159, 1160}]}, 
       {Hue[0.3502263710358284, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1160, 1161}]}, 
       {Hue[0.35526061195262476`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1161, 1162}]}, 
       {Hue[0.35588989206722427`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1162, 1163}]}, 
       {Hue[0.3565191721818239, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1163, 1164}]}, 
       {Hue[0.35777773241102284`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1164, 1165}]}, 
       {Hue[0.360294852869421, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1165, 1166}]}, 
       {Hue[0.3653290937862173, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1166, 1167}]}, 
       {Hue[0.37539757561980985`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1167, 1168}]}, 
       {Hue[0.3759847740911934, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1168, 1169}]}, 
       {Hue[0.376571972562577, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1169, 1170}]}, 
       {Hue[0.3777463695053441, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1170, 1171}]}, 
       {Hue[0.3800951633908784, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1171, 1172}]}, 
       {Hue[0.38479275116194706`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1172, 1173}]}, 
       {Hue[0.3941879267040842, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1173, 1174}]}, 
       {Hue[0.41297827778835855`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1174, 1175}]}, 
       {Hue[0.41361474630580086`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1175, 1176}]}, 
       {Hue[0.4142512148232432, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1176, 1177}]}, 
       {Hue[0.41552415185812763`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1177, 1178}]}, 
       {Hue[0.41807002592789677`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1178, 1179}]}, 
       {Hue[0.42316177406743516`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1179, 1180}]}, 
       {Hue[0.43334527034651177`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1180, 1181}]}, 
       {Hue[0.45371226290466515`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1181, 1182}]}, 
       {Hue[0.4537312841026336, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1182, 1183}]}, 
       {Hue[0.4537503053006021, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1183, 1184}]}, 
       {Hue[0.45378834769653925`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1184, 1185}]}, 
       {Hue[0.4538644324884135, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1185, 1186}]}, 
       {Hue[0.4540166020721617, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1186, 1187}]}, 
       {Hue[0.45432094123965827`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1187, 1188}]}, 
       {Hue[0.4543399624376267, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1188, 1189}]}, 
       {Hue[0.45435898363559524`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1189, 1190}]}, 
       {Hue[0.45439702603153237`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1190, 1191}]}, 
       {Hue[0.45447311082340663`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1191, 1192}]}, 
       {Hue[0.45462528040715483`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1192, 1193}]}, 
       {Hue[0.45464430160512337`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1193, 1194}]}, 
       {Hue[0.45466332280309196`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1194, 1195}]}, 
       {Hue[0.45470136519902893`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1195, 1196}]}, 
       {Hue[0.454777449990903, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1196, 1197}]}, 
       {Hue[0.45479647118887156`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1197, 1198}]}, 
       {Hue[0.45481549238684016`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1198, 1199}]}, 
       {Hue[0.4548535347827773, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1199, 1200}]}, 
       {Hue[0.45487255598074583`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1200, 1201}]}, 
       {Hue[0.45489157717871437`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1201, 1202}]}, 
       {Hue[0.45491059837668296`, 
         NCache[
          Rational[1, 2], 0.5], 1], LineBox[{1202, 1203}]}}}},
    VertexColors->{
      Hue[0.04507038042534869, 
       Rational[1, 2], 1], 
      Hue[0.045656129526931086`, 
       Rational[1, 2], 1], 
      Hue[0.04624187862851364, 
       Rational[1, 2], 1], 
      Hue[0.047413376831678455`, 
       Rational[1, 2], 1], 
      Hue[0.04975637323800807, 
       Rational[1, 2], 1], 
      Hue[0.054442366050667615`, 
       Rational[1, 2], 1], 
      Hue[0.06381435167598638, 
       Rational[1, 2], 1], 
      Hue[0.08255832292662395, 
       Rational[1, 2], 1], 
      Hue[0.08319334207426511, 
       Rational[1, 2], 1], 
      Hue[0.08382836122190629, 
       Rational[1, 2], 1], 
      Hue[0.08509839951718859, 
       Rational[1, 2], 1], 
      Hue[0.08763847610775323, 
       Rational[1, 2], 1], 
      Hue[0.09271862928888232, 
       Rational[1, 2], 1], 
      Hue[0.10287893565114072`, 
       Rational[1, 2], 1], 
      Hue[0.10351395479878188`, 
       Rational[1, 2], 1], 
      Hue[0.10414897394642304`, 
       Rational[1, 2], 1], 
      Hue[0.10541901224170536`, 
       Rational[1, 2], 1], 
      Hue[0.10795908883226986`, 
       Rational[1, 2], 1], 
      Hue[0.11303924201339897`, 
       Rational[1, 2], 1], 
      Hue[0.12319954837565734`, 
       Rational[1, 2], 1], 
      Hue[0.12379248588008265`, 
       Rational[1, 2], 1], 
      Hue[0.12438542338450781`, 
       Rational[1, 2], 1], 
      Hue[0.12557129839335812`, 
       Rational[1, 2], 1], 
      Hue[0.12794304841105877`, 
       Rational[1, 2], 1], 
      Hue[0.13268654844646033`, 
       Rational[1, 2], 1], 
      Hue[0.1332794859508855, 
       Rational[1, 2], 1], 
      Hue[0.13387242345531064`, 
       Rational[1, 2], 1], 
      Hue[0.13505829846416098`, 
       Rational[1, 2], 1], 
      Hue[0.13743004848186177`, 
       Rational[1, 2], 1], 
      Hue[0.14217354851726322`, 
       Rational[1, 2], 1], 
      Hue[0.14276648602168834`, 
       Rational[1, 2], 1], 
      Hue[0.1433594235261135, 
       Rational[1, 2], 1], 
      Hue[0.14454529853496384`, 
       Rational[1, 2], 1], 
      Hue[0.1469170485526646, 
       Rational[1, 2], 1], 
      Hue[0.14750998605708993`, 
       Rational[1, 2], 1], 
      Hue[0.14810292356151508`, 
       Rational[1, 2], 1], 
      Hue[0.1492887985703654, 
       Rational[1, 2], 1], 
      Hue[0.15166054858806602`, 
       Rational[1, 2], 1], 
      Hue[0.1522534860924912, 
       Rational[1, 2], 1], 
      Hue[0.1528464235969164, 
       Rational[1, 2], 1], 
      Hue[0.15403229860576667`, 
       Rational[1, 2], 1], 
      Hue[0.15462523611019183`, 
       Rational[1, 2], 1], 
      Hue[0.155218173614617, 
       Rational[1, 2], 1], 
      Hue[0.15640404862346746`, 
       Rational[1, 2], 1], 
      Hue[0.15699698612789276`, 
       Rational[1, 2], 1], 
      Hue[0.15758992363231794`, 
       Rational[1, 2], 1], 
      Hue[0.15818286113674307`, 
       Rational[1, 2], 1], 
      Hue[0.15877579864116825`, 
       Rational[1, 2], 1], 
      Hue[0.1593687361455934, 
       Rational[1, 2], 1], 
      Hue[0.15996167365001857`, 
       Rational[1, 2], 1], 
      Hue[0.16114754865886888`, 
       Rational[1, 2], 1], 
      Hue[0.16172885508316964`, 
       Rational[1, 2], 1], 
      Hue[0.16231016150747038`, 
       Rational[1, 2], 1], 
      Hue[0.16289146793177126`, 
       Rational[1, 2], 1], 
      Hue[0.16347277435607202`, 
       Rational[1, 2], 1], 
      Hue[0.16405408078037276`, 
       Rational[1, 2], 1], 
      Hue[0.16463538720467363`, 
       Rational[1, 2], 1], 
      Hue[0.1652166936289744, 
       Rational[1, 2], 1], 
      Hue[0.16579800005327514`, 
       Rational[1, 2], 1], 
      Hue[0.16637930647757587`, 
       Rational[1, 2], 1], 
      Hue[0.16696061290187664`, 
       Rational[1, 2], 1], 
      Hue[0.1675419193261775, 
       Rational[1, 2], 1], 
      Hue[0.16812322575047825`, 
       Rational[1, 2], 1], 
      Hue[0.16870453217477901`, 
       Rational[1, 2], 1], 
      Hue[0.1692858385990799, 
       Rational[1, 2], 1], 
      Hue[0.16986714502338063`, 
       Rational[1, 2], 1], 
      Hue[0.1704484514476814, 
       Rational[1, 2], 1], 
      Hue[0.17102975787198213`, 
       Rational[1, 2], 1], 
      Hue[0.17161106429628287`, 
       Rational[1, 2], 1], 
      Hue[0.17219237072058377`, 
       Rational[1, 2], 1], 
      Hue[0.1727736771448845, 
       Rational[1, 2], 1], 
      Hue[0.17335498356918527`, 
       Rational[1, 2], 1], 
      Hue[0.17393628999348615`, 
       Rational[1, 2], 1], 
      Hue[0.17451759641778689`, 
       Rational[1, 2], 1], 
      Hue[0.17509890284208765`, 
       Rational[1, 2], 1], 
      Hue[0.1756802092663884, 
       Rational[1, 2], 1], 
      Hue[0.17626151569068912`, 
       Rational[1, 2], 1], 
      Hue[0.1768428221149899, 
       Rational[1, 2], 1], 
      Hue[0.17742412853929063`, 
       Rational[1, 2], 1], 
      Hue[0.17800543496359136`, 
       Rational[1, 2], 1], 
      Hue[0.1785867413878921, 
       Rational[1, 2], 1], 
      Hue[0.17974935423649374`, 
       Rational[1, 2], 1], 
      Hue[0.1803306606607945, 
       Rational[1, 2], 1], 
      Hue[0.18091196708509538`, 
       Rational[1, 2], 1], 
      Hue[0.18207457993369688`, 
       Rational[1, 2], 1], 
      Hue[0.18265588635799762`, 
       Rational[1, 2], 1], 
      Hue[0.18323719278229836`, 
       Rational[1, 2], 1], 
      Hue[0.18439980563089986`, 
       Rational[1, 2], 1], 
      Hue[0.1849811120552006, 
       Rational[1, 2], 1], 
      Hue[0.18556241847950136`, 
       Rational[1, 2], 1], 
      Hue[0.18672503132810297`, 
       Rational[1, 2], 1], 
      Hue[0.18905025702530612`, 
       Rational[1, 2], 1], 
      Hue[0.18963156344960685`, 
       Rational[1, 2], 1], 
      Hue[0.1902128698739076, 
       Rational[1, 2], 1], 
      Hue[0.19137548272250923`, 
       Rational[1, 2], 1], 
      Hue[0.19370070841971235`, 
       Rational[1, 2], 1], 
      Hue[0.1942820148440131, 
       Rational[1, 2], 1], 
      Hue[0.19486332126831385`, 
       Rational[1, 2], 1], 
      Hue[0.1960259341169155, 
       Rational[1, 2], 1], 
      Hue[0.1983511598141186, 
       Rational[1, 2], 1], 
      Hue[0.1989817362844781, 
       Rational[1, 2], 1], 
      Hue[0.1996123127548376, 
       Rational[1, 2], 1], 
      Hue[0.2008734656955566, 
       Rational[1, 2], 1], 
      Hue[0.2033957715769946, 
       Rational[1, 2], 1], 
      Hue[0.2040263480473541, 
       Rational[1, 2], 1], 
      Hue[0.20465692451771358`, 
       Rational[1, 2], 1], 
      Hue[0.20591807745843257`, 
       Rational[1, 2], 1], 
      Hue[0.20844038333987058`, 
       Rational[1, 2], 1], 
      Hue[0.20907095981023008`, 
       Rational[1, 2], 1], 
      Hue[0.20970153628058957`, 
       Rational[1, 2], 1], 
      Hue[0.21096268922130856`, 
       Rational[1, 2], 1], 
      Hue[0.21348499510274654`, 
       Rational[1, 2], 1], 
      Hue[0.21852960686562253`, 
       Rational[1, 2], 1], 
      Hue[0.21916018333598203`, 
       Rational[1, 2], 1], 
      Hue[0.21979075980634152`, 
       Rational[1, 2], 1], 
      Hue[0.2210519127470605, 
       Rational[1, 2], 1], 
      Hue[0.22357421862849852`, 
       Rational[1, 2], 1], 
      Hue[0.22861883039137448`, 
       Rational[1, 2], 1], 
      Hue[0.22924940686173398`, 
       Rational[1, 2], 1], 
      Hue[0.22987998333209347`, 
       Rational[1, 2], 1], 
      Hue[0.2311411362728125, 
       Rational[1, 2], 1], 
      Hue[0.23366344215425047`, 
       Rational[1, 2], 1], 
      Hue[0.23870805391712646`, 
       Rational[1, 2], 1], 
      Hue[0.23929654874426995`, 
       Rational[1, 2], 1], 
      Hue[0.23988504357141346`, 
       Rational[1, 2], 1], 
      Hue[0.2410620332257006, 
       Rational[1, 2], 1], 
      Hue[0.24341601253427475`, 
       Rational[1, 2], 1], 
      Hue[0.24400450736141824`, 
       Rational[1, 2], 1], 
      Hue[0.24459300218856173`, 
       Rational[1, 2], 1], 
      Hue[0.24576999184284887`, 
       Rational[1, 2], 1], 
      Hue[0.24635848666999252`, 
       Rational[1, 2], 1], 
      Hue[0.246946981497136, 
       Rational[1, 2], 1], 
      Hue[0.248123971151423, 
       Rational[1, 2], 1], 
      Hue[0.24871246597856653`, 
       Rational[1, 2], 1], 
      Hue[0.24930096080571001`, 
       Rational[1, 2], 1], 
      Hue[0.24988945563285353`, 
       Rational[1, 2], 1], 
      Hue[0.7504779504599972, 
       Rational[1, 2], 1], 
      Hue[0.7510664452871407, 
       Rational[1, 2], 1], 
      Hue[0.7516549401142842, 
       Rational[1, 2], 1], 
      Hue[0.7528319297685713, 
       Rational[1, 2], 1], 
      Hue[0.7534204245957148, 
       Rational[1, 2], 1], 
      Hue[0.7540089194228583, 
       Rational[1, 2], 1], 
      Hue[0.7551859090771453, 
       Rational[1, 2], 1], 
      Hue[0.7575398883857194, 
       Rational[1, 2], 1], 
      Hue[0.7581283832128631, 
       Rational[1, 2], 1], 
      Hue[0.7587168780400065, 
       Rational[1, 2], 1], 
      Hue[0.7598938676942936, 
       Rational[1, 2], 1], 
      Hue[0.7622478470028675, 
       Rational[1, 2], 1], 
      Hue[0.7669558056200159, 
       Rational[1, 2], 1], 
      Hue[0.7763717228543123, 
       Rational[1, 2], 1], 
      Hue[0.7770094877275147, 
       Rational[1, 2], 1], 
      Hue[0.777647252600717, 
       Rational[1, 2], 1], 
      Hue[0.7789227823471214, 
       Rational[1, 2], 1], 
      Hue[0.7814738418399304, 
       Rational[1, 2], 1], 
      Hue[0.7865759608255485, 
       Rational[1, 2], 1], 
      Hue[0.7967801987967844, 
       Rational[1, 2], 1], 
      Hue[0.8171886747392565, 
       Rational[1, 2], 1], 
      Hue[0.8178148085323343, 
       Rational[1, 2], 1], 
      Hue[0.8184409423254122, 
       Rational[1, 2], 1], 
      Hue[0.8196932099115679, 
       Rational[1, 2], 1], 
      Hue[0.8221977450838792, 
       Rational[1, 2], 1], 
      Hue[0.827206815428502, 
       Rational[1, 2], 1], 
      Hue[0.8372249561177476, 
       Rational[1, 2], 1], 
      Hue[0.8378510899108255, 
       Rational[1, 2], 1], 
      Hue[0.8384772237039033, 
       Rational[1, 2], 1], 
      Hue[0.8397294912900589, 
       Rational[1, 2], 1], 
      Hue[0.8422340264623704, 
       Rational[1, 2], 1], 
      Hue[0.8472430968069933, 
       Rational[1, 2], 1], 
      Hue[0.8572612374962388, 
       Rational[1, 2], 1], 
      Hue[0.8578452896461009, 
       Rational[1, 2], 1], 
      Hue[0.8584293417959626, 
       Rational[1, 2], 1], 
      Hue[0.8595974460956863, 
       Rational[1, 2], 1], 
      Hue[0.8619336546951339, 
       Rational[1, 2], 1], 
      Hue[0.8666060718940289, 
       Rational[1, 2], 1], 
      Hue[0.8671901240438908, 
       Rational[1, 2], 1], 
      Hue[0.8677741761937527, 
       Rational[1, 2], 1], 
      Hue[0.8689422804934764, 
       Rational[1, 2], 1], 
      Hue[0.8712784890929239, 
       Rational[1, 2], 1], 
      Hue[0.875950906291819, 
       Rational[1, 2], 1], 
      Hue[0.876534958441681, 
       Rational[1, 2], 1], 
      Hue[0.8771190105915428, 
       Rational[1, 2], 1], 
      Hue[0.8782871148912667, 
       Rational[1, 2], 1], 
      Hue[0.8806233234907141, 
       Rational[1, 2], 1], 
      Hue[0.8812073756405759, 
       Rational[1, 2], 1], 
      Hue[0.8817914277904378, 
       Rational[1, 2], 1], 
      Hue[0.8829595320901616, 
       Rational[1, 2], 1], 
      Hue[0.8852957406896091, 
       Rational[1, 2], 1], 
      Hue[0.8858797928394709, 
       Rational[1, 2], 1], 
      Hue[0.8864638449893328, 
       Rational[1, 2], 1], 
      Hue[0.8876319492890566, 
       Rational[1, 2], 1], 
      Hue[0.8882160014389185, 
       Rational[1, 2], 1], 
      Hue[0.8888000535887804, 
       Rational[1, 2], 1], 
      Hue[0.8899681578885041, 
       Rational[1, 2], 1], 
      Hue[0.890552210038366, 
       Rational[1, 2], 1], 
      Hue[0.891136262188228, 
       Rational[1, 2], 1], 
      Hue[0.8923043664879518, 
       Rational[1, 2], 1], 
      Hue[0.8928884186378137, 
       Rational[1, 2], 1], 
      Hue[0.8934724707876756, 
       Rational[1, 2], 1], 
      Hue[0.8946405750873992, 
       Rational[1, 2], 1], 
      Hue[0.8952738972833199, 
       Rational[1, 2], 1], 
      Hue[0.8959072194792405, 
       Rational[1, 2], 1], 
      Hue[0.8965405416751611, 
       Rational[1, 2], 1], 
      Hue[0.8971738638710816, 
       Rational[1, 2], 1], 
      Hue[0.8978071860670023, 
       Rational[1, 2], 1], 
      Hue[0.8984405082629228, 
       Rational[1, 2], 1], 
      Hue[0.8990738304588434, 
       Rational[1, 2], 1], 
      Hue[0.8997071526547641, 
       Rational[1, 2], 1], 
      Hue[0.9003404748506847, 
       Rational[1, 2], 1], 
      Hue[0.9009737970466053, 
       Rational[1, 2], 1], 
      Hue[0.9016071192425259, 
       Rational[1, 2], 1], 
      Hue[0.9022404414384465, 
       Rational[1, 2], 1], 
      Hue[0.9028737636343671, 
       Rational[1, 2], 1], 
      Hue[0.9035070858302877, 
       Rational[1, 2], 1], 
      Hue[0.9041404080262083, 
       Rational[1, 2], 1], 
      Hue[0.9047737302221288, 
       Rational[1, 2], 1], 
      Hue[0.9054070524180494, 
       Rational[1, 2], 1], 
      Hue[0.90604037461397, 
       Rational[1, 2], 1], 
      Hue[0.9066736968098906, 
       Rational[1, 2], 1], 
      Hue[0.9073070190058112, 
       Rational[1, 2], 1], 
      Hue[0.9079403412017318, 
       Rational[1, 2], 1], 
      Hue[0.9085736633976524, 
       Rational[1, 2], 1], 
      Hue[0.909206985593573, 
       Rational[1, 2], 1], 
      Hue[0.9098403077894937, 
       Rational[1, 2], 1], 
      Hue[0.9104736299854144, 
       Rational[1, 2], 1], 
      Hue[0.9111069521813349, 
       Rational[1, 2], 1], 
      Hue[0.9117402743772555, 
       Rational[1, 2], 1], 
      Hue[0.9123735965731761, 
       Rational[1, 2], 1], 
      Hue[0.9130069187690967, 
       Rational[1, 2], 1], 
      Hue[0.9136402409650173, 
       Rational[1, 2], 1], 
      Hue[0.9149068853568585, 
       Rational[1, 2], 1], 
      Hue[0.9155402075527791, 
       Rational[1, 2], 1], 
      Hue[0.9161735297486997, 
       Rational[1, 2], 1], 
      Hue[0.9174401741405409, 
       Rational[1, 2], 1], 
      Hue[0.9180734963364614, 
       Rational[1, 2], 1], 
      Hue[0.918706818532382, 
       Rational[1, 2], 1], 
      Hue[0.9199734629242233, 
       Rational[1, 2], 1], 
      Hue[0.9206067851201439, 
       Rational[1, 2], 1], 
      Hue[0.9212401073160645, 
       Rational[1, 2], 1], 
      Hue[0.9225067517079057, 
       Rational[1, 2], 1], 
      Hue[0.9250400404915881, 
       Rational[1, 2], 1], 
      Hue[0.9256733626875087, 
       Rational[1, 2], 1], 
      Hue[0.9263066848834293, 
       Rational[1, 2], 1], 
      Hue[0.9275733292752706, 
       Rational[1, 2], 1], 
      Hue[0.9301066180589529, 
       Rational[1, 2], 1], 
      Hue[0.9307399402548735, 
       Rational[1, 2], 1], 
      Hue[0.9313732624507941, 
       Rational[1, 2], 1], 
      Hue[0.9326399068426353, 
       Rational[1, 2], 1], 
      Hue[0.9351731956263177, 
       Rational[1, 2], 1], 
      Hue[0.9357644361790224, 
       Rational[1, 2], 1], 
      Hue[0.936355676731727, 
       Rational[1, 2], 1], 
      Hue[0.9375381578371363, 
       Rational[1, 2], 1], 
      Hue[0.9399031200479548, 
       Rational[1, 2], 1], 
      Hue[0.9404943606006595, 
       Rational[1, 2], 1], 
      Hue[0.9410856011533641, 
       Rational[1, 2], 1], 
      Hue[0.9422680822587733, 
       Rational[1, 2], 1], 
      Hue[0.9446330444695918, 
       Rational[1, 2], 1], 
      Hue[0.9452242850222965, 
       Rational[1, 2], 1], 
      Hue[0.9458155255750011, 
       Rational[1, 2], 1], 
      Hue[0.9469980066804105, 
       Rational[1, 2], 1], 
      Hue[0.949362968891229, 
       Rational[1, 2], 1], 
      Hue[0.9540928933128661, 
       Rational[1, 2], 1], 
      Hue[0.9546841338655707, 
       Rational[1, 2], 1], 
      Hue[0.9552753744182754, 
       Rational[1, 2], 1], 
      Hue[0.9564578555236847, 
       Rational[1, 2], 1], 
      Hue[0.9588228177345032, 
       Rational[1, 2], 1], 
      Hue[0.9635527421561403, 
       Rational[1, 2], 1], 
      Hue[0.9730125909994145, 
       Rational[1, 2], 1], 
      Hue[0.9735922004719948, 
       Rational[1, 2], 1], 
      Hue[0.974171809944575, 
       Rational[1, 2], 1], 
      Hue[0.9753310288897354, 
       Rational[1, 2], 1], 
      Hue[0.9776494667800564, 
       Rational[1, 2], 1], 
      Hue[0.9822863425606982, 
       Rational[1, 2], 1], 
      Hue[0.9828659520332785, 
       Rational[1, 2], 1], 
      Hue[0.9834455615058587, 
       Rational[1, 2], 1], 
      Hue[0.9846047804510192, 
       Rational[1, 2], 1], 
      Hue[0.9869232183413401, 
       Rational[1, 2], 1], 
      Hue[0.991560094121982, 
       Rational[1, 2], 1], 
      Hue[0.9921397035945623, 
       Rational[1, 2], 1], 
      Hue[0.9927193130671425, 
       Rational[1, 2], 1], 
      Hue[0.9938785320123029, 
       Rational[1, 2], 1], 
      Hue[0.9944581414848832, 
       Rational[1, 2], 1], 
      Hue[0.9950377509574634, 
       Rational[1, 2], 1], 
      Hue[0.9961969699026239, 
       Rational[1, 2], 1], 
      Hue[0.996776579375204, 
       Rational[1, 2], 1], 
      Hue[0.9973561888477843, 
       Rational[1, 2], 1], 
      Hue[0.9979357983203645, 
       Rational[1, 2], 1], 
      Hue[0.9985154077929448, 
       Rational[1, 2], 1], 
      Hue[0.999095017265525, 
       Rational[1, 2], 1], 
      Hue[0.9996746267381051, 
       Rational[1, 2], 1], 
      Hue[0.5002542362106854, 
       Rational[1, 2], 1], 
      Hue[0.5008338456832657, 
       Rational[1, 2], 1], 
      Hue[0.5014134551558459, 
       Rational[1, 2], 1], 
      Hue[0.5019930646284262, 
       Rational[1, 2], 1], 
      Hue[0.5031522835735867, 
       Rational[1, 2], 1], 
      Hue[0.5054707214639076, 
       Rational[1, 2], 1], 
      Hue[0.5101075972445495, 
       Rational[1, 2], 1], 
      Hue[0.5107364767631885, 
       Rational[1, 2], 1], 
      Hue[0.5113653562818274, 
       Rational[1, 2], 1], 
      Hue[0.5126231153191052, 
       Rational[1, 2], 1], 
      Hue[0.5151386333936612, 
       Rational[1, 2], 1], 
      Hue[0.5201696695427727, 
       Rational[1, 2], 1], 
      Hue[0.530231741840996, 
       Rational[1, 2], 1], 
      Hue[0.5503558864374425, 
       Rational[1, 2], 1], 
      Hue[0.5509426843128655, 
       Rational[1, 2], 1], 
      Hue[0.5515294821882883, 
       Rational[1, 2], 1], 
      Hue[0.5527030779391344, 
       Rational[1, 2], 1], 
      Hue[0.5550502694408264, 
       Rational[1, 2], 1], 
      Hue[0.5597446524442103, 
       Rational[1, 2], 1], 
      Hue[0.5691334184509782, 
       Rational[1, 2], 1], 
      Hue[0.5697202163264011, 
       Rational[1, 2], 1], 
      Hue[0.5703070142018241, 
       Rational[1, 2], 1], 
      Hue[0.5714806099526701, 
       Rational[1, 2], 1], 
      Hue[0.5738278014543621, 
       Rational[1, 2], 1], 
      Hue[0.578522184457746, 
       Rational[1, 2], 1], 
      Hue[0.5879109504645137, 
       Rational[1, 2], 1], 
      Hue[0.5885470183859953, 
       Rational[1, 2], 1], 
      Hue[0.5891830863074771, 
       Rational[1, 2], 1], 
      Hue[0.5904552221504404, 
       Rational[1, 2], 1], 
      Hue[0.5929994938363673, 
       Rational[1, 2], 1], 
      Hue[0.598088037208221, 
       Rational[1, 2], 1], 
      Hue[0.5987241051297028, 
       Rational[1, 2], 1], 
      Hue[0.5993601730511844, 
       Rational[1, 2], 1], 
      Hue[0.6006323088941479, 
       Rational[1, 2], 1], 
      Hue[0.6031765805800747, 
       Rational[1, 2], 1], 
      Hue[0.6082651239519283, 
       Rational[1, 2], 1], 
      Hue[0.60890119187341, 
       Rational[1, 2], 1], 
      Hue[0.6095372597948917, 
       Rational[1, 2], 1], 
      Hue[0.6108093956378551, 
       Rational[1, 2], 1], 
      Hue[0.6133536673237819, 
       Rational[1, 2], 1], 
      Hue[0.6139897352452637, 
       Rational[1, 2], 1], 
      Hue[0.6146258031667454, 
       Rational[1, 2], 1], 
      Hue[0.6158979390097088, 
       Rational[1, 2], 1], 
      Hue[0.6184422106956357, 
       Rational[1, 2], 1], 
      Hue[0.6190782786171174, 
       Rational[1, 2], 1], 
      Hue[0.619714346538599, 
       Rational[1, 2], 1], 
      Hue[0.6209864823815625, 
       Rational[1, 2], 1], 
      Hue[0.6216225503030441, 
       Rational[1, 2], 1], 
      Hue[0.6222586182245259, 
       Rational[1, 2], 1], 
      Hue[0.6235307540674894, 
       Rational[1, 2], 1], 
      Hue[0.6241668219889711, 
       Rational[1, 2], 1], 
      Hue[0.6248028899104527, 
       Rational[1, 2], 1], 
      Hue[0.6260750257534162, 
       Rational[1, 2], 1], 
      Hue[0.6267110936748979, 
       Rational[1, 2], 1], 
      Hue[0.6273471615963796, 
       Rational[1, 2], 1], 
      Hue[0.628619297439343, 
       Rational[1, 2], 1], 
      Hue[0.6292437342807002, 
       Rational[1, 2], 1], 
      Hue[0.6298681711220575, 
       Rational[1, 2], 1], 
      Hue[0.6311170448047723, 
       Rational[1, 2], 1], 
      Hue[0.6317414816461295, 
       Rational[1, 2], 1], 
      Hue[0.6323659184874868, 
       Rational[1, 2], 1], 
      Hue[0.6336147921702014, 
       Rational[1, 2], 1], 
      Hue[0.6342392290115587, 
       Rational[1, 2], 1], 
      Hue[0.6348636658529161, 
       Rational[1, 2], 1], 
      Hue[0.6354881026942734, 
       Rational[1, 2], 1], 
      Hue[0.6361125395356306, 
       Rational[1, 2], 1], 
      Hue[0.6367369763769879, 
       Rational[1, 2], 1], 
      Hue[0.6373614132183453, 
       Rational[1, 2], 1], 
      Hue[0.6379858500597027, 
       Rational[1, 2], 1], 
      Hue[0.6386102869010599, 
       Rational[1, 2], 1], 
      Hue[0.6392347237424172, 
       Rational[1, 2], 1], 
      Hue[0.6398591605837745, 
       Rational[1, 2], 1], 
      Hue[0.6404835974251318, 
       Rational[1, 2], 1], 
      Hue[0.6411080342664891, 
       Rational[1, 2], 1], 
      Hue[0.6417324711078464, 
       Rational[1, 2], 1], 
      Hue[0.6423569079492037, 
       Rational[1, 2], 1], 
      Hue[0.642981344790561, 
       Rational[1, 2], 1], 
      Hue[0.6436057816319183, 
       Rational[1, 2], 1], 
      Hue[0.6442302184732756, 
       Rational[1, 2], 1], 
      Hue[0.6448546553146329, 
       Rational[1, 2], 1], 
      Hue[0.6454790921559903, 
       Rational[1, 2], 1], 
      Hue[0.6461035289973476, 
       Rational[1, 2], 1], 
      Hue[0.6467279658387048, 
       Rational[1, 2], 1], 
      Hue[0.6473524026800621, 
       Rational[1, 2], 1], 
      Hue[0.6479768395214194, 
       Rational[1, 2], 1], 
      Hue[0.6486012763627766, 
       Rational[1, 2], 1], 
      Hue[0.649225713204134, 
       Rational[1, 2], 1], 
      Hue[0.6498501500454913, 
       Rational[1, 2], 1], 
      Hue[0.6504745868868487, 
       Rational[1, 2], 1], 
      Hue[0.6510990237282059, 
       Rational[1, 2], 1], 
      Hue[0.6517234605695632, 
       Rational[1, 2], 1], 
      Hue[0.6523478974109205, 
       Rational[1, 2], 1], 
      Hue[0.653596771093635, 
       Rational[1, 2], 1], 
      Hue[0.6542212079349923, 
       Rational[1, 2], 1], 
      Hue[0.6548456447763498, 
       Rational[1, 2], 1], 
      Hue[0.6560945184590643, 
       Rational[1, 2], 1], 
      Hue[0.6567189553004216, 
       Rational[1, 2], 1], 
      Hue[0.657343392141779, 
       Rational[1, 2], 1], 
      Hue[0.6585922658244936, 
       Rational[1, 2], 1], 
      Hue[0.6592167026658509, 
       Rational[1, 2], 1], 
      Hue[0.6598411395072081, 
       Rational[1, 2], 1], 
      Hue[0.6610900131899228, 
       Rational[1, 2], 1], 
      Hue[0.663587760555352, 
       Rational[1, 2], 1], 
      Hue[0.6642121973967092, 
       Rational[1, 2], 1], 
      Hue[0.6648366342380666, 
       Rational[1, 2], 1], 
      Hue[0.6660855079207813, 
       Rational[1, 2], 1], 
      Hue[0.6685832552862104, 
       Rational[1, 2], 1], 
      Hue[0.6691656104843517, 
       Rational[1, 2], 1], 
      Hue[0.6697479656824932, 
       Rational[1, 2], 1], 
      Hue[0.6709126760787758, 
       Rational[1, 2], 1], 
      Hue[0.6732420968713412, 
       Rational[1, 2], 1], 
      Hue[0.6738244520694825, 
       Rational[1, 2], 1], 
      Hue[0.6744068072676239, 
       Rational[1, 2], 1], 
      Hue[0.6755715176639066, 
       Rational[1, 2], 1], 
      Hue[0.6779009384564718, 
       Rational[1, 2], 1], 
      Hue[0.6784832936546132, 
       Rational[1, 2], 1], 
      Hue[0.6790656488527546, 
       Rational[1, 2], 1], 
      Hue[0.6802303592490373, 
       Rational[1, 2], 1], 
      Hue[0.6825597800416026, 
       Rational[1, 2], 1], 
      Hue[0.6872186216267333, 
       Rational[1, 2], 1], 
      Hue[0.6878009768248747, 
       Rational[1, 2], 1], 
      Hue[0.6883833320230159, 
       Rational[1, 2], 1], 
      Hue[0.6895480424192987, 
       Rational[1, 2], 1], 
      Hue[0.691877463211864, 
       Rational[1, 2], 1], 
      Hue[0.6965363047969948, 
       Rational[1, 2], 1], 
      Hue[0.7058539879672562, 
       Rational[1, 2], 1], 
      Hue[0.7064856132114563, 
       Rational[1, 2], 1], 
      Hue[0.7071172384556563, 
       Rational[1, 2], 1], 
      Hue[0.7083804889440564, 
       Rational[1, 2], 1], 
      Hue[0.7109069899208567, 
       Rational[1, 2], 1], 
      Hue[0.7159599918744571, 
       Rational[1, 2], 1], 
      Hue[0.7260659957816581, 
       Rational[1, 2], 1], 
      Hue[0.7266976210258582, 
       Rational[1, 2], 1], 
      Hue[0.7273292462700582, 
       Rational[1, 2], 1], 
      Hue[0.7285924967584583, 
       Rational[1, 2], 1], 
      Hue[0.7311189977352586, 
       Rational[1, 2], 1], 
      Hue[0.736171999688859, 
       Rational[1, 2], 1], 
      Hue[0.7368036249330591, 
       Rational[1, 2], 1], 
      Hue[0.7374352501772592, 
       Rational[1, 2], 1], 
      Hue[0.7386985006656592, 
       Rational[1, 2], 1], 
      Hue[0.7412250016424595, 
       Rational[1, 2], 1], 
      Hue[0.7418566268866595, 
       Rational[1, 2], 1], 
      Hue[0.7424882521308596, 
       Rational[1, 2], 1], 
      Hue[0.7437515026192598, 
       Rational[1, 2], 1], 
      Hue[0.74627800359606, 
       Rational[1, 2], 1], 
      Hue[0.7468675471970441, 
       Rational[1, 2], 1], 
      Hue[0.7474570907980282, 
       Rational[1, 2], 1], 
      Hue[0.7480466343990122, 
       Rational[1, 2], 1], 
      Hue[0.7486361779999964, 
       Rational[1, 2], 1], 
      Hue[0.7492257216009806, 
       Rational[1, 2], 1], 
      Hue[0.7498152652019645, 
       Rational[1, 2], 1], 
      Hue[0.25040480880294874`, 
       Rational[1, 2], 1], 
      Hue[0.2509943524039328, 
       Rational[1, 2], 1], 
      Hue[0.2515838960049169, 
       Rational[1, 2], 1], 
      Hue[0.25217343960590094`, 
       Rational[1, 2], 1], 
      Hue[0.2533525268078691, 
       Rational[1, 2], 1], 
      Hue[0.25571070121180556`, 
       Rational[1, 2], 1], 
      Hue[0.2563002448127896, 
       Rational[1, 2], 1], 
      Hue[0.25688978841377375`, 
       Rational[1, 2], 1], 
      Hue[0.25806887561574193`, 
       Rational[1, 2], 1], 
      Hue[0.2604270500196783, 
       Rational[1, 2], 1], 
      Hue[0.26514339882755106`, 
       Rational[1, 2], 1], 
      Hue[0.2657329424285351, 
       Rational[1, 2], 1], 
      Hue[0.26632248602951925`, 
       Rational[1, 2], 1], 
      Hue[0.26750157323148743`, 
       Rational[1, 2], 1], 
      Hue[0.2698597476354238, 
       Rational[1, 2], 1], 
      Hue[0.27457609644329656`, 
       Rational[1, 2], 1], 
      Hue[0.28400879405904206`, 
       Rational[1, 2], 1], 
      Hue[0.2846476077060849, 
       Rational[1, 2], 1], 
      Hue[0.2852864213531277, 
       Rational[1, 2], 1], 
      Hue[0.2865640486472133, 
       Rational[1, 2], 1], 
      Hue[0.28911930323538454`, 
       Rational[1, 2], 1], 
      Hue[0.2942298124117271, 
       Rational[1, 2], 1], 
      Hue[0.3044508307644121, 
       Rational[1, 2], 1], 
      Hue[0.3050896444114549, 
       Rational[1, 2], 1], 
      Hue[0.30572845805849774`, 
       Rational[1, 2], 1], 
      Hue[0.3070060853525834, 
       Rational[1, 2], 1], 
      Hue[0.3095613399407546, 
       Rational[1, 2], 1], 
      Hue[0.3146718491170971, 
       Rational[1, 2], 1], 
      Hue[0.32489286746978213`, 
       Rational[1, 2], 1], 
      Hue[0.3255200500367006, 
       Rational[1, 2], 1], 
      Hue[0.32614723260361894`, 
       Rational[1, 2], 1], 
      Hue[0.3274015977374558, 
       Rational[1, 2], 1], 
      Hue[0.3299103280051294, 
       Rational[1, 2], 1], 
      Hue[0.3349277885404768, 
       Rational[1, 2], 1], 
      Hue[0.33555497110739513`, 
       Rational[1, 2], 1], 
      Hue[0.3361821536743136, 
       Rational[1, 2], 1], 
      Hue[0.33743651880815034`, 
       Rational[1, 2], 1], 
      Hue[0.339945249075824, 
       Rational[1, 2], 1], 
      Hue[0.3449627096111713, 
       Rational[1, 2], 1], 
      Hue[0.3455898921780897, 
       Rational[1, 2], 1], 
      Hue[0.3462170747450081, 
       Rational[1, 2], 1], 
      Hue[0.347471439878845, 
       Rational[1, 2], 1], 
      Hue[0.34998017014651855`, 
       Rational[1, 2], 1], 
      Hue[0.350607352713437, 
       Rational[1, 2], 1], 
      Hue[0.35123453528035536`, 
       Rational[1, 2], 1], 
      Hue[0.35248890041419223`, 
       Rational[1, 2], 1], 
      Hue[0.35499763068186585`, 
       Rational[1, 2], 1], 
      Hue[0.3556248132487843, 
       Rational[1, 2], 1], 
      Hue[0.35625199581570266`, 
       Rational[1, 2], 1], 
      Hue[0.3575063609495395, 
       Rational[1, 2], 1], 
      Hue[0.358133543516458, 
       Rational[1, 2], 1], 
      Hue[0.35876072608337634`, 
       Rational[1, 2], 1], 
      Hue[0.36001509121721315`, 
       Rational[1, 2], 1], 
      Hue[0.36064227378413155`, 
       Rational[1, 2], 1], 
      Hue[0.36126945635105, 
       Rational[1, 2], 1], 
      Hue[0.3625238214848868, 
       Rational[1, 2], 1], 
      Hue[0.3631510040518052, 
       Rational[1, 2], 1], 
      Hue[0.3637781866187237, 
       Rational[1, 2], 1], 
      Hue[0.3650325517525605, 
       Rational[1, 2], 1], 
      Hue[0.3656176526762629, 
       Rational[1, 2], 1], 
      Hue[0.36620275359996535`, 
       Rational[1, 2], 1], 
      Hue[0.3673729554473702, 
       Rational[1, 2], 1], 
      Hue[0.36795805637107265`, 
       Rational[1, 2], 1], 
      Hue[0.3685431572947751, 
       Rational[1, 2], 1], 
      Hue[0.36971335914218006`, 
       Rational[1, 2], 1], 
      Hue[0.37029846006588246`, 
       Rational[1, 2], 1], 
      Hue[0.3708835609895849, 
       Rational[1, 2], 1], 
      Hue[0.37146866191328737`, 
       Rational[1, 2], 1], 
      Hue[0.3720537628369898, 
       Rational[1, 2], 1], 
      Hue[0.3726388637606922, 
       Rational[1, 2], 1], 
      Hue[0.37322396468439467`, 
       Rational[1, 2], 1], 
      Hue[0.3738090656080972, 
       Rational[1, 2], 1], 
      Hue[0.3743941665317996, 
       Rational[1, 2], 1], 
      Hue[0.3749792674555021, 
       Rational[1, 2], 1], 
      Hue[0.3755643683792045, 
       Rational[1, 2], 1], 
      Hue[0.37614946930290694`, 
       Rational[1, 2], 1], 
      Hue[0.3767345702266094, 
       Rational[1, 2], 1], 
      Hue[0.3773196711503118, 
       Rational[1, 2], 1], 
      Hue[0.37790477207401424`, 
       Rational[1, 2], 1], 
      Hue[0.3784898729977167, 
       Rational[1, 2], 1], 
      Hue[0.3790749739214192, 
       Rational[1, 2], 1], 
      Hue[0.37966007484512154`, 
       Rational[1, 2], 1], 
      Hue[0.38024517576882405`, 
       Rational[1, 2], 1], 
      Hue[0.3808302766925265, 
       Rational[1, 2], 1], 
      Hue[0.38141537761622896`, 
       Rational[1, 2], 1], 
      Hue[0.3820004785399314, 
       Rational[1, 2], 1], 
      Hue[0.3825855794636338, 
       Rational[1, 2], 1], 
      Hue[0.38317068038733626`, 
       Rational[1, 2], 1], 
      Hue[0.3837557813110387, 
       Rational[1, 2], 1], 
      Hue[0.38434088223474117`, 
       Rational[1, 2], 1], 
      Hue[0.3849259831584437, 
       Rational[1, 2], 1], 
      Hue[0.38551108408214607`, 
       Rational[1, 2], 1], 
      Hue[0.3860961850058485, 
       Rational[1, 2], 1], 
      Hue[0.386681285929551, 
       Rational[1, 2], 1], 
      Hue[0.38726638685325343`, 
       Rational[1, 2], 1], 
      Hue[0.3884365887006583, 
       Rational[1, 2], 1], 
      Hue[0.38902168962436073`, 
       Rational[1, 2], 1], 
      Hue[0.38960679054806313`, 
       Rational[1, 2], 1], 
      Hue[0.39077699239546804`, 
       Rational[1, 2], 1], 
      Hue[0.3913620933191705, 
       Rational[1, 2], 1], 
      Hue[0.391947194242873, 
       Rational[1, 2], 1], 
      Hue[0.39311739609027785`, 
       Rational[1, 2], 1], 
      Hue[0.3937024970139803, 
       Rational[1, 2], 1], 
      Hue[0.39428759793768275`, 
       Rational[1, 2], 1], 
      Hue[0.3954577997850876, 
       Rational[1, 2], 1], 
      Hue[0.3977982034798974, 
       Rational[1, 2], 1], 
      Hue[0.39838330440359987`, 
       Rational[1, 2], 1], 
      Hue[0.3989684053273023, 
       Rational[1, 2], 1], 
      Hue[0.4001386071747072, 
       Rational[1, 2], 1], 
      Hue[0.402479010869517, 
       Rational[1, 2], 1], 
      Hue[0.40311338183927814`, 
       Rational[1, 2], 1], 
      Hue[0.4037477528090393, 
       Rational[1, 2], 1], 
      Hue[0.4050164947485616, 
       Rational[1, 2], 1], 
      Hue[0.40755397862760634`, 
       Rational[1, 2], 1], 
      Hue[0.40818834959736744`, 
       Rational[1, 2], 1], 
      Hue[0.4088227205671286, 
       Rational[1, 2], 1], 
      Hue[0.4100914625066509, 
       Rational[1, 2], 1], 
      Hue[0.41262894638569564`, 
       Rational[1, 2], 1], 
      Hue[0.4132633173554568, 
       Rational[1, 2], 1], 
      Hue[0.413897688325218, 
       Rational[1, 2], 1], 
      Hue[0.4151664302647403, 
       Rational[1, 2], 1], 
      Hue[0.41770391414378494`, 
       Rational[1, 2], 1], 
      Hue[0.4227788819018743, 
       Rational[1, 2], 1], 
      Hue[0.4234132528716354, 
       Rational[1, 2], 1], 
      Hue[0.4240476238413966, 
       Rational[1, 2], 1], 
      Hue[0.4253163657809189, 
       Rational[1, 2], 1], 
      Hue[0.4278538496599636, 
       Rational[1, 2], 1], 
      Hue[0.4329288174180529, 
       Rational[1, 2], 1], 
      Hue[0.44307875293423155`, 
       Rational[1, 2], 1], 
      Hue[0.4436710422607767, 
       Rational[1, 2], 1], 
      Hue[0.44426333158732195`, 
       Rational[1, 2], 1], 
      Hue[0.44544791024041236`, 
       Rational[1, 2], 1], 
      Hue[0.4478170675465931, 
       Rational[1, 2], 1], 
      Hue[0.45255538215895474`, 
       Rational[1, 2], 1], 
      Hue[0.46203201138367794`, 
       Rational[1, 2], 1], 
      Hue[0.46262430071022315`, 
       Rational[1, 2], 1], 
      Hue[0.46321659003676835`, 
       Rational[1, 2], 1], 
      Hue[0.46440116868985876`, 
       Rational[1, 2], 1], 
      Hue[0.46677032599603957`, 
       Rational[1, 2], 1], 
      Hue[0.4715086406084012, 
       Rational[1, 2], 1], 
      Hue[0.4809852698331244, 
       Rational[1, 2], 1], 
      Hue[0.4815659280795452, 
       Rational[1, 2], 1], 
      Hue[0.48214658632596596`, 
       Rational[1, 2], 1], 
      Hue[0.4833079028188076, 
       Rational[1, 2], 1], 
      Hue[0.4856305358044908, 
       Rational[1, 2], 1], 
      Hue[0.4902758017758571, 
       Rational[1, 2], 1], 
      Hue[0.4908564600222779, 
       Rational[1, 2], 1], 
      Hue[0.4914371182686987, 
       Rational[1, 2], 1], 
      Hue[0.4925984347615403, 
       Rational[1, 2], 1], 
      Hue[0.4949210677472235, 
       Rational[1, 2], 1], 
      Hue[0.4955017259936443, 
       Rational[1, 2], 1], 
      Hue[0.49608238424006507`, 
       Rational[1, 2], 1], 
      Hue[0.4972437007329067, 
       Rational[1, 2], 1], 
      Hue[0.4978243589793275, 
       Rational[1, 2], 1], 
      Hue[0.4984050172257483, 
       Rational[1, 2], 1], 
      Hue[0.4989856754721691, 
       Rational[1, 2], 1], 
      Hue[0.4995663337185899, 
       Rational[1, 2], 1], 
      Hue[0.00014699196501062852`, 
       Rational[1, 2], 1], 
      Hue[0.0007276502114314253, 
       Rational[1, 2], 1], 
      Hue[0.001308308457852222, 
       Rational[1, 2], 1], 
      Hue[0.0018889667042730188`, 
       Rational[1, 2], 1], 
      Hue[0.002469624950693816, 
       Rational[1, 2], 1], 
      Hue[0.0030502831971146124`, 
       Rational[1, 2], 1], 
      Hue[0.004211599689956207, 
       Rational[1, 2], 1], 
      Hue[0.004792257936377003, 
       Rational[1, 2], 1], 
      Hue[0.005372916182797801, 
       Rational[1, 2], 1], 
      Hue[0.006534232675639393, 
       Rational[1, 2], 1], 
      Hue[0.00885686566132258, 
       Rational[1, 2], 1], 
      Hue[0.01814739760405533, 
       Rational[1, 2], 1], 
      Hue[0.018777325896534847`, 
       Rational[1, 2], 1], 
      Hue[0.019407254189014365`, 
       Rational[1, 2], 1], 
      Hue[0.020667110773973398`, 
       Rational[1, 2], 1], 
      Hue[0.02318682394389146, 
       Rational[1, 2], 1], 
      Hue[0.028226250283727595`, 
       Rational[1, 2], 1], 
      Hue[0.03830510296339986, 
       Rational[1, 2], 1], 
      Hue[0.03893503125587938, 
       Rational[1, 2], 1], 
      Hue[0.03956495954835889, 
       Rational[1, 2], 1], 
      Hue[0.04082481613331793, 
       Rational[1, 2], 1], 
      Hue[0.04334452930323598, 
       Rational[1, 2], 1], 
      Hue[0.048383955643072116`, 
       Rational[1, 2], 1], 
      Hue[0.058462808322744406`, 
       Rational[1, 2], 1], 
      Hue[0.05905065497200795, 
       Rational[1, 2], 1], 
      Hue[0.05963850162127149, 
       Rational[1, 2], 1], 
      Hue[0.0608141949197986, 
       Rational[1, 2], 1], 
      Hue[0.06316558151685281, 
       Rational[1, 2], 1], 
      Hue[0.06786835471096123, 
       Rational[1, 2], 1], 
      Hue[0.06845620136022479, 
       Rational[1, 2], 1], 
      Hue[0.06904404800948832, 
       Rational[1, 2], 1], 
      Hue[0.07021974130801542, 
       Rational[1, 2], 1], 
      Hue[0.07257112790506963, 
       Rational[1, 2], 1], 
      Hue[0.07727390109917805, 
       Rational[1, 2], 1], 
      Hue[0.0778617477484416, 
       Rational[1, 2], 1], 
      Hue[0.07844959439770514, 
       Rational[1, 2], 1], 
      Hue[0.07962528769623224, 
       Rational[1, 2], 1], 
      Hue[0.08197667429328648, 
       Rational[1, 2], 1], 
      Hue[0.08256452094255003, 
       Rational[1, 2], 1], 
      Hue[0.08315236759181359, 
       Rational[1, 2], 1], 
      Hue[0.08432806089034067, 
       Rational[1, 2], 1], 
      Hue[0.08667944748739487, 
       Rational[1, 2], 1], 
      Hue[0.08726729413665844, 
       Rational[1, 2], 1], 
      Hue[0.08785514078592198, 
       Rational[1, 2], 1], 
      Hue[0.08903083408444905, 
       Rational[1, 2], 1], 
      Hue[0.09138222068150327, 
       Rational[1, 2], 1], 
      Hue[0.09197006733076682, 
       Rational[1, 2], 1], 
      Hue[0.09255791398003038, 
       Rational[1, 2], 1], 
      Hue[0.0937336072785575, 
       Rational[1, 2], 1], 
      Hue[0.09608499387561169, 
       Rational[1, 2], 1], 
      Hue[0.09672211057093397, 
       Rational[1, 2], 1], 
      Hue[0.09735922726625623, 
       Rational[1, 2], 1], 
      Hue[0.09863346065690076, 
       Rational[1, 2], 1], 
      Hue[0.09927057735222307, 
       Rational[1, 2], 1], 
      Hue[0.09990769404754532, 
       Rational[1, 2], 1], 
      Hue[0.10118192743818985`, 
       Rational[1, 2], 1], 
      Hue[0.10181904413351212`, 
       Rational[1, 2], 1], 
      Hue[0.1024561608288344, 
       Rational[1, 2], 1], 
      Hue[0.10309327752415666`, 
       Rational[1, 2], 1], 
      Hue[0.10373039421947897`, 
       Rational[1, 2], 1], 
      Hue[0.10436751091480122`, 
       Rational[1, 2], 1], 
      Hue[0.10500462761012351`, 
       Rational[1, 2], 1], 
      Hue[0.10564174430544576`, 
       Rational[1, 2], 1], 
      Hue[0.10627886100076801`, 
       Rational[1, 2], 1], 
      Hue[0.10691597769609032`, 
       Rational[1, 2], 1], 
      Hue[0.10755309439141254`, 
       Rational[1, 2], 1], 
      Hue[0.10819021108673485`, 
       Rational[1, 2], 1], 
      Hue[0.10882732778205713`, 
       Rational[1, 2], 1], 
      Hue[0.10946444447737938`, 
       Rational[1, 2], 1], 
      Hue[0.11010156117270166`, 
       Rational[1, 2], 1], 
      Hue[0.11073867786802392`, 
       Rational[1, 2], 1], 
      Hue[0.11137579456334618`, 
       Rational[1, 2], 1], 
      Hue[0.11201291125866845`, 
       Rational[1, 2], 1], 
      Hue[0.11265002795399072`, 
       Rational[1, 2], 1], 
      Hue[0.11328714464931298`, 
       Rational[1, 2], 1], 
      Hue[0.11392426134463528`, 
       Rational[1, 2], 1], 
      Hue[0.11456137803995754`, 
       Rational[1, 2], 1], 
      Hue[0.11519849473527982`, 
       Rational[1, 2], 1], 
      Hue[0.11583561143060207`, 
       Rational[1, 2], 1], 
      Hue[0.11647272812592435`, 
       Rational[1, 2], 1], 
      Hue[0.11710984482124662`, 
       Rational[1, 2], 1], 
      Hue[0.11774696151656887`, 
       Rational[1, 2], 1], 
      Hue[0.11838407821189115`, 
       Rational[1, 2], 1], 
      Hue[0.11902119490721341`, 
       Rational[1, 2], 1], 
      Hue[0.1196583116025357, 
       Rational[1, 2], 1], 
      Hue[0.12029542829785798`, 
       Rational[1, 2], 1], 
      Hue[0.1215696616885025, 
       Rational[1, 2], 1], 
      Hue[0.12220677838382481`, 
       Rational[1, 2], 1], 
      Hue[0.12284389507914709`, 
       Rational[1, 2], 1], 
      Hue[0.12411812846979162`, 
       Rational[1, 2], 1], 
      Hue[0.12475524516511388`, 
       Rational[1, 2], 1], 
      Hue[0.12539236186043612`, 
       Rational[1, 2], 1], 
      Hue[0.1266665952510807, 
       Rational[1, 2], 1], 
      Hue[0.12730371194640297`, 
       Rational[1, 2], 1], 
      Hue[0.12794082864172526`, 
       Rational[1, 2], 1], 
      Hue[0.1292150620323698, 
       Rational[1, 2], 1], 
      Hue[0.12985217872769206`, 
       Rational[1, 2], 1], 
      Hue[0.13048929542301432`, 
       Rational[1, 2], 1], 
      Hue[0.13176352881365888`, 
       Rational[1, 2], 1], 
      Hue[0.13240064550898115`, 
       Rational[1, 2], 1], 
      Hue[0.13303776220430344`, 
       Rational[1, 2], 1], 
      Hue[0.13431199559494797`, 
       Rational[1, 2], 1], 
      Hue[0.13686046237623703`, 
       Rational[1, 2], 1], 
      Hue[0.1374859479914349, 
       Rational[1, 2], 1], 
      Hue[0.13811143360663275`, 
       Rational[1, 2], 1], 
      Hue[0.13936240483702853`, 
       Rational[1, 2], 1], 
      Hue[0.14186434729781996`, 
       Rational[1, 2], 1], 
      Hue[0.14248983291301787`, 
       Rational[1, 2], 1], 
      Hue[0.14311531852821574`, 
       Rational[1, 2], 1], 
      Hue[0.14436628975861146`, 
       Rational[1, 2], 1], 
      Hue[0.14686823221940293`, 
       Rational[1, 2], 1], 
      Hue[0.14749371783460077`, 
       Rational[1, 2], 1], 
      Hue[0.14811920344979865`, 
       Rational[1, 2], 1], 
      Hue[0.1493701746801944, 
       Rational[1, 2], 1], 
      Hue[0.15187211714098586`, 
       Rational[1, 2], 1], 
      Hue[0.15687600206256883`, 
       Rational[1, 2], 1], 
      Hue[0.1575014876777667, 
       Rational[1, 2], 1], 
      Hue[0.15812697329296455`, 
       Rational[1, 2], 1], 
      Hue[0.15937794452336027`, 
       Rational[1, 2], 1], 
      Hue[0.16187988698415173`, 
       Rational[1, 2], 1], 
      Hue[0.1668837719057347, 
       Rational[1, 2], 1], 
      Hue[0.1768915417489006, 
       Rational[1, 2], 1], 
      Hue[0.17747494572088252`, 
       Rational[1, 2], 1], 
      Hue[0.1780583496928644, 
       Rational[1, 2], 1], 
      Hue[0.17922515763682823`, 
       Rational[1, 2], 1], 
      Hue[0.18155877352475583`, 
       Rational[1, 2], 1], 
      Hue[0.18622600530061104`, 
       Rational[1, 2], 1], 
      Hue[0.19556046885232148`, 
       Rational[1, 2], 1], 
      Hue[0.1961438728243034, 
       Rational[1, 2], 1], 
      Hue[0.1967272767962853, 
       Rational[1, 2], 1], 
      Hue[0.1978940847402491, 
       Rational[1, 2], 1], 
      Hue[0.2002277006281767, 
       Rational[1, 2], 1], 
      Hue[0.20489493240403192`, 
       Rational[1, 2], 1], 
      Hue[0.2142293959557424, 
       Rational[1, 2], 1], 
      Hue[0.214862069973783, 
       Rational[1, 2], 1], 
      Hue[0.2154947439918236, 
       Rational[1, 2], 1], 
      Hue[0.21676009202790486`, 
       Rational[1, 2], 1], 
      Hue[0.21929078810006736`, 
       Rational[1, 2], 1], 
      Hue[0.22435218024439235`, 
       Rational[1, 2], 1], 
      Hue[0.22498485426243298`, 
       Rational[1, 2], 1], 
      Hue[0.2256175282804736, 
       Rational[1, 2], 1], 
      Hue[0.22688287631655485`, 
       Rational[1, 2], 1], 
      Hue[0.22941357238871735`, 
       Rational[1, 2], 1], 
      Hue[0.2344749645330423, 
       Rational[1, 2], 1], 
      Hue[0.23510763855108294`, 
       Rational[1, 2], 1], 
      Hue[0.23574031256912356`, 
       Rational[1, 2], 1], 
      Hue[0.2370056606052048, 
       Rational[1, 2], 1], 
      Hue[0.2395363566773673, 
       Rational[1, 2], 1], 
      Hue[0.24016903069540793`, 
       Rational[1, 2], 1], 
      Hue[0.24080170471344856`, 
       Rational[1, 2], 1], 
      Hue[0.2420670527495298, 
       Rational[1, 2], 1], 
      Hue[0.2445977488216923, 
       Rational[1, 2], 1], 
      Hue[0.24523042283973293`, 
       Rational[1, 2], 1], 
      Hue[0.24586309685777355`, 
       Rational[1, 2], 1], 
      Hue[0.2471284448938548, 
       Rational[1, 2], 1], 
      Hue[0.24776111891189542`, 
       Rational[1, 2], 1], 
      Hue[0.24839379292993605`, 
       Rational[1, 2], 1], 
      Hue[0.24902646694797667`, 
       Rational[1, 2], 1], 
      Hue[0.2496591409660173, 
       Rational[1, 2], 1], 
      Hue[0.7502918149840578, 
       Rational[1, 2], 1], 
      Hue[0.7509244890020985, 
       Rational[1, 2], 1], 
      Hue[0.7515571630201391, 
       Rational[1, 2], 1], 
      Hue[0.7521898370381798, 
       Rational[1, 2], 1], 
      Hue[0.7528225110562203, 
       Rational[1, 2], 1], 
      Hue[0.753455185074261, 
       Rational[1, 2], 1], 
      Hue[0.7547205331103422, 
       Rational[1, 2], 1], 
      Hue[0.7553111254851669, 
       Rational[1, 2], 1], 
      Hue[0.7559017178599915, 
       Rational[1, 2], 1], 
      Hue[0.7570829026096408, 
       Rational[1, 2], 1], 
      Hue[0.7594452721089394, 
       Rational[1, 2], 1], 
      Hue[0.7641700111075367, 
       Rational[1, 2], 1], 
      Hue[0.7647606034823614, 
       Rational[1, 2], 1], 
      Hue[0.765351195857186, 
       Rational[1, 2], 1], 
      Hue[0.7665323806068354, 
       Rational[1, 2], 1], 
      Hue[0.7688947501061341, 
       Rational[1, 2], 1], 
      Hue[0.7736194891047312, 
       Rational[1, 2], 1], 
      Hue[0.774210081479556, 
       Rational[1, 2], 1], 
      Hue[0.7748006738543806, 
       Rational[1, 2], 1], 
      Hue[0.7759818586040299, 
       Rational[1, 2], 1], 
      Hue[0.7783442281033285, 
       Rational[1, 2], 1], 
      Hue[0.7830689671019258, 
       Rational[1, 2], 1], 
      Hue[0.7925184450991203, 
       Rational[1, 2], 1], 
      Hue[0.7930974063938205, 
       Rational[1, 2], 1], 
      Hue[0.7936763676885208, 
       Rational[1, 2], 1], 
      Hue[0.7948342902779213, 
       Rational[1, 2], 1], 
      Hue[0.7971501354567223, 
       Rational[1, 2], 1], 
      Hue[0.8017818258143243, 
       Rational[1, 2], 1], 
      Hue[0.8023607871090246, 
       Rational[1, 2], 1], 
      Hue[0.8029397484037248, 
       Rational[1, 2], 1], 
      Hue[0.8040976709931253, 
       Rational[1, 2], 1], 
      Hue[0.8064135161719264, 
       Rational[1, 2], 1], 
      Hue[0.8110452065295284, 
       Rational[1, 2], 1], 
      Hue[0.8116241678242286, 
       Rational[1, 2], 1], 
      Hue[0.8122031291189289, 
       Rational[1, 2], 1], 
      Hue[0.8133610517083294, 
       Rational[1, 2], 1], 
      Hue[0.8156768968871304, 
       Rational[1, 2], 1], 
      Hue[0.8162558581818307, 
       Rational[1, 2], 1], 
      Hue[0.8168348194765309, 
       Rational[1, 2], 1], 
      Hue[0.8179927420659314, 
       Rational[1, 2], 1], 
      Hue[0.8203085872447325, 
       Rational[1, 2], 1], 
      Hue[0.8208875485394328, 
       Rational[1, 2], 1], 
      Hue[0.8214665098341329, 
       Rational[1, 2], 1], 
      Hue[0.8226244324235334, 
       Rational[1, 2], 1], 
      Hue[0.8249402776023345, 
       Rational[1, 2], 1], 
      Hue[0.8255192388970347, 
       Rational[1, 2], 1], 
      Hue[0.826098200191735, 
       Rational[1, 2], 1], 
      Hue[0.8272561227811355, 
       Rational[1, 2], 1], 
      Hue[0.8295719679599365, 
       Rational[1, 2], 1], 
      Hue[0.8302001993006954, 
       Rational[1, 2], 1], 
      Hue[0.8308284306414545, 
       Rational[1, 2], 1], 
      Hue[0.8320848933229724, 
       Rational[1, 2], 1], 
      Hue[0.8327131246637314, 
       Rational[1, 2], 1], 
      Hue[0.8333413560044904, 
       Rational[1, 2], 1], 
      Hue[0.8345978186860082, 
       Rational[1, 2], 1], 
      Hue[0.8352260500267673, 
       Rational[1, 2], 1], 
      Hue[0.8358542813675263, 
       Rational[1, 2], 1], 
      Hue[0.8371107440490442, 
       Rational[1, 2], 1], 
      Hue[0.8377389753898031, 
       Rational[1, 2], 1], 
      Hue[0.8383672067305622, 
       Rational[1, 2], 1], 
      Hue[0.8396236694120801, 
       Rational[1, 2], 1], 
      Hue[0.8402519007528391, 
       Rational[1, 2], 1], 
      Hue[0.8408801320935981, 
       Rational[1, 2], 1], 
      Hue[0.841508363434357, 
       Rational[1, 2], 1], 
      Hue[0.842136594775116, 
       Rational[1, 2], 1], 
      Hue[0.8427648261158749, 
       Rational[1, 2], 1], 
      Hue[0.843393057456634, 
       Rational[1, 2], 1], 
      Hue[0.8440212887973929, 
       Rational[1, 2], 1], 
      Hue[0.8446495201381519, 
       Rational[1, 2], 1], 
      Hue[0.8452777514789109, 
       Rational[1, 2], 1], 
      Hue[0.8459059828196699, 
       Rational[1, 2], 1], 
      Hue[0.8465342141604288, 
       Rational[1, 2], 1], 
      Hue[0.8471624455011878, 
       Rational[1, 2], 1], 
      Hue[0.8477906768419468, 
       Rational[1, 2], 1], 
      Hue[0.8484189081827058, 
       Rational[1, 2], 1], 
      Hue[0.8490471395234647, 
       Rational[1, 2], 1], 
      Hue[0.8496753708642237, 
       Rational[1, 2], 1], 
      Hue[0.8503036022049827, 
       Rational[1, 2], 1], 
      Hue[0.8509318335457416, 
       Rational[1, 2], 1], 
      Hue[0.8515600648865006, 
       Rational[1, 2], 1], 
      Hue[0.8521882962272596, 
       Rational[1, 2], 1], 
      Hue[0.8528165275680185, 
       Rational[1, 2], 1], 
      Hue[0.8534447589087775, 
       Rational[1, 2], 1], 
      Hue[0.8540729902495365, 
       Rational[1, 2], 1], 
      Hue[0.8547012215902955, 
       Rational[1, 2], 1], 
      Hue[0.8553294529310544, 
       Rational[1, 2], 1], 
      Hue[0.8559576842718135, 
       Rational[1, 2], 1], 
      Hue[0.8565859156125724, 
       Rational[1, 2], 1], 
      Hue[0.8572141469533314, 
       Rational[1, 2], 1], 
      Hue[0.8578423782940904, 
       Rational[1, 2], 1], 
      Hue[0.8584706096348493, 
       Rational[1, 2], 1], 
      Hue[0.8597270723163672, 
       Rational[1, 2], 1], 
      Hue[0.8603553036571262, 
       Rational[1, 2], 1], 
      Hue[0.8609835349978852, 
       Rational[1, 2], 1], 
      Hue[0.8622399976794031, 
       Rational[1, 2], 1], 
      Hue[0.862868229020162, 
       Rational[1, 2], 1], 
      Hue[0.863496460360921, 
       Rational[1, 2], 1], 
      Hue[0.864752923042439, 
       Rational[1, 2], 1], 
      Hue[0.865381154383198, 
       Rational[1, 2], 1], 
      Hue[0.866009385723957, 
       Rational[1, 2], 1], 
      Hue[0.8672658484054749, 
       Rational[1, 2], 1], 
      Hue[0.8697787737685108, 
       Rational[1, 2], 1], 
      Hue[0.8703649234660538, 
       Rational[1, 2], 1], 
      Hue[0.8709510731635968, 
       Rational[1, 2], 1], 
      Hue[0.8721233725586828, 
       Rational[1, 2], 1], 
      Hue[0.8744679713488549, 
       Rational[1, 2], 1], 
      Hue[0.8750541210463978, 
       Rational[1, 2], 1], 
      Hue[0.8756402707439409, 
       Rational[1, 2], 1], 
      Hue[0.8768125701390269, 
       Rational[1, 2], 1], 
      Hue[0.8791571689291989, 
       Rational[1, 2], 1], 
      Hue[0.879743318626742, 
       Rational[1, 2], 1], 
      Hue[0.8803294683242849, 
       Rational[1, 2], 1], 
      Hue[0.8815017677193709, 
       Rational[1, 2], 1], 
      Hue[0.883846366509543, 
       Rational[1, 2], 1], 
      Hue[0.8885355640898871, 
       Rational[1, 2], 1], 
      Hue[0.88912171378743, 
       Rational[1, 2], 1], 
      Hue[0.8897078634849731, 
       Rational[1, 2], 1], 
      Hue[0.8908801628800591, 
       Rational[1, 2], 1], 
      Hue[0.8932247616702311, 
       Rational[1, 2], 1], 
      Hue[0.8979139592505752, 
       Rational[1, 2], 1], 
      Hue[0.9072923544112633, 
       Rational[1, 2], 1], 
      Hue[0.907927774154865, 
       Rational[1, 2], 1], 
      Hue[0.9085631938984667, 
       Rational[1, 2], 1], 
      Hue[0.9098340333856701, 
       Rational[1, 2], 1], 
      Hue[0.9123757123600771, 
       Rational[1, 2], 1], 
      Hue[0.917459070308891, 
       Rational[1, 2], 1], 
      Hue[0.9276257862065186, 
       Rational[1, 2], 1], 
      Hue[0.947959218001774, 
       Rational[1, 2], 1], 
      Hue[0.9485525561021597, 
       Rational[1, 2], 1], 
      Hue[0.9491458942025454, 
       Rational[1, 2], 1], 
      Hue[0.9503325704033169, 
       Rational[1, 2], 1], 
      Hue[0.95270592280486, 
       Rational[1, 2], 1], 
      Hue[0.9574526276079461, 
       Rational[1, 2], 1], 
      Hue[0.9669460372141183, 
       Rational[1, 2], 1], 
      Hue[0.9675393753145042, 
       Rational[1, 2], 1], 
      Hue[0.9681327134148898, 
       Rational[1, 2], 1], 
      Hue[0.9693193896156613, 
       Rational[1, 2], 1], 
      Hue[0.9716927420172045, 
       Rational[1, 2], 1], 
      Hue[0.9764394468202906, 
       Rational[1, 2], 1], 
      Hue[0.9859328564264628, 
       Rational[1, 2], 1], 
      Hue[0.9865145634467241, 
       Rational[1, 2], 1], 
      Hue[0.9870962704669854, 
       Rational[1, 2], 1], 
      Hue[0.9882596845075081, 
       Rational[1, 2], 1], 
      Hue[0.9905865125885537, 
       Rational[1, 2], 1], 
      Hue[0.9911682196088151, 
       Rational[1, 2], 1], 
      Hue[0.9917499266290765, 
       Rational[1, 2], 1], 
      Hue[0.9929133406695991, 
       Rational[1, 2], 1], 
      Hue[0.9952401687506446, 
       Rational[1, 2], 1], 
      Hue[0.9958218757709059, 
       Rational[1, 2], 1], 
      Hue[0.9964035827911673, 
       Rational[1, 2], 1], 
      Hue[0.9975669968316899, 
       Rational[1, 2], 1], 
      Hue[0.9981487038519513, 
       Rational[1, 2], 1], 
      Hue[0.9987304108722127, 
       Rational[1, 2], 1], 
      Hue[0.999312117892474, 
       Rational[1, 2], 1], 
      Hue[0.9998938249127354, 
       Rational[1, 2], 1], 
      Hue[0.5004755319329968, 
       Rational[1, 2], 1], 
      Hue[0.5010572389532582, 
       Rational[1, 2], 1], 
      Hue[0.5016389459735195, 
       Rational[1, 2], 1], 
      Hue[0.5022206529937807, 
       Rational[1, 2], 1], 
      Hue[0.5045474810748263, 
       Rational[1, 2], 1], 
      Hue[0.5051291880950876, 
       Rational[1, 2], 1], 
      Hue[0.505710895115349, 
       Rational[1, 2], 1], 
      Hue[0.5068743091558716, 
       Rational[1, 2], 1], 
      Hue[0.5092011372369171, 
       Rational[1, 2], 1], 
      Hue[0.513854793399008, 
       Rational[1, 2], 1], 
      Hue[0.5231621057231897, 
       Rational[1, 2], 1], 
      Hue[0.5237930827895098, 
       Rational[1, 2], 1], 
      Hue[0.5244240598558298, 
       Rational[1, 2], 1], 
      Hue[0.52568601398847, 
       Rational[1, 2], 1], 
      Hue[0.5282099222537503, 
       Rational[1, 2], 1], 
      Hue[0.5332577387843109, 
       Rational[1, 2], 1], 
      Hue[0.533888715850631, 
       Rational[1, 2], 1], 
      Hue[0.5345196929169511, 
       Rational[1, 2], 1], 
      Hue[0.5357816470495913, 
       Rational[1, 2], 1], 
      Hue[0.5383055553148717, 
       Rational[1, 2], 1], 
      Hue[0.5433533718454322, 
       Rational[1, 2], 1], 
      Hue[0.5439843489117524, 
       Rational[1, 2], 1], 
      Hue[0.5446153259780724, 
       Rational[1, 2], 1], 
      Hue[0.5458772801107126, 
       Rational[1, 2], 1], 
      Hue[0.5484011883759928, 
       Rational[1, 2], 1], 
      Hue[0.5490321654423129, 
       Rational[1, 2], 1], 
      Hue[0.549663142508633, 
       Rational[1, 2], 1], 
      Hue[0.5509250966412733, 
       Rational[1, 2], 1], 
      Hue[0.5534490049065535, 
       Rational[1, 2], 1], 
      Hue[0.5540799819728736, 
       Rational[1, 2], 1], 
      Hue[0.5547109590391937, 
       Rational[1, 2], 1], 
      Hue[0.5559729131718338, 
       Rational[1, 2], 1], 
      Hue[0.5584968214371142, 
       Rational[1, 2], 1], 
      Hue[0.5591277985034342, 
       Rational[1, 2], 1], 
      Hue[0.5597587755697544, 
       Rational[1, 2], 1], 
      Hue[0.5610207297023944, 
       Rational[1, 2], 1], 
      Hue[0.5635446379676747, 
       Rational[1, 2], 1], 
      Hue[0.5641335333907789, 
       Rational[1, 2], 1], 
      Hue[0.564722428813883, 
       Rational[1, 2], 1], 
      Hue[0.5659002196600912, 
       Rational[1, 2], 1], 
      Hue[0.5664891150831952, 
       Rational[1, 2], 1], 
      Hue[0.5670780105062995, 
       Rational[1, 2], 1], 
      Hue[0.5682558013525076, 
       Rational[1, 2], 1], 
      Hue[0.5688446967756118, 
       Rational[1, 2], 1], 
      Hue[0.5694335921987159, 
       Rational[1, 2], 1], 
      Hue[0.5706113830449241, 
       Rational[1, 2], 1], 
      Hue[0.5712002784680281, 
       Rational[1, 2], 1], 
      Hue[0.5717891738911324, 
       Rational[1, 2], 1], 
      Hue[0.5729669647373405, 
       Rational[1, 2], 1], 
      Hue[0.5735558601604447, 
       Rational[1, 2], 1], 
      Hue[0.5741447555835488, 
       Rational[1, 2], 1], 
      Hue[0.5753225464297571, 
       Rational[1, 2], 1], 
      Hue[0.575911441852861, 
       Rational[1, 2], 1], 
      Hue[0.5765003372759653, 
       Rational[1, 2], 1], 
      Hue[0.5776781281221735, 
       Rational[1, 2], 1], 
      Hue[0.5782670235452776, 
       Rational[1, 2], 1], 
      Hue[0.5788559189683817, 
       Rational[1, 2], 1], 
      Hue[0.5794448143914859, 
       Rational[1, 2], 1], 
      Hue[0.5800337098145899, 
       Rational[1, 2], 1], 
      Hue[0.580622605237694, 
       Rational[1, 2], 1], 
      Hue[0.5812115006607981, 
       Rational[1, 2], 1], 
      Hue[0.5818003960839022, 
       Rational[1, 2], 1], 
      Hue[0.5823892915070064, 
       Rational[1, 2], 1], 
      Hue[0.5829781869301105, 
       Rational[1, 2], 1], 
      Hue[0.5835670823532146, 
       Rational[1, 2], 1], 
      Hue[0.5841559777763188, 
       Rational[1, 2], 1], 
      Hue[0.5847448731994228, 
       Rational[1, 2], 1], 
      Hue[0.585333768622527, 
       Rational[1, 2], 1], 
      Hue[0.585922664045631, 
       Rational[1, 2], 1], 
      Hue[0.5865115594687351, 
       Rational[1, 2], 1], 
      Hue[0.5871004548918393, 
       Rational[1, 2], 1], 
      Hue[0.5876893503149434, 
       Rational[1, 2], 1], 
      Hue[0.5882782457380475, 
       Rational[1, 2], 1], 
      Hue[0.5888671411611517, 
       Rational[1, 2], 1], 
      Hue[0.5894560365842557, 
       Rational[1, 2], 1], 
      Hue[0.5900449320073597, 
       Rational[1, 2], 1], 
      Hue[0.5906338274304639, 
       Rational[1, 2], 1], 
      Hue[0.591222722853568, 
       Rational[1, 2], 1], 
      Hue[0.5918116182766722, 
       Rational[1, 2], 1], 
      Hue[0.5924005136997763, 
       Rational[1, 2], 1], 
      Hue[0.5929894091228805, 
       Rational[1, 2], 1], 
      Hue[0.5935783045459846, 
       Rational[1, 2], 1], 
      Hue[0.5941671999690887, 
       Rational[1, 2], 1], 
      Hue[0.5947560953921927, 
       Rational[1, 2], 1], 
      Hue[0.5953449908152968, 
       Rational[1, 2], 1], 
      Hue[0.5965227816615051, 
       Rational[1, 2], 1], 
      Hue[0.5971116770846092, 
       Rational[1, 2], 1], 
      Hue[0.5977005725077134, 
       Rational[1, 2], 1], 
      Hue[0.5988783633539216, 
       Rational[1, 2], 1], 
      Hue[0.5994672587770258, 
       Rational[1, 2], 1], 
      Hue[0.6000561542001298, 
       Rational[1, 2], 1], 
      Hue[0.601233945046338, 
       Rational[1, 2], 1], 
      Hue[0.6018721105155009, 
       Rational[1, 2], 1], 
      Hue[0.6025102759846637, 
       Rational[1, 2], 1], 
      Hue[0.6037866069229894, 
       Rational[1, 2], 1], 
      Hue[0.6063392687996407, 
       Rational[1, 2], 1], 
      Hue[0.6069774342688035, 
       Rational[1, 2], 1], 
      Hue[0.6076155997379664, 
       Rational[1, 2], 1], 
      Hue[0.6088919306762921, 
       Rational[1, 2], 1], 
      Hue[0.6114445925529435, 
       Rational[1, 2], 1], 
      Hue[0.6120827580221063, 
       Rational[1, 2], 1], 
      Hue[0.612720923491269, 
       Rational[1, 2], 1], 
      Hue[0.6139972544295947, 
       Rational[1, 2], 1], 
      Hue[0.6165499163062461, 
       Rational[1, 2], 1], 
      Hue[0.6216552400595488, 
       Rational[1, 2], 1], 
      Hue[0.6222934055287116, 
       Rational[1, 2], 1], 
      Hue[0.6229315709978744, 
       Rational[1, 2], 1], 
      Hue[0.6242079019362001, 
       Rational[1, 2], 1], 
      Hue[0.6267605638128514, 
       Rational[1, 2], 1], 
      Hue[0.6318658875661541, 
       Rational[1, 2], 1], 
      Hue[0.6420765350727594, 
       Rational[1, 2], 1], 
      Hue[0.6427030694617979, 
       Rational[1, 2], 1], 
      Hue[0.6433296038508364, 
       Rational[1, 2], 1], 
      Hue[0.6445826726289132, 
       Rational[1, 2], 1], 
      Hue[0.6470888101850669, 
       Rational[1, 2], 1], 
      Hue[0.6521010852973743, 
       Rational[1, 2], 1], 
      Hue[0.6621256355219893, 
       Rational[1, 2], 1], 
      Hue[0.6821747359712191, 
       Rational[1, 2], 1], 
      Hue[0.6827591887170417, 
       Rational[1, 2], 1], 
      Hue[0.6833436414628641, 
       Rational[1, 2], 1], 
      Hue[0.684512546954509, 
       Rational[1, 2], 1], 
      Hue[0.6868503579377989, 
       Rational[1, 2], 1], 
      Hue[0.6915259799043787, 
       Rational[1, 2], 1], 
      Hue[0.700877223837538, 
       Rational[1, 2], 1], 
      Hue[0.7014616765833603, 
       Rational[1, 2], 1], 
      Hue[0.7020461293291829, 
       Rational[1, 2], 1], 
      Hue[0.7032150348208277, 
       Rational[1, 2], 1], 
      Hue[0.7055528458041176, 
       Rational[1, 2], 1], 
      Hue[0.7102284677706974, 
       Rational[1, 2], 1], 
      Hue[0.7195797117038569, 
       Rational[1, 2], 1], 
      Hue[0.7202134344957382, 
       Rational[1, 2], 1], 
      Hue[0.7208471572876194, 
       Rational[1, 2], 1], 
      Hue[0.7221146028713816, 
       Rational[1, 2], 1], 
      Hue[0.7246494940389063, 
       Rational[1, 2], 1], 
      Hue[0.729719276373956, 
       Rational[1, 2], 1], 
      Hue[0.7303529991658371, 
       Rational[1, 2], 1], 
      Hue[0.7309867219577184, 
       Rational[1, 2], 1], 
      Hue[0.7322541675414806, 
       Rational[1, 2], 1], 
      Hue[0.7347890587090053, 
       Rational[1, 2], 1], 
      Hue[0.7398588410440549, 
       Rational[1, 2], 1], 
      Hue[0.7404925638359361, 
       Rational[1, 2], 1], 
      Hue[0.7411262866278173, 
       Rational[1, 2], 1], 
      Hue[0.7423937322115797, 
       Rational[1, 2], 1], 
      Hue[0.7449286233791044, 
       Rational[1, 2], 1], 
      Hue[0.7455623461709856, 
       Rational[1, 2], 1], 
      Hue[0.7461960689628668, 
       Rational[1, 2], 1], 
      Hue[0.7474635145466291, 
       Rational[1, 2], 1], 
      Hue[0.7480972373385104, 
       Rational[1, 2], 1], 
      Hue[0.7487309601303914, 
       Rational[1, 2], 1], 
      Hue[0.7493646829222727, 
       Rational[1, 2], 1], 
      Hue[0.7499984057141539, 
       Rational[1, 2], 1], 
      Hue[0.2506321285060351, 
       Rational[1, 2], 1], 
      Hue[0.2512658512979163, 
       Rational[1, 2], 1], 
      Hue[0.25253329688167864`, 
       Rational[1, 2], 1], 
      Hue[0.25506818804920334`, 
       Rational[1, 2], 1], 
      Hue[0.26013797038425285`, 
       Rational[1, 2], 1], 
      Hue[0.26072961153291807`, 
       Rational[1, 2], 1], 
      Hue[0.2613212526815833, 
       Rational[1, 2], 1], 
      Hue[0.26250453497891374`, 
       Rational[1, 2], 1], 
      Hue[0.26487109957357463`, 
       Rational[1, 2], 1], 
      Hue[0.2696042287628964, 
       Rational[1, 2], 1], 
      Hue[0.2701958699115617, 
       Rational[1, 2], 1], 
      Hue[0.2707875110602269, 
       Rational[1, 2], 1], 
      Hue[0.27197079335755736`, 
       Rational[1, 2], 1], 
      Hue[0.27433735795221825`, 
       Rational[1, 2], 1], 
      Hue[0.27907048714154004`, 
       Rational[1, 2], 1], 
      Hue[0.2796621282902053, 
       Rational[1, 2], 1], 
      Hue[0.2802537694388705, 
       Rational[1, 2], 1], 
      Hue[0.281437051736201, 
       Rational[1, 2], 1], 
      Hue[0.2838036163308618, 
       Rational[1, 2], 1], 
      Hue[0.28439525747952704`, 
       Rational[1, 2], 1], 
      Hue[0.28498689862819226`, 
       Rational[1, 2], 1], 
      Hue[0.2861701809255227, 
       Rational[1, 2], 1], 
      Hue[0.2885367455201836, 
       Rational[1, 2], 1], 
      Hue[0.2891283866688488, 
       Rational[1, 2], 1], 
      Hue[0.2897200278175141, 
       Rational[1, 2], 1], 
      Hue[0.2909033101148445, 
       Rational[1, 2], 1], 
      Hue[0.2932698747095054, 
       Rational[1, 2], 1], 
      Hue[0.29386151585817066`, 
       Rational[1, 2], 1], 
      Hue[0.29445315700683583`, 
       Rational[1, 2], 1], 
      Hue[0.29563643930416633`, 
       Rational[1, 2], 1], 
      Hue[0.2980030038988271, 
       Rational[1, 2], 1], 
      Hue[0.298583013967368, 
       Rational[1, 2], 1], 
      Hue[0.29916302403590883`, 
       Rational[1, 2], 1], 
      Hue[0.30032304417299044`, 
       Rational[1, 2], 1], 
      Hue[0.30090305424153124`, 
       Rational[1, 2], 1], 
      Hue[0.30148306431007205`, 
       Rational[1, 2], 1], 
      Hue[0.3026430844471537, 
       Rational[1, 2], 1], 
      Hue[0.3032230945156945, 
       Rational[1, 2], 1], 
      Hue[0.3038031045842353, 
       Rational[1, 2], 1], 
      Hue[0.3049631247213169, 
       Rational[1, 2], 1], 
      Hue[0.30728316499548014`, 
       Rational[1, 2], 1], 
      Hue[0.30786317506402094`, 
       Rational[1, 2], 1], 
      Hue[0.3084431851325618, 
       Rational[1, 2], 1], 
      Hue[0.3096032052696434, 
       Rational[1, 2], 1], 
      Hue[0.31018321533818427`, 
       Rational[1, 2], 1], 
      Hue[0.310763225406725, 
       Rational[1, 2], 1], 
      Hue[0.3119232455438066, 
       Rational[1, 2], 1], 
      Hue[0.31250325561234743`, 
       Rational[1, 2], 1], 
      Hue[0.31308326568088823`, 
       Rational[1, 2], 1], 
      Hue[0.3142432858179699, 
       Rational[1, 2], 1], 
      Hue[0.3148232958865107, 
       Rational[1, 2], 1], 
      Hue[0.3154033059550515, 
       Rational[1, 2], 1], 
      Hue[0.3165633260921333, 
       Rational[1, 2], 1], 
      Hue[0.31714333616067425`, 
       Rational[1, 2], 1], 
      Hue[0.31772334622921505`, 
       Rational[1, 2], 1], 
      Hue[0.31830335629775586`, 
       Rational[1, 2], 1], 
      Hue[0.31888336636629666`, 
       Rational[1, 2], 1], 
      Hue[0.31946337643483746`, 
       Rational[1, 2], 1], 
      Hue[0.32004338650337827`, 
       Rational[1, 2], 1], 
      Hue[0.32062339657191913`, 
       Rational[1, 2], 1], 
      Hue[0.32120340664045993`, 
       Rational[1, 2], 1], 
      Hue[0.32178341670900074`, 
       Rational[1, 2], 1], 
      Hue[0.32236342677754154`, 
       Rational[1, 2], 1], 
      Hue[0.3229434368460823, 
       Rational[1, 2], 1], 
      Hue[0.32352344691462315`, 
       Rational[1, 2], 1], 
      Hue[0.32410345698316395`, 
       Rational[1, 2], 1], 
      Hue[0.32468346705170475`, 
       Rational[1, 2], 1], 
      Hue[0.32526347712024556`, 
       Rational[1, 2], 1], 
      Hue[0.3258434871887864, 
       Rational[1, 2], 1], 
      Hue[0.3264234972573272, 
       Rational[1, 2], 1], 
      Hue[0.327003507325868, 
       Rational[1, 2], 1], 
      Hue[0.32758351739440883`, 
       Rational[1, 2], 1], 
      Hue[0.32816352746294963`, 
       Rational[1, 2], 1], 
      Hue[0.32874353753149044`, 
       Rational[1, 2], 1], 
      Hue[0.32932354760003124`, 
       Rational[1, 2], 1], 
      Hue[0.32990355766857205`, 
       Rational[1, 2], 1], 
      Hue[0.33048356773711285`, 
       Rational[1, 2], 1], 
      Hue[0.3310635778056537, 
       Rational[1, 2], 1], 
      Hue[0.3316435878741945, 
       Rational[1, 2], 1], 
      Hue[0.3322235979427353, 
       Rational[1, 2], 1], 
      Hue[0.3328036080112761, 
       Rational[1, 2], 1], 
      Hue[0.3333836180798169, 
       Rational[1, 2], 1], 
      Hue[0.3339636281483577, 
       Rational[1, 2], 1], 
      Hue[0.3351236482854395, 
       Rational[1, 2], 1], 
      Hue[0.33575292840003906`, 
       Rational[1, 2], 1], 
      Hue[0.33638220851463846`, 
       Rational[1, 2], 1], 
      Hue[0.3376407687438376, 
       Rational[1, 2], 1], 
      Hue[0.33827004885843714`, 
       Rational[1, 2], 1], 
      Hue[0.3388993289730367, 
       Rational[1, 2], 1], 
      Hue[0.34015788920223583`, 
       Rational[1, 2], 1], 
      Hue[0.3407871693168354, 
       Rational[1, 2], 1], 
      Hue[0.34141644943143495`, 
       Rational[1, 2], 1], 
      Hue[0.34267500966063397`, 
       Rational[1, 2], 1], 
      Hue[0.34519213011903205`, 
       Rational[1, 2], 1], 
      Hue[0.3458214102336316, 
       Rational[1, 2], 1], 
      Hue[0.34645069034823117`, 
       Rational[1, 2], 1], 
      Hue[0.3477092505774303, 
       Rational[1, 2], 1], 
      Hue[0.3502263710358284, 
       Rational[1, 2], 1], 
      Hue[0.35526061195262476`, 
       Rational[1, 2], 1], 
      Hue[0.35588989206722427`, 
       Rational[1, 2], 1], 
      Hue[0.3565191721818239, 
       Rational[1, 2], 1], 
      Hue[0.35777773241102284`, 
       Rational[1, 2], 1], 
      Hue[0.360294852869421, 
       Rational[1, 2], 1], 
      Hue[0.3653290937862173, 
       Rational[1, 2], 1], 
      Hue[0.37539757561980985`, 
       Rational[1, 2], 1], 
      Hue[0.3759847740911934, 
       Rational[1, 2], 1], 
      Hue[0.376571972562577, 
       Rational[1, 2], 1], 
      Hue[0.3777463695053441, 
       Rational[1, 2], 1], 
      Hue[0.3800951633908784, 
       Rational[1, 2], 1], 
      Hue[0.38479275116194706`, 
       Rational[1, 2], 1], 
      Hue[0.3941879267040842, 
       Rational[1, 2], 1], 
      Hue[0.41297827778835855`, 
       Rational[1, 2], 1], 
      Hue[0.41361474630580086`, 
       Rational[1, 2], 1], 
      Hue[0.4142512148232432, 
       Rational[1, 2], 1], 
      Hue[0.41552415185812763`, 
       Rational[1, 2], 1], 
      Hue[0.41807002592789677`, 
       Rational[1, 2], 1], 
      Hue[0.42316177406743516`, 
       Rational[1, 2], 1], 
      Hue[0.43334527034651177`, 
       Rational[1, 2], 1], 
      Hue[0.45371226290466515`, 
       Rational[1, 2], 1], 
      Hue[0.4537312841026336, 
       Rational[1, 2], 1], 
      Hue[0.4537503053006021, 
       Rational[1, 2], 1], 
      Hue[0.45378834769653925`, 
       Rational[1, 2], 1], 
      Hue[0.4538644324884135, 
       Rational[1, 2], 1], 
      Hue[0.4540166020721617, 
       Rational[1, 2], 1], 
      Hue[0.45432094123965827`, 
       Rational[1, 2], 1], 
      Hue[0.4543399624376267, 
       Rational[1, 2], 1], 
      Hue[0.45435898363559524`, 
       Rational[1, 2], 1], 
      Hue[0.45439702603153237`, 
       Rational[1, 2], 1], 
      Hue[0.45447311082340663`, 
       Rational[1, 2], 1], 
      Hue[0.45462528040715483`, 
       Rational[1, 2], 1], 
      Hue[0.45464430160512337`, 
       Rational[1, 2], 1], 
      Hue[0.45466332280309196`, 
       Rational[1, 2], 1], 
      Hue[0.45470136519902893`, 
       Rational[1, 2], 1], 
      Hue[0.454777449990903, 
       Rational[1, 2], 1], 
      Hue[0.45479647118887156`, 
       Rational[1, 2], 1], 
      Hue[0.45481549238684016`, 
       Rational[1, 2], 1], 
      Hue[0.4548535347827773, 
       Rational[1, 2], 1], 
      Hue[0.45487255598074583`, 
       Rational[1, 2], 1], 
      Hue[0.45489157717871437`, 
       Rational[1, 2], 1], 
      Hue[0.45491059837668296`, 
       Rational[1, 2], 1], 
      Hue[0.4549296195746514, 
       Rational[1, 2], 1], 
      Hue[0.04507038042534869, 
       Rational[1, 2], 1], 
      Hue[0.045656129526931086`, 
       Rational[1, 2], 1], 
      Hue[0.045656129526931086`, 
       Rational[1, 2], 1], 
      Hue[0.04624187862851364, 
       Rational[1, 2], 1], 
      Hue[0.04624187862851364, 
       Rational[1, 2], 1], 
      Hue[0.047413376831678455`, 
       Rational[1, 2], 1], 
      Hue[0.047413376831678455`, 
       Rational[1, 2], 1], 
      Hue[0.04975637323800807, 
       Rational[1, 2], 1], 
      Hue[0.04975637323800807, 
       Rational[1, 2], 1], 
      Hue[0.054442366050667615`, 
       Rational[1, 2], 1], 
      Hue[0.054442366050667615`, 
       Rational[1, 2], 1], 
      Hue[0.06381435167598638, 
       Rational[1, 2], 1], 
      Hue[0.06381435167598638, 
       Rational[1, 2], 1], 
      Hue[0.08255832292662395, 
       Rational[1, 2], 1], 
      Hue[0.08255832292662395, 
       Rational[1, 2], 1], 
      Hue[0.08319334207426511, 
       Rational[1, 2], 1], 
      Hue[0.08319334207426511, 
       Rational[1, 2], 1], 
      Hue[0.08382836122190629, 
       Rational[1, 2], 1], 
      Hue[0.08382836122190629, 
       Rational[1, 2], 1], 
      Hue[0.08509839951718859, 
       Rational[1, 2], 1], 
      Hue[0.08509839951718859, 
       Rational[1, 2], 1], 
      Hue[0.08763847610775323, 
       Rational[1, 2], 1], 
      Hue[0.08763847610775323, 
       Rational[1, 2], 1], 
      Hue[0.09271862928888232, 
       Rational[1, 2], 1], 
      Hue[0.09271862928888232, 
       Rational[1, 2], 1], 
      Hue[0.10287893565114072`, 
       Rational[1, 2], 1], 
      Hue[0.10287893565114072`, 
       Rational[1, 2], 1], 
      Hue[0.10351395479878188`, 
       Rational[1, 2], 1], 
      Hue[0.10351395479878188`, 
       Rational[1, 2], 1], 
      Hue[0.10414897394642304`, 
       Rational[1, 2], 1], 
      Hue[0.10414897394642304`, 
       Rational[1, 2], 1], 
      Hue[0.10541901224170536`, 
       Rational[1, 2], 1], 
      Hue[0.10541901224170536`, 
       Rational[1, 2], 1], 
      Hue[0.10795908883226986`, 
       Rational[1, 2], 1], 
      Hue[0.10795908883226986`, 
       Rational[1, 2], 1], 
      Hue[0.11303924201339897`, 
       Rational[1, 2], 1], 
      Hue[0.11303924201339897`, 
       Rational[1, 2], 1], 
      Hue[0.12319954837565734`, 
       Rational[1, 2], 1], 
      Hue[0.12319954837565734`, 
       Rational[1, 2], 1], 
      Hue[0.12379248588008265`, 
       Rational[1, 2], 1], 
      Hue[0.12379248588008265`, 
       Rational[1, 2], 1], 
      Hue[0.12438542338450781`, 
       Rational[1, 2], 1], 
      Hue[0.12438542338450781`, 
       Rational[1, 2], 1], 
      Hue[0.12557129839335812`, 
       Rational[1, 2], 1], 
      Hue[0.12557129839335812`, 
       Rational[1, 2], 1], 
      Hue[0.12794304841105877`, 
       Rational[1, 2], 1], 
      Hue[0.12794304841105877`, 
       Rational[1, 2], 1], 
      Hue[0.13268654844646033`, 
       Rational[1, 2], 1], 
      Hue[0.13268654844646033`, 
       Rational[1, 2], 1], 
      Hue[0.1332794859508855, 
       Rational[1, 2], 1], 
      Hue[0.1332794859508855, 
       Rational[1, 2], 1], 
      Hue[0.13387242345531064`, 
       Rational[1, 2], 1], 
      Hue[0.13387242345531064`, 
       Rational[1, 2], 1], 
      Hue[0.13505829846416098`, 
       Rational[1, 2], 1], 
      Hue[0.13505829846416098`, 
       Rational[1, 2], 1], 
      Hue[0.13743004848186177`, 
       Rational[1, 2], 1], 
      Hue[0.13743004848186177`, 
       Rational[1, 2], 1], 
      Hue[0.14217354851726322`, 
       Rational[1, 2], 1], 
      Hue[0.14217354851726322`, 
       Rational[1, 2], 1], 
      Hue[0.14276648602168834`, 
       Rational[1, 2], 1], 
      Hue[0.14276648602168834`, 
       Rational[1, 2], 1], 
      Hue[0.1433594235261135, 
       Rational[1, 2], 1], 
      Hue[0.1433594235261135, 
       Rational[1, 2], 1], 
      Hue[0.14454529853496384`, 
       Rational[1, 2], 1], 
      Hue[0.14454529853496384`, 
       Rational[1, 2], 1], 
      Hue[0.1469170485526646, 
       Rational[1, 2], 1], 
      Hue[0.1469170485526646, 
       Rational[1, 2], 1], 
      Hue[0.14750998605708993`, 
       Rational[1, 2], 1], 
      Hue[0.14750998605708993`, 
       Rational[1, 2], 1], 
      Hue[0.14810292356151508`, 
       Rational[1, 2], 1], 
      Hue[0.14810292356151508`, 
       Rational[1, 2], 1], 
      Hue[0.1492887985703654, 
       Rational[1, 2], 1], 
      Hue[0.1492887985703654, 
       Rational[1, 2], 1], 
      Hue[0.15166054858806602`, 
       Rational[1, 2], 1], 
      Hue[0.15166054858806602`, 
       Rational[1, 2], 1], 
      Hue[0.1522534860924912, 
       Rational[1, 2], 1], 
      Hue[0.1522534860924912, 
       Rational[1, 2], 1], 
      Hue[0.1528464235969164, 
       Rational[1, 2], 1], 
      Hue[0.1528464235969164, 
       Rational[1, 2], 1], 
      Hue[0.15403229860576667`, 
       Rational[1, 2], 1], 
      Hue[0.15403229860576667`, 
       Rational[1, 2], 1], 
      Hue[0.15462523611019183`, 
       Rational[1, 2], 1], 
      Hue[0.15462523611019183`, 
       Rational[1, 2], 1], 
      Hue[0.155218173614617, 
       Rational[1, 2], 1], 
      Hue[0.155218173614617, 
       Rational[1, 2], 1], 
      Hue[0.15640404862346746`, 
       Rational[1, 2], 1], 
      Hue[0.15640404862346746`, 
       Rational[1, 2], 1], 
      Hue[0.15699698612789276`, 
       Rational[1, 2], 1], 
      Hue[0.15699698612789276`, 
       Rational[1, 2], 1], 
      Hue[0.15758992363231794`, 
       Rational[1, 2], 1], 
      Hue[0.15758992363231794`, 
       Rational[1, 2], 1], 
      Hue[0.15818286113674307`, 
       Rational[1, 2], 1], 
      Hue[0.15818286113674307`, 
       Rational[1, 2], 1], 
      Hue[0.15877579864116825`, 
       Rational[1, 2], 1], 
      Hue[0.15877579864116825`, 
       Rational[1, 2], 1], 
      Hue[0.1593687361455934, 
       Rational[1, 2], 1], 
      Hue[0.1593687361455934, 
       Rational[1, 2], 1], 
      Hue[0.15996167365001857`, 
       Rational[1, 2], 1], 
      Hue[0.15996167365001857`, 
       Rational[1, 2], 1], 
      Hue[0.16114754865886888`, 
       Rational[1, 2], 1], 
      Hue[0.16114754865886888`, 
       Rational[1, 2], 1], 
      Hue[0.16172885508316964`, 
       Rational[1, 2], 1], 
      Hue[0.16172885508316964`, 
       Rational[1, 2], 1], 
      Hue[0.16231016150747038`, 
       Rational[1, 2], 1], 
      Hue[0.16231016150747038`, 
       Rational[1, 2], 1], 
      Hue[0.16289146793177126`, 
       Rational[1, 2], 1], 
      Hue[0.16289146793177126`, 
       Rational[1, 2], 1], 
      Hue[0.16347277435607202`, 
       Rational[1, 2], 1], 
      Hue[0.16347277435607202`, 
       Rational[1, 2], 1], 
      Hue[0.16405408078037276`, 
       Rational[1, 2], 1], 
      Hue[0.16405408078037276`, 
       Rational[1, 2], 1], 
      Hue[0.16463538720467363`, 
       Rational[1, 2], 1], 
      Hue[0.16463538720467363`, 
       Rational[1, 2], 1], 
      Hue[0.1652166936289744, 
       Rational[1, 2], 1], 
      Hue[0.1652166936289744, 
       Rational[1, 2], 1], 
      Hue[0.16579800005327514`, 
       Rational[1, 2], 1], 
      Hue[0.16579800005327514`, 
       Rational[1, 2], 1], 
      Hue[0.16637930647757587`, 
       Rational[1, 2], 1], 
      Hue[0.16637930647757587`, 
       Rational[1, 2], 1], 
      Hue[0.16696061290187664`, 
       Rational[1, 2], 1], 
      Hue[0.16696061290187664`, 
       Rational[1, 2], 1], 
      Hue[0.1675419193261775, 
       Rational[1, 2], 1], 
      Hue[0.1675419193261775, 
       Rational[1, 2], 1], 
      Hue[0.16812322575047825`, 
       Rational[1, 2], 1], 
      Hue[0.16812322575047825`, 
       Rational[1, 2], 1], 
      Hue[0.16870453217477901`, 
       Rational[1, 2], 1], 
      Hue[0.16870453217477901`, 
       Rational[1, 2], 1], 
      Hue[0.1692858385990799, 
       Rational[1, 2], 1], 
      Hue[0.1692858385990799, 
       Rational[1, 2], 1], 
      Hue[0.16986714502338063`, 
       Rational[1, 2], 1], 
      Hue[0.16986714502338063`, 
       Rational[1, 2], 1], 
      Hue[0.1704484514476814, 
       Rational[1, 2], 1], 
      Hue[0.1704484514476814, 
       Rational[1, 2], 1], 
      Hue[0.17102975787198213`, 
       Rational[1, 2], 1], 
      Hue[0.17102975787198213`, 
       Rational[1, 2], 1], 
      Hue[0.17161106429628287`, 
       Rational[1, 2], 1], 
      Hue[0.17161106429628287`, 
       Rational[1, 2], 1], 
      Hue[0.17219237072058377`, 
       Rational[1, 2], 1], 
      Hue[0.17219237072058377`, 
       Rational[1, 2], 1], 
      Hue[0.1727736771448845, 
       Rational[1, 2], 1], 
      Hue[0.1727736771448845, 
       Rational[1, 2], 1], 
      Hue[0.17335498356918527`, 
       Rational[1, 2], 1], 
      Hue[0.17335498356918527`, 
       Rational[1, 2], 1], 
      Hue[0.17393628999348615`, 
       Rational[1, 2], 1], 
      Hue[0.17393628999348615`, 
       Rational[1, 2], 1], 
      Hue[0.17451759641778689`, 
       Rational[1, 2], 1], 
      Hue[0.17451759641778689`, 
       Rational[1, 2], 1], 
      Hue[0.17509890284208765`, 
       Rational[1, 2], 1], 
      Hue[0.17509890284208765`, 
       Rational[1, 2], 1], 
      Hue[0.1756802092663884, 
       Rational[1, 2], 1], 
      Hue[0.1756802092663884, 
       Rational[1, 2], 1], 
      Hue[0.17626151569068912`, 
       Rational[1, 2], 1], 
      Hue[0.17626151569068912`, 
       Rational[1, 2], 1], 
      Hue[0.1768428221149899, 
       Rational[1, 2], 1], 
      Hue[0.1768428221149899, 
       Rational[1, 2], 1], 
      Hue[0.17742412853929063`, 
       Rational[1, 2], 1], 
      Hue[0.17742412853929063`, 
       Rational[1, 2], 1], 
      Hue[0.17800543496359136`, 
       Rational[1, 2], 1], 
      Hue[0.17800543496359136`, 
       Rational[1, 2], 1], 
      Hue[0.1785867413878921, 
       Rational[1, 2], 1], 
      Hue[0.1785867413878921, 
       Rational[1, 2], 1], 
      Hue[0.17974935423649374`, 
       Rational[1, 2], 1], 
      Hue[0.17974935423649374`, 
       Rational[1, 2], 1], 
      Hue[0.1803306606607945, 
       Rational[1, 2], 1], 
      Hue[0.1803306606607945, 
       Rational[1, 2], 1], 
      Hue[0.18091196708509538`, 
       Rational[1, 2], 1], 
      Hue[0.18091196708509538`, 
       Rational[1, 2], 1], 
      Hue[0.18207457993369688`, 
       Rational[1, 2], 1], 
      Hue[0.18207457993369688`, 
       Rational[1, 2], 1], 
      Hue[0.18265588635799762`, 
       Rational[1, 2], 1], 
      Hue[0.18265588635799762`, 
       Rational[1, 2], 1], 
      Hue[0.18323719278229836`, 
       Rational[1, 2], 1], 
      Hue[0.18323719278229836`, 
       Rational[1, 2], 1], 
      Hue[0.18439980563089986`, 
       Rational[1, 2], 1], 
      Hue[0.18439980563089986`, 
       Rational[1, 2], 1], 
      Hue[0.1849811120552006, 
       Rational[1, 2], 1], 
      Hue[0.1849811120552006, 
       Rational[1, 2], 1], 
      Hue[0.18556241847950136`, 
       Rational[1, 2], 1], 
      Hue[0.18556241847950136`, 
       Rational[1, 2], 1], 
      Hue[0.18672503132810297`, 
       Rational[1, 2], 1], 
      Hue[0.18672503132810297`, 
       Rational[1, 2], 1], 
      Hue[0.18905025702530612`, 
       Rational[1, 2], 1], 
      Hue[0.18905025702530612`, 
       Rational[1, 2], 1], 
      Hue[0.18963156344960685`, 
       Rational[1, 2], 1], 
      Hue[0.18963156344960685`, 
       Rational[1, 2], 1], 
      Hue[0.1902128698739076, 
       Rational[1, 2], 1], 
      Hue[0.1902128698739076, 
       Rational[1, 2], 1], 
      Hue[0.19137548272250923`, 
       Rational[1, 2], 1], 
      Hue[0.19137548272250923`, 
       Rational[1, 2], 1], 
      Hue[0.19370070841971235`, 
       Rational[1, 2], 1], 
      Hue[0.19370070841971235`, 
       Rational[1, 2], 1], 
      Hue[0.1942820148440131, 
       Rational[1, 2], 1], 
      Hue[0.1942820148440131, 
       Rational[1, 2], 1], 
      Hue[0.19486332126831385`, 
       Rational[1, 2], 1], 
      Hue[0.19486332126831385`, 
       Rational[1, 2], 1], 
      Hue[0.1960259341169155, 
       Rational[1, 2], 1], 
      Hue[0.1960259341169155, 
       Rational[1, 2], 1], 
      Hue[0.1983511598141186, 
       Rational[1, 2], 1], 
      Hue[0.1983511598141186, 
       Rational[1, 2], 1], 
      Hue[0.1989817362844781, 
       Rational[1, 2], 1], 
      Hue[0.1989817362844781, 
       Rational[1, 2], 1], 
      Hue[0.1996123127548376, 
       Rational[1, 2], 1], 
      Hue[0.1996123127548376, 
       Rational[1, 2], 1], 
      Hue[0.2008734656955566, 
       Rational[1, 2], 1], 
      Hue[0.2008734656955566, 
       Rational[1, 2], 1], 
      Hue[0.2033957715769946, 
       Rational[1, 2], 1], 
      Hue[0.2033957715769946, 
       Rational[1, 2], 1], 
      Hue[0.2040263480473541, 
       Rational[1, 2], 1], 
      Hue[0.2040263480473541, 
       Rational[1, 2], 1], 
      Hue[0.20465692451771358`, 
       Rational[1, 2], 1], 
      Hue[0.20465692451771358`, 
       Rational[1, 2], 1], 
      Hue[0.20591807745843257`, 
       Rational[1, 2], 1], 
      Hue[0.20591807745843257`, 
       Rational[1, 2], 1], 
      Hue[0.20844038333987058`, 
       Rational[1, 2], 1], 
      Hue[0.20844038333987058`, 
       Rational[1, 2], 1], 
      Hue[0.20907095981023008`, 
       Rational[1, 2], 1], 
      Hue[0.20907095981023008`, 
       Rational[1, 2], 1], 
      Hue[0.20970153628058957`, 
       Rational[1, 2], 1], 
      Hue[0.20970153628058957`, 
       Rational[1, 2], 1], 
      Hue[0.21096268922130856`, 
       Rational[1, 2], 1], 
      Hue[0.21096268922130856`, 
       Rational[1, 2], 1], 
      Hue[0.21348499510274654`, 
       Rational[1, 2], 1], 
      Hue[0.21348499510274654`, 
       Rational[1, 2], 1], 
      Hue[0.21852960686562253`, 
       Rational[1, 2], 1], 
      Hue[0.21852960686562253`, 
       Rational[1, 2], 1], 
      Hue[0.21916018333598203`, 
       Rational[1, 2], 1], 
      Hue[0.21916018333598203`, 
       Rational[1, 2], 1], 
      Hue[0.21979075980634152`, 
       Rational[1, 2], 1], 
      Hue[0.21979075980634152`, 
       Rational[1, 2], 1], 
      Hue[0.2210519127470605, 
       Rational[1, 2], 1], 
      Hue[0.2210519127470605, 
       Rational[1, 2], 1], 
      Hue[0.22357421862849852`, 
       Rational[1, 2], 1], 
      Hue[0.22357421862849852`, 
       Rational[1, 2], 1], 
      Hue[0.22861883039137448`, 
       Rational[1, 2], 1], 
      Hue[0.22861883039137448`, 
       Rational[1, 2], 1], 
      Hue[0.22924940686173398`, 
       Rational[1, 2], 1], 
      Hue[0.22924940686173398`, 
       Rational[1, 2], 1], 
      Hue[0.22987998333209347`, 
       Rational[1, 2], 1], 
      Hue[0.22987998333209347`, 
       Rational[1, 2], 1], 
      Hue[0.2311411362728125, 
       Rational[1, 2], 1], 
      Hue[0.2311411362728125, 
       Rational[1, 2], 1], 
      Hue[0.23366344215425047`, 
       Rational[1, 2], 1], 
      Hue[0.23366344215425047`, 
       Rational[1, 2], 1], 
      Hue[0.23870805391712646`, 
       Rational[1, 2], 1], 
      Hue[0.23870805391712646`, 
       Rational[1, 2], 1], 
      Hue[0.23929654874426995`, 
       Rational[1, 2], 1], 
      Hue[0.23929654874426995`, 
       Rational[1, 2], 1], 
      Hue[0.23988504357141346`, 
       Rational[1, 2], 1], 
      Hue[0.23988504357141346`, 
       Rational[1, 2], 1], 
      Hue[0.2410620332257006, 
       Rational[1, 2], 1], 
      Hue[0.2410620332257006, 
       Rational[1, 2], 1], 
      Hue[0.24341601253427475`, 
       Rational[1, 2], 1], 
      Hue[0.24341601253427475`, 
       Rational[1, 2], 1], 
      Hue[0.24400450736141824`, 
       Rational[1, 2], 1], 
      Hue[0.24400450736141824`, 
       Rational[1, 2], 1], 
      Hue[0.24459300218856173`, 
       Rational[1, 2], 1], 
      Hue[0.24459300218856173`, 
       Rational[1, 2], 1], 
      Hue[0.24576999184284887`, 
       Rational[1, 2], 1], 
      Hue[0.24576999184284887`, 
       Rational[1, 2], 1], 
      Hue[0.24635848666999252`, 
       Rational[1, 2], 1], 
      Hue[0.24635848666999252`, 
       Rational[1, 2], 1], 
      Hue[0.246946981497136, 
       Rational[1, 2], 1], 
      Hue[0.246946981497136, 
       Rational[1, 2], 1], 
      Hue[0.248123971151423, 
       Rational[1, 2], 1], 
      Hue[0.248123971151423, 
       Rational[1, 2], 1], 
      Hue[0.24871246597856653`, 
       Rational[1, 2], 1], 
      Hue[0.24871246597856653`, 
       Rational[1, 2], 1], 
      Hue[0.24930096080571001`, 
       Rational[1, 2], 1], 
      Hue[0.24930096080571001`, 
       Rational[1, 2], 1], 
      Hue[0.24988945563285353`, 
       Rational[1, 2], 1], 
      Hue[0.24988945563285353`, 
       Rational[1, 2], 1], 
      Hue[0.7504779504599972, 
       Rational[1, 2], 1], 
      Hue[0.7504779504599972, 
       Rational[1, 2], 1], 
      Hue[0.7510664452871407, 
       Rational[1, 2], 1], 
      Hue[0.7510664452871407, 
       Rational[1, 2], 1], 
      Hue[0.7516549401142842, 
       Rational[1, 2], 1], 
      Hue[0.7516549401142842, 
       Rational[1, 2], 1], 
      Hue[0.7528319297685713, 
       Rational[1, 2], 1], 
      Hue[0.7528319297685713, 
       Rational[1, 2], 1], 
      Hue[0.7534204245957148, 
       Rational[1, 2], 1], 
      Hue[0.7534204245957148, 
       Rational[1, 2], 1], 
      Hue[0.7540089194228583, 
       Rational[1, 2], 1], 
      Hue[0.7540089194228583, 
       Rational[1, 2], 1], 
      Hue[0.7551859090771453, 
       Rational[1, 2], 1], 
      Hue[0.7551859090771453, 
       Rational[1, 2], 1], 
      Hue[0.7575398883857194, 
       Rational[1, 2], 1], 
      Hue[0.7575398883857194, 
       Rational[1, 2], 1], 
      Hue[0.7581283832128631, 
       Rational[1, 2], 1], 
      Hue[0.7581283832128631, 
       Rational[1, 2], 1], 
      Hue[0.7587168780400065, 
       Rational[1, 2], 1], 
      Hue[0.7587168780400065, 
       Rational[1, 2], 1], 
      Hue[0.7598938676942936, 
       Rational[1, 2], 1], 
      Hue[0.7598938676942936, 
       Rational[1, 2], 1], 
      Hue[0.7622478470028675, 
       Rational[1, 2], 1], 
      Hue[0.7622478470028675, 
       Rational[1, 2], 1], 
      Hue[0.7669558056200159, 
       Rational[1, 2], 1], 
      Hue[0.7669558056200159, 
       Rational[1, 2], 1], 
      Hue[0.7763717228543123, 
       Rational[1, 2], 1], 
      Hue[0.7763717228543123, 
       Rational[1, 2], 1], 
      Hue[0.7770094877275147, 
       Rational[1, 2], 1], 
      Hue[0.7770094877275147, 
       Rational[1, 2], 1], 
      Hue[0.777647252600717, 
       Rational[1, 2], 1], 
      Hue[0.777647252600717, 
       Rational[1, 2], 1], 
      Hue[0.7789227823471214, 
       Rational[1, 2], 1], 
      Hue[0.7789227823471214, 
       Rational[1, 2], 1], 
      Hue[0.7814738418399304, 
       Rational[1, 2], 1], 
      Hue[0.7814738418399304, 
       Rational[1, 2], 1], 
      Hue[0.7865759608255485, 
       Rational[1, 2], 1], 
      Hue[0.7865759608255485, 
       Rational[1, 2], 1], 
      Hue[0.7967801987967844, 
       Rational[1, 2], 1], 
      Hue[0.7967801987967844, 
       Rational[1, 2], 1], 
      Hue[0.8171886747392565, 
       Rational[1, 2], 1], 
      Hue[0.8171886747392565, 
       Rational[1, 2], 1], 
      Hue[0.8178148085323343, 
       Rational[1, 2], 1], 
      Hue[0.8178148085323343, 
       Rational[1, 2], 1], 
      Hue[0.8184409423254122, 
       Rational[1, 2], 1], 
      Hue[0.8184409423254122, 
       Rational[1, 2], 1], 
      Hue[0.8196932099115679, 
       Rational[1, 2], 1], 
      Hue[0.8196932099115679, 
       Rational[1, 2], 1], 
      Hue[0.8221977450838792, 
       Rational[1, 2], 1], 
      Hue[0.8221977450838792, 
       Rational[1, 2], 1], 
      Hue[0.827206815428502, 
       Rational[1, 2], 1], 
      Hue[0.827206815428502, 
       Rational[1, 2], 1], 
      Hue[0.8372249561177476, 
       Rational[1, 2], 1], 
      Hue[0.8372249561177476, 
       Rational[1, 2], 1], 
      Hue[0.8378510899108255, 
       Rational[1, 2], 1], 
      Hue[0.8378510899108255, 
       Rational[1, 2], 1], 
      Hue[0.8384772237039033, 
       Rational[1, 2], 1], 
      Hue[0.8384772237039033, 
       Rational[1, 2], 1], 
      Hue[0.8397294912900589, 
       Rational[1, 2], 1], 
      Hue[0.8397294912900589, 
       Rational[1, 2], 1], 
      Hue[0.8422340264623704, 
       Rational[1, 2], 1], 
      Hue[0.8422340264623704, 
       Rational[1, 2], 1], 
      Hue[0.8472430968069933, 
       Rational[1, 2], 1], 
      Hue[0.8472430968069933, 
       Rational[1, 2], 1], 
      Hue[0.8572612374962388, 
       Rational[1, 2], 1], 
      Hue[0.8572612374962388, 
       Rational[1, 2], 1], 
      Hue[0.8578452896461009, 
       Rational[1, 2], 1], 
      Hue[0.8578452896461009, 
       Rational[1, 2], 1], 
      Hue[0.8584293417959626, 
       Rational[1, 2], 1], 
      Hue[0.8584293417959626, 
       Rational[1, 2], 1], 
      Hue[0.8595974460956863, 
       Rational[1, 2], 1], 
      Hue[0.8595974460956863, 
       Rational[1, 2], 1], 
      Hue[0.8619336546951339, 
       Rational[1, 2], 1], 
      Hue[0.8619336546951339, 
       Rational[1, 2], 1], 
      Hue[0.8666060718940289, 
       Rational[1, 2], 1], 
      Hue[0.8666060718940289, 
       Rational[1, 2], 1], 
      Hue[0.8671901240438908, 
       Rational[1, 2], 1], 
      Hue[0.8671901240438908, 
       Rational[1, 2], 1], 
      Hue[0.8677741761937527, 
       Rational[1, 2], 1], 
      Hue[0.8677741761937527, 
       Rational[1, 2], 1], 
      Hue[0.8689422804934764, 
       Rational[1, 2], 1], 
      Hue[0.8689422804934764, 
       Rational[1, 2], 1], 
      Hue[0.8712784890929239, 
       Rational[1, 2], 1], 
      Hue[0.8712784890929239, 
       Rational[1, 2], 1], 
      Hue[0.875950906291819, 
       Rational[1, 2], 1], 
      Hue[0.875950906291819, 
       Rational[1, 2], 1], 
      Hue[0.876534958441681, 
       Rational[1, 2], 1], 
      Hue[0.876534958441681, 
       Rational[1, 2], 1], 
      Hue[0.8771190105915428, 
       Rational[1, 2], 1], 
      Hue[0.8771190105915428, 
       Rational[1, 2], 1], 
      Hue[0.8782871148912667, 
       Rational[1, 2], 1], 
      Hue[0.8782871148912667, 
       Rational[1, 2], 1], 
      Hue[0.8806233234907141, 
       Rational[1, 2], 1], 
      Hue[0.8806233234907141, 
       Rational[1, 2], 1], 
      Hue[0.8812073756405759, 
       Rational[1, 2], 1], 
      Hue[0.8812073756405759, 
       Rational[1, 2], 1], 
      Hue[0.8817914277904378, 
       Rational[1, 2], 1], 
      Hue[0.8817914277904378, 
       Rational[1, 2], 1], 
      Hue[0.8829595320901616, 
       Rational[1, 2], 1], 
      Hue[0.8829595320901616, 
       Rational[1, 2], 1], 
      Hue[0.8852957406896091, 
       Rational[1, 2], 1], 
      Hue[0.8852957406896091, 
       Rational[1, 2], 1], 
      Hue[0.8858797928394709, 
       Rational[1, 2], 1], 
      Hue[0.8858797928394709, 
       Rational[1, 2], 1], 
      Hue[0.8864638449893328, 
       Rational[1, 2], 1], 
      Hue[0.8864638449893328, 
       Rational[1, 2], 1], 
      Hue[0.8876319492890566, 
       Rational[1, 2], 1], 
      Hue[0.8876319492890566, 
       Rational[1, 2], 1], 
      Hue[0.8882160014389185, 
       Rational[1, 2], 1], 
      Hue[0.8882160014389185, 
       Rational[1, 2], 1], 
      Hue[0.8888000535887804, 
       Rational[1, 2], 1], 
      Hue[0.8888000535887804, 
       Rational[1, 2], 1], 
      Hue[0.8899681578885041, 
       Rational[1, 2], 1], 
      Hue[0.8899681578885041, 
       Rational[1, 2], 1], 
      Hue[0.890552210038366, 
       Rational[1, 2], 1], 
      Hue[0.890552210038366, 
       Rational[1, 2], 1], 
      Hue[0.891136262188228, 
       Rational[1, 2], 1], 
      Hue[0.891136262188228, 
       Rational[1, 2], 1], 
      Hue[0.8923043664879518, 
       Rational[1, 2], 1], 
      Hue[0.8923043664879518, 
       Rational[1, 2], 1], 
      Hue[0.8928884186378137, 
       Rational[1, 2], 1], 
      Hue[0.8928884186378137, 
       Rational[1, 2], 1], 
      Hue[0.8934724707876756, 
       Rational[1, 2], 1], 
      Hue[0.8934724707876756, 
       Rational[1, 2], 1], 
      Hue[0.8946405750873992, 
       Rational[1, 2], 1], 
      Hue[0.8946405750873992, 
       Rational[1, 2], 1], 
      Hue[0.8952738972833199, 
       Rational[1, 2], 1], 
      Hue[0.8952738972833199, 
       Rational[1, 2], 1], 
      Hue[0.8959072194792405, 
       Rational[1, 2], 1], 
      Hue[0.8959072194792405, 
       Rational[1, 2], 1], 
      Hue[0.8965405416751611, 
       Rational[1, 2], 1], 
      Hue[0.8965405416751611, 
       Rational[1, 2], 1], 
      Hue[0.8971738638710816, 
       Rational[1, 2], 1], 
      Hue[0.8971738638710816, 
       Rational[1, 2], 1], 
      Hue[0.8978071860670023, 
       Rational[1, 2], 1], 
      Hue[0.8978071860670023, 
       Rational[1, 2], 1], 
      Hue[0.8984405082629228, 
       Rational[1, 2], 1], 
      Hue[0.8984405082629228, 
       Rational[1, 2], 1], 
      Hue[0.8990738304588434, 
       Rational[1, 2], 1], 
      Hue[0.8990738304588434, 
       Rational[1, 2], 1], 
      Hue[0.8997071526547641, 
       Rational[1, 2], 1], 
      Hue[0.8997071526547641, 
       Rational[1, 2], 1], 
      Hue[0.9003404748506847, 
       Rational[1, 2], 1], 
      Hue[0.9003404748506847, 
       Rational[1, 2], 1], 
      Hue[0.9009737970466053, 
       Rational[1, 2], 1], 
      Hue[0.9009737970466053, 
       Rational[1, 2], 1], 
      Hue[0.9016071192425259, 
       Rational[1, 2], 1], 
      Hue[0.9016071192425259, 
       Rational[1, 2], 1], 
      Hue[0.9022404414384465, 
       Rational[1, 2], 1], 
      Hue[0.9022404414384465, 
       Rational[1, 2], 1], 
      Hue[0.9028737636343671, 
       Rational[1, 2], 1], 
      Hue[0.9028737636343671, 
       Rational[1, 2], 1], 
      Hue[0.9035070858302877, 
       Rational[1, 2], 1], 
      Hue[0.9035070858302877, 
       Rational[1, 2], 1], 
      Hue[0.9041404080262083, 
       Rational[1, 2], 1], 
      Hue[0.9041404080262083, 
       Rational[1, 2], 1], 
      Hue[0.9047737302221288, 
       Rational[1, 2], 1], 
      Hue[0.9047737302221288, 
       Rational[1, 2], 1], 
      Hue[0.9054070524180494, 
       Rational[1, 2], 1], 
      Hue[0.9054070524180494, 
       Rational[1, 2], 1], 
      Hue[0.90604037461397, 
       Rational[1, 2], 1], 
      Hue[0.90604037461397, 
       Rational[1, 2], 1], 
      Hue[0.9066736968098906, 
       Rational[1, 2], 1], 
      Hue[0.9066736968098906, 
       Rational[1, 2], 1], 
      Hue[0.9073070190058112, 
       Rational[1, 2], 1], 
      Hue[0.9073070190058112, 
       Rational[1, 2], 1], 
      Hue[0.9079403412017318, 
       Rational[1, 2], 1], 
      Hue[0.9079403412017318, 
       Rational[1, 2], 1], 
      Hue[0.9085736633976524, 
       Rational[1, 2], 1], 
      Hue[0.9085736633976524, 
       Rational[1, 2], 1], 
      Hue[0.909206985593573, 
       Rational[1, 2], 1], 
      Hue[0.909206985593573, 
       Rational[1, 2], 1], 
      Hue[0.9098403077894937, 
       Rational[1, 2], 1], 
      Hue[0.9098403077894937, 
       Rational[1, 2], 1], 
      Hue[0.9104736299854144, 
       Rational[1, 2], 1], 
      Hue[0.9104736299854144, 
       Rational[1, 2], 1], 
      Hue[0.9111069521813349, 
       Rational[1, 2], 1], 
      Hue[0.9111069521813349, 
       Rational[1, 2], 1], 
      Hue[0.9117402743772555, 
       Rational[1, 2], 1], 
      Hue[0.9117402743772555, 
       Rational[1, 2], 1], 
      Hue[0.9123735965731761, 
       Rational[1, 2], 1], 
      Hue[0.9123735965731761, 
       Rational[1, 2], 1], 
      Hue[0.9130069187690967, 
       Rational[1, 2], 1], 
      Hue[0.9130069187690967, 
       Rational[1, 2], 1], 
      Hue[0.9136402409650173, 
       Rational[1, 2], 1], 
      Hue[0.9136402409650173, 
       Rational[1, 2], 1], 
      Hue[0.9149068853568585, 
       Rational[1, 2], 1], 
      Hue[0.9149068853568585, 
       Rational[1, 2], 1], 
      Hue[0.9155402075527791, 
       Rational[1, 2], 1], 
      Hue[0.9155402075527791, 
       Rational[1, 2], 1], 
      Hue[0.9161735297486997, 
       Rational[1, 2], 1], 
      Hue[0.9161735297486997, 
       Rational[1, 2], 1], 
      Hue[0.9174401741405409, 
       Rational[1, 2], 1], 
      Hue[0.9174401741405409, 
       Rational[1, 2], 1], 
      Hue[0.9180734963364614, 
       Rational[1, 2], 1], 
      Hue[0.9180734963364614, 
       Rational[1, 2], 1], 
      Hue[0.918706818532382, 
       Rational[1, 2], 1], 
      Hue[0.918706818532382, 
       Rational[1, 2], 1], 
      Hue[0.9199734629242233, 
       Rational[1, 2], 1], 
      Hue[0.9199734629242233, 
       Rational[1, 2], 1], 
      Hue[0.9206067851201439, 
       Rational[1, 2], 1], 
      Hue[0.9206067851201439, 
       Rational[1, 2], 1], 
      Hue[0.9212401073160645, 
       Rational[1, 2], 1], 
      Hue[0.9212401073160645, 
       Rational[1, 2], 1], 
      Hue[0.9225067517079057, 
       Rational[1, 2], 1], 
      Hue[0.9225067517079057, 
       Rational[1, 2], 1], 
      Hue[0.9250400404915881, 
       Rational[1, 2], 1], 
      Hue[0.9250400404915881, 
       Rational[1, 2], 1], 
      Hue[0.9256733626875087, 
       Rational[1, 2], 1], 
      Hue[0.9256733626875087, 
       Rational[1, 2], 1], 
      Hue[0.9263066848834293, 
       Rational[1, 2], 1], 
      Hue[0.9263066848834293, 
       Rational[1, 2], 1], 
      Hue[0.9275733292752706, 
       Rational[1, 2], 1], 
      Hue[0.9275733292752706, 
       Rational[1, 2], 1], 
      Hue[0.9301066180589529, 
       Rational[1, 2], 1], 
      Hue[0.9301066180589529, 
       Rational[1, 2], 1], 
      Hue[0.9307399402548735, 
       Rational[1, 2], 1], 
      Hue[0.9307399402548735, 
       Rational[1, 2], 1], 
      Hue[0.9313732624507941, 
       Rational[1, 2], 1], 
      Hue[0.9313732624507941, 
       Rational[1, 2], 1], 
      Hue[0.9326399068426353, 
       Rational[1, 2], 1], 
      Hue[0.9326399068426353, 
       Rational[1, 2], 1], 
      Hue[0.9351731956263177, 
       Rational[1, 2], 1], 
      Hue[0.9351731956263177, 
       Rational[1, 2], 1], 
      Hue[0.9357644361790224, 
       Rational[1, 2], 1], 
      Hue[0.9357644361790224, 
       Rational[1, 2], 1], 
      Hue[0.936355676731727, 
       Rational[1, 2], 1], 
      Hue[0.936355676731727, 
       Rational[1, 2], 1], 
      Hue[0.9375381578371363, 
       Rational[1, 2], 1], 
      Hue[0.9375381578371363, 
       Rational[1, 2], 1], 
      Hue[0.9399031200479548, 
       Rational[1, 2], 1], 
      Hue[0.9399031200479548, 
       Rational[1, 2], 1], 
      Hue[0.9404943606006595, 
       Rational[1, 2], 1], 
      Hue[0.9404943606006595, 
       Rational[1, 2], 1], 
      Hue[0.9410856011533641, 
       Rational[1, 2], 1], 
      Hue[0.9410856011533641, 
       Rational[1, 2], 1], 
      Hue[0.9422680822587733, 
       Rational[1, 2], 1], 
      Hue[0.9422680822587733, 
       Rational[1, 2], 1], 
      Hue[0.9446330444695918, 
       Rational[1, 2], 1], 
      Hue[0.9446330444695918, 
       Rational[1, 2], 1], 
      Hue[0.9452242850222965, 
       Rational[1, 2], 1], 
      Hue[0.9452242850222965, 
       Rational[1, 2], 1], 
      Hue[0.9458155255750011, 
       Rational[1, 2], 1], 
      Hue[0.9458155255750011, 
       Rational[1, 2], 1], 
      Hue[0.9469980066804105, 
       Rational[1, 2], 1], 
      Hue[0.9469980066804105, 
       Rational[1, 2], 1], 
      Hue[0.949362968891229, 
       Rational[1, 2], 1], 
      Hue[0.949362968891229, 
       Rational[1, 2], 1], 
      Hue[0.9540928933128661, 
       Rational[1, 2], 1], 
      Hue[0.9540928933128661, 
       Rational[1, 2], 1], 
      Hue[0.9546841338655707, 
       Rational[1, 2], 1], 
      Hue[0.9546841338655707, 
       Rational[1, 2], 1], 
      Hue[0.9552753744182754, 
       Rational[1, 2], 1], 
      Hue[0.9552753744182754, 
       Rational[1, 2], 1], 
      Hue[0.9564578555236847, 
       Rational[1, 2], 1], 
      Hue[0.9564578555236847, 
       Rational[1, 2], 1], 
      Hue[0.9588228177345032, 
       Rational[1, 2], 1], 
      Hue[0.9588228177345032, 
       Rational[1, 2], 1], 
      Hue[0.9635527421561403, 
       Rational[1, 2], 1], 
      Hue[0.9635527421561403, 
       Rational[1, 2], 1], 
      Hue[0.9730125909994145, 
       Rational[1, 2], 1], 
      Hue[0.9730125909994145, 
       Rational[1, 2], 1], 
      Hue[0.9735922004719948, 
       Rational[1, 2], 1], 
      Hue[0.9735922004719948, 
       Rational[1, 2], 1], 
      Hue[0.974171809944575, 
       Rational[1, 2], 1], 
      Hue[0.974171809944575, 
       Rational[1, 2], 1], 
      Hue[0.9753310288897354, 
       Rational[1, 2], 1], 
      Hue[0.9753310288897354, 
       Rational[1, 2], 1], 
      Hue[0.9776494667800564, 
       Rational[1, 2], 1], 
      Hue[0.9776494667800564, 
       Rational[1, 2], 1], 
      Hue[0.9822863425606982, 
       Rational[1, 2], 1], 
      Hue[0.9822863425606982, 
       Rational[1, 2], 1], 
      Hue[0.9828659520332785, 
       Rational[1, 2], 1], 
      Hue[0.9828659520332785, 
       Rational[1, 2], 1], 
      Hue[0.9834455615058587, 
       Rational[1, 2], 1], 
      Hue[0.9834455615058587, 
       Rational[1, 2], 1], 
      Hue[0.9846047804510192, 
       Rational[1, 2], 1], 
      Hue[0.9846047804510192, 
       Rational[1, 2], 1], 
      Hue[0.9869232183413401, 
       Rational[1, 2], 1], 
      Hue[0.9869232183413401, 
       Rational[1, 2], 1], 
      Hue[0.991560094121982, 
       Rational[1, 2], 1], 
      Hue[0.991560094121982, 
       Rational[1, 2], 1], 
      Hue[0.9921397035945623, 
       Rational[1, 2], 1], 
      Hue[0.9921397035945623, 
       Rational[1, 2], 1], 
      Hue[0.9927193130671425, 
       Rational[1, 2], 1], 
      Hue[0.9927193130671425, 
       Rational[1, 2], 1], 
      Hue[0.9938785320123029, 
       Rational[1, 2], 1], 
      Hue[0.9938785320123029, 
       Rational[1, 2], 1], 
      Hue[0.9944581414848832, 
       Rational[1, 2], 1], 
      Hue[0.9944581414848832, 
       Rational[1, 2], 1], 
      Hue[0.9950377509574634, 
       Rational[1, 2], 1], 
      Hue[0.9950377509574634, 
       Rational[1, 2], 1], 
      Hue[0.9961969699026239, 
       Rational[1, 2], 1], 
      Hue[0.9961969699026239, 
       Rational[1, 2], 1], 
      Hue[0.996776579375204, 
       Rational[1, 2], 1], 
      Hue[0.996776579375204, 
       Rational[1, 2], 1], 
      Hue[0.9973561888477843, 
       Rational[1, 2], 1], 
      Hue[0.9973561888477843, 
       Rational[1, 2], 1], 
      Hue[0.9979357983203645, 
       Rational[1, 2], 1], 
      Hue[0.9979357983203645, 
       Rational[1, 2], 1], 
      Hue[0.9985154077929448, 
       Rational[1, 2], 1], 
      Hue[0.9985154077929448, 
       Rational[1, 2], 1], 
      Hue[0.999095017265525, 
       Rational[1, 2], 1], 
      Hue[0.999095017265525, 
       Rational[1, 2], 1], 
      Hue[0.9996746267381051, 
       Rational[1, 2], 1], 
      Hue[0.9996746267381051, 
       Rational[1, 2], 1], 
      Hue[0.5002542362106854, 
       Rational[1, 2], 1], 
      Hue[0.5002542362106854, 
       Rational[1, 2], 1], 
      Hue[0.5008338456832657, 
       Rational[1, 2], 1], 
      Hue[0.5008338456832657, 
       Rational[1, 2], 1], 
      Hue[0.5014134551558459, 
       Rational[1, 2], 1], 
      Hue[0.5014134551558459, 
       Rational[1, 2], 1], 
      Hue[0.5019930646284262, 
       Rational[1, 2], 1], 
      Hue[0.5019930646284262, 
       Rational[1, 2], 1], 
      Hue[0.5031522835735867, 
       Rational[1, 2], 1], 
      Hue[0.5031522835735867, 
       Rational[1, 2], 1], 
      Hue[0.5054707214639076, 
       Rational[1, 2], 1], 
      Hue[0.5054707214639076, 
       Rational[1, 2], 1], 
      Hue[0.5101075972445495, 
       Rational[1, 2], 1], 
      Hue[0.5101075972445495, 
       Rational[1, 2], 1], 
      Hue[0.5107364767631885, 
       Rational[1, 2], 1], 
      Hue[0.5107364767631885, 
       Rational[1, 2], 1], 
      Hue[0.5113653562818274, 
       Rational[1, 2], 1], 
      Hue[0.5113653562818274, 
       Rational[1, 2], 1], 
      Hue[0.5126231153191052, 
       Rational[1, 2], 1], 
      Hue[0.5126231153191052, 
       Rational[1, 2], 1], 
      Hue[0.5151386333936612, 
       Rational[1, 2], 1], 
      Hue[0.5151386333936612, 
       Rational[1, 2], 1], 
      Hue[0.5201696695427727, 
       Rational[1, 2], 1], 
      Hue[0.5201696695427727, 
       Rational[1, 2], 1], 
      Hue[0.530231741840996, 
       Rational[1, 2], 1], 
      Hue[0.530231741840996, 
       Rational[1, 2], 1], 
      Hue[0.5503558864374425, 
       Rational[1, 2], 1], 
      Hue[0.5503558864374425, 
       Rational[1, 2], 1], 
      Hue[0.5509426843128655, 
       Rational[1, 2], 1], 
      Hue[0.5509426843128655, 
       Rational[1, 2], 1], 
      Hue[0.5515294821882883, 
       Rational[1, 2], 1], 
      Hue[0.5515294821882883, 
       Rational[1, 2], 1], 
      Hue[0.5527030779391344, 
       Rational[1, 2], 1], 
      Hue[0.5527030779391344, 
       Rational[1, 2], 1], 
      Hue[0.5550502694408264, 
       Rational[1, 2], 1], 
      Hue[0.5550502694408264, 
       Rational[1, 2], 1], 
      Hue[0.5597446524442103, 
       Rational[1, 2], 1], 
      Hue[0.5597446524442103, 
       Rational[1, 2], 1], 
      Hue[0.5691334184509782, 
       Rational[1, 2], 1], 
      Hue[0.5691334184509782, 
       Rational[1, 2], 1], 
      Hue[0.5697202163264011, 
       Rational[1, 2], 1], 
      Hue[0.5697202163264011, 
       Rational[1, 2], 1], 
      Hue[0.5703070142018241, 
       Rational[1, 2], 1], 
      Hue[0.5703070142018241, 
       Rational[1, 2], 1], 
      Hue[0.5714806099526701, 
       Rational[1, 2], 1], 
      Hue[0.5714806099526701, 
       Rational[1, 2], 1], 
      Hue[0.5738278014543621, 
       Rational[1, 2], 1], 
      Hue[0.5738278014543621, 
       Rational[1, 2], 1], 
      Hue[0.578522184457746, 
       Rational[1, 2], 1], 
      Hue[0.578522184457746, 
       Rational[1, 2], 1], 
      Hue[0.5879109504645137, 
       Rational[1, 2], 1], 
      Hue[0.5879109504645137, 
       Rational[1, 2], 1], 
      Hue[0.5885470183859953, 
       Rational[1, 2], 1], 
      Hue[0.5885470183859953, 
       Rational[1, 2], 1], 
      Hue[0.5891830863074771, 
       Rational[1, 2], 1], 
      Hue[0.5891830863074771, 
       Rational[1, 2], 1], 
      Hue[0.5904552221504404, 
       Rational[1, 2], 1], 
      Hue[0.5904552221504404, 
       Rational[1, 2], 1], 
      Hue[0.5929994938363673, 
       Rational[1, 2], 1], 
      Hue[0.5929994938363673, 
       Rational[1, 2], 1], 
      Hue[0.598088037208221, 
       Rational[1, 2], 1], 
      Hue[0.598088037208221, 
       Rational[1, 2], 1], 
      Hue[0.5987241051297028, 
       Rational[1, 2], 1], 
      Hue[0.5987241051297028, 
       Rational[1, 2], 1], 
      Hue[0.5993601730511844, 
       Rational[1, 2], 1], 
      Hue[0.5993601730511844, 
       Rational[1, 2], 1], 
      Hue[0.6006323088941479, 
       Rational[1, 2], 1], 
      Hue[0.6006323088941479, 
       Rational[1, 2], 1], 
      Hue[0.6031765805800747, 
       Rational[1, 2], 1], 
      Hue[0.6031765805800747, 
       Rational[1, 2], 1], 
      Hue[0.6082651239519283, 
       Rational[1, 2], 1], 
      Hue[0.6082651239519283, 
       Rational[1, 2], 1], 
      Hue[0.60890119187341, 
       Rational[1, 2], 1], 
      Hue[0.60890119187341, 
       Rational[1, 2], 1], 
      Hue[0.6095372597948917, 
       Rational[1, 2], 1], 
      Hue[0.6095372597948917, 
       Rational[1, 2], 1], 
      Hue[0.6108093956378551, 
       Rational[1, 2], 1], 
      Hue[0.6108093956378551, 
       Rational[1, 2], 1], 
      Hue[0.6133536673237819, 
       Rational[1, 2], 1], 
      Hue[0.6133536673237819, 
       Rational[1, 2], 1], 
      Hue[0.6139897352452637, 
       Rational[1, 2], 1], 
      Hue[0.6139897352452637, 
       Rational[1, 2], 1], 
      Hue[0.6146258031667454, 
       Rational[1, 2], 1], 
      Hue[0.6146258031667454, 
       Rational[1, 2], 1], 
      Hue[0.6158979390097088, 
       Rational[1, 2], 1], 
      Hue[0.6158979390097088, 
       Rational[1, 2], 1], 
      Hue[0.6184422106956357, 
       Rational[1, 2], 1], 
      Hue[0.6184422106956357, 
       Rational[1, 2], 1], 
      Hue[0.6190782786171174, 
       Rational[1, 2], 1], 
      Hue[0.6190782786171174, 
       Rational[1, 2], 1], 
      Hue[0.619714346538599, 
       Rational[1, 2], 1], 
      Hue[0.619714346538599, 
       Rational[1, 2], 1], 
      Hue[0.6209864823815625, 
       Rational[1, 2], 1], 
      Hue[0.6209864823815625, 
       Rational[1, 2], 1], 
      Hue[0.6216225503030441, 
       Rational[1, 2], 1], 
      Hue[0.6216225503030441, 
       Rational[1, 2], 1], 
      Hue[0.6222586182245259, 
       Rational[1, 2], 1], 
      Hue[0.6222586182245259, 
       Rational[1, 2], 1], 
      Hue[0.6235307540674894, 
       Rational[1, 2], 1], 
      Hue[0.6235307540674894, 
       Rational[1, 2], 1], 
      Hue[0.6241668219889711, 
       Rational[1, 2], 1], 
      Hue[0.6241668219889711, 
       Rational[1, 2], 1], 
      Hue[0.6248028899104527, 
       Rational[1, 2], 1], 
      Hue[0.6248028899104527, 
       Rational[1, 2], 1], 
      Hue[0.6260750257534162, 
       Rational[1, 2], 1], 
      Hue[0.6260750257534162, 
       Rational[1, 2], 1], 
      Hue[0.6267110936748979, 
       Rational[1, 2], 1], 
      Hue[0.6267110936748979, 
       Rational[1, 2], 1], 
      Hue[0.6273471615963796, 
       Rational[1, 2], 1], 
      Hue[0.6273471615963796, 
       Rational[1, 2], 1], 
      Hue[0.628619297439343, 
       Rational[1, 2], 1], 
      Hue[0.628619297439343, 
       Rational[1, 2], 1], 
      Hue[0.6292437342807002, 
       Rational[1, 2], 1], 
      Hue[0.6292437342807002, 
       Rational[1, 2], 1], 
      Hue[0.6298681711220575, 
       Rational[1, 2], 1], 
      Hue[0.6298681711220575, 
       Rational[1, 2], 1], 
      Hue[0.6311170448047723, 
       Rational[1, 2], 1], 
      Hue[0.6311170448047723, 
       Rational[1, 2], 1], 
      Hue[0.6317414816461295, 
       Rational[1, 2], 1], 
      Hue[0.6317414816461295, 
       Rational[1, 2], 1], 
      Hue[0.6323659184874868, 
       Rational[1, 2], 1], 
      Hue[0.6323659184874868, 
       Rational[1, 2], 1], 
      Hue[0.6336147921702014, 
       Rational[1, 2], 1], 
      Hue[0.6336147921702014, 
       Rational[1, 2], 1], 
      Hue[0.6342392290115587, 
       Rational[1, 2], 1], 
      Hue[0.6342392290115587, 
       Rational[1, 2], 1], 
      Hue[0.6348636658529161, 
       Rational[1, 2], 1], 
      Hue[0.6348636658529161, 
       Rational[1, 2], 1], 
      Hue[0.6354881026942734, 
       Rational[1, 2], 1], 
      Hue[0.6354881026942734, 
       Rational[1, 2], 1], 
      Hue[0.6361125395356306, 
       Rational[1, 2], 1], 
      Hue[0.6361125395356306, 
       Rational[1, 2], 1], 
      Hue[0.6367369763769879, 
       Rational[1, 2], 1], 
      Hue[0.6367369763769879, 
       Rational[1, 2], 1], 
      Hue[0.6373614132183453, 
       Rational[1, 2], 1], 
      Hue[0.6373614132183453, 
       Rational[1, 2], 1], 
      Hue[0.6379858500597027, 
       Rational[1, 2], 1], 
      Hue[0.6379858500597027, 
       Rational[1, 2], 1], 
      Hue[0.6386102869010599, 
       Rational[1, 2], 1], 
      Hue[0.6386102869010599, 
       Rational[1, 2], 1], 
      Hue[0.6392347237424172, 
       Rational[1, 2], 1], 
      Hue[0.6392347237424172, 
       Rational[1, 2], 1], 
      Hue[0.6398591605837745, 
       Rational[1, 2], 1], 
      Hue[0.6398591605837745, 
       Rational[1, 2], 1], 
      Hue[0.6404835974251318, 
       Rational[1, 2], 1], 
      Hue[0.6404835974251318, 
       Rational[1, 2], 1], 
      Hue[0.6411080342664891, 
       Rational[1, 2], 1], 
      Hue[0.6411080342664891, 
       Rational[1, 2], 1], 
      Hue[0.6417324711078464, 
       Rational[1, 2], 1], 
      Hue[0.6417324711078464, 
       Rational[1, 2], 1], 
      Hue[0.6423569079492037, 
       Rational[1, 2], 1], 
      Hue[0.6423569079492037, 
       Rational[1, 2], 1], 
      Hue[0.642981344790561, 
       Rational[1, 2], 1], 
      Hue[0.642981344790561, 
       Rational[1, 2], 1], 
      Hue[0.6436057816319183, 
       Rational[1, 2], 1], 
      Hue[0.6436057816319183, 
       Rational[1, 2], 1], 
      Hue[0.6442302184732756, 
       Rational[1, 2], 1], 
      Hue[0.6442302184732756, 
       Rational[1, 2], 1], 
      Hue[0.6448546553146329, 
       Rational[1, 2], 1], 
      Hue[0.6448546553146329, 
       Rational[1, 2], 1], 
      Hue[0.6454790921559903, 
       Rational[1, 2], 1], 
      Hue[0.6454790921559903, 
       Rational[1, 2], 1], 
      Hue[0.6461035289973476, 
       Rational[1, 2], 1], 
      Hue[0.6461035289973476, 
       Rational[1, 2], 1], 
      Hue[0.6467279658387048, 
       Rational[1, 2], 1], 
      Hue[0.6467279658387048, 
       Rational[1, 2], 1], 
      Hue[0.6473524026800621, 
       Rational[1, 2], 1], 
      Hue[0.6473524026800621, 
       Rational[1, 2], 1], 
      Hue[0.6479768395214194, 
       Rational[1, 2], 1], 
      Hue[0.6479768395214194, 
       Rational[1, 2], 1], 
      Hue[0.6486012763627766, 
       Rational[1, 2], 1], 
      Hue[0.6486012763627766, 
       Rational[1, 2], 1], 
      Hue[0.649225713204134, 
       Rational[1, 2], 1], 
      Hue[0.649225713204134, 
       Rational[1, 2], 1], 
      Hue[0.6498501500454913, 
       Rational[1, 2], 1], 
      Hue[0.6498501500454913, 
       Rational[1, 2], 1], 
      Hue[0.6504745868868487, 
       Rational[1, 2], 1], 
      Hue[0.6504745868868487, 
       Rational[1, 2], 1], 
      Hue[0.6510990237282059, 
       Rational[1, 2], 1], 
      Hue[0.6510990237282059, 
       Rational[1, 2], 1], 
      Hue[0.6517234605695632, 
       Rational[1, 2], 1], 
      Hue[0.6517234605695632, 
       Rational[1, 2], 1], 
      Hue[0.6523478974109205, 
       Rational[1, 2], 1], 
      Hue[0.6523478974109205, 
       Rational[1, 2], 1], 
      Hue[0.653596771093635, 
       Rational[1, 2], 1], 
      Hue[0.653596771093635, 
       Rational[1, 2], 1], 
      Hue[0.6542212079349923, 
       Rational[1, 2], 1], 
      Hue[0.6542212079349923, 
       Rational[1, 2], 1], 
      Hue[0.6548456447763498, 
       Rational[1, 2], 1], 
      Hue[0.6548456447763498, 
       Rational[1, 2], 1], 
      Hue[0.6560945184590643, 
       Rational[1, 2], 1], 
      Hue[0.6560945184590643, 
       Rational[1, 2], 1], 
      Hue[0.6567189553004216, 
       Rational[1, 2], 1], 
      Hue[0.6567189553004216, 
       Rational[1, 2], 1], 
      Hue[0.657343392141779, 
       Rational[1, 2], 1], 
      Hue[0.657343392141779, 
       Rational[1, 2], 1], 
      Hue[0.6585922658244936, 
       Rational[1, 2], 1], 
      Hue[0.6585922658244936, 
       Rational[1, 2], 1], 
      Hue[0.6592167026658509, 
       Rational[1, 2], 1], 
      Hue[0.6592167026658509, 
       Rational[1, 2], 1], 
      Hue[0.6598411395072081, 
       Rational[1, 2], 1], 
      Hue[0.6598411395072081, 
       Rational[1, 2], 1], 
      Hue[0.6610900131899228, 
       Rational[1, 2], 1], 
      Hue[0.6610900131899228, 
       Rational[1, 2], 1], 
      Hue[0.663587760555352, 
       Rational[1, 2], 1], 
      Hue[0.663587760555352, 
       Rational[1, 2], 1], 
      Hue[0.6642121973967092, 
       Rational[1, 2], 1], 
      Hue[0.6642121973967092, 
       Rational[1, 2], 1], 
      Hue[0.6648366342380666, 
       Rational[1, 2], 1], 
      Hue[0.6648366342380666, 
       Rational[1, 2], 1], 
      Hue[0.6660855079207813, 
       Rational[1, 2], 1], 
      Hue[0.6660855079207813, 
       Rational[1, 2], 1], 
      Hue[0.6685832552862104, 
       Rational[1, 2], 1], 
      Hue[0.6685832552862104, 
       Rational[1, 2], 1], 
      Hue[0.6691656104843517, 
       Rational[1, 2], 1], 
      Hue[0.6691656104843517, 
       Rational[1, 2], 1], 
      Hue[0.6697479656824932, 
       Rational[1, 2], 1], 
      Hue[0.6697479656824932, 
       Rational[1, 2], 1], 
      Hue[0.6709126760787758, 
       Rational[1, 2], 1], 
      Hue[0.6709126760787758, 
       Rational[1, 2], 1], 
      Hue[0.6732420968713412, 
       Rational[1, 2], 1], 
      Hue[0.6732420968713412, 
       Rational[1, 2], 1], 
      Hue[0.6738244520694825, 
       Rational[1, 2], 1], 
      Hue[0.6738244520694825, 
       Rational[1, 2], 1], 
      Hue[0.6744068072676239, 
       Rational[1, 2], 1], 
      Hue[0.6744068072676239, 
       Rational[1, 2], 1], 
      Hue[0.6755715176639066, 
       Rational[1, 2], 1], 
      Hue[0.6755715176639066, 
       Rational[1, 2], 1], 
      Hue[0.6779009384564718, 
       Rational[1, 2], 1], 
      Hue[0.6779009384564718, 
       Rational[1, 2], 1], 
      Hue[0.6784832936546132, 
       Rational[1, 2], 1], 
      Hue[0.6784832936546132, 
       Rational[1, 2], 1], 
      Hue[0.6790656488527546, 
       Rational[1, 2], 1], 
      Hue[0.6790656488527546, 
       Rational[1, 2], 1], 
      Hue[0.6802303592490373, 
       Rational[1, 2], 1], 
      Hue[0.6802303592490373, 
       Rational[1, 2], 1], 
      Hue[0.6825597800416026, 
       Rational[1, 2], 1], 
      Hue[0.6825597800416026, 
       Rational[1, 2], 1], 
      Hue[0.6872186216267333, 
       Rational[1, 2], 1], 
      Hue[0.6872186216267333, 
       Rational[1, 2], 1], 
      Hue[0.6878009768248747, 
       Rational[1, 2], 1], 
      Hue[0.6878009768248747, 
       Rational[1, 2], 1], 
      Hue[0.6883833320230159, 
       Rational[1, 2], 1], 
      Hue[0.6883833320230159, 
       Rational[1, 2], 1], 
      Hue[0.6895480424192987, 
       Rational[1, 2], 1], 
      Hue[0.6895480424192987, 
       Rational[1, 2], 1], 
      Hue[0.691877463211864, 
       Rational[1, 2], 1], 
      Hue[0.691877463211864, 
       Rational[1, 2], 1], 
      Hue[0.6965363047969948, 
       Rational[1, 2], 1], 
      Hue[0.6965363047969948, 
       Rational[1, 2], 1], 
      Hue[0.7058539879672562, 
       Rational[1, 2], 1], 
      Hue[0.7058539879672562, 
       Rational[1, 2], 1], 
      Hue[0.7064856132114563, 
       Rational[1, 2], 1], 
      Hue[0.7064856132114563, 
       Rational[1, 2], 1], 
      Hue[0.7071172384556563, 
       Rational[1, 2], 1], 
      Hue[0.7071172384556563, 
       Rational[1, 2], 1], 
      Hue[0.7083804889440564, 
       Rational[1, 2], 1], 
      Hue[0.7083804889440564, 
       Rational[1, 2], 1], 
      Hue[0.7109069899208567, 
       Rational[1, 2], 1], 
      Hue[0.7109069899208567, 
       Rational[1, 2], 1], 
      Hue[0.7159599918744571, 
       Rational[1, 2], 1], 
      Hue[0.7159599918744571, 
       Rational[1, 2], 1], 
      Hue[0.7260659957816581, 
       Rational[1, 2], 1], 
      Hue[0.7260659957816581, 
       Rational[1, 2], 1], 
      Hue[0.7266976210258582, 
       Rational[1, 2], 1], 
      Hue[0.7266976210258582, 
       Rational[1, 2], 1], 
      Hue[0.7273292462700582, 
       Rational[1, 2], 1], 
      Hue[0.7273292462700582, 
       Rational[1, 2], 1], 
      Hue[0.7285924967584583, 
       Rational[1, 2], 1], 
      Hue[0.7285924967584583, 
       Rational[1, 2], 1], 
      Hue[0.7311189977352586, 
       Rational[1, 2], 1], 
      Hue[0.7311189977352586, 
       Rational[1, 2], 1], 
      Hue[0.736171999688859, 
       Rational[1, 2], 1], 
      Hue[0.736171999688859, 
       Rational[1, 2], 1], 
      Hue[0.7368036249330591, 
       Rational[1, 2], 1], 
      Hue[0.7368036249330591, 
       Rational[1, 2], 1], 
      Hue[0.7374352501772592, 
       Rational[1, 2], 1], 
      Hue[0.7374352501772592, 
       Rational[1, 2], 1], 
      Hue[0.7386985006656592, 
       Rational[1, 2], 1], 
      Hue[0.7386985006656592, 
       Rational[1, 2], 1], 
      Hue[0.7412250016424595, 
       Rational[1, 2], 1], 
      Hue[0.7412250016424595, 
       Rational[1, 2], 1], 
      Hue[0.7418566268866595, 
       Rational[1, 2], 1], 
      Hue[0.7418566268866595, 
       Rational[1, 2], 1], 
      Hue[0.7424882521308596, 
       Rational[1, 2], 1], 
      Hue[0.7424882521308596, 
       Rational[1, 2], 1], 
      Hue[0.7437515026192598, 
       Rational[1, 2], 1], 
      Hue[0.7437515026192598, 
       Rational[1, 2], 1], 
      Hue[0.74627800359606, 
       Rational[1, 2], 1], 
      Hue[0.74627800359606, 
       Rational[1, 2], 1], 
      Hue[0.7468675471970441, 
       Rational[1, 2], 1], 
      Hue[0.7468675471970441, 
       Rational[1, 2], 1], 
      Hue[0.7474570907980282, 
       Rational[1, 2], 1], 
      Hue[0.7474570907980282, 
       Rational[1, 2], 1], 
      Hue[0.7480466343990122, 
       Rational[1, 2], 1], 
      Hue[0.7480466343990122, 
       Rational[1, 2], 1], 
      Hue[0.7486361779999964, 
       Rational[1, 2], 1], 
      Hue[0.7486361779999964, 
       Rational[1, 2], 1], 
      Hue[0.7492257216009806, 
       Rational[1, 2], 1], 
      Hue[0.7492257216009806, 
       Rational[1, 2], 1], 
      Hue[0.7498152652019645, 
       Rational[1, 2], 1], 
      Hue[0.7498152652019645, 
       Rational[1, 2], 1], 
      Hue[0.25040480880294874`, 
       Rational[1, 2], 1], 
      Hue[0.25040480880294874`, 
       Rational[1, 2], 1], 
      Hue[0.2509943524039328, 
       Rational[1, 2], 1], 
      Hue[0.2509943524039328, 
       Rational[1, 2], 1], 
      Hue[0.2515838960049169, 
       Rational[1, 2], 1], 
      Hue[0.2515838960049169, 
       Rational[1, 2], 1], 
      Hue[0.25217343960590094`, 
       Rational[1, 2], 1], 
      Hue[0.25217343960590094`, 
       Rational[1, 2], 1], 
      Hue[0.2533525268078691, 
       Rational[1, 2], 1], 
      Hue[0.2533525268078691, 
       Rational[1, 2], 1], 
      Hue[0.25571070121180556`, 
       Rational[1, 2], 1], 
      Hue[0.25571070121180556`, 
       Rational[1, 2], 1], 
      Hue[0.2563002448127896, 
       Rational[1, 2], 1], 
      Hue[0.2563002448127896, 
       Rational[1, 2], 1], 
      Hue[0.25688978841377375`, 
       Rational[1, 2], 1], 
      Hue[0.25688978841377375`, 
       Rational[1, 2], 1], 
      Hue[0.25806887561574193`, 
       Rational[1, 2], 1], 
      Hue[0.25806887561574193`, 
       Rational[1, 2], 1], 
      Hue[0.2604270500196783, 
       Rational[1, 2], 1], 
      Hue[0.2604270500196783, 
       Rational[1, 2], 1], 
      Hue[0.26514339882755106`, 
       Rational[1, 2], 1], 
      Hue[0.26514339882755106`, 
       Rational[1, 2], 1], 
      Hue[0.2657329424285351, 
       Rational[1, 2], 1], 
      Hue[0.2657329424285351, 
       Rational[1, 2], 1], 
      Hue[0.26632248602951925`, 
       Rational[1, 2], 1], 
      Hue[0.26632248602951925`, 
       Rational[1, 2], 1], 
      Hue[0.26750157323148743`, 
       Rational[1, 2], 1], 
      Hue[0.26750157323148743`, 
       Rational[1, 2], 1], 
      Hue[0.2698597476354238, 
       Rational[1, 2], 1], 
      Hue[0.2698597476354238, 
       Rational[1, 2], 1], 
      Hue[0.27457609644329656`, 
       Rational[1, 2], 1], 
      Hue[0.27457609644329656`, 
       Rational[1, 2], 1], 
      Hue[0.28400879405904206`, 
       Rational[1, 2], 1], 
      Hue[0.28400879405904206`, 
       Rational[1, 2], 1], 
      Hue[0.2846476077060849, 
       Rational[1, 2], 1], 
      Hue[0.2846476077060849, 
       Rational[1, 2], 1], 
      Hue[0.2852864213531277, 
       Rational[1, 2], 1], 
      Hue[0.2852864213531277, 
       Rational[1, 2], 1], 
      Hue[0.2865640486472133, 
       Rational[1, 2], 1], 
      Hue[0.2865640486472133, 
       Rational[1, 2], 1], 
      Hue[0.28911930323538454`, 
       Rational[1, 2], 1], 
      Hue[0.28911930323538454`, 
       Rational[1, 2], 1], 
      Hue[0.2942298124117271, 
       Rational[1, 2], 1], 
      Hue[0.2942298124117271, 
       Rational[1, 2], 1], 
      Hue[0.3044508307644121, 
       Rational[1, 2], 1], 
      Hue[0.3044508307644121, 
       Rational[1, 2], 1], 
      Hue[0.3050896444114549, 
       Rational[1, 2], 1], 
      Hue[0.3050896444114549, 
       Rational[1, 2], 1], 
      Hue[0.30572845805849774`, 
       Rational[1, 2], 1], 
      Hue[0.30572845805849774`, 
       Rational[1, 2], 1], 
      Hue[0.3070060853525834, 
       Rational[1, 2], 1], 
      Hue[0.3070060853525834, 
       Rational[1, 2], 1], 
      Hue[0.3095613399407546, 
       Rational[1, 2], 1], 
      Hue[0.3095613399407546, 
       Rational[1, 2], 1], 
      Hue[0.3146718491170971, 
       Rational[1, 2], 1], 
      Hue[0.3146718491170971, 
       Rational[1, 2], 1], 
      Hue[0.32489286746978213`, 
       Rational[1, 2], 1], 
      Hue[0.32489286746978213`, 
       Rational[1, 2], 1], 
      Hue[0.3255200500367006, 
       Rational[1, 2], 1], 
      Hue[0.3255200500367006, 
       Rational[1, 2], 1], 
      Hue[0.32614723260361894`, 
       Rational[1, 2], 1], 
      Hue[0.32614723260361894`, 
       Rational[1, 2], 1], 
      Hue[0.3274015977374558, 
       Rational[1, 2], 1], 
      Hue[0.3274015977374558, 
       Rational[1, 2], 1], 
      Hue[0.3299103280051294, 
       Rational[1, 2], 1], 
      Hue[0.3299103280051294, 
       Rational[1, 2], 1], 
      Hue[0.3349277885404768, 
       Rational[1, 2], 1], 
      Hue[0.3349277885404768, 
       Rational[1, 2], 1], 
      Hue[0.33555497110739513`, 
       Rational[1, 2], 1], 
      Hue[0.33555497110739513`, 
       Rational[1, 2], 1], 
      Hue[0.3361821536743136, 
       Rational[1, 2], 1], 
      Hue[0.3361821536743136, 
       Rational[1, 2], 1], 
      Hue[0.33743651880815034`, 
       Rational[1, 2], 1], 
      Hue[0.33743651880815034`, 
       Rational[1, 2], 1], 
      Hue[0.339945249075824, 
       Rational[1, 2], 1], 
      Hue[0.339945249075824, 
       Rational[1, 2], 1], 
      Hue[0.3449627096111713, 
       Rational[1, 2], 1], 
      Hue[0.3449627096111713, 
       Rational[1, 2], 1], 
      Hue[0.3455898921780897, 
       Rational[1, 2], 1], 
      Hue[0.3455898921780897, 
       Rational[1, 2], 1], 
      Hue[0.3462170747450081, 
       Rational[1, 2], 1], 
      Hue[0.3462170747450081, 
       Rational[1, 2], 1], 
      Hue[0.347471439878845, 
       Rational[1, 2], 1], 
      Hue[0.347471439878845, 
       Rational[1, 2], 1], 
      Hue[0.34998017014651855`, 
       Rational[1, 2], 1], 
      Hue[0.34998017014651855`, 
       Rational[1, 2], 1], 
      Hue[0.350607352713437, 
       Rational[1, 2], 1], 
      Hue[0.350607352713437, 
       Rational[1, 2], 1], 
      Hue[0.35123453528035536`, 
       Rational[1, 2], 1], 
      Hue[0.35123453528035536`, 
       Rational[1, 2], 1], 
      Hue[0.35248890041419223`, 
       Rational[1, 2], 1], 
      Hue[0.35248890041419223`, 
       Rational[1, 2], 1], 
      Hue[0.35499763068186585`, 
       Rational[1, 2], 1], 
      Hue[0.35499763068186585`, 
       Rational[1, 2], 1], 
      Hue[0.3556248132487843, 
       Rational[1, 2], 1], 
      Hue[0.3556248132487843, 
       Rational[1, 2], 1], 
      Hue[0.35625199581570266`, 
       Rational[1, 2], 1], 
      Hue[0.35625199581570266`, 
       Rational[1, 2], 1], 
      Hue[0.3575063609495395, 
       Rational[1, 2], 1], 
      Hue[0.3575063609495395, 
       Rational[1, 2], 1], 
      Hue[0.358133543516458, 
       Rational[1, 2], 1], 
      Hue[0.358133543516458, 
       Rational[1, 2], 1], 
      Hue[0.35876072608337634`, 
       Rational[1, 2], 1], 
      Hue[0.35876072608337634`, 
       Rational[1, 2], 1], 
      Hue[0.36001509121721315`, 
       Rational[1, 2], 1], 
      Hue[0.36001509121721315`, 
       Rational[1, 2], 1], 
      Hue[0.36064227378413155`, 
       Rational[1, 2], 1], 
      Hue[0.36064227378413155`, 
       Rational[1, 2], 1], 
      Hue[0.36126945635105, 
       Rational[1, 2], 1], 
      Hue[0.36126945635105, 
       Rational[1, 2], 1], 
      Hue[0.3625238214848868, 
       Rational[1, 2], 1], 
      Hue[0.3625238214848868, 
       Rational[1, 2], 1], 
      Hue[0.3631510040518052, 
       Rational[1, 2], 1], 
      Hue[0.3631510040518052, 
       Rational[1, 2], 1], 
      Hue[0.3637781866187237, 
       Rational[1, 2], 1], 
      Hue[0.3637781866187237, 
       Rational[1, 2], 1], 
      Hue[0.3650325517525605, 
       Rational[1, 2], 1], 
      Hue[0.3650325517525605, 
       Rational[1, 2], 1], 
      Hue[0.3656176526762629, 
       Rational[1, 2], 1], 
      Hue[0.3656176526762629, 
       Rational[1, 2], 1], 
      Hue[0.36620275359996535`, 
       Rational[1, 2], 1], 
      Hue[0.36620275359996535`, 
       Rational[1, 2], 1], 
      Hue[0.3673729554473702, 
       Rational[1, 2], 1], 
      Hue[0.3673729554473702, 
       Rational[1, 2], 1], 
      Hue[0.36795805637107265`, 
       Rational[1, 2], 1], 
      Hue[0.36795805637107265`, 
       Rational[1, 2], 1], 
      Hue[0.3685431572947751, 
       Rational[1, 2], 1], 
      Hue[0.3685431572947751, 
       Rational[1, 2], 1], 
      Hue[0.36971335914218006`, 
       Rational[1, 2], 1], 
      Hue[0.36971335914218006`, 
       Rational[1, 2], 1], 
      Hue[0.37029846006588246`, 
       Rational[1, 2], 1], 
      Hue[0.37029846006588246`, 
       Rational[1, 2], 1], 
      Hue[0.3708835609895849, 
       Rational[1, 2], 1], 
      Hue[0.3708835609895849, 
       Rational[1, 2], 1], 
      Hue[0.37146866191328737`, 
       Rational[1, 2], 1], 
      Hue[0.37146866191328737`, 
       Rational[1, 2], 1], 
      Hue[0.3720537628369898, 
       Rational[1, 2], 1], 
      Hue[0.3720537628369898, 
       Rational[1, 2], 1], 
      Hue[0.3726388637606922, 
       Rational[1, 2], 1], 
      Hue[0.3726388637606922, 
       Rational[1, 2], 1], 
      Hue[0.37322396468439467`, 
       Rational[1, 2], 1], 
      Hue[0.37322396468439467`, 
       Rational[1, 2], 1], 
      Hue[0.3738090656080972, 
       Rational[1, 2], 1], 
      Hue[0.3738090656080972, 
       Rational[1, 2], 1], 
      Hue[0.3743941665317996, 
       Rational[1, 2], 1], 
      Hue[0.3743941665317996, 
       Rational[1, 2], 1], 
      Hue[0.3749792674555021, 
       Rational[1, 2], 1], 
      Hue[0.3749792674555021, 
       Rational[1, 2], 1], 
      Hue[0.3755643683792045, 
       Rational[1, 2], 1], 
      Hue[0.3755643683792045, 
       Rational[1, 2], 1], 
      Hue[0.37614946930290694`, 
       Rational[1, 2], 1], 
      Hue[0.37614946930290694`, 
       Rational[1, 2], 1], 
      Hue[0.3767345702266094, 
       Rational[1, 2], 1], 
      Hue[0.3767345702266094, 
       Rational[1, 2], 1], 
      Hue[0.3773196711503118, 
       Rational[1, 2], 1], 
      Hue[0.3773196711503118, 
       Rational[1, 2], 1], 
      Hue[0.37790477207401424`, 
       Rational[1, 2], 1], 
      Hue[0.37790477207401424`, 
       Rational[1, 2], 1], 
      Hue[0.3784898729977167, 
       Rational[1, 2], 1], 
      Hue[0.3784898729977167, 
       Rational[1, 2], 1], 
      Hue[0.3790749739214192, 
       Rational[1, 2], 1], 
      Hue[0.3790749739214192, 
       Rational[1, 2], 1], 
      Hue[0.37966007484512154`, 
       Rational[1, 2], 1], 
      Hue[0.37966007484512154`, 
       Rational[1, 2], 1], 
      Hue[0.38024517576882405`, 
       Rational[1, 2], 1], 
      Hue[0.38024517576882405`, 
       Rational[1, 2], 1], 
      Hue[0.3808302766925265, 
       Rational[1, 2], 1], 
      Hue[0.3808302766925265, 
       Rational[1, 2], 1], 
      Hue[0.38141537761622896`, 
       Rational[1, 2], 1], 
      Hue[0.38141537761622896`, 
       Rational[1, 2], 1], 
      Hue[0.3820004785399314, 
       Rational[1, 2], 1], 
      Hue[0.3820004785399314, 
       Rational[1, 2], 1], 
      Hue[0.3825855794636338, 
       Rational[1, 2], 1], 
      Hue[0.3825855794636338, 
       Rational[1, 2], 1], 
      Hue[0.38317068038733626`, 
       Rational[1, 2], 1], 
      Hue[0.38317068038733626`, 
       Rational[1, 2], 1], 
      Hue[0.3837557813110387, 
       Rational[1, 2], 1], 
      Hue[0.3837557813110387, 
       Rational[1, 2], 1], 
      Hue[0.38434088223474117`, 
       Rational[1, 2], 1], 
      Hue[0.38434088223474117`, 
       Rational[1, 2], 1], 
      Hue[0.3849259831584437, 
       Rational[1, 2], 1], 
      Hue[0.3849259831584437, 
       Rational[1, 2], 1], 
      Hue[0.38551108408214607`, 
       Rational[1, 2], 1], 
      Hue[0.38551108408214607`, 
       Rational[1, 2], 1], 
      Hue[0.3860961850058485, 
       Rational[1, 2], 1], 
      Hue[0.3860961850058485, 
       Rational[1, 2], 1], 
      Hue[0.386681285929551, 
       Rational[1, 2], 1], 
      Hue[0.386681285929551, 
       Rational[1, 2], 1], 
      Hue[0.38726638685325343`, 
       Rational[1, 2], 1], 
      Hue[0.38726638685325343`, 
       Rational[1, 2], 1], 
      Hue[0.3884365887006583, 
       Rational[1, 2], 1], 
      Hue[0.3884365887006583, 
       Rational[1, 2], 1], 
      Hue[0.38902168962436073`, 
       Rational[1, 2], 1], 
      Hue[0.38902168962436073`, 
       Rational[1, 2], 1], 
      Hue[0.38960679054806313`, 
       Rational[1, 2], 1], 
      Hue[0.38960679054806313`, 
       Rational[1, 2], 1], 
      Hue[0.39077699239546804`, 
       Rational[1, 2], 1], 
      Hue[0.39077699239546804`, 
       Rational[1, 2], 1], 
      Hue[0.3913620933191705, 
       Rational[1, 2], 1], 
      Hue[0.3913620933191705, 
       Rational[1, 2], 1], 
      Hue[0.391947194242873, 
       Rational[1, 2], 1], 
      Hue[0.391947194242873, 
       Rational[1, 2], 1], 
      Hue[0.39311739609027785`, 
       Rational[1, 2], 1], 
      Hue[0.39311739609027785`, 
       Rational[1, 2], 1], 
      Hue[0.3937024970139803, 
       Rational[1, 2], 1], 
      Hue[0.3937024970139803, 
       Rational[1, 2], 1], 
      Hue[0.39428759793768275`, 
       Rational[1, 2], 1], 
      Hue[0.39428759793768275`, 
       Rational[1, 2], 1], 
      Hue[0.3954577997850876, 
       Rational[1, 2], 1], 
      Hue[0.3954577997850876, 
       Rational[1, 2], 1], 
      Hue[0.3977982034798974, 
       Rational[1, 2], 1], 
      Hue[0.3977982034798974, 
       Rational[1, 2], 1], 
      Hue[0.39838330440359987`, 
       Rational[1, 2], 1], 
      Hue[0.39838330440359987`, 
       Rational[1, 2], 1], 
      Hue[0.3989684053273023, 
       Rational[1, 2], 1], 
      Hue[0.3989684053273023, 
       Rational[1, 2], 1], 
      Hue[0.4001386071747072, 
       Rational[1, 2], 1], 
      Hue[0.4001386071747072, 
       Rational[1, 2], 1], 
      Hue[0.402479010869517, 
       Rational[1, 2], 1], 
      Hue[0.402479010869517, 
       Rational[1, 2], 1], 
      Hue[0.40311338183927814`, 
       Rational[1, 2], 1], 
      Hue[0.40311338183927814`, 
       Rational[1, 2], 1], 
      Hue[0.4037477528090393, 
       Rational[1, 2], 1], 
      Hue[0.4037477528090393, 
       Rational[1, 2], 1], 
      Hue[0.4050164947485616, 
       Rational[1, 2], 1], 
      Hue[0.4050164947485616, 
       Rational[1, 2], 1], 
      Hue[0.40755397862760634`, 
       Rational[1, 2], 1], 
      Hue[0.40755397862760634`, 
       Rational[1, 2], 1], 
      Hue[0.40818834959736744`, 
       Rational[1, 2], 1], 
      Hue[0.40818834959736744`, 
       Rational[1, 2], 1], 
      Hue[0.4088227205671286, 
       Rational[1, 2], 1], 
      Hue[0.4088227205671286, 
       Rational[1, 2], 1], 
      Hue[0.4100914625066509, 
       Rational[1, 2], 1], 
      Hue[0.4100914625066509, 
       Rational[1, 2], 1], 
      Hue[0.41262894638569564`, 
       Rational[1, 2], 1], 
      Hue[0.41262894638569564`, 
       Rational[1, 2], 1], 
      Hue[0.4132633173554568, 
       Rational[1, 2], 1], 
      Hue[0.4132633173554568, 
       Rational[1, 2], 1], 
      Hue[0.413897688325218, 
       Rational[1, 2], 1], 
      Hue[0.413897688325218, 
       Rational[1, 2], 1], 
      Hue[0.4151664302647403, 
       Rational[1, 2], 1], 
      Hue[0.4151664302647403, 
       Rational[1, 2], 1], 
      Hue[0.41770391414378494`, 
       Rational[1, 2], 1], 
      Hue[0.41770391414378494`, 
       Rational[1, 2], 1], 
      Hue[0.4227788819018743, 
       Rational[1, 2], 1], 
      Hue[0.4227788819018743, 
       Rational[1, 2], 1], 
      Hue[0.4234132528716354, 
       Rational[1, 2], 1], 
      Hue[0.4234132528716354, 
       Rational[1, 2], 1], 
      Hue[0.4240476238413966, 
       Rational[1, 2], 1], 
      Hue[0.4240476238413966, 
       Rational[1, 2], 1], 
      Hue[0.4253163657809189, 
       Rational[1, 2], 1], 
      Hue[0.4253163657809189, 
       Rational[1, 2], 1], 
      Hue[0.4278538496599636, 
       Rational[1, 2], 1], 
      Hue[0.4278538496599636, 
       Rational[1, 2], 1], 
      Hue[0.4329288174180529, 
       Rational[1, 2], 1], 
      Hue[0.4329288174180529, 
       Rational[1, 2], 1], 
      Hue[0.44307875293423155`, 
       Rational[1, 2], 1], 
      Hue[0.44307875293423155`, 
       Rational[1, 2], 1], 
      Hue[0.4436710422607767, 
       Rational[1, 2], 1], 
      Hue[0.4436710422607767, 
       Rational[1, 2], 1], 
      Hue[0.44426333158732195`, 
       Rational[1, 2], 1], 
      Hue[0.44426333158732195`, 
       Rational[1, 2], 1], 
      Hue[0.44544791024041236`, 
       Rational[1, 2], 1], 
      Hue[0.44544791024041236`, 
       Rational[1, 2], 1], 
      Hue[0.4478170675465931, 
       Rational[1, 2], 1], 
      Hue[0.4478170675465931, 
       Rational[1, 2], 1], 
      Hue[0.45255538215895474`, 
       Rational[1, 2], 1], 
      Hue[0.45255538215895474`, 
       Rational[1, 2], 1], 
      Hue[0.46203201138367794`, 
       Rational[1, 2], 1], 
      Hue[0.46203201138367794`, 
       Rational[1, 2], 1], 
      Hue[0.46262430071022315`, 
       Rational[1, 2], 1], 
      Hue[0.46262430071022315`, 
       Rational[1, 2], 1], 
      Hue[0.46321659003676835`, 
       Rational[1, 2], 1], 
      Hue[0.46321659003676835`, 
       Rational[1, 2], 1], 
      Hue[0.46440116868985876`, 
       Rational[1, 2], 1], 
      Hue[0.46440116868985876`, 
       Rational[1, 2], 1], 
      Hue[0.46677032599603957`, 
       Rational[1, 2], 1], 
      Hue[0.46677032599603957`, 
       Rational[1, 2], 1], 
      Hue[0.4715086406084012, 
       Rational[1, 2], 1], 
      Hue[0.4715086406084012, 
       Rational[1, 2], 1], 
      Hue[0.4809852698331244, 
       Rational[1, 2], 1], 
      Hue[0.4809852698331244, 
       Rational[1, 2], 1], 
      Hue[0.4815659280795452, 
       Rational[1, 2], 1], 
      Hue[0.4815659280795452, 
       Rational[1, 2], 1], 
      Hue[0.48214658632596596`, 
       Rational[1, 2], 1], 
      Hue[0.48214658632596596`, 
       Rational[1, 2], 1], 
      Hue[0.4833079028188076, 
       Rational[1, 2], 1], 
      Hue[0.4833079028188076, 
       Rational[1, 2], 1], 
      Hue[0.4856305358044908, 
       Rational[1, 2], 1], 
      Hue[0.4856305358044908, 
       Rational[1, 2], 1], 
      Hue[0.4902758017758571, 
       Rational[1, 2], 1], 
      Hue[0.4902758017758571, 
       Rational[1, 2], 1], 
      Hue[0.4908564600222779, 
       Rational[1, 2], 1], 
      Hue[0.4908564600222779, 
       Rational[1, 2], 1], 
      Hue[0.4914371182686987, 
       Rational[1, 2], 1], 
      Hue[0.4914371182686987, 
       Rational[1, 2], 1], 
      Hue[0.4925984347615403, 
       Rational[1, 2], 1], 
      Hue[0.4925984347615403, 
       Rational[1, 2], 1], 
      Hue[0.4949210677472235, 
       Rational[1, 2], 1], 
      Hue[0.4949210677472235, 
       Rational[1, 2], 1], 
      Hue[0.4955017259936443, 
       Rational[1, 2], 1], 
      Hue[0.4955017259936443, 
       Rational[1, 2], 1], 
      Hue[0.49608238424006507`, 
       Rational[1, 2], 1], 
      Hue[0.49608238424006507`, 
       Rational[1, 2], 1], 
      Hue[0.4972437007329067, 
       Rational[1, 2], 1], 
      Hue[0.4972437007329067, 
       Rational[1, 2], 1], 
      Hue[0.4978243589793275, 
       Rational[1, 2], 1], 
      Hue[0.4978243589793275, 
       Rational[1, 2], 1], 
      Hue[0.4984050172257483, 
       Rational[1, 2], 1], 
      Hue[0.4984050172257483, 
       Rational[1, 2], 1], 
      Hue[0.4989856754721691, 
       Rational[1, 2], 1], 
      Hue[0.4989856754721691, 
       Rational[1, 2], 1], 
      Hue[0.4995663337185899, 
       Rational[1, 2], 1], 
      Hue[0.4995663337185899, 
       Rational[1, 2], 1], 
      Hue[0.00014699196501062852`, 
       Rational[1, 2], 1], 
      Hue[0.00014699196501062852`, 
       Rational[1, 2], 1], 
      Hue[0.0007276502114314253, 
       Rational[1, 2], 1], 
      Hue[0.0007276502114314253, 
       Rational[1, 2], 1], 
      Hue[0.001308308457852222, 
       Rational[1, 2], 1], 
      Hue[0.001308308457852222, 
       Rational[1, 2], 1], 
      Hue[0.0018889667042730188`, 
       Rational[1, 2], 1], 
      Hue[0.0018889667042730188`, 
       Rational[1, 2], 1], 
      Hue[0.002469624950693816, 
       Rational[1, 2], 1], 
      Hue[0.002469624950693816, 
       Rational[1, 2], 1], 
      Hue[0.0030502831971146124`, 
       Rational[1, 2], 1], 
      Hue[0.0030502831971146124`, 
       Rational[1, 2], 1], 
      Hue[0.004211599689956207, 
       Rational[1, 2], 1], 
      Hue[0.004211599689956207, 
       Rational[1, 2], 1], 
      Hue[0.004792257936377003, 
       Rational[1, 2], 1], 
      Hue[0.004792257936377003, 
       Rational[1, 2], 1], 
      Hue[0.005372916182797801, 
       Rational[1, 2], 1], 
      Hue[0.005372916182797801, 
       Rational[1, 2], 1], 
      Hue[0.006534232675639393, 
       Rational[1, 2], 1], 
      Hue[0.006534232675639393, 
       Rational[1, 2], 1], 
      Hue[0.00885686566132258, 
       Rational[1, 2], 1], 
      Hue[0.00885686566132258, 
       Rational[1, 2], 1], 
      Hue[0.01814739760405533, 
       Rational[1, 2], 1], 
      Hue[0.01814739760405533, 
       Rational[1, 2], 1], 
      Hue[0.018777325896534847`, 
       Rational[1, 2], 1], 
      Hue[0.018777325896534847`, 
       Rational[1, 2], 1], 
      Hue[0.019407254189014365`, 
       Rational[1, 2], 1], 
      Hue[0.019407254189014365`, 
       Rational[1, 2], 1], 
      Hue[0.020667110773973398`, 
       Rational[1, 2], 1], 
      Hue[0.020667110773973398`, 
       Rational[1, 2], 1], 
      Hue[0.02318682394389146, 
       Rational[1, 2], 1], 
      Hue[0.02318682394389146, 
       Rational[1, 2], 1], 
      Hue[0.028226250283727595`, 
       Rational[1, 2], 1], 
      Hue[0.028226250283727595`, 
       Rational[1, 2], 1], 
      Hue[0.03830510296339986, 
       Rational[1, 2], 1], 
      Hue[0.03830510296339986, 
       Rational[1, 2], 1], 
      Hue[0.03893503125587938, 
       Rational[1, 2], 1], 
      Hue[0.03893503125587938, 
       Rational[1, 2], 1], 
      Hue[0.03956495954835889, 
       Rational[1, 2], 1], 
      Hue[0.03956495954835889, 
       Rational[1, 2], 1], 
      Hue[0.04082481613331793, 
       Rational[1, 2], 1], 
      Hue[0.04082481613331793, 
       Rational[1, 2], 1], 
      Hue[0.04334452930323598, 
       Rational[1, 2], 1], 
      Hue[0.04334452930323598, 
       Rational[1, 2], 1], 
      Hue[0.048383955643072116`, 
       Rational[1, 2], 1], 
      Hue[0.048383955643072116`, 
       Rational[1, 2], 1], 
      Hue[0.058462808322744406`, 
       Rational[1, 2], 1], 
      Hue[0.058462808322744406`, 
       Rational[1, 2], 1], 
      Hue[0.05905065497200795, 
       Rational[1, 2], 1], 
      Hue[0.05905065497200795, 
       Rational[1, 2], 1], 
      Hue[0.05963850162127149, 
       Rational[1, 2], 1], 
      Hue[0.05963850162127149, 
       Rational[1, 2], 1], 
      Hue[0.0608141949197986, 
       Rational[1, 2], 1], 
      Hue[0.0608141949197986, 
       Rational[1, 2], 1], 
      Hue[0.06316558151685281, 
       Rational[1, 2], 1], 
      Hue[0.06316558151685281, 
       Rational[1, 2], 1], 
      Hue[0.06786835471096123, 
       Rational[1, 2], 1], 
      Hue[0.06786835471096123, 
       Rational[1, 2], 1], 
      Hue[0.06845620136022479, 
       Rational[1, 2], 1], 
      Hue[0.06845620136022479, 
       Rational[1, 2], 1], 
      Hue[0.06904404800948832, 
       Rational[1, 2], 1], 
      Hue[0.06904404800948832, 
       Rational[1, 2], 1], 
      Hue[0.07021974130801542, 
       Rational[1, 2], 1], 
      Hue[0.07021974130801542, 
       Rational[1, 2], 1], 
      Hue[0.07257112790506963, 
       Rational[1, 2], 1], 
      Hue[0.07257112790506963, 
       Rational[1, 2], 1], 
      Hue[0.07727390109917805, 
       Rational[1, 2], 1], 
      Hue[0.07727390109917805, 
       Rational[1, 2], 1], 
      Hue[0.0778617477484416, 
       Rational[1, 2], 1], 
      Hue[0.0778617477484416, 
       Rational[1, 2], 1], 
      Hue[0.07844959439770514, 
       Rational[1, 2], 1], 
      Hue[0.07844959439770514, 
       Rational[1, 2], 1], 
      Hue[0.07962528769623224, 
       Rational[1, 2], 1], 
      Hue[0.07962528769623224, 
       Rational[1, 2], 1], 
      Hue[0.08197667429328648, 
       Rational[1, 2], 1], 
      Hue[0.08197667429328648, 
       Rational[1, 2], 1], 
      Hue[0.08256452094255003, 
       Rational[1, 2], 1], 
      Hue[0.08256452094255003, 
       Rational[1, 2], 1], 
      Hue[0.08315236759181359, 
       Rational[1, 2], 1], 
      Hue[0.08315236759181359, 
       Rational[1, 2], 1], 
      Hue[0.08432806089034067, 
       Rational[1, 2], 1], 
      Hue[0.08432806089034067, 
       Rational[1, 2], 1], 
      Hue[0.08667944748739487, 
       Rational[1, 2], 1], 
      Hue[0.08667944748739487, 
       Rational[1, 2], 1], 
      Hue[0.08726729413665844, 
       Rational[1, 2], 1], 
      Hue[0.08726729413665844, 
       Rational[1, 2], 1], 
      Hue[0.08785514078592198, 
       Rational[1, 2], 1], 
      Hue[0.08785514078592198, 
       Rational[1, 2], 1], 
      Hue[0.08903083408444905, 
       Rational[1, 2], 1], 
      Hue[0.08903083408444905, 
       Rational[1, 2], 1], 
      Hue[0.09138222068150327, 
       Rational[1, 2], 1], 
      Hue[0.09138222068150327, 
       Rational[1, 2], 1], 
      Hue[0.09197006733076682, 
       Rational[1, 2], 1], 
      Hue[0.09197006733076682, 
       Rational[1, 2], 1], 
      Hue[0.09255791398003038, 
       Rational[1, 2], 1], 
      Hue[0.09255791398003038, 
       Rational[1, 2], 1], 
      Hue[0.0937336072785575, 
       Rational[1, 2], 1], 
      Hue[0.0937336072785575, 
       Rational[1, 2], 1], 
      Hue[0.09608499387561169, 
       Rational[1, 2], 1], 
      Hue[0.09608499387561169, 
       Rational[1, 2], 1], 
      Hue[0.09672211057093397, 
       Rational[1, 2], 1], 
      Hue[0.09672211057093397, 
       Rational[1, 2], 1], 
      Hue[0.09735922726625623, 
       Rational[1, 2], 1], 
      Hue[0.09735922726625623, 
       Rational[1, 2], 1], 
      Hue[0.09863346065690076, 
       Rational[1, 2], 1], 
      Hue[0.09863346065690076, 
       Rational[1, 2], 1], 
      Hue[0.09927057735222307, 
       Rational[1, 2], 1], 
      Hue[0.09927057735222307, 
       Rational[1, 2], 1], 
      Hue[0.09990769404754532, 
       Rational[1, 2], 1], 
      Hue[0.09990769404754532, 
       Rational[1, 2], 1], 
      Hue[0.10118192743818985`, 
       Rational[1, 2], 1], 
      Hue[0.10118192743818985`, 
       Rational[1, 2], 1], 
      Hue[0.10181904413351212`, 
       Rational[1, 2], 1], 
      Hue[0.10181904413351212`, 
       Rational[1, 2], 1], 
      Hue[0.1024561608288344, 
       Rational[1, 2], 1], 
      Hue[0.1024561608288344, 
       Rational[1, 2], 1], 
      Hue[0.10309327752415666`, 
       Rational[1, 2], 1], 
      Hue[0.10309327752415666`, 
       Rational[1, 2], 1], 
      Hue[0.10373039421947897`, 
       Rational[1, 2], 1], 
      Hue[0.10373039421947897`, 
       Rational[1, 2], 1], 
      Hue[0.10436751091480122`, 
       Rational[1, 2], 1], 
      Hue[0.10436751091480122`, 
       Rational[1, 2], 1], 
      Hue[0.10500462761012351`, 
       Rational[1, 2], 1], 
      Hue[0.10500462761012351`, 
       Rational[1, 2], 1], 
      Hue[0.10564174430544576`, 
       Rational[1, 2], 1], 
      Hue[0.10564174430544576`, 
       Rational[1, 2], 1], 
      Hue[0.10627886100076801`, 
       Rational[1, 2], 1], 
      Hue[0.10627886100076801`, 
       Rational[1, 2], 1], 
      Hue[0.10691597769609032`, 
       Rational[1, 2], 1], 
      Hue[0.10691597769609032`, 
       Rational[1, 2], 1], 
      Hue[0.10755309439141254`, 
       Rational[1, 2], 1], 
      Hue[0.10755309439141254`, 
       Rational[1, 2], 1], 
      Hue[0.10819021108673485`, 
       Rational[1, 2], 1], 
      Hue[0.10819021108673485`, 
       Rational[1, 2], 1], 
      Hue[0.10882732778205713`, 
       Rational[1, 2], 1], 
      Hue[0.10882732778205713`, 
       Rational[1, 2], 1], 
      Hue[0.10946444447737938`, 
       Rational[1, 2], 1], 
      Hue[0.10946444447737938`, 
       Rational[1, 2], 1], 
      Hue[0.11010156117270166`, 
       Rational[1, 2], 1], 
      Hue[0.11010156117270166`, 
       Rational[1, 2], 1], 
      Hue[0.11073867786802392`, 
       Rational[1, 2], 1], 
      Hue[0.11073867786802392`, 
       Rational[1, 2], 1], 
      Hue[0.11137579456334618`, 
       Rational[1, 2], 1], 
      Hue[0.11137579456334618`, 
       Rational[1, 2], 1], 
      Hue[0.11201291125866845`, 
       Rational[1, 2], 1], 
      Hue[0.11201291125866845`, 
       Rational[1, 2], 1], 
      Hue[0.11265002795399072`, 
       Rational[1, 2], 1], 
      Hue[0.11265002795399072`, 
       Rational[1, 2], 1], 
      Hue[0.11328714464931298`, 
       Rational[1, 2], 1], 
      Hue[0.11328714464931298`, 
       Rational[1, 2], 1], 
      Hue[0.11392426134463528`, 
       Rational[1, 2], 1], 
      Hue[0.11392426134463528`, 
       Rational[1, 2], 1], 
      Hue[0.11456137803995754`, 
       Rational[1, 2], 1], 
      Hue[0.11456137803995754`, 
       Rational[1, 2], 1], 
      Hue[0.11519849473527982`, 
       Rational[1, 2], 1], 
      Hue[0.11519849473527982`, 
       Rational[1, 2], 1], 
      Hue[0.11583561143060207`, 
       Rational[1, 2], 1], 
      Hue[0.11583561143060207`, 
       Rational[1, 2], 1], 
      Hue[0.11647272812592435`, 
       Rational[1, 2], 1], 
      Hue[0.11647272812592435`, 
       Rational[1, 2], 1], 
      Hue[0.11710984482124662`, 
       Rational[1, 2], 1], 
      Hue[0.11710984482124662`, 
       Rational[1, 2], 1], 
      Hue[0.11774696151656887`, 
       Rational[1, 2], 1], 
      Hue[0.11774696151656887`, 
       Rational[1, 2], 1], 
      Hue[0.11838407821189115`, 
       Rational[1, 2], 1], 
      Hue[0.11838407821189115`, 
       Rational[1, 2], 1], 
      Hue[0.11902119490721341`, 
       Rational[1, 2], 1], 
      Hue[0.11902119490721341`, 
       Rational[1, 2], 1], 
      Hue[0.1196583116025357, 
       Rational[1, 2], 1], 
      Hue[0.1196583116025357, 
       Rational[1, 2], 1], 
      Hue[0.12029542829785798`, 
       Rational[1, 2], 1], 
      Hue[0.12029542829785798`, 
       Rational[1, 2], 1], 
      Hue[0.1215696616885025, 
       Rational[1, 2], 1], 
      Hue[0.1215696616885025, 
       Rational[1, 2], 1], 
      Hue[0.12220677838382481`, 
       Rational[1, 2], 1], 
      Hue[0.12220677838382481`, 
       Rational[1, 2], 1], 
      Hue[0.12284389507914709`, 
       Rational[1, 2], 1], 
      Hue[0.12284389507914709`, 
       Rational[1, 2], 1], 
      Hue[0.12411812846979162`, 
       Rational[1, 2], 1], 
      Hue[0.12411812846979162`, 
       Rational[1, 2], 1], 
      Hue[0.12475524516511388`, 
       Rational[1, 2], 1], 
      Hue[0.12475524516511388`, 
       Rational[1, 2], 1], 
      Hue[0.12539236186043612`, 
       Rational[1, 2], 1], 
      Hue[0.12539236186043612`, 
       Rational[1, 2], 1], 
      Hue[0.1266665952510807, 
       Rational[1, 2], 1], 
      Hue[0.1266665952510807, 
       Rational[1, 2], 1], 
      Hue[0.12730371194640297`, 
       Rational[1, 2], 1], 
      Hue[0.12730371194640297`, 
       Rational[1, 2], 1], 
      Hue[0.12794082864172526`, 
       Rational[1, 2], 1], 
      Hue[0.12794082864172526`, 
       Rational[1, 2], 1], 
      Hue[0.1292150620323698, 
       Rational[1, 2], 1], 
      Hue[0.1292150620323698, 
       Rational[1, 2], 1], 
      Hue[0.12985217872769206`, 
       Rational[1, 2], 1], 
      Hue[0.12985217872769206`, 
       Rational[1, 2], 1], 
      Hue[0.13048929542301432`, 
       Rational[1, 2], 1], 
      Hue[0.13048929542301432`, 
       Rational[1, 2], 1], 
      Hue[0.13176352881365888`, 
       Rational[1, 2], 1], 
      Hue[0.13176352881365888`, 
       Rational[1, 2], 1], 
      Hue[0.13240064550898115`, 
       Rational[1, 2], 1], 
      Hue[0.13240064550898115`, 
       Rational[1, 2], 1], 
      Hue[0.13303776220430344`, 
       Rational[1, 2], 1], 
      Hue[0.13303776220430344`, 
       Rational[1, 2], 1], 
      Hue[0.13431199559494797`, 
       Rational[1, 2], 1], 
      Hue[0.13431199559494797`, 
       Rational[1, 2], 1], 
      Hue[0.13686046237623703`, 
       Rational[1, 2], 1], 
      Hue[0.13686046237623703`, 
       Rational[1, 2], 1], 
      Hue[0.1374859479914349, 
       Rational[1, 2], 1], 
      Hue[0.1374859479914349, 
       Rational[1, 2], 1], 
      Hue[0.13811143360663275`, 
       Rational[1, 2], 1], 
      Hue[0.13811143360663275`, 
       Rational[1, 2], 1], 
      Hue[0.13936240483702853`, 
       Rational[1, 2], 1], 
      Hue[0.13936240483702853`, 
       Rational[1, 2], 1], 
      Hue[0.14186434729781996`, 
       Rational[1, 2], 1], 
      Hue[0.14186434729781996`, 
       Rational[1, 2], 1], 
      Hue[0.14248983291301787`, 
       Rational[1, 2], 1], 
      Hue[0.14248983291301787`, 
       Rational[1, 2], 1], 
      Hue[0.14311531852821574`, 
       Rational[1, 2], 1], 
      Hue[0.14311531852821574`, 
       Rational[1, 2], 1], 
      Hue[0.14436628975861146`, 
       Rational[1, 2], 1], 
      Hue[0.14436628975861146`, 
       Rational[1, 2], 1], 
      Hue[0.14686823221940293`, 
       Rational[1, 2], 1], 
      Hue[0.14686823221940293`, 
       Rational[1, 2], 1], 
      Hue[0.14749371783460077`, 
       Rational[1, 2], 1], 
      Hue[0.14749371783460077`, 
       Rational[1, 2], 1], 
      Hue[0.14811920344979865`, 
       Rational[1, 2], 1], 
      Hue[0.14811920344979865`, 
       Rational[1, 2], 1], 
      Hue[0.1493701746801944, 
       Rational[1, 2], 1], 
      Hue[0.1493701746801944, 
       Rational[1, 2], 1], 
      Hue[0.15187211714098586`, 
       Rational[1, 2], 1], 
      Hue[0.15187211714098586`, 
       Rational[1, 2], 1], 
      Hue[0.15687600206256883`, 
       Rational[1, 2], 1], 
      Hue[0.15687600206256883`, 
       Rational[1, 2], 1], 
      Hue[0.1575014876777667, 
       Rational[1, 2], 1], 
      Hue[0.1575014876777667, 
       Rational[1, 2], 1], 
      Hue[0.15812697329296455`, 
       Rational[1, 2], 1], 
      Hue[0.15812697329296455`, 
       Rational[1, 2], 1], 
      Hue[0.15937794452336027`, 
       Rational[1, 2], 1], 
      Hue[0.15937794452336027`, 
       Rational[1, 2], 1], 
      Hue[0.16187988698415173`, 
       Rational[1, 2], 1], 
      Hue[0.16187988698415173`, 
       Rational[1, 2], 1], 
      Hue[0.1668837719057347, 
       Rational[1, 2], 1], 
      Hue[0.1668837719057347, 
       Rational[1, 2], 1], 
      Hue[0.1768915417489006, 
       Rational[1, 2], 1], 
      Hue[0.1768915417489006, 
       Rational[1, 2], 1], 
      Hue[0.17747494572088252`, 
       Rational[1, 2], 1], 
      Hue[0.17747494572088252`, 
       Rational[1, 2], 1], 
      Hue[0.1780583496928644, 
       Rational[1, 2], 1], 
      Hue[0.1780583496928644, 
       Rational[1, 2], 1], 
      Hue[0.17922515763682823`, 
       Rational[1, 2], 1], 
      Hue[0.17922515763682823`, 
       Rational[1, 2], 1], 
      Hue[0.18155877352475583`, 
       Rational[1, 2], 1], 
      Hue[0.18155877352475583`, 
       Rational[1, 2], 1], 
      Hue[0.18622600530061104`, 
       Rational[1, 2], 1], 
      Hue[0.18622600530061104`, 
       Rational[1, 2], 1], 
      Hue[0.19556046885232148`, 
       Rational[1, 2], 1], 
      Hue[0.19556046885232148`, 
       Rational[1, 2], 1], 
      Hue[0.1961438728243034, 
       Rational[1, 2], 1], 
      Hue[0.1961438728243034, 
       Rational[1, 2], 1], 
      Hue[0.1967272767962853, 
       Rational[1, 2], 1], 
      Hue[0.1967272767962853, 
       Rational[1, 2], 1], 
      Hue[0.1978940847402491, 
       Rational[1, 2], 1], 
      Hue[0.1978940847402491, 
       Rational[1, 2], 1], 
      Hue[0.2002277006281767, 
       Rational[1, 2], 1], 
      Hue[0.2002277006281767, 
       Rational[1, 2], 1], 
      Hue[0.20489493240403192`, 
       Rational[1, 2], 1], 
      Hue[0.20489493240403192`, 
       Rational[1, 2], 1], 
      Hue[0.2142293959557424, 
       Rational[1, 2], 1], 
      Hue[0.2142293959557424, 
       Rational[1, 2], 1], 
      Hue[0.214862069973783, 
       Rational[1, 2], 1], 
      Hue[0.214862069973783, 
       Rational[1, 2], 1], 
      Hue[0.2154947439918236, 
       Rational[1, 2], 1], 
      Hue[0.2154947439918236, 
       Rational[1, 2], 1], 
      Hue[0.21676009202790486`, 
       Rational[1, 2], 1], 
      Hue[0.21676009202790486`, 
       Rational[1, 2], 1], 
      Hue[0.21929078810006736`, 
       Rational[1, 2], 1], 
      Hue[0.21929078810006736`, 
       Rational[1, 2], 1], 
      Hue[0.22435218024439235`, 
       Rational[1, 2], 1], 
      Hue[0.22435218024439235`, 
       Rational[1, 2], 1], 
      Hue[0.22498485426243298`, 
       Rational[1, 2], 1], 
      Hue[0.22498485426243298`, 
       Rational[1, 2], 1], 
      Hue[0.2256175282804736, 
       Rational[1, 2], 1], 
      Hue[0.2256175282804736, 
       Rational[1, 2], 1], 
      Hue[0.22688287631655485`, 
       Rational[1, 2], 1], 
      Hue[0.22688287631655485`, 
       Rational[1, 2], 1], 
      Hue[0.22941357238871735`, 
       Rational[1, 2], 1], 
      Hue[0.22941357238871735`, 
       Rational[1, 2], 1], 
      Hue[0.2344749645330423, 
       Rational[1, 2], 1], 
      Hue[0.2344749645330423, 
       Rational[1, 2], 1], 
      Hue[0.23510763855108294`, 
       Rational[1, 2], 1], 
      Hue[0.23510763855108294`, 
       Rational[1, 2], 1], 
      Hue[0.23574031256912356`, 
       Rational[1, 2], 1], 
      Hue[0.23574031256912356`, 
       Rational[1, 2], 1], 
      Hue[0.2370056606052048, 
       Rational[1, 2], 1], 
      Hue[0.2370056606052048, 
       Rational[1, 2], 1], 
      Hue[0.2395363566773673, 
       Rational[1, 2], 1], 
      Hue[0.2395363566773673, 
       Rational[1, 2], 1], 
      Hue[0.24016903069540793`, 
       Rational[1, 2], 1], 
      Hue[0.24016903069540793`, 
       Rational[1, 2], 1], 
      Hue[0.24080170471344856`, 
       Rational[1, 2], 1], 
      Hue[0.24080170471344856`, 
       Rational[1, 2], 1], 
      Hue[0.2420670527495298, 
       Rational[1, 2], 1], 
      Hue[0.2420670527495298, 
       Rational[1, 2], 1], 
      Hue[0.2445977488216923, 
       Rational[1, 2], 1], 
      Hue[0.2445977488216923, 
       Rational[1, 2], 1], 
      Hue[0.24523042283973293`, 
       Rational[1, 2], 1], 
      Hue[0.24523042283973293`, 
       Rational[1, 2], 1], 
      Hue[0.24586309685777355`, 
       Rational[1, 2], 1], 
      Hue[0.24586309685777355`, 
       Rational[1, 2], 1], 
      Hue[0.2471284448938548, 
       Rational[1, 2], 1], 
      Hue[0.2471284448938548, 
       Rational[1, 2], 1], 
      Hue[0.24776111891189542`, 
       Rational[1, 2], 1], 
      Hue[0.24776111891189542`, 
       Rational[1, 2], 1], 
      Hue[0.24839379292993605`, 
       Rational[1, 2], 1], 
      Hue[0.24839379292993605`, 
       Rational[1, 2], 1], 
      Hue[0.24902646694797667`, 
       Rational[1, 2], 1], 
      Hue[0.24902646694797667`, 
       Rational[1, 2], 1], 
      Hue[0.2496591409660173, 
       Rational[1, 2], 1], 
      Hue[0.2496591409660173, 
       Rational[1, 2], 1], 
      Hue[0.7502918149840578, 
       Rational[1, 2], 1], 
      Hue[0.7502918149840578, 
       Rational[1, 2], 1], 
      Hue[0.7509244890020985, 
       Rational[1, 2], 1], 
      Hue[0.7509244890020985, 
       Rational[1, 2], 1], 
      Hue[0.7515571630201391, 
       Rational[1, 2], 1], 
      Hue[0.7515571630201391, 
       Rational[1, 2], 1], 
      Hue[0.7521898370381798, 
       Rational[1, 2], 1], 
      Hue[0.7521898370381798, 
       Rational[1, 2], 1], 
      Hue[0.7528225110562203, 
       Rational[1, 2], 1], 
      Hue[0.7528225110562203, 
       Rational[1, 2], 1], 
      Hue[0.753455185074261, 
       Rational[1, 2], 1], 
      Hue[0.753455185074261, 
       Rational[1, 2], 1], 
      Hue[0.7547205331103422, 
       Rational[1, 2], 1], 
      Hue[0.7547205331103422, 
       Rational[1, 2], 1], 
      Hue[0.7553111254851669, 
       Rational[1, 2], 1], 
      Hue[0.7553111254851669, 
       Rational[1, 2], 1], 
      Hue[0.7559017178599915, 
       Rational[1, 2], 1], 
      Hue[0.7559017178599915, 
       Rational[1, 2], 1], 
      Hue[0.7570829026096408, 
       Rational[1, 2], 1], 
      Hue[0.7570829026096408, 
       Rational[1, 2], 1], 
      Hue[0.7594452721089394, 
       Rational[1, 2], 1], 
      Hue[0.7594452721089394, 
       Rational[1, 2], 1], 
      Hue[0.7641700111075367, 
       Rational[1, 2], 1], 
      Hue[0.7641700111075367, 
       Rational[1, 2], 1], 
      Hue[0.7647606034823614, 
       Rational[1, 2], 1], 
      Hue[0.7647606034823614, 
       Rational[1, 2], 1], 
      Hue[0.765351195857186, 
       Rational[1, 2], 1], 
      Hue[0.765351195857186, 
       Rational[1, 2], 1], 
      Hue[0.7665323806068354, 
       Rational[1, 2], 1], 
      Hue[0.7665323806068354, 
       Rational[1, 2], 1], 
      Hue[0.7688947501061341, 
       Rational[1, 2], 1], 
      Hue[0.7688947501061341, 
       Rational[1, 2], 1], 
      Hue[0.7736194891047312, 
       Rational[1, 2], 1], 
      Hue[0.7736194891047312, 
       Rational[1, 2], 1], 
      Hue[0.774210081479556, 
       Rational[1, 2], 1], 
      Hue[0.774210081479556, 
       Rational[1, 2], 1], 
      Hue[0.7748006738543806, 
       Rational[1, 2], 1], 
      Hue[0.7748006738543806, 
       Rational[1, 2], 1], 
      Hue[0.7759818586040299, 
       Rational[1, 2], 1], 
      Hue[0.7759818586040299, 
       Rational[1, 2], 1], 
      Hue[0.7783442281033285, 
       Rational[1, 2], 1], 
      Hue[0.7783442281033285, 
       Rational[1, 2], 1], 
      Hue[0.7830689671019258, 
       Rational[1, 2], 1], 
      Hue[0.7830689671019258, 
       Rational[1, 2], 1], 
      Hue[0.7925184450991203, 
       Rational[1, 2], 1], 
      Hue[0.7925184450991203, 
       Rational[1, 2], 1], 
      Hue[0.7930974063938205, 
       Rational[1, 2], 1], 
      Hue[0.7930974063938205, 
       Rational[1, 2], 1], 
      Hue[0.7936763676885208, 
       Rational[1, 2], 1], 
      Hue[0.7936763676885208, 
       Rational[1, 2], 1], 
      Hue[0.7948342902779213, 
       Rational[1, 2], 1], 
      Hue[0.7948342902779213, 
       Rational[1, 2], 1], 
      Hue[0.7971501354567223, 
       Rational[1, 2], 1], 
      Hue[0.7971501354567223, 
       Rational[1, 2], 1], 
      Hue[0.8017818258143243, 
       Rational[1, 2], 1], 
      Hue[0.8017818258143243, 
       Rational[1, 2], 1], 
      Hue[0.8023607871090246, 
       Rational[1, 2], 1], 
      Hue[0.8023607871090246, 
       Rational[1, 2], 1], 
      Hue[0.8029397484037248, 
       Rational[1, 2], 1], 
      Hue[0.8029397484037248, 
       Rational[1, 2], 1], 
      Hue[0.8040976709931253, 
       Rational[1, 2], 1], 
      Hue[0.8040976709931253, 
       Rational[1, 2], 1], 
      Hue[0.8064135161719264, 
       Rational[1, 2], 1], 
      Hue[0.8064135161719264, 
       Rational[1, 2], 1], 
      Hue[0.8110452065295284, 
       Rational[1, 2], 1], 
      Hue[0.8110452065295284, 
       Rational[1, 2], 1], 
      Hue[0.8116241678242286, 
       Rational[1, 2], 1], 
      Hue[0.8116241678242286, 
       Rational[1, 2], 1], 
      Hue[0.8122031291189289, 
       Rational[1, 2], 1], 
      Hue[0.8122031291189289, 
       Rational[1, 2], 1], 
      Hue[0.8133610517083294, 
       Rational[1, 2], 1], 
      Hue[0.8133610517083294, 
       Rational[1, 2], 1], 
      Hue[0.8156768968871304, 
       Rational[1, 2], 1], 
      Hue[0.8156768968871304, 
       Rational[1, 2], 1], 
      Hue[0.8162558581818307, 
       Rational[1, 2], 1], 
      Hue[0.8162558581818307, 
       Rational[1, 2], 1], 
      Hue[0.8168348194765309, 
       Rational[1, 2], 1], 
      Hue[0.8168348194765309, 
       Rational[1, 2], 1], 
      Hue[0.8179927420659314, 
       Rational[1, 2], 1], 
      Hue[0.8179927420659314, 
       Rational[1, 2], 1], 
      Hue[0.8203085872447325, 
       Rational[1, 2], 1], 
      Hue[0.8203085872447325, 
       Rational[1, 2], 1], 
      Hue[0.8208875485394328, 
       Rational[1, 2], 1], 
      Hue[0.8208875485394328, 
       Rational[1, 2], 1], 
      Hue[0.8214665098341329, 
       Rational[1, 2], 1], 
      Hue[0.8214665098341329, 
       Rational[1, 2], 1], 
      Hue[0.8226244324235334, 
       Rational[1, 2], 1], 
      Hue[0.8226244324235334, 
       Rational[1, 2], 1], 
      Hue[0.8249402776023345, 
       Rational[1, 2], 1], 
      Hue[0.8249402776023345, 
       Rational[1, 2], 1], 
      Hue[0.8255192388970347, 
       Rational[1, 2], 1], 
      Hue[0.8255192388970347, 
       Rational[1, 2], 1], 
      Hue[0.826098200191735, 
       Rational[1, 2], 1], 
      Hue[0.826098200191735, 
       Rational[1, 2], 1], 
      Hue[0.8272561227811355, 
       Rational[1, 2], 1], 
      Hue[0.8272561227811355, 
       Rational[1, 2], 1], 
      Hue[0.8295719679599365, 
       Rational[1, 2], 1], 
      Hue[0.8295719679599365, 
       Rational[1, 2], 1], 
      Hue[0.8302001993006954, 
       Rational[1, 2], 1], 
      Hue[0.8302001993006954, 
       Rational[1, 2], 1], 
      Hue[0.8308284306414545, 
       Rational[1, 2], 1], 
      Hue[0.8308284306414545, 
       Rational[1, 2], 1], 
      Hue[0.8320848933229724, 
       Rational[1, 2], 1], 
      Hue[0.8320848933229724, 
       Rational[1, 2], 1], 
      Hue[0.8327131246637314, 
       Rational[1, 2], 1], 
      Hue[0.8327131246637314, 
       Rational[1, 2], 1], 
      Hue[0.8333413560044904, 
       Rational[1, 2], 1], 
      Hue[0.8333413560044904, 
       Rational[1, 2], 1], 
      Hue[0.8345978186860082, 
       Rational[1, 2], 1], 
      Hue[0.8345978186860082, 
       Rational[1, 2], 1], 
      Hue[0.8352260500267673, 
       Rational[1, 2], 1], 
      Hue[0.8352260500267673, 
       Rational[1, 2], 1], 
      Hue[0.8358542813675263, 
       Rational[1, 2], 1], 
      Hue[0.8358542813675263, 
       Rational[1, 2], 1], 
      Hue[0.8371107440490442, 
       Rational[1, 2], 1], 
      Hue[0.8371107440490442, 
       Rational[1, 2], 1], 
      Hue[0.8377389753898031, 
       Rational[1, 2], 1], 
      Hue[0.8377389753898031, 
       Rational[1, 2], 1], 
      Hue[0.8383672067305622, 
       Rational[1, 2], 1], 
      Hue[0.8383672067305622, 
       Rational[1, 2], 1], 
      Hue[0.8396236694120801, 
       Rational[1, 2], 1], 
      Hue[0.8396236694120801, 
       Rational[1, 2], 1], 
      Hue[0.8402519007528391, 
       Rational[1, 2], 1], 
      Hue[0.8402519007528391, 
       Rational[1, 2], 1], 
      Hue[0.8408801320935981, 
       Rational[1, 2], 1], 
      Hue[0.8408801320935981, 
       Rational[1, 2], 1], 
      Hue[0.841508363434357, 
       Rational[1, 2], 1], 
      Hue[0.841508363434357, 
       Rational[1, 2], 1], 
      Hue[0.842136594775116, 
       Rational[1, 2], 1], 
      Hue[0.842136594775116, 
       Rational[1, 2], 1], 
      Hue[0.8427648261158749, 
       Rational[1, 2], 1], 
      Hue[0.8427648261158749, 
       Rational[1, 2], 1], 
      Hue[0.843393057456634, 
       Rational[1, 2], 1], 
      Hue[0.843393057456634, 
       Rational[1, 2], 1], 
      Hue[0.8440212887973929, 
       Rational[1, 2], 1], 
      Hue[0.8440212887973929, 
       Rational[1, 2], 1], 
      Hue[0.8446495201381519, 
       Rational[1, 2], 1], 
      Hue[0.8446495201381519, 
       Rational[1, 2], 1], 
      Hue[0.8452777514789109, 
       Rational[1, 2], 1], 
      Hue[0.8452777514789109, 
       Rational[1, 2], 1], 
      Hue[0.8459059828196699, 
       Rational[1, 2], 1], 
      Hue[0.8459059828196699, 
       Rational[1, 2], 1], 
      Hue[0.8465342141604288, 
       Rational[1, 2], 1], 
      Hue[0.8465342141604288, 
       Rational[1, 2], 1], 
      Hue[0.8471624455011878, 
       Rational[1, 2], 1], 
      Hue[0.8471624455011878, 
       Rational[1, 2], 1], 
      Hue[0.8477906768419468, 
       Rational[1, 2], 1], 
      Hue[0.8477906768419468, 
       Rational[1, 2], 1], 
      Hue[0.8484189081827058, 
       Rational[1, 2], 1], 
      Hue[0.8484189081827058, 
       Rational[1, 2], 1], 
      Hue[0.8490471395234647, 
       Rational[1, 2], 1], 
      Hue[0.8490471395234647, 
       Rational[1, 2], 1], 
      Hue[0.8496753708642237, 
       Rational[1, 2], 1], 
      Hue[0.8496753708642237, 
       Rational[1, 2], 1], 
      Hue[0.8503036022049827, 
       Rational[1, 2], 1], 
      Hue[0.8503036022049827, 
       Rational[1, 2], 1], 
      Hue[0.8509318335457416, 
       Rational[1, 2], 1], 
      Hue[0.8509318335457416, 
       Rational[1, 2], 1], 
      Hue[0.8515600648865006, 
       Rational[1, 2], 1], 
      Hue[0.8515600648865006, 
       Rational[1, 2], 1], 
      Hue[0.8521882962272596, 
       Rational[1, 2], 1], 
      Hue[0.8521882962272596, 
       Rational[1, 2], 1], 
      Hue[0.8528165275680185, 
       Rational[1, 2], 1], 
      Hue[0.8528165275680185, 
       Rational[1, 2], 1], 
      Hue[0.8534447589087775, 
       Rational[1, 2], 1], 
      Hue[0.8534447589087775, 
       Rational[1, 2], 1], 
      Hue[0.8540729902495365, 
       Rational[1, 2], 1], 
      Hue[0.8540729902495365, 
       Rational[1, 2], 1], 
      Hue[0.8547012215902955, 
       Rational[1, 2], 1], 
      Hue[0.8547012215902955, 
       Rational[1, 2], 1], 
      Hue[0.8553294529310544, 
       Rational[1, 2], 1], 
      Hue[0.8553294529310544, 
       Rational[1, 2], 1], 
      Hue[0.8559576842718135, 
       Rational[1, 2], 1], 
      Hue[0.8559576842718135, 
       Rational[1, 2], 1], 
      Hue[0.8565859156125724, 
       Rational[1, 2], 1], 
      Hue[0.8565859156125724, 
       Rational[1, 2], 1], 
      Hue[0.8572141469533314, 
       Rational[1, 2], 1], 
      Hue[0.8572141469533314, 
       Rational[1, 2], 1], 
      Hue[0.8578423782940904, 
       Rational[1, 2], 1], 
      Hue[0.8578423782940904, 
       Rational[1, 2], 1], 
      Hue[0.8584706096348493, 
       Rational[1, 2], 1], 
      Hue[0.8584706096348493, 
       Rational[1, 2], 1], 
      Hue[0.8597270723163672, 
       Rational[1, 2], 1], 
      Hue[0.8597270723163672, 
       Rational[1, 2], 1], 
      Hue[0.8603553036571262, 
       Rational[1, 2], 1], 
      Hue[0.8603553036571262, 
       Rational[1, 2], 1], 
      Hue[0.8609835349978852, 
       Rational[1, 2], 1], 
      Hue[0.8609835349978852, 
       Rational[1, 2], 1], 
      Hue[0.8622399976794031, 
       Rational[1, 2], 1], 
      Hue[0.8622399976794031, 
       Rational[1, 2], 1], 
      Hue[0.862868229020162, 
       Rational[1, 2], 1], 
      Hue[0.862868229020162, 
       Rational[1, 2], 1], 
      Hue[0.863496460360921, 
       Rational[1, 2], 1], 
      Hue[0.863496460360921, 
       Rational[1, 2], 1], 
      Hue[0.864752923042439, 
       Rational[1, 2], 1], 
      Hue[0.864752923042439, 
       Rational[1, 2], 1], 
      Hue[0.865381154383198, 
       Rational[1, 2], 1], 
      Hue[0.865381154383198, 
       Rational[1, 2], 1], 
      Hue[0.866009385723957, 
       Rational[1, 2], 1], 
      Hue[0.866009385723957, 
       Rational[1, 2], 1], 
      Hue[0.8672658484054749, 
       Rational[1, 2], 1], 
      Hue[0.8672658484054749, 
       Rational[1, 2], 1], 
      Hue[0.8697787737685108, 
       Rational[1, 2], 1], 
      Hue[0.8697787737685108, 
       Rational[1, 2], 1], 
      Hue[0.8703649234660538, 
       Rational[1, 2], 1], 
      Hue[0.8703649234660538, 
       Rational[1, 2], 1], 
      Hue[0.8709510731635968, 
       Rational[1, 2], 1], 
      Hue[0.8709510731635968, 
       Rational[1, 2], 1], 
      Hue[0.8721233725586828, 
       Rational[1, 2], 1], 
      Hue[0.8721233725586828, 
       Rational[1, 2], 1], 
      Hue[0.8744679713488549, 
       Rational[1, 2], 1], 
      Hue[0.8744679713488549, 
       Rational[1, 2], 1], 
      Hue[0.8750541210463978, 
       Rational[1, 2], 1], 
      Hue[0.8750541210463978, 
       Rational[1, 2], 1], 
      Hue[0.8756402707439409, 
       Rational[1, 2], 1], 
      Hue[0.8756402707439409, 
       Rational[1, 2], 1], 
      Hue[0.8768125701390269, 
       Rational[1, 2], 1], 
      Hue[0.8768125701390269, 
       Rational[1, 2], 1], 
      Hue[0.8791571689291989, 
       Rational[1, 2], 1], 
      Hue[0.8791571689291989, 
       Rational[1, 2], 1], 
      Hue[0.879743318626742, 
       Rational[1, 2], 1], 
      Hue[0.879743318626742, 
       Rational[1, 2], 1], 
      Hue[0.8803294683242849, 
       Rational[1, 2], 1], 
      Hue[0.8803294683242849, 
       Rational[1, 2], 1], 
      Hue[0.8815017677193709, 
       Rational[1, 2], 1], 
      Hue[0.8815017677193709, 
       Rational[1, 2], 1], 
      Hue[0.883846366509543, 
       Rational[1, 2], 1], 
      Hue[0.883846366509543, 
       Rational[1, 2], 1], 
      Hue[0.8885355640898871, 
       Rational[1, 2], 1], 
      Hue[0.8885355640898871, 
       Rational[1, 2], 1], 
      Hue[0.88912171378743, 
       Rational[1, 2], 1], 
      Hue[0.88912171378743, 
       Rational[1, 2], 1], 
      Hue[0.8897078634849731, 
       Rational[1, 2], 1], 
      Hue[0.8897078634849731, 
       Rational[1, 2], 1], 
      Hue[0.8908801628800591, 
       Rational[1, 2], 1], 
      Hue[0.8908801628800591, 
       Rational[1, 2], 1], 
      Hue[0.8932247616702311, 
       Rational[1, 2], 1], 
      Hue[0.8932247616702311, 
       Rational[1, 2], 1], 
      Hue[0.8979139592505752, 
       Rational[1, 2], 1], 
      Hue[0.8979139592505752, 
       Rational[1, 2], 1], 
      Hue[0.9072923544112633, 
       Rational[1, 2], 1], 
      Hue[0.9072923544112633, 
       Rational[1, 2], 1], 
      Hue[0.907927774154865, 
       Rational[1, 2], 1], 
      Hue[0.907927774154865, 
       Rational[1, 2], 1], 
      Hue[0.9085631938984667, 
       Rational[1, 2], 1], 
      Hue[0.9085631938984667, 
       Rational[1, 2], 1], 
      Hue[0.9098340333856701, 
       Rational[1, 2], 1], 
      Hue[0.9098340333856701, 
       Rational[1, 2], 1], 
      Hue[0.9123757123600771, 
       Rational[1, 2], 1], 
      Hue[0.9123757123600771, 
       Rational[1, 2], 1], 
      Hue[0.917459070308891, 
       Rational[1, 2], 1], 
      Hue[0.917459070308891, 
       Rational[1, 2], 1], 
      Hue[0.9276257862065186, 
       Rational[1, 2], 1], 
      Hue[0.9276257862065186, 
       Rational[1, 2], 1], 
      Hue[0.947959218001774, 
       Rational[1, 2], 1], 
      Hue[0.947959218001774, 
       Rational[1, 2], 1], 
      Hue[0.9485525561021597, 
       Rational[1, 2], 1], 
      Hue[0.9485525561021597, 
       Rational[1, 2], 1], 
      Hue[0.9491458942025454, 
       Rational[1, 2], 1], 
      Hue[0.9491458942025454, 
       Rational[1, 2], 1], 
      Hue[0.9503325704033169, 
       Rational[1, 2], 1], 
      Hue[0.9503325704033169, 
       Rational[1, 2], 1], 
      Hue[0.95270592280486, 
       Rational[1, 2], 1], 
      Hue[0.95270592280486, 
       Rational[1, 2], 1], 
      Hue[0.9574526276079461, 
       Rational[1, 2], 1], 
      Hue[0.9574526276079461, 
       Rational[1, 2], 1], 
      Hue[0.9669460372141183, 
       Rational[1, 2], 1], 
      Hue[0.9669460372141183, 
       Rational[1, 2], 1], 
      Hue[0.9675393753145042, 
       Rational[1, 2], 1], 
      Hue[0.9675393753145042, 
       Rational[1, 2], 1], 
      Hue[0.9681327134148898, 
       Rational[1, 2], 1], 
      Hue[0.9681327134148898, 
       Rational[1, 2], 1], 
      Hue[0.9693193896156613, 
       Rational[1, 2], 1], 
      Hue[0.9693193896156613, 
       Rational[1, 2], 1], 
      Hue[0.9716927420172045, 
       Rational[1, 2], 1], 
      Hue[0.9716927420172045, 
       Rational[1, 2], 1], 
      Hue[0.9764394468202906, 
       Rational[1, 2], 1], 
      Hue[0.9764394468202906, 
       Rational[1, 2], 1], 
      Hue[0.9859328564264628, 
       Rational[1, 2], 1], 
      Hue[0.9859328564264628, 
       Rational[1, 2], 1], 
      Hue[0.9865145634467241, 
       Rational[1, 2], 1], 
      Hue[0.9865145634467241, 
       Rational[1, 2], 1], 
      Hue[0.9870962704669854, 
       Rational[1, 2], 1], 
      Hue[0.9870962704669854, 
       Rational[1, 2], 1], 
      Hue[0.9882596845075081, 
       Rational[1, 2], 1], 
      Hue[0.9882596845075081, 
       Rational[1, 2], 1], 
      Hue[0.9905865125885537, 
       Rational[1, 2], 1], 
      Hue[0.9905865125885537, 
       Rational[1, 2], 1], 
      Hue[0.9911682196088151, 
       Rational[1, 2], 1], 
      Hue[0.9911682196088151, 
       Rational[1, 2], 1], 
      Hue[0.9917499266290765, 
       Rational[1, 2], 1], 
      Hue[0.9917499266290765, 
       Rational[1, 2], 1], 
      Hue[0.9929133406695991, 
       Rational[1, 2], 1], 
      Hue[0.9929133406695991, 
       Rational[1, 2], 1], 
      Hue[0.9952401687506446, 
       Rational[1, 2], 1], 
      Hue[0.9952401687506446, 
       Rational[1, 2], 1], 
      Hue[0.9958218757709059, 
       Rational[1, 2], 1], 
      Hue[0.9958218757709059, 
       Rational[1, 2], 1], 
      Hue[0.9964035827911673, 
       Rational[1, 2], 1], 
      Hue[0.9964035827911673, 
       Rational[1, 2], 1], 
      Hue[0.9975669968316899, 
       Rational[1, 2], 1], 
      Hue[0.9975669968316899, 
       Rational[1, 2], 1], 
      Hue[0.9981487038519513, 
       Rational[1, 2], 1], 
      Hue[0.9981487038519513, 
       Rational[1, 2], 1], 
      Hue[0.9987304108722127, 
       Rational[1, 2], 1], 
      Hue[0.9987304108722127, 
       Rational[1, 2], 1], 
      Hue[0.999312117892474, 
       Rational[1, 2], 1], 
      Hue[0.999312117892474, 
       Rational[1, 2], 1], 
      Hue[0.9998938249127354, 
       Rational[1, 2], 1], 
      Hue[0.9998938249127354, 
       Rational[1, 2], 1], 
      Hue[0.5004755319329968, 
       Rational[1, 2], 1], 
      Hue[0.5004755319329968, 
       Rational[1, 2], 1], 
      Hue[0.5010572389532582, 
       Rational[1, 2], 1], 
      Hue[0.5010572389532582, 
       Rational[1, 2], 1], 
      Hue[0.5016389459735195, 
       Rational[1, 2], 1], 
      Hue[0.5016389459735195, 
       Rational[1, 2], 1], 
      Hue[0.5022206529937807, 
       Rational[1, 2], 1], 
      Hue[0.5022206529937807, 
       Rational[1, 2], 1], 
      Hue[0.5045474810748263, 
       Rational[1, 2], 1], 
      Hue[0.5045474810748263, 
       Rational[1, 2], 1], 
      Hue[0.5051291880950876, 
       Rational[1, 2], 1], 
      Hue[0.5051291880950876, 
       Rational[1, 2], 1], 
      Hue[0.505710895115349, 
       Rational[1, 2], 1], 
      Hue[0.505710895115349, 
       Rational[1, 2], 1], 
      Hue[0.5068743091558716, 
       Rational[1, 2], 1], 
      Hue[0.5068743091558716, 
       Rational[1, 2], 1], 
      Hue[0.5092011372369171, 
       Rational[1, 2], 1], 
      Hue[0.5092011372369171, 
       Rational[1, 2], 1], 
      Hue[0.513854793399008, 
       Rational[1, 2], 1], 
      Hue[0.513854793399008, 
       Rational[1, 2], 1], 
      Hue[0.5231621057231897, 
       Rational[1, 2], 1], 
      Hue[0.5231621057231897, 
       Rational[1, 2], 1], 
      Hue[0.5237930827895098, 
       Rational[1, 2], 1], 
      Hue[0.5237930827895098, 
       Rational[1, 2], 1], 
      Hue[0.5244240598558298, 
       Rational[1, 2], 1], 
      Hue[0.5244240598558298, 
       Rational[1, 2], 1], 
      Hue[0.52568601398847, 
       Rational[1, 2], 1], 
      Hue[0.52568601398847, 
       Rational[1, 2], 1], 
      Hue[0.5282099222537503, 
       Rational[1, 2], 1], 
      Hue[0.5282099222537503, 
       Rational[1, 2], 1], 
      Hue[0.5332577387843109, 
       Rational[1, 2], 1], 
      Hue[0.5332577387843109, 
       Rational[1, 2], 1], 
      Hue[0.533888715850631, 
       Rational[1, 2], 1], 
      Hue[0.533888715850631, 
       Rational[1, 2], 1], 
      Hue[0.5345196929169511, 
       Rational[1, 2], 1], 
      Hue[0.5345196929169511, 
       Rational[1, 2], 1], 
      Hue[0.5357816470495913, 
       Rational[1, 2], 1], 
      Hue[0.5357816470495913, 
       Rational[1, 2], 1], 
      Hue[0.5383055553148717, 
       Rational[1, 2], 1], 
      Hue[0.5383055553148717, 
       Rational[1, 2], 1], 
      Hue[0.5433533718454322, 
       Rational[1, 2], 1], 
      Hue[0.5433533718454322, 
       Rational[1, 2], 1], 
      Hue[0.5439843489117524, 
       Rational[1, 2], 1], 
      Hue[0.5439843489117524, 
       Rational[1, 2], 1], 
      Hue[0.5446153259780724, 
       Rational[1, 2], 1], 
      Hue[0.5446153259780724, 
       Rational[1, 2], 1], 
      Hue[0.5458772801107126, 
       Rational[1, 2], 1], 
      Hue[0.5458772801107126, 
       Rational[1, 2], 1], 
      Hue[0.5484011883759928, 
       Rational[1, 2], 1], 
      Hue[0.5484011883759928, 
       Rational[1, 2], 1], 
      Hue[0.5490321654423129, 
       Rational[1, 2], 1], 
      Hue[0.5490321654423129, 
       Rational[1, 2], 1], 
      Hue[0.549663142508633, 
       Rational[1, 2], 1], 
      Hue[0.549663142508633, 
       Rational[1, 2], 1], 
      Hue[0.5509250966412733, 
       Rational[1, 2], 1], 
      Hue[0.5509250966412733, 
       Rational[1, 2], 1], 
      Hue[0.5534490049065535, 
       Rational[1, 2], 1], 
      Hue[0.5534490049065535, 
       Rational[1, 2], 1], 
      Hue[0.5540799819728736, 
       Rational[1, 2], 1], 
      Hue[0.5540799819728736, 
       Rational[1, 2], 1], 
      Hue[0.5547109590391937, 
       Rational[1, 2], 1], 
      Hue[0.5547109590391937, 
       Rational[1, 2], 1], 
      Hue[0.5559729131718338, 
       Rational[1, 2], 1], 
      Hue[0.5559729131718338, 
       Rational[1, 2], 1], 
      Hue[0.5584968214371142, 
       Rational[1, 2], 1], 
      Hue[0.5584968214371142, 
       Rational[1, 2], 1], 
      Hue[0.5591277985034342, 
       Rational[1, 2], 1], 
      Hue[0.5591277985034342, 
       Rational[1, 2], 1], 
      Hue[0.5597587755697544, 
       Rational[1, 2], 1], 
      Hue[0.5597587755697544, 
       Rational[1, 2], 1], 
      Hue[0.5610207297023944, 
       Rational[1, 2], 1], 
      Hue[0.5610207297023944, 
       Rational[1, 2], 1], 
      Hue[0.5635446379676747, 
       Rational[1, 2], 1], 
      Hue[0.5635446379676747, 
       Rational[1, 2], 1], 
      Hue[0.5641335333907789, 
       Rational[1, 2], 1], 
      Hue[0.5641335333907789, 
       Rational[1, 2], 1], 
      Hue[0.564722428813883, 
       Rational[1, 2], 1], 
      Hue[0.564722428813883, 
       Rational[1, 2], 1], 
      Hue[0.5659002196600912, 
       Rational[1, 2], 1], 
      Hue[0.5659002196600912, 
       Rational[1, 2], 1], 
      Hue[0.5664891150831952, 
       Rational[1, 2], 1], 
      Hue[0.5664891150831952, 
       Rational[1, 2], 1], 
      Hue[0.5670780105062995, 
       Rational[1, 2], 1], 
      Hue[0.5670780105062995, 
       Rational[1, 2], 1], 
      Hue[0.5682558013525076, 
       Rational[1, 2], 1], 
      Hue[0.5682558013525076, 
       Rational[1, 2], 1], 
      Hue[0.5688446967756118, 
       Rational[1, 2], 1], 
      Hue[0.5688446967756118, 
       Rational[1, 2], 1], 
      Hue[0.5694335921987159, 
       Rational[1, 2], 1], 
      Hue[0.5694335921987159, 
       Rational[1, 2], 1], 
      Hue[0.5706113830449241, 
       Rational[1, 2], 1], 
      Hue[0.5706113830449241, 
       Rational[1, 2], 1], 
      Hue[0.5712002784680281, 
       Rational[1, 2], 1], 
      Hue[0.5712002784680281, 
       Rational[1, 2], 1], 
      Hue[0.5717891738911324, 
       Rational[1, 2], 1], 
      Hue[0.5717891738911324, 
       Rational[1, 2], 1], 
      Hue[0.5729669647373405, 
       Rational[1, 2], 1], 
      Hue[0.5729669647373405, 
       Rational[1, 2], 1], 
      Hue[0.5735558601604447, 
       Rational[1, 2], 1], 
      Hue[0.5735558601604447, 
       Rational[1, 2], 1], 
      Hue[0.5741447555835488, 
       Rational[1, 2], 1], 
      Hue[0.5741447555835488, 
       Rational[1, 2], 1], 
      Hue[0.5753225464297571, 
       Rational[1, 2], 1], 
      Hue[0.5753225464297571, 
       Rational[1, 2], 1], 
      Hue[0.575911441852861, 
       Rational[1, 2], 1], 
      Hue[0.575911441852861, 
       Rational[1, 2], 1], 
      Hue[0.5765003372759653, 
       Rational[1, 2], 1], 
      Hue[0.5765003372759653, 
       Rational[1, 2], 1], 
      Hue[0.5776781281221735, 
       Rational[1, 2], 1], 
      Hue[0.5776781281221735, 
       Rational[1, 2], 1], 
      Hue[0.5782670235452776, 
       Rational[1, 2], 1], 
      Hue[0.5782670235452776, 
       Rational[1, 2], 1], 
      Hue[0.5788559189683817, 
       Rational[1, 2], 1], 
      Hue[0.5788559189683817, 
       Rational[1, 2], 1], 
      Hue[0.5794448143914859, 
       Rational[1, 2], 1], 
      Hue[0.5794448143914859, 
       Rational[1, 2], 1], 
      Hue[0.5800337098145899, 
       Rational[1, 2], 1], 
      Hue[0.5800337098145899, 
       Rational[1, 2], 1], 
      Hue[0.580622605237694, 
       Rational[1, 2], 1], 
      Hue[0.580622605237694, 
       Rational[1, 2], 1], 
      Hue[0.5812115006607981, 
       Rational[1, 2], 1], 
      Hue[0.5812115006607981, 
       Rational[1, 2], 1], 
      Hue[0.5818003960839022, 
       Rational[1, 2], 1], 
      Hue[0.5818003960839022, 
       Rational[1, 2], 1], 
      Hue[0.5823892915070064, 
       Rational[1, 2], 1], 
      Hue[0.5823892915070064, 
       Rational[1, 2], 1], 
      Hue[0.5829781869301105, 
       Rational[1, 2], 1], 
      Hue[0.5829781869301105, 
       Rational[1, 2], 1], 
      Hue[0.5835670823532146, 
       Rational[1, 2], 1], 
      Hue[0.5835670823532146, 
       Rational[1, 2], 1], 
      Hue[0.5841559777763188, 
       Rational[1, 2], 1], 
      Hue[0.5841559777763188, 
       Rational[1, 2], 1], 
      Hue[0.5847448731994228, 
       Rational[1, 2], 1], 
      Hue[0.5847448731994228, 
       Rational[1, 2], 1], 
      Hue[0.585333768622527, 
       Rational[1, 2], 1], 
      Hue[0.585333768622527, 
       Rational[1, 2], 1], 
      Hue[0.585922664045631, 
       Rational[1, 2], 1], 
      Hue[0.585922664045631, 
       Rational[1, 2], 1], 
      Hue[0.5865115594687351, 
       Rational[1, 2], 1], 
      Hue[0.5865115594687351, 
       Rational[1, 2], 1], 
      Hue[0.5871004548918393, 
       Rational[1, 2], 1], 
      Hue[0.5871004548918393, 
       Rational[1, 2], 1], 
      Hue[0.5876893503149434, 
       Rational[1, 2], 1], 
      Hue[0.5876893503149434, 
       Rational[1, 2], 1], 
      Hue[0.5882782457380475, 
       Rational[1, 2], 1], 
      Hue[0.5882782457380475, 
       Rational[1, 2], 1], 
      Hue[0.5888671411611517, 
       Rational[1, 2], 1], 
      Hue[0.5888671411611517, 
       Rational[1, 2], 1], 
      Hue[0.5894560365842557, 
       Rational[1, 2], 1], 
      Hue[0.5894560365842557, 
       Rational[1, 2], 1], 
      Hue[0.5900449320073597, 
       Rational[1, 2], 1], 
      Hue[0.5900449320073597, 
       Rational[1, 2], 1], 
      Hue[0.5906338274304639, 
       Rational[1, 2], 1], 
      Hue[0.5906338274304639, 
       Rational[1, 2], 1], 
      Hue[0.591222722853568, 
       Rational[1, 2], 1], 
      Hue[0.591222722853568, 
       Rational[1, 2], 1], 
      Hue[0.5918116182766722, 
       Rational[1, 2], 1], 
      Hue[0.5918116182766722, 
       Rational[1, 2], 1], 
      Hue[0.5924005136997763, 
       Rational[1, 2], 1], 
      Hue[0.5924005136997763, 
       Rational[1, 2], 1], 
      Hue[0.5929894091228805, 
       Rational[1, 2], 1], 
      Hue[0.5929894091228805, 
       Rational[1, 2], 1], 
      Hue[0.5935783045459846, 
       Rational[1, 2], 1], 
      Hue[0.5935783045459846, 
       Rational[1, 2], 1], 
      Hue[0.5941671999690887, 
       Rational[1, 2], 1], 
      Hue[0.5941671999690887, 
       Rational[1, 2], 1], 
      Hue[0.5947560953921927, 
       Rational[1, 2], 1], 
      Hue[0.5947560953921927, 
       Rational[1, 2], 1], 
      Hue[0.5953449908152968, 
       Rational[1, 2], 1], 
      Hue[0.5953449908152968, 
       Rational[1, 2], 1], 
      Hue[0.5965227816615051, 
       Rational[1, 2], 1], 
      Hue[0.5965227816615051, 
       Rational[1, 2], 1], 
      Hue[0.5971116770846092, 
       Rational[1, 2], 1], 
      Hue[0.5971116770846092, 
       Rational[1, 2], 1], 
      Hue[0.5977005725077134, 
       Rational[1, 2], 1], 
      Hue[0.5977005725077134, 
       Rational[1, 2], 1], 
      Hue[0.5988783633539216, 
       Rational[1, 2], 1], 
      Hue[0.5988783633539216, 
       Rational[1, 2], 1], 
      Hue[0.5994672587770258, 
       Rational[1, 2], 1], 
      Hue[0.5994672587770258, 
       Rational[1, 2], 1], 
      Hue[0.6000561542001298, 
       Rational[1, 2], 1], 
      Hue[0.6000561542001298, 
       Rational[1, 2], 1], 
      Hue[0.601233945046338, 
       Rational[1, 2], 1], 
      Hue[0.601233945046338, 
       Rational[1, 2], 1], 
      Hue[0.6018721105155009, 
       Rational[1, 2], 1], 
      Hue[0.6018721105155009, 
       Rational[1, 2], 1], 
      Hue[0.6025102759846637, 
       Rational[1, 2], 1], 
      Hue[0.6025102759846637, 
       Rational[1, 2], 1], 
      Hue[0.6037866069229894, 
       Rational[1, 2], 1], 
      Hue[0.6037866069229894, 
       Rational[1, 2], 1], 
      Hue[0.6063392687996407, 
       Rational[1, 2], 1], 
      Hue[0.6063392687996407, 
       Rational[1, 2], 1], 
      Hue[0.6069774342688035, 
       Rational[1, 2], 1], 
      Hue[0.6069774342688035, 
       Rational[1, 2], 1], 
      Hue[0.6076155997379664, 
       Rational[1, 2], 1], 
      Hue[0.6076155997379664, 
       Rational[1, 2], 1], 
      Hue[0.6088919306762921, 
       Rational[1, 2], 1], 
      Hue[0.6088919306762921, 
       Rational[1, 2], 1], 
      Hue[0.6114445925529435, 
       Rational[1, 2], 1], 
      Hue[0.6114445925529435, 
       Rational[1, 2], 1], 
      Hue[0.6120827580221063, 
       Rational[1, 2], 1], 
      Hue[0.6120827580221063, 
       Rational[1, 2], 1], 
      Hue[0.612720923491269, 
       Rational[1, 2], 1], 
      Hue[0.612720923491269, 
       Rational[1, 2], 1], 
      Hue[0.6139972544295947, 
       Rational[1, 2], 1], 
      Hue[0.6139972544295947, 
       Rational[1, 2], 1], 
      Hue[0.6165499163062461, 
       Rational[1, 2], 1], 
      Hue[0.6165499163062461, 
       Rational[1, 2], 1], 
      Hue[0.6216552400595488, 
       Rational[1, 2], 1], 
      Hue[0.6216552400595488, 
       Rational[1, 2], 1], 
      Hue[0.6222934055287116, 
       Rational[1, 2], 1], 
      Hue[0.6222934055287116, 
       Rational[1, 2], 1], 
      Hue[0.6229315709978744, 
       Rational[1, 2], 1], 
      Hue[0.6229315709978744, 
       Rational[1, 2], 1], 
      Hue[0.6242079019362001, 
       Rational[1, 2], 1], 
      Hue[0.6242079019362001, 
       Rational[1, 2], 1], 
      Hue[0.6267605638128514, 
       Rational[1, 2], 1], 
      Hue[0.6267605638128514, 
       Rational[1, 2], 1], 
      Hue[0.6318658875661541, 
       Rational[1, 2], 1], 
      Hue[0.6318658875661541, 
       Rational[1, 2], 1], 
      Hue[0.6420765350727594, 
       Rational[1, 2], 1], 
      Hue[0.6420765350727594, 
       Rational[1, 2], 1], 
      Hue[0.6427030694617979, 
       Rational[1, 2], 1], 
      Hue[0.6427030694617979, 
       Rational[1, 2], 1], 
      Hue[0.6433296038508364, 
       Rational[1, 2], 1], 
      Hue[0.6433296038508364, 
       Rational[1, 2], 1], 
      Hue[0.6445826726289132, 
       Rational[1, 2], 1], 
      Hue[0.6445826726289132, 
       Rational[1, 2], 1], 
      Hue[0.6470888101850669, 
       Rational[1, 2], 1], 
      Hue[0.6470888101850669, 
       Rational[1, 2], 1], 
      Hue[0.6521010852973743, 
       Rational[1, 2], 1], 
      Hue[0.6521010852973743, 
       Rational[1, 2], 1], 
      Hue[0.6621256355219893, 
       Rational[1, 2], 1], 
      Hue[0.6621256355219893, 
       Rational[1, 2], 1], 
      Hue[0.6821747359712191, 
       Rational[1, 2], 1], 
      Hue[0.6821747359712191, 
       Rational[1, 2], 1], 
      Hue[0.6827591887170417, 
       Rational[1, 2], 1], 
      Hue[0.6827591887170417, 
       Rational[1, 2], 1], 
      Hue[0.6833436414628641, 
       Rational[1, 2], 1], 
      Hue[0.6833436414628641, 
       Rational[1, 2], 1], 
      Hue[0.684512546954509, 
       Rational[1, 2], 1], 
      Hue[0.684512546954509, 
       Rational[1, 2], 1], 
      Hue[0.6868503579377989, 
       Rational[1, 2], 1], 
      Hue[0.6868503579377989, 
       Rational[1, 2], 1], 
      Hue[0.6915259799043787, 
       Rational[1, 2], 1], 
      Hue[0.6915259799043787, 
       Rational[1, 2], 1], 
      Hue[0.700877223837538, 
       Rational[1, 2], 1], 
      Hue[0.700877223837538, 
       Rational[1, 2], 1], 
      Hue[0.7014616765833603, 
       Rational[1, 2], 1], 
      Hue[0.7014616765833603, 
       Rational[1, 2], 1], 
      Hue[0.7020461293291829, 
       Rational[1, 2], 1], 
      Hue[0.7020461293291829, 
       Rational[1, 2], 1], 
      Hue[0.7032150348208277, 
       Rational[1, 2], 1], 
      Hue[0.7032150348208277, 
       Rational[1, 2], 1], 
      Hue[0.7055528458041176, 
       Rational[1, 2], 1], 
      Hue[0.7055528458041176, 
       Rational[1, 2], 1], 
      Hue[0.7102284677706974, 
       Rational[1, 2], 1], 
      Hue[0.7102284677706974, 
       Rational[1, 2], 1], 
      Hue[0.7195797117038569, 
       Rational[1, 2], 1], 
      Hue[0.7195797117038569, 
       Rational[1, 2], 1], 
      Hue[0.7202134344957382, 
       Rational[1, 2], 1], 
      Hue[0.7202134344957382, 
       Rational[1, 2], 1], 
      Hue[0.7208471572876194, 
       Rational[1, 2], 1], 
      Hue[0.7208471572876194, 
       Rational[1, 2], 1], 
      Hue[0.7221146028713816, 
       Rational[1, 2], 1], 
      Hue[0.7221146028713816, 
       Rational[1, 2], 1], 
      Hue[0.7246494940389063, 
       Rational[1, 2], 1], 
      Hue[0.7246494940389063, 
       Rational[1, 2], 1], 
      Hue[0.729719276373956, 
       Rational[1, 2], 1], 
      Hue[0.729719276373956, 
       Rational[1, 2], 1], 
      Hue[0.7303529991658371, 
       Rational[1, 2], 1], 
      Hue[0.7303529991658371, 
       Rational[1, 2], 1], 
      Hue[0.7309867219577184, 
       Rational[1, 2], 1], 
      Hue[0.7309867219577184, 
       Rational[1, 2], 1], 
      Hue[0.7322541675414806, 
       Rational[1, 2], 1], 
      Hue[0.7322541675414806, 
       Rational[1, 2], 1], 
      Hue[0.7347890587090053, 
       Rational[1, 2], 1], 
      Hue[0.7347890587090053, 
       Rational[1, 2], 1], 
      Hue[0.7398588410440549, 
       Rational[1, 2], 1], 
      Hue[0.7398588410440549, 
       Rational[1, 2], 1], 
      Hue[0.7404925638359361, 
       Rational[1, 2], 1], 
      Hue[0.7404925638359361, 
       Rational[1, 2], 1], 
      Hue[0.7411262866278173, 
       Rational[1, 2], 1], 
      Hue[0.7411262866278173, 
       Rational[1, 2], 1], 
      Hue[0.7423937322115797, 
       Rational[1, 2], 1], 
      Hue[0.7423937322115797, 
       Rational[1, 2], 1], 
      Hue[0.7449286233791044, 
       Rational[1, 2], 1], 
      Hue[0.7449286233791044, 
       Rational[1, 2], 1], 
      Hue[0.7455623461709856, 
       Rational[1, 2], 1], 
      Hue[0.7455623461709856, 
       Rational[1, 2], 1], 
      Hue[0.7461960689628668, 
       Rational[1, 2], 1], 
      Hue[0.7461960689628668, 
       Rational[1, 2], 1], 
      Hue[0.7474635145466291, 
       Rational[1, 2], 1], 
      Hue[0.7474635145466291, 
       Rational[1, 2], 1], 
      Hue[0.7480972373385104, 
       Rational[1, 2], 1], 
      Hue[0.7480972373385104, 
       Rational[1, 2], 1], 
      Hue[0.7487309601303914, 
       Rational[1, 2], 1], 
      Hue[0.7487309601303914, 
       Rational[1, 2], 1], 
      Hue[0.7493646829222727, 
       Rational[1, 2], 1], 
      Hue[0.7493646829222727, 
       Rational[1, 2], 1], 
      Hue[0.7499984057141539, 
       Rational[1, 2], 1], 
      Hue[0.7499984057141539, 
       Rational[1, 2], 1], 
      Hue[0.2506321285060351, 
       Rational[1, 2], 1], 
      Hue[0.2506321285060351, 
       Rational[1, 2], 1], 
      Hue[0.2512658512979163, 
       Rational[1, 2], 1], 
      Hue[0.2512658512979163, 
       Rational[1, 2], 1], 
      Hue[0.25253329688167864`, 
       Rational[1, 2], 1], 
      Hue[0.25253329688167864`, 
       Rational[1, 2], 1], 
      Hue[0.25506818804920334`, 
       Rational[1, 2], 1], 
      Hue[0.25506818804920334`, 
       Rational[1, 2], 1], 
      Hue[0.26013797038425285`, 
       Rational[1, 2], 1], 
      Hue[0.26013797038425285`, 
       Rational[1, 2], 1], 
      Hue[0.26072961153291807`, 
       Rational[1, 2], 1], 
      Hue[0.26072961153291807`, 
       Rational[1, 2], 1], 
      Hue[0.2613212526815833, 
       Rational[1, 2], 1], 
      Hue[0.2613212526815833, 
       Rational[1, 2], 1], 
      Hue[0.26250453497891374`, 
       Rational[1, 2], 1], 
      Hue[0.26250453497891374`, 
       Rational[1, 2], 1], 
      Hue[0.26487109957357463`, 
       Rational[1, 2], 1], 
      Hue[0.26487109957357463`, 
       Rational[1, 2], 1], 
      Hue[0.2696042287628964, 
       Rational[1, 2], 1], 
      Hue[0.2696042287628964, 
       Rational[1, 2], 1], 
      Hue[0.2701958699115617, 
       Rational[1, 2], 1], 
      Hue[0.2701958699115617, 
       Rational[1, 2], 1], 
      Hue[0.2707875110602269, 
       Rational[1, 2], 1], 
      Hue[0.2707875110602269, 
       Rational[1, 2], 1], 
      Hue[0.27197079335755736`, 
       Rational[1, 2], 1], 
      Hue[0.27197079335755736`, 
       Rational[1, 2], 1], 
      Hue[0.27433735795221825`, 
       Rational[1, 2], 1], 
      Hue[0.27433735795221825`, 
       Rational[1, 2], 1], 
      Hue[0.27907048714154004`, 
       Rational[1, 2], 1], 
      Hue[0.27907048714154004`, 
       Rational[1, 2], 1], 
      Hue[0.2796621282902053, 
       Rational[1, 2], 1], 
      Hue[0.2796621282902053, 
       Rational[1, 2], 1], 
      Hue[0.2802537694388705, 
       Rational[1, 2], 1], 
      Hue[0.2802537694388705, 
       Rational[1, 2], 1], 
      Hue[0.281437051736201, 
       Rational[1, 2], 1], 
      Hue[0.281437051736201, 
       Rational[1, 2], 1], 
      Hue[0.2838036163308618, 
       Rational[1, 2], 1], 
      Hue[0.2838036163308618, 
       Rational[1, 2], 1], 
      Hue[0.28439525747952704`, 
       Rational[1, 2], 1], 
      Hue[0.28439525747952704`, 
       Rational[1, 2], 1], 
      Hue[0.28498689862819226`, 
       Rational[1, 2], 1], 
      Hue[0.28498689862819226`, 
       Rational[1, 2], 1], 
      Hue[0.2861701809255227, 
       Rational[1, 2], 1], 
      Hue[0.2861701809255227, 
       Rational[1, 2], 1], 
      Hue[0.2885367455201836, 
       Rational[1, 2], 1], 
      Hue[0.2885367455201836, 
       Rational[1, 2], 1], 
      Hue[0.2891283866688488, 
       Rational[1, 2], 1], 
      Hue[0.2891283866688488, 
       Rational[1, 2], 1], 
      Hue[0.2897200278175141, 
       Rational[1, 2], 1], 
      Hue[0.2897200278175141, 
       Rational[1, 2], 1], 
      Hue[0.2909033101148445, 
       Rational[1, 2], 1], 
      Hue[0.2909033101148445, 
       Rational[1, 2], 1], 
      Hue[0.2932698747095054, 
       Rational[1, 2], 1], 
      Hue[0.2932698747095054, 
       Rational[1, 2], 1], 
      Hue[0.29386151585817066`, 
       Rational[1, 2], 1], 
      Hue[0.29386151585817066`, 
       Rational[1, 2], 1], 
      Hue[0.29445315700683583`, 
       Rational[1, 2], 1], 
      Hue[0.29445315700683583`, 
       Rational[1, 2], 1], 
      Hue[0.29563643930416633`, 
       Rational[1, 2], 1], 
      Hue[0.29563643930416633`, 
       Rational[1, 2], 1], 
      Hue[0.2980030038988271, 
       Rational[1, 2], 1], 
      Hue[0.2980030038988271, 
       Rational[1, 2], 1], 
      Hue[0.298583013967368, 
       Rational[1, 2], 1], 
      Hue[0.298583013967368, 
       Rational[1, 2], 1], 
      Hue[0.29916302403590883`, 
       Rational[1, 2], 1], 
      Hue[0.29916302403590883`, 
       Rational[1, 2], 1], 
      Hue[0.30032304417299044`, 
       Rational[1, 2], 1], 
      Hue[0.30032304417299044`, 
       Rational[1, 2], 1], 
      Hue[0.30090305424153124`, 
       Rational[1, 2], 1], 
      Hue[0.30090305424153124`, 
       Rational[1, 2], 1], 
      Hue[0.30148306431007205`, 
       Rational[1, 2], 1], 
      Hue[0.30148306431007205`, 
       Rational[1, 2], 1], 
      Hue[0.3026430844471537, 
       Rational[1, 2], 1], 
      Hue[0.3026430844471537, 
       Rational[1, 2], 1], 
      Hue[0.3032230945156945, 
       Rational[1, 2], 1], 
      Hue[0.3032230945156945, 
       Rational[1, 2], 1], 
      Hue[0.3038031045842353, 
       Rational[1, 2], 1], 
      Hue[0.3038031045842353, 
       Rational[1, 2], 1], 
      Hue[0.3049631247213169, 
       Rational[1, 2], 1], 
      Hue[0.3049631247213169, 
       Rational[1, 2], 1], 
      Hue[0.30728316499548014`, 
       Rational[1, 2], 1], 
      Hue[0.30728316499548014`, 
       Rational[1, 2], 1], 
      Hue[0.30786317506402094`, 
       Rational[1, 2], 1], 
      Hue[0.30786317506402094`, 
       Rational[1, 2], 1], 
      Hue[0.3084431851325618, 
       Rational[1, 2], 1], 
      Hue[0.3084431851325618, 
       Rational[1, 2], 1], 
      Hue[0.3096032052696434, 
       Rational[1, 2], 1], 
      Hue[0.3096032052696434, 
       Rational[1, 2], 1], 
      Hue[0.31018321533818427`, 
       Rational[1, 2], 1], 
      Hue[0.31018321533818427`, 
       Rational[1, 2], 1], 
      Hue[0.310763225406725, 
       Rational[1, 2], 1], 
      Hue[0.310763225406725, 
       Rational[1, 2], 1], 
      Hue[0.3119232455438066, 
       Rational[1, 2], 1], 
      Hue[0.3119232455438066, 
       Rational[1, 2], 1], 
      Hue[0.31250325561234743`, 
       Rational[1, 2], 1], 
      Hue[0.31250325561234743`, 
       Rational[1, 2], 1], 
      Hue[0.31308326568088823`, 
       Rational[1, 2], 1], 
      Hue[0.31308326568088823`, 
       Rational[1, 2], 1], 
      Hue[0.3142432858179699, 
       Rational[1, 2], 1], 
      Hue[0.3142432858179699, 
       Rational[1, 2], 1], 
      Hue[0.3148232958865107, 
       Rational[1, 2], 1], 
      Hue[0.3148232958865107, 
       Rational[1, 2], 1], 
      Hue[0.3154033059550515, 
       Rational[1, 2], 1], 
      Hue[0.3154033059550515, 
       Rational[1, 2], 1], 
      Hue[0.3165633260921333, 
       Rational[1, 2], 1], 
      Hue[0.3165633260921333, 
       Rational[1, 2], 1], 
      Hue[0.31714333616067425`, 
       Rational[1, 2], 1], 
      Hue[0.31714333616067425`, 
       Rational[1, 2], 1], 
      Hue[0.31772334622921505`, 
       Rational[1, 2], 1], 
      Hue[0.31772334622921505`, 
       Rational[1, 2], 1], 
      Hue[0.31830335629775586`, 
       Rational[1, 2], 1], 
      Hue[0.31830335629775586`, 
       Rational[1, 2], 1], 
      Hue[0.31888336636629666`, 
       Rational[1, 2], 1], 
      Hue[0.31888336636629666`, 
       Rational[1, 2], 1], 
      Hue[0.31946337643483746`, 
       Rational[1, 2], 1], 
      Hue[0.31946337643483746`, 
       Rational[1, 2], 1], 
      Hue[0.32004338650337827`, 
       Rational[1, 2], 1], 
      Hue[0.32004338650337827`, 
       Rational[1, 2], 1], 
      Hue[0.32062339657191913`, 
       Rational[1, 2], 1], 
      Hue[0.32062339657191913`, 
       Rational[1, 2], 1], 
      Hue[0.32120340664045993`, 
       Rational[1, 2], 1], 
      Hue[0.32120340664045993`, 
       Rational[1, 2], 1], 
      Hue[0.32178341670900074`, 
       Rational[1, 2], 1], 
      Hue[0.32178341670900074`, 
       Rational[1, 2], 1], 
      Hue[0.32236342677754154`, 
       Rational[1, 2], 1], 
      Hue[0.32236342677754154`, 
       Rational[1, 2], 1], 
      Hue[0.3229434368460823, 
       Rational[1, 2], 1], 
      Hue[0.3229434368460823, 
       Rational[1, 2], 1], 
      Hue[0.32352344691462315`, 
       Rational[1, 2], 1], 
      Hue[0.32352344691462315`, 
       Rational[1, 2], 1], 
      Hue[0.32410345698316395`, 
       Rational[1, 2], 1], 
      Hue[0.32410345698316395`, 
       Rational[1, 2], 1], 
      Hue[0.32468346705170475`, 
       Rational[1, 2], 1], 
      Hue[0.32468346705170475`, 
       Rational[1, 2], 1], 
      Hue[0.32526347712024556`, 
       Rational[1, 2], 1], 
      Hue[0.32526347712024556`, 
       Rational[1, 2], 1], 
      Hue[0.3258434871887864, 
       Rational[1, 2], 1], 
      Hue[0.3258434871887864, 
       Rational[1, 2], 1], 
      Hue[0.3264234972573272, 
       Rational[1, 2], 1], 
      Hue[0.3264234972573272, 
       Rational[1, 2], 1], 
      Hue[0.327003507325868, 
       Rational[1, 2], 1], 
      Hue[0.327003507325868, 
       Rational[1, 2], 1], 
      Hue[0.32758351739440883`, 
       Rational[1, 2], 1], 
      Hue[0.32758351739440883`, 
       Rational[1, 2], 1], 
      Hue[0.32816352746294963`, 
       Rational[1, 2], 1], 
      Hue[0.32816352746294963`, 
       Rational[1, 2], 1], 
      Hue[0.32874353753149044`, 
       Rational[1, 2], 1], 
      Hue[0.32874353753149044`, 
       Rational[1, 2], 1], 
      Hue[0.32932354760003124`, 
       Rational[1, 2], 1], 
      Hue[0.32932354760003124`, 
       Rational[1, 2], 1], 
      Hue[0.32990355766857205`, 
       Rational[1, 2], 1], 
      Hue[0.32990355766857205`, 
       Rational[1, 2], 1], 
      Hue[0.33048356773711285`, 
       Rational[1, 2], 1], 
      Hue[0.33048356773711285`, 
       Rational[1, 2], 1], 
      Hue[0.3310635778056537, 
       Rational[1, 2], 1], 
      Hue[0.3310635778056537, 
       Rational[1, 2], 1], 
      Hue[0.3316435878741945, 
       Rational[1, 2], 1], 
      Hue[0.3316435878741945, 
       Rational[1, 2], 1], 
      Hue[0.3322235979427353, 
       Rational[1, 2], 1], 
      Hue[0.3322235979427353, 
       Rational[1, 2], 1], 
      Hue[0.3328036080112761, 
       Rational[1, 2], 1], 
      Hue[0.3328036080112761, 
       Rational[1, 2], 1], 
      Hue[0.3333836180798169, 
       Rational[1, 2], 1], 
      Hue[0.3333836180798169, 
       Rational[1, 2], 1], 
      Hue[0.3339636281483577, 
       Rational[1, 2], 1], 
      Hue[0.3339636281483577, 
       Rational[1, 2], 1], 
      Hue[0.3351236482854395, 
       Rational[1, 2], 1], 
      Hue[0.3351236482854395, 
       Rational[1, 2], 1], 
      Hue[0.33575292840003906`, 
       Rational[1, 2], 1], 
      Hue[0.33575292840003906`, 
       Rational[1, 2], 1], 
      Hue[0.33638220851463846`, 
       Rational[1, 2], 1], 
      Hue[0.33638220851463846`, 
       Rational[1, 2], 1], 
      Hue[0.3376407687438376, 
       Rational[1, 2], 1], 
      Hue[0.3376407687438376, 
       Rational[1, 2], 1], 
      Hue[0.33827004885843714`, 
       Rational[1, 2], 1], 
      Hue[0.33827004885843714`, 
       Rational[1, 2], 1], 
      Hue[0.3388993289730367, 
       Rational[1, 2], 1], 
      Hue[0.3388993289730367, 
       Rational[1, 2], 1], 
      Hue[0.34015788920223583`, 
       Rational[1, 2], 1], 
      Hue[0.34015788920223583`, 
       Rational[1, 2], 1], 
      Hue[0.3407871693168354, 
       Rational[1, 2], 1], 
      Hue[0.3407871693168354, 
       Rational[1, 2], 1], 
      Hue[0.34141644943143495`, 
       Rational[1, 2], 1], 
      Hue[0.34141644943143495`, 
       Rational[1, 2], 1], 
      Hue[0.34267500966063397`, 
       Rational[1, 2], 1], 
      Hue[0.34267500966063397`, 
       Rational[1, 2], 1], 
      Hue[0.34519213011903205`, 
       Rational[1, 2], 1], 
      Hue[0.34519213011903205`, 
       Rational[1, 2], 1], 
      Hue[0.3458214102336316, 
       Rational[1, 2], 1], 
      Hue[0.3458214102336316, 
       Rational[1, 2], 1], 
      Hue[0.34645069034823117`, 
       Rational[1, 2], 1], 
      Hue[0.34645069034823117`, 
       Rational[1, 2], 1], 
      Hue[0.3477092505774303, 
       Rational[1, 2], 1], 
      Hue[0.3477092505774303, 
       Rational[1, 2], 1], 
      Hue[0.3502263710358284, 
       Rational[1, 2], 1], 
      Hue[0.3502263710358284, 
       Rational[1, 2], 1], 
      Hue[0.35526061195262476`, 
       Rational[1, 2], 1], 
      Hue[0.35526061195262476`, 
       Rational[1, 2], 1], 
      Hue[0.35588989206722427`, 
       Rational[1, 2], 1], 
      Hue[0.35588989206722427`, 
       Rational[1, 2], 1], 
      Hue[0.3565191721818239, 
       Rational[1, 2], 1], 
      Hue[0.3565191721818239, 
       Rational[1, 2], 1], 
      Hue[0.35777773241102284`, 
       Rational[1, 2], 1], 
      Hue[0.35777773241102284`, 
       Rational[1, 2], 1], 
      Hue[0.360294852869421, 
       Rational[1, 2], 1], 
      Hue[0.360294852869421, 
       Rational[1, 2], 1], 
      Hue[0.3653290937862173, 
       Rational[1, 2], 1], 
      Hue[0.3653290937862173, 
       Rational[1, 2], 1], 
      Hue[0.37539757561980985`, 
       Rational[1, 2], 1], 
      Hue[0.37539757561980985`, 
       Rational[1, 2], 1], 
      Hue[0.3759847740911934, 
       Rational[1, 2], 1], 
      Hue[0.3759847740911934, 
       Rational[1, 2], 1], 
      Hue[0.376571972562577, 
       Rational[1, 2], 1], 
      Hue[0.376571972562577, 
       Rational[1, 2], 1], 
      Hue[0.3777463695053441, 
       Rational[1, 2], 1], 
      Hue[0.3777463695053441, 
       Rational[1, 2], 1], 
      Hue[0.3800951633908784, 
       Rational[1, 2], 1], 
      Hue[0.3800951633908784, 
       Rational[1, 2], 1], 
      Hue[0.38479275116194706`, 
       Rational[1, 2], 1], 
      Hue[0.38479275116194706`, 
       Rational[1, 2], 1], 
      Hue[0.3941879267040842, 
       Rational[1, 2], 1], 
      Hue[0.3941879267040842, 
       Rational[1, 2], 1], 
      Hue[0.41297827778835855`, 
       Rational[1, 2], 1], 
      Hue[0.41297827778835855`, 
       Rational[1, 2], 1], 
      Hue[0.41361474630580086`, 
       Rational[1, 2], 1], 
      Hue[0.41361474630580086`, 
       Rational[1, 2], 1], 
      Hue[0.4142512148232432, 
       Rational[1, 2], 1], 
      Hue[0.4142512148232432, 
       Rational[1, 2], 1], 
      Hue[0.41552415185812763`, 
       Rational[1, 2], 1], 
      Hue[0.41552415185812763`, 
       Rational[1, 2], 1], 
      Hue[0.41807002592789677`, 
       Rational[1, 2], 1], 
      Hue[0.41807002592789677`, 
       Rational[1, 2], 1], 
      Hue[0.42316177406743516`, 
       Rational[1, 2], 1], 
      Hue[0.42316177406743516`, 
       Rational[1, 2], 1], 
      Hue[0.43334527034651177`, 
       Rational[1, 2], 1], 
      Hue[0.43334527034651177`, 
       Rational[1, 2], 1], 
      Hue[0.45371226290466515`, 
       Rational[1, 2], 1], 
      Hue[0.45371226290466515`, 
       Rational[1, 2], 1], 
      Hue[0.4537312841026336, 
       Rational[1, 2], 1], 
      Hue[0.4537312841026336, 
       Rational[1, 2], 1], 
      Hue[0.4537503053006021, 
       Rational[1, 2], 1], 
      Hue[0.4537503053006021, 
       Rational[1, 2], 1], 
      Hue[0.45378834769653925`, 
       Rational[1, 2], 1], 
      Hue[0.45378834769653925`, 
       Rational[1, 2], 1], 
      Hue[0.4538644324884135, 
       Rational[1, 2], 1], 
      Hue[0.4538644324884135, 
       Rational[1, 2], 1], 
      Hue[0.4540166020721617, 
       Rational[1, 2], 1], 
      Hue[0.4540166020721617, 
       Rational[1, 2], 1], 
      Hue[0.45432094123965827`, 
       Rational[1, 2], 1], 
      Hue[0.45432094123965827`, 
       Rational[1, 2], 1], 
      Hue[0.4543399624376267, 
       Rational[1, 2], 1], 
      Hue[0.4543399624376267, 
       Rational[1, 2], 1], 
      Hue[0.45435898363559524`, 
       Rational[1, 2], 1], 
      Hue[0.45435898363559524`, 
       Rational[1, 2], 1], 
      Hue[0.45439702603153237`, 
       Rational[1, 2], 1], 
      Hue[0.45439702603153237`, 
       Rational[1, 2], 1], 
      Hue[0.45447311082340663`, 
       Rational[1, 2], 1], 
      Hue[0.45447311082340663`, 
       Rational[1, 2], 1], 
      Hue[0.45462528040715483`, 
       Rational[1, 2], 1], 
      Hue[0.45462528040715483`, 
       Rational[1, 2], 1], 
      Hue[0.45464430160512337`, 
       Rational[1, 2], 1], 
      Hue[0.45464430160512337`, 
       Rational[1, 2], 1], 
      Hue[0.45466332280309196`, 
       Rational[1, 2], 1], 
      Hue[0.45466332280309196`, 
       Rational[1, 2], 1], 
      Hue[0.45470136519902893`, 
       Rational[1, 2], 1], 
      Hue[0.45470136519902893`, 
       Rational[1, 2], 1], 
      Hue[0.454777449990903, 
       Rational[1, 2], 1], 
      Hue[0.454777449990903, 
       Rational[1, 2], 1], 
      Hue[0.45479647118887156`, 
       Rational[1, 2], 1], 
      Hue[0.45479647118887156`, 
       Rational[1, 2], 1], 
      Hue[0.45481549238684016`, 
       Rational[1, 2], 1], 
      Hue[0.45481549238684016`, 
       Rational[1, 2], 1], 
      Hue[0.4548535347827773, 
       Rational[1, 2], 1], 
      Hue[0.4548535347827773, 
       Rational[1, 2], 1], 
      Hue[0.45487255598074583`, 
       Rational[1, 2], 1], 
      Hue[0.45487255598074583`, 
       Rational[1, 2], 1], 
      Hue[0.45489157717871437`, 
       Rational[1, 2], 1], 
      Hue[0.45489157717871437`, 
       Rational[1, 2], 1], 
      Hue[0.45491059837668296`, 
       Rational[1, 2], 1], 
      Hue[0.45491059837668296`, 
       Rational[1, 2], 1], 
      Hue[0.4549296195746514, 
       Rational[1, 2], 1]}]},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->{True, False},
  AxesOrigin->{0, 0},
  DisplayFunction->Identity,
  Epilog->LineBox[CompressedData["
1:eJwUV3c81e8XFxfZCpEiIyMtRZH0PSollKRBmZWoKLJSFFGZyYpIZolLiMzk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    "]],
  Frame->True,
  PlotRange->All,
  PlotRangeClipping->True,
  PlotRangePadding->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{3.3918575873125*^9},
 CellLabel->"Out[48]=",
 ImageCache->GraphicsData["CompressedBitmap", "\<\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\
\>"]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"spinorlist", "=", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"x", ",", 
       RowBox[{"spinorfunction", "[", "x", "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"x", ",", 
       RowBox[{"-", "3"}], ",", "3", ",", "0.1"}], "}"}]}], "]"}]}], 
  ";"}], "\n", 
 RowBox[{"QListSpinorCombinedPlot", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"spinorlist", ",", 
     RowBox[{"Sin", "[", 
      RowBox[{"4", " ", "x"}], "]"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", 
     RowBox[{"-", "2"}], ",", "3"}], "}"}], ",", 
   RowBox[{"PlotRange", "\[Rule]", "All"}], ",", 
   RowBox[{"QCurveStyle", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"RGBColor", "[", 
       RowBox[{"0", ",", "0.7", ",", "0.9"}], "]"}], ",", 
      RowBox[{"Thickness", "[", "0.03", "]"}]}], "}"}]}], ",", 
   RowBox[{"PlotStyle", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"RGBColor", "[", 
       RowBox[{"0.2", ",", "0.2", ",", "0.5"}], "]"}], ",", 
      RowBox[{"Thickness", "[", "0.02", "]"}]}], "}"}]}]}], "]"}]}], "Input",
 CellChangeTimes->{3.391857565875*^9, 3.391857607078125*^9},
 CellLabel->"In[51]:="],

Cell[BoxData[
 GraphicsBox[{{GraphicsComplexBox[CompressedData["
1:eJyFlWlIlEEYx9erdLECLUIrKUURpUsrRcTnBZVAi0BLrfxg0WrZYZmoHR9M
RCMh8AiPLDQiythyVUhLmRVNW7wyxQtyt8S1dNE1dVUUO+ad5+NDA8vy433n
P7sz/H+z53xKlMpaoVCU/P38+xZDG+Hpn2nX1QeJfKyzItXtHHVhP3Txscas
ijpVrmFfwJ+PVRZTpcs3WHRQxscK26ZP8dbNvgY5bpn1uCxuWla2ijwLe/V2
p2ZENSzyFphvb3lkc+y4yJtnCz7T0drOHyJvjqXt06dnRE3Dbz5m2a3i1JbT
wVNivolVp7d5Wn+dgFL+/k8G2XEn3C8bQMXXMzLn9qNVmR1j4Mfzx9nugJeR
0W4TsM7z9Mw6f+8oCzVBCZ8/wtyD03TtpWY4wN/vZ4ZfY+mDqjko5s91rCZh
suDasBkecm5kcSkeDjEDs4AbGBExfswldhZeyM/hpnosKDfUDJXyfNBU5MXk
fTLDETkfXI0eylOBZngirw+e6wPRl2JMIJ+JAYYs/nWxzkY4LP9+WBo9+7S4
zAAX5f8HQz5tyquFBqiQ/z/YfKx/ER4/Ab3y/kBCbXHFN98psOJ5ZvDaHNeo
zDbBY3l/4cN88IWsmmk4JO8/PEjydIpcm4Qe+XzgTFOul6Pfd7GeBVz9FoK8
NYMibxkete7SmDtbRN4KJB4/GXVDXS3yVsFyZ67Ov0sn8tbAtBj/XBnWL/LW
obtlKSskGfMUUn33wa2f6wehm7+vkEJmZs7taB6Acp5vJXk15GyfCeyDJD7f
Wnp/98o7B61OrGcjZVQG3E/WaEWereSUF2d0aagR69tK15NGKnPtC0SenZTa
sd++Z8szkbdB8mhyylX3vBF5GyXHcHd95L1akWcvBUj5bg2ttXjeWjx37AvF
2B+KsU8UY78oxr5RjP2jGPtIMfaTYuwrxdhfirHPFGO/Kca+U4z9pxh9QDH6
gWL0BcXoD4rRJxT/b6BvKEb/UIw+ohj9RDH6imL0F8XoM4rRbxSj7yhG/1GM
PqQY/Ugx+pJi9CfF6FOK0a8Uo28pRv9SjD6mGP1MMfqaYvQ3xehzitHvFKPv
KUb/U4z3AcV4P1CM9wXFeH8g/wG9YwNs
     "], {
      {GrayLevel[0.5], EdgeForm[None], 
       GraphicsGroupBox[
        PolygonBox[{{1, 52, 53, 2}, {2, 54, 55, 3}, {3, 56, 57, 4}, {4, 58, 
         59, 5}, {11, 72, 73, 12}, {5, 60, 61, 6}, {7, 64, 65, 8}, {6, 62, 63,
          7}, {9, 68, 69, 10}, {10, 70, 71, 11}, {49, 148, 149, 50}, {12, 74, 
         75, 13}, {14, 78, 79, 15}, {15, 80, 81, 16}, {16, 82, 83, 17}, {17, 
         84, 85, 18}, {18, 86, 87, 19}, {19, 88, 89, 20}, {20, 90, 91, 21}, {
         21, 92, 93, 22}, {22, 94, 95, 23}, {23, 96, 97, 24}, {8, 66, 67, 
         9}, {24, 98, 99, 25}, {25, 100, 101, 26}, {26, 102, 103, 27}, {31, 
         112, 113, 32}, {28, 106, 107, 29}, {29, 108, 109, 30}, {30, 110, 111,
          31}, {32, 114, 115, 33}, {36, 122, 123, 37}, {33, 116, 117, 34}, {
         34, 118, 119, 35}, {35, 120, 121, 36}, {37, 124, 125, 38}, {13, 76, 
         77, 14}, {38, 126, 127, 39}, {46, 142, 143, 47}, {41, 132, 133, 
         42}, {42, 134, 135, 43}, {27, 104, 105, 28}, {40, 130, 131, 41}, {43,
          136, 137, 44}, {44, 138, 139, 45}, {47, 144, 145, 48}, {48, 146, 
         147, 49}, {45, 140, 141, 46}, {39, 128, 129, 40}, {50, 150, 151, 
         51}}]]}, {{}, {}, 
       {Hue[0, 0, 
         NCache[
          Rational[1, 4], 0.25]], Thickness[0.02], LineBox[{1, 2}], 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{2, 3}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{3, 4}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{4, 5}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{5, 6}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{6, 7}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{7, 8}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{8, 9}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{9, 10}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{10, 11}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{11, 12}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{12, 13}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{13, 14}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{14, 15}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{15, 16}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{16, 17}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{17, 18}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{18, 19}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{19, 20}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{20, 21}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{21, 22}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{22, 23}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{23, 24}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{24, 25}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{25, 26}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{26, 27}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{27, 28}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{28, 29}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{29, 30}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{30, 31}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{31, 32}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{32, 33}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{33, 34}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{34, 35}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{35, 36}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{36, 37}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{37, 38}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{38, 39}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{39, 40}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{40, 41}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{41, 42}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{42, 43}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{43, 44}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{44, 45}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{45, 46}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{46, 47}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{47, 48}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{48, 49}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{49, 50}]}, 
        {Hue[0, 0, 
          NCache[
           Rational[1, 4], 0.25]], LineBox[{50, 51}]}}}},
     VertexColors->{
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]], 
       Hue[0, 0, 
        Rational[1, 4]]}], GraphicsComplexBox[CompressedData["
1:eJyFlVtIlEEYhldbNTqQWF5YkYgmkUVLW3Zhy/eD5YUEheEptUhR85QpRZSg
KF6IyGYmWqa1WVpUrGkEWuasGWuWB9RVCTfPiodSM3E1xa3mn+/yo4Flefj/
eWd3hvcZt8jkgGhbhUJR/Pfz71sMg3fhc5/N9UaI4WOdJWy/au4dMUIrH2us
yyvxQpjyI6j5WGX2np5Rc5p3cI+PFdZoXAqpelQKctwym7nj0hSR8lrkLbHy
J03eVQEtIm+RndZVFtT2m0TeLxaRoNZYb5hF3k/m5N5+ItZ7GKx8zLH5ivQk
beeImP+dBfWFTZrLRuAuf3+K5YYtNK8GDUE0X2+C7ck/UJ4x3QeHef4oa/Rz
zM7pN8I6zxtknrlFvmn2eijm87+yfe79BRYPE6j4+93seUWrc9vqMBTy5y0s
Y3B/ck/VJGg517HQmPCitZAfgBvo7z96yiV4Dirl5+DM0sceB8+CTp4PJ5+q
rmVHTYO3nA/hV7pr1gZHoUxeH7z8qpOylH0gn8kQBG65WS7VV8NR+ffDm4LP
qh0zX+CS/P9AF2CNz1QPQKn8/8G6rfvg2cBx6JD3Byq7fJc0blNgw/Pmofb6
vH2K7zTcl/cXMjp2jzr1TsIRef+hxHFurEYzDu3y+QAkpmdZnIfEekugaZ7u
GlCZRN4y/C4LdNrrYBB5K/B+a+bbXbHPRN4q9HjGWc/bfBJ5a9DgWrlQfLxL
5K3DRENb2Uttj8hTSA9eOIwtuPdCG39fIeXpboV2pJqghOfbSMrFuEM7XTsh
ls+3lR6621RY6lrEehsku4u3P9jqDSJPKVlm+89o4l+J9ZWSUTv/bSonU+TZ
SY6RK415Zp3Is5faz+nTzJf1Is9Bytrkoz6WWiPyNkr5Hh5pUYYaPG8Dnjv2
hWLsD8XYJ4qxXxRj3yjG/lGMfaQY+0kx9pVi7C/F2GeKsd8UY98pxv5TjD6g
GP1AMfqCYvQHxegTiv830DcUo38oRh9RjH6iGH1FMfqLYvQZxeg3itF3FKP/
KEYfUox+pBh9STH6k2L0KcXoV4rRtxSjfylGH1OMfqYYfU0x+pti9DnF6HeK
0fcUo/8pxvuAYrwfKMb7gmK8P5D/AIHmCDk=
     "], {
      {GrayLevel[0.5], EdgeForm[None], 
       GraphicsGroupBox[
        PolygonBox[{{1, 52, 53, 2}, {2, 54, 55, 3}, {3, 56, 57, 4}, {4, 58, 
         59, 5}, {11, 72, 73, 12}, {5, 60, 61, 6}, {7, 64, 65, 8}, {6, 62, 63,
          7}, {9, 68, 69, 10}, {10, 70, 71, 11}, {49, 148, 149, 50}, {12, 74, 
         75, 13}, {14, 78, 79, 15}, {15, 80, 81, 16}, {16, 82, 83, 17}, {17, 
         84, 85, 18}, {18, 86, 87, 19}, {19, 88, 89, 20}, {20, 90, 91, 21}, {
         21, 92, 93, 22}, {22, 94, 95, 23}, {23, 96, 97, 24}, {8, 66, 67, 
         9}, {24, 98, 99, 25}, {25, 100, 101, 26}, {26, 102, 103, 27}, {31, 
         112, 113, 32}, {28, 106, 107, 29}, {29, 108, 109, 30}, {30, 110, 111,
          31}, {32, 114, 115, 33}, {36, 122, 123, 37}, {33, 116, 117, 34}, {
         34, 118, 119, 35}, {35, 120, 121, 36}, {37, 124, 125, 38}, {13, 76, 
         77, 14}, {38, 126, 127, 39}, {46, 142, 143, 47}, {41, 132, 133, 
         42}, {42, 134, 135, 43}, {27, 104, 105, 28}, {40, 130, 131, 41}, {43,
          136, 137, 44}, {44, 138, 139, 45}, {47, 144, 145, 48}, {48, 146, 
         147, 49}, {45, 140, 141, 46}, {39, 128, 129, 40}, {50, 150, 151, 
         51}}]]}, {{}, {}, 
       {Hue[0.6816901138162093, 1, 1], Thickness[0.02], LineBox[{1, 2}], 
        {Hue[0.6976056081253988, 1, 1], LineBox[{2, 3}]}, 
        {Hue[0.7135211024345884, 1, 1], LineBox[{3, 4}]}, 
        {Hue[0.7294365967437779, 1, 1], LineBox[{4, 5}]}, 
        {Hue[0.7453520910529674, 1, 1], LineBox[{5, 6}]}, 
        {Hue[0.26126758536215705`, 1, 1], LineBox[{6, 7}]}, 
        {Hue[0.2771830796713466, 1, 1], LineBox[{7, 8}]}, 
        {Hue[0.29309857398053607`, 1, 1], LineBox[{8, 9}]}, 
        {Hue[0.30901406828972566`, 1, 1], LineBox[{9, 10}]}, 
        {Hue[0.3249295625989152, 1, 1], LineBox[{10, 11}]}, 
        {Hue[0.3408450569081047, 1, 1], LineBox[{11, 12}]}, 
        {Hue[0.3567605512172942, 1, 1], LineBox[{12, 13}]}, 
        {Hue[0.37267604552648376`, 1, 1], LineBox[{13, 14}]}, 
        {Hue[0.3885915398356733, 1, 1], LineBox[{14, 15}]}, 
        {Hue[0.4045070341448629, 1, 1], LineBox[{15, 16}]}, 
        {Hue[0.9204225284540524, 1, 1], LineBox[{16, 17}]}, 
        {Hue[0.9363380227632419, 1, 1], LineBox[{17, 18}]}, 
        {Hue[0.9522535170724314, 1, 1], LineBox[{18, 19}]}, 
        {Hue[0.9681690113816209, 1, 1], LineBox[{19, 20}]}, 
        {Hue[0.9840845056908105, 1, 1], LineBox[{20, 21}]}, 
        {Hue[0., 1, 1], LineBox[{21, 22}]}, 
        {Hue[0.015915494309189558`, 1, 1], LineBox[{22, 23}]}, 
        {Hue[0.031830988618379095`, 1, 1], LineBox[{23, 24}]}, 
        {Hue[0.047746482927568626`, 1, 1], LineBox[{24, 25}]}, 
        {Hue[0.06366197723675816, 1, 1], LineBox[{25, 26}]}, 
        {Hue[0.07957747154594771, 1, 1], LineBox[{26, 27}]}, 
        {Hue[0.5954929658551372, 1, 1], LineBox[{27, 28}]}, 
        {Hue[0.6114084601643268, 1, 1], LineBox[{28, 29}]}, 
        {Hue[0.6273239544735163, 1, 1], LineBox[{29, 30}]}, 
        {Hue[0.6432394487827058, 1, 1], LineBox[{30, 31}]}, 
        {Hue[0.6591549430918953, 1, 1], LineBox[{31, 32}]}, 
        {Hue[0.6750704374010849, 1, 1], LineBox[{32, 33}]}, 
        {Hue[0.6909859317102744, 1, 1], LineBox[{33, 34}]}, 
        {Hue[0.7069014260194639, 1, 1], LineBox[{34, 35}]}, 
        {Hue[0.7228169203286534, 1, 1], LineBox[{35, 36}]}, 
        {Hue[0.738732414637843, 1, 1], LineBox[{36, 37}]}, 
        {Hue[0.2546479089470326, 1, 1], LineBox[{37, 38}]}, 
        {Hue[0.27056340325622213`, 1, 1], LineBox[{38, 39}]}, 
        {Hue[0.28647889756541167`, 1, 1], LineBox[{39, 40}]}, 
        {Hue[0.3023943918746012, 1, 1], LineBox[{40, 41}]}, 
        {Hue[0.31830988618379075`, 1, 1], LineBox[{41, 42}]}, 
        {Hue[0.33422538049298023`, 1, 1], LineBox[{42, 43}]}, 
        {Hue[0.35014087480216977`, 1, 1], LineBox[{43, 44}]}, 
        {Hue[0.36605636911135936`, 1, 1], LineBox[{44, 45}]}, 
        {Hue[0.3819718634205489, 1, 1], LineBox[{45, 46}]}, 
        {Hue[0.39788735772973843`, 1, 1], LineBox[{46, 47}]}, 
        {Hue[0.41380285203892797`, 1, 1], LineBox[{47, 48}]}, 
        {Hue[0.9297183463481175, 1, 1], LineBox[{48, 49}]}, 
        {Hue[0.9456338406573069, 1, 1], LineBox[{49, 50}]}, 
        {Hue[0.9615493349664965, 1, 1], LineBox[{50, 51}]}}}},
     VertexColors->{
       Hue[0.6816901138162093, 1, 1], 
       Hue[0.6976056081253988, 1, 1], 
       Hue[0.7135211024345884, 1, 1], 
       Hue[0.7294365967437779, 1, 1], 
       Hue[0.7453520910529674, 1, 1], 
       Hue[0.26126758536215705`, 1, 1], 
       Hue[0.2771830796713466, 1, 1], 
       Hue[0.29309857398053607`, 1, 1], 
       Hue[0.30901406828972566`, 1, 1], 
       Hue[0.3249295625989152, 1, 1], 
       Hue[0.3408450569081047, 1, 1], 
       Hue[0.3567605512172942, 1, 1], 
       Hue[0.37267604552648376`, 1, 1], 
       Hue[0.3885915398356733, 1, 1], 
       Hue[0.4045070341448629, 1, 1], 
       Hue[0.9204225284540524, 1, 1], 
       Hue[0.9363380227632419, 1, 1], 
       Hue[0.9522535170724314, 1, 1], 
       Hue[0.9681690113816209, 1, 1], 
       Hue[0.9840845056908105, 1, 1], 
       Hue[0., 1, 1], 
       Hue[0.015915494309189558`, 1, 1], 
       Hue[0.031830988618379095`, 1, 1], 
       Hue[0.047746482927568626`, 1, 1], 
       Hue[0.06366197723675816, 1, 1], 
       Hue[0.07957747154594771, 1, 1], 
       Hue[0.5954929658551372, 1, 1], 
       Hue[0.6114084601643268, 1, 1], 
       Hue[0.6273239544735163, 1, 1], 
       Hue[0.6432394487827058, 1, 1], 
       Hue[0.6591549430918953, 1, 1], 
       Hue[0.6750704374010849, 1, 1], 
       Hue[0.6909859317102744, 1, 1], 
       Hue[0.7069014260194639, 1, 1], 
       Hue[0.7228169203286534, 1, 1], 
       Hue[0.738732414637843, 1, 1], 
       Hue[0.2546479089470326, 1, 1], 
       Hue[0.27056340325622213`, 1, 1], 
       Hue[0.28647889756541167`, 1, 1], 
       Hue[0.3023943918746012, 1, 1], 
       Hue[0.31830988618379075`, 1, 1], 
       Hue[0.33422538049298023`, 1, 1], 
       Hue[0.35014087480216977`, 1, 1], 
       Hue[0.36605636911135936`, 1, 1], 
       Hue[0.3819718634205489, 1, 1], 
       Hue[0.39788735772973843`, 1, 1], 
       Hue[0.41380285203892797`, 1, 1], 
       Hue[0.9297183463481175, 1, 1], 
       Hue[0.9456338406573069, 1, 1], 
       Hue[0.9615493349664965, 1, 1], 
       Hue[0.9774648292756861, 1, 1], 
       Hue[0.6816901138162093, 1, 1], 
       Hue[0.6976056081253988, 1, 1], 
       Hue[0.6976056081253988, 1, 1], 
       Hue[0.7135211024345884, 1, 1], 
       Hue[0.7135211024345884, 1, 1], 
       Hue[0.7294365967437779, 1, 1], 
       Hue[0.7294365967437779, 1, 1], 
       Hue[0.7453520910529674, 1, 1], 
       Hue[0.7453520910529674, 1, 1], 
       Hue[0.26126758536215705`, 1, 1], 
       Hue[0.26126758536215705`, 1, 1], 
       Hue[0.2771830796713466, 1, 1], 
       Hue[0.2771830796713466, 1, 1], 
       Hue[0.29309857398053607`, 1, 1], 
       Hue[0.29309857398053607`, 1, 1], 
       Hue[0.30901406828972566`, 1, 1], 
       Hue[0.30901406828972566`, 1, 1], 
       Hue[0.3249295625989152, 1, 1], 
       Hue[0.3249295625989152, 1, 1], 
       Hue[0.3408450569081047, 1, 1], 
       Hue[0.3408450569081047, 1, 1], 
       Hue[0.3567605512172942, 1, 1], 
       Hue[0.3567605512172942, 1, 1], 
       Hue[0.37267604552648376`, 1, 1], 
       Hue[0.37267604552648376`, 1, 1], 
       Hue[0.3885915398356733, 1, 1], 
       Hue[0.3885915398356733, 1, 1], 
       Hue[0.4045070341448629, 1, 1], 
       Hue[0.4045070341448629, 1, 1], 
       Hue[0.9204225284540524, 1, 1], 
       Hue[0.9204225284540524, 1, 1], 
       Hue[0.9363380227632419, 1, 1], 
       Hue[0.9363380227632419, 1, 1], 
       Hue[0.9522535170724314, 1, 1], 
       Hue[0.9522535170724314, 1, 1], 
       Hue[0.9681690113816209, 1, 1], 
       Hue[0.9681690113816209, 1, 1], 
       Hue[0.9840845056908105, 1, 1], 
       Hue[0.9840845056908105, 1, 1], 
       Hue[0., 1, 1], 
       Hue[0., 1, 1], 
       Hue[0.015915494309189558`, 1, 1], 
       Hue[0.015915494309189558`, 1, 1], 
       Hue[0.031830988618379095`, 1, 1], 
       Hue[0.031830988618379095`, 1, 1], 
       Hue[0.047746482927568626`, 1, 1], 
       Hue[0.047746482927568626`, 1, 1], 
       Hue[0.06366197723675816, 1, 1], 
       Hue[0.06366197723675816, 1, 1], 
       Hue[0.07957747154594771, 1, 1], 
       Hue[0.07957747154594771, 1, 1], 
       Hue[0.5954929658551372, 1, 1], 
       Hue[0.5954929658551372, 1, 1], 
       Hue[0.6114084601643268, 1, 1], 
       Hue[0.6114084601643268, 1, 1], 
       Hue[0.6273239544735163, 1, 1], 
       Hue[0.6273239544735163, 1, 1], 
       Hue[0.6432394487827058, 1, 1], 
       Hue[0.6432394487827058, 1, 1], 
       Hue[0.6591549430918953, 1, 1], 
       Hue[0.6591549430918953, 1, 1], 
       Hue[0.6750704374010849, 1, 1], 
       Hue[0.6750704374010849, 1, 1], 
       Hue[0.6909859317102744, 1, 1], 
       Hue[0.6909859317102744, 1, 1], 
       Hue[0.7069014260194639, 1, 1], 
       Hue[0.7069014260194639, 1, 1], 
       Hue[0.7228169203286534, 1, 1], 
       Hue[0.7228169203286534, 1, 1], 
       Hue[0.738732414637843, 1, 1], 
       Hue[0.738732414637843, 1, 1], 
       Hue[0.2546479089470326, 1, 1], 
       Hue[0.2546479089470326, 1, 1], 
       Hue[0.27056340325622213`, 1, 1], 
       Hue[0.27056340325622213`, 1, 1], 
       Hue[0.28647889756541167`, 1, 1], 
       Hue[0.28647889756541167`, 1, 1], 
       Hue[0.3023943918746012, 1, 1], 
       Hue[0.3023943918746012, 1, 1], 
       Hue[0.31830988618379075`, 1, 1], 
       Hue[0.31830988618379075`, 1, 1], 
       Hue[0.33422538049298023`, 1, 1], 
       Hue[0.33422538049298023`, 1, 1], 
       Hue[0.35014087480216977`, 1, 1], 
       Hue[0.35014087480216977`, 1, 1], 
       Hue[0.36605636911135936`, 1, 1], 
       Hue[0.36605636911135936`, 1, 1], 
       Hue[0.3819718634205489, 1, 1], 
       Hue[0.3819718634205489, 1, 1], 
       Hue[0.39788735772973843`, 1, 1], 
       Hue[0.39788735772973843`, 1, 1], 
       Hue[0.41380285203892797`, 1, 1], 
       Hue[0.41380285203892797`, 1, 1], 
       Hue[0.9297183463481175, 1, 1], 
       Hue[0.9297183463481175, 1, 1], 
       Hue[0.9456338406573069, 1, 1], 
       Hue[0.9456338406573069, 1, 1], 
       Hue[0.9615493349664965, 1, 1], 
       Hue[0.9615493349664965, 1, 1], 
       Hue[0.9774648292756861, 1, 1]}], GraphicsComplexBox[CompressedData["
1:eJx1lXtMjWEcx08pia0/GmtjaZ1ouZV1UVGe12xKM1Zrkyxnky60bsaYDCVx
1shGJSklFMZIrGb8YkXpJtHljO6nnFZKKh2nTnje5/eP7ffbzs4+e9/n+5zz
Pvt+Xvv98UERpgqFIvPv59+3mErj1+iWVeF1EMnHCFEO+rTtb5qgns8MqFMW
B4dkNIIbHwNMh+12bFj6FnL46CEwLMTq0GwxyHHT4GBm86VTVSnypoDVerNx
zSeRNwGLxvcNBgZ0i7yf4HTKP2poR7/I+wH5hbaqSU8tzPEZhc8rrU/U+/WK
9cMwO6lszhlphWv8fh2kFriE+Dy5BxF8vwHYdjM8bVzVAq48vw9sq+cW3gnu
ByPP64KUwlulNW7DkM3Xd0BPtMZysGAU1vP7W8Ciy7y7onYErvLrtaCKN8T6
jgzAJc4V4DjFPmput4Piv7krX2dV7ZuWrElsgwJ5PVuuHd/47FU/bJDz2eu4
rRPninSQJ+/PAlKXdTprdSCfSTdLTj5irJ7Sgof8+1lSkNHpgF0nRMv/j834
ZCXH7ayDG/L/ZzZFc6Ear0Jokp8PG35R0rbZtwFMeN4Y+630Meld2wq58vNl
Ok+vwNGkdnCXnz+LPJ8fs8W/FRrl82HpN/fYu5o2i/2mWH+myvLk8yqRN83O
PI5qiRh7JPL0zC70fptFRqbIM7ChXxn6XdpSkTfD+tR+zsc7ykWekQVXr84q
vFwu8hRST+/prDTHMmjg9yukw9+/GcpKHsB1nm8ivdNdaU2oyYMovt5Ucg+z
T49/mCz2myfFrHhvvi7oqMgzk2xfaiwSI9RifzPpqTJ3wMrjosgzlxKsR4qz
tWqRN19yOxhbnup4VuRZSMoLzqYuucdE3gLJpfLDhPfeSDzvSjxv7AvF2B+K
sU8UY78oxr5RjP2jGPtIMfaTYuwrxdhfirHPFGO/Kca+U4z9pxh9QDH6gWL0
BcXoD4rRJxSjTyhGv1CMvqEY/UMx+ohi9BPF6CuK0V8Uo88oRr9RjL6jGP1H
MfqQYvQjxehLitGfFKNPKUa/Uoy+pRj9SzH6mGL0M8Xoa4rR3xSjzylGv1OM
vqcY/U8xvg8oxvcD8h9SmxvR
     "], {
      {GrayLevel[0.5], EdgeForm[None], 
       GraphicsGroupBox[{PolygonBox[{{90, 20, 21}, {21, 22, 91}}], 
         PolygonBox[{{56, 3, 4, 57}, {54, 2, 3, 55}, {52, 1, 2, 53}, {60, 5, 
          6, 61}, {62, 6, 7, 63}, {68, 9, 10, 69}, {70, 10, 11, 71}, {64, 7, 
          8, 65}, {58, 4, 5, 59}, {72, 11, 12, 73}, {74, 12, 13, 75}, {66, 8, 
          9, 67}, {80, 15, 16, 81}, {82, 16, 17, 83}, {84, 17, 18, 85}, {146, 
          49, 50, 147}, {86, 18, 19, 87}, {94, 23, 24, 95}, {88, 19, 20, 
          89}, {78, 14, 15, 79}, {92, 22, 23, 93}, {96, 24, 25, 97}, {98, 25, 
          26, 99}, {100, 26, 27, 101}, {110, 31, 32, 111}, {104, 28, 29, 
          105}, {106, 29, 30, 107}, {108, 30, 31, 109}, {112, 32, 33, 113}, {
          120, 36, 37, 121}, {114, 33, 34, 115}, {116, 34, 35, 117}, {118, 35,
           36, 119}, {122, 37, 38, 123}, {76, 13, 14, 77}, {124, 38, 39, 
          125}, {140, 46, 47, 141}, {130, 41, 42, 131}, {132, 42, 43, 133}, {
          102, 27, 28, 103}, {128, 40, 41, 129}, {134, 43, 44, 135}, {136, 44,
           45, 137}, {142, 47, 48, 143}, {144, 48, 49, 145}, {138, 45, 46, 
          139}, {126, 39, 40, 127}, {148, 50, 51, 149}}]}]}, {{}, {}, 
       {Hue[0.8633802276324187, 
         NCache[
          Rational[1, 2], 0.5], 1], Thickness[0.02], LineBox[{1, 2}], 
        {Hue[0.8952112162507977, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{2, 3}]}, 
        {Hue[0.9270422048691769, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{3, 4}]}, 
        {Hue[0.9588731934875558, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{4, 5}]}, 
        {Hue[0.9907041821059349, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{5, 6}]}, 
        {Hue[0.522535170724314, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{6, 7}]}, 
        {Hue[0.5543661593426931, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{7, 8}]}, 
        {Hue[0.5861971479610721, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{8, 9}]}, 
        {Hue[0.6180281365794511, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{9, 10}]}, 
        {Hue[0.6498591251978303, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{10, 11}]}, 
        {Hue[0.6816901138162094, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{11, 12}]}, 
        {Hue[0.7135211024345884, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{12, 13}]}, 
        {Hue[0.7453520910529675, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{13, 14}]}, 
        {Hue[0.2771830796713466, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{14, 15}]}, 
        {Hue[0.30901406828972566`, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{15, 16}]}, 
        {Hue[0.34084505690810474`, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{16, 17}]}, 
        {Hue[0.3726760455264838, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{17, 18}]}, 
        {Hue[0.4045070341448629, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{18, 19}]}, 
        {Hue[0.43633802276324196`, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{19, 20}]}, 
        {Hue[0.468169011381621, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{20, 21}]}, 
        {Hue[0., 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{21, 22}]}, 
        {Hue[0.031830988618379116`, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{22, 23}]}, 
        {Hue[0.06366197723675819, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{23, 24}]}, 
        {Hue[0.09549296585513725, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{24, 25}]}, 
        {Hue[0.12732395447351633`, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{25, 26}]}, 
        {Hue[0.15915494309189543`, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{26, 27}]}, 
        {Hue[0.19098593171027448`, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{27, 28}]}, 
        {Hue[0.22281692032865355`, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{28, 29}]}, 
        {Hue[0.7546479089470326, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{29, 30}]}, 
        {Hue[0.7864788975654117, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{30, 31}]}, 
        {Hue[0.8183098861837907, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{31, 32}]}, 
        {Hue[0.8501408748021698, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{32, 33}]}, 
        {Hue[0.8819718634205489, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{33, 34}]}, 
        {Hue[0.913802852038928, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{34, 35}]}, 
        {Hue[0.945633840657307, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{35, 36}]}, 
        {Hue[0.9774648292756861, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{36, 37}]}, 
        {Hue[0.5092958178940651, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{37, 38}]}, 
        {Hue[0.5411268065124442, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{38, 39}]}, 
        {Hue[0.5729577951308232, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{39, 40}]}, 
        {Hue[0.6047887837492023, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{40, 41}]}, 
        {Hue[0.6366197723675815, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{41, 42}]}, 
        {Hue[0.6684507609859605, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{42, 43}]}, 
        {Hue[0.7002817496043395, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{43, 44}]}, 
        {Hue[0.7321127382227186, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{44, 45}]}, 
        {Hue[0.26394372684109774`, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{45, 46}]}, 
        {Hue[0.29577471545947687`, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{46, 47}]}, 
        {Hue[0.32760570407785594`, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{47, 48}]}, 
        {Hue[0.3594366926962349, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{48, 49}]}, 
        {Hue[0.39126768131461404`, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{49, 50}]}, 
        {Hue[0.4230986699329931, 
          NCache[
           Rational[1, 2], 0.5], 1], LineBox[{50, 51}]}}}},
     VertexColors->{
       Hue[0.8633802276324187, 
        Rational[1, 2], 1], 
       Hue[0.8952112162507977, 
        Rational[1, 2], 1], 
       Hue[0.9270422048691769, 
        Rational[1, 2], 1], 
       Hue[0.9588731934875558, 
        Rational[1, 2], 1], 
       Hue[0.9907041821059349, 
        Rational[1, 2], 1], 
       Hue[0.522535170724314, 
        Rational[1, 2], 1], 
       Hue[0.5543661593426931, 
        Rational[1, 2], 1], 
       Hue[0.5861971479610721, 
        Rational[1, 2], 1], 
       Hue[0.6180281365794511, 
        Rational[1, 2], 1], 
       Hue[0.6498591251978303, 
        Rational[1, 2], 1], 
       Hue[0.6816901138162094, 
        Rational[1, 2], 1], 
       Hue[0.7135211024345884, 
        Rational[1, 2], 1], 
       Hue[0.7453520910529675, 
        Rational[1, 2], 1], 
       Hue[0.2771830796713466, 
        Rational[1, 2], 1], 
       Hue[0.30901406828972566`, 
        Rational[1, 2], 1], 
       Hue[0.34084505690810474`, 
        Rational[1, 2], 1], 
       Hue[0.3726760455264838, 
        Rational[1, 2], 1], 
       Hue[0.4045070341448629, 
        Rational[1, 2], 1], 
       Hue[0.43633802276324196`, 
        Rational[1, 2], 1], 
       Hue[0.468169011381621, 
        Rational[1, 2], 1], 
       Hue[0., 
        Rational[1, 2], 1], 
       Hue[0.031830988618379116`, 
        Rational[1, 2], 1], 
       Hue[0.06366197723675819, 
        Rational[1, 2], 1], 
       Hue[0.09549296585513725, 
        Rational[1, 2], 1], 
       Hue[0.12732395447351633`, 
        Rational[1, 2], 1], 
       Hue[0.15915494309189543`, 
        Rational[1, 2], 1], 
       Hue[0.19098593171027448`, 
        Rational[1, 2], 1], 
       Hue[0.22281692032865355`, 
        Rational[1, 2], 1], 
       Hue[0.7546479089470326, 
        Rational[1, 2], 1], 
       Hue[0.7864788975654117, 
        Rational[1, 2], 1], 
       Hue[0.8183098861837907, 
        Rational[1, 2], 1], 
       Hue[0.8501408748021698, 
        Rational[1, 2], 1], 
       Hue[0.8819718634205489, 
        Rational[1, 2], 1], 
       Hue[0.913802852038928, 
        Rational[1, 2], 1], 
       Hue[0.945633840657307, 
        Rational[1, 2], 1], 
       Hue[0.9774648292756861, 
        Rational[1, 2], 1], 
       Hue[0.5092958178940651, 
        Rational[1, 2], 1], 
       Hue[0.5411268065124442, 
        Rational[1, 2], 1], 
       Hue[0.5729577951308232, 
        Rational[1, 2], 1], 
       Hue[0.6047887837492023, 
        Rational[1, 2], 1], 
       Hue[0.6366197723675815, 
        Rational[1, 2], 1], 
       Hue[0.6684507609859605, 
        Rational[1, 2], 1], 
       Hue[0.7002817496043395, 
        Rational[1, 2], 1], 
       Hue[0.7321127382227186, 
        Rational[1, 2], 1], 
       Hue[0.26394372684109774`, 
        Rational[1, 2], 1], 
       Hue[0.29577471545947687`, 
        Rational[1, 2], 1], 
       Hue[0.32760570407785594`, 
        Rational[1, 2], 1], 
       Hue[0.3594366926962349, 
        Rational[1, 2], 1], 
       Hue[0.39126768131461404`, 
        Rational[1, 2], 1], 
       Hue[0.4230986699329931, 
        Rational[1, 2], 1], 
       Hue[0.4549296585513722, 
        Rational[1, 2], 1], 
       Hue[0.8633802276324187, 
        Rational[1, 2], 1], 
       Hue[0.8952112162507977, 
        Rational[1, 2], 1], 
       Hue[0.8952112162507977, 
        Rational[1, 2], 1], 
       Hue[0.9270422048691769, 
        Rational[1, 2], 1], 
       Hue[0.9270422048691769, 
        Rational[1, 2], 1], 
       Hue[0.9588731934875558, 
        Rational[1, 2], 1], 
       Hue[0.9588731934875558, 
        Rational[1, 2], 1], 
       Hue[0.9907041821059349, 
        Rational[1, 2], 1], 
       Hue[0.9907041821059349, 
        Rational[1, 2], 1], 
       Hue[0.522535170724314, 
        Rational[1, 2], 1], 
       Hue[0.522535170724314, 
        Rational[1, 2], 1], 
       Hue[0.5543661593426931, 
        Rational[1, 2], 1], 
       Hue[0.5543661593426931, 
        Rational[1, 2], 1], 
       Hue[0.5861971479610721, 
        Rational[1, 2], 1], 
       Hue[0.5861971479610721, 
        Rational[1, 2], 1], 
       Hue[0.6180281365794511, 
        Rational[1, 2], 1], 
       Hue[0.6180281365794511, 
        Rational[1, 2], 1], 
       Hue[0.6498591251978303, 
        Rational[1, 2], 1], 
       Hue[0.6498591251978303, 
        Rational[1, 2], 1], 
       Hue[0.6816901138162094, 
        Rational[1, 2], 1], 
       Hue[0.6816901138162094, 
        Rational[1, 2], 1], 
       Hue[0.7135211024345884, 
        Rational[1, 2], 1], 
       Hue[0.7135211024345884, 
        Rational[1, 2], 1], 
       Hue[0.7453520910529675, 
        Rational[1, 2], 1], 
       Hue[0.7453520910529675, 
        Rational[1, 2], 1], 
       Hue[0.2771830796713466, 
        Rational[1, 2], 1], 
       Hue[0.2771830796713466, 
        Rational[1, 2], 1], 
       Hue[0.30901406828972566`, 
        Rational[1, 2], 1], 
       Hue[0.30901406828972566`, 
        Rational[1, 2], 1], 
       Hue[0.34084505690810474`, 
        Rational[1, 2], 1], 
       Hue[0.34084505690810474`, 
        Rational[1, 2], 1], 
       Hue[0.3726760455264838, 
        Rational[1, 2], 1], 
       Hue[0.3726760455264838, 
        Rational[1, 2], 1], 
       Hue[0.4045070341448629, 
        Rational[1, 2], 1], 
       Hue[0.4045070341448629, 
        Rational[1, 2], 1], 
       Hue[0.43633802276324196`, 
        Rational[1, 2], 1], 
       Hue[0.43633802276324196`, 
        Rational[1, 2], 1], 
       Hue[0.468169011381621, 
        Rational[1, 2], 1], 
       Hue[0.468169011381621, 
        Rational[1, 2], 1], 
       Hue[0.031830988618379116`, 
        Rational[1, 2], 1], 
       Hue[0.031830988618379116`, 
        Rational[1, 2], 1], 
       Hue[0.06366197723675819, 
        Rational[1, 2], 1], 
       Hue[0.06366197723675819, 
        Rational[1, 2], 1], 
       Hue[0.09549296585513725, 
        Rational[1, 2], 1], 
       Hue[0.09549296585513725, 
        Rational[1, 2], 1], 
       Hue[0.12732395447351633`, 
        Rational[1, 2], 1], 
       Hue[0.12732395447351633`, 
        Rational[1, 2], 1], 
       Hue[0.15915494309189543`, 
        Rational[1, 2], 1], 
       Hue[0.15915494309189543`, 
        Rational[1, 2], 1], 
       Hue[0.19098593171027448`, 
        Rational[1, 2], 1], 
       Hue[0.19098593171027448`, 
        Rational[1, 2], 1], 
       Hue[0.22281692032865355`, 
        Rational[1, 2], 1], 
       Hue[0.22281692032865355`, 
        Rational[1, 2], 1], 
       Hue[0.7546479089470326, 
        Rational[1, 2], 1], 
       Hue[0.7546479089470326, 
        Rational[1, 2], 1], 
       Hue[0.7864788975654117, 
        Rational[1, 2], 1], 
       Hue[0.7864788975654117, 
        Rational[1, 2], 1], 
       Hue[0.8183098861837907, 
        Rational[1, 2], 1], 
       Hue[0.8183098861837907, 
        Rational[1, 2], 1], 
       Hue[0.8501408748021698, 
        Rational[1, 2], 1], 
       Hue[0.8501408748021698, 
        Rational[1, 2], 1], 
       Hue[0.8819718634205489, 
        Rational[1, 2], 1], 
       Hue[0.8819718634205489, 
        Rational[1, 2], 1], 
       Hue[0.913802852038928, 
        Rational[1, 2], 1], 
       Hue[0.913802852038928, 
        Rational[1, 2], 1], 
       Hue[0.945633840657307, 
        Rational[1, 2], 1], 
       Hue[0.945633840657307, 
        Rational[1, 2], 1], 
       Hue[0.9774648292756861, 
        Rational[1, 2], 1], 
       Hue[0.9774648292756861, 
        Rational[1, 2], 1], 
       Hue[0.5092958178940651, 
        Rational[1, 2], 1], 
       Hue[0.5092958178940651, 
        Rational[1, 2], 1], 
       Hue[0.5411268065124442, 
        Rational[1, 2], 1], 
       Hue[0.5411268065124442, 
        Rational[1, 2], 1], 
       Hue[0.5729577951308232, 
        Rational[1, 2], 1], 
       Hue[0.5729577951308232, 
        Rational[1, 2], 1], 
       Hue[0.6047887837492023, 
        Rational[1, 2], 1], 
       Hue[0.6047887837492023, 
        Rational[1, 2], 1], 
       Hue[0.6366197723675815, 
        Rational[1, 2], 1], 
       Hue[0.6366197723675815, 
        Rational[1, 2], 1], 
       Hue[0.6684507609859605, 
        Rational[1, 2], 1], 
       Hue[0.6684507609859605, 
        Rational[1, 2], 1], 
       Hue[0.7002817496043395, 
        Rational[1, 2], 1], 
       Hue[0.7002817496043395, 
        Rational[1, 2], 1], 
       Hue[0.7321127382227186, 
        Rational[1, 2], 1], 
       Hue[0.7321127382227186, 
        Rational[1, 2], 1], 
       Hue[0.26394372684109774`, 
        Rational[1, 2], 1], 
       Hue[0.26394372684109774`, 
        Rational[1, 2], 1], 
       Hue[0.29577471545947687`, 
        Rational[1, 2], 1], 
       Hue[0.29577471545947687`, 
        Rational[1, 2], 1], 
       Hue[0.32760570407785594`, 
        Rational[1, 2], 1], 
       Hue[0.32760570407785594`, 
        Rational[1, 2], 1], 
       Hue[0.3594366926962349, 
        Rational[1, 2], 1], 
       Hue[0.3594366926962349, 
        Rational[1, 2], 1], 
       Hue[0.39126768131461404`, 
        Rational[1, 2], 1], 
       Hue[0.39126768131461404`, 
        Rational[1, 2], 1], 
       Hue[0.4230986699329931, 
        Rational[1, 2], 1], 
       Hue[0.4230986699329931, 
        Rational[1, 2], 1], 
       Hue[0.4549296585513722, 
        Rational[1, 2], 1]}]}, {{}, {}, 
    {RGBColor[0, 0.7, 0.9], Thickness[0.03], LineBox[CompressedData["
1:eJwVmnc8Vf8fx4l7jVKUElnRUElGSdb7aCJtSaEy2lIqUVJSaZhlRAr1LUmk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      "]]}}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  DisplayFunction->Identity,
  Epilog->{
    RGBColor[0.2, 0.2, 0.5], 
    Thickness[0.02], 
    LineBox[{{-2., 0.2780774759126713}, {-1.9, 
      0.3058275026065413}, {-1.7999999999999998`, 0.2854066843122354}, {-1.7, 
      0.20284659219329154`}, {-1.5999999999999999`, 
      0.05456379526725478}, {-1.5, 0.14877224618449109`}, {-1.4, 
      0.38121645256137043`}, {-1.2999999999999998`, 
      0.6044603802351721}, {-1.2, 0.7748719824824932}, {-1.0999999999999999`, 
      0.8538167306628126}, {-0.9999999999999999, 
      0.8198670175812729}, {-0.8999999999999999, 
      0.6832285630841473}, {-0.7999999999999998, 
      0.5130778836938318}, {-0.6999999999999998, 
      0.480924756527288}, {-0.5999999999999999, 
      0.6599518277814447}, {-0.4999999999999999, 
      0.8832701932227283}, {-0.39999999999999986`, 
      1.0372989529894387`}, {-0.2999999999999998, 
      1.0873627322572972`}, {-0.19999999999999984`, 
      1.0528398227088864`}, {-0.09999999999999984, 0.9946699312672349}, {
      0, 0.9793821813312402}, {0.10000000000000017`, 1.0164854335406204`}, {
      0.20000000000000018`, 1.048957193381096}, {0.30000000000000016`, 
      1.0137529750921943`}, {0.4000000000000002, 0.885548270925885}, {
      0.5000000000000002, 0.6898611545618348}, {0.6000000000000002, 
      0.5187476246781831}, {0.7000000000000002, 0.5178742609290009}, {
      0.8000000000000003, 0.6675136810282826}, {0.9000000000000002, 
      0.8178566034526504}, {1.0000000000000002`, 0.890874254226621}, {
      1.1000000000000003`, 0.8652955968642496}, {1.2000000000000002`, 
      0.7495475153366311}, {1.3000000000000003`, 0.5686093094317457}, {
      1.4000000000000004`, 0.3545628143045022}, {1.5000000000000002`, 
      0.13903626106505543`}, {1.6000000000000003`, 0.052104512500728714`}, {
      1.7000000000000002`, 0.20154424940389476`}, {1.8000000000000003`, 
      0.30102792788588206`}, {1.9000000000000004`, 0.3502346286250287}, {
      2.0000000000000004`, 0.3547859379970888}, {2.1, 0.32404217304040006`}, {
      2.2, 0.2690997342750755}, {2.3000000000000003`, 0.20117303329695863`}, {
      2.4000000000000004`, 0.13041431094393705`}, {2.5000000000000004`, 
      0.06530916093518283}, {2.6000000000000005`, 0.01663408619638268}, {2.7, 
      0.0352768311494306}, {2.8000000000000003`, 0.060170576683383346`}, {
      2.9000000000000004`, 0.07228540608982231}, {3.0000000000000004`, 
      0.07329893823339338}}]},
  PlotRange->All,
  PlotRangeClipping->True,
  PlotRangePadding->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{{3.391857601109375*^9, 3.391857608015625*^9}},
 CellLabel->"Out[52]="]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"If", "[", 
   RowBox[{"msgonindet", ",", 
    RowBox[{"On", "[", 
     RowBox[{"Arg", "::", "\"\<indet\>\""}], "]"}]}], "]"}], ";"}]], "Input"]
}, Open  ]]
},
WindowSize->{944, 815},
WindowMargins->{{15, Automatic}, {Automatic, 36}},
Magnification->1.,
FrontEndVersion->"6.0 for Microsoft Windows (32-bit) (April 20, 2007)",
StyleDefinitions->"HelpBrowser.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "VQM`ArgColorPlot`"->{
  Cell[590, 23, 69, 1, 37, "Subsection",
   CellTags->"VQM`ArgColorPlot`"],
  Cell[662, 26, 379, 7, 39, "Text",
   CellTags->"VQM`ArgColorPlot`"],
  Cell[1044, 35, 13565, 316, 531, "Text",
   CellTags->"VQM`ArgColorPlot`"]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"VQM`ArgColorPlot`", 1774520, 46877}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[590, 23, 69, 1, 37, "Subsection",
 CellTags->"VQM`ArgColorPlot`"],
Cell[662, 26, 379, 7, 39, "Text",
 CellTags->"VQM`ArgColorPlot`"],
Cell[1044, 35, 13565, 316, 531, "Text",
 CellTags->"VQM`ArgColorPlot`"],
Cell[14612, 353, 45, 0, 39, "MathCaption"],
Cell[14660, 355, 131, 2, 27, "Input"],
Cell[14794, 359, 41, 0, 39, "MathCaption"],
Cell[14838, 361, 309, 10, 48, "Input"],
Cell[CellGroupData[{
Cell[15172, 375, 416, 13, 42, "Input"],
Cell[15591, 390, 135600, 2885, 245, 131189, 2810, "CachedBoxData", "BoxData", \
"Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[151228, 3280, 941, 27, 70, "Input"],
Cell[152172, 3309, 140338, 2814, 240, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[292547, 6128, 1553, 47, 92, "Input"],
Cell[294103, 6177, 10558, 242, 245, "Output"],
Cell[304664, 6421, 3378, 79, 243, "Output"],
Cell[308045, 6502, 10743, 248, 245, "Output"]
}, Open  ]],
Cell[318803, 6753, 386, 12, 27, "Input"],
Cell[CellGroupData[{
Cell[319214, 6769, 1831, 59, 108, "Input"],
Cell[321048, 6830, 39677, 907, 236, "Output"],
Cell[360728, 7739, 31130, 765, 226, "Output"],
Cell[391861, 8506, 19648, 432, 226, "Output"],
Cell[411512, 8940, 15788, 320, 236, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[427337, 9265, 1378, 40, 51, "Input"],
Cell[428718, 9307, 260502, 5598, 238, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[689257, 14910, 1435, 44, 62, "Input"],
Cell[690695, 14956, 1017016, 30263, 234, 1006060, 30081, "CachedBoxData", \
"BoxData", "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1707748, 45224, 1202, 35, 68, "Input"],
Cell[1708953, 45261, 65034, 1594, 233, "Output"]
}, Open  ]],
Cell[1774002, 46858, 178, 5, 27, "Input"]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
