(* Content-type: application/mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 6.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       145,          7]
NotebookDataLength[    784272,      18024]
NotebookOptionsPosition[    776679,      17803]
NotebookOutlinePosition[    777079,      17821]
CellTagsIndexPosition[    777036,      17818]
WindowFrame->Normal
ContainsDynamic->False*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell[TextData[{
 "The package VQM`FastFourier",
 StyleBox["`", "MB"]
}], "Title"],

Cell["\<\
Bernd Thaller
Institute of Mathematics
University of Graz
Austria
04-10-2004\
\>", "Subsubtitle"],

Cell[CellGroupData[{

Cell["Abstract", "Subsection"],

Cell[TextData[{
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " has a very efficient algorithm for calculating the Fourier transformation \
of a list. We use it for an approximate calculation of the Fourier \
transformation of a function. "
}], "Text"],

Cell[TextData[{
 "This package VQM`FastFourier` provides commands for the numerical Fourier \
transform of complex-valued functions that are given as a list of complex \
values on a suitable grid of space points. This package is part of the VQM \
packages which can be obtained here:  ",
 ButtonBox["http://www.uni-graz.at/imawww/vqm/software.html",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    URL["http://www.uni-graz.at/imawww/vqm/software.html"], None}],
 " (free download). The VQM packages are part of the Visual Quantum Mechanics \
project, see ",
 ButtonBox["http://www.uni-graz.at/imawww/vqm/",
  BaseStyle->"Hyperlink",
  ButtonData:>{
    URL["http://www.uni-graz.at/imawww/vqm/"], None}],
 ". In particular, these packages are distributed with the book ",
 StyleBox["Advanced Visual Quantum Mechanics",
  FontSlant->"Italic"],
 ", Springer-Verlag New York, 2004."
}], "Text"],

Cell[TextData[{
 "This notebook describes the application of ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 "'s fast Fourier algorithm to the Fourier transform of a complex-valued \
function. The package VQM`FastFourier` implements this method. The"
}], "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Fourier transform of a list", "Section"],

Cell[TextData[{
 "The discrete Fourier transform of a list of ",
 StyleBox["n",
  FontSlant->"Italic"],
 " complex numbers ",
 StyleBox["{", "InlineFormula",
  FontFamily->"Times New Roman"],
 Cell[BoxData[
  FormBox[
   SubscriptBox["a", "1"], TraditionalForm]], "InlineFormula",
  FontFamily->"Times New Roman"],
 StyleBox[", ", "InlineFormula",
  FontFamily->"Times New Roman"],
 Cell[BoxData[
  FormBox[
   SubscriptBox["a", "2"], TraditionalForm]], "InlineFormula",
  FontFamily->"Times New Roman"],
 StyleBox[", ... , ", "InlineFormula",
  FontFamily->"Times New Roman"],
 Cell[BoxData[
  FormBox[
   SubscriptBox["a", "n"], TraditionalForm]], "InlineFormula",
  FontFamily->"Times New Roman"],
 StyleBox["}", "InlineFormula",
  FontFamily->"Times New Roman"],
 " is again a list of ",
 StyleBox["n",
  FontSlant->"Italic"],
 " complex numbers which is defined by"
}], "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    StyleBox[
     RowBox[{"(", "1", ")"}],
     FontFamily->"Times New Roman"], 
    StyleBox[
     RowBox[{"       ", 
      StyleBox[" ", "DisplayFormula"]}]], 
    StyleBox[
     SubscriptBox[
      OverscriptBox[
       StyleBox["a",
        FontSlant->"Italic"], "\[Wedge]"], 
      StyleBox["s",
       FontSlant->"Italic"]], "DisplayFormula",
     FontFamily->"Times New Roman"]}], 
   StyleBox[" ", "DisplayFormula",
    FontFamily->"Times New Roman"], 
   StyleBox["=", "DisplayFormula",
    FontFamily->"Times New Roman"], 
   StyleBox[" ", "DisplayFormula",
    FontFamily->"Times New Roman"], 
   StyleBox[
    RowBox[{
     StyleBox[
      FractionBox["1", 
       SqrtBox[
        StyleBox["n",
         FontSlant->"Italic"]]], "DisplayFormula"], 
     RowBox[{
      StyleBox[
       UnderoverscriptBox[
        StyleBox["\[Sum]",
         FontWeight->"Plain",
         FontSlant->"Plain",
         FontTracking->"Plain",
         FontVariations->{"CompatibilityType"->0,
         "Masked"->False,
         "Outline"->False,
         "RotationAngle"->0,
         "Shadow"->False,
         "StrikeThrough"->False,
         "Underline"->False}], 
        RowBox[{"r", "=", "1"}], "n"], "DisplayFormula",
       FontSize->14], 
      RowBox[{
       StyleBox[
        SubscriptBox[
         StyleBox["a",
          FontSlant->"Italic"], 
         StyleBox["r",
          FontSlant->"Italic"]], "DisplayFormula"], 
       RowBox[{
        StyleBox["Exp", "InlineFormula",
         FontWeight->"Plain",
         FontSlant->"Plain",
         FontTracking->"Plain",
         FontVariations->{"CompatibilityType"->0,
         "Masked"->False,
         "Outline"->False,
         "RotationAngle"->0,
         "Shadow"->False,
         "StrikeThrough"->False,
         "Underline"->False}], 
        StyleBox["[", "InlineFormula",
         FontWeight->"Plain",
         FontSlant->"Plain",
         FontTracking->"Plain",
         FontVariations->{"CompatibilityType"->0,
         "Masked"->False,
         "Outline"->False,
         "RotationAngle"->0,
         "Shadow"->False,
         "StrikeThrough"->False,
         "Underline"->False}], 
        RowBox[{
         RowBox[{"-", "\[ImaginaryI]"}], 
         StyleBox[
          FractionBox[
           RowBox[{
            StyleBox["2",
             FontWeight->"Plain",
             FontSlant->"Plain",
             FontTracking->"Plain",
             FontVariations->{"Outline"->False,
             "Shadow"->False,
             "Underline"->False}], "\[Pi]"}], 
           RowBox[{"n", " "}]],
          FontSize->10,
          FontSlant->"Italic"], 
         RowBox[{
          StyleBox["(",
           FontSlant->"Italic"], 
          RowBox[{
           StyleBox["s",
            FontSlant->"Italic"], 
           StyleBox["-",
            FontSlant->"Italic"], 
           StyleBox["1",
            FontWeight->"Plain",
            FontSlant->"Plain",
            FontTracking->"Plain",
            FontVariations->{"Outline"->False,
            "Shadow"->False,
            "Underline"->False}]}], 
          StyleBox[")",
           FontSlant->"Italic"]}], 
         RowBox[{
          StyleBox["(",
           FontSlant->"Italic"], 
          RowBox[{
           StyleBox["r",
            FontSlant->"Italic"], 
           StyleBox["-",
            FontSlant->"Italic"], 
           StyleBox["1",
            FontWeight->"Plain",
            FontSlant->"Plain",
            FontTracking->"Plain",
            FontVariations->{"Outline"->False,
            "Shadow"->False,
            "Underline"->False}]}], 
          StyleBox[")",
           FontSlant->"Italic"]}]}], "]"}]}]}]}],
    FontFamily->"Times New Roman"]}], 
  StyleBox[",", "DisplayFormula",
   FontFamily->"Times New Roman"], 
  StyleBox["    ", "DisplayFormula",
   FontFamily->"Times New Roman"], 
  StyleBox[
   RowBox[{
    StyleBox["s",
     FontSlant->"Italic"], 
    StyleBox[" ", "DisplayFormula"], 
    StyleBox["=", "DisplayFormula"], 
    StyleBox[" ", "DisplayFormula"], 
    StyleBox["1", "DisplayFormula"]}],
   FontFamily->"Times New Roman"], 
  StyleBox[",", "DisplayFormula",
   FontFamily->"Times New Roman"], 
  StyleBox[" ", "DisplayFormula",
   FontFamily->"Times New Roman"], 
  StyleBox["...", "DisplayFormula",
   FontFamily->"Times New Roman"], 
  StyleBox[" ", "DisplayFormula",
   FontFamily->"Times New Roman"], 
  StyleBox[",", "DisplayFormula",
   FontFamily->"Times New Roman"], 
  StyleBox[" ", "DisplayFormula",
   FontFamily->"Times New Roman"], 
  StyleBox[
   RowBox[{
    StyleBox["n",
     FontSlant->"Italic"], 
    StyleBox[" ", "DisplayFormula"], 
    StyleBox[".", "DisplayFormula"]}],
   FontFamily->"Times New Roman"]}]], "DisplayFormula"],

Cell[TextData[{
 "In ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 ", this operation is implemented by the command ",
 StyleBox["InverseFourier", "InlineInput"],
 ":"
}], "Text"],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"{", 
    RowBox[{
     SubscriptBox[
      OverscriptBox[
       StyleBox["a",
        FontSlant->"Italic"], "\[Wedge]"], "1"], ",", "...", " ", ",", " ", 
     SubscriptBox[
      OverscriptBox[
       StyleBox["a",
        FontSlant->"Italic"], "\[Wedge]"], 
      StyleBox["n",
       FontSlant->"Italic"]]}], "}"}], "DisplayFormula",
   FontFamily->"Times New Roman"], 
  StyleBox[" ", "DisplayFormula",
   FontFamily->"Times New Roman"], 
  StyleBox["=", "DisplayFormula",
   FontFamily->"Times New Roman"], 
  StyleBox[" ", "DisplayFormula",
   FontFamily->"Times New Roman"], 
  RowBox[{
   StyleBox["InverseFourier", "InlineInput"], 
   StyleBox["[", "Input~",
    FontFamily->"Times New Roman"], 
   StyleBox[
    RowBox[{"{", 
     RowBox[{
      SubscriptBox[
       StyleBox["a",
        FontSlant->"Italic"], "1"], ",", "...", " ", ",", " ", 
      SubscriptBox[
       StyleBox["a",
        FontSlant->"Italic"], 
       StyleBox["n",
        FontSlant->"Italic"]]}], "}"}], "DisplayFormula",
    FontFamily->"Times New Roman"], 
   StyleBox["]", "Inputz",
    FontFamily->"Times New Roman"]}]}]], "DisplayFormula"],

Cell[TextData[{
 "As the name suggests this function is inverse to ",
 StyleBox["Fourier[]", "InlineInput"],
 ", which is defined by a reversed sign in the exponent. (The name is only a \
matter of convention. For the purpose of quantum mechanics, the definition \
with a minus in the exponent is more common)."
}], "Text"],

Cell[TextData[{
 "The formula (1) has a certain symmetry. If we multiply the whole equation \
(1) by the ",
 StyleBox["s",
  FontSlant->"Italic"],
 "-dependent phase factor ",
 Cell[BoxData[
  RowBox[{
   StyleBox["Exp", "InlineFormula",
    FontFamily->"Times New Roman",
    FontVariations->{"CompatibilityType"->0}], 
   StyleBox["[", "InlineFormula",
    FontFamily->"Times New Roman",
    FontVariations->{"CompatibilityType"->0}], 
   RowBox[{
    StyleBox["2",
     FontFamily->"Times New Roman",
     FontSize->10], 
    StyleBox["\[Pi]",
     FontFamily->"Times New Roman",
     FontSize->10], 
    StyleBox[" ",
     FontFamily->"Times New Roman",
     FontSize->10], "\[ImaginaryI]", 
    StyleBox[
     FractionBox[
      RowBox[{
       StyleBox["m",
        FontSlant->"Italic"], 
       RowBox[{"(", 
        RowBox[{
         StyleBox["s",
          FontSlant->"Italic"], "-", "1"}], ")"}]}], 
      StyleBox["n",
       FontSlant->"Italic"]], "Text",
     FontFamily->"Times New Roman",
     FontSize->10]}], 
   StyleBox["]", "InlineFormula",
    FontFamily->"Times New Roman"]}]]],
 ", this is the same as a cyclic permutation of the coefficients ",
 Cell[BoxData[
  StyleBox[
   SubscriptBox["a", "r"], "InlineFormula",
   FontFamily->"Times New Roman"]]],
 ":"
}], "Text"],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{
    StyleBox[
     SubscriptBox[
      OverscriptBox[
       StyleBox["a",
        FontWeight->"Plain",
        FontSlant->"Italic",
        FontTracking->"Plain",
        FontVariations->{"CompatibilityType"->0,
        "Masked"->False,
        "Outline"->False,
        "RotationAngle"->0,
        "Shadow"->False,
        "StrikeThrough"->False,
        "Underline"->False}], "\[Wedge]"], "s"], "DisplayFormula",
     FontWeight->"Plain"], 
    RowBox[{
     StyleBox["Exp", "InlineFormula",
      FontWeight->"Plain",
      FontVariations->{"CompatibilityType"->0}], 
     StyleBox["[", "InlineFormula",
      FontWeight->"Plain",
      FontVariations->{"CompatibilityType"->0}], 
     RowBox[{
      StyleBox["2",
       FontSize->10,
       FontWeight->"Plain"], 
      StyleBox["\[Pi]",
       FontSize->10,
       FontWeight->"Plain"], " ", 
      StyleBox["\[ImaginaryI]", "DisplayFormula"], 
      StyleBox[
       FractionBox[
        RowBox[{
         StyleBox["m",
          FontSlant->"Italic"], 
         RowBox[{"(", 
          RowBox[{
           StyleBox["s",
            FontSlant->"Italic"], "-", "1"}], ")"}]}], 
        StyleBox["n",
         FontSlant->"Italic"]], "Text",
       FontSize->10,
       FontWeight->"Plain"]}], 
     StyleBox["]", "InlineFormula",
      FontWeight->"Plain"]}]}],
   FontFamily->"Times New Roman"], 
  StyleBox["=", "DisplayFormula",
   FontFamily->"Times New Roman",
   FontWeight->"Plain",
   FontVariations->{"CompatibilityType"->0}], 
  StyleBox[" ", "DisplayFormula",
   FontFamily->"Times New Roman",
   FontWeight->"Plain",
   FontVariations->{"CompatibilityType"->0}], 
  RowBox[{
   RowBox[{
    StyleBox[
     FractionBox["1", 
      SqrtBox[
       StyleBox["n",
        FontSlant->"Italic"]]],
     FontFamily->"Times New Roman",
     FontWeight->"Plain",
     FontSlant->"Plain",
     FontTracking->"Plain",
     FontVariations->{"CompatibilityType"->0,
     "Masked"->False,
     "Outline"->False,
     "RotationAngle"->0,
     "Shadow"->False,
     "StrikeThrough"->False,
     "Underline"->False}], 
    RowBox[{
     StyleBox[
      UnderoverscriptBox[
       StyleBox["\[Sum]",
        FontWeight->"Plain",
        FontSlant->"Plain",
        FontTracking->"Plain",
        FontVariations->{"CompatibilityType"->0,
        "Masked"->False,
        "Outline"->False,
        "RotationAngle"->0,
        "Shadow"->False,
        "StrikeThrough"->False,
        "Underline"->False}], 
       RowBox[{"r", "=", "1"}], "n"], "DisplayFormula",
      FontFamily->"Times New Roman",
      FontWeight->"Plain",
      FontSlant->"Plain",
      FontTracking->"Plain",
      FontVariations->{"Outline"->False,
      "Shadow"->False,
      "Underline"->False}], 
     RowBox[{
      StyleBox[
       SubscriptBox[
        StyleBox["a",
         FontWeight->"Plain",
         FontSlant->"Italic",
         FontTracking->"Plain",
         FontVariations->{"CompatibilityType"->0,
         "Masked"->False,
         "Outline"->False,
         "RotationAngle"->0,
         "Shadow"->False,
         "StrikeThrough"->False,
         "Underline"->False}], "r"], "DisplayFormula",
       FontFamily->"Times New Roman",
       FontWeight->"Plain",
       FontSlant->"Plain",
       FontTracking->"Plain",
       FontVariations->{"Outline"->False,
       "Shadow"->False,
       "Underline"->False}], 
      RowBox[{
       StyleBox["Exp", "InlineFormula",
        FontFamily->"Times New Roman",
        FontWeight->"Plain",
        FontSlant->"Plain",
        FontTracking->"Plain",
        FontVariations->{"CompatibilityType"->0,
        "Masked"->False,
        "Outline"->False,
        "RotationAngle"->0,
        "Shadow"->False,
        "StrikeThrough"->False,
        "Underline"->False}], 
       StyleBox["[", "InlineFormula",
        FontFamily->"Times New Roman",
        FontWeight->"Plain",
        FontSlant->"Plain",
        FontTracking->"Plain",
        FontVariations->{"CompatibilityType"->0,
        "Masked"->False,
        "Outline"->False,
        "RotationAngle"->0,
        "Shadow"->False,
        "StrikeThrough"->False,
        "Underline"->False}], 
       RowBox[{
        RowBox[{"-", "2"}], "\[Pi]", 
        StyleBox[" ",
         FontFamily->"Times New Roman"], "\[ImaginaryI]", 
        StyleBox[
         FractionBox[
          RowBox[{
           RowBox[{"(", 
            RowBox[{"r", "-", "1", "-", "m"}], ")"}], 
           RowBox[{"(", 
            RowBox[{"s", "-", "1"}], ")"}]}], "n"],
         FontSize->10]}], "]"}]}]}]}], 
   StyleBox["=",
    FontFamily->"Times New Roman",
    FontWeight->"Plain",
    FontSlant->"Plain",
    FontTracking->"Plain",
    FontVariations->{"CompatibilityType"->0,
    "Masked"->False,
    "Outline"->False,
    "RotationAngle"->0,
    "Shadow"->False,
    "StrikeThrough"->False,
    "Underline"->False}], 
   StyleBox[" ",
    FontFamily->"Times New Roman",
    FontWeight->"Plain",
    FontSlant->"Plain",
    FontTracking->"Plain",
    FontVariations->{"CompatibilityType"->0,
    "Masked"->False,
    "Outline"->False,
    "RotationAngle"->0,
    "Shadow"->False,
    "StrikeThrough"->False,
    "Underline"->False}], 
   StyleBox[
    RowBox[{
     StyleBox[
      FractionBox["1", 
       SqrtBox["n"]],
      FontWeight->"Plain",
      FontSlant->"Plain",
      FontTracking->"Plain",
      FontVariations->{"CompatibilityType"->0,
      "Masked"->False,
      "Outline"->False,
      "RotationAngle"->0,
      "Shadow"->False,
      "StrikeThrough"->False,
      "Underline"->False}], 
     RowBox[{
      StyleBox[
       UnderoverscriptBox[
        StyleBox["\[Sum]",
         FontWeight->"Plain",
         FontSlant->"Plain",
         FontTracking->"Plain",
         FontVariations->{"CompatibilityType"->0,
         "Masked"->False,
         "Outline"->False,
         "RotationAngle"->0,
         "Shadow"->False,
         "StrikeThrough"->False,
         "Underline"->False}], 
        RowBox[{"r", "=", "1"}], "n"], "DisplayFormula",
       FontWeight->"Plain",
       FontSlant->"Plain",
       FontTracking->"Plain",
       FontVariations->{"Outline"->False,
       "Shadow"->False,
       "Underline"->False}], 
      RowBox[{
       StyleBox[
        SubscriptBox[
         StyleBox["b",
          FontWeight->"Plain",
          FontSlant->"Italic",
          FontTracking->"Plain",
          FontVariations->{"CompatibilityType"->0,
          "Masked"->False,
          "Outline"->False,
          "RotationAngle"->0,
          "Shadow"->False,
          "StrikeThrough"->False,
          "Underline"->False}], "r"], "DisplayFormula",
        FontWeight->"Plain",
        FontSlant->"Plain",
        FontTracking->"Plain",
        FontVariations->{"Outline"->False,
        "Shadow"->False,
        "Underline"->False}], 
       RowBox[{
        StyleBox["Exp", "InlineFormula",
         FontWeight->"Plain",
         FontSlant->"Plain",
         FontTracking->"Plain",
         FontVariations->{"CompatibilityType"->0,
         "Masked"->False,
         "Outline"->False,
         "RotationAngle"->0,
         "Shadow"->False,
         "StrikeThrough"->False,
         "Underline"->False}], 
        StyleBox["[", "InlineFormula",
         FontWeight->"Plain",
         FontSlant->"Plain",
         FontTracking->"Plain",
         FontVariations->{"CompatibilityType"->0,
         "Masked"->False,
         "Outline"->False,
         "RotationAngle"->0,
         "Shadow"->False,
         "StrikeThrough"->False,
         "Underline"->False}], 
        RowBox[{
         RowBox[{"-", "\[ImaginaryI]"}], 
         StyleBox[
          FractionBox[
           RowBox[{
            StyleBox["2",
             FontWeight->"Plain",
             FontSlant->"Plain",
             FontTracking->"Plain",
             FontVariations->{"Outline"->False,
             "Shadow"->False,
             "Underline"->False}], "\[Pi]"}], 
           RowBox[{"n", " "}]],
          FontSize->10,
          FontSlant->"Italic"], 
         RowBox[{
          StyleBox["(",
           FontSlant->"Italic"], 
          RowBox[{
           StyleBox["s",
            FontSlant->"Italic"], 
           StyleBox["-",
            FontSlant->"Italic"], 
           StyleBox["1",
            FontWeight->"Plain",
            FontSlant->"Plain",
            FontTracking->"Plain",
            FontVariations->{"Outline"->False,
            "Shadow"->False,
            "Underline"->False}]}], 
          StyleBox[")",
           FontSlant->"Italic"]}], 
         RowBox[{
          StyleBox["(",
           FontSlant->"Italic"], 
          RowBox[{
           StyleBox["r",
            FontSlant->"Italic"], 
           StyleBox["-",
            FontSlant->"Italic"], 
           StyleBox["1",
            FontWeight->"Plain",
            FontSlant->"Plain",
            FontTracking->"Plain",
            FontVariations->{"Outline"->False,
            "Shadow"->False,
            "Underline"->False}]}], 
          StyleBox[")",
           FontSlant->"Italic"]}]}], "]"}]}]}]}],
    FontFamily->"Times New Roman"]}]}]], "DisplayFormula",
 FontFamily->"Times New Roman"],

Cell[TextData[{
 "where ",
 Cell[BoxData[
  StyleBox[
   SubscriptBox[
    StyleBox["b",
     FontFamily->"Times New Roman",
     FontSlant->"Italic",
     FontVariations->{"CompatibilityType"->0}], 
    RowBox[{"r", " "}]],
   FontFamily->"Times New Roman"]], "InlineFormula"],
 StyleBox["= ", "InlineFormula"],
 Cell[BoxData[
  StyleBox[
   SubscriptBox[
    StyleBox["a",
     FontFamily->"Times New Roman",
     FontSlant->"Italic",
     FontVariations->{"CompatibilityType"->0}], 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"r", "+", "m"}], ")"}], 
     RowBox[{"(", 
      RowBox[{"mod", " ", "n"}], ")"}]}]],
   FontFamily->"Times New Roman"]], "InlineFormula"],
 StyleBox[", or", "InlineFormula"]
}], "Text"],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"{", 
    RowBox[{
     SubscriptBox["b", "1"], ",", "...", " ", ",", 
     SubscriptBox["b", "n"]}], "}"}],
   FontFamily->"Times New Roman"], 
  StyleBox[" ",
   FontFamily->"Times New Roman"], 
  StyleBox["=",
   FontFamily->"Times New Roman"], " ", 
  RowBox[{
   RowBox[{
    StyleBox["RotateLeft",
     FontWeight->"Bold"], 
    StyleBox["[",
     FontFamily->"Times New Roman"], 
    StyleBox[
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        SubscriptBox["a", "1"], ",", "...", " ", ",", 
        SubscriptBox["a", "n"]}], "}"}], ",", "m"}],
     FontFamily->"Times New Roman"], 
    StyleBox["]",
     FontFamily->"Times New Roman"]}], "."}]}]], "DisplayFormula"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Fourier transform of a function", "Section"],

Cell[TextData[{
 "For an integrable function \[Psi]",
 StyleBox["(",
  FontFamily->"Times New Roman"],
 StyleBox["x",
  FontFamily->"Times New Roman",
  FontSlant->"Italic"],
 StyleBox[")",
  FontFamily->"Times New Roman"],
 ", the Fourier transform is given by"
}], "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"(", "2", ")"}], "        ", 
   OverscriptBox["\[Psi]", "\[Wedge]"], 
   RowBox[{"(", 
    StyleBox["k",
     FontSlant->"Italic"], ")"}]}], " ", "=", " ", 
  RowBox[{
   FractionBox["1", 
    SqrtBox[
     RowBox[{"2", "\[Pi]"}]]], 
   RowBox[{
    SubsuperscriptBox["\[Integral]", 
     RowBox[{"-", "\[Infinity]"}], "\[Infinity]"], 
    RowBox[{"\[Psi]", 
     RowBox[{"(", 
      StyleBox["x",
       FontSlant->"Italic"], ")"}], "  ", 
     SuperscriptBox["e", 
      RowBox[{
       RowBox[{"-", "\[ImaginaryI]"}], " ", 
       StyleBox["k",
        FontSlant->"Italic"], 
       StyleBox[" ",
        FontSlant->"Italic"], 
       StyleBox["x",
        FontSlant->"Italic"]}]], 
     RowBox[{
      RowBox[{"\[DifferentialD]", 
       StyleBox["x",
        FontSlant->"Italic"]}], 
      StyleBox[".",
       FontSlant->"Italic"]}]}]}]}]}]], "DisplayFormula",
 FontFamily->"Times New Roman"],

Cell[TextData[{
 "We may approximate this integral by a finite Riemann sum. To this purpose \
we assume that the function ",
 StyleBox["\[Psi]",
  FontFamily->"Times New Roman"],
 " is reasonably well localized within a finite interval ",
 StyleBox["[",
  FontFamily->"Times New Roman"],
 Cell[BoxData[
  FormBox[
   SubscriptBox["x", "0"], TraditionalForm]],
  FontFamily->"Times New Roman"],
 StyleBox[", ",
  FontFamily->"Times New Roman"],
 Cell[BoxData[
  FormBox[
   SubscriptBox["x", "n"], TraditionalForm]],
  FontFamily->"Times New Roman"],
 StyleBox["]",
  FontFamily->"Times New Roman"],
 ". We divide this interval into ",
 StyleBox["n",
  FontSlant->"Italic"],
 " subintervals by introducing equally spaced points ",
 Cell[BoxData[
  FormBox[
   StyleBox[
    SubscriptBox["x", "r"],
    FontFamily->"Times New Roman"], TraditionalForm]]],
 ","
}], "Text"],

Cell[BoxData[
 StyleBox[
  RowBox[{
   RowBox[{
    RowBox[{
     SubscriptBox[
      StyleBox["x",
       FontFamily->"Times New Roman",
       FontSlant->"Italic"], "0"], " ", "<", " ", 
     SubscriptBox[
      StyleBox["x",
       FontSlant->"Italic"], 
      RowBox[{"1", " "}]], "<", " ", 
     SubscriptBox[
      StyleBox["x",
       FontSlant->"Italic"], 
      RowBox[{"2", " "}]], "<"}], " ", "..."}], " ", "<", " ", 
   SubscriptBox[
    StyleBox["x",
     FontSlant->"Italic"], "n"]}],
  FontFamily->"Times New Roman"]], "DisplayFormula",
 FontFamily->"Times New Roman"],

Cell["and define", "Text"],

Cell[TextData[{
 Cell[BoxData[
  RowBox[{
   StyleBox[
    RowBox[{"d", 
     StyleBox["x",
      FontSlant->"Italic"]}]], 
   StyleBox[" ",
    FontSlant->"Italic"]}]],
  FontFamily->"Times New Roman"],
 "= ",
 Cell[BoxData[
  RowBox[{
   SubscriptBox[
    StyleBox["x",
     FontSlant->"Italic"], "r"], " "}]],
  FontFamily->"Times New Roman"],
 "-",
 Cell[BoxData[
  RowBox[{" ", 
   SubscriptBox[
    StyleBox["x",
     FontSlant->"Italic"], 
    RowBox[{"r", "-", "1"}]], " "}]],
  FontFamily->"Times New Roman"],
 "."
}], "DisplayFormula",
 FontFamily->"Times New Roman"],

Cell["\<\
The Riemann sum approximating the Fourier integral is then given by\
\>", "Text"],

Cell[BoxData[
 StyleBox[
  RowBox[{
   RowBox[{
    RowBox[{"(", "3", ")"}], "        ", 
    OverscriptBox["\[Psi]", "\[Wedge]"], 
    RowBox[{"(", 
     StyleBox["k",
      FontSlant->"Italic"], ")"}]}], " ", "\[TildeTilde]", " ", 
   RowBox[{
    FractionBox["1", 
     SqrtBox[
      RowBox[{"2", "\[Pi]"}]]], " ", 
    RowBox[{
     UnderoverscriptBox["\[Sum]", 
      RowBox[{"r", "=", "1"}], "n"], 
     RowBox[{"\[Psi]", 
      RowBox[{"(", 
       StyleBox[
        SubscriptBox["x", "r"],
        FontSlant->"Italic"], ")"}], 
      SuperscriptBox["e", 
       RowBox[{
        RowBox[{"-", "\[ImaginaryI]"}], " ", "k", " ", 
        StyleBox[
         SubscriptBox["x", "r"],
         FontSlant->"Italic"]}]], 
      RowBox[{"d", 
       StyleBox["x",
        FontSlant->"Italic"]}]}]}]}]}],
  FontFamily->"Times New Roman"]], "DisplayFormula",
 FontFamily->"Times New Roman"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Applying the command InverseFourier", "Section"],

Cell["This sum (3) looks already similar to (1). We set", "Text"],

Cell[BoxData[
 RowBox[{
  SubscriptBox[
   StyleBox["k",
    FontSlant->"Italic"], "s"], " ", "=", " ", 
  StyleBox[
   RowBox[{
    RowBox[{
     FractionBox[
      RowBox[{"2", "\[Pi]"}], 
      RowBox[{"n", " ", 
       StyleBox[
        RowBox[{"d", 
         StyleBox["x",
          FontSlant->"Italic"]}]]}]], 
     RowBox[{"(", 
      RowBox[{"s", "-", "1"}], ")"}], "  ", 
     StyleBox["and", "Text"], "  ", 
     SubscriptBox[
      StyleBox["x",
       FontSlant->"Italic"], "r"]}], " ", "=", " ", 
    RowBox[{
     SubscriptBox[
      StyleBox["x",
       FontSlant->"Italic"], "1"], " ", "+", " ", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"r", "-", "1"}], ")"}], 
      RowBox[{"d", 
       StyleBox["x",
        FontSlant->"Italic"]}]}]}]}],
   FontFamily->"Times New Roman"]}]], "DisplayFormula",
 FontFamily->"Times New Roman",
 FontWeight->"Plain"],

Cell["and obtain", "Text"],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{
    RowBox[{"(", "4", ")"}], "        ", 
    OverscriptBox["\[Psi]", "\[Wedge]"], 
    RowBox[{"(", 
     StyleBox[
      SubscriptBox["k", "s"],
      FontSlant->"Italic"], ")"}]}],
   FontFamily->"Times New Roman"], 
  StyleBox[" ",
   FontFamily->"Times New Roman"], 
  StyleBox["\[TildeTilde]",
   FontFamily->"Times New Roman"], 
  StyleBox[" ",
   FontFamily->"Times New Roman"], 
  RowBox[{
   StyleBox[
    FractionBox["1", 
     SqrtBox[
      RowBox[{"2", "\[Pi]"}]]],
    FontFamily->"Times New Roman"], 
   StyleBox[" ",
    FontFamily->"Times New Roman"], 
   RowBox[{
    StyleBox[
     UnderoverscriptBox["\[Sum]", 
      RowBox[{"r", "=", "1"}], "n"],
     FontFamily->"Times New Roman"], 
    RowBox[{
     StyleBox["\[Psi]",
      FontFamily->"Times New Roman"], 
     StyleBox[
      RowBox[{"(", 
       StyleBox[
        SubscriptBox["x", "r"],
        FontSlant->"Italic"], ")"}],
      FontFamily->"Times New Roman"], 
     StyleBox[
      SuperscriptBox["e", 
       RowBox[{
        RowBox[{"-", "\[ImaginaryI]"}], " ", 
        StyleBox[
         SubscriptBox["k", "s"],
         FontSlant->"Italic"], 
        StyleBox[
         SubscriptBox["x", 
          StyleBox["1",
           FontSlant->"Plain"]],
         FontSlant->"Italic"]}]],
      FontFamily->"Times New Roman"], 
     RowBox[{
      StyleBox["Exp", "InlineFormula",
       FontFamily->"Times New Roman",
       FontWeight->"Plain",
       FontSlant->"Plain",
       FontTracking->"Plain",
       FontVariations->{"CompatibilityType"->0,
       "Masked"->False,
       "Outline"->False,
       "RotationAngle"->0,
       "Shadow"->False,
       "StrikeThrough"->False,
       "Underline"->False}], 
      StyleBox["[", "InlineFormula",
       FontFamily->"Times New Roman",
       FontWeight->"Plain",
       FontSlant->"Plain",
       FontTracking->"Plain",
       FontVariations->{"CompatibilityType"->0,
       "Masked"->False,
       "Outline"->False,
       "RotationAngle"->0,
       "Shadow"->False,
       "StrikeThrough"->False,
       "Underline"->False}], 
      RowBox[{
       RowBox[{"-", "\[ImaginaryI]"}], 
       StyleBox[
        FractionBox[
         RowBox[{
          StyleBox["2",
           FontWeight->"Plain",
           FontSlant->"Plain",
           FontTracking->"Plain",
           FontVariations->{"Outline"->False,
           "Shadow"->False,
           "Underline"->False}], "\[Pi]"}], 
         RowBox[{"n", " "}]],
        FontSize->10,
        FontSlant->"Italic"], 
       RowBox[{
        StyleBox["(",
         FontSlant->"Italic"], 
        RowBox[{
         StyleBox["s",
          FontSlant->"Italic"], 
         StyleBox["-",
          FontSlant->"Italic"], 
         StyleBox["1",
          FontWeight->"Plain",
          FontSlant->"Plain",
          FontTracking->"Plain",
          FontVariations->{"Outline"->False,
          "Shadow"->False,
          "Underline"->False}]}], 
        StyleBox[")",
         FontSlant->"Italic"]}], 
       RowBox[{
        StyleBox["(",
         FontSlant->"Italic"], 
        RowBox[{
         StyleBox["r",
          FontSlant->"Italic"], 
         StyleBox["-",
          FontSlant->"Italic"], 
         StyleBox["1",
          FontWeight->"Plain",
          FontSlant->"Plain",
          FontTracking->"Plain",
          FontVariations->{"Outline"->False,
          "Shadow"->False,
          "Underline"->False}]}], 
        StyleBox[")",
         FontSlant->"Italic"]}]}], "]"}], " ", 
     StyleBox[
      RowBox[{"d", 
       StyleBox["x",
        FontSlant->"Italic"]}],
      FontFamily->"Times New Roman"]}]}]}]}]], "DisplayFormula",
 FontFamily->"Times New Roman"],

Cell[TextData[{
 "We see that the built-in discrete Fourier transform is obviously suitable \
to calculate (a numerical approximation of) the Fourier integral ",
 Cell[BoxData[
  RowBox[{
   OverscriptBox["\[Psi]", "\[Wedge]"], 
   RowBox[{"(", 
    StyleBox["k",
     FontSlant->"Italic"], ")"}]}]]],
 "at the points ",
 Cell[BoxData[
  SubscriptBox[
   StyleBox["k",
    FontWeight->"Plain",
    FontSlant->"Italic"], "s"]]],
 " which are given by "
}], "Text"],

Cell[TextData[{
 "{",
 Cell[BoxData[
  FormBox[
   SubscriptBox["k", "1"], TraditionalForm]], "InlineInput",
  FontWeight->"Plain"],
 ",",
 StyleBox[" ", "InlineInput"],
 StyleBox[Cell[BoxData[
  FormBox[
   SubscriptBox["k", "2"], TraditionalForm]], "InlineInput",
  FontWeight->"Plain"], "InlineInput"],
 ",",
 StyleBox[" ", "InlineInput"],
 StyleBox[Cell[BoxData[
  FormBox[
   SubscriptBox["k", "3"], TraditionalForm]], "InlineInput",
  FontWeight->"Plain"], "InlineInput"],
 ", ... ,",
 StyleBox[" ", "InlineInput"],
 StyleBox[Cell[BoxData[
  FormBox[
   SubscriptBox["k", "n"], TraditionalForm]], "InlineInput",
  FontWeight->"Plain"], "InlineInput"],
 "} = {0,",
 StyleBox[" ", "InlineInput"],
 StyleBox[Cell[BoxData[
  FormBox[
   FractionBox[
    RowBox[{"2", "\[Pi]"}], 
    RowBox[{"n", " ", 
     StyleBox[
      RowBox[{"d", 
       StyleBox["x",
        FontSlant->"Italic"]}]]}]], TraditionalForm]], "InlineInput",
  FontWeight->"Plain"], "InlineInput"],
 ",",
 StyleBox[" ", "InlineInput"],
 StyleBox[Cell[BoxData[
  FormBox[
   FractionBox[
    RowBox[{"4", "\[Pi]"}], 
    RowBox[{"n", " ", 
     StyleBox[
      RowBox[{"d", 
       StyleBox["x",
        FontSlant->"Italic"]}]]}]], TraditionalForm]], "InlineInput",
  FontWeight->"Plain"], "InlineInput"],
 ", ... ,",
 StyleBox[" ", "InlineInput"],
 StyleBox[Cell[BoxData[
  FormBox[
   FractionBox[
    RowBox[{"2", 
     RowBox[{"(", 
      RowBox[{"n", "-", "1"}], ")"}], "\[Pi]"}], 
    RowBox[{"n", " ", 
     StyleBox[
      RowBox[{"d", 
       StyleBox["x",
        FontSlant->"Italic"]}]]}]], TraditionalForm]], "InlineInput",
  FontWeight->"Plain"], "InlineInput"],
 "} ,"
}], "DisplayFormula",
 FontFamily->"Times New Roman"],

Cell[TextData[{
 "These points form a grid in ",
 StyleBox["k",
  FontSlant->"Italic"],
 "-space where the distance between neighbouring points is"
}], "Text"],

Cell[TextData[{
 Cell[BoxData[
  StyleBox[
   RowBox[{"d", 
    StyleBox["k",
     FontSlant->"Italic"]}]]],
  FontFamily->"Times New Roman"],
 " = ",
 Cell[BoxData[
  FormBox[
   FractionBox[
    RowBox[{"2", "\[Pi]"}], 
    RowBox[{"n", " ", 
     StyleBox[
      RowBox[{"d", 
       StyleBox["x",
        FontSlant->"Italic"]}]]}]], TraditionalForm]]],
 "= ",
 Cell[BoxData[
  FormBox[
   FractionBox[
    RowBox[{"2", "\[Pi]"}], 
    StyleBox[
     RowBox[{
      SubscriptBox["x", 
       StyleBox["n",
        FontSlant->"Italic"]], "-", 
      SubscriptBox["x", 
       StyleBox["0",
        FontSlant->"Plain"]]}],
     FontSlant->"Italic"]], TraditionalForm]]]
}], "DisplayFormula",
 FontFamily->"Times New Roman"],

Cell[TextData[{
 "and the length of the interval in Fourier space is ",
 StyleBox[" ", "InlineFormula"],
 StyleBox["n", "InlineFormula",
  FontFamily->"Times New Roman",
  FontSlant->"Italic"],
 StyleBox[" d", "InlineFormula",
  FontFamily->"Times New Roman"],
 StyleBox["k", "InlineFormula",
  FontFamily->"Times New Roman",
  FontSlant->"Italic"],
 StyleBox[" = 2\[Pi]/d", "InlineFormula",
  FontFamily->"Times New Roman"],
 StyleBox["x", "InlineFormula",
  FontFamily->"Times New Roman",
  FontSlant->"Italic"],
 ". It depends on the distance ",
 StyleBox["d", "InlineFormula",
  FontFamily->"Times New Roman"],
 StyleBox["x", "InlineFormula",
  FontFamily->"Times New Roman",
  FontSlant->"Italic"],
 " between the grid points in ",
 StyleBox["x",
  FontSlant->"Italic"],
 "-space."
}], "Text"],

Cell[TextData[{
 "Eq. (4) can be described as the (inverse) Fourier transform of a list, \
except that the result is multiplied by an ",
 StyleBox["s",
  FontSlant->"Italic"],
 "-dependent phase factor ",
 Cell[BoxData[
  StyleBox[
   SuperscriptBox["e", 
    RowBox[{"\[ImaginaryI]", " ", 
     StyleBox[
      SubscriptBox["k", "s"],
      FontSlant->"Italic"], 
     StyleBox[
      SubscriptBox["x", 
       StyleBox["1",
        FontSize->7]],
      FontSlant->"Italic"]}]],
   FontFamily->"Times New Roman"]]],
 ". "
}], "Text"],

Cell[BoxData[
 StyleBox[
  RowBox[{
   RowBox[{
    RowBox[{
     FractionBox["1", 
      SqrtBox[
       RowBox[{"2", "\[Pi]"}]]], " ", 
     RowBox[{
      UnderoverscriptBox["\[Sum]", 
       RowBox[{"r", "=", "1"}], "n"], 
      RowBox[{"\[Psi]", 
       RowBox[{"(", 
        StyleBox[
         SubscriptBox["x", "r"],
         FontSlant->"Italic"], ")"}], 
       SuperscriptBox["e", 
        RowBox[{
         RowBox[{"-", "\[ImaginaryI]"}], " ", 
         StyleBox[
          SubscriptBox["k", "s"],
          FontSlant->"Italic"], 
         StyleBox[
          SubscriptBox["x", 
           StyleBox["1",
            FontSlant->"Plain"]],
          FontSlant->"Italic"]}]], 
       RowBox[{
        StyleBox["Exp", "InlineFormula",
         FontWeight->"Plain",
         FontSlant->"Plain",
         FontTracking->"Plain",
         FontVariations->{"CompatibilityType"->0,
         "Masked"->False,
         "Outline"->False,
         "RotationAngle"->0,
         "Shadow"->False,
         "StrikeThrough"->False,
         "Underline"->False}], 
        StyleBox["[", "InlineFormula",
         FontWeight->"Plain",
         FontSlant->"Plain",
         FontTracking->"Plain",
         FontVariations->{"CompatibilityType"->0,
         "Masked"->False,
         "Outline"->False,
         "RotationAngle"->0,
         "Shadow"->False,
         "StrikeThrough"->False,
         "Underline"->False}], 
        RowBox[{
         RowBox[{"-", "\[ImaginaryI]"}], 
         StyleBox[
          FractionBox[
           RowBox[{
            StyleBox["2",
             FontWeight->"Plain",
             FontSlant->"Plain",
             FontTracking->"Plain",
             FontVariations->{"Outline"->False,
             "Shadow"->False,
             "Underline"->False}], "\[Pi]"}], 
           RowBox[{"n", " "}]],
          FontSize->10,
          FontSlant->"Italic"], 
         RowBox[{
          StyleBox["(",
           FontSlant->"Italic"], 
          RowBox[{
           StyleBox["s",
            FontSlant->"Italic"], 
           StyleBox["-",
            FontSlant->"Italic"], 
           StyleBox["1",
            FontWeight->"Plain",
            FontSlant->"Plain",
            FontTracking->"Plain",
            FontVariations->{"Outline"->False,
            "Shadow"->False,
            "Underline"->False}]}], 
          StyleBox[")",
           FontSlant->"Italic"]}], 
         RowBox[{
          StyleBox["(",
           FontSlant->"Italic"], 
          RowBox[{
           StyleBox["r",
            FontSlant->"Italic"], 
           StyleBox["-",
            FontSlant->"Italic"], 
           StyleBox["1",
            FontWeight->"Plain",
            FontSlant->"Plain",
            FontTracking->"Plain",
            FontVariations->{"Outline"->False,
            "Shadow"->False,
            "Underline"->False}]}], 
          StyleBox[")",
           FontSlant->"Italic"]}]}], "]"}], 
       RowBox[{"d", 
        StyleBox["x",
         FontSlant->"Italic"]}]}]}]}], "\n", "            ", "=", " ", 
    RowBox[{
     SqrtBox[
      FractionBox["n", 
       RowBox[{"2", " ", "\[Pi]"}]]], " ", "dx", " ", 
     FractionBox["1", 
      SqrtBox["n"]], " ", 
     RowBox[{
      UnderoverscriptBox["\[Sum]", 
       RowBox[{"r", "=", "1"}], "n"], 
      RowBox[{"\[Psi]", 
       RowBox[{"(", 
        StyleBox[
         SubscriptBox["x", "r"],
         FontSlant->"Italic"], ")"}], 
       SuperscriptBox["e", 
        RowBox[{
         RowBox[{"-", "\[ImaginaryI]"}], " ", 
         StyleBox[
          SubscriptBox["k", "s"],
          FontSlant->"Italic"], 
         StyleBox[
          SubscriptBox["x", 
           StyleBox["1",
            FontSlant->"Plain"]],
          FontSlant->"Italic"]}]], 
       RowBox[{
        StyleBox["Exp", "InlineFormula",
         FontWeight->"Plain",
         FontSlant->"Plain",
         FontTracking->"Plain",
         FontVariations->{"CompatibilityType"->0,
         "Masked"->False,
         "Outline"->False,
         "RotationAngle"->0,
         "Shadow"->False,
         "StrikeThrough"->False,
         "Underline"->False}], 
        StyleBox["[", "InlineFormula",
         FontWeight->"Plain",
         FontSlant->"Plain",
         FontTracking->"Plain",
         FontVariations->{"CompatibilityType"->0,
         "Masked"->False,
         "Outline"->False,
         "RotationAngle"->0,
         "Shadow"->False,
         "StrikeThrough"->False,
         "Underline"->False}], 
        RowBox[{
         RowBox[{"-", "\[ImaginaryI]"}], 
         StyleBox[
          FractionBox[
           RowBox[{
            StyleBox["2",
             FontWeight->"Plain",
             FontSlant->"Plain",
             FontTracking->"Plain",
             FontVariations->{"Outline"->False,
             "Shadow"->False,
             "Underline"->False}], "\[Pi]"}], 
           RowBox[{"n", " "}]],
          FontSize->10,
          FontSlant->"Italic"], 
         RowBox[{
          StyleBox["(",
           FontSlant->"Italic"], 
          RowBox[{
           StyleBox["s",
            FontSlant->"Italic"], 
           StyleBox["-",
            FontSlant->"Italic"], 
           StyleBox["1",
            FontWeight->"Plain",
            FontSlant->"Plain",
            FontTracking->"Plain",
            FontVariations->{"Outline"->False,
            "Shadow"->False,
            "Underline"->False}]}], 
          StyleBox[")",
           FontSlant->"Italic"]}], 
         RowBox[{
          StyleBox["(",
           FontSlant->"Italic"], 
          RowBox[{
           StyleBox["r",
            FontSlant->"Italic"], 
           StyleBox["-",
            FontSlant->"Italic"], 
           StyleBox["1",
            FontWeight->"Plain",
            FontSlant->"Plain",
            FontTracking->"Plain",
            FontVariations->{"Outline"->False,
            "Shadow"->False,
            "Underline"->False}]}], 
          StyleBox[")",
           FontSlant->"Italic"]}]}], "]"}], 
       RowBox[{"d", 
        StyleBox["x",
         FontSlant->"Italic"]}]}]}]}]}], "\n"}],
  FontFamily->"Times New Roman"]], "DisplayFormula"],

Cell[TextData[{
 "The element with index ",
 StyleBox["s",
  FontSlant->"Italic"],
 " in the list ",
 StyleBox["InverseFourier",
  FontFamily->"Courier",
  FontWeight->"Bold"],
 "[",
 Cell[BoxData[
  StyleBox[
   RowBox[{"{", 
    RowBox[{
     RowBox[{"\[Psi]", 
      RowBox[{"(", 
       StyleBox[
        SubscriptBox["x", "1"],
        FontSlant->"Italic"], ")"}]}], ",", " ", "...", " ", ",", 
     RowBox[{"\[Psi]", 
      RowBox[{"(", 
       StyleBox[
        SubscriptBox["x", "n"],
        FontSlant->"Italic"], ")"}]}]}], "}"}],
   FontFamily->"Times New Roman"]]],
 "] has to be multiplied by the phase factor  ",
 Cell[BoxData[
  StyleBox[
   SuperscriptBox["e", 
    RowBox[{"\[ImaginaryI]", " ", 
     StyleBox[
      SubscriptBox["k", "s"],
      FontSlant->"Italic"], 
     StyleBox[
      SubscriptBox["x", 
       StyleBox["1",
        FontSize->7,
        FontSlant->"Plain"]],
      FontSlant->"Italic"]}]],
   FontFamily->"Times New Roman"]]],
 ". This multiplication by an ",
 StyleBox["s",
  FontSlant->"Italic"],
 "-dependent phase factor can be achieved more effectively by a cyclic \
permutation of the elements ",
 Cell[BoxData[
  StyleBox[
   RowBox[{"{", 
    RowBox[{
     RowBox[{"\[Psi]", 
      RowBox[{"(", 
       StyleBox[
        SubscriptBox["x", "1"],
        FontSlant->"Italic"], ")"}]}], ",", " ", "...", " ", ",", 
     RowBox[{"\[Psi]", 
      RowBox[{"(", 
       StyleBox[
        SubscriptBox["x", "n"],
        FontSlant->"Italic"], ")"}]}]}], "}"}], "InlineFormula",
   FontFamily->"Times New Roman"]]],
 ". We define"
}], "Text"],

Cell[BoxData[
 StyleBox[
  RowBox[{
   RowBox[{"ind", "[", 
    StyleBox["x",
     FontSlant->"Italic"], "]"}], " ", ":=", " ", 
   StyleBox[
    FractionBox[
     RowBox[{
      StyleBox["x",
       FontSlant->"Italic"], 
      StyleBox["-",
       FontSlant->"Italic"], 
      SubscriptBox[
       StyleBox["x",
        FontSlant->"Italic"], "0"]}], 
     StyleBox[
      RowBox[{"d", 
       StyleBox["x",
        FontSlant->"Italic"]}]]],
    FontSize->10]}],
  FontFamily->"Times New Roman",
  FontWeight->"Plain"]], "DisplayFormula"],

Cell[TextData[{
 "This is the index number of the point x in the sequence ",
 StyleBox["{", "InlineFormula",
  FontFamily->"Times New Roman"],
 Cell[BoxData[
  RowBox[{
   SubscriptBox[
    StyleBox["x",
     FontSlant->"Italic"], "1"], 
   StyleBox[",", "Text"], 
   SubscriptBox[
    StyleBox["x",
     FontSlant->"Italic"], "2"], ",", " ", "...", " ", ",", 
   SubscriptBox[
    StyleBox["x",
     FontSlant->"Italic"], "n"]}]], "InlineFormula",
  FontFamily->"Times New Roman"],
 StyleBox["}", "InlineFormula",
  FontFamily->"Times New Roman"],
 ". Hence ",
 StyleBox["ind[", "InlineFormula"],
 Cell[BoxData[
  SubscriptBox[
   StyleBox["x",
    FontSlant->"Italic"], "1"]], "InlineFormula"],
 StyleBox["]=1", "InlineFormula"],
 ", and ",
 StyleBox["ind[0] = -", "InlineFormula"],
 Cell[BoxData[
  FormBox[
   SubscriptBox["x", "0"], TraditionalForm]]],
 StyleBox["/d", "InlineFormula"],
 StyleBox["x", "InlineFormula",
  FontSlant->"Italic"],
 ". In ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 ", the cyclic permutation is performed by the command ",
 StyleBox["RotateLeft",
  FontFamily->"Courier",
  FontWeight->"Bold"],
 ":"
}], "Text"],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"{", 
    RowBox[{
     SubscriptBox["a", "1"], ",", " ", "...", " ", ",", 
     SubscriptBox["a", "n"]}], "}"}],
   FontFamily->"Times New Roman",
   FontWeight->"Plain"], 
  StyleBox[" ",
   FontFamily->"Times New Roman"], 
  StyleBox["=",
   FontFamily->"Times New Roman"], " ", 
  RowBox[{
   StyleBox["RotateLeft",
    FontFamily->"Courier",
    FontWeight->"Bold"], 
   StyleBox["[",
    FontWeight->"Plain"], 
   StyleBox[
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"\[Psi]", 
        RowBox[{"(", 
         StyleBox[
          SubscriptBox["x", "1"],
          FontSlant->"Italic"], ")"}]}], ",", " ", "...", " ", ",", 
       RowBox[{"\[Psi]", 
        RowBox[{"(", 
         StyleBox[
          SubscriptBox["x", "n"],
          FontSlant->"Italic"], ")"}]}]}], "}"}], ",", 
     RowBox[{"ind", "[", "0", "]"}]}],
    FontFamily->"Times New Roman",
    FontWeight->"Plain"], 
   StyleBox["]",
    FontWeight->"Plain"]}]}]], "DisplayFormula"],

Cell["\<\
This has precisely the same effect as the multiplication with the phase \
factor in (4). Hence the Fourier transform of the function is given by\
\>", "Text"],

Cell[BoxData[{
 StyleBox[
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{"4", "'"}], ")"}], "        ", 
    OverscriptBox["\[Psi]", "\[Wedge]"], 
    RowBox[{"(", 
     StyleBox[
      SubscriptBox[
       StyleBox["k",
        FontWeight->"Plain",
        FontSlant->"Italic"], "s"],
      FontSlant->"Italic"], ")"}]}], " ", "\[TildeTilde]", " ", 
   RowBox[{
    SqrtBox[
     FractionBox["n", 
      RowBox[{"2", " ", "\[Pi]"}]]], " ", 
    StyleBox[
     RowBox[{"d", 
      StyleBox["x",
       FontSlant->"Italic"]}]], " ", 
    FractionBox["1", 
     SqrtBox["n"]], "  ", 
    RowBox[{
     UnderoverscriptBox["\[Sum]", 
      RowBox[{"r", "=", "1"}], "n"], 
     RowBox[{
      StyleBox[
       SubscriptBox["a", "r"],
       FontSlant->"Italic"], 
      RowBox[{
       StyleBox["Exp", "InlineFormula",
        FontWeight->"Plain",
        FontSlant->"Plain",
        FontTracking->"Plain",
        FontVariations->{"CompatibilityType"->0,
        "Masked"->False,
        "Outline"->False,
        "RotationAngle"->0,
        "Shadow"->False,
        "StrikeThrough"->False,
        "Underline"->False}], 
       StyleBox["[", "InlineFormula",
        FontWeight->"Plain",
        FontSlant->"Plain",
        FontTracking->"Plain",
        FontVariations->{"CompatibilityType"->0,
        "Masked"->False,
        "Outline"->False,
        "RotationAngle"->0,
        "Shadow"->False,
        "StrikeThrough"->False,
        "Underline"->False}], 
       RowBox[{
        RowBox[{"-", "\[ImaginaryI]"}], 
        StyleBox[
         FractionBox[
          RowBox[{
           StyleBox["2",
            FontWeight->"Plain",
            FontSlant->"Plain",
            FontTracking->"Plain",
            FontVariations->{"Outline"->False,
            "Shadow"->False,
            "Underline"->False}], "\[Pi]"}], 
          RowBox[{"n", " "}]],
         FontSize->10,
         FontSlant->"Italic"], 
        RowBox[{
         StyleBox["(",
          FontSlant->"Italic"], 
         RowBox[{
          StyleBox["s",
           FontSlant->"Italic"], 
          StyleBox["-",
           FontSlant->"Italic"], 
          StyleBox["1",
           FontWeight->"Plain",
           FontSlant->"Plain",
           FontTracking->"Plain",
           FontVariations->{"Outline"->False,
           "Shadow"->False,
           "Underline"->False}]}], 
         StyleBox[")",
          FontSlant->"Italic"]}], 
        RowBox[{
         StyleBox["(",
          FontSlant->"Italic"], 
         RowBox[{
          StyleBox["r",
           FontSlant->"Italic"], 
          StyleBox["-",
           FontSlant->"Italic"], 
          StyleBox["1",
           FontWeight->"Plain",
           FontSlant->"Plain",
           FontTracking->"Plain",
           FontVariations->{"Outline"->False,
           "Shadow"->False,
           "Underline"->False}]}], 
         StyleBox[")",
          FontSlant->"Italic"]}]}], "]"}]}]}]}]}],
  FontFamily->"Times New Roman"], "\n", 
 RowBox[{
  StyleBox["                             ",
   FontFamily->"Times New Roman"], 
  RowBox[{
   StyleBox["=",
    FontFamily->"Times New Roman"], 
   StyleBox["  ",
    FontFamily->"Times New Roman"], 
   RowBox[{
    StyleBox[
     SqrtBox[
      FractionBox["n", 
       RowBox[{"2", " ", "\[Pi]"}]]],
     FontFamily->"Times New Roman"], 
    StyleBox[" ",
     FontFamily->"Times New Roman"], 
    StyleBox[
     RowBox[{"d", 
      StyleBox["x",
       FontSlant->"Italic"]}],
     FontFamily->"Times New Roman"], 
    StyleBox[
     StyleBox[
      RowBox[{" ", 
       StyleBox[" ",
        FontSlant->"Italic"]}]],
     FontFamily->"Times New Roman"], 
    RowBox[{
     RowBox[{
      StyleBox["InverseFourier",
       FontWeight->"Bold",
       FontSlant->"Plain"], 
      StyleBox["[",
       FontSlant->"Italic"], 
      StyleBox[
       RowBox[{"{", 
        RowBox[{
         SubscriptBox["a", "1"], ",", " ", "...", " ", ",", 
         SubscriptBox["a", "n"]}], "}"}],
       FontFamily->"Times New Roman"], 
      StyleBox["]",
       FontSlant->"Italic"]}], 
     StyleBox[".",
      FontSlant->"Italic"]}]}]}]}]}], "DisplayFormula"],

Cell[TextData[{
 "Discrete Fourier sums can be treated as periodic functions. In particular, \
if ",
 StyleBox["p", "InlineFormula",
  FontFamily->"Times New Roman",
  FontSlant->"Italic"],
 StyleBox[" = 2\[Pi]/d", "InlineFormula",
  FontFamily->"Times New Roman"],
 StyleBox["x", "InlineFormula",
  FontFamily->"Times New Roman",
  FontSlant->"Italic"],
 " is the length of the interval in ",
 StyleBox["k",
  FontSlant->"Italic"],
 "-space, we can use Eq. (3) and find, using ",
 Cell[BoxData[
  StyleBox[
   SuperscriptBox["e", 
    RowBox[{"\[ImaginaryI]", " ", 
     RowBox[{"(", 
      RowBox[{"\[Alpha]", 
       StyleBox[" ",
        FontSlant->"Italic"], 
       StyleBox["\[PlusMinus]",
        FontSlant->"Plain"], 
       StyleBox[" ",
        FontSlant->"Italic"], 
       RowBox[{"2", "\[Pi]"}]}], 
      StyleBox[")",
       FontSlant->"Italic"]}]}]],
   FontFamily->"Times New Roman"]]],
 " = ",
 Cell[BoxData[
  StyleBox[
   SuperscriptBox["e", 
    RowBox[{"\[ImaginaryI]", " ", "\[Alpha]"}]],
   FontFamily->"Times New Roman"]]]
}], "Text"],

Cell[BoxData[
 StyleBox[
  RowBox[{" ", 
   RowBox[{
    RowBox[{
     OverscriptBox["\[Psi]", "\[Wedge]"], 
     RowBox[{"(", 
      RowBox[{
       StyleBox[
        SubscriptBox[
         StyleBox["k",
          FontWeight->"Plain",
          FontSlant->"Italic"], "s"],
        FontSlant->"Italic"], "\[PlusMinus]", "p"}], ")"}]}], "=", " ", 
    RowBox[{
     RowBox[{
      FractionBox[
       StyleBox[
        RowBox[{"d", 
         StyleBox["x",
          FontSlant->"Italic"]}]], 
       SqrtBox[
        RowBox[{"2", "\[Pi]"}]]], " ", 
      RowBox[{
       UnderoverscriptBox["\[Sum]", 
        RowBox[{"r", "=", "1"}], "n"], 
       RowBox[{
        StyleBox[
         SubscriptBox["a", "r"],
         FontSlant->"Italic"], 
        SuperscriptBox["e", 
         RowBox[{
          RowBox[{"-", "\[ImaginaryI]"}], " ", 
          RowBox[{"(", 
           RowBox[{
            StyleBox[
             SubscriptBox["k", "s"],
             FontSlant->"Italic"], 
            StyleBox[" ",
             FontSlant->"Italic"], 
            StyleBox["\[PlusMinus]",
             FontSlant->"Plain"], 
            StyleBox[" ",
             FontSlant->"Italic"], 
            StyleBox["p",
             FontSlant->"Italic"]}], 
           StyleBox[")",
            FontSlant->"Italic"]}], 
          StyleBox["dx",
           FontSlant->"Italic"], 
          StyleBox[" ",
           FontSlant->"Italic"], 
          RowBox[{
           StyleBox["(",
            FontSlant->"Italic"], 
           RowBox[{
            StyleBox["r",
             FontSlant->"Italic"], 
            StyleBox["-",
             FontSlant->"Italic"], 
            StyleBox["1",
             FontSlant->"Plain"]}], 
           StyleBox[")",
            FontSlant->"Italic"]}]}]]}]}]}], " ", "=", " ", 
     RowBox[{
      RowBox[{
       FractionBox[
        StyleBox[
         RowBox[{"d", 
          StyleBox["x",
           FontSlant->"Italic"]}]], 
        SqrtBox[
         RowBox[{"2", "\[Pi]"}]]], " ", 
       RowBox[{
        UnderoverscriptBox["\[Sum]", 
         RowBox[{"r", "=", "1"}], "n"], 
        RowBox[{
         StyleBox[
          SubscriptBox["a", "r"],
          FontSlant->"Italic"], 
         SuperscriptBox["e", 
          RowBox[{
           RowBox[{
            RowBox[{"-", "\[ImaginaryI]"}], " ", 
            StyleBox[
             SubscriptBox["k", "s"],
             FontSlant->"Italic"], 
            StyleBox[" ",
             FontSlant->"Italic"], 
            StyleBox["dx",
             FontSlant->"Italic"], 
            StyleBox[" ",
             FontSlant->"Italic"], 
            RowBox[{
             StyleBox["(",
              FontSlant->"Italic"], 
             RowBox[{
              StyleBox["r",
               FontSlant->"Italic"], 
              StyleBox["-",
               FontSlant->"Italic"], 
              StyleBox["1",
               FontSlant->"Plain"]}], 
             StyleBox[")",
              FontSlant->"Italic"]}]}], 
           StyleBox[" ",
            FontSlant->"Italic"]}]]}]}]}], "=", " ", 
      RowBox[{
       OverscriptBox["\[Psi]", "\[Wedge]"], 
       RowBox[{
        RowBox[{"(", 
         StyleBox[
          SubscriptBox[
           StyleBox["k",
            FontWeight->"Plain",
            FontSlant->"Italic"], "s"],
          FontSlant->"Italic"], ")"}], "."}]}]}]}]}]}],
  FontFamily->"Times New Roman"]], "DisplayFormula"],

Cell[TextData[{
 "Usually, we are interested in an interval in momentum space, which is \
symmetric with respect to the origin. Using the periodicity in ",
 StyleBox["k",
  FontSlant->"Italic"],
 "-space, we can correct this again by a ",
 StyleBox["RotateLeft", "InlineInput",
  FontFamily->"Courier",
  FontWeight->"Bold"],
 " command. Since ",
 Cell[BoxData[
  RowBox[{
   StyleBox[" ", "InlineFormula",
    FontFamily->"Times New Roman"], 
   StyleBox[
    RowBox[{
     OverscriptBox["\[Psi]", "\[Wedge]"], 
     RowBox[{"(", 
      StyleBox[
       SubscriptBox[
        StyleBox["k",
         FontWeight->"Plain",
         FontSlant->"Italic"], "s"],
       FontSlant->"Italic"], ")"}]}],
    FontFamily->"Times New Roman"], " "}]]],
 "is periodic, we obtain the values of ",
 Cell[BoxData[
  StyleBox[
   OverscriptBox["\[Psi]", "\[Wedge]"],
   FontFamily->"Times New Roman"]]],
 " at the ",
 StyleBox["n",
  FontFamily->"Times New Roman",
  FontSlant->"Italic"],
 " points:"
}], "Text"],

Cell[TextData[{
 StyleBox["{",
  FontFamily->"Times New Roman"],
 Cell[BoxData[
  FormBox[
   RowBox[{"-", 
    FractionBox["\[Pi]", 
     StyleBox[
      RowBox[{"d", 
       StyleBox["x",
        FontSlant->"Italic"]}]]]}], TraditionalForm]],
  FontFamily->"Times New Roman"],
 Cell[BoxData[
  FormBox[
   RowBox[{"+", " ", 
    StyleBox[
     RowBox[{"d", 
      StyleBox["k",
       FontSlant->"Italic"]}]]}], TraditionalForm]],
  FontFamily->"Times New Roman"],
 StyleBox[", ... , -",
  FontFamily->"Times New Roman"],
 Cell[BoxData[
  FormBox[
   StyleBox[
    RowBox[{"d", 
     StyleBox["k",
      FontSlant->"Italic"]}]], TraditionalForm]],
  FontFamily->"Times New Roman"],
 StyleBox[", 0, ",
  FontFamily->"Times New Roman"],
 Cell[BoxData[
  FormBox[
   StyleBox[
    RowBox[{"d", 
     StyleBox["k",
      FontSlant->"Italic"]}]], TraditionalForm]],
  FontFamily->"Times New Roman"],
 StyleBox[", ",
  FontFamily->"Times New Roman"],
 Cell[BoxData[
  FormBox[
   RowBox[{"2", 
    StyleBox[
     RowBox[{"d", 
      StyleBox["k",
       FontSlant->"Italic"]}]]}], TraditionalForm]],
  FontFamily->"Times New Roman"],
 StyleBox[", ... , ",
  FontFamily->"Times New Roman"],
 Cell[BoxData[
  FormBox[
   FractionBox["\[Pi]", 
    RowBox[{" ", 
     StyleBox[
      RowBox[{"d", 
       StyleBox["x",
        FontSlant->"Italic"]}]]}]], TraditionalForm]],
  FontFamily->"Times New Roman"],
 StyleBox["}",
  FontFamily->"Times New Roman"]
}], "DisplayFormula"],

Cell[TextData[{
 "(which are obtained from",
 StyleBox[" ", "InlineInput"],
 StyleBox["{", "InlineInput",
  FontFamily->"Times New Roman"],
 StyleBox[Cell[BoxData[
  FormBox[
   SubscriptBox["k", "1"], TraditionalForm]], "InlineInput",
  FontFamily->"Times New Roman",
  FontWeight->"Plain"], "InlineInput"],
 StyleBox[", ", "InlineInput",
  FontFamily->"Times New Roman"],
 StyleBox[Cell[BoxData[
  FormBox[
   SubscriptBox["k", "2"], TraditionalForm]], "InlineInput",
  FontFamily->"Times New Roman",
  FontWeight->"Plain"], "InlineInput"],
 StyleBox[", ", "InlineInput",
  FontFamily->"Times New Roman"],
 StyleBox[Cell[BoxData[
  FormBox[
   SubscriptBox["k", "3"], TraditionalForm]], "InlineInput",
  FontFamily->"Times New Roman",
  FontWeight->"Plain"], "InlineInput"],
 StyleBox[", ... , ", "InlineInput",
  FontFamily->"Times New Roman"],
 StyleBox[Cell[BoxData[
  FormBox[
   SubscriptBox["k", "n"], TraditionalForm]], "InlineInput",
  FontFamily->"Times New Roman",
  FontWeight->"Plain"], "InlineInput"],
 StyleBox["}", "InlineInput",
  FontFamily->"Times New Roman"],
 StyleBox[" ", "InlineInput"],
 "by shifting these points ",
 StyleBox["n", "InlineFormula",
  FontFamily->"Times New Roman",
  FontSlant->"Italic"],
 StyleBox["/2 + 1", "InlineFormula",
  FontFamily->"Times New Roman"],
 " places towards the left) by performing the cyclic permutation"
}], "Text"],

Cell[TextData[{
 StyleBox["RotateLeft",
  FontFamily->"Courier",
  FontWeight->"Bold"],
 "[",
 StyleBox["{",
  FontFamily->"Times New Roman"],
 Cell[BoxData[
  RowBox[{" ", 
   RowBox[{
    OverscriptBox["\[Psi]", "\[Wedge]"], 
    RowBox[{"(", 
     StyleBox[
      SubscriptBox[
       StyleBox["k",
        FontWeight->"Plain",
        FontSlant->"Italic"], 
       StyleBox["1",
        FontSlant->"Plain"]],
      FontSlant->"Italic"], ")"}]}]}]],
  FontFamily->"Times New Roman"],
 StyleBox[",",
  FontFamily->"Times New Roman"],
 Cell[BoxData[
  RowBox[{" ", 
   RowBox[{
    OverscriptBox["\[Psi]", "\[Wedge]"], 
    RowBox[{"(", 
     StyleBox[
      SubscriptBox[
       StyleBox["k",
        FontWeight->"Plain",
        FontSlant->"Italic"], 
       StyleBox["2",
        FontSlant->"Plain"]],
      FontSlant->"Italic"], ")"}]}]}]],
  FontFamily->"Times New Roman"],
 StyleBox[", ... , ",
  FontFamily->"Times New Roman"],
 Cell[BoxData[
  RowBox[{" ", 
   RowBox[{
    OverscriptBox["\[Psi]", "\[Wedge]"], 
    RowBox[{"(", 
     StyleBox[
      SubscriptBox[
       StyleBox["k",
        FontWeight->"Plain",
        FontSlant->"Italic"], 
       StyleBox["n",
        FontSlant->"Italic"]],
      FontSlant->"Italic"], ")"}]}]}]],
  FontFamily->"Times New Roman"],
 StyleBox["}, ",
  FontFamily->"Times New Roman"],
 StyleBox["n",
  FontFamily->"Times New Roman",
  FontSlant->"Italic"],
 StyleBox["/2+1",
  FontFamily->"Times New Roman"],
 "] ."
}], "DisplayFormula"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Practical implementation", "Section"],

Cell[TextData[{
 "Assume that we are given a function ",
 StyleBox["f", "InlineFormula",
  FontFamily->"Times New Roman",
  FontSlant->"Italic"],
 StyleBox["[", "InlineFormula",
  FontFamily->"Times New Roman"],
 StyleBox["x", "InlineFormula",
  FontFamily->"Times New Roman",
  FontSlant->"Italic"],
 StyleBox["]", "InlineFormula",
  FontFamily->"Times New Roman"],
 " which vanishes outside an interval ",
 StyleBox["[",
  FontFamily->"Times New Roman"],
 StyleBox["a",
  FontFamily->"Times New Roman",
  FontSlant->"Italic"],
 StyleBox[", ",
  FontFamily->"Times New Roman"],
 StyleBox["b",
  FontFamily->"Times New Roman",
  FontSlant->"Italic"],
 StyleBox["]",
  FontFamily->"Times New Roman"],
 ". We want to determine a discrete approximation for the Fourier transform, \
suitable for plotting it quickly.\nWe first choose a grid of points on the \
x-axis. We assume that the grid points are equally spaced by ",
 StyleBox["d",
  FontFamily->"Times New Roman"],
 StyleBox["x",
  FontFamily->"Times New Roman",
  FontSlant->"Italic"],
 " and given by the numbers ",
 StyleBox["n",
  FontFamily->"Times New Roman",
  FontSlant->"Italic"],
 StyleBox[" d",
  FontFamily->"Times New Roman"],
 StyleBox["x",
  FontFamily->"Times New Roman",
  FontSlant->"Italic"],
 ". Since it is best to choose ",
 StyleBox["n",
  FontSlant->"Italic"],
 " as a power of 2, we calculate d",
 StyleBox["x",
  FontSlant->"Italic"],
 ":"
}], "Text"],

Cell[BoxData[
 RowBox[{"dx", "=", 
  FractionBox[
   RowBox[{"b", "-", "a"}], "n"]}]], "Input",
 Evaluatable->False],

Cell[BoxData[
 RowBox[{"xgrid", "=", 
  RowBox[{"Table", "[", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"Floor", "[", 
       FractionBox["a", "dx"], "]"}], " ", "dx"}], "+", 
     RowBox[{"k", " ", "dx"}]}], ",", 
    RowBox[{"{", 
     RowBox[{"k", ",", "1", ",", "n"}], "}"}]}], "]"}]}]], "Input",
 Evaluatable->False],

Cell[TextData[{
 "Of course, the size of d",
 StyleBox["x",
  FontSlant->"Italic"],
 " depends on the variation of the given function. d",
 StyleBox["x",
  FontSlant->"Italic"],
 " determines the size of the interval in ",
 StyleBox["k",
  FontSlant->"Italic"],
 "-space, where the values of the Fourier transform are calculated. The \
spacing of the points in ",
 StyleBox["k",
  FontSlant->"Italic"],
 "-space depends on the size of the interval ",
 StyleBox["[",
  FontFamily->"Times New Roman"],
 StyleBox["a",
  FontFamily->"Times New Roman",
  FontSlant->"Italic"],
 StyleBox[",",
  FontFamily->"Times New Roman"],
 StyleBox["b",
  FontFamily->"Times New Roman",
  FontSlant->"Italic"],
 StyleBox["]",
  FontFamily->"Times New Roman"],
 ". If we want to evaluate the ",
 StyleBox["k",
  FontSlant->"Italic"],
 "-values at points which are closer together, we have to choose the interval \
in ",
 StyleBox["x",
  FontSlant->"Italic"],
 "-space larger. The points in xgrid are inside the significant interval ",
 StyleBox["[",
  FontFamily->"Times New Roman"],
 StyleBox["a",
  FontFamily->"Times New Roman",
  FontSlant->"Italic"],
 StyleBox[",",
  FontFamily->"Times New Roman"],
 StyleBox["b",
  FontFamily->"Times New Roman",
  FontSlant->"Italic"],
 StyleBox["]",
  FontFamily->"Times New Roman"],
 ". We determine the function values at these points:"
}], "Text"],

Cell[BoxData[
 RowBox[{"fvals", "=", 
  RowBox[{"f", "/@", "xgrid"}]}]], "Input",
 Evaluatable->False],

Cell["Next we rotate the list of function values", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"ind", "[", "x_", "]"}], ":=", 
  RowBox[{
   RowBox[{"Floor", "[", 
    FractionBox[
     RowBox[{"x", "-", "a"}], "dx"], "]"}], "+", "1"}]}]], "Input",
 Evaluatable->False],

Cell[BoxData[
 RowBox[{"RotateLeft", "[", 
  RowBox[{"fvals", ",", 
   RowBox[{
    RowBox[{"ind", "[", "0", "]"}], "-", "1"}]}], "]"}]], "Input",
 Evaluatable->False],

Cell["From this, we calculate the discrete Fourier transform", "Text"],

Cell[BoxData[
 RowBox[{
  SqrtBox[
   FractionBox["n", 
    RowBox[{"2", " ", "\[Pi]"}]]], " ", "dx", " ", 
  RowBox[{"InverseFourier", "[", 
   RowBox[{"RotateLeft", "[", 
    RowBox[{"fvals", ",", 
     RowBox[{
      RowBox[{"ind", "[", "0", "]"}], "-", "1"}]}], "]"}], "]"}]}]], "Input",
 Evaluatable->False],

Cell[TextData[{
 "The result has been multiplied by the appropriate factor. Finally, the \
result is rotated by ",
 StyleBox["n",
  FontSlant->"Italic"],
 "/2"
}], "Text"],

Cell[BoxData[
 RowBox[{"RotateLeft", "[", 
  RowBox[{
   RowBox[{
    SqrtBox[
     FractionBox["n", 
      RowBox[{"2", " ", "\[Pi]"}]]], " ", "dx", " ", 
    RowBox[{"InverseFourier", "[", 
     RowBox[{"RotateLeft", "[", 
      RowBox[{"fvals", ",", 
       RowBox[{
        RowBox[{"ind", "[", "0", "]"}], "-", "1"}]}], "]"}], "]"}]}], ",", 
   RowBox[{"Round", "[", 
    RowBox[{
     FractionBox["n", "2"], "+", "1"}], "]"}]}], "]"}]], "Input",
 Evaluatable->False]
}, Open  ]],

Cell[CellGroupData[{

Cell["Explicit example", "Section"],

Cell["This is the function we want to transform:", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"f", "[", "x_", "]"}], "=", 
   RowBox[{
    SuperscriptBox[
     RowBox[{"(", 
      FractionBox["1", "\[Pi]"], ")"}], 
     RowBox[{"1", "/", "4"}]], " ", 
    RowBox[{"Exp", "[", 
     RowBox[{
      RowBox[{"-", 
       FractionBox[
        SuperscriptBox["x", "2"], "2"]}], "+", 
      RowBox[{"I", " ", "2", " ", "x"}]}], "]"}]}]}], ";"}]], "Input"],

Cell["We choose an interval and grid points in x-space:", "Text"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"n", "=", "512"}], ";"}], "\n", 
 RowBox[{
  RowBox[{"a", "=", 
   RowBox[{"-", "10"}]}], ";", 
  RowBox[{"b", "=", "14"}], ";"}]}], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"dx", "=", 
   FractionBox[
    RowBox[{"b", "-", "a"}], "n"]}], ";"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"xgrid", "=", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"Floor", "[", 
        FractionBox["a", "dx"], "]"}], " ", "dx"}], "+", 
      RowBox[{"k", " ", "dx"}]}], ",", 
     RowBox[{"{", 
      RowBox[{"k", ",", "1", ",", "n"}], "}"}]}], "]"}]}], ";"}]], "Input"],

Cell["\<\
Note that 0 is an element of xgrid. Now we evaluate the function on the grid \
points. This gives the discretization of the function f:\
\>", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"fvals", "=", 
   RowBox[{"f", "/@", "xgrid"}]}], ";"}]], "Input"],

Cell["\<\
The following definition is useful for identifying the value of f at a \
particular point:\
\>", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"ind", "[", "x_", "]"}], ":=", 
  RowBox[{
   RowBox[{"Floor", "[", 
    FractionBox[
     RowBox[{"x", "-", "a"}], "dx"], "]"}], "+", "1"}]}]], "Input"],

Cell["\<\
Next, we compute the Fourier transform of this list, as described above:\
\>", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"kvals", "=", 
   RowBox[{"Chop", "[", 
    RowBox[{"RotateLeft", "[", 
     RowBox[{
      RowBox[{
       SqrtBox[
        FractionBox["n", 
         RowBox[{"2", " ", "\[Pi]"}]]], " ", "dx", " ", 
       RowBox[{"InverseFourier", "[", 
        RowBox[{"RotateLeft", "[", 
         RowBox[{"fvals", ",", 
          RowBox[{
           RowBox[{"ind", "[", "0", "]"}], "-", "1"}]}], "]"}], "]"}]}], ",", 
      
      RowBox[{"Round", "[", 
       FractionBox["n", "2"], "]"}]}], "]"}], "]"}]}], ";"}]], "Input"],

Cell["\<\
This list may be interpreted as a discretization of the Fourier transform of \
the function f. The grid in k-space is given by:\
\>", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"kleft", "=", 
   RowBox[{"-", 
    FractionBox["\[Pi]", "dx"]}]}], ";", 
  RowBox[{"kright", "=", 
   FractionBox["\[Pi]", "dx"]}], ";", 
  RowBox[{"dk", "=", 
   FractionBox[
    RowBox[{"2", " ", "\[Pi]"}], 
    RowBox[{"n", " ", "dx"}]]}], ";"}]], "Input"],

Cell["In order to visualize the result, we load the package", "Text"],

Cell[BoxData[
 RowBox[{"<<", "\"\<VQM`ArgColorPlot`\>\""}]], "Input"],

Cell["This is the original function (or rather its discretization)", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"QListArgColorPlot", "[", 
  RowBox[{
   RowBox[{"Take", "[", 
    RowBox[{"fvals", ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"ind", "[", 
        RowBox[{"-", "3"}], "]"}], ",", 
       RowBox[{"ind", "[", "3", "]"}]}], "}"}]}], "]"}], ",", 
   RowBox[{"Frame", "\[Rule]", "True"}], ",", 
   RowBox[{"FrameTicks", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"QNiceTicks", "[", 
       RowBox[{
        RowBox[{"-", "3"}], ",", "3", ",", "dx"}], "]"}], ",", "Automatic"}], 
     "}"}]}]}], "]"}]], "Input",
 CellChangeTimes->{3.3927236133838687`*^9}],

Cell[BoxData[
 GraphicsBox[GraphicsComplexBox[CompressedData["
1:eJyF13tM1WUYB/Cj4bCcEx0qqUt0xAwdU0AEFM5XRfSoIHfO4apO06kzndlt
lo/TFZatC1o2UFKTNMWpqaWYe/OCKSLlbWIhhGhSWzF15fBSb+Oh9v7+eJ/z
z9mH74Bzfr/391yGzH4hfW5Xl8u1qovL9e97x6vNHfWHWni23yo3/wB97i99
4uL21ezucD+MDEl4+012AB5fvBAVXLuWHYjrrYVB0a+8yw5C/cKkXnfPvs8e
hJtZl9eVB5Wwg1Hh6VHZ/sUG9lC80/vVz8OLN7JDsLPv3Op9h0rZoZhQVvYY
8eXsYdjvKlm56IMt7DDU9VlSl/3VNvYIHL1T812vwAp2OI4diHvOu2sHeyQi
Xl88sJ12sUchSa1I+3bjHnYEnnrGPyj2wV52JBLCY/2bzu9nR6Hb/NueIWcO
sEfjxbqQwQd/O8SORvVfV/rExRxmj8HqIxOLw/dWsWOw//mKB62Zx9ix2FLw
ZbeaEsWOw5zCR5UlgcfZY1G2oX308MoT7HFITPWLOVdwih2P3dsWDFsXepqd
gPMne97N8T/DduOjT0ZUt/19lg38Wrzo6dZetR0mYPNrc1xbo+s4H4/lmx4u
/Ozw95yPx/H8mj0Lmn/gfALuNc0a7Nf3IucT0Nbd1/9g6iXOJ2L3/eDT29df
5nwiPl5f5W1uvMJ5Ig7fqJmbN+oq54mY3PjeyVtv1XM+Ce1lEXvQfI3zSbg0
7lN3T/dPnCehZfmOzGtlDZwnYd7R2qoh7dc5n4ytAy40bg9r4nwyGv0+dPmu
sF1TcL5+QUrayp85n4I3wirO7Atp5tyD4f5doktPscmDhsgIv/BZNzifipeX
NiSv+5NNU9Eyv6bhxzUtnE+Dp8fV+Ck9bnI+DTv7HTvht5btmo6BMV9vSHnU
mU/Hsm+ayhfPu8V5Mk729iyrqmZTMjbV3vEvHfAL5ynofzAkb81sNqXgQGm3
FtfmznwGwp69d7XyXGc+A6GzjyD+9848Fd6iFTUvdb3dYaQiwNd3ScyTbHLk
KtXx+2nm30ea+f8pzfx8Ks3x+dPN74d08/tTunl9VLrj+mWY1xcZ5vWnDPP+
qAzH/cs07y8yzftPmeb5UJmO85Nlni9kmeePsszzqbIc5zfbPN/INs8/ZZvP
h8p2PD855vOFHPP5oxzz+VQ5jufXaz7f8JrPP3nN+qC8jvrhM+sLfGb9IZ9Z
n5TPUb9yzfqGXLP+Ua5ZH1Wuo37mmfUVeWb9pTyzPqs8R/3ON+s78s36T/lm
f1D5jv5RYPYXFJj9hwrM/qQKHP2r0OxvKDT7HxWa/VEVOvpnkdlfUWT2Xyoy
+7MqcvTvmWZ/D54Js/+38ft/L9jdXXCA4EDBQYIHCQ4WPFRwiOBQwcMEhwke
IThc8EjBowRHCI4UHCV4tOBowWMExwiOFRwneKzgcYLjBScIdguG3SRYz4NW
k2A9L1pNgvU8aTUJ1vOm1SRYz6NWk2A9r1pNgvU8azUJ1vOu1SRYz8NWk2A9
L1tNgvU8bTUJ1vO21SRYz+NWk2A9r1tNgvU8bzUJ1vO+1RBMgpVgvS9YDcEk
WAnW+4bVEEyClWC9r1gNwSRYCdb7jtUQTIKVYL0vWQ3BJFgJ1vuW1RBMgpVg
va9ZDcEkWAnW+57VEEyClWC9L1oNwSRYCdb7ptUQTIKVYL2vWg3BJFgJ1vuu
1RBMgpVgvS9bDcEkWAnW+7bVEEyClWC9r1sNwSRYCdb7vtXB//sfjcU4AQ==

   "], {
    {GrayLevel[0.5], EdgeForm[None], 
     GraphicsGroupBox[PolygonBox[CompressedData["
1:eJwt1Xe8z3UUx/Hzsffe+9p7huyZSipuUUnoUklcIkRCpYyKNhFFIbKKpiha
mkQUhRapNFRapF7vzvnj+Xi8H9/v+ZzP53s+v/u4GVmjMrNzmNk05ELCrbgN
ep4Td2BGvC+KxXgUxaJmOm6P2jy4C3cjb6yZiVnIHWbjzqhVzRzMRT4UxgIs
RBEUxEOYh0LIj/twPwqEB/Bg1KrHPbg3arVmPh6O3lWxBa+gWuzxCBbFt1XB
y9gctSWxHCtQCsWxFI+jRHgCy6K2DFZjDcrGmiexEqVjZo9hSfRSzVqsQzlU
xkvYFGepiOfwPCqhPJ7BBlQIG/Fs1KrHejwdtVrzAl6M3jrDKjwVZ9UMXsVW
ZKAVjuEHtEYNvIm3UBN18AF2oC6q43W8EbW18R7ej1qt2Y63USu8g3ejVj12
4kPUQ2N8hgNogvr4CHvQIGp2YXe807O9+BgN0Qj78Wn00rNPsC/e6Ru34bU4
u/Y4iENoimb4El+hOc7AUXyLlvHsaxxGi3AE30StZvYjfsKZseY7fB+z1R6f
44vYSzU/4zjaoCNO4R90Qnv8ib/QAW3xG06gXfgdf0StevyCX6NWa/7Gyeit
nqfxLzoH/dEndDGXg5wTXcndkIecF93N5SPnx1nmCpALooe5QuTCONtcEXJR
nEO+EFXIVdHbXDVyBvqQe6IUuTTOM1eGXBa9zHsUIxfHueTzUYFcEReYq0Su
nHwv1ZQgl0zeWz3KkcsnX6s9q5NrIJM8QHdCboMryBehNrkOLjZXl1wPfc3V
JzdAP3MNyY1wibnG5Ca4lDxI90DuhMHme9Yk10q+V3/9jsgtcbm5VuTWyc+m
Hk3JzXCZ+RnbktthoLn25A7J91JNc3KL5L11p7nIuZPfrc7QWXePK8lZumPd
LYaQh2rmmjWuIo/WGtXievOaHrrr5LXX6A40ewwzX9NTd4mrzfXS7JPXqqa3
7h7XkrN1B5o9RpFH6Jv1rRhJHq470Oxxnbl+mnXyWvXI1CyT12pNf80yeW/1
HKjZJP8WfUOWvhVjzGfQVbNJPouxOrPOihvMDdNZMc7ccJ0F482N0FkxwVy2
vgU3mu8xVLNM3lvPRmuWmEiepD3UGzeRJ6tG73Cz+bPx6p38nZ5NUi2mmPcY
o97Je+nZZK3FVHNTlJP/z/8f+RYh/wccRsM7
       "]]]}, {{}, {}, {
      {Hue[0.045070341448627996`, 1, 1], LineBox[{1, 2}]}, 
      {Hue[0.059991117363493184`, 1, 1], LineBox[{2, 3}]}, 
      {Hue[0.07491189327835837, 1, 1], LineBox[{3, 4}]}, 
      {Hue[0.08983266919322357, 1, 1], LineBox[{4, 5}]}, 
      {Hue[0.10475344510808872`, 1, 1], LineBox[{5, 6}]}, 
      {Hue[0.11967422102295393`, 1, 1], LineBox[{6, 7}]}, 
      {Hue[0.13459499693781912`, 1, 1], LineBox[{7, 8}]}, 
      {Hue[0.1495157728526843, 1, 1], LineBox[{8, 9}]}, 
      {Hue[0.16443654876754946`, 1, 1], LineBox[{9, 10}]}, 
      {Hue[0.1793573246824147, 1, 1], LineBox[{10, 11}]}, 
      {Hue[0.19427810059727987`, 1, 1], LineBox[{11, 12}]}, 
      {Hue[0.20919887651214505`, 1, 1], LineBox[{12, 13}]}, 
      {Hue[0.22411965242701024`, 1, 1], LineBox[{13, 14}]}, 
      {Hue[0.23904042834187544`, 1, 1], LineBox[{14, 15}]}, 
      {Hue[0.2539612042567406, 1, 1], LineBox[{15, 16}]}, 
      {Hue[0.2688819801716058, 1, 1], LineBox[{16, 17}]}, 
      {Hue[0.283802756086471, 1, 1], LineBox[{17, 18}]}, 
      {Hue[0.29872353200133617`, 1, 1], LineBox[{18, 19}]}, 
      {Hue[0.3136443079162014, 1, 1], LineBox[{19, 20}]}, 
      {Hue[0.32856508383106653`, 1, 1], LineBox[{20, 21}]}, 
      {Hue[0.34348585974593177`, 1, 1], LineBox[{21, 22}]}, 
      {Hue[0.358406635660797, 1, 1], LineBox[{22, 23}]}, 
      {Hue[0.3733274115756622, 1, 1], LineBox[{23, 24}]}, 
      {Hue[0.38824818749052736`, 1, 1], LineBox[{24, 25}]}, 
      {Hue[0.40316896340539254`, 1, 1], LineBox[{25, 26}]}, 
      {Hue[0.4180897393202577, 1, 1], LineBox[{26, 27}]}, 
      {Hue[0.4330105152351229, 1, 1], LineBox[{27, 28}]}, 
      {Hue[0.44793129114998803`, 1, 1], LineBox[{28, 29}]}, 
      {Hue[0.4628520670648533, 1, 1], LineBox[{29, 30}]}, 
      {Hue[0.4777728429797184, 1, 1], LineBox[{30, 31}]}, 
      {Hue[0.4926936188945837, 1, 1], LineBox[{31, 32}]}, 
      {Hue[0.5076143948094487, 1, 1], LineBox[{32, 33}]}, 
      {Hue[0.522535170724314, 1, 1], LineBox[{33, 34}]}, 
      {Hue[0.5374559466391792, 1, 1], LineBox[{34, 35}]}, 
      {Hue[0.5523767225540444, 1, 1], LineBox[{35, 36}]}, 
      {Hue[0.5672974984689095, 1, 1], LineBox[{36, 37}]}, 
      {Hue[0.5822182743837747, 1, 1], LineBox[{37, 38}]}, 
      {Hue[0.5971390502986399, 1, 1], LineBox[{38, 39}]}, 
      {Hue[0.6120598262135051, 1, 1], LineBox[{39, 40}]}, 
      {Hue[0.6269806021283703, 1, 1], LineBox[{40, 41}]}, 
      {Hue[0.6419013780432354, 1, 1], LineBox[{41, 42}]}, 
      {Hue[0.6568221539581007, 1, 1], LineBox[{42, 43}]}, 
      {Hue[0.6717429298729658, 1, 1], LineBox[{43, 44}]}, 
      {Hue[0.6866637057878311, 1, 1], LineBox[{44, 45}]}, 
      {Hue[0.7015844817026962, 1, 1], LineBox[{45, 46}]}, 
      {Hue[0.7165052576175615, 1, 1], LineBox[{46, 47}]}, 
      {Hue[0.7314260335324265, 1, 1], LineBox[{47, 48}]}, 
      {Hue[0.7463468094472918, 1, 1], LineBox[{48, 49}]}, 
      {Hue[0.761267585362157, 1, 1], LineBox[{49, 50}]}, 
      {Hue[0.7761883612770222, 1, 1], LineBox[{50, 51}]}, 
      {Hue[0.7911091371918874, 1, 1], LineBox[{51, 52}]}, 
      {Hue[0.8060299131067525, 1, 1], LineBox[{52, 53}]}, 
      {Hue[0.8209506890216177, 1, 1], LineBox[{53, 54}]}, 
      {Hue[0.8358714649364829, 1, 1], LineBox[{54, 55}]}, 
      {Hue[0.8507922408513481, 1, 1], LineBox[{55, 56}]}, 
      {Hue[0.8657130167662133, 1, 1], LineBox[{56, 57}]}, 
      {Hue[0.8806337926810786, 1, 1], LineBox[{57, 58}]}, 
      {Hue[0.8955545685959436, 1, 1], LineBox[{58, 59}]}, 
      {Hue[0.9104753445108089, 1, 1], LineBox[{59, 60}]}, 
      {Hue[0.9253961204256741, 1, 1], LineBox[{60, 61}]}, 
      {Hue[0.9403168963405393, 1, 1], LineBox[{61, 62}]}, 
      {Hue[0.9552376722554045, 1, 1], LineBox[{62, 63}]}, 
      {Hue[0.9701584481702696, 1, 1], LineBox[{63, 64}]}, 
      {Hue[0.9850792240851348, 1, 1], LineBox[{64, 65}]}, 
      {Hue[0., 1, 1], LineBox[{65, 66}]}, 
      {Hue[0.014920775914865188`, 1, 1], LineBox[{66, 67}]}, 
      {Hue[0.029841551829730376`, 1, 1], LineBox[{67, 68}]}, 
      {Hue[0.044762327744595556`, 1, 1], LineBox[{68, 69}]}, 
      {Hue[0.05968310365946075, 1, 1], LineBox[{69, 70}]}, 
      {Hue[0.07460387957432593, 1, 1], LineBox[{70, 71}]}, 
      {Hue[0.08952465548919111, 1, 1], LineBox[{71, 72}]}, 
      {Hue[0.10444543140405632`, 1, 1], LineBox[{72, 73}]}, 
      {Hue[0.11936620731892149`, 1, 1], LineBox[{73, 74}]}, 
      {Hue[0.1342869832337867, 1, 1], LineBox[{74, 75}]}, 
      {Hue[0.1492077591486519, 1, 1], LineBox[{75, 76}]}, 
      {Hue[0.16412853506351707`, 1, 1], LineBox[{76, 77}]}, 
      {Hue[0.17904931097838225`, 1, 1], LineBox[{77, 78}]}, 
      {Hue[0.19397008689324746`, 1, 1], LineBox[{78, 79}]}, 
      {Hue[0.20889086280811264`, 1, 1], LineBox[{79, 80}]}, 
      {Hue[0.22381163872297782`, 1, 1], LineBox[{80, 81}]}, 
      {Hue[0.238732414637843, 1, 1], LineBox[{81, 82}]}, 
      {Hue[0.2536531905527082, 1, 1], LineBox[{82, 83}]}, 
      {Hue[0.2685739664675734, 1, 1], LineBox[{83, 84}]}, 
      {Hue[0.2834947423824386, 1, 1], LineBox[{84, 85}]}, 
      {Hue[0.2984155182973038, 1, 1], LineBox[{85, 86}]}, 
      {Hue[0.31333629421216896`, 1, 1], LineBox[{86, 87}]}, 
      {Hue[0.32825707012703415`, 1, 1], LineBox[{87, 88}]}, 
      {Hue[0.3431778460418993, 1, 1], LineBox[{88, 89}]}, 
      {Hue[0.3580986219567645, 1, 1], LineBox[{89, 90}]}, 
      {Hue[0.37301939787162974`, 1, 1], LineBox[{90, 91}]}, 
      {Hue[0.3879401737864949, 1, 1], LineBox[{91, 92}]}, 
      {Hue[0.4028609497013601, 1, 1], LineBox[{92, 93}]}, 
      {Hue[0.4177817256162253, 1, 1], LineBox[{93, 94}]}, 
      {Hue[0.43270250153109047`, 1, 1], LineBox[{94, 95}]}, 
      {Hue[0.44762327744595565`, 1, 1], LineBox[{95, 96}]}, 
      {Hue[0.46254405336082083`, 1, 1], LineBox[{96, 97}]}, 
      {Hue[0.477464829275686, 1, 1], LineBox[{97, 98}]}, 
      {Hue[0.49238560519055125`, 1, 1], LineBox[{98, 99}]}, 
      {Hue[0.5073063811054164, 1, 1], LineBox[{99, 100}]}, 
      {Hue[0.5222271570202817, 1, 1], LineBox[{100, 101}]}, 
      {Hue[0.5371479329351467, 1, 1], LineBox[{101, 102}]}, 
      {Hue[0.552068708850012, 1, 1], LineBox[{102, 103}]}, 
      {Hue[0.5669894847648771, 1, 1], LineBox[{103, 104}]}, 
      {Hue[0.5819102606797423, 1, 1], LineBox[{104, 105}]}, 
      {Hue[0.5968310365946075, 1, 1], LineBox[{105, 106}]}, 
      {Hue[0.6117518125094726, 1, 1], LineBox[{106, 107}]}, 
      {Hue[0.6266725884243378, 1, 1], LineBox[{107, 108}]}, 
      {Hue[0.641593364339203, 1, 1], LineBox[{108, 109}]}, 
      {Hue[0.6565141402540682, 1, 1], LineBox[{109, 110}]}, 
      {Hue[0.6714349161689335, 1, 1], LineBox[{110, 111}]}, 
      {Hue[0.6863556920837985, 1, 1], LineBox[{111, 112}]}, 
      {Hue[0.7012764679986638, 1, 1], LineBox[{112, 113}]}, 
      {Hue[0.716197243913529, 1, 1], LineBox[{113, 114}]}, 
      {Hue[0.7311180198283942, 1, 1], LineBox[{114, 115}]}, 
      {Hue[0.7460387957432594, 1, 1], LineBox[{115, 116}]}, 
      {Hue[0.7609595716581246, 1, 1], LineBox[{116, 117}]}, 
      {Hue[0.7758803475729897, 1, 1], LineBox[{117, 118}]}, 
      {Hue[0.7908011234878549, 1, 1], LineBox[{118, 119}]}, 
      {Hue[0.8057218994027201, 1, 1], LineBox[{119, 120}]}, 
      {Hue[0.8206426753175853, 1, 1], LineBox[{120, 121}]}, 
      {Hue[0.8355634512324506, 1, 1], LineBox[{121, 122}]}, 
      {Hue[0.8504842271473156, 1, 1], LineBox[{122, 123}]}, 
      {Hue[0.8654050030621809, 1, 1], LineBox[{123, 124}]}, 
      {Hue[0.8803257789770461, 1, 1], LineBox[{124, 125}]}, 
      {Hue[0.8952465548919113, 1, 1], LineBox[{125, 126}]}, 
      {Hue[0.9101673308067765, 1, 1], LineBox[{126, 127}]}, 
      {Hue[0.9250881067216417, 1, 1], LineBox[{127, 128}]}, 
      {Hue[0.9400088826365068, 1, 1], LineBox[{128, 129}]}}}},
   VertexColors->{
     Hue[0.045070341448627996`, 1, 1], 
     Hue[0.059991117363493184`, 1, 1], 
     Hue[0.07491189327835837, 1, 1], 
     Hue[0.08983266919322357, 1, 1], 
     Hue[0.10475344510808872`, 1, 1], 
     Hue[0.11967422102295393`, 1, 1], 
     Hue[0.13459499693781912`, 1, 1], 
     Hue[0.1495157728526843, 1, 1], 
     Hue[0.16443654876754946`, 1, 1], 
     Hue[0.1793573246824147, 1, 1], 
     Hue[0.19427810059727987`, 1, 1], 
     Hue[0.20919887651214505`, 1, 1], 
     Hue[0.22411965242701024`, 1, 1], 
     Hue[0.23904042834187544`, 1, 1], 
     Hue[0.2539612042567406, 1, 1], 
     Hue[0.2688819801716058, 1, 1], 
     Hue[0.283802756086471, 1, 1], 
     Hue[0.29872353200133617`, 1, 1], 
     Hue[0.3136443079162014, 1, 1], 
     Hue[0.32856508383106653`, 1, 1], 
     Hue[0.34348585974593177`, 1, 1], 
     Hue[0.358406635660797, 1, 1], 
     Hue[0.3733274115756622, 1, 1], 
     Hue[0.38824818749052736`, 1, 1], 
     Hue[0.40316896340539254`, 1, 1], 
     Hue[0.4180897393202577, 1, 1], 
     Hue[0.4330105152351229, 1, 1], 
     Hue[0.44793129114998803`, 1, 1], 
     Hue[0.4628520670648533, 1, 1], 
     Hue[0.4777728429797184, 1, 1], 
     Hue[0.4926936188945837, 1, 1], 
     Hue[0.5076143948094487, 1, 1], 
     Hue[0.522535170724314, 1, 1], 
     Hue[0.5374559466391792, 1, 1], 
     Hue[0.5523767225540444, 1, 1], 
     Hue[0.5672974984689095, 1, 1], 
     Hue[0.5822182743837747, 1, 1], 
     Hue[0.5971390502986399, 1, 1], 
     Hue[0.6120598262135051, 1, 1], 
     Hue[0.6269806021283703, 1, 1], 
     Hue[0.6419013780432354, 1, 1], 
     Hue[0.6568221539581007, 1, 1], 
     Hue[0.6717429298729658, 1, 1], 
     Hue[0.6866637057878311, 1, 1], 
     Hue[0.7015844817026962, 1, 1], 
     Hue[0.7165052576175615, 1, 1], 
     Hue[0.7314260335324265, 1, 1], 
     Hue[0.7463468094472918, 1, 1], 
     Hue[0.761267585362157, 1, 1], 
     Hue[0.7761883612770222, 1, 1], 
     Hue[0.7911091371918874, 1, 1], 
     Hue[0.8060299131067525, 1, 1], 
     Hue[0.8209506890216177, 1, 1], 
     Hue[0.8358714649364829, 1, 1], 
     Hue[0.8507922408513481, 1, 1], 
     Hue[0.8657130167662133, 1, 1], 
     Hue[0.8806337926810786, 1, 1], 
     Hue[0.8955545685959436, 1, 1], 
     Hue[0.9104753445108089, 1, 1], 
     Hue[0.9253961204256741, 1, 1], 
     Hue[0.9403168963405393, 1, 1], 
     Hue[0.9552376722554045, 1, 1], 
     Hue[0.9701584481702696, 1, 1], 
     Hue[0.9850792240851348, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.014920775914865188`, 1, 1], 
     Hue[0.029841551829730376`, 1, 1], 
     Hue[0.044762327744595556`, 1, 1], 
     Hue[0.05968310365946075, 1, 1], 
     Hue[0.07460387957432593, 1, 1], 
     Hue[0.08952465548919111, 1, 1], 
     Hue[0.10444543140405632`, 1, 1], 
     Hue[0.11936620731892149`, 1, 1], 
     Hue[0.1342869832337867, 1, 1], 
     Hue[0.1492077591486519, 1, 1], 
     Hue[0.16412853506351707`, 1, 1], 
     Hue[0.17904931097838225`, 1, 1], 
     Hue[0.19397008689324746`, 1, 1], 
     Hue[0.20889086280811264`, 1, 1], 
     Hue[0.22381163872297782`, 1, 1], 
     Hue[0.238732414637843, 1, 1], 
     Hue[0.2536531905527082, 1, 1], 
     Hue[0.2685739664675734, 1, 1], 
     Hue[0.2834947423824386, 1, 1], 
     Hue[0.2984155182973038, 1, 1], 
     Hue[0.31333629421216896`, 1, 1], 
     Hue[0.32825707012703415`, 1, 1], 
     Hue[0.3431778460418993, 1, 1], 
     Hue[0.3580986219567645, 1, 1], 
     Hue[0.37301939787162974`, 1, 1], 
     Hue[0.3879401737864949, 1, 1], 
     Hue[0.4028609497013601, 1, 1], 
     Hue[0.4177817256162253, 1, 1], 
     Hue[0.43270250153109047`, 1, 1], 
     Hue[0.44762327744595565`, 1, 1], 
     Hue[0.46254405336082083`, 1, 1], 
     Hue[0.477464829275686, 1, 1], 
     Hue[0.49238560519055125`, 1, 1], 
     Hue[0.5073063811054164, 1, 1], 
     Hue[0.5222271570202817, 1, 1], 
     Hue[0.5371479329351467, 1, 1], 
     Hue[0.552068708850012, 1, 1], 
     Hue[0.5669894847648771, 1, 1], 
     Hue[0.5819102606797423, 1, 1], 
     Hue[0.5968310365946075, 1, 1], 
     Hue[0.6117518125094726, 1, 1], 
     Hue[0.6266725884243378, 1, 1], 
     Hue[0.641593364339203, 1, 1], 
     Hue[0.6565141402540682, 1, 1], 
     Hue[0.6714349161689335, 1, 1], 
     Hue[0.6863556920837985, 1, 1], 
     Hue[0.7012764679986638, 1, 1], 
     Hue[0.716197243913529, 1, 1], 
     Hue[0.7311180198283942, 1, 1], 
     Hue[0.7460387957432594, 1, 1], 
     Hue[0.7609595716581246, 1, 1], 
     Hue[0.7758803475729897, 1, 1], 
     Hue[0.7908011234878549, 1, 1], 
     Hue[0.8057218994027201, 1, 1], 
     Hue[0.8206426753175853, 1, 1], 
     Hue[0.8355634512324506, 1, 1], 
     Hue[0.8504842271473156, 1, 1], 
     Hue[0.8654050030621809, 1, 1], 
     Hue[0.8803257789770461, 1, 1], 
     Hue[0.8952465548919113, 1, 1], 
     Hue[0.9101673308067765, 1, 1], 
     Hue[0.9250881067216417, 1, 1], 
     Hue[0.9400088826365068, 1, 1], 
     Hue[0.954929658551372, 1, 1], 
     Hue[0.045070341448627996`, 1, 1], 
     Hue[0.059991117363493184`, 1, 1], 
     Hue[0.059991117363493184`, 1, 1], 
     Hue[0.07491189327835837, 1, 1], 
     Hue[0.07491189327835837, 1, 1], 
     Hue[0.08983266919322357, 1, 1], 
     Hue[0.08983266919322357, 1, 1], 
     Hue[0.10475344510808872`, 1, 1], 
     Hue[0.10475344510808872`, 1, 1], 
     Hue[0.11967422102295393`, 1, 1], 
     Hue[0.11967422102295393`, 1, 1], 
     Hue[0.13459499693781912`, 1, 1], 
     Hue[0.13459499693781912`, 1, 1], 
     Hue[0.1495157728526843, 1, 1], 
     Hue[0.1495157728526843, 1, 1], 
     Hue[0.16443654876754946`, 1, 1], 
     Hue[0.16443654876754946`, 1, 1], 
     Hue[0.1793573246824147, 1, 1], 
     Hue[0.1793573246824147, 1, 1], 
     Hue[0.19427810059727987`, 1, 1], 
     Hue[0.19427810059727987`, 1, 1], 
     Hue[0.20919887651214505`, 1, 1], 
     Hue[0.20919887651214505`, 1, 1], 
     Hue[0.22411965242701024`, 1, 1], 
     Hue[0.22411965242701024`, 1, 1], 
     Hue[0.23904042834187544`, 1, 1], 
     Hue[0.23904042834187544`, 1, 1], 
     Hue[0.2539612042567406, 1, 1], 
     Hue[0.2539612042567406, 1, 1], 
     Hue[0.2688819801716058, 1, 1], 
     Hue[0.2688819801716058, 1, 1], 
     Hue[0.283802756086471, 1, 1], 
     Hue[0.283802756086471, 1, 1], 
     Hue[0.29872353200133617`, 1, 1], 
     Hue[0.29872353200133617`, 1, 1], 
     Hue[0.3136443079162014, 1, 1], 
     Hue[0.3136443079162014, 1, 1], 
     Hue[0.32856508383106653`, 1, 1], 
     Hue[0.32856508383106653`, 1, 1], 
     Hue[0.34348585974593177`, 1, 1], 
     Hue[0.34348585974593177`, 1, 1], 
     Hue[0.358406635660797, 1, 1], 
     Hue[0.358406635660797, 1, 1], 
     Hue[0.3733274115756622, 1, 1], 
     Hue[0.3733274115756622, 1, 1], 
     Hue[0.38824818749052736`, 1, 1], 
     Hue[0.38824818749052736`, 1, 1], 
     Hue[0.40316896340539254`, 1, 1], 
     Hue[0.40316896340539254`, 1, 1], 
     Hue[0.4180897393202577, 1, 1], 
     Hue[0.4180897393202577, 1, 1], 
     Hue[0.4330105152351229, 1, 1], 
     Hue[0.4330105152351229, 1, 1], 
     Hue[0.44793129114998803`, 1, 1], 
     Hue[0.44793129114998803`, 1, 1], 
     Hue[0.4628520670648533, 1, 1], 
     Hue[0.4628520670648533, 1, 1], 
     Hue[0.4777728429797184, 1, 1], 
     Hue[0.4777728429797184, 1, 1], 
     Hue[0.4926936188945837, 1, 1], 
     Hue[0.4926936188945837, 1, 1], 
     Hue[0.5076143948094487, 1, 1], 
     Hue[0.5076143948094487, 1, 1], 
     Hue[0.522535170724314, 1, 1], 
     Hue[0.522535170724314, 1, 1], 
     Hue[0.5374559466391792, 1, 1], 
     Hue[0.5374559466391792, 1, 1], 
     Hue[0.5523767225540444, 1, 1], 
     Hue[0.5523767225540444, 1, 1], 
     Hue[0.5672974984689095, 1, 1], 
     Hue[0.5672974984689095, 1, 1], 
     Hue[0.5822182743837747, 1, 1], 
     Hue[0.5822182743837747, 1, 1], 
     Hue[0.5971390502986399, 1, 1], 
     Hue[0.5971390502986399, 1, 1], 
     Hue[0.6120598262135051, 1, 1], 
     Hue[0.6120598262135051, 1, 1], 
     Hue[0.6269806021283703, 1, 1], 
     Hue[0.6269806021283703, 1, 1], 
     Hue[0.6419013780432354, 1, 1], 
     Hue[0.6419013780432354, 1, 1], 
     Hue[0.6568221539581007, 1, 1], 
     Hue[0.6568221539581007, 1, 1], 
     Hue[0.6717429298729658, 1, 1], 
     Hue[0.6717429298729658, 1, 1], 
     Hue[0.6866637057878311, 1, 1], 
     Hue[0.6866637057878311, 1, 1], 
     Hue[0.7015844817026962, 1, 1], 
     Hue[0.7015844817026962, 1, 1], 
     Hue[0.7165052576175615, 1, 1], 
     Hue[0.7165052576175615, 1, 1], 
     Hue[0.7314260335324265, 1, 1], 
     Hue[0.7314260335324265, 1, 1], 
     Hue[0.7463468094472918, 1, 1], 
     Hue[0.7463468094472918, 1, 1], 
     Hue[0.761267585362157, 1, 1], 
     Hue[0.761267585362157, 1, 1], 
     Hue[0.7761883612770222, 1, 1], 
     Hue[0.7761883612770222, 1, 1], 
     Hue[0.7911091371918874, 1, 1], 
     Hue[0.7911091371918874, 1, 1], 
     Hue[0.8060299131067525, 1, 1], 
     Hue[0.8060299131067525, 1, 1], 
     Hue[0.8209506890216177, 1, 1], 
     Hue[0.8209506890216177, 1, 1], 
     Hue[0.8358714649364829, 1, 1], 
     Hue[0.8358714649364829, 1, 1], 
     Hue[0.8507922408513481, 1, 1], 
     Hue[0.8507922408513481, 1, 1], 
     Hue[0.8657130167662133, 1, 1], 
     Hue[0.8657130167662133, 1, 1], 
     Hue[0.8806337926810786, 1, 1], 
     Hue[0.8806337926810786, 1, 1], 
     Hue[0.8955545685959436, 1, 1], 
     Hue[0.8955545685959436, 1, 1], 
     Hue[0.9104753445108089, 1, 1], 
     Hue[0.9104753445108089, 1, 1], 
     Hue[0.9253961204256741, 1, 1], 
     Hue[0.9253961204256741, 1, 1], 
     Hue[0.9403168963405393, 1, 1], 
     Hue[0.9403168963405393, 1, 1], 
     Hue[0.9552376722554045, 1, 1], 
     Hue[0.9552376722554045, 1, 1], 
     Hue[0.9701584481702696, 1, 1], 
     Hue[0.9701584481702696, 1, 1], 
     Hue[0.9850792240851348, 1, 1], 
     Hue[0.9850792240851348, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.014920775914865188`, 1, 1], 
     Hue[0.014920775914865188`, 1, 1], 
     Hue[0.029841551829730376`, 1, 1], 
     Hue[0.029841551829730376`, 1, 1], 
     Hue[0.044762327744595556`, 1, 1], 
     Hue[0.044762327744595556`, 1, 1], 
     Hue[0.05968310365946075, 1, 1], 
     Hue[0.05968310365946075, 1, 1], 
     Hue[0.07460387957432593, 1, 1], 
     Hue[0.07460387957432593, 1, 1], 
     Hue[0.08952465548919111, 1, 1], 
     Hue[0.08952465548919111, 1, 1], 
     Hue[0.10444543140405632`, 1, 1], 
     Hue[0.10444543140405632`, 1, 1], 
     Hue[0.11936620731892149`, 1, 1], 
     Hue[0.11936620731892149`, 1, 1], 
     Hue[0.1342869832337867, 1, 1], 
     Hue[0.1342869832337867, 1, 1], 
     Hue[0.1492077591486519, 1, 1], 
     Hue[0.1492077591486519, 1, 1], 
     Hue[0.16412853506351707`, 1, 1], 
     Hue[0.16412853506351707`, 1, 1], 
     Hue[0.17904931097838225`, 1, 1], 
     Hue[0.17904931097838225`, 1, 1], 
     Hue[0.19397008689324746`, 1, 1], 
     Hue[0.19397008689324746`, 1, 1], 
     Hue[0.20889086280811264`, 1, 1], 
     Hue[0.20889086280811264`, 1, 1], 
     Hue[0.22381163872297782`, 1, 1], 
     Hue[0.22381163872297782`, 1, 1], 
     Hue[0.238732414637843, 1, 1], 
     Hue[0.238732414637843, 1, 1], 
     Hue[0.2536531905527082, 1, 1], 
     Hue[0.2536531905527082, 1, 1], 
     Hue[0.2685739664675734, 1, 1], 
     Hue[0.2685739664675734, 1, 1], 
     Hue[0.2834947423824386, 1, 1], 
     Hue[0.2834947423824386, 1, 1], 
     Hue[0.2984155182973038, 1, 1], 
     Hue[0.2984155182973038, 1, 1], 
     Hue[0.31333629421216896`, 1, 1], 
     Hue[0.31333629421216896`, 1, 1], 
     Hue[0.32825707012703415`, 1, 1], 
     Hue[0.32825707012703415`, 1, 1], 
     Hue[0.3431778460418993, 1, 1], 
     Hue[0.3431778460418993, 1, 1], 
     Hue[0.3580986219567645, 1, 1], 
     Hue[0.3580986219567645, 1, 1], 
     Hue[0.37301939787162974`, 1, 1], 
     Hue[0.37301939787162974`, 1, 1], 
     Hue[0.3879401737864949, 1, 1], 
     Hue[0.3879401737864949, 1, 1], 
     Hue[0.4028609497013601, 1, 1], 
     Hue[0.4028609497013601, 1, 1], 
     Hue[0.4177817256162253, 1, 1], 
     Hue[0.4177817256162253, 1, 1], 
     Hue[0.43270250153109047`, 1, 1], 
     Hue[0.43270250153109047`, 1, 1], 
     Hue[0.44762327744595565`, 1, 1], 
     Hue[0.44762327744595565`, 1, 1], 
     Hue[0.46254405336082083`, 1, 1], 
     Hue[0.46254405336082083`, 1, 1], 
     Hue[0.477464829275686, 1, 1], 
     Hue[0.477464829275686, 1, 1], 
     Hue[0.49238560519055125`, 1, 1], 
     Hue[0.49238560519055125`, 1, 1], 
     Hue[0.5073063811054164, 1, 1], 
     Hue[0.5073063811054164, 1, 1], 
     Hue[0.5222271570202817, 1, 1], 
     Hue[0.5222271570202817, 1, 1], 
     Hue[0.5371479329351467, 1, 1], 
     Hue[0.5371479329351467, 1, 1], 
     Hue[0.552068708850012, 1, 1], 
     Hue[0.552068708850012, 1, 1], 
     Hue[0.5669894847648771, 1, 1], 
     Hue[0.5669894847648771, 1, 1], 
     Hue[0.5819102606797423, 1, 1], 
     Hue[0.5819102606797423, 1, 1], 
     Hue[0.5968310365946075, 1, 1], 
     Hue[0.5968310365946075, 1, 1], 
     Hue[0.6117518125094726, 1, 1], 
     Hue[0.6117518125094726, 1, 1], 
     Hue[0.6266725884243378, 1, 1], 
     Hue[0.6266725884243378, 1, 1], 
     Hue[0.641593364339203, 1, 1], 
     Hue[0.641593364339203, 1, 1], 
     Hue[0.6565141402540682, 1, 1], 
     Hue[0.6565141402540682, 1, 1], 
     Hue[0.6714349161689335, 1, 1], 
     Hue[0.6714349161689335, 1, 1], 
     Hue[0.6863556920837985, 1, 1], 
     Hue[0.6863556920837985, 1, 1], 
     Hue[0.7012764679986638, 1, 1], 
     Hue[0.7012764679986638, 1, 1], 
     Hue[0.716197243913529, 1, 1], 
     Hue[0.716197243913529, 1, 1], 
     Hue[0.7311180198283942, 1, 1], 
     Hue[0.7311180198283942, 1, 1], 
     Hue[0.7460387957432594, 1, 1], 
     Hue[0.7460387957432594, 1, 1], 
     Hue[0.7609595716581246, 1, 1], 
     Hue[0.7609595716581246, 1, 1], 
     Hue[0.7758803475729897, 1, 1], 
     Hue[0.7758803475729897, 1, 1], 
     Hue[0.7908011234878549, 1, 1], 
     Hue[0.7908011234878549, 1, 1], 
     Hue[0.8057218994027201, 1, 1], 
     Hue[0.8057218994027201, 1, 1], 
     Hue[0.8206426753175853, 1, 1], 
     Hue[0.8206426753175853, 1, 1], 
     Hue[0.8355634512324506, 1, 1], 
     Hue[0.8355634512324506, 1, 1], 
     Hue[0.8504842271473156, 1, 1], 
     Hue[0.8504842271473156, 1, 1], 
     Hue[0.8654050030621809, 1, 1], 
     Hue[0.8654050030621809, 1, 1], 
     Hue[0.8803257789770461, 1, 1], 
     Hue[0.8803257789770461, 1, 1], 
     Hue[0.8952465548919113, 1, 1], 
     Hue[0.8952465548919113, 1, 1], 
     Hue[0.9101673308067765, 1, 1], 
     Hue[0.9101673308067765, 1, 1], 
     Hue[0.9250881067216417, 1, 1], 
     Hue[0.9250881067216417, 1, 1], 
     Hue[0.9400088826365068, 1, 1], 
     Hue[0.9400088826365068, 1, 1], 
     Hue[0.954929658551372, 1, 1]}],
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  Epilog->LineBox[{{1, 0.008344251073246413}, {2, 0.009593628447188283}, {
     3, 0.01100586460097308}, {4, 0.012598277675673227`}, {
     5, 0.014389441126572085`}, {6, 0.016399190718953214`}, {
     7, 0.018648617716584232`}, {8, 0.021160046397946573`}, {
     9, 0.023956994029747086`}, {10, 0.0270641114559759}, {
     11, 0.03050710252787448}, {12, 0.03431262070944409}, {
     13, 0.038508141347880824`}, {14, 0.043121808301270076`}, {
     15, 0.04818225386893847}, {16, 0.05371839127405521}, {
     17, 0.05975917930334211}, {18, 0.06633335911384833}, {
     19, 0.07346916366912468}, {20, 0.0811940007628494}, {
     21, 0.08953411112162063}, {22, 0.09851420364336069}, {
     23, 0.10815707041522597`}, {24, 0.11848318475531519`}, {
     25, 0.12951028612472462`}, {26, 0.14125295634834167`}, {
     27, 0.1537221921509343}, {28, 0.16692497954557173`}, {
     29, 0.18086387608971444`}, {30, 0.19553660743580398`}, {
     31, 0.2109356849334289}, {32, 0.2270480512752962}, {
     33, 0.24385476130642741`}, {34, 0.2613307051237464}, {
     35, 0.2794443804718406}, {36, 0.29815772118264694`}, {
     37, 0.3174259880071437}, {38, 0.33719772764369327`}, {
     39, 0.35741480508142714`}, {40, 0.3780125135522741}, {
     41, 0.3989197654295719}, {42, 0.4200593663358202}, {
     43, 0.44134837354154155`}, {44, 0.4626985384691962}, {
     45, 0.4840168317813948}, {46, 0.5052060481546798}, {
     47, 0.5261654864445083}, {48, 0.546791699561078}, {
     49, 0.5669793070276936}, {50, 0.5866218619123392}, {
     51, 0.6056127626376299}, {52, 0.6238461991120364}, {
     53, 0.641218121712225}, {54, 0.6576272209061723}, {
     55, 0.6729759047600129}, {56, 0.6871712612353379}, {
     57, 0.70012599207068}, {58, 0.711759305159426}, {
     59, 0.7219977526896632}, {60, 0.7307760028976727}, {
     61, 0.7380375340988992}, {62, 0.7437352406861113}, {
     63, 0.7478319420070576}, {64, 0.7503007864315618}, {
     65, 0.7511255444649425}, {66, 0.7503007864315618}, {
     67, 0.7478319420070576}, {68, 0.7437352406861113}, {
     69, 0.7380375340988992}, {70, 0.7307760028976727}, {
     71, 0.7219977526896632}, {72, 0.711759305159426}, {
     73, 0.70012599207068}, {74, 0.6871712612353379}, {
     75, 0.6729759047600129}, {76, 0.6576272209061723}, {
     77, 0.641218121712225}, {78, 0.6238461991120364}, {
     79, 0.6056127626376299}, {80, 0.5866218619123392}, {
     81, 0.5669793070276936}, {82, 0.546791699561078}, {
     83, 0.5261654864445083}, {84, 0.5052060481546798}, {
     85, 0.4840168317813948}, {86, 0.4626985384691962}, {
     87, 0.44134837354154155`}, {88, 0.4200593663358202}, {
     89, 0.3989197654295719}, {90, 0.3780125135522741}, {
     91, 0.35741480508142714`}, {92, 0.33719772764369327`}, {
     93, 0.3174259880071437}, {94, 0.29815772118264694`}, {
     95, 0.2794443804718406}, {96, 0.2613307051237464}, {
     97, 0.24385476130642741`}, {98, 0.2270480512752962}, {
     99, 0.2109356849334289}, {100, 0.19553660743580398`}, {
     101, 0.18086387608971444`}, {102, 0.16692497954557173`}, {
     103, 0.1537221921509343}, {104, 0.14125295634834167`}, {
     105, 0.12951028612472462`}, {106, 0.11848318475531519`}, {
     107, 0.10815707041522597`}, {108, 0.09851420364336069}, {
     109, 0.08953411112162063}, {110, 0.0811940007628494}, {
     111, 0.07346916366912468}, {112, 0.06633335911384833}, {
     113, 0.05975917930334211}, {114, 0.05371839127405521}, {
     115, 0.04818225386893847}, {116, 0.043121808301270076`}, {
     117, 0.038508141347880824`}, {118, 0.03431262070944409}, {
     119, 0.03050710252787448}, {120, 0.0270641114559759}, {
     121, 0.023956994029747086`}, {122, 0.021160046397946573`}, {
     123, 0.018648617716584232`}, {124, 0.016399190718953214`}, {
     125, 0.014389441126572085`}, {126, 0.012598277675673227`}, {
     127, 0.01100586460097308}, {128, 0.009593628447188283}, {
     129, 0.008344251073246413}}],
  Frame->True,
  FrameTicks->{{{1., 
      FormBox[
       RowBox[{"-", "3"}], TraditionalForm]}, {22.333333333333332`, 
      FormBox[
       RowBox[{"-", "2"}], TraditionalForm]}, {43.666666666666664`, 
      FormBox[
       RowBox[{"-", "1"}], TraditionalForm]}, {65., 
      FormBox["0", TraditionalForm]}, {86.33333333333333, 
      FormBox["1", TraditionalForm]}, {107.66666666666666`, 
      FormBox["2", TraditionalForm]}, {129., 
      FormBox["3", TraditionalForm]}}, Automatic},
  PlotRange->{{0., 129.}, {0., 0.7511255444649425}},
  PlotRangeClipping->True,
  PlotRangePadding->{
    Scaled[0.02], 
    Scaled[0.02]}]], "Output",
 CellChangeTimes->{
  3.3927236144554095`*^9, {3.392726505442445*^9, 3.39184983246875*^9}, 
   3.391850037765625*^9, 3.39247235826097*^9},
 ImageCache->GraphicsData["CompressedBitmap", "\<\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\
\>"]]
}, Open  ]],

Cell["\<\
The following function is needed in order to extract parts of the \
Fourier-transformed list:\
\>", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"indk", "[", "k_", "]"}], ":=", 
   RowBox[{
    RowBox[{"Floor", "[", 
     FractionBox[
      RowBox[{"k", "-", "kleft"}], "dk"], "]"}], "+", "1"}]}], ";"}]], "Input"],

Cell[TextData[{
 "Now we can plot the interesting part of the list ",
 StyleBox["kvals", "InlineInput"]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"QListArgColorPlot", "[", 
  RowBox[{
   RowBox[{"Take", "[", 
    RowBox[{"kvals", ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"indk", "[", 
        RowBox[{"-", "1"}], "]"}], ",", 
       RowBox[{"indk", "[", "5", "]"}]}], "}"}]}], "]"}], ",", 
   RowBox[{"Frame", "\[Rule]", "True"}], ",", 
   RowBox[{"FrameTicks", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"QNiceTicks", "[", 
       RowBox[{
        RowBox[{"-", "1"}], ",", "5", ",", "dk"}], "]"}], ",", "Automatic"}], 
     "}"}]}], ",", 
   RowBox[{"PlotRange", "\[Rule]", "All"}]}], "]"}]], "Input",
 CellChangeTimes->{3.3927236571668253`*^9}],

Cell[BoxData[
 GraphicsBox[GraphicsComplexBox[CompressedData["
1:eJxTTMoPSmViYGBwA2IQDQEf7P36RMy+Lay1hwo4CAe1uh481Q/lczgs9+1f
m3FzPpQv4FA0jecd0+m1UL6Iw8ckrwm/WXdB+RIOXaEOn6daHoXyZRwM9xev
l9e6AOUrOBRP35bV/PcalK/kIMDE7uDccQ/KV3E4aNJckvTuEZSv5pAis9zv
+OunUL6Gg1D9Yo1Zm59D+VoO12NaRe69hfF1HPrts+QZ+p9B+XoO3/sF5k27
/RjKN3AIihHqaA16AOUbOqx1OJ2hyncTyjdyEDiy4+3VLxehfGMHy382lxM4
TkD5Jg6+r4uvX+fYB+WbOnxXtGuwa9sI5Zs5yH+pVM2RWAzlmzts1Z7jdjZ3
MpRv4XDO+mpcQWmTPSz8GVCBA34+BwG+AAG+CAG+BAG+DAG+AgG+EgG+CgG+
GgG+BgG+FgG+DgG+HgG+AQG+IQG+EQG+MQG+CQG+KQG+GQG+OQG+BZwPAC0x
fEI=
   "], {
    {GrayLevel[0.5], EdgeForm[None], 
     GraphicsGroupBox[
      PolygonBox[{{1, 25, 26, 2}, {2, 27, 28, 3}, {3, 29, 30, 4}, {4, 31, 32, 
       5}, {10, 43, 44, 11}, {11, 45, 46, 12}, {7, 37, 38, 8}, {5, 33, 34, 
       6}, {9, 41, 42, 10}, {6, 35, 36, 7}, {8, 39, 40, 9}, {12, 47, 48, 
       13}, {20, 63, 64, 21}, {14, 51, 52, 15}, {15, 53, 54, 16}, {16, 55, 56,
        17}, {22, 67, 68, 23}, {13, 49, 50, 14}, {19, 61, 62, 20}, {17, 57, 
       58, 18}, {21, 65, 66, 22}, {18, 59, 60, 19}, {23, 69, 70, 
       24}}]]}, {{}, {}, {
      {Hue[0, 1, 1], LineBox[{1, 2}]}, 
      {Hue[0, 1, 1], LineBox[{2, 3}]}, 
      {Hue[0, 1, 1], LineBox[{3, 4}]}, 
      {Hue[0, 1, 1], LineBox[{4, 5}]}, 
      {Hue[0, 1, 1], LineBox[{5, 6}]}, 
      {Hue[0, 1, 1], LineBox[{6, 7}]}, 
      {Hue[0, 1, 1], LineBox[{7, 8}]}, 
      {Hue[0, 1, 1], LineBox[{8, 9}]}, 
      {Hue[0, 1, 1], LineBox[{9, 10}]}, 
      {Hue[0, 1, 1], LineBox[{10, 11}]}, 
      {Hue[0, 1, 1], LineBox[{11, 12}]}, 
      {Hue[0, 1, 1], LineBox[{12, 13}]}, 
      {Hue[0, 1, 1], LineBox[{13, 14}]}, 
      {Hue[0, 1, 1], LineBox[{14, 15}]}, 
      {Hue[0, 1, 1], LineBox[{15, 16}]}, 
      {Hue[0, 1, 1], LineBox[{16, 17}]}, 
      {Hue[0, 1, 1], LineBox[{17, 18}]}, 
      {Hue[0, 1, 1], LineBox[{18, 19}]}, 
      {Hue[0, 1, 1], LineBox[{19, 20}]}, 
      {Hue[0, 1, 1], LineBox[{20, 21}]}, 
      {Hue[0, 1, 1], LineBox[{21, 22}]}, 
      {Hue[0, 1, 1], LineBox[{22, 23}]}, 
      {Hue[0, 1, 1], LineBox[{23, 24}]}}}},
   VertexColors->{
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1], 
     Hue[0, 1, 1]}],
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{5., 0},
  Epilog->LineBox[{{1, 0.007234536898326717}, {2, 0.01552344315920443}, {
     3, 0.031102786629913528`}, {4, 0.05818947940705933}, {
     5, 0.10165378830641793`}, {6, 0.165819877409984}, {
     7, 0.25257103094098726`}, {8, 0.3592232267495696}, {
     9, 0.47706681492493974`}, {10, 0.591599681437991}, {
     11, 0.6850315605341959}, {12, 0.7406740945552619}, {
     13, 0.7477865601348651}, {14, 0.7049561132025576}, {
     15, 0.6205552184296711}, {16, 0.5100734384203724}, {
     17, 0.3914884112719033}, {18, 0.28056858267797313`}, {
     19, 0.18775568308227422`}, {20, 0.11732243547514294`}, {
     21, 0.0684546530811828}, {22, 0.037295703482392854`}, {
     23, 0.01897354835004324}, {24, 0.009013059515067836}}],
  Frame->True,
  FrameTicks->{{{1., 
      FormBox[
       RowBox[{"-", "1"}], TraditionalForm]}, {4.819718634205488, 
      FormBox["0", TraditionalForm]}, {8.639437268410976, 
      FormBox["1", TraditionalForm]}, {12.459155902616464`, 
      FormBox["2", TraditionalForm]}, {16.278874536821952`, 
      FormBox["3", TraditionalForm]}, {20.09859317102744, 
      FormBox["4", TraditionalForm]}, {23.91831180523293, 
      FormBox["5", TraditionalForm]}}, Automatic},
  PlotRange->All,
  PlotRangeClipping->True,
  PlotRangePadding->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{
  3.3927236577176175`*^9, {3.3927265055626173`*^9, 3.39184983271875*^9}, 
   3.391850038*^9, 3.3924723589892797`*^9},
 ImageCache->GraphicsData["CompressedBitmap", "\<\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\
\>"]]
}, Open  ]],

Cell["\<\
We can easily combine the computation above in a single function. The \
following computes the Fourier transform of a numerically given function \
xlist and returns the list of Fourier-transformed values together with the \
information about the grid in k-space:\
\>", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"fourierList", "[", 
    RowBox[{"xlist_", ",", 
     RowBox[{"{", 
      RowBox[{"xleft_", ",", "xright_"}], "}"}]}], "]"}], ":=", 
   RowBox[{"Module", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
      "n", ",", "dx", ",", "a", ",", "b", ",", "klist", ",", "kleft", ",", 
       "kright", ",", "dk"}], "}"}], ",", 
     RowBox[{
      RowBox[{"n", "=", 
       RowBox[{
        RowBox[{"Length", "[", "xlist", "]"}], "-", "1"}]}], ";", 
      RowBox[{"dx", "=", 
       FractionBox[
        RowBox[{"xright", "-", "xleft"}], "n"]}], ";", 
      RowBox[{"a", "=", 
       RowBox[{"-", 
        FractionBox["xleft", "dx"]}]}], ";", 
      RowBox[{"b", "=", 
       FractionBox["xright", "dx"]}], ";", 
      RowBox[{"klist", "=", 
       RowBox[{
        SqrtBox[
         FractionBox["n", 
          RowBox[{"2", " ", "\[Pi]"}]]], " ", "dx", " ", 
        RowBox[{"RotateLeft", "[", 
         RowBox[{
          RowBox[{"InverseFourier", "[", 
           RowBox[{"RotateLeft", "[", 
            RowBox[{"xlist", ",", 
             RowBox[{"Round", "[", "a", "]"}]}], "]"}], "]"}], ",", 
          RowBox[{"Round", "[", 
           RowBox[{
            FractionBox["n", "2"], "+", "1"}], "]"}]}], "]"}]}]}], ";", 
      RowBox[{"kleft", "=", 
       RowBox[{"-", 
        FractionBox["\[Pi]", "dx"]}]}], ";", 
      RowBox[{"kright", "=", 
       FractionBox["\[Pi]", "dx"]}], ";", 
      RowBox[{"{", 
       RowBox[{"klist", ",", 
        RowBox[{"{", 
         RowBox[{"kleft", ",", "kright"}], "}"}]}], "}"}]}]}], "]"}]}], 
  ";"}]], "Input"],

Cell[TextData[{
 "This function is implemented as ",
 StyleBox["QFourierList", "InlineInput"],
 " in the package ",
 StyleBox["VQM`FastFourier`", "InlineInput"]
}], "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["The package VQM`FastFourier`", "Section"],

Cell["\<\
The results described above form the theoretical background of the package \
VQM`FastFourier`. Here we show how to obtain the Fourier transform of a \
function f with the help of this package. The package VQM`ArgColorPlot` will \
be useful for the visualization of the function and its Fourier transform.\
\>", "Text"],

Cell["This loads all declarations of the VQM packages:", "Text"],

Cell["Needs[\"VQM`\"]", "Input",
 CellChangeTimes->{{3.391849985859375*^9, 3.391849989125*^9}}],

Cell["\<\
The most important function defined by VQM`FastFourier` is QFourierList. \
(Note that all symbols defined by the VQM packages start with the letter Q.)\
\>", "Text"],

Cell["\<\
The first argument of QFourierList is a list flist of complex numbers. You \
may interpret this list as the discretization of a function defined in some \
interval [a,b]. The second argument of QFourierList specifies this interval. \
Hence the function is called like this: QFourierList[flist,{a,b}]. In order \
to have an example, we generate a list of complex values from a function f. \
The package VQM`FastFourier` provides some auxiliary functions for this.\
\>", "Text"],

Cell["\<\
We repeat the example above with the help of this package and consider the \
function\
\>", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"f", "[", "x_", "]"}], "=", 
   RowBox[{
    SuperscriptBox[
     RowBox[{"(", 
      FractionBox["1", "\[Pi]"], ")"}], 
     RowBox[{"1", "/", "4"}]], " ", 
    RowBox[{"Exp", "[", 
     RowBox[{
      RowBox[{"-", 
       FractionBox[
        SuperscriptBox["x", "2"], "2"]}], "+", 
      RowBox[{"I", " ", "2", " ", "x"}]}], "]"}]}]}], ";"}]], "Input"],

Cell["in the interval [-10,14]:", "Text"],

Cell[CellGroupData[{

Cell["QArgColorPlot[f[x],{x,-10,14},PlotRange->All]", "Input"],

Cell[BoxData[
 GraphicsBox[GraphicsComplexBox[CompressedData["
1:eJxEl3c419///+09XtYLL3skK6GM5HnOyaqkQpQyEoWQCgkVkkSDjCRJRkZW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   "], {
    {GrayLevel[0.5], EdgeForm[None], 
     GraphicsGroupBox[{PolygonBox[{{14, 1233, 15}, {1187, 3578, 1188}}], 
       PolygonBox[CompressedData["
1:eJwtnAnUTtXbh5/znGceMmdOMlSSSkgSmg0lqaRQkiRkCkmESCqpkIQkKiRl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        "]]}]}, {{}, {}, {
      {Hue[0.816901294068976, 1, 1], LineBox[{1, 2}]}, 
      {Hue[0.8192442904753057, 1, 1], LineBox[{2, 3}]}, 
      {Hue[0.8215872868816358, 1, 1], LineBox[{3, 4}]}, 
      {Hue[0.8262732796942951, 1, 1], LineBox[{4, 5}]}, 
      {Hue[0.8356452653196136, 1, 1], LineBox[{5, 6}]}, 
      {Hue[0.8543892365702518, 1, 1], LineBox[{6, 7}]}, 
      {Hue[0.8918771790715269, 1, 1], LineBox[{7, 8}]}, 
      {Hue[0.9668530640740771, 1, 1], LineBox[{8, 9}]}, 
      {Hue[0.12941796587021068`, 1, 1], LineBox[{9, 10}]}, 
      {Hue[0.2812099670030569, 1, 1], LineBox[{10, 11}]}, 
      {Hue[0.28353519270025984`, 1, 1], LineBox[{11, 12}]}, 
      {Hue[0.28586041839746285`, 1, 1], LineBox[{12, 13}]}, 
      {Hue[0.29051086979186935`, 1, 1], LineBox[{13, 14}]}, 
      {Hue[0.29981177258068187`, 1, 1], LineBox[{14, 15}]}, 
      {Hue[0.31841357815830684`, 1, 1], LineBox[{15, 16}]}, 
      {Hue[0.3207388038555099, 1, 1], LineBox[{16, 17}]}, 
      {Hue[0.32306402955271285`, 1, 1], LineBox[{17, 18}]}, 
      {Hue[0.3277144809471194, 1, 1], LineBox[{18, 19}]}, 
      {Hue[0.33701538373593193`, 1, 1], LineBox[{19, 20}]}, 
      {Hue[0.34631628652474383`, 1, 1], LineBox[{20, 21}]}, 
      {Hue[0.35561718931355635`, 1, 1], LineBox[{21, 22}]}, 
      {Hue[0.35794241501075935`, 1, 1], LineBox[{22, 23}]}, 
      {Hue[0.3602676407079629, 1, 1], LineBox[{23, 24}]}, 
      {Hue[0.36491809210236886`, 1, 1], LineBox[{24, 25}]}, 
      {Hue[0.3695685434967748, 1, 1], LineBox[{25, 26}]}, 
      {Hue[0.37421899489118077`, 1, 1], LineBox[{26, 27}]}, 
      {Hue[0.37886944628558666`, 1, 1], LineBox[{27, 28}]}, 
      {Hue[0.3835198976799933, 1, 1], LineBox[{28, 29}]}, 
      {Hue[0.38817034907439985`, 1, 1], LineBox[{29, 30}]}, 
      {Hue[0.3928208004688058, 1, 1], LineBox[{30, 31}]}, 
      {Hue[0.39747125186321175`, 1, 1], LineBox[{31, 32}]}, 
      {Hue[0.4021217032576183, 1, 1], LineBox[{32, 33}]}, 
      {Hue[0.4067721546520248, 1, 1], LineBox[{33, 34}]}, 
      {Hue[0.41142260604643077`, 1, 1], LineBox[{34, 35}]}, 
      {Hue[0.4160730574408368, 1, 1], LineBox[{35, 36}]}, 
      {Hue[0.4207235088352433, 1, 1], LineBox[{36, 37}]}, 
      {Hue[0.4230487345324463, 1, 1], LineBox[{37, 38}]}, 
      {Hue[0.42537396022964985`, 1, 1], LineBox[{38, 39}]}, 
      {Hue[0.4300244116240558, 1, 1], LineBox[{39, 40}]}, 
      {Hue[0.4325467175054938, 1, 1], LineBox[{40, 41}]}, 
      {Hue[0.43506902338693176`, 1, 1], LineBox[{41, 42}]}, 
      {Hue[0.43759132926836974`, 1, 1], LineBox[{42, 43}]}, 
      {Hue[0.44011363514980767`, 1, 1], LineBox[{43, 44}]}, 
      {Hue[0.44263594103124576`, 1, 1], LineBox[{44, 45}]}, 
      {Hue[0.44515824691268374`, 1, 1], LineBox[{45, 46}]}, 
      {Hue[0.4502028586755597, 1, 1], LineBox[{46, 47}]}, 
      {Hue[0.45272516455699774`, 1, 1], LineBox[{47, 48}]}, 
      {Hue[0.4552474704384357, 1, 1], LineBox[{48, 49}]}, 
      {Hue[0.4577697763198737, 1, 1], LineBox[{49, 50}]}, 
      {Hue[0.4602920822013117, 1, 1], LineBox[{50, 51}]}, 
      {Hue[0.46281438808274966`, 1, 1], LineBox[{51, 52}]}, 
      {Hue[0.46533669396418764`, 1, 1], LineBox[{52, 53}]}, 
      {Hue[0.47038130572706366`, 1, 1], LineBox[{53, 54}]}, 
      {Hue[0.47290361160850164`, 1, 1], LineBox[{54, 55}]}, 
      {Hue[0.4754259174899396, 1, 1], LineBox[{55, 56}]}, 
      {Hue[0.4779482233713776, 1, 1], LineBox[{56, 57}]}, 
      {Hue[0.4804705292528156, 1, 1], LineBox[{57, 58}]}, 
      {Hue[0.4829928351342536, 1, 1], LineBox[{58, 59}]}, 
      {Hue[0.4855151410156916, 1, 1], LineBox[{59, 60}]}, 
      {Hue[0.49055975277856756`, 1, 1], LineBox[{60, 61}]}, 
      {Hue[0.49308205866000554`, 1, 1], LineBox[{61, 62}]}, 
      {Hue[0.4956043645414435, 1, 1], LineBox[{62, 63}]}, 
      {Hue[0.49812667042288156`, 1, 1], LineBox[{63, 64}]}, 
      {Hue[0.5006489763043194, 1, 1], LineBox[{64, 65}]}, 
      {Hue[0.5031712821857575, 1, 1], LineBox[{65, 66}]}, 
      {Hue[0.5056935880671953, 1, 1], LineBox[{66, 67}]}, 
      {Hue[0.5107381998300711, 1, 1], LineBox[{67, 68}]}, 
      {Hue[0.5132605057115088, 1, 1], LineBox[{68, 69}]}, 
      {Hue[0.5157828115929468, 1, 1], LineBox[{69, 70}]}, 
      {Hue[0.5183051174743848, 1, 1], LineBox[{70, 71}]}, 
      {Hue[0.5208274233558228, 1, 1], LineBox[{71, 72}]}, 
      {Hue[0.5233497292372608, 1, 1], LineBox[{72, 73}]}, 
      {Hue[0.5258720351186987, 1, 1], LineBox[{73, 74}]}, 
      {Hue[0.5309166468815747, 1, 1], LineBox[{74, 75}]}, 
      {Hue[0.5334389527630128, 1, 1], LineBox[{75, 76}]}, 
      {Hue[0.5359612586444507, 1, 1], LineBox[{76, 77}]}, 
      {Hue[0.5384835645258887, 1, 1], LineBox[{77, 78}]}, 
      {Hue[0.5410058704073267, 1, 1], LineBox[{78, 79}]}, 
      {Hue[0.5435281762887647, 1, 1], LineBox[{79, 80}]}, 
      {Hue[0.5460504821702027, 1, 1], LineBox[{80, 81}]}, 
      {Hue[0.5510950939330787, 1, 1], LineBox[{81, 82}]}, 
      {Hue[0.5536173998145166, 1, 1], LineBox[{82, 83}]}, 
      {Hue[0.5561397056959547, 1, 1], LineBox[{83, 84}]}, 
      {Hue[0.5586620115773926, 1, 1], LineBox[{84, 85}]}, 
      {Hue[0.5611843174588307, 1, 1], LineBox[{85, 86}]}, 
      {Hue[0.5637066233402686, 1, 1], LineBox[{86, 87}]}, 
      {Hue[0.5662289292217066, 1, 1], LineBox[{87, 88}]}, 
      {Hue[0.5712735409845826, 1, 1], LineBox[{88, 89}]}, 
      {Hue[0.5737958468660206, 1, 1], LineBox[{89, 90}]}, 
      {Hue[0.5763181527474586, 1, 1], LineBox[{90, 91}]}, 
      {Hue[0.5788404586288965, 1, 1], LineBox[{91, 92}]}, 
      {Hue[0.5813627645103345, 1, 1], LineBox[{92, 93}]}, 
      {Hue[0.5838850703917726, 1, 1], LineBox[{93, 94}]}, 
      {Hue[0.5864073762732105, 1, 1], LineBox[{94, 95}]}, 
      {Hue[0.5914519880360865, 1, 1], LineBox[{95, 96}]}, 
      {Hue[0.5938059673446605, 1, 1], LineBox[{96, 97}]}, 
      {Hue[0.5961599466532348, 1, 1], LineBox[{97, 98}]}, 
      {Hue[0.6008679052703831, 1, 1], LineBox[{98, 99}]}, 
      {Hue[0.6032218845789571, 1, 1], LineBox[{99, 100}]}, 
      {Hue[0.6055758638875314, 1, 1], LineBox[{100, 101}]}, 
      {Hue[0.6102838225046796, 1, 1], LineBox[{101, 102}]}, 
      {Hue[0.6126378018132537, 1, 1], LineBox[{102, 103}]}, 
      {Hue[0.614991781121828, 1, 1], LineBox[{103, 104}]}, 
      {Hue[0.6196997397389763, 1, 1], LineBox[{104, 105}]}, 
      {Hue[0.6220537190475501, 1, 1], LineBox[{105, 106}]}, 
      {Hue[0.6244076983561242, 1, 1], LineBox[{106, 107}]}, 
      {Hue[0.6291156569732725, 1, 1], LineBox[{107, 108}]}, 
      {Hue[0.6314696362818468, 1, 1], LineBox[{108, 109}]}, 
      {Hue[0.6338236155904208, 1, 1], LineBox[{109, 110}]}, 
      {Hue[0.6385315742075688, 1, 1], LineBox[{110, 111}]}, 
      {Hue[0.6408855535161428, 1, 1], LineBox[{111, 112}]}, 
      {Hue[0.6432395328247171, 1, 1], LineBox[{112, 113}]}, 
      {Hue[0.6479474914418654, 1, 1], LineBox[{113, 114}]}, 
      {Hue[0.6503014707504393, 1, 1], LineBox[{114, 115}]}, 
      {Hue[0.6526554500590136, 1, 1], LineBox[{115, 116}]}, 
      {Hue[0.6573634086761619, 1, 1], LineBox[{116, 117}]}, 
      {Hue[0.659717387984736, 1, 1], LineBox[{117, 118}]}, 
      {Hue[0.6620713672933103, 1, 1], LineBox[{118, 119}]}, 
      {Hue[0.6667793259104585, 1, 1], LineBox[{119, 120}]}, 
      {Hue[0.6691333052190325, 1, 1], LineBox[{120, 121}]}, 
      {Hue[0.6714872845276068, 1, 1], LineBox[{121, 122}]}, 
      {Hue[0.6761952431447551, 1, 1], LineBox[{122, 123}]}, 
      {Hue[0.678549222453329, 1, 1], LineBox[{123, 124}]}, 
      {Hue[0.6809032017619033, 1, 1], LineBox[{124, 125}]}, 
      {Hue[0.6856111603790516, 1, 1], LineBox[{125, 126}]}, 
      {Hue[0.6879651396876256, 1, 1], LineBox[{126, 127}]}, 
      {Hue[0.6903191189961999, 1, 1], LineBox[{127, 128}]}, 
      {Hue[0.6950270776133483, 1, 1], LineBox[{128, 129}]}, 
      {Hue[0.6973810569219222, 1, 1], LineBox[{129, 130}]}, 
      {Hue[0.6997350362304962, 1, 1], LineBox[{130, 131}]}, 
      {Hue[0.7044429948476445, 1, 1], LineBox[{131, 132}]}, 
      {Hue[0.7067969741562188, 1, 1], LineBox[{132, 133}]}, 
      {Hue[0.7091509534647928, 1, 1], LineBox[{133, 134}]}, 
      {Hue[0.7138589120819407, 1, 1], LineBox[{134, 135}]}, 
      {Hue[0.7162128913905148, 1, 1], LineBox[{135, 136}]}, 
      {Hue[0.7185668706990891, 1, 1], LineBox[{136, 137}]}, 
      {Hue[0.7232748293162373, 1, 1], LineBox[{137, 138}]}, 
      {Hue[0.7256288086248113, 1, 1], LineBox[{138, 139}]}, 
      {Hue[0.7279827879333856, 1, 1], LineBox[{139, 140}]}, 
      {Hue[0.7326907465505339, 1, 1], LineBox[{140, 141}]}, 
      {Hue[0.735044725859108, 1, 1], LineBox[{141, 142}]}, 
      {Hue[0.7373987051676822, 1, 1], LineBox[{142, 143}]}, 
      {Hue[0.7421066637848305, 1, 1], LineBox[{143, 144}]}, 
      {Hue[0.7446577232776395, 1, 1], LineBox[{144, 145}]}, 
      {Hue[0.7472087827704486, 1, 1], LineBox[{145, 146}]}, 
      {Hue[0.7523109017560665, 1, 1], LineBox[{146, 147}]}, 
      {Hue[0.7548619612488755, 1, 1], LineBox[{147, 148}]}, 
      {Hue[0.7574130207416845, 1, 1], LineBox[{148, 149}]}, 
      {Hue[0.7625151397273026, 1, 1], LineBox[{149, 150}]}, 
      {Hue[0.7650661992201115, 1, 1], LineBox[{150, 151}]}, 
      {Hue[0.7676172587129205, 1, 1], LineBox[{151, 152}]}, 
      {Hue[0.7727193776985386, 1, 1], LineBox[{152, 153}]}, 
      {Hue[0.7752704371913476, 1, 1], LineBox[{153, 154}]}, 
      {Hue[0.7778214966841566, 1, 1], LineBox[{154, 155}]}, 
      {Hue[0.7829236156697745, 1, 1], LineBox[{155, 156}]}, 
      {Hue[0.7854746751625836, 1, 1], LineBox[{156, 157}]}, 
      {Hue[0.7880257346553926, 1, 1], LineBox[{157, 158}]}, 
      {Hue[0.7931278536410106, 1, 1], LineBox[{158, 159}]}, 
      {Hue[0.7956789131338196, 1, 1], LineBox[{159, 160}]}, 
      {Hue[0.7982299726266286, 1, 1], LineBox[{160, 161}]}, 
      {Hue[0.8033320916122466, 1, 1], LineBox[{161, 162}]}, 
      {Hue[0.8058831511050557, 1, 1], LineBox[{162, 163}]}, 
      {Hue[0.8084342105978647, 1, 1], LineBox[{163, 164}]}, 
      {Hue[0.8135363295834827, 1, 1], LineBox[{164, 165}]}, 
      {Hue[0.8160873890762916, 1, 1], LineBox[{165, 166}]}, 
      {Hue[0.8186384485691007, 1, 1], LineBox[{166, 167}]}, 
      {Hue[0.8237405675547187, 1, 1], LineBox[{167, 168}]}, 
      {Hue[0.8262916270475277, 1, 1], LineBox[{168, 169}]}, 
      {Hue[0.8288426865403367, 1, 1], LineBox[{169, 170}]}, 
      {Hue[0.8339448055259547, 1, 1], LineBox[{170, 171}]}, 
      {Hue[0.8364958650187637, 1, 1], LineBox[{171, 172}]}, 
      {Hue[0.8390469245115727, 1, 1], LineBox[{172, 173}]}, 
      {Hue[0.8441490434971908, 1, 1], LineBox[{173, 174}]}, 
      {Hue[0.8467001029899998, 1, 1], LineBox[{174, 175}]}, 
      {Hue[0.8492511624828087, 1, 1], LineBox[{175, 176}]}, 
      {Hue[0.8543532814684268, 1, 1], LineBox[{176, 177}]}, 
      {Hue[0.8569043409612358, 1, 1], LineBox[{177, 178}]}, 
      {Hue[0.8594554004540448, 1, 1], LineBox[{178, 179}]}, 
      {Hue[0.8645575194396629, 1, 1], LineBox[{179, 180}]}, 
      {Hue[0.8671085789324718, 1, 1], LineBox[{180, 181}]}, 
      {Hue[0.8696596384252808, 1, 1], LineBox[{181, 182}]}, 
      {Hue[0.8747617574108988, 1, 1], LineBox[{182, 183}]}, 
      {Hue[0.8773128169037079, 1, 1], LineBox[{183, 184}]}, 
      {Hue[0.8798638763965169, 1, 1], LineBox[{184, 185}]}, 
      {Hue[0.8849659953821348, 1, 1], LineBox[{185, 186}]}, 
      {Hue[0.8875170548749438, 1, 1], LineBox[{186, 187}]}, 
      {Hue[0.8900681143677529, 1, 1], LineBox[{187, 188}]}, 
      {Hue[0.8951702333533709, 1, 1], LineBox[{188, 189}]}, 
      {Hue[0.89772129284618, 1, 1], LineBox[{189, 190}]}, 
      {Hue[0.9002723523389888, 1, 1], LineBox[{190, 191}]}, 
      {Hue[0.9053744713246066, 1, 1], LineBox[{191, 192}]}, 
      {Hue[0.9078790064969182, 1, 1], LineBox[{192, 193}]}, 
      {Hue[0.9103835416692296, 1, 1], LineBox[{193, 194}]}, 
      {Hue[0.9153926120138524, 1, 1], LineBox[{194, 195}]}, 
      {Hue[0.9178971471861637, 1, 1], LineBox[{195, 196}]}, 
      {Hue[0.920401682358475, 1, 1], LineBox[{196, 197}]}, 
      {Hue[0.9254107527030977, 1, 1], LineBox[{197, 198}]}, 
      {Hue[0.9279152878754091, 1, 1], LineBox[{198, 199}]}, 
      {Hue[0.9304198230477204, 1, 1], LineBox[{199, 200}]}, 
      {Hue[0.9354288933923434, 1, 1], LineBox[{200, 201}]}, 
      {Hue[0.937933428564655, 1, 1], LineBox[{201, 202}]}, 
      {Hue[0.9404379637369664, 1, 1], LineBox[{202, 203}]}, 
      {Hue[0.945447034081589, 1, 1], LineBox[{203, 204}]}, 
      {Hue[0.9479515692539005, 1, 1], LineBox[{204, 205}]}, 
      {Hue[0.9504561044262118, 1, 1], LineBox[{205, 206}]}, 
      {Hue[0.9554651747708345, 1, 1], LineBox[{206, 207}]}, 
      {Hue[0.9579697099431458, 1, 1], LineBox[{207, 208}]}, 
      {Hue[0.9604742451154572, 1, 1], LineBox[{208, 209}]}, 
      {Hue[0.9654833154600801, 1, 1], LineBox[{209, 210}]}, 
      {Hue[0.9679878506323918, 1, 1], LineBox[{210, 211}]}, 
      {Hue[0.9704923858047031, 1, 1], LineBox[{211, 212}]}, 
      {Hue[0.9755014561493258, 1, 1], LineBox[{212, 213}]}, 
      {Hue[0.9780059913216371, 1, 1], LineBox[{213, 214}]}, 
      {Hue[0.9805105264939485, 1, 1], LineBox[{214, 215}]}, 
      {Hue[0.9855195968385713, 1, 1], LineBox[{215, 216}]}, 
      {Hue[0.9880241320108826, 1, 1], LineBox[{216, 217}]}, 
      {Hue[0.9905286671831939, 1, 1], LineBox[{217, 218}]}, 
      {Hue[0.9955377375278166, 1, 1], LineBox[{218, 219}]}, 
      {Hue[0.998042272700128, 1, 1], LineBox[{219, 220}]}, 
      {Hue[0.0005468078724393148, 1, 1], LineBox[{220, 221}]}, 
      {Hue[0.005555878217062295, 1, 1], LineBox[{221, 222}]}, 
      {Hue[0.008060413389373928, 1, 1], LineBox[{222, 223}]}, 
      {Hue[0.010564948561685276`, 1, 1], LineBox[{223, 224}]}, 
      {Hue[0.015574018906307979`, 1, 1], LineBox[{224, 225}]}, 
      {Hue[0.018078554078619323`, 1, 1], LineBox[{225, 226}]}, 
      {Hue[0.02058308925093067, 1, 1], LineBox[{226, 227}]}, 
      {Hue[0.025592159595553377`, 1, 1], LineBox[{227, 228}]}, 
      {Hue[0.02809669476786473, 1, 1], LineBox[{228, 229}]}, 
      {Hue[0.030601229940176073`, 1, 1], LineBox[{229, 230}]}, 
      {Hue[0.03561030028479876, 1, 1], LineBox[{230, 231}]}, 
      {Hue[0.03811483545711011, 1, 1], LineBox[{231, 232}]}, 
      {Hue[0.04061937062942147, 1, 1], LineBox[{232, 233}]}, 
      {Hue[0.045628440974044455`, 1, 1], LineBox[{233, 234}]}, 
      {Hue[0.048132976146356074`, 1, 1], LineBox[{234, 235}]}, 
      {Hue[0.05063751131866743, 1, 1], LineBox[{235, 236}]}, 
      {Hue[0.05564658166329014, 1, 1], LineBox[{236, 237}]}, 
      {Hue[0.05815111683560147, 1, 1], LineBox[{237, 238}]}, 
      {Hue[0.06065565200791283, 1, 1], LineBox[{238, 239}]}, 
      {Hue[0.06566472235253552, 1, 1], LineBox[{239, 240}]}, 
      {Hue[0.06800093095198288, 1, 1], LineBox[{240, 241}]}, 
      {Hue[0.0703371395514305, 1, 1], LineBox[{241, 242}]}, 
      {Hue[0.07500955675032554, 1, 1], LineBox[{242, 243}]}, 
      {Hue[0.07734576534977317, 1, 1], LineBox[{243, 244}]}, 
      {Hue[0.0796819739492208, 1, 1], LineBox[{244, 245}]}, 
      {Hue[0.08435439114811583, 1, 1], LineBox[{245, 246}]}, 
      {Hue[0.08669059974756345, 1, 1], LineBox[{246, 247}]}, 
      {Hue[0.08902680834701081, 1, 1], LineBox[{247, 248}]}, 
      {Hue[0.09369922554590611, 1, 1], LineBox[{248, 249}]}, 
      {Hue[0.1030440599436961, 1, 1], LineBox[{249, 250}]}, 
      {Hue[0.10538026854314346`, 1, 1], LineBox[{250, 251}]}, 
      {Hue[0.1077164771425911, 1, 1], LineBox[{251, 252}]}, 
      {Hue[0.1123888943414861, 1, 1], LineBox[{252, 253}]}, 
      {Hue[0.11472510294093347`, 1, 1], LineBox[{253, 254}]}, 
      {Hue[0.11706131154038113`, 1, 1], LineBox[{254, 255}]}, 
      {Hue[0.12173372873927613`, 1, 1], LineBox[{255, 256}]}, 
      {Hue[0.12406993733872347`, 1, 1], LineBox[{256, 257}]}, 
      {Hue[0.12640614593817112`, 1, 1], LineBox[{257, 258}]}, 
      {Hue[0.13107856313706612`, 1, 1], LineBox[{258, 259}]}, 
      {Hue[0.14042339753485641`, 1, 1], LineBox[{259, 260}]}, 
      {Hue[0.14275960613430405`, 1, 1], LineBox[{260, 261}]}, 
      {Hue[0.1450958147337514, 1, 1], LineBox[{261, 262}]}, 
      {Hue[0.14976823193264668`, 1, 1], LineBox[{262, 263}]}, 
      {Hue[0.15210444053209407`, 1, 1], LineBox[{263, 264}]}, 
      {Hue[0.1544406491315417, 1, 1], LineBox[{264, 265}]}, 
      {Hue[0.1591130663304367, 1, 1], LineBox[{265, 266}]}, 
      {Hue[0.16144927492988406`, 1, 1], LineBox[{266, 267}]}, 
      {Hue[0.1637854835293317, 1, 1], LineBox[{267, 268}]}, 
      {Hue[0.1684579007282267, 1, 1], LineBox[{268, 269}]}, 
      {Hue[0.1778027351260167, 1, 1], LineBox[{269, 270}]}, 
      {Hue[0.18013894372546407`, 1, 1], LineBox[{270, 271}]}, 
      {Hue[0.1824751523249117, 1, 1], LineBox[{271, 272}]}, 
      {Hue[0.1871475695238067, 1, 1], LineBox[{272, 273}]}, 
      {Hue[0.18948377812325434`, 1, 1], LineBox[{273, 274}]}, 
      {Hue[0.19181998672270198`, 1, 1], LineBox[{274, 275}]}, 
      {Hue[0.196492403921597, 1, 1], LineBox[{275, 276}]}, 
      {Hue[0.19882861252104464`, 1, 1], LineBox[{276, 277}]}, 
      {Hue[0.201164821120492, 1, 1], LineBox[{277, 278}]}, 
      {Hue[0.20583723831938727`, 1, 1], LineBox[{278, 279}]}, 
      {Hue[0.2151820727171773, 1, 1], LineBox[{279, 280}]}, 
      {Hue[0.21771536150085966`, 1, 1], LineBox[{280, 281}]}, 
      {Hue[0.22024865028454202`, 1, 1], LineBox[{281, 282}]}, 
      {Hue[0.22531522785190675`, 1, 1], LineBox[{282, 283}]}, 
      {Hue[0.2278485166355891, 1, 1], LineBox[{283, 284}]}, 
      {Hue[0.23038180541927147`, 1, 1], LineBox[{284, 285}]}, 
      {Hue[0.23544838298663648`, 1, 1], LineBox[{285, 286}]}, 
      {Hue[0.23798167177031915`, 1, 1], LineBox[{286, 287}]}, 
      {Hue[0.2405149605540015, 1, 1], LineBox[{287, 288}]}, 
      {Hue[0.24558153812136624`, 1, 1], LineBox[{288, 289}]}, 
      {Hue[0.2557146932560957, 1, 1], LineBox[{289, 290}]}, 
      {Hue[0.25824798203977806`, 1, 1], LineBox[{290, 291}]}, 
      {Hue[0.2607812708234604, 1, 1], LineBox[{291, 292}]}, 
      {Hue[0.26584784839082515`, 1, 1], LineBox[{292, 293}]}, 
      {Hue[0.2683811371745075, 1, 1], LineBox[{293, 294}]}, 
      {Hue[0.2709144259581899, 1, 1], LineBox[{294, 295}]}, 
      {Hue[0.27598100352555494`, 1, 1], LineBox[{295, 296}]}, 
      {Hue[0.2785142923092376, 1, 1], LineBox[{296, 297}]}, 
      {Hue[0.28104758109291994`, 1, 1], LineBox[{297, 298}]}, 
      {Hue[0.28611415866028467`, 1, 1], LineBox[{298, 299}]}, 
      {Hue[0.2962473137950141, 1, 1], LineBox[{299, 300}]}, 
      {Hue[0.2987806025786965, 1, 1], LineBox[{300, 301}]}, 
      {Hue[0.30131389136237885`, 1, 1], LineBox[{301, 302}]}, 
      {Hue[0.3063804689297436, 1, 1], LineBox[{302, 303}]}, 
      {Hue[0.30891375771342594`, 1, 1], LineBox[{303, 304}]}, 
      {Hue[0.3114470464971083, 1, 1], LineBox[{304, 305}]}, 
      {Hue[0.3165136240644733, 1, 1], LineBox[{305, 306}]}, 
      {Hue[0.31904691284815595`, 1, 1], LineBox[{306, 307}]}, 
      {Hue[0.32158020163183837`, 1, 1], LineBox[{307, 308}]}, 
      {Hue[0.3266467791992031, 1, 1], LineBox[{308, 309}]}, 
      {Hue[0.33677993433393255`, 1, 1], LineBox[{309, 310}]}, 
      {Hue[0.3393132231176149, 1, 1], LineBox[{310, 311}]}, 
      {Hue[0.3418465119012973, 1, 1], LineBox[{311, 312}]}, 
      {Hue[0.3469130894686623, 1, 1], LineBox[{312, 313}]}, 
      {Hue[0.3494463782523449, 1, 1], LineBox[{313, 314}]}, 
      {Hue[0.3519796670360273, 1, 1], LineBox[{314, 315}]}, 
      {Hue[0.357046244603392, 1, 1], LineBox[{315, 316}]}, 
      {Hue[0.3595795333870744, 1, 1], LineBox[{316, 317}]}, 
      {Hue[0.36211282217075674`, 1, 1], LineBox[{317, 318}]}, 
      {Hue[0.36717939973812147`, 1, 1], LineBox[{318, 319}]}, 
      {Hue[0.37731255487285126`, 1, 1], LineBox[{319, 320}]}, 
      {Hue[0.3796775170836699, 1, 1], LineBox[{320, 321}]}, 
      {Hue[0.38204247929448826`, 1, 1], LineBox[{321, 322}]}, 
      {Hue[0.3867724037161256, 1, 1], LineBox[{322, 323}]}, 
      {Hue[0.389137365926944, 1, 1], LineBox[{323, 324}]}, 
      {Hue[0.39150232813776265`, 1, 1], LineBox[{324, 325}]}, 
      {Hue[0.39623225255939964`, 1, 1], LineBox[{325, 326}]}, 
      {Hue[0.39859721477021803`, 1, 1], LineBox[{326, 327}]}, 
      {Hue[0.4009621769810367, 1, 1], LineBox[{327, 328}]}, 
      {Hue[0.40569210140267375`, 1, 1], LineBox[{328, 329}]}, 
      {Hue[0.4151519502459478, 1, 1], LineBox[{329, 330}]}, 
      {Hue[0.4175169124567662, 1, 1], LineBox[{330, 331}]}, 
      {Hue[0.41988187466758486`, 1, 1], LineBox[{331, 332}]}, 
      {Hue[0.42461179908922186`, 1, 1], LineBox[{332, 333}]}, 
      {Hue[0.42697676130004053`, 1, 1], LineBox[{333, 334}]}, 
      {Hue[0.4293417235108592, 1, 1], LineBox[{334, 335}]}, 
      {Hue[0.43407164793249625`, 1, 1], LineBox[{335, 336}]}, 
      {Hue[0.43643661014331486`, 1, 1], LineBox[{336, 337}]}, 
      {Hue[0.43880157235413325`, 1, 1], LineBox[{337, 338}]}, 
      {Hue[0.4435314967757706, 1, 1], LineBox[{338, 339}]}, 
      {Hue[0.45299134561904464`, 1, 1], LineBox[{339, 340}]}, 
      {Hue[0.455356307829863, 1, 1], LineBox[{340, 341}]}, 
      {Hue[0.4577212700406817, 1, 1], LineBox[{341, 342}]}, 
      {Hue[0.46245119446231875`, 1, 1], LineBox[{342, 343}]}, 
      {Hue[0.46481615667313736`, 1, 1], LineBox[{343, 344}]}, 
      {Hue[0.467181118883956, 1, 1], LineBox[{344, 345}]}, 
      {Hue[0.4719110433055931, 1, 1], LineBox[{345, 346}]}, 
      {Hue[0.47427600551641175`, 1, 1], LineBox[{346, 347}]}, 
      {Hue[0.47664096772723014`, 1, 1], LineBox[{347, 348}]}, 
      {Hue[0.4813708921488674, 1, 1], LineBox[{348, 349}]}, 
      {Hue[0.4908307409921415, 1, 1], LineBox[{349, 350}]}, 
      {Hue[0.49319570320295986`, 1, 1], LineBox[{350, 351}]}, 
      {Hue[0.4955606654137785, 1, 1], LineBox[{351, 352}]}, 
      {Hue[0.5002905898354155, 1, 1], LineBox[{352, 353}]}, 
      {Hue[0.5026555520462341, 1, 1], LineBox[{353, 354}]}, 
      {Hue[0.5050205142570529, 1, 1], LineBox[{354, 355}]}, 
      {Hue[0.50975043867869, 1, 1], LineBox[{355, 356}]}, 
      {Hue[0.5121154008895086, 1, 1], LineBox[{356, 357}]}, 
      {Hue[0.514480363100327, 1, 1], LineBox[{357, 358}]}, 
      {Hue[0.5192102875219642, 1, 1], LineBox[{358, 359}]}, 
      {Hue[0.5286701363652384, 1, 1], LineBox[{359, 360}]}, 
      {Hue[0.5309885742555593, 1, 1], LineBox[{360, 361}]}, 
      {Hue[0.5333070121458803, 1, 1], LineBox[{361, 362}]}, 
      {Hue[0.537943887926522, 1, 1], LineBox[{362, 363}]}, 
      {Hue[0.540262325816843, 1, 1], LineBox[{363, 364}]}, 
      {Hue[0.5425807637071637, 1, 1], LineBox[{364, 365}]}, 
      {Hue[0.5472176394878057, 1, 1], LineBox[{365, 366}]}, 
      {Hue[0.5495360773781268, 1, 1], LineBox[{366, 367}]}, 
      {Hue[0.5518545152684478, 1, 1], LineBox[{367, 368}]}, 
      {Hue[0.5564913910490895, 1, 1], LineBox[{368, 369}]}, 
      {Hue[0.5657651426103731, 1, 1], LineBox[{369, 370}]}, 
      {Hue[0.5680835805006943, 1, 1], LineBox[{370, 371}]}, 
      {Hue[0.5704020183910152, 1, 1], LineBox[{371, 372}]}, 
      {Hue[0.5750388941716569, 1, 1], LineBox[{372, 373}]}, 
      {Hue[0.5773573320619779, 1, 1], LineBox[{373, 374}]}, 
      {Hue[0.5796757699522987, 1, 1], LineBox[{374, 375}]}, 
      {Hue[0.5843126457329406, 1, 1], LineBox[{375, 376}]}, 
      {Hue[0.5866310836232617, 1, 1], LineBox[{376, 377}]}, 
      {Hue[0.5889495215135827, 1, 1], LineBox[{377, 378}]}, 
      {Hue[0.5935863972942244, 1, 1], LineBox[{378, 379}]}, 
      {Hue[0.6028601488555081, 1, 1], LineBox[{379, 380}]}, 
      {Hue[0.6051785867458291, 1, 1], LineBox[{380, 381}]}, 
      {Hue[0.6074970246361502, 1, 1], LineBox[{381, 382}]}, 
      {Hue[0.6121339004167918, 1, 1], LineBox[{382, 383}]}, 
      {Hue[0.6144523383071128, 1, 1], LineBox[{383, 384}]}, 
      {Hue[0.6167707761974335, 1, 1], LineBox[{384, 385}]}, 
      {Hue[0.6214076519780756, 1, 1], LineBox[{385, 386}]}, 
      {Hue[0.6237260898683965, 1, 1], LineBox[{386, 387}]}, 
      {Hue[0.6260445277587176, 1, 1], LineBox[{387, 388}]}, 
      {Hue[0.6306814035393593, 1, 1], LineBox[{388, 389}]}, 
      {Hue[0.639955155100643, 1, 1], LineBox[{389, 390}]}, 
      {Hue[0.642273592990964, 1, 1], LineBox[{390, 391}]}, 
      {Hue[0.644592030881285, 1, 1], LineBox[{391, 392}]}, 
      {Hue[0.6492289066619267, 1, 1], LineBox[{392, 393}]}, 
      {Hue[0.6585026582232103, 1, 1], LineBox[{393, 394}]}, 
      {Hue[0.6608210961135315, 1, 1], LineBox[{394, 395}]}, 
      {Hue[0.6631395340038524, 1, 1], LineBox[{395, 396}]}, 
      {Hue[0.6677764097844944, 1, 1], LineBox[{396, 397}]}, 
      {Hue[0.6770501613457782, 1, 1], LineBox[{397, 398}]}, 
      {Hue[0.6795656794203342, 1, 1], LineBox[{398, 399}]}, 
      {Hue[0.6820811974948899, 1, 1], LineBox[{399, 400}]}, 
      {Hue[0.6871122336440013, 1, 1], LineBox[{400, 401}]}, 
      {Hue[0.6896277517185571, 1, 1], LineBox[{401, 402}]}, 
      {Hue[0.6921432697931131, 1, 1], LineBox[{402, 403}]}, 
      {Hue[0.6971743059422248, 1, 1], LineBox[{403, 404}]}, 
      {Hue[0.6996898240167806, 1, 1], LineBox[{404, 405}]}, 
      {Hue[0.7022053420913362, 1, 1], LineBox[{405, 406}]}, 
      {Hue[0.707236378240448, 1, 1], LineBox[{406, 407}]}, 
      {Hue[0.7172984505386711, 1, 1], LineBox[{407, 408}]}, 
      {Hue[0.7198139686132269, 1, 1], LineBox[{408, 409}]}, 
      {Hue[0.7223294866877826, 1, 1], LineBox[{409, 410}]}, 
      {Hue[0.7273605228368943, 1, 1], LineBox[{410, 411}]}, 
      {Hue[0.7298760409114502, 1, 1], LineBox[{411, 412}]}, 
      {Hue[0.732391558986006, 1, 1], LineBox[{412, 413}]}, 
      {Hue[0.7374225951351177, 1, 1], LineBox[{413, 414}]}, 
      {Hue[0.7399381132096735, 1, 1], LineBox[{414, 415}]}, 
      {Hue[0.7424536312842293, 1, 1], LineBox[{415, 416}]}, 
      {Hue[0.747484667433341, 1, 1], LineBox[{416, 417}]}, 
      {Hue[0.7575467397315642, 1, 1], LineBox[{417, 418}]}, 
      {Hue[0.7600622578061199, 1, 1], LineBox[{418, 419}]}, 
      {Hue[0.7625777758806758, 1, 1], LineBox[{419, 420}]}, 
      {Hue[0.7676088120297874, 1, 1], LineBox[{420, 421}]}, 
      {Hue[0.7776708843280107, 1, 1], LineBox[{421, 422}]}, 
      {Hue[0.7801864024025665, 1, 1], LineBox[{422, 423}]}, 
      {Hue[0.7827019204771223, 1, 1], LineBox[{423, 424}]}, 
      {Hue[0.787732956626234, 1, 1], LineBox[{424, 425}]}, 
      {Hue[0.7977950289244572, 1, 1], LineBox[{425, 426}]}, 
      {Hue[0.800310546999013, 1, 1], LineBox[{426, 427}]}, 
      {Hue[0.8028260650735689, 1, 1], LineBox[{427, 428}]}, 
      {Hue[0.8078571012226805, 1, 1], LineBox[{428, 429}]}, 
      {Hue[0.8179191735209038, 1, 1], LineBox[{429, 430}]}, 
      {Hue[0.8204346915954596, 1, 1], LineBox[{430, 431}]}, 
      {Hue[0.8229502096700154, 1, 1], LineBox[{431, 432}]}, 
      {Hue[0.827981245819127, 1, 1], LineBox[{432, 433}]}, 
      {Hue[0.8380433181173503, 1, 1], LineBox[{433, 434}]}, 
      {Hue[0.8403905096190423, 1, 1], LineBox[{434, 435}]}, 
      {Hue[0.8427377011207342, 1, 1], LineBox[{435, 436}]}, 
      {Hue[0.8474320841241183, 1, 1], LineBox[{436, 437}]}, 
      {Hue[0.856820850130886, 1, 1], LineBox[{437, 438}]}, 
      {Hue[0.8591680416325781, 1, 1], LineBox[{438, 439}]}, 
      {Hue[0.8615152331342699, 1, 1], LineBox[{439, 440}]}, 
      {Hue[0.8662096161376538, 1, 1], LineBox[{440, 441}]}, 
      {Hue[0.8755983821444216, 1, 1], LineBox[{441, 442}]}, 
      {Hue[0.8779455736461137, 1, 1], LineBox[{442, 443}]}, 
      {Hue[0.8802927651478055, 1, 1], LineBox[{443, 444}]}, 
      {Hue[0.8849871481511894, 1, 1], LineBox[{444, 445}]}, 
      {Hue[0.8943759141579573, 1, 1], LineBox[{445, 446}]}, 
      {Hue[0.8967231056596492, 1, 1], LineBox[{446, 447}]}, 
      {Hue[0.8990702971613411, 1, 1], LineBox[{447, 448}]}, 
      {Hue[0.903764680164725, 1, 1], LineBox[{448, 449}]}, 
      {Hue[0.9131534461714927, 1, 1], LineBox[{449, 450}]}, 
      {Hue[0.9155006376731848, 1, 1], LineBox[{450, 451}]}, 
      {Hue[0.9178478291748767, 1, 1], LineBox[{451, 452}]}, 
      {Hue[0.9225422121782606, 1, 1], LineBox[{452, 453}]}, 
      {Hue[0.9319309781850283, 1, 1], LineBox[{453, 454}]}, 
      {Hue[0.9342781696867204, 1, 1], LineBox[{454, 455}]}, 
      {Hue[0.9366253611884122, 1, 1], LineBox[{455, 456}]}, 
      {Hue[0.9413197441917962, 1, 1], LineBox[{456, 457}]}, 
      {Hue[0.9507085101985641, 1, 1], LineBox[{457, 458}]}, 
      {Hue[0.953055701700256, 1, 1], LineBox[{458, 459}]}, 
      {Hue[0.9554028932019479, 1, 1], LineBox[{459, 460}]}, 
      {Hue[0.960097276205332, 1, 1], LineBox[{460, 461}]}, 
      {Hue[0.9694860422120998, 1, 1], LineBox[{461, 462}]}, 
      {Hue[0.9882635742256354, 1, 1], LineBox[{462, 463}]}, 
      {Hue[0.9908078459115621, 1, 1], LineBox[{463, 464}]}, 
      {Hue[0.993352117597489, 1, 1], LineBox[{464, 465}]}, 
      {Hue[0.9984406609693426, 1, 1], LineBox[{465, 466}]}, 
      {Hue[0.00861774771305003, 1, 1], LineBox[{466, 467}]}, 
      {Hue[0.011162019398976914`, 1, 1], LineBox[{467, 468}]}, 
      {Hue[0.013706291084903658`, 1, 1], LineBox[{468, 469}]}, 
      {Hue[0.018794834456757423`, 1, 1], LineBox[{469, 470}]}, 
      {Hue[0.02897192120046467, 1, 1], LineBox[{470, 471}]}, 
      {Hue[0.03151619288639141, 1, 1], LineBox[{471, 472}]}, 
      {Hue[0.034060464572318296`, 1, 1], LineBox[{472, 473}]}, 
      {Hue[0.03914900794417192, 1, 1], LineBox[{473, 474}]}, 
      {Hue[0.04932609468787931, 1, 1], LineBox[{474, 475}]}, 
      {Hue[0.06968026817529394, 1, 1], LineBox[{475, 476}]}, 
      {Hue[0.07222453986122068, 1, 1], LineBox[{476, 477}]}, 
      {Hue[0.07476881154714757, 1, 1], LineBox[{477, 478}]}, 
      {Hue[0.07985735491900121, 1, 1], LineBox[{478, 479}]}, 
      {Hue[0.09003444166270858, 1, 1], LineBox[{479, 480}]}, 
      {Hue[0.09257871334863547, 1, 1], LineBox[{480, 481}]}, 
      {Hue[0.09512298503456218, 1, 1], LineBox[{481, 482}]}, 
      {Hue[0.10021152840641596`, 1, 1], LineBox[{482, 483}]}, 
      {Hue[0.11038861515012321`, 1, 1], LineBox[{483, 484}]}, 
      {Hue[0.1129328868360501, 1, 1], LineBox[{484, 485}]}, 
      {Hue[0.115477158521977, 1, 1], LineBox[{485, 486}]}, 
      {Hue[0.1205657018938306, 1, 1], LineBox[{486, 487}]}, 
      {Hue[0.130742788637538, 1, 1], LineBox[{487, 488}]}, 
      {Hue[0.15109696212495263`, 1, 1], LineBox[{488, 489}]}, 
      {Hue[0.15359470949038187`, 1, 1], LineBox[{489, 490}]}, 
      {Hue[0.15609245685581094`, 1, 1], LineBox[{490, 491}]}, 
      {Hue[0.1610879515866694, 1, 1], LineBox[{491, 492}]}, 
      {Hue[0.1710789410483863, 1, 1], LineBox[{492, 493}]}, 
      {Hue[0.17357668841381552`, 1, 1], LineBox[{493, 494}]}, 
      {Hue[0.17607443577924475`, 1, 1], LineBox[{494, 495}]}, 
      {Hue[0.1810699305101032, 1, 1], LineBox[{495, 496}]}, 
      {Hue[0.1910609199718201, 1, 1], LineBox[{496, 497}]}, 
      {Hue[0.19355866733724933`, 1, 1], LineBox[{497, 498}]}, 
      {Hue[0.19605641470267854`, 1, 1], LineBox[{498, 499}]}, 
      {Hue[0.20105190943353698`, 1, 1], LineBox[{499, 500}]}, 
      {Hue[0.2110428988952539, 1, 1], LineBox[{500, 501}]}, 
      {Hue[0.2310248778186877, 1, 1], LineBox[{501, 502}]}, 
      {Hue[0.23352262518411693`, 1, 1], LineBox[{502, 503}]}, 
      {Hue[0.23602037254954614`, 1, 1], LineBox[{503, 504}]}, 
      {Hue[0.2410158672804046, 1, 1], LineBox[{504, 505}]}, 
      {Hue[0.2510068567421215, 1, 1], LineBox[{505, 506}]}, 
      {Hue[0.27098883566555515`, 1, 1], LineBox[{506, 507}]}, 
      {Hue[0.2734865830309844, 1, 1], LineBox[{507, 508}]}, 
      {Hue[0.27598433039641346`, 1, 1], LineBox[{508, 509}]}, 
      {Hue[0.28097982512727193`, 1, 1], LineBox[{509, 510}]}, 
      {Hue[0.2909708145889888, 1, 1], LineBox[{510, 511}]}, 
      {Hue[0.3109527935124226, 1, 1], LineBox[{511, 512}]}, 
      {Hue[0.313282214304988, 1, 1], LineBox[{512, 513}]}, 
      {Hue[0.31561163509755336`, 1, 1], LineBox[{513, 514}]}, 
      {Hue[0.32027047668268416`, 1, 1], LineBox[{514, 515}]}, 
      {Hue[0.3295881598529455, 1, 1], LineBox[{515, 516}]}, 
      {Hue[0.3482235261934684, 1, 1], LineBox[{516, 517}]}, 
      {Hue[0.3505529469860338, 1, 1], LineBox[{517, 518}]}, 
      {Hue[0.35288236777859916`, 1, 1], LineBox[{518, 519}]}, 
      {Hue[0.3575412093637299, 1, 1], LineBox[{519, 520}]}, 
      {Hue[0.3668588925339913, 1, 1], LineBox[{520, 521}]}, 
      {Hue[0.38549425887451416`, 1, 1], LineBox[{521, 522}]}, 
      {Hue[0.3878236796670795, 1, 1], LineBox[{522, 523}]}, 
      {Hue[0.39015310045964485`, 1, 1], LineBox[{523, 524}]}, 
      {Hue[0.3948119420447756, 1, 1], LineBox[{524, 525}]}, 
      {Hue[0.40412962521503704`, 1, 1], LineBox[{525, 526}]}, 
      {Hue[0.4227649915555599, 1, 1], LineBox[{526, 527}]}, 
      {Hue[0.46003572423660566`, 1, 1], LineBox[{527, 528}]}, 
      {Hue[0.4625622252134059, 1, 1], LineBox[{528, 529}]}, 
      {Hue[0.46508872619020614`, 1, 1], LineBox[{529, 530}]}, 
      {Hue[0.4701417281438066, 1, 1], LineBox[{530, 531}]}, 
      {Hue[0.48024773205100757`, 1, 1], LineBox[{531, 532}]}, 
      {Hue[0.5004597398654094, 1, 1], LineBox[{532, 533}]}, 
      {Hue[0.5408837554942132, 1, 1], LineBox[{533, 534}]}, 
      {Hue[0.5434102564710135, 1, 1], LineBox[{534, 535}]}, 
      {Hue[0.5459367574478137, 1, 1], LineBox[{535, 536}]}, 
      {Hue[0.5509897594014141, 1, 1], LineBox[{536, 537}]}, 
      {Hue[0.5610957633086151, 1, 1], LineBox[{537, 538}]}, 
      {Hue[0.581307771123017, 1, 1], LineBox[{538, 539}]}, 
      {Hue[0.6217317867518208, 1, 1], LineBox[{539, 540}]}, 
      {Hue[0.6240899611557571, 1, 1], LineBox[{540, 541}]}, 
      {Hue[0.6264481355596936, 1, 1], LineBox[{541, 542}]}, 
      {Hue[0.6311644843675663, 1, 1], LineBox[{542, 543}]}, 
      {Hue[0.6405971819833118, 1, 1], LineBox[{543, 544}]}, 
      {Hue[0.6594625772148028, 1, 1], LineBox[{544, 545}]}, 
      {Hue[0.6971933676777848, 1, 1], LineBox[{545, 546}]}, 
      {Hue[0.6995515420817213, 1, 1], LineBox[{546, 547}]}, 
      {Hue[0.7019097164856576, 1, 1], LineBox[{547, 548}]}, 
      {Hue[0.7066260652935304, 1, 1], LineBox[{548, 549}]}, 
      {Hue[0.716058762909276, 1, 1], LineBox[{549, 550}]}, 
      {Hue[0.734924158140767, 1, 1], LineBox[{550, 551}]}, 
      {Hue[0.772654948603749, 1, 1], LineBox[{551, 552}]}, 
      {Hue[0.7752102031919202, 1, 1], LineBox[{552, 553}]}, 
      {Hue[0.7777654577800914, 1, 1], LineBox[{553, 554}]}, 
      {Hue[0.782875966956434, 1, 1], LineBox[{554, 555}]}, 
      {Hue[0.793096985309119, 1, 1], LineBox[{555, 556}]}, 
      {Hue[0.8135390220144891, 1, 1], LineBox[{556, 557}]}, 
      {Hue[0.8160942766026603, 1, 1], LineBox[{557, 558}]}, 
      {Hue[0.8186495311908315, 1, 1], LineBox[{558, 559}]}, 
      {Hue[0.8237600403671741, 1, 1], LineBox[{559, 560}]}, 
      {Hue[0.8339810587198591, 1, 1], LineBox[{560, 561}]}, 
      {Hue[0.8544230954252292, 1, 1], LineBox[{561, 562}]}, 
      {Hue[0.8569783500134004, 1, 1], LineBox[{562, 563}]}, 
      {Hue[0.8595336046015717, 1, 1], LineBox[{563, 564}]}, 
      {Hue[0.8646441137779142, 1, 1], LineBox[{564, 565}]}, 
      {Hue[0.8748651321305992, 1, 1], LineBox[{565, 566}]}, 
      {Hue[0.8774203867187704, 1, 1], LineBox[{566, 567}]}, 
      {Hue[0.8799756413069417, 1, 1], LineBox[{567, 568}]}, 
      {Hue[0.8850861504832842, 1, 1], LineBox[{568, 569}]}, 
      {Hue[0.8953071688359693, 1, 1], LineBox[{569, 570}]}, 
      {Hue[0.8978624234241405, 1, 1], LineBox[{570, 571}]}, 
      {Hue[0.9004176780123118, 1, 1], LineBox[{571, 572}]}, 
      {Hue[0.9055281871886542, 1, 1], LineBox[{572, 573}]}, 
      {Hue[0.9157492055413393, 1, 1], LineBox[{573, 574}]}, 
      {Hue[0.9183044601295106, 1, 1], LineBox[{574, 575}]}, 
      {Hue[0.9208597147176818, 1, 1], LineBox[{575, 576}]}, 
      {Hue[0.9259702238940244, 1, 1], LineBox[{576, 577}]}, 
      {Hue[0.9361912422467094, 1, 1], LineBox[{577, 578}]}, 
      {Hue[0.938699972514383, 1, 1], LineBox[{578, 579}]}, 
      {Hue[0.9412087027820567, 1, 1], LineBox[{579, 580}]}, 
      {Hue[0.946226163317404, 1, 1], LineBox[{580, 581}]}, 
      {Hue[0.9487348935850776, 1, 1], LineBox[{581, 582}]}, 
      {Hue[0.9512436238527512, 1, 1], LineBox[{582, 583}]}, 
      {Hue[0.9562610843880985, 1, 1], LineBox[{583, 584}]}, 
      {Hue[0.9587698146557722, 1, 1], LineBox[{584, 585}]}, 
      {Hue[0.9612785449234458, 1, 1], LineBox[{585, 586}]}, 
      {Hue[0.9637872751911194, 1, 1], LineBox[{586, 587}]}, 
      {Hue[0.9662960054587931, 1, 1], LineBox[{587, 588}]}, 
      {Hue[0.9688047357264667, 1, 1], LineBox[{588, 589}]}, 
      {Hue[0.9713134659941404, 1, 1], LineBox[{589, 590}]}, 
      {Hue[0.973822196261814, 1, 1], LineBox[{590, 591}]}, 
      {Hue[0.9763309265294876, 1, 1], LineBox[{591, 592}]}, 
      {Hue[0.9788396567971613, 1, 1], LineBox[{592, 593}]}, 
      {Hue[0.981348387064835, 1, 1], LineBox[{593, 594}]}, 
      {Hue[0.9838571173325086, 1, 1], LineBox[{594, 595}]}, 
      {Hue[0.9863658476001822, 1, 1], LineBox[{595, 596}]}, 
      {Hue[0.9888745778678558, 1, 1], LineBox[{596, 597}]}, 
      {Hue[0.9913833081355296, 1, 1], LineBox[{597, 598}]}, 
      {Hue[0.9938920384032032, 1, 1], LineBox[{598, 599}]}, 
      {Hue[0.9964007686708768, 1, 1], LineBox[{599, 600}]}, 
      {Hue[0.9989094989385504, 1, 1], LineBox[{600, 601}]}, 
      {Hue[0.0014182292062240736`, 1, 1], LineBox[{601, 602}]}, 
      {Hue[0.003926959473897716, 1, 1], LineBox[{602, 603}]}, 
      {Hue[0.0064356897415713595`, 1, 1], LineBox[{603, 604}]}, 
      {Hue[0.008944420009244999, 1, 1], LineBox[{604, 605}]}, 
      {Hue[0.011453150276918643`, 1, 1], LineBox[{605, 606}]}, 
      {Hue[0.013961880544592286`, 1, 1], LineBox[{606, 607}]}, 
      {Hue[0.016470610812265928`, 1, 1], LineBox[{607, 608}]}, 
      {Hue[0.018979341079939573`, 1, 1], LineBox[{608, 609}]}, 
      {Hue[0.021488071347613213`, 1, 1], LineBox[{609, 610}]}, 
      {Hue[0.023996801615286858`, 1, 1], LineBox[{610, 611}]}, 
      {Hue[0.026505531882960495`, 1, 1], LineBox[{611, 612}]}, 
      {Hue[0.02901426215063414, 1, 1], LineBox[{612, 613}]}, 
      {Hue[0.031522992418307784`, 1, 1], LineBox[{613, 614}]}, 
      {Hue[0.03654045295365507, 1, 1], LineBox[{614, 615}]}, 
      {Hue[0.0390491832213287, 1, 1], LineBox[{615, 616}]}, 
      {Hue[0.041557913489002354`, 1, 1], LineBox[{616, 617}]}, 
      {Hue[0.04657537402434964, 1, 1], LineBox[{617, 618}]}, 
      {Hue[0.04908410429202328, 1, 1], LineBox[{618, 619}]}, 
      {Hue[0.05159283455969691, 1, 1], LineBox[{619, 620}]}, 
      {Hue[0.05661029509504419, 1, 1], LineBox[{620, 621}]}, 
      {Hue[0.05911902536271785, 1, 1], LineBox[{621, 622}]}, 
      {Hue[0.061627755630391495`, 1, 1], LineBox[{622, 623}]}, 
      {Hue[0.06664521616573878, 1, 1], LineBox[{623, 624}]}, 
      {Hue[0.06915394643341242, 1, 1], LineBox[{624, 625}]}, 
      {Hue[0.07166267670108606, 1, 1], LineBox[{625, 626}]}, 
      {Hue[0.07668013723643334, 1, 1], LineBox[{626, 627}]}, 
      {Hue[0.07918886750410699, 1, 1], LineBox[{627, 628}]}, 
      {Hue[0.08169759777178064, 1, 1], LineBox[{628, 629}]}, 
      {Hue[0.08671505830712793, 1, 1], LineBox[{629, 630}]}, 
      {Hue[0.0967499793778225, 1, 1], LineBox[{630, 631}]}, 
      {Hue[0.09909038307263227, 1, 1], LineBox[{631, 632}]}, 
      {Hue[0.10143078676744204`, 1, 1], LineBox[{632, 633}]}, 
      {Hue[0.10611159415706164`, 1, 1], LineBox[{633, 634}]}, 
      {Hue[0.11547320893630074`, 1, 1], LineBox[{634, 635}]}, 
      {Hue[0.134196438494779, 1, 1], LineBox[{635, 636}]}, 
      {Hue[0.1365368421895888, 1, 1], LineBox[{636, 637}]}, 
      {Hue[0.13887724588439856`, 1, 1], LineBox[{637, 638}]}, 
      {Hue[0.14355805327401813`, 1, 1], LineBox[{638, 639}]}, 
      {Hue[0.15291966805325727`, 1, 1], LineBox[{639, 640}]}, 
      {Hue[0.1716428976117355, 1, 1], LineBox[{640, 641}]}, 
      {Hue[0.1739833013065453, 1, 1], LineBox[{641, 642}]}, 
      {Hue[0.17632370500135508`, 1, 1], LineBox[{642, 643}]}, 
      {Hue[0.18100451239097462`, 1, 1], LineBox[{643, 644}]}, 
      {Hue[0.19036612717021376`, 1, 1], LineBox[{644, 645}]}, 
      {Hue[0.20908935672869203`, 1, 1], LineBox[{645, 646}]}, 
      {Hue[0.24653581584564851`, 1, 1], LineBox[{646, 647}]}, 
      {Hue[0.24907329972469316`, 1, 1], LineBox[{647, 648}]}, 
      {Hue[0.25161078360373784`, 1, 1], LineBox[{648, 649}]}, 
      {Hue[0.25668575136182714`, 1, 1], LineBox[{649, 650}]}, 
      {Hue[0.2668356868780058, 1, 1], LineBox[{650, 651}]}, 
      {Hue[0.2871355579103631, 1, 1], LineBox[{651, 652}]}, 
      {Hue[0.32773529997507767`, 1, 1], LineBox[{652, 653}]}, 
      {Hue[0.3302727838541223, 1, 1], LineBox[{653, 654}]}, 
      {Hue[0.332810267733167, 1, 1], LineBox[{654, 655}]}, 
      {Hue[0.3378852354912563, 1, 1], LineBox[{655, 656}]}, 
      {Hue[0.348035171007435, 1, 1], LineBox[{656, 657}]}, 
      {Hue[0.3683350420397923, 1, 1], LineBox[{657, 658}]}, 
      {Hue[0.40893478410450684`, 1, 1], LineBox[{658, 659}]}, 
      {Hue[0.41130394141068766`, 1, 1], LineBox[{659, 660}]}, 
      {Hue[0.41367309871686847`, 1, 1], LineBox[{660, 661}]}, 
      {Hue[0.4184114133292301, 1, 1], LineBox[{661, 662}]}, 
      {Hue[0.4278880425539533, 1, 1], LineBox[{662, 663}]}, 
      {Hue[0.4468413010033997, 1, 1], LineBox[{663, 664}]}, 
      {Hue[0.4847478179022925, 1, 1], LineBox[{664, 665}]}, 
      {Hue[0.48711697520847325`, 1, 1], LineBox[{665, 666}]}, 
      {Hue[0.48948613251465406`, 1, 1], LineBox[{666, 667}]}, 
      {Hue[0.4942244471270157, 1, 1], LineBox[{667, 668}]}, 
      {Hue[0.5037010763517389, 1, 1], LineBox[{668, 669}]}, 
      {Hue[0.5226543348011852, 1, 1], LineBox[{669, 670}]}, 
      {Hue[0.5605608517000781, 1, 1], LineBox[{670, 671}]}, 
      {Hue[0.5628834846857612, 1, 1], LineBox[{671, 672}]}, 
      {Hue[0.5652061176714445, 1, 1], LineBox[{672, 673}]}, 
      {Hue[0.5698513836428108, 1, 1], LineBox[{673, 674}]}, 
      {Hue[0.5791419155855435, 1, 1], LineBox[{674, 675}]}, 
      {Hue[0.597722979471009, 1, 1], LineBox[{675, 676}]}, 
      {Hue[0.6000456124566923, 1, 1], LineBox[{676, 677}]}, 
      {Hue[0.6023682454423754, 1, 1], LineBox[{677, 678}]}, 
      {Hue[0.6070135114137418, 1, 1], LineBox[{678, 679}]}, 
      {Hue[0.6163040433564745, 1, 1], LineBox[{679, 680}]}, 
      {Hue[0.63488510724194, 1, 1], LineBox[{680, 681}]}, 
      {Hue[0.6372077402276233, 1, 1], LineBox[{681, 682}]}, 
      {Hue[0.6395303732133064, 1, 1], LineBox[{682, 683}]}, 
      {Hue[0.6441756391846729, 1, 1], LineBox[{683, 684}]}, 
      {Hue[0.6534661711274056, 1, 1], LineBox[{684, 685}]}, 
      {Hue[0.672047235012871, 1, 1], LineBox[{685, 686}]}, 
      {Hue[0.6743698679985541, 1, 1], LineBox[{686, 687}]}, 
      {Hue[0.6766925009842375, 1, 1], LineBox[{687, 688}]}, 
      {Hue[0.6813377669556038, 1, 1], LineBox[{688, 689}]}, 
      {Hue[0.6906282988983365, 1, 1], LineBox[{689, 690}]}, 
      {Hue[0.709209362783802, 1, 1], LineBox[{690, 691}]}, 
      {Hue[0.71172907595372, 1, 1], LineBox[{691, 692}]}, 
      {Hue[0.7142487891236381, 1, 1], LineBox[{692, 693}]}, 
      {Hue[0.7192882154634741, 1, 1], LineBox[{693, 694}]}, 
      {Hue[0.7293670681431466, 1, 1], LineBox[{694, 695}]}, 
      {Hue[0.7318867813130645, 1, 1], LineBox[{695, 696}]}, 
      {Hue[0.7344064944829827, 1, 1], LineBox[{696, 697}]}, 
      {Hue[0.7394459208228188, 1, 1], LineBox[{697, 698}]}, 
      {Hue[0.749524773502491, 1, 1], LineBox[{698, 699}]}, 
      {Hue[0.7520444866724092, 1, 1], LineBox[{699, 700}]}, 
      {Hue[0.7545641998423271, 1, 1], LineBox[{700, 701}]}, 
      {Hue[0.7596036261821633, 1, 1], LineBox[{701, 702}]}, 
      {Hue[0.7696824788618355, 1, 1], LineBox[{702, 703}]}, 
      {Hue[0.7898401842211801, 1, 1], LineBox[{703, 704}]}, 
      {Hue[0.7923598973910981, 1, 1], LineBox[{704, 705}]}, 
      {Hue[0.7948796105610161, 1, 1], LineBox[{705, 706}]}, 
      {Hue[0.7999190369008522, 1, 1], LineBox[{706, 707}]}, 
      {Hue[0.8099978895805245, 1, 1], LineBox[{707, 708}]}, 
      {Hue[0.8125176027504426, 1, 1], LineBox[{708, 709}]}, 
      {Hue[0.8150373159203607, 1, 1], LineBox[{709, 710}]}, 
      {Hue[0.8200767422601969, 1, 1], LineBox[{710, 711}]}, 
      {Hue[0.8301555949398691, 1, 1], LineBox[{711, 712}]}, 
      {Hue[0.8326753081097872, 1, 1], LineBox[{712, 713}]}, 
      {Hue[0.8351950212797054, 1, 1], LineBox[{713, 714}]}, 
      {Hue[0.8402344476195414, 1, 1], LineBox[{714, 715}]}, 
      {Hue[0.8503133002992137, 1, 1], LineBox[{715, 716}]}, 
      {Hue[0.8704710056585583, 1, 1], LineBox[{716, 717}]}, 
      {Hue[0.8728223922556125, 1, 1], LineBox[{717, 718}]}, 
      {Hue[0.8751737788526667, 1, 1], LineBox[{718, 719}]}, 
      {Hue[0.8798765520467752, 1, 1], LineBox[{719, 720}]}, 
      {Hue[0.8892820984349921, 1, 1], LineBox[{720, 721}]}, 
      {Hue[0.8916334850320461, 1, 1], LineBox[{721, 722}]}, 
      {Hue[0.8939848716291003, 1, 1], LineBox[{722, 723}]}, 
      {Hue[0.8986876448232088, 1, 1], LineBox[{723, 724}]}, 
      {Hue[0.9080931912114256, 1, 1], LineBox[{724, 725}]}, 
      {Hue[0.9104445778084798, 1, 1], LineBox[{725, 726}]}, 
      {Hue[0.912795964405534, 1, 1], LineBox[{726, 727}]}, 
      {Hue[0.9174987375996424, 1, 1], LineBox[{727, 728}]}, 
      {Hue[0.9269042839878592, 1, 1], LineBox[{728, 729}]}, 
      {Hue[0.9457153767642927, 1, 1], LineBox[{729, 730}]}, 
      {Hue[0.9480667633613469, 1, 1], LineBox[{730, 731}]}, 
      {Hue[0.9504181499584011, 1, 1], LineBox[{731, 732}]}, 
      {Hue[0.9551209231525096, 1, 1], LineBox[{732, 733}]}, 
      {Hue[0.9645264695407264, 1, 1], LineBox[{733, 734}]}, 
      {Hue[0.9668778561377805, 1, 1], LineBox[{734, 735}]}, 
      {Hue[0.9692292427348348, 1, 1], LineBox[{735, 736}]}, 
      {Hue[0.9739320159289432, 1, 1], LineBox[{736, 737}]}, 
      {Hue[0.9833375623171601, 1, 1], LineBox[{737, 738}]}, 
      {Hue[0.9856889489142143, 1, 1], LineBox[{738, 739}]}, 
      {Hue[0.9880403355112685, 1, 1], LineBox[{739, 740}]}, 
      {Hue[0.9927431087053771, 1, 1], LineBox[{740, 741}]}, 
      {Hue[0.0021486550935938373`, 1, 1], LineBox[{741, 742}]}, 
      {Hue[0.004500041690647965, 1, 1], LineBox[{742, 743}]}, 
      {Hue[0.006851428287702235, 1, 1], LineBox[{743, 744}]}, 
      {Hue[0.011554201481810633`, 1, 1], LineBox[{744, 745}]}, 
      {Hue[0.020959747870027433`, 1, 1], LineBox[{745, 746}]}, 
      {Hue[0.02350821465131656, 1, 1], LineBox[{746, 747}]}, 
      {Hue[0.02605668143260555, 1, 1], LineBox[{747, 748}]}, 
      {Hue[0.031153614995183685`, 1, 1], LineBox[{748, 749}]}, 
      {Hue[0.041347482120340076`, 1, 1], LineBox[{749, 750}]}, 
      {Hue[0.04389594890162907, 1, 1], LineBox[{750, 751}]}, 
      {Hue[0.04644441568291821, 1, 1], LineBox[{751, 752}]}, 
      {Hue[0.051541349245496464`, 1, 1], LineBox[{752, 753}]}, 
      {Hue[0.06173521637065272, 1, 1], LineBox[{753, 754}]}, 
      {Hue[0.06428368315194186, 1, 1], LineBox[{754, 755}]}, 
      {Hue[0.066832149933231, 1, 1], LineBox[{755, 756}]}, 
      {Hue[0.07192908349580913, 1, 1], LineBox[{756, 757}]}, 
      {Hue[0.08212295062096552, 1, 1], LineBox[{757, 758}]}, 
      {Hue[0.08467141740225466, 1, 1], LineBox[{758, 759}]}, 
      {Hue[0.08721988418354365, 1, 1], LineBox[{759, 760}]}, 
      {Hue[0.09231681774612178, 1, 1], LineBox[{760, 761}]}, 
      {Hue[0.10251068487127818`, 1, 1], LineBox[{761, 762}]}, 
      {Hue[0.10505915165256717`, 1, 1], LineBox[{762, 763}]}, 
      {Hue[0.10760761843385629`, 1, 1], LineBox[{763, 764}]}, 
      {Hue[0.11270455199643457`, 1, 1], LineBox[{764, 765}]}, 
      {Hue[0.12289841912159083`, 1, 1], LineBox[{765, 766}]}, 
      {Hue[0.12544688590287997`, 1, 1], LineBox[{766, 767}]}, 
      {Hue[0.1279953526841691, 1, 1], LineBox[{767, 768}]}, 
      {Hue[0.13309228624674724`, 1, 1], LineBox[{768, 769}]}, 
      {Hue[0.14328615337190362`, 1, 1], LineBox[{769, 770}]}, 
      {Hue[0.14583462015319276`, 1, 1], LineBox[{770, 771}]}, 
      {Hue[0.14838308693448174`, 1, 1], LineBox[{771, 772}]}, 
      {Hue[0.1534800204970599, 1, 1], LineBox[{772, 773}]}, 
      {Hue[0.16367388762221627`, 1, 1], LineBox[{773, 774}]}, 
      {Hue[0.16622235440350527`, 1, 1], LineBox[{774, 775}]}, 
      {Hue[0.1687708211847944, 1, 1], LineBox[{775, 776}]}, 
      {Hue[0.17386775474737268`, 1, 1], LineBox[{776, 777}]}, 
      {Hue[0.18406162187252892`, 1, 1], LineBox[{777, 778}]}, 
      {Hue[0.18656356433332041`, 1, 1], LineBox[{778, 779}]}, 
      {Hue[0.18906550679411188`, 1, 1], LineBox[{779, 780}]}, 
      {Hue[0.19406939171569484`, 1, 1], LineBox[{780, 781}]}, 
      {Hue[0.20407716155886074`, 1, 1], LineBox[{781, 782}]}, 
      {Hue[0.2065791040196522, 1, 1], LineBox[{782, 783}]}, 
      {Hue[0.2090810464804437, 1, 1], LineBox[{783, 784}]}, 
      {Hue[0.21408493140202664`, 1, 1], LineBox[{784, 785}]}, 
      {Hue[0.22409270124519243`, 1, 1], LineBox[{785, 786}]}, 
      {Hue[0.22659464370598376`, 1, 1], LineBox[{786, 787}]}, 
      {Hue[0.22909658616677522`, 1, 1], LineBox[{787, 788}]}, 
      {Hue[0.2341004710883582, 1, 1], LineBox[{788, 789}]}, 
      {Hue[0.24410824093152408`, 1, 1], LineBox[{789, 790}]}, 
      {Hue[0.24661018339231555`, 1, 1], LineBox[{790, 791}]}, 
      {Hue[0.24911212585310705`, 1, 1], LineBox[{791, 792}]}, 
      {Hue[0.25411601077469, 1, 1], LineBox[{792, 793}]}, 
      {Hue[0.2641237806178559, 1, 1], LineBox[{793, 794}]}, 
      {Hue[0.2666257230786474, 1, 1], LineBox[{794, 795}]}, 
      {Hue[0.26912766553943873`, 1, 1], LineBox[{795, 796}]}, 
      {Hue[0.2741315504610215, 1, 1], LineBox[{796, 797}]}, 
      {Hue[0.276633492921813, 1, 1], LineBox[{797, 798}]}, 
      {Hue[0.2791354353826045, 1, 1], LineBox[{798, 799}]}, 
      {Hue[0.2841393203041874, 1, 1], LineBox[{799, 800}]}, 
      {Hue[0.2866412627649789, 1, 1], LineBox[{800, 801}]}, 
      {Hue[0.28914320522577036`, 1, 1], LineBox[{801, 802}]}, 
      {Hue[0.29414709014735335`, 1, 1], LineBox[{802, 803}]}, 
      {Hue[0.3041548599905192, 1, 1], LineBox[{803, 804}]}, 
      {Hue[0.3066568024513107, 1, 1], LineBox[{804, 805}]}, 
      {Hue[0.3091587449121022, 1, 1], LineBox[{805, 806}]}, 
      {Hue[0.31416262983368515`, 1, 1], LineBox[{806, 807}]}, 
      {Hue[0.31666457229447664`, 1, 1], LineBox[{807, 808}]}, 
      {Hue[0.31916651475526814`, 1, 1], LineBox[{808, 809}]}, 
      {Hue[0.324170399676851, 1, 1], LineBox[{809, 810}]}, 
      {Hue[0.3266723421376425, 1, 1], LineBox[{810, 811}]}, 
      {Hue[0.329174284598434, 1, 1], LineBox[{811, 812}]}, 
      {Hue[0.33417816952001694`, 1, 1], LineBox[{812, 813}]}, 
      {Hue[0.3441859393631829, 1, 1], LineBox[{813, 814}]}, 
      {Hue[0.3465195552511106, 1, 1], LineBox[{814, 815}]}, 
      {Hue[0.3488531711390381, 1, 1], LineBox[{815, 816}]}, 
      {Hue[0.35352040291489345`, 1, 1], LineBox[{816, 817}]}, 
      {Hue[0.3558540188028212, 1, 1], LineBox[{817, 818}]}, 
      {Hue[0.3581876346907486, 1, 1], LineBox[{818, 819}]}, 
      {Hue[0.36285486646660386`, 1, 1], LineBox[{819, 820}]}, 
      {Hue[0.3651884823545317, 1, 1], LineBox[{820, 821}]}, 
      {Hue[0.3675220982424591, 1, 1], LineBox[{821, 822}]}, 
      {Hue[0.37218933001831445`, 1, 1], LineBox[{822, 823}]}, 
      {Hue[0.38152379357002464`, 1, 1], LineBox[{823, 824}]}, 
      {Hue[0.38385740945795216`, 1, 1], LineBox[{824, 825}]}, 
      {Hue[0.38619102534587996`, 1, 1], LineBox[{825, 826}]}, 
      {Hue[0.39085825712173494`, 1, 1], LineBox[{826, 827}]}, 
      {Hue[0.3931918730096627, 1, 1], LineBox[{827, 828}]}, 
      {Hue[0.3955254888975901, 1, 1], LineBox[{828, 829}]}, 
      {Hue[0.4001927206734454, 1, 1], LineBox[{829, 830}]}, 
      {Hue[0.40252633656137315`, 1, 1], LineBox[{830, 831}]}, 
      {Hue[0.4048599524493006, 1, 1], LineBox[{831, 832}]}, 
      {Hue[0.4095271842251559, 1, 1], LineBox[{832, 833}]}, 
      {Hue[0.41886164777686646`, 1, 1], LineBox[{833, 834}]}, 
      {Hue[0.42119526366479415`, 1, 1], LineBox[{834, 835}]}, 
      {Hue[0.4235288795527217, 1, 1], LineBox[{835, 836}]}, 
      {Hue[0.42819611132857693`, 1, 1], LineBox[{836, 837}]}, 
      {Hue[0.4305297272165047, 1, 1], LineBox[{837, 838}]}, 
      {Hue[0.4328633431044322, 1, 1], LineBox[{838, 839}]}, 
      {Hue[0.4375305748802874, 1, 1], LineBox[{839, 840}]}, 
      {Hue[0.4398641907682152, 1, 1], LineBox[{840, 841}]}, 
      {Hue[0.4421978066561428, 1, 1], LineBox[{841, 842}]}, 
      {Hue[0.44686503843199804`, 1, 1], LineBox[{842, 843}]}, 
      {Hue[0.4561995019837082, 1, 1], LineBox[{843, 844}]}, 
      {Hue[0.45853311787163575`, 1, 1], LineBox[{844, 845}]}, 
      {Hue[0.46086673375956344`, 1, 1], LineBox[{845, 846}]}, 
      {Hue[0.4655339655354184, 1, 1], LineBox[{846, 847}]}, 
      {Hue[0.46786758142334617`, 1, 1], LineBox[{847, 848}]}, 
      {Hue[0.4702011973112737, 1, 1], LineBox[{848, 849}]}, 
      {Hue[0.4748684290871289, 1, 1], LineBox[{849, 850}]}, 
      {Hue[0.4772020449750567, 1, 1], LineBox[{850, 851}]}, 
      {Hue[0.47953566086298427`, 1, 1], LineBox[{851, 852}]}, 
      {Hue[0.4842028926388394, 1, 1], LineBox[{852, 853}]}, 
      {Hue[0.49353735619054995`, 1, 1], LineBox[{853, 854}]}, 
      {Hue[0.49606805226271244`, 1, 1], LineBox[{854, 855}]}, 
      {Hue[0.49859874833487494`, 1, 1], LineBox[{855, 856}]}, 
      {Hue[0.5036601404791998, 1, 1], LineBox[{856, 857}]}, 
      {Hue[0.5061908365513623, 1, 1], LineBox[{857, 858}]}, 
      {Hue[0.5087215326235248, 1, 1], LineBox[{858, 859}]}, 
      {Hue[0.5137829247678498, 1, 1], LineBox[{859, 860}]}, 
      {Hue[0.5163136208400123, 1, 1], LineBox[{860, 861}]}, 
      {Hue[0.5188443169121748, 1, 1], LineBox[{861, 862}]}, 
      {Hue[0.5239057090564998, 1, 1], LineBox[{862, 863}]}, 
      {Hue[0.5340284933451498, 1, 1], LineBox[{863, 864}]}, 
      {Hue[0.5365591894173123, 1, 1], LineBox[{864, 865}]}, 
      {Hue[0.5390898854894748, 1, 1], LineBox[{865, 866}]}, 
      {Hue[0.5441512776337998, 1, 1], LineBox[{866, 867}]}, 
      {Hue[0.5466819737059623, 1, 1], LineBox[{867, 868}]}, 
      {Hue[0.5492126697781248, 1, 1], LineBox[{868, 869}]}, 
      {Hue[0.5542740619224498, 1, 1], LineBox[{869, 870}]}, 
      {Hue[0.5568047579946123, 1, 1], LineBox[{870, 871}]}, 
      {Hue[0.5593354540667748, 1, 1], LineBox[{871, 872}]}, 
      {Hue[0.5643968462110998, 1, 1], LineBox[{872, 873}]}, 
      {Hue[0.5745196304997497, 1, 1], LineBox[{873, 874}]}, 
      {Hue[0.5770503265719122, 1, 1], LineBox[{874, 875}]}, 
      {Hue[0.5795810226440747, 1, 1], LineBox[{875, 876}]}, 
      {Hue[0.5846424147883997, 1, 1], LineBox[{876, 877}]}, 
      {Hue[0.5871731108605622, 1, 1], LineBox[{877, 878}]}, 
      {Hue[0.5897038069327247, 1, 1], LineBox[{878, 879}]}, 
      {Hue[0.5947651990770497, 1, 1], LineBox[{879, 880}]}, 
      {Hue[0.5972958951492122, 1, 1], LineBox[{880, 881}]}, 
      {Hue[0.5998265912213747, 1, 1], LineBox[{881, 882}]}, 
      {Hue[0.6048879833656997, 1, 1], LineBox[{882, 883}]}, 
      {Hue[0.6150107676543496, 1, 1], LineBox[{883, 884}]}, 
      {Hue[0.6175414637265121, 1, 1], LineBox[{884, 885}]}, 
      {Hue[0.6200721597986746, 1, 1], LineBox[{885, 886}]}, 
      {Hue[0.6251335519429996, 1, 1], LineBox[{886, 887}]}, 
      {Hue[0.6276642480151621, 1, 1], LineBox[{887, 888}]}, 
      {Hue[0.6301949440873246, 1, 1], LineBox[{888, 889}]}, 
      {Hue[0.6352563362316496, 1, 1], LineBox[{889, 890}]}, 
      {Hue[0.6377870323038121, 1, 1], LineBox[{890, 891}]}, 
      {Hue[0.6403177283759746, 1, 1], LineBox[{891, 892}]}, 
      {Hue[0.6453791205202996, 1, 1], LineBox[{892, 893}]}, 
      {Hue[0.6555019048089492, 1, 1], LineBox[{893, 894}]}, 
      {Hue[0.6578642743082477, 1, 1], LineBox[{894, 895}]}, 
      {Hue[0.6602266438075466, 1, 1], LineBox[{895, 896}]}, 
      {Hue[0.6649513828061435, 1, 1], LineBox[{896, 897}]}, 
      {Hue[0.6673137523054423, 1, 1], LineBox[{897, 898}]}, 
      {Hue[0.6696761218047409, 1, 1], LineBox[{898, 899}]}, 
      {Hue[0.674400860803338, 1, 1], LineBox[{899, 900}]}, 
      {Hue[0.676763230302637, 1, 1], LineBox[{900, 901}]}, 
      {Hue[0.6791255998019354, 1, 1], LineBox[{901, 902}]}, 
      {Hue[0.6838503388005327, 1, 1], LineBox[{902, 903}]}, 
      {Hue[0.6932998167977273, 1, 1], LineBox[{903, 904}]}, 
      {Hue[0.695662186297026, 1, 1], LineBox[{904, 905}]}, 
      {Hue[0.6980245557963245, 1, 1], LineBox[{905, 906}]}, 
      {Hue[0.7027492947949219, 1, 1], LineBox[{906, 907}]}, 
      {Hue[0.7051116642942206, 1, 1], LineBox[{907, 908}]}, 
      {Hue[0.7074740337935191, 1, 1], LineBox[{908, 909}]}, 
      {Hue[0.7121987727921164, 1, 1], LineBox[{909, 910}]}, 
      {Hue[0.7145611422914152, 1, 1], LineBox[{910, 911}]}, 
      {Hue[0.7169235117907138, 1, 1], LineBox[{911, 912}]}, 
      {Hue[0.7216482507893109, 1, 1], LineBox[{912, 913}]}, 
      {Hue[0.7310977287865053, 1, 1], LineBox[{913, 914}]}, 
      {Hue[0.7334600982858038, 1, 1], LineBox[{914, 915}]}, 
      {Hue[0.7358224677851026, 1, 1], LineBox[{915, 916}]}, 
      {Hue[0.7405472067836996, 1, 1], LineBox[{916, 917}]}, 
      {Hue[0.7429095762829985, 1, 1], LineBox[{917, 918}]}, 
      {Hue[0.7452719457822969, 1, 1], LineBox[{918, 919}]}, 
      {Hue[0.7499966847808942, 1, 1], LineBox[{919, 920}]}, 
      {Hue[0.752359054280193, 1, 1], LineBox[{920, 921}]}, 
      {Hue[0.7547214237794915, 1, 1], LineBox[{921, 922}]}, 
      {Hue[0.7594461627780887, 1, 1], LineBox[{922, 923}]}, 
      {Hue[0.7688956407752834, 1, 1], LineBox[{923, 924}]}, 
      {Hue[0.7712580102745822, 1, 1], LineBox[{924, 925}]}, 
      {Hue[0.7736203797738807, 1, 1], LineBox[{925, 926}]}, 
      {Hue[0.7783451187724779, 1, 1], LineBox[{926, 927}]}, 
      {Hue[0.7807074882717767, 1, 1], LineBox[{927, 928}]}, 
      {Hue[0.7830698577710753, 1, 1], LineBox[{928, 929}]}, 
      {Hue[0.7877945967696725, 1, 1], LineBox[{929, 930}]}, 
      {Hue[0.7901569662689714, 1, 1], LineBox[{930, 931}]}, 
      {Hue[0.7925193357682698, 1, 1], LineBox[{931, 932}]}, 
      {Hue[0.7972440747668671, 1, 1], LineBox[{932, 933}]}, 
      {Hue[0.8066935527640614, 1, 1], LineBox[{933, 934}]}, 
      {Hue[0.8090093979428622, 1, 1], LineBox[{934, 935}]}, 
      {Hue[0.8113252431216633, 1, 1], LineBox[{935, 936}]}, 
      {Hue[0.8159569334792656, 1, 1], LineBox[{936, 937}]}, 
      {Hue[0.8182727786580668, 1, 1], LineBox[{937, 938}]}, 
      {Hue[0.8205886238368676, 1, 1], LineBox[{938, 939}]}, 
      {Hue[0.8252203141944696, 1, 1], LineBox[{939, 940}]}, 
      {Hue[0.8275361593732707, 1, 1], LineBox[{940, 941}]}, 
      {Hue[0.8298520045520715, 1, 1], LineBox[{941, 942}]}, 
      {Hue[0.8344836949096734, 1, 1], LineBox[{942, 943}]}, 
      {Hue[0.8437470756248775, 1, 1], LineBox[{943, 944}]}, 
      {Hue[0.8460629208036786, 1, 1], LineBox[{944, 945}]}, 
      {Hue[0.8483787659824795, 1, 1], LineBox[{945, 946}]}, 
      {Hue[0.8530104563400814, 1, 1], LineBox[{946, 947}]}, 
      {Hue[0.8553263015188826, 1, 1], LineBox[{947, 948}]}, 
      {Hue[0.8576421466976837, 1, 1], LineBox[{948, 949}]}, 
      {Hue[0.8622738370552856, 1, 1], LineBox[{949, 950}]}, 
      {Hue[0.8645896822340865, 1, 1], LineBox[{950, 951}]}, 
      {Hue[0.8669055274128876, 1, 1], LineBox[{951, 952}]}, 
      {Hue[0.8715372177704899, 1, 1], LineBox[{952, 953}]}, 
      {Hue[0.8808005984856938, 1, 1], LineBox[{953, 954}]}, 
      {Hue[0.8831164436644949, 1, 1], LineBox[{954, 955}]}, 
      {Hue[0.8854322888432958, 1, 1], LineBox[{955, 956}]}, 
      {Hue[0.8900639792008977, 1, 1], LineBox[{956, 957}]}, 
      {Hue[0.892379824379699, 1, 1], LineBox[{957, 958}]}, 
      {Hue[0.8946956695585001, 1, 1], LineBox[{958, 959}]}, 
      {Hue[0.899327359916102, 1, 1], LineBox[{959, 960}]}, 
      {Hue[0.9016432050949029, 1, 1], LineBox[{960, 961}]}, 
      {Hue[0.903959050273704, 1, 1], LineBox[{961, 962}]}, 
      {Hue[0.9085907406313063, 1, 1], LineBox[{962, 963}]}, 
      {Hue[0.9178541213465102, 1, 1], LineBox[{963, 964}]}, 
      {Hue[0.9201699665253114, 1, 1], LineBox[{964, 965}]}, 
      {Hue[0.9224858117041121, 1, 1], LineBox[{965, 966}]}, 
      {Hue[0.9271175020617142, 1, 1], LineBox[{966, 967}]}, 
      {Hue[0.9294333472405153, 1, 1], LineBox[{967, 968}]}, 
      {Hue[0.9317491924193162, 1, 1], LineBox[{968, 969}]}, 
      {Hue[0.9363808827769181, 1, 1], LineBox[{969, 970}]}, 
      {Hue[0.9386967279557192, 1, 1], LineBox[{970, 971}]}, 
      {Hue[0.9410125731345201, 1, 1], LineBox[{971, 972}]}, 
      {Hue[0.945644263492122, 1, 1], LineBox[{972, 973}]}, 
      {Hue[0.9479601086709232, 1, 1], LineBox[{973, 974}]}, 
      {Hue[0.9502759538497243, 1, 1], LineBox[{974, 975}]}, 
      {Hue[0.9549076442073263, 1, 1], LineBox[{975, 976}]}, 
      {Hue[0.9574205695703621, 1, 1], LineBox[{976, 977}]}, 
      {Hue[0.9599334949333983, 1, 1], LineBox[{977, 978}]}, 
      {Hue[0.96495934565947, 1, 1], LineBox[{978, 979}]}, 
      {Hue[0.9674722710225059, 1, 1], LineBox[{979, 980}]}, 
      {Hue[0.9699851963855417, 1, 1], LineBox[{980, 981}]}, 
      {Hue[0.9750110471116135, 1, 1], LineBox[{981, 982}]}, 
      {Hue[0.9775239724746493, 1, 1], LineBox[{982, 983}]}, 
      {Hue[0.9800368978376851, 1, 1], LineBox[{983, 984}]}, 
      {Hue[0.9850627485637571, 1, 1], LineBox[{984, 985}]}, 
      {Hue[0.9875756739267929, 1, 1], LineBox[{985, 986}]}, 
      {Hue[0.9900885992898291, 1, 1], LineBox[{986, 987}]}, 
      {Hue[0.9951144500159008, 1, 1], LineBox[{987, 988}]}, 
      {Hue[0.9976273753789366, 1, 1], LineBox[{988, 989}]}, 
      {Hue[0.00014030074197249827`, 1, 1], LineBox[{989, 990}]}, 
      {Hue[0.005166151468044201, 1, 1], LineBox[{990, 991}]}, 
      {Hue[0.007679076831080055, 1, 1], LineBox[{991, 992}]}, 
      {Hue[0.010192002194115906`, 1, 1], LineBox[{992, 993}]}, 
      {Hue[0.015217852920187897`, 1, 1], LineBox[{993, 994}]}, 
      {Hue[0.017730778283223746`, 1, 1], LineBox[{994, 995}]}, 
      {Hue[0.02024370364625988, 1, 1], LineBox[{995, 996}]}, 
      {Hue[0.025269554372331585`, 1, 1], LineBox[{996, 997}]}, 
      {Hue[0.027782479735367443`, 1, 1], LineBox[{997, 998}]}, 
      {Hue[0.030295405098403293`, 1, 1], LineBox[{998, 999}]}, 
      {Hue[0.035321255824475, 1, 1], LineBox[{999, 1000}]}, 
      {Hue[0.037834181187510856`, 1, 1], LineBox[{1000, 1001}]}, 
      {Hue[0.04034710655054671, 1, 1], LineBox[{1001, 1002}]}, 
      {Hue[0.045372957276618414`, 1, 1], LineBox[{1002, 1003}]}, 
      {Hue[0.04788588263965426, 1, 1], LineBox[{1003, 1004}]}, 
      {Hue[0.05039880800269011, 1, 1], LineBox[{1004, 1005}]}, 
      {Hue[0.0554246587287621, 1, 1], LineBox[{1005, 1006}]}, 
      {Hue[0.05793758409179796, 1, 1], LineBox[{1006, 1007}]}, 
      {Hue[0.06045050945483409, 1, 1], LineBox[{1007, 1008}]}, 
      {Hue[0.06547636018090579, 1, 1], LineBox[{1008, 1009}]}, 
      {Hue[0.06798928554394164, 1, 1], LineBox[{1009, 1010}]}, 
      {Hue[0.0705022109069775, 1, 1], LineBox[{1010, 1011}]}, 
      {Hue[0.0755280616330492, 1, 1], LineBox[{1011, 1012}]}, 
      {Hue[0.07804098699608507, 1, 1], LineBox[{1012, 1013}]}, 
      {Hue[0.08055391235912089, 1, 1], LineBox[{1013, 1014}]}, 
      {Hue[0.08557976308519263, 1, 1], LineBox[{1014, 1015}]}, 
      {Hue[0.08809268844822847, 1, 1], LineBox[{1015, 1016}]}, 
      {Hue[0.09060561381126431, 1, 1], LineBox[{1016, 1017}]}, 
      {Hue[0.09563146453733631, 1, 1], LineBox[{1017, 1018}]}, 
      {Hue[0.09814438990037216, 1, 1], LineBox[{1018, 1019}]}, 
      {Hue[0.10065731526340828`, 1, 1], LineBox[{1019, 1020}]}, 
      {Hue[0.10568316598948, 1, 1], LineBox[{1020, 1021}]}, 
      {Hue[0.10819609135251586`, 1, 1], LineBox[{1021, 1022}]}, 
      {Hue[0.1107090167155517, 1, 1], LineBox[{1022, 1023}]}, 
      {Hue[0.1157348674416234, 1, 1], LineBox[{1023, 1024}]}, 
      {Hue[0.11807946623179555`, 1, 1], LineBox[{1024, 1025}]}, 
      {Hue[0.12042406502196741`, 1, 1], LineBox[{1025, 1026}]}, 
      {Hue[0.12511326260231143`, 1, 1], LineBox[{1026, 1027}]}, 
      {Hue[0.12745786139248358`, 1, 1], LineBox[{1027, 1028}]}, 
      {Hue[0.1298024601826557, 1, 1], LineBox[{1028, 1029}]}, 
      {Hue[0.1344916577629997, 1, 1], LineBox[{1029, 1030}]}, 
      {Hue[0.1368362565531716, 1, 1], LineBox[{1030, 1031}]}, 
      {Hue[0.13918085534334373`, 1, 1], LineBox[{1031, 1032}]}, 
      {Hue[0.14387005292368804`, 1, 1], LineBox[{1032, 1033}]}, 
      {Hue[0.14621465171386017`, 1, 1], LineBox[{1033, 1034}]}, 
      {Hue[0.14855925050403201`, 1, 1], LineBox[{1034, 1035}]}, 
      {Hue[0.15324844808437604`, 1, 1], LineBox[{1035, 1036}]}, 
      {Hue[0.15559304687454817`, 1, 1], LineBox[{1036, 1037}]}, 
      {Hue[0.15793764566472007`, 1, 1], LineBox[{1037, 1038}]}, 
      {Hue[0.16262684324506407`, 1, 1], LineBox[{1038, 1039}]}, 
      {Hue[0.16497144203523623`, 1, 1], LineBox[{1039, 1040}]}, 
      {Hue[0.16731604082540807`, 1, 1], LineBox[{1040, 1041}]}, 
      {Hue[0.17200523840575208`, 1, 1], LineBox[{1041, 1042}]}, 
      {Hue[0.17434983719592423`, 1, 1], LineBox[{1042, 1043}]}, 
      {Hue[0.1766944359860961, 1, 1], LineBox[{1043, 1044}]}, 
      {Hue[0.1813836335664401, 1, 1], LineBox[{1044, 1045}]}, 
      {Hue[0.18372823235661226`, 1, 1], LineBox[{1045, 1046}]}, 
      {Hue[0.18607283114678438`, 1, 1], LineBox[{1046, 1047}]}, 
      {Hue[0.1907620287271284, 1, 1], LineBox[{1047, 1048}]}, 
      {Hue[0.19310662751730026`, 1, 1], LineBox[{1048, 1049}]}, 
      {Hue[0.19545122630747241`, 1, 1], LineBox[{1049, 1050}]}, 
      {Hue[0.2001404238878167, 1, 1], LineBox[{1050, 1051}]}, 
      {Hue[0.20248502267798885`, 1, 1], LineBox[{1051, 1052}]}, 
      {Hue[0.20482962146816072`, 1, 1], LineBox[{1052, 1053}]}, 
      {Hue[0.20951881904850472`, 1, 1], LineBox[{1053, 1054}]}, 
      {Hue[0.21186341783867682`, 1, 1], LineBox[{1054, 1055}]}, 
      {Hue[0.21420801662884872`, 1, 1], LineBox[{1055, 1056}]}, 
      {Hue[0.21889721420919273`, 1, 1], LineBox[{1056, 1057}]}, 
      {Hue[0.22124181299936488`, 1, 1], LineBox[{1057, 1058}]}, 
      {Hue[0.22358641178953675`, 1, 1], LineBox[{1058, 1059}]}, 
      {Hue[0.22827560936988076`, 1, 1], LineBox[{1059, 1060}]}, 
      {Hue[0.2306202081600529, 1, 1], LineBox[{1060, 1061}]}, 
      {Hue[0.23296480695022476`, 1, 1], LineBox[{1061, 1062}]}, 
      {Hue[0.23765400453056879`, 1, 1], LineBox[{1062, 1063}]}, 
      {Hue[0.2399986033207409, 1, 1], LineBox[{1063, 1064}]}, 
      {Hue[0.24234320211091306`, 1, 1], LineBox[{1064, 1065}]}, 
      {Hue[0.24703239969125707`, 1, 1], LineBox[{1065, 1066}]}, 
      {Hue[0.24937699848142894`, 1, 1], LineBox[{1066, 1067}]}, 
      {Hue[0.2517215972716011, 1, 1], LineBox[{1067, 1068}]}, 
      {Hue[0.2564107948519454, 1, 1], LineBox[{1068, 1069}]}, 
      {Hue[0.2587553936421175, 1, 1], LineBox[{1069, 1070}]}, 
      {Hue[0.2610999924322894, 1, 1], LineBox[{1070, 1071}]}, 
      {Hue[0.2657891900126334, 1, 1], LineBox[{1071, 1072}]}, 
      {Hue[0.2683308689870403, 1, 1], LineBox[{1072, 1073}]}, 
      {Hue[0.2708725479614471, 1, 1], LineBox[{1073, 1074}]}, 
      {Hue[0.27595590591026087`, 1, 1], LineBox[{1074, 1075}]}, 
      {Hue[0.27849758488466775`, 1, 1], LineBox[{1075, 1076}]}, 
      {Hue[0.28103926385907463`, 1, 1], LineBox[{1076, 1077}]}, 
      {Hue[0.2861226218078886, 1, 1], LineBox[{1077, 1078}]}, 
      {Hue[0.2886643007822955, 1, 1], LineBox[{1078, 1079}]}, 
      {Hue[0.29120597975670265`, 1, 1], LineBox[{1079, 1080}]}, 
      {Hue[0.2962893377055164, 1, 1], LineBox[{1080, 1081}]}, 
      {Hue[0.29883101667992323`, 1, 1], LineBox[{1081, 1082}]}, 
      {Hue[0.3013726956543301, 1, 1], LineBox[{1082, 1083}]}, 
      {Hue[0.3064560536031439, 1, 1], LineBox[{1083, 1084}]}, 
      {Hue[0.30899773257755075`, 1, 1], LineBox[{1084, 1085}]}, 
      {Hue[0.3115394115519576, 1, 1], LineBox[{1085, 1086}]}, 
      {Hue[0.3166227695007716, 1, 1], LineBox[{1086, 1087}]}, 
      {Hue[0.3191644484751785, 1, 1], LineBox[{1087, 1088}]}, 
      {Hue[0.32170612744958565`, 1, 1], LineBox[{1088, 1089}]}, 
      {Hue[0.32678948539839936`, 1, 1], LineBox[{1089, 1090}]}, 
      {Hue[0.32933116437280624`, 1, 1], LineBox[{1090, 1091}]}, 
      {Hue[0.3318728433472131, 1, 1], LineBox[{1091, 1092}]}, 
      {Hue[0.3369562012960269, 1, 1], LineBox[{1092, 1093}]}, 
      {Hue[0.3394978802704337, 1, 1], LineBox[{1093, 1094}]}, 
      {Hue[0.3420395592448406, 1, 1], LineBox[{1094, 1095}]}, 
      {Hue[0.3471229171936546, 1, 1], LineBox[{1095, 1096}]}, 
      {Hue[0.3496645961680615, 1, 1], LineBox[{1096, 1097}]}, 
      {Hue[0.35220627514246866`, 1, 1], LineBox[{1097, 1098}]}, 
      {Hue[0.35728963309128237`, 1, 1], LineBox[{1098, 1099}]}, 
      {Hue[0.35983131206568925`, 1, 1], LineBox[{1099, 1100}]}, 
      {Hue[0.3623729910400961, 1, 1], LineBox[{1100, 1101}]}, 
      {Hue[0.36745634898890983`, 1, 1], LineBox[{1101, 1102}]}, 
      {Hue[0.3699980279633167, 1, 1], LineBox[{1102, 1103}]}, 
      {Hue[0.3725397069377236, 1, 1], LineBox[{1103, 1104}]}, 
      {Hue[0.37762306488653763`, 1, 1], LineBox[{1104, 1105}]}, 
      {Hue[0.38016474386094445`, 1, 1], LineBox[{1105, 1106}]}, 
      {Hue[0.3827064228353516, 1, 1], LineBox[{1106, 1107}]}, 
      {Hue[0.38778978078416537`, 1, 1], LineBox[{1107, 1108}]}, 
      {Hue[0.39033145975857225`, 1, 1], LineBox[{1108, 1109}]}, 
      {Hue[0.39287313873297913`, 1, 1], LineBox[{1109, 1110}]}, 
      {Hue[0.39795649668179284`, 1, 1], LineBox[{1110, 1111}]}, 
      {Hue[0.4004981756561997, 1, 1], LineBox[{1111, 1112}]}, 
      {Hue[0.4030398546306066, 1, 1], LineBox[{1112, 1113}]}, 
      {Hue[0.4081232125794206, 1, 1], LineBox[{1113, 1114}]}, 
      {Hue[0.41066489155382746`, 1, 1], LineBox[{1114, 1115}]}, 
      {Hue[0.4132065705282346, 1, 1], LineBox[{1115, 1116}]}, 
      {Hue[0.4157482495026415, 1, 1], LineBox[{1116, 1117}]}, 
      {Hue[0.4182899284770484, 1, 1], LineBox[{1117, 1118}]}, 
      {Hue[0.42083160745145526`, 1, 1], LineBox[{1118, 1119}]}, 
      {Hue[0.4233732864258621, 1, 1], LineBox[{1119, 1120}]}, 
      {Hue[0.42845664437467584`, 1, 1], LineBox[{1120, 1121}]}, 
      {Hue[0.430829996776219, 1, 1], LineBox[{1121, 1122}]}, 
      {Hue[0.4332033491777619, 1, 1], LineBox[{1122, 1123}]}, 
      {Hue[0.4379500539808479, 1, 1], LineBox[{1123, 1124}]}, 
      {Hue[0.4403234063823911, 1, 1], LineBox[{1124, 1125}]}, 
      {Hue[0.4426967587839343, 1, 1], LineBox[{1125, 1126}]}, 
      {Hue[0.4474434635870203, 1, 1], LineBox[{1126, 1127}]}, 
      {Hue[0.44981681598856316`, 1, 1], LineBox[{1127, 1128}]}, 
      {Hue[0.45219016839010634`, 1, 1], LineBox[{1128, 1129}]}, 
      {Hue[0.45693687319319265`, 1, 1], LineBox[{1129, 1130}]}, 
      {Hue[0.45931022559473583`, 1, 1], LineBox[{1130, 1131}]}, 
      {Hue[0.4616835779962787, 1, 1], LineBox[{1131, 1132}]}, 
      {Hue[0.4664302827993647, 1, 1], LineBox[{1132, 1133}]}, 
      {Hue[0.4688036352009079, 1, 1], LineBox[{1133, 1134}]}, 
      {Hue[0.4711769876024508, 1, 1], LineBox[{1134, 1135}]}, 
      {Hue[0.47592369240553684`, 1, 1], LineBox[{1135, 1136}]}, 
      {Hue[0.47829704480707996`, 1, 1], LineBox[{1136, 1137}]}, 
      {Hue[0.48067039720862315`, 1, 1], LineBox[{1137, 1138}]}, 
      {Hue[0.4854171020117092, 1, 1], LineBox[{1138, 1139}]}, 
      {Hue[0.4877904544132521, 1, 1], LineBox[{1139, 1140}]}, 
      {Hue[0.4901638068147952, 1, 1], LineBox[{1140, 1141}]}, 
      {Hue[0.4949105116178816, 1, 1], LineBox[{1141, 1142}]}, 
      {Hue[0.4972838640194247, 1, 1], LineBox[{1142, 1143}]}, 
      {Hue[0.4996572164209676, 1, 1], LineBox[{1143, 1144}]}, 
      {Hue[0.5044039212240535, 1, 1], LineBox[{1144, 1145}]}, 
      {Hue[0.5067772736255967, 1, 1], LineBox[{1145, 1146}]}, 
      {Hue[0.5091506260271397, 1, 1], LineBox[{1146, 1147}]}, 
      {Hue[0.5138973308302257, 1, 1], LineBox[{1147, 1148}]}, 
      {Hue[0.5162706832317687, 1, 1], LineBox[{1148, 1149}]}, 
      {Hue[0.518644035633312, 1, 1], LineBox[{1149, 1150}]}, 
      {Hue[0.523390740436398, 1, 1], LineBox[{1150, 1151}]}, 
      {Hue[0.5257640928379409, 1, 1], LineBox[{1151, 1152}]}, 
      {Hue[0.528137445239484, 1, 1], LineBox[{1152, 1153}]}, 
      {Hue[0.5328841500425704, 1, 1], LineBox[{1153, 1154}]}, 
      {Hue[0.5352575024441135, 1, 1], LineBox[{1154, 1155}]}, 
      {Hue[0.5376308548456564, 1, 1], LineBox[{1155, 1156}]}, 
      {Hue[0.5423775596487425, 1, 1], LineBox[{1156, 1157}]}, 
      {Hue[0.5447509120502856, 1, 1], LineBox[{1157, 1158}]}, 
      {Hue[0.5471242644518288, 1, 1], LineBox[{1158, 1159}]}, 
      {Hue[0.5518709692549145, 1, 1], LineBox[{1159, 1160}]}, 
      {Hue[0.5542443216564572, 1, 1], LineBox[{1160, 1161}]}, 
      {Hue[0.5566176740580002, 1, 1], LineBox[{1161, 1162}]}, 
      {Hue[0.5613643788610867, 1, 1], LineBox[{1162, 1163}]}, 
      {Hue[0.5637377312626297, 1, 1], LineBox[{1163, 1164}]}, 
      {Hue[0.566111083664173, 1, 1], LineBox[{1164, 1165}]}, 
      {Hue[0.5708577884672592, 1, 1], LineBox[{1165, 1166}]}, 
      {Hue[0.5732311408688024, 1, 1], LineBox[{1166, 1167}]}, 
      {Hue[0.5756044932703456, 1, 1], LineBox[{1167, 1168}]}, 
      {Hue[0.5803511980734313, 1, 1], LineBox[{1168, 1169}]}, 
      {Hue[0.5850048542355217, 1, 1], LineBox[{1169, 1170}]}, 
      {Hue[0.5896585103976129, 1, 1], LineBox[{1170, 1171}]}, 
      {Hue[0.5943121665597039, 1, 1], LineBox[{1171, 1172}]}, 
      {Hue[0.5989658227217949, 1, 1], LineBox[{1172, 1173}]}, 
      {Hue[0.6036194788838858, 1, 1], LineBox[{1173, 1174}]}, 
      {Hue[0.6082731350459769, 1, 1], LineBox[{1174, 1175}]}, 
      {Hue[0.6129267912080679, 1, 1], LineBox[{1175, 1176}]}, 
      {Hue[0.6175804473701589, 1, 1], LineBox[{1176, 1177}]}, 
      {Hue[0.6222341035322498, 1, 1], LineBox[{1177, 1178}]}, 
      {Hue[0.626887759694341, 1, 1], LineBox[{1178, 1179}]}, 
      {Hue[0.6315414158564319, 1, 1], LineBox[{1179, 1180}]}, 
      {Hue[0.6361950720185223, 1, 1], LineBox[{1180, 1181}]}, 
      {Hue[0.6455023843427038, 1, 1], LineBox[{1181, 1182}]}, 
      {Hue[0.6548096966668858, 1, 1], LineBox[{1182, 1183}]}, 
      {Hue[0.6641170089910678, 1, 1], LineBox[{1183, 1184}]}, 
      {Hue[0.6734243213152493, 1, 1], LineBox[{1184, 1185}]}, 
      {Hue[0.6757511493962948, 1, 1], LineBox[{1185, 1186}]}, 
      {Hue[0.6780779774773398, 1, 1], LineBox[{1186, 1187}]}, 
      {Hue[0.6827316336394308, 1, 1], LineBox[{1187, 1188}]}, 
      {Hue[0.6920389459636127, 1, 1], LineBox[{1188, 1189}]}, 
      {Hue[0.6943657740446583, 1, 1], LineBox[{1189, 1190}]}, 
      {Hue[0.6966926021257038, 1, 1], LineBox[{1190, 1191}]}, 
      {Hue[0.7013462582877947, 1, 1], LineBox[{1191, 1192}]}, 
      {Hue[0.7106535706119763, 1, 1], LineBox[{1192, 1193}]}, 
      {Hue[0.7292681952603397, 1, 1], LineBox[{1193, 1194}]}, 
      {Hue[0.7317921035256203, 1, 1], LineBox[{1194, 1195}]}, 
      {Hue[0.7343160117909008, 1, 1], LineBox[{1195, 1196}]}, 
      {Hue[0.7393638283214612, 1, 1], LineBox[{1196, 1197}]}, 
      {Hue[0.7494594613825827, 1, 1], LineBox[{1197, 1198}]}, 
      {Hue[0.7696507275048251, 1, 1], LineBox[{1198, 1199}]}, 
      {Hue[0.8100332597493104, 1, 1], LineBox[{1199, 1200}]}, 
      {Hue[0.8907983242382805, 1, 1], LineBox[{1200, 1201}]}, 
      {Hue[0.04155555255293395, 1, 1], LineBox[{1201, 1202}]}, 
      {Hue[0.20492591265861984`, 1, 1], LineBox[{1202, 1203}]}, 
      {Hue[0.36531871625245843`, 1, 1], LineBox[{1203, 1204}]}, 
      {Hue[0.5149386191830096, 1, 1], LineBox[{1204, 1205}]}, 
      {Hue[0.6771716539045933, 1, 1], LineBox[{1205, 1206}]}, 
      {Hue[0.8286317879628902, 1, 1], LineBox[{1206, 1207}]}, 
      {Hue[0.9771143655093397, 1, 1], LineBox[{1207, 1208}]}, 
      {Hue[0.1382100748468212, 1, 1], LineBox[{1208, 1209}]}, 
      {Hue[0.28853288352101586`, 1, 1], LineBox[{1209, 1210}]}, 
      {Hue[0.45146882398624244`, 1, 1], LineBox[{1210, 1211}]}, 
      {Hue[0.45154490877811665`, 1, 1], LineBox[{1211, 1212}]}, 
      {Hue[0.4516209935699909, 1, 1], LineBox[{1212, 1213}]}, 
      {Hue[0.45177316315373883`, 1, 1], LineBox[{1213, 1214}]}, 
      {Hue[0.4520775023212352, 1, 1], LineBox[{1214, 1215}]}, 
      {Hue[0.4526861806562286, 1, 1], LineBox[{1215, 1216}]}, 
      {Hue[0.4539035373262148, 1, 1], LineBox[{1216, 1217}]}, 
      {Hue[0.45397962211808907`, 1, 1], LineBox[{1217, 1218}]}, 
      {Hue[0.45405570690996333`, 1, 1], LineBox[{1218, 1219}]}, 
      {Hue[0.45420787649371125`, 1, 1], LineBox[{1219, 1220}]}, 
      {Hue[0.45451221566120764`, 1, 1], LineBox[{1220, 1221}]}, 
      {Hue[0.4551208939962005, 1, 1], LineBox[{1221, 1222}]}, 
      {Hue[0.4551969787880747, 1, 1], LineBox[{1222, 1223}]}, 
      {Hue[0.45527306357994896`, 1, 1], LineBox[{1223, 1224}]}, 
      {Hue[0.4554252331636969, 1, 1], LineBox[{1224, 1225}]}, 
      {Hue[0.45572957233119327`, 1, 1], LineBox[{1225, 1226}]}, 
      {Hue[0.45580565712306753`, 1, 1], LineBox[{1226, 1227}]}, 
      {Hue[0.45588174191494174`, 1, 1], LineBox[{1227, 1228}]}, 
      {Hue[0.45603391149869027`, 1, 1], LineBox[{1228, 1229}]}, 
      {Hue[0.4561099962905645, 1, 1], LineBox[{1229, 1230}]}, 
      {Hue[0.45618608108243874`, 1, 1], LineBox[{1230, 1231}]}, 
      {Hue[0.456262165874313, 1, 1], LineBox[{1231, 1232}]}}}},
   VertexColors->{
     Hue[0.816901294068976, 1, 1], 
     Hue[0.8192442904753057, 1, 1], 
     Hue[0.8215872868816358, 1, 1], 
     Hue[0.8262732796942951, 1, 1], 
     Hue[0.8356452653196136, 1, 1], 
     Hue[0.8543892365702518, 1, 1], 
     Hue[0.8918771790715269, 1, 1], 
     Hue[0.9668530640740771, 1, 1], 
     Hue[0.12941796587021068`, 1, 1], 
     Hue[0.2812099670030569, 1, 1], 
     Hue[0.28353519270025984`, 1, 1], 
     Hue[0.28586041839746285`, 1, 1], 
     Hue[0.29051086979186935`, 1, 1], 
     Hue[0.29981177258068187`, 1, 1], 
     Hue[0.31841357815830684`, 1, 1], 
     Hue[0.3207388038555099, 1, 1], 
     Hue[0.32306402955271285`, 1, 1], 
     Hue[0.3277144809471194, 1, 1], 
     Hue[0.33701538373593193`, 1, 1], 
     Hue[0.34631628652474383`, 1, 1], 
     Hue[0.35561718931355635`, 1, 1], 
     Hue[0.35794241501075935`, 1, 1], 
     Hue[0.3602676407079629, 1, 1], 
     Hue[0.36491809210236886`, 1, 1], 
     Hue[0.3695685434967748, 1, 1], 
     Hue[0.37421899489118077`, 1, 1], 
     Hue[0.37886944628558666`, 1, 1], 
     Hue[0.3835198976799933, 1, 1], 
     Hue[0.38817034907439985`, 1, 1], 
     Hue[0.3928208004688058, 1, 1], 
     Hue[0.39747125186321175`, 1, 1], 
     Hue[0.4021217032576183, 1, 1], 
     Hue[0.4067721546520248, 1, 1], 
     Hue[0.41142260604643077`, 1, 1], 
     Hue[0.4160730574408368, 1, 1], 
     Hue[0.4207235088352433, 1, 1], 
     Hue[0.4230487345324463, 1, 1], 
     Hue[0.42537396022964985`, 1, 1], 
     Hue[0.4300244116240558, 1, 1], 
     Hue[0.4325467175054938, 1, 1], 
     Hue[0.43506902338693176`, 1, 1], 
     Hue[0.43759132926836974`, 1, 1], 
     Hue[0.44011363514980767`, 1, 1], 
     Hue[0.44263594103124576`, 1, 1], 
     Hue[0.44515824691268374`, 1, 1], 
     Hue[0.4502028586755597, 1, 1], 
     Hue[0.45272516455699774`, 1, 1], 
     Hue[0.4552474704384357, 1, 1], 
     Hue[0.4577697763198737, 1, 1], 
     Hue[0.4602920822013117, 1, 1], 
     Hue[0.46281438808274966`, 1, 1], 
     Hue[0.46533669396418764`, 1, 1], 
     Hue[0.47038130572706366`, 1, 1], 
     Hue[0.47290361160850164`, 1, 1], 
     Hue[0.4754259174899396, 1, 1], 
     Hue[0.4779482233713776, 1, 1], 
     Hue[0.4804705292528156, 1, 1], 
     Hue[0.4829928351342536, 1, 1], 
     Hue[0.4855151410156916, 1, 1], 
     Hue[0.49055975277856756`, 1, 1], 
     Hue[0.49308205866000554`, 1, 1], 
     Hue[0.4956043645414435, 1, 1], 
     Hue[0.49812667042288156`, 1, 1], 
     Hue[0.5006489763043194, 1, 1], 
     Hue[0.5031712821857575, 1, 1], 
     Hue[0.5056935880671953, 1, 1], 
     Hue[0.5107381998300711, 1, 1], 
     Hue[0.5132605057115088, 1, 1], 
     Hue[0.5157828115929468, 1, 1], 
     Hue[0.5183051174743848, 1, 1], 
     Hue[0.5208274233558228, 1, 1], 
     Hue[0.5233497292372608, 1, 1], 
     Hue[0.5258720351186987, 1, 1], 
     Hue[0.5309166468815747, 1, 1], 
     Hue[0.5334389527630128, 1, 1], 
     Hue[0.5359612586444507, 1, 1], 
     Hue[0.5384835645258887, 1, 1], 
     Hue[0.5410058704073267, 1, 1], 
     Hue[0.5435281762887647, 1, 1], 
     Hue[0.5460504821702027, 1, 1], 
     Hue[0.5510950939330787, 1, 1], 
     Hue[0.5536173998145166, 1, 1], 
     Hue[0.5561397056959547, 1, 1], 
     Hue[0.5586620115773926, 1, 1], 
     Hue[0.5611843174588307, 1, 1], 
     Hue[0.5637066233402686, 1, 1], 
     Hue[0.5662289292217066, 1, 1], 
     Hue[0.5712735409845826, 1, 1], 
     Hue[0.5737958468660206, 1, 1], 
     Hue[0.5763181527474586, 1, 1], 
     Hue[0.5788404586288965, 1, 1], 
     Hue[0.5813627645103345, 1, 1], 
     Hue[0.5838850703917726, 1, 1], 
     Hue[0.5864073762732105, 1, 1], 
     Hue[0.5914519880360865, 1, 1], 
     Hue[0.5938059673446605, 1, 1], 
     Hue[0.5961599466532348, 1, 1], 
     Hue[0.6008679052703831, 1, 1], 
     Hue[0.6032218845789571, 1, 1], 
     Hue[0.6055758638875314, 1, 1], 
     Hue[0.6102838225046796, 1, 1], 
     Hue[0.6126378018132537, 1, 1], 
     Hue[0.614991781121828, 1, 1], 
     Hue[0.6196997397389763, 1, 1], 
     Hue[0.6220537190475501, 1, 1], 
     Hue[0.6244076983561242, 1, 1], 
     Hue[0.6291156569732725, 1, 1], 
     Hue[0.6314696362818468, 1, 1], 
     Hue[0.6338236155904208, 1, 1], 
     Hue[0.6385315742075688, 1, 1], 
     Hue[0.6408855535161428, 1, 1], 
     Hue[0.6432395328247171, 1, 1], 
     Hue[0.6479474914418654, 1, 1], 
     Hue[0.6503014707504393, 1, 1], 
     Hue[0.6526554500590136, 1, 1], 
     Hue[0.6573634086761619, 1, 1], 
     Hue[0.659717387984736, 1, 1], 
     Hue[0.6620713672933103, 1, 1], 
     Hue[0.6667793259104585, 1, 1], 
     Hue[0.6691333052190325, 1, 1], 
     Hue[0.6714872845276068, 1, 1], 
     Hue[0.6761952431447551, 1, 1], 
     Hue[0.678549222453329, 1, 1], 
     Hue[0.6809032017619033, 1, 1], 
     Hue[0.6856111603790516, 1, 1], 
     Hue[0.6879651396876256, 1, 1], 
     Hue[0.6903191189961999, 1, 1], 
     Hue[0.6950270776133483, 1, 1], 
     Hue[0.6973810569219222, 1, 1], 
     Hue[0.6997350362304962, 1, 1], 
     Hue[0.7044429948476445, 1, 1], 
     Hue[0.7067969741562188, 1, 1], 
     Hue[0.7091509534647928, 1, 1], 
     Hue[0.7138589120819407, 1, 1], 
     Hue[0.7162128913905148, 1, 1], 
     Hue[0.7185668706990891, 1, 1], 
     Hue[0.7232748293162373, 1, 1], 
     Hue[0.7256288086248113, 1, 1], 
     Hue[0.7279827879333856, 1, 1], 
     Hue[0.7326907465505339, 1, 1], 
     Hue[0.735044725859108, 1, 1], 
     Hue[0.7373987051676822, 1, 1], 
     Hue[0.7421066637848305, 1, 1], 
     Hue[0.7446577232776395, 1, 1], 
     Hue[0.7472087827704486, 1, 1], 
     Hue[0.7523109017560665, 1, 1], 
     Hue[0.7548619612488755, 1, 1], 
     Hue[0.7574130207416845, 1, 1], 
     Hue[0.7625151397273026, 1, 1], 
     Hue[0.7650661992201115, 1, 1], 
     Hue[0.7676172587129205, 1, 1], 
     Hue[0.7727193776985386, 1, 1], 
     Hue[0.7752704371913476, 1, 1], 
     Hue[0.7778214966841566, 1, 1], 
     Hue[0.7829236156697745, 1, 1], 
     Hue[0.7854746751625836, 1, 1], 
     Hue[0.7880257346553926, 1, 1], 
     Hue[0.7931278536410106, 1, 1], 
     Hue[0.7956789131338196, 1, 1], 
     Hue[0.7982299726266286, 1, 1], 
     Hue[0.8033320916122466, 1, 1], 
     Hue[0.8058831511050557, 1, 1], 
     Hue[0.8084342105978647, 1, 1], 
     Hue[0.8135363295834827, 1, 1], 
     Hue[0.8160873890762916, 1, 1], 
     Hue[0.8186384485691007, 1, 1], 
     Hue[0.8237405675547187, 1, 1], 
     Hue[0.8262916270475277, 1, 1], 
     Hue[0.8288426865403367, 1, 1], 
     Hue[0.8339448055259547, 1, 1], 
     Hue[0.8364958650187637, 1, 1], 
     Hue[0.8390469245115727, 1, 1], 
     Hue[0.8441490434971908, 1, 1], 
     Hue[0.8467001029899998, 1, 1], 
     Hue[0.8492511624828087, 1, 1], 
     Hue[0.8543532814684268, 1, 1], 
     Hue[0.8569043409612358, 1, 1], 
     Hue[0.8594554004540448, 1, 1], 
     Hue[0.8645575194396629, 1, 1], 
     Hue[0.8671085789324718, 1, 1], 
     Hue[0.8696596384252808, 1, 1], 
     Hue[0.8747617574108988, 1, 1], 
     Hue[0.8773128169037079, 1, 1], 
     Hue[0.8798638763965169, 1, 1], 
     Hue[0.8849659953821348, 1, 1], 
     Hue[0.8875170548749438, 1, 1], 
     Hue[0.8900681143677529, 1, 1], 
     Hue[0.8951702333533709, 1, 1], 
     Hue[0.89772129284618, 1, 1], 
     Hue[0.9002723523389888, 1, 1], 
     Hue[0.9053744713246066, 1, 1], 
     Hue[0.9078790064969182, 1, 1], 
     Hue[0.9103835416692296, 1, 1], 
     Hue[0.9153926120138524, 1, 1], 
     Hue[0.9178971471861637, 1, 1], 
     Hue[0.920401682358475, 1, 1], 
     Hue[0.9254107527030977, 1, 1], 
     Hue[0.9279152878754091, 1, 1], 
     Hue[0.9304198230477204, 1, 1], 
     Hue[0.9354288933923434, 1, 1], 
     Hue[0.937933428564655, 1, 1], 
     Hue[0.9404379637369664, 1, 1], 
     Hue[0.945447034081589, 1, 1], 
     Hue[0.9479515692539005, 1, 1], 
     Hue[0.9504561044262118, 1, 1], 
     Hue[0.9554651747708345, 1, 1], 
     Hue[0.9579697099431458, 1, 1], 
     Hue[0.9604742451154572, 1, 1], 
     Hue[0.9654833154600801, 1, 1], 
     Hue[0.9679878506323918, 1, 1], 
     Hue[0.9704923858047031, 1, 1], 
     Hue[0.9755014561493258, 1, 1], 
     Hue[0.9780059913216371, 1, 1], 
     Hue[0.9805105264939485, 1, 1], 
     Hue[0.9855195968385713, 1, 1], 
     Hue[0.9880241320108826, 1, 1], 
     Hue[0.9905286671831939, 1, 1], 
     Hue[0.9955377375278166, 1, 1], 
     Hue[0.998042272700128, 1, 1], 
     Hue[0.0005468078724393148, 1, 1], 
     Hue[0.005555878217062295, 1, 1], 
     Hue[0.008060413389373928, 1, 1], 
     Hue[0.010564948561685276`, 1, 1], 
     Hue[0.015574018906307979`, 1, 1], 
     Hue[0.018078554078619323`, 1, 1], 
     Hue[0.02058308925093067, 1, 1], 
     Hue[0.025592159595553377`, 1, 1], 
     Hue[0.02809669476786473, 1, 1], 
     Hue[0.030601229940176073`, 1, 1], 
     Hue[0.03561030028479876, 1, 1], 
     Hue[0.03811483545711011, 1, 1], 
     Hue[0.04061937062942147, 1, 1], 
     Hue[0.045628440974044455`, 1, 1], 
     Hue[0.048132976146356074`, 1, 1], 
     Hue[0.05063751131866743, 1, 1], 
     Hue[0.05564658166329014, 1, 1], 
     Hue[0.05815111683560147, 1, 1], 
     Hue[0.06065565200791283, 1, 1], 
     Hue[0.06566472235253552, 1, 1], 
     Hue[0.06800093095198288, 1, 1], 
     Hue[0.0703371395514305, 1, 1], 
     Hue[0.07500955675032554, 1, 1], 
     Hue[0.07734576534977317, 1, 1], 
     Hue[0.0796819739492208, 1, 1], 
     Hue[0.08435439114811583, 1, 1], 
     Hue[0.08669059974756345, 1, 1], 
     Hue[0.08902680834701081, 1, 1], 
     Hue[0.09369922554590611, 1, 1], 
     Hue[0.1030440599436961, 1, 1], 
     Hue[0.10538026854314346`, 1, 1], 
     Hue[0.1077164771425911, 1, 1], 
     Hue[0.1123888943414861, 1, 1], 
     Hue[0.11472510294093347`, 1, 1], 
     Hue[0.11706131154038113`, 1, 1], 
     Hue[0.12173372873927613`, 1, 1], 
     Hue[0.12406993733872347`, 1, 1], 
     Hue[0.12640614593817112`, 1, 1], 
     Hue[0.13107856313706612`, 1, 1], 
     Hue[0.14042339753485641`, 1, 1], 
     Hue[0.14275960613430405`, 1, 1], 
     Hue[0.1450958147337514, 1, 1], 
     Hue[0.14976823193264668`, 1, 1], 
     Hue[0.15210444053209407`, 1, 1], 
     Hue[0.1544406491315417, 1, 1], 
     Hue[0.1591130663304367, 1, 1], 
     Hue[0.16144927492988406`, 1, 1], 
     Hue[0.1637854835293317, 1, 1], 
     Hue[0.1684579007282267, 1, 1], 
     Hue[0.1778027351260167, 1, 1], 
     Hue[0.18013894372546407`, 1, 1], 
     Hue[0.1824751523249117, 1, 1], 
     Hue[0.1871475695238067, 1, 1], 
     Hue[0.18948377812325434`, 1, 1], 
     Hue[0.19181998672270198`, 1, 1], 
     Hue[0.196492403921597, 1, 1], 
     Hue[0.19882861252104464`, 1, 1], 
     Hue[0.201164821120492, 1, 1], 
     Hue[0.20583723831938727`, 1, 1], 
     Hue[0.2151820727171773, 1, 1], 
     Hue[0.21771536150085966`, 1, 1], 
     Hue[0.22024865028454202`, 1, 1], 
     Hue[0.22531522785190675`, 1, 1], 
     Hue[0.2278485166355891, 1, 1], 
     Hue[0.23038180541927147`, 1, 1], 
     Hue[0.23544838298663648`, 1, 1], 
     Hue[0.23798167177031915`, 1, 1], 
     Hue[0.2405149605540015, 1, 1], 
     Hue[0.24558153812136624`, 1, 1], 
     Hue[0.2557146932560957, 1, 1], 
     Hue[0.25824798203977806`, 1, 1], 
     Hue[0.2607812708234604, 1, 1], 
     Hue[0.26584784839082515`, 1, 1], 
     Hue[0.2683811371745075, 1, 1], 
     Hue[0.2709144259581899, 1, 1], 
     Hue[0.27598100352555494`, 1, 1], 
     Hue[0.2785142923092376, 1, 1], 
     Hue[0.28104758109291994`, 1, 1], 
     Hue[0.28611415866028467`, 1, 1], 
     Hue[0.2962473137950141, 1, 1], 
     Hue[0.2987806025786965, 1, 1], 
     Hue[0.30131389136237885`, 1, 1], 
     Hue[0.3063804689297436, 1, 1], 
     Hue[0.30891375771342594`, 1, 1], 
     Hue[0.3114470464971083, 1, 1], 
     Hue[0.3165136240644733, 1, 1], 
     Hue[0.31904691284815595`, 1, 1], 
     Hue[0.32158020163183837`, 1, 1], 
     Hue[0.3266467791992031, 1, 1], 
     Hue[0.33677993433393255`, 1, 1], 
     Hue[0.3393132231176149, 1, 1], 
     Hue[0.3418465119012973, 1, 1], 
     Hue[0.3469130894686623, 1, 1], 
     Hue[0.3494463782523449, 1, 1], 
     Hue[0.3519796670360273, 1, 1], 
     Hue[0.357046244603392, 1, 1], 
     Hue[0.3595795333870744, 1, 1], 
     Hue[0.36211282217075674`, 1, 1], 
     Hue[0.36717939973812147`, 1, 1], 
     Hue[0.37731255487285126`, 1, 1], 
     Hue[0.3796775170836699, 1, 1], 
     Hue[0.38204247929448826`, 1, 1], 
     Hue[0.3867724037161256, 1, 1], 
     Hue[0.389137365926944, 1, 1], 
     Hue[0.39150232813776265`, 1, 1], 
     Hue[0.39623225255939964`, 1, 1], 
     Hue[0.39859721477021803`, 1, 1], 
     Hue[0.4009621769810367, 1, 1], 
     Hue[0.40569210140267375`, 1, 1], 
     Hue[0.4151519502459478, 1, 1], 
     Hue[0.4175169124567662, 1, 1], 
     Hue[0.41988187466758486`, 1, 1], 
     Hue[0.42461179908922186`, 1, 1], 
     Hue[0.42697676130004053`, 1, 1], 
     Hue[0.4293417235108592, 1, 1], 
     Hue[0.43407164793249625`, 1, 1], 
     Hue[0.43643661014331486`, 1, 1], 
     Hue[0.43880157235413325`, 1, 1], 
     Hue[0.4435314967757706, 1, 1], 
     Hue[0.45299134561904464`, 1, 1], 
     Hue[0.455356307829863, 1, 1], 
     Hue[0.4577212700406817, 1, 1], 
     Hue[0.46245119446231875`, 1, 1], 
     Hue[0.46481615667313736`, 1, 1], 
     Hue[0.467181118883956, 1, 1], 
     Hue[0.4719110433055931, 1, 1], 
     Hue[0.47427600551641175`, 1, 1], 
     Hue[0.47664096772723014`, 1, 1], 
     Hue[0.4813708921488674, 1, 1], 
     Hue[0.4908307409921415, 1, 1], 
     Hue[0.49319570320295986`, 1, 1], 
     Hue[0.4955606654137785, 1, 1], 
     Hue[0.5002905898354155, 1, 1], 
     Hue[0.5026555520462341, 1, 1], 
     Hue[0.5050205142570529, 1, 1], 
     Hue[0.50975043867869, 1, 1], 
     Hue[0.5121154008895086, 1, 1], 
     Hue[0.514480363100327, 1, 1], 
     Hue[0.5192102875219642, 1, 1], 
     Hue[0.5286701363652384, 1, 1], 
     Hue[0.5309885742555593, 1, 1], 
     Hue[0.5333070121458803, 1, 1], 
     Hue[0.537943887926522, 1, 1], 
     Hue[0.540262325816843, 1, 1], 
     Hue[0.5425807637071637, 1, 1], 
     Hue[0.5472176394878057, 1, 1], 
     Hue[0.5495360773781268, 1, 1], 
     Hue[0.5518545152684478, 1, 1], 
     Hue[0.5564913910490895, 1, 1], 
     Hue[0.5657651426103731, 1, 1], 
     Hue[0.5680835805006943, 1, 1], 
     Hue[0.5704020183910152, 1, 1], 
     Hue[0.5750388941716569, 1, 1], 
     Hue[0.5773573320619779, 1, 1], 
     Hue[0.5796757699522987, 1, 1], 
     Hue[0.5843126457329406, 1, 1], 
     Hue[0.5866310836232617, 1, 1], 
     Hue[0.5889495215135827, 1, 1], 
     Hue[0.5935863972942244, 1, 1], 
     Hue[0.6028601488555081, 1, 1], 
     Hue[0.6051785867458291, 1, 1], 
     Hue[0.6074970246361502, 1, 1], 
     Hue[0.6121339004167918, 1, 1], 
     Hue[0.6144523383071128, 1, 1], 
     Hue[0.6167707761974335, 1, 1], 
     Hue[0.6214076519780756, 1, 1], 
     Hue[0.6237260898683965, 1, 1], 
     Hue[0.6260445277587176, 1, 1], 
     Hue[0.6306814035393593, 1, 1], 
     Hue[0.639955155100643, 1, 1], 
     Hue[0.642273592990964, 1, 1], 
     Hue[0.644592030881285, 1, 1], 
     Hue[0.6492289066619267, 1, 1], 
     Hue[0.6585026582232103, 1, 1], 
     Hue[0.6608210961135315, 1, 1], 
     Hue[0.6631395340038524, 1, 1], 
     Hue[0.6677764097844944, 1, 1], 
     Hue[0.6770501613457782, 1, 1], 
     Hue[0.6795656794203342, 1, 1], 
     Hue[0.6820811974948899, 1, 1], 
     Hue[0.6871122336440013, 1, 1], 
     Hue[0.6896277517185571, 1, 1], 
     Hue[0.6921432697931131, 1, 1], 
     Hue[0.6971743059422248, 1, 1], 
     Hue[0.6996898240167806, 1, 1], 
     Hue[0.7022053420913362, 1, 1], 
     Hue[0.707236378240448, 1, 1], 
     Hue[0.7172984505386711, 1, 1], 
     Hue[0.7198139686132269, 1, 1], 
     Hue[0.7223294866877826, 1, 1], 
     Hue[0.7273605228368943, 1, 1], 
     Hue[0.7298760409114502, 1, 1], 
     Hue[0.732391558986006, 1, 1], 
     Hue[0.7374225951351177, 1, 1], 
     Hue[0.7399381132096735, 1, 1], 
     Hue[0.7424536312842293, 1, 1], 
     Hue[0.747484667433341, 1, 1], 
     Hue[0.7575467397315642, 1, 1], 
     Hue[0.7600622578061199, 1, 1], 
     Hue[0.7625777758806758, 1, 1], 
     Hue[0.7676088120297874, 1, 1], 
     Hue[0.7776708843280107, 1, 1], 
     Hue[0.7801864024025665, 1, 1], 
     Hue[0.7827019204771223, 1, 1], 
     Hue[0.787732956626234, 1, 1], 
     Hue[0.7977950289244572, 1, 1], 
     Hue[0.800310546999013, 1, 1], 
     Hue[0.8028260650735689, 1, 1], 
     Hue[0.8078571012226805, 1, 1], 
     Hue[0.8179191735209038, 1, 1], 
     Hue[0.8204346915954596, 1, 1], 
     Hue[0.8229502096700154, 1, 1], 
     Hue[0.827981245819127, 1, 1], 
     Hue[0.8380433181173503, 1, 1], 
     Hue[0.8403905096190423, 1, 1], 
     Hue[0.8427377011207342, 1, 1], 
     Hue[0.8474320841241183, 1, 1], 
     Hue[0.856820850130886, 1, 1], 
     Hue[0.8591680416325781, 1, 1], 
     Hue[0.8615152331342699, 1, 1], 
     Hue[0.8662096161376538, 1, 1], 
     Hue[0.8755983821444216, 1, 1], 
     Hue[0.8779455736461137, 1, 1], 
     Hue[0.8802927651478055, 1, 1], 
     Hue[0.8849871481511894, 1, 1], 
     Hue[0.8943759141579573, 1, 1], 
     Hue[0.8967231056596492, 1, 1], 
     Hue[0.8990702971613411, 1, 1], 
     Hue[0.903764680164725, 1, 1], 
     Hue[0.9131534461714927, 1, 1], 
     Hue[0.9155006376731848, 1, 1], 
     Hue[0.9178478291748767, 1, 1], 
     Hue[0.9225422121782606, 1, 1], 
     Hue[0.9319309781850283, 1, 1], 
     Hue[0.9342781696867204, 1, 1], 
     Hue[0.9366253611884122, 1, 1], 
     Hue[0.9413197441917962, 1, 1], 
     Hue[0.9507085101985641, 1, 1], 
     Hue[0.953055701700256, 1, 1], 
     Hue[0.9554028932019479, 1, 1], 
     Hue[0.960097276205332, 1, 1], 
     Hue[0.9694860422120998, 1, 1], 
     Hue[0.9882635742256354, 1, 1], 
     Hue[0.9908078459115621, 1, 1], 
     Hue[0.993352117597489, 1, 1], 
     Hue[0.9984406609693426, 1, 1], 
     Hue[0.00861774771305003, 1, 1], 
     Hue[0.011162019398976914`, 1, 1], 
     Hue[0.013706291084903658`, 1, 1], 
     Hue[0.018794834456757423`, 1, 1], 
     Hue[0.02897192120046467, 1, 1], 
     Hue[0.03151619288639141, 1, 1], 
     Hue[0.034060464572318296`, 1, 1], 
     Hue[0.03914900794417192, 1, 1], 
     Hue[0.04932609468787931, 1, 1], 
     Hue[0.06968026817529394, 1, 1], 
     Hue[0.07222453986122068, 1, 1], 
     Hue[0.07476881154714757, 1, 1], 
     Hue[0.07985735491900121, 1, 1], 
     Hue[0.09003444166270858, 1, 1], 
     Hue[0.09257871334863547, 1, 1], 
     Hue[0.09512298503456218, 1, 1], 
     Hue[0.10021152840641596`, 1, 1], 
     Hue[0.11038861515012321`, 1, 1], 
     Hue[0.1129328868360501, 1, 1], 
     Hue[0.115477158521977, 1, 1], 
     Hue[0.1205657018938306, 1, 1], 
     Hue[0.130742788637538, 1, 1], 
     Hue[0.15109696212495263`, 1, 1], 
     Hue[0.15359470949038187`, 1, 1], 
     Hue[0.15609245685581094`, 1, 1], 
     Hue[0.1610879515866694, 1, 1], 
     Hue[0.1710789410483863, 1, 1], 
     Hue[0.17357668841381552`, 1, 1], 
     Hue[0.17607443577924475`, 1, 1], 
     Hue[0.1810699305101032, 1, 1], 
     Hue[0.1910609199718201, 1, 1], 
     Hue[0.19355866733724933`, 1, 1], 
     Hue[0.19605641470267854`, 1, 1], 
     Hue[0.20105190943353698`, 1, 1], 
     Hue[0.2110428988952539, 1, 1], 
     Hue[0.2310248778186877, 1, 1], 
     Hue[0.23352262518411693`, 1, 1], 
     Hue[0.23602037254954614`, 1, 1], 
     Hue[0.2410158672804046, 1, 1], 
     Hue[0.2510068567421215, 1, 1], 
     Hue[0.27098883566555515`, 1, 1], 
     Hue[0.2734865830309844, 1, 1], 
     Hue[0.27598433039641346`, 1, 1], 
     Hue[0.28097982512727193`, 1, 1], 
     Hue[0.2909708145889888, 1, 1], 
     Hue[0.3109527935124226, 1, 1], 
     Hue[0.313282214304988, 1, 1], 
     Hue[0.31561163509755336`, 1, 1], 
     Hue[0.32027047668268416`, 1, 1], 
     Hue[0.3295881598529455, 1, 1], 
     Hue[0.3482235261934684, 1, 1], 
     Hue[0.3505529469860338, 1, 1], 
     Hue[0.35288236777859916`, 1, 1], 
     Hue[0.3575412093637299, 1, 1], 
     Hue[0.3668588925339913, 1, 1], 
     Hue[0.38549425887451416`, 1, 1], 
     Hue[0.3878236796670795, 1, 1], 
     Hue[0.39015310045964485`, 1, 1], 
     Hue[0.3948119420447756, 1, 1], 
     Hue[0.40412962521503704`, 1, 1], 
     Hue[0.4227649915555599, 1, 1], 
     Hue[0.46003572423660566`, 1, 1], 
     Hue[0.4625622252134059, 1, 1], 
     Hue[0.46508872619020614`, 1, 1], 
     Hue[0.4701417281438066, 1, 1], 
     Hue[0.48024773205100757`, 1, 1], 
     Hue[0.5004597398654094, 1, 1], 
     Hue[0.5408837554942132, 1, 1], 
     Hue[0.5434102564710135, 1, 1], 
     Hue[0.5459367574478137, 1, 1], 
     Hue[0.5509897594014141, 1, 1], 
     Hue[0.5610957633086151, 1, 1], 
     Hue[0.581307771123017, 1, 1], 
     Hue[0.6217317867518208, 1, 1], 
     Hue[0.6240899611557571, 1, 1], 
     Hue[0.6264481355596936, 1, 1], 
     Hue[0.6311644843675663, 1, 1], 
     Hue[0.6405971819833118, 1, 1], 
     Hue[0.6594625772148028, 1, 1], 
     Hue[0.6971933676777848, 1, 1], 
     Hue[0.6995515420817213, 1, 1], 
     Hue[0.7019097164856576, 1, 1], 
     Hue[0.7066260652935304, 1, 1], 
     Hue[0.716058762909276, 1, 1], 
     Hue[0.734924158140767, 1, 1], 
     Hue[0.772654948603749, 1, 1], 
     Hue[0.7752102031919202, 1, 1], 
     Hue[0.7777654577800914, 1, 1], 
     Hue[0.782875966956434, 1, 1], 
     Hue[0.793096985309119, 1, 1], 
     Hue[0.8135390220144891, 1, 1], 
     Hue[0.8160942766026603, 1, 1], 
     Hue[0.8186495311908315, 1, 1], 
     Hue[0.8237600403671741, 1, 1], 
     Hue[0.8339810587198591, 1, 1], 
     Hue[0.8544230954252292, 1, 1], 
     Hue[0.8569783500134004, 1, 1], 
     Hue[0.8595336046015717, 1, 1], 
     Hue[0.8646441137779142, 1, 1], 
     Hue[0.8748651321305992, 1, 1], 
     Hue[0.8774203867187704, 1, 1], 
     Hue[0.8799756413069417, 1, 1], 
     Hue[0.8850861504832842, 1, 1], 
     Hue[0.8953071688359693, 1, 1], 
     Hue[0.8978624234241405, 1, 1], 
     Hue[0.9004176780123118, 1, 1], 
     Hue[0.9055281871886542, 1, 1], 
     Hue[0.9157492055413393, 1, 1], 
     Hue[0.9183044601295106, 1, 1], 
     Hue[0.9208597147176818, 1, 1], 
     Hue[0.9259702238940244, 1, 1], 
     Hue[0.9361912422467094, 1, 1], 
     Hue[0.938699972514383, 1, 1], 
     Hue[0.9412087027820567, 1, 1], 
     Hue[0.946226163317404, 1, 1], 
     Hue[0.9487348935850776, 1, 1], 
     Hue[0.9512436238527512, 1, 1], 
     Hue[0.9562610843880985, 1, 1], 
     Hue[0.9587698146557722, 1, 1], 
     Hue[0.9612785449234458, 1, 1], 
     Hue[0.9637872751911194, 1, 1], 
     Hue[0.9662960054587931, 1, 1], 
     Hue[0.9688047357264667, 1, 1], 
     Hue[0.9713134659941404, 1, 1], 
     Hue[0.973822196261814, 1, 1], 
     Hue[0.9763309265294876, 1, 1], 
     Hue[0.9788396567971613, 1, 1], 
     Hue[0.981348387064835, 1, 1], 
     Hue[0.9838571173325086, 1, 1], 
     Hue[0.9863658476001822, 1, 1], 
     Hue[0.9888745778678558, 1, 1], 
     Hue[0.9913833081355296, 1, 1], 
     Hue[0.9938920384032032, 1, 1], 
     Hue[0.9964007686708768, 1, 1], 
     Hue[0.9989094989385504, 1, 1], 
     Hue[0.0014182292062240736`, 1, 1], 
     Hue[0.003926959473897716, 1, 1], 
     Hue[0.0064356897415713595`, 1, 1], 
     Hue[0.008944420009244999, 1, 1], 
     Hue[0.011453150276918643`, 1, 1], 
     Hue[0.013961880544592286`, 1, 1], 
     Hue[0.016470610812265928`, 1, 1], 
     Hue[0.018979341079939573`, 1, 1], 
     Hue[0.021488071347613213`, 1, 1], 
     Hue[0.023996801615286858`, 1, 1], 
     Hue[0.026505531882960495`, 1, 1], 
     Hue[0.02901426215063414, 1, 1], 
     Hue[0.031522992418307784`, 1, 1], 
     Hue[0.03654045295365507, 1, 1], 
     Hue[0.0390491832213287, 1, 1], 
     Hue[0.041557913489002354`, 1, 1], 
     Hue[0.04657537402434964, 1, 1], 
     Hue[0.04908410429202328, 1, 1], 
     Hue[0.05159283455969691, 1, 1], 
     Hue[0.05661029509504419, 1, 1], 
     Hue[0.05911902536271785, 1, 1], 
     Hue[0.061627755630391495`, 1, 1], 
     Hue[0.06664521616573878, 1, 1], 
     Hue[0.06915394643341242, 1, 1], 
     Hue[0.07166267670108606, 1, 1], 
     Hue[0.07668013723643334, 1, 1], 
     Hue[0.07918886750410699, 1, 1], 
     Hue[0.08169759777178064, 1, 1], 
     Hue[0.08671505830712793, 1, 1], 
     Hue[0.0967499793778225, 1, 1], 
     Hue[0.09909038307263227, 1, 1], 
     Hue[0.10143078676744204`, 1, 1], 
     Hue[0.10611159415706164`, 1, 1], 
     Hue[0.11547320893630074`, 1, 1], 
     Hue[0.134196438494779, 1, 1], 
     Hue[0.1365368421895888, 1, 1], 
     Hue[0.13887724588439856`, 1, 1], 
     Hue[0.14355805327401813`, 1, 1], 
     Hue[0.15291966805325727`, 1, 1], 
     Hue[0.1716428976117355, 1, 1], 
     Hue[0.1739833013065453, 1, 1], 
     Hue[0.17632370500135508`, 1, 1], 
     Hue[0.18100451239097462`, 1, 1], 
     Hue[0.19036612717021376`, 1, 1], 
     Hue[0.20908935672869203`, 1, 1], 
     Hue[0.24653581584564851`, 1, 1], 
     Hue[0.24907329972469316`, 1, 1], 
     Hue[0.25161078360373784`, 1, 1], 
     Hue[0.25668575136182714`, 1, 1], 
     Hue[0.2668356868780058, 1, 1], 
     Hue[0.2871355579103631, 1, 1], 
     Hue[0.32773529997507767`, 1, 1], 
     Hue[0.3302727838541223, 1, 1], 
     Hue[0.332810267733167, 1, 1], 
     Hue[0.3378852354912563, 1, 1], 
     Hue[0.348035171007435, 1, 1], 
     Hue[0.3683350420397923, 1, 1], 
     Hue[0.40893478410450684`, 1, 1], 
     Hue[0.41130394141068766`, 1, 1], 
     Hue[0.41367309871686847`, 1, 1], 
     Hue[0.4184114133292301, 1, 1], 
     Hue[0.4278880425539533, 1, 1], 
     Hue[0.4468413010033997, 1, 1], 
     Hue[0.4847478179022925, 1, 1], 
     Hue[0.48711697520847325`, 1, 1], 
     Hue[0.48948613251465406`, 1, 1], 
     Hue[0.4942244471270157, 1, 1], 
     Hue[0.5037010763517389, 1, 1], 
     Hue[0.5226543348011852, 1, 1], 
     Hue[0.5605608517000781, 1, 1], 
     Hue[0.5628834846857612, 1, 1], 
     Hue[0.5652061176714445, 1, 1], 
     Hue[0.5698513836428108, 1, 1], 
     Hue[0.5791419155855435, 1, 1], 
     Hue[0.597722979471009, 1, 1], 
     Hue[0.6000456124566923, 1, 1], 
     Hue[0.6023682454423754, 1, 1], 
     Hue[0.6070135114137418, 1, 1], 
     Hue[0.6163040433564745, 1, 1], 
     Hue[0.63488510724194, 1, 1], 
     Hue[0.6372077402276233, 1, 1], 
     Hue[0.6395303732133064, 1, 1], 
     Hue[0.6441756391846729, 1, 1], 
     Hue[0.6534661711274056, 1, 1], 
     Hue[0.672047235012871, 1, 1], 
     Hue[0.6743698679985541, 1, 1], 
     Hue[0.6766925009842375, 1, 1], 
     Hue[0.6813377669556038, 1, 1], 
     Hue[0.6906282988983365, 1, 1], 
     Hue[0.709209362783802, 1, 1], 
     Hue[0.71172907595372, 1, 1], 
     Hue[0.7142487891236381, 1, 1], 
     Hue[0.7192882154634741, 1, 1], 
     Hue[0.7293670681431466, 1, 1], 
     Hue[0.7318867813130645, 1, 1], 
     Hue[0.7344064944829827, 1, 1], 
     Hue[0.7394459208228188, 1, 1], 
     Hue[0.749524773502491, 1, 1], 
     Hue[0.7520444866724092, 1, 1], 
     Hue[0.7545641998423271, 1, 1], 
     Hue[0.7596036261821633, 1, 1], 
     Hue[0.7696824788618355, 1, 1], 
     Hue[0.7898401842211801, 1, 1], 
     Hue[0.7923598973910981, 1, 1], 
     Hue[0.7948796105610161, 1, 1], 
     Hue[0.7999190369008522, 1, 1], 
     Hue[0.8099978895805245, 1, 1], 
     Hue[0.8125176027504426, 1, 1], 
     Hue[0.8150373159203607, 1, 1], 
     Hue[0.8200767422601969, 1, 1], 
     Hue[0.8301555949398691, 1, 1], 
     Hue[0.8326753081097872, 1, 1], 
     Hue[0.8351950212797054, 1, 1], 
     Hue[0.8402344476195414, 1, 1], 
     Hue[0.8503133002992137, 1, 1], 
     Hue[0.8704710056585583, 1, 1], 
     Hue[0.8728223922556125, 1, 1], 
     Hue[0.8751737788526667, 1, 1], 
     Hue[0.8798765520467752, 1, 1], 
     Hue[0.8892820984349921, 1, 1], 
     Hue[0.8916334850320461, 1, 1], 
     Hue[0.8939848716291003, 1, 1], 
     Hue[0.8986876448232088, 1, 1], 
     Hue[0.9080931912114256, 1, 1], 
     Hue[0.9104445778084798, 1, 1], 
     Hue[0.912795964405534, 1, 1], 
     Hue[0.9174987375996424, 1, 1], 
     Hue[0.9269042839878592, 1, 1], 
     Hue[0.9457153767642927, 1, 1], 
     Hue[0.9480667633613469, 1, 1], 
     Hue[0.9504181499584011, 1, 1], 
     Hue[0.9551209231525096, 1, 1], 
     Hue[0.9645264695407264, 1, 1], 
     Hue[0.9668778561377805, 1, 1], 
     Hue[0.9692292427348348, 1, 1], 
     Hue[0.9739320159289432, 1, 1], 
     Hue[0.9833375623171601, 1, 1], 
     Hue[0.9856889489142143, 1, 1], 
     Hue[0.9880403355112685, 1, 1], 
     Hue[0.9927431087053771, 1, 1], 
     Hue[0.0021486550935938373`, 1, 1], 
     Hue[0.004500041690647965, 1, 1], 
     Hue[0.006851428287702235, 1, 1], 
     Hue[0.011554201481810633`, 1, 1], 
     Hue[0.020959747870027433`, 1, 1], 
     Hue[0.02350821465131656, 1, 1], 
     Hue[0.02605668143260555, 1, 1], 
     Hue[0.031153614995183685`, 1, 1], 
     Hue[0.041347482120340076`, 1, 1], 
     Hue[0.04389594890162907, 1, 1], 
     Hue[0.04644441568291821, 1, 1], 
     Hue[0.051541349245496464`, 1, 1], 
     Hue[0.06173521637065272, 1, 1], 
     Hue[0.06428368315194186, 1, 1], 
     Hue[0.066832149933231, 1, 1], 
     Hue[0.07192908349580913, 1, 1], 
     Hue[0.08212295062096552, 1, 1], 
     Hue[0.08467141740225466, 1, 1], 
     Hue[0.08721988418354365, 1, 1], 
     Hue[0.09231681774612178, 1, 1], 
     Hue[0.10251068487127818`, 1, 1], 
     Hue[0.10505915165256717`, 1, 1], 
     Hue[0.10760761843385629`, 1, 1], 
     Hue[0.11270455199643457`, 1, 1], 
     Hue[0.12289841912159083`, 1, 1], 
     Hue[0.12544688590287997`, 1, 1], 
     Hue[0.1279953526841691, 1, 1], 
     Hue[0.13309228624674724`, 1, 1], 
     Hue[0.14328615337190362`, 1, 1], 
     Hue[0.14583462015319276`, 1, 1], 
     Hue[0.14838308693448174`, 1, 1], 
     Hue[0.1534800204970599, 1, 1], 
     Hue[0.16367388762221627`, 1, 1], 
     Hue[0.16622235440350527`, 1, 1], 
     Hue[0.1687708211847944, 1, 1], 
     Hue[0.17386775474737268`, 1, 1], 
     Hue[0.18406162187252892`, 1, 1], 
     Hue[0.18656356433332041`, 1, 1], 
     Hue[0.18906550679411188`, 1, 1], 
     Hue[0.19406939171569484`, 1, 1], 
     Hue[0.20407716155886074`, 1, 1], 
     Hue[0.2065791040196522, 1, 1], 
     Hue[0.2090810464804437, 1, 1], 
     Hue[0.21408493140202664`, 1, 1], 
     Hue[0.22409270124519243`, 1, 1], 
     Hue[0.22659464370598376`, 1, 1], 
     Hue[0.22909658616677522`, 1, 1], 
     Hue[0.2341004710883582, 1, 1], 
     Hue[0.24410824093152408`, 1, 1], 
     Hue[0.24661018339231555`, 1, 1], 
     Hue[0.24911212585310705`, 1, 1], 
     Hue[0.25411601077469, 1, 1], 
     Hue[0.2641237806178559, 1, 1], 
     Hue[0.2666257230786474, 1, 1], 
     Hue[0.26912766553943873`, 1, 1], 
     Hue[0.2741315504610215, 1, 1], 
     Hue[0.276633492921813, 1, 1], 
     Hue[0.2791354353826045, 1, 1], 
     Hue[0.2841393203041874, 1, 1], 
     Hue[0.2866412627649789, 1, 1], 
     Hue[0.28914320522577036`, 1, 1], 
     Hue[0.29414709014735335`, 1, 1], 
     Hue[0.3041548599905192, 1, 1], 
     Hue[0.3066568024513107, 1, 1], 
     Hue[0.3091587449121022, 1, 1], 
     Hue[0.31416262983368515`, 1, 1], 
     Hue[0.31666457229447664`, 1, 1], 
     Hue[0.31916651475526814`, 1, 1], 
     Hue[0.324170399676851, 1, 1], 
     Hue[0.3266723421376425, 1, 1], 
     Hue[0.329174284598434, 1, 1], 
     Hue[0.33417816952001694`, 1, 1], 
     Hue[0.3441859393631829, 1, 1], 
     Hue[0.3465195552511106, 1, 1], 
     Hue[0.3488531711390381, 1, 1], 
     Hue[0.35352040291489345`, 1, 1], 
     Hue[0.3558540188028212, 1, 1], 
     Hue[0.3581876346907486, 1, 1], 
     Hue[0.36285486646660386`, 1, 1], 
     Hue[0.3651884823545317, 1, 1], 
     Hue[0.3675220982424591, 1, 1], 
     Hue[0.37218933001831445`, 1, 1], 
     Hue[0.38152379357002464`, 1, 1], 
     Hue[0.38385740945795216`, 1, 1], 
     Hue[0.38619102534587996`, 1, 1], 
     Hue[0.39085825712173494`, 1, 1], 
     Hue[0.3931918730096627, 1, 1], 
     Hue[0.3955254888975901, 1, 1], 
     Hue[0.4001927206734454, 1, 1], 
     Hue[0.40252633656137315`, 1, 1], 
     Hue[0.4048599524493006, 1, 1], 
     Hue[0.4095271842251559, 1, 1], 
     Hue[0.41886164777686646`, 1, 1], 
     Hue[0.42119526366479415`, 1, 1], 
     Hue[0.4235288795527217, 1, 1], 
     Hue[0.42819611132857693`, 1, 1], 
     Hue[0.4305297272165047, 1, 1], 
     Hue[0.4328633431044322, 1, 1], 
     Hue[0.4375305748802874, 1, 1], 
     Hue[0.4398641907682152, 1, 1], 
     Hue[0.4421978066561428, 1, 1], 
     Hue[0.44686503843199804`, 1, 1], 
     Hue[0.4561995019837082, 1, 1], 
     Hue[0.45853311787163575`, 1, 1], 
     Hue[0.46086673375956344`, 1, 1], 
     Hue[0.4655339655354184, 1, 1], 
     Hue[0.46786758142334617`, 1, 1], 
     Hue[0.4702011973112737, 1, 1], 
     Hue[0.4748684290871289, 1, 1], 
     Hue[0.4772020449750567, 1, 1], 
     Hue[0.47953566086298427`, 1, 1], 
     Hue[0.4842028926388394, 1, 1], 
     Hue[0.49353735619054995`, 1, 1], 
     Hue[0.49606805226271244`, 1, 1], 
     Hue[0.49859874833487494`, 1, 1], 
     Hue[0.5036601404791998, 1, 1], 
     Hue[0.5061908365513623, 1, 1], 
     Hue[0.5087215326235248, 1, 1], 
     Hue[0.5137829247678498, 1, 1], 
     Hue[0.5163136208400123, 1, 1], 
     Hue[0.5188443169121748, 1, 1], 
     Hue[0.5239057090564998, 1, 1], 
     Hue[0.5340284933451498, 1, 1], 
     Hue[0.5365591894173123, 1, 1], 
     Hue[0.5390898854894748, 1, 1], 
     Hue[0.5441512776337998, 1, 1], 
     Hue[0.5466819737059623, 1, 1], 
     Hue[0.5492126697781248, 1, 1], 
     Hue[0.5542740619224498, 1, 1], 
     Hue[0.5568047579946123, 1, 1], 
     Hue[0.5593354540667748, 1, 1], 
     Hue[0.5643968462110998, 1, 1], 
     Hue[0.5745196304997497, 1, 1], 
     Hue[0.5770503265719122, 1, 1], 
     Hue[0.5795810226440747, 1, 1], 
     Hue[0.5846424147883997, 1, 1], 
     Hue[0.5871731108605622, 1, 1], 
     Hue[0.5897038069327247, 1, 1], 
     Hue[0.5947651990770497, 1, 1], 
     Hue[0.5972958951492122, 1, 1], 
     Hue[0.5998265912213747, 1, 1], 
     Hue[0.6048879833656997, 1, 1], 
     Hue[0.6150107676543496, 1, 1], 
     Hue[0.6175414637265121, 1, 1], 
     Hue[0.6200721597986746, 1, 1], 
     Hue[0.6251335519429996, 1, 1], 
     Hue[0.6276642480151621, 1, 1], 
     Hue[0.6301949440873246, 1, 1], 
     Hue[0.6352563362316496, 1, 1], 
     Hue[0.6377870323038121, 1, 1], 
     Hue[0.6403177283759746, 1, 1], 
     Hue[0.6453791205202996, 1, 1], 
     Hue[0.6555019048089492, 1, 1], 
     Hue[0.6578642743082477, 1, 1], 
     Hue[0.6602266438075466, 1, 1], 
     Hue[0.6649513828061435, 1, 1], 
     Hue[0.6673137523054423, 1, 1], 
     Hue[0.6696761218047409, 1, 1], 
     Hue[0.674400860803338, 1, 1], 
     Hue[0.676763230302637, 1, 1], 
     Hue[0.6791255998019354, 1, 1], 
     Hue[0.6838503388005327, 1, 1], 
     Hue[0.6932998167977273, 1, 1], 
     Hue[0.695662186297026, 1, 1], 
     Hue[0.6980245557963245, 1, 1], 
     Hue[0.7027492947949219, 1, 1], 
     Hue[0.7051116642942206, 1, 1], 
     Hue[0.7074740337935191, 1, 1], 
     Hue[0.7121987727921164, 1, 1], 
     Hue[0.7145611422914152, 1, 1], 
     Hue[0.7169235117907138, 1, 1], 
     Hue[0.7216482507893109, 1, 1], 
     Hue[0.7310977287865053, 1, 1], 
     Hue[0.7334600982858038, 1, 1], 
     Hue[0.7358224677851026, 1, 1], 
     Hue[0.7405472067836996, 1, 1], 
     Hue[0.7429095762829985, 1, 1], 
     Hue[0.7452719457822969, 1, 1], 
     Hue[0.7499966847808942, 1, 1], 
     Hue[0.752359054280193, 1, 1], 
     Hue[0.7547214237794915, 1, 1], 
     Hue[0.7594461627780887, 1, 1], 
     Hue[0.7688956407752834, 1, 1], 
     Hue[0.7712580102745822, 1, 1], 
     Hue[0.7736203797738807, 1, 1], 
     Hue[0.7783451187724779, 1, 1], 
     Hue[0.7807074882717767, 1, 1], 
     Hue[0.7830698577710753, 1, 1], 
     Hue[0.7877945967696725, 1, 1], 
     Hue[0.7901569662689714, 1, 1], 
     Hue[0.7925193357682698, 1, 1], 
     Hue[0.7972440747668671, 1, 1], 
     Hue[0.8066935527640614, 1, 1], 
     Hue[0.8090093979428622, 1, 1], 
     Hue[0.8113252431216633, 1, 1], 
     Hue[0.8159569334792656, 1, 1], 
     Hue[0.8182727786580668, 1, 1], 
     Hue[0.8205886238368676, 1, 1], 
     Hue[0.8252203141944696, 1, 1], 
     Hue[0.8275361593732707, 1, 1], 
     Hue[0.8298520045520715, 1, 1], 
     Hue[0.8344836949096734, 1, 1], 
     Hue[0.8437470756248775, 1, 1], 
     Hue[0.8460629208036786, 1, 1], 
     Hue[0.8483787659824795, 1, 1], 
     Hue[0.8530104563400814, 1, 1], 
     Hue[0.8553263015188826, 1, 1], 
     Hue[0.8576421466976837, 1, 1], 
     Hue[0.8622738370552856, 1, 1], 
     Hue[0.8645896822340865, 1, 1], 
     Hue[0.8669055274128876, 1, 1], 
     Hue[0.8715372177704899, 1, 1], 
     Hue[0.8808005984856938, 1, 1], 
     Hue[0.8831164436644949, 1, 1], 
     Hue[0.8854322888432958, 1, 1], 
     Hue[0.8900639792008977, 1, 1], 
     Hue[0.892379824379699, 1, 1], 
     Hue[0.8946956695585001, 1, 1], 
     Hue[0.899327359916102, 1, 1], 
     Hue[0.9016432050949029, 1, 1], 
     Hue[0.903959050273704, 1, 1], 
     Hue[0.9085907406313063, 1, 1], 
     Hue[0.9178541213465102, 1, 1], 
     Hue[0.9201699665253114, 1, 1], 
     Hue[0.9224858117041121, 1, 1], 
     Hue[0.9271175020617142, 1, 1], 
     Hue[0.9294333472405153, 1, 1], 
     Hue[0.9317491924193162, 1, 1], 
     Hue[0.9363808827769181, 1, 1], 
     Hue[0.9386967279557192, 1, 1], 
     Hue[0.9410125731345201, 1, 1], 
     Hue[0.945644263492122, 1, 1], 
     Hue[0.9479601086709232, 1, 1], 
     Hue[0.9502759538497243, 1, 1], 
     Hue[0.9549076442073263, 1, 1], 
     Hue[0.9574205695703621, 1, 1], 
     Hue[0.9599334949333983, 1, 1], 
     Hue[0.96495934565947, 1, 1], 
     Hue[0.9674722710225059, 1, 1], 
     Hue[0.9699851963855417, 1, 1], 
     Hue[0.9750110471116135, 1, 1], 
     Hue[0.9775239724746493, 1, 1], 
     Hue[0.9800368978376851, 1, 1], 
     Hue[0.9850627485637571, 1, 1], 
     Hue[0.9875756739267929, 1, 1], 
     Hue[0.9900885992898291, 1, 1], 
     Hue[0.9951144500159008, 1, 1], 
     Hue[0.9976273753789366, 1, 1], 
     Hue[0.00014030074197249827`, 1, 1], 
     Hue[0.005166151468044201, 1, 1], 
     Hue[0.007679076831080055, 1, 1], 
     Hue[0.010192002194115906`, 1, 1], 
     Hue[0.015217852920187897`, 1, 1], 
     Hue[0.017730778283223746`, 1, 1], 
     Hue[0.02024370364625988, 1, 1], 
     Hue[0.025269554372331585`, 1, 1], 
     Hue[0.027782479735367443`, 1, 1], 
     Hue[0.030295405098403293`, 1, 1], 
     Hue[0.035321255824475, 1, 1], 
     Hue[0.037834181187510856`, 1, 1], 
     Hue[0.04034710655054671, 1, 1], 
     Hue[0.045372957276618414`, 1, 1], 
     Hue[0.04788588263965426, 1, 1], 
     Hue[0.05039880800269011, 1, 1], 
     Hue[0.0554246587287621, 1, 1], 
     Hue[0.05793758409179796, 1, 1], 
     Hue[0.06045050945483409, 1, 1], 
     Hue[0.06547636018090579, 1, 1], 
     Hue[0.06798928554394164, 1, 1], 
     Hue[0.0705022109069775, 1, 1], 
     Hue[0.0755280616330492, 1, 1], 
     Hue[0.07804098699608507, 1, 1], 
     Hue[0.08055391235912089, 1, 1], 
     Hue[0.08557976308519263, 1, 1], 
     Hue[0.08809268844822847, 1, 1], 
     Hue[0.09060561381126431, 1, 1], 
     Hue[0.09563146453733631, 1, 1], 
     Hue[0.09814438990037216, 1, 1], 
     Hue[0.10065731526340828`, 1, 1], 
     Hue[0.10568316598948, 1, 1], 
     Hue[0.10819609135251586`, 1, 1], 
     Hue[0.1107090167155517, 1, 1], 
     Hue[0.1157348674416234, 1, 1], 
     Hue[0.11807946623179555`, 1, 1], 
     Hue[0.12042406502196741`, 1, 1], 
     Hue[0.12511326260231143`, 1, 1], 
     Hue[0.12745786139248358`, 1, 1], 
     Hue[0.1298024601826557, 1, 1], 
     Hue[0.1344916577629997, 1, 1], 
     Hue[0.1368362565531716, 1, 1], 
     Hue[0.13918085534334373`, 1, 1], 
     Hue[0.14387005292368804`, 1, 1], 
     Hue[0.14621465171386017`, 1, 1], 
     Hue[0.14855925050403201`, 1, 1], 
     Hue[0.15324844808437604`, 1, 1], 
     Hue[0.15559304687454817`, 1, 1], 
     Hue[0.15793764566472007`, 1, 1], 
     Hue[0.16262684324506407`, 1, 1], 
     Hue[0.16497144203523623`, 1, 1], 
     Hue[0.16731604082540807`, 1, 1], 
     Hue[0.17200523840575208`, 1, 1], 
     Hue[0.17434983719592423`, 1, 1], 
     Hue[0.1766944359860961, 1, 1], 
     Hue[0.1813836335664401, 1, 1], 
     Hue[0.18372823235661226`, 1, 1], 
     Hue[0.18607283114678438`, 1, 1], 
     Hue[0.1907620287271284, 1, 1], 
     Hue[0.19310662751730026`, 1, 1], 
     Hue[0.19545122630747241`, 1, 1], 
     Hue[0.2001404238878167, 1, 1], 
     Hue[0.20248502267798885`, 1, 1], 
     Hue[0.20482962146816072`, 1, 1], 
     Hue[0.20951881904850472`, 1, 1], 
     Hue[0.21186341783867682`, 1, 1], 
     Hue[0.21420801662884872`, 1, 1], 
     Hue[0.21889721420919273`, 1, 1], 
     Hue[0.22124181299936488`, 1, 1], 
     Hue[0.22358641178953675`, 1, 1], 
     Hue[0.22827560936988076`, 1, 1], 
     Hue[0.2306202081600529, 1, 1], 
     Hue[0.23296480695022476`, 1, 1], 
     Hue[0.23765400453056879`, 1, 1], 
     Hue[0.2399986033207409, 1, 1], 
     Hue[0.24234320211091306`, 1, 1], 
     Hue[0.24703239969125707`, 1, 1], 
     Hue[0.24937699848142894`, 1, 1], 
     Hue[0.2517215972716011, 1, 1], 
     Hue[0.2564107948519454, 1, 1], 
     Hue[0.2587553936421175, 1, 1], 
     Hue[0.2610999924322894, 1, 1], 
     Hue[0.2657891900126334, 1, 1], 
     Hue[0.2683308689870403, 1, 1], 
     Hue[0.2708725479614471, 1, 1], 
     Hue[0.27595590591026087`, 1, 1], 
     Hue[0.27849758488466775`, 1, 1], 
     Hue[0.28103926385907463`, 1, 1], 
     Hue[0.2861226218078886, 1, 1], 
     Hue[0.2886643007822955, 1, 1], 
     Hue[0.29120597975670265`, 1, 1], 
     Hue[0.2962893377055164, 1, 1], 
     Hue[0.29883101667992323`, 1, 1], 
     Hue[0.3013726956543301, 1, 1], 
     Hue[0.3064560536031439, 1, 1], 
     Hue[0.30899773257755075`, 1, 1], 
     Hue[0.3115394115519576, 1, 1], 
     Hue[0.3166227695007716, 1, 1], 
     Hue[0.3191644484751785, 1, 1], 
     Hue[0.32170612744958565`, 1, 1], 
     Hue[0.32678948539839936`, 1, 1], 
     Hue[0.32933116437280624`, 1, 1], 
     Hue[0.3318728433472131, 1, 1], 
     Hue[0.3369562012960269, 1, 1], 
     Hue[0.3394978802704337, 1, 1], 
     Hue[0.3420395592448406, 1, 1], 
     Hue[0.3471229171936546, 1, 1], 
     Hue[0.3496645961680615, 1, 1], 
     Hue[0.35220627514246866`, 1, 1], 
     Hue[0.35728963309128237`, 1, 1], 
     Hue[0.35983131206568925`, 1, 1], 
     Hue[0.3623729910400961, 1, 1], 
     Hue[0.36745634898890983`, 1, 1], 
     Hue[0.3699980279633167, 1, 1], 
     Hue[0.3725397069377236, 1, 1], 
     Hue[0.37762306488653763`, 1, 1], 
     Hue[0.38016474386094445`, 1, 1], 
     Hue[0.3827064228353516, 1, 1], 
     Hue[0.38778978078416537`, 1, 1], 
     Hue[0.39033145975857225`, 1, 1], 
     Hue[0.39287313873297913`, 1, 1], 
     Hue[0.39795649668179284`, 1, 1], 
     Hue[0.4004981756561997, 1, 1], 
     Hue[0.4030398546306066, 1, 1], 
     Hue[0.4081232125794206, 1, 1], 
     Hue[0.41066489155382746`, 1, 1], 
     Hue[0.4132065705282346, 1, 1], 
     Hue[0.4157482495026415, 1, 1], 
     Hue[0.4182899284770484, 1, 1], 
     Hue[0.42083160745145526`, 1, 1], 
     Hue[0.4233732864258621, 1, 1], 
     Hue[0.42845664437467584`, 1, 1], 
     Hue[0.430829996776219, 1, 1], 
     Hue[0.4332033491777619, 1, 1], 
     Hue[0.4379500539808479, 1, 1], 
     Hue[0.4403234063823911, 1, 1], 
     Hue[0.4426967587839343, 1, 1], 
     Hue[0.4474434635870203, 1, 1], 
     Hue[0.44981681598856316`, 1, 1], 
     Hue[0.45219016839010634`, 1, 1], 
     Hue[0.45693687319319265`, 1, 1], 
     Hue[0.45931022559473583`, 1, 1], 
     Hue[0.4616835779962787, 1, 1], 
     Hue[0.4664302827993647, 1, 1], 
     Hue[0.4688036352009079, 1, 1], 
     Hue[0.4711769876024508, 1, 1], 
     Hue[0.47592369240553684`, 1, 1], 
     Hue[0.47829704480707996`, 1, 1], 
     Hue[0.48067039720862315`, 1, 1], 
     Hue[0.4854171020117092, 1, 1], 
     Hue[0.4877904544132521, 1, 1], 
     Hue[0.4901638068147952, 1, 1], 
     Hue[0.4949105116178816, 1, 1], 
     Hue[0.4972838640194247, 1, 1], 
     Hue[0.4996572164209676, 1, 1], 
     Hue[0.5044039212240535, 1, 1], 
     Hue[0.5067772736255967, 1, 1], 
     Hue[0.5091506260271397, 1, 1], 
     Hue[0.5138973308302257, 1, 1], 
     Hue[0.5162706832317687, 1, 1], 
     Hue[0.518644035633312, 1, 1], 
     Hue[0.523390740436398, 1, 1], 
     Hue[0.5257640928379409, 1, 1], 
     Hue[0.528137445239484, 1, 1], 
     Hue[0.5328841500425704, 1, 1], 
     Hue[0.5352575024441135, 1, 1], 
     Hue[0.5376308548456564, 1, 1], 
     Hue[0.5423775596487425, 1, 1], 
     Hue[0.5447509120502856, 1, 1], 
     Hue[0.5471242644518288, 1, 1], 
     Hue[0.5518709692549145, 1, 1], 
     Hue[0.5542443216564572, 1, 1], 
     Hue[0.5566176740580002, 1, 1], 
     Hue[0.5613643788610867, 1, 1], 
     Hue[0.5637377312626297, 1, 1], 
     Hue[0.566111083664173, 1, 1], 
     Hue[0.5708577884672592, 1, 1], 
     Hue[0.5732311408688024, 1, 1], 
     Hue[0.5756044932703456, 1, 1], 
     Hue[0.5803511980734313, 1, 1], 
     Hue[0.5850048542355217, 1, 1], 
     Hue[0.5896585103976129, 1, 1], 
     Hue[0.5943121665597039, 1, 1], 
     Hue[0.5989658227217949, 1, 1], 
     Hue[0.6036194788838858, 1, 1], 
     Hue[0.6082731350459769, 1, 1], 
     Hue[0.6129267912080679, 1, 1], 
     Hue[0.6175804473701589, 1, 1], 
     Hue[0.6222341035322498, 1, 1], 
     Hue[0.626887759694341, 1, 1], 
     Hue[0.6315414158564319, 1, 1], 
     Hue[0.6361950720185223, 1, 1], 
     Hue[0.6455023843427038, 1, 1], 
     Hue[0.6548096966668858, 1, 1], 
     Hue[0.6641170089910678, 1, 1], 
     Hue[0.6734243213152493, 1, 1], 
     Hue[0.6757511493962948, 1, 1], 
     Hue[0.6780779774773398, 1, 1], 
     Hue[0.6827316336394308, 1, 1], 
     Hue[0.6920389459636127, 1, 1], 
     Hue[0.6943657740446583, 1, 1], 
     Hue[0.6966926021257038, 1, 1], 
     Hue[0.7013462582877947, 1, 1], 
     Hue[0.7106535706119763, 1, 1], 
     Hue[0.7292681952603397, 1, 1], 
     Hue[0.7317921035256203, 1, 1], 
     Hue[0.7343160117909008, 1, 1], 
     Hue[0.7393638283214612, 1, 1], 
     Hue[0.7494594613825827, 1, 1], 
     Hue[0.7696507275048251, 1, 1], 
     Hue[0.8100332597493104, 1, 1], 
     Hue[0.8907983242382805, 1, 1], 
     Hue[0.04155555255293395, 1, 1], 
     Hue[0.20492591265861984`, 1, 1], 
     Hue[0.36531871625245843`, 1, 1], 
     Hue[0.5149386191830096, 1, 1], 
     Hue[0.6771716539045933, 1, 1], 
     Hue[0.8286317879628902, 1, 1], 
     Hue[0.9771143655093397, 1, 1], 
     Hue[0.1382100748468212, 1, 1], 
     Hue[0.28853288352101586`, 1, 1], 
     Hue[0.45146882398624244`, 1, 1], 
     Hue[0.45154490877811665`, 1, 1], 
     Hue[0.4516209935699909, 1, 1], 
     Hue[0.45177316315373883`, 1, 1], 
     Hue[0.4520775023212352, 1, 1], 
     Hue[0.4526861806562286, 1, 1], 
     Hue[0.4539035373262148, 1, 1], 
     Hue[0.45397962211808907`, 1, 1], 
     Hue[0.45405570690996333`, 1, 1], 
     Hue[0.45420787649371125`, 1, 1], 
     Hue[0.45451221566120764`, 1, 1], 
     Hue[0.4551208939962005, 1, 1], 
     Hue[0.4551969787880747, 1, 1], 
     Hue[0.45527306357994896`, 1, 1], 
     Hue[0.4554252331636969, 1, 1], 
     Hue[0.45572957233119327`, 1, 1], 
     Hue[0.45580565712306753`, 1, 1], 
     Hue[0.45588174191494174`, 1, 1], 
     Hue[0.45603391149869027`, 1, 1], 
     Hue[0.4561099962905645, 1, 1], 
     Hue[0.45618608108243874`, 1, 1], 
     Hue[0.456262165874313, 1, 1], 
     Hue[0.45633825066618666`, 1, 1], 
     Hue[0.31841357815830684`, 1, 1], 
     Hue[0.31841357815830684`, 1, 1], 
     Hue[0.3207388038555099, 1, 1], 
     Hue[0.3207388038555099, 1, 1], 
     Hue[0.32306402955271285`, 1, 1], 
     Hue[0.32306402955271285`, 1, 1], 
     Hue[0.3277144809471194, 1, 1], 
     Hue[0.3277144809471194, 1, 1], 
     Hue[0.33701538373593193`, 1, 1], 
     Hue[0.33701538373593193`, 1, 1], 
     Hue[0.34631628652474383`, 1, 1], 
     Hue[0.34631628652474383`, 1, 1], 
     Hue[0.35561718931355635`, 1, 1], 
     Hue[0.35561718931355635`, 1, 1], 
     Hue[0.35794241501075935`, 1, 1], 
     Hue[0.35794241501075935`, 1, 1], 
     Hue[0.3602676407079629, 1, 1], 
     Hue[0.3602676407079629, 1, 1], 
     Hue[0.36491809210236886`, 1, 1], 
     Hue[0.36491809210236886`, 1, 1], 
     Hue[0.3695685434967748, 1, 1], 
     Hue[0.3695685434967748, 1, 1], 
     Hue[0.37421899489118077`, 1, 1], 
     Hue[0.37421899489118077`, 1, 1], 
     Hue[0.37886944628558666`, 1, 1], 
     Hue[0.37886944628558666`, 1, 1], 
     Hue[0.3835198976799933, 1, 1], 
     Hue[0.3835198976799933, 1, 1], 
     Hue[0.38817034907439985`, 1, 1], 
     Hue[0.38817034907439985`, 1, 1], 
     Hue[0.3928208004688058, 1, 1], 
     Hue[0.3928208004688058, 1, 1], 
     Hue[0.39747125186321175`, 1, 1], 
     Hue[0.39747125186321175`, 1, 1], 
     Hue[0.4021217032576183, 1, 1], 
     Hue[0.4021217032576183, 1, 1], 
     Hue[0.4067721546520248, 1, 1], 
     Hue[0.4067721546520248, 1, 1], 
     Hue[0.41142260604643077`, 1, 1], 
     Hue[0.41142260604643077`, 1, 1], 
     Hue[0.4160730574408368, 1, 1], 
     Hue[0.4160730574408368, 1, 1], 
     Hue[0.4207235088352433, 1, 1], 
     Hue[0.4207235088352433, 1, 1], 
     Hue[0.4230487345324463, 1, 1], 
     Hue[0.4230487345324463, 1, 1], 
     Hue[0.42537396022964985`, 1, 1], 
     Hue[0.42537396022964985`, 1, 1], 
     Hue[0.4300244116240558, 1, 1], 
     Hue[0.4300244116240558, 1, 1], 
     Hue[0.4325467175054938, 1, 1], 
     Hue[0.4325467175054938, 1, 1], 
     Hue[0.43506902338693176`, 1, 1], 
     Hue[0.43506902338693176`, 1, 1], 
     Hue[0.43759132926836974`, 1, 1], 
     Hue[0.43759132926836974`, 1, 1], 
     Hue[0.44011363514980767`, 1, 1], 
     Hue[0.44011363514980767`, 1, 1], 
     Hue[0.44263594103124576`, 1, 1], 
     Hue[0.44263594103124576`, 1, 1], 
     Hue[0.44515824691268374`, 1, 1], 
     Hue[0.44515824691268374`, 1, 1], 
     Hue[0.4502028586755597, 1, 1], 
     Hue[0.4502028586755597, 1, 1], 
     Hue[0.45272516455699774`, 1, 1], 
     Hue[0.45272516455699774`, 1, 1], 
     Hue[0.4552474704384357, 1, 1], 
     Hue[0.4552474704384357, 1, 1], 
     Hue[0.4577697763198737, 1, 1], 
     Hue[0.4577697763198737, 1, 1], 
     Hue[0.4602920822013117, 1, 1], 
     Hue[0.4602920822013117, 1, 1], 
     Hue[0.46281438808274966`, 1, 1], 
     Hue[0.46281438808274966`, 1, 1], 
     Hue[0.46533669396418764`, 1, 1], 
     Hue[0.46533669396418764`, 1, 1], 
     Hue[0.47038130572706366`, 1, 1], 
     Hue[0.47038130572706366`, 1, 1], 
     Hue[0.47290361160850164`, 1, 1], 
     Hue[0.47290361160850164`, 1, 1], 
     Hue[0.4754259174899396, 1, 1], 
     Hue[0.4754259174899396, 1, 1], 
     Hue[0.4779482233713776, 1, 1], 
     Hue[0.4779482233713776, 1, 1], 
     Hue[0.4804705292528156, 1, 1], 
     Hue[0.4804705292528156, 1, 1], 
     Hue[0.4829928351342536, 1, 1], 
     Hue[0.4829928351342536, 1, 1], 
     Hue[0.4855151410156916, 1, 1], 
     Hue[0.4855151410156916, 1, 1], 
     Hue[0.49055975277856756`, 1, 1], 
     Hue[0.49055975277856756`, 1, 1], 
     Hue[0.49308205866000554`, 1, 1], 
     Hue[0.49308205866000554`, 1, 1], 
     Hue[0.4956043645414435, 1, 1], 
     Hue[0.4956043645414435, 1, 1], 
     Hue[0.49812667042288156`, 1, 1], 
     Hue[0.49812667042288156`, 1, 1], 
     Hue[0.5006489763043194, 1, 1], 
     Hue[0.5006489763043194, 1, 1], 
     Hue[0.5031712821857575, 1, 1], 
     Hue[0.5031712821857575, 1, 1], 
     Hue[0.5056935880671953, 1, 1], 
     Hue[0.5056935880671953, 1, 1], 
     Hue[0.5107381998300711, 1, 1], 
     Hue[0.5107381998300711, 1, 1], 
     Hue[0.5132605057115088, 1, 1], 
     Hue[0.5132605057115088, 1, 1], 
     Hue[0.5157828115929468, 1, 1], 
     Hue[0.5157828115929468, 1, 1], 
     Hue[0.5183051174743848, 1, 1], 
     Hue[0.5183051174743848, 1, 1], 
     Hue[0.5208274233558228, 1, 1], 
     Hue[0.5208274233558228, 1, 1], 
     Hue[0.5233497292372608, 1, 1], 
     Hue[0.5233497292372608, 1, 1], 
     Hue[0.5258720351186987, 1, 1], 
     Hue[0.5258720351186987, 1, 1], 
     Hue[0.5309166468815747, 1, 1], 
     Hue[0.5309166468815747, 1, 1], 
     Hue[0.5334389527630128, 1, 1], 
     Hue[0.5334389527630128, 1, 1], 
     Hue[0.5359612586444507, 1, 1], 
     Hue[0.5359612586444507, 1, 1], 
     Hue[0.5384835645258887, 1, 1], 
     Hue[0.5384835645258887, 1, 1], 
     Hue[0.5410058704073267, 1, 1], 
     Hue[0.5410058704073267, 1, 1], 
     Hue[0.5435281762887647, 1, 1], 
     Hue[0.5435281762887647, 1, 1], 
     Hue[0.5460504821702027, 1, 1], 
     Hue[0.5460504821702027, 1, 1], 
     Hue[0.5510950939330787, 1, 1], 
     Hue[0.5510950939330787, 1, 1], 
     Hue[0.5536173998145166, 1, 1], 
     Hue[0.5536173998145166, 1, 1], 
     Hue[0.5561397056959547, 1, 1], 
     Hue[0.5561397056959547, 1, 1], 
     Hue[0.5586620115773926, 1, 1], 
     Hue[0.5586620115773926, 1, 1], 
     Hue[0.5611843174588307, 1, 1], 
     Hue[0.5611843174588307, 1, 1], 
     Hue[0.5637066233402686, 1, 1], 
     Hue[0.5637066233402686, 1, 1], 
     Hue[0.5662289292217066, 1, 1], 
     Hue[0.5662289292217066, 1, 1], 
     Hue[0.5712735409845826, 1, 1], 
     Hue[0.5712735409845826, 1, 1], 
     Hue[0.5737958468660206, 1, 1], 
     Hue[0.5737958468660206, 1, 1], 
     Hue[0.5763181527474586, 1, 1], 
     Hue[0.5763181527474586, 1, 1], 
     Hue[0.5788404586288965, 1, 1], 
     Hue[0.5788404586288965, 1, 1], 
     Hue[0.5813627645103345, 1, 1], 
     Hue[0.5813627645103345, 1, 1], 
     Hue[0.5838850703917726, 1, 1], 
     Hue[0.5838850703917726, 1, 1], 
     Hue[0.5864073762732105, 1, 1], 
     Hue[0.5864073762732105, 1, 1], 
     Hue[0.5914519880360865, 1, 1], 
     Hue[0.5914519880360865, 1, 1], 
     Hue[0.5938059673446605, 1, 1], 
     Hue[0.5938059673446605, 1, 1], 
     Hue[0.5961599466532348, 1, 1], 
     Hue[0.5961599466532348, 1, 1], 
     Hue[0.6008679052703831, 1, 1], 
     Hue[0.6008679052703831, 1, 1], 
     Hue[0.6032218845789571, 1, 1], 
     Hue[0.6032218845789571, 1, 1], 
     Hue[0.6055758638875314, 1, 1], 
     Hue[0.6055758638875314, 1, 1], 
     Hue[0.6102838225046796, 1, 1], 
     Hue[0.6102838225046796, 1, 1], 
     Hue[0.6126378018132537, 1, 1], 
     Hue[0.6126378018132537, 1, 1], 
     Hue[0.614991781121828, 1, 1], 
     Hue[0.614991781121828, 1, 1], 
     Hue[0.6196997397389763, 1, 1], 
     Hue[0.6196997397389763, 1, 1], 
     Hue[0.6220537190475501, 1, 1], 
     Hue[0.6220537190475501, 1, 1], 
     Hue[0.6244076983561242, 1, 1], 
     Hue[0.6244076983561242, 1, 1], 
     Hue[0.6291156569732725, 1, 1], 
     Hue[0.6291156569732725, 1, 1], 
     Hue[0.6314696362818468, 1, 1], 
     Hue[0.6314696362818468, 1, 1], 
     Hue[0.6338236155904208, 1, 1], 
     Hue[0.6338236155904208, 1, 1], 
     Hue[0.6385315742075688, 1, 1], 
     Hue[0.6385315742075688, 1, 1], 
     Hue[0.6408855535161428, 1, 1], 
     Hue[0.6408855535161428, 1, 1], 
     Hue[0.6432395328247171, 1, 1], 
     Hue[0.6432395328247171, 1, 1], 
     Hue[0.6479474914418654, 1, 1], 
     Hue[0.6479474914418654, 1, 1], 
     Hue[0.6503014707504393, 1, 1], 
     Hue[0.6503014707504393, 1, 1], 
     Hue[0.6526554500590136, 1, 1], 
     Hue[0.6526554500590136, 1, 1], 
     Hue[0.6573634086761619, 1, 1], 
     Hue[0.6573634086761619, 1, 1], 
     Hue[0.659717387984736, 1, 1], 
     Hue[0.659717387984736, 1, 1], 
     Hue[0.6620713672933103, 1, 1], 
     Hue[0.6620713672933103, 1, 1], 
     Hue[0.6667793259104585, 1, 1], 
     Hue[0.6667793259104585, 1, 1], 
     Hue[0.6691333052190325, 1, 1], 
     Hue[0.6691333052190325, 1, 1], 
     Hue[0.6714872845276068, 1, 1], 
     Hue[0.6714872845276068, 1, 1], 
     Hue[0.6761952431447551, 1, 1], 
     Hue[0.6761952431447551, 1, 1], 
     Hue[0.678549222453329, 1, 1], 
     Hue[0.678549222453329, 1, 1], 
     Hue[0.6809032017619033, 1, 1], 
     Hue[0.6809032017619033, 1, 1], 
     Hue[0.6856111603790516, 1, 1], 
     Hue[0.6856111603790516, 1, 1], 
     Hue[0.6879651396876256, 1, 1], 
     Hue[0.6879651396876256, 1, 1], 
     Hue[0.6903191189961999, 1, 1], 
     Hue[0.6903191189961999, 1, 1], 
     Hue[0.6950270776133483, 1, 1], 
     Hue[0.6950270776133483, 1, 1], 
     Hue[0.6973810569219222, 1, 1], 
     Hue[0.6973810569219222, 1, 1], 
     Hue[0.6997350362304962, 1, 1], 
     Hue[0.6997350362304962, 1, 1], 
     Hue[0.7044429948476445, 1, 1], 
     Hue[0.7044429948476445, 1, 1], 
     Hue[0.7067969741562188, 1, 1], 
     Hue[0.7067969741562188, 1, 1], 
     Hue[0.7091509534647928, 1, 1], 
     Hue[0.7091509534647928, 1, 1], 
     Hue[0.7138589120819407, 1, 1], 
     Hue[0.7138589120819407, 1, 1], 
     Hue[0.7162128913905148, 1, 1], 
     Hue[0.7162128913905148, 1, 1], 
     Hue[0.7185668706990891, 1, 1], 
     Hue[0.7185668706990891, 1, 1], 
     Hue[0.7232748293162373, 1, 1], 
     Hue[0.7232748293162373, 1, 1], 
     Hue[0.7256288086248113, 1, 1], 
     Hue[0.7256288086248113, 1, 1], 
     Hue[0.7279827879333856, 1, 1], 
     Hue[0.7279827879333856, 1, 1], 
     Hue[0.7326907465505339, 1, 1], 
     Hue[0.7326907465505339, 1, 1], 
     Hue[0.735044725859108, 1, 1], 
     Hue[0.735044725859108, 1, 1], 
     Hue[0.7373987051676822, 1, 1], 
     Hue[0.7373987051676822, 1, 1], 
     Hue[0.7421066637848305, 1, 1], 
     Hue[0.7421066637848305, 1, 1], 
     Hue[0.7446577232776395, 1, 1], 
     Hue[0.7446577232776395, 1, 1], 
     Hue[0.7472087827704486, 1, 1], 
     Hue[0.7472087827704486, 1, 1], 
     Hue[0.7523109017560665, 1, 1], 
     Hue[0.7523109017560665, 1, 1], 
     Hue[0.7548619612488755, 1, 1], 
     Hue[0.7548619612488755, 1, 1], 
     Hue[0.7574130207416845, 1, 1], 
     Hue[0.7574130207416845, 1, 1], 
     Hue[0.7625151397273026, 1, 1], 
     Hue[0.7625151397273026, 1, 1], 
     Hue[0.7650661992201115, 1, 1], 
     Hue[0.7650661992201115, 1, 1], 
     Hue[0.7676172587129205, 1, 1], 
     Hue[0.7676172587129205, 1, 1], 
     Hue[0.7727193776985386, 1, 1], 
     Hue[0.7727193776985386, 1, 1], 
     Hue[0.7752704371913476, 1, 1], 
     Hue[0.7752704371913476, 1, 1], 
     Hue[0.7778214966841566, 1, 1], 
     Hue[0.7778214966841566, 1, 1], 
     Hue[0.7829236156697745, 1, 1], 
     Hue[0.7829236156697745, 1, 1], 
     Hue[0.7854746751625836, 1, 1], 
     Hue[0.7854746751625836, 1, 1], 
     Hue[0.7880257346553926, 1, 1], 
     Hue[0.7880257346553926, 1, 1], 
     Hue[0.7931278536410106, 1, 1], 
     Hue[0.7931278536410106, 1, 1], 
     Hue[0.7956789131338196, 1, 1], 
     Hue[0.7956789131338196, 1, 1], 
     Hue[0.7982299726266286, 1, 1], 
     Hue[0.7982299726266286, 1, 1], 
     Hue[0.8033320916122466, 1, 1], 
     Hue[0.8033320916122466, 1, 1], 
     Hue[0.8058831511050557, 1, 1], 
     Hue[0.8058831511050557, 1, 1], 
     Hue[0.8084342105978647, 1, 1], 
     Hue[0.8084342105978647, 1, 1], 
     Hue[0.8135363295834827, 1, 1], 
     Hue[0.8135363295834827, 1, 1], 
     Hue[0.8160873890762916, 1, 1], 
     Hue[0.8160873890762916, 1, 1], 
     Hue[0.8186384485691007, 1, 1], 
     Hue[0.8186384485691007, 1, 1], 
     Hue[0.8237405675547187, 1, 1], 
     Hue[0.8237405675547187, 1, 1], 
     Hue[0.8262916270475277, 1, 1], 
     Hue[0.8262916270475277, 1, 1], 
     Hue[0.8288426865403367, 1, 1], 
     Hue[0.8288426865403367, 1, 1], 
     Hue[0.8339448055259547, 1, 1], 
     Hue[0.8339448055259547, 1, 1], 
     Hue[0.8364958650187637, 1, 1], 
     Hue[0.8364958650187637, 1, 1], 
     Hue[0.8390469245115727, 1, 1], 
     Hue[0.8390469245115727, 1, 1], 
     Hue[0.8441490434971908, 1, 1], 
     Hue[0.8441490434971908, 1, 1], 
     Hue[0.8467001029899998, 1, 1], 
     Hue[0.8467001029899998, 1, 1], 
     Hue[0.8492511624828087, 1, 1], 
     Hue[0.8492511624828087, 1, 1], 
     Hue[0.8543532814684268, 1, 1], 
     Hue[0.8543532814684268, 1, 1], 
     Hue[0.8569043409612358, 1, 1], 
     Hue[0.8569043409612358, 1, 1], 
     Hue[0.8594554004540448, 1, 1], 
     Hue[0.8594554004540448, 1, 1], 
     Hue[0.8645575194396629, 1, 1], 
     Hue[0.8645575194396629, 1, 1], 
     Hue[0.8671085789324718, 1, 1], 
     Hue[0.8671085789324718, 1, 1], 
     Hue[0.8696596384252808, 1, 1], 
     Hue[0.8696596384252808, 1, 1], 
     Hue[0.8747617574108988, 1, 1], 
     Hue[0.8747617574108988, 1, 1], 
     Hue[0.8773128169037079, 1, 1], 
     Hue[0.8773128169037079, 1, 1], 
     Hue[0.8798638763965169, 1, 1], 
     Hue[0.8798638763965169, 1, 1], 
     Hue[0.8849659953821348, 1, 1], 
     Hue[0.8849659953821348, 1, 1], 
     Hue[0.8875170548749438, 1, 1], 
     Hue[0.8875170548749438, 1, 1], 
     Hue[0.8900681143677529, 1, 1], 
     Hue[0.8900681143677529, 1, 1], 
     Hue[0.8951702333533709, 1, 1], 
     Hue[0.8951702333533709, 1, 1], 
     Hue[0.89772129284618, 1, 1], 
     Hue[0.89772129284618, 1, 1], 
     Hue[0.9002723523389888, 1, 1], 
     Hue[0.9002723523389888, 1, 1], 
     Hue[0.9053744713246066, 1, 1], 
     Hue[0.9053744713246066, 1, 1], 
     Hue[0.9078790064969182, 1, 1], 
     Hue[0.9078790064969182, 1, 1], 
     Hue[0.9103835416692296, 1, 1], 
     Hue[0.9103835416692296, 1, 1], 
     Hue[0.9153926120138524, 1, 1], 
     Hue[0.9153926120138524, 1, 1], 
     Hue[0.9178971471861637, 1, 1], 
     Hue[0.9178971471861637, 1, 1], 
     Hue[0.920401682358475, 1, 1], 
     Hue[0.920401682358475, 1, 1], 
     Hue[0.9254107527030977, 1, 1], 
     Hue[0.9254107527030977, 1, 1], 
     Hue[0.9279152878754091, 1, 1], 
     Hue[0.9279152878754091, 1, 1], 
     Hue[0.9304198230477204, 1, 1], 
     Hue[0.9304198230477204, 1, 1], 
     Hue[0.9354288933923434, 1, 1], 
     Hue[0.9354288933923434, 1, 1], 
     Hue[0.937933428564655, 1, 1], 
     Hue[0.937933428564655, 1, 1], 
     Hue[0.9404379637369664, 1, 1], 
     Hue[0.9404379637369664, 1, 1], 
     Hue[0.945447034081589, 1, 1], 
     Hue[0.945447034081589, 1, 1], 
     Hue[0.9479515692539005, 1, 1], 
     Hue[0.9479515692539005, 1, 1], 
     Hue[0.9504561044262118, 1, 1], 
     Hue[0.9504561044262118, 1, 1], 
     Hue[0.9554651747708345, 1, 1], 
     Hue[0.9554651747708345, 1, 1], 
     Hue[0.9579697099431458, 1, 1], 
     Hue[0.9579697099431458, 1, 1], 
     Hue[0.9604742451154572, 1, 1], 
     Hue[0.9604742451154572, 1, 1], 
     Hue[0.9654833154600801, 1, 1], 
     Hue[0.9654833154600801, 1, 1], 
     Hue[0.9679878506323918, 1, 1], 
     Hue[0.9679878506323918, 1, 1], 
     Hue[0.9704923858047031, 1, 1], 
     Hue[0.9704923858047031, 1, 1], 
     Hue[0.9755014561493258, 1, 1], 
     Hue[0.9755014561493258, 1, 1], 
     Hue[0.9780059913216371, 1, 1], 
     Hue[0.9780059913216371, 1, 1], 
     Hue[0.9805105264939485, 1, 1], 
     Hue[0.9805105264939485, 1, 1], 
     Hue[0.9855195968385713, 1, 1], 
     Hue[0.9855195968385713, 1, 1], 
     Hue[0.9880241320108826, 1, 1], 
     Hue[0.9880241320108826, 1, 1], 
     Hue[0.9905286671831939, 1, 1], 
     Hue[0.9905286671831939, 1, 1], 
     Hue[0.9955377375278166, 1, 1], 
     Hue[0.9955377375278166, 1, 1], 
     Hue[0.998042272700128, 1, 1], 
     Hue[0.998042272700128, 1, 1], 
     Hue[0.0005468078724393148, 1, 1], 
     Hue[0.0005468078724393148, 1, 1], 
     Hue[0.005555878217062295, 1, 1], 
     Hue[0.005555878217062295, 1, 1], 
     Hue[0.008060413389373928, 1, 1], 
     Hue[0.008060413389373928, 1, 1], 
     Hue[0.010564948561685276`, 1, 1], 
     Hue[0.010564948561685276`, 1, 1], 
     Hue[0.015574018906307979`, 1, 1], 
     Hue[0.015574018906307979`, 1, 1], 
     Hue[0.018078554078619323`, 1, 1], 
     Hue[0.018078554078619323`, 1, 1], 
     Hue[0.02058308925093067, 1, 1], 
     Hue[0.02058308925093067, 1, 1], 
     Hue[0.025592159595553377`, 1, 1], 
     Hue[0.025592159595553377`, 1, 1], 
     Hue[0.02809669476786473, 1, 1], 
     Hue[0.02809669476786473, 1, 1], 
     Hue[0.030601229940176073`, 1, 1], 
     Hue[0.030601229940176073`, 1, 1], 
     Hue[0.03561030028479876, 1, 1], 
     Hue[0.03561030028479876, 1, 1], 
     Hue[0.03811483545711011, 1, 1], 
     Hue[0.03811483545711011, 1, 1], 
     Hue[0.04061937062942147, 1, 1], 
     Hue[0.04061937062942147, 1, 1], 
     Hue[0.045628440974044455`, 1, 1], 
     Hue[0.045628440974044455`, 1, 1], 
     Hue[0.048132976146356074`, 1, 1], 
     Hue[0.048132976146356074`, 1, 1], 
     Hue[0.05063751131866743, 1, 1], 
     Hue[0.05063751131866743, 1, 1], 
     Hue[0.05564658166329014, 1, 1], 
     Hue[0.05564658166329014, 1, 1], 
     Hue[0.05815111683560147, 1, 1], 
     Hue[0.05815111683560147, 1, 1], 
     Hue[0.06065565200791283, 1, 1], 
     Hue[0.06065565200791283, 1, 1], 
     Hue[0.06566472235253552, 1, 1], 
     Hue[0.06566472235253552, 1, 1], 
     Hue[0.06800093095198288, 1, 1], 
     Hue[0.06800093095198288, 1, 1], 
     Hue[0.0703371395514305, 1, 1], 
     Hue[0.0703371395514305, 1, 1], 
     Hue[0.07500955675032554, 1, 1], 
     Hue[0.07500955675032554, 1, 1], 
     Hue[0.07734576534977317, 1, 1], 
     Hue[0.07734576534977317, 1, 1], 
     Hue[0.0796819739492208, 1, 1], 
     Hue[0.0796819739492208, 1, 1], 
     Hue[0.08435439114811583, 1, 1], 
     Hue[0.08435439114811583, 1, 1], 
     Hue[0.08669059974756345, 1, 1], 
     Hue[0.08669059974756345, 1, 1], 
     Hue[0.08902680834701081, 1, 1], 
     Hue[0.08902680834701081, 1, 1], 
     Hue[0.09369922554590611, 1, 1], 
     Hue[0.09369922554590611, 1, 1], 
     Hue[0.1030440599436961, 1, 1], 
     Hue[0.1030440599436961, 1, 1], 
     Hue[0.10538026854314346`, 1, 1], 
     Hue[0.10538026854314346`, 1, 1], 
     Hue[0.1077164771425911, 1, 1], 
     Hue[0.1077164771425911, 1, 1], 
     Hue[0.1123888943414861, 1, 1], 
     Hue[0.1123888943414861, 1, 1], 
     Hue[0.11472510294093347`, 1, 1], 
     Hue[0.11472510294093347`, 1, 1], 
     Hue[0.11706131154038113`, 1, 1], 
     Hue[0.11706131154038113`, 1, 1], 
     Hue[0.12173372873927613`, 1, 1], 
     Hue[0.12173372873927613`, 1, 1], 
     Hue[0.12406993733872347`, 1, 1], 
     Hue[0.12406993733872347`, 1, 1], 
     Hue[0.12640614593817112`, 1, 1], 
     Hue[0.12640614593817112`, 1, 1], 
     Hue[0.13107856313706612`, 1, 1], 
     Hue[0.13107856313706612`, 1, 1], 
     Hue[0.14042339753485641`, 1, 1], 
     Hue[0.14042339753485641`, 1, 1], 
     Hue[0.14275960613430405`, 1, 1], 
     Hue[0.14275960613430405`, 1, 1], 
     Hue[0.1450958147337514, 1, 1], 
     Hue[0.1450958147337514, 1, 1], 
     Hue[0.14976823193264668`, 1, 1], 
     Hue[0.14976823193264668`, 1, 1], 
     Hue[0.15210444053209407`, 1, 1], 
     Hue[0.15210444053209407`, 1, 1], 
     Hue[0.1544406491315417, 1, 1], 
     Hue[0.1544406491315417, 1, 1], 
     Hue[0.1591130663304367, 1, 1], 
     Hue[0.1591130663304367, 1, 1], 
     Hue[0.16144927492988406`, 1, 1], 
     Hue[0.16144927492988406`, 1, 1], 
     Hue[0.1637854835293317, 1, 1], 
     Hue[0.1637854835293317, 1, 1], 
     Hue[0.1684579007282267, 1, 1], 
     Hue[0.1684579007282267, 1, 1], 
     Hue[0.1778027351260167, 1, 1], 
     Hue[0.1778027351260167, 1, 1], 
     Hue[0.18013894372546407`, 1, 1], 
     Hue[0.18013894372546407`, 1, 1], 
     Hue[0.1824751523249117, 1, 1], 
     Hue[0.1824751523249117, 1, 1], 
     Hue[0.1871475695238067, 1, 1], 
     Hue[0.1871475695238067, 1, 1], 
     Hue[0.18948377812325434`, 1, 1], 
     Hue[0.18948377812325434`, 1, 1], 
     Hue[0.19181998672270198`, 1, 1], 
     Hue[0.19181998672270198`, 1, 1], 
     Hue[0.196492403921597, 1, 1], 
     Hue[0.196492403921597, 1, 1], 
     Hue[0.19882861252104464`, 1, 1], 
     Hue[0.19882861252104464`, 1, 1], 
     Hue[0.201164821120492, 1, 1], 
     Hue[0.201164821120492, 1, 1], 
     Hue[0.20583723831938727`, 1, 1], 
     Hue[0.20583723831938727`, 1, 1], 
     Hue[0.2151820727171773, 1, 1], 
     Hue[0.2151820727171773, 1, 1], 
     Hue[0.21771536150085966`, 1, 1], 
     Hue[0.21771536150085966`, 1, 1], 
     Hue[0.22024865028454202`, 1, 1], 
     Hue[0.22024865028454202`, 1, 1], 
     Hue[0.22531522785190675`, 1, 1], 
     Hue[0.22531522785190675`, 1, 1], 
     Hue[0.2278485166355891, 1, 1], 
     Hue[0.2278485166355891, 1, 1], 
     Hue[0.23038180541927147`, 1, 1], 
     Hue[0.23038180541927147`, 1, 1], 
     Hue[0.23544838298663648`, 1, 1], 
     Hue[0.23544838298663648`, 1, 1], 
     Hue[0.23798167177031915`, 1, 1], 
     Hue[0.23798167177031915`, 1, 1], 
     Hue[0.2405149605540015, 1, 1], 
     Hue[0.2405149605540015, 1, 1], 
     Hue[0.24558153812136624`, 1, 1], 
     Hue[0.24558153812136624`, 1, 1], 
     Hue[0.2557146932560957, 1, 1], 
     Hue[0.2557146932560957, 1, 1], 
     Hue[0.25824798203977806`, 1, 1], 
     Hue[0.25824798203977806`, 1, 1], 
     Hue[0.2607812708234604, 1, 1], 
     Hue[0.2607812708234604, 1, 1], 
     Hue[0.26584784839082515`, 1, 1], 
     Hue[0.26584784839082515`, 1, 1], 
     Hue[0.2683811371745075, 1, 1], 
     Hue[0.2683811371745075, 1, 1], 
     Hue[0.2709144259581899, 1, 1], 
     Hue[0.2709144259581899, 1, 1], 
     Hue[0.27598100352555494`, 1, 1], 
     Hue[0.27598100352555494`, 1, 1], 
     Hue[0.2785142923092376, 1, 1], 
     Hue[0.2785142923092376, 1, 1], 
     Hue[0.28104758109291994`, 1, 1], 
     Hue[0.28104758109291994`, 1, 1], 
     Hue[0.28611415866028467`, 1, 1], 
     Hue[0.28611415866028467`, 1, 1], 
     Hue[0.2962473137950141, 1, 1], 
     Hue[0.2962473137950141, 1, 1], 
     Hue[0.2987806025786965, 1, 1], 
     Hue[0.2987806025786965, 1, 1], 
     Hue[0.30131389136237885`, 1, 1], 
     Hue[0.30131389136237885`, 1, 1], 
     Hue[0.3063804689297436, 1, 1], 
     Hue[0.3063804689297436, 1, 1], 
     Hue[0.30891375771342594`, 1, 1], 
     Hue[0.30891375771342594`, 1, 1], 
     Hue[0.3114470464971083, 1, 1], 
     Hue[0.3114470464971083, 1, 1], 
     Hue[0.3165136240644733, 1, 1], 
     Hue[0.3165136240644733, 1, 1], 
     Hue[0.31904691284815595`, 1, 1], 
     Hue[0.31904691284815595`, 1, 1], 
     Hue[0.32158020163183837`, 1, 1], 
     Hue[0.32158020163183837`, 1, 1], 
     Hue[0.3266467791992031, 1, 1], 
     Hue[0.3266467791992031, 1, 1], 
     Hue[0.33677993433393255`, 1, 1], 
     Hue[0.33677993433393255`, 1, 1], 
     Hue[0.3393132231176149, 1, 1], 
     Hue[0.3393132231176149, 1, 1], 
     Hue[0.3418465119012973, 1, 1], 
     Hue[0.3418465119012973, 1, 1], 
     Hue[0.3469130894686623, 1, 1], 
     Hue[0.3469130894686623, 1, 1], 
     Hue[0.3494463782523449, 1, 1], 
     Hue[0.3494463782523449, 1, 1], 
     Hue[0.3519796670360273, 1, 1], 
     Hue[0.3519796670360273, 1, 1], 
     Hue[0.357046244603392, 1, 1], 
     Hue[0.357046244603392, 1, 1], 
     Hue[0.3595795333870744, 1, 1], 
     Hue[0.3595795333870744, 1, 1], 
     Hue[0.36211282217075674`, 1, 1], 
     Hue[0.36211282217075674`, 1, 1], 
     Hue[0.36717939973812147`, 1, 1], 
     Hue[0.36717939973812147`, 1, 1], 
     Hue[0.37731255487285126`, 1, 1], 
     Hue[0.37731255487285126`, 1, 1], 
     Hue[0.3796775170836699, 1, 1], 
     Hue[0.3796775170836699, 1, 1], 
     Hue[0.38204247929448826`, 1, 1], 
     Hue[0.38204247929448826`, 1, 1], 
     Hue[0.3867724037161256, 1, 1], 
     Hue[0.3867724037161256, 1, 1], 
     Hue[0.389137365926944, 1, 1], 
     Hue[0.389137365926944, 1, 1], 
     Hue[0.39150232813776265`, 1, 1], 
     Hue[0.39150232813776265`, 1, 1], 
     Hue[0.39623225255939964`, 1, 1], 
     Hue[0.39623225255939964`, 1, 1], 
     Hue[0.39859721477021803`, 1, 1], 
     Hue[0.39859721477021803`, 1, 1], 
     Hue[0.4009621769810367, 1, 1], 
     Hue[0.4009621769810367, 1, 1], 
     Hue[0.40569210140267375`, 1, 1], 
     Hue[0.40569210140267375`, 1, 1], 
     Hue[0.4151519502459478, 1, 1], 
     Hue[0.4151519502459478, 1, 1], 
     Hue[0.4175169124567662, 1, 1], 
     Hue[0.4175169124567662, 1, 1], 
     Hue[0.41988187466758486`, 1, 1], 
     Hue[0.41988187466758486`, 1, 1], 
     Hue[0.42461179908922186`, 1, 1], 
     Hue[0.42461179908922186`, 1, 1], 
     Hue[0.42697676130004053`, 1, 1], 
     Hue[0.42697676130004053`, 1, 1], 
     Hue[0.4293417235108592, 1, 1], 
     Hue[0.4293417235108592, 1, 1], 
     Hue[0.43407164793249625`, 1, 1], 
     Hue[0.43407164793249625`, 1, 1], 
     Hue[0.43643661014331486`, 1, 1], 
     Hue[0.43643661014331486`, 1, 1], 
     Hue[0.43880157235413325`, 1, 1], 
     Hue[0.43880157235413325`, 1, 1], 
     Hue[0.4435314967757706, 1, 1], 
     Hue[0.4435314967757706, 1, 1], 
     Hue[0.45299134561904464`, 1, 1], 
     Hue[0.45299134561904464`, 1, 1], 
     Hue[0.455356307829863, 1, 1], 
     Hue[0.455356307829863, 1, 1], 
     Hue[0.4577212700406817, 1, 1], 
     Hue[0.4577212700406817, 1, 1], 
     Hue[0.46245119446231875`, 1, 1], 
     Hue[0.46245119446231875`, 1, 1], 
     Hue[0.46481615667313736`, 1, 1], 
     Hue[0.46481615667313736`, 1, 1], 
     Hue[0.467181118883956, 1, 1], 
     Hue[0.467181118883956, 1, 1], 
     Hue[0.4719110433055931, 1, 1], 
     Hue[0.4719110433055931, 1, 1], 
     Hue[0.47427600551641175`, 1, 1], 
     Hue[0.47427600551641175`, 1, 1], 
     Hue[0.47664096772723014`, 1, 1], 
     Hue[0.47664096772723014`, 1, 1], 
     Hue[0.4813708921488674, 1, 1], 
     Hue[0.4813708921488674, 1, 1], 
     Hue[0.4908307409921415, 1, 1], 
     Hue[0.4908307409921415, 1, 1], 
     Hue[0.49319570320295986`, 1, 1], 
     Hue[0.49319570320295986`, 1, 1], 
     Hue[0.4955606654137785, 1, 1], 
     Hue[0.4955606654137785, 1, 1], 
     Hue[0.5002905898354155, 1, 1], 
     Hue[0.5002905898354155, 1, 1], 
     Hue[0.5026555520462341, 1, 1], 
     Hue[0.5026555520462341, 1, 1], 
     Hue[0.5050205142570529, 1, 1], 
     Hue[0.5050205142570529, 1, 1], 
     Hue[0.50975043867869, 1, 1], 
     Hue[0.50975043867869, 1, 1], 
     Hue[0.5121154008895086, 1, 1], 
     Hue[0.5121154008895086, 1, 1], 
     Hue[0.514480363100327, 1, 1], 
     Hue[0.514480363100327, 1, 1], 
     Hue[0.5192102875219642, 1, 1], 
     Hue[0.5192102875219642, 1, 1], 
     Hue[0.5286701363652384, 1, 1], 
     Hue[0.5286701363652384, 1, 1], 
     Hue[0.5309885742555593, 1, 1], 
     Hue[0.5309885742555593, 1, 1], 
     Hue[0.5333070121458803, 1, 1], 
     Hue[0.5333070121458803, 1, 1], 
     Hue[0.537943887926522, 1, 1], 
     Hue[0.537943887926522, 1, 1], 
     Hue[0.540262325816843, 1, 1], 
     Hue[0.540262325816843, 1, 1], 
     Hue[0.5425807637071637, 1, 1], 
     Hue[0.5425807637071637, 1, 1], 
     Hue[0.5472176394878057, 1, 1], 
     Hue[0.5472176394878057, 1, 1], 
     Hue[0.5495360773781268, 1, 1], 
     Hue[0.5495360773781268, 1, 1], 
     Hue[0.5518545152684478, 1, 1], 
     Hue[0.5518545152684478, 1, 1], 
     Hue[0.5564913910490895, 1, 1], 
     Hue[0.5564913910490895, 1, 1], 
     Hue[0.5657651426103731, 1, 1], 
     Hue[0.5657651426103731, 1, 1], 
     Hue[0.5680835805006943, 1, 1], 
     Hue[0.5680835805006943, 1, 1], 
     Hue[0.5704020183910152, 1, 1], 
     Hue[0.5704020183910152, 1, 1], 
     Hue[0.5750388941716569, 1, 1], 
     Hue[0.5750388941716569, 1, 1], 
     Hue[0.5773573320619779, 1, 1], 
     Hue[0.5773573320619779, 1, 1], 
     Hue[0.5796757699522987, 1, 1], 
     Hue[0.5796757699522987, 1, 1], 
     Hue[0.5843126457329406, 1, 1], 
     Hue[0.5843126457329406, 1, 1], 
     Hue[0.5866310836232617, 1, 1], 
     Hue[0.5866310836232617, 1, 1], 
     Hue[0.5889495215135827, 1, 1], 
     Hue[0.5889495215135827, 1, 1], 
     Hue[0.5935863972942244, 1, 1], 
     Hue[0.5935863972942244, 1, 1], 
     Hue[0.6028601488555081, 1, 1], 
     Hue[0.6028601488555081, 1, 1], 
     Hue[0.6051785867458291, 1, 1], 
     Hue[0.6051785867458291, 1, 1], 
     Hue[0.6074970246361502, 1, 1], 
     Hue[0.6074970246361502, 1, 1], 
     Hue[0.6121339004167918, 1, 1], 
     Hue[0.6121339004167918, 1, 1], 
     Hue[0.6144523383071128, 1, 1], 
     Hue[0.6144523383071128, 1, 1], 
     Hue[0.6167707761974335, 1, 1], 
     Hue[0.6167707761974335, 1, 1], 
     Hue[0.6214076519780756, 1, 1], 
     Hue[0.6214076519780756, 1, 1], 
     Hue[0.6237260898683965, 1, 1], 
     Hue[0.6237260898683965, 1, 1], 
     Hue[0.6260445277587176, 1, 1], 
     Hue[0.6260445277587176, 1, 1], 
     Hue[0.6306814035393593, 1, 1], 
     Hue[0.6306814035393593, 1, 1], 
     Hue[0.639955155100643, 1, 1], 
     Hue[0.639955155100643, 1, 1], 
     Hue[0.642273592990964, 1, 1], 
     Hue[0.642273592990964, 1, 1], 
     Hue[0.644592030881285, 1, 1], 
     Hue[0.644592030881285, 1, 1], 
     Hue[0.6492289066619267, 1, 1], 
     Hue[0.6492289066619267, 1, 1], 
     Hue[0.6585026582232103, 1, 1], 
     Hue[0.6585026582232103, 1, 1], 
     Hue[0.6608210961135315, 1, 1], 
     Hue[0.6608210961135315, 1, 1], 
     Hue[0.6631395340038524, 1, 1], 
     Hue[0.6631395340038524, 1, 1], 
     Hue[0.6677764097844944, 1, 1], 
     Hue[0.6677764097844944, 1, 1], 
     Hue[0.6770501613457782, 1, 1], 
     Hue[0.6770501613457782, 1, 1], 
     Hue[0.6795656794203342, 1, 1], 
     Hue[0.6795656794203342, 1, 1], 
     Hue[0.6820811974948899, 1, 1], 
     Hue[0.6820811974948899, 1, 1], 
     Hue[0.6871122336440013, 1, 1], 
     Hue[0.6871122336440013, 1, 1], 
     Hue[0.6896277517185571, 1, 1], 
     Hue[0.6896277517185571, 1, 1], 
     Hue[0.6921432697931131, 1, 1], 
     Hue[0.6921432697931131, 1, 1], 
     Hue[0.6971743059422248, 1, 1], 
     Hue[0.6971743059422248, 1, 1], 
     Hue[0.6996898240167806, 1, 1], 
     Hue[0.6996898240167806, 1, 1], 
     Hue[0.7022053420913362, 1, 1], 
     Hue[0.7022053420913362, 1, 1], 
     Hue[0.707236378240448, 1, 1], 
     Hue[0.707236378240448, 1, 1], 
     Hue[0.7172984505386711, 1, 1], 
     Hue[0.7172984505386711, 1, 1], 
     Hue[0.7198139686132269, 1, 1], 
     Hue[0.7198139686132269, 1, 1], 
     Hue[0.7223294866877826, 1, 1], 
     Hue[0.7223294866877826, 1, 1], 
     Hue[0.7273605228368943, 1, 1], 
     Hue[0.7273605228368943, 1, 1], 
     Hue[0.7298760409114502, 1, 1], 
     Hue[0.7298760409114502, 1, 1], 
     Hue[0.732391558986006, 1, 1], 
     Hue[0.732391558986006, 1, 1], 
     Hue[0.7374225951351177, 1, 1], 
     Hue[0.7374225951351177, 1, 1], 
     Hue[0.7399381132096735, 1, 1], 
     Hue[0.7399381132096735, 1, 1], 
     Hue[0.7424536312842293, 1, 1], 
     Hue[0.7424536312842293, 1, 1], 
     Hue[0.747484667433341, 1, 1], 
     Hue[0.747484667433341, 1, 1], 
     Hue[0.7575467397315642, 1, 1], 
     Hue[0.7575467397315642, 1, 1], 
     Hue[0.7600622578061199, 1, 1], 
     Hue[0.7600622578061199, 1, 1], 
     Hue[0.7625777758806758, 1, 1], 
     Hue[0.7625777758806758, 1, 1], 
     Hue[0.7676088120297874, 1, 1], 
     Hue[0.7676088120297874, 1, 1], 
     Hue[0.7776708843280107, 1, 1], 
     Hue[0.7776708843280107, 1, 1], 
     Hue[0.7801864024025665, 1, 1], 
     Hue[0.7801864024025665, 1, 1], 
     Hue[0.7827019204771223, 1, 1], 
     Hue[0.7827019204771223, 1, 1], 
     Hue[0.787732956626234, 1, 1], 
     Hue[0.787732956626234, 1, 1], 
     Hue[0.7977950289244572, 1, 1], 
     Hue[0.7977950289244572, 1, 1], 
     Hue[0.800310546999013, 1, 1], 
     Hue[0.800310546999013, 1, 1], 
     Hue[0.8028260650735689, 1, 1], 
     Hue[0.8028260650735689, 1, 1], 
     Hue[0.8078571012226805, 1, 1], 
     Hue[0.8078571012226805, 1, 1], 
     Hue[0.8179191735209038, 1, 1], 
     Hue[0.8179191735209038, 1, 1], 
     Hue[0.8204346915954596, 1, 1], 
     Hue[0.8204346915954596, 1, 1], 
     Hue[0.8229502096700154, 1, 1], 
     Hue[0.8229502096700154, 1, 1], 
     Hue[0.827981245819127, 1, 1], 
     Hue[0.827981245819127, 1, 1], 
     Hue[0.8380433181173503, 1, 1], 
     Hue[0.8380433181173503, 1, 1], 
     Hue[0.8403905096190423, 1, 1], 
     Hue[0.8403905096190423, 1, 1], 
     Hue[0.8427377011207342, 1, 1], 
     Hue[0.8427377011207342, 1, 1], 
     Hue[0.8474320841241183, 1, 1], 
     Hue[0.8474320841241183, 1, 1], 
     Hue[0.856820850130886, 1, 1], 
     Hue[0.856820850130886, 1, 1], 
     Hue[0.8591680416325781, 1, 1], 
     Hue[0.8591680416325781, 1, 1], 
     Hue[0.8615152331342699, 1, 1], 
     Hue[0.8615152331342699, 1, 1], 
     Hue[0.8662096161376538, 1, 1], 
     Hue[0.8662096161376538, 1, 1], 
     Hue[0.8755983821444216, 1, 1], 
     Hue[0.8755983821444216, 1, 1], 
     Hue[0.8779455736461137, 1, 1], 
     Hue[0.8779455736461137, 1, 1], 
     Hue[0.8802927651478055, 1, 1], 
     Hue[0.8802927651478055, 1, 1], 
     Hue[0.8849871481511894, 1, 1], 
     Hue[0.8849871481511894, 1, 1], 
     Hue[0.8943759141579573, 1, 1], 
     Hue[0.8943759141579573, 1, 1], 
     Hue[0.8967231056596492, 1, 1], 
     Hue[0.8967231056596492, 1, 1], 
     Hue[0.8990702971613411, 1, 1], 
     Hue[0.8990702971613411, 1, 1], 
     Hue[0.903764680164725, 1, 1], 
     Hue[0.903764680164725, 1, 1], 
     Hue[0.9131534461714927, 1, 1], 
     Hue[0.9131534461714927, 1, 1], 
     Hue[0.9155006376731848, 1, 1], 
     Hue[0.9155006376731848, 1, 1], 
     Hue[0.9178478291748767, 1, 1], 
     Hue[0.9178478291748767, 1, 1], 
     Hue[0.9225422121782606, 1, 1], 
     Hue[0.9225422121782606, 1, 1], 
     Hue[0.9319309781850283, 1, 1], 
     Hue[0.9319309781850283, 1, 1], 
     Hue[0.9342781696867204, 1, 1], 
     Hue[0.9342781696867204, 1, 1], 
     Hue[0.9366253611884122, 1, 1], 
     Hue[0.9366253611884122, 1, 1], 
     Hue[0.9413197441917962, 1, 1], 
     Hue[0.9413197441917962, 1, 1], 
     Hue[0.9507085101985641, 1, 1], 
     Hue[0.9507085101985641, 1, 1], 
     Hue[0.953055701700256, 1, 1], 
     Hue[0.953055701700256, 1, 1], 
     Hue[0.9554028932019479, 1, 1], 
     Hue[0.9554028932019479, 1, 1], 
     Hue[0.960097276205332, 1, 1], 
     Hue[0.960097276205332, 1, 1], 
     Hue[0.9694860422120998, 1, 1], 
     Hue[0.9694860422120998, 1, 1], 
     Hue[0.9882635742256354, 1, 1], 
     Hue[0.9882635742256354, 1, 1], 
     Hue[0.9908078459115621, 1, 1], 
     Hue[0.9908078459115621, 1, 1], 
     Hue[0.993352117597489, 1, 1], 
     Hue[0.993352117597489, 1, 1], 
     Hue[0.9984406609693426, 1, 1], 
     Hue[0.9984406609693426, 1, 1], 
     Hue[0.00861774771305003, 1, 1], 
     Hue[0.00861774771305003, 1, 1], 
     Hue[0.011162019398976914`, 1, 1], 
     Hue[0.011162019398976914`, 1, 1], 
     Hue[0.013706291084903658`, 1, 1], 
     Hue[0.013706291084903658`, 1, 1], 
     Hue[0.018794834456757423`, 1, 1], 
     Hue[0.018794834456757423`, 1, 1], 
     Hue[0.02897192120046467, 1, 1], 
     Hue[0.02897192120046467, 1, 1], 
     Hue[0.03151619288639141, 1, 1], 
     Hue[0.03151619288639141, 1, 1], 
     Hue[0.034060464572318296`, 1, 1], 
     Hue[0.034060464572318296`, 1, 1], 
     Hue[0.03914900794417192, 1, 1], 
     Hue[0.03914900794417192, 1, 1], 
     Hue[0.04932609468787931, 1, 1], 
     Hue[0.04932609468787931, 1, 1], 
     Hue[0.06968026817529394, 1, 1], 
     Hue[0.06968026817529394, 1, 1], 
     Hue[0.07222453986122068, 1, 1], 
     Hue[0.07222453986122068, 1, 1], 
     Hue[0.07476881154714757, 1, 1], 
     Hue[0.07476881154714757, 1, 1], 
     Hue[0.07985735491900121, 1, 1], 
     Hue[0.07985735491900121, 1, 1], 
     Hue[0.09003444166270858, 1, 1], 
     Hue[0.09003444166270858, 1, 1], 
     Hue[0.09257871334863547, 1, 1], 
     Hue[0.09257871334863547, 1, 1], 
     Hue[0.09512298503456218, 1, 1], 
     Hue[0.09512298503456218, 1, 1], 
     Hue[0.10021152840641596`, 1, 1], 
     Hue[0.10021152840641596`, 1, 1], 
     Hue[0.11038861515012321`, 1, 1], 
     Hue[0.11038861515012321`, 1, 1], 
     Hue[0.1129328868360501, 1, 1], 
     Hue[0.1129328868360501, 1, 1], 
     Hue[0.115477158521977, 1, 1], 
     Hue[0.115477158521977, 1, 1], 
     Hue[0.1205657018938306, 1, 1], 
     Hue[0.1205657018938306, 1, 1], 
     Hue[0.130742788637538, 1, 1], 
     Hue[0.130742788637538, 1, 1], 
     Hue[0.15109696212495263`, 1, 1], 
     Hue[0.15109696212495263`, 1, 1], 
     Hue[0.15359470949038187`, 1, 1], 
     Hue[0.15359470949038187`, 1, 1], 
     Hue[0.15609245685581094`, 1, 1], 
     Hue[0.15609245685581094`, 1, 1], 
     Hue[0.1610879515866694, 1, 1], 
     Hue[0.1610879515866694, 1, 1], 
     Hue[0.1710789410483863, 1, 1], 
     Hue[0.1710789410483863, 1, 1], 
     Hue[0.17357668841381552`, 1, 1], 
     Hue[0.17357668841381552`, 1, 1], 
     Hue[0.17607443577924475`, 1, 1], 
     Hue[0.17607443577924475`, 1, 1], 
     Hue[0.1810699305101032, 1, 1], 
     Hue[0.1810699305101032, 1, 1], 
     Hue[0.1910609199718201, 1, 1], 
     Hue[0.1910609199718201, 1, 1], 
     Hue[0.19355866733724933`, 1, 1], 
     Hue[0.19355866733724933`, 1, 1], 
     Hue[0.19605641470267854`, 1, 1], 
     Hue[0.19605641470267854`, 1, 1], 
     Hue[0.20105190943353698`, 1, 1], 
     Hue[0.20105190943353698`, 1, 1], 
     Hue[0.2110428988952539, 1, 1], 
     Hue[0.2110428988952539, 1, 1], 
     Hue[0.2310248778186877, 1, 1], 
     Hue[0.2310248778186877, 1, 1], 
     Hue[0.23352262518411693`, 1, 1], 
     Hue[0.23352262518411693`, 1, 1], 
     Hue[0.23602037254954614`, 1, 1], 
     Hue[0.23602037254954614`, 1, 1], 
     Hue[0.2410158672804046, 1, 1], 
     Hue[0.2410158672804046, 1, 1], 
     Hue[0.2510068567421215, 1, 1], 
     Hue[0.2510068567421215, 1, 1], 
     Hue[0.27098883566555515`, 1, 1], 
     Hue[0.27098883566555515`, 1, 1], 
     Hue[0.2734865830309844, 1, 1], 
     Hue[0.2734865830309844, 1, 1], 
     Hue[0.27598433039641346`, 1, 1], 
     Hue[0.27598433039641346`, 1, 1], 
     Hue[0.28097982512727193`, 1, 1], 
     Hue[0.28097982512727193`, 1, 1], 
     Hue[0.2909708145889888, 1, 1], 
     Hue[0.2909708145889888, 1, 1], 
     Hue[0.3109527935124226, 1, 1], 
     Hue[0.3109527935124226, 1, 1], 
     Hue[0.313282214304988, 1, 1], 
     Hue[0.313282214304988, 1, 1], 
     Hue[0.31561163509755336`, 1, 1], 
     Hue[0.31561163509755336`, 1, 1], 
     Hue[0.32027047668268416`, 1, 1], 
     Hue[0.32027047668268416`, 1, 1], 
     Hue[0.3295881598529455, 1, 1], 
     Hue[0.3295881598529455, 1, 1], 
     Hue[0.3482235261934684, 1, 1], 
     Hue[0.3482235261934684, 1, 1], 
     Hue[0.3505529469860338, 1, 1], 
     Hue[0.3505529469860338, 1, 1], 
     Hue[0.35288236777859916`, 1, 1], 
     Hue[0.35288236777859916`, 1, 1], 
     Hue[0.3575412093637299, 1, 1], 
     Hue[0.3575412093637299, 1, 1], 
     Hue[0.3668588925339913, 1, 1], 
     Hue[0.3668588925339913, 1, 1], 
     Hue[0.38549425887451416`, 1, 1], 
     Hue[0.38549425887451416`, 1, 1], 
     Hue[0.3878236796670795, 1, 1], 
     Hue[0.3878236796670795, 1, 1], 
     Hue[0.39015310045964485`, 1, 1], 
     Hue[0.39015310045964485`, 1, 1], 
     Hue[0.3948119420447756, 1, 1], 
     Hue[0.3948119420447756, 1, 1], 
     Hue[0.40412962521503704`, 1, 1], 
     Hue[0.40412962521503704`, 1, 1], 
     Hue[0.4227649915555599, 1, 1], 
     Hue[0.4227649915555599, 1, 1], 
     Hue[0.46003572423660566`, 1, 1], 
     Hue[0.46003572423660566`, 1, 1], 
     Hue[0.4625622252134059, 1, 1], 
     Hue[0.4625622252134059, 1, 1], 
     Hue[0.46508872619020614`, 1, 1], 
     Hue[0.46508872619020614`, 1, 1], 
     Hue[0.4701417281438066, 1, 1], 
     Hue[0.4701417281438066, 1, 1], 
     Hue[0.48024773205100757`, 1, 1], 
     Hue[0.48024773205100757`, 1, 1], 
     Hue[0.5004597398654094, 1, 1], 
     Hue[0.5004597398654094, 1, 1], 
     Hue[0.5408837554942132, 1, 1], 
     Hue[0.5408837554942132, 1, 1], 
     Hue[0.5434102564710135, 1, 1], 
     Hue[0.5434102564710135, 1, 1], 
     Hue[0.5459367574478137, 1, 1], 
     Hue[0.5459367574478137, 1, 1], 
     Hue[0.5509897594014141, 1, 1], 
     Hue[0.5509897594014141, 1, 1], 
     Hue[0.5610957633086151, 1, 1], 
     Hue[0.5610957633086151, 1, 1], 
     Hue[0.581307771123017, 1, 1], 
     Hue[0.581307771123017, 1, 1], 
     Hue[0.6217317867518208, 1, 1], 
     Hue[0.6217317867518208, 1, 1], 
     Hue[0.6240899611557571, 1, 1], 
     Hue[0.6240899611557571, 1, 1], 
     Hue[0.6264481355596936, 1, 1], 
     Hue[0.6264481355596936, 1, 1], 
     Hue[0.6311644843675663, 1, 1], 
     Hue[0.6311644843675663, 1, 1], 
     Hue[0.6405971819833118, 1, 1], 
     Hue[0.6405971819833118, 1, 1], 
     Hue[0.6594625772148028, 1, 1], 
     Hue[0.6594625772148028, 1, 1], 
     Hue[0.6971933676777848, 1, 1], 
     Hue[0.6971933676777848, 1, 1], 
     Hue[0.6995515420817213, 1, 1], 
     Hue[0.6995515420817213, 1, 1], 
     Hue[0.7019097164856576, 1, 1], 
     Hue[0.7019097164856576, 1, 1], 
     Hue[0.7066260652935304, 1, 1], 
     Hue[0.7066260652935304, 1, 1], 
     Hue[0.716058762909276, 1, 1], 
     Hue[0.716058762909276, 1, 1], 
     Hue[0.734924158140767, 1, 1], 
     Hue[0.734924158140767, 1, 1], 
     Hue[0.772654948603749, 1, 1], 
     Hue[0.772654948603749, 1, 1], 
     Hue[0.7752102031919202, 1, 1], 
     Hue[0.7752102031919202, 1, 1], 
     Hue[0.7777654577800914, 1, 1], 
     Hue[0.7777654577800914, 1, 1], 
     Hue[0.782875966956434, 1, 1], 
     Hue[0.782875966956434, 1, 1], 
     Hue[0.793096985309119, 1, 1], 
     Hue[0.793096985309119, 1, 1], 
     Hue[0.8135390220144891, 1, 1], 
     Hue[0.8135390220144891, 1, 1], 
     Hue[0.8160942766026603, 1, 1], 
     Hue[0.8160942766026603, 1, 1], 
     Hue[0.8186495311908315, 1, 1], 
     Hue[0.8186495311908315, 1, 1], 
     Hue[0.8237600403671741, 1, 1], 
     Hue[0.8237600403671741, 1, 1], 
     Hue[0.8339810587198591, 1, 1], 
     Hue[0.8339810587198591, 1, 1], 
     Hue[0.8544230954252292, 1, 1], 
     Hue[0.8544230954252292, 1, 1], 
     Hue[0.8569783500134004, 1, 1], 
     Hue[0.8569783500134004, 1, 1], 
     Hue[0.8595336046015717, 1, 1], 
     Hue[0.8595336046015717, 1, 1], 
     Hue[0.8646441137779142, 1, 1], 
     Hue[0.8646441137779142, 1, 1], 
     Hue[0.8748651321305992, 1, 1], 
     Hue[0.8748651321305992, 1, 1], 
     Hue[0.8774203867187704, 1, 1], 
     Hue[0.8774203867187704, 1, 1], 
     Hue[0.8799756413069417, 1, 1], 
     Hue[0.8799756413069417, 1, 1], 
     Hue[0.8850861504832842, 1, 1], 
     Hue[0.8850861504832842, 1, 1], 
     Hue[0.8953071688359693, 1, 1], 
     Hue[0.8953071688359693, 1, 1], 
     Hue[0.8978624234241405, 1, 1], 
     Hue[0.8978624234241405, 1, 1], 
     Hue[0.9004176780123118, 1, 1], 
     Hue[0.9004176780123118, 1, 1], 
     Hue[0.9055281871886542, 1, 1], 
     Hue[0.9055281871886542, 1, 1], 
     Hue[0.9157492055413393, 1, 1], 
     Hue[0.9157492055413393, 1, 1], 
     Hue[0.9183044601295106, 1, 1], 
     Hue[0.9183044601295106, 1, 1], 
     Hue[0.9208597147176818, 1, 1], 
     Hue[0.9208597147176818, 1, 1], 
     Hue[0.9259702238940244, 1, 1], 
     Hue[0.9259702238940244, 1, 1], 
     Hue[0.9361912422467094, 1, 1], 
     Hue[0.9361912422467094, 1, 1], 
     Hue[0.938699972514383, 1, 1], 
     Hue[0.938699972514383, 1, 1], 
     Hue[0.9412087027820567, 1, 1], 
     Hue[0.9412087027820567, 1, 1], 
     Hue[0.946226163317404, 1, 1], 
     Hue[0.946226163317404, 1, 1], 
     Hue[0.9487348935850776, 1, 1], 
     Hue[0.9487348935850776, 1, 1], 
     Hue[0.9512436238527512, 1, 1], 
     Hue[0.9512436238527512, 1, 1], 
     Hue[0.9562610843880985, 1, 1], 
     Hue[0.9562610843880985, 1, 1], 
     Hue[0.9587698146557722, 1, 1], 
     Hue[0.9587698146557722, 1, 1], 
     Hue[0.9612785449234458, 1, 1], 
     Hue[0.9612785449234458, 1, 1], 
     Hue[0.9637872751911194, 1, 1], 
     Hue[0.9637872751911194, 1, 1], 
     Hue[0.9662960054587931, 1, 1], 
     Hue[0.9662960054587931, 1, 1], 
     Hue[0.9688047357264667, 1, 1], 
     Hue[0.9688047357264667, 1, 1], 
     Hue[0.9713134659941404, 1, 1], 
     Hue[0.9713134659941404, 1, 1], 
     Hue[0.973822196261814, 1, 1], 
     Hue[0.973822196261814, 1, 1], 
     Hue[0.9763309265294876, 1, 1], 
     Hue[0.9763309265294876, 1, 1], 
     Hue[0.9788396567971613, 1, 1], 
     Hue[0.9788396567971613, 1, 1], 
     Hue[0.981348387064835, 1, 1], 
     Hue[0.981348387064835, 1, 1], 
     Hue[0.9838571173325086, 1, 1], 
     Hue[0.9838571173325086, 1, 1], 
     Hue[0.9863658476001822, 1, 1], 
     Hue[0.9863658476001822, 1, 1], 
     Hue[0.9888745778678558, 1, 1], 
     Hue[0.9888745778678558, 1, 1], 
     Hue[0.9913833081355296, 1, 1], 
     Hue[0.9913833081355296, 1, 1], 
     Hue[0.9938920384032032, 1, 1], 
     Hue[0.9938920384032032, 1, 1], 
     Hue[0.9964007686708768, 1, 1], 
     Hue[0.9964007686708768, 1, 1], 
     Hue[0.9989094989385504, 1, 1], 
     Hue[0.9989094989385504, 1, 1], 
     Hue[0.0014182292062240736`, 1, 1], 
     Hue[0.0014182292062240736`, 1, 1], 
     Hue[0.003926959473897716, 1, 1], 
     Hue[0.003926959473897716, 1, 1], 
     Hue[0.0064356897415713595`, 1, 1], 
     Hue[0.0064356897415713595`, 1, 1], 
     Hue[0.008944420009244999, 1, 1], 
     Hue[0.008944420009244999, 1, 1], 
     Hue[0.011453150276918643`, 1, 1], 
     Hue[0.011453150276918643`, 1, 1], 
     Hue[0.013961880544592286`, 1, 1], 
     Hue[0.013961880544592286`, 1, 1], 
     Hue[0.016470610812265928`, 1, 1], 
     Hue[0.016470610812265928`, 1, 1], 
     Hue[0.018979341079939573`, 1, 1], 
     Hue[0.018979341079939573`, 1, 1], 
     Hue[0.021488071347613213`, 1, 1], 
     Hue[0.021488071347613213`, 1, 1], 
     Hue[0.023996801615286858`, 1, 1], 
     Hue[0.023996801615286858`, 1, 1], 
     Hue[0.026505531882960495`, 1, 1], 
     Hue[0.026505531882960495`, 1, 1], 
     Hue[0.02901426215063414, 1, 1], 
     Hue[0.02901426215063414, 1, 1], 
     Hue[0.031522992418307784`, 1, 1], 
     Hue[0.031522992418307784`, 1, 1], 
     Hue[0.03654045295365507, 1, 1], 
     Hue[0.03654045295365507, 1, 1], 
     Hue[0.0390491832213287, 1, 1], 
     Hue[0.0390491832213287, 1, 1], 
     Hue[0.041557913489002354`, 1, 1], 
     Hue[0.041557913489002354`, 1, 1], 
     Hue[0.04657537402434964, 1, 1], 
     Hue[0.04657537402434964, 1, 1], 
     Hue[0.04908410429202328, 1, 1], 
     Hue[0.04908410429202328, 1, 1], 
     Hue[0.05159283455969691, 1, 1], 
     Hue[0.05159283455969691, 1, 1], 
     Hue[0.05661029509504419, 1, 1], 
     Hue[0.05661029509504419, 1, 1], 
     Hue[0.05911902536271785, 1, 1], 
     Hue[0.05911902536271785, 1, 1], 
     Hue[0.061627755630391495`, 1, 1], 
     Hue[0.061627755630391495`, 1, 1], 
     Hue[0.06664521616573878, 1, 1], 
     Hue[0.06664521616573878, 1, 1], 
     Hue[0.06915394643341242, 1, 1], 
     Hue[0.06915394643341242, 1, 1], 
     Hue[0.07166267670108606, 1, 1], 
     Hue[0.07166267670108606, 1, 1], 
     Hue[0.07668013723643334, 1, 1], 
     Hue[0.07668013723643334, 1, 1], 
     Hue[0.07918886750410699, 1, 1], 
     Hue[0.07918886750410699, 1, 1], 
     Hue[0.08169759777178064, 1, 1], 
     Hue[0.08169759777178064, 1, 1], 
     Hue[0.08671505830712793, 1, 1], 
     Hue[0.08671505830712793, 1, 1], 
     Hue[0.0967499793778225, 1, 1], 
     Hue[0.0967499793778225, 1, 1], 
     Hue[0.09909038307263227, 1, 1], 
     Hue[0.09909038307263227, 1, 1], 
     Hue[0.10143078676744204`, 1, 1], 
     Hue[0.10143078676744204`, 1, 1], 
     Hue[0.10611159415706164`, 1, 1], 
     Hue[0.10611159415706164`, 1, 1], 
     Hue[0.11547320893630074`, 1, 1], 
     Hue[0.11547320893630074`, 1, 1], 
     Hue[0.134196438494779, 1, 1], 
     Hue[0.134196438494779, 1, 1], 
     Hue[0.1365368421895888, 1, 1], 
     Hue[0.1365368421895888, 1, 1], 
     Hue[0.13887724588439856`, 1, 1], 
     Hue[0.13887724588439856`, 1, 1], 
     Hue[0.14355805327401813`, 1, 1], 
     Hue[0.14355805327401813`, 1, 1], 
     Hue[0.15291966805325727`, 1, 1], 
     Hue[0.15291966805325727`, 1, 1], 
     Hue[0.1716428976117355, 1, 1], 
     Hue[0.1716428976117355, 1, 1], 
     Hue[0.1739833013065453, 1, 1], 
     Hue[0.1739833013065453, 1, 1], 
     Hue[0.17632370500135508`, 1, 1], 
     Hue[0.17632370500135508`, 1, 1], 
     Hue[0.18100451239097462`, 1, 1], 
     Hue[0.18100451239097462`, 1, 1], 
     Hue[0.19036612717021376`, 1, 1], 
     Hue[0.19036612717021376`, 1, 1], 
     Hue[0.20908935672869203`, 1, 1], 
     Hue[0.20908935672869203`, 1, 1], 
     Hue[0.24653581584564851`, 1, 1], 
     Hue[0.24653581584564851`, 1, 1], 
     Hue[0.24907329972469316`, 1, 1], 
     Hue[0.24907329972469316`, 1, 1], 
     Hue[0.25161078360373784`, 1, 1], 
     Hue[0.25161078360373784`, 1, 1], 
     Hue[0.25668575136182714`, 1, 1], 
     Hue[0.25668575136182714`, 1, 1], 
     Hue[0.2668356868780058, 1, 1], 
     Hue[0.2668356868780058, 1, 1], 
     Hue[0.2871355579103631, 1, 1], 
     Hue[0.2871355579103631, 1, 1], 
     Hue[0.32773529997507767`, 1, 1], 
     Hue[0.32773529997507767`, 1, 1], 
     Hue[0.3302727838541223, 1, 1], 
     Hue[0.3302727838541223, 1, 1], 
     Hue[0.332810267733167, 1, 1], 
     Hue[0.332810267733167, 1, 1], 
     Hue[0.3378852354912563, 1, 1], 
     Hue[0.3378852354912563, 1, 1], 
     Hue[0.348035171007435, 1, 1], 
     Hue[0.348035171007435, 1, 1], 
     Hue[0.3683350420397923, 1, 1], 
     Hue[0.3683350420397923, 1, 1], 
     Hue[0.40893478410450684`, 1, 1], 
     Hue[0.40893478410450684`, 1, 1], 
     Hue[0.41130394141068766`, 1, 1], 
     Hue[0.41130394141068766`, 1, 1], 
     Hue[0.41367309871686847`, 1, 1], 
     Hue[0.41367309871686847`, 1, 1], 
     Hue[0.4184114133292301, 1, 1], 
     Hue[0.4184114133292301, 1, 1], 
     Hue[0.4278880425539533, 1, 1], 
     Hue[0.4278880425539533, 1, 1], 
     Hue[0.4468413010033997, 1, 1], 
     Hue[0.4468413010033997, 1, 1], 
     Hue[0.4847478179022925, 1, 1], 
     Hue[0.4847478179022925, 1, 1], 
     Hue[0.48711697520847325`, 1, 1], 
     Hue[0.48711697520847325`, 1, 1], 
     Hue[0.48948613251465406`, 1, 1], 
     Hue[0.48948613251465406`, 1, 1], 
     Hue[0.4942244471270157, 1, 1], 
     Hue[0.4942244471270157, 1, 1], 
     Hue[0.5037010763517389, 1, 1], 
     Hue[0.5037010763517389, 1, 1], 
     Hue[0.5226543348011852, 1, 1], 
     Hue[0.5226543348011852, 1, 1], 
     Hue[0.5605608517000781, 1, 1], 
     Hue[0.5605608517000781, 1, 1], 
     Hue[0.5628834846857612, 1, 1], 
     Hue[0.5628834846857612, 1, 1], 
     Hue[0.5652061176714445, 1, 1], 
     Hue[0.5652061176714445, 1, 1], 
     Hue[0.5698513836428108, 1, 1], 
     Hue[0.5698513836428108, 1, 1], 
     Hue[0.5791419155855435, 1, 1], 
     Hue[0.5791419155855435, 1, 1], 
     Hue[0.597722979471009, 1, 1], 
     Hue[0.597722979471009, 1, 1], 
     Hue[0.6000456124566923, 1, 1], 
     Hue[0.6000456124566923, 1, 1], 
     Hue[0.6023682454423754, 1, 1], 
     Hue[0.6023682454423754, 1, 1], 
     Hue[0.6070135114137418, 1, 1], 
     Hue[0.6070135114137418, 1, 1], 
     Hue[0.6163040433564745, 1, 1], 
     Hue[0.6163040433564745, 1, 1], 
     Hue[0.63488510724194, 1, 1], 
     Hue[0.63488510724194, 1, 1], 
     Hue[0.6372077402276233, 1, 1], 
     Hue[0.6372077402276233, 1, 1], 
     Hue[0.6395303732133064, 1, 1], 
     Hue[0.6395303732133064, 1, 1], 
     Hue[0.6441756391846729, 1, 1], 
     Hue[0.6441756391846729, 1, 1], 
     Hue[0.6534661711274056, 1, 1], 
     Hue[0.6534661711274056, 1, 1], 
     Hue[0.672047235012871, 1, 1], 
     Hue[0.672047235012871, 1, 1], 
     Hue[0.6743698679985541, 1, 1], 
     Hue[0.6743698679985541, 1, 1], 
     Hue[0.6766925009842375, 1, 1], 
     Hue[0.6766925009842375, 1, 1], 
     Hue[0.6813377669556038, 1, 1], 
     Hue[0.6813377669556038, 1, 1], 
     Hue[0.6906282988983365, 1, 1], 
     Hue[0.6906282988983365, 1, 1], 
     Hue[0.709209362783802, 1, 1], 
     Hue[0.709209362783802, 1, 1], 
     Hue[0.71172907595372, 1, 1], 
     Hue[0.71172907595372, 1, 1], 
     Hue[0.7142487891236381, 1, 1], 
     Hue[0.7142487891236381, 1, 1], 
     Hue[0.7192882154634741, 1, 1], 
     Hue[0.7192882154634741, 1, 1], 
     Hue[0.7293670681431466, 1, 1], 
     Hue[0.7293670681431466, 1, 1], 
     Hue[0.7318867813130645, 1, 1], 
     Hue[0.7318867813130645, 1, 1], 
     Hue[0.7344064944829827, 1, 1], 
     Hue[0.7344064944829827, 1, 1], 
     Hue[0.7394459208228188, 1, 1], 
     Hue[0.7394459208228188, 1, 1], 
     Hue[0.749524773502491, 1, 1], 
     Hue[0.749524773502491, 1, 1], 
     Hue[0.7520444866724092, 1, 1], 
     Hue[0.7520444866724092, 1, 1], 
     Hue[0.7545641998423271, 1, 1], 
     Hue[0.7545641998423271, 1, 1], 
     Hue[0.7596036261821633, 1, 1], 
     Hue[0.7596036261821633, 1, 1], 
     Hue[0.7696824788618355, 1, 1], 
     Hue[0.7696824788618355, 1, 1], 
     Hue[0.7898401842211801, 1, 1], 
     Hue[0.7898401842211801, 1, 1], 
     Hue[0.7923598973910981, 1, 1], 
     Hue[0.7923598973910981, 1, 1], 
     Hue[0.7948796105610161, 1, 1], 
     Hue[0.7948796105610161, 1, 1], 
     Hue[0.7999190369008522, 1, 1], 
     Hue[0.7999190369008522, 1, 1], 
     Hue[0.8099978895805245, 1, 1], 
     Hue[0.8099978895805245, 1, 1], 
     Hue[0.8125176027504426, 1, 1], 
     Hue[0.8125176027504426, 1, 1], 
     Hue[0.8150373159203607, 1, 1], 
     Hue[0.8150373159203607, 1, 1], 
     Hue[0.8200767422601969, 1, 1], 
     Hue[0.8200767422601969, 1, 1], 
     Hue[0.8301555949398691, 1, 1], 
     Hue[0.8301555949398691, 1, 1], 
     Hue[0.8326753081097872, 1, 1], 
     Hue[0.8326753081097872, 1, 1], 
     Hue[0.8351950212797054, 1, 1], 
     Hue[0.8351950212797054, 1, 1], 
     Hue[0.8402344476195414, 1, 1], 
     Hue[0.8402344476195414, 1, 1], 
     Hue[0.8503133002992137, 1, 1], 
     Hue[0.8503133002992137, 1, 1], 
     Hue[0.8704710056585583, 1, 1], 
     Hue[0.8704710056585583, 1, 1], 
     Hue[0.8728223922556125, 1, 1], 
     Hue[0.8728223922556125, 1, 1], 
     Hue[0.8751737788526667, 1, 1], 
     Hue[0.8751737788526667, 1, 1], 
     Hue[0.8798765520467752, 1, 1], 
     Hue[0.8798765520467752, 1, 1], 
     Hue[0.8892820984349921, 1, 1], 
     Hue[0.8892820984349921, 1, 1], 
     Hue[0.8916334850320461, 1, 1], 
     Hue[0.8916334850320461, 1, 1], 
     Hue[0.8939848716291003, 1, 1], 
     Hue[0.8939848716291003, 1, 1], 
     Hue[0.8986876448232088, 1, 1], 
     Hue[0.8986876448232088, 1, 1], 
     Hue[0.9080931912114256, 1, 1], 
     Hue[0.9080931912114256, 1, 1], 
     Hue[0.9104445778084798, 1, 1], 
     Hue[0.9104445778084798, 1, 1], 
     Hue[0.912795964405534, 1, 1], 
     Hue[0.912795964405534, 1, 1], 
     Hue[0.9174987375996424, 1, 1], 
     Hue[0.9174987375996424, 1, 1], 
     Hue[0.9269042839878592, 1, 1], 
     Hue[0.9269042839878592, 1, 1], 
     Hue[0.9457153767642927, 1, 1], 
     Hue[0.9457153767642927, 1, 1], 
     Hue[0.9480667633613469, 1, 1], 
     Hue[0.9480667633613469, 1, 1], 
     Hue[0.9504181499584011, 1, 1], 
     Hue[0.9504181499584011, 1, 1], 
     Hue[0.9551209231525096, 1, 1], 
     Hue[0.9551209231525096, 1, 1], 
     Hue[0.9645264695407264, 1, 1], 
     Hue[0.9645264695407264, 1, 1], 
     Hue[0.9668778561377805, 1, 1], 
     Hue[0.9668778561377805, 1, 1], 
     Hue[0.9692292427348348, 1, 1], 
     Hue[0.9692292427348348, 1, 1], 
     Hue[0.9739320159289432, 1, 1], 
     Hue[0.9739320159289432, 1, 1], 
     Hue[0.9833375623171601, 1, 1], 
     Hue[0.9833375623171601, 1, 1], 
     Hue[0.9856889489142143, 1, 1], 
     Hue[0.9856889489142143, 1, 1], 
     Hue[0.9880403355112685, 1, 1], 
     Hue[0.9880403355112685, 1, 1], 
     Hue[0.9927431087053771, 1, 1], 
     Hue[0.9927431087053771, 1, 1], 
     Hue[0.0021486550935938373`, 1, 1], 
     Hue[0.0021486550935938373`, 1, 1], 
     Hue[0.004500041690647965, 1, 1], 
     Hue[0.004500041690647965, 1, 1], 
     Hue[0.006851428287702235, 1, 1], 
     Hue[0.006851428287702235, 1, 1], 
     Hue[0.011554201481810633`, 1, 1], 
     Hue[0.011554201481810633`, 1, 1], 
     Hue[0.020959747870027433`, 1, 1], 
     Hue[0.020959747870027433`, 1, 1], 
     Hue[0.02350821465131656, 1, 1], 
     Hue[0.02350821465131656, 1, 1], 
     Hue[0.02605668143260555, 1, 1], 
     Hue[0.02605668143260555, 1, 1], 
     Hue[0.031153614995183685`, 1, 1], 
     Hue[0.031153614995183685`, 1, 1], 
     Hue[0.041347482120340076`, 1, 1], 
     Hue[0.041347482120340076`, 1, 1], 
     Hue[0.04389594890162907, 1, 1], 
     Hue[0.04389594890162907, 1, 1], 
     Hue[0.04644441568291821, 1, 1], 
     Hue[0.04644441568291821, 1, 1], 
     Hue[0.051541349245496464`, 1, 1], 
     Hue[0.051541349245496464`, 1, 1], 
     Hue[0.06173521637065272, 1, 1], 
     Hue[0.06173521637065272, 1, 1], 
     Hue[0.06428368315194186, 1, 1], 
     Hue[0.06428368315194186, 1, 1], 
     Hue[0.066832149933231, 1, 1], 
     Hue[0.066832149933231, 1, 1], 
     Hue[0.07192908349580913, 1, 1], 
     Hue[0.07192908349580913, 1, 1], 
     Hue[0.08212295062096552, 1, 1], 
     Hue[0.08212295062096552, 1, 1], 
     Hue[0.08467141740225466, 1, 1], 
     Hue[0.08467141740225466, 1, 1], 
     Hue[0.08721988418354365, 1, 1], 
     Hue[0.08721988418354365, 1, 1], 
     Hue[0.09231681774612178, 1, 1], 
     Hue[0.09231681774612178, 1, 1], 
     Hue[0.10251068487127818`, 1, 1], 
     Hue[0.10251068487127818`, 1, 1], 
     Hue[0.10505915165256717`, 1, 1], 
     Hue[0.10505915165256717`, 1, 1], 
     Hue[0.10760761843385629`, 1, 1], 
     Hue[0.10760761843385629`, 1, 1], 
     Hue[0.11270455199643457`, 1, 1], 
     Hue[0.11270455199643457`, 1, 1], 
     Hue[0.12289841912159083`, 1, 1], 
     Hue[0.12289841912159083`, 1, 1], 
     Hue[0.12544688590287997`, 1, 1], 
     Hue[0.12544688590287997`, 1, 1], 
     Hue[0.1279953526841691, 1, 1], 
     Hue[0.1279953526841691, 1, 1], 
     Hue[0.13309228624674724`, 1, 1], 
     Hue[0.13309228624674724`, 1, 1], 
     Hue[0.14328615337190362`, 1, 1], 
     Hue[0.14328615337190362`, 1, 1], 
     Hue[0.14583462015319276`, 1, 1], 
     Hue[0.14583462015319276`, 1, 1], 
     Hue[0.14838308693448174`, 1, 1], 
     Hue[0.14838308693448174`, 1, 1], 
     Hue[0.1534800204970599, 1, 1], 
     Hue[0.1534800204970599, 1, 1], 
     Hue[0.16367388762221627`, 1, 1], 
     Hue[0.16367388762221627`, 1, 1], 
     Hue[0.16622235440350527`, 1, 1], 
     Hue[0.16622235440350527`, 1, 1], 
     Hue[0.1687708211847944, 1, 1], 
     Hue[0.1687708211847944, 1, 1], 
     Hue[0.17386775474737268`, 1, 1], 
     Hue[0.17386775474737268`, 1, 1], 
     Hue[0.18406162187252892`, 1, 1], 
     Hue[0.18406162187252892`, 1, 1], 
     Hue[0.18656356433332041`, 1, 1], 
     Hue[0.18656356433332041`, 1, 1], 
     Hue[0.18906550679411188`, 1, 1], 
     Hue[0.18906550679411188`, 1, 1], 
     Hue[0.19406939171569484`, 1, 1], 
     Hue[0.19406939171569484`, 1, 1], 
     Hue[0.20407716155886074`, 1, 1], 
     Hue[0.20407716155886074`, 1, 1], 
     Hue[0.2065791040196522, 1, 1], 
     Hue[0.2065791040196522, 1, 1], 
     Hue[0.2090810464804437, 1, 1], 
     Hue[0.2090810464804437, 1, 1], 
     Hue[0.21408493140202664`, 1, 1], 
     Hue[0.21408493140202664`, 1, 1], 
     Hue[0.22409270124519243`, 1, 1], 
     Hue[0.22409270124519243`, 1, 1], 
     Hue[0.22659464370598376`, 1, 1], 
     Hue[0.22659464370598376`, 1, 1], 
     Hue[0.22909658616677522`, 1, 1], 
     Hue[0.22909658616677522`, 1, 1], 
     Hue[0.2341004710883582, 1, 1], 
     Hue[0.2341004710883582, 1, 1], 
     Hue[0.24410824093152408`, 1, 1], 
     Hue[0.24410824093152408`, 1, 1], 
     Hue[0.24661018339231555`, 1, 1], 
     Hue[0.24661018339231555`, 1, 1], 
     Hue[0.24911212585310705`, 1, 1], 
     Hue[0.24911212585310705`, 1, 1], 
     Hue[0.25411601077469, 1, 1], 
     Hue[0.25411601077469, 1, 1], 
     Hue[0.2641237806178559, 1, 1], 
     Hue[0.2641237806178559, 1, 1], 
     Hue[0.2666257230786474, 1, 1], 
     Hue[0.2666257230786474, 1, 1], 
     Hue[0.26912766553943873`, 1, 1], 
     Hue[0.26912766553943873`, 1, 1], 
     Hue[0.2741315504610215, 1, 1], 
     Hue[0.2741315504610215, 1, 1], 
     Hue[0.276633492921813, 1, 1], 
     Hue[0.276633492921813, 1, 1], 
     Hue[0.2791354353826045, 1, 1], 
     Hue[0.2791354353826045, 1, 1], 
     Hue[0.2841393203041874, 1, 1], 
     Hue[0.2841393203041874, 1, 1], 
     Hue[0.2866412627649789, 1, 1], 
     Hue[0.2866412627649789, 1, 1], 
     Hue[0.28914320522577036`, 1, 1], 
     Hue[0.28914320522577036`, 1, 1], 
     Hue[0.29414709014735335`, 1, 1], 
     Hue[0.29414709014735335`, 1, 1], 
     Hue[0.3041548599905192, 1, 1], 
     Hue[0.3041548599905192, 1, 1], 
     Hue[0.3066568024513107, 1, 1], 
     Hue[0.3066568024513107, 1, 1], 
     Hue[0.3091587449121022, 1, 1], 
     Hue[0.3091587449121022, 1, 1], 
     Hue[0.31416262983368515`, 1, 1], 
     Hue[0.31416262983368515`, 1, 1], 
     Hue[0.31666457229447664`, 1, 1], 
     Hue[0.31666457229447664`, 1, 1], 
     Hue[0.31916651475526814`, 1, 1], 
     Hue[0.31916651475526814`, 1, 1], 
     Hue[0.324170399676851, 1, 1], 
     Hue[0.324170399676851, 1, 1], 
     Hue[0.3266723421376425, 1, 1], 
     Hue[0.3266723421376425, 1, 1], 
     Hue[0.329174284598434, 1, 1], 
     Hue[0.329174284598434, 1, 1], 
     Hue[0.33417816952001694`, 1, 1], 
     Hue[0.33417816952001694`, 1, 1], 
     Hue[0.3441859393631829, 1, 1], 
     Hue[0.3441859393631829, 1, 1], 
     Hue[0.3465195552511106, 1, 1], 
     Hue[0.3465195552511106, 1, 1], 
     Hue[0.3488531711390381, 1, 1], 
     Hue[0.3488531711390381, 1, 1], 
     Hue[0.35352040291489345`, 1, 1], 
     Hue[0.35352040291489345`, 1, 1], 
     Hue[0.3558540188028212, 1, 1], 
     Hue[0.3558540188028212, 1, 1], 
     Hue[0.3581876346907486, 1, 1], 
     Hue[0.3581876346907486, 1, 1], 
     Hue[0.36285486646660386`, 1, 1], 
     Hue[0.36285486646660386`, 1, 1], 
     Hue[0.3651884823545317, 1, 1], 
     Hue[0.3651884823545317, 1, 1], 
     Hue[0.3675220982424591, 1, 1], 
     Hue[0.3675220982424591, 1, 1], 
     Hue[0.37218933001831445`, 1, 1], 
     Hue[0.37218933001831445`, 1, 1], 
     Hue[0.38152379357002464`, 1, 1], 
     Hue[0.38152379357002464`, 1, 1], 
     Hue[0.38385740945795216`, 1, 1], 
     Hue[0.38385740945795216`, 1, 1], 
     Hue[0.38619102534587996`, 1, 1], 
     Hue[0.38619102534587996`, 1, 1], 
     Hue[0.39085825712173494`, 1, 1], 
     Hue[0.39085825712173494`, 1, 1], 
     Hue[0.3931918730096627, 1, 1], 
     Hue[0.3931918730096627, 1, 1], 
     Hue[0.3955254888975901, 1, 1], 
     Hue[0.3955254888975901, 1, 1], 
     Hue[0.4001927206734454, 1, 1], 
     Hue[0.4001927206734454, 1, 1], 
     Hue[0.40252633656137315`, 1, 1], 
     Hue[0.40252633656137315`, 1, 1], 
     Hue[0.4048599524493006, 1, 1], 
     Hue[0.4048599524493006, 1, 1], 
     Hue[0.4095271842251559, 1, 1], 
     Hue[0.4095271842251559, 1, 1], 
     Hue[0.41886164777686646`, 1, 1], 
     Hue[0.41886164777686646`, 1, 1], 
     Hue[0.42119526366479415`, 1, 1], 
     Hue[0.42119526366479415`, 1, 1], 
     Hue[0.4235288795527217, 1, 1], 
     Hue[0.4235288795527217, 1, 1], 
     Hue[0.42819611132857693`, 1, 1], 
     Hue[0.42819611132857693`, 1, 1], 
     Hue[0.4305297272165047, 1, 1], 
     Hue[0.4305297272165047, 1, 1], 
     Hue[0.4328633431044322, 1, 1], 
     Hue[0.4328633431044322, 1, 1], 
     Hue[0.4375305748802874, 1, 1], 
     Hue[0.4375305748802874, 1, 1], 
     Hue[0.4398641907682152, 1, 1], 
     Hue[0.4398641907682152, 1, 1], 
     Hue[0.4421978066561428, 1, 1], 
     Hue[0.4421978066561428, 1, 1], 
     Hue[0.44686503843199804`, 1, 1], 
     Hue[0.44686503843199804`, 1, 1], 
     Hue[0.4561995019837082, 1, 1], 
     Hue[0.4561995019837082, 1, 1], 
     Hue[0.45853311787163575`, 1, 1], 
     Hue[0.45853311787163575`, 1, 1], 
     Hue[0.46086673375956344`, 1, 1], 
     Hue[0.46086673375956344`, 1, 1], 
     Hue[0.4655339655354184, 1, 1], 
     Hue[0.4655339655354184, 1, 1], 
     Hue[0.46786758142334617`, 1, 1], 
     Hue[0.46786758142334617`, 1, 1], 
     Hue[0.4702011973112737, 1, 1], 
     Hue[0.4702011973112737, 1, 1], 
     Hue[0.4748684290871289, 1, 1], 
     Hue[0.4748684290871289, 1, 1], 
     Hue[0.4772020449750567, 1, 1], 
     Hue[0.4772020449750567, 1, 1], 
     Hue[0.47953566086298427`, 1, 1], 
     Hue[0.47953566086298427`, 1, 1], 
     Hue[0.4842028926388394, 1, 1], 
     Hue[0.4842028926388394, 1, 1], 
     Hue[0.49353735619054995`, 1, 1], 
     Hue[0.49353735619054995`, 1, 1], 
     Hue[0.49606805226271244`, 1, 1], 
     Hue[0.49606805226271244`, 1, 1], 
     Hue[0.49859874833487494`, 1, 1], 
     Hue[0.49859874833487494`, 1, 1], 
     Hue[0.5036601404791998, 1, 1], 
     Hue[0.5036601404791998, 1, 1], 
     Hue[0.5061908365513623, 1, 1], 
     Hue[0.5061908365513623, 1, 1], 
     Hue[0.5087215326235248, 1, 1], 
     Hue[0.5087215326235248, 1, 1], 
     Hue[0.5137829247678498, 1, 1], 
     Hue[0.5137829247678498, 1, 1], 
     Hue[0.5163136208400123, 1, 1], 
     Hue[0.5163136208400123, 1, 1], 
     Hue[0.5188443169121748, 1, 1], 
     Hue[0.5188443169121748, 1, 1], 
     Hue[0.5239057090564998, 1, 1], 
     Hue[0.5239057090564998, 1, 1], 
     Hue[0.5340284933451498, 1, 1], 
     Hue[0.5340284933451498, 1, 1], 
     Hue[0.5365591894173123, 1, 1], 
     Hue[0.5365591894173123, 1, 1], 
     Hue[0.5390898854894748, 1, 1], 
     Hue[0.5390898854894748, 1, 1], 
     Hue[0.5441512776337998, 1, 1], 
     Hue[0.5441512776337998, 1, 1], 
     Hue[0.5466819737059623, 1, 1], 
     Hue[0.5466819737059623, 1, 1], 
     Hue[0.5492126697781248, 1, 1], 
     Hue[0.5492126697781248, 1, 1], 
     Hue[0.5542740619224498, 1, 1], 
     Hue[0.5542740619224498, 1, 1], 
     Hue[0.5568047579946123, 1, 1], 
     Hue[0.5568047579946123, 1, 1], 
     Hue[0.5593354540667748, 1, 1], 
     Hue[0.5593354540667748, 1, 1], 
     Hue[0.5643968462110998, 1, 1], 
     Hue[0.5643968462110998, 1, 1], 
     Hue[0.5745196304997497, 1, 1], 
     Hue[0.5745196304997497, 1, 1], 
     Hue[0.5770503265719122, 1, 1], 
     Hue[0.5770503265719122, 1, 1], 
     Hue[0.5795810226440747, 1, 1], 
     Hue[0.5795810226440747, 1, 1], 
     Hue[0.5846424147883997, 1, 1], 
     Hue[0.5846424147883997, 1, 1], 
     Hue[0.5871731108605622, 1, 1], 
     Hue[0.5871731108605622, 1, 1], 
     Hue[0.5897038069327247, 1, 1], 
     Hue[0.5897038069327247, 1, 1], 
     Hue[0.5947651990770497, 1, 1], 
     Hue[0.5947651990770497, 1, 1], 
     Hue[0.5972958951492122, 1, 1], 
     Hue[0.5972958951492122, 1, 1], 
     Hue[0.5998265912213747, 1, 1], 
     Hue[0.5998265912213747, 1, 1], 
     Hue[0.6048879833656997, 1, 1], 
     Hue[0.6048879833656997, 1, 1], 
     Hue[0.6150107676543496, 1, 1], 
     Hue[0.6150107676543496, 1, 1], 
     Hue[0.6175414637265121, 1, 1], 
     Hue[0.6175414637265121, 1, 1], 
     Hue[0.6200721597986746, 1, 1], 
     Hue[0.6200721597986746, 1, 1], 
     Hue[0.6251335519429996, 1, 1], 
     Hue[0.6251335519429996, 1, 1], 
     Hue[0.6276642480151621, 1, 1], 
     Hue[0.6276642480151621, 1, 1], 
     Hue[0.6301949440873246, 1, 1], 
     Hue[0.6301949440873246, 1, 1], 
     Hue[0.6352563362316496, 1, 1], 
     Hue[0.6352563362316496, 1, 1], 
     Hue[0.6377870323038121, 1, 1], 
     Hue[0.6377870323038121, 1, 1], 
     Hue[0.6403177283759746, 1, 1], 
     Hue[0.6403177283759746, 1, 1], 
     Hue[0.6453791205202996, 1, 1], 
     Hue[0.6453791205202996, 1, 1], 
     Hue[0.6555019048089492, 1, 1], 
     Hue[0.6555019048089492, 1, 1], 
     Hue[0.6578642743082477, 1, 1], 
     Hue[0.6578642743082477, 1, 1], 
     Hue[0.6602266438075466, 1, 1], 
     Hue[0.6602266438075466, 1, 1], 
     Hue[0.6649513828061435, 1, 1], 
     Hue[0.6649513828061435, 1, 1], 
     Hue[0.6673137523054423, 1, 1], 
     Hue[0.6673137523054423, 1, 1], 
     Hue[0.6696761218047409, 1, 1], 
     Hue[0.6696761218047409, 1, 1], 
     Hue[0.674400860803338, 1, 1], 
     Hue[0.674400860803338, 1, 1], 
     Hue[0.676763230302637, 1, 1], 
     Hue[0.676763230302637, 1, 1], 
     Hue[0.6791255998019354, 1, 1], 
     Hue[0.6791255998019354, 1, 1], 
     Hue[0.6838503388005327, 1, 1], 
     Hue[0.6838503388005327, 1, 1], 
     Hue[0.6932998167977273, 1, 1], 
     Hue[0.6932998167977273, 1, 1], 
     Hue[0.695662186297026, 1, 1], 
     Hue[0.695662186297026, 1, 1], 
     Hue[0.6980245557963245, 1, 1], 
     Hue[0.6980245557963245, 1, 1], 
     Hue[0.7027492947949219, 1, 1], 
     Hue[0.7027492947949219, 1, 1], 
     Hue[0.7051116642942206, 1, 1], 
     Hue[0.7051116642942206, 1, 1], 
     Hue[0.7074740337935191, 1, 1], 
     Hue[0.7074740337935191, 1, 1], 
     Hue[0.7121987727921164, 1, 1], 
     Hue[0.7121987727921164, 1, 1], 
     Hue[0.7145611422914152, 1, 1], 
     Hue[0.7145611422914152, 1, 1], 
     Hue[0.7169235117907138, 1, 1], 
     Hue[0.7169235117907138, 1, 1], 
     Hue[0.7216482507893109, 1, 1], 
     Hue[0.7216482507893109, 1, 1], 
     Hue[0.7310977287865053, 1, 1], 
     Hue[0.7310977287865053, 1, 1], 
     Hue[0.7334600982858038, 1, 1], 
     Hue[0.7334600982858038, 1, 1], 
     Hue[0.7358224677851026, 1, 1], 
     Hue[0.7358224677851026, 1, 1], 
     Hue[0.7405472067836996, 1, 1], 
     Hue[0.7405472067836996, 1, 1], 
     Hue[0.7429095762829985, 1, 1], 
     Hue[0.7429095762829985, 1, 1], 
     Hue[0.7452719457822969, 1, 1], 
     Hue[0.7452719457822969, 1, 1], 
     Hue[0.7499966847808942, 1, 1], 
     Hue[0.7499966847808942, 1, 1], 
     Hue[0.752359054280193, 1, 1], 
     Hue[0.752359054280193, 1, 1], 
     Hue[0.7547214237794915, 1, 1], 
     Hue[0.7547214237794915, 1, 1], 
     Hue[0.7594461627780887, 1, 1], 
     Hue[0.7594461627780887, 1, 1], 
     Hue[0.7688956407752834, 1, 1], 
     Hue[0.7688956407752834, 1, 1], 
     Hue[0.7712580102745822, 1, 1], 
     Hue[0.7712580102745822, 1, 1], 
     Hue[0.7736203797738807, 1, 1], 
     Hue[0.7736203797738807, 1, 1], 
     Hue[0.7783451187724779, 1, 1], 
     Hue[0.7783451187724779, 1, 1], 
     Hue[0.7807074882717767, 1, 1], 
     Hue[0.7807074882717767, 1, 1], 
     Hue[0.7830698577710753, 1, 1], 
     Hue[0.7830698577710753, 1, 1], 
     Hue[0.7877945967696725, 1, 1], 
     Hue[0.7877945967696725, 1, 1], 
     Hue[0.7901569662689714, 1, 1], 
     Hue[0.7901569662689714, 1, 1], 
     Hue[0.7925193357682698, 1, 1], 
     Hue[0.7925193357682698, 1, 1], 
     Hue[0.7972440747668671, 1, 1], 
     Hue[0.7972440747668671, 1, 1], 
     Hue[0.8066935527640614, 1, 1], 
     Hue[0.8066935527640614, 1, 1], 
     Hue[0.8090093979428622, 1, 1], 
     Hue[0.8090093979428622, 1, 1], 
     Hue[0.8113252431216633, 1, 1], 
     Hue[0.8113252431216633, 1, 1], 
     Hue[0.8159569334792656, 1, 1], 
     Hue[0.8159569334792656, 1, 1], 
     Hue[0.8182727786580668, 1, 1], 
     Hue[0.8182727786580668, 1, 1], 
     Hue[0.8205886238368676, 1, 1], 
     Hue[0.8205886238368676, 1, 1], 
     Hue[0.8252203141944696, 1, 1], 
     Hue[0.8252203141944696, 1, 1], 
     Hue[0.8275361593732707, 1, 1], 
     Hue[0.8275361593732707, 1, 1], 
     Hue[0.8298520045520715, 1, 1], 
     Hue[0.8298520045520715, 1, 1], 
     Hue[0.8344836949096734, 1, 1], 
     Hue[0.8344836949096734, 1, 1], 
     Hue[0.8437470756248775, 1, 1], 
     Hue[0.8437470756248775, 1, 1], 
     Hue[0.8460629208036786, 1, 1], 
     Hue[0.8460629208036786, 1, 1], 
     Hue[0.8483787659824795, 1, 1], 
     Hue[0.8483787659824795, 1, 1], 
     Hue[0.8530104563400814, 1, 1], 
     Hue[0.8530104563400814, 1, 1], 
     Hue[0.8553263015188826, 1, 1], 
     Hue[0.8553263015188826, 1, 1], 
     Hue[0.8576421466976837, 1, 1], 
     Hue[0.8576421466976837, 1, 1], 
     Hue[0.8622738370552856, 1, 1], 
     Hue[0.8622738370552856, 1, 1], 
     Hue[0.8645896822340865, 1, 1], 
     Hue[0.8645896822340865, 1, 1], 
     Hue[0.8669055274128876, 1, 1], 
     Hue[0.8669055274128876, 1, 1], 
     Hue[0.8715372177704899, 1, 1], 
     Hue[0.8715372177704899, 1, 1], 
     Hue[0.8808005984856938, 1, 1], 
     Hue[0.8808005984856938, 1, 1], 
     Hue[0.8831164436644949, 1, 1], 
     Hue[0.8831164436644949, 1, 1], 
     Hue[0.8854322888432958, 1, 1], 
     Hue[0.8854322888432958, 1, 1], 
     Hue[0.8900639792008977, 1, 1], 
     Hue[0.8900639792008977, 1, 1], 
     Hue[0.892379824379699, 1, 1], 
     Hue[0.892379824379699, 1, 1], 
     Hue[0.8946956695585001, 1, 1], 
     Hue[0.8946956695585001, 1, 1], 
     Hue[0.899327359916102, 1, 1], 
     Hue[0.899327359916102, 1, 1], 
     Hue[0.9016432050949029, 1, 1], 
     Hue[0.9016432050949029, 1, 1], 
     Hue[0.903959050273704, 1, 1], 
     Hue[0.903959050273704, 1, 1], 
     Hue[0.9085907406313063, 1, 1], 
     Hue[0.9085907406313063, 1, 1], 
     Hue[0.9178541213465102, 1, 1], 
     Hue[0.9178541213465102, 1, 1], 
     Hue[0.9201699665253114, 1, 1], 
     Hue[0.9201699665253114, 1, 1], 
     Hue[0.9224858117041121, 1, 1], 
     Hue[0.9224858117041121, 1, 1], 
     Hue[0.9271175020617142, 1, 1], 
     Hue[0.9271175020617142, 1, 1], 
     Hue[0.9294333472405153, 1, 1], 
     Hue[0.9294333472405153, 1, 1], 
     Hue[0.9317491924193162, 1, 1], 
     Hue[0.9317491924193162, 1, 1], 
     Hue[0.9363808827769181, 1, 1], 
     Hue[0.9363808827769181, 1, 1], 
     Hue[0.9386967279557192, 1, 1], 
     Hue[0.9386967279557192, 1, 1], 
     Hue[0.9410125731345201, 1, 1], 
     Hue[0.9410125731345201, 1, 1], 
     Hue[0.945644263492122, 1, 1], 
     Hue[0.945644263492122, 1, 1], 
     Hue[0.9479601086709232, 1, 1], 
     Hue[0.9479601086709232, 1, 1], 
     Hue[0.9502759538497243, 1, 1], 
     Hue[0.9502759538497243, 1, 1], 
     Hue[0.9549076442073263, 1, 1], 
     Hue[0.9549076442073263, 1, 1], 
     Hue[0.9574205695703621, 1, 1], 
     Hue[0.9574205695703621, 1, 1], 
     Hue[0.9599334949333983, 1, 1], 
     Hue[0.9599334949333983, 1, 1], 
     Hue[0.96495934565947, 1, 1], 
     Hue[0.96495934565947, 1, 1], 
     Hue[0.9674722710225059, 1, 1], 
     Hue[0.9674722710225059, 1, 1], 
     Hue[0.9699851963855417, 1, 1], 
     Hue[0.9699851963855417, 1, 1], 
     Hue[0.9750110471116135, 1, 1], 
     Hue[0.9750110471116135, 1, 1], 
     Hue[0.9775239724746493, 1, 1], 
     Hue[0.9775239724746493, 1, 1], 
     Hue[0.9800368978376851, 1, 1], 
     Hue[0.9800368978376851, 1, 1], 
     Hue[0.9850627485637571, 1, 1], 
     Hue[0.9850627485637571, 1, 1], 
     Hue[0.9875756739267929, 1, 1], 
     Hue[0.9875756739267929, 1, 1], 
     Hue[0.9900885992898291, 1, 1], 
     Hue[0.9900885992898291, 1, 1], 
     Hue[0.9951144500159008, 1, 1], 
     Hue[0.9951144500159008, 1, 1], 
     Hue[0.9976273753789366, 1, 1], 
     Hue[0.9976273753789366, 1, 1], 
     Hue[0.00014030074197249827`, 1, 1], 
     Hue[0.00014030074197249827`, 1, 1], 
     Hue[0.005166151468044201, 1, 1], 
     Hue[0.005166151468044201, 1, 1], 
     Hue[0.007679076831080055, 1, 1], 
     Hue[0.007679076831080055, 1, 1], 
     Hue[0.010192002194115906`, 1, 1], 
     Hue[0.010192002194115906`, 1, 1], 
     Hue[0.015217852920187897`, 1, 1], 
     Hue[0.015217852920187897`, 1, 1], 
     Hue[0.017730778283223746`, 1, 1], 
     Hue[0.017730778283223746`, 1, 1], 
     Hue[0.02024370364625988, 1, 1], 
     Hue[0.02024370364625988, 1, 1], 
     Hue[0.025269554372331585`, 1, 1], 
     Hue[0.025269554372331585`, 1, 1], 
     Hue[0.027782479735367443`, 1, 1], 
     Hue[0.027782479735367443`, 1, 1], 
     Hue[0.030295405098403293`, 1, 1], 
     Hue[0.030295405098403293`, 1, 1], 
     Hue[0.035321255824475, 1, 1], 
     Hue[0.035321255824475, 1, 1], 
     Hue[0.037834181187510856`, 1, 1], 
     Hue[0.037834181187510856`, 1, 1], 
     Hue[0.04034710655054671, 1, 1], 
     Hue[0.04034710655054671, 1, 1], 
     Hue[0.045372957276618414`, 1, 1], 
     Hue[0.045372957276618414`, 1, 1], 
     Hue[0.04788588263965426, 1, 1], 
     Hue[0.04788588263965426, 1, 1], 
     Hue[0.05039880800269011, 1, 1], 
     Hue[0.05039880800269011, 1, 1], 
     Hue[0.0554246587287621, 1, 1], 
     Hue[0.0554246587287621, 1, 1], 
     Hue[0.05793758409179796, 1, 1], 
     Hue[0.05793758409179796, 1, 1], 
     Hue[0.06045050945483409, 1, 1], 
     Hue[0.06045050945483409, 1, 1], 
     Hue[0.06547636018090579, 1, 1], 
     Hue[0.06547636018090579, 1, 1], 
     Hue[0.06798928554394164, 1, 1], 
     Hue[0.06798928554394164, 1, 1], 
     Hue[0.0705022109069775, 1, 1], 
     Hue[0.0705022109069775, 1, 1], 
     Hue[0.0755280616330492, 1, 1], 
     Hue[0.0755280616330492, 1, 1], 
     Hue[0.07804098699608507, 1, 1], 
     Hue[0.07804098699608507, 1, 1], 
     Hue[0.08055391235912089, 1, 1], 
     Hue[0.08055391235912089, 1, 1], 
     Hue[0.08557976308519263, 1, 1], 
     Hue[0.08557976308519263, 1, 1], 
     Hue[0.08809268844822847, 1, 1], 
     Hue[0.08809268844822847, 1, 1], 
     Hue[0.09060561381126431, 1, 1], 
     Hue[0.09060561381126431, 1, 1], 
     Hue[0.09563146453733631, 1, 1], 
     Hue[0.09563146453733631, 1, 1], 
     Hue[0.09814438990037216, 1, 1], 
     Hue[0.09814438990037216, 1, 1], 
     Hue[0.10065731526340828`, 1, 1], 
     Hue[0.10065731526340828`, 1, 1], 
     Hue[0.10568316598948, 1, 1], 
     Hue[0.10568316598948, 1, 1], 
     Hue[0.10819609135251586`, 1, 1], 
     Hue[0.10819609135251586`, 1, 1], 
     Hue[0.1107090167155517, 1, 1], 
     Hue[0.1107090167155517, 1, 1], 
     Hue[0.1157348674416234, 1, 1], 
     Hue[0.1157348674416234, 1, 1], 
     Hue[0.11807946623179555`, 1, 1], 
     Hue[0.11807946623179555`, 1, 1], 
     Hue[0.12042406502196741`, 1, 1], 
     Hue[0.12042406502196741`, 1, 1], 
     Hue[0.12511326260231143`, 1, 1], 
     Hue[0.12511326260231143`, 1, 1], 
     Hue[0.12745786139248358`, 1, 1], 
     Hue[0.12745786139248358`, 1, 1], 
     Hue[0.1298024601826557, 1, 1], 
     Hue[0.1298024601826557, 1, 1], 
     Hue[0.1344916577629997, 1, 1], 
     Hue[0.1344916577629997, 1, 1], 
     Hue[0.1368362565531716, 1, 1], 
     Hue[0.1368362565531716, 1, 1], 
     Hue[0.13918085534334373`, 1, 1], 
     Hue[0.13918085534334373`, 1, 1], 
     Hue[0.14387005292368804`, 1, 1], 
     Hue[0.14387005292368804`, 1, 1], 
     Hue[0.14621465171386017`, 1, 1], 
     Hue[0.14621465171386017`, 1, 1], 
     Hue[0.14855925050403201`, 1, 1], 
     Hue[0.14855925050403201`, 1, 1], 
     Hue[0.15324844808437604`, 1, 1], 
     Hue[0.15324844808437604`, 1, 1], 
     Hue[0.15559304687454817`, 1, 1], 
     Hue[0.15559304687454817`, 1, 1], 
     Hue[0.15793764566472007`, 1, 1], 
     Hue[0.15793764566472007`, 1, 1], 
     Hue[0.16262684324506407`, 1, 1], 
     Hue[0.16262684324506407`, 1, 1], 
     Hue[0.16497144203523623`, 1, 1], 
     Hue[0.16497144203523623`, 1, 1], 
     Hue[0.16731604082540807`, 1, 1], 
     Hue[0.16731604082540807`, 1, 1], 
     Hue[0.17200523840575208`, 1, 1], 
     Hue[0.17200523840575208`, 1, 1], 
     Hue[0.17434983719592423`, 1, 1], 
     Hue[0.17434983719592423`, 1, 1], 
     Hue[0.1766944359860961, 1, 1], 
     Hue[0.1766944359860961, 1, 1], 
     Hue[0.1813836335664401, 1, 1], 
     Hue[0.1813836335664401, 1, 1], 
     Hue[0.18372823235661226`, 1, 1], 
     Hue[0.18372823235661226`, 1, 1], 
     Hue[0.18607283114678438`, 1, 1], 
     Hue[0.18607283114678438`, 1, 1], 
     Hue[0.1907620287271284, 1, 1], 
     Hue[0.1907620287271284, 1, 1], 
     Hue[0.19310662751730026`, 1, 1], 
     Hue[0.19310662751730026`, 1, 1], 
     Hue[0.19545122630747241`, 1, 1], 
     Hue[0.19545122630747241`, 1, 1], 
     Hue[0.2001404238878167, 1, 1], 
     Hue[0.2001404238878167, 1, 1], 
     Hue[0.20248502267798885`, 1, 1], 
     Hue[0.20248502267798885`, 1, 1], 
     Hue[0.20482962146816072`, 1, 1], 
     Hue[0.20482962146816072`, 1, 1], 
     Hue[0.20951881904850472`, 1, 1], 
     Hue[0.20951881904850472`, 1, 1], 
     Hue[0.21186341783867682`, 1, 1], 
     Hue[0.21186341783867682`, 1, 1], 
     Hue[0.21420801662884872`, 1, 1], 
     Hue[0.21420801662884872`, 1, 1], 
     Hue[0.21889721420919273`, 1, 1], 
     Hue[0.21889721420919273`, 1, 1], 
     Hue[0.22124181299936488`, 1, 1], 
     Hue[0.22124181299936488`, 1, 1], 
     Hue[0.22358641178953675`, 1, 1], 
     Hue[0.22358641178953675`, 1, 1], 
     Hue[0.22827560936988076`, 1, 1], 
     Hue[0.22827560936988076`, 1, 1], 
     Hue[0.2306202081600529, 1, 1], 
     Hue[0.2306202081600529, 1, 1], 
     Hue[0.23296480695022476`, 1, 1], 
     Hue[0.23296480695022476`, 1, 1], 
     Hue[0.23765400453056879`, 1, 1], 
     Hue[0.23765400453056879`, 1, 1], 
     Hue[0.2399986033207409, 1, 1], 
     Hue[0.2399986033207409, 1, 1], 
     Hue[0.24234320211091306`, 1, 1], 
     Hue[0.24234320211091306`, 1, 1], 
     Hue[0.24703239969125707`, 1, 1], 
     Hue[0.24703239969125707`, 1, 1], 
     Hue[0.24937699848142894`, 1, 1], 
     Hue[0.24937699848142894`, 1, 1], 
     Hue[0.2517215972716011, 1, 1], 
     Hue[0.2517215972716011, 1, 1], 
     Hue[0.2564107948519454, 1, 1], 
     Hue[0.2564107948519454, 1, 1], 
     Hue[0.2587553936421175, 1, 1], 
     Hue[0.2587553936421175, 1, 1], 
     Hue[0.2610999924322894, 1, 1], 
     Hue[0.2610999924322894, 1, 1], 
     Hue[0.2657891900126334, 1, 1], 
     Hue[0.2657891900126334, 1, 1], 
     Hue[0.2683308689870403, 1, 1], 
     Hue[0.2683308689870403, 1, 1], 
     Hue[0.2708725479614471, 1, 1], 
     Hue[0.2708725479614471, 1, 1], 
     Hue[0.27595590591026087`, 1, 1], 
     Hue[0.27595590591026087`, 1, 1], 
     Hue[0.27849758488466775`, 1, 1], 
     Hue[0.27849758488466775`, 1, 1], 
     Hue[0.28103926385907463`, 1, 1], 
     Hue[0.28103926385907463`, 1, 1], 
     Hue[0.2861226218078886, 1, 1], 
     Hue[0.2861226218078886, 1, 1], 
     Hue[0.2886643007822955, 1, 1], 
     Hue[0.2886643007822955, 1, 1], 
     Hue[0.29120597975670265`, 1, 1], 
     Hue[0.29120597975670265`, 1, 1], 
     Hue[0.2962893377055164, 1, 1], 
     Hue[0.2962893377055164, 1, 1], 
     Hue[0.29883101667992323`, 1, 1], 
     Hue[0.29883101667992323`, 1, 1], 
     Hue[0.3013726956543301, 1, 1], 
     Hue[0.3013726956543301, 1, 1], 
     Hue[0.3064560536031439, 1, 1], 
     Hue[0.3064560536031439, 1, 1], 
     Hue[0.30899773257755075`, 1, 1], 
     Hue[0.30899773257755075`, 1, 1], 
     Hue[0.3115394115519576, 1, 1], 
     Hue[0.3115394115519576, 1, 1], 
     Hue[0.3166227695007716, 1, 1], 
     Hue[0.3166227695007716, 1, 1], 
     Hue[0.3191644484751785, 1, 1], 
     Hue[0.3191644484751785, 1, 1], 
     Hue[0.32170612744958565`, 1, 1], 
     Hue[0.32170612744958565`, 1, 1], 
     Hue[0.32678948539839936`, 1, 1], 
     Hue[0.32678948539839936`, 1, 1], 
     Hue[0.32933116437280624`, 1, 1], 
     Hue[0.32933116437280624`, 1, 1], 
     Hue[0.3318728433472131, 1, 1], 
     Hue[0.3318728433472131, 1, 1], 
     Hue[0.3369562012960269, 1, 1], 
     Hue[0.3369562012960269, 1, 1], 
     Hue[0.3394978802704337, 1, 1], 
     Hue[0.3394978802704337, 1, 1], 
     Hue[0.3420395592448406, 1, 1], 
     Hue[0.3420395592448406, 1, 1], 
     Hue[0.3471229171936546, 1, 1], 
     Hue[0.3471229171936546, 1, 1], 
     Hue[0.3496645961680615, 1, 1], 
     Hue[0.3496645961680615, 1, 1], 
     Hue[0.35220627514246866`, 1, 1], 
     Hue[0.35220627514246866`, 1, 1], 
     Hue[0.35728963309128237`, 1, 1], 
     Hue[0.35728963309128237`, 1, 1], 
     Hue[0.35983131206568925`, 1, 1], 
     Hue[0.35983131206568925`, 1, 1], 
     Hue[0.3623729910400961, 1, 1], 
     Hue[0.3623729910400961, 1, 1], 
     Hue[0.36745634898890983`, 1, 1], 
     Hue[0.36745634898890983`, 1, 1], 
     Hue[0.3699980279633167, 1, 1], 
     Hue[0.3699980279633167, 1, 1], 
     Hue[0.3725397069377236, 1, 1], 
     Hue[0.3725397069377236, 1, 1], 
     Hue[0.37762306488653763`, 1, 1], 
     Hue[0.37762306488653763`, 1, 1], 
     Hue[0.38016474386094445`, 1, 1], 
     Hue[0.38016474386094445`, 1, 1], 
     Hue[0.3827064228353516, 1, 1], 
     Hue[0.3827064228353516, 1, 1], 
     Hue[0.38778978078416537`, 1, 1], 
     Hue[0.38778978078416537`, 1, 1], 
     Hue[0.39033145975857225`, 1, 1], 
     Hue[0.39033145975857225`, 1, 1], 
     Hue[0.39287313873297913`, 1, 1], 
     Hue[0.39287313873297913`, 1, 1], 
     Hue[0.39795649668179284`, 1, 1], 
     Hue[0.39795649668179284`, 1, 1], 
     Hue[0.4004981756561997, 1, 1], 
     Hue[0.4004981756561997, 1, 1], 
     Hue[0.4030398546306066, 1, 1], 
     Hue[0.4030398546306066, 1, 1], 
     Hue[0.4081232125794206, 1, 1], 
     Hue[0.4081232125794206, 1, 1], 
     Hue[0.41066489155382746`, 1, 1], 
     Hue[0.41066489155382746`, 1, 1], 
     Hue[0.4132065705282346, 1, 1], 
     Hue[0.4132065705282346, 1, 1], 
     Hue[0.4157482495026415, 1, 1], 
     Hue[0.4157482495026415, 1, 1], 
     Hue[0.4182899284770484, 1, 1], 
     Hue[0.4182899284770484, 1, 1], 
     Hue[0.42083160745145526`, 1, 1], 
     Hue[0.42083160745145526`, 1, 1], 
     Hue[0.4233732864258621, 1, 1], 
     Hue[0.4233732864258621, 1, 1], 
     Hue[0.42845664437467584`, 1, 1], 
     Hue[0.42845664437467584`, 1, 1], 
     Hue[0.430829996776219, 1, 1], 
     Hue[0.430829996776219, 1, 1], 
     Hue[0.4332033491777619, 1, 1], 
     Hue[0.4332033491777619, 1, 1], 
     Hue[0.4379500539808479, 1, 1], 
     Hue[0.4379500539808479, 1, 1], 
     Hue[0.4403234063823911, 1, 1], 
     Hue[0.4403234063823911, 1, 1], 
     Hue[0.4426967587839343, 1, 1], 
     Hue[0.4426967587839343, 1, 1], 
     Hue[0.4474434635870203, 1, 1], 
     Hue[0.4474434635870203, 1, 1], 
     Hue[0.44981681598856316`, 1, 1], 
     Hue[0.44981681598856316`, 1, 1], 
     Hue[0.45219016839010634`, 1, 1], 
     Hue[0.45219016839010634`, 1, 1], 
     Hue[0.45693687319319265`, 1, 1], 
     Hue[0.45693687319319265`, 1, 1], 
     Hue[0.45931022559473583`, 1, 1], 
     Hue[0.45931022559473583`, 1, 1], 
     Hue[0.4616835779962787, 1, 1], 
     Hue[0.4616835779962787, 1, 1], 
     Hue[0.4664302827993647, 1, 1], 
     Hue[0.4664302827993647, 1, 1], 
     Hue[0.4688036352009079, 1, 1], 
     Hue[0.4688036352009079, 1, 1], 
     Hue[0.4711769876024508, 1, 1], 
     Hue[0.4711769876024508, 1, 1], 
     Hue[0.47592369240553684`, 1, 1], 
     Hue[0.47592369240553684`, 1, 1], 
     Hue[0.47829704480707996`, 1, 1], 
     Hue[0.47829704480707996`, 1, 1], 
     Hue[0.48067039720862315`, 1, 1], 
     Hue[0.48067039720862315`, 1, 1], 
     Hue[0.4854171020117092, 1, 1], 
     Hue[0.4854171020117092, 1, 1], 
     Hue[0.4877904544132521, 1, 1], 
     Hue[0.4877904544132521, 1, 1], 
     Hue[0.4901638068147952, 1, 1], 
     Hue[0.4901638068147952, 1, 1], 
     Hue[0.4949105116178816, 1, 1], 
     Hue[0.4949105116178816, 1, 1], 
     Hue[0.4972838640194247, 1, 1], 
     Hue[0.4972838640194247, 1, 1], 
     Hue[0.4996572164209676, 1, 1], 
     Hue[0.4996572164209676, 1, 1], 
     Hue[0.5044039212240535, 1, 1], 
     Hue[0.5044039212240535, 1, 1], 
     Hue[0.5067772736255967, 1, 1], 
     Hue[0.5067772736255967, 1, 1], 
     Hue[0.5091506260271397, 1, 1], 
     Hue[0.5091506260271397, 1, 1], 
     Hue[0.5138973308302257, 1, 1], 
     Hue[0.5138973308302257, 1, 1], 
     Hue[0.5162706832317687, 1, 1], 
     Hue[0.5162706832317687, 1, 1], 
     Hue[0.518644035633312, 1, 1], 
     Hue[0.518644035633312, 1, 1], 
     Hue[0.523390740436398, 1, 1], 
     Hue[0.523390740436398, 1, 1], 
     Hue[0.5257640928379409, 1, 1], 
     Hue[0.5257640928379409, 1, 1], 
     Hue[0.528137445239484, 1, 1], 
     Hue[0.528137445239484, 1, 1], 
     Hue[0.5328841500425704, 1, 1], 
     Hue[0.5328841500425704, 1, 1], 
     Hue[0.5352575024441135, 1, 1], 
     Hue[0.5352575024441135, 1, 1], 
     Hue[0.5376308548456564, 1, 1], 
     Hue[0.5376308548456564, 1, 1], 
     Hue[0.5423775596487425, 1, 1], 
     Hue[0.5423775596487425, 1, 1], 
     Hue[0.5447509120502856, 1, 1], 
     Hue[0.5447509120502856, 1, 1], 
     Hue[0.5471242644518288, 1, 1], 
     Hue[0.5471242644518288, 1, 1], 
     Hue[0.5518709692549145, 1, 1], 
     Hue[0.5518709692549145, 1, 1], 
     Hue[0.5542443216564572, 1, 1], 
     Hue[0.5542443216564572, 1, 1], 
     Hue[0.5566176740580002, 1, 1], 
     Hue[0.5566176740580002, 1, 1], 
     Hue[0.5613643788610867, 1, 1], 
     Hue[0.5613643788610867, 1, 1], 
     Hue[0.5637377312626297, 1, 1], 
     Hue[0.5637377312626297, 1, 1], 
     Hue[0.566111083664173, 1, 1], 
     Hue[0.566111083664173, 1, 1], 
     Hue[0.5708577884672592, 1, 1], 
     Hue[0.5708577884672592, 1, 1], 
     Hue[0.5732311408688024, 1, 1], 
     Hue[0.5732311408688024, 1, 1], 
     Hue[0.5756044932703456, 1, 1], 
     Hue[0.5756044932703456, 1, 1], 
     Hue[0.5803511980734313, 1, 1], 
     Hue[0.5803511980734313, 1, 1], 
     Hue[0.5850048542355217, 1, 1], 
     Hue[0.5850048542355217, 1, 1], 
     Hue[0.5896585103976129, 1, 1], 
     Hue[0.5896585103976129, 1, 1], 
     Hue[0.5943121665597039, 1, 1], 
     Hue[0.5943121665597039, 1, 1], 
     Hue[0.5989658227217949, 1, 1], 
     Hue[0.5989658227217949, 1, 1], 
     Hue[0.6036194788838858, 1, 1], 
     Hue[0.6036194788838858, 1, 1], 
     Hue[0.6082731350459769, 1, 1], 
     Hue[0.6082731350459769, 1, 1], 
     Hue[0.6129267912080679, 1, 1], 
     Hue[0.6129267912080679, 1, 1], 
     Hue[0.6175804473701589, 1, 1], 
     Hue[0.6175804473701589, 1, 1], 
     Hue[0.6222341035322498, 1, 1], 
     Hue[0.6222341035322498, 1, 1], 
     Hue[0.626887759694341, 1, 1], 
     Hue[0.626887759694341, 1, 1], 
     Hue[0.6315414158564319, 1, 1], 
     Hue[0.6315414158564319, 1, 1], 
     Hue[0.6361950720185223, 1, 1], 
     Hue[0.6361950720185223, 1, 1], 
     Hue[0.6455023843427038, 1, 1], 
     Hue[0.6455023843427038, 1, 1], 
     Hue[0.6548096966668858, 1, 1], 
     Hue[0.6548096966668858, 1, 1], 
     Hue[0.6641170089910678, 1, 1], 
     Hue[0.6641170089910678, 1, 1], 
     Hue[0.6734243213152493, 1, 1], 
     Hue[0.6734243213152493, 1, 1], 
     Hue[0.6757511493962948, 1, 1], 
     Hue[0.6757511493962948, 1, 1], 
     Hue[0.6780779774773398, 1, 1], 
     Hue[0.6780779774773398, 1, 1], 
     Hue[0.6827316336394308, 1, 1], 
     Hue[0.6827316336394308, 1, 1]}],
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  Epilog->LineBox[CompressedData["
1:eJwUV3c4198Xt/ceH3tLVkIZ4X3vzaqkQpQyEoWQCgmVlUSDjKxkZmQle+Vz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    "]],
  PlotRange->All,
  PlotRangeClipping->True,
  PlotRangePadding->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{
  3.3927237629789762`*^9, 3.3927238194601917`*^9, 3.392724146159963*^9, 
   3.392724904119856*^9, {3.392726508226448*^9, 3.39184984146875*^9}, 
   3.391850041875*^9, 3.3924723631853456`*^9}]
}, Open  ]],

Cell["Let us first generate a suitable grid in x-space:", "Text"],

Cell["a = -10.; b = 14.; n = 512;", "Input"],

Cell["xgrid = QGrid[a,b,n];", "Input"],

Cell[TextData[{
 "This command divides the interval [-10,10] by equally spaced points ",
 Cell[BoxData[
  StyleBox[
   RowBox[{
    SubscriptBox[
     StyleBox["x",
      FontFamily->"Times New Roman",
      FontSlant->"Italic"], "0"], " ", "=", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"a", "<", " ", 
        SubscriptBox[
         StyleBox["x",
          FontFamily->"Times New Roman",
          FontSlant->"Italic"], 
         RowBox[{"1", " "}]], "<", " ", 
        SubscriptBox[
         StyleBox["x",
          FontSlant->"Italic"], 
         RowBox[{"2", " "}]], "<"}], " ", "..."}], " ", "<", " ", 
      SubscriptBox[
       StyleBox["x",
        FontSlant->"Italic"], "n"]}], "=", " ", "b"}]}],
   FontFamily->"Times New Roman"]],
  FontFamily->"Times New Roman"],
 " and returns the list {",
 Cell[BoxData[
  RowBox[{" ", 
   RowBox[{
    SubscriptBox[
     StyleBox["x",
      FontFamily->"Times New Roman",
      FontSlant->"Italic"], "1"], ",", " ", 
    SubscriptBox[
     StyleBox["x",
      FontSlant->"Italic"], "2"], ",", " ", "...", " ", ",", " ", 
    SubscriptBox[
     StyleBox["x",
      FontSlant->"Italic"], "n"]}]}]],
  FontFamily->"Times New Roman"],
 "}. The number n should be even. You can obtain information about the space \
grid via the following functions."
}], "Text"],

Cell[CellGroupData[{

Cell["{QLeftBorder[a,b,n],QRightBorder[a,b,n],QSpaceStep[a,b,n]}", "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"-", "9.984375`"}], ",", "13.96875`", ",", "0.046875`"}], 
  "}"}]], "Output",
 CellChangeTimes->{
  3.3927261456250525`*^9, {3.392726508947485*^9, 3.3918498455*^9}, 
   3.391850043296875*^9, 3.392472364434189*^9}]
}, Open  ]],

Cell[TextData[{
 StyleBox["QIndexPosition[", "InlineInput"],
 StyleBox["x",
  FontSlant->"Italic"],
 StyleBox[",", "InlineInput"],
 StyleBox["a",
  FontSlant->"Italic"],
 StyleBox[",", "InlineInput"],
 StyleBox["b",
  FontSlant->"Italic"],
 StyleBox[",", "InlineInput"],
 StyleBox["n",
  FontSlant->"Italic"],
 StyleBox["]", "InlineInput"],
 " lets you access the element of xgrid corresponding to a special value of \
",
 StyleBox["x",
  FontSlant->"Italic"]
}], "Text"],

Cell[CellGroupData[{

Cell["QIndexPosition[5.,a,b,n]", "Input"],

Cell[BoxData["321"], "Output",
 CellChangeTimes->{
  3.392726148539243*^9, {3.3927265110104513`*^9, 3.391849845671875*^9}, 
   3.39185004421875*^9, 3.392472365017824*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["xgrid[[321]]", "Input"],

Cell[BoxData["5.015625`"], "Output",
 CellChangeTimes->{
  3.3927261527553053`*^9, {3.392726511631344*^9, 3.39184984575*^9}, 
   3.3918500443125*^9, 3.392472365087134*^9}]
}, Open  ]],

Cell["We can now define a discretization of the function f by", "Text"],

Cell["fvals = f /@ QGrid[a,b,n];", "Input",
 CellChangeTimes->{3.3927266517528286`*^9, 3.392727838989992*^9}],

Cell["\<\
Usually, the function in position space will be the result of a numerical \
computation, and the list fvals of complex numbers is the object to start \
with. \
\>", "Text"],

Cell[CellGroupData[{

Cell["QListArgColorPlot[fvals,PlotRange->All]", "Input"],

Cell[BoxData[
 GraphicsBox[GraphicsComplexBox[CompressedData["
1:eJyF22t0VFWWB/AjhKaVEUvkNeoMF0QFtOlCHvIKbECBCIQikBBeyc37ndxK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   "], {
    {GrayLevel[0.5], EdgeForm[None], 
     GraphicsGroupBox[{PolygonBox[{{33, 513, 34}, {394, 1234, 395}}], 
       PolygonBox[CompressedData["
1:eJwtmHfYiFUYh8973tdKRLKyJUlChSQZpZJNKLJJ9k7ITlaS7JXIKmWVrJSR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        "]]}]}, {{}, {}, {
      {Hue[0.821874730133715, 1, 1], LineBox[{1, 2}]}, 
      {Hue[0.8367955060485802, 1, 1], LineBox[{2, 3}]}, 
      {Hue[0.8517162819634454, 1, 1], LineBox[{3, 4}]}, 
      {Hue[0.8666370578783106, 1, 1], LineBox[{4, 5}]}, 
      {Hue[0.8815578337931758, 1, 1], LineBox[{5, 6}]}, 
      {Hue[0.8964786097080409, 1, 1], LineBox[{6, 7}]}, 
      {Hue[0.9113993856229061, 1, 1], LineBox[{7, 8}]}, 
      {Hue[0.9263201615377713, 1, 1], LineBox[{8, 9}]}, 
      {Hue[0.9412409374526365, 1, 1], LineBox[{9, 10}]}, 
      {Hue[0.9561617133675017, 1, 1], LineBox[{10, 11}]}, 
      {Hue[0.9710824892823668, 1, 1], LineBox[{11, 12}]}, 
      {Hue[0.986003265197232, 1, 1], LineBox[{12, 13}]}, 
      {Hue[0.0009240411120972668, 1, 1], LineBox[{13, 14}]}, 
      {Hue[0.015844817026962454`, 1, 1], LineBox[{14, 15}]}, 
      {Hue[0.03076559294182764, 1, 1], LineBox[{15, 16}]}, 
      {Hue[0.04568636885669283, 1, 1], LineBox[{16, 17}]}, 
      {Hue[0.06060714477155802, 1, 1], LineBox[{17, 18}]}, 
      {Hue[0.07552792068642322, 1, 1], LineBox[{18, 19}]}, 
      {Hue[0.09044869660128839, 1, 1], LineBox[{19, 20}]}, 
      {Hue[0.1053694725161536, 1, 1], LineBox[{20, 21}]}, 
      {Hue[0.12029024843101876`, 1, 1], LineBox[{21, 22}]}, 
      {Hue[0.13521102434588395`, 1, 1], LineBox[{22, 23}]}, 
      {Hue[0.15013180026074915`, 1, 1], LineBox[{23, 24}]}, 
      {Hue[0.16505257617561434`, 1, 1], LineBox[{24, 25}]}, 
      {Hue[0.17997335209047952`, 1, 1], LineBox[{25, 26}]}, 
      {Hue[0.19489412800534475`, 1, 1], LineBox[{26, 27}]}, 
      {Hue[0.2098149039202099, 1, 1], LineBox[{27, 28}]}, 
      {Hue[0.2247356798350751, 1, 1], LineBox[{28, 29}]}, 
      {Hue[0.23965645574994027`, 1, 1], LineBox[{29, 30}]}, 
      {Hue[0.25457723166480545`, 1, 1], LineBox[{30, 31}]}, 
      {Hue[0.26949800757967063`, 1, 1], LineBox[{31, 32}]}, 
      {Hue[0.28441878349453587`, 1, 1], LineBox[{32, 33}]}, 
      {Hue[0.29933955940940105`, 1, 1], LineBox[{33, 34}]}, 
      {Hue[0.3142603353242662, 1, 1], LineBox[{34, 35}]}, 
      {Hue[0.32918111123913146`, 1, 1], LineBox[{35, 36}]}, 
      {Hue[0.34410188715399664`, 1, 1], LineBox[{36, 37}]}, 
      {Hue[0.35902266306886177`, 1, 1], LineBox[{37, 38}]}, 
      {Hue[0.373943438983727, 1, 1], LineBox[{38, 39}]}, 
      {Hue[0.3888642148985922, 1, 1], LineBox[{39, 40}]}, 
      {Hue[0.40378499081345737`, 1, 1], LineBox[{40, 41}]}, 
      {Hue[0.41870576672832255`, 1, 1], LineBox[{41, 42}]}, 
      {Hue[0.4336265426431878, 1, 1], LineBox[{42, 43}]}, 
      {Hue[0.44854731855805297`, 1, 1], LineBox[{43, 44}]}, 
      {Hue[0.46346809447291815`, 1, 1], LineBox[{44, 45}]}, 
      {Hue[0.4783888703877833, 1, 1], LineBox[{45, 46}]}, 
      {Hue[0.4933096463026485, 1, 1], LineBox[{46, 47}]}, 
      {Hue[0.5082304222175136, 1, 1], LineBox[{47, 48}]}, 
      {Hue[0.5231511981323789, 1, 1], LineBox[{48, 49}]}, 
      {Hue[0.5380719740472439, 1, 1], LineBox[{49, 50}]}, 
      {Hue[0.5529927499621092, 1, 1], LineBox[{50, 51}]}, 
      {Hue[0.5679135258769743, 1, 1], LineBox[{51, 52}]}, 
      {Hue[0.5828343017918396, 1, 1], LineBox[{52, 53}]}, 
      {Hue[0.5977550777067047, 1, 1], LineBox[{53, 54}]}, 
      {Hue[0.61267585362157, 1, 1], LineBox[{54, 55}]}, 
      {Hue[0.6275966295364351, 1, 1], LineBox[{55, 56}]}, 
      {Hue[0.6425174054513003, 1, 1], LineBox[{56, 57}]}, 
      {Hue[0.6574381813661655, 1, 1], LineBox[{57, 58}]}, 
      {Hue[0.6723589572810307, 1, 1], LineBox[{58, 59}]}, 
      {Hue[0.6872797331958959, 1, 1], LineBox[{59, 60}]}, 
      {Hue[0.702200509110761, 1, 1], LineBox[{60, 61}]}, 
      {Hue[0.7171212850256263, 1, 1], LineBox[{61, 62}]}, 
      {Hue[0.7320420609404914, 1, 1], LineBox[{62, 63}]}, 
      {Hue[0.7469628368553567, 1, 1], LineBox[{63, 64}]}, 
      {Hue[0.7618836127702218, 1, 1], LineBox[{64, 65}]}, 
      {Hue[0.776804388685087, 1, 1], LineBox[{65, 66}]}, 
      {Hue[0.7917251645999522, 1, 1], LineBox[{66, 67}]}, 
      {Hue[0.8066459405148174, 1, 1], LineBox[{67, 68}]}, 
      {Hue[0.8215667164296826, 1, 1], LineBox[{68, 69}]}, 
      {Hue[0.8364874923445478, 1, 1], LineBox[{69, 70}]}, 
      {Hue[0.851408268259413, 1, 1], LineBox[{70, 71}]}, 
      {Hue[0.8663290441742781, 1, 1], LineBox[{71, 72}]}, 
      {Hue[0.8812498200891433, 1, 1], LineBox[{72, 73}]}, 
      {Hue[0.8961705960040085, 1, 1], LineBox[{73, 74}]}, 
      {Hue[0.9110913719188737, 1, 1], LineBox[{74, 75}]}, 
      {Hue[0.9260121478337389, 1, 1], LineBox[{75, 76}]}, 
      {Hue[0.940932923748604, 1, 1], LineBox[{76, 77}]}, 
      {Hue[0.9558536996634692, 1, 1], LineBox[{77, 78}]}, 
      {Hue[0.9707744755783345, 1, 1], LineBox[{78, 79}]}, 
      {Hue[0.9856952514931997, 1, 1], LineBox[{79, 80}]}, 
      {Hue[0.0006160274080648445, 1, 1], LineBox[{80, 81}]}, 
      {Hue[0.015536803322930032`, 1, 1], LineBox[{81, 82}]}, 
      {Hue[0.030457579237795223`, 1, 1], LineBox[{82, 83}]}, 
      {Hue[0.0453783551526604, 1, 1], LineBox[{83, 84}]}, 
      {Hue[0.0602991310675256, 1, 1], LineBox[{84, 85}]}, 
      {Hue[0.07521990698239078, 1, 1], LineBox[{85, 86}]}, 
      {Hue[0.09014068289725596, 1, 1], LineBox[{86, 87}]}, 
      {Hue[0.10506145881212117`, 1, 1], LineBox[{87, 88}]}, 
      {Hue[0.11998223472698637`, 1, 1], LineBox[{88, 89}]}, 
      {Hue[0.13490301064185153`, 1, 1], LineBox[{89, 90}]}, 
      {Hue[0.14982378655671671`, 1, 1], LineBox[{90, 91}]}, 
      {Hue[0.16474456247158192`, 1, 1], LineBox[{91, 92}]}, 
      {Hue[0.1796653383864471, 1, 1], LineBox[{92, 93}]}, 
      {Hue[0.19458611430131229`, 1, 1], LineBox[{93, 94}]}, 
      {Hue[0.20950689021617747`, 1, 1], LineBox[{94, 95}]}, 
      {Hue[0.22442766613104267`, 1, 1], LineBox[{95, 96}]}, 
      {Hue[0.23934844204590786`, 1, 1], LineBox[{96, 97}]}, 
      {Hue[0.25426921796077306`, 1, 1], LineBox[{97, 98}]}, 
      {Hue[0.26918999387563824`, 1, 1], LineBox[{98, 99}]}, 
      {Hue[0.2841107697905034, 1, 1], LineBox[{99, 100}]}, 
      {Hue[0.2990315457053686, 1, 1], LineBox[{100, 101}]}, 
      {Hue[0.3139523216202338, 1, 1], LineBox[{101, 102}]}, 
      {Hue[0.32887309753509897`, 1, 1], LineBox[{102, 103}]}, 
      {Hue[0.34379387344996415`, 1, 1], LineBox[{103, 104}]}, 
      {Hue[0.3587146493648294, 1, 1], LineBox[{104, 105}]}, 
      {Hue[0.37363542527969457`, 1, 1], LineBox[{105, 106}]}, 
      {Hue[0.38855620119455975`, 1, 1], LineBox[{106, 107}]}, 
      {Hue[0.40347697710942493`, 1, 1], LineBox[{107, 108}]}, 
      {Hue[0.4183977530242901, 1, 1], LineBox[{108, 109}]}, 
      {Hue[0.4333185289391553, 1, 1], LineBox[{109, 110}]}, 
      {Hue[0.44823930485402047`, 1, 1], LineBox[{110, 111}]}, 
      {Hue[0.4631600807688857, 1, 1], LineBox[{111, 112}]}, 
      {Hue[0.4780808566837509, 1, 1], LineBox[{112, 113}]}, 
      {Hue[0.49300163259861607`, 1, 1], LineBox[{113, 114}]}, 
      {Hue[0.5079224085134812, 1, 1], LineBox[{114, 115}]}, 
      {Hue[0.5228431844283464, 1, 1], LineBox[{115, 116}]}, 
      {Hue[0.5377639603432116, 1, 1], LineBox[{116, 117}]}, 
      {Hue[0.5526847362580768, 1, 1], LineBox[{117, 118}]}, 
      {Hue[0.567605512172942, 1, 1], LineBox[{118, 119}]}, 
      {Hue[0.582526288087807, 1, 1], LineBox[{119, 120}]}, 
      {Hue[0.5974470640026723, 1, 1], LineBox[{120, 121}]}, 
      {Hue[0.6123678399175374, 1, 1], LineBox[{121, 122}]}, 
      {Hue[0.6272886158324027, 1, 1], LineBox[{122, 123}]}, 
      {Hue[0.6422093917472679, 1, 1], LineBox[{123, 124}]}, 
      {Hue[0.6571301676621331, 1, 1], LineBox[{124, 125}]}, 
      {Hue[0.6720509435769983, 1, 1], LineBox[{125, 126}]}, 
      {Hue[0.6869717194918634, 1, 1], LineBox[{126, 127}]}, 
      {Hue[0.7018924954067287, 1, 1], LineBox[{127, 128}]}, 
      {Hue[0.7168132713215938, 1, 1], LineBox[{128, 129}]}, 
      {Hue[0.7317340472364591, 1, 1], LineBox[{129, 130}]}, 
      {Hue[0.7466548231513241, 1, 1], LineBox[{130, 131}]}, 
      {Hue[0.7615755990661894, 1, 1], LineBox[{131, 132}]}, 
      {Hue[0.7764963749810546, 1, 1], LineBox[{132, 133}]}, 
      {Hue[0.7914171508959198, 1, 1], LineBox[{133, 134}]}, 
      {Hue[0.806337926810785, 1, 1], LineBox[{134, 135}]}, 
      {Hue[0.8212587027256502, 1, 1], LineBox[{135, 136}]}, 
      {Hue[0.8361794786405153, 1, 1], LineBox[{136, 137}]}, 
      {Hue[0.8511002545553805, 1, 1], LineBox[{137, 138}]}, 
      {Hue[0.8660210304702457, 1, 1], LineBox[{138, 139}]}, 
      {Hue[0.8809418063851109, 1, 1], LineBox[{139, 140}]}, 
      {Hue[0.8958625822999761, 1, 1], LineBox[{140, 141}]}, 
      {Hue[0.9107833582148412, 1, 1], LineBox[{141, 142}]}, 
      {Hue[0.9257041341297065, 1, 1], LineBox[{142, 143}]}, 
      {Hue[0.9406249100445717, 1, 1], LineBox[{143, 144}]}, 
      {Hue[0.9555456859594369, 1, 1], LineBox[{144, 145}]}, 
      {Hue[0.9704664618743021, 1, 1], LineBox[{145, 146}]}, 
      {Hue[0.9853872377891673, 1, 1], LineBox[{146, 147}]}, 
      {Hue[0.0003080137040324222, 1, 1], LineBox[{147, 148}]}, 
      {Hue[0.01522878961889761, 1, 1], LineBox[{148, 149}]}, 
      {Hue[0.0301495655337628, 1, 1], LineBox[{149, 150}]}, 
      {Hue[0.045070341448627996`, 1, 1], LineBox[{150, 151}]}, 
      {Hue[0.059991117363493184`, 1, 1], LineBox[{151, 152}]}, 
      {Hue[0.07491189327835837, 1, 1], LineBox[{152, 153}]}, 
      {Hue[0.08983266919322357, 1, 1], LineBox[{153, 154}]}, 
      {Hue[0.10475344510808872`, 1, 1], LineBox[{154, 155}]}, 
      {Hue[0.11967422102295393`, 1, 1], LineBox[{155, 156}]}, 
      {Hue[0.13459499693781912`, 1, 1], LineBox[{156, 157}]}, 
      {Hue[0.1495157728526843, 1, 1], LineBox[{157, 158}]}, 
      {Hue[0.16443654876754946`, 1, 1], LineBox[{158, 159}]}, 
      {Hue[0.1793573246824147, 1, 1], LineBox[{159, 160}]}, 
      {Hue[0.19427810059727987`, 1, 1], LineBox[{160, 161}]}, 
      {Hue[0.20919887651214505`, 1, 1], LineBox[{161, 162}]}, 
      {Hue[0.22411965242701024`, 1, 1], LineBox[{162, 163}]}, 
      {Hue[0.23904042834187544`, 1, 1], LineBox[{163, 164}]}, 
      {Hue[0.2539612042567406, 1, 1], LineBox[{164, 165}]}, 
      {Hue[0.2688819801716058, 1, 1], LineBox[{165, 166}]}, 
      {Hue[0.283802756086471, 1, 1], LineBox[{166, 167}]}, 
      {Hue[0.29872353200133617`, 1, 1], LineBox[{167, 168}]}, 
      {Hue[0.3136443079162014, 1, 1], LineBox[{168, 169}]}, 
      {Hue[0.32856508383106653`, 1, 1], LineBox[{169, 170}]}, 
      {Hue[0.34348585974593177`, 1, 1], LineBox[{170, 171}]}, 
      {Hue[0.358406635660797, 1, 1], LineBox[{171, 172}]}, 
      {Hue[0.3733274115756622, 1, 1], LineBox[{172, 173}]}, 
      {Hue[0.38824818749052736`, 1, 1], LineBox[{173, 174}]}, 
      {Hue[0.40316896340539254`, 1, 1], LineBox[{174, 175}]}, 
      {Hue[0.4180897393202577, 1, 1], LineBox[{175, 176}]}, 
      {Hue[0.4330105152351229, 1, 1], LineBox[{176, 177}]}, 
      {Hue[0.44793129114998803`, 1, 1], LineBox[{177, 178}]}, 
      {Hue[0.4628520670648533, 1, 1], LineBox[{178, 179}]}, 
      {Hue[0.4777728429797184, 1, 1], LineBox[{179, 180}]}, 
      {Hue[0.4926936188945837, 1, 1], LineBox[{180, 181}]}, 
      {Hue[0.5076143948094487, 1, 1], LineBox[{181, 182}]}, 
      {Hue[0.522535170724314, 1, 1], LineBox[{182, 183}]}, 
      {Hue[0.5374559466391792, 1, 1], LineBox[{183, 184}]}, 
      {Hue[0.5523767225540444, 1, 1], LineBox[{184, 185}]}, 
      {Hue[0.5672974984689095, 1, 1], LineBox[{185, 186}]}, 
      {Hue[0.5822182743837747, 1, 1], LineBox[{186, 187}]}, 
      {Hue[0.5971390502986399, 1, 1], LineBox[{187, 188}]}, 
      {Hue[0.6120598262135051, 1, 1], LineBox[{188, 189}]}, 
      {Hue[0.6269806021283703, 1, 1], LineBox[{189, 190}]}, 
      {Hue[0.6419013780432354, 1, 1], LineBox[{190, 191}]}, 
      {Hue[0.6568221539581007, 1, 1], LineBox[{191, 192}]}, 
      {Hue[0.6717429298729658, 1, 1], LineBox[{192, 193}]}, 
      {Hue[0.6866637057878311, 1, 1], LineBox[{193, 194}]}, 
      {Hue[0.7015844817026962, 1, 1], LineBox[{194, 195}]}, 
      {Hue[0.7165052576175615, 1, 1], LineBox[{195, 196}]}, 
      {Hue[0.7314260335324265, 1, 1], LineBox[{196, 197}]}, 
      {Hue[0.7463468094472918, 1, 1], LineBox[{197, 198}]}, 
      {Hue[0.761267585362157, 1, 1], LineBox[{198, 199}]}, 
      {Hue[0.7761883612770222, 1, 1], LineBox[{199, 200}]}, 
      {Hue[0.7911091371918874, 1, 1], LineBox[{200, 201}]}, 
      {Hue[0.8060299131067525, 1, 1], LineBox[{201, 202}]}, 
      {Hue[0.8209506890216177, 1, 1], LineBox[{202, 203}]}, 
      {Hue[0.8358714649364829, 1, 1], LineBox[{203, 204}]}, 
      {Hue[0.8507922408513481, 1, 1], LineBox[{204, 205}]}, 
      {Hue[0.8657130167662133, 1, 1], LineBox[{205, 206}]}, 
      {Hue[0.8806337926810786, 1, 1], LineBox[{206, 207}]}, 
      {Hue[0.8955545685959436, 1, 1], LineBox[{207, 208}]}, 
      {Hue[0.9104753445108089, 1, 1], LineBox[{208, 209}]}, 
      {Hue[0.9253961204256741, 1, 1], LineBox[{209, 210}]}, 
      {Hue[0.9403168963405393, 1, 1], LineBox[{210, 211}]}, 
      {Hue[0.9552376722554045, 1, 1], LineBox[{211, 212}]}, 
      {Hue[0.9701584481702696, 1, 1], LineBox[{212, 213}]}, 
      {Hue[0.9850792240851348, 1, 1], LineBox[{213, 214}]}, 
      {Hue[0., 1, 1], LineBox[{214, 215}]}, 
      {Hue[0.014920775914865188`, 1, 1], LineBox[{215, 216}]}, 
      {Hue[0.029841551829730376`, 1, 1], LineBox[{216, 217}]}, 
      {Hue[0.044762327744595556`, 1, 1], LineBox[{217, 218}]}, 
      {Hue[0.05968310365946075, 1, 1], LineBox[{218, 219}]}, 
      {Hue[0.07460387957432593, 1, 1], LineBox[{219, 220}]}, 
      {Hue[0.08952465548919111, 1, 1], LineBox[{220, 221}]}, 
      {Hue[0.10444543140405632`, 1, 1], LineBox[{221, 222}]}, 
      {Hue[0.11936620731892149`, 1, 1], LineBox[{222, 223}]}, 
      {Hue[0.1342869832337867, 1, 1], LineBox[{223, 224}]}, 
      {Hue[0.1492077591486519, 1, 1], LineBox[{224, 225}]}, 
      {Hue[0.16412853506351707`, 1, 1], LineBox[{225, 226}]}, 
      {Hue[0.17904931097838225`, 1, 1], LineBox[{226, 227}]}, 
      {Hue[0.19397008689324746`, 1, 1], LineBox[{227, 228}]}, 
      {Hue[0.20889086280811264`, 1, 1], LineBox[{228, 229}]}, 
      {Hue[0.22381163872297782`, 1, 1], LineBox[{229, 230}]}, 
      {Hue[0.238732414637843, 1, 1], LineBox[{230, 231}]}, 
      {Hue[0.2536531905527082, 1, 1], LineBox[{231, 232}]}, 
      {Hue[0.2685739664675734, 1, 1], LineBox[{232, 233}]}, 
      {Hue[0.2834947423824386, 1, 1], LineBox[{233, 234}]}, 
      {Hue[0.2984155182973038, 1, 1], LineBox[{234, 235}]}, 
      {Hue[0.31333629421216896`, 1, 1], LineBox[{235, 236}]}, 
      {Hue[0.32825707012703415`, 1, 1], LineBox[{236, 237}]}, 
      {Hue[0.3431778460418993, 1, 1], LineBox[{237, 238}]}, 
      {Hue[0.3580986219567645, 1, 1], LineBox[{238, 239}]}, 
      {Hue[0.37301939787162974`, 1, 1], LineBox[{239, 240}]}, 
      {Hue[0.3879401737864949, 1, 1], LineBox[{240, 241}]}, 
      {Hue[0.4028609497013601, 1, 1], LineBox[{241, 242}]}, 
      {Hue[0.4177817256162253, 1, 1], LineBox[{242, 243}]}, 
      {Hue[0.43270250153109047`, 1, 1], LineBox[{243, 244}]}, 
      {Hue[0.44762327744595565`, 1, 1], LineBox[{244, 245}]}, 
      {Hue[0.46254405336082083`, 1, 1], LineBox[{245, 246}]}, 
      {Hue[0.477464829275686, 1, 1], LineBox[{246, 247}]}, 
      {Hue[0.49238560519055125`, 1, 1], LineBox[{247, 248}]}, 
      {Hue[0.5073063811054164, 1, 1], LineBox[{248, 249}]}, 
      {Hue[0.5222271570202817, 1, 1], LineBox[{249, 250}]}, 
      {Hue[0.5371479329351467, 1, 1], LineBox[{250, 251}]}, 
      {Hue[0.552068708850012, 1, 1], LineBox[{251, 252}]}, 
      {Hue[0.5669894847648771, 1, 1], LineBox[{252, 253}]}, 
      {Hue[0.5819102606797423, 1, 1], LineBox[{253, 254}]}, 
      {Hue[0.5968310365946075, 1, 1], LineBox[{254, 255}]}, 
      {Hue[0.6117518125094726, 1, 1], LineBox[{255, 256}]}, 
      {Hue[0.6266725884243378, 1, 1], LineBox[{256, 257}]}, 
      {Hue[0.641593364339203, 1, 1], LineBox[{257, 258}]}, 
      {Hue[0.6565141402540682, 1, 1], LineBox[{258, 259}]}, 
      {Hue[0.6714349161689335, 1, 1], LineBox[{259, 260}]}, 
      {Hue[0.6863556920837985, 1, 1], LineBox[{260, 261}]}, 
      {Hue[0.7012764679986638, 1, 1], LineBox[{261, 262}]}, 
      {Hue[0.716197243913529, 1, 1], LineBox[{262, 263}]}, 
      {Hue[0.7311180198283942, 1, 1], LineBox[{263, 264}]}, 
      {Hue[0.7460387957432594, 1, 1], LineBox[{264, 265}]}, 
      {Hue[0.7609595716581246, 1, 1], LineBox[{265, 266}]}, 
      {Hue[0.7758803475729897, 1, 1], LineBox[{266, 267}]}, 
      {Hue[0.7908011234878549, 1, 1], LineBox[{267, 268}]}, 
      {Hue[0.8057218994027201, 1, 1], LineBox[{268, 269}]}, 
      {Hue[0.8206426753175853, 1, 1], LineBox[{269, 270}]}, 
      {Hue[0.8355634512324506, 1, 1], LineBox[{270, 271}]}, 
      {Hue[0.8504842271473156, 1, 1], LineBox[{271, 272}]}, 
      {Hue[0.8654050030621809, 1, 1], LineBox[{272, 273}]}, 
      {Hue[0.8803257789770461, 1, 1], LineBox[{273, 274}]}, 
      {Hue[0.8952465548919113, 1, 1], LineBox[{274, 275}]}, 
      {Hue[0.9101673308067765, 1, 1], LineBox[{275, 276}]}, 
      {Hue[0.9250881067216417, 1, 1], LineBox[{276, 277}]}, 
      {Hue[0.9400088826365068, 1, 1], LineBox[{277, 278}]}, 
      {Hue[0.954929658551372, 1, 1], LineBox[{278, 279}]}, 
      {Hue[0.9698504344662372, 1, 1], LineBox[{279, 280}]}, 
      {Hue[0.9847712103811024, 1, 1], LineBox[{280, 281}]}, 
      {Hue[0.9996919862959676, 1, 1], LineBox[{281, 282}]}, 
      {Hue[0.014612762210832766`, 1, 1], LineBox[{282, 283}]}, 
      {Hue[0.029533538125697957`, 1, 1], LineBox[{283, 284}]}, 
      {Hue[0.044454314040563145`, 1, 1], LineBox[{284, 285}]}, 
      {Hue[0.05937508995542833, 1, 1], LineBox[{285, 286}]}, 
      {Hue[0.07429586587029352, 1, 1], LineBox[{286, 287}]}, 
      {Hue[0.08921664178515872, 1, 1], LineBox[{287, 288}]}, 
      {Hue[0.1041374177000239, 1, 1], LineBox[{288, 289}]}, 
      {Hue[0.11905819361488908`, 1, 1], LineBox[{289, 290}]}, 
      {Hue[0.13397896952975427`, 1, 1], LineBox[{290, 291}]}, 
      {Hue[0.14889974544461948`, 1, 1], LineBox[{291, 292}]}, 
      {Hue[0.16382052135948466`, 1, 1], LineBox[{292, 293}]}, 
      {Hue[0.17874129727434984`, 1, 1], LineBox[{293, 294}]}, 
      {Hue[0.19366207318921502`, 1, 1], LineBox[{294, 295}]}, 
      {Hue[0.20858284910408023`, 1, 1], LineBox[{295, 296}]}, 
      {Hue[0.2235036250189454, 1, 1], LineBox[{296, 297}]}, 
      {Hue[0.2384244009338106, 1, 1], LineBox[{297, 298}]}, 
      {Hue[0.2533451768486758, 1, 1], LineBox[{298, 299}]}, 
      {Hue[0.268265952763541, 1, 1], LineBox[{299, 300}]}, 
      {Hue[0.28318672867840616`, 1, 1], LineBox[{300, 301}]}, 
      {Hue[0.29810750459327134`, 1, 1], LineBox[{301, 302}]}, 
      {Hue[0.3130282805081365, 1, 1], LineBox[{302, 303}]}, 
      {Hue[0.3279490564230017, 1, 1], LineBox[{303, 304}]}, 
      {Hue[0.34286983233786694`, 1, 1], LineBox[{304, 305}]}, 
      {Hue[0.3577906082527321, 1, 1], LineBox[{305, 306}]}, 
      {Hue[0.37271138416759725`, 1, 1], LineBox[{306, 307}]}, 
      {Hue[0.38763216008246254`, 1, 1], LineBox[{307, 308}]}, 
      {Hue[0.4025529359973277, 1, 1], LineBox[{308, 309}]}, 
      {Hue[0.4174737119121929, 1, 1], LineBox[{309, 310}]}, 
      {Hue[0.432394487827058, 1, 1], LineBox[{310, 311}]}, 
      {Hue[0.4473152637419232, 1, 1], LineBox[{311, 312}]}, 
      {Hue[0.4622360396567884, 1, 1], LineBox[{312, 313}]}, 
      {Hue[0.47715681557165357`, 1, 1], LineBox[{313, 314}]}, 
      {Hue[0.49207759148651875`, 1, 1], LineBox[{314, 315}]}, 
      {Hue[0.5069983674013839, 1, 1], LineBox[{315, 316}]}, 
      {Hue[0.5219191433162491, 1, 1], LineBox[{316, 317}]}, 
      {Hue[0.5368399192311143, 1, 1], LineBox[{317, 318}]}, 
      {Hue[0.5517606951459795, 1, 1], LineBox[{318, 319}]}, 
      {Hue[0.5666814710608448, 1, 1], LineBox[{319, 320}]}, 
      {Hue[0.5816022469757098, 1, 1], LineBox[{320, 321}]}, 
      {Hue[0.5965230228905751, 1, 1], LineBox[{321, 322}]}, 
      {Hue[0.6114437988054402, 1, 1], LineBox[{322, 323}]}, 
      {Hue[0.6263645747203055, 1, 1], LineBox[{323, 324}]}, 
      {Hue[0.6412853506351706, 1, 1], LineBox[{324, 325}]}, 
      {Hue[0.6562061265500359, 1, 1], LineBox[{325, 326}]}, 
      {Hue[0.671126902464901, 1, 1], LineBox[{326, 327}]}, 
      {Hue[0.6860476783797662, 1, 1], LineBox[{327, 328}]}, 
      {Hue[0.7009684542946314, 1, 1], LineBox[{328, 329}]}, 
      {Hue[0.7158892302094966, 1, 1], LineBox[{329, 330}]}, 
      {Hue[0.7308100061243618, 1, 1], LineBox[{330, 331}]}, 
      {Hue[0.7457307820392269, 1, 1], LineBox[{331, 332}]}, 
      {Hue[0.7606515579540921, 1, 1], LineBox[{332, 333}]}, 
      {Hue[0.7755723338689573, 1, 1], LineBox[{333, 334}]}, 
      {Hue[0.7904931097838226, 1, 1], LineBox[{334, 335}]}, 
      {Hue[0.8054138856986877, 1, 1], LineBox[{335, 336}]}, 
      {Hue[0.820334661613553, 1, 1], LineBox[{336, 337}]}, 
      {Hue[0.835255437528418, 1, 1], LineBox[{337, 338}]}, 
      {Hue[0.8501762134432833, 1, 1], LineBox[{338, 339}]}, 
      {Hue[0.8650969893581485, 1, 1], LineBox[{339, 340}]}, 
      {Hue[0.8800177652730137, 1, 1], LineBox[{340, 341}]}, 
      {Hue[0.8949385411878789, 1, 1], LineBox[{341, 342}]}, 
      {Hue[0.909859317102744, 1, 1], LineBox[{342, 343}]}, 
      {Hue[0.9247800930176092, 1, 1], LineBox[{343, 344}]}, 
      {Hue[0.9397008689324744, 1, 1], LineBox[{344, 345}]}, 
      {Hue[0.9546216448473396, 1, 1], LineBox[{345, 346}]}, 
      {Hue[0.9695424207622048, 1, 1], LineBox[{346, 347}]}, 
      {Hue[0.9844631966770699, 1, 1], LineBox[{347, 348}]}, 
      {Hue[0.9993839725919351, 1, 1], LineBox[{348, 349}]}, 
      {Hue[0.014304748506800346`, 1, 1], LineBox[{349, 350}]}, 
      {Hue[0.02922552442166553, 1, 1], LineBox[{350, 351}]}, 
      {Hue[0.04414630033653072, 1, 1], LineBox[{351, 352}]}, 
      {Hue[0.059067076251395914`, 1, 1], LineBox[{352, 353}]}, 
      {Hue[0.0739878521662611, 1, 1], LineBox[{353, 354}]}, 
      {Hue[0.08890862808112629, 1, 1], LineBox[{354, 355}]}, 
      {Hue[0.10382940399599148`, 1, 1], LineBox[{355, 356}]}, 
      {Hue[0.11875017991085667`, 1, 1], LineBox[{356, 357}]}, 
      {Hue[0.13367095582572183`, 1, 1], LineBox[{357, 358}]}, 
      {Hue[0.14859173174058704`, 1, 1], LineBox[{358, 359}]}, 
      {Hue[0.16351250765545225`, 1, 1], LineBox[{359, 360}]}, 
      {Hue[0.17843328357031743`, 1, 1], LineBox[{360, 361}]}, 
      {Hue[0.1933540594851826, 1, 1], LineBox[{361, 362}]}, 
      {Hue[0.2082748354000478, 1, 1], LineBox[{362, 363}]}, 
      {Hue[0.223195611314913, 1, 1], LineBox[{363, 364}]}, 
      {Hue[0.23811638722977818`, 1, 1], LineBox[{364, 365}]}, 
      {Hue[0.25303716314464336`, 1, 1], LineBox[{365, 366}]}, 
      {Hue[0.26795793905950854`, 1, 1], LineBox[{366, 367}]}, 
      {Hue[0.2828787149743737, 1, 1], LineBox[{367, 368}]}, 
      {Hue[0.29779949088923896`, 1, 1], LineBox[{368, 369}]}, 
      {Hue[0.31272026680410414`, 1, 1], LineBox[{369, 370}]}, 
      {Hue[0.3276410427189693, 1, 1], LineBox[{370, 371}]}, 
      {Hue[0.3425618186338345, 1, 1], LineBox[{371, 372}]}, 
      {Hue[0.3574825945486997, 1, 1], LineBox[{372, 373}]}, 
      {Hue[0.37240337046356486`, 1, 1], LineBox[{373, 374}]}, 
      {Hue[0.38732414637843005`, 1, 1], LineBox[{374, 375}]}, 
      {Hue[0.4022449222932953, 1, 1], LineBox[{375, 376}]}, 
      {Hue[0.41716569820816046`, 1, 1], LineBox[{376, 377}]}, 
      {Hue[0.43208647412302564`, 1, 1], LineBox[{377, 378}]}, 
      {Hue[0.4470072500378908, 1, 1], LineBox[{378, 379}]}, 
      {Hue[0.461928025952756, 1, 1], LineBox[{379, 380}]}, 
      {Hue[0.4768488018676212, 1, 1], LineBox[{380, 381}]}, 
      {Hue[0.49176957778248637`, 1, 1], LineBox[{381, 382}]}, 
      {Hue[0.5066903536973515, 1, 1], LineBox[{382, 383}]}, 
      {Hue[0.5216111296122167, 1, 1], LineBox[{383, 384}]}, 
      {Hue[0.5365319055270819, 1, 1], LineBox[{384, 385}]}, 
      {Hue[0.551452681441947, 1, 1], LineBox[{385, 386}]}, 
      {Hue[0.5663734573568122, 1, 1], LineBox[{386, 387}]}, 
      {Hue[0.5812942332716775, 1, 1], LineBox[{387, 388}]}, 
      {Hue[0.5962150091865426, 1, 1], LineBox[{388, 389}]}, 
      {Hue[0.6111357851014079, 1, 1], LineBox[{389, 390}]}, 
      {Hue[0.6260565610162729, 1, 1], LineBox[{390, 391}]}, 
      {Hue[0.6409773369311382, 1, 1], LineBox[{391, 392}]}, 
      {Hue[0.6558981128460033, 1, 1], LineBox[{392, 393}]}, 
      {Hue[0.6708188887608686, 1, 1], LineBox[{393, 394}]}, 
      {Hue[0.6857396646757338, 1, 1], LineBox[{394, 395}]}, 
      {Hue[0.700660440590599, 1, 1], LineBox[{395, 396}]}, 
      {Hue[0.7155812165054641, 1, 1], LineBox[{396, 397}]}, 
      {Hue[0.7305019924203293, 1, 1], LineBox[{397, 398}]}, 
      {Hue[0.7454227683351946, 1, 1], LineBox[{398, 399}]}, 
      {Hue[0.7603435442500597, 1, 1], LineBox[{399, 400}]}, 
      {Hue[0.775264320164925, 1, 1], LineBox[{400, 401}]}, 
      {Hue[0.79018509607979, 1, 1], LineBox[{401, 402}]}, 
      {Hue[0.8051058719946552, 1, 1], LineBox[{402, 403}]}, 
      {Hue[0.8200266479095205, 1, 1], LineBox[{403, 404}]}, 
      {Hue[0.8349474238243857, 1, 1], LineBox[{404, 405}]}, 
      {Hue[0.8498681997392509, 1, 1], LineBox[{405, 406}]}, 
      {Hue[0.864788975654116, 1, 1], LineBox[{406, 407}]}, 
      {Hue[0.8797097515689812, 1, 1], LineBox[{407, 408}]}, 
      {Hue[0.8946305274838464, 1, 1], LineBox[{408, 409}]}, 
      {Hue[0.9095513033987116, 1, 1], LineBox[{409, 410}]}, 
      {Hue[0.9244720793135768, 1, 1], LineBox[{410, 411}]}, 
      {Hue[0.939392855228442, 1, 1], LineBox[{411, 412}]}, 
      {Hue[0.9543136311433071, 1, 1], LineBox[{412, 413}]}, 
      {Hue[0.9692344070581723, 1, 1], LineBox[{413, 414}]}, 
      {Hue[0.9841551829730375, 1, 1], LineBox[{414, 415}]}, 
      {Hue[0.9990759588879027, 1, 1], LineBox[{415, 416}]}, 
      {Hue[0.013996734802767922`, 1, 1], LineBox[{416, 417}]}, 
      {Hue[0.028917510717633117`, 1, 1], LineBox[{417, 418}]}, 
      {Hue[0.0438382866324983, 1, 1], LineBox[{418, 419}]}, 
      {Hue[0.05875906254736348, 1, 1], LineBox[{419, 420}]}, 
      {Hue[0.07367983846222867, 1, 1], LineBox[{420, 421}]}, 
      {Hue[0.08860061437709388, 1, 1], LineBox[{421, 422}]}, 
      {Hue[0.10352139029195906`, 1, 1], LineBox[{422, 423}]}, 
      {Hue[0.11844216620682425`, 1, 1], LineBox[{423, 424}]}, 
      {Hue[0.13336294212168942`, 1, 1], LineBox[{424, 425}]}, 
      {Hue[0.14828371803655466`, 1, 1], LineBox[{425, 426}]}, 
      {Hue[0.1632044939514198, 1, 1], LineBox[{426, 427}]}, 
      {Hue[0.17812526986628502`, 1, 1], LineBox[{427, 428}]}, 
      {Hue[0.19304604578115017`, 1, 1], LineBox[{428, 429}]}, 
      {Hue[0.20796682169601538`, 1, 1], LineBox[{429, 430}]}, 
      {Hue[0.22288759761088056`, 1, 1], LineBox[{430, 431}]}, 
      {Hue[0.23780837352574577`, 1, 1], LineBox[{431, 432}]}, 
      {Hue[0.252729149440611, 1, 1], LineBox[{432, 433}]}, 
      {Hue[0.26764992535547616`, 1, 1], LineBox[{433, 434}]}, 
      {Hue[0.28257070127034134`, 1, 1], LineBox[{434, 435}]}, 
      {Hue[0.29749147718520647`, 1, 1], LineBox[{435, 436}]}, 
      {Hue[0.3124122531000717, 1, 1], LineBox[{436, 437}]}, 
      {Hue[0.3273330290149369, 1, 1], LineBox[{437, 438}]}, 
      {Hue[0.34225380492980206`, 1, 1], LineBox[{438, 439}]}, 
      {Hue[0.35717458084466724`, 1, 1], LineBox[{439, 440}]}, 
      {Hue[0.3720953567595324, 1, 1], LineBox[{440, 441}]}, 
      {Hue[0.3870161326743976, 1, 1], LineBox[{441, 442}]}, 
      {Hue[0.4019369085892628, 1, 1], LineBox[{442, 443}]}, 
      {Hue[0.41685768450412797`, 1, 1], LineBox[{443, 444}]}, 
      {Hue[0.4317784604189932, 1, 1], LineBox[{444, 445}]}, 
      {Hue[0.4466992363338584, 1, 1], LineBox[{445, 446}]}, 
      {Hue[0.46162001224872357`, 1, 1], LineBox[{446, 447}]}, 
      {Hue[0.47654078816358875`, 1, 1], LineBox[{447, 448}]}, 
      {Hue[0.4914615640784539, 1, 1], LineBox[{448, 449}]}, 
      {Hue[0.5063823399933192, 1, 1], LineBox[{449, 450}]}, 
      {Hue[0.5213031159081842, 1, 1], LineBox[{450, 451}]}, 
      {Hue[0.5362238918230495, 1, 1], LineBox[{451, 452}]}, 
      {Hue[0.5511446677379146, 1, 1], LineBox[{452, 453}]}, 
      {Hue[0.5660654436527799, 1, 1], LineBox[{453, 454}]}, 
      {Hue[0.5809862195676451, 1, 1], LineBox[{454, 455}]}, 
      {Hue[0.5959069954825102, 1, 1], LineBox[{455, 456}]}, 
      {Hue[0.6108277713973754, 1, 1], LineBox[{456, 457}]}, 
      {Hue[0.6257485473122406, 1, 1], LineBox[{457, 458}]}, 
      {Hue[0.6406693232271058, 1, 1], LineBox[{458, 459}]}, 
      {Hue[0.655590099141971, 1, 1], LineBox[{459, 460}]}, 
      {Hue[0.6705108750568362, 1, 1], LineBox[{460, 461}]}, 
      {Hue[0.6854316509717013, 1, 1], LineBox[{461, 462}]}, 
      {Hue[0.7003524268865666, 1, 1], LineBox[{462, 463}]}, 
      {Hue[0.7152732028014317, 1, 1], LineBox[{463, 464}]}, 
      {Hue[0.730193978716297, 1, 1], LineBox[{464, 465}]}, 
      {Hue[0.7451147546311621, 1, 1], LineBox[{465, 466}]}, 
      {Hue[0.7600355305460273, 1, 1], LineBox[{466, 467}]}, 
      {Hue[0.7749563064608925, 1, 1], LineBox[{467, 468}]}, 
      {Hue[0.7898770823757577, 1, 1], LineBox[{468, 469}]}, 
      {Hue[0.8047978582906228, 1, 1], LineBox[{469, 470}]}, 
      {Hue[0.8197186342054881, 1, 1], LineBox[{470, 471}]}, 
      {Hue[0.8346394101203533, 1, 1], LineBox[{471, 472}]}, 
      {Hue[0.8495601860352184, 1, 1], LineBox[{472, 473}]}, 
      {Hue[0.8644809619500836, 1, 1], LineBox[{473, 474}]}, 
      {Hue[0.8794017378649488, 1, 1], LineBox[{474, 475}]}, 
      {Hue[0.894322513779814, 1, 1], LineBox[{475, 476}]}, 
      {Hue[0.9092432896946792, 1, 1], LineBox[{476, 477}]}, 
      {Hue[0.9241640656095443, 1, 1], LineBox[{477, 478}]}, 
      {Hue[0.9390848415244095, 1, 1], LineBox[{478, 479}]}, 
      {Hue[0.9540056174392747, 1, 1], LineBox[{479, 480}]}, 
      {Hue[0.96892639335414, 1, 1], LineBox[{480, 481}]}, 
      {Hue[0.9838471692690052, 1, 1], LineBox[{481, 482}]}, 
      {Hue[0.9987679451838704, 1, 1], LineBox[{482, 483}]}, 
      {Hue[0.013688721098735502`, 1, 1], LineBox[{483, 484}]}, 
      {Hue[0.02860949701360068, 1, 1], LineBox[{484, 485}]}, 
      {Hue[0.043530272928465875`, 1, 1], LineBox[{485, 486}]}, 
      {Hue[0.05845104884333107, 1, 1], LineBox[{486, 487}]}, 
      {Hue[0.07337182475819626, 1, 1], LineBox[{487, 488}]}, 
      {Hue[0.08829260067306144, 1, 1], LineBox[{488, 489}]}, 
      {Hue[0.1032133765879266, 1, 1], LineBox[{489, 490}]}, 
      {Hue[0.11813415250279183`, 1, 1], LineBox[{490, 491}]}, 
      {Hue[0.133054928417657, 1, 1], LineBox[{491, 492}]}, 
      {Hue[0.1479757043325222, 1, 1], LineBox[{492, 493}]}, 
      {Hue[0.16289648024738737`, 1, 1], LineBox[{493, 494}]}, 
      {Hue[0.17781725616225258`, 1, 1], LineBox[{494, 495}]}, 
      {Hue[0.1927380320771178, 1, 1], LineBox[{495, 496}]}, 
      {Hue[0.20765880799198297`, 1, 1], LineBox[{496, 497}]}, 
      {Hue[0.22257958390684815`, 1, 1], LineBox[{497, 498}]}, 
      {Hue[0.23750035982171333`, 1, 1], LineBox[{498, 499}]}, 
      {Hue[0.25242113573657854`, 1, 1], LineBox[{499, 500}]}, 
      {Hue[0.2673419116514437, 1, 1], LineBox[{500, 501}]}, 
      {Hue[0.2822626875663089, 1, 1], LineBox[{501, 502}]}, 
      {Hue[0.2971834634811741, 1, 1], LineBox[{502, 503}]}, 
      {Hue[0.31210423939603926`, 1, 1], LineBox[{503, 504}]}, 
      {Hue[0.3270250153109045, 1, 1], LineBox[{504, 505}]}, 
      {Hue[0.3419457912257697, 1, 1], LineBox[{505, 506}]}, 
      {Hue[0.35686656714063486`, 1, 1], LineBox[{506, 507}]}, 
      {Hue[0.37178734305550004`, 1, 1], LineBox[{507, 508}]}, 
      {Hue[0.3867081189703652, 1, 1], LineBox[{508, 509}]}, 
      {Hue[0.4016288948852304, 1, 1], LineBox[{509, 510}]}, 
      {Hue[0.4165496708000956, 1, 1], LineBox[{510, 511}]}, 
      {Hue[0.4314704467149608, 1, 1], LineBox[{511, 512}]}}}},
   VertexColors->{
     Hue[0.821874730133715, 1, 1], 
     Hue[0.8367955060485802, 1, 1], 
     Hue[0.8517162819634454, 1, 1], 
     Hue[0.8666370578783106, 1, 1], 
     Hue[0.8815578337931758, 1, 1], 
     Hue[0.8964786097080409, 1, 1], 
     Hue[0.9113993856229061, 1, 1], 
     Hue[0.9263201615377713, 1, 1], 
     Hue[0.9412409374526365, 1, 1], 
     Hue[0.9561617133675017, 1, 1], 
     Hue[0.9710824892823668, 1, 1], 
     Hue[0.986003265197232, 1, 1], 
     Hue[0.0009240411120972668, 1, 1], 
     Hue[0.015844817026962454`, 1, 1], 
     Hue[0.03076559294182764, 1, 1], 
     Hue[0.04568636885669283, 1, 1], 
     Hue[0.06060714477155802, 1, 1], 
     Hue[0.07552792068642322, 1, 1], 
     Hue[0.09044869660128839, 1, 1], 
     Hue[0.1053694725161536, 1, 1], 
     Hue[0.12029024843101876`, 1, 1], 
     Hue[0.13521102434588395`, 1, 1], 
     Hue[0.15013180026074915`, 1, 1], 
     Hue[0.16505257617561434`, 1, 1], 
     Hue[0.17997335209047952`, 1, 1], 
     Hue[0.19489412800534475`, 1, 1], 
     Hue[0.2098149039202099, 1, 1], 
     Hue[0.2247356798350751, 1, 1], 
     Hue[0.23965645574994027`, 1, 1], 
     Hue[0.25457723166480545`, 1, 1], 
     Hue[0.26949800757967063`, 1, 1], 
     Hue[0.28441878349453587`, 1, 1], 
     Hue[0.29933955940940105`, 1, 1], 
     Hue[0.3142603353242662, 1, 1], 
     Hue[0.32918111123913146`, 1, 1], 
     Hue[0.34410188715399664`, 1, 1], 
     Hue[0.35902266306886177`, 1, 1], 
     Hue[0.373943438983727, 1, 1], 
     Hue[0.3888642148985922, 1, 1], 
     Hue[0.40378499081345737`, 1, 1], 
     Hue[0.41870576672832255`, 1, 1], 
     Hue[0.4336265426431878, 1, 1], 
     Hue[0.44854731855805297`, 1, 1], 
     Hue[0.46346809447291815`, 1, 1], 
     Hue[0.4783888703877833, 1, 1], 
     Hue[0.4933096463026485, 1, 1], 
     Hue[0.5082304222175136, 1, 1], 
     Hue[0.5231511981323789, 1, 1], 
     Hue[0.5380719740472439, 1, 1], 
     Hue[0.5529927499621092, 1, 1], 
     Hue[0.5679135258769743, 1, 1], 
     Hue[0.5828343017918396, 1, 1], 
     Hue[0.5977550777067047, 1, 1], 
     Hue[0.61267585362157, 1, 1], 
     Hue[0.6275966295364351, 1, 1], 
     Hue[0.6425174054513003, 1, 1], 
     Hue[0.6574381813661655, 1, 1], 
     Hue[0.6723589572810307, 1, 1], 
     Hue[0.6872797331958959, 1, 1], 
     Hue[0.702200509110761, 1, 1], 
     Hue[0.7171212850256263, 1, 1], 
     Hue[0.7320420609404914, 1, 1], 
     Hue[0.7469628368553567, 1, 1], 
     Hue[0.7618836127702218, 1, 1], 
     Hue[0.776804388685087, 1, 1], 
     Hue[0.7917251645999522, 1, 1], 
     Hue[0.8066459405148174, 1, 1], 
     Hue[0.8215667164296826, 1, 1], 
     Hue[0.8364874923445478, 1, 1], 
     Hue[0.851408268259413, 1, 1], 
     Hue[0.8663290441742781, 1, 1], 
     Hue[0.8812498200891433, 1, 1], 
     Hue[0.8961705960040085, 1, 1], 
     Hue[0.9110913719188737, 1, 1], 
     Hue[0.9260121478337389, 1, 1], 
     Hue[0.940932923748604, 1, 1], 
     Hue[0.9558536996634692, 1, 1], 
     Hue[0.9707744755783345, 1, 1], 
     Hue[0.9856952514931997, 1, 1], 
     Hue[0.0006160274080648445, 1, 1], 
     Hue[0.015536803322930032`, 1, 1], 
     Hue[0.030457579237795223`, 1, 1], 
     Hue[0.0453783551526604, 1, 1], 
     Hue[0.0602991310675256, 1, 1], 
     Hue[0.07521990698239078, 1, 1], 
     Hue[0.09014068289725596, 1, 1], 
     Hue[0.10506145881212117`, 1, 1], 
     Hue[0.11998223472698637`, 1, 1], 
     Hue[0.13490301064185153`, 1, 1], 
     Hue[0.14982378655671671`, 1, 1], 
     Hue[0.16474456247158192`, 1, 1], 
     Hue[0.1796653383864471, 1, 1], 
     Hue[0.19458611430131229`, 1, 1], 
     Hue[0.20950689021617747`, 1, 1], 
     Hue[0.22442766613104267`, 1, 1], 
     Hue[0.23934844204590786`, 1, 1], 
     Hue[0.25426921796077306`, 1, 1], 
     Hue[0.26918999387563824`, 1, 1], 
     Hue[0.2841107697905034, 1, 1], 
     Hue[0.2990315457053686, 1, 1], 
     Hue[0.3139523216202338, 1, 1], 
     Hue[0.32887309753509897`, 1, 1], 
     Hue[0.34379387344996415`, 1, 1], 
     Hue[0.3587146493648294, 1, 1], 
     Hue[0.37363542527969457`, 1, 1], 
     Hue[0.38855620119455975`, 1, 1], 
     Hue[0.40347697710942493`, 1, 1], 
     Hue[0.4183977530242901, 1, 1], 
     Hue[0.4333185289391553, 1, 1], 
     Hue[0.44823930485402047`, 1, 1], 
     Hue[0.4631600807688857, 1, 1], 
     Hue[0.4780808566837509, 1, 1], 
     Hue[0.49300163259861607`, 1, 1], 
     Hue[0.5079224085134812, 1, 1], 
     Hue[0.5228431844283464, 1, 1], 
     Hue[0.5377639603432116, 1, 1], 
     Hue[0.5526847362580768, 1, 1], 
     Hue[0.567605512172942, 1, 1], 
     Hue[0.582526288087807, 1, 1], 
     Hue[0.5974470640026723, 1, 1], 
     Hue[0.6123678399175374, 1, 1], 
     Hue[0.6272886158324027, 1, 1], 
     Hue[0.6422093917472679, 1, 1], 
     Hue[0.6571301676621331, 1, 1], 
     Hue[0.6720509435769983, 1, 1], 
     Hue[0.6869717194918634, 1, 1], 
     Hue[0.7018924954067287, 1, 1], 
     Hue[0.7168132713215938, 1, 1], 
     Hue[0.7317340472364591, 1, 1], 
     Hue[0.7466548231513241, 1, 1], 
     Hue[0.7615755990661894, 1, 1], 
     Hue[0.7764963749810546, 1, 1], 
     Hue[0.7914171508959198, 1, 1], 
     Hue[0.806337926810785, 1, 1], 
     Hue[0.8212587027256502, 1, 1], 
     Hue[0.8361794786405153, 1, 1], 
     Hue[0.8511002545553805, 1, 1], 
     Hue[0.8660210304702457, 1, 1], 
     Hue[0.8809418063851109, 1, 1], 
     Hue[0.8958625822999761, 1, 1], 
     Hue[0.9107833582148412, 1, 1], 
     Hue[0.9257041341297065, 1, 1], 
     Hue[0.9406249100445717, 1, 1], 
     Hue[0.9555456859594369, 1, 1], 
     Hue[0.9704664618743021, 1, 1], 
     Hue[0.9853872377891673, 1, 1], 
     Hue[0.0003080137040324222, 1, 1], 
     Hue[0.01522878961889761, 1, 1], 
     Hue[0.0301495655337628, 1, 1], 
     Hue[0.045070341448627996`, 1, 1], 
     Hue[0.059991117363493184`, 1, 1], 
     Hue[0.07491189327835837, 1, 1], 
     Hue[0.08983266919322357, 1, 1], 
     Hue[0.10475344510808872`, 1, 1], 
     Hue[0.11967422102295393`, 1, 1], 
     Hue[0.13459499693781912`, 1, 1], 
     Hue[0.1495157728526843, 1, 1], 
     Hue[0.16443654876754946`, 1, 1], 
     Hue[0.1793573246824147, 1, 1], 
     Hue[0.19427810059727987`, 1, 1], 
     Hue[0.20919887651214505`, 1, 1], 
     Hue[0.22411965242701024`, 1, 1], 
     Hue[0.23904042834187544`, 1, 1], 
     Hue[0.2539612042567406, 1, 1], 
     Hue[0.2688819801716058, 1, 1], 
     Hue[0.283802756086471, 1, 1], 
     Hue[0.29872353200133617`, 1, 1], 
     Hue[0.3136443079162014, 1, 1], 
     Hue[0.32856508383106653`, 1, 1], 
     Hue[0.34348585974593177`, 1, 1], 
     Hue[0.358406635660797, 1, 1], 
     Hue[0.3733274115756622, 1, 1], 
     Hue[0.38824818749052736`, 1, 1], 
     Hue[0.40316896340539254`, 1, 1], 
     Hue[0.4180897393202577, 1, 1], 
     Hue[0.4330105152351229, 1, 1], 
     Hue[0.44793129114998803`, 1, 1], 
     Hue[0.4628520670648533, 1, 1], 
     Hue[0.4777728429797184, 1, 1], 
     Hue[0.4926936188945837, 1, 1], 
     Hue[0.5076143948094487, 1, 1], 
     Hue[0.522535170724314, 1, 1], 
     Hue[0.5374559466391792, 1, 1], 
     Hue[0.5523767225540444, 1, 1], 
     Hue[0.5672974984689095, 1, 1], 
     Hue[0.5822182743837747, 1, 1], 
     Hue[0.5971390502986399, 1, 1], 
     Hue[0.6120598262135051, 1, 1], 
     Hue[0.6269806021283703, 1, 1], 
     Hue[0.6419013780432354, 1, 1], 
     Hue[0.6568221539581007, 1, 1], 
     Hue[0.6717429298729658, 1, 1], 
     Hue[0.6866637057878311, 1, 1], 
     Hue[0.7015844817026962, 1, 1], 
     Hue[0.7165052576175615, 1, 1], 
     Hue[0.7314260335324265, 1, 1], 
     Hue[0.7463468094472918, 1, 1], 
     Hue[0.761267585362157, 1, 1], 
     Hue[0.7761883612770222, 1, 1], 
     Hue[0.7911091371918874, 1, 1], 
     Hue[0.8060299131067525, 1, 1], 
     Hue[0.8209506890216177, 1, 1], 
     Hue[0.8358714649364829, 1, 1], 
     Hue[0.8507922408513481, 1, 1], 
     Hue[0.8657130167662133, 1, 1], 
     Hue[0.8806337926810786, 1, 1], 
     Hue[0.8955545685959436, 1, 1], 
     Hue[0.9104753445108089, 1, 1], 
     Hue[0.9253961204256741, 1, 1], 
     Hue[0.9403168963405393, 1, 1], 
     Hue[0.9552376722554045, 1, 1], 
     Hue[0.9701584481702696, 1, 1], 
     Hue[0.9850792240851348, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.014920775914865188`, 1, 1], 
     Hue[0.029841551829730376`, 1, 1], 
     Hue[0.044762327744595556`, 1, 1], 
     Hue[0.05968310365946075, 1, 1], 
     Hue[0.07460387957432593, 1, 1], 
     Hue[0.08952465548919111, 1, 1], 
     Hue[0.10444543140405632`, 1, 1], 
     Hue[0.11936620731892149`, 1, 1], 
     Hue[0.1342869832337867, 1, 1], 
     Hue[0.1492077591486519, 1, 1], 
     Hue[0.16412853506351707`, 1, 1], 
     Hue[0.17904931097838225`, 1, 1], 
     Hue[0.19397008689324746`, 1, 1], 
     Hue[0.20889086280811264`, 1, 1], 
     Hue[0.22381163872297782`, 1, 1], 
     Hue[0.238732414637843, 1, 1], 
     Hue[0.2536531905527082, 1, 1], 
     Hue[0.2685739664675734, 1, 1], 
     Hue[0.2834947423824386, 1, 1], 
     Hue[0.2984155182973038, 1, 1], 
     Hue[0.31333629421216896`, 1, 1], 
     Hue[0.32825707012703415`, 1, 1], 
     Hue[0.3431778460418993, 1, 1], 
     Hue[0.3580986219567645, 1, 1], 
     Hue[0.37301939787162974`, 1, 1], 
     Hue[0.3879401737864949, 1, 1], 
     Hue[0.4028609497013601, 1, 1], 
     Hue[0.4177817256162253, 1, 1], 
     Hue[0.43270250153109047`, 1, 1], 
     Hue[0.44762327744595565`, 1, 1], 
     Hue[0.46254405336082083`, 1, 1], 
     Hue[0.477464829275686, 1, 1], 
     Hue[0.49238560519055125`, 1, 1], 
     Hue[0.5073063811054164, 1, 1], 
     Hue[0.5222271570202817, 1, 1], 
     Hue[0.5371479329351467, 1, 1], 
     Hue[0.552068708850012, 1, 1], 
     Hue[0.5669894847648771, 1, 1], 
     Hue[0.5819102606797423, 1, 1], 
     Hue[0.5968310365946075, 1, 1], 
     Hue[0.6117518125094726, 1, 1], 
     Hue[0.6266725884243378, 1, 1], 
     Hue[0.641593364339203, 1, 1], 
     Hue[0.6565141402540682, 1, 1], 
     Hue[0.6714349161689335, 1, 1], 
     Hue[0.6863556920837985, 1, 1], 
     Hue[0.7012764679986638, 1, 1], 
     Hue[0.716197243913529, 1, 1], 
     Hue[0.7311180198283942, 1, 1], 
     Hue[0.7460387957432594, 1, 1], 
     Hue[0.7609595716581246, 1, 1], 
     Hue[0.7758803475729897, 1, 1], 
     Hue[0.7908011234878549, 1, 1], 
     Hue[0.8057218994027201, 1, 1], 
     Hue[0.8206426753175853, 1, 1], 
     Hue[0.8355634512324506, 1, 1], 
     Hue[0.8504842271473156, 1, 1], 
     Hue[0.8654050030621809, 1, 1], 
     Hue[0.8803257789770461, 1, 1], 
     Hue[0.8952465548919113, 1, 1], 
     Hue[0.9101673308067765, 1, 1], 
     Hue[0.9250881067216417, 1, 1], 
     Hue[0.9400088826365068, 1, 1], 
     Hue[0.954929658551372, 1, 1], 
     Hue[0.9698504344662372, 1, 1], 
     Hue[0.9847712103811024, 1, 1], 
     Hue[0.9996919862959676, 1, 1], 
     Hue[0.014612762210832766`, 1, 1], 
     Hue[0.029533538125697957`, 1, 1], 
     Hue[0.044454314040563145`, 1, 1], 
     Hue[0.05937508995542833, 1, 1], 
     Hue[0.07429586587029352, 1, 1], 
     Hue[0.08921664178515872, 1, 1], 
     Hue[0.1041374177000239, 1, 1], 
     Hue[0.11905819361488908`, 1, 1], 
     Hue[0.13397896952975427`, 1, 1], 
     Hue[0.14889974544461948`, 1, 1], 
     Hue[0.16382052135948466`, 1, 1], 
     Hue[0.17874129727434984`, 1, 1], 
     Hue[0.19366207318921502`, 1, 1], 
     Hue[0.20858284910408023`, 1, 1], 
     Hue[0.2235036250189454, 1, 1], 
     Hue[0.2384244009338106, 1, 1], 
     Hue[0.2533451768486758, 1, 1], 
     Hue[0.268265952763541, 1, 1], 
     Hue[0.28318672867840616`, 1, 1], 
     Hue[0.29810750459327134`, 1, 1], 
     Hue[0.3130282805081365, 1, 1], 
     Hue[0.3279490564230017, 1, 1], 
     Hue[0.34286983233786694`, 1, 1], 
     Hue[0.3577906082527321, 1, 1], 
     Hue[0.37271138416759725`, 1, 1], 
     Hue[0.38763216008246254`, 1, 1], 
     Hue[0.4025529359973277, 1, 1], 
     Hue[0.4174737119121929, 1, 1], 
     Hue[0.432394487827058, 1, 1], 
     Hue[0.4473152637419232, 1, 1], 
     Hue[0.4622360396567884, 1, 1], 
     Hue[0.47715681557165357`, 1, 1], 
     Hue[0.49207759148651875`, 1, 1], 
     Hue[0.5069983674013839, 1, 1], 
     Hue[0.5219191433162491, 1, 1], 
     Hue[0.5368399192311143, 1, 1], 
     Hue[0.5517606951459795, 1, 1], 
     Hue[0.5666814710608448, 1, 1], 
     Hue[0.5816022469757098, 1, 1], 
     Hue[0.5965230228905751, 1, 1], 
     Hue[0.6114437988054402, 1, 1], 
     Hue[0.6263645747203055, 1, 1], 
     Hue[0.6412853506351706, 1, 1], 
     Hue[0.6562061265500359, 1, 1], 
     Hue[0.671126902464901, 1, 1], 
     Hue[0.6860476783797662, 1, 1], 
     Hue[0.7009684542946314, 1, 1], 
     Hue[0.7158892302094966, 1, 1], 
     Hue[0.7308100061243618, 1, 1], 
     Hue[0.7457307820392269, 1, 1], 
     Hue[0.7606515579540921, 1, 1], 
     Hue[0.7755723338689573, 1, 1], 
     Hue[0.7904931097838226, 1, 1], 
     Hue[0.8054138856986877, 1, 1], 
     Hue[0.820334661613553, 1, 1], 
     Hue[0.835255437528418, 1, 1], 
     Hue[0.8501762134432833, 1, 1], 
     Hue[0.8650969893581485, 1, 1], 
     Hue[0.8800177652730137, 1, 1], 
     Hue[0.8949385411878789, 1, 1], 
     Hue[0.909859317102744, 1, 1], 
     Hue[0.9247800930176092, 1, 1], 
     Hue[0.9397008689324744, 1, 1], 
     Hue[0.9546216448473396, 1, 1], 
     Hue[0.9695424207622048, 1, 1], 
     Hue[0.9844631966770699, 1, 1], 
     Hue[0.9993839725919351, 1, 1], 
     Hue[0.014304748506800346`, 1, 1], 
     Hue[0.02922552442166553, 1, 1], 
     Hue[0.04414630033653072, 1, 1], 
     Hue[0.059067076251395914`, 1, 1], 
     Hue[0.0739878521662611, 1, 1], 
     Hue[0.08890862808112629, 1, 1], 
     Hue[0.10382940399599148`, 1, 1], 
     Hue[0.11875017991085667`, 1, 1], 
     Hue[0.13367095582572183`, 1, 1], 
     Hue[0.14859173174058704`, 1, 1], 
     Hue[0.16351250765545225`, 1, 1], 
     Hue[0.17843328357031743`, 1, 1], 
     Hue[0.1933540594851826, 1, 1], 
     Hue[0.2082748354000478, 1, 1], 
     Hue[0.223195611314913, 1, 1], 
     Hue[0.23811638722977818`, 1, 1], 
     Hue[0.25303716314464336`, 1, 1], 
     Hue[0.26795793905950854`, 1, 1], 
     Hue[0.2828787149743737, 1, 1], 
     Hue[0.29779949088923896`, 1, 1], 
     Hue[0.31272026680410414`, 1, 1], 
     Hue[0.3276410427189693, 1, 1], 
     Hue[0.3425618186338345, 1, 1], 
     Hue[0.3574825945486997, 1, 1], 
     Hue[0.37240337046356486`, 1, 1], 
     Hue[0.38732414637843005`, 1, 1], 
     Hue[0.4022449222932953, 1, 1], 
     Hue[0.41716569820816046`, 1, 1], 
     Hue[0.43208647412302564`, 1, 1], 
     Hue[0.4470072500378908, 1, 1], 
     Hue[0.461928025952756, 1, 1], 
     Hue[0.4768488018676212, 1, 1], 
     Hue[0.49176957778248637`, 1, 1], 
     Hue[0.5066903536973515, 1, 1], 
     Hue[0.5216111296122167, 1, 1], 
     Hue[0.5365319055270819, 1, 1], 
     Hue[0.551452681441947, 1, 1], 
     Hue[0.5663734573568122, 1, 1], 
     Hue[0.5812942332716775, 1, 1], 
     Hue[0.5962150091865426, 1, 1], 
     Hue[0.6111357851014079, 1, 1], 
     Hue[0.6260565610162729, 1, 1], 
     Hue[0.6409773369311382, 1, 1], 
     Hue[0.6558981128460033, 1, 1], 
     Hue[0.6708188887608686, 1, 1], 
     Hue[0.6857396646757338, 1, 1], 
     Hue[0.700660440590599, 1, 1], 
     Hue[0.7155812165054641, 1, 1], 
     Hue[0.7305019924203293, 1, 1], 
     Hue[0.7454227683351946, 1, 1], 
     Hue[0.7603435442500597, 1, 1], 
     Hue[0.775264320164925, 1, 1], 
     Hue[0.79018509607979, 1, 1], 
     Hue[0.8051058719946552, 1, 1], 
     Hue[0.8200266479095205, 1, 1], 
     Hue[0.8349474238243857, 1, 1], 
     Hue[0.8498681997392509, 1, 1], 
     Hue[0.864788975654116, 1, 1], 
     Hue[0.8797097515689812, 1, 1], 
     Hue[0.8946305274838464, 1, 1], 
     Hue[0.9095513033987116, 1, 1], 
     Hue[0.9244720793135768, 1, 1], 
     Hue[0.939392855228442, 1, 1], 
     Hue[0.9543136311433071, 1, 1], 
     Hue[0.9692344070581723, 1, 1], 
     Hue[0.9841551829730375, 1, 1], 
     Hue[0.9990759588879027, 1, 1], 
     Hue[0.013996734802767922`, 1, 1], 
     Hue[0.028917510717633117`, 1, 1], 
     Hue[0.0438382866324983, 1, 1], 
     Hue[0.05875906254736348, 1, 1], 
     Hue[0.07367983846222867, 1, 1], 
     Hue[0.08860061437709388, 1, 1], 
     Hue[0.10352139029195906`, 1, 1], 
     Hue[0.11844216620682425`, 1, 1], 
     Hue[0.13336294212168942`, 1, 1], 
     Hue[0.14828371803655466`, 1, 1], 
     Hue[0.1632044939514198, 1, 1], 
     Hue[0.17812526986628502`, 1, 1], 
     Hue[0.19304604578115017`, 1, 1], 
     Hue[0.20796682169601538`, 1, 1], 
     Hue[0.22288759761088056`, 1, 1], 
     Hue[0.23780837352574577`, 1, 1], 
     Hue[0.252729149440611, 1, 1], 
     Hue[0.26764992535547616`, 1, 1], 
     Hue[0.28257070127034134`, 1, 1], 
     Hue[0.29749147718520647`, 1, 1], 
     Hue[0.3124122531000717, 1, 1], 
     Hue[0.3273330290149369, 1, 1], 
     Hue[0.34225380492980206`, 1, 1], 
     Hue[0.35717458084466724`, 1, 1], 
     Hue[0.3720953567595324, 1, 1], 
     Hue[0.3870161326743976, 1, 1], 
     Hue[0.4019369085892628, 1, 1], 
     Hue[0.41685768450412797`, 1, 1], 
     Hue[0.4317784604189932, 1, 1], 
     Hue[0.4466992363338584, 1, 1], 
     Hue[0.46162001224872357`, 1, 1], 
     Hue[0.47654078816358875`, 1, 1], 
     Hue[0.4914615640784539, 1, 1], 
     Hue[0.5063823399933192, 1, 1], 
     Hue[0.5213031159081842, 1, 1], 
     Hue[0.5362238918230495, 1, 1], 
     Hue[0.5511446677379146, 1, 1], 
     Hue[0.5660654436527799, 1, 1], 
     Hue[0.5809862195676451, 1, 1], 
     Hue[0.5959069954825102, 1, 1], 
     Hue[0.6108277713973754, 1, 1], 
     Hue[0.6257485473122406, 1, 1], 
     Hue[0.6406693232271058, 1, 1], 
     Hue[0.655590099141971, 1, 1], 
     Hue[0.6705108750568362, 1, 1], 
     Hue[0.6854316509717013, 1, 1], 
     Hue[0.7003524268865666, 1, 1], 
     Hue[0.7152732028014317, 1, 1], 
     Hue[0.730193978716297, 1, 1], 
     Hue[0.7451147546311621, 1, 1], 
     Hue[0.7600355305460273, 1, 1], 
     Hue[0.7749563064608925, 1, 1], 
     Hue[0.7898770823757577, 1, 1], 
     Hue[0.8047978582906228, 1, 1], 
     Hue[0.8197186342054881, 1, 1], 
     Hue[0.8346394101203533, 1, 1], 
     Hue[0.8495601860352184, 1, 1], 
     Hue[0.8644809619500836, 1, 1], 
     Hue[0.8794017378649488, 1, 1], 
     Hue[0.894322513779814, 1, 1], 
     Hue[0.9092432896946792, 1, 1], 
     Hue[0.9241640656095443, 1, 1], 
     Hue[0.9390848415244095, 1, 1], 
     Hue[0.9540056174392747, 1, 1], 
     Hue[0.96892639335414, 1, 1], 
     Hue[0.9838471692690052, 1, 1], 
     Hue[0.9987679451838704, 1, 1], 
     Hue[0.013688721098735502`, 1, 1], 
     Hue[0.02860949701360068, 1, 1], 
     Hue[0.043530272928465875`, 1, 1], 
     Hue[0.05845104884333107, 1, 1], 
     Hue[0.07337182475819626, 1, 1], 
     Hue[0.08829260067306144, 1, 1], 
     Hue[0.1032133765879266, 1, 1], 
     Hue[0.11813415250279183`, 1, 1], 
     Hue[0.133054928417657, 1, 1], 
     Hue[0.1479757043325222, 1, 1], 
     Hue[0.16289648024738737`, 1, 1], 
     Hue[0.17781725616225258`, 1, 1], 
     Hue[0.1927380320771178, 1, 1], 
     Hue[0.20765880799198297`, 1, 1], 
     Hue[0.22257958390684815`, 1, 1], 
     Hue[0.23750035982171333`, 1, 1], 
     Hue[0.25242113573657854`, 1, 1], 
     Hue[0.2673419116514437, 1, 1], 
     Hue[0.2822626875663089, 1, 1], 
     Hue[0.2971834634811741, 1, 1], 
     Hue[0.31210423939603926`, 1, 1], 
     Hue[0.3270250153109045, 1, 1], 
     Hue[0.3419457912257697, 1, 1], 
     Hue[0.35686656714063486`, 1, 1], 
     Hue[0.37178734305550004`, 1, 1], 
     Hue[0.3867081189703652, 1, 1], 
     Hue[0.4016288948852304, 1, 1], 
     Hue[0.4165496708000956, 1, 1], 
     Hue[0.4314704467149608, 1, 1], 
     Hue[0.446391222629826, 1, 1], 
     Hue[0.3142603353242662, 1, 1], 
     Hue[0.3142603353242662, 1, 1], 
     Hue[0.32918111123913146`, 1, 1], 
     Hue[0.32918111123913146`, 1, 1], 
     Hue[0.34410188715399664`, 1, 1], 
     Hue[0.34410188715399664`, 1, 1], 
     Hue[0.35902266306886177`, 1, 1], 
     Hue[0.35902266306886177`, 1, 1], 
     Hue[0.373943438983727, 1, 1], 
     Hue[0.373943438983727, 1, 1], 
     Hue[0.3888642148985922, 1, 1], 
     Hue[0.3888642148985922, 1, 1], 
     Hue[0.40378499081345737`, 1, 1], 
     Hue[0.40378499081345737`, 1, 1], 
     Hue[0.41870576672832255`, 1, 1], 
     Hue[0.41870576672832255`, 1, 1], 
     Hue[0.4336265426431878, 1, 1], 
     Hue[0.4336265426431878, 1, 1], 
     Hue[0.44854731855805297`, 1, 1], 
     Hue[0.44854731855805297`, 1, 1], 
     Hue[0.46346809447291815`, 1, 1], 
     Hue[0.46346809447291815`, 1, 1], 
     Hue[0.4783888703877833, 1, 1], 
     Hue[0.4783888703877833, 1, 1], 
     Hue[0.4933096463026485, 1, 1], 
     Hue[0.4933096463026485, 1, 1], 
     Hue[0.5082304222175136, 1, 1], 
     Hue[0.5082304222175136, 1, 1], 
     Hue[0.5231511981323789, 1, 1], 
     Hue[0.5231511981323789, 1, 1], 
     Hue[0.5380719740472439, 1, 1], 
     Hue[0.5380719740472439, 1, 1], 
     Hue[0.5529927499621092, 1, 1], 
     Hue[0.5529927499621092, 1, 1], 
     Hue[0.5679135258769743, 1, 1], 
     Hue[0.5679135258769743, 1, 1], 
     Hue[0.5828343017918396, 1, 1], 
     Hue[0.5828343017918396, 1, 1], 
     Hue[0.5977550777067047, 1, 1], 
     Hue[0.5977550777067047, 1, 1], 
     Hue[0.61267585362157, 1, 1], 
     Hue[0.61267585362157, 1, 1], 
     Hue[0.6275966295364351, 1, 1], 
     Hue[0.6275966295364351, 1, 1], 
     Hue[0.6425174054513003, 1, 1], 
     Hue[0.6425174054513003, 1, 1], 
     Hue[0.6574381813661655, 1, 1], 
     Hue[0.6574381813661655, 1, 1], 
     Hue[0.6723589572810307, 1, 1], 
     Hue[0.6723589572810307, 1, 1], 
     Hue[0.6872797331958959, 1, 1], 
     Hue[0.6872797331958959, 1, 1], 
     Hue[0.702200509110761, 1, 1], 
     Hue[0.702200509110761, 1, 1], 
     Hue[0.7171212850256263, 1, 1], 
     Hue[0.7171212850256263, 1, 1], 
     Hue[0.7320420609404914, 1, 1], 
     Hue[0.7320420609404914, 1, 1], 
     Hue[0.7469628368553567, 1, 1], 
     Hue[0.7469628368553567, 1, 1], 
     Hue[0.7618836127702218, 1, 1], 
     Hue[0.7618836127702218, 1, 1], 
     Hue[0.776804388685087, 1, 1], 
     Hue[0.776804388685087, 1, 1], 
     Hue[0.7917251645999522, 1, 1], 
     Hue[0.7917251645999522, 1, 1], 
     Hue[0.8066459405148174, 1, 1], 
     Hue[0.8066459405148174, 1, 1], 
     Hue[0.8215667164296826, 1, 1], 
     Hue[0.8215667164296826, 1, 1], 
     Hue[0.8364874923445478, 1, 1], 
     Hue[0.8364874923445478, 1, 1], 
     Hue[0.851408268259413, 1, 1], 
     Hue[0.851408268259413, 1, 1], 
     Hue[0.8663290441742781, 1, 1], 
     Hue[0.8663290441742781, 1, 1], 
     Hue[0.8812498200891433, 1, 1], 
     Hue[0.8812498200891433, 1, 1], 
     Hue[0.8961705960040085, 1, 1], 
     Hue[0.8961705960040085, 1, 1], 
     Hue[0.9110913719188737, 1, 1], 
     Hue[0.9110913719188737, 1, 1], 
     Hue[0.9260121478337389, 1, 1], 
     Hue[0.9260121478337389, 1, 1], 
     Hue[0.940932923748604, 1, 1], 
     Hue[0.940932923748604, 1, 1], 
     Hue[0.9558536996634692, 1, 1], 
     Hue[0.9558536996634692, 1, 1], 
     Hue[0.9707744755783345, 1, 1], 
     Hue[0.9707744755783345, 1, 1], 
     Hue[0.9856952514931997, 1, 1], 
     Hue[0.9856952514931997, 1, 1], 
     Hue[0.0006160274080648445, 1, 1], 
     Hue[0.0006160274080648445, 1, 1], 
     Hue[0.015536803322930032`, 1, 1], 
     Hue[0.015536803322930032`, 1, 1], 
     Hue[0.030457579237795223`, 1, 1], 
     Hue[0.030457579237795223`, 1, 1], 
     Hue[0.0453783551526604, 1, 1], 
     Hue[0.0453783551526604, 1, 1], 
     Hue[0.0602991310675256, 1, 1], 
     Hue[0.0602991310675256, 1, 1], 
     Hue[0.07521990698239078, 1, 1], 
     Hue[0.07521990698239078, 1, 1], 
     Hue[0.09014068289725596, 1, 1], 
     Hue[0.09014068289725596, 1, 1], 
     Hue[0.10506145881212117`, 1, 1], 
     Hue[0.10506145881212117`, 1, 1], 
     Hue[0.11998223472698637`, 1, 1], 
     Hue[0.11998223472698637`, 1, 1], 
     Hue[0.13490301064185153`, 1, 1], 
     Hue[0.13490301064185153`, 1, 1], 
     Hue[0.14982378655671671`, 1, 1], 
     Hue[0.14982378655671671`, 1, 1], 
     Hue[0.16474456247158192`, 1, 1], 
     Hue[0.16474456247158192`, 1, 1], 
     Hue[0.1796653383864471, 1, 1], 
     Hue[0.1796653383864471, 1, 1], 
     Hue[0.19458611430131229`, 1, 1], 
     Hue[0.19458611430131229`, 1, 1], 
     Hue[0.20950689021617747`, 1, 1], 
     Hue[0.20950689021617747`, 1, 1], 
     Hue[0.22442766613104267`, 1, 1], 
     Hue[0.22442766613104267`, 1, 1], 
     Hue[0.23934844204590786`, 1, 1], 
     Hue[0.23934844204590786`, 1, 1], 
     Hue[0.25426921796077306`, 1, 1], 
     Hue[0.25426921796077306`, 1, 1], 
     Hue[0.26918999387563824`, 1, 1], 
     Hue[0.26918999387563824`, 1, 1], 
     Hue[0.2841107697905034, 1, 1], 
     Hue[0.2841107697905034, 1, 1], 
     Hue[0.2990315457053686, 1, 1], 
     Hue[0.2990315457053686, 1, 1], 
     Hue[0.3139523216202338, 1, 1], 
     Hue[0.3139523216202338, 1, 1], 
     Hue[0.32887309753509897`, 1, 1], 
     Hue[0.32887309753509897`, 1, 1], 
     Hue[0.34379387344996415`, 1, 1], 
     Hue[0.34379387344996415`, 1, 1], 
     Hue[0.3587146493648294, 1, 1], 
     Hue[0.3587146493648294, 1, 1], 
     Hue[0.37363542527969457`, 1, 1], 
     Hue[0.37363542527969457`, 1, 1], 
     Hue[0.38855620119455975`, 1, 1], 
     Hue[0.38855620119455975`, 1, 1], 
     Hue[0.40347697710942493`, 1, 1], 
     Hue[0.40347697710942493`, 1, 1], 
     Hue[0.4183977530242901, 1, 1], 
     Hue[0.4183977530242901, 1, 1], 
     Hue[0.4333185289391553, 1, 1], 
     Hue[0.4333185289391553, 1, 1], 
     Hue[0.44823930485402047`, 1, 1], 
     Hue[0.44823930485402047`, 1, 1], 
     Hue[0.4631600807688857, 1, 1], 
     Hue[0.4631600807688857, 1, 1], 
     Hue[0.4780808566837509, 1, 1], 
     Hue[0.4780808566837509, 1, 1], 
     Hue[0.49300163259861607`, 1, 1], 
     Hue[0.49300163259861607`, 1, 1], 
     Hue[0.5079224085134812, 1, 1], 
     Hue[0.5079224085134812, 1, 1], 
     Hue[0.5228431844283464, 1, 1], 
     Hue[0.5228431844283464, 1, 1], 
     Hue[0.5377639603432116, 1, 1], 
     Hue[0.5377639603432116, 1, 1], 
     Hue[0.5526847362580768, 1, 1], 
     Hue[0.5526847362580768, 1, 1], 
     Hue[0.567605512172942, 1, 1], 
     Hue[0.567605512172942, 1, 1], 
     Hue[0.582526288087807, 1, 1], 
     Hue[0.582526288087807, 1, 1], 
     Hue[0.5974470640026723, 1, 1], 
     Hue[0.5974470640026723, 1, 1], 
     Hue[0.6123678399175374, 1, 1], 
     Hue[0.6123678399175374, 1, 1], 
     Hue[0.6272886158324027, 1, 1], 
     Hue[0.6272886158324027, 1, 1], 
     Hue[0.6422093917472679, 1, 1], 
     Hue[0.6422093917472679, 1, 1], 
     Hue[0.6571301676621331, 1, 1], 
     Hue[0.6571301676621331, 1, 1], 
     Hue[0.6720509435769983, 1, 1], 
     Hue[0.6720509435769983, 1, 1], 
     Hue[0.6869717194918634, 1, 1], 
     Hue[0.6869717194918634, 1, 1], 
     Hue[0.7018924954067287, 1, 1], 
     Hue[0.7018924954067287, 1, 1], 
     Hue[0.7168132713215938, 1, 1], 
     Hue[0.7168132713215938, 1, 1], 
     Hue[0.7317340472364591, 1, 1], 
     Hue[0.7317340472364591, 1, 1], 
     Hue[0.7466548231513241, 1, 1], 
     Hue[0.7466548231513241, 1, 1], 
     Hue[0.7615755990661894, 1, 1], 
     Hue[0.7615755990661894, 1, 1], 
     Hue[0.7764963749810546, 1, 1], 
     Hue[0.7764963749810546, 1, 1], 
     Hue[0.7914171508959198, 1, 1], 
     Hue[0.7914171508959198, 1, 1], 
     Hue[0.806337926810785, 1, 1], 
     Hue[0.806337926810785, 1, 1], 
     Hue[0.8212587027256502, 1, 1], 
     Hue[0.8212587027256502, 1, 1], 
     Hue[0.8361794786405153, 1, 1], 
     Hue[0.8361794786405153, 1, 1], 
     Hue[0.8511002545553805, 1, 1], 
     Hue[0.8511002545553805, 1, 1], 
     Hue[0.8660210304702457, 1, 1], 
     Hue[0.8660210304702457, 1, 1], 
     Hue[0.8809418063851109, 1, 1], 
     Hue[0.8809418063851109, 1, 1], 
     Hue[0.8958625822999761, 1, 1], 
     Hue[0.8958625822999761, 1, 1], 
     Hue[0.9107833582148412, 1, 1], 
     Hue[0.9107833582148412, 1, 1], 
     Hue[0.9257041341297065, 1, 1], 
     Hue[0.9257041341297065, 1, 1], 
     Hue[0.9406249100445717, 1, 1], 
     Hue[0.9406249100445717, 1, 1], 
     Hue[0.9555456859594369, 1, 1], 
     Hue[0.9555456859594369, 1, 1], 
     Hue[0.9704664618743021, 1, 1], 
     Hue[0.9704664618743021, 1, 1], 
     Hue[0.9853872377891673, 1, 1], 
     Hue[0.9853872377891673, 1, 1], 
     Hue[0.0003080137040324222, 1, 1], 
     Hue[0.0003080137040324222, 1, 1], 
     Hue[0.01522878961889761, 1, 1], 
     Hue[0.01522878961889761, 1, 1], 
     Hue[0.0301495655337628, 1, 1], 
     Hue[0.0301495655337628, 1, 1], 
     Hue[0.045070341448627996`, 1, 1], 
     Hue[0.045070341448627996`, 1, 1], 
     Hue[0.059991117363493184`, 1, 1], 
     Hue[0.059991117363493184`, 1, 1], 
     Hue[0.07491189327835837, 1, 1], 
     Hue[0.07491189327835837, 1, 1], 
     Hue[0.08983266919322357, 1, 1], 
     Hue[0.08983266919322357, 1, 1], 
     Hue[0.10475344510808872`, 1, 1], 
     Hue[0.10475344510808872`, 1, 1], 
     Hue[0.11967422102295393`, 1, 1], 
     Hue[0.11967422102295393`, 1, 1], 
     Hue[0.13459499693781912`, 1, 1], 
     Hue[0.13459499693781912`, 1, 1], 
     Hue[0.1495157728526843, 1, 1], 
     Hue[0.1495157728526843, 1, 1], 
     Hue[0.16443654876754946`, 1, 1], 
     Hue[0.16443654876754946`, 1, 1], 
     Hue[0.1793573246824147, 1, 1], 
     Hue[0.1793573246824147, 1, 1], 
     Hue[0.19427810059727987`, 1, 1], 
     Hue[0.19427810059727987`, 1, 1], 
     Hue[0.20919887651214505`, 1, 1], 
     Hue[0.20919887651214505`, 1, 1], 
     Hue[0.22411965242701024`, 1, 1], 
     Hue[0.22411965242701024`, 1, 1], 
     Hue[0.23904042834187544`, 1, 1], 
     Hue[0.23904042834187544`, 1, 1], 
     Hue[0.2539612042567406, 1, 1], 
     Hue[0.2539612042567406, 1, 1], 
     Hue[0.2688819801716058, 1, 1], 
     Hue[0.2688819801716058, 1, 1], 
     Hue[0.283802756086471, 1, 1], 
     Hue[0.283802756086471, 1, 1], 
     Hue[0.29872353200133617`, 1, 1], 
     Hue[0.29872353200133617`, 1, 1], 
     Hue[0.3136443079162014, 1, 1], 
     Hue[0.3136443079162014, 1, 1], 
     Hue[0.32856508383106653`, 1, 1], 
     Hue[0.32856508383106653`, 1, 1], 
     Hue[0.34348585974593177`, 1, 1], 
     Hue[0.34348585974593177`, 1, 1], 
     Hue[0.358406635660797, 1, 1], 
     Hue[0.358406635660797, 1, 1], 
     Hue[0.3733274115756622, 1, 1], 
     Hue[0.3733274115756622, 1, 1], 
     Hue[0.38824818749052736`, 1, 1], 
     Hue[0.38824818749052736`, 1, 1], 
     Hue[0.40316896340539254`, 1, 1], 
     Hue[0.40316896340539254`, 1, 1], 
     Hue[0.4180897393202577, 1, 1], 
     Hue[0.4180897393202577, 1, 1], 
     Hue[0.4330105152351229, 1, 1], 
     Hue[0.4330105152351229, 1, 1], 
     Hue[0.44793129114998803`, 1, 1], 
     Hue[0.44793129114998803`, 1, 1], 
     Hue[0.4628520670648533, 1, 1], 
     Hue[0.4628520670648533, 1, 1], 
     Hue[0.4777728429797184, 1, 1], 
     Hue[0.4777728429797184, 1, 1], 
     Hue[0.4926936188945837, 1, 1], 
     Hue[0.4926936188945837, 1, 1], 
     Hue[0.5076143948094487, 1, 1], 
     Hue[0.5076143948094487, 1, 1], 
     Hue[0.522535170724314, 1, 1], 
     Hue[0.522535170724314, 1, 1], 
     Hue[0.5374559466391792, 1, 1], 
     Hue[0.5374559466391792, 1, 1], 
     Hue[0.5523767225540444, 1, 1], 
     Hue[0.5523767225540444, 1, 1], 
     Hue[0.5672974984689095, 1, 1], 
     Hue[0.5672974984689095, 1, 1], 
     Hue[0.5822182743837747, 1, 1], 
     Hue[0.5822182743837747, 1, 1], 
     Hue[0.5971390502986399, 1, 1], 
     Hue[0.5971390502986399, 1, 1], 
     Hue[0.6120598262135051, 1, 1], 
     Hue[0.6120598262135051, 1, 1], 
     Hue[0.6269806021283703, 1, 1], 
     Hue[0.6269806021283703, 1, 1], 
     Hue[0.6419013780432354, 1, 1], 
     Hue[0.6419013780432354, 1, 1], 
     Hue[0.6568221539581007, 1, 1], 
     Hue[0.6568221539581007, 1, 1], 
     Hue[0.6717429298729658, 1, 1], 
     Hue[0.6717429298729658, 1, 1], 
     Hue[0.6866637057878311, 1, 1], 
     Hue[0.6866637057878311, 1, 1], 
     Hue[0.7015844817026962, 1, 1], 
     Hue[0.7015844817026962, 1, 1], 
     Hue[0.7165052576175615, 1, 1], 
     Hue[0.7165052576175615, 1, 1], 
     Hue[0.7314260335324265, 1, 1], 
     Hue[0.7314260335324265, 1, 1], 
     Hue[0.7463468094472918, 1, 1], 
     Hue[0.7463468094472918, 1, 1], 
     Hue[0.761267585362157, 1, 1], 
     Hue[0.761267585362157, 1, 1], 
     Hue[0.7761883612770222, 1, 1], 
     Hue[0.7761883612770222, 1, 1], 
     Hue[0.7911091371918874, 1, 1], 
     Hue[0.7911091371918874, 1, 1], 
     Hue[0.8060299131067525, 1, 1], 
     Hue[0.8060299131067525, 1, 1], 
     Hue[0.8209506890216177, 1, 1], 
     Hue[0.8209506890216177, 1, 1], 
     Hue[0.8358714649364829, 1, 1], 
     Hue[0.8358714649364829, 1, 1], 
     Hue[0.8507922408513481, 1, 1], 
     Hue[0.8507922408513481, 1, 1], 
     Hue[0.8657130167662133, 1, 1], 
     Hue[0.8657130167662133, 1, 1], 
     Hue[0.8806337926810786, 1, 1], 
     Hue[0.8806337926810786, 1, 1], 
     Hue[0.8955545685959436, 1, 1], 
     Hue[0.8955545685959436, 1, 1], 
     Hue[0.9104753445108089, 1, 1], 
     Hue[0.9104753445108089, 1, 1], 
     Hue[0.9253961204256741, 1, 1], 
     Hue[0.9253961204256741, 1, 1], 
     Hue[0.9403168963405393, 1, 1], 
     Hue[0.9403168963405393, 1, 1], 
     Hue[0.9552376722554045, 1, 1], 
     Hue[0.9552376722554045, 1, 1], 
     Hue[0.9701584481702696, 1, 1], 
     Hue[0.9701584481702696, 1, 1], 
     Hue[0.9850792240851348, 1, 1], 
     Hue[0.9850792240851348, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.014920775914865188`, 1, 1], 
     Hue[0.014920775914865188`, 1, 1], 
     Hue[0.029841551829730376`, 1, 1], 
     Hue[0.029841551829730376`, 1, 1], 
     Hue[0.044762327744595556`, 1, 1], 
     Hue[0.044762327744595556`, 1, 1], 
     Hue[0.05968310365946075, 1, 1], 
     Hue[0.05968310365946075, 1, 1], 
     Hue[0.07460387957432593, 1, 1], 
     Hue[0.07460387957432593, 1, 1], 
     Hue[0.08952465548919111, 1, 1], 
     Hue[0.08952465548919111, 1, 1], 
     Hue[0.10444543140405632`, 1, 1], 
     Hue[0.10444543140405632`, 1, 1], 
     Hue[0.11936620731892149`, 1, 1], 
     Hue[0.11936620731892149`, 1, 1], 
     Hue[0.1342869832337867, 1, 1], 
     Hue[0.1342869832337867, 1, 1], 
     Hue[0.1492077591486519, 1, 1], 
     Hue[0.1492077591486519, 1, 1], 
     Hue[0.16412853506351707`, 1, 1], 
     Hue[0.16412853506351707`, 1, 1], 
     Hue[0.17904931097838225`, 1, 1], 
     Hue[0.17904931097838225`, 1, 1], 
     Hue[0.19397008689324746`, 1, 1], 
     Hue[0.19397008689324746`, 1, 1], 
     Hue[0.20889086280811264`, 1, 1], 
     Hue[0.20889086280811264`, 1, 1], 
     Hue[0.22381163872297782`, 1, 1], 
     Hue[0.22381163872297782`, 1, 1], 
     Hue[0.238732414637843, 1, 1], 
     Hue[0.238732414637843, 1, 1], 
     Hue[0.2536531905527082, 1, 1], 
     Hue[0.2536531905527082, 1, 1], 
     Hue[0.2685739664675734, 1, 1], 
     Hue[0.2685739664675734, 1, 1], 
     Hue[0.2834947423824386, 1, 1], 
     Hue[0.2834947423824386, 1, 1], 
     Hue[0.2984155182973038, 1, 1], 
     Hue[0.2984155182973038, 1, 1], 
     Hue[0.31333629421216896`, 1, 1], 
     Hue[0.31333629421216896`, 1, 1], 
     Hue[0.32825707012703415`, 1, 1], 
     Hue[0.32825707012703415`, 1, 1], 
     Hue[0.3431778460418993, 1, 1], 
     Hue[0.3431778460418993, 1, 1], 
     Hue[0.3580986219567645, 1, 1], 
     Hue[0.3580986219567645, 1, 1], 
     Hue[0.37301939787162974`, 1, 1], 
     Hue[0.37301939787162974`, 1, 1], 
     Hue[0.3879401737864949, 1, 1], 
     Hue[0.3879401737864949, 1, 1], 
     Hue[0.4028609497013601, 1, 1], 
     Hue[0.4028609497013601, 1, 1], 
     Hue[0.4177817256162253, 1, 1], 
     Hue[0.4177817256162253, 1, 1], 
     Hue[0.43270250153109047`, 1, 1], 
     Hue[0.43270250153109047`, 1, 1], 
     Hue[0.44762327744595565`, 1, 1], 
     Hue[0.44762327744595565`, 1, 1], 
     Hue[0.46254405336082083`, 1, 1], 
     Hue[0.46254405336082083`, 1, 1], 
     Hue[0.477464829275686, 1, 1], 
     Hue[0.477464829275686, 1, 1], 
     Hue[0.49238560519055125`, 1, 1], 
     Hue[0.49238560519055125`, 1, 1], 
     Hue[0.5073063811054164, 1, 1], 
     Hue[0.5073063811054164, 1, 1], 
     Hue[0.5222271570202817, 1, 1], 
     Hue[0.5222271570202817, 1, 1], 
     Hue[0.5371479329351467, 1, 1], 
     Hue[0.5371479329351467, 1, 1], 
     Hue[0.552068708850012, 1, 1], 
     Hue[0.552068708850012, 1, 1], 
     Hue[0.5669894847648771, 1, 1], 
     Hue[0.5669894847648771, 1, 1], 
     Hue[0.5819102606797423, 1, 1], 
     Hue[0.5819102606797423, 1, 1], 
     Hue[0.5968310365946075, 1, 1], 
     Hue[0.5968310365946075, 1, 1], 
     Hue[0.6117518125094726, 1, 1], 
     Hue[0.6117518125094726, 1, 1], 
     Hue[0.6266725884243378, 1, 1], 
     Hue[0.6266725884243378, 1, 1], 
     Hue[0.641593364339203, 1, 1], 
     Hue[0.641593364339203, 1, 1], 
     Hue[0.6565141402540682, 1, 1], 
     Hue[0.6565141402540682, 1, 1], 
     Hue[0.6714349161689335, 1, 1], 
     Hue[0.6714349161689335, 1, 1], 
     Hue[0.6863556920837985, 1, 1], 
     Hue[0.6863556920837985, 1, 1], 
     Hue[0.7012764679986638, 1, 1], 
     Hue[0.7012764679986638, 1, 1], 
     Hue[0.716197243913529, 1, 1], 
     Hue[0.716197243913529, 1, 1], 
     Hue[0.7311180198283942, 1, 1], 
     Hue[0.7311180198283942, 1, 1], 
     Hue[0.7460387957432594, 1, 1], 
     Hue[0.7460387957432594, 1, 1], 
     Hue[0.7609595716581246, 1, 1], 
     Hue[0.7609595716581246, 1, 1], 
     Hue[0.7758803475729897, 1, 1], 
     Hue[0.7758803475729897, 1, 1], 
     Hue[0.7908011234878549, 1, 1], 
     Hue[0.7908011234878549, 1, 1], 
     Hue[0.8057218994027201, 1, 1], 
     Hue[0.8057218994027201, 1, 1], 
     Hue[0.8206426753175853, 1, 1], 
     Hue[0.8206426753175853, 1, 1], 
     Hue[0.8355634512324506, 1, 1], 
     Hue[0.8355634512324506, 1, 1], 
     Hue[0.8504842271473156, 1, 1], 
     Hue[0.8504842271473156, 1, 1], 
     Hue[0.8654050030621809, 1, 1], 
     Hue[0.8654050030621809, 1, 1], 
     Hue[0.8803257789770461, 1, 1], 
     Hue[0.8803257789770461, 1, 1], 
     Hue[0.8952465548919113, 1, 1], 
     Hue[0.8952465548919113, 1, 1], 
     Hue[0.9101673308067765, 1, 1], 
     Hue[0.9101673308067765, 1, 1], 
     Hue[0.9250881067216417, 1, 1], 
     Hue[0.9250881067216417, 1, 1], 
     Hue[0.9400088826365068, 1, 1], 
     Hue[0.9400088826365068, 1, 1], 
     Hue[0.954929658551372, 1, 1], 
     Hue[0.954929658551372, 1, 1], 
     Hue[0.9698504344662372, 1, 1], 
     Hue[0.9698504344662372, 1, 1], 
     Hue[0.9847712103811024, 1, 1], 
     Hue[0.9847712103811024, 1, 1], 
     Hue[0.9996919862959676, 1, 1], 
     Hue[0.9996919862959676, 1, 1], 
     Hue[0.014612762210832766`, 1, 1], 
     Hue[0.014612762210832766`, 1, 1], 
     Hue[0.029533538125697957`, 1, 1], 
     Hue[0.029533538125697957`, 1, 1], 
     Hue[0.044454314040563145`, 1, 1], 
     Hue[0.044454314040563145`, 1, 1], 
     Hue[0.05937508995542833, 1, 1], 
     Hue[0.05937508995542833, 1, 1], 
     Hue[0.07429586587029352, 1, 1], 
     Hue[0.07429586587029352, 1, 1], 
     Hue[0.08921664178515872, 1, 1], 
     Hue[0.08921664178515872, 1, 1], 
     Hue[0.1041374177000239, 1, 1], 
     Hue[0.1041374177000239, 1, 1], 
     Hue[0.11905819361488908`, 1, 1], 
     Hue[0.11905819361488908`, 1, 1], 
     Hue[0.13397896952975427`, 1, 1], 
     Hue[0.13397896952975427`, 1, 1], 
     Hue[0.14889974544461948`, 1, 1], 
     Hue[0.14889974544461948`, 1, 1], 
     Hue[0.16382052135948466`, 1, 1], 
     Hue[0.16382052135948466`, 1, 1], 
     Hue[0.17874129727434984`, 1, 1], 
     Hue[0.17874129727434984`, 1, 1], 
     Hue[0.19366207318921502`, 1, 1], 
     Hue[0.19366207318921502`, 1, 1], 
     Hue[0.20858284910408023`, 1, 1], 
     Hue[0.20858284910408023`, 1, 1], 
     Hue[0.2235036250189454, 1, 1], 
     Hue[0.2235036250189454, 1, 1], 
     Hue[0.2384244009338106, 1, 1], 
     Hue[0.2384244009338106, 1, 1], 
     Hue[0.2533451768486758, 1, 1], 
     Hue[0.2533451768486758, 1, 1], 
     Hue[0.268265952763541, 1, 1], 
     Hue[0.268265952763541, 1, 1], 
     Hue[0.28318672867840616`, 1, 1], 
     Hue[0.28318672867840616`, 1, 1], 
     Hue[0.29810750459327134`, 1, 1], 
     Hue[0.29810750459327134`, 1, 1], 
     Hue[0.3130282805081365, 1, 1], 
     Hue[0.3130282805081365, 1, 1], 
     Hue[0.3279490564230017, 1, 1], 
     Hue[0.3279490564230017, 1, 1], 
     Hue[0.34286983233786694`, 1, 1], 
     Hue[0.34286983233786694`, 1, 1], 
     Hue[0.3577906082527321, 1, 1], 
     Hue[0.3577906082527321, 1, 1], 
     Hue[0.37271138416759725`, 1, 1], 
     Hue[0.37271138416759725`, 1, 1], 
     Hue[0.38763216008246254`, 1, 1], 
     Hue[0.38763216008246254`, 1, 1], 
     Hue[0.4025529359973277, 1, 1], 
     Hue[0.4025529359973277, 1, 1], 
     Hue[0.4174737119121929, 1, 1], 
     Hue[0.4174737119121929, 1, 1], 
     Hue[0.432394487827058, 1, 1], 
     Hue[0.432394487827058, 1, 1], 
     Hue[0.4473152637419232, 1, 1], 
     Hue[0.4473152637419232, 1, 1], 
     Hue[0.4622360396567884, 1, 1], 
     Hue[0.4622360396567884, 1, 1], 
     Hue[0.47715681557165357`, 1, 1], 
     Hue[0.47715681557165357`, 1, 1], 
     Hue[0.49207759148651875`, 1, 1], 
     Hue[0.49207759148651875`, 1, 1], 
     Hue[0.5069983674013839, 1, 1], 
     Hue[0.5069983674013839, 1, 1], 
     Hue[0.5219191433162491, 1, 1], 
     Hue[0.5219191433162491, 1, 1], 
     Hue[0.5368399192311143, 1, 1], 
     Hue[0.5368399192311143, 1, 1], 
     Hue[0.5517606951459795, 1, 1], 
     Hue[0.5517606951459795, 1, 1], 
     Hue[0.5666814710608448, 1, 1], 
     Hue[0.5666814710608448, 1, 1], 
     Hue[0.5816022469757098, 1, 1], 
     Hue[0.5816022469757098, 1, 1], 
     Hue[0.5965230228905751, 1, 1], 
     Hue[0.5965230228905751, 1, 1], 
     Hue[0.6114437988054402, 1, 1], 
     Hue[0.6114437988054402, 1, 1], 
     Hue[0.6263645747203055, 1, 1], 
     Hue[0.6263645747203055, 1, 1], 
     Hue[0.6412853506351706, 1, 1], 
     Hue[0.6412853506351706, 1, 1], 
     Hue[0.6562061265500359, 1, 1], 
     Hue[0.6562061265500359, 1, 1], 
     Hue[0.671126902464901, 1, 1], 
     Hue[0.671126902464901, 1, 1], 
     Hue[0.6860476783797662, 1, 1], 
     Hue[0.6860476783797662, 1, 1], 
     Hue[0.7009684542946314, 1, 1], 
     Hue[0.7009684542946314, 1, 1], 
     Hue[0.7158892302094966, 1, 1], 
     Hue[0.7158892302094966, 1, 1], 
     Hue[0.7308100061243618, 1, 1], 
     Hue[0.7308100061243618, 1, 1], 
     Hue[0.7457307820392269, 1, 1], 
     Hue[0.7457307820392269, 1, 1], 
     Hue[0.7606515579540921, 1, 1], 
     Hue[0.7606515579540921, 1, 1], 
     Hue[0.7755723338689573, 1, 1], 
     Hue[0.7755723338689573, 1, 1], 
     Hue[0.7904931097838226, 1, 1], 
     Hue[0.7904931097838226, 1, 1], 
     Hue[0.8054138856986877, 1, 1], 
     Hue[0.8054138856986877, 1, 1], 
     Hue[0.820334661613553, 1, 1], 
     Hue[0.820334661613553, 1, 1], 
     Hue[0.835255437528418, 1, 1], 
     Hue[0.835255437528418, 1, 1], 
     Hue[0.8501762134432833, 1, 1], 
     Hue[0.8501762134432833, 1, 1], 
     Hue[0.8650969893581485, 1, 1], 
     Hue[0.8650969893581485, 1, 1], 
     Hue[0.8800177652730137, 1, 1], 
     Hue[0.8800177652730137, 1, 1], 
     Hue[0.8949385411878789, 1, 1], 
     Hue[0.8949385411878789, 1, 1], 
     Hue[0.909859317102744, 1, 1], 
     Hue[0.909859317102744, 1, 1], 
     Hue[0.9247800930176092, 1, 1], 
     Hue[0.9247800930176092, 1, 1], 
     Hue[0.9397008689324744, 1, 1], 
     Hue[0.9397008689324744, 1, 1], 
     Hue[0.9546216448473396, 1, 1], 
     Hue[0.9546216448473396, 1, 1], 
     Hue[0.9695424207622048, 1, 1], 
     Hue[0.9695424207622048, 1, 1], 
     Hue[0.9844631966770699, 1, 1], 
     Hue[0.9844631966770699, 1, 1], 
     Hue[0.9993839725919351, 1, 1], 
     Hue[0.9993839725919351, 1, 1], 
     Hue[0.014304748506800346`, 1, 1], 
     Hue[0.014304748506800346`, 1, 1], 
     Hue[0.02922552442166553, 1, 1], 
     Hue[0.02922552442166553, 1, 1], 
     Hue[0.04414630033653072, 1, 1], 
     Hue[0.04414630033653072, 1, 1], 
     Hue[0.059067076251395914`, 1, 1], 
     Hue[0.059067076251395914`, 1, 1], 
     Hue[0.0739878521662611, 1, 1], 
     Hue[0.0739878521662611, 1, 1], 
     Hue[0.08890862808112629, 1, 1], 
     Hue[0.08890862808112629, 1, 1], 
     Hue[0.10382940399599148`, 1, 1], 
     Hue[0.10382940399599148`, 1, 1], 
     Hue[0.11875017991085667`, 1, 1], 
     Hue[0.11875017991085667`, 1, 1], 
     Hue[0.13367095582572183`, 1, 1], 
     Hue[0.13367095582572183`, 1, 1], 
     Hue[0.14859173174058704`, 1, 1], 
     Hue[0.14859173174058704`, 1, 1], 
     Hue[0.16351250765545225`, 1, 1], 
     Hue[0.16351250765545225`, 1, 1], 
     Hue[0.17843328357031743`, 1, 1], 
     Hue[0.17843328357031743`, 1, 1], 
     Hue[0.1933540594851826, 1, 1], 
     Hue[0.1933540594851826, 1, 1], 
     Hue[0.2082748354000478, 1, 1], 
     Hue[0.2082748354000478, 1, 1], 
     Hue[0.223195611314913, 1, 1], 
     Hue[0.223195611314913, 1, 1], 
     Hue[0.23811638722977818`, 1, 1], 
     Hue[0.23811638722977818`, 1, 1], 
     Hue[0.25303716314464336`, 1, 1], 
     Hue[0.25303716314464336`, 1, 1], 
     Hue[0.26795793905950854`, 1, 1], 
     Hue[0.26795793905950854`, 1, 1], 
     Hue[0.2828787149743737, 1, 1], 
     Hue[0.2828787149743737, 1, 1], 
     Hue[0.29779949088923896`, 1, 1], 
     Hue[0.29779949088923896`, 1, 1], 
     Hue[0.31272026680410414`, 1, 1], 
     Hue[0.31272026680410414`, 1, 1], 
     Hue[0.3276410427189693, 1, 1], 
     Hue[0.3276410427189693, 1, 1], 
     Hue[0.3425618186338345, 1, 1], 
     Hue[0.3425618186338345, 1, 1], 
     Hue[0.3574825945486997, 1, 1], 
     Hue[0.3574825945486997, 1, 1], 
     Hue[0.37240337046356486`, 1, 1], 
     Hue[0.37240337046356486`, 1, 1], 
     Hue[0.38732414637843005`, 1, 1], 
     Hue[0.38732414637843005`, 1, 1], 
     Hue[0.4022449222932953, 1, 1], 
     Hue[0.4022449222932953, 1, 1], 
     Hue[0.41716569820816046`, 1, 1], 
     Hue[0.41716569820816046`, 1, 1], 
     Hue[0.43208647412302564`, 1, 1], 
     Hue[0.43208647412302564`, 1, 1], 
     Hue[0.4470072500378908, 1, 1], 
     Hue[0.4470072500378908, 1, 1], 
     Hue[0.461928025952756, 1, 1], 
     Hue[0.461928025952756, 1, 1], 
     Hue[0.4768488018676212, 1, 1], 
     Hue[0.4768488018676212, 1, 1], 
     Hue[0.49176957778248637`, 1, 1], 
     Hue[0.49176957778248637`, 1, 1], 
     Hue[0.5066903536973515, 1, 1], 
     Hue[0.5066903536973515, 1, 1], 
     Hue[0.5216111296122167, 1, 1], 
     Hue[0.5216111296122167, 1, 1], 
     Hue[0.5365319055270819, 1, 1], 
     Hue[0.5365319055270819, 1, 1], 
     Hue[0.551452681441947, 1, 1], 
     Hue[0.551452681441947, 1, 1], 
     Hue[0.5663734573568122, 1, 1], 
     Hue[0.5663734573568122, 1, 1], 
     Hue[0.5812942332716775, 1, 1], 
     Hue[0.5812942332716775, 1, 1], 
     Hue[0.5962150091865426, 1, 1], 
     Hue[0.5962150091865426, 1, 1], 
     Hue[0.6111357851014079, 1, 1], 
     Hue[0.6111357851014079, 1, 1], 
     Hue[0.6260565610162729, 1, 1], 
     Hue[0.6260565610162729, 1, 1], 
     Hue[0.6409773369311382, 1, 1], 
     Hue[0.6409773369311382, 1, 1], 
     Hue[0.6558981128460033, 1, 1], 
     Hue[0.6558981128460033, 1, 1], 
     Hue[0.6708188887608686, 1, 1], 
     Hue[0.6708188887608686, 1, 1], 
     Hue[0.6857396646757338, 1, 1], 
     Hue[0.6857396646757338, 1, 1]}],
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  Epilog->LineBox[{{1, 1.6935340580135376`*^-22}, {
     2, 2.7013089434515773`*^-22}, {3, 4.299325410417724*^-22}, {
     4, 6.827663648881313*^-22}, {5, 1.0819063225216787`*^-21}, {
     6, 1.7106176111419957`*^-21}, {7, 2.6987456705425006`*^-21}, {
     8, 4.2483150697988076`*^-21}, {9, 6.672939957808189*^-21}, {
     10, 1.0458356839810648`*^-20}, {11, 1.6355183996580148`*^-20}, {
     12, 2.5520733498368177`*^-20}, {13, 3.973531041929608*^-20}, {
     14, 6.173135468013424*^-20}, {15, 9.569312660785609*^-20}, {
     16, 1.480135317255462*^-19}, {17, 2.2843773098606825`*^-19}, {
     18, 3.5178716801562995`*^-19}, {19, 5.405525115964513*^-19}, {
     20, 8.28784318139607*^-19}, {21, 1.2679172195507736`*^-18}, {
     22, 1.9354680865679464`*^-18}, {23, 2.947995858920429*^-18}, {
     24, 4.4803656938206634`*^-18}, {25, 6.794316861745609*^-18}, {
     26, 1.0280728028685746`*^-17}, {27, 1.5522000150346483`*^-17}, {
     28, 2.3383915662751886`*^-17}, {29, 3.5150582663928854`*^-17}, {
     30, 5.2722204387178404`*^-17}, {31, 7.89042383666834*^-17}, {
     32, 1.1782917677830467`*^-16}, {33, 1.7557031748998872`*^-16}, {
     34, 2.6103280195919997`*^-16}, {35, 3.872441052560456*^-16}, {
     36, 5.732186151020675*^-16}, {37, 8.466452871414991*^-16}, {
     38, 1.2477524881167413`*^-15}, {39, 1.8348523969390077`*^-15}, {
     40, 2.6922758914159714`*^-15}, {41, 3.941701481299362*^-15}, {
     42, 5.758291398208811*^-15}, {43, 8.393619661572887*^-15}, {
     44, 1.2208172959972521`*^-14}, {45, 1.7717310670629466`*^-14}, {
     46, 2.5656101625092018`*^-14}, {47, 3.707057082579622*^-14}, {
     48, 5.344580459233516*^-14}, {49, 7.688536936935771*^-14}, {
     50, 1.1036199191656932`*^-13}, {51, 1.580669580110904*^-13}, {
     52, 2.2589592677155978`*^-13}, {53, 3.221227843234126*^-13}, {
     54, 4.583320509256901*^-13}, {55, 6.507059110734258*^-13}, {
     56, 9.217964159673821*^-13}, {57, 1.3029598043873206`*^-12}, {
     58, 1.8376921525727373`*^-12}, {59, 2.5861888981760644`*^-12}, {
     60, 3.631562016207424*^-12}, {61, 5.0882966930636755`*^-12}, {
     62, 7.113725909546953*^-12}, {63, 9.923561575543488*^-12}, {
     64, 1.3812864432954321`*^-11}, {65, 1.9184287299213337`*^-11}, {
     66, 2.6586020152748382`*^-11}, {67, 3.6762643200274067`*^-11}, {
     68, 5.072310938789075*^-11}, {69, 6.983139974211626*^-11}, {
     70, 9.592711117259212*^-11}, {71, 1.3148545822382016`*^-10}, {
     72, 1.798290384585907*^-10}, {73, 2.454074030632512*^-10}, {
     74, 3.3416521269515235`*^-10}, {75, 4.540258224893414*^-10}, {
     76, 6.155248775304267*^-10}, {77, 8.326383436888574*^-10}, {
     78, 1.1238618846365146`*^-9}, {79, 1.5136142859631868`*^-9}, {
     80, 2.0340575357737896`*^-9}, {81, 2.727451195769395*^-9}, {
     82, 3.6491900559248614`*^-9}, {83, 4.871713241369462*^-9}, {
     84, 6.4895217274559454`*^-9}, {85, 8.625601785122393*^-9}, {
     86, 1.143962682793732*^-8}, {87, 1.5138402582064123`*^-8}, {
     88, 1.9989134166329445`*^-8}, {89, 2.6336232960777236`*^-8}, {
     90, 3.4622551267281476`*^-8}, {91, 4.541613984984655*^-8}, {
     92, 5.9443874652812666`*^-8}, {93, 7.763360267872042*^-8}, {
     94, 1.0116682283297703`*^-7}, {95, 1.3154435665042106`*^-7}, {
     96, 1.7066798897368368`*^-7}, {97, 2.2094168439275953`*^-7}, {
     98, 2.853967280337101*^-7}, {99, 3.6784601760141807`*^-7}, {
     100, 4.730737680877867*^-7}, {101, 6.070681041288068*^-7}, {
     102, 7.773054327613681*^-7}, {103, 9.930971386469107*^-7}, {
     104, 1.2660110541651635`*^-6}, {105, 1.6103823595269838`*^-6}, {
     106, 2.0439310962054014`*^-6}, {107, 2.5885063646882575`*^-6}, {
     108, 3.2709805595093306`*^-6}, {109, 4.124320705162408*^-6}, {
     110, 5.188868129033276*^-6}, {111, 6.513862372589593*^-6}, {
     112, 8.159250426441251*^-6}, {113, 0.0000101978280994376}, {
     114, 0.00001271776660975753}, {115, 0.00001582558431978149}, {
     116, 0.000019649630916271522`}, {117, 0.00002434415923693149}, {
     118, 0.000030094068318430215`}, {119, 0.00003712041002129977}, {
     120, 0.00004568676067906624}, {121, 0.00005610656849737733}, {
     122, 0.00006875159673421496}, {123, 0.00008406159182671469}, {
     124, 0.00010255531435391963`}, {125, 0.00012484307875301263`}, {
     126, 0.00015164095470627406`}, {127, 0.00018378678870435632`}, {
     128, 0.00022225820803473448`}, {129, 0.00026819277085811307`}, {
     130, 0.00032291042458717297`}, {131, 0.00038793842989333545`}, {
     132, 0.0004650388987209338}, {133, 0.0005562390810359108}, {
     134, 0.0006638645160090288}, {135, 0.0007905751382558167}, {
     136, 0.0009394043979607872}, {137, 0.001113801414564607}, {
     138, 0.001317676136604145}, {139, 0.0015554474247579002`}, {
     140, 0.0018320939107596115`}, {141, 0.0021532074113328}, {
     142, 0.0025250485935685565`}, {143, 0.002954604496319363}, {
     144, 0.003449647411549908}, {145, 0.004018794520776898}, {
     146, 0.004671567565649339}, {147, 0.0054184517096023945`}, {
     148, 0.006270952620948514}, {149, 0.007241650678706022}, {
     150, 0.008344251073246413}, {151, 0.009593628447188283}, {
     152, 0.01100586460097308}, {153, 0.012598277675673225`}, {
     154, 0.014389441126572085`}, {155, 0.016399190718953217`}, {
     156, 0.018648617716584236`}, {157, 0.021160046397946577`}, {
     158, 0.023956994029747086`}, {159, 0.0270641114559759}, {
     160, 0.03050710252787448}, {161, 0.03431262070944408}, {
     162, 0.03850814134788083}, {163, 0.043121808301270076`}, {
     164, 0.04818225386893848}, {165, 0.05371839127405522}, {
     166, 0.05975917930334211}, {167, 0.06633335911384833}, {
     168, 0.07346916366912468}, {169, 0.0811940007628494}, {
     170, 0.08953411112162063}, {171, 0.09851420364336069}, {
     172, 0.10815707041522599`}, {173, 0.1184831847553152}, {
     174, 0.12951028612472462`}, {175, 0.14125295634834165`}, {
     176, 0.1537221921509343}, {177, 0.1669249795455717}, {
     178, 0.1808638760897144}, {179, 0.19553660743580398`}, {
     180, 0.21093568493342887`}, {181, 0.2270480512752962}, {
     182, 0.24385476130642741`}, {183, 0.2613307051237464}, {
     184, 0.2794443804718406}, {185, 0.2981577211826469}, {
     186, 0.31742598800714367`}, {187, 0.33719772764369327`}, {
     188, 0.3574148050814272}, {189, 0.3780125135522741}, {
     190, 0.3989197654295719}, {191, 0.4200593663358202}, {
     192, 0.44134837354154155`}, {193, 0.46269853846919623`}, {
     194, 0.48401683178139476`}, {195, 0.5052060481546798}, {
     196, 0.5261654864445083}, {197, 0.5467916995610779}, {
     198, 0.5669793070276937}, {199, 0.5866218619123392}, {
     200, 0.6056127626376299}, {201, 0.6238461991120364}, {
     202, 0.6412181217122251}, {203, 0.6576272209061722}, {
     204, 0.6729759047600129}, {205, 0.6871712612353379}, {
     206, 0.70012599207068}, {207, 0.711759305159426}, {
     208, 0.7219977526896632}, {209, 0.7307760028976727}, {
     210, 0.7380375340988992}, {211, 0.7437352406861114}, {
     212, 0.7478319420070575}, {213, 0.7503007864315618}, {
     214, 0.7511255444649425}, {215, 0.7503007864315618}, {
     216, 0.7478319420070575}, {217, 0.7437352406861114}, {
     218, 0.7380375340988992}, {219, 0.7307760028976727}, {
     220, 0.7219977526896632}, {221, 0.711759305159426}, {
     222, 0.70012599207068}, {223, 0.6871712612353379}, {
     224, 0.6729759047600129}, {225, 0.6576272209061722}, {
     226, 0.6412181217122251}, {227, 0.6238461991120364}, {
     228, 0.6056127626376299}, {229, 0.5866218619123392}, {
     230, 0.5669793070276937}, {231, 0.5467916995610779}, {
     232, 0.5261654864445083}, {233, 0.5052060481546798}, {
     234, 0.48401683178139476`}, {235, 0.46269853846919623`}, {
     236, 0.44134837354154155`}, {237, 0.4200593663358202}, {
     238, 0.3989197654295719}, {239, 0.3780125135522741}, {
     240, 0.3574148050814272}, {241, 0.33719772764369327`}, {
     242, 0.31742598800714367`}, {243, 0.2981577211826469}, {
     244, 0.2794443804718406}, {245, 0.2613307051237464}, {
     246, 0.24385476130642741`}, {247, 0.2270480512752962}, {
     248, 0.21093568493342887`}, {249, 0.19553660743580398`}, {
     250, 0.1808638760897144}, {251, 0.1669249795455717}, {
     252, 0.1537221921509343}, {253, 0.14125295634834165`}, {
     254, 0.12951028612472462`}, {255, 0.1184831847553152}, {
     256, 0.10815707041522599`}, {257, 0.09851420364336069}, {
     258, 0.08953411112162063}, {259, 0.0811940007628494}, {
     260, 0.07346916366912468}, {261, 0.06633335911384833}, {
     262, 0.05975917930334211}, {263, 0.05371839127405522}, {
     264, 0.04818225386893848}, {265, 0.043121808301270076`}, {
     266, 0.03850814134788083}, {267, 0.03431262070944408}, {
     268, 0.03050710252787448}, {269, 0.0270641114559759}, {
     270, 0.023956994029747086`}, {271, 0.021160046397946577`}, {
     272, 0.018648617716584236`}, {273, 0.016399190718953217`}, {
     274, 0.014389441126572085`}, {275, 0.012598277675673225`}, {
     276, 0.01100586460097308}, {277, 0.009593628447188283}, {
     278, 0.008344251073246413}, {279, 0.007241650678706022}, {
     280, 0.006270952620948514}, {281, 0.0054184517096023945`}, {
     282, 0.004671567565649339}, {283, 0.004018794520776898}, {
     284, 0.003449647411549908}, {285, 0.002954604496319363}, {
     286, 0.0025250485935685565`}, {287, 0.0021532074113328}, {
     288, 0.0018320939107596115`}, {289, 0.0015554474247579002`}, {
     290, 0.001317676136604145}, {291, 0.001113801414564607}, {
     292, 0.0009394043979607872}, {293, 0.0007905751382558167}, {
     294, 0.0006638645160090288}, {295, 0.0005562390810359108}, {
     296, 0.0004650388987209338}, {297, 0.00038793842989333545`}, {
     298, 0.00032291042458717297`}, {299, 0.00026819277085811307`}, {
     300, 0.00022225820803473448`}, {301, 0.00018378678870435632`}, {
     302, 0.00015164095470627406`}, {303, 0.00012484307875301263`}, {
     304, 0.00010255531435391963`}, {305, 0.00008406159182671469}, {
     306, 0.00006875159673421496}, {307, 0.00005610656849737733}, {
     308, 0.00004568676067906624}, {309, 0.00003712041002129977}, {
     310, 0.000030094068318430215`}, {311, 0.00002434415923693149}, {
     312, 0.000019649630916271522`}, {313, 0.00001582558431978149}, {
     314, 0.00001271776660975753}, {315, 0.0000101978280994376}, {
     316, 8.159250426441251*^-6}, {317, 6.513862372589593*^-6}, {
     318, 5.188868129033276*^-6}, {319, 4.124320705162408*^-6}, {
     320, 3.2709805595093306`*^-6}, {321, 2.5885063646882575`*^-6}, {
     322, 2.0439310962054014`*^-6}, {323, 1.6103823595269838`*^-6}, {
     324, 1.2660110541651635`*^-6}, {325, 9.930971386469107*^-7}, {
     326, 7.773054327613681*^-7}, {327, 6.070681041288068*^-7}, {
     328, 4.730737680877867*^-7}, {329, 3.6784601760141807`*^-7}, {
     330, 2.853967280337101*^-7}, {331, 2.2094168439275953`*^-7}, {
     332, 1.7066798897368368`*^-7}, {333, 1.3154435665042106`*^-7}, {
     334, 1.0116682283297703`*^-7}, {335, 7.763360267872042*^-8}, {
     336, 5.9443874652812666`*^-8}, {337, 4.541613984984655*^-8}, {
     338, 3.4622551267281476`*^-8}, {339, 2.6336232960777236`*^-8}, {
     340, 1.9989134166329445`*^-8}, {341, 1.5138402582064123`*^-8}, {
     342, 1.143962682793732*^-8}, {343, 8.625601785122393*^-9}, {
     344, 6.4895217274559454`*^-9}, {345, 4.871713241369462*^-9}, {
     346, 3.6491900559248614`*^-9}, {347, 2.727451195769395*^-9}, {
     348, 2.0340575357737896`*^-9}, {349, 1.5136142859631868`*^-9}, {
     350, 1.1238618846365146`*^-9}, {351, 8.326383436888574*^-10}, {
     352, 6.155248775304267*^-10}, {353, 4.540258224893414*^-10}, {
     354, 3.3416521269515235`*^-10}, {355, 2.454074030632512*^-10}, {
     356, 1.798290384585907*^-10}, {357, 1.3148545822382016`*^-10}, {
     358, 9.592711117259212*^-11}, {359, 6.983139974211626*^-11}, {
     360, 5.072310938789075*^-11}, {361, 3.6762643200274067`*^-11}, {
     362, 2.6586020152748382`*^-11}, {363, 1.9184287299213337`*^-11}, {
     364, 1.3812864432954321`*^-11}, {365, 9.923561575543488*^-12}, {
     366, 7.113725909546953*^-12}, {367, 5.0882966930636755`*^-12}, {
     368, 3.631562016207424*^-12}, {369, 2.5861888981760644`*^-12}, {
     370, 1.8376921525727373`*^-12}, {371, 1.3029598043873206`*^-12}, {
     372, 9.217964159673821*^-13}, {373, 6.507059110734258*^-13}, {
     374, 4.583320509256901*^-13}, {375, 3.221227843234126*^-13}, {
     376, 2.2589592677155978`*^-13}, {377, 1.580669580110904*^-13}, {
     378, 1.1036199191656932`*^-13}, {379, 7.688536936935771*^-14}, {
     380, 5.344580459233516*^-14}, {381, 3.707057082579622*^-14}, {
     382, 2.5656101625092018`*^-14}, {383, 1.7717310670629466`*^-14}, {
     384, 1.2208172959972521`*^-14}, {385, 8.393619661572887*^-15}, {
     386, 5.758291398208811*^-15}, {387, 3.941701481299362*^-15}, {
     388, 2.6922758914159714`*^-15}, {389, 1.8348523969390077`*^-15}, {
     390, 1.2477524881167413`*^-15}, {391, 8.466452871414991*^-16}, {
     392, 5.732186151020675*^-16}, {393, 3.872441052560456*^-16}, {
     394, 2.6103280195919997`*^-16}, {395, 1.7557031748998872`*^-16}, {
     396, 1.1782917677830467`*^-16}, {397, 7.89042383666834*^-17}, {
     398, 5.2722204387178404`*^-17}, {399, 3.5150582663928854`*^-17}, {
     400, 2.3383915662751886`*^-17}, {401, 1.5522000150346483`*^-17}, {
     402, 1.0280728028685746`*^-17}, {403, 6.794316861745609*^-18}, {
     404, 4.4803656938206634`*^-18}, {405, 2.947995858920429*^-18}, {
     406, 1.9354680865679464`*^-18}, {407, 1.2679172195507736`*^-18}, {
     408, 8.28784318139607*^-19}, {409, 5.405525115964513*^-19}, {
     410, 3.5178716801562995`*^-19}, {411, 2.2843773098606825`*^-19}, {
     412, 1.480135317255462*^-19}, {413, 9.569312660785609*^-20}, {
     414, 6.173135468013424*^-20}, {415, 3.973531041929608*^-20}, {
     416, 2.5520733498368177`*^-20}, {417, 1.6355183996580148`*^-20}, {
     418, 1.0458356839810648`*^-20}, {419, 6.672939957808189*^-21}, {
     420, 4.2483150697988076`*^-21}, {421, 2.6987456705425006`*^-21}, {
     422, 1.7106176111419957`*^-21}, {423, 1.0819063225216787`*^-21}, {
     424, 6.827663648881313*^-22}, {425, 4.299325410417724*^-22}, {
     426, 2.7013089434515773`*^-22}, {427, 1.6935340580135376`*^-22}, {
     428, 1.0593984988613142`*^-22}, {429, 6.61257348117829*^-23}, {
     430, 4.1183894115290864`*^-23}, {431, 2.5593521487963007`*^-23}, {
     432, 1.5870054665998174`*^-23}, {433, 9.81911935945938*^-24}, {
     434, 6.061950635874868*^-24}, {435, 3.734203583090401*^-24}, {
     436, 2.295246463259056*^-24}, {437, 1.4076879793215982`*^-24}, {
     438, 8.614482972582932*^-25}, {439, 5.260145671545396*^-25}, {
     440, 3.2048821599769365`*^-25}, {441, 1.9483729858668534`*^-25}, {
     442, 1.1818922176310188`*^-25}, {443, 7.153677971608614*^-26}, {
     444, 4.32042667669742*^-26}, {445, 2.6035722349781514`*^-26}, {
     446, 1.5655190751051853`*^-26}, {447, 9.392751563620501*^-27}, {
     448, 5.623064264604205*^-27}, {449, 3.3589148787591285`*^-27}, {
     450, 2.002030497666969*^-27}, {451, 1.1906609749581223`*^-27}, {
     452, 7.065636479083231*^-28}, {453, 4.183696766366554*^-28}, {
     454, 2.4718086567467686`*^-28}, {455, 1.45718683193495*^-28}, {
     456, 8.571589552496148*^-29}, {457, 5.030987465496924*^-29}, {
     458, 2.9463941490925873`*^-29}, {459, 1.72176625272489*^-29}, {
     460, 1.003929645001987*^-29}, {461, 5.8408776059798*^-30}, {
     462, 3.3907726587569325`*^-30}, {463, 1.9641062564806244`*^-30}, {
     464, 1.1352121356313588`*^-30}, {465, 6.546886486358888*^-31}, {
     466, 3.767370531463794*^-31}, {467, 2.1631548094373588`*^-31}, {
     468, 1.2393175684591396`*^-31}, {469, 7.084730763722833*^-32}, {
     470, 4.041195258768626*^-32}, {471, 2.3000753230122232`*^-32}, {
     472, 1.3062311070148853`*^-32}, {473, 7.401908719577692*^-33}, {
     474, 4.185170619230736*^-33}, {475, 2.361175945738152*^-33}, {
     476, 1.3291967548855846`*^-33}, {477, 7.466136644208733*^-34}, {
     478, 4.184545895196625*^-34}, {479, 2.3401649831940875`*^-34}, {
     480, 1.3058411716283284`*^-34}, {481, 7.270762819113856*^-35}, {
     482, 4.0393858307673845`*^-35}, {483, 2.2392183140385735`*^-35}, {
     484, 1.2385777581113318`*^-35}, {485, 6.835902439064113*^-36}, {
     486, 3.764559577505862*^-36}, {487, 2.0686081576420188`*^-36}, {
     488, 1.1341957903967286`*^-36}, {489, 6.205025473738304*^-37}, {
     490, 3.3872312400148805`*^-37}, {491, 1.8449808517896503`*^-37}, {
     492, 1.0027314115932059`*^-37}, {493, 5.437798570904108*^-38}, {
     494, 2.9424382145018097`*^-38}, {495, 1.588683291340805*^-38}, {
     496, 8.558803257863102*^-39}, {497, 4.600812116570227*^-39}, {
     498, 2.4677530254376193`*^-39}, {499, 1.3207317776463552`*^-39}, {
     500, 7.052990540688031*^-40}, {501, 3.758181194442919*^-40}, {
     502, 1.9981489925650604`*^-40}, {503, 1.0600436807047588`*^-40}, {
     504, 5.611324624114468*^-41}, {505, 2.9638264289361857`*^-41}, {
     506, 1.5620174396337128`*^-41}, {507, 8.214189732783182*^-42}, {
     508, 4.310119599512813*^-42}, {509, 2.2566263363745864`*^-42}, {
     510, 1.1788966212439895`*^-42}, {511, 6.145221386529972*^-43}, {
     512, 3.1962819575724666`*^-43}}],
  PlotRange->All,
  PlotRangeClipping->True,
  PlotRangePadding->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{
  3.3927261647926145`*^9, {3.3927265173495665`*^9, 3.391849847171875*^9}, 
   3.3918500469375*^9, 3.392472365965917*^9},
 ImageCache->GraphicsData["CompressedBitmap", "\<\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\
\>"]]
}, Open  ]],

Cell[TextData[{
 "The option ",
 StyleBox["QHorizontalRange", "InlineInput"],
 " lets you specify the plot range on the horizontal axis:"
}], "Text"],

Cell[CellGroupData[{

Cell["\<\
QListArgColorPlot[fvals,PlotRange->All, \
QHorizontalRange->QSpaceInterval[a,b,n]]\
\>", "Input"],

Cell[BoxData[
 GraphicsBox[GraphicsComplexBox[CompressedData["
1:eJyF2n90VPWVAPAv+R0CyeT3JFh9IGCALjsq0qjFd9XdCtuKo6vdUCs+qGgq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   "], {
    {GrayLevel[0.5], EdgeForm[None], 
     GraphicsGroupBox[{PolygonBox[{{71, 513, 72}, {356, 1082, 357}}], 
       PolygonBox[CompressedData["
1:eJwll3fYDmQUh5/x2nvvvffeM3vvvVcqFQ2hMkpCkqRCJSEJSShSIYmsZBdZ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        "]]}]}, {{}, {}, {
      {Hue[0., 1, 1], LineBox[{1, 2}]}, 
      {Hue[0., 1, 1], LineBox[{2, 3}]}, 
      {Hue[0., 1, 1], LineBox[{3, 4}]}, 
      {Hue[0., 1, 1], LineBox[{4, 5}]}, 
      {Hue[0., 1, 1], LineBox[{5, 6}]}, 
      {Hue[0., 1, 1], LineBox[{6, 7}]}, 
      {Hue[0., 1, 1], LineBox[{7, 8}]}, 
      {Hue[0., 1, 1], LineBox[{8, 9}]}, 
      {Hue[0., 1, 1], LineBox[{9, 10}]}, 
      {Hue[0., 1, 1], LineBox[{10, 11}]}, 
      {Hue[0., 1, 1], LineBox[{11, 12}]}, 
      {Hue[0., 1, 1], LineBox[{12, 13}]}, 
      {Hue[0., 1, 1], LineBox[{13, 14}]}, 
      {Hue[0., 1, 1], LineBox[{14, 15}]}, 
      {Hue[0., 1, 1], LineBox[{15, 16}]}, 
      {Hue[0., 1, 1], LineBox[{16, 17}]}, 
      {Hue[0., 1, 1], LineBox[{17, 18}]}, 
      {Hue[0., 1, 1], LineBox[{18, 19}]}, 
      {Hue[0., 1, 1], LineBox[{19, 20}]}, 
      {Hue[0., 1, 1], LineBox[{20, 21}]}, 
      {Hue[0., 1, 1], LineBox[{21, 22}]}, 
      {Hue[0., 1, 1], LineBox[{22, 23}]}, 
      {Hue[0., 1, 1], LineBox[{23, 24}]}, 
      {Hue[0., 1, 1], LineBox[{24, 25}]}, 
      {Hue[0., 1, 1], LineBox[{25, 26}]}, 
      {Hue[0., 1, 1], LineBox[{26, 27}]}, 
      {Hue[0., 1, 1], LineBox[{27, 28}]}, 
      {Hue[0., 1, 1], LineBox[{28, 29}]}, 
      {Hue[0., 1, 1], LineBox[{29, 30}]}, 
      {Hue[0., 1, 1], LineBox[{30, 31}]}, 
      {Hue[0., 1, 1], LineBox[{31, 32}]}, 
      {Hue[0., 1, 1], LineBox[{32, 33}]}, 
      {Hue[0., 1, 1], LineBox[{33, 34}]}, 
      {Hue[0., 1, 1], LineBox[{34, 35}]}, 
      {Hue[0., 1, 1], LineBox[{35, 36}]}, 
      {Hue[0., 1, 1], LineBox[{36, 37}]}, 
      {Hue[0., 1, 1], LineBox[{37, 38}]}, 
      {Hue[0., 1, 1], LineBox[{38, 39}]}, 
      {Hue[0., 1, 1], LineBox[{39, 40}]}, 
      {Hue[0., 1, 1], LineBox[{40, 41}]}, 
      {Hue[0., 1, 1], LineBox[{41, 42}]}, 
      {Hue[0., 1, 1], LineBox[{42, 43}]}, 
      {Hue[0., 1, 1], LineBox[{43, 44}]}, 
      {Hue[0., 1, 1], LineBox[{44, 45}]}, 
      {Hue[0., 1, 1], LineBox[{45, 46}]}, 
      {Hue[0., 1, 1], LineBox[{46, 47}]}, 
      {Hue[0., 1, 1], LineBox[{47, 48}]}, 
      {Hue[0., 1, 1], LineBox[{48, 49}]}, 
      {Hue[0., 1, 1], LineBox[{49, 50}]}, 
      {Hue[0., 1, 1], LineBox[{50, 51}]}, 
      {Hue[0., 1, 1], LineBox[{51, 52}]}, 
      {Hue[0., 1, 1], LineBox[{52, 53}]}, 
      {Hue[0., 1, 1], LineBox[{53, 54}]}, 
      {Hue[0., 1, 1], LineBox[{54, 55}]}, 
      {Hue[0., 1, 1], LineBox[{55, 56}]}, 
      {Hue[0., 1, 1], LineBox[{56, 57}]}, 
      {Hue[0., 1, 1], LineBox[{57, 58}]}, 
      {Hue[0., 1, 1], LineBox[{58, 59}]}, 
      {Hue[0., 1, 1], LineBox[{59, 60}]}, 
      {Hue[0., 1, 1], LineBox[{60, 61}]}, 
      {Hue[0., 1, 1], LineBox[{61, 62}]}, 
      {Hue[0., 1, 1], LineBox[{62, 63}]}, 
      {Hue[0., 1, 1], LineBox[{63, 64}]}, 
      {Hue[0., 1, 1], LineBox[{64, 65}]}, 
      {Hue[0., 1, 1], LineBox[{65, 66}]}, 
      {Hue[0., 1, 1], LineBox[{66, 67}]}, 
      {Hue[0., 1, 1], LineBox[{67, 68}]}, 
      {Hue[0., 1, 1], LineBox[{68, 69}]}, 
      {Hue[0., 1, 1], LineBox[{69, 70}]}, 
      {Hue[0., 1, 1], LineBox[{70, 71}]}, 
      {Hue[0., 1, 1], LineBox[{71, 72}]}, 
      {Hue[0.8812498200891433, 1, 1], LineBox[{72, 73}]}, 
      {Hue[0.8961705960040085, 1, 1], LineBox[{73, 74}]}, 
      {Hue[0.9110913719188737, 1, 1], LineBox[{74, 75}]}, 
      {Hue[0.9260121478337389, 1, 1], LineBox[{75, 76}]}, 
      {Hue[0.940932923748604, 1, 1], LineBox[{76, 77}]}, 
      {Hue[0.9558536996634692, 1, 1], LineBox[{77, 78}]}, 
      {Hue[0.9707744755783345, 1, 1], LineBox[{78, 79}]}, 
      {Hue[0.9856952514931997, 1, 1], LineBox[{79, 80}]}, 
      {Hue[0., 1, 1], LineBox[{80, 81}]}, 
      {Hue[0.015536803322930032`, 1, 1], LineBox[{81, 82}]}, 
      {Hue[0.030457579237795223`, 1, 1], LineBox[{82, 83}]}, 
      {Hue[0.0453783551526604, 1, 1], LineBox[{83, 84}]}, 
      {Hue[0.0602991310675256, 1, 1], LineBox[{84, 85}]}, 
      {Hue[0.07521990698239078, 1, 1], LineBox[{85, 86}]}, 
      {Hue[0.09014068289725596, 1, 1], LineBox[{86, 87}]}, 
      {Hue[0.10506145881212117`, 1, 1], LineBox[{87, 88}]}, 
      {Hue[0.11998223472698637`, 1, 1], LineBox[{88, 89}]}, 
      {Hue[0.13490301064185153`, 1, 1], LineBox[{89, 90}]}, 
      {Hue[0.14982378655671671`, 1, 1], LineBox[{90, 91}]}, 
      {Hue[0.16474456247158192`, 1, 1], LineBox[{91, 92}]}, 
      {Hue[0.1796653383864471, 1, 1], LineBox[{92, 93}]}, 
      {Hue[0.19458611430131229`, 1, 1], LineBox[{93, 94}]}, 
      {Hue[0.20950689021617747`, 1, 1], LineBox[{94, 95}]}, 
      {Hue[0.22442766613104267`, 1, 1], LineBox[{95, 96}]}, 
      {Hue[0.23934844204590786`, 1, 1], LineBox[{96, 97}]}, 
      {Hue[0.25426921796077306`, 1, 1], LineBox[{97, 98}]}, 
      {Hue[0.26918999387563824`, 1, 1], LineBox[{98, 99}]}, 
      {Hue[0.2841107697905034, 1, 1], LineBox[{99, 100}]}, 
      {Hue[0.2990315457053686, 1, 1], LineBox[{100, 101}]}, 
      {Hue[0.3139523216202338, 1, 1], LineBox[{101, 102}]}, 
      {Hue[0.32887309753509897`, 1, 1], LineBox[{102, 103}]}, 
      {Hue[0.34379387344996415`, 1, 1], LineBox[{103, 104}]}, 
      {Hue[0.3587146493648294, 1, 1], LineBox[{104, 105}]}, 
      {Hue[0.37363542527969457`, 1, 1], LineBox[{105, 106}]}, 
      {Hue[0.38855620119455975`, 1, 1], LineBox[{106, 107}]}, 
      {Hue[0.40347697710942493`, 1, 1], LineBox[{107, 108}]}, 
      {Hue[0.4183977530242901, 1, 1], LineBox[{108, 109}]}, 
      {Hue[0.4333185289391553, 1, 1], LineBox[{109, 110}]}, 
      {Hue[0.44823930485402047`, 1, 1], LineBox[{110, 111}]}, 
      {Hue[0.4631600807688857, 1, 1], LineBox[{111, 112}]}, 
      {Hue[0.4780808566837509, 1, 1], LineBox[{112, 113}]}, 
      {Hue[0.49300163259861607`, 1, 1], LineBox[{113, 114}]}, 
      {Hue[0.5079224085134812, 1, 1], LineBox[{114, 115}]}, 
      {Hue[0.5228431844283464, 1, 1], LineBox[{115, 116}]}, 
      {Hue[0.5377639603432116, 1, 1], LineBox[{116, 117}]}, 
      {Hue[0.5526847362580768, 1, 1], LineBox[{117, 118}]}, 
      {Hue[0.567605512172942, 1, 1], LineBox[{118, 119}]}, 
      {Hue[0.582526288087807, 1, 1], LineBox[{119, 120}]}, 
      {Hue[0.5974470640026723, 1, 1], LineBox[{120, 121}]}, 
      {Hue[0.6123678399175374, 1, 1], LineBox[{121, 122}]}, 
      {Hue[0.6272886158324027, 1, 1], LineBox[{122, 123}]}, 
      {Hue[0.6422093917472679, 1, 1], LineBox[{123, 124}]}, 
      {Hue[0.6571301676621331, 1, 1], LineBox[{124, 125}]}, 
      {Hue[0.6720509435769983, 1, 1], LineBox[{125, 126}]}, 
      {Hue[0.6869717194918634, 1, 1], LineBox[{126, 127}]}, 
      {Hue[0.7018924954067287, 1, 1], LineBox[{127, 128}]}, 
      {Hue[0.7168132713215938, 1, 1], LineBox[{128, 129}]}, 
      {Hue[0.7317340472364591, 1, 1], LineBox[{129, 130}]}, 
      {Hue[0.7466548231513241, 1, 1], LineBox[{130, 131}]}, 
      {Hue[0.7615755990661894, 1, 1], LineBox[{131, 132}]}, 
      {Hue[0.7764963749810546, 1, 1], LineBox[{132, 133}]}, 
      {Hue[0.7914171508959198, 1, 1], LineBox[{133, 134}]}, 
      {Hue[0.806337926810785, 1, 1], LineBox[{134, 135}]}, 
      {Hue[0.8212587027256502, 1, 1], LineBox[{135, 136}]}, 
      {Hue[0.8361794786405153, 1, 1], LineBox[{136, 137}]}, 
      {Hue[0.8511002545553805, 1, 1], LineBox[{137, 138}]}, 
      {Hue[0.8660210304702457, 1, 1], LineBox[{138, 139}]}, 
      {Hue[0.8809418063851109, 1, 1], LineBox[{139, 140}]}, 
      {Hue[0.8958625822999761, 1, 1], LineBox[{140, 141}]}, 
      {Hue[0.9107833582148412, 1, 1], LineBox[{141, 142}]}, 
      {Hue[0.9257041341297065, 1, 1], LineBox[{142, 143}]}, 
      {Hue[0.9406249100445717, 1, 1], LineBox[{143, 144}]}, 
      {Hue[0.9555456859594369, 1, 1], LineBox[{144, 145}]}, 
      {Hue[0.9704664618743021, 1, 1], LineBox[{145, 146}]}, 
      {Hue[0.9853872377891673, 1, 1], LineBox[{146, 147}]}, 
      {Hue[0.0003080137040324222, 1, 1], LineBox[{147, 148}]}, 
      {Hue[0.01522878961889761, 1, 1], LineBox[{148, 149}]}, 
      {Hue[0.0301495655337628, 1, 1], LineBox[{149, 150}]}, 
      {Hue[0.045070341448627996`, 1, 1], LineBox[{150, 151}]}, 
      {Hue[0.059991117363493184`, 1, 1], LineBox[{151, 152}]}, 
      {Hue[0.07491189327835837, 1, 1], LineBox[{152, 153}]}, 
      {Hue[0.08983266919322357, 1, 1], LineBox[{153, 154}]}, 
      {Hue[0.10475344510808872`, 1, 1], LineBox[{154, 155}]}, 
      {Hue[0.11967422102295393`, 1, 1], LineBox[{155, 156}]}, 
      {Hue[0.13459499693781912`, 1, 1], LineBox[{156, 157}]}, 
      {Hue[0.1495157728526843, 1, 1], LineBox[{157, 158}]}, 
      {Hue[0.16443654876754946`, 1, 1], LineBox[{158, 159}]}, 
      {Hue[0.1793573246824147, 1, 1], LineBox[{159, 160}]}, 
      {Hue[0.19427810059727987`, 1, 1], LineBox[{160, 161}]}, 
      {Hue[0.20919887651214505`, 1, 1], LineBox[{161, 162}]}, 
      {Hue[0.22411965242701024`, 1, 1], LineBox[{162, 163}]}, 
      {Hue[0.23904042834187544`, 1, 1], LineBox[{163, 164}]}, 
      {Hue[0.2539612042567406, 1, 1], LineBox[{164, 165}]}, 
      {Hue[0.2688819801716058, 1, 1], LineBox[{165, 166}]}, 
      {Hue[0.283802756086471, 1, 1], LineBox[{166, 167}]}, 
      {Hue[0.29872353200133617`, 1, 1], LineBox[{167, 168}]}, 
      {Hue[0.3136443079162014, 1, 1], LineBox[{168, 169}]}, 
      {Hue[0.32856508383106653`, 1, 1], LineBox[{169, 170}]}, 
      {Hue[0.34348585974593177`, 1, 1], LineBox[{170, 171}]}, 
      {Hue[0.358406635660797, 1, 1], LineBox[{171, 172}]}, 
      {Hue[0.3733274115756622, 1, 1], LineBox[{172, 173}]}, 
      {Hue[0.38824818749052736`, 1, 1], LineBox[{173, 174}]}, 
      {Hue[0.40316896340539254`, 1, 1], LineBox[{174, 175}]}, 
      {Hue[0.4180897393202577, 1, 1], LineBox[{175, 176}]}, 
      {Hue[0.4330105152351229, 1, 1], LineBox[{176, 177}]}, 
      {Hue[0.44793129114998803`, 1, 1], LineBox[{177, 178}]}, 
      {Hue[0.4628520670648533, 1, 1], LineBox[{178, 179}]}, 
      {Hue[0.4777728429797184, 1, 1], LineBox[{179, 180}]}, 
      {Hue[0.4926936188945837, 1, 1], LineBox[{180, 181}]}, 
      {Hue[0.5076143948094487, 1, 1], LineBox[{181, 182}]}, 
      {Hue[0.522535170724314, 1, 1], LineBox[{182, 183}]}, 
      {Hue[0.5374559466391792, 1, 1], LineBox[{183, 184}]}, 
      {Hue[0.5523767225540444, 1, 1], LineBox[{184, 185}]}, 
      {Hue[0.5672974984689095, 1, 1], LineBox[{185, 186}]}, 
      {Hue[0.5822182743837747, 1, 1], LineBox[{186, 187}]}, 
      {Hue[0.5971390502986399, 1, 1], LineBox[{187, 188}]}, 
      {Hue[0.6120598262135051, 1, 1], LineBox[{188, 189}]}, 
      {Hue[0.6269806021283703, 1, 1], LineBox[{189, 190}]}, 
      {Hue[0.6419013780432354, 1, 1], LineBox[{190, 191}]}, 
      {Hue[0.6568221539581007, 1, 1], LineBox[{191, 192}]}, 
      {Hue[0.6717429298729658, 1, 1], LineBox[{192, 193}]}, 
      {Hue[0.6866637057878311, 1, 1], LineBox[{193, 194}]}, 
      {Hue[0.7015844817026962, 1, 1], LineBox[{194, 195}]}, 
      {Hue[0.7165052576175615, 1, 1], LineBox[{195, 196}]}, 
      {Hue[0.7314260335324265, 1, 1], LineBox[{196, 197}]}, 
      {Hue[0.7463468094472918, 1, 1], LineBox[{197, 198}]}, 
      {Hue[0.761267585362157, 1, 1], LineBox[{198, 199}]}, 
      {Hue[0.7761883612770222, 1, 1], LineBox[{199, 200}]}, 
      {Hue[0.7911091371918874, 1, 1], LineBox[{200, 201}]}, 
      {Hue[0.8060299131067525, 1, 1], LineBox[{201, 202}]}, 
      {Hue[0.8209506890216177, 1, 1], LineBox[{202, 203}]}, 
      {Hue[0.8358714649364829, 1, 1], LineBox[{203, 204}]}, 
      {Hue[0.8507922408513481, 1, 1], LineBox[{204, 205}]}, 
      {Hue[0.8657130167662133, 1, 1], LineBox[{205, 206}]}, 
      {Hue[0.8806337926810786, 1, 1], LineBox[{206, 207}]}, 
      {Hue[0.8955545685959436, 1, 1], LineBox[{207, 208}]}, 
      {Hue[0.9104753445108089, 1, 1], LineBox[{208, 209}]}, 
      {Hue[0.9253961204256741, 1, 1], LineBox[{209, 210}]}, 
      {Hue[0.9403168963405393, 1, 1], LineBox[{210, 211}]}, 
      {Hue[0.9552376722554045, 1, 1], LineBox[{211, 212}]}, 
      {Hue[0.9701584481702696, 1, 1], LineBox[{212, 213}]}, 
      {Hue[0.9850792240851348, 1, 1], LineBox[{213, 214}]}, 
      {Hue[0., 1, 1], LineBox[{214, 215}]}, 
      {Hue[0.014920775914865188`, 1, 1], LineBox[{215, 216}]}, 
      {Hue[0.029841551829730376`, 1, 1], LineBox[{216, 217}]}, 
      {Hue[0.044762327744595556`, 1, 1], LineBox[{217, 218}]}, 
      {Hue[0.05968310365946075, 1, 1], LineBox[{218, 219}]}, 
      {Hue[0.07460387957432593, 1, 1], LineBox[{219, 220}]}, 
      {Hue[0.08952465548919111, 1, 1], LineBox[{220, 221}]}, 
      {Hue[0.10444543140405632`, 1, 1], LineBox[{221, 222}]}, 
      {Hue[0.11936620731892149`, 1, 1], LineBox[{222, 223}]}, 
      {Hue[0.1342869832337867, 1, 1], LineBox[{223, 224}]}, 
      {Hue[0.1492077591486519, 1, 1], LineBox[{224, 225}]}, 
      {Hue[0.16412853506351707`, 1, 1], LineBox[{225, 226}]}, 
      {Hue[0.17904931097838225`, 1, 1], LineBox[{226, 227}]}, 
      {Hue[0.19397008689324746`, 1, 1], LineBox[{227, 228}]}, 
      {Hue[0.20889086280811264`, 1, 1], LineBox[{228, 229}]}, 
      {Hue[0.22381163872297782`, 1, 1], LineBox[{229, 230}]}, 
      {Hue[0.238732414637843, 1, 1], LineBox[{230, 231}]}, 
      {Hue[0.2536531905527082, 1, 1], LineBox[{231, 232}]}, 
      {Hue[0.2685739664675734, 1, 1], LineBox[{232, 233}]}, 
      {Hue[0.2834947423824386, 1, 1], LineBox[{233, 234}]}, 
      {Hue[0.2984155182973038, 1, 1], LineBox[{234, 235}]}, 
      {Hue[0.31333629421216896`, 1, 1], LineBox[{235, 236}]}, 
      {Hue[0.32825707012703415`, 1, 1], LineBox[{236, 237}]}, 
      {Hue[0.3431778460418993, 1, 1], LineBox[{237, 238}]}, 
      {Hue[0.3580986219567645, 1, 1], LineBox[{238, 239}]}, 
      {Hue[0.37301939787162974`, 1, 1], LineBox[{239, 240}]}, 
      {Hue[0.3879401737864949, 1, 1], LineBox[{240, 241}]}, 
      {Hue[0.4028609497013601, 1, 1], LineBox[{241, 242}]}, 
      {Hue[0.4177817256162253, 1, 1], LineBox[{242, 243}]}, 
      {Hue[0.43270250153109047`, 1, 1], LineBox[{243, 244}]}, 
      {Hue[0.44762327744595565`, 1, 1], LineBox[{244, 245}]}, 
      {Hue[0.46254405336082083`, 1, 1], LineBox[{245, 246}]}, 
      {Hue[0.477464829275686, 1, 1], LineBox[{246, 247}]}, 
      {Hue[0.49238560519055125`, 1, 1], LineBox[{247, 248}]}, 
      {Hue[0.5073063811054164, 1, 1], LineBox[{248, 249}]}, 
      {Hue[0.5222271570202817, 1, 1], LineBox[{249, 250}]}, 
      {Hue[0.5371479329351467, 1, 1], LineBox[{250, 251}]}, 
      {Hue[0.552068708850012, 1, 1], LineBox[{251, 252}]}, 
      {Hue[0.5669894847648771, 1, 1], LineBox[{252, 253}]}, 
      {Hue[0.5819102606797423, 1, 1], LineBox[{253, 254}]}, 
      {Hue[0.5968310365946075, 1, 1], LineBox[{254, 255}]}, 
      {Hue[0.6117518125094726, 1, 1], LineBox[{255, 256}]}, 
      {Hue[0.6266725884243378, 1, 1], LineBox[{256, 257}]}, 
      {Hue[0.641593364339203, 1, 1], LineBox[{257, 258}]}, 
      {Hue[0.6565141402540682, 1, 1], LineBox[{258, 259}]}, 
      {Hue[0.6714349161689335, 1, 1], LineBox[{259, 260}]}, 
      {Hue[0.6863556920837985, 1, 1], LineBox[{260, 261}]}, 
      {Hue[0.7012764679986638, 1, 1], LineBox[{261, 262}]}, 
      {Hue[0.716197243913529, 1, 1], LineBox[{262, 263}]}, 
      {Hue[0.7311180198283942, 1, 1], LineBox[{263, 264}]}, 
      {Hue[0.7460387957432594, 1, 1], LineBox[{264, 265}]}, 
      {Hue[0.7609595716581246, 1, 1], LineBox[{265, 266}]}, 
      {Hue[0.7758803475729897, 1, 1], LineBox[{266, 267}]}, 
      {Hue[0.7908011234878549, 1, 1], LineBox[{267, 268}]}, 
      {Hue[0.8057218994027201, 1, 1], LineBox[{268, 269}]}, 
      {Hue[0.8206426753175853, 1, 1], LineBox[{269, 270}]}, 
      {Hue[0.8355634512324506, 1, 1], LineBox[{270, 271}]}, 
      {Hue[0.8504842271473156, 1, 1], LineBox[{271, 272}]}, 
      {Hue[0.8654050030621809, 1, 1], LineBox[{272, 273}]}, 
      {Hue[0.8803257789770461, 1, 1], LineBox[{273, 274}]}, 
      {Hue[0.8952465548919113, 1, 1], LineBox[{274, 275}]}, 
      {Hue[0.9101673308067765, 1, 1], LineBox[{275, 276}]}, 
      {Hue[0.9250881067216417, 1, 1], LineBox[{276, 277}]}, 
      {Hue[0.9400088826365068, 1, 1], LineBox[{277, 278}]}, 
      {Hue[0.954929658551372, 1, 1], LineBox[{278, 279}]}, 
      {Hue[0.9698504344662372, 1, 1], LineBox[{279, 280}]}, 
      {Hue[0.9847712103811024, 1, 1], LineBox[{280, 281}]}, 
      {Hue[0.9996919862959676, 1, 1], LineBox[{281, 282}]}, 
      {Hue[0.014612762210832766`, 1, 1], LineBox[{282, 283}]}, 
      {Hue[0.029533538125697957`, 1, 1], LineBox[{283, 284}]}, 
      {Hue[0.044454314040563145`, 1, 1], LineBox[{284, 285}]}, 
      {Hue[0.05937508995542833, 1, 1], LineBox[{285, 286}]}, 
      {Hue[0.07429586587029352, 1, 1], LineBox[{286, 287}]}, 
      {Hue[0.08921664178515872, 1, 1], LineBox[{287, 288}]}, 
      {Hue[0.1041374177000239, 1, 1], LineBox[{288, 289}]}, 
      {Hue[0.11905819361488908`, 1, 1], LineBox[{289, 290}]}, 
      {Hue[0.13397896952975427`, 1, 1], LineBox[{290, 291}]}, 
      {Hue[0.14889974544461948`, 1, 1], LineBox[{291, 292}]}, 
      {Hue[0.16382052135948466`, 1, 1], LineBox[{292, 293}]}, 
      {Hue[0.17874129727434984`, 1, 1], LineBox[{293, 294}]}, 
      {Hue[0.19366207318921502`, 1, 1], LineBox[{294, 295}]}, 
      {Hue[0.20858284910408023`, 1, 1], LineBox[{295, 296}]}, 
      {Hue[0.2235036250189454, 1, 1], LineBox[{296, 297}]}, 
      {Hue[0.2384244009338106, 1, 1], LineBox[{297, 298}]}, 
      {Hue[0.2533451768486758, 1, 1], LineBox[{298, 299}]}, 
      {Hue[0.268265952763541, 1, 1], LineBox[{299, 300}]}, 
      {Hue[0.28318672867840616`, 1, 1], LineBox[{300, 301}]}, 
      {Hue[0.29810750459327134`, 1, 1], LineBox[{301, 302}]}, 
      {Hue[0.3130282805081365, 1, 1], LineBox[{302, 303}]}, 
      {Hue[0.3279490564230017, 1, 1], LineBox[{303, 304}]}, 
      {Hue[0.34286983233786694`, 1, 1], LineBox[{304, 305}]}, 
      {Hue[0.3577906082527321, 1, 1], LineBox[{305, 306}]}, 
      {Hue[0.37271138416759725`, 1, 1], LineBox[{306, 307}]}, 
      {Hue[0.38763216008246254`, 1, 1], LineBox[{307, 308}]}, 
      {Hue[0.4025529359973277, 1, 1], LineBox[{308, 309}]}, 
      {Hue[0.4174737119121929, 1, 1], LineBox[{309, 310}]}, 
      {Hue[0.432394487827058, 1, 1], LineBox[{310, 311}]}, 
      {Hue[0.4473152637419232, 1, 1], LineBox[{311, 312}]}, 
      {Hue[0.4622360396567884, 1, 1], LineBox[{312, 313}]}, 
      {Hue[0.47715681557165357`, 1, 1], LineBox[{313, 314}]}, 
      {Hue[0.49207759148651875`, 1, 1], LineBox[{314, 315}]}, 
      {Hue[0.5069983674013839, 1, 1], LineBox[{315, 316}]}, 
      {Hue[0.5219191433162491, 1, 1], LineBox[{316, 317}]}, 
      {Hue[0.5368399192311143, 1, 1], LineBox[{317, 318}]}, 
      {Hue[0.5517606951459795, 1, 1], LineBox[{318, 319}]}, 
      {Hue[0.5666814710608448, 1, 1], LineBox[{319, 320}]}, 
      {Hue[0.5816022469757098, 1, 1], LineBox[{320, 321}]}, 
      {Hue[0.5965230228905751, 1, 1], LineBox[{321, 322}]}, 
      {Hue[0.6114437988054402, 1, 1], LineBox[{322, 323}]}, 
      {Hue[0.6263645747203055, 1, 1], LineBox[{323, 324}]}, 
      {Hue[0.6412853506351706, 1, 1], LineBox[{324, 325}]}, 
      {Hue[0.6562061265500359, 1, 1], LineBox[{325, 326}]}, 
      {Hue[0.671126902464901, 1, 1], LineBox[{326, 327}]}, 
      {Hue[0.6860476783797662, 1, 1], LineBox[{327, 328}]}, 
      {Hue[0.7009684542946314, 1, 1], LineBox[{328, 329}]}, 
      {Hue[0.7158892302094966, 1, 1], LineBox[{329, 330}]}, 
      {Hue[0.7308100061243618, 1, 1], LineBox[{330, 331}]}, 
      {Hue[0.7457307820392269, 1, 1], LineBox[{331, 332}]}, 
      {Hue[0.7606515579540921, 1, 1], LineBox[{332, 333}]}, 
      {Hue[0.7755723338689573, 1, 1], LineBox[{333, 334}]}, 
      {Hue[0.7904931097838226, 1, 1], LineBox[{334, 335}]}, 
      {Hue[0.8054138856986877, 1, 1], LineBox[{335, 336}]}, 
      {Hue[0.820334661613553, 1, 1], LineBox[{336, 337}]}, 
      {Hue[0.835255437528418, 1, 1], LineBox[{337, 338}]}, 
      {Hue[0.8501762134432833, 1, 1], LineBox[{338, 339}]}, 
      {Hue[0.8650969893581485, 1, 1], LineBox[{339, 340}]}, 
      {Hue[0.8800177652730137, 1, 1], LineBox[{340, 341}]}, 
      {Hue[0.8949385411878789, 1, 1], LineBox[{341, 342}]}, 
      {Hue[0.909859317102744, 1, 1], LineBox[{342, 343}]}, 
      {Hue[0.9247800930176092, 1, 1], LineBox[{343, 344}]}, 
      {Hue[0.9397008689324744, 1, 1], LineBox[{344, 345}]}, 
      {Hue[0.9546216448473396, 1, 1], LineBox[{345, 346}]}, 
      {Hue[0.9695424207622048, 1, 1], LineBox[{346, 347}]}, 
      {Hue[0.9844631966770699, 1, 1], LineBox[{347, 348}]}, 
      {Hue[0., 1, 1], LineBox[{348, 349}]}, 
      {Hue[0.014304748506800346`, 1, 1], LineBox[{349, 350}]}, 
      {Hue[0.02922552442166553, 1, 1], LineBox[{350, 351}]}, 
      {Hue[0.04414630033653072, 1, 1], LineBox[{351, 352}]}, 
      {Hue[0.059067076251395914`, 1, 1], LineBox[{352, 353}]}, 
      {Hue[0.0739878521662611, 1, 1], LineBox[{353, 354}]}, 
      {Hue[0.08890862808112629, 1, 1], LineBox[{354, 355}]}, 
      {Hue[0.10382940399599148`, 1, 1], LineBox[{355, 356}]}, 
      {Hue[0.11875017991085667`, 1, 1], LineBox[{356, 357}]}, 
      {Hue[0., 1, 1], LineBox[{357, 358}]}, 
      {Hue[0., 1, 1], LineBox[{358, 359}]}, 
      {Hue[0., 1, 1], LineBox[{359, 360}]}, 
      {Hue[0., 1, 1], LineBox[{360, 361}]}, 
      {Hue[0., 1, 1], LineBox[{361, 362}]}, 
      {Hue[0., 1, 1], LineBox[{362, 363}]}, 
      {Hue[0., 1, 1], LineBox[{363, 364}]}, 
      {Hue[0., 1, 1], LineBox[{364, 365}]}, 
      {Hue[0., 1, 1], LineBox[{365, 366}]}, 
      {Hue[0., 1, 1], LineBox[{366, 367}]}, 
      {Hue[0., 1, 1], LineBox[{367, 368}]}, 
      {Hue[0., 1, 1], LineBox[{368, 369}]}, 
      {Hue[0., 1, 1], LineBox[{369, 370}]}, 
      {Hue[0., 1, 1], LineBox[{370, 371}]}, 
      {Hue[0., 1, 1], LineBox[{371, 372}]}, 
      {Hue[0., 1, 1], LineBox[{372, 373}]}, 
      {Hue[0., 1, 1], LineBox[{373, 374}]}, 
      {Hue[0., 1, 1], LineBox[{374, 375}]}, 
      {Hue[0., 1, 1], LineBox[{375, 376}]}, 
      {Hue[0., 1, 1], LineBox[{376, 377}]}, 
      {Hue[0., 1, 1], LineBox[{377, 378}]}, 
      {Hue[0., 1, 1], LineBox[{378, 379}]}, 
      {Hue[0., 1, 1], LineBox[{379, 380}]}, 
      {Hue[0., 1, 1], LineBox[{380, 381}]}, 
      {Hue[0., 1, 1], LineBox[{381, 382}]}, 
      {Hue[0., 1, 1], LineBox[{382, 383}]}, 
      {Hue[0., 1, 1], LineBox[{383, 384}]}, 
      {Hue[0., 1, 1], LineBox[{384, 385}]}, 
      {Hue[0., 1, 1], LineBox[{385, 386}]}, 
      {Hue[0., 1, 1], LineBox[{386, 387}]}, 
      {Hue[0., 1, 1], LineBox[{387, 388}]}, 
      {Hue[0., 1, 1], LineBox[{388, 389}]}, 
      {Hue[0., 1, 1], LineBox[{389, 390}]}, 
      {Hue[0., 1, 1], LineBox[{390, 391}]}, 
      {Hue[0., 1, 1], LineBox[{391, 392}]}, 
      {Hue[0., 1, 1], LineBox[{392, 393}]}, 
      {Hue[0., 1, 1], LineBox[{393, 394}]}, 
      {Hue[0., 1, 1], LineBox[{394, 395}]}, 
      {Hue[0., 1, 1], LineBox[{395, 396}]}, 
      {Hue[0., 1, 1], LineBox[{396, 397}]}, 
      {Hue[0., 1, 1], LineBox[{397, 398}]}, 
      {Hue[0., 1, 1], LineBox[{398, 399}]}, 
      {Hue[0., 1, 1], LineBox[{399, 400}]}, 
      {Hue[0., 1, 1], LineBox[{400, 401}]}, 
      {Hue[0., 1, 1], LineBox[{401, 402}]}, 
      {Hue[0., 1, 1], LineBox[{402, 403}]}, 
      {Hue[0., 1, 1], LineBox[{403, 404}]}, 
      {Hue[0., 1, 1], LineBox[{404, 405}]}, 
      {Hue[0., 1, 1], LineBox[{405, 406}]}, 
      {Hue[0., 1, 1], LineBox[{406, 407}]}, 
      {Hue[0., 1, 1], LineBox[{407, 408}]}, 
      {Hue[0., 1, 1], LineBox[{408, 409}]}, 
      {Hue[0., 1, 1], LineBox[{409, 410}]}, 
      {Hue[0., 1, 1], LineBox[{410, 411}]}, 
      {Hue[0., 1, 1], LineBox[{411, 412}]}, 
      {Hue[0., 1, 1], LineBox[{412, 413}]}, 
      {Hue[0., 1, 1], LineBox[{413, 414}]}, 
      {Hue[0., 1, 1], LineBox[{414, 415}]}, 
      {Hue[0., 1, 1], LineBox[{415, 416}]}, 
      {Hue[0., 1, 1], LineBox[{416, 417}]}, 
      {Hue[0., 1, 1], LineBox[{417, 418}]}, 
      {Hue[0., 1, 1], LineBox[{418, 419}]}, 
      {Hue[0., 1, 1], LineBox[{419, 420}]}, 
      {Hue[0., 1, 1], LineBox[{420, 421}]}, 
      {Hue[0., 1, 1], LineBox[{421, 422}]}, 
      {Hue[0., 1, 1], LineBox[{422, 423}]}, 
      {Hue[0., 1, 1], LineBox[{423, 424}]}, 
      {Hue[0., 1, 1], LineBox[{424, 425}]}, 
      {Hue[0., 1, 1], LineBox[{425, 426}]}, 
      {Hue[0., 1, 1], LineBox[{426, 427}]}, 
      {Hue[0., 1, 1], LineBox[{427, 428}]}, 
      {Hue[0., 1, 1], LineBox[{428, 429}]}, 
      {Hue[0., 1, 1], LineBox[{429, 430}]}, 
      {Hue[0., 1, 1], LineBox[{430, 431}]}, 
      {Hue[0., 1, 1], LineBox[{431, 432}]}, 
      {Hue[0., 1, 1], LineBox[{432, 433}]}, 
      {Hue[0., 1, 1], LineBox[{433, 434}]}, 
      {Hue[0., 1, 1], LineBox[{434, 435}]}, 
      {Hue[0., 1, 1], LineBox[{435, 436}]}, 
      {Hue[0., 1, 1], LineBox[{436, 437}]}, 
      {Hue[0., 1, 1], LineBox[{437, 438}]}, 
      {Hue[0., 1, 1], LineBox[{438, 439}]}, 
      {Hue[0., 1, 1], LineBox[{439, 440}]}, 
      {Hue[0., 1, 1], LineBox[{440, 441}]}, 
      {Hue[0., 1, 1], LineBox[{441, 442}]}, 
      {Hue[0., 1, 1], LineBox[{442, 443}]}, 
      {Hue[0., 1, 1], LineBox[{443, 444}]}, 
      {Hue[0., 1, 1], LineBox[{444, 445}]}, 
      {Hue[0., 1, 1], LineBox[{445, 446}]}, 
      {Hue[0., 1, 1], LineBox[{446, 447}]}, 
      {Hue[0., 1, 1], LineBox[{447, 448}]}, 
      {Hue[0., 1, 1], LineBox[{448, 449}]}, 
      {Hue[0., 1, 1], LineBox[{449, 450}]}, 
      {Hue[0., 1, 1], LineBox[{450, 451}]}, 
      {Hue[0., 1, 1], LineBox[{451, 452}]}, 
      {Hue[0., 1, 1], LineBox[{452, 453}]}, 
      {Hue[0., 1, 1], LineBox[{453, 454}]}, 
      {Hue[0., 1, 1], LineBox[{454, 455}]}, 
      {Hue[0., 1, 1], LineBox[{455, 456}]}, 
      {Hue[0., 1, 1], LineBox[{456, 457}]}, 
      {Hue[0., 1, 1], LineBox[{457, 458}]}, 
      {Hue[0., 1, 1], LineBox[{458, 459}]}, 
      {Hue[0., 1, 1], LineBox[{459, 460}]}, 
      {Hue[0., 1, 1], LineBox[{460, 461}]}, 
      {Hue[0., 1, 1], LineBox[{461, 462}]}, 
      {Hue[0., 1, 1], LineBox[{462, 463}]}, 
      {Hue[0., 1, 1], LineBox[{463, 464}]}, 
      {Hue[0., 1, 1], LineBox[{464, 465}]}, 
      {Hue[0., 1, 1], LineBox[{465, 466}]}, 
      {Hue[0., 1, 1], LineBox[{466, 467}]}, 
      {Hue[0., 1, 1], LineBox[{467, 468}]}, 
      {Hue[0., 1, 1], LineBox[{468, 469}]}, 
      {Hue[0., 1, 1], LineBox[{469, 470}]}, 
      {Hue[0., 1, 1], LineBox[{470, 471}]}, 
      {Hue[0., 1, 1], LineBox[{471, 472}]}, 
      {Hue[0., 1, 1], LineBox[{472, 473}]}, 
      {Hue[0., 1, 1], LineBox[{473, 474}]}, 
      {Hue[0., 1, 1], LineBox[{474, 475}]}, 
      {Hue[0., 1, 1], LineBox[{475, 476}]}, 
      {Hue[0., 1, 1], LineBox[{476, 477}]}, 
      {Hue[0., 1, 1], LineBox[{477, 478}]}, 
      {Hue[0., 1, 1], LineBox[{478, 479}]}, 
      {Hue[0., 1, 1], LineBox[{479, 480}]}, 
      {Hue[0., 1, 1], LineBox[{480, 481}]}, 
      {Hue[0., 1, 1], LineBox[{481, 482}]}, 
      {Hue[0., 1, 1], LineBox[{482, 483}]}, 
      {Hue[0., 1, 1], LineBox[{483, 484}]}, 
      {Hue[0., 1, 1], LineBox[{484, 485}]}, 
      {Hue[0., 1, 1], LineBox[{485, 486}]}, 
      {Hue[0., 1, 1], LineBox[{486, 487}]}, 
      {Hue[0., 1, 1], LineBox[{487, 488}]}, 
      {Hue[0., 1, 1], LineBox[{488, 489}]}, 
      {Hue[0., 1, 1], LineBox[{489, 490}]}, 
      {Hue[0., 1, 1], LineBox[{490, 491}]}, 
      {Hue[0., 1, 1], LineBox[{491, 492}]}, 
      {Hue[0., 1, 1], LineBox[{492, 493}]}, 
      {Hue[0., 1, 1], LineBox[{493, 494}]}, 
      {Hue[0., 1, 1], LineBox[{494, 495}]}, 
      {Hue[0., 1, 1], LineBox[{495, 496}]}, 
      {Hue[0., 1, 1], LineBox[{496, 497}]}, 
      {Hue[0., 1, 1], LineBox[{497, 498}]}, 
      {Hue[0., 1, 1], LineBox[{498, 499}]}, 
      {Hue[0., 1, 1], LineBox[{499, 500}]}, 
      {Hue[0., 1, 1], LineBox[{500, 501}]}, 
      {Hue[0., 1, 1], LineBox[{501, 502}]}, 
      {Hue[0., 1, 1], LineBox[{502, 503}]}, 
      {Hue[0., 1, 1], LineBox[{503, 504}]}, 
      {Hue[0., 1, 1], LineBox[{504, 505}]}, 
      {Hue[0., 1, 1], LineBox[{505, 506}]}, 
      {Hue[0., 1, 1], LineBox[{506, 507}]}, 
      {Hue[0., 1, 1], LineBox[{507, 508}]}, 
      {Hue[0., 1, 1], LineBox[{508, 509}]}, 
      {Hue[0., 1, 1], LineBox[{509, 510}]}, 
      {Hue[0., 1, 1], LineBox[{510, 511}]}, 
      {Hue[0., 1, 1], LineBox[{511, 512}]}}}},
   VertexColors->{
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.8812498200891433, 1, 1], 
     Hue[0.8961705960040085, 1, 1], 
     Hue[0.9110913719188737, 1, 1], 
     Hue[0.9260121478337389, 1, 1], 
     Hue[0.940932923748604, 1, 1], 
     Hue[0.9558536996634692, 1, 1], 
     Hue[0.9707744755783345, 1, 1], 
     Hue[0.9856952514931997, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.015536803322930032`, 1, 1], 
     Hue[0.030457579237795223`, 1, 1], 
     Hue[0.0453783551526604, 1, 1], 
     Hue[0.0602991310675256, 1, 1], 
     Hue[0.07521990698239078, 1, 1], 
     Hue[0.09014068289725596, 1, 1], 
     Hue[0.10506145881212117`, 1, 1], 
     Hue[0.11998223472698637`, 1, 1], 
     Hue[0.13490301064185153`, 1, 1], 
     Hue[0.14982378655671671`, 1, 1], 
     Hue[0.16474456247158192`, 1, 1], 
     Hue[0.1796653383864471, 1, 1], 
     Hue[0.19458611430131229`, 1, 1], 
     Hue[0.20950689021617747`, 1, 1], 
     Hue[0.22442766613104267`, 1, 1], 
     Hue[0.23934844204590786`, 1, 1], 
     Hue[0.25426921796077306`, 1, 1], 
     Hue[0.26918999387563824`, 1, 1], 
     Hue[0.2841107697905034, 1, 1], 
     Hue[0.2990315457053686, 1, 1], 
     Hue[0.3139523216202338, 1, 1], 
     Hue[0.32887309753509897`, 1, 1], 
     Hue[0.34379387344996415`, 1, 1], 
     Hue[0.3587146493648294, 1, 1], 
     Hue[0.37363542527969457`, 1, 1], 
     Hue[0.38855620119455975`, 1, 1], 
     Hue[0.40347697710942493`, 1, 1], 
     Hue[0.4183977530242901, 1, 1], 
     Hue[0.4333185289391553, 1, 1], 
     Hue[0.44823930485402047`, 1, 1], 
     Hue[0.4631600807688857, 1, 1], 
     Hue[0.4780808566837509, 1, 1], 
     Hue[0.49300163259861607`, 1, 1], 
     Hue[0.5079224085134812, 1, 1], 
     Hue[0.5228431844283464, 1, 1], 
     Hue[0.5377639603432116, 1, 1], 
     Hue[0.5526847362580768, 1, 1], 
     Hue[0.567605512172942, 1, 1], 
     Hue[0.582526288087807, 1, 1], 
     Hue[0.5974470640026723, 1, 1], 
     Hue[0.6123678399175374, 1, 1], 
     Hue[0.6272886158324027, 1, 1], 
     Hue[0.6422093917472679, 1, 1], 
     Hue[0.6571301676621331, 1, 1], 
     Hue[0.6720509435769983, 1, 1], 
     Hue[0.6869717194918634, 1, 1], 
     Hue[0.7018924954067287, 1, 1], 
     Hue[0.7168132713215938, 1, 1], 
     Hue[0.7317340472364591, 1, 1], 
     Hue[0.7466548231513241, 1, 1], 
     Hue[0.7615755990661894, 1, 1], 
     Hue[0.7764963749810546, 1, 1], 
     Hue[0.7914171508959198, 1, 1], 
     Hue[0.806337926810785, 1, 1], 
     Hue[0.8212587027256502, 1, 1], 
     Hue[0.8361794786405153, 1, 1], 
     Hue[0.8511002545553805, 1, 1], 
     Hue[0.8660210304702457, 1, 1], 
     Hue[0.8809418063851109, 1, 1], 
     Hue[0.8958625822999761, 1, 1], 
     Hue[0.9107833582148412, 1, 1], 
     Hue[0.9257041341297065, 1, 1], 
     Hue[0.9406249100445717, 1, 1], 
     Hue[0.9555456859594369, 1, 1], 
     Hue[0.9704664618743021, 1, 1], 
     Hue[0.9853872377891673, 1, 1], 
     Hue[0.0003080137040324222, 1, 1], 
     Hue[0.01522878961889761, 1, 1], 
     Hue[0.0301495655337628, 1, 1], 
     Hue[0.045070341448627996`, 1, 1], 
     Hue[0.059991117363493184`, 1, 1], 
     Hue[0.07491189327835837, 1, 1], 
     Hue[0.08983266919322357, 1, 1], 
     Hue[0.10475344510808872`, 1, 1], 
     Hue[0.11967422102295393`, 1, 1], 
     Hue[0.13459499693781912`, 1, 1], 
     Hue[0.1495157728526843, 1, 1], 
     Hue[0.16443654876754946`, 1, 1], 
     Hue[0.1793573246824147, 1, 1], 
     Hue[0.19427810059727987`, 1, 1], 
     Hue[0.20919887651214505`, 1, 1], 
     Hue[0.22411965242701024`, 1, 1], 
     Hue[0.23904042834187544`, 1, 1], 
     Hue[0.2539612042567406, 1, 1], 
     Hue[0.2688819801716058, 1, 1], 
     Hue[0.283802756086471, 1, 1], 
     Hue[0.29872353200133617`, 1, 1], 
     Hue[0.3136443079162014, 1, 1], 
     Hue[0.32856508383106653`, 1, 1], 
     Hue[0.34348585974593177`, 1, 1], 
     Hue[0.358406635660797, 1, 1], 
     Hue[0.3733274115756622, 1, 1], 
     Hue[0.38824818749052736`, 1, 1], 
     Hue[0.40316896340539254`, 1, 1], 
     Hue[0.4180897393202577, 1, 1], 
     Hue[0.4330105152351229, 1, 1], 
     Hue[0.44793129114998803`, 1, 1], 
     Hue[0.4628520670648533, 1, 1], 
     Hue[0.4777728429797184, 1, 1], 
     Hue[0.4926936188945837, 1, 1], 
     Hue[0.5076143948094487, 1, 1], 
     Hue[0.522535170724314, 1, 1], 
     Hue[0.5374559466391792, 1, 1], 
     Hue[0.5523767225540444, 1, 1], 
     Hue[0.5672974984689095, 1, 1], 
     Hue[0.5822182743837747, 1, 1], 
     Hue[0.5971390502986399, 1, 1], 
     Hue[0.6120598262135051, 1, 1], 
     Hue[0.6269806021283703, 1, 1], 
     Hue[0.6419013780432354, 1, 1], 
     Hue[0.6568221539581007, 1, 1], 
     Hue[0.6717429298729658, 1, 1], 
     Hue[0.6866637057878311, 1, 1], 
     Hue[0.7015844817026962, 1, 1], 
     Hue[0.7165052576175615, 1, 1], 
     Hue[0.7314260335324265, 1, 1], 
     Hue[0.7463468094472918, 1, 1], 
     Hue[0.761267585362157, 1, 1], 
     Hue[0.7761883612770222, 1, 1], 
     Hue[0.7911091371918874, 1, 1], 
     Hue[0.8060299131067525, 1, 1], 
     Hue[0.8209506890216177, 1, 1], 
     Hue[0.8358714649364829, 1, 1], 
     Hue[0.8507922408513481, 1, 1], 
     Hue[0.8657130167662133, 1, 1], 
     Hue[0.8806337926810786, 1, 1], 
     Hue[0.8955545685959436, 1, 1], 
     Hue[0.9104753445108089, 1, 1], 
     Hue[0.9253961204256741, 1, 1], 
     Hue[0.9403168963405393, 1, 1], 
     Hue[0.9552376722554045, 1, 1], 
     Hue[0.9701584481702696, 1, 1], 
     Hue[0.9850792240851348, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.014920775914865188`, 1, 1], 
     Hue[0.029841551829730376`, 1, 1], 
     Hue[0.044762327744595556`, 1, 1], 
     Hue[0.05968310365946075, 1, 1], 
     Hue[0.07460387957432593, 1, 1], 
     Hue[0.08952465548919111, 1, 1], 
     Hue[0.10444543140405632`, 1, 1], 
     Hue[0.11936620731892149`, 1, 1], 
     Hue[0.1342869832337867, 1, 1], 
     Hue[0.1492077591486519, 1, 1], 
     Hue[0.16412853506351707`, 1, 1], 
     Hue[0.17904931097838225`, 1, 1], 
     Hue[0.19397008689324746`, 1, 1], 
     Hue[0.20889086280811264`, 1, 1], 
     Hue[0.22381163872297782`, 1, 1], 
     Hue[0.238732414637843, 1, 1], 
     Hue[0.2536531905527082, 1, 1], 
     Hue[0.2685739664675734, 1, 1], 
     Hue[0.2834947423824386, 1, 1], 
     Hue[0.2984155182973038, 1, 1], 
     Hue[0.31333629421216896`, 1, 1], 
     Hue[0.32825707012703415`, 1, 1], 
     Hue[0.3431778460418993, 1, 1], 
     Hue[0.3580986219567645, 1, 1], 
     Hue[0.37301939787162974`, 1, 1], 
     Hue[0.3879401737864949, 1, 1], 
     Hue[0.4028609497013601, 1, 1], 
     Hue[0.4177817256162253, 1, 1], 
     Hue[0.43270250153109047`, 1, 1], 
     Hue[0.44762327744595565`, 1, 1], 
     Hue[0.46254405336082083`, 1, 1], 
     Hue[0.477464829275686, 1, 1], 
     Hue[0.49238560519055125`, 1, 1], 
     Hue[0.5073063811054164, 1, 1], 
     Hue[0.5222271570202817, 1, 1], 
     Hue[0.5371479329351467, 1, 1], 
     Hue[0.552068708850012, 1, 1], 
     Hue[0.5669894847648771, 1, 1], 
     Hue[0.5819102606797423, 1, 1], 
     Hue[0.5968310365946075, 1, 1], 
     Hue[0.6117518125094726, 1, 1], 
     Hue[0.6266725884243378, 1, 1], 
     Hue[0.641593364339203, 1, 1], 
     Hue[0.6565141402540682, 1, 1], 
     Hue[0.6714349161689335, 1, 1], 
     Hue[0.6863556920837985, 1, 1], 
     Hue[0.7012764679986638, 1, 1], 
     Hue[0.716197243913529, 1, 1], 
     Hue[0.7311180198283942, 1, 1], 
     Hue[0.7460387957432594, 1, 1], 
     Hue[0.7609595716581246, 1, 1], 
     Hue[0.7758803475729897, 1, 1], 
     Hue[0.7908011234878549, 1, 1], 
     Hue[0.8057218994027201, 1, 1], 
     Hue[0.8206426753175853, 1, 1], 
     Hue[0.8355634512324506, 1, 1], 
     Hue[0.8504842271473156, 1, 1], 
     Hue[0.8654050030621809, 1, 1], 
     Hue[0.8803257789770461, 1, 1], 
     Hue[0.8952465548919113, 1, 1], 
     Hue[0.9101673308067765, 1, 1], 
     Hue[0.9250881067216417, 1, 1], 
     Hue[0.9400088826365068, 1, 1], 
     Hue[0.954929658551372, 1, 1], 
     Hue[0.9698504344662372, 1, 1], 
     Hue[0.9847712103811024, 1, 1], 
     Hue[0.9996919862959676, 1, 1], 
     Hue[0.014612762210832766`, 1, 1], 
     Hue[0.029533538125697957`, 1, 1], 
     Hue[0.044454314040563145`, 1, 1], 
     Hue[0.05937508995542833, 1, 1], 
     Hue[0.07429586587029352, 1, 1], 
     Hue[0.08921664178515872, 1, 1], 
     Hue[0.1041374177000239, 1, 1], 
     Hue[0.11905819361488908`, 1, 1], 
     Hue[0.13397896952975427`, 1, 1], 
     Hue[0.14889974544461948`, 1, 1], 
     Hue[0.16382052135948466`, 1, 1], 
     Hue[0.17874129727434984`, 1, 1], 
     Hue[0.19366207318921502`, 1, 1], 
     Hue[0.20858284910408023`, 1, 1], 
     Hue[0.2235036250189454, 1, 1], 
     Hue[0.2384244009338106, 1, 1], 
     Hue[0.2533451768486758, 1, 1], 
     Hue[0.268265952763541, 1, 1], 
     Hue[0.28318672867840616`, 1, 1], 
     Hue[0.29810750459327134`, 1, 1], 
     Hue[0.3130282805081365, 1, 1], 
     Hue[0.3279490564230017, 1, 1], 
     Hue[0.34286983233786694`, 1, 1], 
     Hue[0.3577906082527321, 1, 1], 
     Hue[0.37271138416759725`, 1, 1], 
     Hue[0.38763216008246254`, 1, 1], 
     Hue[0.4025529359973277, 1, 1], 
     Hue[0.4174737119121929, 1, 1], 
     Hue[0.432394487827058, 1, 1], 
     Hue[0.4473152637419232, 1, 1], 
     Hue[0.4622360396567884, 1, 1], 
     Hue[0.47715681557165357`, 1, 1], 
     Hue[0.49207759148651875`, 1, 1], 
     Hue[0.5069983674013839, 1, 1], 
     Hue[0.5219191433162491, 1, 1], 
     Hue[0.5368399192311143, 1, 1], 
     Hue[0.5517606951459795, 1, 1], 
     Hue[0.5666814710608448, 1, 1], 
     Hue[0.5816022469757098, 1, 1], 
     Hue[0.5965230228905751, 1, 1], 
     Hue[0.6114437988054402, 1, 1], 
     Hue[0.6263645747203055, 1, 1], 
     Hue[0.6412853506351706, 1, 1], 
     Hue[0.6562061265500359, 1, 1], 
     Hue[0.671126902464901, 1, 1], 
     Hue[0.6860476783797662, 1, 1], 
     Hue[0.7009684542946314, 1, 1], 
     Hue[0.7158892302094966, 1, 1], 
     Hue[0.7308100061243618, 1, 1], 
     Hue[0.7457307820392269, 1, 1], 
     Hue[0.7606515579540921, 1, 1], 
     Hue[0.7755723338689573, 1, 1], 
     Hue[0.7904931097838226, 1, 1], 
     Hue[0.8054138856986877, 1, 1], 
     Hue[0.820334661613553, 1, 1], 
     Hue[0.835255437528418, 1, 1], 
     Hue[0.8501762134432833, 1, 1], 
     Hue[0.8650969893581485, 1, 1], 
     Hue[0.8800177652730137, 1, 1], 
     Hue[0.8949385411878789, 1, 1], 
     Hue[0.909859317102744, 1, 1], 
     Hue[0.9247800930176092, 1, 1], 
     Hue[0.9397008689324744, 1, 1], 
     Hue[0.9546216448473396, 1, 1], 
     Hue[0.9695424207622048, 1, 1], 
     Hue[0.9844631966770699, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.014304748506800346`, 1, 1], 
     Hue[0.02922552442166553, 1, 1], 
     Hue[0.04414630033653072, 1, 1], 
     Hue[0.059067076251395914`, 1, 1], 
     Hue[0.0739878521662611, 1, 1], 
     Hue[0.08890862808112629, 1, 1], 
     Hue[0.10382940399599148`, 1, 1], 
     Hue[0.11875017991085667`, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.8812498200891433, 1, 1], 
     Hue[0.8812498200891433, 1, 1], 
     Hue[0.8961705960040085, 1, 1], 
     Hue[0.8961705960040085, 1, 1], 
     Hue[0.9110913719188737, 1, 1], 
     Hue[0.9110913719188737, 1, 1], 
     Hue[0.9260121478337389, 1, 1], 
     Hue[0.9260121478337389, 1, 1], 
     Hue[0.940932923748604, 1, 1], 
     Hue[0.940932923748604, 1, 1], 
     Hue[0.9558536996634692, 1, 1], 
     Hue[0.9558536996634692, 1, 1], 
     Hue[0.9707744755783345, 1, 1], 
     Hue[0.9707744755783345, 1, 1], 
     Hue[0.9856952514931997, 1, 1], 
     Hue[0.9856952514931997, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.015536803322930032`, 1, 1], 
     Hue[0.015536803322930032`, 1, 1], 
     Hue[0.030457579237795223`, 1, 1], 
     Hue[0.030457579237795223`, 1, 1], 
     Hue[0.0453783551526604, 1, 1], 
     Hue[0.0453783551526604, 1, 1], 
     Hue[0.0602991310675256, 1, 1], 
     Hue[0.0602991310675256, 1, 1], 
     Hue[0.07521990698239078, 1, 1], 
     Hue[0.07521990698239078, 1, 1], 
     Hue[0.09014068289725596, 1, 1], 
     Hue[0.09014068289725596, 1, 1], 
     Hue[0.10506145881212117`, 1, 1], 
     Hue[0.10506145881212117`, 1, 1], 
     Hue[0.11998223472698637`, 1, 1], 
     Hue[0.11998223472698637`, 1, 1], 
     Hue[0.13490301064185153`, 1, 1], 
     Hue[0.13490301064185153`, 1, 1], 
     Hue[0.14982378655671671`, 1, 1], 
     Hue[0.14982378655671671`, 1, 1], 
     Hue[0.16474456247158192`, 1, 1], 
     Hue[0.16474456247158192`, 1, 1], 
     Hue[0.1796653383864471, 1, 1], 
     Hue[0.1796653383864471, 1, 1], 
     Hue[0.19458611430131229`, 1, 1], 
     Hue[0.19458611430131229`, 1, 1], 
     Hue[0.20950689021617747`, 1, 1], 
     Hue[0.20950689021617747`, 1, 1], 
     Hue[0.22442766613104267`, 1, 1], 
     Hue[0.22442766613104267`, 1, 1], 
     Hue[0.23934844204590786`, 1, 1], 
     Hue[0.23934844204590786`, 1, 1], 
     Hue[0.25426921796077306`, 1, 1], 
     Hue[0.25426921796077306`, 1, 1], 
     Hue[0.26918999387563824`, 1, 1], 
     Hue[0.26918999387563824`, 1, 1], 
     Hue[0.2841107697905034, 1, 1], 
     Hue[0.2841107697905034, 1, 1], 
     Hue[0.2990315457053686, 1, 1], 
     Hue[0.2990315457053686, 1, 1], 
     Hue[0.3139523216202338, 1, 1], 
     Hue[0.3139523216202338, 1, 1], 
     Hue[0.32887309753509897`, 1, 1], 
     Hue[0.32887309753509897`, 1, 1], 
     Hue[0.34379387344996415`, 1, 1], 
     Hue[0.34379387344996415`, 1, 1], 
     Hue[0.3587146493648294, 1, 1], 
     Hue[0.3587146493648294, 1, 1], 
     Hue[0.37363542527969457`, 1, 1], 
     Hue[0.37363542527969457`, 1, 1], 
     Hue[0.38855620119455975`, 1, 1], 
     Hue[0.38855620119455975`, 1, 1], 
     Hue[0.40347697710942493`, 1, 1], 
     Hue[0.40347697710942493`, 1, 1], 
     Hue[0.4183977530242901, 1, 1], 
     Hue[0.4183977530242901, 1, 1], 
     Hue[0.4333185289391553, 1, 1], 
     Hue[0.4333185289391553, 1, 1], 
     Hue[0.44823930485402047`, 1, 1], 
     Hue[0.44823930485402047`, 1, 1], 
     Hue[0.4631600807688857, 1, 1], 
     Hue[0.4631600807688857, 1, 1], 
     Hue[0.4780808566837509, 1, 1], 
     Hue[0.4780808566837509, 1, 1], 
     Hue[0.49300163259861607`, 1, 1], 
     Hue[0.49300163259861607`, 1, 1], 
     Hue[0.5079224085134812, 1, 1], 
     Hue[0.5079224085134812, 1, 1], 
     Hue[0.5228431844283464, 1, 1], 
     Hue[0.5228431844283464, 1, 1], 
     Hue[0.5377639603432116, 1, 1], 
     Hue[0.5377639603432116, 1, 1], 
     Hue[0.5526847362580768, 1, 1], 
     Hue[0.5526847362580768, 1, 1], 
     Hue[0.567605512172942, 1, 1], 
     Hue[0.567605512172942, 1, 1], 
     Hue[0.582526288087807, 1, 1], 
     Hue[0.582526288087807, 1, 1], 
     Hue[0.5974470640026723, 1, 1], 
     Hue[0.5974470640026723, 1, 1], 
     Hue[0.6123678399175374, 1, 1], 
     Hue[0.6123678399175374, 1, 1], 
     Hue[0.6272886158324027, 1, 1], 
     Hue[0.6272886158324027, 1, 1], 
     Hue[0.6422093917472679, 1, 1], 
     Hue[0.6422093917472679, 1, 1], 
     Hue[0.6571301676621331, 1, 1], 
     Hue[0.6571301676621331, 1, 1], 
     Hue[0.6720509435769983, 1, 1], 
     Hue[0.6720509435769983, 1, 1], 
     Hue[0.6869717194918634, 1, 1], 
     Hue[0.6869717194918634, 1, 1], 
     Hue[0.7018924954067287, 1, 1], 
     Hue[0.7018924954067287, 1, 1], 
     Hue[0.7168132713215938, 1, 1], 
     Hue[0.7168132713215938, 1, 1], 
     Hue[0.7317340472364591, 1, 1], 
     Hue[0.7317340472364591, 1, 1], 
     Hue[0.7466548231513241, 1, 1], 
     Hue[0.7466548231513241, 1, 1], 
     Hue[0.7615755990661894, 1, 1], 
     Hue[0.7615755990661894, 1, 1], 
     Hue[0.7764963749810546, 1, 1], 
     Hue[0.7764963749810546, 1, 1], 
     Hue[0.7914171508959198, 1, 1], 
     Hue[0.7914171508959198, 1, 1], 
     Hue[0.806337926810785, 1, 1], 
     Hue[0.806337926810785, 1, 1], 
     Hue[0.8212587027256502, 1, 1], 
     Hue[0.8212587027256502, 1, 1], 
     Hue[0.8361794786405153, 1, 1], 
     Hue[0.8361794786405153, 1, 1], 
     Hue[0.8511002545553805, 1, 1], 
     Hue[0.8511002545553805, 1, 1], 
     Hue[0.8660210304702457, 1, 1], 
     Hue[0.8660210304702457, 1, 1], 
     Hue[0.8809418063851109, 1, 1], 
     Hue[0.8809418063851109, 1, 1], 
     Hue[0.8958625822999761, 1, 1], 
     Hue[0.8958625822999761, 1, 1], 
     Hue[0.9107833582148412, 1, 1], 
     Hue[0.9107833582148412, 1, 1], 
     Hue[0.9257041341297065, 1, 1], 
     Hue[0.9257041341297065, 1, 1], 
     Hue[0.9406249100445717, 1, 1], 
     Hue[0.9406249100445717, 1, 1], 
     Hue[0.9555456859594369, 1, 1], 
     Hue[0.9555456859594369, 1, 1], 
     Hue[0.9704664618743021, 1, 1], 
     Hue[0.9704664618743021, 1, 1], 
     Hue[0.9853872377891673, 1, 1], 
     Hue[0.9853872377891673, 1, 1], 
     Hue[0.0003080137040324222, 1, 1], 
     Hue[0.0003080137040324222, 1, 1], 
     Hue[0.01522878961889761, 1, 1], 
     Hue[0.01522878961889761, 1, 1], 
     Hue[0.0301495655337628, 1, 1], 
     Hue[0.0301495655337628, 1, 1], 
     Hue[0.045070341448627996`, 1, 1], 
     Hue[0.045070341448627996`, 1, 1], 
     Hue[0.059991117363493184`, 1, 1], 
     Hue[0.059991117363493184`, 1, 1], 
     Hue[0.07491189327835837, 1, 1], 
     Hue[0.07491189327835837, 1, 1], 
     Hue[0.08983266919322357, 1, 1], 
     Hue[0.08983266919322357, 1, 1], 
     Hue[0.10475344510808872`, 1, 1], 
     Hue[0.10475344510808872`, 1, 1], 
     Hue[0.11967422102295393`, 1, 1], 
     Hue[0.11967422102295393`, 1, 1], 
     Hue[0.13459499693781912`, 1, 1], 
     Hue[0.13459499693781912`, 1, 1], 
     Hue[0.1495157728526843, 1, 1], 
     Hue[0.1495157728526843, 1, 1], 
     Hue[0.16443654876754946`, 1, 1], 
     Hue[0.16443654876754946`, 1, 1], 
     Hue[0.1793573246824147, 1, 1], 
     Hue[0.1793573246824147, 1, 1], 
     Hue[0.19427810059727987`, 1, 1], 
     Hue[0.19427810059727987`, 1, 1], 
     Hue[0.20919887651214505`, 1, 1], 
     Hue[0.20919887651214505`, 1, 1], 
     Hue[0.22411965242701024`, 1, 1], 
     Hue[0.22411965242701024`, 1, 1], 
     Hue[0.23904042834187544`, 1, 1], 
     Hue[0.23904042834187544`, 1, 1], 
     Hue[0.2539612042567406, 1, 1], 
     Hue[0.2539612042567406, 1, 1], 
     Hue[0.2688819801716058, 1, 1], 
     Hue[0.2688819801716058, 1, 1], 
     Hue[0.283802756086471, 1, 1], 
     Hue[0.283802756086471, 1, 1], 
     Hue[0.29872353200133617`, 1, 1], 
     Hue[0.29872353200133617`, 1, 1], 
     Hue[0.3136443079162014, 1, 1], 
     Hue[0.3136443079162014, 1, 1], 
     Hue[0.32856508383106653`, 1, 1], 
     Hue[0.32856508383106653`, 1, 1], 
     Hue[0.34348585974593177`, 1, 1], 
     Hue[0.34348585974593177`, 1, 1], 
     Hue[0.358406635660797, 1, 1], 
     Hue[0.358406635660797, 1, 1], 
     Hue[0.3733274115756622, 1, 1], 
     Hue[0.3733274115756622, 1, 1], 
     Hue[0.38824818749052736`, 1, 1], 
     Hue[0.38824818749052736`, 1, 1], 
     Hue[0.40316896340539254`, 1, 1], 
     Hue[0.40316896340539254`, 1, 1], 
     Hue[0.4180897393202577, 1, 1], 
     Hue[0.4180897393202577, 1, 1], 
     Hue[0.4330105152351229, 1, 1], 
     Hue[0.4330105152351229, 1, 1], 
     Hue[0.44793129114998803`, 1, 1], 
     Hue[0.44793129114998803`, 1, 1], 
     Hue[0.4628520670648533, 1, 1], 
     Hue[0.4628520670648533, 1, 1], 
     Hue[0.4777728429797184, 1, 1], 
     Hue[0.4777728429797184, 1, 1], 
     Hue[0.4926936188945837, 1, 1], 
     Hue[0.4926936188945837, 1, 1], 
     Hue[0.5076143948094487, 1, 1], 
     Hue[0.5076143948094487, 1, 1], 
     Hue[0.522535170724314, 1, 1], 
     Hue[0.522535170724314, 1, 1], 
     Hue[0.5374559466391792, 1, 1], 
     Hue[0.5374559466391792, 1, 1], 
     Hue[0.5523767225540444, 1, 1], 
     Hue[0.5523767225540444, 1, 1], 
     Hue[0.5672974984689095, 1, 1], 
     Hue[0.5672974984689095, 1, 1], 
     Hue[0.5822182743837747, 1, 1], 
     Hue[0.5822182743837747, 1, 1], 
     Hue[0.5971390502986399, 1, 1], 
     Hue[0.5971390502986399, 1, 1], 
     Hue[0.6120598262135051, 1, 1], 
     Hue[0.6120598262135051, 1, 1], 
     Hue[0.6269806021283703, 1, 1], 
     Hue[0.6269806021283703, 1, 1], 
     Hue[0.6419013780432354, 1, 1], 
     Hue[0.6419013780432354, 1, 1], 
     Hue[0.6568221539581007, 1, 1], 
     Hue[0.6568221539581007, 1, 1], 
     Hue[0.6717429298729658, 1, 1], 
     Hue[0.6717429298729658, 1, 1], 
     Hue[0.6866637057878311, 1, 1], 
     Hue[0.6866637057878311, 1, 1], 
     Hue[0.7015844817026962, 1, 1], 
     Hue[0.7015844817026962, 1, 1], 
     Hue[0.7165052576175615, 1, 1], 
     Hue[0.7165052576175615, 1, 1], 
     Hue[0.7314260335324265, 1, 1], 
     Hue[0.7314260335324265, 1, 1], 
     Hue[0.7463468094472918, 1, 1], 
     Hue[0.7463468094472918, 1, 1], 
     Hue[0.761267585362157, 1, 1], 
     Hue[0.761267585362157, 1, 1], 
     Hue[0.7761883612770222, 1, 1], 
     Hue[0.7761883612770222, 1, 1], 
     Hue[0.7911091371918874, 1, 1], 
     Hue[0.7911091371918874, 1, 1], 
     Hue[0.8060299131067525, 1, 1], 
     Hue[0.8060299131067525, 1, 1], 
     Hue[0.8209506890216177, 1, 1], 
     Hue[0.8209506890216177, 1, 1], 
     Hue[0.8358714649364829, 1, 1], 
     Hue[0.8358714649364829, 1, 1], 
     Hue[0.8507922408513481, 1, 1], 
     Hue[0.8507922408513481, 1, 1], 
     Hue[0.8657130167662133, 1, 1], 
     Hue[0.8657130167662133, 1, 1], 
     Hue[0.8806337926810786, 1, 1], 
     Hue[0.8806337926810786, 1, 1], 
     Hue[0.8955545685959436, 1, 1], 
     Hue[0.8955545685959436, 1, 1], 
     Hue[0.9104753445108089, 1, 1], 
     Hue[0.9104753445108089, 1, 1], 
     Hue[0.9253961204256741, 1, 1], 
     Hue[0.9253961204256741, 1, 1], 
     Hue[0.9403168963405393, 1, 1], 
     Hue[0.9403168963405393, 1, 1], 
     Hue[0.9552376722554045, 1, 1], 
     Hue[0.9552376722554045, 1, 1], 
     Hue[0.9701584481702696, 1, 1], 
     Hue[0.9701584481702696, 1, 1], 
     Hue[0.9850792240851348, 1, 1], 
     Hue[0.9850792240851348, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.014920775914865188`, 1, 1], 
     Hue[0.014920775914865188`, 1, 1], 
     Hue[0.029841551829730376`, 1, 1], 
     Hue[0.029841551829730376`, 1, 1], 
     Hue[0.044762327744595556`, 1, 1], 
     Hue[0.044762327744595556`, 1, 1], 
     Hue[0.05968310365946075, 1, 1], 
     Hue[0.05968310365946075, 1, 1], 
     Hue[0.07460387957432593, 1, 1], 
     Hue[0.07460387957432593, 1, 1], 
     Hue[0.08952465548919111, 1, 1], 
     Hue[0.08952465548919111, 1, 1], 
     Hue[0.10444543140405632`, 1, 1], 
     Hue[0.10444543140405632`, 1, 1], 
     Hue[0.11936620731892149`, 1, 1], 
     Hue[0.11936620731892149`, 1, 1], 
     Hue[0.1342869832337867, 1, 1], 
     Hue[0.1342869832337867, 1, 1], 
     Hue[0.1492077591486519, 1, 1], 
     Hue[0.1492077591486519, 1, 1], 
     Hue[0.16412853506351707`, 1, 1], 
     Hue[0.16412853506351707`, 1, 1], 
     Hue[0.17904931097838225`, 1, 1], 
     Hue[0.17904931097838225`, 1, 1], 
     Hue[0.19397008689324746`, 1, 1], 
     Hue[0.19397008689324746`, 1, 1], 
     Hue[0.20889086280811264`, 1, 1], 
     Hue[0.20889086280811264`, 1, 1], 
     Hue[0.22381163872297782`, 1, 1], 
     Hue[0.22381163872297782`, 1, 1], 
     Hue[0.238732414637843, 1, 1], 
     Hue[0.238732414637843, 1, 1], 
     Hue[0.2536531905527082, 1, 1], 
     Hue[0.2536531905527082, 1, 1], 
     Hue[0.2685739664675734, 1, 1], 
     Hue[0.2685739664675734, 1, 1], 
     Hue[0.2834947423824386, 1, 1], 
     Hue[0.2834947423824386, 1, 1], 
     Hue[0.2984155182973038, 1, 1], 
     Hue[0.2984155182973038, 1, 1], 
     Hue[0.31333629421216896`, 1, 1], 
     Hue[0.31333629421216896`, 1, 1], 
     Hue[0.32825707012703415`, 1, 1], 
     Hue[0.32825707012703415`, 1, 1], 
     Hue[0.3431778460418993, 1, 1], 
     Hue[0.3431778460418993, 1, 1], 
     Hue[0.3580986219567645, 1, 1], 
     Hue[0.3580986219567645, 1, 1], 
     Hue[0.37301939787162974`, 1, 1], 
     Hue[0.37301939787162974`, 1, 1], 
     Hue[0.3879401737864949, 1, 1], 
     Hue[0.3879401737864949, 1, 1], 
     Hue[0.4028609497013601, 1, 1], 
     Hue[0.4028609497013601, 1, 1], 
     Hue[0.4177817256162253, 1, 1], 
     Hue[0.4177817256162253, 1, 1], 
     Hue[0.43270250153109047`, 1, 1], 
     Hue[0.43270250153109047`, 1, 1], 
     Hue[0.44762327744595565`, 1, 1], 
     Hue[0.44762327744595565`, 1, 1], 
     Hue[0.46254405336082083`, 1, 1], 
     Hue[0.46254405336082083`, 1, 1], 
     Hue[0.477464829275686, 1, 1], 
     Hue[0.477464829275686, 1, 1], 
     Hue[0.49238560519055125`, 1, 1], 
     Hue[0.49238560519055125`, 1, 1], 
     Hue[0.5073063811054164, 1, 1], 
     Hue[0.5073063811054164, 1, 1], 
     Hue[0.5222271570202817, 1, 1], 
     Hue[0.5222271570202817, 1, 1], 
     Hue[0.5371479329351467, 1, 1], 
     Hue[0.5371479329351467, 1, 1], 
     Hue[0.552068708850012, 1, 1], 
     Hue[0.552068708850012, 1, 1], 
     Hue[0.5669894847648771, 1, 1], 
     Hue[0.5669894847648771, 1, 1], 
     Hue[0.5819102606797423, 1, 1], 
     Hue[0.5819102606797423, 1, 1], 
     Hue[0.5968310365946075, 1, 1], 
     Hue[0.5968310365946075, 1, 1], 
     Hue[0.6117518125094726, 1, 1], 
     Hue[0.6117518125094726, 1, 1], 
     Hue[0.6266725884243378, 1, 1], 
     Hue[0.6266725884243378, 1, 1], 
     Hue[0.641593364339203, 1, 1], 
     Hue[0.641593364339203, 1, 1], 
     Hue[0.6565141402540682, 1, 1], 
     Hue[0.6565141402540682, 1, 1], 
     Hue[0.6714349161689335, 1, 1], 
     Hue[0.6714349161689335, 1, 1], 
     Hue[0.6863556920837985, 1, 1], 
     Hue[0.6863556920837985, 1, 1], 
     Hue[0.7012764679986638, 1, 1], 
     Hue[0.7012764679986638, 1, 1], 
     Hue[0.716197243913529, 1, 1], 
     Hue[0.716197243913529, 1, 1], 
     Hue[0.7311180198283942, 1, 1], 
     Hue[0.7311180198283942, 1, 1], 
     Hue[0.7460387957432594, 1, 1], 
     Hue[0.7460387957432594, 1, 1], 
     Hue[0.7609595716581246, 1, 1], 
     Hue[0.7609595716581246, 1, 1], 
     Hue[0.7758803475729897, 1, 1], 
     Hue[0.7758803475729897, 1, 1], 
     Hue[0.7908011234878549, 1, 1], 
     Hue[0.7908011234878549, 1, 1], 
     Hue[0.8057218994027201, 1, 1], 
     Hue[0.8057218994027201, 1, 1], 
     Hue[0.8206426753175853, 1, 1], 
     Hue[0.8206426753175853, 1, 1], 
     Hue[0.8355634512324506, 1, 1], 
     Hue[0.8355634512324506, 1, 1], 
     Hue[0.8504842271473156, 1, 1], 
     Hue[0.8504842271473156, 1, 1], 
     Hue[0.8654050030621809, 1, 1], 
     Hue[0.8654050030621809, 1, 1], 
     Hue[0.8803257789770461, 1, 1], 
     Hue[0.8803257789770461, 1, 1], 
     Hue[0.8952465548919113, 1, 1], 
     Hue[0.8952465548919113, 1, 1], 
     Hue[0.9101673308067765, 1, 1], 
     Hue[0.9101673308067765, 1, 1], 
     Hue[0.9250881067216417, 1, 1], 
     Hue[0.9250881067216417, 1, 1], 
     Hue[0.9400088826365068, 1, 1], 
     Hue[0.9400088826365068, 1, 1], 
     Hue[0.954929658551372, 1, 1], 
     Hue[0.954929658551372, 1, 1], 
     Hue[0.9698504344662372, 1, 1], 
     Hue[0.9698504344662372, 1, 1], 
     Hue[0.9847712103811024, 1, 1], 
     Hue[0.9847712103811024, 1, 1], 
     Hue[0.9996919862959676, 1, 1], 
     Hue[0.9996919862959676, 1, 1], 
     Hue[0.014612762210832766`, 1, 1], 
     Hue[0.014612762210832766`, 1, 1], 
     Hue[0.029533538125697957`, 1, 1], 
     Hue[0.029533538125697957`, 1, 1], 
     Hue[0.044454314040563145`, 1, 1], 
     Hue[0.044454314040563145`, 1, 1], 
     Hue[0.05937508995542833, 1, 1], 
     Hue[0.05937508995542833, 1, 1], 
     Hue[0.07429586587029352, 1, 1], 
     Hue[0.07429586587029352, 1, 1], 
     Hue[0.08921664178515872, 1, 1], 
     Hue[0.08921664178515872, 1, 1], 
     Hue[0.1041374177000239, 1, 1], 
     Hue[0.1041374177000239, 1, 1], 
     Hue[0.11905819361488908`, 1, 1], 
     Hue[0.11905819361488908`, 1, 1], 
     Hue[0.13397896952975427`, 1, 1], 
     Hue[0.13397896952975427`, 1, 1], 
     Hue[0.14889974544461948`, 1, 1], 
     Hue[0.14889974544461948`, 1, 1], 
     Hue[0.16382052135948466`, 1, 1], 
     Hue[0.16382052135948466`, 1, 1], 
     Hue[0.17874129727434984`, 1, 1], 
     Hue[0.17874129727434984`, 1, 1], 
     Hue[0.19366207318921502`, 1, 1], 
     Hue[0.19366207318921502`, 1, 1], 
     Hue[0.20858284910408023`, 1, 1], 
     Hue[0.20858284910408023`, 1, 1], 
     Hue[0.2235036250189454, 1, 1], 
     Hue[0.2235036250189454, 1, 1], 
     Hue[0.2384244009338106, 1, 1], 
     Hue[0.2384244009338106, 1, 1], 
     Hue[0.2533451768486758, 1, 1], 
     Hue[0.2533451768486758, 1, 1], 
     Hue[0.268265952763541, 1, 1], 
     Hue[0.268265952763541, 1, 1], 
     Hue[0.28318672867840616`, 1, 1], 
     Hue[0.28318672867840616`, 1, 1], 
     Hue[0.29810750459327134`, 1, 1], 
     Hue[0.29810750459327134`, 1, 1], 
     Hue[0.3130282805081365, 1, 1], 
     Hue[0.3130282805081365, 1, 1], 
     Hue[0.3279490564230017, 1, 1], 
     Hue[0.3279490564230017, 1, 1], 
     Hue[0.34286983233786694`, 1, 1], 
     Hue[0.34286983233786694`, 1, 1], 
     Hue[0.3577906082527321, 1, 1], 
     Hue[0.3577906082527321, 1, 1], 
     Hue[0.37271138416759725`, 1, 1], 
     Hue[0.37271138416759725`, 1, 1], 
     Hue[0.38763216008246254`, 1, 1], 
     Hue[0.38763216008246254`, 1, 1], 
     Hue[0.4025529359973277, 1, 1], 
     Hue[0.4025529359973277, 1, 1], 
     Hue[0.4174737119121929, 1, 1], 
     Hue[0.4174737119121929, 1, 1], 
     Hue[0.432394487827058, 1, 1], 
     Hue[0.432394487827058, 1, 1], 
     Hue[0.4473152637419232, 1, 1], 
     Hue[0.4473152637419232, 1, 1], 
     Hue[0.4622360396567884, 1, 1], 
     Hue[0.4622360396567884, 1, 1], 
     Hue[0.47715681557165357`, 1, 1], 
     Hue[0.47715681557165357`, 1, 1], 
     Hue[0.49207759148651875`, 1, 1], 
     Hue[0.49207759148651875`, 1, 1], 
     Hue[0.5069983674013839, 1, 1], 
     Hue[0.5069983674013839, 1, 1], 
     Hue[0.5219191433162491, 1, 1], 
     Hue[0.5219191433162491, 1, 1], 
     Hue[0.5368399192311143, 1, 1], 
     Hue[0.5368399192311143, 1, 1], 
     Hue[0.5517606951459795, 1, 1], 
     Hue[0.5517606951459795, 1, 1], 
     Hue[0.5666814710608448, 1, 1], 
     Hue[0.5666814710608448, 1, 1], 
     Hue[0.5816022469757098, 1, 1], 
     Hue[0.5816022469757098, 1, 1], 
     Hue[0.5965230228905751, 1, 1], 
     Hue[0.5965230228905751, 1, 1], 
     Hue[0.6114437988054402, 1, 1], 
     Hue[0.6114437988054402, 1, 1], 
     Hue[0.6263645747203055, 1, 1], 
     Hue[0.6263645747203055, 1, 1], 
     Hue[0.6412853506351706, 1, 1], 
     Hue[0.6412853506351706, 1, 1], 
     Hue[0.6562061265500359, 1, 1], 
     Hue[0.6562061265500359, 1, 1], 
     Hue[0.671126902464901, 1, 1], 
     Hue[0.671126902464901, 1, 1], 
     Hue[0.6860476783797662, 1, 1], 
     Hue[0.6860476783797662, 1, 1], 
     Hue[0.7009684542946314, 1, 1], 
     Hue[0.7009684542946314, 1, 1], 
     Hue[0.7158892302094966, 1, 1], 
     Hue[0.7158892302094966, 1, 1], 
     Hue[0.7308100061243618, 1, 1], 
     Hue[0.7308100061243618, 1, 1], 
     Hue[0.7457307820392269, 1, 1], 
     Hue[0.7457307820392269, 1, 1], 
     Hue[0.7606515579540921, 1, 1], 
     Hue[0.7606515579540921, 1, 1], 
     Hue[0.7755723338689573, 1, 1], 
     Hue[0.7755723338689573, 1, 1], 
     Hue[0.7904931097838226, 1, 1], 
     Hue[0.7904931097838226, 1, 1], 
     Hue[0.8054138856986877, 1, 1], 
     Hue[0.8054138856986877, 1, 1], 
     Hue[0.820334661613553, 1, 1], 
     Hue[0.820334661613553, 1, 1], 
     Hue[0.835255437528418, 1, 1], 
     Hue[0.835255437528418, 1, 1], 
     Hue[0.8501762134432833, 1, 1], 
     Hue[0.8501762134432833, 1, 1], 
     Hue[0.8650969893581485, 1, 1], 
     Hue[0.8650969893581485, 1, 1], 
     Hue[0.8800177652730137, 1, 1], 
     Hue[0.8800177652730137, 1, 1], 
     Hue[0.8949385411878789, 1, 1], 
     Hue[0.8949385411878789, 1, 1], 
     Hue[0.909859317102744, 1, 1], 
     Hue[0.909859317102744, 1, 1], 
     Hue[0.9247800930176092, 1, 1], 
     Hue[0.9247800930176092, 1, 1], 
     Hue[0.9397008689324744, 1, 1], 
     Hue[0.9397008689324744, 1, 1], 
     Hue[0.9546216448473396, 1, 1], 
     Hue[0.9546216448473396, 1, 1], 
     Hue[0.9695424207622048, 1, 1], 
     Hue[0.9695424207622048, 1, 1], 
     Hue[0.9844631966770699, 1, 1], 
     Hue[0.9844631966770699, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.014304748506800346`, 1, 1], 
     Hue[0.014304748506800346`, 1, 1], 
     Hue[0.02922552442166553, 1, 1], 
     Hue[0.02922552442166553, 1, 1], 
     Hue[0.04414630033653072, 1, 1], 
     Hue[0.04414630033653072, 1, 1], 
     Hue[0.059067076251395914`, 1, 1], 
     Hue[0.059067076251395914`, 1, 1], 
     Hue[0.0739878521662611, 1, 1], 
     Hue[0.0739878521662611, 1, 1], 
     Hue[0.08890862808112629, 1, 1], 
     Hue[0.08890862808112629, 1, 1], 
     Hue[0.10382940399599148`, 1, 1], 
     Hue[0.10382940399599148`, 1, 1], 
     Hue[0.11875017991085667`, 1, 1], 
     Hue[0.11875017991085667`, 1, 1]}],
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  Epilog->LineBox[{{-9.984375, 0.}, {-9.9375, 0.}, {-9.890625, 0.}, {-9.84375,
      0.}, {-9.796875, 0.}, {-9.75, 0.}, {-9.703125, 0.}, {-9.65625, 
     0.}, {-9.609375, 0.}, {-9.5625, 0.}, {-9.515625, 0.}, {-9.46875, 
     0.}, {-9.421875, 0.}, {-9.375, 0.}, {-9.328125, 0.}, {-9.28125, 
     0.}, {-9.234375, 0.}, {-9.1875, 0.}, {-9.140625, 0.}, {-9.09375, 
     0.}, {-9.046875, 0.}, {-9., 0.}, {-8.953125, 0.}, {-8.90625, 
     0.}, {-8.859375, 0.}, {-8.8125, 0.}, {-8.765625, 0.}, {-8.71875, 
     0.}, {-8.671875, 0.}, {-8.625, 0.}, {-8.578125, 0.}, {-8.53125, 
     0.}, {-8.484375, 0.}, {-8.4375, 0.}, {-8.390625, 0.}, {-8.34375, 
     0.}, {-8.296875, 0.}, {-8.25, 0.}, {-8.203125, 0.}, {-8.15625, 
     0.}, {-8.109375, 0.}, {-8.0625, 0.}, {-8.015625, 0.}, {-7.96875, 
     0.}, {-7.921875, 0.}, {-7.875, 0.}, {-7.828125, 0.}, {-7.78125, 
     0.}, {-7.734375, 0.}, {-7.6875, 0.}, {-7.640625, 0.}, {-7.59375, 
     0.}, {-7.546875, 0.}, {-7.5, 0.}, {-7.453125, 0.}, {-7.40625, 
     0.}, {-7.359375, 0.}, {-7.3125, 0.}, {-7.265625, 0.}, {-7.21875, 
     0.}, {-7.171875, 0.}, {-7.125, 0.}, {-7.078125, 0.}, {-7.03125, 
     0.}, {-6.984375, 0.}, {-6.9375, 0.}, {-6.890625, 0.}, {-6.84375, 
     0.}, {-6.796875, 0.}, {-6.75, 0.}, {-6.703125, 0.}, {-6.65625, 
     1.798290384585907*^-10}, {-6.609375, 2.454074030632512*^-10}, {-6.5625, 
     3.3416521269515235`*^-10}, {-6.515625, 
     4.540258224893414*^-10}, {-6.46875, 6.155248775304267*^-10}, {-6.421875, 
     8.326383436888574*^-10}, {-6.375, 1.1238618846365146`*^-9}, {-6.328125, 
     1.5136142859631868`*^-9}, {-6.28125, 
     2.0340422990181602`*^-9}, {-6.234375, 2.727451195769395*^-9}, {-6.1875, 
     3.6491900559248614`*^-9}, {-6.140625, 4.871713241369462*^-9}, {-6.09375, 
     6.4895217274559454`*^-9}, {-6.046875, 8.625601785122393*^-9}, {-6., 
     1.143962682793732*^-8}, {-5.953125, 1.5138402582064123`*^-8}, {-5.90625, 
     1.9989134166329445`*^-8}, {-5.859375, 2.6336232960777236`*^-8}, {-5.8125,
      3.4622551267281476`*^-8}, {-5.765625, 4.541613984984655*^-8}, {-5.71875,
      5.9443874652812666`*^-8}, {-5.671875, 7.763360267872042*^-8}, {-5.625, 
     1.0116682283297703`*^-7}, {-5.578125, 
     1.3154435665042106`*^-7}, {-5.53125, 
     1.7066798897368368`*^-7}, {-5.484375, 2.2094168439275953`*^-7}, {-5.4375,
      2.853967280337101*^-7}, {-5.390625, 3.6784601760141807`*^-7}, {-5.34375,
      4.730737680877867*^-7}, {-5.296875, 6.070681041288068*^-7}, {-5.25, 
     7.773054327613681*^-7}, {-5.203125, 9.930971386469107*^-7}, {-5.15625, 
     1.2660110541651635`*^-6}, {-5.109375, 1.6103823595269838`*^-6}, {-5.0625,
      2.0439310962054014`*^-6}, {-5.015625, 
     2.5885063646882575`*^-6}, {-4.96875, 
     3.2709805595093306`*^-6}, {-4.921875, 4.124320705162408*^-6}, {-4.875, 
     5.188868129033276*^-6}, {-4.828125, 6.513862372589593*^-6}, {-4.78125, 
     8.159250426441251*^-6}, {-4.734375, 0.0000101978280994376}, {-4.6875, 
     0.00001271776660975753}, {-4.640625, 0.00001582558431978149}, {-4.59375, 
     0.000019649630916271522`}, {-4.546875, 0.00002434415923693149}, {-4.5, 
     0.000030094068318430215`}, {-4.453125, 
     0.00003712041002129977}, {-4.40625, 0.00004568676067906624}, {-4.359375, 
     0.00005610656849737733}, {-4.3125, 0.00006875159673421496}, {-4.265625, 
     0.00008406159182671469}, {-4.21875, 0.00010255531435391963`}, {-4.171875,
      0.00012484307875301263`}, {-4.125, 0.00015164095470627406`}, {-4.078125,
      0.00018378678870435632`}, {-4.03125, 
     0.00022225820803473448`}, {-3.984375, 0.00026819277085811307`}, {-3.9375,
      0.00032291042458717297`}, {-3.890625, 
     0.00038793842989333545`}, {-3.84375, 0.0004650388987209338}, {-3.796875, 
     0.0005562390810359108}, {-3.75, 0.0006638645160090288}, {-3.703125, 
     0.0007905751382558167}, {-3.65625, 0.0009394043979607872}, {-3.609375, 
     0.001113801414564607}, {-3.5625, 0.001317676136604145}, {-3.515625, 
     0.0015554474247579002`}, {-3.46875, 0.0018320939107596115`}, {-3.421875, 
     0.0021532074113328}, {-3.375, 0.0025250485935685565`}, {-3.328125, 
     0.002954604496319363}, {-3.28125, 0.003449647411549908}, {-3.234375, 
     0.004018794520776898}, {-3.1875, 0.004671567565649339}, {-3.140625, 
     0.0054184517096023945`}, {-3.09375, 0.006270952620948514}, {-3.046875, 
     0.007241650678706022}, {-3., 0.008344251073246413}, {-2.953125, 
     0.009593628447188283}, {-2.90625, 0.01100586460097308}, {-2.859375, 
     0.012598277675673225`}, {-2.8125, 0.014389441126572085`}, {-2.765625, 
     0.016399190718953217`}, {-2.71875, 0.018648617716584236`}, {-2.671875, 
     0.021160046397946577`}, {-2.625, 0.023956994029747086`}, {-2.578125, 
     0.0270641114559759}, {-2.53125, 0.03050710252787448}, {-2.484375, 
     0.03431262070944408}, {-2.4375, 0.03850814134788083}, {-2.390625, 
     0.043121808301270076`}, {-2.34375, 0.04818225386893848}, {-2.296875, 
     0.05371839127405522}, {-2.25, 0.05975917930334211}, {-2.203125, 
     0.06633335911384833}, {-2.15625, 0.07346916366912468}, {-2.109375, 
     0.0811940007628494}, {-2.0625, 0.08953411112162063}, {-2.015625, 
     0.09851420364336069}, {-1.96875, 0.10815707041522599`}, {-1.921875, 
     0.1184831847553152}, {-1.875, 0.12951028612472462`}, {-1.828125, 
     0.14125295634834165`}, {-1.78125, 0.1537221921509343}, {-1.734375, 
     0.1669249795455717}, {-1.6875, 0.1808638760897144}, {-1.640625, 
     0.19553660743580398`}, {-1.59375, 0.21093568493342887`}, {-1.546875, 
     0.2270480512752962}, {-1.5, 0.24385476130642741`}, {-1.453125, 
     0.2613307051237464}, {-1.40625, 0.2794443804718406}, {-1.359375, 
     0.2981577211826469}, {-1.3125, 0.31742598800714367`}, {-1.265625, 
     0.33719772764369327`}, {-1.21875, 0.3574148050814272}, {-1.171875, 
     0.3780125135522741}, {-1.125, 0.3989197654295719}, {-1.078125, 
     0.4200593663358202}, {-1.03125, 0.44134837354154155`}, {-0.984375, 
     0.46269853846919623`}, {-0.9375, 0.48401683178139476`}, {-0.890625, 
     0.5052060481546798}, {-0.84375, 0.5261654864445083}, {-0.796875, 
     0.5467916995610779}, {-0.75, 0.5669793070276937}, {-0.703125, 
     0.5866218619123392}, {-0.65625, 0.6056127626376299}, {-0.609375, 
     0.6238461991120364}, {-0.5625, 0.6412181217122251}, {-0.515625, 
     0.6576272209061722}, {-0.46875, 0.6729759047600129}, {-0.421875, 
     0.6871712612353379}, {-0.375, 0.70012599207068}, {-0.328125, 
     0.711759305159426}, {-0.28125, 0.7219977526896632}, {-0.234375, 
     0.7307760028976727}, {-0.1875, 0.7380375340988992}, {-0.140625, 
     0.7437352406861114}, {-0.09375, 0.7478319420070575}, {-0.046875, 
     0.7503007864315618}, {0, 0.7511255444649425}, {0.046875, 
     0.7503007864315618}, {0.09375, 0.7478319420070575}, {0.140625, 
     0.7437352406861114}, {0.1875, 0.7380375340988992}, {0.234375, 
     0.7307760028976727}, {0.28125, 0.7219977526896632}, {0.328125, 
     0.711759305159426}, {0.375, 0.70012599207068}, {0.421875, 
     0.6871712612353379}, {0.46875, 0.6729759047600129}, {0.515625, 
     0.6576272209061722}, {0.5625, 0.6412181217122251}, {0.609375, 
     0.6238461991120364}, {0.65625, 0.6056127626376299}, {0.703125, 
     0.5866218619123392}, {0.75, 0.5669793070276937}, {0.796875, 
     0.5467916995610779}, {0.84375, 0.5261654864445083}, {0.890625, 
     0.5052060481546798}, {0.9375, 0.48401683178139476`}, {0.984375, 
     0.46269853846919623`}, {1.03125, 0.44134837354154155`}, {1.078125, 
     0.4200593663358202}, {1.125, 0.3989197654295719}, {1.171875, 
     0.3780125135522741}, {1.21875, 0.3574148050814272}, {1.265625, 
     0.33719772764369327`}, {1.3125, 0.31742598800714367`}, {1.359375, 
     0.2981577211826469}, {1.40625, 0.2794443804718406}, {1.453125, 
     0.2613307051237464}, {1.5, 0.24385476130642741`}, {1.546875, 
     0.2270480512752962}, {1.59375, 0.21093568493342887`}, {1.640625, 
     0.19553660743580398`}, {1.6875, 0.1808638760897144}, {1.734375, 
     0.1669249795455717}, {1.78125, 0.1537221921509343}, {1.828125, 
     0.14125295634834165`}, {1.875, 0.12951028612472462`}, {1.921875, 
     0.1184831847553152}, {1.96875, 0.10815707041522599`}, {2.015625, 
     0.09851420364336069}, {2.0625, 0.08953411112162063}, {2.109375, 
     0.0811940007628494}, {2.15625, 0.07346916366912468}, {2.203125, 
     0.06633335911384833}, {2.25, 0.05975917930334211}, {2.296875, 
     0.05371839127405522}, {2.34375, 0.04818225386893848}, {2.390625, 
     0.043121808301270076`}, {2.4375, 0.03850814134788083}, {2.484375, 
     0.03431262070944408}, {2.53125, 0.03050710252787448}, {2.578125, 
     0.0270641114559759}, {2.625, 0.023956994029747086`}, {2.671875, 
     0.021160046397946577`}, {2.71875, 0.018648617716584236`}, {2.765625, 
     0.016399190718953217`}, {2.8125, 0.014389441126572085`}, {2.859375, 
     0.012598277675673225`}, {2.90625, 0.01100586460097308}, {2.953125, 
     0.009593628447188283}, {3., 0.008344251073246413}, {3.046875, 
     0.007241650678706022}, {3.09375, 0.006270952620948514}, {3.140625, 
     0.0054184517096023945`}, {3.1875, 0.004671567565649339}, {3.234375, 
     0.004018794520776898}, {3.28125, 0.003449647411549908}, {3.328125, 
     0.002954604496319363}, {3.375, 0.0025250485935685565`}, {3.421875, 
     0.0021532074113328}, {3.46875, 0.0018320939107596115`}, {3.515625, 
     0.0015554474247579002`}, {3.5625, 0.001317676136604145}, {3.609375, 
     0.001113801414564607}, {3.65625, 0.0009394043979607872}, {3.703125, 
     0.0007905751382558167}, {3.75, 0.0006638645160090288}, {3.796875, 
     0.0005562390810359108}, {3.84375, 0.0004650388987209338}, {3.890625, 
     0.00038793842989333545`}, {3.9375, 0.00032291042458717297`}, {3.984375, 
     0.00026819277085811307`}, {4.03125, 0.00022225820803473448`}, {4.078125, 
     0.00018378678870435632`}, {4.125, 0.00015164095470627406`}, {4.171875, 
     0.00012484307875301263`}, {4.21875, 0.00010255531435391963`}, {4.265625, 
     0.00008406159182671469}, {4.3125, 0.00006875159673421496}, {4.359375, 
     0.00005610656849737733}, {4.40625, 0.00004568676067906624}, {4.453125, 
     0.00003712041002129977}, {4.5, 0.000030094068318430215`}, {4.546875, 
     0.00002434415923693149}, {4.59375, 0.000019649630916271522`}, {4.640625, 
     0.00001582558431978149}, {4.6875, 0.00001271776660975753}, {4.734375, 
     0.0000101978280994376}, {4.78125, 8.159250426441251*^-6}, {4.828125, 
     6.513862372589593*^-6}, {4.875, 5.188868129033276*^-6}, {4.921875, 
     4.124320705162408*^-6}, {4.96875, 3.2709805595093306`*^-6}, {5.015625, 
     2.5885063646882575`*^-6}, {5.0625, 2.0439310962054014`*^-6}, {5.109375, 
     1.6103823595269838`*^-6}, {5.15625, 1.2660110541651635`*^-6}, {5.203125, 
     9.930971386469107*^-7}, {5.25, 7.773054327613681*^-7}, {5.296875, 
     6.070681041288068*^-7}, {5.34375, 4.730737680877867*^-7}, {5.390625, 
     3.6784601760141807`*^-7}, {5.4375, 2.853967280337101*^-7}, {5.484375, 
     2.2094168439275953`*^-7}, {5.53125, 1.7066798897368368`*^-7}, {5.578125, 
     1.3154435665042106`*^-7}, {5.625, 1.0116682283297703`*^-7}, {5.671875, 
     7.763360267872042*^-8}, {5.71875, 5.9443874652812666`*^-8}, {5.765625, 
     4.541613984984655*^-8}, {5.8125, 3.4622551267281476`*^-8}, {5.859375, 
     2.6336232960777236`*^-8}, {5.90625, 1.9989134166329445`*^-8}, {5.953125, 
     1.5138402582064123`*^-8}, {6., 1.143962682793732*^-8}, {6.046875, 
     8.625601785122393*^-9}, {6.09375, 6.4895217274559454`*^-9}, {6.140625, 
     4.871713241369462*^-9}, {6.1875, 3.6491900559248614`*^-9}, {6.234375, 
     2.727451195769395*^-9}, {6.28125, 2.0340422990181602`*^-9}, {6.328125, 
     1.5136142859631868`*^-9}, {6.375, 1.1238618846365146`*^-9}, {6.421875, 
     8.326383436888574*^-10}, {6.46875, 6.155248775304267*^-10}, {6.515625, 
     4.540258224893414*^-10}, {6.5625, 3.3416521269515235`*^-10}, {6.609375, 
     2.454074030632512*^-10}, {6.65625, 1.798290384585907*^-10}, {6.703125, 
     0.}, {6.75, 0.}, {6.796875, 0.}, {6.84375, 0.}, {6.890625, 0.}, {6.9375, 
     0.}, {6.984375, 0.}, {7.03125, 0.}, {7.078125, 0.}, {7.125, 0.}, {
     7.171875, 0.}, {7.21875, 0.}, {7.265625, 0.}, {7.3125, 0.}, {7.359375, 
     0.}, {7.40625, 0.}, {7.453125, 0.}, {7.5, 0.}, {7.546875, 0.}, {7.59375, 
     0.}, {7.640625, 0.}, {7.6875, 0.}, {7.734375, 0.}, {7.78125, 0.}, {
     7.828125, 0.}, {7.875, 0.}, {7.921875, 0.}, {7.96875, 0.}, {8.015625, 
     0.}, {8.0625, 0.}, {8.109375, 0.}, {8.15625, 0.}, {8.203125, 0.}, {8.25, 
     0.}, {8.296875, 0.}, {8.34375, 0.}, {8.390625, 0.}, {8.4375, 0.}, {
     8.484375, 0.}, {8.53125, 0.}, {8.578125, 0.}, {8.625, 0.}, {8.671875, 
     0.}, {8.71875, 0.}, {8.765625, 0.}, {8.8125, 0.}, {8.859375, 0.}, {
     8.90625, 0.}, {8.953125, 0.}, {9., 0.}, {9.046875, 0.}, {9.09375, 0.}, {
     9.140625, 0.}, {9.1875, 0.}, {9.234375, 0.}, {9.28125, 0.}, {9.328125, 
     0.}, {9.375, 0.}, {9.421875, 0.}, {9.46875, 0.}, {9.515625, 0.}, {9.5625,
      0.}, {9.609375, 0.}, {9.65625, 0.}, {9.703125, 0.}, {9.75, 0.}, {
     9.796875, 0.}, {9.84375, 0.}, {9.890625, 0.}, {9.9375, 0.}, {9.984375, 
     0.}, {10.03125, 0.}, {10.078125, 0.}, {10.125, 0.}, {10.171875, 0.}, {
     10.21875, 0.}, {10.265625, 0.}, {10.3125, 0.}, {10.359375, 0.}, {
     10.40625, 0.}, {10.453125, 0.}, {10.5, 0.}, {10.546875, 0.}, {10.59375, 
     0.}, {10.640625, 0.}, {10.6875, 0.}, {10.734375, 0.}, {10.78125, 0.}, {
     10.828125, 0.}, {10.875, 0.}, {10.921875, 0.}, {10.96875, 0.}, {
     11.015625, 0.}, {11.0625, 0.}, {11.109375, 0.}, {11.15625, 0.}, {
     11.203125, 0.}, {11.25, 0.}, {11.296875, 0.}, {11.34375, 0.}, {11.390625,
      0.}, {11.4375, 0.}, {11.484375, 0.}, {11.53125, 0.}, {11.578125, 0.}, {
     11.625, 0.}, {11.671875, 0.}, {11.71875, 0.}, {11.765625, 0.}, {11.8125, 
     0.}, {11.859375, 0.}, {11.90625, 0.}, {11.953125, 0.}, {12., 0.}, {
     12.046875, 0.}, {12.09375, 0.}, {12.140625, 0.}, {12.1875, 0.}, {
     12.234375, 0.}, {12.28125, 0.}, {12.328125, 0.}, {12.375, 0.}, {
     12.421875, 0.}, {12.46875, 0.}, {12.515625, 0.}, {12.5625, 0.}, {
     12.609375, 0.}, {12.65625, 0.}, {12.703125, 0.}, {12.75, 0.}, {12.796875,
      0.}, {12.84375, 0.}, {12.890625, 0.}, {12.9375, 0.}, {12.984375, 0.}, {
     13.03125, 0.}, {13.078125, 0.}, {13.125, 0.}, {13.171875, 0.}, {13.21875,
      0.}, {13.265625, 0.}, {13.3125, 0.}, {13.359375, 0.}, {13.40625, 0.}, {
     13.453125, 0.}, {13.5, 0.}, {13.546875, 0.}, {13.59375, 0.}, {13.640625, 
     0.}, {13.6875, 0.}, {13.734375, 0.}, {13.78125, 0.}, {13.828125, 0.}, {
     13.875, 0.}, {13.921875, 0.}, {13.96875, 0.}}],
  PlotRange->All,
  PlotRangeClipping->True,
  PlotRangePadding->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{
  3.392726195356563*^9, {3.392726519472619*^9, 3.391849849078125*^9}, 
   3.39185004884375*^9, 3.392472367823471*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
QListArgColorPlot[fvals,PlotRange->All, \
QHorizontalRange->{QSpaceInterval[a,b,n],{-5,5}}]\
\>", "Input"],

Cell[BoxData[
 GraphicsBox[GraphicsComplexBox[CompressedData["
1:eJyF2Ht0zVcWB/CT5Ob9uvemHqOzpj8GU8xYHqGY4bdH17Q1LU0tOtTQy9Aq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   "], {
    {GrayLevel[0.5], EdgeForm[None], 
     GraphicsGroupBox[PolygonBox[CompressedData["
1:eJwt13f4TmUcx/HvfeyVbJGZMkN2UoTsEJIQ9aOSnVIqZRSV9t5ZlZURaVMS
7amhrCKjQjuaen+uz/PH67rePefc4znPfVCNvHF9xmYR8SnyI+FzfIEs99lW
bEMBFMP3+AHFc/bjAEogH77CltxYzbEZX+auaY7t2IGCKIp9+C43dyHsxC4U
zt3zNb7JXdMaB/EjjkIR7MHe3Fz67Cf8jJKooD3wpfKhIn00fsPvKJXzBw6h
NMrgL/yNsrk5fsGvubG65zD+zN2re/7BvygXXiM/axXAMbnP/sMRlEdTtOba
KWiW+0wPPSXvVd/5W+zOfTfNUZBrhVAprDBdBJXpqtoXXQrVwkrTZVCdPgHV
6RqoTVfRM6JLJo/VPWXpcronrLz2gpp0LRxLV8HxYVW1VvLcuqeiviuO0706
E3SJ5LW0Zk1dQx26rvakvaAe3RjN6OY4iW4g9IkSHlNLayeP1XcuShdLXqs+
6tH1k8fqmbahT9Wc4TXqaGzyvQ21ptZCo/CaLeiWaBLWij45+bfSPU3UyXvV
nhpqbPJc+s6V9FskPyuteRrdFi3oAbiBnoFzNTc60mdojbBOdGe0ptuhH302
Tqdboj3dIXms7ulCd9UZCutGd0cbrY3e9FnaQ1gfum/y3LqnB32mnlF4jf70
OWgf/qwn3St5Ln02QHvXHujuGEmPQg99Fwymz9N3ChtCD0WnsPPpC/Qdw/Lo
YehCd8VF9MXoRvfCBPoy9A6vMZoeoz2HjaXHoWd4zAj6kuS9aY8D6UHJe9M9
4+lLk+fWnJfTE/WM6E24mj+krsn8Z2Bf/TfXrkG/sMn0tfpN6EGYRd+CwXq2
mERflTxW91xHT0H/sKn0ND3j8BmYSd+IgWE30Tcnz617ptPXJ58drXErfZue
cdjt9B0YEnYnfZeeuZ4l5tBz9Uz17HEvfZ9+g7D76QeQRw/Dw/QjGB6e4276
nuSxOsPtdBaTz6LGPEg/lDxWz/AK+srkZ6E159HzMSLsCfpJ/Ub0hXicnp28
V635KP1Y8jXd8xS9ACPDFtKLdObCFtNLMFpnASvoZ3QmwlbSqzBee8Nr9Drt
kR6DZfTy5LG651l6tc5I2HP085igs4JX6DU6M2Fr6VeT59Y9L9Av6syG13id
Xo9J4c9eol9OnkufvUFv0Jmhp+Fz+gtM19nDW/TbOoNh79DvYnLYe/T7OpNh
H9Af4jp6Cj6hN2GqnjWO5myXwuzwGpu59qXOWNhX9BbcEB7zKf1Z8t60x430
m8l70z1b6W2YEbad3oGZYV/T3+iM62zjW3q3znjYHnovZoXto7/TOxX2Pf0D
bg3bTx/QOxB2kP4Rt4f9RP+sdyLsF/pX3Bn2G/273pGwP+hDuDu8x530ruS9
6rPD9J96B/Tu4B/6X71Denegf9QkPBCW0fnwYHjMX9z7d/JYjfmPPpI8Vr/R
R/THyb+VxuRnbAE8pHcRhekieCSsKF0Mj4YVp0vgsbCj6JKZf2vNUZAulHku
vSNPs9bS5LOvM1Caa2Uwh56L8nQFzKPnoxJdGU+EHUtXwZP0EtSj6+Pp8JiK
9DGZx+qeqnQ1PBVWna6BBfQinEDXxuKwOnTdzHPrnpr0cVgYXqMBfSKWhjWk
G2FZWGP6JCzXu4t29Ol4nl6BZnRzPBPWgm6JlfQqtKZPwbPhOZrQTTOP1R5q
0cdn3rvGtKJPzjxWz7AsXS7zs9Sa7ekOeCGsI30GXqRX4zS6bea9as029KmZ
r+nvxOH6szD570aN6cS1zniJfhnd6O54hV6DnnQvrA3rTZ+FV8P60H3xWlg/
+mysC+tPn4PXwwbQ52J92EB6EN4IG0yfhw1hQ+ih2Bh2Pn0B3gzvsQd9Zua9
6rM8ehjeot/GRfTFeId+FyPpUXgvbDQ9Bu+HxwynL8w8VmNG0JdkHqtn1IXu
mvlZaYz+B2ccPqA/xAT6MnwUdjk9ER+HXUFfiU/CJtFXZf63guYYT1+aeS79
m2EyfS0+o/8Hdd96jA==
       "]]]}, {{}, {}, {
      {Hue[0.4183977530242901, 1, 1], LineBox[{1, 2}]}, 
      {Hue[0.4333185289391553, 1, 1], LineBox[{2, 3}]}, 
      {Hue[0.44823930485402047`, 1, 1], LineBox[{3, 4}]}, 
      {Hue[0.4631600807688857, 1, 1], LineBox[{4, 5}]}, 
      {Hue[0.4780808566837509, 1, 1], LineBox[{5, 6}]}, 
      {Hue[0.49300163259861607`, 1, 1], LineBox[{6, 7}]}, 
      {Hue[0.5079224085134812, 1, 1], LineBox[{7, 8}]}, 
      {Hue[0.5228431844283464, 1, 1], LineBox[{8, 9}]}, 
      {Hue[0.5377639603432116, 1, 1], LineBox[{9, 10}]}, 
      {Hue[0.5526847362580768, 1, 1], LineBox[{10, 11}]}, 
      {Hue[0.567605512172942, 1, 1], LineBox[{11, 12}]}, 
      {Hue[0.582526288087807, 1, 1], LineBox[{12, 13}]}, 
      {Hue[0.5974470640026723, 1, 1], LineBox[{13, 14}]}, 
      {Hue[0.6123678399175374, 1, 1], LineBox[{14, 15}]}, 
      {Hue[0.6272886158324027, 1, 1], LineBox[{15, 16}]}, 
      {Hue[0.6422093917472679, 1, 1], LineBox[{16, 17}]}, 
      {Hue[0.6571301676621331, 1, 1], LineBox[{17, 18}]}, 
      {Hue[0.6720509435769983, 1, 1], LineBox[{18, 19}]}, 
      {Hue[0.6869717194918634, 1, 1], LineBox[{19, 20}]}, 
      {Hue[0.7018924954067287, 1, 1], LineBox[{20, 21}]}, 
      {Hue[0.7168132713215938, 1, 1], LineBox[{21, 22}]}, 
      {Hue[0.7317340472364591, 1, 1], LineBox[{22, 23}]}, 
      {Hue[0.7466548231513241, 1, 1], LineBox[{23, 24}]}, 
      {Hue[0.7615755990661894, 1, 1], LineBox[{24, 25}]}, 
      {Hue[0.7764963749810546, 1, 1], LineBox[{25, 26}]}, 
      {Hue[0.7914171508959198, 1, 1], LineBox[{26, 27}]}, 
      {Hue[0.806337926810785, 1, 1], LineBox[{27, 28}]}, 
      {Hue[0.8212587027256502, 1, 1], LineBox[{28, 29}]}, 
      {Hue[0.8361794786405153, 1, 1], LineBox[{29, 30}]}, 
      {Hue[0.8511002545553805, 1, 1], LineBox[{30, 31}]}, 
      {Hue[0.8660210304702457, 1, 1], LineBox[{31, 32}]}, 
      {Hue[0.8809418063851109, 1, 1], LineBox[{32, 33}]}, 
      {Hue[0.8958625822999761, 1, 1], LineBox[{33, 34}]}, 
      {Hue[0.9107833582148412, 1, 1], LineBox[{34, 35}]}, 
      {Hue[0.9257041341297065, 1, 1], LineBox[{35, 36}]}, 
      {Hue[0.9406249100445717, 1, 1], LineBox[{36, 37}]}, 
      {Hue[0.9555456859594369, 1, 1], LineBox[{37, 38}]}, 
      {Hue[0.9704664618743021, 1, 1], LineBox[{38, 39}]}, 
      {Hue[0.9853872377891673, 1, 1], LineBox[{39, 40}]}, 
      {Hue[0.0003080137040324222, 1, 1], LineBox[{40, 41}]}, 
      {Hue[0.01522878961889761, 1, 1], LineBox[{41, 42}]}, 
      {Hue[0.0301495655337628, 1, 1], LineBox[{42, 43}]}, 
      {Hue[0.045070341448627996`, 1, 1], LineBox[{43, 44}]}, 
      {Hue[0.059991117363493184`, 1, 1], LineBox[{44, 45}]}, 
      {Hue[0.07491189327835837, 1, 1], LineBox[{45, 46}]}, 
      {Hue[0.08983266919322357, 1, 1], LineBox[{46, 47}]}, 
      {Hue[0.10475344510808872`, 1, 1], LineBox[{47, 48}]}, 
      {Hue[0.11967422102295393`, 1, 1], LineBox[{48, 49}]}, 
      {Hue[0.13459499693781912`, 1, 1], LineBox[{49, 50}]}, 
      {Hue[0.1495157728526843, 1, 1], LineBox[{50, 51}]}, 
      {Hue[0.16443654876754946`, 1, 1], LineBox[{51, 52}]}, 
      {Hue[0.1793573246824147, 1, 1], LineBox[{52, 53}]}, 
      {Hue[0.19427810059727987`, 1, 1], LineBox[{53, 54}]}, 
      {Hue[0.20919887651214505`, 1, 1], LineBox[{54, 55}]}, 
      {Hue[0.22411965242701024`, 1, 1], LineBox[{55, 56}]}, 
      {Hue[0.23904042834187544`, 1, 1], LineBox[{56, 57}]}, 
      {Hue[0.2539612042567406, 1, 1], LineBox[{57, 58}]}, 
      {Hue[0.2688819801716058, 1, 1], LineBox[{58, 59}]}, 
      {Hue[0.283802756086471, 1, 1], LineBox[{59, 60}]}, 
      {Hue[0.29872353200133617`, 1, 1], LineBox[{60, 61}]}, 
      {Hue[0.3136443079162014, 1, 1], LineBox[{61, 62}]}, 
      {Hue[0.32856508383106653`, 1, 1], LineBox[{62, 63}]}, 
      {Hue[0.34348585974593177`, 1, 1], LineBox[{63, 64}]}, 
      {Hue[0.358406635660797, 1, 1], LineBox[{64, 65}]}, 
      {Hue[0.3733274115756622, 1, 1], LineBox[{65, 66}]}, 
      {Hue[0.38824818749052736`, 1, 1], LineBox[{66, 67}]}, 
      {Hue[0.40316896340539254`, 1, 1], LineBox[{67, 68}]}, 
      {Hue[0.4180897393202577, 1, 1], LineBox[{68, 69}]}, 
      {Hue[0.4330105152351229, 1, 1], LineBox[{69, 70}]}, 
      {Hue[0.44793129114998803`, 1, 1], LineBox[{70, 71}]}, 
      {Hue[0.4628520670648533, 1, 1], LineBox[{71, 72}]}, 
      {Hue[0.4777728429797184, 1, 1], LineBox[{72, 73}]}, 
      {Hue[0.4926936188945837, 1, 1], LineBox[{73, 74}]}, 
      {Hue[0.5076143948094487, 1, 1], LineBox[{74, 75}]}, 
      {Hue[0.522535170724314, 1, 1], LineBox[{75, 76}]}, 
      {Hue[0.5374559466391792, 1, 1], LineBox[{76, 77}]}, 
      {Hue[0.5523767225540444, 1, 1], LineBox[{77, 78}]}, 
      {Hue[0.5672974984689095, 1, 1], LineBox[{78, 79}]}, 
      {Hue[0.5822182743837747, 1, 1], LineBox[{79, 80}]}, 
      {Hue[0.5971390502986399, 1, 1], LineBox[{80, 81}]}, 
      {Hue[0.6120598262135051, 1, 1], LineBox[{81, 82}]}, 
      {Hue[0.6269806021283703, 1, 1], LineBox[{82, 83}]}, 
      {Hue[0.6419013780432354, 1, 1], LineBox[{83, 84}]}, 
      {Hue[0.6568221539581007, 1, 1], LineBox[{84, 85}]}, 
      {Hue[0.6717429298729658, 1, 1], LineBox[{85, 86}]}, 
      {Hue[0.6866637057878311, 1, 1], LineBox[{86, 87}]}, 
      {Hue[0.7015844817026962, 1, 1], LineBox[{87, 88}]}, 
      {Hue[0.7165052576175615, 1, 1], LineBox[{88, 89}]}, 
      {Hue[0.7314260335324265, 1, 1], LineBox[{89, 90}]}, 
      {Hue[0.7463468094472918, 1, 1], LineBox[{90, 91}]}, 
      {Hue[0.761267585362157, 1, 1], LineBox[{91, 92}]}, 
      {Hue[0.7761883612770222, 1, 1], LineBox[{92, 93}]}, 
      {Hue[0.7911091371918874, 1, 1], LineBox[{93, 94}]}, 
      {Hue[0.8060299131067525, 1, 1], LineBox[{94, 95}]}, 
      {Hue[0.8209506890216177, 1, 1], LineBox[{95, 96}]}, 
      {Hue[0.8358714649364829, 1, 1], LineBox[{96, 97}]}, 
      {Hue[0.8507922408513481, 1, 1], LineBox[{97, 98}]}, 
      {Hue[0.8657130167662133, 1, 1], LineBox[{98, 99}]}, 
      {Hue[0.8806337926810786, 1, 1], LineBox[{99, 100}]}, 
      {Hue[0.8955545685959436, 1, 1], LineBox[{100, 101}]}, 
      {Hue[0.9104753445108089, 1, 1], LineBox[{101, 102}]}, 
      {Hue[0.9253961204256741, 1, 1], LineBox[{102, 103}]}, 
      {Hue[0.9403168963405393, 1, 1], LineBox[{103, 104}]}, 
      {Hue[0.9552376722554045, 1, 1], LineBox[{104, 105}]}, 
      {Hue[0.9701584481702696, 1, 1], LineBox[{105, 106}]}, 
      {Hue[0.9850792240851348, 1, 1], LineBox[{106, 107}]}, 
      {Hue[0., 1, 1], LineBox[{107, 108}]}, 
      {Hue[0.014920775914865188`, 1, 1], LineBox[{108, 109}]}, 
      {Hue[0.029841551829730376`, 1, 1], LineBox[{109, 110}]}, 
      {Hue[0.044762327744595556`, 1, 1], LineBox[{110, 111}]}, 
      {Hue[0.05968310365946075, 1, 1], LineBox[{111, 112}]}, 
      {Hue[0.07460387957432593, 1, 1], LineBox[{112, 113}]}, 
      {Hue[0.08952465548919111, 1, 1], LineBox[{113, 114}]}, 
      {Hue[0.10444543140405632`, 1, 1], LineBox[{114, 115}]}, 
      {Hue[0.11936620731892149`, 1, 1], LineBox[{115, 116}]}, 
      {Hue[0.1342869832337867, 1, 1], LineBox[{116, 117}]}, 
      {Hue[0.1492077591486519, 1, 1], LineBox[{117, 118}]}, 
      {Hue[0.16412853506351707`, 1, 1], LineBox[{118, 119}]}, 
      {Hue[0.17904931097838225`, 1, 1], LineBox[{119, 120}]}, 
      {Hue[0.19397008689324746`, 1, 1], LineBox[{120, 121}]}, 
      {Hue[0.20889086280811264`, 1, 1], LineBox[{121, 122}]}, 
      {Hue[0.22381163872297782`, 1, 1], LineBox[{122, 123}]}, 
      {Hue[0.238732414637843, 1, 1], LineBox[{123, 124}]}, 
      {Hue[0.2536531905527082, 1, 1], LineBox[{124, 125}]}, 
      {Hue[0.2685739664675734, 1, 1], LineBox[{125, 126}]}, 
      {Hue[0.2834947423824386, 1, 1], LineBox[{126, 127}]}, 
      {Hue[0.2984155182973038, 1, 1], LineBox[{127, 128}]}, 
      {Hue[0.31333629421216896`, 1, 1], LineBox[{128, 129}]}, 
      {Hue[0.32825707012703415`, 1, 1], LineBox[{129, 130}]}, 
      {Hue[0.3431778460418993, 1, 1], LineBox[{130, 131}]}, 
      {Hue[0.3580986219567645, 1, 1], LineBox[{131, 132}]}, 
      {Hue[0.37301939787162974`, 1, 1], LineBox[{132, 133}]}, 
      {Hue[0.3879401737864949, 1, 1], LineBox[{133, 134}]}, 
      {Hue[0.4028609497013601, 1, 1], LineBox[{134, 135}]}, 
      {Hue[0.4177817256162253, 1, 1], LineBox[{135, 136}]}, 
      {Hue[0.43270250153109047`, 1, 1], LineBox[{136, 137}]}, 
      {Hue[0.44762327744595565`, 1, 1], LineBox[{137, 138}]}, 
      {Hue[0.46254405336082083`, 1, 1], LineBox[{138, 139}]}, 
      {Hue[0.477464829275686, 1, 1], LineBox[{139, 140}]}, 
      {Hue[0.49238560519055125`, 1, 1], LineBox[{140, 141}]}, 
      {Hue[0.5073063811054164, 1, 1], LineBox[{141, 142}]}, 
      {Hue[0.5222271570202817, 1, 1], LineBox[{142, 143}]}, 
      {Hue[0.5371479329351467, 1, 1], LineBox[{143, 144}]}, 
      {Hue[0.552068708850012, 1, 1], LineBox[{144, 145}]}, 
      {Hue[0.5669894847648771, 1, 1], LineBox[{145, 146}]}, 
      {Hue[0.5819102606797423, 1, 1], LineBox[{146, 147}]}, 
      {Hue[0.5968310365946075, 1, 1], LineBox[{147, 148}]}, 
      {Hue[0.6117518125094726, 1, 1], LineBox[{148, 149}]}, 
      {Hue[0.6266725884243378, 1, 1], LineBox[{149, 150}]}, 
      {Hue[0.641593364339203, 1, 1], LineBox[{150, 151}]}, 
      {Hue[0.6565141402540682, 1, 1], LineBox[{151, 152}]}, 
      {Hue[0.6714349161689335, 1, 1], LineBox[{152, 153}]}, 
      {Hue[0.6863556920837985, 1, 1], LineBox[{153, 154}]}, 
      {Hue[0.7012764679986638, 1, 1], LineBox[{154, 155}]}, 
      {Hue[0.716197243913529, 1, 1], LineBox[{155, 156}]}, 
      {Hue[0.7311180198283942, 1, 1], LineBox[{156, 157}]}, 
      {Hue[0.7460387957432594, 1, 1], LineBox[{157, 158}]}, 
      {Hue[0.7609595716581246, 1, 1], LineBox[{158, 159}]}, 
      {Hue[0.7758803475729897, 1, 1], LineBox[{159, 160}]}, 
      {Hue[0.7908011234878549, 1, 1], LineBox[{160, 161}]}, 
      {Hue[0.8057218994027201, 1, 1], LineBox[{161, 162}]}, 
      {Hue[0.8206426753175853, 1, 1], LineBox[{162, 163}]}, 
      {Hue[0.8355634512324506, 1, 1], LineBox[{163, 164}]}, 
      {Hue[0.8504842271473156, 1, 1], LineBox[{164, 165}]}, 
      {Hue[0.8654050030621809, 1, 1], LineBox[{165, 166}]}, 
      {Hue[0.8803257789770461, 1, 1], LineBox[{166, 167}]}, 
      {Hue[0.8952465548919113, 1, 1], LineBox[{167, 168}]}, 
      {Hue[0.9101673308067765, 1, 1], LineBox[{168, 169}]}, 
      {Hue[0.9250881067216417, 1, 1], LineBox[{169, 170}]}, 
      {Hue[0.9400088826365068, 1, 1], LineBox[{170, 171}]}, 
      {Hue[0.954929658551372, 1, 1], LineBox[{171, 172}]}, 
      {Hue[0.9698504344662372, 1, 1], LineBox[{172, 173}]}, 
      {Hue[0.9847712103811024, 1, 1], LineBox[{173, 174}]}, 
      {Hue[0.9996919862959676, 1, 1], LineBox[{174, 175}]}, 
      {Hue[0.014612762210832766`, 1, 1], LineBox[{175, 176}]}, 
      {Hue[0.029533538125697957`, 1, 1], LineBox[{176, 177}]}, 
      {Hue[0.044454314040563145`, 1, 1], LineBox[{177, 178}]}, 
      {Hue[0.05937508995542833, 1, 1], LineBox[{178, 179}]}, 
      {Hue[0.07429586587029352, 1, 1], LineBox[{179, 180}]}, 
      {Hue[0.08921664178515872, 1, 1], LineBox[{180, 181}]}, 
      {Hue[0.1041374177000239, 1, 1], LineBox[{181, 182}]}, 
      {Hue[0.11905819361488908`, 1, 1], LineBox[{182, 183}]}, 
      {Hue[0.13397896952975427`, 1, 1], LineBox[{183, 184}]}, 
      {Hue[0.14889974544461948`, 1, 1], LineBox[{184, 185}]}, 
      {Hue[0.16382052135948466`, 1, 1], LineBox[{185, 186}]}, 
      {Hue[0.17874129727434984`, 1, 1], LineBox[{186, 187}]}, 
      {Hue[0.19366207318921502`, 1, 1], LineBox[{187, 188}]}, 
      {Hue[0.20858284910408023`, 1, 1], LineBox[{188, 189}]}, 
      {Hue[0.2235036250189454, 1, 1], LineBox[{189, 190}]}, 
      {Hue[0.2384244009338106, 1, 1], LineBox[{190, 191}]}, 
      {Hue[0.2533451768486758, 1, 1], LineBox[{191, 192}]}, 
      {Hue[0.268265952763541, 1, 1], LineBox[{192, 193}]}, 
      {Hue[0.28318672867840616`, 1, 1], LineBox[{193, 194}]}, 
      {Hue[0.29810750459327134`, 1, 1], LineBox[{194, 195}]}, 
      {Hue[0.3130282805081365, 1, 1], LineBox[{195, 196}]}, 
      {Hue[0.3279490564230017, 1, 1], LineBox[{196, 197}]}, 
      {Hue[0.34286983233786694`, 1, 1], LineBox[{197, 198}]}, 
      {Hue[0.3577906082527321, 1, 1], LineBox[{198, 199}]}, 
      {Hue[0.37271138416759725`, 1, 1], LineBox[{199, 200}]}, 
      {Hue[0.38763216008246254`, 1, 1], LineBox[{200, 201}]}, 
      {Hue[0.4025529359973277, 1, 1], LineBox[{201, 202}]}, 
      {Hue[0.4174737119121929, 1, 1], LineBox[{202, 203}]}, 
      {Hue[0.432394487827058, 1, 1], LineBox[{203, 204}]}, 
      {Hue[0.4473152637419232, 1, 1], LineBox[{204, 205}]}, 
      {Hue[0.4622360396567884, 1, 1], LineBox[{205, 206}]}, 
      {Hue[0.47715681557165357`, 1, 1], LineBox[{206, 207}]}, 
      {Hue[0.49207759148651875`, 1, 1], LineBox[{207, 208}]}, 
      {Hue[0.5069983674013839, 1, 1], LineBox[{208, 209}]}, 
      {Hue[0.5219191433162491, 1, 1], LineBox[{209, 210}]}, 
      {Hue[0.5368399192311143, 1, 1], LineBox[{210, 211}]}, 
      {Hue[0.5517606951459795, 1, 1], LineBox[{211, 212}]}, 
      {Hue[0.5666814710608448, 1, 1], LineBox[{212, 213}]}}}},
   VertexColors->{
     Hue[0.4183977530242901, 1, 1], 
     Hue[0.4333185289391553, 1, 1], 
     Hue[0.44823930485402047`, 1, 1], 
     Hue[0.4631600807688857, 1, 1], 
     Hue[0.4780808566837509, 1, 1], 
     Hue[0.49300163259861607`, 1, 1], 
     Hue[0.5079224085134812, 1, 1], 
     Hue[0.5228431844283464, 1, 1], 
     Hue[0.5377639603432116, 1, 1], 
     Hue[0.5526847362580768, 1, 1], 
     Hue[0.567605512172942, 1, 1], 
     Hue[0.582526288087807, 1, 1], 
     Hue[0.5974470640026723, 1, 1], 
     Hue[0.6123678399175374, 1, 1], 
     Hue[0.6272886158324027, 1, 1], 
     Hue[0.6422093917472679, 1, 1], 
     Hue[0.6571301676621331, 1, 1], 
     Hue[0.6720509435769983, 1, 1], 
     Hue[0.6869717194918634, 1, 1], 
     Hue[0.7018924954067287, 1, 1], 
     Hue[0.7168132713215938, 1, 1], 
     Hue[0.7317340472364591, 1, 1], 
     Hue[0.7466548231513241, 1, 1], 
     Hue[0.7615755990661894, 1, 1], 
     Hue[0.7764963749810546, 1, 1], 
     Hue[0.7914171508959198, 1, 1], 
     Hue[0.806337926810785, 1, 1], 
     Hue[0.8212587027256502, 1, 1], 
     Hue[0.8361794786405153, 1, 1], 
     Hue[0.8511002545553805, 1, 1], 
     Hue[0.8660210304702457, 1, 1], 
     Hue[0.8809418063851109, 1, 1], 
     Hue[0.8958625822999761, 1, 1], 
     Hue[0.9107833582148412, 1, 1], 
     Hue[0.9257041341297065, 1, 1], 
     Hue[0.9406249100445717, 1, 1], 
     Hue[0.9555456859594369, 1, 1], 
     Hue[0.9704664618743021, 1, 1], 
     Hue[0.9853872377891673, 1, 1], 
     Hue[0.0003080137040324222, 1, 1], 
     Hue[0.01522878961889761, 1, 1], 
     Hue[0.0301495655337628, 1, 1], 
     Hue[0.045070341448627996`, 1, 1], 
     Hue[0.059991117363493184`, 1, 1], 
     Hue[0.07491189327835837, 1, 1], 
     Hue[0.08983266919322357, 1, 1], 
     Hue[0.10475344510808872`, 1, 1], 
     Hue[0.11967422102295393`, 1, 1], 
     Hue[0.13459499693781912`, 1, 1], 
     Hue[0.1495157728526843, 1, 1], 
     Hue[0.16443654876754946`, 1, 1], 
     Hue[0.1793573246824147, 1, 1], 
     Hue[0.19427810059727987`, 1, 1], 
     Hue[0.20919887651214505`, 1, 1], 
     Hue[0.22411965242701024`, 1, 1], 
     Hue[0.23904042834187544`, 1, 1], 
     Hue[0.2539612042567406, 1, 1], 
     Hue[0.2688819801716058, 1, 1], 
     Hue[0.283802756086471, 1, 1], 
     Hue[0.29872353200133617`, 1, 1], 
     Hue[0.3136443079162014, 1, 1], 
     Hue[0.32856508383106653`, 1, 1], 
     Hue[0.34348585974593177`, 1, 1], 
     Hue[0.358406635660797, 1, 1], 
     Hue[0.3733274115756622, 1, 1], 
     Hue[0.38824818749052736`, 1, 1], 
     Hue[0.40316896340539254`, 1, 1], 
     Hue[0.4180897393202577, 1, 1], 
     Hue[0.4330105152351229, 1, 1], 
     Hue[0.44793129114998803`, 1, 1], 
     Hue[0.4628520670648533, 1, 1], 
     Hue[0.4777728429797184, 1, 1], 
     Hue[0.4926936188945837, 1, 1], 
     Hue[0.5076143948094487, 1, 1], 
     Hue[0.522535170724314, 1, 1], 
     Hue[0.5374559466391792, 1, 1], 
     Hue[0.5523767225540444, 1, 1], 
     Hue[0.5672974984689095, 1, 1], 
     Hue[0.5822182743837747, 1, 1], 
     Hue[0.5971390502986399, 1, 1], 
     Hue[0.6120598262135051, 1, 1], 
     Hue[0.6269806021283703, 1, 1], 
     Hue[0.6419013780432354, 1, 1], 
     Hue[0.6568221539581007, 1, 1], 
     Hue[0.6717429298729658, 1, 1], 
     Hue[0.6866637057878311, 1, 1], 
     Hue[0.7015844817026962, 1, 1], 
     Hue[0.7165052576175615, 1, 1], 
     Hue[0.7314260335324265, 1, 1], 
     Hue[0.7463468094472918, 1, 1], 
     Hue[0.761267585362157, 1, 1], 
     Hue[0.7761883612770222, 1, 1], 
     Hue[0.7911091371918874, 1, 1], 
     Hue[0.8060299131067525, 1, 1], 
     Hue[0.8209506890216177, 1, 1], 
     Hue[0.8358714649364829, 1, 1], 
     Hue[0.8507922408513481, 1, 1], 
     Hue[0.8657130167662133, 1, 1], 
     Hue[0.8806337926810786, 1, 1], 
     Hue[0.8955545685959436, 1, 1], 
     Hue[0.9104753445108089, 1, 1], 
     Hue[0.9253961204256741, 1, 1], 
     Hue[0.9403168963405393, 1, 1], 
     Hue[0.9552376722554045, 1, 1], 
     Hue[0.9701584481702696, 1, 1], 
     Hue[0.9850792240851348, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.014920775914865188`, 1, 1], 
     Hue[0.029841551829730376`, 1, 1], 
     Hue[0.044762327744595556`, 1, 1], 
     Hue[0.05968310365946075, 1, 1], 
     Hue[0.07460387957432593, 1, 1], 
     Hue[0.08952465548919111, 1, 1], 
     Hue[0.10444543140405632`, 1, 1], 
     Hue[0.11936620731892149`, 1, 1], 
     Hue[0.1342869832337867, 1, 1], 
     Hue[0.1492077591486519, 1, 1], 
     Hue[0.16412853506351707`, 1, 1], 
     Hue[0.17904931097838225`, 1, 1], 
     Hue[0.19397008689324746`, 1, 1], 
     Hue[0.20889086280811264`, 1, 1], 
     Hue[0.22381163872297782`, 1, 1], 
     Hue[0.238732414637843, 1, 1], 
     Hue[0.2536531905527082, 1, 1], 
     Hue[0.2685739664675734, 1, 1], 
     Hue[0.2834947423824386, 1, 1], 
     Hue[0.2984155182973038, 1, 1], 
     Hue[0.31333629421216896`, 1, 1], 
     Hue[0.32825707012703415`, 1, 1], 
     Hue[0.3431778460418993, 1, 1], 
     Hue[0.3580986219567645, 1, 1], 
     Hue[0.37301939787162974`, 1, 1], 
     Hue[0.3879401737864949, 1, 1], 
     Hue[0.4028609497013601, 1, 1], 
     Hue[0.4177817256162253, 1, 1], 
     Hue[0.43270250153109047`, 1, 1], 
     Hue[0.44762327744595565`, 1, 1], 
     Hue[0.46254405336082083`, 1, 1], 
     Hue[0.477464829275686, 1, 1], 
     Hue[0.49238560519055125`, 1, 1], 
     Hue[0.5073063811054164, 1, 1], 
     Hue[0.5222271570202817, 1, 1], 
     Hue[0.5371479329351467, 1, 1], 
     Hue[0.552068708850012, 1, 1], 
     Hue[0.5669894847648771, 1, 1], 
     Hue[0.5819102606797423, 1, 1], 
     Hue[0.5968310365946075, 1, 1], 
     Hue[0.6117518125094726, 1, 1], 
     Hue[0.6266725884243378, 1, 1], 
     Hue[0.641593364339203, 1, 1], 
     Hue[0.6565141402540682, 1, 1], 
     Hue[0.6714349161689335, 1, 1], 
     Hue[0.6863556920837985, 1, 1], 
     Hue[0.7012764679986638, 1, 1], 
     Hue[0.716197243913529, 1, 1], 
     Hue[0.7311180198283942, 1, 1], 
     Hue[0.7460387957432594, 1, 1], 
     Hue[0.7609595716581246, 1, 1], 
     Hue[0.7758803475729897, 1, 1], 
     Hue[0.7908011234878549, 1, 1], 
     Hue[0.8057218994027201, 1, 1], 
     Hue[0.8206426753175853, 1, 1], 
     Hue[0.8355634512324506, 1, 1], 
     Hue[0.8504842271473156, 1, 1], 
     Hue[0.8654050030621809, 1, 1], 
     Hue[0.8803257789770461, 1, 1], 
     Hue[0.8952465548919113, 1, 1], 
     Hue[0.9101673308067765, 1, 1], 
     Hue[0.9250881067216417, 1, 1], 
     Hue[0.9400088826365068, 1, 1], 
     Hue[0.954929658551372, 1, 1], 
     Hue[0.9698504344662372, 1, 1], 
     Hue[0.9847712103811024, 1, 1], 
     Hue[0.9996919862959676, 1, 1], 
     Hue[0.014612762210832766`, 1, 1], 
     Hue[0.029533538125697957`, 1, 1], 
     Hue[0.044454314040563145`, 1, 1], 
     Hue[0.05937508995542833, 1, 1], 
     Hue[0.07429586587029352, 1, 1], 
     Hue[0.08921664178515872, 1, 1], 
     Hue[0.1041374177000239, 1, 1], 
     Hue[0.11905819361488908`, 1, 1], 
     Hue[0.13397896952975427`, 1, 1], 
     Hue[0.14889974544461948`, 1, 1], 
     Hue[0.16382052135948466`, 1, 1], 
     Hue[0.17874129727434984`, 1, 1], 
     Hue[0.19366207318921502`, 1, 1], 
     Hue[0.20858284910408023`, 1, 1], 
     Hue[0.2235036250189454, 1, 1], 
     Hue[0.2384244009338106, 1, 1], 
     Hue[0.2533451768486758, 1, 1], 
     Hue[0.268265952763541, 1, 1], 
     Hue[0.28318672867840616`, 1, 1], 
     Hue[0.29810750459327134`, 1, 1], 
     Hue[0.3130282805081365, 1, 1], 
     Hue[0.3279490564230017, 1, 1], 
     Hue[0.34286983233786694`, 1, 1], 
     Hue[0.3577906082527321, 1, 1], 
     Hue[0.37271138416759725`, 1, 1], 
     Hue[0.38763216008246254`, 1, 1], 
     Hue[0.4025529359973277, 1, 1], 
     Hue[0.4174737119121929, 1, 1], 
     Hue[0.432394487827058, 1, 1], 
     Hue[0.4473152637419232, 1, 1], 
     Hue[0.4622360396567884, 1, 1], 
     Hue[0.47715681557165357`, 1, 1], 
     Hue[0.49207759148651875`, 1, 1], 
     Hue[0.5069983674013839, 1, 1], 
     Hue[0.5219191433162491, 1, 1], 
     Hue[0.5368399192311143, 1, 1], 
     Hue[0.5517606951459795, 1, 1], 
     Hue[0.5666814710608448, 1, 1], 
     Hue[0.5816022469757098, 1, 1], 
     Hue[0.4183977530242901, 1, 1], 
     Hue[0.4333185289391553, 1, 1], 
     Hue[0.4333185289391553, 1, 1], 
     Hue[0.44823930485402047`, 1, 1], 
     Hue[0.44823930485402047`, 1, 1], 
     Hue[0.4631600807688857, 1, 1], 
     Hue[0.4631600807688857, 1, 1], 
     Hue[0.4780808566837509, 1, 1], 
     Hue[0.4780808566837509, 1, 1], 
     Hue[0.49300163259861607`, 1, 1], 
     Hue[0.49300163259861607`, 1, 1], 
     Hue[0.5079224085134812, 1, 1], 
     Hue[0.5079224085134812, 1, 1], 
     Hue[0.5228431844283464, 1, 1], 
     Hue[0.5228431844283464, 1, 1], 
     Hue[0.5377639603432116, 1, 1], 
     Hue[0.5377639603432116, 1, 1], 
     Hue[0.5526847362580768, 1, 1], 
     Hue[0.5526847362580768, 1, 1], 
     Hue[0.567605512172942, 1, 1], 
     Hue[0.567605512172942, 1, 1], 
     Hue[0.582526288087807, 1, 1], 
     Hue[0.582526288087807, 1, 1], 
     Hue[0.5974470640026723, 1, 1], 
     Hue[0.5974470640026723, 1, 1], 
     Hue[0.6123678399175374, 1, 1], 
     Hue[0.6123678399175374, 1, 1], 
     Hue[0.6272886158324027, 1, 1], 
     Hue[0.6272886158324027, 1, 1], 
     Hue[0.6422093917472679, 1, 1], 
     Hue[0.6422093917472679, 1, 1], 
     Hue[0.6571301676621331, 1, 1], 
     Hue[0.6571301676621331, 1, 1], 
     Hue[0.6720509435769983, 1, 1], 
     Hue[0.6720509435769983, 1, 1], 
     Hue[0.6869717194918634, 1, 1], 
     Hue[0.6869717194918634, 1, 1], 
     Hue[0.7018924954067287, 1, 1], 
     Hue[0.7018924954067287, 1, 1], 
     Hue[0.7168132713215938, 1, 1], 
     Hue[0.7168132713215938, 1, 1], 
     Hue[0.7317340472364591, 1, 1], 
     Hue[0.7317340472364591, 1, 1], 
     Hue[0.7466548231513241, 1, 1], 
     Hue[0.7466548231513241, 1, 1], 
     Hue[0.7615755990661894, 1, 1], 
     Hue[0.7615755990661894, 1, 1], 
     Hue[0.7764963749810546, 1, 1], 
     Hue[0.7764963749810546, 1, 1], 
     Hue[0.7914171508959198, 1, 1], 
     Hue[0.7914171508959198, 1, 1], 
     Hue[0.806337926810785, 1, 1], 
     Hue[0.806337926810785, 1, 1], 
     Hue[0.8212587027256502, 1, 1], 
     Hue[0.8212587027256502, 1, 1], 
     Hue[0.8361794786405153, 1, 1], 
     Hue[0.8361794786405153, 1, 1], 
     Hue[0.8511002545553805, 1, 1], 
     Hue[0.8511002545553805, 1, 1], 
     Hue[0.8660210304702457, 1, 1], 
     Hue[0.8660210304702457, 1, 1], 
     Hue[0.8809418063851109, 1, 1], 
     Hue[0.8809418063851109, 1, 1], 
     Hue[0.8958625822999761, 1, 1], 
     Hue[0.8958625822999761, 1, 1], 
     Hue[0.9107833582148412, 1, 1], 
     Hue[0.9107833582148412, 1, 1], 
     Hue[0.9257041341297065, 1, 1], 
     Hue[0.9257041341297065, 1, 1], 
     Hue[0.9406249100445717, 1, 1], 
     Hue[0.9406249100445717, 1, 1], 
     Hue[0.9555456859594369, 1, 1], 
     Hue[0.9555456859594369, 1, 1], 
     Hue[0.9704664618743021, 1, 1], 
     Hue[0.9704664618743021, 1, 1], 
     Hue[0.9853872377891673, 1, 1], 
     Hue[0.9853872377891673, 1, 1], 
     Hue[0.0003080137040324222, 1, 1], 
     Hue[0.0003080137040324222, 1, 1], 
     Hue[0.01522878961889761, 1, 1], 
     Hue[0.01522878961889761, 1, 1], 
     Hue[0.0301495655337628, 1, 1], 
     Hue[0.0301495655337628, 1, 1], 
     Hue[0.045070341448627996`, 1, 1], 
     Hue[0.045070341448627996`, 1, 1], 
     Hue[0.059991117363493184`, 1, 1], 
     Hue[0.059991117363493184`, 1, 1], 
     Hue[0.07491189327835837, 1, 1], 
     Hue[0.07491189327835837, 1, 1], 
     Hue[0.08983266919322357, 1, 1], 
     Hue[0.08983266919322357, 1, 1], 
     Hue[0.10475344510808872`, 1, 1], 
     Hue[0.10475344510808872`, 1, 1], 
     Hue[0.11967422102295393`, 1, 1], 
     Hue[0.11967422102295393`, 1, 1], 
     Hue[0.13459499693781912`, 1, 1], 
     Hue[0.13459499693781912`, 1, 1], 
     Hue[0.1495157728526843, 1, 1], 
     Hue[0.1495157728526843, 1, 1], 
     Hue[0.16443654876754946`, 1, 1], 
     Hue[0.16443654876754946`, 1, 1], 
     Hue[0.1793573246824147, 1, 1], 
     Hue[0.1793573246824147, 1, 1], 
     Hue[0.19427810059727987`, 1, 1], 
     Hue[0.19427810059727987`, 1, 1], 
     Hue[0.20919887651214505`, 1, 1], 
     Hue[0.20919887651214505`, 1, 1], 
     Hue[0.22411965242701024`, 1, 1], 
     Hue[0.22411965242701024`, 1, 1], 
     Hue[0.23904042834187544`, 1, 1], 
     Hue[0.23904042834187544`, 1, 1], 
     Hue[0.2539612042567406, 1, 1], 
     Hue[0.2539612042567406, 1, 1], 
     Hue[0.2688819801716058, 1, 1], 
     Hue[0.2688819801716058, 1, 1], 
     Hue[0.283802756086471, 1, 1], 
     Hue[0.283802756086471, 1, 1], 
     Hue[0.29872353200133617`, 1, 1], 
     Hue[0.29872353200133617`, 1, 1], 
     Hue[0.3136443079162014, 1, 1], 
     Hue[0.3136443079162014, 1, 1], 
     Hue[0.32856508383106653`, 1, 1], 
     Hue[0.32856508383106653`, 1, 1], 
     Hue[0.34348585974593177`, 1, 1], 
     Hue[0.34348585974593177`, 1, 1], 
     Hue[0.358406635660797, 1, 1], 
     Hue[0.358406635660797, 1, 1], 
     Hue[0.3733274115756622, 1, 1], 
     Hue[0.3733274115756622, 1, 1], 
     Hue[0.38824818749052736`, 1, 1], 
     Hue[0.38824818749052736`, 1, 1], 
     Hue[0.40316896340539254`, 1, 1], 
     Hue[0.40316896340539254`, 1, 1], 
     Hue[0.4180897393202577, 1, 1], 
     Hue[0.4180897393202577, 1, 1], 
     Hue[0.4330105152351229, 1, 1], 
     Hue[0.4330105152351229, 1, 1], 
     Hue[0.44793129114998803`, 1, 1], 
     Hue[0.44793129114998803`, 1, 1], 
     Hue[0.4628520670648533, 1, 1], 
     Hue[0.4628520670648533, 1, 1], 
     Hue[0.4777728429797184, 1, 1], 
     Hue[0.4777728429797184, 1, 1], 
     Hue[0.4926936188945837, 1, 1], 
     Hue[0.4926936188945837, 1, 1], 
     Hue[0.5076143948094487, 1, 1], 
     Hue[0.5076143948094487, 1, 1], 
     Hue[0.522535170724314, 1, 1], 
     Hue[0.522535170724314, 1, 1], 
     Hue[0.5374559466391792, 1, 1], 
     Hue[0.5374559466391792, 1, 1], 
     Hue[0.5523767225540444, 1, 1], 
     Hue[0.5523767225540444, 1, 1], 
     Hue[0.5672974984689095, 1, 1], 
     Hue[0.5672974984689095, 1, 1], 
     Hue[0.5822182743837747, 1, 1], 
     Hue[0.5822182743837747, 1, 1], 
     Hue[0.5971390502986399, 1, 1], 
     Hue[0.5971390502986399, 1, 1], 
     Hue[0.6120598262135051, 1, 1], 
     Hue[0.6120598262135051, 1, 1], 
     Hue[0.6269806021283703, 1, 1], 
     Hue[0.6269806021283703, 1, 1], 
     Hue[0.6419013780432354, 1, 1], 
     Hue[0.6419013780432354, 1, 1], 
     Hue[0.6568221539581007, 1, 1], 
     Hue[0.6568221539581007, 1, 1], 
     Hue[0.6717429298729658, 1, 1], 
     Hue[0.6717429298729658, 1, 1], 
     Hue[0.6866637057878311, 1, 1], 
     Hue[0.6866637057878311, 1, 1], 
     Hue[0.7015844817026962, 1, 1], 
     Hue[0.7015844817026962, 1, 1], 
     Hue[0.7165052576175615, 1, 1], 
     Hue[0.7165052576175615, 1, 1], 
     Hue[0.7314260335324265, 1, 1], 
     Hue[0.7314260335324265, 1, 1], 
     Hue[0.7463468094472918, 1, 1], 
     Hue[0.7463468094472918, 1, 1], 
     Hue[0.761267585362157, 1, 1], 
     Hue[0.761267585362157, 1, 1], 
     Hue[0.7761883612770222, 1, 1], 
     Hue[0.7761883612770222, 1, 1], 
     Hue[0.7911091371918874, 1, 1], 
     Hue[0.7911091371918874, 1, 1], 
     Hue[0.8060299131067525, 1, 1], 
     Hue[0.8060299131067525, 1, 1], 
     Hue[0.8209506890216177, 1, 1], 
     Hue[0.8209506890216177, 1, 1], 
     Hue[0.8358714649364829, 1, 1], 
     Hue[0.8358714649364829, 1, 1], 
     Hue[0.8507922408513481, 1, 1], 
     Hue[0.8507922408513481, 1, 1], 
     Hue[0.8657130167662133, 1, 1], 
     Hue[0.8657130167662133, 1, 1], 
     Hue[0.8806337926810786, 1, 1], 
     Hue[0.8806337926810786, 1, 1], 
     Hue[0.8955545685959436, 1, 1], 
     Hue[0.8955545685959436, 1, 1], 
     Hue[0.9104753445108089, 1, 1], 
     Hue[0.9104753445108089, 1, 1], 
     Hue[0.9253961204256741, 1, 1], 
     Hue[0.9253961204256741, 1, 1], 
     Hue[0.9403168963405393, 1, 1], 
     Hue[0.9403168963405393, 1, 1], 
     Hue[0.9552376722554045, 1, 1], 
     Hue[0.9552376722554045, 1, 1], 
     Hue[0.9701584481702696, 1, 1], 
     Hue[0.9701584481702696, 1, 1], 
     Hue[0.9850792240851348, 1, 1], 
     Hue[0.9850792240851348, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.014920775914865188`, 1, 1], 
     Hue[0.014920775914865188`, 1, 1], 
     Hue[0.029841551829730376`, 1, 1], 
     Hue[0.029841551829730376`, 1, 1], 
     Hue[0.044762327744595556`, 1, 1], 
     Hue[0.044762327744595556`, 1, 1], 
     Hue[0.05968310365946075, 1, 1], 
     Hue[0.05968310365946075, 1, 1], 
     Hue[0.07460387957432593, 1, 1], 
     Hue[0.07460387957432593, 1, 1], 
     Hue[0.08952465548919111, 1, 1], 
     Hue[0.08952465548919111, 1, 1], 
     Hue[0.10444543140405632`, 1, 1], 
     Hue[0.10444543140405632`, 1, 1], 
     Hue[0.11936620731892149`, 1, 1], 
     Hue[0.11936620731892149`, 1, 1], 
     Hue[0.1342869832337867, 1, 1], 
     Hue[0.1342869832337867, 1, 1], 
     Hue[0.1492077591486519, 1, 1], 
     Hue[0.1492077591486519, 1, 1], 
     Hue[0.16412853506351707`, 1, 1], 
     Hue[0.16412853506351707`, 1, 1], 
     Hue[0.17904931097838225`, 1, 1], 
     Hue[0.17904931097838225`, 1, 1], 
     Hue[0.19397008689324746`, 1, 1], 
     Hue[0.19397008689324746`, 1, 1], 
     Hue[0.20889086280811264`, 1, 1], 
     Hue[0.20889086280811264`, 1, 1], 
     Hue[0.22381163872297782`, 1, 1], 
     Hue[0.22381163872297782`, 1, 1], 
     Hue[0.238732414637843, 1, 1], 
     Hue[0.238732414637843, 1, 1], 
     Hue[0.2536531905527082, 1, 1], 
     Hue[0.2536531905527082, 1, 1], 
     Hue[0.2685739664675734, 1, 1], 
     Hue[0.2685739664675734, 1, 1], 
     Hue[0.2834947423824386, 1, 1], 
     Hue[0.2834947423824386, 1, 1], 
     Hue[0.2984155182973038, 1, 1], 
     Hue[0.2984155182973038, 1, 1], 
     Hue[0.31333629421216896`, 1, 1], 
     Hue[0.31333629421216896`, 1, 1], 
     Hue[0.32825707012703415`, 1, 1], 
     Hue[0.32825707012703415`, 1, 1], 
     Hue[0.3431778460418993, 1, 1], 
     Hue[0.3431778460418993, 1, 1], 
     Hue[0.3580986219567645, 1, 1], 
     Hue[0.3580986219567645, 1, 1], 
     Hue[0.37301939787162974`, 1, 1], 
     Hue[0.37301939787162974`, 1, 1], 
     Hue[0.3879401737864949, 1, 1], 
     Hue[0.3879401737864949, 1, 1], 
     Hue[0.4028609497013601, 1, 1], 
     Hue[0.4028609497013601, 1, 1], 
     Hue[0.4177817256162253, 1, 1], 
     Hue[0.4177817256162253, 1, 1], 
     Hue[0.43270250153109047`, 1, 1], 
     Hue[0.43270250153109047`, 1, 1], 
     Hue[0.44762327744595565`, 1, 1], 
     Hue[0.44762327744595565`, 1, 1], 
     Hue[0.46254405336082083`, 1, 1], 
     Hue[0.46254405336082083`, 1, 1], 
     Hue[0.477464829275686, 1, 1], 
     Hue[0.477464829275686, 1, 1], 
     Hue[0.49238560519055125`, 1, 1], 
     Hue[0.49238560519055125`, 1, 1], 
     Hue[0.5073063811054164, 1, 1], 
     Hue[0.5073063811054164, 1, 1], 
     Hue[0.5222271570202817, 1, 1], 
     Hue[0.5222271570202817, 1, 1], 
     Hue[0.5371479329351467, 1, 1], 
     Hue[0.5371479329351467, 1, 1], 
     Hue[0.552068708850012, 1, 1], 
     Hue[0.552068708850012, 1, 1], 
     Hue[0.5669894847648771, 1, 1], 
     Hue[0.5669894847648771, 1, 1], 
     Hue[0.5819102606797423, 1, 1], 
     Hue[0.5819102606797423, 1, 1], 
     Hue[0.5968310365946075, 1, 1], 
     Hue[0.5968310365946075, 1, 1], 
     Hue[0.6117518125094726, 1, 1], 
     Hue[0.6117518125094726, 1, 1], 
     Hue[0.6266725884243378, 1, 1], 
     Hue[0.6266725884243378, 1, 1], 
     Hue[0.641593364339203, 1, 1], 
     Hue[0.641593364339203, 1, 1], 
     Hue[0.6565141402540682, 1, 1], 
     Hue[0.6565141402540682, 1, 1], 
     Hue[0.6714349161689335, 1, 1], 
     Hue[0.6714349161689335, 1, 1], 
     Hue[0.6863556920837985, 1, 1], 
     Hue[0.6863556920837985, 1, 1], 
     Hue[0.7012764679986638, 1, 1], 
     Hue[0.7012764679986638, 1, 1], 
     Hue[0.716197243913529, 1, 1], 
     Hue[0.716197243913529, 1, 1], 
     Hue[0.7311180198283942, 1, 1], 
     Hue[0.7311180198283942, 1, 1], 
     Hue[0.7460387957432594, 1, 1], 
     Hue[0.7460387957432594, 1, 1], 
     Hue[0.7609595716581246, 1, 1], 
     Hue[0.7609595716581246, 1, 1], 
     Hue[0.7758803475729897, 1, 1], 
     Hue[0.7758803475729897, 1, 1], 
     Hue[0.7908011234878549, 1, 1], 
     Hue[0.7908011234878549, 1, 1], 
     Hue[0.8057218994027201, 1, 1], 
     Hue[0.8057218994027201, 1, 1], 
     Hue[0.8206426753175853, 1, 1], 
     Hue[0.8206426753175853, 1, 1], 
     Hue[0.8355634512324506, 1, 1], 
     Hue[0.8355634512324506, 1, 1], 
     Hue[0.8504842271473156, 1, 1], 
     Hue[0.8504842271473156, 1, 1], 
     Hue[0.8654050030621809, 1, 1], 
     Hue[0.8654050030621809, 1, 1], 
     Hue[0.8803257789770461, 1, 1], 
     Hue[0.8803257789770461, 1, 1], 
     Hue[0.8952465548919113, 1, 1], 
     Hue[0.8952465548919113, 1, 1], 
     Hue[0.9101673308067765, 1, 1], 
     Hue[0.9101673308067765, 1, 1], 
     Hue[0.9250881067216417, 1, 1], 
     Hue[0.9250881067216417, 1, 1], 
     Hue[0.9400088826365068, 1, 1], 
     Hue[0.9400088826365068, 1, 1], 
     Hue[0.954929658551372, 1, 1], 
     Hue[0.954929658551372, 1, 1], 
     Hue[0.9698504344662372, 1, 1], 
     Hue[0.9698504344662372, 1, 1], 
     Hue[0.9847712103811024, 1, 1], 
     Hue[0.9847712103811024, 1, 1], 
     Hue[0.9996919862959676, 1, 1], 
     Hue[0.9996919862959676, 1, 1], 
     Hue[0.014612762210832766`, 1, 1], 
     Hue[0.014612762210832766`, 1, 1], 
     Hue[0.029533538125697957`, 1, 1], 
     Hue[0.029533538125697957`, 1, 1], 
     Hue[0.044454314040563145`, 1, 1], 
     Hue[0.044454314040563145`, 1, 1], 
     Hue[0.05937508995542833, 1, 1], 
     Hue[0.05937508995542833, 1, 1], 
     Hue[0.07429586587029352, 1, 1], 
     Hue[0.07429586587029352, 1, 1], 
     Hue[0.08921664178515872, 1, 1], 
     Hue[0.08921664178515872, 1, 1], 
     Hue[0.1041374177000239, 1, 1], 
     Hue[0.1041374177000239, 1, 1], 
     Hue[0.11905819361488908`, 1, 1], 
     Hue[0.11905819361488908`, 1, 1], 
     Hue[0.13397896952975427`, 1, 1], 
     Hue[0.13397896952975427`, 1, 1], 
     Hue[0.14889974544461948`, 1, 1], 
     Hue[0.14889974544461948`, 1, 1], 
     Hue[0.16382052135948466`, 1, 1], 
     Hue[0.16382052135948466`, 1, 1], 
     Hue[0.17874129727434984`, 1, 1], 
     Hue[0.17874129727434984`, 1, 1], 
     Hue[0.19366207318921502`, 1, 1], 
     Hue[0.19366207318921502`, 1, 1], 
     Hue[0.20858284910408023`, 1, 1], 
     Hue[0.20858284910408023`, 1, 1], 
     Hue[0.2235036250189454, 1, 1], 
     Hue[0.2235036250189454, 1, 1], 
     Hue[0.2384244009338106, 1, 1], 
     Hue[0.2384244009338106, 1, 1], 
     Hue[0.2533451768486758, 1, 1], 
     Hue[0.2533451768486758, 1, 1], 
     Hue[0.268265952763541, 1, 1], 
     Hue[0.268265952763541, 1, 1], 
     Hue[0.28318672867840616`, 1, 1], 
     Hue[0.28318672867840616`, 1, 1], 
     Hue[0.29810750459327134`, 1, 1], 
     Hue[0.29810750459327134`, 1, 1], 
     Hue[0.3130282805081365, 1, 1], 
     Hue[0.3130282805081365, 1, 1], 
     Hue[0.3279490564230017, 1, 1], 
     Hue[0.3279490564230017, 1, 1], 
     Hue[0.34286983233786694`, 1, 1], 
     Hue[0.34286983233786694`, 1, 1], 
     Hue[0.3577906082527321, 1, 1], 
     Hue[0.3577906082527321, 1, 1], 
     Hue[0.37271138416759725`, 1, 1], 
     Hue[0.37271138416759725`, 1, 1], 
     Hue[0.38763216008246254`, 1, 1], 
     Hue[0.38763216008246254`, 1, 1], 
     Hue[0.4025529359973277, 1, 1], 
     Hue[0.4025529359973277, 1, 1], 
     Hue[0.4174737119121929, 1, 1], 
     Hue[0.4174737119121929, 1, 1], 
     Hue[0.432394487827058, 1, 1], 
     Hue[0.432394487827058, 1, 1], 
     Hue[0.4473152637419232, 1, 1], 
     Hue[0.4473152637419232, 1, 1], 
     Hue[0.4622360396567884, 1, 1], 
     Hue[0.4622360396567884, 1, 1], 
     Hue[0.47715681557165357`, 1, 1], 
     Hue[0.47715681557165357`, 1, 1], 
     Hue[0.49207759148651875`, 1, 1], 
     Hue[0.49207759148651875`, 1, 1], 
     Hue[0.5069983674013839, 1, 1], 
     Hue[0.5069983674013839, 1, 1], 
     Hue[0.5219191433162491, 1, 1], 
     Hue[0.5219191433162491, 1, 1], 
     Hue[0.5368399192311143, 1, 1], 
     Hue[0.5368399192311143, 1, 1], 
     Hue[0.5517606951459795, 1, 1], 
     Hue[0.5517606951459795, 1, 1], 
     Hue[0.5666814710608448, 1, 1], 
     Hue[0.5666814710608448, 1, 1], 
     Hue[0.5816022469757098, 1, 1]}],
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  Epilog->LineBox[{{-4.96875, 3.2709805595093306`*^-6}, {-4.921875, 
     4.124320705162408*^-6}, {-4.875, 5.188868129033276*^-6}, {-4.828125, 
     6.513862372589593*^-6}, {-4.78125, 8.159250426441251*^-6}, {-4.734375, 
     0.0000101978280994376}, {-4.6875, 0.00001271776660975753}, {-4.640625, 
     0.00001582558431978149}, {-4.59375, 
     0.000019649630916271522`}, {-4.546875, 0.00002434415923693149}, {-4.5, 
     0.000030094068318430215`}, {-4.453125, 
     0.00003712041002129977}, {-4.40625, 0.00004568676067906624}, {-4.359375, 
     0.00005610656849737733}, {-4.3125, 0.00006875159673421496}, {-4.265625, 
     0.00008406159182671469}, {-4.21875, 0.00010255531435391963`}, {-4.171875,
      0.00012484307875301263`}, {-4.125, 0.00015164095470627406`}, {-4.078125,
      0.00018378678870435632`}, {-4.03125, 
     0.00022225820803473448`}, {-3.984375, 0.00026819277085811307`}, {-3.9375,
      0.00032291042458717297`}, {-3.890625, 
     0.00038793842989333545`}, {-3.84375, 0.0004650388987209338}, {-3.796875, 
     0.0005562390810359108}, {-3.75, 0.0006638645160090288}, {-3.703125, 
     0.0007905751382558167}, {-3.65625, 0.0009394043979607872}, {-3.609375, 
     0.001113801414564607}, {-3.5625, 0.001317676136604145}, {-3.515625, 
     0.0015554474247579002`}, {-3.46875, 0.0018320939107596115`}, {-3.421875, 
     0.0021532074113328}, {-3.375, 0.0025250485935685565`}, {-3.328125, 
     0.002954604496319363}, {-3.28125, 0.003449647411549908}, {-3.234375, 
     0.004018794520776898}, {-3.1875, 0.004671567565649339}, {-3.140625, 
     0.0054184517096023945`}, {-3.09375, 0.006270952620948514}, {-3.046875, 
     0.007241650678706022}, {-3., 0.008344251073246413}, {-2.953125, 
     0.009593628447188283}, {-2.90625, 0.01100586460097308}, {-2.859375, 
     0.012598277675673225`}, {-2.8125, 0.014389441126572085`}, {-2.765625, 
     0.016399190718953217`}, {-2.71875, 0.018648617716584236`}, {-2.671875, 
     0.021160046397946577`}, {-2.625, 0.023956994029747086`}, {-2.578125, 
     0.0270641114559759}, {-2.53125, 0.03050710252787448}, {-2.484375, 
     0.03431262070944408}, {-2.4375, 0.03850814134788083}, {-2.390625, 
     0.043121808301270076`}, {-2.34375, 0.04818225386893848}, {-2.296875, 
     0.05371839127405522}, {-2.25, 0.05975917930334211}, {-2.203125, 
     0.06633335911384833}, {-2.15625, 0.07346916366912468}, {-2.109375, 
     0.0811940007628494}, {-2.0625, 0.08953411112162063}, {-2.015625, 
     0.09851420364336069}, {-1.96875, 0.10815707041522599`}, {-1.921875, 
     0.1184831847553152}, {-1.875, 0.12951028612472462`}, {-1.828125, 
     0.14125295634834165`}, {-1.78125, 0.1537221921509343}, {-1.734375, 
     0.1669249795455717}, {-1.6875, 0.1808638760897144}, {-1.640625, 
     0.19553660743580398`}, {-1.59375, 0.21093568493342887`}, {-1.546875, 
     0.2270480512752962}, {-1.5, 0.24385476130642741`}, {-1.453125, 
     0.2613307051237464}, {-1.40625, 0.2794443804718406}, {-1.359375, 
     0.2981577211826469}, {-1.3125, 0.31742598800714367`}, {-1.265625, 
     0.33719772764369327`}, {-1.21875, 0.3574148050814272}, {-1.171875, 
     0.3780125135522741}, {-1.125, 0.3989197654295719}, {-1.078125, 
     0.4200593663358202}, {-1.03125, 0.44134837354154155`}, {-0.984375, 
     0.46269853846919623`}, {-0.9375, 0.48401683178139476`}, {-0.890625, 
     0.5052060481546798}, {-0.84375, 0.5261654864445083}, {-0.796875, 
     0.5467916995610779}, {-0.75, 0.5669793070276937}, {-0.703125, 
     0.5866218619123392}, {-0.65625, 0.6056127626376299}, {-0.609375, 
     0.6238461991120364}, {-0.5625, 0.6412181217122251}, {-0.515625, 
     0.6576272209061722}, {-0.46875, 0.6729759047600129}, {-0.421875, 
     0.6871712612353379}, {-0.375, 0.70012599207068}, {-0.328125, 
     0.711759305159426}, {-0.28125, 0.7219977526896632}, {-0.234375, 
     0.7307760028976727}, {-0.1875, 0.7380375340988992}, {-0.140625, 
     0.7437352406861114}, {-0.09375, 0.7478319420070575}, {-0.046875, 
     0.7503007864315618}, {0, 0.7511255444649425}, {0.046875, 
     0.7503007864315618}, {0.09375, 0.7478319420070575}, {0.140625, 
     0.7437352406861114}, {0.1875, 0.7380375340988992}, {0.234375, 
     0.7307760028976727}, {0.28125, 0.7219977526896632}, {0.328125, 
     0.711759305159426}, {0.375, 0.70012599207068}, {0.421875, 
     0.6871712612353379}, {0.46875, 0.6729759047600129}, {0.515625, 
     0.6576272209061722}, {0.5625, 0.6412181217122251}, {0.609375, 
     0.6238461991120364}, {0.65625, 0.6056127626376299}, {0.703125, 
     0.5866218619123392}, {0.75, 0.5669793070276937}, {0.796875, 
     0.5467916995610779}, {0.84375, 0.5261654864445083}, {0.890625, 
     0.5052060481546798}, {0.9375, 0.48401683178139476`}, {0.984375, 
     0.46269853846919623`}, {1.03125, 0.44134837354154155`}, {1.078125, 
     0.4200593663358202}, {1.125, 0.3989197654295719}, {1.171875, 
     0.3780125135522741}, {1.21875, 0.3574148050814272}, {1.265625, 
     0.33719772764369327`}, {1.3125, 0.31742598800714367`}, {1.359375, 
     0.2981577211826469}, {1.40625, 0.2794443804718406}, {1.453125, 
     0.2613307051237464}, {1.5, 0.24385476130642741`}, {1.546875, 
     0.2270480512752962}, {1.59375, 0.21093568493342887`}, {1.640625, 
     0.19553660743580398`}, {1.6875, 0.1808638760897144}, {1.734375, 
     0.1669249795455717}, {1.78125, 0.1537221921509343}, {1.828125, 
     0.14125295634834165`}, {1.875, 0.12951028612472462`}, {1.921875, 
     0.1184831847553152}, {1.96875, 0.10815707041522599`}, {2.015625, 
     0.09851420364336069}, {2.0625, 0.08953411112162063}, {2.109375, 
     0.0811940007628494}, {2.15625, 0.07346916366912468}, {2.203125, 
     0.06633335911384833}, {2.25, 0.05975917930334211}, {2.296875, 
     0.05371839127405522}, {2.34375, 0.04818225386893848}, {2.390625, 
     0.043121808301270076`}, {2.4375, 0.03850814134788083}, {2.484375, 
     0.03431262070944408}, {2.53125, 0.03050710252787448}, {2.578125, 
     0.0270641114559759}, {2.625, 0.023956994029747086`}, {2.671875, 
     0.021160046397946577`}, {2.71875, 0.018648617716584236`}, {2.765625, 
     0.016399190718953217`}, {2.8125, 0.014389441126572085`}, {2.859375, 
     0.012598277675673225`}, {2.90625, 0.01100586460097308}, {2.953125, 
     0.009593628447188283}, {3., 0.008344251073246413}, {3.046875, 
     0.007241650678706022}, {3.09375, 0.006270952620948514}, {3.140625, 
     0.0054184517096023945`}, {3.1875, 0.004671567565649339}, {3.234375, 
     0.004018794520776898}, {3.28125, 0.003449647411549908}, {3.328125, 
     0.002954604496319363}, {3.375, 0.0025250485935685565`}, {3.421875, 
     0.0021532074113328}, {3.46875, 0.0018320939107596115`}, {3.515625, 
     0.0015554474247579002`}, {3.5625, 0.001317676136604145}, {3.609375, 
     0.001113801414564607}, {3.65625, 0.0009394043979607872}, {3.703125, 
     0.0007905751382558167}, {3.75, 0.0006638645160090288}, {3.796875, 
     0.0005562390810359108}, {3.84375, 0.0004650388987209338}, {3.890625, 
     0.00038793842989333545`}, {3.9375, 0.00032291042458717297`}, {3.984375, 
     0.00026819277085811307`}, {4.03125, 0.00022225820803473448`}, {4.078125, 
     0.00018378678870435632`}, {4.125, 0.00015164095470627406`}, {4.171875, 
     0.00012484307875301263`}, {4.21875, 0.00010255531435391963`}, {4.265625, 
     0.00008406159182671469}, {4.3125, 0.00006875159673421496}, {4.359375, 
     0.00005610656849737733}, {4.40625, 0.00004568676067906624}, {4.453125, 
     0.00003712041002129977}, {4.5, 0.000030094068318430215`}, {4.546875, 
     0.00002434415923693149}, {4.59375, 0.000019649630916271522`}, {4.640625, 
     0.00001582558431978149}, {4.6875, 0.00001271776660975753}, {4.734375, 
     0.0000101978280994376}, {4.78125, 8.159250426441251*^-6}, {4.828125, 
     6.513862372589593*^-6}, {4.875, 5.188868129033276*^-6}, {4.921875, 
     4.124320705162408*^-6}, {4.96875, 3.2709805595093306`*^-6}}],
  PlotRange->All,
  PlotRangeClipping->True,
  PlotRangePadding->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{
  3.3927262244784384`*^9, {3.392726520103526*^9, 3.391849850609375*^9}, 
   3.39185004978125*^9, 3.3924723688741426`*^9}]
}, Open  ]],

Cell["\<\
The points in x-space determine the interval and the grid in k-space, as \
explained above. You can obtain a list of the points in k-space by \
QFourierGrid[-10.,10,100]. Information about the grid in Fourier space can be \
obtained by\
\>", "Text"],

Cell[CellGroupData[{

Cell["\<\
{kleft,kright,dk}={QFourierLeftBorder[a,b,n],QFourierRightBorder[a,b,n],\
QFourierStep[a,b,n]}\
\>", "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"-", "66.7588438887831`"}], ",", "67.02064327658225`", ",", 
   "0.2617993877991494`"}], "}"}]], "Output",
 CellChangeTimes->{
  3.392726245779067*^9, {3.392726520173627*^9, 3.391849850921875*^9}, 
   3.39185004990625*^9, 3.3924723692100697`*^9}]
}, Open  ]],

Cell["Hence ", "Text"],

Cell[CellGroupData[{

Cell["QFourierGrid[a,b,n]==Table[n,{n,kleft,kright,dk}]", "Input"],

Cell[BoxData["True"], "Output",
 CellChangeTimes->{
  3.3927263115336175`*^9, {3.392726520223699*^9, 3.39184985096875*^9}, 
   3.391850049953125*^9, 3.3924723696130953`*^9}]
}, Open  ]],

Cell["\<\
If the function in x-space is given on the points of QGrid, then QFourierList \
returns the values of the Fourier transform on the points QFourierGrid in \
k-space.\
\>", "Text"],

Cell["Ff = QFourierList[fvals,{a,b}];", "Input"],

Cell["\<\
QFourierList returns the list of Fourier values and the interval in k-space: \
{Fflist,{kleft,kright}}.\
\>", "Text"],

Cell["Ffvals = Ff[[1]]; kinterval = Ff[[2]];", "Input"],

Cell[CellGroupData[{

Cell["\<\
QListArgColorPlot[Ffvals,PlotRange->All, \
QHorizontalRange->{kinterval,{-1,5}},Frame->True]\
\>", "Input"],

Cell[BoxData[
 GraphicsBox[GraphicsComplexBox[CompressedData["
1:eJxTTMoPSmViYGBwBmIQzcA9KyxL9OX+j7svz1l3q9++IUaw/+OOB/tjcmxj
Yl/Ot2cA8y/st5n2/+n7m2vtGaDg+pnLQqzCu+wZYsHy9u0/3fktXI5C+Q/s
OdbcPG1hdMEear595ru179g5r9sfAJv3wb7oeBd7//R79gxbrn7Sff7F3oN/
a/K974+g6n/av5f+v/7Qt6f2Dklxu5Y5/7U385r2tHL/c6h+Bgf2BDam2z+f
2zt0fD7Af4HJwdCN18131jP7B5tB5rE4vP9wMKnz6WP7hvvbVXz+sTk8lG1+
uCnqAdR8TocbYTqHNkndtF9gVtPRocPtUH1N5caVvxeh9vE6PIs7yJEtcsL+
QK/DmyNR/A4zp0ecfCy+z34D2H4Bh+sH7ZnZ+jfaNxQxrWg4Iugg9Ytb5a/i
Yqh7hBym7FHk7iyfbC8w78nNebeFHSSmbNi2vq4JFh77YeEIC29cfFj44+QT
AtD4wc9/gJsPjT9cfFh84uLD4hcnHxrfuPiw+MfFh6UHXHxY+sDFh6UXXHxY
+sHFh6UnXHxY+sLFh6U3XHxY+sPFh6VHXHxY+sTFh6VXXHxY+oXxAeBFSaQ=

   "], {
    {GrayLevel[0.5], EdgeForm[None], 
     GraphicsGroupBox[
      PolygonBox[{{1, 24, 25, 2}, {3, 28, 29, 4}, {2, 26, 27, 3}, {21, 64, 65,
        22}, {5, 32, 33, 6}, {7, 36, 37, 8}, {4, 30, 31, 5}, {9, 40, 41, 
       10}, {6, 34, 35, 7}, {10, 42, 43, 11}, {11, 44, 45, 12}, {12, 46, 47, 
       13}, {16, 54, 55, 17}, {15, 52, 53, 16}, {14, 50, 51, 15}, {8, 38, 39, 
       9}, {17, 56, 57, 18}, {13, 48, 49, 14}, {18, 58, 59, 19}, {19, 60, 61, 
       20}, {20, 62, 63, 21}, {22, 66, 67, 23}}]]}, {{}, {}, {
      {Hue[0.005859375000000259, 1, 1], LineBox[{1, 2}]}, 
      {Hue[0.003906250000000009, 1, 1], LineBox[{2, 3}]}, 
      {Hue[0.001953125000000028, 1, 1], LineBox[{3, 4}]}, 
      {Hue[0., 1, 1], LineBox[{4, 5}]}, 
      {Hue[0.998046875, 1, 1], LineBox[{5, 6}]}, 
      {Hue[0.99609375, 1, 1], LineBox[{6, 7}]}, 
      {Hue[0.994140625, 1, 1], LineBox[{7, 8}]}, 
      {Hue[0.9921875, 1, 1], LineBox[{8, 9}]}, 
      {Hue[0.990234375, 1, 1], LineBox[{9, 10}]}, 
      {Hue[0.98828125, 1, 1], LineBox[{10, 11}]}, 
      {Hue[0.986328125, 1, 1], LineBox[{11, 12}]}, 
      {Hue[0.984375, 1, 1], LineBox[{12, 13}]}, 
      {Hue[0.982421875, 1, 1], LineBox[{13, 14}]}, 
      {Hue[0.98046875, 1, 1], LineBox[{14, 15}]}, 
      {Hue[0.978515625, 1, 1], LineBox[{15, 16}]}, 
      {Hue[0.9765625, 1, 1], LineBox[{16, 17}]}, 
      {Hue[0.974609375, 1, 1], LineBox[{17, 18}]}, 
      {Hue[0.97265625, 1, 1], LineBox[{18, 19}]}, 
      {Hue[0.970703125, 1, 1], LineBox[{19, 20}]}, 
      {Hue[0.96875, 1, 1], LineBox[{20, 21}]}, 
      {Hue[0.966796875, 1, 1], LineBox[{21, 22}]}, 
      {Hue[0.96484375, 1, 1], LineBox[{22, 23}]}}}},
   VertexColors->{
     Hue[0.005859375000000259, 1, 1], 
     Hue[0.003906250000000009, 1, 1], 
     Hue[0.001953125000000028, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.998046875, 1, 1], 
     Hue[0.99609375, 1, 1], 
     Hue[0.994140625, 1, 1], 
     Hue[0.9921875, 1, 1], 
     Hue[0.990234375, 1, 1], 
     Hue[0.98828125, 1, 1], 
     Hue[0.986328125, 1, 1], 
     Hue[0.984375, 1, 1], 
     Hue[0.982421875, 1, 1], 
     Hue[0.98046875, 1, 1], 
     Hue[0.978515625, 1, 1], 
     Hue[0.9765625, 1, 1], 
     Hue[0.974609375, 1, 1], 
     Hue[0.97265625, 1, 1], 
     Hue[0.970703125, 1, 1], 
     Hue[0.96875, 1, 1], 
     Hue[0.966796875, 1, 1], 
     Hue[0.96484375, 1, 1], 
     Hue[0.9628906249999999, 1, 1], 
     Hue[0.005859375000000259, 1, 1], 
     Hue[0.003906250000000009, 1, 1], 
     Hue[0.003906250000000009, 1, 1], 
     Hue[0.001953125000000028, 1, 1], 
     Hue[0.001953125000000028, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.998046875, 1, 1], 
     Hue[0.998046875, 1, 1], 
     Hue[0.99609375, 1, 1], 
     Hue[0.99609375, 1, 1], 
     Hue[0.994140625, 1, 1], 
     Hue[0.994140625, 1, 1], 
     Hue[0.9921875, 1, 1], 
     Hue[0.9921875, 1, 1], 
     Hue[0.990234375, 1, 1], 
     Hue[0.990234375, 1, 1], 
     Hue[0.98828125, 1, 1], 
     Hue[0.98828125, 1, 1], 
     Hue[0.986328125, 1, 1], 
     Hue[0.986328125, 1, 1], 
     Hue[0.984375, 1, 1], 
     Hue[0.984375, 1, 1], 
     Hue[0.982421875, 1, 1], 
     Hue[0.982421875, 1, 1], 
     Hue[0.98046875, 1, 1], 
     Hue[0.98046875, 1, 1], 
     Hue[0.978515625, 1, 1], 
     Hue[0.978515625, 1, 1], 
     Hue[0.9765625, 1, 1], 
     Hue[0.9765625, 1, 1], 
     Hue[0.974609375, 1, 1], 
     Hue[0.974609375, 1, 1], 
     Hue[0.97265625, 1, 1], 
     Hue[0.97265625, 1, 1], 
     Hue[0.970703125, 1, 1], 
     Hue[0.970703125, 1, 1], 
     Hue[0.96875, 1, 1], 
     Hue[0.96875, 1, 1], 
     Hue[0.966796875, 1, 1], 
     Hue[0.966796875, 1, 1], 
     Hue[0.96484375, 1, 1], 
     Hue[0.96484375, 1, 1], 
     Hue[0.9628906249999999, 1, 1]}],
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  Epilog->LineBox[{{-0.7838641826095625, 
     0.015553821717245927`}, {-0.5225761217397036, 
     0.031163653139952494`}, {-0.2612880608698447, 0.05830335314366805}, {
     0, 0.10185271939899408`}, {0.2612880608698589, 0.1661443781485554}, {
     0.5225761217397178, 0.25306529910329845`}, {0.7838641826095625, 
     0.3599262076238348}, {1.0451522434794214`, 0.4780004094746949}, {
     1.3064403043492803`, 0.5927574107558735}, {1.567728365219125, 
     0.6863721311027556}, {1.8290164260889838`, 0.7421235546228846}, {
     2.0903044869588427`, 0.7492499389218218}, {2.3515925478286874`, 
     0.7063356750679245}, {2.6128806086985463`, 0.621769612203506}, {
     2.874168669568405, 0.5110716251883184}, {3.13545673043825, 
     0.39225453340746474`}, {3.396744791308109, 0.28111764056971084`}, {
     3.6580328521779677`, 0.18812311103351148`}, {3.9193209130478124`, 
     0.117552029282335}, {4.180608973917671, 0.06858861522028491}, {
     4.44189703478753, 0.03736868920349341}, {4.703185095657375, 
     0.019010678581648024`}, {4.964473156527234, 0.00903069759631063}}],
  Frame->True,
  PlotRange->All,
  PlotRangeClipping->True,
  PlotRangePadding->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{
  3.3927263249328847`*^9, {3.3927265203338575`*^9, 3.39184985134375*^9}, 
   3.391850050875*^9, 3.392472369774249*^9},
 ImageCache->GraphicsData["CompressedBitmap", "\<\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\
\>"]]
}, Open  ]],

Cell["This is the discretization of a shifted Gaussian function.", "Text"],

Cell[TextData[{
 "We can (approximately) reverse the whole process by ",
 StyleBox["QInverseFourierList", "InlineInput"],
 "."
}], "Text"],

Cell["result = QInverseFourierList[Ff];", "Input"],

Cell[CellGroupData[{

Cell["\<\
QListArgColorPlot[result[[1]],PlotRange->All,QHorizontalRange->result[[2]]]\
\>", "Input"],

Cell[BoxData[
 GraphicsBox[GraphicsComplexBox[CompressedData["
1:eJyFmgdQVVe3xxFBMFgREFGjGMWOsZeEs7dJNDGJ9YuJ3dgSa9RoTOxdsXeD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   "], {
    {GrayLevel[0.5], EdgeForm[None], 
     GraphicsGroupBox[{PolygonBox[{{113, 513, 114}, {400, 1086, 401}}], 
       PolygonBox[CompressedData["
1:eJwtl3f8FlIbh894mipJKpUW7aFdmihKSKXICqES2nsaSYUkhGyykpXs7JCV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        "]]}]}, {{}, {}, {
      {Hue[0., 1, 1], LineBox[{1, 2}]}, 
      {Hue[0., 1, 1], LineBox[{2, 3}]}, 
      {Hue[0., 1, 1], LineBox[{3, 4}]}, 
      {Hue[0., 1, 1], LineBox[{4, 5}]}, 
      {Hue[0., 1, 1], LineBox[{5, 6}]}, 
      {Hue[0., 1, 1], LineBox[{6, 7}]}, 
      {Hue[0., 1, 1], LineBox[{7, 8}]}, 
      {Hue[0., 1, 1], LineBox[{8, 9}]}, 
      {Hue[0., 1, 1], LineBox[{9, 10}]}, 
      {Hue[0., 1, 1], LineBox[{10, 11}]}, 
      {Hue[0., 1, 1], LineBox[{11, 12}]}, 
      {Hue[0., 1, 1], LineBox[{12, 13}]}, 
      {Hue[0., 1, 1], LineBox[{13, 14}]}, 
      {Hue[0., 1, 1], LineBox[{14, 15}]}, 
      {Hue[0., 1, 1], LineBox[{15, 16}]}, 
      {Hue[0., 1, 1], LineBox[{16, 17}]}, 
      {Hue[0., 1, 1], LineBox[{17, 18}]}, 
      {Hue[0., 1, 1], LineBox[{18, 19}]}, 
      {Hue[0., 1, 1], LineBox[{19, 20}]}, 
      {Hue[0., 1, 1], LineBox[{20, 21}]}, 
      {Hue[0., 1, 1], LineBox[{21, 22}]}, 
      {Hue[0., 1, 1], LineBox[{22, 23}]}, 
      {Hue[0., 1, 1], LineBox[{23, 24}]}, 
      {Hue[0., 1, 1], LineBox[{24, 25}]}, 
      {Hue[0., 1, 1], LineBox[{25, 26}]}, 
      {Hue[0., 1, 1], LineBox[{26, 27}]}, 
      {Hue[0., 1, 1], LineBox[{27, 28}]}, 
      {Hue[0., 1, 1], LineBox[{28, 29}]}, 
      {Hue[0., 1, 1], LineBox[{29, 30}]}, 
      {Hue[0., 1, 1], LineBox[{30, 31}]}, 
      {Hue[0., 1, 1], LineBox[{31, 32}]}, 
      {Hue[0., 1, 1], LineBox[{32, 33}]}, 
      {Hue[0., 1, 1], LineBox[{33, 34}]}, 
      {Hue[0., 1, 1], LineBox[{34, 35}]}, 
      {Hue[0., 1, 1], LineBox[{35, 36}]}, 
      {Hue[0., 1, 1], LineBox[{36, 37}]}, 
      {Hue[0., 1, 1], LineBox[{37, 38}]}, 
      {Hue[0., 1, 1], LineBox[{38, 39}]}, 
      {Hue[0., 1, 1], LineBox[{39, 40}]}, 
      {Hue[0., 1, 1], LineBox[{40, 41}]}, 
      {Hue[0., 1, 1], LineBox[{41, 42}]}, 
      {Hue[0., 1, 1], LineBox[{42, 43}]}, 
      {Hue[0., 1, 1], LineBox[{43, 44}]}, 
      {Hue[0., 1, 1], LineBox[{44, 45}]}, 
      {Hue[0., 1, 1], LineBox[{45, 46}]}, 
      {Hue[0., 1, 1], LineBox[{46, 47}]}, 
      {Hue[0., 1, 1], LineBox[{47, 48}]}, 
      {Hue[0., 1, 1], LineBox[{48, 49}]}, 
      {Hue[0., 1, 1], LineBox[{49, 50}]}, 
      {Hue[0., 1, 1], LineBox[{50, 51}]}, 
      {Hue[0., 1, 1], LineBox[{51, 52}]}, 
      {Hue[0., 1, 1], LineBox[{52, 53}]}, 
      {Hue[0., 1, 1], LineBox[{53, 54}]}, 
      {Hue[0., 1, 1], LineBox[{54, 55}]}, 
      {Hue[0., 1, 1], LineBox[{55, 56}]}, 
      {Hue[0., 1, 1], LineBox[{56, 57}]}, 
      {Hue[0., 1, 1], LineBox[{57, 58}]}, 
      {Hue[0., 1, 1], LineBox[{58, 59}]}, 
      {Hue[0., 1, 1], LineBox[{59, 60}]}, 
      {Hue[0., 1, 1], LineBox[{60, 61}]}, 
      {Hue[0., 1, 1], LineBox[{61, 62}]}, 
      {Hue[0., 1, 1], LineBox[{62, 63}]}, 
      {Hue[0., 1, 1], LineBox[{63, 64}]}, 
      {Hue[0., 1, 1], LineBox[{64, 65}]}, 
      {Hue[0., 1, 1], LineBox[{65, 66}]}, 
      {Hue[0., 1, 1], LineBox[{66, 67}]}, 
      {Hue[0., 1, 1], LineBox[{67, 68}]}, 
      {Hue[0., 1, 1], LineBox[{68, 69}]}, 
      {Hue[0., 1, 1], LineBox[{69, 70}]}, 
      {Hue[0., 1, 1], LineBox[{70, 71}]}, 
      {Hue[0., 1, 1], LineBox[{71, 72}]}, 
      {Hue[0., 1, 1], LineBox[{72, 73}]}, 
      {Hue[0., 1, 1], LineBox[{73, 74}]}, 
      {Hue[0., 1, 1], LineBox[{74, 75}]}, 
      {Hue[0., 1, 1], LineBox[{75, 76}]}, 
      {Hue[0., 1, 1], LineBox[{76, 77}]}, 
      {Hue[0., 1, 1], LineBox[{77, 78}]}, 
      {Hue[0., 1, 1], LineBox[{78, 79}]}, 
      {Hue[0., 1, 1], LineBox[{79, 80}]}, 
      {Hue[0., 1, 1], LineBox[{80, 81}]}, 
      {Hue[0., 1, 1], LineBox[{81, 82}]}, 
      {Hue[0., 1, 1], LineBox[{82, 83}]}, 
      {Hue[0., 1, 1], LineBox[{83, 84}]}, 
      {Hue[0., 1, 1], LineBox[{84, 85}]}, 
      {Hue[0., 1, 1], LineBox[{85, 86}]}, 
      {Hue[0., 1, 1], LineBox[{86, 87}]}, 
      {Hue[0., 1, 1], LineBox[{87, 88}]}, 
      {Hue[0., 1, 1], LineBox[{88, 89}]}, 
      {Hue[0., 1, 1], LineBox[{89, 90}]}, 
      {Hue[0., 1, 1], LineBox[{90, 91}]}, 
      {Hue[0., 1, 1], LineBox[{91, 92}]}, 
      {Hue[0., 1, 1], LineBox[{92, 93}]}, 
      {Hue[0., 1, 1], LineBox[{93, 94}]}, 
      {Hue[0., 1, 1], LineBox[{94, 95}]}, 
      {Hue[0., 1, 1], LineBox[{95, 96}]}, 
      {Hue[0., 1, 1], LineBox[{96, 97}]}, 
      {Hue[0., 1, 1], LineBox[{97, 98}]}, 
      {Hue[0., 1, 1], LineBox[{98, 99}]}, 
      {Hue[0., 1, 1], LineBox[{99, 100}]}, 
      {Hue[0., 1, 1], LineBox[{100, 101}]}, 
      {Hue[0., 1, 1], LineBox[{101, 102}]}, 
      {Hue[0., 1, 1], LineBox[{102, 103}]}, 
      {Hue[0., 1, 1], LineBox[{103, 104}]}, 
      {Hue[0., 1, 1], LineBox[{104, 105}]}, 
      {Hue[0., 1, 1], LineBox[{105, 106}]}, 
      {Hue[0., 1, 1], LineBox[{106, 107}]}, 
      {Hue[0., 1, 1], LineBox[{107, 108}]}, 
      {Hue[0., 1, 1], LineBox[{108, 109}]}, 
      {Hue[0., 1, 1], LineBox[{109, 110}]}, 
      {Hue[0., 1, 1], LineBox[{110, 111}]}, 
      {Hue[0., 1, 1], LineBox[{111, 112}]}, 
      {Hue[0., 1, 1], LineBox[{112, 113}]}, 
      {Hue[0., 1, 1], LineBox[{113, 114}]}, 
      {Hue[0.5, 1, 1], LineBox[{114, 115}]}, 
      {Hue[0.6058591852983447, 1, 1], LineBox[{115, 116}]}, 
      {Hue[0.6227331009286026, 1, 1], LineBox[{116, 117}]}, 
      {Hue[0.6396070296781868, 1, 1], LineBox[{117, 118}]}, 
      {Hue[0.656480876666192, 1, 1], LineBox[{118, 119}]}, 
      {Hue[0.6733547902917163, 1, 1], LineBox[{119, 120}]}, 
      {Hue[0.6902286993955555, 1, 1], LineBox[{120, 121}]}, 
      {Hue[0.7071025954512155, 1, 1], LineBox[{121, 122}]}, 
      {Hue[0.7239764961464545, 1, 1], LineBox[{122, 123}]}, 
      {Hue[0.7408504066523738, 1, 1], LineBox[{123, 124}]}, 
      {Hue[0.7577243013630727, 1, 1], LineBox[{124, 125}]}, 
      {Hue[0.774598206239555, 1, 1], LineBox[{125, 126}]}, 
      {Hue[0.7914721068391426, 1, 1], LineBox[{126, 127}]}, 
      {Hue[0.8083460062146112, 1, 1], LineBox[{127, 128}]}, 
      {Hue[0.8252199058922096, 1, 1], LineBox[{128, 129}]}, 
      {Hue[0.8420938081369235, 1, 1], LineBox[{129, 130}]}, 
      {Hue[0.8589677083674467, 1, 1], LineBox[{130, 131}]}, 
      {Hue[0.875841609788686, 1, 1], LineBox[{131, 132}]}, 
      {Hue[0.8927155107206101, 1, 1], LineBox[{132, 133}]}, 
      {Hue[0.9095894115722645, 1, 1], LineBox[{133, 134}]}, 
      {Hue[0.9264633121110101, 1, 1], LineBox[{134, 135}]}, 
      {Hue[0.9433372133570396, 1, 1], LineBox[{135, 136}]}, 
      {Hue[0.9602111141693217, 1, 1], LineBox[{136, 137}]}, 
      {Hue[0.977085015216567, 1, 1], LineBox[{137, 138}]}, 
      {Hue[0.9939589160497319, 1, 1], LineBox[{138, 139}]}, 
      {Hue[0.01083281711493061, 1, 1], LineBox[{139, 140}]}, 
      {Hue[0.027706717946513252`, 1, 1], LineBox[{140, 141}]}, 
      {Hue[0.04458061887268499, 1, 1], LineBox[{141, 142}]}, 
      {Hue[0.06145451978097727, 1, 1], LineBox[{142, 143}]}, 
      {Hue[0.07832842071614594, 1, 1], LineBox[{143, 144}]}, 
      {Hue[0.09520232160259494, 1, 1], LineBox[{144, 145}]}, 
      {Hue[0.11207622253581707`, 1, 1], LineBox[{145, 146}]}, 
      {Hue[0.12895012343739437`, 1, 1], LineBox[{146, 147}]}, 
      {Hue[0.14582402437012018`, 1, 1], LineBox[{147, 148}]}, 
      {Hue[0.1626979252814971, 1, 1], LineBox[{148, 149}]}, 
      {Hue[0.1795718261961194, 1, 1], LineBox[{149, 150}]}, 
      {Hue[0.19644572710819364`, 1, 1], LineBox[{150, 151}]}, 
      {Hue[0.2133196280245769, 1, 1], LineBox[{151, 152}]}, 
      {Hue[0.2301935289414818, 1, 1], LineBox[{152, 153}]}, 
      {Hue[0.24706742985462202`, 1, 1], LineBox[{153, 154}]}, 
      {Hue[0.2639413307673858, 1, 1], LineBox[{154, 155}]}, 
      {Hue[0.28081523168399963`, 1, 1], LineBox[{155, 156}]}, 
      {Hue[0.2976891325983772, 1, 1], LineBox[{156, 157}]}, 
      {Hue[0.31456303351318465`, 1, 1], LineBox[{157, 158}]}, 
      {Hue[0.3314369344284545, 1, 1], LineBox[{158, 159}]}, 
      {Hue[0.3483108353435908, 1, 1], LineBox[{159, 160}]}, 
      {Hue[0.3651847362579608, 1, 1], LineBox[{160, 161}]}, 
      {Hue[0.3820586371728739, 1, 1], LineBox[{161, 162}]}, 
      {Hue[0.39893253808768075`, 1, 1], LineBox[{162, 163}]}, 
      {Hue[0.4158064390027232, 1, 1], LineBox[{163, 164}]}, 
      {Hue[0.4326803399174645, 1, 1], LineBox[{164, 165}]}, 
      {Hue[0.4495542408324696, 1, 1], LineBox[{165, 166}]}, 
      {Hue[0.4664281417471684, 1, 1], LineBox[{166, 167}]}, 
      {Hue[0.4833020426621788, 1, 1], LineBox[{167, 168}]}, 
      {Hue[0.5001759435770288, 1, 1], LineBox[{168, 169}]}, 
      {Hue[0.5170498444919072, 1, 1], LineBox[{169, 170}]}, 
      {Hue[0.5339237454066981, 1, 1], LineBox[{170, 171}]}, 
      {Hue[0.55079764632164, 1, 1], LineBox[{171, 172}]}, 
      {Hue[0.5676715472364586, 1, 1], LineBox[{172, 173}]}, 
      {Hue[0.5845454481513088, 1, 1], LineBox[{173, 174}]}, 
      {Hue[0.601419349066175, 1, 1], LineBox[{174, 175}]}, 
      {Hue[0.6182932499810705, 1, 1], LineBox[{175, 176}]}, 
      {Hue[0.6351671508959293, 1, 1], LineBox[{176, 177}]}, 
      {Hue[0.6520410518107687, 1, 1], LineBox[{177, 178}]}, 
      {Hue[0.6689149527256437, 1, 1], LineBox[{178, 179}]}, 
      {Hue[0.6857888536405153, 1, 1], LineBox[{179, 180}]}, 
      {Hue[0.7026627545553767, 1, 1], LineBox[{180, 181}]}, 
      {Hue[0.7195366554702447, 1, 1], LineBox[{181, 182}]}, 
      {Hue[0.7364105563850998, 1, 1], LineBox[{182, 183}]}, 
      {Hue[0.753284457299976, 1, 1], LineBox[{183, 184}]}, 
      {Hue[0.7701583582148419, 1, 1], LineBox[{184, 185}]}, 
      {Hue[0.7870322591297062, 1, 1], LineBox[{185, 186}]}, 
      {Hue[0.803906160044568, 1, 1], LineBox[{186, 187}]}, 
      {Hue[0.8207800609594362, 1, 1], LineBox[{187, 188}]}, 
      {Hue[0.8376539618743016, 1, 1], LineBox[{188, 189}]}, 
      {Hue[0.8545278627891637, 1, 1], LineBox[{189, 190}]}, 
      {Hue[0.8714017637040312, 1, 1], LineBox[{190, 191}]}, 
      {Hue[0.8882756646188987, 1, 1], LineBox[{191, 192}]}, 
      {Hue[0.9051495655337625, 1, 1], LineBox[{192, 193}]}, 
      {Hue[0.9220234664486272, 1, 1], LineBox[{193, 194}]}, 
      {Hue[0.9388973673634933, 1, 1], LineBox[{194, 195}]}, 
      {Hue[0.9557712682783588, 1, 1], LineBox[{195, 196}]}, 
      {Hue[0.9726451691932227, 1, 1], LineBox[{196, 197}]}, 
      {Hue[0.9895190701080886, 1, 1], LineBox[{197, 198}]}, 
      {Hue[0.006392971022953186, 1, 1], LineBox[{198, 199}]}, 
      {Hue[0.023266871937819053`, 1, 1], LineBox[{199, 200}]}, 
      {Hue[0.040140772852684185`, 1, 1], LineBox[{200, 201}]}, 
      {Hue[0.057014673767549456`, 1, 1], LineBox[{201, 202}]}, 
      {Hue[0.07388857468241454, 1, 1], LineBox[{202, 203}]}, 
      {Hue[0.09076247559727997, 1, 1], LineBox[{203, 204}]}, 
      {Hue[0.10763637651214496`, 1, 1], LineBox[{204, 205}]}, 
      {Hue[0.12451027742701012`, 1, 1], LineBox[{205, 206}]}, 
      {Hue[0.1413841783418754, 1, 1], LineBox[{206, 207}]}, 
      {Hue[0.15825807925674074`, 1, 1], LineBox[{207, 208}]}, 
      {Hue[0.17513198017160578`, 1, 1], LineBox[{208, 209}]}, 
      {Hue[0.19200588108647093`, 1, 1], LineBox[{209, 210}]}, 
      {Hue[0.20887978200133625`, 1, 1], LineBox[{210, 211}]}, 
      {Hue[0.22575368291620135`, 1, 1], LineBox[{211, 212}]}, 
      {Hue[0.24262758383106653`, 1, 1], LineBox[{212, 213}]}, 
      {Hue[0.2595014847459317, 1, 1], LineBox[{213, 214}]}, 
      {Hue[0.2763753856607969, 1, 1], LineBox[{214, 215}]}, 
      {Hue[0.2932492865756621, 1, 1], LineBox[{215, 216}]}, 
      {Hue[0.31012318749052736`, 1, 1], LineBox[{216, 217}]}, 
      {Hue[0.3269970884053926, 1, 1], LineBox[{217, 218}]}, 
      {Hue[0.3438709893202577, 1, 1], LineBox[{218, 219}]}, 
      {Hue[0.3607448902351229, 1, 1], LineBox[{219, 220}]}, 
      {Hue[0.3776187911499881, 1, 1], LineBox[{220, 221}]}, 
      {Hue[0.3944926920648533, 1, 1], LineBox[{221, 222}]}, 
      {Hue[0.41136659297971845`, 1, 1], LineBox[{222, 223}]}, 
      {Hue[0.4282404938945837, 1, 1], LineBox[{223, 224}]}, 
      {Hue[0.44511439480944887`, 1, 1], LineBox[{224, 225}]}, 
      {Hue[0.46198829572431405`, 1, 1], LineBox[{225, 226}]}, 
      {Hue[0.4788621966391792, 1, 1], LineBox[{226, 227}]}, 
      {Hue[0.4957360975540444, 1, 1], LineBox[{227, 228}]}, 
      {Hue[0.5126099984689094, 1, 1], LineBox[{228, 229}]}, 
      {Hue[0.5294838993837747, 1, 1], LineBox[{229, 230}]}, 
      {Hue[0.54635780029864, 1, 1], LineBox[{230, 231}]}, 
      {Hue[0.5632317012135051, 1, 1], LineBox[{231, 232}]}, 
      {Hue[0.5801056021283704, 1, 1], LineBox[{232, 233}]}, 
      {Hue[0.5969795030432354, 1, 1], LineBox[{233, 234}]}, 
      {Hue[0.6138534039581007, 1, 1], LineBox[{234, 235}]}, 
      {Hue[0.6307273048729658, 1, 1], LineBox[{235, 236}]}, 
      {Hue[0.6476012057878311, 1, 1], LineBox[{236, 237}]}, 
      {Hue[0.6644751067026962, 1, 1], LineBox[{237, 238}]}, 
      {Hue[0.6813490076175615, 1, 1], LineBox[{238, 239}]}, 
      {Hue[0.6982229085324266, 1, 1], LineBox[{239, 240}]}, 
      {Hue[0.7150968094472918, 1, 1], LineBox[{240, 241}]}, 
      {Hue[0.731970710362157, 1, 1], LineBox[{241, 242}]}, 
      {Hue[0.7488446112770222, 1, 1], LineBox[{242, 243}]}, 
      {Hue[0.7657185121918874, 1, 1], LineBox[{243, 244}]}, 
      {Hue[0.7825924131067525, 1, 1], LineBox[{244, 245}]}, 
      {Hue[0.7994663140216177, 1, 1], LineBox[{245, 246}]}, 
      {Hue[0.8163402149364829, 1, 1], LineBox[{246, 247}]}, 
      {Hue[0.8332141158513481, 1, 1], LineBox[{247, 248}]}, 
      {Hue[0.8500880167662133, 1, 1], LineBox[{248, 249}]}, 
      {Hue[0.8669619176810786, 1, 1], LineBox[{249, 250}]}, 
      {Hue[0.8838358185959436, 1, 1], LineBox[{250, 251}]}, 
      {Hue[0.9007097195108089, 1, 1], LineBox[{251, 252}]}, 
      {Hue[0.9175836204256741, 1, 1], LineBox[{252, 253}]}, 
      {Hue[0.9344575213405393, 1, 1], LineBox[{253, 254}]}, 
      {Hue[0.9513314222554045, 1, 1], LineBox[{254, 255}]}, 
      {Hue[0.9682053231702696, 1, 1], LineBox[{255, 256}]}, 
      {Hue[0.9850792240851348, 1, 1], LineBox[{256, 257}]}, 
      {Hue[0.001953125000000001, 1, 1], LineBox[{257, 258}]}, 
      {Hue[0.018827025914865188`, 1, 1], LineBox[{258, 259}]}, 
      {Hue[0.03570092682973037, 1, 1], LineBox[{259, 260}]}, 
      {Hue[0.052574827744595556`, 1, 1], LineBox[{260, 261}]}, 
      {Hue[0.06944872865946075, 1, 1], LineBox[{261, 262}]}, 
      {Hue[0.08632262957432595, 1, 1], LineBox[{262, 263}]}, 
      {Hue[0.10319653048919113`, 1, 1], LineBox[{263, 264}]}, 
      {Hue[0.12007043140405632`, 1, 1], LineBox[{264, 265}]}, 
      {Hue[0.1369443323189215, 1, 1], LineBox[{265, 266}]}, 
      {Hue[0.15381823323378668`, 1, 1], LineBox[{266, 267}]}, 
      {Hue[0.1706921341486519, 1, 1], LineBox[{267, 268}]}, 
      {Hue[0.18756603506351707`, 1, 1], LineBox[{268, 269}]}, 
      {Hue[0.20443993597838225`, 1, 1], LineBox[{269, 270}]}, 
      {Hue[0.22131383689324746`, 1, 1], LineBox[{270, 271}]}, 
      {Hue[0.23818773780811264`, 1, 1], LineBox[{271, 272}]}, 
      {Hue[0.2550616387229778, 1, 1], LineBox[{272, 273}]}, 
      {Hue[0.271935539637843, 1, 1], LineBox[{273, 274}]}, 
      {Hue[0.2888094405527082, 1, 1], LineBox[{274, 275}]}, 
      {Hue[0.30568334146757337`, 1, 1], LineBox[{275, 276}]}, 
      {Hue[0.3225572423824386, 1, 1], LineBox[{276, 277}]}, 
      {Hue[0.3394311432973038, 1, 1], LineBox[{277, 278}]}, 
      {Hue[0.35630504421216896`, 1, 1], LineBox[{278, 279}]}, 
      {Hue[0.37317894512703415`, 1, 1], LineBox[{279, 280}]}, 
      {Hue[0.3900528460418993, 1, 1], LineBox[{280, 281}]}, 
      {Hue[0.4069267469567645, 1, 1], LineBox[{281, 282}]}, 
      {Hue[0.4238006478716297, 1, 1], LineBox[{282, 283}]}, 
      {Hue[0.44067454878649487`, 1, 1], LineBox[{283, 284}]}, 
      {Hue[0.45754844970136005`, 1, 1], LineBox[{284, 285}]}, 
      {Hue[0.47442235061622523`, 1, 1], LineBox[{285, 286}]}, 
      {Hue[0.49129625153109047`, 1, 1], LineBox[{286, 287}]}, 
      {Hue[0.5081701524459556, 1, 1], LineBox[{287, 288}]}, 
      {Hue[0.5250440533608207, 1, 1], LineBox[{288, 289}]}, 
      {Hue[0.541917954275686, 1, 1], LineBox[{289, 290}]}, 
      {Hue[0.5587918551905511, 1, 1], LineBox[{290, 291}]}, 
      {Hue[0.5756657561054164, 1, 1], LineBox[{291, 292}]}, 
      {Hue[0.5925396570202817, 1, 1], LineBox[{292, 293}]}, 
      {Hue[0.6094135579351467, 1, 1], LineBox[{293, 294}]}, 
      {Hue[0.6262874588500119, 1, 1], LineBox[{294, 295}]}, 
      {Hue[0.6431613597648771, 1, 1], LineBox[{295, 296}]}, 
      {Hue[0.6600352606797423, 1, 1], LineBox[{296, 297}]}, 
      {Hue[0.6769091615946075, 1, 1], LineBox[{297, 298}]}, 
      {Hue[0.6937830625094727, 1, 1], LineBox[{298, 299}]}, 
      {Hue[0.7106569634243378, 1, 1], LineBox[{299, 300}]}, 
      {Hue[0.727530864339203, 1, 1], LineBox[{300, 301}]}, 
      {Hue[0.7444047652540682, 1, 1], LineBox[{301, 302}]}, 
      {Hue[0.7612786661689335, 1, 1], LineBox[{302, 303}]}, 
      {Hue[0.7781525670837985, 1, 1], LineBox[{303, 304}]}, 
      {Hue[0.7950264679986638, 1, 1], LineBox[{304, 305}]}, 
      {Hue[0.8119003689135291, 1, 1], LineBox[{305, 306}]}, 
      {Hue[0.8287742698283941, 1, 1], LineBox[{306, 307}]}, 
      {Hue[0.8456481707432593, 1, 1], LineBox[{307, 308}]}, 
      {Hue[0.8625220716581248, 1, 1], LineBox[{308, 309}]}, 
      {Hue[0.8793959725729897, 1, 1], LineBox[{309, 310}]}, 
      {Hue[0.8962698734878554, 1, 1], LineBox[{310, 311}]}, 
      {Hue[0.9131437744027202, 1, 1], LineBox[{311, 312}]}, 
      {Hue[0.9300176753175851, 1, 1], LineBox[{312, 313}]}, 
      {Hue[0.9468915762324507, 1, 1], LineBox[{313, 314}]}, 
      {Hue[0.9637654771473162, 1, 1], LineBox[{314, 315}]}, 
      {Hue[0.9806393780621809, 1, 1], LineBox[{315, 316}]}, 
      {Hue[0.997513278977046, 1, 1], LineBox[{316, 317}]}, 
      {Hue[0.01438717989191182, 1, 1], LineBox[{317, 318}]}, 
      {Hue[0.03126108080677669, 1, 1], LineBox[{318, 319}]}, 
      {Hue[0.048134981721641074`, 1, 1], LineBox[{319, 320}]}, 
      {Hue[0.06500888263650717, 1, 1], LineBox[{320, 321}]}, 
      {Hue[0.08188278355137264, 1, 1], LineBox[{321, 322}]}, 
      {Hue[0.09875668446623698, 1, 1], LineBox[{322, 323}]}, 
      {Hue[0.11563058538110169`, 1, 1], LineBox[{323, 324}]}, 
      {Hue[0.13250448629596936`, 1, 1], LineBox[{324, 325}]}, 
      {Hue[0.14937838721083332`, 1, 1], LineBox[{325, 326}]}, 
      {Hue[0.16625228812570111`, 1, 1], LineBox[{326, 327}]}, 
      {Hue[0.18312618904056366`, 1, 1], LineBox[{327, 328}]}, 
      {Hue[0.20000008995542723`, 1, 1], LineBox[{328, 329}]}, 
      {Hue[0.21687399087029446`, 1, 1], LineBox[{329, 330}]}, 
      {Hue[0.23374789178516156`, 1, 1], LineBox[{330, 331}]}, 
      {Hue[0.2506217927000209, 1, 1], LineBox[{331, 332}]}, 
      {Hue[0.267495693614888, 1, 1], LineBox[{332, 333}]}, 
      {Hue[0.2843695945297534, 1, 1], LineBox[{333, 334}]}, 
      {Hue[0.3012434954446202, 1, 1], LineBox[{334, 335}]}, 
      {Hue[0.3181173963594849, 1, 1], LineBox[{335, 336}]}, 
      {Hue[0.33499129727434906`, 1, 1], LineBox[{336, 337}]}, 
      {Hue[0.3518651981892264, 1, 1], LineBox[{337, 338}]}, 
      {Hue[0.3687390991040822, 1, 1], LineBox[{338, 339}]}, 
      {Hue[0.3856130000189433, 1, 1], LineBox[{339, 340}]}, 
      {Hue[0.4024869009338191, 1, 1], LineBox[{340, 341}]}, 
      {Hue[0.41936080184869673`, 1, 1], LineBox[{341, 342}]}, 
      {Hue[0.43623470276358123`, 1, 1], LineBox[{342, 343}]}, 
      {Hue[0.4531086036784016, 1, 1], LineBox[{343, 344}]}, 
      {Hue[0.46998250459322793`, 1, 1], LineBox[{344, 345}]}, 
      {Hue[0.4868564055081243, 1, 1], LineBox[{345, 346}]}, 
      {Hue[0.5037303064230843, 1, 1], LineBox[{346, 347}]}, 
      {Hue[0.5206042073378334, 1, 1], LineBox[{347, 348}]}, 
      {Hue[0.5374781082527711, 1, 1], LineBox[{348, 349}]}, 
      {Hue[0.554352009167685, 1, 1], LineBox[{349, 350}]}, 
      {Hue[0.5712259100824209, 1, 1], LineBox[{350, 351}]}, 
      {Hue[0.5880998109971458, 1, 1], LineBox[{351, 352}]}, 
      {Hue[0.6049737119121952, 1, 1], LineBox[{352, 353}]}, 
      {Hue[0.6218476128271425, 1, 1], LineBox[{353, 354}]}, 
      {Hue[0.6387215137422474, 1, 1], LineBox[{354, 355}]}, 
      {Hue[0.655595414656956, 1, 1], LineBox[{355, 356}]}, 
      {Hue[0.6724693155719632, 1, 1], LineBox[{356, 357}]}, 
      {Hue[0.6893432164865616, 1, 1], LineBox[{357, 358}]}, 
      {Hue[0.7062171174015373, 1, 1], LineBox[{358, 359}]}, 
      {Hue[0.7230910183162487, 1, 1], LineBox[{359, 360}]}, 
      {Hue[0.7399649192304052, 1, 1], LineBox[{360, 361}]}, 
      {Hue[0.7568388201460909, 1, 1], LineBox[{361, 362}]}, 
      {Hue[0.7737127210634289, 1, 1], LineBox[{362, 363}]}, 
      {Hue[0.7905866219758477, 1, 1], LineBox[{363, 364}]}, 
      {Hue[0.8074605228909263, 1, 1], LineBox[{364, 365}]}, 
      {Hue[0.8243344238026495, 1, 1], LineBox[{365, 366}]}, 
      {Hue[0.8412083247197057, 1, 1], LineBox[{366, 367}]}, 
      {Hue[0.8580822256375282, 1, 1], LineBox[{367, 368}]}, 
      {Hue[0.874956126559485, 1, 1], LineBox[{368, 369}]}, 
      {Hue[0.8918300274636765, 1, 1], LineBox[{369, 370}]}, 
      {Hue[0.9087039283913463, 1, 1], LineBox[{370, 371}]}, 
      {Hue[0.9255778292982721, 1, 1], LineBox[{371, 372}]}, 
      {Hue[0.9424517302066693, 1, 1], LineBox[{372, 373}]}, 
      {Hue[0.9593256310977288, 1, 1], LineBox[{373, 374}]}, 
      {Hue[0.9761995320802711, 1, 1], LineBox[{374, 375}]}, 
      {Hue[0.9930734329972324, 1, 1], LineBox[{375, 376}]}, 
      {Hue[0.009947333879442764, 1, 1], LineBox[{376, 377}]}, 
      {Hue[0.0268212347484224, 1, 1], LineBox[{377, 378}]}, 
      {Hue[0.04369513568401856, 1, 1], LineBox[{378, 379}]}, 
      {Hue[0.06056903675627041, 1, 1], LineBox[{379, 380}]}, 
      {Hue[0.07744293768879483, 1, 1], LineBox[{380, 381}]}, 
      {Hue[0.09431683820258967, 1, 1], LineBox[{381, 382}]}, 
      {Hue[0.11119073889870218`, 1, 1], LineBox[{382, 383}]}, 
      {Hue[0.1280646399305784, 1, 1], LineBox[{383, 384}]}, 
      {Hue[0.14493854189656583`, 1, 1], LineBox[{384, 385}]}, 
      {Hue[0.16181244205002082`, 1, 1], LineBox[{385, 386}]}, 
      {Hue[0.17868634385560458`, 1, 1], LineBox[{386, 387}]}, 
      {Hue[0.1955602447326109, 1, 1], LineBox[{387, 388}]}, 
      {Hue[0.21243414538737151`, 1, 1], LineBox[{388, 389}]}, 
      {Hue[0.2293080462819191, 1, 1], LineBox[{389, 390}]}, 
      {Hue[0.25, 1, 1], LineBox[{390, 391}]}, 
      {Hue[0.26305583920610676`, 1, 1], LineBox[{391, 392}]}, 
      {Hue[0.27992975012920396`, 1, 1], LineBox[{392, 393}]}, 
      {Hue[0.2968036537992501, 1, 1], LineBox[{393, 394}]}, 
      {Hue[0.3136775686453699, 1, 1], LineBox[{394, 395}]}, 
      {Hue[0.330551450707842, 1, 1], LineBox[{395, 396}]}, 
      {Hue[0.34742535537280034`, 1, 1], LineBox[{396, 397}]}, 
      {Hue[0.3642992132658118, 1, 1], LineBox[{397, 398}]}, 
      {Hue[0.3811731274121047, 1, 1], LineBox[{398, 399}]}, 
      {Hue[0.39804707607115386`, 1, 1], LineBox[{399, 400}]}, 
      {Hue[0.5, 1, 1], LineBox[{400, 401}]}, 
      {Hue[0., 1, 1], LineBox[{401, 402}]}, 
      {Hue[0., 1, 1], LineBox[{402, 403}]}, 
      {Hue[0., 1, 1], LineBox[{403, 404}]}, 
      {Hue[0., 1, 1], LineBox[{404, 405}]}, 
      {Hue[0., 1, 1], LineBox[{405, 406}]}, 
      {Hue[0., 1, 1], LineBox[{406, 407}]}, 
      {Hue[0., 1, 1], LineBox[{407, 408}]}, 
      {Hue[0., 1, 1], LineBox[{408, 409}]}, 
      {Hue[0., 1, 1], LineBox[{409, 410}]}, 
      {Hue[0., 1, 1], LineBox[{410, 411}]}, 
      {Hue[0., 1, 1], LineBox[{411, 412}]}, 
      {Hue[0., 1, 1], LineBox[{412, 413}]}, 
      {Hue[0., 1, 1], LineBox[{413, 414}]}, 
      {Hue[0., 1, 1], LineBox[{414, 415}]}, 
      {Hue[0., 1, 1], LineBox[{415, 416}]}, 
      {Hue[0., 1, 1], LineBox[{416, 417}]}, 
      {Hue[0., 1, 1], LineBox[{417, 418}]}, 
      {Hue[0., 1, 1], LineBox[{418, 419}]}, 
      {Hue[0., 1, 1], LineBox[{419, 420}]}, 
      {Hue[0., 1, 1], LineBox[{420, 421}]}, 
      {Hue[0., 1, 1], LineBox[{421, 422}]}, 
      {Hue[0., 1, 1], LineBox[{422, 423}]}, 
      {Hue[0., 1, 1], LineBox[{423, 424}]}, 
      {Hue[0., 1, 1], LineBox[{424, 425}]}, 
      {Hue[0., 1, 1], LineBox[{425, 426}]}, 
      {Hue[0., 1, 1], LineBox[{426, 427}]}, 
      {Hue[0., 1, 1], LineBox[{427, 428}]}, 
      {Hue[0., 1, 1], LineBox[{428, 429}]}, 
      {Hue[0., 1, 1], LineBox[{429, 430}]}, 
      {Hue[0., 1, 1], LineBox[{430, 431}]}, 
      {Hue[0., 1, 1], LineBox[{431, 432}]}, 
      {Hue[0., 1, 1], LineBox[{432, 433}]}, 
      {Hue[0., 1, 1], LineBox[{433, 434}]}, 
      {Hue[0., 1, 1], LineBox[{434, 435}]}, 
      {Hue[0., 1, 1], LineBox[{435, 436}]}, 
      {Hue[0., 1, 1], LineBox[{436, 437}]}, 
      {Hue[0., 1, 1], LineBox[{437, 438}]}, 
      {Hue[0., 1, 1], LineBox[{438, 439}]}, 
      {Hue[0., 1, 1], LineBox[{439, 440}]}, 
      {Hue[0., 1, 1], LineBox[{440, 441}]}, 
      {Hue[0., 1, 1], LineBox[{441, 442}]}, 
      {Hue[0., 1, 1], LineBox[{442, 443}]}, 
      {Hue[0., 1, 1], LineBox[{443, 444}]}, 
      {Hue[0., 1, 1], LineBox[{444, 445}]}, 
      {Hue[0., 1, 1], LineBox[{445, 446}]}, 
      {Hue[0., 1, 1], LineBox[{446, 447}]}, 
      {Hue[0., 1, 1], LineBox[{447, 448}]}, 
      {Hue[0., 1, 1], LineBox[{448, 449}]}, 
      {Hue[0., 1, 1], LineBox[{449, 450}]}, 
      {Hue[0., 1, 1], LineBox[{450, 451}]}, 
      {Hue[0., 1, 1], LineBox[{451, 452}]}, 
      {Hue[0., 1, 1], LineBox[{452, 453}]}, 
      {Hue[0., 1, 1], LineBox[{453, 454}]}, 
      {Hue[0., 1, 1], LineBox[{454, 455}]}, 
      {Hue[0., 1, 1], LineBox[{455, 456}]}, 
      {Hue[0., 1, 1], LineBox[{456, 457}]}, 
      {Hue[0., 1, 1], LineBox[{457, 458}]}, 
      {Hue[0., 1, 1], LineBox[{458, 459}]}, 
      {Hue[0., 1, 1], LineBox[{459, 460}]}, 
      {Hue[0., 1, 1], LineBox[{460, 461}]}, 
      {Hue[0., 1, 1], LineBox[{461, 462}]}, 
      {Hue[0., 1, 1], LineBox[{462, 463}]}, 
      {Hue[0., 1, 1], LineBox[{463, 464}]}, 
      {Hue[0., 1, 1], LineBox[{464, 465}]}, 
      {Hue[0., 1, 1], LineBox[{465, 466}]}, 
      {Hue[0., 1, 1], LineBox[{466, 467}]}, 
      {Hue[0., 1, 1], LineBox[{467, 468}]}, 
      {Hue[0., 1, 1], LineBox[{468, 469}]}, 
      {Hue[0., 1, 1], LineBox[{469, 470}]}, 
      {Hue[0., 1, 1], LineBox[{470, 471}]}, 
      {Hue[0., 1, 1], LineBox[{471, 472}]}, 
      {Hue[0., 1, 1], LineBox[{472, 473}]}, 
      {Hue[0., 1, 1], LineBox[{473, 474}]}, 
      {Hue[0., 1, 1], LineBox[{474, 475}]}, 
      {Hue[0., 1, 1], LineBox[{475, 476}]}, 
      {Hue[0., 1, 1], LineBox[{476, 477}]}, 
      {Hue[0., 1, 1], LineBox[{477, 478}]}, 
      {Hue[0., 1, 1], LineBox[{478, 479}]}, 
      {Hue[0., 1, 1], LineBox[{479, 480}]}, 
      {Hue[0., 1, 1], LineBox[{480, 481}]}, 
      {Hue[0., 1, 1], LineBox[{481, 482}]}, 
      {Hue[0., 1, 1], LineBox[{482, 483}]}, 
      {Hue[0., 1, 1], LineBox[{483, 484}]}, 
      {Hue[0., 1, 1], LineBox[{484, 485}]}, 
      {Hue[0., 1, 1], LineBox[{485, 486}]}, 
      {Hue[0., 1, 1], LineBox[{486, 487}]}, 
      {Hue[0., 1, 1], LineBox[{487, 488}]}, 
      {Hue[0., 1, 1], LineBox[{488, 489}]}, 
      {Hue[0., 1, 1], LineBox[{489, 490}]}, 
      {Hue[0., 1, 1], LineBox[{490, 491}]}, 
      {Hue[0., 1, 1], LineBox[{491, 492}]}, 
      {Hue[0., 1, 1], LineBox[{492, 493}]}, 
      {Hue[0., 1, 1], LineBox[{493, 494}]}, 
      {Hue[0., 1, 1], LineBox[{494, 495}]}, 
      {Hue[0., 1, 1], LineBox[{495, 496}]}, 
      {Hue[0., 1, 1], LineBox[{496, 497}]}, 
      {Hue[0., 1, 1], LineBox[{497, 498}]}, 
      {Hue[0., 1, 1], LineBox[{498, 499}]}, 
      {Hue[0., 1, 1], LineBox[{499, 500}]}, 
      {Hue[0., 1, 1], LineBox[{500, 501}]}, 
      {Hue[0., 1, 1], LineBox[{501, 502}]}, 
      {Hue[0., 1, 1], LineBox[{502, 503}]}, 
      {Hue[0., 1, 1], LineBox[{503, 504}]}, 
      {Hue[0., 1, 1], LineBox[{504, 505}]}, 
      {Hue[0., 1, 1], LineBox[{505, 506}]}, 
      {Hue[0., 1, 1], LineBox[{506, 507}]}, 
      {Hue[0., 1, 1], LineBox[{507, 508}]}, 
      {Hue[0., 1, 1], LineBox[{508, 509}]}, 
      {Hue[0., 1, 1], LineBox[{509, 510}]}, 
      {Hue[0., 1, 1], LineBox[{510, 511}]}, 
      {Hue[0., 1, 1], LineBox[{511, 512}]}}}},
   VertexColors->{
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.5, 1, 1], 
     Hue[0.6058591852983447, 1, 1], 
     Hue[0.6227331009286026, 1, 1], 
     Hue[0.6396070296781868, 1, 1], 
     Hue[0.656480876666192, 1, 1], 
     Hue[0.6733547902917163, 1, 1], 
     Hue[0.6902286993955555, 1, 1], 
     Hue[0.7071025954512155, 1, 1], 
     Hue[0.7239764961464545, 1, 1], 
     Hue[0.7408504066523738, 1, 1], 
     Hue[0.7577243013630727, 1, 1], 
     Hue[0.774598206239555, 1, 1], 
     Hue[0.7914721068391426, 1, 1], 
     Hue[0.8083460062146112, 1, 1], 
     Hue[0.8252199058922096, 1, 1], 
     Hue[0.8420938081369235, 1, 1], 
     Hue[0.8589677083674467, 1, 1], 
     Hue[0.875841609788686, 1, 1], 
     Hue[0.8927155107206101, 1, 1], 
     Hue[0.9095894115722645, 1, 1], 
     Hue[0.9264633121110101, 1, 1], 
     Hue[0.9433372133570396, 1, 1], 
     Hue[0.9602111141693217, 1, 1], 
     Hue[0.977085015216567, 1, 1], 
     Hue[0.9939589160497319, 1, 1], 
     Hue[0.01083281711493061, 1, 1], 
     Hue[0.027706717946513252`, 1, 1], 
     Hue[0.04458061887268499, 1, 1], 
     Hue[0.06145451978097727, 1, 1], 
     Hue[0.07832842071614594, 1, 1], 
     Hue[0.09520232160259494, 1, 1], 
     Hue[0.11207622253581707`, 1, 1], 
     Hue[0.12895012343739437`, 1, 1], 
     Hue[0.14582402437012018`, 1, 1], 
     Hue[0.1626979252814971, 1, 1], 
     Hue[0.1795718261961194, 1, 1], 
     Hue[0.19644572710819364`, 1, 1], 
     Hue[0.2133196280245769, 1, 1], 
     Hue[0.2301935289414818, 1, 1], 
     Hue[0.24706742985462202`, 1, 1], 
     Hue[0.2639413307673858, 1, 1], 
     Hue[0.28081523168399963`, 1, 1], 
     Hue[0.2976891325983772, 1, 1], 
     Hue[0.31456303351318465`, 1, 1], 
     Hue[0.3314369344284545, 1, 1], 
     Hue[0.3483108353435908, 1, 1], 
     Hue[0.3651847362579608, 1, 1], 
     Hue[0.3820586371728739, 1, 1], 
     Hue[0.39893253808768075`, 1, 1], 
     Hue[0.4158064390027232, 1, 1], 
     Hue[0.4326803399174645, 1, 1], 
     Hue[0.4495542408324696, 1, 1], 
     Hue[0.4664281417471684, 1, 1], 
     Hue[0.4833020426621788, 1, 1], 
     Hue[0.5001759435770288, 1, 1], 
     Hue[0.5170498444919072, 1, 1], 
     Hue[0.5339237454066981, 1, 1], 
     Hue[0.55079764632164, 1, 1], 
     Hue[0.5676715472364586, 1, 1], 
     Hue[0.5845454481513088, 1, 1], 
     Hue[0.601419349066175, 1, 1], 
     Hue[0.6182932499810705, 1, 1], 
     Hue[0.6351671508959293, 1, 1], 
     Hue[0.6520410518107687, 1, 1], 
     Hue[0.6689149527256437, 1, 1], 
     Hue[0.6857888536405153, 1, 1], 
     Hue[0.7026627545553767, 1, 1], 
     Hue[0.7195366554702447, 1, 1], 
     Hue[0.7364105563850998, 1, 1], 
     Hue[0.753284457299976, 1, 1], 
     Hue[0.7701583582148419, 1, 1], 
     Hue[0.7870322591297062, 1, 1], 
     Hue[0.803906160044568, 1, 1], 
     Hue[0.8207800609594362, 1, 1], 
     Hue[0.8376539618743016, 1, 1], 
     Hue[0.8545278627891637, 1, 1], 
     Hue[0.8714017637040312, 1, 1], 
     Hue[0.8882756646188987, 1, 1], 
     Hue[0.9051495655337625, 1, 1], 
     Hue[0.9220234664486272, 1, 1], 
     Hue[0.9388973673634933, 1, 1], 
     Hue[0.9557712682783588, 1, 1], 
     Hue[0.9726451691932227, 1, 1], 
     Hue[0.9895190701080886, 1, 1], 
     Hue[0.006392971022953186, 1, 1], 
     Hue[0.023266871937819053`, 1, 1], 
     Hue[0.040140772852684185`, 1, 1], 
     Hue[0.057014673767549456`, 1, 1], 
     Hue[0.07388857468241454, 1, 1], 
     Hue[0.09076247559727997, 1, 1], 
     Hue[0.10763637651214496`, 1, 1], 
     Hue[0.12451027742701012`, 1, 1], 
     Hue[0.1413841783418754, 1, 1], 
     Hue[0.15825807925674074`, 1, 1], 
     Hue[0.17513198017160578`, 1, 1], 
     Hue[0.19200588108647093`, 1, 1], 
     Hue[0.20887978200133625`, 1, 1], 
     Hue[0.22575368291620135`, 1, 1], 
     Hue[0.24262758383106653`, 1, 1], 
     Hue[0.2595014847459317, 1, 1], 
     Hue[0.2763753856607969, 1, 1], 
     Hue[0.2932492865756621, 1, 1], 
     Hue[0.31012318749052736`, 1, 1], 
     Hue[0.3269970884053926, 1, 1], 
     Hue[0.3438709893202577, 1, 1], 
     Hue[0.3607448902351229, 1, 1], 
     Hue[0.3776187911499881, 1, 1], 
     Hue[0.3944926920648533, 1, 1], 
     Hue[0.41136659297971845`, 1, 1], 
     Hue[0.4282404938945837, 1, 1], 
     Hue[0.44511439480944887`, 1, 1], 
     Hue[0.46198829572431405`, 1, 1], 
     Hue[0.4788621966391792, 1, 1], 
     Hue[0.4957360975540444, 1, 1], 
     Hue[0.5126099984689094, 1, 1], 
     Hue[0.5294838993837747, 1, 1], 
     Hue[0.54635780029864, 1, 1], 
     Hue[0.5632317012135051, 1, 1], 
     Hue[0.5801056021283704, 1, 1], 
     Hue[0.5969795030432354, 1, 1], 
     Hue[0.6138534039581007, 1, 1], 
     Hue[0.6307273048729658, 1, 1], 
     Hue[0.6476012057878311, 1, 1], 
     Hue[0.6644751067026962, 1, 1], 
     Hue[0.6813490076175615, 1, 1], 
     Hue[0.6982229085324266, 1, 1], 
     Hue[0.7150968094472918, 1, 1], 
     Hue[0.731970710362157, 1, 1], 
     Hue[0.7488446112770222, 1, 1], 
     Hue[0.7657185121918874, 1, 1], 
     Hue[0.7825924131067525, 1, 1], 
     Hue[0.7994663140216177, 1, 1], 
     Hue[0.8163402149364829, 1, 1], 
     Hue[0.8332141158513481, 1, 1], 
     Hue[0.8500880167662133, 1, 1], 
     Hue[0.8669619176810786, 1, 1], 
     Hue[0.8838358185959436, 1, 1], 
     Hue[0.9007097195108089, 1, 1], 
     Hue[0.9175836204256741, 1, 1], 
     Hue[0.9344575213405393, 1, 1], 
     Hue[0.9513314222554045, 1, 1], 
     Hue[0.9682053231702696, 1, 1], 
     Hue[0.9850792240851348, 1, 1], 
     Hue[0.001953125000000001, 1, 1], 
     Hue[0.018827025914865188`, 1, 1], 
     Hue[0.03570092682973037, 1, 1], 
     Hue[0.052574827744595556`, 1, 1], 
     Hue[0.06944872865946075, 1, 1], 
     Hue[0.08632262957432595, 1, 1], 
     Hue[0.10319653048919113`, 1, 1], 
     Hue[0.12007043140405632`, 1, 1], 
     Hue[0.1369443323189215, 1, 1], 
     Hue[0.15381823323378668`, 1, 1], 
     Hue[0.1706921341486519, 1, 1], 
     Hue[0.18756603506351707`, 1, 1], 
     Hue[0.20443993597838225`, 1, 1], 
     Hue[0.22131383689324746`, 1, 1], 
     Hue[0.23818773780811264`, 1, 1], 
     Hue[0.2550616387229778, 1, 1], 
     Hue[0.271935539637843, 1, 1], 
     Hue[0.2888094405527082, 1, 1], 
     Hue[0.30568334146757337`, 1, 1], 
     Hue[0.3225572423824386, 1, 1], 
     Hue[0.3394311432973038, 1, 1], 
     Hue[0.35630504421216896`, 1, 1], 
     Hue[0.37317894512703415`, 1, 1], 
     Hue[0.3900528460418993, 1, 1], 
     Hue[0.4069267469567645, 1, 1], 
     Hue[0.4238006478716297, 1, 1], 
     Hue[0.44067454878649487`, 1, 1], 
     Hue[0.45754844970136005`, 1, 1], 
     Hue[0.47442235061622523`, 1, 1], 
     Hue[0.49129625153109047`, 1, 1], 
     Hue[0.5081701524459556, 1, 1], 
     Hue[0.5250440533608207, 1, 1], 
     Hue[0.541917954275686, 1, 1], 
     Hue[0.5587918551905511, 1, 1], 
     Hue[0.5756657561054164, 1, 1], 
     Hue[0.5925396570202817, 1, 1], 
     Hue[0.6094135579351467, 1, 1], 
     Hue[0.6262874588500119, 1, 1], 
     Hue[0.6431613597648771, 1, 1], 
     Hue[0.6600352606797423, 1, 1], 
     Hue[0.6769091615946075, 1, 1], 
     Hue[0.6937830625094727, 1, 1], 
     Hue[0.7106569634243378, 1, 1], 
     Hue[0.727530864339203, 1, 1], 
     Hue[0.7444047652540682, 1, 1], 
     Hue[0.7612786661689335, 1, 1], 
     Hue[0.7781525670837985, 1, 1], 
     Hue[0.7950264679986638, 1, 1], 
     Hue[0.8119003689135291, 1, 1], 
     Hue[0.8287742698283941, 1, 1], 
     Hue[0.8456481707432593, 1, 1], 
     Hue[0.8625220716581248, 1, 1], 
     Hue[0.8793959725729897, 1, 1], 
     Hue[0.8962698734878554, 1, 1], 
     Hue[0.9131437744027202, 1, 1], 
     Hue[0.9300176753175851, 1, 1], 
     Hue[0.9468915762324507, 1, 1], 
     Hue[0.9637654771473162, 1, 1], 
     Hue[0.9806393780621809, 1, 1], 
     Hue[0.997513278977046, 1, 1], 
     Hue[0.01438717989191182, 1, 1], 
     Hue[0.03126108080677669, 1, 1], 
     Hue[0.048134981721641074`, 1, 1], 
     Hue[0.06500888263650717, 1, 1], 
     Hue[0.08188278355137264, 1, 1], 
     Hue[0.09875668446623698, 1, 1], 
     Hue[0.11563058538110169`, 1, 1], 
     Hue[0.13250448629596936`, 1, 1], 
     Hue[0.14937838721083332`, 1, 1], 
     Hue[0.16625228812570111`, 1, 1], 
     Hue[0.18312618904056366`, 1, 1], 
     Hue[0.20000008995542723`, 1, 1], 
     Hue[0.21687399087029446`, 1, 1], 
     Hue[0.23374789178516156`, 1, 1], 
     Hue[0.2506217927000209, 1, 1], 
     Hue[0.267495693614888, 1, 1], 
     Hue[0.2843695945297534, 1, 1], 
     Hue[0.3012434954446202, 1, 1], 
     Hue[0.3181173963594849, 1, 1], 
     Hue[0.33499129727434906`, 1, 1], 
     Hue[0.3518651981892264, 1, 1], 
     Hue[0.3687390991040822, 1, 1], 
     Hue[0.3856130000189433, 1, 1], 
     Hue[0.4024869009338191, 1, 1], 
     Hue[0.41936080184869673`, 1, 1], 
     Hue[0.43623470276358123`, 1, 1], 
     Hue[0.4531086036784016, 1, 1], 
     Hue[0.46998250459322793`, 1, 1], 
     Hue[0.4868564055081243, 1, 1], 
     Hue[0.5037303064230843, 1, 1], 
     Hue[0.5206042073378334, 1, 1], 
     Hue[0.5374781082527711, 1, 1], 
     Hue[0.554352009167685, 1, 1], 
     Hue[0.5712259100824209, 1, 1], 
     Hue[0.5880998109971458, 1, 1], 
     Hue[0.6049737119121952, 1, 1], 
     Hue[0.6218476128271425, 1, 1], 
     Hue[0.6387215137422474, 1, 1], 
     Hue[0.655595414656956, 1, 1], 
     Hue[0.6724693155719632, 1, 1], 
     Hue[0.6893432164865616, 1, 1], 
     Hue[0.7062171174015373, 1, 1], 
     Hue[0.7230910183162487, 1, 1], 
     Hue[0.7399649192304052, 1, 1], 
     Hue[0.7568388201460909, 1, 1], 
     Hue[0.7737127210634289, 1, 1], 
     Hue[0.7905866219758477, 1, 1], 
     Hue[0.8074605228909263, 1, 1], 
     Hue[0.8243344238026495, 1, 1], 
     Hue[0.8412083247197057, 1, 1], 
     Hue[0.8580822256375282, 1, 1], 
     Hue[0.874956126559485, 1, 1], 
     Hue[0.8918300274636765, 1, 1], 
     Hue[0.9087039283913463, 1, 1], 
     Hue[0.9255778292982721, 1, 1], 
     Hue[0.9424517302066693, 1, 1], 
     Hue[0.9593256310977288, 1, 1], 
     Hue[0.9761995320802711, 1, 1], 
     Hue[0.9930734329972324, 1, 1], 
     Hue[0.009947333879442764, 1, 1], 
     Hue[0.0268212347484224, 1, 1], 
     Hue[0.04369513568401856, 1, 1], 
     Hue[0.06056903675627041, 1, 1], 
     Hue[0.07744293768879483, 1, 1], 
     Hue[0.09431683820258967, 1, 1], 
     Hue[0.11119073889870218`, 1, 1], 
     Hue[0.1280646399305784, 1, 1], 
     Hue[0.14493854189656583`, 1, 1], 
     Hue[0.16181244205002082`, 1, 1], 
     Hue[0.17868634385560458`, 1, 1], 
     Hue[0.1955602447326109, 1, 1], 
     Hue[0.21243414538737151`, 1, 1], 
     Hue[0.2293080462819191, 1, 1], 
     Hue[0.25, 1, 1], 
     Hue[0.26305583920610676`, 1, 1], 
     Hue[0.27992975012920396`, 1, 1], 
     Hue[0.2968036537992501, 1, 1], 
     Hue[0.3136775686453699, 1, 1], 
     Hue[0.330551450707842, 1, 1], 
     Hue[0.34742535537280034`, 1, 1], 
     Hue[0.3642992132658118, 1, 1], 
     Hue[0.3811731274121047, 1, 1], 
     Hue[0.39804707607115386`, 1, 1], 
     Hue[0.5, 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0., 1, 1], 
     Hue[0.5, 1, 1], 
     Hue[0.5, 1, 1], 
     Hue[0.6058591852983447, 1, 1], 
     Hue[0.6058591852983447, 1, 1], 
     Hue[0.6227331009286026, 1, 1], 
     Hue[0.6227331009286026, 1, 1], 
     Hue[0.6396070296781868, 1, 1], 
     Hue[0.6396070296781868, 1, 1], 
     Hue[0.656480876666192, 1, 1], 
     Hue[0.656480876666192, 1, 1], 
     Hue[0.6733547902917163, 1, 1], 
     Hue[0.6733547902917163, 1, 1], 
     Hue[0.6902286993955555, 1, 1], 
     Hue[0.6902286993955555, 1, 1], 
     Hue[0.7071025954512155, 1, 1], 
     Hue[0.7071025954512155, 1, 1], 
     Hue[0.7239764961464545, 1, 1], 
     Hue[0.7239764961464545, 1, 1], 
     Hue[0.7408504066523738, 1, 1], 
     Hue[0.7408504066523738, 1, 1], 
     Hue[0.7577243013630727, 1, 1], 
     Hue[0.7577243013630727, 1, 1], 
     Hue[0.774598206239555, 1, 1], 
     Hue[0.774598206239555, 1, 1], 
     Hue[0.7914721068391426, 1, 1], 
     Hue[0.7914721068391426, 1, 1], 
     Hue[0.8083460062146112, 1, 1], 
     Hue[0.8083460062146112, 1, 1], 
     Hue[0.8252199058922096, 1, 1], 
     Hue[0.8252199058922096, 1, 1], 
     Hue[0.8420938081369235, 1, 1], 
     Hue[0.8420938081369235, 1, 1], 
     Hue[0.8589677083674467, 1, 1], 
     Hue[0.8589677083674467, 1, 1], 
     Hue[0.875841609788686, 1, 1], 
     Hue[0.875841609788686, 1, 1], 
     Hue[0.8927155107206101, 1, 1], 
     Hue[0.8927155107206101, 1, 1], 
     Hue[0.9095894115722645, 1, 1], 
     Hue[0.9095894115722645, 1, 1], 
     Hue[0.9264633121110101, 1, 1], 
     Hue[0.9264633121110101, 1, 1], 
     Hue[0.9433372133570396, 1, 1], 
     Hue[0.9433372133570396, 1, 1], 
     Hue[0.9602111141693217, 1, 1], 
     Hue[0.9602111141693217, 1, 1], 
     Hue[0.977085015216567, 1, 1], 
     Hue[0.977085015216567, 1, 1], 
     Hue[0.9939589160497319, 1, 1], 
     Hue[0.9939589160497319, 1, 1], 
     Hue[0.01083281711493061, 1, 1], 
     Hue[0.01083281711493061, 1, 1], 
     Hue[0.027706717946513252`, 1, 1], 
     Hue[0.027706717946513252`, 1, 1], 
     Hue[0.04458061887268499, 1, 1], 
     Hue[0.04458061887268499, 1, 1], 
     Hue[0.06145451978097727, 1, 1], 
     Hue[0.06145451978097727, 1, 1], 
     Hue[0.07832842071614594, 1, 1], 
     Hue[0.07832842071614594, 1, 1], 
     Hue[0.09520232160259494, 1, 1], 
     Hue[0.09520232160259494, 1, 1], 
     Hue[0.11207622253581707`, 1, 1], 
     Hue[0.11207622253581707`, 1, 1], 
     Hue[0.12895012343739437`, 1, 1], 
     Hue[0.12895012343739437`, 1, 1], 
     Hue[0.14582402437012018`, 1, 1], 
     Hue[0.14582402437012018`, 1, 1], 
     Hue[0.1626979252814971, 1, 1], 
     Hue[0.1626979252814971, 1, 1], 
     Hue[0.1795718261961194, 1, 1], 
     Hue[0.1795718261961194, 1, 1], 
     Hue[0.19644572710819364`, 1, 1], 
     Hue[0.19644572710819364`, 1, 1], 
     Hue[0.2133196280245769, 1, 1], 
     Hue[0.2133196280245769, 1, 1], 
     Hue[0.2301935289414818, 1, 1], 
     Hue[0.2301935289414818, 1, 1], 
     Hue[0.24706742985462202`, 1, 1], 
     Hue[0.24706742985462202`, 1, 1], 
     Hue[0.2639413307673858, 1, 1], 
     Hue[0.2639413307673858, 1, 1], 
     Hue[0.28081523168399963`, 1, 1], 
     Hue[0.28081523168399963`, 1, 1], 
     Hue[0.2976891325983772, 1, 1], 
     Hue[0.2976891325983772, 1, 1], 
     Hue[0.31456303351318465`, 1, 1], 
     Hue[0.31456303351318465`, 1, 1], 
     Hue[0.3314369344284545, 1, 1], 
     Hue[0.3314369344284545, 1, 1], 
     Hue[0.3483108353435908, 1, 1], 
     Hue[0.3483108353435908, 1, 1], 
     Hue[0.3651847362579608, 1, 1], 
     Hue[0.3651847362579608, 1, 1], 
     Hue[0.3820586371728739, 1, 1], 
     Hue[0.3820586371728739, 1, 1], 
     Hue[0.39893253808768075`, 1, 1], 
     Hue[0.39893253808768075`, 1, 1], 
     Hue[0.4158064390027232, 1, 1], 
     Hue[0.4158064390027232, 1, 1], 
     Hue[0.4326803399174645, 1, 1], 
     Hue[0.4326803399174645, 1, 1], 
     Hue[0.4495542408324696, 1, 1], 
     Hue[0.4495542408324696, 1, 1], 
     Hue[0.4664281417471684, 1, 1], 
     Hue[0.4664281417471684, 1, 1], 
     Hue[0.4833020426621788, 1, 1], 
     Hue[0.4833020426621788, 1, 1], 
     Hue[0.5001759435770288, 1, 1], 
     Hue[0.5001759435770288, 1, 1], 
     Hue[0.5170498444919072, 1, 1], 
     Hue[0.5170498444919072, 1, 1], 
     Hue[0.5339237454066981, 1, 1], 
     Hue[0.5339237454066981, 1, 1], 
     Hue[0.55079764632164, 1, 1], 
     Hue[0.55079764632164, 1, 1], 
     Hue[0.5676715472364586, 1, 1], 
     Hue[0.5676715472364586, 1, 1], 
     Hue[0.5845454481513088, 1, 1], 
     Hue[0.5845454481513088, 1, 1], 
     Hue[0.601419349066175, 1, 1], 
     Hue[0.601419349066175, 1, 1], 
     Hue[0.6182932499810705, 1, 1], 
     Hue[0.6182932499810705, 1, 1], 
     Hue[0.6351671508959293, 1, 1], 
     Hue[0.6351671508959293, 1, 1], 
     Hue[0.6520410518107687, 1, 1], 
     Hue[0.6520410518107687, 1, 1], 
     Hue[0.6689149527256437, 1, 1], 
     Hue[0.6689149527256437, 1, 1], 
     Hue[0.6857888536405153, 1, 1], 
     Hue[0.6857888536405153, 1, 1], 
     Hue[0.7026627545553767, 1, 1], 
     Hue[0.7026627545553767, 1, 1], 
     Hue[0.7195366554702447, 1, 1], 
     Hue[0.7195366554702447, 1, 1], 
     Hue[0.7364105563850998, 1, 1], 
     Hue[0.7364105563850998, 1, 1], 
     Hue[0.753284457299976, 1, 1], 
     Hue[0.753284457299976, 1, 1], 
     Hue[0.7701583582148419, 1, 1], 
     Hue[0.7701583582148419, 1, 1], 
     Hue[0.7870322591297062, 1, 1], 
     Hue[0.7870322591297062, 1, 1], 
     Hue[0.803906160044568, 1, 1], 
     Hue[0.803906160044568, 1, 1], 
     Hue[0.8207800609594362, 1, 1], 
     Hue[0.8207800609594362, 1, 1], 
     Hue[0.8376539618743016, 1, 1], 
     Hue[0.8376539618743016, 1, 1], 
     Hue[0.8545278627891637, 1, 1], 
     Hue[0.8545278627891637, 1, 1], 
     Hue[0.8714017637040312, 1, 1], 
     Hue[0.8714017637040312, 1, 1], 
     Hue[0.8882756646188987, 1, 1], 
     Hue[0.8882756646188987, 1, 1], 
     Hue[0.9051495655337625, 1, 1], 
     Hue[0.9051495655337625, 1, 1], 
     Hue[0.9220234664486272, 1, 1], 
     Hue[0.9220234664486272, 1, 1], 
     Hue[0.9388973673634933, 1, 1], 
     Hue[0.9388973673634933, 1, 1], 
     Hue[0.9557712682783588, 1, 1], 
     Hue[0.9557712682783588, 1, 1], 
     Hue[0.9726451691932227, 1, 1], 
     Hue[0.9726451691932227, 1, 1], 
     Hue[0.9895190701080886, 1, 1], 
     Hue[0.9895190701080886, 1, 1], 
     Hue[0.006392971022953186, 1, 1], 
     Hue[0.006392971022953186, 1, 1], 
     Hue[0.023266871937819053`, 1, 1], 
     Hue[0.023266871937819053`, 1, 1], 
     Hue[0.040140772852684185`, 1, 1], 
     Hue[0.040140772852684185`, 1, 1], 
     Hue[0.057014673767549456`, 1, 1], 
     Hue[0.057014673767549456`, 1, 1], 
     Hue[0.07388857468241454, 1, 1], 
     Hue[0.07388857468241454, 1, 1], 
     Hue[0.09076247559727997, 1, 1], 
     Hue[0.09076247559727997, 1, 1], 
     Hue[0.10763637651214496`, 1, 1], 
     Hue[0.10763637651214496`, 1, 1], 
     Hue[0.12451027742701012`, 1, 1], 
     Hue[0.12451027742701012`, 1, 1], 
     Hue[0.1413841783418754, 1, 1], 
     Hue[0.1413841783418754, 1, 1], 
     Hue[0.15825807925674074`, 1, 1], 
     Hue[0.15825807925674074`, 1, 1], 
     Hue[0.17513198017160578`, 1, 1], 
     Hue[0.17513198017160578`, 1, 1], 
     Hue[0.19200588108647093`, 1, 1], 
     Hue[0.19200588108647093`, 1, 1], 
     Hue[0.20887978200133625`, 1, 1], 
     Hue[0.20887978200133625`, 1, 1], 
     Hue[0.22575368291620135`, 1, 1], 
     Hue[0.22575368291620135`, 1, 1], 
     Hue[0.24262758383106653`, 1, 1], 
     Hue[0.24262758383106653`, 1, 1], 
     Hue[0.2595014847459317, 1, 1], 
     Hue[0.2595014847459317, 1, 1], 
     Hue[0.2763753856607969, 1, 1], 
     Hue[0.2763753856607969, 1, 1], 
     Hue[0.2932492865756621, 1, 1], 
     Hue[0.2932492865756621, 1, 1], 
     Hue[0.31012318749052736`, 1, 1], 
     Hue[0.31012318749052736`, 1, 1], 
     Hue[0.3269970884053926, 1, 1], 
     Hue[0.3269970884053926, 1, 1], 
     Hue[0.3438709893202577, 1, 1], 
     Hue[0.3438709893202577, 1, 1], 
     Hue[0.3607448902351229, 1, 1], 
     Hue[0.3607448902351229, 1, 1], 
     Hue[0.3776187911499881, 1, 1], 
     Hue[0.3776187911499881, 1, 1], 
     Hue[0.3944926920648533, 1, 1], 
     Hue[0.3944926920648533, 1, 1], 
     Hue[0.41136659297971845`, 1, 1], 
     Hue[0.41136659297971845`, 1, 1], 
     Hue[0.4282404938945837, 1, 1], 
     Hue[0.4282404938945837, 1, 1], 
     Hue[0.44511439480944887`, 1, 1], 
     Hue[0.44511439480944887`, 1, 1], 
     Hue[0.46198829572431405`, 1, 1], 
     Hue[0.46198829572431405`, 1, 1], 
     Hue[0.4788621966391792, 1, 1], 
     Hue[0.4788621966391792, 1, 1], 
     Hue[0.4957360975540444, 1, 1], 
     Hue[0.4957360975540444, 1, 1], 
     Hue[0.5126099984689094, 1, 1], 
     Hue[0.5126099984689094, 1, 1], 
     Hue[0.5294838993837747, 1, 1], 
     Hue[0.5294838993837747, 1, 1], 
     Hue[0.54635780029864, 1, 1], 
     Hue[0.54635780029864, 1, 1], 
     Hue[0.5632317012135051, 1, 1], 
     Hue[0.5632317012135051, 1, 1], 
     Hue[0.5801056021283704, 1, 1], 
     Hue[0.5801056021283704, 1, 1], 
     Hue[0.5969795030432354, 1, 1], 
     Hue[0.5969795030432354, 1, 1], 
     Hue[0.6138534039581007, 1, 1], 
     Hue[0.6138534039581007, 1, 1], 
     Hue[0.6307273048729658, 1, 1], 
     Hue[0.6307273048729658, 1, 1], 
     Hue[0.6476012057878311, 1, 1], 
     Hue[0.6476012057878311, 1, 1], 
     Hue[0.6644751067026962, 1, 1], 
     Hue[0.6644751067026962, 1, 1], 
     Hue[0.6813490076175615, 1, 1], 
     Hue[0.6813490076175615, 1, 1], 
     Hue[0.6982229085324266, 1, 1], 
     Hue[0.6982229085324266, 1, 1], 
     Hue[0.7150968094472918, 1, 1], 
     Hue[0.7150968094472918, 1, 1], 
     Hue[0.731970710362157, 1, 1], 
     Hue[0.731970710362157, 1, 1], 
     Hue[0.7488446112770222, 1, 1], 
     Hue[0.7488446112770222, 1, 1], 
     Hue[0.7657185121918874, 1, 1], 
     Hue[0.7657185121918874, 1, 1], 
     Hue[0.7825924131067525, 1, 1], 
     Hue[0.7825924131067525, 1, 1], 
     Hue[0.7994663140216177, 1, 1], 
     Hue[0.7994663140216177, 1, 1], 
     Hue[0.8163402149364829, 1, 1], 
     Hue[0.8163402149364829, 1, 1], 
     Hue[0.8332141158513481, 1, 1], 
     Hue[0.8332141158513481, 1, 1], 
     Hue[0.8500880167662133, 1, 1], 
     Hue[0.8500880167662133, 1, 1], 
     Hue[0.8669619176810786, 1, 1], 
     Hue[0.8669619176810786, 1, 1], 
     Hue[0.8838358185959436, 1, 1], 
     Hue[0.8838358185959436, 1, 1], 
     Hue[0.9007097195108089, 1, 1], 
     Hue[0.9007097195108089, 1, 1], 
     Hue[0.9175836204256741, 1, 1], 
     Hue[0.9175836204256741, 1, 1], 
     Hue[0.9344575213405393, 1, 1], 
     Hue[0.9344575213405393, 1, 1], 
     Hue[0.9513314222554045, 1, 1], 
     Hue[0.9513314222554045, 1, 1], 
     Hue[0.9682053231702696, 1, 1], 
     Hue[0.9682053231702696, 1, 1], 
     Hue[0.9850792240851348, 1, 1], 
     Hue[0.9850792240851348, 1, 1], 
     Hue[0.001953125000000001, 1, 1], 
     Hue[0.001953125000000001, 1, 1], 
     Hue[0.018827025914865188`, 1, 1], 
     Hue[0.018827025914865188`, 1, 1], 
     Hue[0.03570092682973037, 1, 1], 
     Hue[0.03570092682973037, 1, 1], 
     Hue[0.052574827744595556`, 1, 1], 
     Hue[0.052574827744595556`, 1, 1], 
     Hue[0.06944872865946075, 1, 1], 
     Hue[0.06944872865946075, 1, 1], 
     Hue[0.08632262957432595, 1, 1], 
     Hue[0.08632262957432595, 1, 1], 
     Hue[0.10319653048919113`, 1, 1], 
     Hue[0.10319653048919113`, 1, 1], 
     Hue[0.12007043140405632`, 1, 1], 
     Hue[0.12007043140405632`, 1, 1], 
     Hue[0.1369443323189215, 1, 1], 
     Hue[0.1369443323189215, 1, 1], 
     Hue[0.15381823323378668`, 1, 1], 
     Hue[0.15381823323378668`, 1, 1], 
     Hue[0.1706921341486519, 1, 1], 
     Hue[0.1706921341486519, 1, 1], 
     Hue[0.18756603506351707`, 1, 1], 
     Hue[0.18756603506351707`, 1, 1], 
     Hue[0.20443993597838225`, 1, 1], 
     Hue[0.20443993597838225`, 1, 1], 
     Hue[0.22131383689324746`, 1, 1], 
     Hue[0.22131383689324746`, 1, 1], 
     Hue[0.23818773780811264`, 1, 1], 
     Hue[0.23818773780811264`, 1, 1], 
     Hue[0.2550616387229778, 1, 1], 
     Hue[0.2550616387229778, 1, 1], 
     Hue[0.271935539637843, 1, 1], 
     Hue[0.271935539637843, 1, 1], 
     Hue[0.2888094405527082, 1, 1], 
     Hue[0.2888094405527082, 1, 1], 
     Hue[0.30568334146757337`, 1, 1], 
     Hue[0.30568334146757337`, 1, 1], 
     Hue[0.3225572423824386, 1, 1], 
     Hue[0.3225572423824386, 1, 1], 
     Hue[0.3394311432973038, 1, 1], 
     Hue[0.3394311432973038, 1, 1], 
     Hue[0.35630504421216896`, 1, 1], 
     Hue[0.35630504421216896`, 1, 1], 
     Hue[0.37317894512703415`, 1, 1], 
     Hue[0.37317894512703415`, 1, 1], 
     Hue[0.3900528460418993, 1, 1], 
     Hue[0.3900528460418993, 1, 1], 
     Hue[0.4069267469567645, 1, 1], 
     Hue[0.4069267469567645, 1, 1], 
     Hue[0.4238006478716297, 1, 1], 
     Hue[0.4238006478716297, 1, 1], 
     Hue[0.44067454878649487`, 1, 1], 
     Hue[0.44067454878649487`, 1, 1], 
     Hue[0.45754844970136005`, 1, 1], 
     Hue[0.45754844970136005`, 1, 1], 
     Hue[0.47442235061622523`, 1, 1], 
     Hue[0.47442235061622523`, 1, 1], 
     Hue[0.49129625153109047`, 1, 1], 
     Hue[0.49129625153109047`, 1, 1], 
     Hue[0.5081701524459556, 1, 1], 
     Hue[0.5081701524459556, 1, 1], 
     Hue[0.5250440533608207, 1, 1], 
     Hue[0.5250440533608207, 1, 1], 
     Hue[0.541917954275686, 1, 1], 
     Hue[0.541917954275686, 1, 1], 
     Hue[0.5587918551905511, 1, 1], 
     Hue[0.5587918551905511, 1, 1], 
     Hue[0.5756657561054164, 1, 1], 
     Hue[0.5756657561054164, 1, 1], 
     Hue[0.5925396570202817, 1, 1], 
     Hue[0.5925396570202817, 1, 1], 
     Hue[0.6094135579351467, 1, 1], 
     Hue[0.6094135579351467, 1, 1], 
     Hue[0.6262874588500119, 1, 1], 
     Hue[0.6262874588500119, 1, 1], 
     Hue[0.6431613597648771, 1, 1], 
     Hue[0.6431613597648771, 1, 1], 
     Hue[0.6600352606797423, 1, 1], 
     Hue[0.6600352606797423, 1, 1], 
     Hue[0.6769091615946075, 1, 1], 
     Hue[0.6769091615946075, 1, 1], 
     Hue[0.6937830625094727, 1, 1], 
     Hue[0.6937830625094727, 1, 1], 
     Hue[0.7106569634243378, 1, 1], 
     Hue[0.7106569634243378, 1, 1], 
     Hue[0.727530864339203, 1, 1], 
     Hue[0.727530864339203, 1, 1], 
     Hue[0.7444047652540682, 1, 1], 
     Hue[0.7444047652540682, 1, 1], 
     Hue[0.7612786661689335, 1, 1], 
     Hue[0.7612786661689335, 1, 1], 
     Hue[0.7781525670837985, 1, 1], 
     Hue[0.7781525670837985, 1, 1], 
     Hue[0.7950264679986638, 1, 1], 
     Hue[0.7950264679986638, 1, 1], 
     Hue[0.8119003689135291, 1, 1], 
     Hue[0.8119003689135291, 1, 1], 
     Hue[0.8287742698283941, 1, 1], 
     Hue[0.8287742698283941, 1, 1], 
     Hue[0.8456481707432593, 1, 1], 
     Hue[0.8456481707432593, 1, 1], 
     Hue[0.8625220716581248, 1, 1], 
     Hue[0.8625220716581248, 1, 1], 
     Hue[0.8793959725729897, 1, 1], 
     Hue[0.8793959725729897, 1, 1], 
     Hue[0.8962698734878554, 1, 1], 
     Hue[0.8962698734878554, 1, 1], 
     Hue[0.9131437744027202, 1, 1], 
     Hue[0.9131437744027202, 1, 1], 
     Hue[0.9300176753175851, 1, 1], 
     Hue[0.9300176753175851, 1, 1], 
     Hue[0.9468915762324507, 1, 1], 
     Hue[0.9468915762324507, 1, 1], 
     Hue[0.9637654771473162, 1, 1], 
     Hue[0.9637654771473162, 1, 1], 
     Hue[0.9806393780621809, 1, 1], 
     Hue[0.9806393780621809, 1, 1], 
     Hue[0.997513278977046, 1, 1], 
     Hue[0.997513278977046, 1, 1], 
     Hue[0.01438717989191182, 1, 1], 
     Hue[0.01438717989191182, 1, 1], 
     Hue[0.03126108080677669, 1, 1], 
     Hue[0.03126108080677669, 1, 1], 
     Hue[0.048134981721641074`, 1, 1], 
     Hue[0.048134981721641074`, 1, 1], 
     Hue[0.06500888263650717, 1, 1], 
     Hue[0.06500888263650717, 1, 1], 
     Hue[0.08188278355137264, 1, 1], 
     Hue[0.08188278355137264, 1, 1], 
     Hue[0.09875668446623698, 1, 1], 
     Hue[0.09875668446623698, 1, 1], 
     Hue[0.11563058538110169`, 1, 1], 
     Hue[0.11563058538110169`, 1, 1], 
     Hue[0.13250448629596936`, 1, 1], 
     Hue[0.13250448629596936`, 1, 1], 
     Hue[0.14937838721083332`, 1, 1], 
     Hue[0.14937838721083332`, 1, 1], 
     Hue[0.16625228812570111`, 1, 1], 
     Hue[0.16625228812570111`, 1, 1], 
     Hue[0.18312618904056366`, 1, 1], 
     Hue[0.18312618904056366`, 1, 1], 
     Hue[0.20000008995542723`, 1, 1], 
     Hue[0.20000008995542723`, 1, 1], 
     Hue[0.21687399087029446`, 1, 1], 
     Hue[0.21687399087029446`, 1, 1], 
     Hue[0.23374789178516156`, 1, 1], 
     Hue[0.23374789178516156`, 1, 1], 
     Hue[0.2506217927000209, 1, 1], 
     Hue[0.2506217927000209, 1, 1], 
     Hue[0.267495693614888, 1, 1], 
     Hue[0.267495693614888, 1, 1], 
     Hue[0.2843695945297534, 1, 1], 
     Hue[0.2843695945297534, 1, 1], 
     Hue[0.3012434954446202, 1, 1], 
     Hue[0.3012434954446202, 1, 1], 
     Hue[0.3181173963594849, 1, 1], 
     Hue[0.3181173963594849, 1, 1], 
     Hue[0.33499129727434906`, 1, 1], 
     Hue[0.33499129727434906`, 1, 1], 
     Hue[0.3518651981892264, 1, 1], 
     Hue[0.3518651981892264, 1, 1], 
     Hue[0.3687390991040822, 1, 1], 
     Hue[0.3687390991040822, 1, 1], 
     Hue[0.3856130000189433, 1, 1], 
     Hue[0.3856130000189433, 1, 1], 
     Hue[0.4024869009338191, 1, 1], 
     Hue[0.4024869009338191, 1, 1], 
     Hue[0.41936080184869673`, 1, 1], 
     Hue[0.41936080184869673`, 1, 1], 
     Hue[0.43623470276358123`, 1, 1], 
     Hue[0.43623470276358123`, 1, 1], 
     Hue[0.4531086036784016, 1, 1], 
     Hue[0.4531086036784016, 1, 1], 
     Hue[0.46998250459322793`, 1, 1], 
     Hue[0.46998250459322793`, 1, 1], 
     Hue[0.4868564055081243, 1, 1], 
     Hue[0.4868564055081243, 1, 1], 
     Hue[0.5037303064230843, 1, 1], 
     Hue[0.5037303064230843, 1, 1], 
     Hue[0.5206042073378334, 1, 1], 
     Hue[0.5206042073378334, 1, 1], 
     Hue[0.5374781082527711, 1, 1], 
     Hue[0.5374781082527711, 1, 1], 
     Hue[0.554352009167685, 1, 1], 
     Hue[0.554352009167685, 1, 1], 
     Hue[0.5712259100824209, 1, 1], 
     Hue[0.5712259100824209, 1, 1], 
     Hue[0.5880998109971458, 1, 1], 
     Hue[0.5880998109971458, 1, 1], 
     Hue[0.6049737119121952, 1, 1], 
     Hue[0.6049737119121952, 1, 1], 
     Hue[0.6218476128271425, 1, 1], 
     Hue[0.6218476128271425, 1, 1], 
     Hue[0.6387215137422474, 1, 1], 
     Hue[0.6387215137422474, 1, 1], 
     Hue[0.655595414656956, 1, 1], 
     Hue[0.655595414656956, 1, 1], 
     Hue[0.6724693155719632, 1, 1], 
     Hue[0.6724693155719632, 1, 1], 
     Hue[0.6893432164865616, 1, 1], 
     Hue[0.6893432164865616, 1, 1], 
     Hue[0.7062171174015373, 1, 1], 
     Hue[0.7062171174015373, 1, 1], 
     Hue[0.7230910183162487, 1, 1], 
     Hue[0.7230910183162487, 1, 1], 
     Hue[0.7399649192304052, 1, 1], 
     Hue[0.7399649192304052, 1, 1], 
     Hue[0.7568388201460909, 1, 1], 
     Hue[0.7568388201460909, 1, 1], 
     Hue[0.7737127210634289, 1, 1], 
     Hue[0.7737127210634289, 1, 1], 
     Hue[0.7905866219758477, 1, 1], 
     Hue[0.7905866219758477, 1, 1], 
     Hue[0.8074605228909263, 1, 1], 
     Hue[0.8074605228909263, 1, 1], 
     Hue[0.8243344238026495, 1, 1], 
     Hue[0.8243344238026495, 1, 1], 
     Hue[0.8412083247197057, 1, 1], 
     Hue[0.8412083247197057, 1, 1], 
     Hue[0.8580822256375282, 1, 1], 
     Hue[0.8580822256375282, 1, 1], 
     Hue[0.874956126559485, 1, 1], 
     Hue[0.874956126559485, 1, 1], 
     Hue[0.8918300274636765, 1, 1], 
     Hue[0.8918300274636765, 1, 1], 
     Hue[0.9087039283913463, 1, 1], 
     Hue[0.9087039283913463, 1, 1], 
     Hue[0.9255778292982721, 1, 1], 
     Hue[0.9255778292982721, 1, 1], 
     Hue[0.9424517302066693, 1, 1], 
     Hue[0.9424517302066693, 1, 1], 
     Hue[0.9593256310977288, 1, 1], 
     Hue[0.9593256310977288, 1, 1], 
     Hue[0.9761995320802711, 1, 1], 
     Hue[0.9761995320802711, 1, 1], 
     Hue[0.9930734329972324, 1, 1], 
     Hue[0.9930734329972324, 1, 1], 
     Hue[0.009947333879442764, 1, 1], 
     Hue[0.009947333879442764, 1, 1], 
     Hue[0.0268212347484224, 1, 1], 
     Hue[0.0268212347484224, 1, 1], 
     Hue[0.04369513568401856, 1, 1], 
     Hue[0.04369513568401856, 1, 1], 
     Hue[0.06056903675627041, 1, 1], 
     Hue[0.06056903675627041, 1, 1], 
     Hue[0.07744293768879483, 1, 1], 
     Hue[0.07744293768879483, 1, 1], 
     Hue[0.09431683820258967, 1, 1], 
     Hue[0.09431683820258967, 1, 1], 
     Hue[0.11119073889870218`, 1, 1], 
     Hue[0.11119073889870218`, 1, 1], 
     Hue[0.1280646399305784, 1, 1], 
     Hue[0.1280646399305784, 1, 1], 
     Hue[0.14493854189656583`, 1, 1], 
     Hue[0.14493854189656583`, 1, 1], 
     Hue[0.16181244205002082`, 1, 1], 
     Hue[0.16181244205002082`, 1, 1], 
     Hue[0.17868634385560458`, 1, 1], 
     Hue[0.17868634385560458`, 1, 1], 
     Hue[0.1955602447326109, 1, 1], 
     Hue[0.1955602447326109, 1, 1], 
     Hue[0.21243414538737151`, 1, 1], 
     Hue[0.21243414538737151`, 1, 1], 
     Hue[0.2293080462819191, 1, 1], 
     Hue[0.2293080462819191, 1, 1], 
     Hue[0.25, 1, 1], 
     Hue[0.25, 1, 1], 
     Hue[0.26305583920610676`, 1, 1], 
     Hue[0.26305583920610676`, 1, 1], 
     Hue[0.27992975012920396`, 1, 1], 
     Hue[0.27992975012920396`, 1, 1], 
     Hue[0.2968036537992501, 1, 1], 
     Hue[0.2968036537992501, 1, 1], 
     Hue[0.3136775686453699, 1, 1], 
     Hue[0.3136775686453699, 1, 1], 
     Hue[0.330551450707842, 1, 1], 
     Hue[0.330551450707842, 1, 1], 
     Hue[0.34742535537280034`, 1, 1], 
     Hue[0.34742535537280034`, 1, 1], 
     Hue[0.3642992132658118, 1, 1], 
     Hue[0.3642992132658118, 1, 1], 
     Hue[0.3811731274121047, 1, 1], 
     Hue[0.3811731274121047, 1, 1], 
     Hue[0.39804707607115386`, 1, 1], 
     Hue[0.39804707607115386`, 1, 1], 
     Hue[0.5, 1, 1], 
     Hue[0.5, 1, 1]}],
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  Epilog->LineBox[{{-11.97651663405088, 0.}, {-11.92954990215264, 
     0.}, {-11.882583170254403`, 0.}, {-11.835616438356164`, 
     0.}, {-11.788649706457925`, 0.}, {-11.741682974559685`, 
     0.}, {-11.694716242661446`, 0.}, {-11.647749510763209`, 
     0.}, {-11.60078277886497, 0.}, {-11.55381604696673, 
     0.}, {-11.506849315068493`, 0.}, {-11.459882583170254`, 
     0.}, {-11.412915851272015`, 0.}, {-11.365949119373775`, 
     0.}, {-11.318982387475536`, 0.}, {-11.272015655577299`, 
     0.}, {-11.22504892367906, 0.}, {-11.17808219178082, 
     0.}, {-11.131115459882583`, 0.}, {-11.084148727984344`, 
     0.}, {-11.037181996086105`, 0.}, {-10.990215264187865`, 
     0.}, {-10.943248532289626`, 0.}, {-10.896281800391389`, 
     0.}, {-10.84931506849315, 0.}, {-10.80234833659491, 
     0.}, {-10.755381604696673`, 0.}, {-10.708414872798434`, 
     0.}, {-10.661448140900195`, 0.}, {-10.614481409001955`, 
     0.}, {-10.567514677103716`, 0.}, {-10.520547945205479`, 
     0.}, {-10.47358121330724, 0.}, {-10.426614481409, 
     0.}, {-10.379647749510763`, 0.}, {-10.332681017612524`, 
     0.}, {-10.285714285714285`, 0.}, {-10.238747553816046`, 
     0.}, {-10.191780821917806`, 0.}, {-10.144814090019569`, 
     0.}, {-10.09784735812133, 0.}, {-10.05088062622309, 
     0.}, {-10.003913894324853`, 0.}, {-9.956947162426614, 
     0.}, {-9.909980430528375, 0.}, {-9.863013698630136, 
     0.}, {-9.816046966731896, 0.}, {-9.769080234833659, 
     0.}, {-9.72211350293542, 0.}, {-9.67514677103718, 
     0.}, {-9.628180039138943, 0.}, {-9.581213307240704, 
     0.}, {-9.534246575342465, 0.}, {-9.487279843444226, 
     0.}, {-9.440313111545986, 0.}, {-9.393346379647749, 
     0.}, {-9.34637964774951, 0.}, {-9.29941291585127, 
     0.}, {-9.252446183953033, 0.}, {-9.205479452054794, 
     0.}, {-9.158512720156555, 0.}, {-9.111545988258316, 
     0.}, {-9.064579256360076, 0.}, {-9.017612524461839, 
     0.}, {-8.9706457925636, 0.}, {-8.92367906066536, 
     0.}, {-8.876712328767123, 0.}, {-8.829745596868884, 
     0.}, {-8.782778864970645, 0.}, {-8.735812133072406, 
     0.}, {-8.688845401174166, 0.}, {-8.641878669275929, 
     0.}, {-8.59491193737769, 0.}, {-8.54794520547945, 
     0.}, {-8.500978473581213, 0.}, {-8.454011741682974, 
     0.}, {-8.407045009784735, 0.}, {-8.360078277886496, 
     0.}, {-8.313111545988257, 0.}, {-8.266144814090019, 
     0.}, {-8.21917808219178, 0.}, {-8.17221135029354, 
     0.}, {-8.125244618395303, 0.}, {-8.078277886497064, 
     0.}, {-8.031311154598825, 0.}, {-7.984344422700586, 
     0.}, {-7.9373776908023475`, 0.}, {-7.890410958904109, 
     0.}, {-7.84344422700587, 0.}, {-7.796477495107631, 
     0.}, {-7.7495107632093925`, 0.}, {-7.702544031311154, 
     0.}, {-7.655577299412915, 0.}, {-7.608610567514676, 
     0.}, {-7.5616438356164375`, 0.}, {-7.514677103718199, 
     0.}, {-7.46771037181996, 0.}, {-7.420743639921721, 
     0.}, {-7.3737769080234825`, 0.}, {-7.326810176125244, 
     0.}, {-7.279843444227005, 0.}, {-7.232876712328766, 
     0.}, {-7.1859099804305275`, 0.}, {-7.138943248532289, 
     0.}, {-7.09197651663405, 0.}, {-7.045009784735811, 
     0.}, {-6.9980430528375726`, 0.}, {-6.951076320939334, 
     0.}, {-6.904109589041095, 0.}, {-6.857142857142856, 
     0.}, {-6.810176125244618, 0.}, {-6.763209393346379, 
     0.}, {-6.71624266144814, 0.}, {-6.669275929549901, 
     1.1146378458122364`*^-10}, {-6.622309197651663, 
     1.798290711340169*^-10}, {-6.575342465753424, 
     2.4540735289285193`*^-10}, {-6.528375733855185, 
     3.341651880264987*^-10}, {-6.481409001956946, 
     4.5402579199620793`*^-10}, {-6.434442270058708, 
     6.155249816035679*^-10}, {-6.387475538160469, 
     8.326383372720491*^-10}, {-6.34050880626223, 
     1.1238619075720716`*^-9}, {-6.293542074363991, 
     1.5136142550940394`*^-9}, {-6.246575342465753, 
     2.0340576674756893`*^-9}, {-6.199608610567514, 
     2.7274511648907347`*^-9}, {-6.152641878669275, 
     3.649190077364248*^-9}, {-6.105675146771036, 
     4.871713297785885*^-9}, {-6.058708414872798, 
     6.489521747111615*^-9}, {-6.011741682974559, 
     8.625601807063124*^-9}, {-5.96477495107632, 
     1.1439626874631107`*^-8}, {-5.917808219178081, 
     1.5138402538008152`*^-8}, {-5.870841487279843, 
     1.998913420433536*^-8}, {-5.823874755381604, 
     2.6336232928766557`*^-8}, {-5.776908023483365, 
     3.462255126936783*^-8}, {-5.729941291585126, 
     4.541613983512785*^-8}, {-5.682974559686888, 
     5.9443874787860965`*^-8}, {-5.636007827788649, 
     7.763360272936063*^-8}, {-5.58904109589041, 
     1.0116682281523681`*^-7}, {-5.542074363992171, 
     1.3154435662530026`*^-7}, {-5.495107632093933, 
     1.7066798902598574`*^-7}, {-5.448140900195694, 
     2.2094168439938245`*^-7}, {-5.401174168297455, 
     2.8539672807312715`*^-7}, {-5.354207436399216, 
     3.6784601757501764`*^-7}, {-5.307240704500978, 
     4.7307376807669645`*^-7}, {-5.260273972602739, 
     6.070681041163823*^-7}, {-5.2133072407045, 
     7.773054328157751*^-7}, {-5.166340508806261, 
     9.930971385911677*^-7}, {-5.119373776908023, 
     1.266011054208007*^-6}, {-5.0724070450097845`, 
     1.6103823595010388`*^-6}, {-5.025440313111545, 
     2.0439310961983656`*^-6}, {-4.978473581213306, 
     2.5885063646478003`*^-6}, {-4.931506849315068, 
     3.2709805595384457`*^-6}, {-4.8845401174168295`, 
     4.124320705128717*^-6}, {-4.83757338551859, 
     5.188868129072776*^-6}, {-4.790606653620351, 
     6.5138623724850525`*^-6}, {-4.743639921722113, 
     8.159250426462896*^-6}, {-4.6966731898238745`, 
     0.000010197828099432689`}, {-4.649706457925635, 
     0.000012717766609814107`}, {-4.602739726027396, 
     0.000015825584319755882`}, {-4.555772994129158, 
     0.000019649630916281903`}, {-4.5088062622309195`, 
     0.00002434415923696738}, {-4.46183953033268, 
     0.000030094068318513763`}, {-4.414872798434441, 
     0.000037120410021240786`}, {-4.367906066536203, 
     0.00004568676067911134}, {-4.3209393346379645`, 
     0.00005610656849736523}, {-4.273972602739725, 
     0.00006875159673425522}, {-4.227005870841486, 
     0.00008406159182672121}, {-4.180039138943248, 
     0.00010255531435397157`}, {-4.1330724070450096`, 
     0.00012484307875294798`}, {-4.08610567514677, 
     0.00015164095470630992`}, {-4.039138943248531, 
     0.000183786788704313}, {-3.992172211350293, 
     0.00022225820803475224`}, {-3.9452054794520546`, 
     0.00026819277085812234`}, {-3.8982387475538154`, 
     0.0003229104245872094}, {-3.851272015655576, 
     0.0003879384298932832}, {-3.804305283757339, 
     0.0004650388987209761}, {-3.7573385518590996`, 
     0.0005562390810359168}, {-3.7103718199608604`, 
     0.0006638645160091038}, {-3.663405088062621, 
     0.0007905751382557082}, {-3.616438356164382, 
     0.0009394043979607694}, {-3.5694716242661446`, 
     0.0011138014145646142`}, {-3.5225048923679054`, 
     0.001317676136604144}, {-3.4755381604696662`, 
     0.0015554474247578382`}, {-3.428571428571429, 
     0.0018320939107596239`}, {-3.3816046966731896`, 
     0.002153207411332789}, {-3.3346379647749504`, 
     0.002525048593568606}, {-3.2876712328767113`, 
     0.0029546044963193368`}, {-3.240704500978472, 
     0.0034496474115498954`}, {-3.1937377690802347`, 
     0.004018794520776894}, {-3.1467710371819955`, 
     0.00467156756564932}, {-3.0998043052837563`, 
     0.0054184517096023364`}, {-3.052837573385519, 
     0.006270952620948491}, {-3.0058708414872797`, 
     0.007241650678706041}, {-2.9589041095890405`, 
     0.00834425107324641}, {-2.9119373776908013`, 
     0.00959362844718826}, {-2.864970645792562, 
     0.011005864600972984`}, {-2.8180039138943247`, 
     0.01259827767567321}, {-2.7710371819960855`, 
     0.014389441126572075`}, {-2.7240704500978463`, 
     0.01639919071895315}, {-2.677103718199609, 
     0.018648617716584145`}, {-2.6301369863013697`, 
     0.02116004639794654}, {-2.5831702544031305`, 
     0.02395699402974709}, {-2.5362035225048913`, 
     0.027064111455975863`}, {-2.489236790606652, 
     0.03050710252787441}, {-2.4422700587084147`, 
     0.034312620709444026`}, {-2.3953033268101755`, 
     0.038508141347880734`}, {-2.3483365949119364`, 
     0.04312180830126996}, {-2.301369863013699, 
     0.04818225386893843}, {-2.2544031311154598`, 
     0.05371839127405519}, {-2.2074363992172206`, 
     0.05975917930334213}, {-2.1604696673189814`, 
     0.0663333591138482}, {-2.113502935420742, 
     0.07346916366912455}, {-2.0665362035225048`, 
     0.08119400076284937}, {-2.0195694716242656`, 
     0.08953411112162052}, {-1.9726027397260264`, 
     0.09851420364336058}, {-1.925636007827789, 
     0.10815707041522583`}, {-1.8786692759295498`, 
     0.11848318475531514`}, {-1.8317025440313106`, 
     0.12951028612472454`}, {-1.7847358121330714`, 
     0.14125295634834156`}, {-1.7377690802348322`, 
     0.1537221921509342}, {-1.6908023483365948`, 
     0.16692497954557164`}, {-1.6438356164383556`, 
     0.18086387608971433`}, {-1.5968688845401164`, 
     0.1955366074358038}, {-1.549902152641879, 
     0.21093568493342876`}, {-1.5029354207436398`, 
     0.2270480512752961}, {-1.4559686888454006`, 
     0.2438547613064273}, {-1.4090019569471615`, 
     0.2613307051237463}, {-1.3620352250489223`, 
     0.27944438047184034`}, {-1.3150684931506849`, 
     0.2981577211826468}, {-1.2681017612524457`, 
     0.3174259880071435}, {-1.2211350293542065`, 
     0.33719772764369305`}, {-1.174168297455969, 
     0.3574148050814269}, {-1.1272015655577299`, 
     0.378012513552274}, {-1.0802348336594907`, 
     0.39891976542957175`}, {-1.0332681017612515`, 
     0.42005936633582003`}, {-0.9863013698630123, 
     0.4413483735415413}, {-0.9393346379647749, 
     0.462698538469196}, {-0.8923679060665357, 
     0.48401683178139454`}, {-0.8454011741682965, 
     0.5052060481546795}, {-0.7984344422700591, 
     0.5261654864445081}, {-0.7514677103718199, 
     0.5467916995610776}, {-0.7045009784735807, 
     0.5669793070276935}, {-0.6575342465753415, 
     0.586621861912339}, {-0.6105675146771024, 
     0.6056127626376295}, {-0.5636007827788649, 
     0.6238461991120362}, {-0.5166340508806258, 
     0.641218121712225}, {-0.46966731898238656`, 
     0.6576272209061719}, {-0.42270058708414915`, 
     0.6729759047600126}, {-0.37573385518590996`, 
     0.6871712612353377}, {-0.32876712328767077`, 
     0.7001259920706797}, {-0.2818003913894316, 
     0.7117593051594256}, {-0.2348336594911924, 
     0.7219977526896628}, {-0.18786692759295498`, 
     0.7307760028976726}, {-0.1409001956947158, 
     0.7380375340988988}, {-0.0939334637964766, 
     0.7437352406861111}, {-0.04696673189823919, 0.7478319420070572}, {
     0, 0.7503007864315615}, {0.04696673189823919, 0.7511255444649424}, {
     0.09393346379647838, 0.7503007864315617}, {0.14090019569471757`, 
     0.7478319420070572}, {0.18786692759295498`, 0.7437352406861113}, {
     0.23483365949119417`, 0.7380375340988988}, {0.28180039138943336`, 
     0.7307760028976724}, {0.32876712328767077`, 0.7219977526896627}, {
     0.37573385518590996`, 0.711759305159426}, {0.42270058708414915`, 
     0.7001259920706797}, {0.46966731898238834`, 0.6871712612353376}, {
     0.5166340508806275, 0.6729759047600126}, {0.5636007827788649, 
     0.657627220906172}, {0.6105675146771041, 0.641218121712225}, {
     0.6575342465753433, 0.6238461991120361}, {0.7045009784735807, 
     0.6056127626376296}, {0.7514677103718199, 0.5866218619123391}, {
     0.7984344422700591, 0.5669793070276934}, {0.8454011741682983, 
     0.5467916995610778}, {0.8923679060665375, 0.5261654864445079}, {
     0.9393346379647749, 0.5052060481546796}, {0.9863013698630141, 
     0.48401683178139465`}, {1.0332681017612533`, 0.462698538469196}, {
     1.0802348336594907`, 0.44134837354154133`}, {1.1272015655577299`, 
     0.42005936633582003`}, {1.174168297455969, 0.3989197654295718}, {
     1.2211350293542083`, 0.3780125135522739}, {1.2681017612524474`, 
     0.357414805081427}, {1.3150684931506849`, 0.3371977276436931}, {
     1.362035225048924, 0.3174259880071436}, {1.4090019569471632`, 
     0.2981577211826467}, {1.4559686888454006`, 0.27944438047184045`}, {
     1.5029354207436398`, 0.26133070512374634`}, {1.549902152641879, 
     0.24385476130642736`}, {1.5968688845401182`, 0.22704805127529612`}, {
     1.6438356164383574`, 0.2109356849334286}, {1.6908023483365948`, 
     0.1955366074358039}, {1.737769080234834, 0.1808638760897143}, {
     1.7847358121330732`, 0.1669249795455716}, {1.8317025440313106`, 
     0.1537221921509341}, {1.8786692759295498`, 0.14125295634834162`}, {
     1.925636007827789, 0.1295102861247246}, {1.9726027397260282`, 
     0.1184831847553151}, {2.0195694716242674`, 0.10815707041522586`}, {
     2.0665362035225048`, 0.09851420364336061}, {2.113502935420744, 
     0.0895341111216205}, {2.160469667318983, 0.08119400076284929}, {
     2.2074363992172206`, 0.07346916366912458}, {2.2544031311154598`, 
     0.0663333591138483}, {2.301369863013699, 0.05975917930334214}, {
     2.348336594911938, 0.05371839127405509}, {2.3953033268101773`, 
     0.04818225386893834}, {2.4422700587084147`, 0.043121808301270055`}, {
     2.489236790606654, 0.03850814134788078}, {2.536203522504893, 
     0.034312620709444006`}, {2.5831702544031305`, 0.030507102527874353`}, {
     2.6301369863013697`, 0.027064111455975884`}, {2.677103718199609, 
     0.023956994029747044`}, {2.724070450097848, 0.021160046397946514`}, {
     2.7710371819960873`, 0.018648617716584166`}, {2.8180039138943247`, 
     0.01639919071895321}, {2.864970645792564, 0.014389441126572047`}, {
     2.911937377690803, 0.012598277675673165`}, {2.9589041095890405`, 
     0.011005864600973006`}, {3.0058708414872797`, 0.009593628447188297}, {
     3.052837573385519, 0.008344251073246404}, {3.099804305283758, 
     0.007241650678706}, {3.1467710371819972`, 0.006270952620948439}, {
     3.1937377690802347`, 0.005418451709602366}, {3.240704500978474, 
     0.0046715675656493144`}, {3.287671232876713, 0.004018794520776871}, {
     3.3346379647749504`, 0.003449647411549889}, {3.3816046966731896`, 
     0.002954604496319346}, {3.428571428571429, 0.002525048593568576}, {
     3.475538160469668, 0.0021532074113327364`}, {3.522504892367907, 
     0.0018320939107595911`}, {3.5694716242661446`, 0.0015554474247579078`}, {
     3.616438356164384, 0.0013176761366041083`}, {3.663405088062623, 
     0.0011138014145645216`}, {3.7103718199608604`, 0.0009394043979608097}, {
     3.7573385518590996`, 0.0007905751382557879}, {3.804305283757339, 
     0.0006638645160091275}, {3.851272015655578, 0.0005562390810358164}, {
     3.898238747553817, 0.00046503889872092583`}, {3.9452054794520546`, 
     0.0003879384298933052}, {3.9921722113502938`, 0.00032291042458716635`}, {
     4.039138943248533, 0.0002681927708581025}, {4.08610567514677, 
     0.0002222582080347793}, {4.133072407045011, 0.00018378678870435648`}, {
     4.180039138943249, 0.00015164095470630507`}, {4.227005870841486, 
     0.00012484307875294145`}, {4.273972602739727, 0.00010255531435398183`}, {
     4.3209393346379645`, 0.0000840615918267504}, {4.367906066536202, 
     0.00006875159673425208}, {4.414872798434443, 0.00005610656849735432}, {
     4.46183953033268, 0.000045686760679110036`}, {4.508806262230921, 
     0.00003712041002131874}, {4.555772994129159, 0.00003009406831851638}, {
     4.602739726027396, 0.000024344159236883366`}, {4.649706457925637, 
     0.000019649630916338512`}, {4.6966731898238745`, 
     0.000015825584319779938`}, {4.7436399217221155`, 
     0.000012717766609783351`}, {4.790606653620353, 
     0.000010197828099446573`}, {4.83757338551859, 8.159250426534814*^-6}, {
     4.884540117416831, 6.513862372558555*^-6}, {4.931506849315069, 
     5.188868129056099*^-6}, {4.978473581213306, 4.124320705156627*^-6}, {
     5.025440313111547, 3.270980559567462*^-6}, {5.0724070450097845`, 
     2.588506364737744*^-6}, {5.119373776908022, 2.043931096290343*^-6}, {
     5.166340508806263, 1.610382359543991*^-6}, {5.2133072407045, 
     1.2660110542325776`*^-6}, {5.260273972602741, 9.93097138638118*^-7}, {
     5.307240704500979, 7.773054327926097*^-7}, {5.354207436399216, 
     6.07068104123621*^-7}, {5.401174168297457, 4.730737681638085*^-7}, {
     5.448140900195694, 3.678460175811697*^-7}, {5.495107632093935, 
     2.8539672795152917`*^-7}, {5.542074363992173, 2.209416843410984*^-7}, {
     5.58904109589041, 1.7066798907222866`*^-7}, {5.636007827788651, 
     1.3154435668035579`*^-7}, {5.682974559686889, 1.0116682282254496`*^-7}, {
     5.729941291585126, 7.763360264006796*^-8}, {5.776908023483367, 
     5.944387477012292*^-8}, {5.823874755381604, 4.5416139825348516`*^-8}, {
     5.870841487279842, 3.462255125897259*^-8}, {5.917808219178083, 
     2.6336232967917442`*^-8}, {5.96477495107632, 1.9989134174029878`*^-8}, {
     6.011741682974561, 1.513840257998916*^-8}, {6.0587084148727985`, 
     1.1439626880071873`*^-8}, {6.105675146771036, 8.625601757067139*^-9}, {
     6.152641878669277, 6.489521768056408*^-9}, {6.199608610567514, 
     4.87171319258195*^-9}, {6.246575342465755, 3.6491900414652467`*^-9}, {
     6.293542074363993, 2.7266664351694674`*^-9}, {6.34050880626223, 
     2.0340577012663213`*^-9}, {6.387475538160471, 1.513614378340835*^-9}, {
     6.4344422700587085`, 1.1238618997549755`*^-9}, {6.481409001956946, 
     8.32638285955028*^-10}, {6.528375733855187, 6.155249353634929*^-10}, {
     6.575342465753424, 4.5402581771459657`*^-10}, {6.622309197651662, 
     3.3416517814527953`*^-10}, {6.669275929549903, 2.454074049910307*^-10}, {
     6.71624266144814, 1.7982901895185695`*^-10}, {6.763209393346381, 
     1.1314185110662897`*^-10}, {6.8101761252446185`, 0.}, {6.857142857142856,
      0.}, {6.904109589041097, 0.}, {6.951076320939334, 0.}, {
     6.998043052837575, 0.}, {7.045009784735813, 0.}, {7.09197651663405, 
     0.}, {7.138943248532291, 0.}, {7.185909980430528, 0.}, {
     7.232876712328766, 0.}, {7.279843444227007, 0.}, {7.326810176125244, 
     0.}, {7.373776908023482, 0.}, {7.420743639921723, 0.}, {7.46771037181996,
      0.}, {7.514677103718201, 0.}, {7.561643835616438, 0.}, {
     7.608610567514676, 0.}, {7.655577299412917, 0.}, {7.702544031311154, 
     0.}, {7.749510763209395, 0.}, {7.7964774951076325`, 0.}, {
     7.84344422700587, 0.}, {7.890410958904111, 0.}, {7.937377690802348, 
     0.}, {7.984344422700586, 0.}, {8.031311154598827, 0.}, {
     8.078277886497064, 0.}, {8.125244618395302, 0.}, {8.172211350293543, 
     0.}, {8.21917808219178, 0.}, {8.26614481409002, 0.}, {8.313111545988258, 
     0.}, {8.360078277886496, 0.}, {8.407045009784737, 0.}, {
     8.454011741682974, 0.}, {8.500978473581215, 0.}, {8.547945205479452, 
     0.}, {8.59491193737769, 0.}, {8.64187866927593, 0.}, {8.688845401174168, 
     0.}, {8.735812133072406, 0.}, {8.782778864970647, 0.}, {
     8.829745596868884, 0.}, {8.876712328767121, 0.}, {8.923679060665362, 
     0.}, {8.9706457925636, 0.}, {9.01761252446184, 0.}, {9.064579256360078, 
     0.}, {9.111545988258316, 0.}, {9.158512720156557, 0.}, {
     9.205479452054794, 0.}, {9.252446183953035, 0.}, {9.299412915851272, 
     0.}, {9.34637964774951, 0.}, {9.39334637964775, 0.}, {9.440313111545988, 
     0.}, {9.487279843444226, 0.}, {9.534246575342467, 0.}, {
     9.581213307240704, 0.}, {9.628180039138941, 0.}, {9.675146771037182, 
     0.}, {9.72211350293542, 0.}, {9.76908023483366, 0.}, {9.816046966731898, 
     0.}, {9.863013698630136, 0.}, {9.909980430528377, 0.}, {
     9.956947162426614, 0.}, {10.003913894324855`, 0.}, {10.050880626223092`, 
     0.}, {10.09784735812133, 0.}, {10.14481409001957, 0.}, {
     10.191780821917808`, 0.}, {10.238747553816046`, 0.}, {
     10.285714285714286`, 0.}, {10.332681017612524`, 0.}, {
     10.379647749510761`, 0.}, {10.426614481409002`, 0.}, {10.47358121330724, 
     0.}, {10.52054794520548, 0.}, {10.567514677103718`, 0.}, {
     10.614481409001955`, 0.}, {10.661448140900196`, 0.}, {
     10.708414872798434`, 0.}, {10.755381604696675`, 0.}, {
     10.802348336594912`, 0.}, {10.84931506849315, 0.}, {10.89628180039139, 
     0.}, {10.943248532289628`, 0.}, {10.990215264187865`, 0.}, {
     11.037181996086106`, 0.}, {11.084148727984344`, 0.}, {
     11.131115459882581`, 0.}, {11.178082191780822`, 0.}, {11.22504892367906, 
     0.}, {11.2720156555773, 0.}, {11.318982387475538`, 0.}, {
     11.365949119373775`, 0.}, {11.412915851272016`, 0.}, {
     11.459882583170254`, 0.}, {11.506849315068495`, 0.}, {
     11.553816046966732`, 0.}, {11.60078277886497, 0.}, {11.64774951076321, 
     0.}, {11.694716242661448`, 0.}, {11.741682974559685`, 0.}, {
     11.788649706457926`, 0.}, {11.835616438356164`, 0.}, {
     11.882583170254401`, 0.}, {11.929549902152642`, 0.}, {11.97651663405088, 
     0.}, {12.023483365949119`, 0.}}],
  PlotRange->All,
  PlotRangeClipping->True,
  PlotRangePadding->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{
  3.3927263526527443`*^9, {3.3927265211850815`*^9, 3.391849853515625*^9}, 
   3.391850052109375*^9, 3.392472370896297*^9},
 ImageCache->GraphicsData["CompressedBitmap", "\<\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\
\>"]]
}, Open  ]],

Cell[TextData[{
 "Now the interval in ",
 StyleBox["x",
  FontSlant->"Italic"],
 "-space is symmetric with respect to ",
 StyleBox["x ",
  FontSlant->"Italic"],
 "= 0."
}], "Text"],

Cell["\<\
Computing the Fourier transformation again gives the same result as before \
(up to numerical errors)\
\>", "Text"],

Cell["{testFfvals, testkinterval} = QFourierList[result];", "Input"],

Cell[CellGroupData[{

Cell["\<\
QListArgColorPlot[testFfvals,PlotRange->All,QHorizontalRange->{testkinterval,{\
-1,5}},Frame->True]\
\>", "Input"],

Cell[BoxData[
 GraphicsBox[GraphicsComplexBox[CompressedData["
1:eJxTTMoPSmViYGBwBmIQzcA9KyxL9OX+3/bJc89tqLVviBHs/7jjwf7Xuy/P
WXer354BzL+wPy7HNib25Xx7BiiwnPb/6fuba+0ZYsHy9tfOXBZiFd4F5T+w
b/vpzm/hctQear4925qbpy2MLtgfAJv3wT793dp37JzX7Rm2XP2k+/yLff7x
Lvb+6feg6n/au/FvTb73/ZG9Q1LcrmXOf+3fSv9ff+jbU6h+BgcTr2lPK/c/
t3fo+HyA/wKTA0sCG9Ptn8/tH2wGmcfioOfG6+Y765l9w/3tKj7/2BzefziY
1Pn0MdR8Tof7ss0PN0U9sF9gVtPRocPtcCVM59AmqZtQ+3gdKq+p3Ljy96L9
gV6HN0ei+B2exB3kyBY5Yb8BbL+Aw/TpEScfi++zbyhiWtFwRNDh2kF7Zrb+
jVD3CDlI/uJW+au42F5g3pOb824LO0zco8jdWT4ZFh77YeEIC29cfFj44+QT
AtD4wc9/gJsPjT9cfFh84uLD4hcnHxrfuPiw+MfFh6UHXHxY+sDFh6UXXHxY
+sHFh6UnXHxY+sLFh6U3XHxY+sPFh6VHXHxY+sTFh6VXXHxY+oXxAfiOSes=

   "], {
    {GrayLevel[0.5], EdgeForm[None], 
     GraphicsGroupBox[
      PolygonBox[{{1, 24, 25, 2}, {3, 28, 29, 4}, {2, 26, 27, 3}, {21, 64, 65,
        22}, {5, 32, 33, 6}, {7, 36, 37, 8}, {4, 30, 31, 5}, {9, 40, 41, 
       10}, {6, 34, 35, 7}, {10, 42, 43, 11}, {11, 44, 45, 12}, {12, 46, 47, 
       13}, {16, 54, 55, 17}, {15, 52, 53, 16}, {14, 50, 51, 15}, {8, 38, 39, 
       9}, {17, 56, 57, 18}, {13, 48, 49, 14}, {18, 58, 59, 19}, {19, 60, 61, 
       20}, {20, 62, 63, 21}, {22, 66, 67, 23}}]]}, {{}, {}, {
      {Hue[0.013671875000000699`, 1, 1], LineBox[{1, 2}]}, 
      {Hue[0.009765625000000425, 1, 1], LineBox[{2, 3}]}, 
      {Hue[0.005859375000000088, 1, 1], LineBox[{3, 4}]}, 
      {Hue[0.0019531250000000178`, 1, 1], LineBox[{4, 5}]}, 
      {Hue[0.998046875, 1, 1], LineBox[{5, 6}]}, 
      {Hue[0.994140625, 1, 1], LineBox[{6, 7}]}, 
      {Hue[0.990234375, 1, 1], LineBox[{7, 8}]}, 
      {Hue[0.986328125, 1, 1], LineBox[{8, 9}]}, 
      {Hue[0.982421875, 1, 1], LineBox[{9, 10}]}, 
      {Hue[0.978515625, 1, 1], LineBox[{10, 11}]}, 
      {Hue[0.974609375, 1, 1], LineBox[{11, 12}]}, 
      {Hue[0.970703125, 1, 1], LineBox[{12, 13}]}, 
      {Hue[0.966796875, 1, 1], LineBox[{13, 14}]}, 
      {Hue[0.962890625, 1, 1], LineBox[{14, 15}]}, 
      {Hue[0.958984375, 1, 1], LineBox[{15, 16}]}, 
      {Hue[0.955078125, 1, 1], LineBox[{16, 17}]}, 
      {Hue[0.951171875, 1, 1], LineBox[{17, 18}]}, 
      {Hue[0.947265625, 1, 1], LineBox[{18, 19}]}, 
      {Hue[0.943359375, 1, 1], LineBox[{19, 20}]}, 
      {Hue[0.939453125, 1, 1], LineBox[{20, 21}]}, 
      {Hue[0.935546875, 1, 1], LineBox[{21, 22}]}, 
      {Hue[0.931640625, 1, 1], LineBox[{22, 23}]}}}},
   VertexColors->{
     Hue[0.013671875000000699`, 1, 1], 
     Hue[0.009765625000000425, 1, 1], 
     Hue[0.005859375000000088, 1, 1], 
     Hue[0.0019531250000000178`, 1, 1], 
     Hue[0.998046875, 1, 1], 
     Hue[0.994140625, 1, 1], 
     Hue[0.990234375, 1, 1], 
     Hue[0.986328125, 1, 1], 
     Hue[0.982421875, 1, 1], 
     Hue[0.978515625, 1, 1], 
     Hue[0.974609375, 1, 1], 
     Hue[0.970703125, 1, 1], 
     Hue[0.966796875, 1, 1], 
     Hue[0.962890625, 1, 1], 
     Hue[0.958984375, 1, 1], 
     Hue[0.955078125, 1, 1], 
     Hue[0.951171875, 1, 1], 
     Hue[0.947265625, 1, 1], 
     Hue[0.943359375, 1, 1], 
     Hue[0.939453125, 1, 1], 
     Hue[0.935546875, 1, 1], 
     Hue[0.931640625, 1, 1], 
     Hue[0.9277343749999999, 1, 1], 
     Hue[0.013671875000000699`, 1, 1], 
     Hue[0.009765625000000425, 1, 1], 
     Hue[0.009765625000000425, 1, 1], 
     Hue[0.005859375000000088, 1, 1], 
     Hue[0.005859375000000088, 1, 1], 
     Hue[0.0019531250000000178`, 1, 1], 
     Hue[0.0019531250000000178`, 1, 1], 
     Hue[0.998046875, 1, 1], 
     Hue[0.998046875, 1, 1], 
     Hue[0.994140625, 1, 1], 
     Hue[0.994140625, 1, 1], 
     Hue[0.990234375, 1, 1], 
     Hue[0.990234375, 1, 1], 
     Hue[0.986328125, 1, 1], 
     Hue[0.986328125, 1, 1], 
     Hue[0.982421875, 1, 1], 
     Hue[0.982421875, 1, 1], 
     Hue[0.978515625, 1, 1], 
     Hue[0.978515625, 1, 1], 
     Hue[0.974609375, 1, 1], 
     Hue[0.974609375, 1, 1], 
     Hue[0.970703125, 1, 1], 
     Hue[0.970703125, 1, 1], 
     Hue[0.966796875, 1, 1], 
     Hue[0.966796875, 1, 1], 
     Hue[0.962890625, 1, 1], 
     Hue[0.962890625, 1, 1], 
     Hue[0.958984375, 1, 1], 
     Hue[0.958984375, 1, 1], 
     Hue[0.955078125, 1, 1], 
     Hue[0.955078125, 1, 1], 
     Hue[0.951171875, 1, 1], 
     Hue[0.951171875, 1, 1], 
     Hue[0.947265625, 1, 1], 
     Hue[0.947265625, 1, 1], 
     Hue[0.943359375, 1, 1], 
     Hue[0.943359375, 1, 1], 
     Hue[0.939453125, 1, 1], 
     Hue[0.939453125, 1, 1], 
     Hue[0.935546875, 1, 1], 
     Hue[0.935546875, 1, 1], 
     Hue[0.931640625, 1, 1], 
     Hue[0.931640625, 1, 1], 
     Hue[0.9277343749999999, 1, 1]}],
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  Epilog->LineBox[{{-0.7838641826095625, 
     0.007248694504781379}, {-0.5225761217397036, 
     0.015553821717245916`}, {-0.2612880608698447, 0.0311636531399525}, {
     0, 0.05830335314366803}, {0.2612880608698589, 0.10185271939899407`}, {
     0.5225761217397178, 0.16614437814855537`}, {0.7838641826095625, 
     0.25306529910329834`}, {1.0451522434794214`, 0.3599262076238347}, {
     1.3064403043492803`, 0.47800040947469474`}, {1.567728365219125, 
     0.5927574107558733}, {1.8290164260889838`, 0.6863721311027554}, {
     2.0903044869588427`, 0.7421235546228844}, {2.3515925478286874`, 
     0.7492499389218215}, {2.6128806086985463`, 0.7063356750679242}, {
     2.874168669568405, 0.621769612203506}, {3.13545673043825, 
     0.5110716251883182}, {3.396744791308109, 0.3922545334074645}, {
     3.6580328521779677`, 0.2811176405697107}, {3.9193209130478124`, 
     0.18812311103351143`}, {4.180608973917671, 0.11755202928233498`}, {
     4.44189703478753, 0.0685886152202849}, {4.703185095657375, 
     0.037368689203493406`}, {4.964473156527234, 0.019010678581648014`}}],
  Frame->True,
  PlotRange->All,
  PlotRangeClipping->True,
  PlotRangePadding->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{
  3.3927263673438687`*^9, {3.3927265215956717`*^9, 3.391849854640625*^9}, 
   3.391850052734375*^9, 3.3924723716804934`*^9},
 ImageCache->GraphicsData["CompressedBitmap", "\<\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\
\>"]]
}, Open  ]],

Cell["The numerical error is:", "Text"],

Cell[CellGroupData[{

Cell["Max[Abs[testFfvals-Ffvals]]", "Input"],

Cell[BoxData["0.1332854140412082`"], "Output",
 CellChangeTimes->{{3.3927265218360176`*^9, 3.3918498608125*^9}, 
   3.391850052796875*^9, 3.392472371984517*^9}]
}, Open  ]]
}, Open  ]]
}, Open  ]]
},
WindowToolbars->"EditBar",
WindowSize->{615, 767},
WindowMargins->{{0, Automatic}, {Automatic, 19}},
ShowSelection->True,
Magnification->1.,
FrontEndVersion->"6.0 for Linux x86 (64-bit) (April 20, 2007)",
StyleDefinitions->"Default.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[590, 23, 81, 3, 141, "Title"],
Cell[674, 28, 107, 6, 124, "Subsubtitle"],
Cell[CellGroupData[{
Cell[806, 38, 30, 0, 38, "Subsection"],
Cell[839, 40, 256, 6, 71, "Text"],
Cell[1098, 48, 888, 19, 171, "Text"],
Cell[1989, 69, 267, 6, 71, "Text"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2293, 80, 46, 0, 73, "Section"],
Cell[2342, 82, 882, 29, 51, "Text"],
Cell[3227, 113, 4767, 158, 54, "DisplayFormula"],
Cell[7997, 273, 185, 7, 31, "Text"],
Cell[8185, 282, 1177, 39, 34, "DisplayFormula"],
Cell[9365, 323, 323, 6, 91, "Text"],
Cell[9691, 331, 1293, 45, 80, "Text"],
Cell[10987, 378, 9138, 309, 91, "DisplayFormula"],
Cell[20128, 689, 719, 26, 33, "Text"],
Cell[20850, 717, 728, 26, 24, "DisplayFormula"]
}, Open  ]],
Cell[CellGroupData[{
Cell[21615, 748, 50, 0, 73, "Section"],
Cell[21668, 750, 274, 10, 31, "Text"],
Cell[21945, 762, 945, 34, 45, "DisplayFormula"],
Cell[22893, 798, 869, 30, 91, "Text"],
Cell[23765, 830, 583, 21, 25, "DisplayFormula"],
Cell[24351, 853, 26, 0, 31, "Text"],
Cell[24380, 855, 577, 27, 25, "DisplayFormula"],
Cell[24960, 884, 91, 2, 31, "Text"],
Cell[25054, 888, 887, 31, 52, "DisplayFormula"]
}, Open  ]],
Cell[CellGroupData[{
Cell[25978, 924, 54, 0, 73, "Section"],
Cell[26035, 926, 65, 0, 31, "Text"],
Cell[26103, 928, 875, 33, 40, "DisplayFormula"],
Cell[26981, 963, 26, 0, 31, "Text"],
Cell[27010, 965, 3698, 131, 52, "DisplayFormula"],
Cell[30711, 1098, 463, 16, 77, "Text"],
Cell[31177, 1116, 1706, 64, 32, "DisplayFormula"],
Cell[32886, 1182, 159, 5, 51, "Text"],
Cell[33048, 1189, 724, 32, 35, "DisplayFormula"],
Cell[33775, 1223, 798, 26, 51, "Text"],
Cell[34576, 1251, 534, 20, 53, "Text"],
Cell[35113, 1273, 6091, 202, 131, "DisplayFormula"],
Cell[41207, 1477, 1581, 60, 93, "Text"],
Cell[42791, 1539, 539, 22, 36, "DisplayFormula"],
Cell[43333, 1563, 1153, 42, 71, "Text"],
Cell[44489, 1607, 1017, 37, 24, "DisplayFormula"],
Cell[45509, 1646, 168, 3, 51, "Text"],
Cell[45680, 1651, 4123, 147, 105, "DisplayFormula"],
Cell[49806, 1800, 1059, 37, 71, "Text"],
Cell[50868, 1839, 3391, 117, 52, "DisplayFormula"],
Cell[54262, 1958, 995, 35, 107, "Text"],
Cell[55260, 1995, 1469, 61, 29, "DisplayFormula"],
Cell[56732, 2058, 1379, 41, 51, "Text"],
Cell[58114, 2101, 1484, 61, 33, "DisplayFormula"]
}, Open  ]],
Cell[CellGroupData[{
Cell[59635, 2167, 43, 0, 73, "Section"],
Cell[59681, 2169, 1431, 49, 131, "Text"],
Cell[61115, 2220, 116, 4, 46, "Input",
 Evaluatable->False],
Cell[61234, 2226, 333, 11, 44, "Input",
 Evaluatable->False],
Cell[61570, 2239, 1373, 48, 151, "Text"],
Cell[62946, 2289, 102, 3, 32, "Input",
 Evaluatable->False],
Cell[63051, 2294, 58, 0, 31, "Text"],
Cell[63112, 2296, 208, 7, 44, "Input",
 Evaluatable->False],
Cell[63323, 2305, 167, 5, 32, "Input",
 Evaluatable->False],
Cell[63493, 2312, 70, 0, 31, "Text"],
Cell[63566, 2314, 312, 10, 56, "Input",
 Evaluatable->False],
Cell[63881, 2326, 171, 6, 51, "Text"],
Cell[64055, 2334, 471, 15, 120, "Input",
 Evaluatable->False]
}, Open  ]],
Cell[CellGroupData[{
Cell[64563, 2354, 35, 0, 73, "Section"],
Cell[64601, 2356, 58, 0, 31, "Text"],
Cell[64662, 2358, 401, 14, 50, "Input"],
Cell[65066, 2374, 65, 0, 31, "Text"],
Cell[65134, 2376, 176, 6, 55, "Input"],
Cell[65313, 2384, 115, 4, 46, "Input"],
Cell[65431, 2390, 339, 11, 44, "Input"],
Cell[65773, 2403, 159, 3, 51, "Text"],
Cell[65935, 2408, 100, 3, 32, "Input"],
Cell[66038, 2413, 114, 3, 31, "Text"],
Cell[66155, 2418, 187, 6, 44, "Input"],
Cell[66345, 2426, 96, 2, 31, "Text"],
Cell[66444, 2430, 546, 17, 169, "Input"],
Cell[66993, 2449, 151, 3, 51, "Text"],
Cell[67147, 2454, 294, 10, 46, "Input"],
Cell[67444, 2466, 69, 0, 31, "Text"],
Cell[67516, 2468, 69, 1, 32, "Input"],
Cell[67588, 2471, 76, 0, 31, "Text"],
Cell[CellGroupData[{
Cell[67689, 2475, 600, 18, 77, "Input"],
Cell[68292, 2495, 37983, 790, 261, 29847, 654, "CachedBoxData", "BoxData", \
"Output"]
}, Open  ]],
Cell[106290, 3288, 117, 3, 51, "Text"],
Cell[106410, 3293, 215, 7, 46, "Input"],
Cell[106628, 3302, 115, 3, 31, "Text"],
Cell[CellGroupData[{
Cell[106768, 3309, 652, 19, 99, "Input"],
Cell[107423, 3330, 9380, 218, 261, 4869, 142, "CachedBoxData", "BoxData", \
"Output"]
}, Open  ]],
Cell[116818, 3551, 286, 5, 91, "Text"],
Cell[117107, 3558, 1601, 48, 293, "Input"],
Cell[118711, 3608, 172, 5, 31, "Text"]
}, Open  ]],
Cell[CellGroupData[{
Cell[118920, 3618, 47, 0, 73, "Section"],
Cell[118970, 3620, 328, 5, 91, "Text"],
Cell[119301, 3627, 64, 0, 31, "Text"],
Cell[119368, 3629, 95, 1, 32, "Input"],
Cell[119466, 3632, 175, 3, 51, "Text"],
Cell[119644, 3637, 486, 7, 131, "Text"],
Cell[120133, 3646, 109, 3, 51, "Text"],
Cell[120245, 3651, 401, 14, 50, "Input"],
Cell[120649, 3667, 41, 0, 31, "Text"],
Cell[CellGroupData[{
Cell[120715, 3671, 62, 0, 32, "Input"],
Cell[120780, 3673, 281334, 5984, 246, "Output"]
}, Open  ]],
Cell[402129, 9660, 65, 0, 31, "Text"],
Cell[402197, 9662, 44, 0, 32, "Input"],
Cell[402244, 9664, 38, 0, 32, "Input"],
Cell[402285, 9666, 1322, 44, 91, "Text"],
Cell[CellGroupData[{
Cell[403632, 9714, 75, 0, 32, "Input"],
Cell[403710, 9716, 265, 7, 31, "Output"]
}, Open  ]],
Cell[403990, 9726, 471, 18, 51, "Text"],
Cell[CellGroupData[{
Cell[404486, 9748, 41, 0, 32, "Input"],
Cell[404530, 9750, 170, 3, 31, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[404737, 9758, 29, 0, 32, "Input"],
Cell[404769, 9760, 171, 3, 31, "Output"]
}, Open  ]],
Cell[404955, 9766, 71, 0, 31, "Text"],
Cell[405029, 9768, 109, 1, 32, "Input"],
Cell[405141, 9771, 182, 4, 51, "Text"],
Cell[CellGroupData[{
Cell[405348, 9779, 56, 0, 32, "Input"],
Cell[405407, 9781, 107651, 2224, 241, 103804, 2159, "CachedBoxData", \
"BoxData", "Output"]
}, Open  ]],
Cell[513073, 12008, 149, 4, 51, "Text"],
Cell[CellGroupData[{
Cell[513247, 12016, 107, 3, 32, "Input"],
Cell[513357, 12021, 86783, 1933, 246, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[600177, 13959, 116, 3, 32, "Input"],
Cell[600296, 13964, 50118, 1058, 248, "Output"]
}, Open  ]],
Cell[650429, 15025, 259, 5, 71, "Text"],
Cell[CellGroupData[{
Cell[650713, 15034, 119, 3, 32, "Input"],
Cell[650835, 15039, 297, 7, 31, "Output"]
}, Open  ]],
Cell[651147, 15049, 22, 0, 31, "Text"],
Cell[CellGroupData[{
Cell[651194, 15053, 66, 0, 32, "Input"],
Cell[651263, 15055, 173, 3, 31, "Output"]
}, Open  ]],
Cell[651451, 15061, 188, 4, 51, "Text"],
Cell[651642, 15067, 48, 0, 32, "Input"],
Cell[651693, 15069, 127, 3, 51, "Text"],
Cell[651823, 15074, 55, 0, 32, "Input"],
Cell[CellGroupData[{
Cell[651903, 15078, 117, 3, 32, "Input"],
Cell[652023, 15083, 11352, 230, 244, 5664, 134, "CachedBoxData", "BoxData", \
"Output"]
}, Open  ]],
Cell[663390, 15316, 74, 0, 31, "Text"],
Cell[663467, 15318, 138, 4, 31, "Text"],
Cell[663608, 15324, 50, 0, 32, "Input"],
Cell[CellGroupData[{
Cell[663683, 15328, 100, 2, 32, "Input"],
Cell[663786, 15332, 99307, 2179, 248, 95673, 2117, "CachedBoxData", \
"BoxData", "Output"]
}, Open  ]],
Cell[763108, 17514, 180, 8, 31, "Text"],
Cell[763291, 17524, 125, 3, 51, "Text"],
Cell[763419, 17529, 68, 0, 32, "Input"],
Cell[CellGroupData[{
Cell[763512, 17533, 124, 3, 32, "Input"],
Cell[763639, 17538, 12714, 249, 244, 5813, 134, "CachedBoxData", "BoxData", \
"Output"]
}, Open  ]],
Cell[776368, 17790, 39, 0, 31, "Text"],
Cell[CellGroupData[{
Cell[776432, 17794, 44, 0, 32, "Input"],
Cell[776479, 17796, 160, 2, 31, "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
