(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.0'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     14971,        381]*)
(*NotebookOutlinePosition[     19347,        512]*)
(*  CellTagsIndexPosition[     18882,        494]*)
(*WindowFrame->Normal*)



Notebook[{
Cell["QVectorToArrow", "Section",
  CellTags->"QVectorToArrow"],

Cell["\<\
QVectorToArrow[pt1,pt2] gives a collection of lines representing a three-dimensional arrow from point pt1 to point pt2. pt1 is optional, default is {0,0,0}. If the vector is shorter than QMinLength, then it is represented by a point. Use with Graphics3D. Example: Show[Graphics3D[QVectorToArrow[{0,0,0},{1,1,1}]]]. The following options control the appearance of the arrow: QArrowHead, QArrowShaft, QArrowShape, QArrowScale, QNeedleStyle, QHeadColor, QShaftColor, QLinePointSize, QMinLength. Package: VQM`VisualizeVector`.\
\>", "Usage",
  CellTags->"QVectorToArrow"],

Cell[BoxData[
    FormBox[
      StyleBox[
        FrameBox[
          StyleBox[GridBox[{
                {
                  FormBox["QArrowHead",
                    
                    "InputForm"], \(QArrowHead\ is\ an\ option\ for\ \
QVectorToArrow . \ 
                        QArrowHead -> \(True\ \((default)\)\ causes\ \
the\ vector\ to\ be\ drawn\ with\ an\ arrowhead . \ QArrowHead -> 
                      Automatic\ draws\ the\ vector\ as\ a\ line\ \
segment\ and\ with\ a\ point\ of\ size\ 2*
                        QLinePointSize\ instead\ of\ the\ arrowhead . \
\ Otherwise\ the\ vector\ is\ just\ represented\ by\ a\ line\ segment \
. \ \(Package : \ \(\(VQM`VisualizeVector`\)\(.\)\)\)\)\)},
                {
                  FormBox["QArrowScale",
                    "InputForm"], \(QArrowScale -> 
                    k\ is\ an\ option\ for\ QVectorToArrow . \ 
                        It\ scales\ the\ length\ of\ the\ arrow\ \
representing\ the\ vector\ by\ the\ factor\ k . \ \(Package : \ \
\(\(VQM`VisualizeVector`\)\(.\)\)\)\)},
                {
                  FormBox["QArrowShaft",
                    
                    "InputForm"], \(QArrowShaft\ is\ an\ option\ for\ \
QVectorToArrow . \ 
                        QArrowShaft -> \(True\ causes\ the\ vector\ \
to\ be\ drawn\ with\ a\ shaft . \ With\ QArrowShaft -> 
                      False\ \((default)\)\ the\ shaft\ is\ drawn\ as\
\ a\ line\ with\ thickness\ QLinePointSize . \ \(Package : \ \
\(\(VQM`VisualizeVector`\)\(.\)\)\)\)\)},
                {
                  FormBox["QArrowShape",
                    
                    "InputForm"], \(QArrowShape -> {n, hfac, rfac, 
                        sfac}\ is\ an\ option\ for\ QVectorToArrow\ \
that\ controls\ the\ appearance\ of\ the\ arrow' 
                      s\ head . \ 
                        Default\ value\ of\ QArrowShape\ is\ {6, 1/4, 
                        1/\((2\ GoldenRatio)\), 1/2}, \ 
                  or\ {6, 1/2, 1/4, 1/2}\ if\ QNeedleStyle -> 
                      True . \ 
                          The\ head\ of\ the\ arrow\ is\ a\ cone\ \
with\ half\ height\ h\  = \ \(hfac . 
                          length \((vector)\)\ and\ radius\ r = 
                      rfac*h\ drawn\ using\ n\ polygons . \ 
                          If\ hfac\ \((rfac)\)\ is\ negative\), \ 
                  then\ h = \(-hfac\)\ \((r = \(-rfac\))\) . \ 
                          If\ the\ option\ QArrowShaft -> 
                      True\ is\ set, \ 
                  then\ the\ shaft\ is\ a\ cylinder\ \((approx . \ 
                            by\ n\ polygons)\)\ with\ radius\ r*
                      sfac . \ If\ QNeedleStyle -> True, \ 
                  then\ the\ shaft\ is\ a\ cone\ with\ apex\ at\ the\ \
origin . \ \(Package : \ \(\(VQM`VisualizeVector`\)\(.\)\)\)\)},
                {
                  FormBox["QHeadColor",
                    "InputForm"], \(QHeadColor -> 
                    GrayLevel[
                        0]\ is\ an\ option\ for\ QVectorToArrow . \ 
                        It\ sets\ the\ color\ for\ the\ head\ of\ the\
\ arrow . \ \(Package : \ \(\(VQM`VisualizeVector`\)\(.\)\)\)\)},
                {
                  FormBox["QLinePointSize",
                    "InputForm"], \(QLinePointSize -> 
                    0.001\ is\ an\ option\ for\ QVectorToArrow . \ 
                        It\ sets\ the\ thickness\ of\ the\ lines\ and\
\ points\ representing\ the\ arrow . \ \(Package : \ \
\(\(VQM`VisualizeVector`\)\(.\)\)\)\)},
                {
                  FormBox["QMinLength",
                    
                    "InputForm"], \(QMinLength\ is\ an\ option\ for\ \
QVectorToArrow . \ QMinLength -> 
                    0.1\ \((default)\)\ specifies\ that\ vectors\ of\ \
length\ less\ than\ 0.1\ are\ to\ be\ drawn\ as\ points\ of\ size\ \
QLinePointSize . \ \(Package : \ \
\(\(VQM`VisualizeVector`\)\(.\)\)\)\)},
                {
                  FormBox["QNeedleStyle",
                    
                    "InputForm"], \(QNeedleStyle\ is\ an\ option\ for\
\ QVectorToArrow . \ QNeedleStyle -> 
                    True\ causes\ the\ vector\ to\ be\ drawn\ as\ a\ \
needle . \ The\ needle\ is\ a\ double - 
                      cone\ whose\ style\ is\ controlled\ by\ the\ \
options\ QArrowShape, \ QShaftColor, \ 
                  QHeadColor . \ \(Package : \ \
\(\(VQM`VisualizeVector`\)\(.\)\)\)\)},
                {
                  FormBox["QShaftColor",
                    "InputForm"], \(QShaftColor -> 
                    GrayLevel[
                        0]\ is\ an\ option\ for\ QVectorToArrow . \ 
                        It\ sets\ the\ color\ for\ the\ shaft\ of\ \
the\ arrow, \ in\ case\ that\ VectorShaft -> 
                    
                    True . \ \(Package : \ \(\(VQM`VisualizeVector`\)\
\(.\)\)\)\)}
                },
              ColumnAlignments->{Right, Left}],
            GridBoxOptions->{RowLines->False}]],
        "2ColumnBox"], TraditionalForm]], "Text",
  CellTags->"QVectorToArrow"],

Cell[TextData[{
  "This command is contained in the package ",
  ButtonBox["VQM`VisualizeVector`",
    ButtonStyle->"AddOnsLink"]
}], "Text",
  CellTags->"QVectorToArrow"],

Cell["QVisualizeVector", "Section",
  CellTags->"QVisualizeVector"],

Cell["\<\
QVisualizeVector[3vector] converts a vector into an arrow graphics and displays it together with other graphics elements whose appearance is controlled by the options QDrawUnitSphere, QDrawAxes, QCoordinateCube, QCoordinateCircles. Package: VQM`VisualizeVector`.\
\>", "Usage",
  CellTags->"QVisualizeVector"],

Cell[BoxData[
    FormBox[
      StyleBox[
        FrameBox[
          StyleBox[GridBox[{
                {
                  FormBox["QCoordinateCirclesColor",
                    "InputForm"], \(QCoordinateCircles -> 
                    RGBColor[0, 0, 1] . \ 
                        Option\ for\ QVisualizeVector\ etc . \ 
                        Draws\ the\ coordinate\ circles\ in\ the\ \
given\ color\ \((provided\ the\ option\ QCoordinateCircles\ is\ set\ \
to\ True . \ \(Package : \ \(\(VQM`VisualizeVector`\)\(.\)\)\)\)\)},
                {
                  FormBox["QDrawUnitSphere",
                    "InputForm"], \(QDrawUnitSphere -> 
                    n\ is\ an\ option\ for\ QVisualizeVector . \ 
                        It\ draws\ the\ outline\ of\ the\ unit\ \
sphere\ by\ plotting\ n\ \((default : \ 
                          20)\)\ circles . \ \(Package : \ \
\(\(VQM`VisualizeVector`\)\(.\)\)\)\)}
                },
              ColumnAlignments->{Right, Left}],
            GridBoxOptions->{RowLines->False}]],
        "2ColumnBox"], TraditionalForm]], "Text",
  CellTags->"QVisualizeVector"],

Cell[TextData[{
  "This command is contained in the package ",
  ButtonBox["VQM`VisualizeVector`",
    ButtonStyle->"AddOnsLink"]
}], "Text",
  CellTags->"QVisualizeVector"],

Cell["polyDisk", "Section",
  CellTags->"polyDisk"],

Cell["\<\
polyDisk[r] is a regular octagon with radius r in the xy-plane. polyDisk[r,n] is a regular polygon with n sides. Package: VQM`VisualizeVector`.\
\>", "Usage",
  CellTags->"polyDisk"],

Cell[TextData[{
  "This command is contained in the package ",
  ButtonBox["VQM`VisualizeVector`",
    ButtonStyle->"AddOnsLink"]
}], "Text",
  CellTags->"polyDisk"],

Cell["polyCone", "Section",
  CellTags->"polyCone"],

Cell["\<\
polyCone[r, h] is a set of 8 triangles approximating the shape of a half-cone with apex at the origin and pointing in the z-direction. polyCone[r,h,n] approximates the half-cone by n triangles. Package: VQM`VisualizeVector`.\
\>", "Usage",
  CellTags->"polyCone"],

Cell[TextData[{
  "This command is contained in the package ",
  ButtonBox["VQM`VisualizeVector`",
    ButtonStyle->"AddOnsLink"]
}], "Text",
  CellTags->"polyCone"],

Cell["polyCylinder", "Section",
  CellTags->"polyCylinder"],

Cell["\<\
polyCylinder[r1, r2, h, n(optional)] represents a cylindrical shape, symmetric around the z-axis. r1 is the radius at the bottom, r2 the radius at the top. h is the height. The cylinder is approximated by n polygons (default: 8). Package: VQM`VisualizeVector`.\
\>", "Usage",
  CellTags->"polyCylinder"],

Cell[TextData[{
  "This command is contained in the package ",
  ButtonBox["VQM`VisualizeVector`",
    ButtonStyle->"AddOnsLink"]
}], "Text",
  CellTags->"polyCylinder"],

Cell["arrowCone", "Section",
  CellTags->"arrowCone"],

Cell["\<\
arrowCone[{pt1,pt2}, c, opts] represents a cone with color c, translated and rotated, so that it has apex at pt2 and points in the direction from pt1 to pt2. Its appearance is controlled by the option QArrowShape. Package: VQM`VisualizeVector`.\
\>", "Usage",
  CellTags->"arrowCone"],

Cell[TextData[{
  "This command is contained in the package ",
  ButtonBox["VQM`VisualizeVector`",
    ButtonStyle->"AddOnsLink"]
}], "Text",
  CellTags->"arrowCone"],

Cell["arrowWithShaft", "Section",
  CellTags->"arrowWithShaft"],

Cell["\<\
arrowWithShaft[{pt1,pt2}, colorhead, colorshaft, opts] represents an arrow from pt1 to pt2 (a collection of polygons). Its appearance is controlled by the option QArrowShape. Package: VQM`VisualizeVector`.\
\>", "Usage",
  CellTags->"arrowWithShaft"],

Cell[TextData[{
  "This command is contained in the package ",
  ButtonBox["VQM`VisualizeVector`",
    ButtonStyle->"AddOnsLink"]
}], "Text",
  CellTags->"arrowWithShaft"],

Cell["doubleHead", "Section",
  CellTags->"doubleHead"],

Cell["\<\
doubleHead[{pt1,pt2}, color1, color2, opts] is actually a double-cone connecting the points pt1 with pt2. The part closer to pt1 has color1. The shape is controlled by the option QArrowShape. Package: VQM`VisualizeVector`.\
\>", "Usage",
  CellTags->"doubleHead"],

Cell[TextData[{
  "This command is contained in the package ",
  ButtonBox["VQM`VisualizeVector`",
    ButtonStyle->"AddOnsLink"]
}], "Text",
  CellTags->"doubleHead"],

Cell["unitSphere", "Section",
  CellTags->"unitSphere"],

Cell["\<\
unitSphere[theta] represents a sphere with radius 1 and center at the origin as a collection of circles parallel to the xy-plane. The polar angles of the circles are multiples of theta. Package: VQM`VisualizeVector`.\
\>", "Usage",
  CellTags->"unitSphere"],

Cell[TextData[{
  "This command is contained in the package ",
  ButtonBox["VQM`VisualizeVector`",
    ButtonStyle->"AddOnsLink"]
}], "Text",
  CellTags->"unitSphere"],

Cell["QRoundSphere", "Section",
  CellTags->"QRoundSphere"],

Cell["\<\
QRoundSphere[r,n,m,opts] is a sphere with radus r, represented by as a wireframe. QRoundSphere[r,n,m] is similar to Sphere[r,n,m] in the standard package Graphics`Shapes`, but QRoundSphere uses circles (generated by ParametricPlot3D) instead of polygons. The options are passed to ParametricPlot3D (useful for PlotPoints). Package: VQM`VisualizeVector`.\
\>", "Usage",
  CellTags->"QRoundSphere"],

Cell[TextData[{
  "This command is contained in the package ",
  ButtonBox["VQM`VisualizeVector`",
    ButtonStyle->"AddOnsLink"]
}], "Text",
  CellTags->"QRoundSphere"],

Cell["coordinateCube", "Section",
  CellTags->"coordinateCube"],

Cell["\<\
coordinateCube[pt] represents a rectangular shape with edges parallel to the coordinate axes. It has one point at the origin and one at the point pt. Package: VQM`VisualizeVector`.\
\>", "Usage",
  CellTags->"coordinateCube"],

Cell[TextData[{
  "This command is contained in the package ",
  ButtonBox["VQM`VisualizeVector`",
    ButtonStyle->"AddOnsLink"]
}], "Text",
  CellTags->"coordinateCube"],

Cell["coordinateCircles", "Section",
  CellTags->"coordinateCircles"],

Cell["\<\
coordinateCircles[pt] are two circular arcs representing the polar angle and the azimuthal angle of a point pt. Package: VQM`VisualizeVector`.\
\>", "Usage",
  CellTags->"coordinateCircles"],

Cell[TextData[{
  "This command is contained in the package ",
  ButtonBox["VQM`VisualizeVector`",
    ButtonStyle->"AddOnsLink"]
}], "Text",
  CellTags->"coordinateCircles"],

Cell["graphicElements", "Section",
  CellTags->"graphicElements"],

Cell["\<\
graphicElements[pt, opts] defines the graphic elements for the visualization of the position pt according to the values of the options QDrawUnitSphere, QDrawAxes, QCoordinateCube, QCoordinateCircles, QCoordinateCirclesColor. Package: VQM`VisualizeVector`.\
\>", "Usage",
  CellTags->"graphicElements"],

Cell[TextData[{
  "This command is contained in the package ",
  ButtonBox["VQM`VisualizeVector`",
    ButtonStyle->"AddOnsLink"]
}], "Text",
  CellTags->"graphicElements"]
},
FrontEndVersion->"5.0 for Macintosh",
ScreenRectangle->{{0, 1280}, {0, 961}},
StyleDefinitions -> "HelpBrowser.nb"
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{
  "QVectorToArrow"->{
    Cell[1754, 51, 63, 1, 70, "Section",
      CellTags->"QVectorToArrow"],
    Cell[1820, 54, 577, 3, 70, "Usage",
      CellTags->"QVectorToArrow"],
    Cell[2400, 59, 5076, 106, 70, "Text",
      CellTags->"QVectorToArrow"],
    Cell[7479, 167, 171, 5, 70, "Text",
      CellTags->"QVectorToArrow"]},
  "QVisualizeVector"->{
    Cell[7653, 174, 67, 1, 70, "Section",
      CellTags->"QVisualizeVector"],
    Cell[7723, 177, 319, 3, 70, "Usage",
      CellTags->"QVisualizeVector"],
    Cell[8045, 182, 1124, 25, 70, "Text",
      CellTags->"QVisualizeVector"],
    Cell[9172, 209, 173, 5, 70, "Text",
      CellTags->"QVisualizeVector"]},
  "polyDisk"->{
    Cell[9348, 216, 51, 1, 70, "Section",
      CellTags->"polyDisk"],
    Cell[9402, 219, 192, 3, 70, "Usage",
      CellTags->"polyDisk"],
    Cell[9597, 224, 165, 5, 70, "Text",
      CellTags->"polyDisk"]},
  "polyCone"->{
    Cell[9765, 231, 51, 1, 70, "Section",
      CellTags->"polyCone"],
    Cell[9819, 234, 273, 3, 70, "Usage",
      CellTags->"polyCone"],
    Cell[10095, 239, 165, 5, 70, "Text",
      CellTags->"polyCone"]},
  "polyCylinder"->{
    Cell[10263, 246, 59, 1, 70, "Section",
      CellTags->"polyCylinder"],
    Cell[10325, 249, 313, 3, 70, "Usage",
      CellTags->"polyCylinder"],
    Cell[10641, 254, 169, 5, 70, "Text",
      CellTags->"polyCylinder"]},
  "arrowCone"->{
    Cell[10813, 261, 53, 1, 70, "Section",
      CellTags->"arrowCone"],
    Cell[10869, 264, 294, 3, 70, "Usage",
      CellTags->"arrowCone"],
    Cell[11166, 269, 166, 5, 70, "Text",
      CellTags->"arrowCone"]},
  "arrowWithShaft"->{
    Cell[11335, 276, 63, 1, 70, "Section",
      CellTags->"arrowWithShaft"],
    Cell[11401, 279, 260, 3, 70, "Usage",
      CellTags->"arrowWithShaft"],
    Cell[11664, 284, 171, 5, 70, "Text",
      CellTags->"arrowWithShaft"]},
  "doubleHead"->{
    Cell[11838, 291, 55, 1, 70, "Section",
      CellTags->"doubleHead"],
    Cell[11896, 294, 273, 3, 70, "Usage",
      CellTags->"doubleHead"],
    Cell[12172, 299, 167, 5, 70, "Text",
      CellTags->"doubleHead"]},
  "unitSphere"->{
    Cell[12342, 306, 55, 1, 70, "Section",
      CellTags->"unitSphere"],
    Cell[12400, 309, 267, 3, 70, "Usage",
      CellTags->"unitSphere"],
    Cell[12670, 314, 167, 5, 70, "Text",
      CellTags->"unitSphere"]},
  "QRoundSphere"->{
    Cell[12840, 321, 59, 1, 70, "Section",
      CellTags->"QRoundSphere"],
    Cell[12902, 324, 406, 3, 70, "Usage",
      CellTags->"QRoundSphere"],
    Cell[13311, 329, 169, 5, 70, "Text",
      CellTags->"QRoundSphere"]},
  "coordinateCube"->{
    Cell[13483, 336, 63, 1, 70, "Section",
      CellTags->"coordinateCube"],
    Cell[13549, 339, 235, 3, 70, "Usage",
      CellTags->"coordinateCube"],
    Cell[13787, 344, 171, 5, 70, "Text",
      CellTags->"coordinateCube"]},
  "coordinateCircles"->{
    Cell[13961, 351, 69, 1, 70, "Section",
      CellTags->"coordinateCircles"],
    Cell[14033, 354, 200, 3, 70, "Usage",
      CellTags->"coordinateCircles"],
    Cell[14236, 359, 174, 5, 70, "Text",
      CellTags->"coordinateCircles"]},
  "graphicElements"->{
    Cell[14413, 366, 65, 1, 70, "Section",
      CellTags->"graphicElements"],
    Cell[14481, 369, 311, 3, 70, "Usage",
      CellTags->"graphicElements"],
    Cell[14795, 374, 172, 5, 70, "Text",
      CellTags->"graphicElements"]}
  }
*)

(*CellTagsIndex
CellTagsIndex->{
  {"QVectorToArrow", 15526, 395},
  {"QVisualizeVector", 15854, 404},
  {"polyDisk", 16184, 413},
  {"polyCone", 16411, 420},
  {"polyCylinder", 16643, 427},
  {"arrowCone", 16886, 434},
  {"arrowWithShaft", 17125, 441},
  {"doubleHead", 17375, 448},
  {"unitSphere", 17613, 455},
  {"QRoundSphere", 17853, 462},
  {"coordinateCube", 18101, 469},
  {"coordinateCircles", 18358, 476},
  {"graphicElements", 18622, 483}
  }
*)

(*NotebookFileOutline
Notebook[{
Cell[1754, 51, 63, 1, 70, "Section",
  CellTags->"QVectorToArrow"],
Cell[1820, 54, 577, 3, 70, "Usage",
  CellTags->"QVectorToArrow"],
Cell[2400, 59, 5076, 106, 70, "Text",
  CellTags->"QVectorToArrow"],
Cell[7479, 167, 171, 5, 70, "Text",
  CellTags->"QVectorToArrow"],
Cell[7653, 174, 67, 1, 70, "Section",
  CellTags->"QVisualizeVector"],
Cell[7723, 177, 319, 3, 70, "Usage",
  CellTags->"QVisualizeVector"],
Cell[8045, 182, 1124, 25, 70, "Text",
  CellTags->"QVisualizeVector"],
Cell[9172, 209, 173, 5, 70, "Text",
  CellTags->"QVisualizeVector"],
Cell[9348, 216, 51, 1, 70, "Section",
  CellTags->"polyDisk"],
Cell[9402, 219, 192, 3, 70, "Usage",
  CellTags->"polyDisk"],
Cell[9597, 224, 165, 5, 70, "Text",
  CellTags->"polyDisk"],
Cell[9765, 231, 51, 1, 70, "Section",
  CellTags->"polyCone"],
Cell[9819, 234, 273, 3, 70, "Usage",
  CellTags->"polyCone"],
Cell[10095, 239, 165, 5, 70, "Text",
  CellTags->"polyCone"],
Cell[10263, 246, 59, 1, 70, "Section",
  CellTags->"polyCylinder"],
Cell[10325, 249, 313, 3, 70, "Usage",
  CellTags->"polyCylinder"],
Cell[10641, 254, 169, 5, 70, "Text",
  CellTags->"polyCylinder"],
Cell[10813, 261, 53, 1, 70, "Section",
  CellTags->"arrowCone"],
Cell[10869, 264, 294, 3, 70, "Usage",
  CellTags->"arrowCone"],
Cell[11166, 269, 166, 5, 70, "Text",
  CellTags->"arrowCone"],
Cell[11335, 276, 63, 1, 70, "Section",
  CellTags->"arrowWithShaft"],
Cell[11401, 279, 260, 3, 70, "Usage",
  CellTags->"arrowWithShaft"],
Cell[11664, 284, 171, 5, 70, "Text",
  CellTags->"arrowWithShaft"],
Cell[11838, 291, 55, 1, 70, "Section",
  CellTags->"doubleHead"],
Cell[11896, 294, 273, 3, 70, "Usage",
  CellTags->"doubleHead"],
Cell[12172, 299, 167, 5, 70, "Text",
  CellTags->"doubleHead"],
Cell[12342, 306, 55, 1, 70, "Section",
  CellTags->"unitSphere"],
Cell[12400, 309, 267, 3, 70, "Usage",
  CellTags->"unitSphere"],
Cell[12670, 314, 167, 5, 70, "Text",
  CellTags->"unitSphere"],
Cell[12840, 321, 59, 1, 70, "Section",
  CellTags->"QRoundSphere"],
Cell[12902, 324, 406, 3, 70, "Usage",
  CellTags->"QRoundSphere"],
Cell[13311, 329, 169, 5, 70, "Text",
  CellTags->"QRoundSphere"],
Cell[13483, 336, 63, 1, 70, "Section",
  CellTags->"coordinateCube"],
Cell[13549, 339, 235, 3, 70, "Usage",
  CellTags->"coordinateCube"],
Cell[13787, 344, 171, 5, 70, "Text",
  CellTags->"coordinateCube"],
Cell[13961, 351, 69, 1, 70, "Section",
  CellTags->"coordinateCircles"],
Cell[14033, 354, 200, 3, 70, "Usage",
  CellTags->"coordinateCircles"],
Cell[14236, 359, 174, 5, 70, "Text",
  CellTags->"coordinateCircles"],
Cell[14413, 366, 65, 1, 70, "Section",
  CellTags->"graphicElements"],
Cell[14481, 369, 311, 3, 70, "Usage",
  CellTags->"graphicElements"],
Cell[14795, 374, 172, 5, 70, "Text",
  CellTags->"graphicElements"]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)
